//
//   $Id: COMPRESS.C 1.4 2002/10/15 19:58:14Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	COMPRESS.C  created 05/31/91  by Drew Gislason

   Copyright (c) 1991-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Rules:
	1) Start of DGROUP (DS:0) must contain at least 6 zeros.
	2) There can be no other data fixups besides DGROUP

	Operation:
	1) Must compress the file
	2) Must modify EXE to reflect new size
	3) Must modify MAP segemnt list to reflect new size

	Compressed Format:
	Offset 0 - [DGROUP Size] (word)
	Offset 2 - [Blk Size] (word) (# of 0s)
	Offset 4 - [Next Blk] (word) (FFFFH = no more blocks)

	Example:
	24,				[Total current DGROUP size]  24 bytes (4+8+4+8)
	16, 12,			[1st Blk, 16 0s, ptr to next Blk]
	"12345678",		[Stuff]
	20, 0FFFFH,		[2nd Blk, 20 0s, ptr to next Blk, FFFF=end list]
	"12345678"		[More Stuff]

	Would Expand to:
	db 0				[Where DGROUP size used to be]
	db 16 dup (0)	[Expanded 1st Blk]
	db "12345678"	[Stuff]
	db 20 dup (0)	[Expanded 2nd Blk]
	db "12345678"	[More Stuff]

**********************************************************/
#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dosver.h"
#include "datalght.h"
#include "fatworld.h"

/* in MAP file, how far do we look before giving up? */
#define LINES_TO_LOOK 16
#define M_START	1			/* byte offset to start field */
#define M_STOP		8			/* byte offset to stop field */
#define M_LENGTH	15			/* byte offset to length field */

/*
   EXE HEADER
   Description: Every EXE begins with this structure

   Reference: Adv MS-DOS, p 32  -or-  DOS Ref, p 62
*/
struct exehdr
{
   unsigned sig;     /* 00  Mark Zimbowski does it again...  */
   unsigned imgmod;  /* 02  image len mod 512 bytes (a page) */
   unsigned imglen;  /* 04  image len in 512-byte pages      */
   unsigned nreloc;  /* 06  number of relocatable items      */
   unsigned hdrlen;  /* 08  size of header in paragraphs     */
   unsigned minpar;  /* 0A  min paragraphs of RAM required   */
   unsigned maxpar;  /* 0C  max paragraphs of RAM needed     */
   unsigned ssdsp;   /* 0E  SS = this + startseg             */
   unsigned spofs;   /* 10  value placed in SP upon execute  */
   unsigned cksum;   /* 12  ignored by operating system      */
   unsigned ipofs;   /* 14  value placed in IP upon execute  */
   unsigned csdsp;   /* 16  CS = this + startseg             */
   unsigned reldsp;  /* 18  location of 1st relocatable item */
   unsigned ovlnum;  /* 1A  Overlay no. (normally 0)         */
};


char name[66];
struct exehdr header;
long map_start_data;			/* start of _DATA according to MAP */
uchar *dgroup;					/* in memory copy of initialized data */
long beg_text;					/* size of EXE header/beginning of CODE */
long beg_dgroup;				/* beginning of DGROUP in EXE file */
unsigned init_data_size;	/* INIT DATA size after compression */
unsigned org_idata_size;	/* INIT DATA size before compression */
char line1[82];

bool fIsFat32World;

void error(int n);


/*
	Return the size and start of all INITIALIZED data,
	that is, segments "_DATA" through "_BSS".
*/
unsigned get_init_DATA_size(long *start)
{
	int i;
	char * p;
	FILE * fp;
	unsigned size = 0;
	long end;

	p = strrchr(name,'.');
	strcpy(p,".MAP");
	fp = fopen(name, "r+");
	if(!fp)
	{
		error(5);
	}

	/* find the line with "_DATA" on it */
	for(i=0; i<LINES_TO_LOOK; ++i)
	{
		if(fgets(line1, 80, fp) == NULL)
			error(6);
		p = strchr(line1,'_');
		if(p && strncmp(p, "_DATA", 5)==0)
			break;
	}

	/* "_DATA" not found, bad map file */
	if(i == LINES_TO_LOOK)
		error(7);

	/* read in the start and size */
	sscanf(line1+M_START, "%lx", start);

	/* add in the size of ALL segments until BSS */
	while(1)
	{
		if(fgets(line1, 80, fp) == NULL)
			error(6);

		p = strchr(line1,'_');
		if(p && strncmp(p, "_BSS", 4)==0)
			break;
	}
	sscanf(line1+M_START,"%lx", &end);

	fclose(fp);

	size = (unsigned)(end - *start);

	return size;
}


/*
	Read in all of DGROUP.  Set the location and size
	in global variables.

	Return FALSE if there are any problems.
*/
int read_dgroup(FILE *fp)
{
	unsigned len;

	/* no relocatable items, we can't find DGROUP */
	if(!header.nreloc)
		return FALSE;

	/* determine beginning of EXE code (_TEXT) */
	beg_text = (long)header.hdrlen * 0x10;

	/* determine the beginning and size of DGROUP from .MAP */
	org_idata_size = init_data_size = get_init_DATA_size(&map_start_data);
	beg_dgroup = beg_text + map_start_data;

	/* allocate the memory */
	dgroup = malloc(init_data_size);
	if(!dgroup)
		return FALSE;

	/* actually read in the bytes */
	fseek(fp, beg_dgroup, SEEK_SET);
	len = fread(dgroup, 1, init_data_size, fp);

	/* kludge because of differences in Microsoft or Borland */
	if(len > init_data_size || len < init_data_size-5)
		return FALSE;

	return TRUE;
}


/*
	write out the new DGROUP and truncate the file
*/
int write_dgroup(FILE *fp)
{
	long end_text;

	fseek(fp, beg_dgroup, SEEK_SET);
	if(fwrite(dgroup, 1, init_data_size, fp) != init_data_size)
		return FALSE;

	/* truncate the file */
	_write(fp->fd, "", 0);

	/* update the EXE header */
	end_text = beg_dgroup + init_data_size;
	header.imglen = (unsigned)(end_text / 0x200);
	header.imgmod = (unsigned)(end_text % 0x200);
	if(header.imgmod)
		++header.imglen;

	/* write ou the new EXE header */
	fseek(fp, 0L, SEEK_SET);
	fwrite(&header, 1, sizeof(header), fp);

	return TRUE;
}


/*
	count the # of zeros at this position.
*/
int zero_count(uchar *p, unsigned len)
{
	int count = 0;

	while(*p==0 && len)
	{
		++count;
		--len;
		++p;
	}

	return count;
}


/*
	find the next block of zeros.  Returns NULL if
	none left.
*/
uchar * next_block(uchar *pos, unsigned left, int minblk)
{
	while(left > minblk)
	{
		if(*pos == 0 && zero_count(pos,left) >= minblk)
			return pos;
		++pos;
		--left;
	}
	return NULL;
}


/*
	Compress the DGROUP that is in memory
*/
int compress(void)
{
	/* minimum 5 in a row */
	#define MINBLK 5
	int len, left;
	int count;
	uchar *pos, *next;

	/* there must be at least 6 zeros at DS:0 */
	if(zero_count(dgroup, init_data_size) < 6)
		return FALSE;

	len = left = init_data_size - 2;
	pos = dgroup + 2;

	/* find all the blocks of zeros */
	while(1)
	{
		/* how many zeros in this block? */
		count = zero_count(pos, left);

		/* adjust sizes */
		len  -= count - 4;
		left -= count;

		/* reduce the block of zeros to a 4-byte structure */
		if(left)
			memmove(pos+4, pos+count, left);
		*(unsigned *)pos = count;

		next = next_block(pos+4, left, MINBLK);
		if(!next)
		{
			*(unsigned *)(pos+2) = 0xFFFF;
			break;
		}

		/* move on to next block */
		left -= next-pos-4;
		*(unsigned *)(pos+2) = next-pos-4;
		pos = next;
	}

	/* report the lengths */
	*(unsigned *)dgroup = len;
	init_data_size = len+2;

	return TRUE;
}


/*
	write the new size into the map file.
*/
void set_diffsize(char *oldval, int diff)
{
	char temp[6];
	long val;

	sscanf(oldval,"%lx",&val);
	val += diff;
	sprintf(temp,"%05X", val);
	strncpy(oldval,temp,5);
}

void set_newsize(char *oldval, long size)
{
	char temp[6];

	sprintf(temp,"%05lX", size);
	strncpy(oldval,temp,5);
}


/*
	Modify the MAP file.  This will update all segments
	from _DATA through _BSS segments to reflect the new
	size of the initialized DATA.  _BSS appears larger
	and _DATA appears smaller.
*/
int mod_map(FILE *fp)
{
	char *p;
	long pos, pos2;
	int i;
	long map_start_bss = map_start_data + init_data_size;
	int founddata = FALSE;		/* TRUE if _DATA has been found */

	/* find the line with "_DATA" on it */
	for(i=0; i<LINES_TO_LOOK; ++i)
	{
		fgetpos(fp, &pos);
		if(fgets(line1, 80, fp) == NULL)
		{
			printf("Error reading MAP\n");
			return FALSE;
		}
		fgetpos(fp, &pos2);

		p = strchr(line1,'_');

		/* update "STOP" and "LENGTH" fields for _DATA */
		if(!founddata && p && strncmp(p, "_DATA", 5)==0)
		{
			set_newsize(line1+M_STOP, map_start_bss-1);
			set_newsize(line1+M_LENGTH, (long)init_data_size);

			/* write out the _DATA segment in the MAP */
			fsetpos(fp, &pos);
			fwrite(line1, 1, strlen(line1), fp);
			fsetpos(fp, &pos2);

			founddata = 1;
		}

		/* update "START" and "LENGTH" fields for _BSS */
		else if(founddata && p && strncmp(p, "_BSS", 4)==0)
		{
			set_newsize(line1+M_START, map_start_bss);
			set_diffsize(line1+M_LENGTH, org_idata_size - init_data_size);

			fsetpos(fp, &pos);
			fwrite(line1, 1, strlen(line1), fp);
			fsetpos(fp, &pos2);

			/* done processing .MAP */
			break;
		}

		/* update "START", "END" and "LENGTH" fields INIT data segment(s) */
		else if(founddata)
		{
			set_newsize(line1+M_START, map_start_bss);
			set_newsize(line1+M_STOP, map_start_bss);
			set_newsize(line1+M_LENGTH, 0);

			fsetpos(fp, &pos);
			fwrite(line1, 1, strlen(line1), fp);
			fsetpos(fp, &pos2);
		}
	}

	/* _DATA not found, bad map file */
	if(i == LINES_TO_LOOK)
		return FALSE;

	return TRUE;
}


void DisplaySignOnMessage( void )
{
	char * szSignOn = "Compress %s %s\n%s\n\n";


	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	Compress the DGROUP of an EXE file and modify its MAP
	to reflect this compression.
*/
main(int argc, char *argv[])
{
	FILE *fp;
	char *p;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* show help */
	if(argc < 2)
	{
		/* display sign-on */
		DisplaySignOnMessage();

		printf("Usage = COMPRESS <file>\n");
		printf("Note: the first fixup of the file must be DGROUP\n");
		return 1;
	}

	/* append EXE name if needed */
	strcpy(name,argv[1]);
	p = strrchr(name,'.');
	if(!p)
		strcat(name,".EXE");

	/* open the file */
	fp = fopen(name, "rb+");
	if(!fp)
	{
		printf("cannot open %s\n",name);
		return 2;
	}

	/* make sure this is an EXE */
	if(fread(&header, 1, sizeof(header), fp) != sizeof(header)
		|| (header.sig != 0x5A4D))
	{
		error(1);
	}

	/* read in the DGROUP */
	if(!read_dgroup(fp))
	{
		error(2);
	}

	/* make sure DS:0 has at least 6 zeros */
	if(!compress())
	{
		error(3);
	}

	/* write out the modified DGROUP */
	if(!write_dgroup(fp))
	{
		error(4);
	}
	fclose(fp);

	/* modify the MAP file */
	p = strrchr(name,'.');
	strcpy(p,".MAP");
	fp = fopen(name, "r+");
	if(!fp)
		printf("Warning! no MAP file!\n");
	else
	{
		if(!mod_map(fp))
		{
			printf("Bad MAP file\n");
			return 7;
		}
		fclose(fp);
	}

	return 0;
}

char *errmsg[] =
{
	"",								/* 0 */
	"Not an EXE!",
	"Cannot read DGROUP",
	"Cannot compress DGROUP",
	"Failed to write DGROUP",	/* 4 */
	"No MAP File!",
	"Error reading MAP",
	"Segment _DATA not found",	/* 7 */

};


void error(int n)
{
	printf("%s\n",errmsg[n]);
	exit(n);
}
