//
//   $Id: SAVE.C 1.4 2002/10/15 19:58:14Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	SAVE.C

	Copyright (c) 1994-2002 Datalight, Inc.
	Datalight Proprietary and Confidential
	All Rights Reserved

	Revision History
		03/16/94 - rls - original
		06/20/97 - ms  - revised

	Description
		this program stores information in the system environment

*/
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <dos.h>
#include "dostruct.h"
#include "save.h"
#include "save.pt"

#pragma argsused

/*
	(main)

	entry point for SAVE program
*/
int main(int argc, char *argv[])
{
	int iCommand;									// command to execute

	/*
		If user didn't specify the right number of arguments, prompt
		on usage
	*/
	if (argc < 3)
	{
		printf(USAGE);
		return 1;
	}

	/*
		Determine which command was used
	*/
	iCommand = WhichCommand(strupr(argv[1]));

	/*
		Execute correct operation depending on input command
	*/
	ExecuteCommand(iCommand, argv);

	return 0;
}

/*
	ExecuteCommand

	interpret & execute the command input
*/
int ExecuteCommand(int iCommand, char **argv)
{
	int result;									// result code
	union REGS inregs, outregs;			// registers for DOS calls
	int iVersion;								// version number
	unsigned uRevision;						// revision number
	char String[128];							// temporary string array
	char EnvString[128];						// environment string array

	switch(iCommand)
	{
		case 1 :		// DATE
		{
			printf("Saving the Date\n");
			inregs.x.ax = GET_SYSTEM_DATE;
			intdos(&inregs, &outregs);
			sprintf(String, "%d-%02d-%d", outregs.h.dh, outregs.h.dl, outregs.x.cx);
			break;
		}

		case 2 : 	// TIME
		{
			printf("Saving the Time\n");
			inregs.x.ax = GET_SYSTEM_TIME;
			intdos(&inregs, &inregs);
			sprintf(String, "%02d:%02d:%02d", inregs.h.ch, inregs.h.cl,
				inregs.h.dh);
			break;
		}

		case 3: 		// CWD
		{
			printf("Saving the Current Working Directory\n");
			String[0]= '\\';
			inregs.x.ax = GET_CWD;
			inregs.x.si = (unsigned) String+1;
			inregs.h.dl = 0;
			intdos(&inregs, &outregs);
			break;
		}

		case 4 : 	// DOS
		{
			printf("Saving the Current DOS\n");
			sprintf(String, "%s-DOS", (IsROMDOS()) ? "ROM" : "MS");
			break;
		}

		case 5 : 	// VER
		{
			printf("Saving the Current DOS Version\n");
			iVersion = GetDOSVersion();
			sprintf(String, "%d.%d",iVersion/100, iVersion % 100);
			break;
		}

		case 6 : 	// DRIVE
		{
			printf("Saving the Current Drive\n");
			String[0] = 'A' + GetDefaultDrive();
			String[1] = ':';
			String[2] = 0;
			break;
		}

		case 7 : 	// REV
		{
			uRevision = GetRevision();
			sprintf(String, "%d.%d", uRevision >> 8, uRevision & 0xFF);
			break;
		}

		default : 	// no match
		{
			strcpy(String, argv[1]);
			break;
		}

	}
	/*
		Copy the modified string to the Environment string
	*/
	sprintf(EnvString, "%s=%s", argv[2], String);
	printf("%s\n", EnvString);

	/*
		Set the environment string
	*/
	result = StoreToEnvironment(EnvString);
	if (result!=0)
	{
		return 1;
	}

	return 0;
}

/*
	WhichCommand

	scans the command list to match input string
*/
int WhichCommand(char *lpInput)
{
	int i;										 			// ubiquitous counter

	// list of available commands
	static char *szCommandList[] = COMMAND_LIST;

	/*
		search the command list for a matching string
	*/
	for (i=NUMBER_OF_COMMANDS;i>0;i--)
	{
		if (strcmp(lpInput, szCommandList[i-1])==0)
		{
			break;
		}
	}

#ifdef DEBUG
	if (i>0)
		printf("---matched with %s\r\n", szCommandList[i-1]);
	else
		printf("---no match\r\n");
#endif
	/*
		return the matching command number (or zero for no match)
	*/
	return i;
}

/*
	StoreToEnvironment

	syntax : StoreToEnvironment [variable=[string]]

	Add or delete an environment variable, or show the
	whole environment.
*/
int StoreToEnvironment(char *szString)
{
	int result;							// result code
	extern unsigned _psp;			// externally available PSP segment?
	char far *lpEnv;					// pointer to environment
	char far *lpEP;					// pointer to environment
	unsigned uEnvSeg;					// environment segment
	unsigned uEnvSize;				// environment size
	char far *lpEquals;				// pointer to EQUALS in string

	/*
		find the environment
	*/
	uEnvSeg  = GetEnvironmentSegment();
	uEnvSize = *(unsigned far *) MK_FP(uEnvSeg-1,3);
	uEnvSize *= 16;
	lpEnv = (void far *) MK_FP(uEnvSeg, 0);

	/*
		Prepare the environment set string by UPPERCASEing the left side
	*/
	strupr(szString);

#ifdef DEBUG
	printf("storing %s\r\n", szString);
#endif

	/*
		If the environment variable already exists, remove it
	*/
	lpEP = MatchEnvEntry(lpEnv,szString);
	if (lpEP != 0)
	{
		RemoveFromEnvironment(lpEP);
	}

	/*
		Check to see if it was a null assignment (i.e. a remove)
		If so, we're done!
	*/
	lpEquals = strchr(szString, EQUALS);
	if (*(++lpEquals)==0)
	{
		return 0;
	}

	/*
		Otherwise, add the environment variable and assigned value
	*/
	result = AddToEnvironment(lpEnv, szString, uEnvSize);
	if (result!=0)
	{
		return 1;
	}

	return 0;
}

/*
	AddToEnvironment

	Add a variable to the environment.
*/
int AddToEnvironment(char far *lpEnv, char *szString, unsigned uEnvSize)
{
	/*
		find the end of the environment area, and determine how much space
		is available
	*/
	while (*lpEnv)
	{
		while (*lpEnv)
		{
			--uEnvSize;
			++lpEnv;
		}
		--uEnvSize;
		++lpEnv;
	}
	--uEnvSize;

	/*
		if there is insufficient space for this variable, exit
	*/
	if (strlen(szString) >= uEnvSize)
	{
		printf("Out of environment space!\n");
		return 1;
	}

	/*
		otherwise, add the string
	*/
	do
	{
		*lpEnv++ = *szString;
	}
	while (*szString++);

	/*
		zero-terminate the enviroment area
	*/
	*lpEnv = 0;

	return 0;
}

/*
	MatchEnvEntry

	Given an environment pointer and a name, find the
	matching environment entry.  Both are assumed to be
	UPPERCASE.
*/
char far * MatchEnvEntry(char far *lpEnv,char *szName)
{
	char *szTemp;
	char far *lpEP;

	/*
		search the environment list for a matching entry
	*/
	while (*lpEnv)
	{
		lpEP = lpEnv;
		szTemp = szName;

		while (*szTemp && *lpEP != EQUALS)
		{
			if (*szTemp != *lpEP)
			{
				break;
			}
			++szTemp;
			++lpEP;
		}

		/*
			If both strings have reached the EQUALS character,
			a match has occurred
		*/
		if (*szTemp == EQUALS && *lpEP == EQUALS)
		{
#ifdef DEBUG
			printf("Matched %s\r\n", lpEnv);
#endif
			return lpEnv;
		}

		/*
			Otherwise, no match occurred; so increment to the next
			environment entry
		*/
		while (*lpEP)
		{
			++lpEP;
		}

		lpEnv = lpEP+1;
	}

	return 0;
}


/*
	RemoveFromEnvironment

	Remove a variable from the environment.  "lpEntry" points
	to the variable entry to delete.
*/
void RemoveFromEnvironment(char far *lpEntry)
{
	char far *lpNext;		// pointer to next entry

	/*
		find the next entry
	*/
	lpNext = lpEntry;
	while (*lpNext)
	{
		++lpNext;
	}

	++lpNext;

	/*
		copy the next entry into the position of the current entry,
		and so on, effectively deleting the current entry
	*/
	while (*lpNext)
	{
		do
		{
			*lpEntry++ = *lpNext;
		}
		while (*lpNext++);
	}
	*lpEntry = 0;
}

/*
	GetEnvironmentSegment

	locate the Environment segment for DOS
*/
unsigned GetEnvironmentSegment(void)
{
	extern unsigned _psp;		// externally-available psp
	struct PSP far *PSP;			// pointer to DOS PSP
	unsigned psp_seg;				// psp segment

	/*
		initialize the psp segment to the current progam
	*/
	psp_seg = _psp;

	/*
		search for PSP chain for the current PSP segment
	*/
	do
	{
		PSP = MK_FP(psp_seg, 0);
		psp_seg = PSP->last_psp;
	}
	while (FP_SEG(PSP) != psp_seg);

	/*
		return the Environment Segment from the PSP structure
	*/
	return (PSP->env_seg);
}

/*
	IsROMDOS

	returns TRUE if the current DOS is ROM-DOS
*/
int IsROMDOS(void)
{
	union REGS inregs, outregs;		// calling registers
	struct SREGS segregs;				// segment registers

	inregs.x.ax = GET_REVISION;		// ROM-DOS internal call
	inregs.x.cx = 0;						//
	inregs.x.di = 3;						// function code 3, get version string
	inregs.x.si = CHECKVAL;				// sanity check
	intdosx(&inregs,&outregs,&segregs);

	/*
		If CX fails to return the value passed in SI, this isn't ROM-DOS!
	*/
	if(outregs.x.cx != CHECKVAL)
		return 0;
	else
		return 1;
}

/*
	GetDOSVersion

	obtain the DOS version number
*/
unsigned GetDOSVersion(void)
{
	unsigned uVersion;					// return value
	union REGS inregs, outregs;		// calling registers

	inregs.x.ax = GET_DOS_VERSION;
	intdos(&inregs,&outregs);
	uVersion = (outregs.h.al * 100) + outregs.h.ah;

	return uVersion;

}

/*
	GetDefaultDrive

	obtain the default disk drive
*/
GetDefaultDrive(void)
{
	union REGS inregs, outregs;				// calling registers

	/*
		Make the call
	*/
	inregs.x.ax = GET_DEFAULT_DRIVE;
	intdos(&inregs,&outregs);

	return outregs.h.al ;
}

/*
	GetRevision

	obtain the revision number
*/
int GetRevision(void)
{
	union REGS inregs, outregs;				// calling registers

	/*
		Make the call
	*/
	inregs.x.ax = GET_REVISION;
	intdos(&inregs, &outregs);

	return (outregs.x.ax);

}
