//
//   $Id: SHOWERR.C 1.4 2002/10/15 19:58:14Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**********************************************************

	SHOWERR.C created  8/10/95  by Drew Gislason
	Copyright (c) 1995-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.

	Usage:
		SHOWERR <filename> [<ignoremsgs>]

	Looks for lines with "error" or "warning" or "Fatal".  Will
	show all lines that have error or warning.  Optionally,
	a file may be provided that has messages to be ignored,
	for instance, the assembler spits out:

		Error messages:    None
		Warning messages:  None

	which you obviously don't want to see since there were
	no errors.

	Example:
		C>MAKE -B >x.y
		C>SHOWERR x.y ignore.txt

***********************************************************/
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_LINE_LEN 128

// maximum length of ignore lines
#define MAX_IGNORE_LEN		(20*1024)

static char ignoreLines[MAX_IGNORE_LEN+1];

/*
	Does this line match any of the lines in the lineArray?
	If so, return a pointer to the line.
*/
char * LineMatch(char *line, char *lineArray)
{
	int len = strlen(line);
	char * p;

	// look through each line
	while(*lineArray)
	{
		// does this line match?
		if(strncmp(line, lineArray, len) == 0)
			return lineArray;

		// try next line
		p = strchr(lineArray, '\n');
		if(p)
			lineArray = p+1;
		else
			lineArray = lineArray + strlen(lineArray);
	}

	// no match found
	return 0;
}

/*
	Look for the substring "str" at any point on the
	line, ignoring case.
*/
char * stristr(char *str, char *line)
{
	int len = strlen(str);

	// scan the line for the substring
	while(*line)
	{
		// did we match the first character?
		if(toupper(*str) == toupper(*line))
		{
			// did we match the whole substring?
			if(strnicmp(str, line, len) == 0)
				return line;
		}

		// try next character in the line
		++line;
	}

	// not found
	return 0;
}

// put a string to the console, without adding a '\n' at the end
void putsNoCR(char *str)
{
	while(*str)
	{
		putchar(*str);
		++str;
	}
}

int main(int argCount, char *argString[])
{
	FILE * fp;

	// argument count
	if(argCount < 2)
	{
		// signon
		printf("SHOWERROR v1.0\n");

		printf("Usage: SHOWERR <filename> [<ignoremsgs>]\n");
		return 1;
	}

	// get (optional) ignoremsgs files
	ignoreLines[0] = 0;
	if(argCount >= 3)
	{
		FILE * fpIgnore;
		unsigned bytesRead;

		fpIgnore = fopen(argString[2],"r");
		if(fpIgnore)
		{
			bytesRead = fread(ignoreLines, sizeof(char), MAX_IGNORE_LEN, fpIgnore);
			ignoreLines[bytesRead] = 0;
			fclose(fpIgnore);
		}
	}

	// get input file name
	fp = fopen(argString[1],"r");
	if(!fp)
	{
		printf("Failed to open input file %s!\n",argString[1]);
		return 2;
	}

	// loop until end of file on stdin
	while(1)
	{
		char * foundAt;
		char line[MAX_LINE_LEN+1];

		// get a line (break at the end)
		if(!fgets(line, MAX_LINE_LEN, fp))
			break;

		// look for warning or error, ignoring case
		foundAt = stristr("warning", line);
		if(!foundAt)
			foundAt = stristr("error", line);
		if(!foundAt)
			foundAt = stristr("fatal", line);

		// line contains error? check against lines to ignore
		if (foundAt)
		{
			// if we shouldn't ignore it, spit it to the screen
			if(!LineMatch(line, ignoreLines))
				putsNoCR(line);
		}
	}

	// everything worked!
	return 0;
}
