//
//   $Id: COMBAT.C 1.10 2003/01/28 17:27:02Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:COMBAT.C

	Copyright (c) 1989-2002 Datalight, Inc.
	Datalight Confidential and Proprietary
   All Rights Reserved.

	Batch processing, including getting the "command line"

	$DOCHISTORY
	03/01/1999 dg  Added LFNs to batch files
	07/16/1998 dg  %path%; now works on cmdline
	07/16/1998 dg  now includes "datalght.h"
	08/20/1997 de  SPR 1361: revised to fix batch processing bug
	07/20/1989 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__ 1

// ansi includes
#include <stdio.h>
#include <string.h>
#include <dos.h>
#include <dir.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "command.h"
#include "comprot.h"

int fIsCall;
struct BatchControlBlock far * pCurBCB = 0;


/*
	$DOCPROC:GetCommand

	Get a command from the DOS prompt, BATch file or pipe. Place it
	into "szLine". Uses "BATCH_LINELEN" as maximum line length.

	Returns FALSE if one command is over (end of batch).
*/
bool PASCAL GetCommand(char szLine[])
{
	int iLen, fd;

do_line:

	// unless an @ appears at the start of a line or !echo, show it
	//
	fDontShowLine = FALSE;
	fIsCall = FALSE;

	// do we get our input from a pipe?
	//
	if(c_pipe_ptr)
		strcpy(szLine, c_pipe_ptr);

	// get command line from user
	//
	else if(!c_in_batch)
	{
		//
		//
		if( c_one_only )
		{
			return FALSE;
		}

		// restore old string F3
		//
		memcpy(szLine, c_lastline, sizeof(c_lastline));
		szLine[0] = LINELEN;

		// only display prompt if echo is on
		if(c_echo)
			prompt_string();

		// get a string from the user
		getstr(szLine, GET_DOSKEY);

		// save this string for F3
		memcpy(c_lastline, szLine, sizeof(c_lastline));

		// convert from DOS get string to asciiz string
		iLen = szLine[1];
		memcpy(szLine, szLine+2, iLen);
		szLine[iLen] = 0;

		// expand things like %env%
		BatchExpand(szLine, BATCH_LINELEN, FALSE);
	}

	/* get command line from .BAT file */
	else
	{
open_again:
		fd = dos_faropen(pCurBCB->batName, RD);
		if(fd < 0)
		{
			// can't reopen this one! give up and go to parent batch file
			// (if any)
			PrintError(ENOBATCH);
			DestroyCurBCB();
			if(!c_in_batch)
				return FALSE;	// no longer in batch file tree
			goto open_again;
		}

read_again:
		dos_seek(fd, c_bat_pos, SEEK_SET);
		iLen = ReadBatchLine(fd, szLine);

		if(iLen < 0)
		{
			// we are done with this batch file
			//
			DestroyCurBCB();
			iLen = 0;
			dos_close(fd);

			// only this one batch file was to be run, terminate
			//
			if(c_one_only)
				return FALSE;

			// end of this batch file, get another command
			//

//  NOT compatible MSDOS 7.1 formatting
//
//			if(c_echo)
//				printf("\n");

			goto do_line;
		}

		// process the file read into a command line
		//
		else
		{
			c_bat_pos += iLen;

         // we've moved forward, now step past the end of line
         //
         if( szLine[iLen] == '\r' )       /* assume cr/lf */
         {
            c_bat_pos += 2;
         }
         else if( szLine[iLen] == ctrl('Z') )
         {
            c_bat_pos++;
         }
		}

		// skip all preceding whitespace
		//
		szLine[iLen] = NULL;
		strcpy(szLine, LFNSkipWhite(szLine));

		// ignore labels, these are not printed nor counted as a line
		//
		if(c_in_batch && szLine[0] == ':')
			goto read_again;

		// we've got our command, close the file
		//
		dos_close(fd);

		// Process batch substitutions (ie %1, %path%, ...)
		//
		if(BatchExpand(szLine, BATCH_LINELEN, TRUE) && c_echo)
		{
			prompt_string();
			printf("%s\n",szLine);
		}
	}

	// we got a line
	return TRUE;
}


/*
	$DOCPROC:BatchExpand

	This routine does 2 things:
	1) Handle leading '@'. Sets "fDontShowLine" if bInBatch and
		string starts with an '@'.
	2) Expand the '%' symbol.
		"%%"		- expands into "%" (bInBatch only)
		"%1"		- expands into the argument (bInBatch only)
		"%env%"	- expands into environment variable
		"%"		- by itself is thrown out

	bInBatch is TRUE for batch files, FALSE for normal cmdline.
	This is because the rules differ slightly for batch files vs.
	command line.

	Returns TRUE if the line should be displayed.
*/
int PASCAL BatchExpand(char *str, int maxSZLen, bool bInBatch)
{
	int show = TRUE;
	int len, alen, nlen;
	char *strEnd, *arg, far *farg;

	// valid parameters
	assert(strlen(str) <= maxSZLen);

	// truncate string. cannot be greater than maxSZLen
	if(strlen(str) > maxSZLen)
		str[maxSZLen] = 0;

	strEnd = str+maxSZLen-1;			// remember max non-NUL char position

	/* if the BATch line begins with '@', don't display it */
	if(bInBatch && *str == '@')
	{
		memmove(str, str+1, strlen(str));		/* remove the '@' symbol */
		show = FALSE;
		fDontShowLine = TRUE;
	}

	/*
		Look for "%" symbols.  They can mean one of 4 of things:
		"%%"		- expands into "%"
		"%1"		- expands into the argument
		"%env%"	- expands into environment variable
		"%"		- by itself is thrown out
	*/
	while(1)
	{
		/* look for the next */
		str = intl_strchr(str, '%');
		if(!str)
			break;

		arg = str+1;				// next character in string
		len = strlen(str);		// length excluding the NULL

		/* convert "%%" to "%" */
		if(*arg == '%')
		{
			memmove(str, str+1, len);
			++str;
			len--;
		}

		/* convert "%1" to argument */
		else if(*arg>='0' && *arg<='9')
		{
			// when at the cmdline, simply leave the %1 in the string
			if(!bInBatch)
			{
				str += 2;
				continue;
			}

			// convert the %1 to argument it represents
			// get ptr to and length of %n arg
			alen = BatchArgLen(*arg-'0', &farg);

			// expand, but not past maxlen of string
			nlen = 2;
			if((str+alen+len-nlen) > strEnd)
				alen = strEnd - (str+len-nlen);

			// open up string and copy in %n
			memmove(str+alen, str+nlen, 1+len-nlen);
			if(alen)
				farmemcpy((char far*)(str), farg, alen);

			// on to next string postition
			str += alen;
		}

		/* put env string in place of "%name%" */
		else if((arg = intl_strchr(arg, '%')) != NULL)
		{
			*arg = NULL;				/* make asciiz string out of "name" */

			// remember length of "%name%"
			nlen = ((unsigned)arg-(unsigned)str) + 1;
			arg  = GetEnvironmentVariable(str+1);

			/* copy env string into where "%name%" was */
			if(arg)
			{
				alen = strlen(arg);
				if((str+alen+len-nlen) > strEnd)		 // limit expansion to sizeof(str)
					alen = strEnd - (str+len-nlen);

				memmove(str+alen, str+nlen, 1+len-nlen);
				memmove(str, arg, alen);
				str += alen;
			}

			/* the "name" was not in the environment, throw it out */
			else
				memmove(str, str+nlen, 1+len-nlen);
		}

		/* a "%" by itself is thrown out */
		else
		{
			if(bInBatch)
				strcpy(str, str+1);
			else
				++str;
		}
	}

	/* return whether or not to show this line */
	return show;
}


/*
	Return length of an argument and a pointer to that argument.
*/
int PASCAL BatchArgLen(int num, char far **p)
{
	char far * arg;
	int  len=0, i;
	bool fInQuotes = FALSE;
	bool fIsSpace = FALSE;

	arg = FarLFNSkipWhite(pCurBCB->batArgs);
	for(i=0; i<num; ++i)
		arg = FarLFNNextArg(arg);
	*p = arg;

	while((*arg) && (!fIsSpace))
	{
		if( *arg == '\"' )
		{
			fInQuotes = !fInQuotes;
		}

		++len;
		++arg;

		fIsSpace = DL_IS_SPACE(*arg) && (!fInQuotes);
	}

	return len;
}

/*
	$DOCPROC:ForExpand

	Assumes line can be expanded up to BATCH_LINELEN.
*/
void PASCAL ForExpand(char *szLine, char *szForName, char cForVariable)
{
	char *psz;
	int	iForNameLen = strlen(szForName);
	int	iLineSize   = strlen(szLine)+1;		// includes the NULL

	psz = szLine;
	while(psz)
	{
		// expand for variable?
		psz = intl_strchr(psz, '%');
		if(psz && *(psz+1) == cForVariable)
		{
			// make an opening big enough (may have to truncate iForNameLen)
			if(iLineSize + iForNameLen - 2 > BATCH_LINELEN)
				iForNameLen = BATCH_LINELEN - iLineSize + 2;
			memmove(psz+iForNameLen, psz + 2, iLineSize - 2 - (psz-szLine));

			// add in the variable name and update line size
			memmove(psz, szForName, iForNameLen);
			iLineSize += iForNameLen - 2;
		}
	}
}

/*
	$DOCPROC:CreateBCB

	Create a batch control block. We are spawning a new batch file.
	Returns NULL if not enough memory for the BCB.
*/
struct BatchControlBlock far * PASCAL CreateBCB(void)
{
	int i;
	unsigned seg;
	struct BatchControlBlock far * bat;

	// Free memory for dos_malloc by reallocating COMMAND's transient block
	//
	farcall(res_openhole);

	// try to allocate a block and if successful initialize the new BCB
	//
	seg = dos_malloc((sizeof(struct BatchControlBlock) + 15)/16);
	bat = MK_FP(seg, 0);
	if(seg)
	{
		farmemset(bat, 0, sizeof(struct BatchControlBlock));

		bat->prevBCBSeg = FP_SEG(pCurBCB);	// link old BCB into the CALL stack
		pCurBCB         = bat;  				// new BCB is the current one
		c_curBCBSeg     = seg;					// let Novel know where BCB is
		c_in_batch      = TRUE;					// let XCOPY know it ran from a .BAT

		for(i = 0; i < MAX_BATCH_ARGS; i++)
			pCurBCB->batArgOffsets[i] = 0xFFFF;
	}
	else
		PrintError(ENOMEM);
	farcall(res_closehole);	// take back all remaining memory

	return bat;
}

/*
	in case we have "CALL"ed a batch file, restore the
	state of the previous batch file.
*/
void PASCAL DestroyCurBCB(void)
{
	unsigned prevSeg, doneSeg;

	if(pCurBCB != 0)
	{
		doneSeg = FP_SEG(pCurBCB);		// get seg of current BCB
		if(doneSeg != 0)					// and see if its valid
		{
			// have valid BCB address -- make prev BCB the current BCB
			//
			prevSeg = pCurBCB->prevBCBSeg;
			pCurBCB = MK_FP(prevSeg, 0);

			c_curBCBSeg = prevSeg;

			dos_free(doneSeg);			// throw away the old BCB
		}
	}
	c_in_batch = (pCurBCB != 0);		// reset flag if no .BAT running now
}

/*
	End all batches. This frees all the memory used to
	remember previous batch files and restores the state
	of echo.
*/
void PASCAL EndAllBatches(void)
{
	while(pCurBCB != 0)
	{
		DestroyCurBCB();
	}
}

/*
	Open a batch file. The filename is already absolute and has been
	verified to exist.

	Returns TRUE if worked. FALSE if failed (not enough memory).
*/
bool PASCAL BatchOpen(char szName[], char szArgs[])
{
	int	fd, i;
	bool	bOldEchoState;
	struct BatchControlBlock far * lpNewBCB;
	char far *lpsz;
	unsigned char far * lpHptr;
	unsigned uSize;
	bool fAddQuotes = FALSE;
	char * pTemp;


	// get old echo state, if no previous .BAT file, make echo ON
	//
	if(c_in_batch)
		bOldEchoState = pCurBCB->batEcho;

   else
	{
		// remember old redirection
		lpHptr = *(void far * far *)MK_FP(_psp, 0x34);
		c_bat_redirect[0] = *lpHptr++;
		c_bat_redirect[1] = *lpHptr;

		// echo is on when starting a new batch file
		bOldEchoState = TRUE;
	}

	// make a new BCB and initialize it if CALLing or starting anew
	//
	if(!c_in_batch || fIsCall)
		lpNewBCB = CreateBCB();

	// already in a batch file
	else
	{
		// we are chaining .BAT files and want to reuse the existing BCB.
		// Reinitialize the name and arg fields. Leave prevBCBSeg as it is.
		// Echo and file position handled below. Ignore unknown fields.
		//
		lpNewBCB = pCurBCB;

		i = sizeof(pCurBCB->batArgOffsets) +
			 sizeof(pCurBCB->batName) + sizeof(pCurBCB->batArgs);
		farmemset(&pCurBCB->batArgOffsets, 0, i);

		for(i = 0; i < MAX_BATCH_ARGS; i++)
			pCurBCB->batArgOffsets[i] = 0xFFFF;
	}

	// if we have a valid (possibly new) BCB
	//
	if(lpNewBCB != 0)
	{
		lpsz = (char far *)(szName);
		farmemcpy(pCurBCB->batName,lpsz,strlen(szName)+1);

		lpsz = (char far *)(szArgs);
		uSize = strlen(pszCmdName);

		pTemp = pszCmdName;
		while( *pTemp )
		{
			if( *pTemp == ' ' )
			{
				fAddQuotes = TRUE;
				uSize += 2;
				break;
			}

			pTemp++;
		}

		if( fAddQuotes )
		{
			farmemcpy(pCurBCB->batArgs,"\"",1);
			farmemcpy(&(pCurBCB->batArgs)[1],pszCmdName,uSize);
			farmemcpy(&(pCurBCB->batArgs)[uSize-1],"\"",1);
		}
		else
		{
			farmemcpy(&(pCurBCB->batArgs)[0],pszCmdName,uSize);
		}
		farmemcpy(&(pCurBCB->batArgs)[uSize],lpsz,strlen(szArgs)+1);

		pCurBCB->batPos  = 0L;
		pCurBCB->batEcho = bOldEchoState;
	}

	// we have opened the new batch file and are no longer CALLing
	//
	fIsCall = FALSE;

	return TRUE;
}

// is this a comment line?
bool PASCAL IsCommentLine(char *szCmdLine)
{
	// blank lines and labels are "comment lines"
	szCmdLine = LFNSkipWhite(szCmdLine);
	if(*szCmdLine==0 || *szCmdLine==':')
		return TRUE;

	// REM is also a comment line
	if(strncmpi(szCmdLine,"REM",3) == 0 && DL_IS_SPACE(szCmdLine[3]))
		return TRUE;

	// not a comment line
	return FALSE;
}
