//
//   $Id: COMCOPY.C 1.11 2002/10/15 20:08:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:COMCOPY.C

	Copyright (c) 1999-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	This module contains the code specific to the COPY command.

   $DOCHISTORY
	03/01/1999 dg  Created by Drew Gislason
***************************************************************************/
// ansi includes
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <stdio.h>
#include <string.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dlcntry.h"
#include "dldosio.h"
#include "dlerrcod.h"
#include "command.h"
#include "comprot.h"

// COPY command variables
bool fCopyOverwritePrompt;
char szCopyDrive[MAXDRIVE];			// drive size didn't change
char szCopyDirectory[DL_MAXPATH];	// must be big enough to hold LFN
char szCopyFileName[DL_MAXPATH];		// must be big enough to hold LFN

char szCopyDelimiters[] = "+:";

/*
	$DOCPROC:SkipCopyWhite
	Skip whitespace between COPY args. A '+' is considered whitespace.
*/
char * PASCAL SkipCopyWhite(char *szArg)
{
	while(iscopyspace(*szArg))
		++szArg;
	return szArg;
}

/*
	$DOCPROC:EndCopyArg

	Just like LFNEndArg, but understands that '+' separates arguments when
	not in quotes, and understands copying to a device name.

	Examples:
	a+b  				is 2 arguments
	"a"+"b"			is 2 arguments
	"a+b"				is 1 argument
	a:x.y				is 1 argument
	con:x.y        is 2 arguments
*/
char * PASCAL EndCopyArg(char *szArg)
{
	char * pszEndArg;

	// special case: handle a:x.y as 1 arg, but handle con:x.y as 2 args
	pszEndArg = LFNEndSpecialArg(szArg, szCopyDelimiters);
	if(*pszEndArg == ':' )
	{
		if((pszEndArg-szArg) == 1)
		{
			pszEndArg = LFNEndSpecialArg(pszEndArg+1, szCopyDelimiters);
		}
		else
		{
			pszEndArg++;
			pszEndArg = LFNSkipWhite(pszEndArg);
		}
	}

	// return end of the argument
	return pszEndArg;
}

/*
	$DOCPROC:NextCopyArg
*/
char * PASCAL NextCopyArg(char *szArg)
{
	return SkipCopyWhite(EndCopyArg(szArg));
}

/*
	$DOCPROC:GetCopyArg

	Get an argument from the copy command.
*/
void PASCAL GetCopyArg(char *szDst, char *szArg)
{
	char *psz;
	int	iLen;

	// copy this argument, use "end copy" rules (plus is whitespace)
	psz = EndCopyArg(szArg);
	iLen = psz - szArg;
	memmove(szDst, szArg, iLen);
	szDst[iLen] = 0;

	// strip off any quotes
	LFNStripArgQuotes(szDst);
}

/*
	$DOCPROC:ProcessCopyArgs

	This multi-purpose function encapsulates the tricky "setup" required
	by a COPY command. It:

	1. Validates # of copy arguments
	2. Determines copy vs. concatinate
	3. Determines # of input arguments
	4. Sets up the destination name

	On Entry
		szArg					the command line arguments
		(global) szTemp	used to create destination name

	On Exit
		piCount				number of source arguments
		pfConcatinate		is this a copy or concatinate?
		szCopyDrive			drive of destination
		szCopyPath			valid path of destination
		szCopyFileName		filename (may contain wildcards) of destination

	Returns TRUE if worked. Returns FALSE and displays message if error.
*/
bool PASCAL ProcessCopyArgs(char *szArgs, int *piArgCount, bool *pfConcatinate)
{
	extern char cmdstr_copied[];

	int	iCount;
	bool	fConcatinate;
	char *psz;
	char *pszDstFileName;
	char *pszTooManyArgs;
	bool	fIsWildSrc, fIsWildDst;
	bool	fInSrc;

	// assume not concatinating and no args
	iCount = 0;
	fConcatinate = FALSE;

	// count source filespecs and find destination
	psz = szArgs;
	pszDstFileName = 0;
	while(*psz)
	{
		// one more arg
		++iCount;

		// find end of arg (maybe + separating them)
		psz = EndCopyArg(psz);
		psz = LFNSkipWhite(psz);

		// concatinating?
		if(*psz == '+')
		{
			++psz;
			psz = LFNSkipWhite(psz);
			fConcatinate = TRUE;
			continue;
		}

		// point destination to last argument
		if(*psz)
		{
			pszDstFileName = psz;
			break;
		}
	}

	// should be no arguments after the filename
	if(pszDstFileName)
	{
		psz = EndCopyArg(pszDstFileName);
		psz = LFNSkipWhite(psz);

		// cannot be a '+' after destination. that's invalid.
		if(*psz == '+')
		{
			PrintError(ENOCOPY);
			printf(cmdstr_copied, 0);
			return FALSE;
		}

		// too many args. set count to 0 to generate invalid # args error below
		if(*psz)
			iCount = 0;
	}

	// should have at least 1 arg
	if(iCount == 0)
	{
		PrintError(ENUMPARM);
		return FALSE;
	}

	// No destination specified. Assume current directory with 1st filespec
	if(!pszDstFileName)
		pszDstFileName = NamePart(szArgs);

	// now set up destination name
	GetCopyArg(szTemp, pszDstFileName);
	if(!ValidateDirectory(szTemp))
		return FALSE;

	// save destination information
	strcpy(szCopyDrive, szSplitDrive);
	strcpy(szCopyDirectory, szSplitDirectory);
	strcpy(szCopyFileName, szSplitFileName);

	// special case: if the source is wild, but not the dest, then
	// it's really concatinate, not copy.
	if(!fConcatinate)
	{
		GetCopyArg(szTemp, szArgs);
		fIsWildSrc = CheckWild(szTemp);
		if( !fIsWildSrc )
		{
			//  If the source is a directory then the wild cards are assumed
			//
			fIsWildSrc = DirExists(szTemp);
		}
		fIsWildDst = CheckWild(szCopyFileName);
		if(fIsWildSrc && !fIsWildDst)
			fConcatinate = TRUE;
	}

	// return count and concatinate results
	*pfConcatinate = fConcatinate;
	*piArgCount    = iCount;

	// everything worked
	return TRUE;
}

/*
	$DOCPROC:BuildName

	Create a wildcard name based on a mask and source.
*/
char * PASCAL BuildName(char *szDst, char *szMask, int iMaskLen,
						char *szSrc, int iSrcLen)
{
	while(iMaskLen)
	{
		// match to end of string
		if(*szMask == '*')
		{
			memmove(szDst, szSrc, iSrcLen);
			szDst += iSrcLen;
			break;
		}

		// just use the mask char
		if(*szMask != '?')
			*szDst++ = *szMask;

		// match single character wildcards too
		else if(iSrcLen)
			*szDst++ = *szSrc;

		// on to next character
		++szMask;
		--iMaskLen;

		if(iSrcLen)
		{
			++szSrc;
			--iSrcLen;
		}
	}

	// done, null terminate the string
	*szDst = 0;

	// return end of dest
	return szDst;
}

/*
	$DOCMODULE:MatchCopyName

	Merge a wildcard name into a destination name.

	On Entry
		szMask		a wildcard mask
		szSrc			a legal filename (already tested to exist)
	On Exit
		szDst			contains merged name

	Examples with mask "ab*.h":
   src name     dst name
	.c           ab.h
	1.c          ab.h
	2long.c      abong.h
	3company     abompany.h
*/
void PASCAL MatchCopyName(char *szDst, char *szMask, char *szSrc)
{
	char *szExtMask;
	char *szExtSrc;
   bool fIsWildSrc;

   fIsWildSrc = CheckWild(szMask);

	// find the extensions of mask and src name
   //
	szExtMask = ExtPart(szMask);
	szExtSrc  = ExtPart(szSrc);

	// build the dst (name part only) 1 char at a time
	szDst = BuildName(szDst, szMask, szExtMask-szMask, szSrc, szExtSrc-szSrc);

   // no mask extension? we're done
   if(*szExtMask != '.')
   {
      if( fLFNPresent && fIsWildSrc )
      {
         // in an LFN environment, FNAME* matches FNAME.TXT even
         // though ExtPart() returns no extension.  So copy the rest
         // (including the zero terminator) and return
         //
         memmove(szDst, szExtSrc, strlen(szExtSrc)+1);
      }
      return;
   }

	// mask has extension, so dest will too
	++szExtMask;
	*szDst++ = '.';

	// skip past '.' on src extension
	if(*szExtSrc == '.')
		++szExtSrc;

	// build the dst extension 1 char at a time
	BuildName(szDst, szExtMask, strlen(szExtMask), szExtSrc, strlen(szExtSrc));
}

/*
	Copy a single file from source to destination filename. This is a
	single file copy only. Will end on ^Z if fAscii mode. Concatinates
	if fConcatinate.

	On Entry
		szSrcName		name of input file
		fAscii			copy in ascii mode (false = binary mode)
		fConcatinate	append this file to dest file
		fShowName		determines whether or not to show files names of
							copied files.
	Uses globals:
	szCopyDrive
	szCopyPath
	szCopyFilename
	fCopyOverwritePrompt

	Returns 0=not copied, 1=copied, -1=error
*/
int PASCAL CopyFile(char *szSrcName, bool fAscii, bool fConcatinate,
		bool fFirst, bool fShowName )
{
	extern char cmdstr_copyprompt[];

	static char cCtrlZ = ctrl('Z');
	char	szDst[DL_MAXPATH];
	int	fdIn, fdOut;				/* file handles */
	unsigned uDate, uTime;			/* preserve file date and time */
	char huge * lpBuff;				/* buffer for copying */
	char huge * lpOrgBuff;			/* original buffer */
	long	lSize;						/* buffer size (up to 640K) */
	ulong	ulCount, ulThisCount;	/* used to count bytes read */
	ulong	n;								/* temporary counter */
	ulong	ulBytesCopied;				/* total bytes copied */
	bool	fGotCtrlZ = FALSE;		/* only used if ascii */
	bool	fIsDevice;
	bool	fDstExists;
	bool	fDstSame;
	char *psz;
	char	c;
	int	iAction;
	int	iErrorCode;
	char * szNamePart;
	unsigned char ucModeData = 0;


	// determine the exact destination name
	strcpy(szDst, szCopyDrive);
	strcat(szDst, szCopyDirectory);
	if(lastchar(szDst) != '\\')
		strcat(szDst, "\\");
	MatchCopyName(szDst+strlen(szDst), szCopyFileName, NamePart(szSrcName));

	fDstExists = FileExists(szDst);
	fDstSame   = FALSE;
	if(fDstExists)
	{
		// get longname so comparison is the same
		GetLongName(szDst);

		// dst cannot be same as src
		GetAbsoluteName( szTemp, szSrcName );
		GetAbsoluteName( szTemp3, szDst );
		if( strcmpi( szTemp, szTemp3 ) == 0 )
		{
			// dst is same as src, that's OK for COPY A+B+C, which appends to A
			if(fConcatinate && fFirst)
				fDstSame = TRUE;

			// dst can't be same as src for most copies
			else
			{
				if(fShowName)
				{
					printf("%s\n", szSrcName);
				}
				PrintError(ENOCOPY);
				return -1;
			}
		}
	}

	// try to open the original file
	if(!SmartCreateOpenFile(OPEN_ACCESS_READONLY, A_NORMAL, FILE_OPEN,
					szSrcName, &fdIn, &iAction))
	{
		PrintError(ENOFILE);
		return -1;
	}

	// file or device as input?
	fIsDevice = isatty(fdIn);

	// ask for overwrite
	if(fDstExists && fCopyOverwritePrompt && !fConcatinate)
	{
		printf(cmdstr_copyprompt, szDst);
		c = GetKey("YNA");
		if(c == 'N')					/* no, don't overwrite this one! */
		{
			dos_close(fdIn);
			return 0;
		}

		/* quit asking! */
		if(c == 'A')
			fCopyOverwritePrompt = FALSE;
	}

	// display filenames only for files (not devices)
	else
	{
		if(fShowName)
		{
			if(!fIsDevice)
			{
				printf("%s\n", szSrcName);
			}
		}
	}

	// if it's a device, it's always ascii (never binary)
	if(fIsDevice)
		fAscii = TRUE;

	// copy A+B+C appends to A, no need to copy
	if(fConcatinate && fDstSame)
	{
		dos_close(fdIn);
		return 1;			// copied 1 file
	}

	// on first file, make sure to truncate if required
	if(fFirst)
		fConcatinate = FALSE;

	// remember original date of file
	if(fConcatinate)
		uDate = 0;
	else
		get_date_time(fdIn, &uDate, &uTime);

	// allocate as much memory as possible for the copy
	lpOrgBuff = lpBuff = UseBigBuffer(&lSize);

	// delay opening destination until we have read source
	// helps to prevent floppy swap madness
	fdOut = INVALID;

	/* copy as much as possible */
	ulBytesCopied = 0;
	while(!fGotCtrlZ)
	{
		// get as much data as possible
		ulCount = 0;
		while(!fGotCtrlZ && ulCount < (ulong)lSize)
		{
			/* remember, reading from CON: might only give a few chars */
			ulThisCount = HugeRead(fdIn, lpBuff, lSize-ulCount);
			if(!ulThisCount)
				break;

			/* we must check for a ^Z in ascii files */
			if(fAscii)
			{
				for(n=0; n<ulThisCount; ++n)
				{
					/* stop processing on a ctrl-Z */
					if(*lpBuff == ctrl('Z'))
					{
						ulThisCount = n;
						fGotCtrlZ = TRUE;    /* simply set the flag for now */
						break;
					}
					++lpBuff;
				}
			}
			else
				lpBuff += ulThisCount;

			/* record size of read */
			ulCount += ulThisCount;
		}

		/* got no characters? we're done */
		if(!ulCount)
			break;

		/* NOW we can open/create the destination file */
		if(fdOut == INVALID)
		{
			if(fConcatinate)
				iAction = FILE_OPEN | FILE_CREATE;
			else
				iAction = FILE_CREATE | FILE_TRUNCATE;
			if(!SmartCreateOpenFile(OPEN_ACCESS_WRITEONLY, A_NORMAL, iAction,
						szDst, &fdOut, &iAction))
			{
				iErrorCode = dlCheckDOSError();
				if( iErrorCode == 3 )
				{
					PrintError(ENOFILE);
				}
				else
				{
					PrintError(ECREAT);
				}
				dos_close(fdIn);
				return -1;
			}

         // IR 3364 9-7-01 TD
         //
         // In the case of a "copy /b FILENAME.TXT PRN" we need to be in
         // RAW mode (otherwise a binary print will be halted with a CTRL-Z).
         //
         if( !fAscii )
         {
            ucModeData = dos_ioctl( fdOut,0,0 );
            if( ucModeData & 0x80 )       // if we're a device
            {
               dos_ioctl( fdOut,1,ucModeData|0x20 );
            }
         }

			// seek to end when concatinating
			if(fConcatinate)
			{
				n = dos_seek(fdOut, 0L, SEEK_END);

				// when concatinating multiple ascii files, remove the ^Z
				if(fAscii && n)
					dos_seek(fdOut, n-1, SEEK_SET);
			}
		}

		/* try to write all those bytes */
		lpBuff = lpOrgBuff;
		n = HugeWrite(fdOut, lpBuff, ulCount);
		if(n != ulCount)
		{
			PrintError(ENOSPACE);
			dos_close(fdIn);

         // IR 3364 9-7-01 TD
         //
         // If we saved off a value before the write, we need to restore it
         //
         if( ucModeData )
         {
            dos_ioctl( fdOut,1,ucModeData );
            ucModeData = 0;
         }

			dos_close(fdOut);
			SmartWildcardDelete(szDst);
			return -2;
		}

		// update bytes copied
		ulBytesCopied += ulCount;
	}

	// append a ^Z on ascii files
	if(fAscii && !fIsDevice)
		dos_write(fdOut, (char far *)&cCtrlZ, 1);

	// preserve orignal date and time (but not on concatination)
	if(uDate)
		set_date_time(fdOut, uDate, uTime);

	// we are done w/ the files
	dos_close(fdOut);
	dos_close(fdIn);

	// if didn't copy a file (if 1st. big_read() returned 0) don't say we did.
	return (ulBytesCopied) ? 1 : 0;
}

/*
	$DOCPROC:HugeRead
	Read from a file (up to the long size).
	Returns # of bytes read. May be 0.
*/
ulong PASCAL HugeRead(int fd, char huge * lpBuff, ulong ulSize)
{
	unsigned nCount, n;
	ulong ulTotal = 0L;

	do
	{
		if(ulSize >= 0xF000)
			nCount = 0xF000;
		else
			nCount = (unsigned)ulSize;
		n = dos_read(fd, (char far *)lpBuff, nCount);
		ulTotal+= n;
		if(n != nCount)
			break;

		lpBuff += nCount;
		ulSize -= nCount;
	} while(ulSize);

	return ulTotal;
}

/*
	$DOCPROC:HugeWrite
	Write to a file (up to the long size). Returns # of bytes written,
	which may be 0 (no disk free space).
*/
ulong PASCAL HugeWrite(int fd, char huge * lpBuff, ulong ulSize)
{
	unsigned nCount, n;
	ulong ulTotal = 0L;

	do
	{
		if(ulSize >= 0xF000)
			nCount = 0xF000;
		else
			nCount = (unsigned)ulSize;
		n = dos_write(fd, (char far *)lpBuff, nCount);
		ulTotal += n;
		if(n != nCount)
			break;

		lpBuff += n;
		ulSize -= n;

	} while(ulSize);

	return ulTotal;
}
