//
//   $Id: COMDIR.C 1.7 2002/10/15 20:08:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
	$DOCMODULE:COMDIR.C

	Copyright (c) 1999-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	This contains all the DIR related functions. DIR requires enough support
	that it warrents its own file.

	DIR	[drive:][path][filespec] [/A:?] [/B] [/L] [/O:?] [/P] [/S] [/V] [/W] [/4]

	/A?  Display only files with a matching attribute as show below
	     A Archive   D Directory    H Hidden     R Read Only
	     S System    X Show Attrs   - Prefix for not
	/B   Display in bare format
	/L   Display in lowercase
	/O?  Sort directory listing by item as shown below
	     N Name      E Extension           D Date and time
	     S Size      G Group directories   - Prefix for reverse
	     A Last access date
	/P   Display one screen full and pause
	/S   Display files in subdirectories as well
	/V   Verbose directory information
	/W   Display in wide format without date, time, or size
	/4   Displays year with 4 digits (default 2 digits)

	Environment variables:
	DIRCMD	allows "default" switches to be specified.
	DIRSIZE	rows [cols] allows for non-PC standard screen sizes, and
				overrides BIOS screen size settings

   $DOCHISTORY
	02/23/1999 dg  Added verbose
	02/21/1999 dg	Created by Drew Gislason
***************************************************************************/
#define __FILENUM__ 8

// ansi includes
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <stdio.h>
#include <string.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlcntry.h"
#include "dl64.h"
#include "command.h"
#include "comprot.h"

// a convenient type for sorting directories
typedef FIND huge * LPFFBLK;

struct WIN32_FIND_DATA lfnFindData;

// DIR command variables
char szPausePath[DL_MAXPATH];			// for use by pause
char szDirPath[DL_MAXPATH];			// temporary pathname
char szRecursePath[DL_MAXPATH];
char szRecurseFileMask[DL_MAXPATH];
int fDirAttrSwitch;			/* TRUE if /A used */
int fDirAttr;					/* /A will show hidden files too */
int fDirNonAttr;				/* /A will show hidden files too */
int iaDirAttrList[6];
int fDirBareFormat;			/* /B Bare format */
int fDirLowerCase;			/* /L Use lower case */
int fDirPause;					/* /P can pause after 1 page */
int fDirrecurse;				/* /S Recurse in to sub-directories */
int fDirsort;					/* /O Sort dir */
int iaDirSortList[6];
int fDirVerbose;				/* /V verbose */
int fDirWide;					/* /W can make it print wide */
int fDir4DigitDate;			/* /4 use 4 digit dates... ignored, always true */

// for displaying numbers in string format
char szDirNum[] = "%*s";

// DIR info about disk
ulong ulSectorsPerCluster;
ulong ulBytesPerSector;
ulong	ulAvailableClusters;
ulong ulTotalClusters;
uquad uqBytesFree;
uquad uqTotalDiskSpace;

// totals added up during DIR and DIR /S
ulong ulThisDirFiles;
ulong ulThisDirDirectories;
uquad uqThisDirBytes;
uquad uqThisDirBytesAllocated;
ulong ulTotalFiles;
ulong ulTotalDirectories;
uquad uqTotalBytes;
uquad uqTotalBytesAllocated;


/*
	Size of directory fields.

	80 Col Example:
   THISIS~1       <DIR>        02-22-1999  6:34a This is a long subdir
   FILENAME EXT         2,337  02-22-1999  5:51a MAKEFILE
           90 file(s)        804,446 bytes
            3 dir(s)   2,531,389,440 bytes free

	40 Col Example:
   COMMAND  COM    28967 04-03-1998 11:28a
   SUBDIR       <DIR>    04-03-1998 11:28a
           1 file(s)     38451 bytes
                       1365504 bytes free
*/
struct DirFields
{
	int	iCols;					// # of columns expected in this dirfield
	int	iShortNameWidth;		// leftmost name
	int	iFileSizeWidth;		//
	char	*szDirStr;				// directory string
	char	*szInBetweenDate;		// spaces between date/time
	int	iCountWidth;			//
	int	iBytesWidth;			// bytes free, etc..
	int	iWideWidth;				// for DIR /w
	char	*szNumFmt32;			// number format for 32 bits
	char	*szNumFmt64;			// number format for 64 bits
};

struct DirFields dirField80Col =
{
	80, 13, 13, "  <DIR>", "  ", 10, 14, 15, "%,*lu", "%,*qd"
};

struct DirFields dirField80Col622 =
{
	80, 13, 13, "<DIR>", "  ", 10, 14, 15, "%,*lu", "%,*qd"
};

struct DirFields dirField40Col =
{
	40, 13, 7, "<DIR>", " ", 6, 12, 13, "%*lu", "%*qd"
};

struct DirFields * dfDir = &dirField80Col;

/*
	Read in and display each directory. Will Recurse into subdirectories
	if /S specified. Does no display itself, but relies on
	DisplayDirectory().

	Warning! This function is recursive, so minimize stack usage.

	Returns TRUE if worked. FALSE if directory invalid.
*/
bool PASCAL RecurseIntoDirectory(char *szPathName)
{
	bool	bFound;
	int	iLen;
	long	lSize;
	long	lCount;
	FIND dp;					/* findfirst/next */
	void far * lpDirBuff;

	/* attempt to change to new directory */
	if(!determine_dir(szPathName))
		return FALSE;

	/* if name or extension is not filled out, do it here */
	//  except if there is a '?' wild card in the name
	//
	if( (intl_strchr(szSplitFileName, '.') == 0) &&
			(intl_strchr(szSplitFileName, '?') == 0) )
	{
		strcat(szSplitFileName, ".*");
	}

	if(*szSplitFileName == '.')
	{
		memmove(szSplitFileName+1, szSplitFileName, strlen(szSplitFileName)+1);
		*szSplitFileName = '*';
	}

	/* find the files in this directory */
	sprintf(szDirPath, "%s%s", szSplitDrive, szSplitFileName);
	lCount = 0;
	bFound = SmartFindFirst((fDirAttrSwitch ? ALL_ATTR : A_SUBDIR),
					szDirPath, &dp);

	/* change back to the old directory */
	INT_set_cwd(szSplitOldCwd);

	/* get a buffer to hold all the filenames in */
	lpDirBuff = UseBigBuffer(&lSize);

	if(bFound)
	{
		// get all the directory entries for this directory
		lCount = GatherDirectory(lpDirBuff, lSize, &dp);

		// done, close the find
		if(!SmartFindClose(&dp))
			assert0();

		// update totals
		ulTotalFiles       += ulThisDirFiles;
		ulTotalDirectories += ulThisDirDirectories;
		AddQuad(&uqTotalBytes, &uqThisDirBytes);
		AddQuad(&uqTotalBytesAllocated, &uqThisDirBytesAllocated);

		// sort them? okie dokie...
		if(fDirsort)
			SortDirectory(lpDirBuff, lCount);
	}

	// display the directory
	sprintf(szDirPath, "%s%s", szSplitDrive, szSplitDirectory);
	DisplayDirectory(lpDirBuff, lCount, szDirPath);

	// recurse into subdirectories?
	if(IsDirRecurse())
	{
		sprintf(szRecursePath, "%s%s", szSplitDrive, szSplitDirectory);

		iLen = strlen(szRecursePath);
		add_path(szRecursePath, "*.*");
		bFound = SmartFindFirst(A_SUBDIR, szRecursePath, &dp);

		while(bFound)
		{
			szRecursePath[iLen] = '\0';
			add_path(szRecursePath, dp.name);

			// Don't try to look "down" into subdirectories
			// named "", ".", or "..".  It will only bring
			// you headaches.
			if(dp.name[0] && (dp.attr & A_SUBDIR)
				&& !IsDotDir(dp.name))
			{
				// save the file mask. we'll restore it later
				strcpy(szRecurseFileMask, szSplitFileName);

				// recurse into new directory with path
				add_path(szRecursePath, szSplitFileName);
				RecurseIntoDirectory(szRecursePath);
				del_path(szRecursePath);
			}

			bFound = SmartFindNext(&dp);
		}
		if(!SmartFindClose(&dp))
			assert0();

		// remove this name
		szRecursePath[iLen] = '\0';
	}

	return TRUE;
}

// any files or dirs this time?
bool PASCAL IsDirFiles(void)
{
	return (ulTotalFiles || ulTotalDirectories);
}

// zero out the totals, this is a new dir command
void PASCAL DirZeroTotals(void)
{
	ZeroQuad(&uqTotalBytes);
	ZeroQuad(&uqTotalBytesAllocated);
	ulTotalFiles = ulTotalDirectories = 0;
}

/*
	Get the freespace.
	Returns TRUE if worked, FALSE if failed.
*/
bool PASCAL DirGetFreeSpace(int drive)
{
	static char szDrive[] = "A:\\";
	static GDF32  gdfFreeSpace32;
	unsigned uSectorsPerCluster;
	unsigned uFreeClusters;
	unsigned uBytesPerSector;
	unsigned uTotalClusters;
	ulong	ulBytesPerCluster;

	// first, try to get freespace from FAT32 method
	*szDrive = drive+'A';
	if(dlGetDriveFreeSpace32(szDrive, &gdfFreeSpace32))
	{
		// set the globals based on the disk data
		ulSectorsPerCluster = gdfFreeSpace32.ulSectorsPerCluster;
		ulBytesPerSector    = gdfFreeSpace32.ulBytesPerSector;
		ulAvailableClusters = gdfFreeSpace32.ulAvailableClusters;
		ulTotalClusters     = gdfFreeSpace32.ulTotalClusters;
	}

	// try old FAT16 style function
	else if(__dlGetFreeSpaceOnDrive(drive+1, &uSectorsPerCluster,
		&uFreeClusters, &uBytesPerSector, &uTotalClusters))
	{
		ulSectorsPerCluster = uSectorsPerCluster;
		ulBytesPerSector    = uBytesPerSector;
		ulAvailableClusters = uFreeClusters;
		ulTotalClusters     = uTotalClusters;
	}

	// both methods failed, give up!
	else
		return FALSE;

	// calculate some totals
	ulBytesPerCluster = ulSectorsPerCluster * ulBytesPerSector;
	uqBytesFree = QuadMultiply(ulAvailableClusters, ulBytesPerCluster);
	uqTotalDiskSpace = QuadMultiply(ulTotalClusters, ulBytesPerCluster);

	// worked
	return TRUE;
}

/*
	display either:
	                      138,936,320 bytes free
   or
            4 dir(s)     138,936,320 bytes free
	depending on "fWithDirCount".

   NOTE: scale the output to MB when the total size
   of the drive is > four gigabytes (4L*1024L*1024L*1024L)
*/
void PASCAL DirDisplayFreeBytes(bool fWithDirCount)
{
	extern char cmdstr_dirdirs[];		// dir(s)
	extern char cmdstr_dirfreestr[];	// bytes free
	extern char cmdstr_dirfreestr2[];	// free
	extern char cmdstr_dirspace[];	// "%-*s"
	char * szMBString;


	// display dircount, or just blanks
	if(fWithDirCount && fIsFat32World )
	{
		printf(dfDir->szNumFmt32, dfDir->iCountWidth, ulTotalDirectories);
		printf(cmdstr_dirdirs);
	}
	else
	{
		//  Spacing is different for DOS 7.1 and DOS 6.22
		//
		if( fIsFat32World )
		{
			printf(cmdstr_dirspace, dfDir->iCountWidth+9, "");
		}
		else
		{
			printf(cmdstr_dirspace, dfDir->iCountWidth+7, "");
		}
	}

	// display bytes free
   //
   if( IsQuadGTFourGig( &uqTotalDiskSpace ) )
   {
      ScaleToMB( &uqBytesFree, &szMBString, TRUE );
		printf("%*s", dfDir->iBytesWidth+3, szMBString );
		printf(cmdstr_dirfreestr2);
	}
	else
	{
		printf(dfDir->szNumFmt64, dfDir->iBytesWidth, uqBytesFree);
		printf(cmdstr_dirfreestr);
	}
	DirDisplayLF();
}

/*
	Summarize the directory. Understands the many modes:

	Might be one of:

   1. Dir /s[/w] (wide or normal with subdir)
   Total files listed:
           87 file(s)        817,502 bytes
            8 dir(s)     138,903,552 bytes free

   2. Dir /s/v (verbose with subdir)
   Total files listed:
           87 file(s)        785,212 bytes
            8 dir(s)       3,178,496 bytes allocated
                         138,936,320 bytes free
                       1,077,411,840 bytes total disk space,  87% in use

   3. Dir [/w] (wide or normal)
            4 dir(s)     138,936,320 bytes free

   4. Dir /v
            8 dir(s)       3,178,496 bytes allocated
                         138,936,320 bytes free
                       1,077,411,840 bytes total disk space,  87% in use

*/
void PASCAL DirDisplaySummary(void)
{
	extern char cmdstr_dirlisted[];	// Total files listed:

	// recurse requires a total header and # files lines
	if(IsDirRecurse())
	{
		// Total files listed:
		DirDisplayLF();
		printf(cmdstr_dirlisted);
		DirDisplayLF();

		// display file bytes (and allocated for verbose)
		if( fIsFat32World )
		{
			DirDisplayFileBytes(ulTotalFiles, uqTotalBytes);
			if(IsDirVerbose())
				DirDisplayAllocated(ulTotalDirectories, uqTotalBytesAllocated);
		}
		else
		{
			DirDisplayFileBytes(ulTotalFiles+ulTotalDirectories, uqTotalBytes);
			if(IsDirVerbose())
				DirDisplayAllocated(ulTotalDirectories, uqTotalBytesAllocated);
		}
	}

	// verbose
	if(IsDirVerbose())
		DirDisplayTotalDiskSpace();

	// normal and wide
	else
		DirDisplayFreeBytes(TRUE);
}

/*
	Displays total disk space for verbose modes.

   Total files listed:
                         138,936,320 bytes free
                       1,077,411,840 bytes total disk space,  87% in use

   NOTE: scale the output to MB when the total size
   of the drive is > four gigabytes (4L*1024L*1024L*1024L)
*/
void PASCAL DirDisplayTotalDiskSpace(void)
{
	extern char cmdstr_dirtotal[]; 	// bytes total disk space, 10% in use
	extern char cmdstr_dirspace[];
	int	iPercentFree, iPercentUsed;
	char * szMBString;

	// display free space
	DirDisplayFreeBytes(FALSE);

	// compute percentage used
	if(ulTotalClusters < ulAvailableClusters || !ulTotalClusters)
		iPercentUsed = 100;		// invalid, use 100%
	else
	{
		// scale for 4 percentage digits (eg 100.8)
		if(ulAvailableClusters <= 4294967)
			iPercentFree = (int)((ulAvailableClusters*1000) / ulTotalClusters);
		else
			iPercentFree = (int)(ulAvailableClusters / (ulTotalClusters/1000));
		iPercentUsed = 1000 - iPercentFree;

		// round up if decimal pt .5 or greater
		if(iPercentUsed % 10 >= 5)
			iPercentUsed += 10;
		iPercentUsed /= 10;
	}

	// display space at start of line
   //
	printf(cmdstr_dirspace, dfDir->iCountWidth+9, "");
   if( IsQuadGTFourGig( &uqTotalDiskSpace ) )
   {
	   ScaleToMB( &uqTotalDiskSpace, &szMBString, TRUE );
		printf("%*s", dfDir->iBytesWidth+3, szMBString );
   }
   else
	{
		printf(dfDir->szNumFmt64, dfDir->iBytesWidth, uqTotalDiskSpace);
	}
	printf(cmdstr_dirtotal, iPercentUsed);
	DirDisplayLF();
}

/*
   Scales the output to MegaBytes, returning a pointer to a scaled
   version string.

   When dealing with "file(s)" and "allocated" dir space, we
   need to scale values greater than four gigabytes (4L*1024L*1024L*1024L)

   When dealing with "total disk space" and "bytes free", we
   need to scale when the total disk space (uqTotalDiskSpace)
   is greater than four gigabytes (4L*1024L*1024L*1024L).
   Therefore, we "force" this compare to happen with fAuto=TRUE.
*/
static bool PASCAL ScaleToMB( uquad * puqBytes, char ** pszMBString, int fAuto )
{
	static char szScaledString[20];

	if( fAuto || IsQuadGTFourGig( puqBytes ) )
	{
		DlScaleQuadToMegabytes( szScaledString, *puqBytes, TRUE );
		*pszMBString = szScaledString;
		return TRUE;
	}

	return FALSE;
}


/*
	display the standard:
	        86 file(s)        785,112 bytes
*/
static void PASCAL DirDisplayFileBytes(ulong ulCount, uquad uqBytes)
{
	extern char cmdstr_dirfiles[];	// file(s)
	extern char cmdstr_dirbytes[];	//  bytes
	char * szMBString;


	// display the line, using appropriate widths and number strings
	// for the # of display columns
	printf(dfDir->szNumFmt32, dfDir->iCountWidth, ulCount);
	printf(cmdstr_dirfiles);
	if( ScaleToMB( &uqBytes, &szMBString, FALSE ) )
	{
		printf("%*s", dfDir->iBytesWidth+3, szMBString );
	}
	else
	{
		printf(dfDir->szNumFmt64, dfDir->iBytesWidth, uqBytes);
		printf(cmdstr_dirbytes);
	}
	DirDisplayLF();
}


/*
	Displays the line:
         4 dir(s)       3,145,728 bytes allocated
*/
static void PASCAL DirDisplayAllocated(ulong ulDirectories, uquad uqBytesAllocated)
{
	extern char cmdstr_dirdirs[];			// dir(s)
	extern char cmdstr_dirallocated[];	// bytes allocated
	extern char cmdstr_dirallocated2[];	// bytes allocated
	extern char cmdstr_dirspace[];	   // "%-*s"
	char * szMBString;

	// display the line, using appropriate widths and number strings
	// for the # of display columns
	if( fIsFat32World )
	{
		printf(dfDir->szNumFmt32, dfDir->iCountWidth, ulDirectories);
		printf(cmdstr_dirdirs);
	}
	else
	{
		printf(cmdstr_dirspace, dfDir->iCountWidth+9, "");
	}
	if( ScaleToMB( &uqBytesAllocated, &szMBString, FALSE ) )
	{
		printf("%*s", dfDir->iBytesWidth+3, szMBString );
		printf(cmdstr_dirallocated2);
	}
	else
	{
		printf(dfDir->szNumFmt64, dfDir->iBytesWidth, uqBytesAllocated);
		printf(cmdstr_dirallocated);
	}
	DirDisplayLF();
}

/*
	Process all arguments to DIR.

	Returns TRUE if worked. Returns FALSE and displays a message if an
	argument is invalid.
*/
bool PASCAL do_dir_args(char *szArgs)
{
	struct ext_switch Attrs =
	{
		6, { {'R', 0x1},  {'H', 0x2}, {'S', 0x4}, {'X', 0x8}, {'D', 0x10},
			  {'A', 0x20} }
	};
	struct ext_switch SortOrder =
	{
		6, { {'N', 0x1},  {'E', 0x2}, {'G', 0x4}, {'S', 0x8}, {'D', 0x10},
			  {'A', 0x20} }
	};

	// determine how big the screen is and prepare for pausing
	DetermineScreenSize();

	// all the switches start out false
	fDirsort = fDirBareFormat = fDirLowerCase = fDirPause =
	fDirrecurse = fDirsort = fDirWide = fDirVerbose =
	fDirAttrSwitch = fDirAttr = fDirNonAttr = FALSE;

	// narrow screen? shrink some columns
	if( fIsFat32World )
	{
		dfDir = &dirField80Col;
	}
	else
	{
		dfDir = &dirField80Col622;
	}
	if(iScreenCols <= 45)
		dfDir = &dirField40Col;

	// use bare format for small screens
	if(iScreenCols < 40 || iScreenRows < 8)
		fDirBareFormat = TRUE;

	// convert maxcols to "units" wide
	iScreenCols /= (dfDir->iWideWidth+1);

	/* look for DIRCMD= in environment */
	stuff_env_var(szArgs,"DIRCMD");

	if(!get_extended_switch(szArgs, 'A', &fDirAttr, &Attrs, iaDirAttrList))
		return FALSE;

	if(!get_extended_switch(szArgs, 'O', &fDirsort, &SortOrder, iaDirSortList))
		return FALSE;

	// for /O with no sort order, assume name
	if( fDirsort && ( iaDirSortList[0] == NULL ) )
		iaDirSortList[0] = SortOrder.switches[0].value;

	// get the other switches
	GetSwitch(szArgs, 'B', &fDirBareFormat);
	GetSwitch(szArgs, 'L', &fDirLowerCase);
	GetSwitch(szArgs, 'P', &fDirPause);
	GetSwitch(szArgs, 'S', &fDirrecurse);
	if( fIsFat32World )
	{
		GetSwitch(szArgs, 'V', &fDirVerbose);
	}
	GetSwitch(szArgs, 'W', &fDirWide);
	GetSwitch(szArgs, '4', &fDir4DigitDate);

	// top bit in switch used for show attributes
	fDirAttrSwitch = (fDirAttr & 0x80);
	fDirNonAttr    = (fDirAttr >> 8) & 0x7F;
	fDirAttr       &= 0x77;		// clear "attr_switch" and "attr_show" bits

	// precedence is bare, wide, verbose, normal
	if(fDirBareFormat)
		fDirWide = fDirVerbose = FALSE;
	if(fDirWide)
		fDirVerbose = FALSE;

	return TRUE;
}

// return a pointer to the long version of this filename. If it is to
// be modified, the contents at the pointer must be copied.
char * PASCAL GetLongFileNameOnly(char szFileName[])
{
	if(fLFNPresent)
	{
		if( SmartExpandPath(FULLPATH_LONGNAME, szFileName,
				lfnFindData.szFileName) )
		{
			return NamePart(lfnFindData.szFileName);
		}
	}

	return szFileName;
}

/*
	round up to next whole cluster size
*/
ulong PASCAL RoundUpToCluster(ulong ulFileSize)
{
	ulong	ulBytesPerCluster = ulSectorsPerCluster * ulBytesPerSector;

	// globals must have been set prior!
	assert(ulBytesPerCluster);

	// round up to next whole cluster size
	if(ulFileSize % ulBytesPerCluster)
		ulFileSize += ulBytesPerCluster - (ulFileSize % ulBytesPerCluster);

	return ulFileSize;
}

/*
	GatherDirectory

	Gather all the directory entries into memory. Does no sorting at this
	point, but DOES count entries and sizes.

	Knows about the attributes to gather, so only those items that should
	be found are saved.

	On Entry
		lpDirBuff		a far buffer for storing entries
		lSizeDirBuff	size of the buffer
		pFoundBlk		results of the findfirst

	Returns the # of entries found that match search criteria (may be 0).
*/
long PASCAL GatherDirectory(void far *lpDirBuff, long lSizeDirBuff,
		PFIND pFoundBlk)
{
	FIND huge * lpBlks = lpDirBuff;
	LPFFBLK huge * lplpPtrArray;
	long	lMaxDirEntries;
	long	lCount;
	bool	bFound;

	// determine # of entries we can read. guarantee enough space for the
	// sorting pointers at end of ffblk array
	lMaxDirEntries = lSizeDirBuff / (sizeof(FIND) + sizeof(void far *));

	// start with no count/sizes
	ulThisDirFiles = ulThisDirDirectories = 0;
	ZeroQuad(&uqThisDirBytes);
	ZeroQuad(&uqThisDirBytesAllocated);

	lCount = 0;
	do
	{
		// don't save "CON:" device files
		if(pFoundBlk->attr & A_DEVICE)
			break;

		// attribute we don't want
		if(pFoundBlk->attr & fDirNonAttr)
			goto do_next;

		// but has to be an attribute we do want
		if( (pFoundBlk->attr & fDirAttr) != fDirAttr )
			goto do_next;

		// OK, save this entry for display

		// directory entry
		if(pFoundBlk->attr & A_SUBDIR)
			++ulThisDirDirectories;

		// file entry
		else
		{
			++ulThisDirFiles;
			AddQuadLong(&uqThisDirBytes, pFoundBlk->size);
			AddQuadLong(&uqThisDirBytesAllocated,
				RoundUpToCluster(pFoundBlk->size));
		}

		// copy the entry into our list
		farmemcpy((void far *)lpBlks, pFoundBlk, sizeof(FIND));

		// on to next entry
		--lMaxDirEntries;
		++lCount;
		++lpBlks;

do_next:
		bFound = SmartFindNext(pFoundBlk);
	} while(bFound && lMaxDirEntries);

	// now create an array of pointers to each ffblk at the end
	lplpPtrArray   = (LPFFBLK huge *)lpBlks;
	lpBlks         = lpDirBuff;
	lMaxDirEntries = lCount;
	for(lCount = 0; lCount < lMaxDirEntries; ++lCount)
		*lplpPtrArray++ = lpBlks++;

	// return the # of items found
	return (lCount);
}

enum sortit { SORT_DONT_SWAP=-1, SORT_DONT_CARE, SORT_SWAP };

/*
	Compare 2 findfirst blocks based on the fDirsort value.
	Return a
		-1 Don't Swap
		0 Equal
		1 Swap
*/
int PASCAL CompareSortType(FIND huge *lpBlk1,
					FIND huge *lpBlk2, int iSortType)
{
	char ename1[DOTNAMESIZE], ename2[DOTNAMESIZE];
	char szShortName1[DOTNAMESIZE], szShortName2[DOTNAMESIZE];
	unsigned uDate1, uDate2;
	int val;
	enum sortit result;
	int reverse;

	reverse = iSortType >> 8;
	if(reverse)
		iSortType = reverse;
	switch(iSortType)
	{
		/* Sort by Name (and Reverse) */
		case 1:

		/* Sort by Extension (and Reverse) */
		case 2:

			// copy in filenames to local space
			farmemcpy(szShortName1, lpBlk1->name, DOTNAMESIZE);
			farmemcpy(szShortName2, lpBlk2->name, DOTNAMESIZE);

			// expand them to the 11 char name with spaces
			name_expand(ename1, szShortName1);
			name_expand(ename2, szShortName2);

			if(iSortType == 1)
				val = strncmp(ename1, ename2, 8);
			else
				val = strncmp(ename1+8, ename2+8, 3);

			if(val == 0)
				return SORT_DONT_CARE;		/* don't care if they are the same */
			else if(val > 0)
				result = SORT_SWAP;
			else
				result = SORT_DONT_SWAP;
			break;

		/* Group DIRs first */
		case 0x04:

			/* 1st entry is a directory */
			if((lpBlk1->attr & FA_DIREC))
			{
				/* don't care if both are directories */
				if(lpBlk2->attr & FA_DIREC)
					result = SORT_DONT_CARE;
				else
					result = SORT_DONT_SWAP;
			}

			/* 1st entry is a file */
			else
			{
				if(lpBlk2->attr & FA_DIREC)
					result = SORT_SWAP;
				else
					result = SORT_DONT_CARE;
			}
			break;

		/* Sort by Size */
		case 0x08:
			if(lpBlk1->size == lpBlk2->size)
				return SORT_DONT_CARE;
			else if(lpBlk1->size > lpBlk2->size)
				result = SORT_SWAP;
			else
				result = SORT_DONT_SWAP;
			break;

		/* Sort by Date/Time (and Reverse) */
		case 0x10:
			if(lpBlk1->date >  lpBlk2->date)
				result = SORT_SWAP;
			else if(lpBlk1->date <  lpBlk2->date)
				result = SORT_DONT_SWAP;
			else if(lpBlk1->time >  lpBlk2->time)
				result = SORT_SWAP;
			else if(lpBlk1->time < lpBlk2->time)
				result = SORT_DONT_SWAP;
			else
				result = SORT_DONT_CARE;
			break;

		/* Sort by Last Access Date */
		case 0x20:

			// get the last access dates
			uDate1 = SmartGetLastAccessDate((FIND far *)lpBlk1);
			uDate2 = SmartGetLastAccessDate((FIND far *)lpBlk2);

			// sort them
			if(uDate1 > uDate2)
				result = SORT_SWAP;
			else if(uDate1 < uDate2)
				result = SORT_DONT_SWAP;
			else
				result = SORT_DONT_CARE;
			break;

		default:
			assert0();					/* Shouldn't oughta get here! */
			break;
	}

	/* reverse the sort */
	if(reverse)
	{
		if(result == SORT_SWAP)
			result = SORT_DONT_SWAP;
		else if(result == SORT_DONT_SWAP)
			result = SORT_SWAP;
	}

	/* result is one of 3 things: SWAP, DONT, DON'T CARE */
	return result;
}

/*
	Compare two blocks for sorting purposes. Checks all the
	different sorting possiblilities (name, ext, group, size, date).

	Returns whether to swap or not.
*/
int PASCAL Compare2Entries(FIND huge *lpBlk1, FIND huge *lpBlk2)
{
	int i;
	enum sortit result;

	/* up to 6 sorting posibilities */
	for(i=0; i<6 && iaDirSortList[i]; ++i)
	{
		result = CompareSortType(lpBlk1, lpBlk2, iaDirSortList[i]);
		if(result != SORT_DONT_CARE)
			return (result);
	}
	return SORT_DONT_SWAP;
}

/*
	Sort based on the options in the sort array.
*/
void PASCAL SortDirectory(void far *lpDirBuff, long lCount)
{
	FIND huge * lpBlks = lpDirBuff;
	LPFFBLK huge * lplpPtrArray;
	LPFFBLK lpSwapTemp;
	long lLimit, lIndex;
	bool inorder;

	// do a simple bubble sort
	for(lLimit = lCount-1; lLimit > 0; --lLimit)
	{
		inorder = TRUE;
		lplpPtrArray = (LPFFBLK huge *)(lpBlks + lCount);
		for(lIndex=0; lIndex<lLimit; ++lIndex)
		{
			// swap these entries?
			if(Compare2Entries(*lplpPtrArray,*(lplpPtrArray+1)) == SORT_SWAP)
			{
				// swap 'em
				lpSwapTemp        = *lplpPtrArray;
				*lplpPtrArray     = *(lplpPtrArray+1);
				*(lplpPtrArray+1) = lpSwapTemp;
				inorder = FALSE;
			}

			// on to next pair of pointers
			++lplpPtrArray;
		}

		// optimization: if we did no swaps, we're done
		if(inorder)
			break;
	}
}

/*
	Display attributes in char form.
*/
void PASCAL DisplayAttributes(uchar attr)
{
	static char szAttrChar[] = "RHSVDA";
	static char szFill[] = "    ";
	int	i;
	uchar	mask = 1;

	// display lower 6 bits in char form
	for(i=0; i<6; ++i)
	{
		printf("%c", (attr & mask) ? szAttrChar[i] : ' ');
		mask = (mask << 1);
	}
	printf( szFill );
}


/*
	This is the standard display directory function. It displays in all
	the many modes, including bare, wide, verbose, and normal.

	See DIR.TXT for examples of all the possible permutations (there are
	a LOT!)

	On Entry
		lpDirBuff		A far buffer holding and array of ffblks
		lCount			# of ffblks in the far buffer
		szPath			the path to this directory, in string form
*/
void PASCAL DisplayDirectory(void far *lpDirBuff, long lCount, char *szPath)
{
	// strings for displaying the directory
	extern char cmdstr_dirof[];
	extern char cmdstr_dirspace[];
	extern char cmdstr_verbose[];

	FIND huge *lpBlks = lpDirBuff;
	LPFFBLK huge * lplpPtrArray;
	FIND dp;
	char	ename[DOTNAMESIZE], ext[4];
	char	date[20],time[20];
	long	i;
	char	*szNamePart;

	// display the directory header
	if(!IsDirBare())
	{
		// get longname path if possible (but not on small screens)
		if(!IsDirShort() &&
				SmartExpandPath(FULLPATH_LONGNAME, szPath, lfnFindData.szFileName))
			strcpy(szPath, lfnFindData.szFileName);

		// make sure pause has a name to use
		strcpy(szPausePath, szPath);

		if(!IsDirRecurse())
		{
			printf(" ");
			// display the current directory if there are files in it
			//
			printf(cmdstr_dirof, szPath);
			DirDisplayLF();
		}

		// display verbose header
		if(lCount)
		{
			// adjust whitespace (/S needs newline, otherwise space)
			if(IsDirRecurse())
			{
				DirDisplayLF();

				// display the current directory if there are files in it
				//
				printf(cmdstr_dirof, szPath);
				DirDisplayLF();
			}

			if(IsDirVerbose())
			{
				printf(cmdstr_verbose);
				DirDisplayLF();
				DirDisplayLF();
			}

			// one more line
			DirDisplayLF();
		}
	}

	// lowercase the directory, too
	if(fDirLowerCase)
		strlwr(szPath);

	// display all directory entries. already sorted in lplpPtrArray
	lplpPtrArray = (LPFFBLK huge *)(lpBlks + lCount);
	for(i=0; i<lCount; ++i)
	{
		// get a local copy of directory entry
		farmemcpy(&dp, (void far *)*lplpPtrArray, sizeof(FIND));

		/* get name strings */
		name_expand(ename, dp.name);

		/* display in lower case */
		if(fDirLowerCase)
		{
			strlwr(dp.name);
			strlwr(ename);
		}

		/* expand the name into name ext form */
		memmove(ext, &ename[8], 3);
		ext[3]  = '\0';
		ename[8] = '\0';

		// expand path to the longname path if possible
		add_path(szPath, dp.name);
		if(!IsDotDir(dp.name) &&
				SmartExpandPath(FULLPATH_LONGNAME, szPath,
					lfnFindData.szFileName))
		{
			strcpy(szPath, lfnFindData.szFileName);
		}

		// DIR Bare takes precedence
		if(IsDirBare())
		{
			/* display in lower case */
			if(fDirLowerCase)
			{
				strlwr(szPath);
			}

			// show full path on /S, show name only if not
			// don't show subdir "." and ".."
			if(!IsDotDir(dp.name))
			{
				printf("%s", IsDirRecurse() ? szPath : LastPart(szPath));
				DirDisplayLF();
			}
		}

		// DIR Wide is next precedence
		else if(IsDirWide())
		{

//  BT 08-18-00
//
//  This is a fix for our widthwise formatting.
//
#if 1
			if(dp.attr & A_SUBDIR)
			{
				sprintf(ename, "[%s]", dp.name);
				printf(cmdstr_dirspace, (dfDir->iWideWidth)-1, ename);
			}
			else
			{
				sprintf(ename, dp.name);
				printf(cmdstr_dirspace, (dfDir->iWideWidth)-3, ename );
			}
#else
			if(dp.attr & A_SUBDIR)
				sprintf(ename, "[%s]", dp.name);
			else
				sprintf(ename, dp.name);
			printf(cmdstr_dirspace, (dfDir->iWideWidth)-1, ename);
#endif
		}

		// Normal or Verbose
		else
		{
			// convert date/times to ascii form
			timetoa(dp.time, time);
			datetoa(dp.date, date);

			// display name
			printf("%8s %3s ",ename,ext);

			// subdir, show only <DIR> symbol */
			if(dp.attr & A_SUBDIR)
			{
				printf(cmdstr_dirspace, dfDir->iFileSizeWidth,
						dfDir->szDirStr);
			}
			// file, show file size
			else
				printf(dfDir->szNumFmt32, dfDir->iFileSizeWidth, dp.size);

			// display bytes allocated to this file
			if(IsDirVerbose() && !IsDirShort())
			{
				printf(dfDir->szInBetweenDate);
				if(dp.attr & A_SUBDIR)
					printf(cmdstr_dirspace, 12, "");
				else
					printf(dfDir->szNumFmt32, 12, RoundUpToCluster(dp.size));
			}

			// display date and time. shrink it if need be */
			printf("%s%s %s",dfDir->szInBetweenDate,date,time);

			// verbose also displays last access date and attributes
			if(IsDirVerbose())
			{
				datetoa(SmartGetLastAccessDate(&dp), date);
				printf("%s%s%s",dfDir->szInBetweenDate,date,dfDir->szInBetweenDate);
				DisplayAttributes(dp.attr);
			}

			// display longname (but not if on small screen)
			if( fLFNPresent && ( IsDirVerbose() || (!IsDirShort()) ))
			{
				// display longname
				szNamePart = LastPart(szPath);
				printf(" %s",szNamePart);

				// the line is longer than 80 columns. Handle that with
				// pause, but let output wrap naturally
				if(IsDirVerbose() || 47+strlen(szNamePart) > 80)
					DirDisplayNoLF();
			}

			// finally done with the line
			DirDisplayLF();
		}

		// make sure to end the line in Wide mode
		if(IsDirWide())
		{
			if(((i+1) % iScreenCols) == 0 || (i+1) == lCount)
				DirDisplayLF();
			else
				printf("%c", 0x09 );
//				printf(" ");
		}

		// remove filename from path
		del_path(szPath);

		// on to next entry
		++lplpPtrArray;
	}

	// no finishing stats for bare directories
	if(!lCount || IsDirBare())
		return;

	// display file totals
	if( fIsFat32World )
	{
		DirDisplayFileBytes(ulThisDirFiles, uqThisDirBytes);
		if(IsDirVerbose())
			DirDisplayAllocated(ulThisDirDirectories, uqThisDirBytesAllocated);
	}
	else
	{
		DirDisplayFileBytes(ulThisDirFiles+ulThisDirDirectories,
				uqThisDirBytes);
		if(IsDirVerbose())
			DirDisplayAllocated(ulThisDirDirectories, uqThisDirBytesAllocated);
	}
}

/*
	This is passed a string which presumably contains a drive letter
	(such as "A:"). Prints the volume label message for this drive.
	Used by both DIR and VOL.

	On Entry		fNoPause		ensures the DirDisplayLF() won't pause.

	Returns the drive # or ERR on failure.
*/
int PASCAL DisplayVolume(char * szArgs, bool fNoPause)
{
	extern char cmdstr_dirvol[];
	extern char cmdstr_dirvolyes[];
	extern char cmdstr_dirvolno[];
	extern char cmdstr_dirserial[];

	int  drive;
	char volume[16];
	char part2[20];
	char serial[10];
	struct drive_serial Drive_Serial;

	/* make sure drive is valid */
	if((drive = cmdgetdrive(szArgs)) < 0)
		return ERR;

	// make sure screen size is correct to handle possible pausing
	if(fNoPause)
		DetermineScreenSize();

	/* DOS1 findfirst of volume label */
	if(get_vol(drive, volume))
		sprintf(part2,cmdstr_dirvolyes,volume);
	else
		sprintf(part2,cmdstr_dirvolno);
	DirDisplayLF();
	printf(cmdstr_dirvol,drive+'A',part2);
	DirDisplayLF();

	if(get_drive_serial(drive, &Drive_Serial))
	{
		sprintf(serial, "%04X-%04X", Drive_Serial.serial2, Drive_Serial.serial1);
		intl_strupr(serial);
		printf(cmdstr_dirserial, serial);
		DirDisplayLF();
	}
	return drive;
}

/*
	Helper function to DirDisplayLF. Shows a continue line after it
	pauses.
*/
void PASCAL DirDisplayContinue(void)
{
	extern char cmdstr_dirpause[];

	/* for screens with more than 8 lines, show a continue line */
	if(iScreenRows >= 8)
	{
		DirDisplayLF();
		printf(cmdstr_dirpause, szPausePath);
		DirDisplayLF();
	}
}

/*
	Possibly display with pause
*/
void PASCAL DirDisplayLF(void)
{
	if(fDirPause)
	{
		if(DisplayScreenLF())
			DirDisplayContinue();
	}
	else
		printf("\n");
}

/*
	Possibly display with pause
*/
void PASCAL DirDisplayNoLF(void)
{
	if(fDirPause)
	{
		if(DisplayScreenNoLF())
			DirDisplayContinue();
	}
}

// return TRUE if bare format
bool PASCAL IsDirBare(void)
{
	return fDirBareFormat;
}

// return if wide format
bool PASCAL IsDirWide(void)
{
	return fDirWide;
}

// return TRUE if verbose mode
bool PASCAL IsDirVerbose(void)
{
	return fDirVerbose;
}

// return TRUE if recurse into subdirs
bool PASCAL IsDirRecurse(void)
{
	return fDirrecurse;
}

// if columns are less than 40, use short form
bool PASCAL IsDirShort(void)
{
	return (dfDir->iCols < 40);
}
