//
//   $Id: COMMAND.H 1.9 2002/10/15 20:08:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:COMMAND.H

	Copyright (c) 1989-2002 Datalight, Inc.
	Datalight Confidential and Proprietary.
   All Rights Reserved.

	DEFINES for command.com processor source files.

   $DOCHISTORY
	02/20/1999 dg  Added Long Filename support
	07/16/1998 dg  Updated to Solid Code (datalght.h)
	08/20/1997 de  SPR 1361: revised to fix batch file processing bug
	11/10/1988 dg  original
*********************************************************************/

#ifndef __COMMAND
#define __COMMAND

/* enable or disable LFN support */
#define LFN 1

//  Show two digit years verses four digit years
//
#define TWO_DIGIT_YEAR	FALSE

#define ERR  -1
#define OK    0
#define INVALID ERR

#define RD		0
#define WR		1
#define RDWR	2

// change HELPVER any time the layout of the COMMAND.HLP file changes.
// the hi word is the macro version (ie DOS 7.10), lo word is helpfile ver
#define HELPVER	0x07100001L

// for determining program type
#define COM  0
#define EXE  1
#define BAT  2

// 128 for 6.22 compatibility, 1024 for 7.1 compatibility
#define BATCH_LINELEN	1024
#define LINELEN			 128
#define NAMESIZE 			 260
#define OLD_NAMESIZE 	  66
#define CMDSIZE			  10

#define S_NOAUTOEXEC 2
#define S_NOEXIT		1

#define GET_NORMAL	0
#define GET_NOBUF		1
#define GET_DOSKEY	2

#define INVALID_DIR		-1L

/* MS-DOS Functions */
#define SET_CUR_DR		0x0E
#define GET_CUR_DR		0x19
#define SET_DTA			0x1A
#define DOS_GET_COUNTRY	0x38
#define MK_DIR				0x39
#define RM_DIR				0x3A
#define CH_DIR				0x3B
#define DOS_CREAT			0x3C
#define DOS_OPEN			0x3D
#define DOS_CLOSE			0x3E

/* FINDFIRST attributes */
#define A_READONLY	0x01
#define A_HIDDEN		0x02
#define A_SYSTEM		0x04
#define A_VOLUME		0x08
#define A_SUBDIR		0x10
#define A_ARCHIVE		0x20
#define A_DEVICE		0x40
#define A_ALLFILES	0x06	/* hidden and system */
#define ALL_ATTR		0x16

/* STANDARD HANDLES */
#define STDIN  0
#define STDOUT 1
#define STDERR 2
#define STDAUX 3
#define STDPRN 4

/* ERROR MESSAGES */
extern char * messages[];
#define ENUMPARM  0
#define EMKDIR    1
#define ERMDIR    2
#define EDRIVE    3
#define ECHDIR    4
#define ENOFILE   5
#define EDATE     6
#define ETIME     7
#define ESYNTAX   8
#define EBADCMD   9
#define ENOMEM   10
#define EBADEXE  11
#define ECREAT   12
#define EINVDR   13
#define ERENAME  14
#define EBADDEV  15
#define ENOSPACE 16
#define ENOBATCH 17
#define ENOENV   18
#define ENOLABEL 19
#define ETERMBAT 20
#define EINVDRV  21
#define ENOWILD  22
#define ENOCOPY  23
#define EACCESS  24
#define EPARMS   25
#define EBADVER  26
#define EONOFF   27
#define ENOPATH  28
#define MSG_DEL_ALL_FILES	  29
#define ECURDIR  30
#define ENOPIPE  31 // couldn't open/create an intermediate pipe file
#define EUSAGE   32
#define ENOHELP  33
#define ENOTROMDOS 34
#define EDIREXISTS 35
#define EINVDIR  36
#define EBADSWITCH	37
#define ETOOMANY		38

#define PSP_ENVSEG	0x2C
#define PSP_VER		0x40

typedef void PASCAL (*PAS_FUNCPTR)(char *);

/* STRUCTURES */
struct cmds
{
	char *cmd;
	PAS_FUNCPTR func;
};

// these macros are intended to simplify the change to the new .BAT
// file state record
//
extern unsigned char c_no_batch_echo;
#define c_echo     ((c_in_batch) ? pCurBCB->batEcho : c_no_batch_echo)
#define c_bat_pos  (pCurBCB->batPos)

/*
	Our new batch file state record, the BCB. Aside from possible problems
	with the four unknown? fields, we are different from MS in that:

	1) we don't use the batArgOffsets.
	2) our batName is a fixed length ASCIIZ string while MS uses a variable
   	length ASCIIZ string
	3) MS batArgs (eg {%0, %1,...%9}) form a contiguos variable length
		ASCIIZ string which immediately follows their batName. The args
		are delimited by CR (0Dh) chars. The offset of the each arg,
		relative	to the start of the BPB, is kept in batArgOffsets.
		A SHIFT op just does (batArgsOffset[i] = batArgsOffset[i+1]) with
		the last actual argument offset replaced by 0FFFFh. MS allows
		more than 10 arguments to a .BAT file and SHIFT must eventually
		expose them all.
*/

#define 	MAX_BATCH_ARGS	10
struct BatchControlBlock
{
	unsigned char unknown1;							// 00: ?
	unsigned char batEcho;							// 01: 1 = ON
	unsigned char unknown2;    		         // 02: ?
	unsigned prevBCBSeg;  		              	// 03: 0 = none, this is the first
	unsigned unknown3;                  		// 05: ?
	unsigned char unknown4;             		// 07: ?
	long batPos;										// 08: the .BAT file lseek offset
	unsigned batArgOffsets[MAX_BATCH_ARGS];	// 0C: 0FFFFh means none
	char batName[NAMESIZE];							// 20: fully canonical ASCIIZ path name
	char batArgs[BATCH_LINELEN];					// 62: the text of %1, %2...
};
extern struct BatchControlBlock far * pCurBCB;

struct rd_country
{
	int	datefmt;
	char	curr[5];
	char	thsep[2];
	char	desep[2];
	char	dtsep[2];
	char	tmsep[2];
	char	currstyle;
	char	digits;
	char	timefmt;
	long	casefmt;
	char	dasep[2];
	char	fill[10];
};

struct exe_parms
{
	unsigned envseg;
	void far * cmdline;
	void far * fcb1;
	void far * fcb2;
};

struct ext_switch
{
	int count;
	struct switches
	{
		char option;
		unsigned value;
	} switches[8];
};


#define DOS5_FAT_ID ( ((long)'1'<<24) | ((long)'T'<<16L) | ('A'<< 8) | 'F')

struct drive_serial
{
	unsigned level;
	unsigned serial1, serial2;		/* Could be considered a long */
	char volume_label[11];
	char Fat_Type[8];
};

/* assert */
#define assert(e)	if(!(e)) _assert(__FILENUM__ * 1000 + __LINE__)
#define assert0() _assert(__FILENUM__ * 1000 + __LINE__)

/* in-line functions */
#define curdisk() (bdos(GET_CUR_DR, 0, 0)&0xFF)
#define isspace(c) (((c)>=9 && (c)<=13) || (c)==' ')
#define isequalspace(c) (isspace(c) || (c)=='=')
#define iscopyspace(c)  (isspace(c) || (c)=='+')

// external variables
extern char szSplitOldCwd[];
extern char szSplitDrive[];
extern char szSplitDirectory[];
extern char szSplitFileName[];

// variables specific to certain commands
extern char szDirPath[];
extern bool fCopyOverwritePrompt;

// items passed between RESIDENT and TRANSIENT command
extern char     c_cmdline[LINELEN+2];	// the whole command
extern char     c_lastline[LINELEN+2]; // previous line as type in by user
extern char     c_exe_name[NAMESIZE];
extern char     c_cmdname[OLD_NAMESIZE];
extern char     c_exe_cmdline[LINELEN+2];
extern bool     c_in_batch;				// if TRUE then we processing a .BAT file
extern bool     c_in_batch_for;
extern int      c_no_exit;
extern int      c_one_only;
extern int      c_ask_autoexec;
extern char *   c_pipe_ptr;
extern int      c_redirect[2];
extern char     c_bat_redirect[2];
extern char     c_ctty[];
extern unsigned c_curBCBSeg;
extern unsigned c_resseg;
extern char     c_new_COMSPEC;
extern int      c_ask_autoexec;
extern char *   c_pipe_ptr;
extern int      c_pipe_num;
extern int      c_pipe_del;
extern char     c_exe_fcb1[16];
extern char     c_exe_fcb2[16];
extern int      c_errorlevel;
extern int      c_iExecError;
extern int      c_ignore;
extern int      c_crit_error;
extern uchar    c_crit_drive;
extern int      c_lfnfor;

// other important items
extern bool     fDontShowLine;
extern bool     fIsCall;

// globals for parsing the command
extern char * pszCmdName;
extern char * pszCmdArgs;
extern char szCmdLine[BATCH_LINELEN];
extern char szTemp[BATCH_LINELEN];
extern char szTemp2[BATCH_LINELEN];
extern char szTemp3[BATCH_LINELEN];

// error messages and commands
extern int maxerr;
extern char * messages[];
extern int MAX_CMD;
extern struct cmds cmds[];

// standard
extern unsigned _psp;
extern int      errno;

// For displaying LFN information
//
extern bool fLFNPresent;

// For displaying 7.1 formats
//
extern bool fIsFat32World;

// screen info
extern int iScreenRows;
extern int iScreenCols;
extern int iScreenCurrentRow;

// RESIDENT functions callable from TRANSIENT command
extern void res_openhole();
extern void res_closehole();


//  If you include the borland CTYPE.H file, other warnings will be
//  created for the macros in this file.  Added here temporarily the
//  prototype for toupper()
//
int toupper(int __ch);
int isalpha(int __c);

int BiosGetCols( void );

#endif	/* __COMMAND */
