//
//   $Id: COMUTIL.C 1.19 2003/01/10 18:23:42Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:COMUTIL.C

	Copyright (c) 1989-2002 Datalight, Inc.
	Datalight Confidential and Proprietary
   All Rights Reserved.

   $DOCHISTORY
	02/21/1999 dg	Added full LFN support
   06/19/1998 dg  Added LFN DIR support
	08/20/1997 de  SPR 1361: revised to fix batch file processing bug
	05/01/1889 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__ 6

// ansi includes
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// datalight includes
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"
#include "dlcntry.h"
#include "dldosio.h"
#include "dlerrcod.h"
#include "command.h"
#include "comprot.h"

#define MAXROWS	128
#define MINROWS	  4
#define MAXCOLS	256
#define MINCOLS	 16

#define valid_on_off_terminator(c) (c == ' ' || c == '\t' || c == '\n' || c == '\r' || c == 0)

// screen size (may be adjusted by the DIRSIZE command)
int iScreenCols;			/* 80 by default */
int iScreenRows;			/* 25 by default */
int iScreenCurrentRow;

/*
	$DOCPROC:DetermineScreenSize

	Determines the screen size, and sets the globals "iScreenRows" and
	"iScreenCols".
*/
void PASCAL DetermineScreenSize(void)
{
	char *szDirSize;

	// default to 80x25
	iScreenCols = 80;
	iScreenRows = 25;

	/* user may set # of lines and columns */
	szDirSize = GetEnvironmentVariable("DIRSIZE");
	if(szDirSize)
	{
		szDirSize = GetNextInteger(szDirSize, &iScreenRows, MINROWS, MAXROWS);
		GetNextInteger(szDirSize, &iScreenCols, MINCOLS, MAXCOLS);
	}

	// no DIRSIZE environment setting, determine # of cols via BIOS
	else
	{
		// call BIOS to get # of screen columns
		iScreenCols = BiosGetCols();

		// if reasonable, set # of columns
		MinMax(&iScreenCols, MINCOLS, MAXCOLS);
	}

	// getting screen size. Assume starting at top of screen.
	iScreenCurrentRow = 0;
}

/*
	$DOCPROC:MinMax

	Make sure the interger is in bounds.
*/
void PASCAL MinMax(int *piValue, int iMinValue, int iMaxValue)
{
	// make sure value is in bounds
	if(*piValue < iMinValue)
		*piValue = iMinValue;
	if(*piValue > iMaxValue)
		*piValue = iMaxValue;
}

/*
	$DOCPROC:GetNextInteger

	Helper function to DetermineScreenSize. Leaves "*piValue" unchanged if
	no more digits in the string.
*/
char * PASCAL GetNextInteger(char *szStr, int * piValue, int iMinValue, int iMaxValue)
{
	// skip to the digits
	while(*szStr && !isdigit(*szStr))
		++szStr;

	// we're on an integer
	if(isdigit(*szStr))
	{
		// get the integer
		*piValue = atoi(szStr);

		// advance string past integer
		while(isdigit(*szStr))
			++szStr;

		// make sure value is in bounds
		MinMax(piValue, iMinValue, iMaxValue);
	}

	// return pointer to string past the integer value
	return szStr;
}

/*
	$DOCPROC:DisplayScreenLF

	When using /P, causes a pause during the DIR or TYPE command.

	Returns TRUE if it paused.
*/
bool PASCAL DisplayScreenLF(void)
{
	printf("\n");
	return DisplayScreenNoLF();
}

/*
	$DOCPROC:DisplayScreenNoLF

	Counts one line for the purposes of /P with DIR or TYPE. Doesn't
	actually display a LF because the last char on the screen was already
	written.

	Returns TRUE if it paused.
*/
bool PASCAL DisplayScreenNoLF(void)
{
	// on to next row
	++iScreenCurrentRow;

	// need to pause?
	if(((iScreenCurrentRow+1) % iScreenRows) == 0)
	{
		iScreenCurrentRow = 0;
		do_pause("");
		return TRUE;
	}

	return FALSE;
}

/*
	$DOCPROC:DisplayWithPause

	Display lines with pause. Understands both screen height and width,
	and will pause for a key ever screenful. Understands tabs as 8 columns.
*/
void PASCAL DisplayWithPause(char huge * lpBuff, ulong ulCount)
{
	char huge * lpLine;
	int	iLineLen;
	int	iTempLen;
	int	iLineCols;

	// determine screen rows and cols
	DetermineScreenSize();

	// now, write 1 line at a time
	lpLine = lpBuff;
	while(ulCount)
	{
		// determine length of THIS line, based on screen columns
		if(ulCount > BATCH_LINELEN)
			iLineLen = BATCH_LINELEN;
		else
			iLineLen = (int)ulCount;
		iTempLen = iLineLen = GetLineLength(lpLine, 0, iLineLen, &iLineCols);

		// don't write the CR/LF pair. We'll do that with DisplayScreenLF
		if(iTempLen > 1 && lpLine[iTempLen-1] == '\n')
			--iTempLen;
		if(iTempLen > 1 && lpLine[iTempLen-1] == '\r')
			--iTempLen;

		// display the line
		HugeWrite(STDOUT, lpLine, iTempLen);

		// display the LF at the end
		if(iLineCols != iScreenCols)
			DisplayScreenLF();
		else
			DisplayScreenNoLF();		// exactly screen columns, don't show LF

		// one to next line
		lpLine  += iLineLen;
		ulCount -= iLineLen;
	}
}

/*
	$DOCPROC:GetLineLength

	Helper to DisplayWithPause. Gets the maximum characters that will
	display on this physical line.

	Return both # of characters and columns.
*/
int PASCAL GetLineLength(char huge *lpLine, int iStartCol, int iMaxLen, int *piLineCols)
{
	int	iLineLen = 0;
	int	iCols = iStartCol;
	#define TABSIZE 8

	// determine the columns up to screen columns
	while(iMaxLen > 0 && *lpLine != '\n' && *lpLine != '\r' && iCols < iScreenCols)
	{
		switch(*lpLine)
		{
			case '\t':	iCols += TABSIZE - (iCols % TABSIZE); break;
			default:		++iCols; break;
		}
		++iLineLen;
		--iMaxLen;
		++lpLine;
	}

	// handle CR/LF (include them in the line, but don't count towards columns)
	if(iMaxLen && *lpLine=='\r')
	{
		++lpLine;
		++iLineLen;
	}
	if(iMaxLen && *lpLine=='\n')
		++iLineLen;

	// return columns and length of line
	*piLineCols = iCols;
	return iLineLen;
}

/*
	$DOCPROC:LFNEndSpecialArg

	Find the end of the current "special" arg. A special arg ends on a
	space, or one of the "special" delimiter characters. Is quote and
	international aware.

	Returns the end of this argument (may be on the trailing null).
*/
char * PASCAL LFNEndSpecialArg(char *szArg, char *szDelimiters)
{
	bool	bInQuote = FALSE;

	// skip past all the text for this arg. understands quotes and
	// the special delimiters
	while(*szArg && (bInQuote ||
		!(DL_IS_SPACE(*szArg) || strchr(szDelimiters, *szArg))))
	{
		// handle double byte characters
		if(IsDoubleByteChar(*szArg))
		{
			++szArg;
		}
		else
		{
			//  handle quotes if not in a 6.22 environment
			//
			if( ( fLFNPresent ) || ( fIsFat32World ) )
			{
				if(*szArg == '"')
				{
					bInQuote = !bInQuote;
				}
			}
		}

		++szArg;
	}

	// return end of the argument
	return szArg;
}

/*
	$DOCPROC:LFNNextSpecialArg

	Find the next argument with special delimiter characters. Is quote
	and international aware.

	Returns the start of the next argument, or on the trailing null if no
	more arguments.
*/
char * PASCAL LFNNextSpecialArg(char *szArg, char *szDelimiters)
{
	szArg = LFNEndSpecialArg(szArg, szDelimiters);
	return LFNSkipSpecialSpaces(szArg, szDelimiters);
}

/*
	$DOCPROC:LFNSkipSpecialSpace

	Skip spaces, treating the special delimiters as spaces too.
*/
char * PASCAL LFNSkipSpecialSpaces(char *szArg, char *szDelimiters)
{
	while(*szArg && (DL_IS_SPACE(*szArg) || strchr(szDelimiters, *szArg)))
		++szArg;
	return szArg;
}

/*
	$DOCPROC:LFNEndEqualArg
*/
char * PASCAL LFNEndEqualArg(char *szArg)
{
	return LFNEndSpecialArg(szArg, "=");
}

/*
	$DOCPROC:LFNNextEqualArg
*/
char * PASCAL LFNNextEqualArg(char *szArg)
{
	return LFNNextSpecialArg(szArg, "=");
}

/*
	$DOCPROC:LFNEndForArg
*/
char * PASCAL LFNEndForArg(char *szArg)
{
	return LFNEndSpecialArg(szArg, ",)");
}

/*
	$DOCPROC:LFNNextForArg
*/
char * PASCAL LFNNextForArg(char *szArg)
{
	return LFNNextSpecialArg(szArg, ",)");
}

/*
	$DOCPROC:StripEndingSpaces

	Removes spaces from the end of the string, if any.
*/
void PASCAL StripEndingSpaces(char *psz)
{
	char * pszEndArg;

	// skip all arguments until the last one
	do
	{
		pszEndArg = LFNEndArg(psz);
		if(*pszEndArg != 0)
		{
			// any trailing spaces? remove them.
			psz = LFNSkipWhite(pszEndArg);
			if(*psz == 0)
				*pszEndArg = 0;
		}
	} while(*pszEndArg);
}

/*
	$DOCPROC:FarLFNSkipWhite

	Skip white space on a far string. Will stop on the NULL.
*/
char far * PASCAL FarLFNSkipWhite(char far *lpsz)
{
	while(*lpsz && DL_IS_SPACE(*lpsz))
		++lpsz;
	return lpsz;
}

/*
	$DOCPROC:FarLFNNextArg

	Find the start of the next argument. Is "quote" aware.

	Returns the start of the next argument, or on the trailing null if no
	more arguments.
*/
char far * PASCAL FarLFNNextArg(char far *lpszArg)
{
	bool bInQuote = FALSE;


	while(*lpszArg && (bInQuote || !DL_IS_SPACE(*lpszArg)))
	{
		//  Don't acknowledge quotes in a 6.22 environment
		//
		if( ( fLFNPresent ) || ( fIsFat32World ) )
		{
			if(*lpszArg == '"')
			{
				bInQuote = !bInQuote;
			}
		}


		++lpszArg;
	}


	return FarLFNSkipWhite(lpszArg);
}

/*
	ROM-DOS specific function
*/
void PASCAL set_dos_ver(unsigned ver)
{
	extern unsigned _psp;

	union REGS r;
	r.x.ax = 0x30DB;			// ROM-DOS internal call
	r.x.si = 0xB2D2;			// sanity check
	r.x.cx = ver;				// version that DOS will report
	r.x.di = 2;					// function code for set version
	intdos(&r,&r);
}

void PASCAL ctrlc_ok_now(void)
{
	union REGS r;
	r.x.ax = 0x30DB;			// ROM-DOS internal call
	r.x.si = 0xB2D2;			// sanity check (magic cookie)
	r.x.di = 1;					// function to say ctrl-C is OK
	intdos(&r,&r);
}

/*
	$DOCPROC:GetDLBIOSVer

	Get the Datalight BIOS internal revision # in string form.
	Return an "" empty string if not the Datalight BIOS.
*/
bool PASCAL GetDLBIOSVer(char *szVer)
{
	union REGS r;

	r.x.ax = 0x2000;			/* Int 15h, fn 2000h, DL BIOS get version */
	int86(0x15,&r,&r);
	if(r.x.cflag && r.h.ah == 0x86)
		return FALSE;
	sprintf(szVer, "%d.%02d.%d",r.h.ah,r.h.al,r.x.cx);
	return TRUE;
}

/*
	Get the Super-Boot information (ROM-DOS kernel only)
*/
bool PASCAL GetSuperBoot(uchar *pSuperID, uchar *pSuperOrder,
																uchar *pSuperBoot)
{
	union REGS r;
	r.x.ax = 0x30DB;
	r.x.cx = r.x.dx = 0;
	r.x.si = 0xB2D2;			/* magic cookie tells RD to process function */
	r.x.di = 6;					/* super-boot function */
	intdos(&r,&r);
	*pSuperID = r.h.cl;
	*pSuperOrder = r.h.ch-1;
	*pSuperBoot = r.h.dl;

	/* returns 0 (false) if superboot not installed */
	return r.h.cl;
}

/*
	"rev" must be big enough to hold a 32 byte string.
*/
unsigned PASCAL get_internal_ver(char *rev)
{
	extern char notrd[];
	extern char notrep[];

	union REGS r;
	struct SREGS sr;

	r.x.ax = 0x30DB;			// ROM-DOS internal call
	r.x.cx = 0;
	r.x.di = 3;					// function code 3, get version string
	r.x.si = 0xB2D2;			// sanity check
	intdosx(&r,&r,&sr);

	/* failed - not ROM-DOS */
	if(r.x.cx != 0xB2D2)
		strcpy(rev, notrd);

	/* report ROM-DOS revision */
	else
	{
		if(r.x.bx != 0xFFFF)
			farmemcpy((void far *)rev, MK_FP(sr.es,r.x.bx), 32);
		else
			strcpy(rev, notrep);
	}

	return r.x.ax;
}

/*
	Print the prompt string.
*/
void PASCAL prompt_string(void)
{
	// temporary places capable of holding longnames (for $p)
	extern char szDirPath[];
	extern char szRecursePath[];

	unsigned char cur_drive;
	char *p;

	/* if in .BAT file and echo is off, no prompt */
	if(!c_echo)
		return;

	/* get current drive */
	cur_drive = INT_get_cur_dr();

	p=GetEnvironmentVariable("PROMPT");
	if(p==NULL)
		p="$n$g";

	/* print the prompt */
	while(*p)
	{
		if(*p=='$')
		{
			++p;
			switch(toupper(*p))
			{
			case 'P':
				// get the current directory, in long form if possible
				GetCurrentShortPath(cur_drive, szDirPath);
				if( fLFNPresent )
				{
					if(SmartExpandPath(FULLPATH_LONGNAME, szDirPath, szRecursePath))
						strcpy(szDirPath, szRecursePath);
				}
				printf("%s",szDirPath);
				break;
			case 'N':	printf("%c",cur_drive+'A'); break;
			case 'D':	showdate(); break;
			case 'T':	showtime(); break;
			case 'V':	do_ver("$"); break;
			case '$':	printf("$"); break;
			case 'G':	printf(">"); break;
			case 'L':	printf("<"); break;
			case 'B':	printf("|"); break;
			case 'Q':	printf("="); break;
			case 'H':	printf("\010 \010"); break;
			case 'E':	printf("\033"); break;
			case '_':	printf("\n"); break;
			default:		break;
			}
		}
		else
			printf("%c",*p);
		++p;
	}
}

/*
	Display a single environment variable. Returns pointer to end of
	environment variable (the null).
*/
char far * PASCAL ShowOneEnvironmentVariable(char far * lpszEnvVar)
{
	while(*lpszEnvVar)
	{
		printf("%c", *lpszEnvVar);
		++lpszEnvVar;
	}
	printf("\n");
	return lpszEnvVar;
}

/*
	Display the whole environment. (Environment is terminated by a
	double null).
*/
void PASCAL env_show(char far *lpszEnvVar)
{
	while(*lpszEnvVar)
	{
		lpszEnvVar = ShowOneEnvironmentVariable(lpszEnvVar);
		++lpszEnvVar;
	}
}

/*
	Remove a variable from the environment.  "ep" points
	to the variable to delete.
*/
void PASCAL env_remove(char far *ep)
{
	char far *next;

	/* find the next one */
	next=ep;
	while(*next)
		++next;
	++next;

	/* copy until done */
	while(*next)
	{
		do
		{
			*ep++ = *next;
		} while(*next++);
	}
	*ep= NULL;
}

/*
	Take the old environment and make a new copy from it
*/
void PASCAL env_make_new_copy(unsigned newseg, unsigned paras)
{
	extern unsigned _psp;

	unsigned far * oldptr;
	unsigned far * newptr;
	unsigned oldseg;
	unsigned i;

	/* fill up new PSP with contents of old one */
	oldseg = *(unsigned far *)MK_FP(_psp,0x2C);
	oldptr = MK_FP(oldseg,0);
	newptr = MK_FP(newseg,0);
	for(i=0; i<paras*8; ++i)
		*newptr++ = *oldptr++;

	/* tell the PSP where the new env is located */
	*(unsigned far *)MK_FP(_psp, 0x2C) = newseg;
}

/*
	Add a variable to the environment.
*/
void PASCAL env_add(char far *env,char *s,unsigned envsize)
{
	extern int c_no_exit;
	extern void res_openhole();
	extern void res_closehole();

	unsigned left;
	unsigned seg, paras;

	/* find the end, and determine how much is left */
again:
	left = envsize;
	while(*env)
	{
		while(*env++)
			--left;
		--left;
	}
	--left;

	/* not enough space! */
	if(strlen(s) >= left)
	{
		/* Non-permanent COMMAND can allocate a BIGGER environment */
		seg = 0;
		if(!c_no_exit)
		{
			farcall(res_openhole);
			paras = ((envsize + strlen(s)) >> 4) + 2;
			seg = dos_realloc(FP_SEG(env), paras);
			if(!seg)
				seg = dos_malloc(paras);
			farcall(res_closehole);
		}

		/* we were able to allocate the memory, a new environment! */
		if(seg)
		{
			env_make_new_copy(seg, paras);
			env = MK_FP(seg, 0);
			envsize = paras * 0x10;
			goto again;
		}

		/* no room in environment! */
		else
		{
			PrintError(ENOENV);

			/* fill up as much of the environment as we can */
			if(!left)
				return;
			s[left-1] = 0;
		}
	}

	/* tack on the string */
	do
	{
		*env++ = *s;
	} while(*s++);
	*env = 0;
}


/*
	Given an environment pointer and a name, find the
	matching environment entry.  Both are assumed to be
	UPPERCASE.
*/
char far * PASCAL env_match(char far *env,char *name)
{
	char *p= name;
	char far *ep;

	while(*env)
	{
		ep=env;
		p= name;
		while(*p && *ep!='=')
		{
			if(*p!=*ep)
				break;
			++p, ++ep;
		}
		if(*p==NULL && *ep=='=')
			return env;
		while(*ep)
			++ep;
		env=ep+1;
	}
	return NULL;
}

/*
	Given an environment variable name, insert the contents of the
	env variable into the dst string.
*/
void PASCAL stuff_env_var(char *dst, char *envname)
{
	char * envptr;
	int n;

	envptr = GetEnvironmentVariable(envname);
	if(envptr != NULL)
	{
		/* stuff envptr before user args, so user can override them */
		n = strlen(envptr);
		memmove(dst+n+1,dst,strlen(dst)+1);
		memmove(dst,envptr,n);
		dst[n]= ' ';
	}
}

/*
	Set ON/OFF.  Used by BREAK, LFNFOR, VERIFY and ECHO.

	Returns FALSE if user specified something other than ON, OFF or "".
*/
int  PASCAL set_on_off(char *szArgs, char * szCmdName, int * piSetting)
{
	static char * szOn  = "on";
	static char * szOff = "off";
	char * szNext;

	strupr(szCmdName);

	if(*szArgs == '\0')
		printf("%s is %s\n",szCmdName,*piSetting ? szOn : szOff);
	else
	{
		//  This is a way to determine if szArgs equal "ON" or "OFF"
		//  which can be followed by white space.  If we are not
		//  setting the state of the echo then return FALSE.
		//
		szArgs = skipequal(szArgs);
		szNext = LFNNextArg(szArgs);

		if(   (*szNext == 0)
		   && (strncmpi(szArgs,szOn,2) == 0)
         && (valid_on_off_terminator(szArgs[2]))
		  )
		{
		}
		else
		{
			if(   (*szNext == 0)
			   && (strncmpi(szArgs,szOff,3) == 0)
            && (valid_on_off_terminator(szArgs[3]))
			  )
			{
			}
			else
			{
				return FALSE;
			}
		}
      // o[n] or o[f]f are the only valid cases to reach here.
      //
		*piSetting = (toupper(szArgs[1]) == 'N');
	}
	return TRUE;
}

void PASCAL showtime(void)
{
	extern struct rd_country c_country;

	union REGS regs;
	unsigned hour,minutes,seconds,hund;
	char * sep = c_country.tmsep;
	char am_pm;

	/* get the time */
	regs.x.ax = 0x2C00;
	int86(0x21, (void *) &regs, (void *)&regs);
	hour = regs.x.cx >> 8;
	minutes  = regs.x.cx & 0xFF;
	seconds  = regs.x.dx >> 8;
	hund = regs.x.dx & 0xFF;

	/* convert for 12/24hr format */
	am_pm = convert_hour(&hour);

	printf("%02d%s%02d%s%02d%s%02d%c",
		hour, sep, minutes, sep, seconds, c_country.desep, hund, am_pm);
}

void PASCAL showdate(void)
{
	extern struct rd_country c_country;

	extern char *day_of_week[];

	union REGS regs;
	unsigned year, month, day;
	char *date_sep = c_country.dtsep;

	/* get date */
	regs.x.ax = 0x2A00;
	int86(0x21, (void *) &regs, (void *)&regs);
	year = regs.x.cx;
	month = regs.x.dx >> 8;
	day = regs.x.dx & 0xFF;

	printf("%s ",day_of_week[regs.x.ax & 0xF]);
	switch(c_country.datefmt)
	{
		case 0: printf("%2d%s%02d%s%04d",
				month, date_sep, day, date_sep, year); break;
		case 1: printf("%2d%s%02d%s%04d",
				day, date_sep, month, date_sep, year); break;
		case 2: printf("%4d%s%02d%s%02d",
				year, date_sep, month, date_sep, day); break;
	}
}

/*
	Set the date and time of this file
*/
void PASCAL set_date_time(int fd, unsigned date, unsigned time)
{
	union REGS r;
	r.x.ax = 0x5701;
	r.x.bx = fd;
	r.x.cx = time;
	r.x.dx = date;
	intdos(&r,&r);
}

/*
	Get the current date and time of this file.
*/
void PASCAL get_date_time(int fd, unsigned * date, unsigned * time)
{
	union REGS r;

	/* preserve date of original file */
	r.x.ax = 0x5700;
	r.x.bx = fd;
	intdos(&r,&r);
	if(r.x.cflag)
		*date = *time = 0;
	else
	{
		*time = r.x.cx;
		*date = r.x.dx;
	}
}


/*
	$DOCPROC:GetSwitch

	Look in the string for the given switch. If found, set *pfSwitch
	either TRUE or FALSE, if a /-S is given. Removes the switch from
	the string. Removes preceding whitespace from the string when done.
*/
void PASCAL GetSwitch(char * szStr, char cOption, bool *pfSwitch)
{
	char *psz = szStr;
	bool	fNegate;
	int	iCount;

	while(1)
	{
		fNegate = FALSE;
		iCount  = 1;

		// any more switches?
		psz = intl_strchr(psz, '/');
		if(!psz)
			break;

		// check for negation
		if(psz[1] == '-')
		{
			fNegate = TRUE;
			++iCount;
		}

		// is it the option we're looking for? set switch value and remove
		// switch from string
		if(toupper(psz[iCount]) == cOption)
		{
			*pfSwitch = !fNegate;
			strcpy(psz, psz+iCount+1);
			break;
		}

		// on to next character
		else
			++psz;
	}

	// remove preceding whitespace
	psz = LFNSkipWhite(szStr);
	if(psz > szStr)
		strcpy(szStr, psz);
}

/*
	Get Extended Switch looks for a switch (/A:DHS) followed
 	by extra characters that add info.  If found,
	set "*val" to true.  Removes switch from string.

	Returns pointer to string (may have moved if switch
	was at the beginning).
*/
int PASCAL get_extended_switch(char * str, char option, int *val, struct ext_switch *exs, int *list)
{
	char *p = str;
	int negate;
	int i, j;
	int found;
	int count;

	while(1)
	{
		negate = FALSE;
		i = 1;

		/* any switches left? */
		p = intl_strchr(p,'/');
		if(!p)
			break;

		/* check for negation */
		if(p[1] == '-')
		{
			negate = TRUE;
			++i;
		}

		/* option found, redefine the extended options */
		if(toupper(p[i]) == option)
		{
			/* we're redefining extended options */
			memset(list, 0, exs->count * sizeof(int));
			++i;

			/* negating this switch */
			if(negate)
				*val = FALSE;

			/* setting extended options */
			else
			{
				/* let user routines KNOW someone chose this option */
				*val |= 0x80;

				/* skip the colon, if there is one */
				if(p[i] == ':')
					++i;

				/* check out the extended options specified */
				for(count = 0; p[i] && p[i] != '/' && !DL_IS_SPACE(p[i]); ++i)
				{
					found = negate = FALSE;

					/* this extended option may be negated */
					if(p[i] == '-')
					{
						negate = TRUE;
						++i;
					}

					/* is it one of the options we're looking for? */
					for(j=0; j<exs->count; ++j)
					{
						if(exs->switches[j].option == toupper(p[i]))
						{
							if(negate)
								list[count] = exs->switches[j].value << 8;
							else
								list[count] = exs->switches[j].value;
						 	*val |= list[count];
							++count;
							found = TRUE;
						}
					}

					/* user shouldn't specify an */
					if(!found)
					{
						PrintError(EPARMS);
						return FALSE;
					}
				}
			}

			/* remove the option from the list */
			strcpy(p,p+i);
		}
		else
			++p;
	}
	return TRUE;
}

/*
	Determine if any unknown switches are left over.
*/
int PASCAL switchesleft(char *str)
{
	if(intl_strchr(str,'/'))
	{
//		PrintError(EPARMS);
		PrintError(EBADSWITCH);
		return TRUE;
	}
	return FALSE;
}

/*
	Read a line from a file. Does NOT strip off the CR/LF pair.
	Returns the length of the line, up to, but not including the
	CR/LF. Stops reading at ^Z or EOF.
*/
int PASCAL ReadBatchLine(int fd, char *str)
{
#if 0
	char	*p;
	int	iLen;

	iLen = dos_read(fd, (void far *)str, BATCH_LINELEN-1);
	if(iLen > 0)
	{
		// lines stop on Ctrl-Z and CR
		str[iLen] = '\0';
		for(p = str; iLen && *p!=0 && *p!=ctrl('Z') && *p!='\r'; ++p)
			--iLen;

		// end of file at Ctrl-Z and 0
		if((*p == 0 || *p == ctrl('Z')) && (p-str) == 0)
			return -1;

		// return line length found
		return (int)(p-str);
	}

	// end of file
	return -1;

#else

#define BATCH_BUFFER_LEN 0x20

   char *szBatchBuffer = str;
   unsigned uCounter;
   int iLen;
   int retval = -1;     // flag for error condition
   bool fBufferFull = FALSE;
   bool fEndOfLineFound = FALSE;

   do
   {
      iLen = dos_read( fd, (void far *)szBatchBuffer, BATCH_BUFFER_LEN );

      if( iLen > 0 )
      {
         fBufferFull = (retval == 1023);

         // In case there's no CRLF in a short buffer, terminate it
         //
         if( iLen < BATCH_BUFFER_LEN )
         {
            szBatchBuffer[ iLen++ ] = 0;
         }

         for( uCounter = 0; uCounter < iLen ; szBatchBuffer++,uCounter++ )
         {
            if(   ( *szBatchBuffer == ctrl('Z') )
               || ( *szBatchBuffer == '\r' )
               || ( *szBatchBuffer == 0 )
              )
            {
               retval += uCounter;
               fEndOfLineFound = TRUE;
               break;
            }
         }
         if(!fEndOfLineFound)
         {
            retval += iLen;
         }
      }
   } while ( !fBufferFull && !fEndOfLineFound && iLen > 0 );

   // if fBufferFull && !fEndOfLineFound, MS reports "Bad Command or File Name"
   //

   // Adjust for starting at -1
   //
   if( fEndOfLineFound || fBufferFull )
   {
      retval++;
   }
   return retval;
#endif
}


/*
	Get a string from DOS, assumes s is big enough to hold
	a full (128 char) command line.
*/
void PASCAL getstr(char *s, int getMode)
{
	/*
		[ locals ]
	*/
	union REGS r;
	unsigned char len=(unsigned char)*s;

	/*
		[ code ]
	*/

	// Check the Mode to make sure it's understood
	assert( (getMode | GET_NOBUF | GET_DOSKEY) == (GET_NOBUF | GET_DOSKEY));

	// If we're getting through DOSKEY
	if( (getMode & GET_DOSKEY) && (len == 0x80) )
	{
		r.x.ax = 0x4810;
		r.x.dx = (unsigned)s;
		int86( 0x2F, &r, &r );
		if( r.x.ax )
			getMode &= ~GET_DOSKEY;
	}

	// If we don't want to try DOSKEY or if DOSKEY failed
	if( !(getMode & GET_DOSKEY) )
	{
		// Clear the input buffer before getting string?
		if( getMode & GET_NOBUF)
			r.x.ax = 0x0C0A;
		else
			r.x.ax = 0x0A00;

		r.x.dx = (unsigned)s;
		intdos(&r,&r);
	}

	// Scroll one line
	printf("\n");

	// Null terminate the string
	len = *(s+1);
	*(s + 3 + len) = '\0';
}


/*
	Get a near pointer to an environment string from our
	environment.  NOTE: THIS FUNCTION IS NOT REENTRANT!
*/
char * PASCAL GetEnvironmentVariable(char *name)
{
	static char envstr[DL_MAXPATH];

	char far *ep;
	unsigned envseg;
	int i=0;

	intl_strupr(name);
	envseg = * (unsigned far *) MK_FP(_psp, 0x2C);
	ep= env_match(MK_FP(envseg, 0), name);
	if(ep==NULL)
		return NULL;

	while(*ep != '=')
		++ep;

	/* copy the env string into envstr */
	++ep;							/* skip past the = */
	while(i < sizeof(envstr)-1 && *ep)
		envstr[i++]= *ep++;
	envstr[i] = 0;

	return envstr;
}

char PASCAL convert_hour(unsigned *hour)
{
	extern struct rd_country c_country;
	char c = ' ';

	/* 12 hour format */
	if(c_country.timefmt == 0)
	{
		if(*hour >= 12)
		{
			*hour -= 12;
			c = 'p';
		}
		else
			c = 'a';

		if(*hour == 0)
			*hour=12;
	}
	return c;
}

/*
	Place the given file time, in international format, into
	the string.
*/
void PASCAL timetoa(unsigned time, char *str)
{
	extern struct rd_country c_country;

	unsigned hour, minute;
	char c;

	hour = time >> 11;
	minute = (time >> 5) & 0x3F;

	/* for those countries with a 12 hour clock */
	c = convert_hour(&hour);

  	sprintf(str, "%2d%s%02d%c", hour, c_country.tmsep, minute, c);
}


/*
	Place the given date, in international format, into the
	string.
*/
void PASCAL datetoa(unsigned date, char *str)
{
	extern struct rd_country c_country;
	extern int fDir4DigitDate;

	static char szMMDD[] = "%02d%s%02d%s";
	static char szYY[] = "%02d";
	static char szYYYY[] = "%04d";
	static char szNoDate[]   = "          ";

	char *sep = c_country.dtsep;
	unsigned year, month, day;

	/* get year/month/day from the DOS file date */
	year = (date >> 9) + 1980;
	month = (date >> 5) & 0xF;
	day  = date & 0x1F;

	/* no date, use blank string */
	if(date == 0)
		strcpy(str, szNoDate);

	/* use international date format */
	else
	{
		if( !fDir4DigitDate )
		{
			if( year >= 2100 )
			{
				year -= 2100;
			}
			else if( year >= 2000 )
			{
				year -= 2000;
			}
			else
			{
				year -= 1900;
			}
		}

		switch(c_country.datefmt)
		{
			/* mm-dd-yy*/
			case 0:
				sprintf( str,szMMDD,month,sep,day,sep );
				if( fDir4DigitDate )
				{
					sprintf( str+strlen(str), szYYYY, year );
				}
				else
				{
					sprintf( str+strlen(str), szYY, year );
				}
				break;

			/* dd-mm-yyyy */
			case 1:
				sprintf(str,szMMDD,day,sep,month,sep);
				if( fDir4DigitDate )
				{
					sprintf( str+strlen(str), szYYYY, year );
				}
				else
				{
					sprintf( str+strlen(str), szYY, year );
				}
				break;

			/* yyyy-mm-dd*/
			case 2:
				if( fDir4DigitDate )
				{
					sprintf( str, szYYYY, year );
				}
				else
				{
					sprintf( str, szYY, year );
				}
				strcat( str, sep );
				sprintf(str+strlen(str),szMMDD,month,sep,day,"");
				break;
		}
	}
}

/*
	$DOCPROC:LFNArgcopy

	Copy a single argument into the destination. Don't go over the
	maximum length. Strips quotes.
*/
void PASCAL LFNArgCopy(char *pszDst, char *pszArg, int nMaxLen)
{
	char *pszEndArg;
	int	iLen;

	// copy the argument
	pszEndArg = LFNEndArg(pszArg);
	iLen = pszEndArg - pszArg;
	if(iLen > nMaxLen-1)
		iLen = nMaxLen-1;
	memmove(pszDst, pszArg, iLen);
	pszDst[iLen] = 0;

	if( ( fLFNPresent ) || ( fIsFat32World ) )
	{
		// remove quotes (if any)
		LFNStripArgQuotes(pszDst);
	}
}

/*
	$DOCPROC:ParmCount

	Determine # of arguments, given the arg string. Assumes the string
	does starts on non-whitespace.
*/
int PASCAL ParmCount(char *s)
{
	int count=0;
	while(s && *s)
	{
		s = LFNNextArg(s);
		++count;
	}
	return count;
}

/*
	Does this string have wildcards?
*/
bool PASCAL CheckWild(char *s)
{
	if(intl_strchr(s,'?') || intl_strchr(s,'*'))
		return TRUE;
	return FALSE;
}

/*
   change a name from "name.ext" form to "name    ext"
   also expands wildcards into "????"
*/
void PASCAL name_expand(char *ename,char *name)
{
	char *p;

	/* for special subdirs "." and ".." */
	if(*name=='.')
	{
		*ename++ = '.';
		name_fillto(ename,name++,10);
		if(*name=='.')
			*ename = '.';
		ename[11] = 0;
		return;
	}
	name=name_fillto(ename,name,8);
	if((p=intl_strchr(name,'.'))!=NULL)
		name= p+1;
	name_fillto(ename+8,name,3);
	ename[11] = 0;
}


/*
	a routine used by name expand to fill the end of
	the string.
*/
char * PASCAL name_fillto(char *ename,char *name,int num)
{
	int i;
	char fill;

	/* file name */
	for(i=0;i<num;++i)
	{
		if(*name=='*')
			fill='?';
		else if(*name=='.' || *name==NULL)
			fill=' ';
		else
		{
			fill = *name;
			++name;
		}
		ename[i]=fill;
	}
	return name;
}

int PASCAL INT_get_cur_dr(void)
{
	return (bdos(GET_CUR_DR,0,0) & 0xFF);
}

void PASCAL INT_set_cur_dr(unsigned int drive)
{
	bdos(SET_CUR_DR,drive,0);
}

/*
	Returns TRUE if worked, FALSE if failed.
*/
int PASCAL INT_set_cwd(char * path)
{
	return INT_dir(CH_DIR, path);
}

/*
	mkdir, rmdir, chdir

	Returns TRUE if worked, FALSE if failed.
*/
int PASCAL INT_dir(unsigned func, char *path)
{
	extern int errno;
	union REGS r;

	r.h.ah = func;
	r.x.dx = (unsigned)path;
	intdos(&r,&r);
	if(r.x.cflag)
	{
		errno = r.x.ax;
		return FALSE;
	}
	return TRUE;
}

/*
	Get or set the DOS break flag.
*/
int PASCAL INT_setbreak(int set,int val)
{
	union REGS r;

	r.x.ax= 0x3300 + set;
	r.x.dx= val;
	intdos(&r,&r);

	return r.h.dl;
}


/*
	Get or Set the DOS verify flag.
*/
int PASCAL INT_setverify(int set,int val)
{
	union REGS r;

	r.h.ah = set ? 0x2E : 0x54;
	r.h.al = val;
	intdos(&r,&r);

	return r.h.al;
}

/*
	Set the DTA to some new address.  Cannot fail.
*/
void PASCAL INT_setdta(void * new_dta)
{
	union REGS r;

	r.x.ax = 0x1A00;
	r.x.dx = (unsigned) new_dta;
	intdos(&r,&r);
}

/*
	GetCurrentShortPath

	Gets the full path, including the drive letter, using the shortname
	functions.

	On Entry		drive (0=A, 1=B, etc...)
	On Exit		szFullPath (eg "F:\SUBDIR)

	Returns TRUE if worked, FALSE if failed.
*/
bool PASCAL GetCurrentShortPath(int drive, char * szFullPath)
{
	/* get current working directory for the drive */
	if(!dlGetCurrentDirectory(drive+1, szFullPath+3))
		return FALSE;

	/* set up the drive part of the path */
	*szFullPath++ = drive + 'A';
	*szFullPath++ = ':';
	*szFullPath   = '\\';

	return TRUE;
}

// return dos version
unsigned PASCAL GetDosVersion(void)
{
	union REGS r;
	r.x.ax = 0x3000;
	intdos(&r,&r);
	return r.x.ax;
}

/*
	Determine current PSP.
*/
unsigned PASCAL get_psp(void)
{
	union REGS regs;
	regs.x.ax = 0x5100;
	intdos(&regs, &regs);
	return regs.x.bx;
}

/*
	This functions removes the " ==  =  " from a CTTY,
	PATH, or PROMPT command.
*/
char * PASCAL skipequal(char *args)
{
	while( DL_IS_SPACE(*args) || *args == '=' )
		++args;
	return args;
}

/*
	Parse a string into an FCB.
	Returns TRUE if worked.  FALSE if bad drive.
*/
int	PASCAL dos1_parse(struct fcb * fcb, char * name)
{
	union REGS r;
	r.x.ax = 0x2900;
	r.x.si = (unsigned) name;
	r.x.di = (unsigned) fcb;
	intdos(&r, &r);
	return r.h.al != 0xFF;
}

int PASCAL get_ext_error(void)
{
	return bdos(0x59,0,0);
}

int PASCAL farstrlen(void far *s)
{
	int i = 0;

	while(((char far *)(s))[i] != '\0')
		++i;

	return i;
}


/*
	Get the volume name of the drive.
	Returns FALSE if failed, TRUE if worked.
*/
int PASCAL get_vol(int drive, char *name)
{
	struct xfcb x, volume;
	union REGS regs;

	INT_setdta((void *) &volume);

	/* set up the extended FCB */
	x.xfcb_flag = 0xFF;
	x.xfcb_attr = FA_LABEL;
	x.xfcb_fcb.fcb_drive = drive+1;
	strcpy(x.xfcb_fcb.fcb_name, "???????????");

	/* DOS 1 findfirst */
	regs.x.ax = 0x1100;
	regs.x.dx = (unsigned) &x;
	intdos(&regs, &regs);
	if(regs.h.al == 0xFF)
		return FALSE;

	/* get the name into string */
	memcpy(name, volume.xfcb_fcb.fcb_name, 11);
	name[11] = '\0';
	return TRUE;
}

char terminators[] = "/[+|=]\t;, \"";

/*
	Is this character legal in a filename?
*/
bool PASCAL filechar(unsigned char c)
{
	// all 2-byte chars allowed in filenames
	if(IsDoubleByteChar(c))
		return TRUE;

	if(c && !intl_strchr(terminators,c))
		return TRUE;

	return FALSE;
}

/*
	Date and time strings are in the similar format. Helper
	function to do_date() and do_time().
*/
void PASCAL get_datetime_str(char *str, char *id, void PASCAL (*dtfunc)(void), char *opt)
{
	extern char cmdstr_curdatetime[];
	extern char cmdstr_newdatetime[];

	if(dtfunc)
	{
		printf(cmdstr_curdatetime,id);
		(*dtfunc)();
		printf("\n");
		*str = LINELEN;
		str[1]=0;
	}
	else
		str[str[1]+2]='\r';

	printf(cmdstr_newdatetime,id,opt);
	getstr(str,GET_NORMAL);
	str[str[1]+2] = 0;
}

/*
	Given a name, get the full name.
*/
bool PASCAL GetAbsoluteName(char szAbsName[], char szName[])
{
	/* get the absolute name from this relative name */
	if(!determine_dir(szName))
		return FALSE;
	INT_set_cwd(szSplitOldCwd);

	/* return the absolute name */
	MakePath(szAbsName, szSplitFileName);
	return TRUE;
}

/*
	Given a full path, get the longname version.
*/
bool PASCAL GetLongName(char szFileName[])
{
	// try to convert to longname. If it doesn't work, ok, then
	// shortname remains.
	if( fLFNPresent )
	{
		return SmartExpandPath(FULLPATH_LONGNAME, szFileName, szFileName);
	}

	return FALSE;
}

/*
	Given any relative name, create an absolute short name
*/
bool PASCAL GetShortName(char szShortName[], char szFileName[])
{
	char szTempName[DL_MAXPATH];

	if( fLFNPresent )
	{
		// try to convert name part to short name
		if(!SmartExpandPath(FULLPATH_SHORTNAME, szFileName, szTempName))
			StrNCopy(szTempName, szFileName, DL_MAXPATH);

		// return the name in absolute short form
		return GetAbsoluteName(szShortName, szTempName);
	}
	else
	{
		// return the name in absolute short form
		return GetAbsoluteName(szShortName, szFileName);
	}
}

void PASCAL MakePath(char *dst, char *name)
{
	strcpy(dst, szSplitDrive);
	strcat(dst, szSplitDirectory);
	add_path(dst, name);
}

/*
	Return the drive for this path
*/
int PASCAL cmdgetdrive(char *path)
{
	struct fcb fcb;
	int drive;

	if(!dos1_parse(&fcb, path))
	{
		PrintError(EDRIVE);		/* not valid drive */
		return ERR;
	}

	/* return the drive (0=A, 1=B, etc) */
	drive = (fcb.fcb_drive) ? fcb.fcb_drive-1 : INT_get_cur_dr();
	return drive;
}

/*
	Parse a path into drive/path/filename.  Puts these pieces into the
	"szSplitDrive", "szSplitDirectory", "szSplitFileName" global strings.
	Changes directory into the new dir. Saves the old directory in
	"szSplitOldCwd".

	szSplitOldCwd is guaranteed to be on the same drive as
	szSplitDrive which may or may not be the current drive.

	Returns TRUE if worked, FALSE if failed.
*/
int PASCAL determine_dir(char *szPath)
{
	int	iFlags;
	extern bool fAssumeWild_szSplitFileName;
	static char szNewPath[DL_MAXPATH];

	/* split up the path */
	iFlags = LFNSplitFileName(szPath, szSplitDrive, szSplitDirectory,
					szSplitFileName);

	/* if name is similar to "CON:", remove the colon */
	if(lastchar(szSplitFileName) == ':')
		*(szSplitFileName+strlen(szSplitFileName)-1) = 0;

	/* determine drive */
	if(!(iFlags & SPLIT_DRIVE))
	{
		strcpy(szSplitDrive, "A:");
		*szSplitDrive = INT_get_cur_dr() + 'A';
	}
	intl_strupr(szSplitDrive);

	/* remember the old CWD for the given drive */
	if(!GetCurrentShortPath((*szSplitDrive-'A'), szSplitOldCwd))
	{
		PrintError(EDRIVE);		/* not valid drive */
		return FALSE;
	}

	/* no directory part */
	if(!(iFlags & (SPLIT_DIRECTORY | SPLIT_FILENAME)))
		/* don't need to chdir */;

	/* can we change into the full directory? */
	else if(SmartChangeDirectory(szPath))
	{
		*szSplitFileName = 0;
	}

	/* nope. Try just drive and path */
	else if(iFlags & SPLIT_DIRECTORY)
	{
		/* make a new string */
		strcpy(szNewPath, szSplitDrive);
		strcat(szNewPath, szSplitDirectory);

		/* no trailing '\\' unless it's the root */
		if(*szNewPath && (lastchar(szNewPath) == '\\'))
			szNewPath[strlen(szNewPath) - 1] = 0;

		if( lastchar(szNewPath) == ':')
		{
			szNewPath[2] = '\\';
			szNewPath[3] = 0x00;
		}

		/* if we can't change into this, it's not a valid directory */
		if(!SmartChangeDirectory(szNewPath))
		{
			printf("\n");
			PrintError(ECHDIR);
			return FALSE;
		}
	}

	/* convert directory to an absolute (shortname) one */
	GetCurrentShortPath((*szSplitDrive-'A'), szNewPath);
	strcpy(szSplitDirectory, szNewPath + 2);

	/* if no filename, assume all */
	if(*szSplitFileName == 0)
	{
		strcpy(szSplitFileName, "*.*");
		fAssumeWild_szSplitFileName = TRUE;
	}
	else
	{
		fAssumeWild_szSplitFileName = FALSE;
	}

	return TRUE;
}

/*
	Print a prompt and get a 'Y' or 'N' answer
	If the user enters anything but 'Y' or 'N' then Prompt again!
*/
int PASCAL get_yn(char *string)
{
	char ans = 0;
	while(1){

		printf(string);

		ans = toupper(bdos(0xC, 0, 1));		/* Clear buffer and get Ans */
		printf("\n");
		if(ans == 'Y')
			return (TRUE);
		if(ans == 'N')
			return (FALSE);
	}
}

/*
	DOS 5 provides the Drive serial with call 69H.
	Drive 0 = Default, 1 = A 2 = B.

	This call fails with any other DOS version.
*/
bool PASCAL get_drive_serial(int drive, struct drive_serial *p)
{
	union REGS regs;
	extern unsigned char osmajor;

	if(osmajor < 5)
		return (FALSE);

	regs.x.ax = 0x6900;
	regs.x.bx = drive + 1;
	regs.x.dx = (unsigned) p;	// we know we're small model
	intdos(&regs, &regs);
	if(regs.x.flags & 1)
		return (FALSE);
	return (TRUE);
}

/*
	Add an item to the path.
*/
void PASCAL add_path(char *path, char *name)
{
	if(path[strlen(path)-1] != '\\')
		strcat(path, "\\");
	strcat(path, name);
}

/*
	Remove an item from this path
*/
void PASCAL del_path(char *path)
{
	int len = strlen(path);
	for( ; len > 0; --len)
	{
		if(path[len] == '\\')
		{
			path[len] =  '\0';
			return;
		}
	}
}

/*
	Return the location of ROM-DOS code (ROM/HMA or conventional RAM)
*/
unsigned PASCAL get_dos_mem_position(void)
{
	union REGS regs;
	regs.x.ax = 0x3306;
	intdos(&regs,&regs);
	return (regs.h.dh);
}

/*
	International strupr - 2 byte character aware
*/
void PASCAL intl_strupr(char *str)
{
	union REGS r;
	r.x.ax = 0x6522;				// convert asciiz to UPPERCASE
	r.x.dx = (unsigned)str;		// DS:DX pts to string
	intdos(&r,&r);
}

/*
	International search - 2 byte character aware
*/
char * PASCAL intl_strchr(char *str, char c)
{
	register char * s = str;
	while(*s)
	{
		if(*s == c)
			return s;
		if(IsDoubleByteChar(*s))
			++s;
		++s;
	}
	return 0;	// not found
}


/*
	Very similar to strchr, but "knows" about quotes. That is, it won't
	find the character when in quotes.
*/
char * PASCAL intl_inpipestr(char *str, char c)
{
	char * p = str;
	int instring=0;
	while(*p)
	{
		//  Only handle quotes in a 7.1 or LFN environment
		//
		if( ( fLFNPresent ) || ( fIsFat32World ) )
		{
			if(*p == '"')
				instring = !instring;
		}

		if(*p==c && !instring)		/* no redirect if "instring" */
			break;
      else if((*p==c) && !strchr(p,'"'))
         // instring is set, so if there's no terminating ", we have a success!
         //
         break;

		if(IsDoubleByteChar(*p))
			++p;
		++p;
	}

	return (*p ? p : 0);				/* return 0 if not found */
}

/*
	International reverse search - 2 byte character aware
*/
char * PASCAL intl_strrchr(char *str, char c)
{
	register char * s;
	register char * p;

	// nothing in string, can't possibly find it
	if(str == 0 || *str == 0)
		return 0;

	s = str + strlen(str) - 1;
	while(s >= str)
	{
		if(*s == c)
		{
			// go forward through string to make sure this
			// isn't the 2nd byte of a 2-byte character
			for(p=str; p < s; ++p)
				if(IsDoubleByteChar(*p))
					++p;

			// we found the character, return it's position.
			if(p == s)
				return s;
		}
		--s;
	}
	return 0;	// not found
}

/*
	Find the filename extension. If no file extension, returns ptr to the
	terminating 0.
*/
char * PASCAL ExtPart(char *szPathName)
{
	char *szExt;

	// look for an extension in the name part
	szExt = intl_strrchr(LastPart(szPathName), '.');
	if(!szExt)
		szExt = szPathName + strlen(szPathName);
	return szExt;
}

/*
	Return the name part of a path. All examples below would point to
	"FileName".

	FileName.txt
	c:FileName.txt
	c:\Subdir\FileName.txt
	..\Subdir\FileName.txt
	..\
	..
	.\
	.
*/
char * PASCAL NamePart(char *szPathName)
{
	char * psz;


	if( *szPathName == 0 )
	{
		return szPathName;
	}

	if( DirExists(szPathName) )
	{
		psz = szPathName + strlen(szPathName);
		return psz;
	}

	// just after the last backslash
	psz = intl_strrchr(szPathName, '\\');
	if(psz)
	{
		++psz;
	}
	else
	{
		// or just after the colon
		//
		if( szPathName[1] == ':' )
			psz = &szPathName[2];

		//  Maybe we have a '.' or '..'
		//
		if( (!psz) && ( szPathName[0] == '.' ) )
		{
			if( szPathName[1] == 0 )
			{
				psz = &szPathName[1];
			}
			else
			{
				if( szPathName[1] == '.' )
				{
					if( szPathName[2] == 0 )
					{
						psz = &szPathName[2];
					}
				}
			}
		}

		if(!psz)
		{
			// string is already a name
			psz = szPathName;
		}
	}
	return psz;
}

/*
	Return the last part (name or dir) of a path.
*/
char * PASCAL LastPart(char *szPathName)
{
	char * psz;


	// just after the last backslash if there is one
	//
	psz = intl_strrchr(szPathName, '\\');
	if(psz)
	{
		psz++;
	}
	else
	{
		psz = szPathName;
	}

	return psz;
}

/*
	Return the last character in the string.
*/
int PASCAL lastchar(char *str)
{
	register char * p = str;
	unsigned char c = 0;

	while(*p)
	{
		c = *p;
		if(IsDoubleByteChar(*p))
			++p;
		++p;
	}
	return c;
}

/*
	International support requires some special treatment,
	such as double byte characters and date/time formats.
*/
void PASCAL InitCountryInfo(void)
{
	extern struct rd_country c_country;
	extern unsigned char far * c_dbcs_ptr;
	extern char printf_thousands_sep;

	union REGS r;
	struct
	{
		char id;
		unsigned char far * ptr;
	} buff;

	// set date/time info
	bdos(0x38, (unsigned)(&c_country), 0);
	printf_thousands_sep = *c_country.thsep;

	// set Double Byte Character Set Ptr
	r.x.ax = 0x6507;
	r.x.bx = r.x.dx = 0xFFFF;
	r.x.cx = sizeof(buff);
	r.x.di = (unsigned)(&buff);
	intdos(&r,&r);
	if(r.x.cflag == 0)
		c_dbcs_ptr = buff.ptr;
}

/*
	Return where the COPY, TYPE, DIR buffer is, and the size of it.
*/
char far * PASCAL UseBigBuffer(long *size)
{
	extern unsigned c_trans_memseg;
	extern unsigned c_trans_memparas;
	char far * buff;

	/* round buffer size to even sector length */
	*size = 16L * c_trans_memparas;
	*size = (*size & ~0x1ff);

	buff  = MK_FP(c_trans_memseg, 0);

	return buff;
}

/*
	Filename contains only dots (eg "." and "..")
*/
bool PASCAL IsDotDir(char *szName)
{
	// skip all the dots
	while(*szName == '.')
		++szName;

	// returns TRUE if only dots found
	return (*szName == 0);
}

/*
	Wait for input from the user. Accepts only "Yes" and "No".
	Returns TRUE for yes, FALSE for no.
*/
bool PASCAL GetYesNo(void)
{
	return (	GetKey("YN") == 'Y');
}

/*
	Get a single key from DOS. Used for any list of questions.

	On Entry		szChars		a list of acceptible characters. 0 means any.
	Returns the UPPERCASED character.
*/
int PASCAL GetKey(char *szChars)
{
	int c;
	while(1)
	{
		c = bdos(0x0c,0,8) & 0xFF;
		c = toupper(c);
		if(!szChars || intl_strchr(szChars, c))
			break;
	}
	printf("%c\n",c);
	return c;
}


/*
	Helper function to do_del. Displays standard error codes.
*/
void PASCAL DeleteWithErrorMsg(char szFileName[])
{
	int iErrorCode;

	if(!SmartWildcardDelete(szFileName))
	{
		// display error codes if there was an error
		iErrorCode = dlCheckDOSError();
		if( iErrorCode == ERR_ACCESS_DENIED )
		{
			PrintError(EACCESS);
		}
		else if(iErrorCode == 3)
		{
			PrintError(ENOPATH);
		}
		else
		{
			PrintError(ENOFILE);
		}
	}
}

/*
	Is this open file a device?
*/


/*
	Determine if the file in question exists. Uses the get file attribute
	function. File exists even if hidden or read only.
*/
bool PASCAL FileExists(char szName[])
{
	unsigned uAttributes;

	// must exist and be a file (not a dir)
	if(SmartGetFileAttributes(szName, &uAttributes) && !(uAttributes & A_SUBDIR))
		return TRUE;
	return FALSE;
}

/*
	Determine if the dir in question exists. Uses the get file attribute
	function. File exists even if hidden or read only.
*/
bool PASCAL DirExists(char szName[])
{
	unsigned uAttributes;

	// must exist and be a dir
	if(SmartGetFileAttributes(szName, &uAttributes) && (uAttributes & A_SUBDIR))
		return TRUE;
	return FALSE;
}

/*
	Copy a string, but don't exceed maxsize. Unlink strncpy, this
	always null terminates the string.
*/
void PASCAL StrNCopy(char *szDst, char *szSrc, int iMaxSize)
{
	strncpy(szDst, szSrc, iMaxSize);
	szDst[iMaxSize-1] = 0;
}

/*
	Append the 2nd string onto the first. Total string size cannot
	exceed iMaxSize, including the NULL. Always null terminates the
	string.
*/
void PASCAL StrNCat(char *szDst, char *szAppend, int iMaxSize)
{
	int	iLen;
	int	iLenAppend;

	// can't do anything, no length
	if(!iMaxSize)
		return;

	// truncate dest string if too large
	iLen = strlen(szDst);
	if(iLen >= iMaxSize-1)
	{
		// no room for appended string, just return
		szDst[iMaxSize-1] = 0;
		return;
	}

	// determine max length of appended string
	iLenAppend = strlen(szAppend);
	if(iLen+iLenAppend > iMaxSize-1)
		iLenAppend = (iMaxSize-1) - iLen;

	// append the string
	memmove(szDst+iLen, szAppend, iLenAppend);
	szDst[iLen+iLenAppend] = 0;
}

/*
	Set the default extension
*/
void PASCAL SetDefaultExtension(char szDest[], char szSrc[], char szExt[])
{
	strcpy(szDest, szSrc);
	strcat(szDest, szExt);
}

/*
	$DOCPROC:ValidateDirectory

	Make sure the directory exists and is valid. Does no checking on the
	filename. Parses the path into the szSplitDrive, szSplitDirectory and
	szSplitFileName globals.

	Returns FALSE and displays error message if failed.
*/
bool PASCAL ValidateDirectory(char *szPath)
{
	if(!determine_dir(szPath))
		return FALSE;
	INT_set_cwd(szSplitOldCwd);
	return TRUE;
}

//  $DOCPROC:WildcardRename
//
//  Rename files or directories (may have wildcards).  The current
//  directory should already be set to the source path and the
//  destination can not have a path.
//
bool PASCAL WildcardRename(char * szOldName, char * szNewName)
{
	bool fSuccessful;
	unsigned uCount = 0;
	FIND dp;
	int iAttr = A_NORMAL;
   char LongFileName[LFN_MAX_PATH];

	//  If this is an LFN enviornment, we can rename directories too.
	//
	if( fIsFat32World || fLFNPresent )
	{
		iAttr = A_SUBDIR;
	}

	//  Search for the file
	//
	fSuccessful = SmartFindFirst(iAttr, szOldName, &dp);
	while( fSuccessful )
	{
   	if( fLFNPresent )
      {
         SmartExpandPath(FULLPATH_LONGNAME, dp.name, LongFileName);
      }
      else
      {
         strcpy( LongFileName, dp.name );
      }

		//  Create the destination name
		//
		MatchCopyName( szTemp, szNewName, LastPart( LongFileName ));
		MatchCopyName( szTemp2, szOldName, LastPart( LongFileName ));

		//  Rename this file
		//
		fSuccessful = SmartRenameFileOrDirectory( szTemp2, szTemp );
		if( fSuccessful )
		{
			//  Update the count
			//
			uCount++;

			//  Check if there are anymore files
			//
			fSuccessful = SmartFindNext(&dp);
		}
		else
		{
			//  Free the search handle
			//
			SmartFindClose(&dp);

			return fSuccessful;
		}
	}

	//  Free the search handle
	//
	SmartFindClose(&dp);

	return (uCount != 0);
}
