//
//   $Id: ATA.H 1.4 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/****************************************************************

	ATA.H						Created 12-03-96 PKG
								Updated 12-04-96 PKG

	ATA read and write routine proto-types and defines

	Copyright (c) 1996-2002 Datalight, Inc.
	All rights reserved.
   Datalight Confidential and Proprietary

 ****************************************************************/

struct _tag_ATACardInfo
{
	unsigned isValid;
	unsigned cylinders;
	unsigned heads;
	unsigned sectors;
};

/* publics in the ATA module */
extern struct _tag_ATACardInfo ATACardInfo[];


/*
	This is the number of BIOS ticks to wait before timing out
	Should be about 3 seconds, however the ATA spec states it
	could take no more than 12 seconds.  I don't want to wait
	that long!
*/
#define MAX_BIOS_TICKS	54

/*
	Set this to 1 to use the 16 bit data port at offset 0
	rather than the alternate even and odd ports
	Set to 0 to use the 8 bit data ports.
*/
#define USE_STANDARD_DATA 1

/*
	If set to 1, ATA read/write routines will read
	access one sector at a time.  Otherwise they
	will check the status for each word of data
*/
#define ATA_READ_SECTOR 1

/* The size of one sector is 512 bytes */
#define SECSIZE 512

/* This is the device control: bits 3 and 1 are set to   */
/* turn IRQs off and set the device in a non-reset state */
#define ATA_DEV_CONTROL		0x0e


/* ATA register offsets */
#define ATA_PORT_DATA		0
#define ATA_ALT_DATA_EVEN	8
#define ATA_ALT_DATA_ODD	9

#define ATA_PORT_FEATURES	1
#define ATA_PORT_SEC_CNT	2
#define ATA_PORT_SEC_NUM	3
#define ATA_PORT_CYL_LOW	4
#define ATA_PORT_CYL_HIG	5
#define ATA_PORT_DRV_HD		6

#define ATA_PORT_STATUS		7
#define ATA_PORT_CMD			7


/* Bit mask for the control register */
#define ATA_PORT_DEV_CTRL	0x0e


/* ATA command codes */
#define ATA_CMD_IDENTIFY	0xEC
#define ATA_CMD_SET_FEAT	0xEF
#define SET_8BIT_MODE		0x01


/* These are hexidecimal, not decimal in table 9 of ATA spec */
#define ATA_CMD_READ			0x20
#define ATA_CMD_WRITE		0x30


/*
	The ATA_DH_CHS will put the Drive/Head register into
	CHS state and set the required bits!
	The DRIVE/HEAD register:
	bit	0		1		2		3		4		5		6		7
			1		L		1		DRV	HS3	HS2	HS1	HS0
 */
#define	ATA_DH_CHS		0xa0


/*
	The ERROR register, each define is the bit that would
	be set in the case of the given error
	bit	0		1		2		3		4		5		6		7
			BBK	UNC	MC		IDNF	MCR	ABRT	TK0NF	AMNF
 */
#define	ATA_ERROR_BAD_BLOCK_DETECT		0x80
#define	ATA_ERROR_UNCORR_DATA			0x40
#define	ATA_ERROR_MEDIA_CHANGED			0x20
#define	ATA_ERROR_ID_NOT_FOUND			0x10
#define	ATA_ERROR_ABORT					0x08
#define	ATA_ERROR_MEDIA_CHANGE_REQ		0x04
#define	ATA_ERROR_TRACK_0_NOT_FOUND	0x02
#define	ATA_ERROR_ADDRESS_NOT_FOUND	0x01


/*
	The status register, each define is the bit that would
	be set to indicate the current state of the drive
	bit	0		1		2		3		4		5		6		7
			ERR	IDX	CORR	DRQ	DSC	DWF	DRDY	BSY
 */
#define	ATA_STATUS_BSY					0x80
#define	ATA_STATUS_DRDY				0x40
#define	ATA_STATUS_WRITE_FAULT		0x20
#define	ATA_STATUS_SEEK_COMPLETE	0x10
#define	ATA_STATUS_DRQ					0x08
#define	ATA_STATUS_CORRECTED_DATA	0x04
#define	ATA_STATUS_INDEX				0x02
#define	ATA_STATUS_ERR					0x01

/* ata.c */
unsigned ATABlockRead(uchar socket, void far *dataPtr,
								ulong start, unsigned len);
unsigned ATABlockWrite(uchar socket, void far *dataPtr,
								ulong start, unsigned len);
ulong SetupATACard(uchar socket);
