//
//   $Id: PCIC.C 1.5 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/************************************************************************
 *  Source   : PCIC.C
 *  Desc     : Contains all PCMCIA socket related functions
 *  Notes    : Compatible with PCIC i82365sl [A-step(82) & B-step(83)]
 *
 *     Created: 6-7-95 PKG
 *     Updated: 12-04-96 PKG
 *
 *  Copyright (c) 1995-2002 Datalight, Inc.
 *  All right reserved.
 *  Datalight Confidential and Proprietary
 ***********************************************************************/
#define __FILENUM__	3

#include <stdlib.h>	/* _rotl, _rotr */
#include <dos.h>		/* outport, inport */
#include "datalght.h"
#include "common.h"
#include "pcic.h"


/* Global variables */
#if SAV_REG_FUNCS
/* The PCIC has a maximum of 0..0xff registers */
static char element[256];
#endif

/* Slight difference in between rev 0x83 and 0x84 */
static unsigned pcic_DF = FALSE;	/* 1 if it is rev 84h */

/* Utilities in TRIUTIL */
void far * dos_getvect(int intnum);
 typedef void (interrupt far * farisr)(void);
void dos_setvect(int intnum, void far * isr);

/* Need a place to save our old IRQ handler and ours */
void interrupt (*oldIRQHandler)();
extern void interrupt ourHandler(void);
 void (*Handler)(char, unsigned);


/*********************/
/* Hardware Routines */
/*********************/


/* ------------------------------------------------------------ *
 * GetPCIC
 *
 * Given a PCIC index number, write to the index register and
 * and return the data.
 *
 * On Entry:
 *    index index register number
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Return data of the given register
 * ------------------------------------------------------------ */
uchar GetPCIC(uchar index, uchar socket)
{
	outportb(PCIC_INDEX, index+(socket*0x40));
	return inportb(PCIC_DATA);

}


/* ------------------------------------------------------------ *
 * CheckHardware
 *
 * Checks to see if the Hardware exists by verifing the PCIC
 * revision number is 82h, 83h, or 84h.  CheckHardware will
 * be used to determine the number of PC card sockets available.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Return TRUE if present, FALSE otherwise
 * ------------------------------------------------------------ */
unsigned CheckHardware(uchar socket)
{
   unsigned data;

	ASSERT(socket<MAX_PCIC_SOCKETS);

   data = GetPCIC(IDENTIFY, socket);

   if (data==PCIC_revDF)
      pcic_DF = TRUE;

   return ((data == PCIC_revA)
				|| (data == PCIC_revB)
				|| (data == PCIC_revDF));

}  /* END-CheckHardware */


/* ------------------------------------------------------------ *
 * CardDetect
 *
 * Verifies if card is inserted in socket.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Return TRUE if Card is inserted.
 * ------------------------------------------------------------ */
unsigned CardDetect(uchar socket)
{
   unsigned data;

	ASSERT(socket<MAX_PCIC_SOCKETS);

	data = GetPCIC(STATUS, socket);

	/* Are CD1 & CD2 active? */
   return((data & _CD1) && (data & _CD2));

}  /* END-CardDetect */


/* ------------------------------------------------------------ *
 * CardWriteProtect
 *
 * Checks the write protect status
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Return TRUE if Card is write protected
 * ------------------------------------------------------------ */
unsigned CardWriteProtect(uchar socket)
{
	ASSERT(socket<MAX_PCIC_SOCKETS);

	/* Is the WP status bit set? */
   return(GetPCIC(STATUS, socket) & _WP);

}  /* END-CardWriteProtect */


/* ------------------------------------------------------------ *
 * ResetSocket
 *
 * Sets the RESET# bit, We don't want this to be active!
 *
 * NOTE: This was added since the RESET signal was found to
 * often default to the RESET# low state, causing the PCIC
 * to behave strangely.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Card has been reset
 *    Returns TRUE on success or FALSE for a bad socket
 * ------------------------------------------------------------ */
unsigned ResetSocket(uchar socket)
{
	ASSERT(socket<MAX_PCIC_SOCKETS);

	/* DeASSERT Reset Signal NOTE: this is active low */
	/* We don't want this floating */
	outportb(PCIC_DATA, GetPCIC(INT_CONTROL, socket) | _RESET);

	return TRUE;

}  /* END-ResetSocket */


/* ------------------------------------------------------------ *
 * EnableChangeIRQ
 *
 * Enables the Card Detect Change Interrupt to the specified
 * IRQ for a socket.  We will only allocate IRQ's 8 and above.
 *
 * NOTE: Handler(unsigned, unsigned) MUST be define prior to
 *			enabling the ChangeIRQ lines see SetCallbackHandler()
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    TRUE if PC Card IRQ is configured to irq for socket
 * ------------------------------------------------------------ */
unsigned EnableChangeIRQ(uchar socket, unsigned irq)
{
	unsigned data;

	ASSERT(socket<MAX_PCIC_SOCKETS);

	/* 0x05 == Card Status Change IRQ */
	data = GetPCIC(INT_CONFIG, socket);

	/* Set the IRQ and the CD Enable */
	data = ((irq<<4) | _CD_ENABLE);

	/* 0x05 == Card Status Change IRQ */
   outportb(PCIC_DATA, data);

#if INCLUDE_IRQ && (0)
	/* We have to be sure that the second 8259 interrupt */
	/* controller is enabling the appropriate IRQ by */
	/* forcing the irq bit low */
	data = inport(SLAVE_8259A_DATA);

	/* We need ALL bits set save the IRQ bit */
	data &= ~(1 << (irq & ~0xf8));
	outport(SLAVE_8259A_DATA, data);

	/* Set the interrupt vector to point to our new handler */
	oldIRQHandler = (void far*) dos_getvect(0x70+(irq-0x08));
	dos_setvect(0x70+(irq-0x08), ourHandler);
#endif

	return TRUE;

} /* END-EnableChangeIRQ */


/* ------------------------------------------------------------ *
 * DisableChangeIRQ
 *
 * Disables the Card Detect Change Interrupt from the specified
 * socket.
 *
 * NOTE: This will disable ALL IRQ's for the particula socket.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    TRUE after PC Card IRQ is disabled
 * ------------------------------------------------------------ */
#ifdef __BORLANDC__
#pragma argsused
#endif
unsigned DisableChangeIRQ(uchar socket, unsigned irq)
{
	unsigned data;

	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(irq < 0xf);

	/* 0x05 == Card Status Change IRQ */
   GetPCIC(INT_CONFIG, socket);
	/* Disable the IRQs */
   outport(PCIC_DATA, 0);

#if INCLUDE_IRQ
	/* Tell the 8259 to disable the IRQ we allocated for */
	/* CD Change Interrupts */
	data = inport(SLAVE_8259A_DATA);

	/* Clear only the IRQ bit */
	data |= (1 <<(irq&~0xf8));
	outport(SLAVE_8259A_DATA, data);

	/* Restore the original IRQ handler */
	dos_setvect(0x70+(irq-0x08), oldIRQHandler);
#endif

	return TRUE;

} /* END-DisableChangeIRQ */


/* ------------------------------------------------------------ *
 * AckCDIRQ
 *
 * Acknowledges a CD change IRQ by reading the CD Source
 * Interrupt register and returning 1 if the specified socket
 * generated the interrupt.
 *
 * NOTE: Handler(unsigned, unsigned) MUST be defined prior to
 *			enabling the ChangeIRQ lines
 *
 * On Entry:
 *    socket is the socket status register to check
 *
 * On Exit:
 *    Return TRUE if socket is the source of an interrupt
 * ------------------------------------------------------------ */
unsigned AckCDIRQ(uchar socket)
{
	ASSERT(socket<MAX_PCIC_SOCKETS);

	/* Return non-zero if CD source is set */
	return (GetPCIC(INT_SOURCE, socket) & _CD_SRC);

} /* END-AckCDIRQ */


/* ------------------------------------------------------------ *
 * SetCallbackHandler
 *
 * Set the function pointer to call upon an insertion or
 * removal of a PC card.
 *
 * On Entry:
 *    nothing
 *
 * On Exit:
 *    Handler is set to the passed function
 * ------------------------------------------------------------ */
#if INCLUDE_IRQ
void SetCallbackHandler(void (*CBHandler)(char, unsigned))
{
	ASSERT(CBHandler);
	Handler = CBHandler;

} /* END-SetCallbackHandler */
#endif

/* ------------------------------------------------------------ *
 * PowerDownSocket
 *
 * Disables all power to socket and disables the output signals
 * Note also that the Vcc & Vpp values are cleared to zero
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *
 * On Exit:
 *    Return TRUE after Vcc, Vpp1 & Vpp2 have been disabled
 * ------------------------------------------------------------ */
unsigned PowerDownSocket(uchar socket)
{
	ASSERT(socket<MAX_PCIC_SOCKETS);

	/* Vpp and Vcc control */
	GetPCIC(POWER, socket);

	/* The DF power control regist is different from */
	/* erlier revisions of the hardware */
   if (pcic_DF)
		/* set Vxx to zero */
      outportb(PCIC_DATA, inportb(PCIC_DATA) & _ZERO_VXX_DF);
   else
		/* disable power */
      outportb(PCIC_DATA, inportb(PCIC_DATA)& _ZERO_VXX);

	/* disable outputs */
   outportb(PCIC_DATA, inportb(PCIC_DATA) & _DISABLE_OUTPUT);

	return TRUE;

} /* END-PowerDownSocket */


/* ------------------------------------------------------------ *
 * PowerUpSocket
 *
 * Turns on power to socket.  Sets Vcc to 5 volts.  Sets Vpp1 & Vpp2 to
 * the same value, either 12, 5 or 0 volts depending on vpp input
 * value.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    vpp is 12, 5 or 0  (decimal)
 * On Exit:
 *    Set Vcc=5 & Vpp1,Vpp2 = vpp, enable Vcc, Vpp1 & Vpp2
 *    Returns TRUE on success or FALSE for a bad socket
 * ------------------------------------------------------------ */
unsigned PowerUpSocket(uchar socket, unsigned vpp)
{
	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(vpp==0 || vpp==12 || vpp==5);

	GetPCIC(POWER, socket);

	/* clear everthing */
	outportb(PCIC_DATA, 0);

	/* The DF power control register is different from */
	/* earlier revisions of the hardware */
   if (pcic_DF)
   {
	   if (vpp==12)
		   outportb(PCIC_DATA, _PWR_VPP_12_DF);		/* Vcc=5, Vpp=12 */
	   else
		   outportb(PCIC_DATA, _PWR_VPP_5_DF);			/* Vcc=5, Vpp=5 */
   }
   else
   {
	   if (vpp==12)
		   outportb(PCIC_DATA, _PWR_VPP_12);			/* Vcc=5, Vpp=12 */
	   else
		   outportb(PCIC_DATA, _PWR_VPP_5);			/* Vcc=5, Vpp=5 */

		/* pwr enable */
	   outportb(PCIC_DATA, inportb(PCIC_DATA) | _PWR_ENABLE);
   }  /* IF - pcic_DF */


	/* Set reset off, we want to remain in control */
	outportb(PCIC_DATA, inportb(PCIC_DATA) | _DISABLE_RESET);

	/* output enable */
	outportb(PCIC_DATA, inportb(PCIC_DATA)|_ENABLE_OUTPUT);

	/*
		need a short delay here to allow Vpp and Vcc to ramp
		30 ms should be enough
	*/
	internal_delay(30);

	return TRUE;

}  /* END-PowerUpSocket */


/* ------------------------------------------------------------ *
 * ConfigureSocket
 *
 * Configure socket interface type to memory or IO.  This just
 * enables the Multiplexer for the IO interface.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    type is 0 (memory - SOCKET_MEM) or 1 (IO - SOCKET_IO)
 *
 * On Exit:
 *    PCIC has been configured
 *    Returns TRUE on success or FALSE for a bad socket or type
 * ------------------------------------------------------------ */
unsigned ConfigureSocket(uchar socket, unsigned type)
{
   unsigned data;

	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(type==SOCKET_MEM || type==SOCKET_IO);

	data = GetPCIC(INT_CONTROL, socket);

   if (type)
      data|=_IO_ENABLE;    /* IO */
   else
      data&=_IO_DISABLE;    /* Memory */

	/* 0x03 == Interrupt & General Control */
   outportb(PCIC_DATA, data);

	return TRUE;

}  /* END-ConfigureSocket */


/* ------------------------------------------------------------ *
 * SetupIoWindow
 *
 * This routine creates a system memory mapped IO window.  It creates
 * a 15 byte wide memory window.  Its main purpose was for Intel's
 * Flash ATA drive.  It is not used in Flash cards.
 *
 * The IO window number WILL be 10 and 11, these number were chosen
 * to keep them different from the memory windows and to allow bit
 * manipulations a bit simpler.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    winNum is 10-11
 *    port is any valid IO address number on the system (ie: 300h)
 * On Exit:
 *    OK
 *       returns TRUE
 *    error
 *       returns FALSE on invalid winNum or socket
 * ------------------------------------------------------------ */
unsigned SetupIoWindow(uchar socket, unsigned winNum, unsigned port)
{
	char ctrlByte;
	union
	{
		unsigned uWord;
		struct
		{
			char lByte;
			char hByte;
		}b;
	} lPort;

	ASSERT(winNum==10 || winNum==11);
	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(!(port%0x08));

	/* Setup our port structure and fix winNum */
   lPort.uWord = port;
   winNum -= 10;

   /* 0 Wait States, IOCS16# from card, 16bit data */
	ctrlByte = GetPCIC(IO_CONTROL, socket) & _rotr(0xF0, winNum*4);
   outportb(PCIC_DATA, ctrlByte | _rotl(_IO_CTRL, winNum*4));

   /* Start Port Address */
	/* Low Byte Addr */
	GetPCIC(IO_START_LOW(winNum), socket);
   outportb(PCIC_DATA, lPort.b.lByte);
   /* High Byte Addr */
	GetPCIC(IO_START_HIGH(winNum), socket);
   outportb(PCIC_DATA, lPort.b.hByte);

   /* Stop Port Address (15 bytes) */
	/* Low Byte Addr */
	GetPCIC(IO_END_LOW(winNum), socket);
   outportb(PCIC_DATA, lPort.b.lByte + 0x0F);	/* Add 15 bytes */
   /* High Byte Addr */
	GetPCIC(IO_END_HIGH(winNum), socket);
   outportb(PCIC_DATA, lPort.b.hByte);

   return TRUE;

}  /* END-SetupIoWindow */


/* ------------------------------------------------------------ *
 * SetupMemWindow
 *
 * This routine creates a system memory mapped memory window.  The cards
 * address space is mapped into the computer's system memory at addr.  It
 * creates a 4K memory window starting at address 0 on the card and at
 * address addr in system memory.  Setting type to 0 will access common
 * memory and setting type to 1 will access attribute (tuple) space.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    type is 0 or 1 (0 = MEM_COMMON, 1 = MEM_ATTRIB)
 *    winNum is 0-4
 *    addr is seg:off (ie: D0000000) Base system memory address
 *		size is the window siaze in bytes
 *
 * On Exit:
 *    OK
 *       returns TRUE
 *    error
 *       returns FALSE on invalid winNum, socket or type
 * ------------------------------------------------------------ */
unsigned SetupMemWindow(uchar socket, unsigned type, unsigned winNum,
								ulong addr, unsigned size)
{
	char winbase;        /* Window base address byte */
	char winoff;         /* Window offset address byte */

	ASSERT(winNum<5);
	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(type==MEM_COMMON || type==MEM_ATTRIB);
	ASSERT(size%4096==0);

	/* The PCIC takes the two's compliment of the card	   */
	/* offset to place at the given system memory address */

	winbase=(addr>>24);           /* set window base address byte */
	winoff =~winbase;             /* perform 2's compliment */
	winoff++;                     /*    on winoff */


	/* Card Memory Address : Set (Attr. or Comm) & no wp */
	/* Low Byte Addr */
	GetPCIC(MEM_OFF_LOW(winNum), socket);
	outportb(PCIC_DATA, winoff);
	/* High Byte Addr set the type of memory to address */
	GetPCIC(MEM_OFF_HIGH(winNum), socket);
	outportb(PCIC_DATA, type?0x7F:0x3F);

	/* Start System Memory Address : Set 16bit & wait states */
	/* Low Byte Addr */
	GetPCIC(MEM_START_LOW(winNum), socket);
	outportb(PCIC_DATA, winbase);
	/* High Byte Addr */
	GetPCIC(MEM_START_HIGH(winNum), socket);

#if ZERO_WAIT_STATE
	outportb(PCIC_DATA, _ZERO_WAIT);	/* Zero wait state */
#else
	outportb(PCIC_DATA, _ONE_WAIT);		/* One wait state */
#endif

	/* Stop System Memory Address : Set zero wait states & 4K size */
	/* Low Byte Addr */
	GetPCIC(MEM_END_LOW(winNum), socket);
	outportb(PCIC_DATA, winbase+(size/4096-1));	/* no smaller than 4k */
	/* High Byte Addr */
	GetPCIC(MEM_END_HIGH(winNum), socket);

	/* Additional wait states for 16 bit access, defined */
	/* in PCIC.H, This is normaly 0, but may need to be  */
	/* 1 or 2 for additional wait states                 */
	outportb(PCIC_DATA, PC_WAIT_STATES<<6);

	return TRUE;

}  /* END-SetupMemWindow */


/* ------------------------------------------------------------ *
 * SlideMemWindow
 *
 * This routine modifies an existing system memory mapped memory window.
 * The card's address space is adjusted to a new offset (cardAddr) using the
 * existing system memory base address.  The window will now view a new
 * location on the card.
 *
 * NOTE: In order to move the address of the memmory window you MUST
 * MUST disable the window and re-enable the window before and after
 * calling SlideMemWindow function  6/8/95  pkg
 *
 * cardAddr MUST ALSO be on a 4k boundary ( cardAddr & 0xffff000)
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    type is 0 or 1 (0 = common, 1 = attribute memory)
 *    winNum is 0-4
 *    cardAddr is an absolute address (NOT seg:off).
 *
 * On Exit:
 *    returns TRUE upon success, FALSE otherwise
 * ------------------------------------------------------------ */
unsigned SlideMemWindow(uchar socket, unsigned type, unsigned winNum,
								ulong cardAddr)
{
   unsigned card_offset;    /* Card offset address byte */
   unsigned system_addr;    /* System memory window address */

	ASSERT(winNum<5);
	ASSERT(socket<MAX_PCIC_SOCKETS);
	ASSERT(type==MEM_COMMON || type==MEM_ATTRIB);

	/* High Byte Addr */
	system_addr = GetPCIC(MEM_START_HIGH(winNum), socket);
   system_addr<<=8;
	/* Low Byte Addr */
   system_addr |= GetPCIC(MEM_START_LOW(winNum), socket);
   system_addr&=0x3FFF;

	/* Loose the lower 12 bits, and calculate the card offset */
   card_offset = (cardAddr>>12);
   card_offset -= system_addr;
   card_offset &= 0x3FFF;

   if (type)                    /* Check for Attribute Window type and */
      card_offset|=0x4000;      /*    set bit if so. */

   /* Card Memory Address : Set (Attr. or Comm) & no wp */
	/* Low Byte Addr */
	GetPCIC(MEM_OFF_LOW(winNum), socket);
   outportb(PCIC_DATA, card_offset);
   card_offset>>=8;
	/* High Byte Addr */
	GetPCIC(MEM_OFF_HIGH(winNum), socket);
   outportb(PCIC_DATA, card_offset);

   return TRUE;

}  /* END-SlideMemWindow */


/* ------------------------------------------------------------ *
 * WindowDisable
 *
 * Disables the window.  All future accesses will return garbage.
 * The window addresses are not modified.
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    window is 0-4=memory, 10-11=IO
 *
 * On Exit:
 *    OK
 *       returns TRUE
 *    error
 *       returns FALSE on invalid window param
 * ------------------------------------------------------------ */
unsigned WindowDisable(uchar socket, unsigned window)
{
   unsigned winbit=1;

	ASSERT(socket<MAX_PCIC_SOCKETS);

   switch (window)
   {
      case 0: /* MEMORY */
      case 1:
      case 2:
      case 3:
		case 4:
	      winbit = _rotl(winbit, window);
	      break;

      case 10: /* IO */
      case 11:
			winbit = _rotl(winbit, window-4);
			break;

		default:
			ERROR();
			return FALSE;
   }

   /* Disable window */
	/* 0x06 == Address Window Enable Reg. */
   outportb(PCIC_DATA, GetPCIC(WIN_CONTROL, socket) &~ winbit);

   return TRUE;

}  /* END-WindowDisable */


/* ------------------------------------------------------------ *
 * WindowEnable
 *
 * Enables an existing window.  The window must already be setup
 * via a call to SetupMemWindow() or SetupIoWindow()
 *
 * On Entry:
 *    socket is 0 to MAX_PCIC_SOCKETS-1
 *    window is 0-4=memory, 10-11=IO
 *
 * On Exit:
 *    OK
 *       returns TRUE
 *    error
 *       returns FALSE on invalid window param
 * ------------------------------------------------------------ */
unsigned WindowEnable(uchar socket, unsigned window)
{
   unsigned winbit=1;

	ASSERT(socket<MAX_PCIC_SOCKETS);

   switch (window)
   {
      case 0: /* MEMORY */
      case 1:
      case 2:
      case 3:
      case 4:
			winbit = _rotl(winbit, window);
			break;

      case 10: /* IO */
      case 11:
			winbit = _rotl(winbit, window-4);
			break;

		default:
			ERROR();
			return FALSE;
   }

   /* Enable window */
	/* 0x06 == Address Window Enable Reg. */
	outportb(PCIC_DATA, GetPCIC(WIN_CONTROL, socket)|(winbit|(MEMCS16<<5)));

   return TRUE;

}  /* END-WindowEnable */


/* ------------------------------------------------------------ *
 * WindowSize
 *
 * WindowSize will simply allow upper software layers to check
 * out the size of a memory window that is used.
 *
 * On Entry:
 *		nothing
 *
 * On Exit:
 * 	returns the size of a memory window
 * ------------------------------------------------------------ */
unsigned PCICWindowSize(void)
{
	return (unsigned) WINDOW_SIZE;

} /* END-WindowSize */


/* The following two functions are for saving and restoring the 		*/
/* PCIC registers when accessing it via this library.  The registers */
/* should be pulled (saved) before accessing the PCIC and pushed 		*/
/* (restored) before exiting program.                                */
/*																							*/
/*	This is usefull in utilities such as the CIS dumper and reading   */
/* writing card images to allow them to work while other PCMCIA      */
/* programs are being used.                                          */
/*																							*/
/* char element[256];      Structure for PCIC registers             */
/*																							*/

/* ------------------------------------------------------------ *
 * pull_registers
 *
 * Saves all the pcic register values.  This helps in allowing utilities
 * to work correctly when Card Services and Socket Services on loaded.
 *
 * On Entry:
 *
 * On Exit:
 *    returns void ; fills the element array
 * ------------------------------------------------------------ */
#if SAV_REG_FUNCS
void pull_registers(void)
{

	int x_ctr;

	/* Loop through all the available PCIC registers and save */
	/* them in the global element[] array */
	for(x_ctr=0;x_ctr<0x80;x_ctr++)
	{
		outportb(PCIC_INDEX,x_ctr);
		element[x_ctr]=inportb(PCIC_DATA);
	}
}	/* END-pull_registers */
#endif /* SAV_REG_FUNCS */


/* ------------------------------------------------------------ *
 * push_registers
 *
 * Restores all the pcic registers to their original values when the
 * pull_registers() was called.  This helps in allowing utilities
 * to work correctly when Card Services and Socket Services on loaded.
 *
 * On Entry:
 *
 * On Exit:
 *    returns void ; PCIC restored from the element array
 * ------------------------------------------------------------ */
#if SAV_REG_FUNCS
void push_registers(void)
{

	int x_ctr;

	/* Write all values from element[] to PCIC */
	for(x_ctr=0;x_ctr<0x36;x_ctr++)
	{
		/* But avoid writing to reserved registeres */
		if ((x_ctr == 0x16) || (x_ctr == 0x17) ||
			(x_ctr == 0x1e) || (x_ctr == 0x1f) ||
			(x_ctr == 0x26) || (x_ctr == 0x27) ||
			(x_ctr == 0x2e) || (x_ctr == 0x2f))
		{
			continue;
		}
		outportb(PCIC_INDEX,x_ctr);
		outportb(PCIC_DATA,element[x_ctr]);

	}

	/* Write all values from element[] to PCIC */
	for(x_ctr=0x40;x_ctr<0x76;x_ctr++)
	{
		/* But avaoid writing to reserved registeres */
		if ((x_ctr == 0x56) || (x_ctr == 0x57) ||
			(x_ctr == 0x5e) || (x_ctr == 0x5f) ||
			(x_ctr == 0x66) || (x_ctr == 0x67) ||
			(x_ctr == 0x6e) || (x_ctr == 0x6f))
		{
			continue;
		}
		outportb(PCIC_INDEX,x_ctr);
		outportb(PCIC_DATA,element[x_ctr]);
	}
}  /* END-push_registers */
#endif /* SAV_REG_FUNCS */

/* End of File : PCIC.C */
