//
//   $Id: PCIC.H 1.4 2002/10/15 20:31:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/************************************************************************
 *  Source   : PCIC.H
 *  Desc     : Contains all defines, and structures for PCIC.C
 *  Notes    : Compatible with PCIC i82365sl [A-step(82) & B-step(83)]
 *
 *     Created: 6-8-95 PKG
 *     Updated: 12-04-96 PKG
 *
 *  Copyright (c) 1995-2002 Datalight, Inc.
 *  All rights reserved.
 *  Datalight Confidential and Proprietary
 ***********************************************************************/
#ifndef _PCIC_H
#define _PCIC_H

/***********************************************************************/
/*   LOOK AT ITEMS BELOW   LOOK AT ITEMS BELOW   LOOK AT ITEMS BELOW   */
/***********************************************************************/

/* Include pull and push register functions? */
#define SAV_REG_FUNCS	0

/* Set this to 1 to include IRQ handling code */
#define INCLUDE_IRQ		0

/* The default IRQ to generate for a Card Change */
#define	CHANGE_IRQ			0x00

/* PC_WAIT_STATES will define the number of additional    */
/* wait states for 16 bit access to the PC card generated */
/* by the PC card.  This can be 0 (standard) to 3         */
#define PC_WAIT_STATES	3

/***********************************************************************/
/*   LOOK AT ITEMS ABOVE   LOOK AT ITEMS ABOVE   LOOK AT ITEMS ABOVE   */
/***********************************************************************/

/* Default memory window size for the PCIC module to use */
/* NOTE:  Smallest window size is 4k */
#define WINDOW_SIZE	(4 * 1024)		/* 4k */
#define IO_WINDOW_SIZE (0x0f)			/* 16 bytes */

/* Maximum number of sockets that we can handle */
#define MAX_PCIC_SOCKETS 2

/* PCIC revision code */
#define PCIC_revA		0x82
#define PCIC_revB		0x83
#define PCIC_revDF	0x84

/* Use zero or one wait states 0 - false   1 - true */
#define ZERO_WAIT_STATE 1

/* If this is set to 0 the MEMCS16# will be driven active */
/* providing hardware decoding of 16 bit access to 8 bit  */
/* This should normally NOT be needed                     */
#define MEMCS16 1

/* This should never happen! */
#if PC_WAIT_STATES>3
#error PC_WAIT_STATES may NOT be more than 3
#endif

/* Number of windows supported by the 82365SL PCIC */
#define NUM_IO_WINDOWS	2
#define NUM_MEM_WINDOWS	5

/* defines pcic io port */
#define  PCIC_INDEX        0x3e0
#define  PCIC_DATA         0x3e1

/* One of these values will be passed to the */
/* call back function */
#define INSERTION		0x80
#define REMOVAL		0x40

/* Memmory types for Memmory Windows */
#define MEM_COMMON	0
#define MEM_ATTRIB	1

/* Configuration types to send to ConfigureSocket */
#define SOCKET_MEM	0
#define SOCKET_IO		1

/* delay needed on IBM laptops */
#define  IBM_PWR_DELAY     0xFFF

/* 8259 ICU addresses */
#define SLAVE_8259A	0xa0
#define SLAVE_8259A_DATA	0xa1
#define MASTER_8259A	0x20
#define MASTER_8259A_DATA	0x21

/* End Of Resource flag */
#define EOR	0xffff

/* Window numbers to be saved in the resource list */
/* need to be different, so we need a simple hash  */
/* algorithim to keep track of each window/socket  */
/* combination */
#define WIN_NUM(socket,window)	(socket*0x20+window+1)

/* INDEX values for 82365SL PCIC */
#define IDENTIFY					(0x00)
#define STATUS						(0x01)
#define POWER						(0x02)
#define INT_SOURCE				(0x04)
#define WIN_CONTROL				(0x06)
#define GEN_CONTROL				(0x16)
#define GLOBAL_CONTROL			(0x1e)

#define INT_CONTROL				(0x03)
#define INT_CONFIG				(0x05)

#define IO_CONTROL				(0x07)

#define IO_START_LOW(win)		(0x08+(win*0x04))
#define IO_START_HIGH(win)		(0x09+(win*0x04))
#define IO_END_LOW(win)			(0x0a+(win*0x04))
#define IO_END_HIGH(win)		(0x0b+(win*0x04))

#define MEM_START_LOW(win)		(0x10+(win*0x08))
#define MEM_START_HIGH(win)	(0x11+(win*0x08))
#define MEM_END_LOW(win)		(0x12+(win*0x08))
#define MEM_END_HIGH(win)		(0x13+(win*0x08))
#define MEM_OFF_LOW(win)		(0x14+(win*0x08))
#define MEM_OFF_HIGH(win)		(0x15+(win*0x08))


/* Some commonly used PCIC flags */
#define _CD1				0x04
#define _CD2				0x08

#define _WP					0x10
#define _RESET				0x40

#define _CD_SRC			0x08
#define _CD_ENABLE		0x08

#define _PWR_ENABLE		0x10

#define _DISABLE_RESET	0x40

#define _ZERO_VXX_DF		~0x1B
#define _ZERO_VXX			~0x10
#define _DISABLE_OUTPUT	~0x80
#define _ENABLE_OUTPUT	0x80

#define _PWR_VPP_12_DF	0x12
#define _PWR_VPP_5_DF	0x11
#define _PWR_VPP_12		0x0a
#define _PWR_VPP_5		0x05

#define _IO_ENABLE		0x20
#define _IO_DISABLE		~0x20

/*
	Turn on a 16 bit data path, but only generat IOCS16# by the
	IOIS16# from the ISA bus and ignore the PCIC settings.  This
	will allow a 16 bit and an 8 bit data path to the IO registers.
*/
#define _IO_CTRL			0x0b

#define _ZERO_WAIT		0x80
#define _ONE_WAIT			0xc0

/* io addresses */
typedef struct __io_tag
{
	unsigned addr;			/* Each socket has 2 io windows */
} _io;

/* memory available */
typedef struct __mem_tag
{
	unsigned long addr;	/* and 5 addresses */
} _mem;


/* PCIC.C */
uchar GetPCIC(uchar index, uchar socket);
unsigned CheckHardware(uchar socket);
unsigned CardDetect(uchar socket);
unsigned CardWriteProtect(uchar socket);
unsigned ResetSocket(uchar socket);
unsigned EnableChangeIRQ(uchar socket, unsigned irq);
unsigned DisableChangeIRQ(uchar socket, unsigned irq);
unsigned AckCDIRQ(uchar socket);
void SetCallbackHandler(void (*CBHandler));
unsigned PowerDownSocket(uchar socket);
unsigned PowerUpSocket(uchar socket, unsigned vpp);
unsigned ConfigureSocket(uchar socket, unsigned type);
unsigned SetupIoWindow(uchar socket, unsigned winNum, unsigned port);
unsigned SetupMemWindow(uchar socket, unsigned type, unsigned winNum,
								ulong addr, unsigned size);
unsigned SlideMemWindow(uchar socket, unsigned type, unsigned winNum,
								ulong cardAddr);
unsigned WindowDisable(uchar socket, unsigned window);
unsigned WindowEnable(uchar socket, unsigned window);
unsigned PCICWindowSize(void);
void pull_registers(void);
void push_registers(void);

#endif   /*  END-_PCIC_H  */

