//
//   $Id: BLOCK.C 1.5 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  BLOCK.C	created 	12/17/96	by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *  	Contains code for caching read sectors from the CDROM
 *	either to conventional memory or XMS memory.
 *
 *  Revision History:
 *
 * 12/23/96 jmb  Rewrote caching code.  Made it simpler
 *					designed it to allow the use of XMS memory
 * 12/04/96 jmb	Corrected increment of pointer in the
 *					routine cd_blkkilld.  Added code to clear
 *					the logical sector number and user fields
 *					in cd_blkkilld
 *********************************************************/
#define __FILENUM__ 7				// for ASSERT MACRO

#include "datalght.h"

#include "cdstruct.h"
#include "cdrom.h"
#include <stdio.h>
#include "xms.h"

//
// PROTOTYPE Includes
//
#include "block.pt"
#include "cddev.pt"
#include "cdutil.pt"
#include "cdmemory.pt"
#include "xms.pt"

//
// external data
//
extern int numBuffers;
extern int bUseExtended;
extern int numDrives;

//
// public data
//
unsigned uXMSBufferHandle;			// XMS handle to the memory for the buffers
RAW_BUF far *lpLocalBuffer;			// this is only set when using XMS
												// for the buffers



/*
**************************************************************************
 CD_BLKINIT -    Allocate from free memory and initialize buffer space
        and control structures used for buffered reads from a CD.
        The argument nalloc defines how many buffers to allocate.
        the buffer block size is defined by LS_SIZE in cdstruct.h.
        A pointer to the allocated buffer control record is returned
        on success, otherwise NULL is returned.

--------------->!!!! alloc.h has been included from TC, needed temporarily
             to define some things for malloc !!!!!!

 Summary -
        #include <cdstruct.h>

         BUFCTL *cd_blkinit(nalloc)
         UCOUNT nalloc;
         {
 Description -    Allocate and initialize CD input buffr structures.

 Returns -    Pointer to buffer control block

**************************************************************************
*/

PBUFFER_CONTROL PASCAL cd_blkinit(void)                                     /*__fn__*/
{
	PBUFFER_CONTROL pRoot;        		// buffer control record pointer
	PSECTOR_BLOCK  pSectorBlockPtr;
	PSECTOR_BLOCK  pTempPtr;    		// block management array pointer
	RAW_BUF far *lpRawBufPtr;        	// buffer space pointer
	UCOUNT i;
	ULONG ulBlockOffset = 0;

	//
  // allocate the blocks and controls. if not successful, return null.
	//
	pRoot           = CDMemBufCtrl(NULL);
	if (!pRoot)
		return(NULL);

	pSectorBlockPtr = CDMemSectorBlock(NULL);
	if (!pSectorBlockPtr)
		return(NULL);

	lpRawBufPtr     = CDMemRawBuffers(NULL);
	if (!lpRawBufPtr)
		return(NULL);

	// if we are using XMS for the buffers, set up the pointer
	// to the local buffer that was allocated at load time
	if (bUseExtended)
		lpLocalBuffer = lpRawBufPtr;
	else
		lpLocalBuffer = NULL;


	pRoot->nlsbuf          = numBuffers;    	// init buffer control record
	pRoot->pBlockHead      = pSectorBlockPtr;
	pRoot->pSectorListHead = NULL;

	pSectorBlockPtr->pPrevBlock = NULL;			// mark as beginning of the list

	for (i=0; i < numBuffers; i++) 					// init management array
	{
		// initialize the drive letter
		pSectorBlockPtr->iDrive         = -1;

		// initialize the sector number
		pSectorBlockPtr->ulSectorNumber = 0;

		// get a pointer to the current block and move pointer to next block
		pTempPtr                        = pSectorBlockPtr++;

		// link to next block
		pTempPtr->pNextBlock            = pSectorBlockPtr;

		// if using conventional memory, store the far pointer to block
		if (!bUseExtended)
		{
			if ((unsigned)lpRawBufPtr >= 0xF000)
				lpRawBufPtr = (RAW_BUF far *)Normalize(lpRawBufPtr);

				// make sure that we got a valid normalized address
				AssertLP(lpRawBufPtr);


			// link in block
			pTempPtr->blockType.lpBlock     = lpRawBufPtr++;
		}

		// if using XMS memory, store the offset of this sector block
		else
		{
			pTempPtr->blockType.ulXMSBlockOffset = ulBlockOffset;
			ulBlockOffset += LS_SIZE;
		}

		// no initial sector number list
		pTempPtr->pNextSector           = NULL;

		// link next block to this block
		pSectorBlockPtr->pPrevBlock     = pTempPtr;
	}

	// at end of list -- set end ptrs to null
	pTempPtr->pNextBlock  = NULL;
	pTempPtr->pNextSector = NULL;

	// set tail ptr
	pRoot->pBlockTail = pTempPtr;

	// return root structure
	return(pRoot);
}

//
// GetSector will read one sector of data from the cdrom if needed, or it will
// return a pointer to the specified data in the cache.
//
// NOTICE - at this time, we only support 1 sector caching.  If multiple
// sectors are needed, they are read directly to user space.

RAW_BUF  far * PASCAL GetSectorsFromCache(PBUFFER_CONTROL pRoot, COUNT iDrive,
									               ULONG ulSectorNumber, INT iNumSectors)
{
	PSECTOR_BLOCK pSectorData;
//	PSECTOR_BLOCK pTemp;

	// validate parameters
	ASSERT(pRoot);
//	ASSERT(iDrive >= 0 && iDrive < MAXDRIVES);
	ASSERT(iDrive >= 0 && iDrive < numDrives);
	ASSERT(ulSectorNumber > 0);
	ASSERT(iNumSectors == 1);

	// if the CDROM driver indicates that the disk has changed, purge
	// all cached sectors for this drive from the cache
	if (CDChanged(iDrive) == YES)
		ResetCache(pRoot, iDrive);

	// get a pointer to the sector if it is in the cache
	pSectorData = FindSectorInCache(pRoot, iDrive, ulSectorNumber, iNumSectors);

	// if it is in the cache, return a pointer to it
	if (pSectorData)
	{
		if (!bUseExtended)
			return pSectorData->blockType.lpBlock;
		else
		{
			// we have to move the data from XMS to the local buffer
			XMSRead(pSectorData->blockType.ulXMSBlockOffset, lpLocalBuffer->buf, LS_SIZE);

			// return a pointer to the local buffer
			return lpLocalBuffer;
		}
	}

	pSectorData = GetLRU(pRoot, iNumSectors);


	// read to buffer in conventional memory
	if (!bUseExtended)
	{
		// if the read failed, return NULL
		if (!CD_CDRead(iDrive, iNumSectors, ulSectorNumber, pSectorData->blockType.lpBlock->buf))
			return NULL;
	}

	// read to temporary buffer and then transfer the data to XMS
	else
	{
		// if the read failed, return NULL
		if (!CD_CDRead(iDrive, iNumSectors, ulSectorNumber, lpLocalBuffer->buf))
			return NULL;
		XMSWrite(pSectorData->blockType.ulXMSBlockOffset, lpLocalBuffer->buf, LS_SIZE);
	}

	// if multiple sector read were supported the following code should be used
	//
	//		pTemp = pSectorData;
	//		for (i = 0; i < iNumSectors; i++)
	//		{
	//			pTemp->ulSectorNumber = ulSectorNumber + i;
	//			pTemp->iDrive = iDrive;
	//			pTemp = pTemp->pNextBlock;
	//		}
	//
	// but since we only support the reading and writing of one block at a
	// time this is the code that is used:
	//
		pSectorData->ulSectorNumber = ulSectorNumber;
	  	pSectorData->iDrive         = iDrive;

	InsertBlock(pRoot, pSectorData, iNumSectors);

	if (!bUseExtended)
		return pSectorData->blockType.lpBlock;
	else
		return lpLocalBuffer;

}


//
// FindSectorInCache will attempt to find the specified sector number in
// the Sector Number list and return a pointer to corresponding Sector Block
// structure.  If not found, NULL will be returned
//
PSECTOR_BLOCK PASCAL FindSectorInCache(PBUFFER_CONTROL pRoot, COUNT iDrive,
                                          ULONG ulSectorNumber, INT iNumSectors)
{
	PSECTOR_BLOCK pSearchPtr;

	// validate parameters
	ASSERT(pRoot);
	ASSERT(iDrive >= 0 && iDrive < numDrives);
	ASSERT(ulSectorNumber > 0);
	ASSERT(iNumSectors == 1);

	// get a pointer to the head
	pSearchPtr = pRoot->pSectorListHead;

	// find the sector - if it is in the list
	while (pSearchPtr && iDrive > pSearchPtr->iDrive)
		pSearchPtr = pSearchPtr->pNextSector;

	while (pSearchPtr && (iDrive         == pSearchPtr->iDrive &&
								 ulSectorNumber !=  pSearchPtr->ulSectorNumber))
		pSearchPtr = pSearchPtr->pNextSector;

	// if we hit the end of the list or the sector was not in the list
	if (!pSearchPtr || (iDrive != pSearchPtr->iDrive ||
							  ulSectorNumber != pSearchPtr->ulSectorNumber))

		return NULL;


	// we should modify our routine to be able to find several sectors, but
	// for now the requirements are to find only 1 sector.  We have
	// found it, so return the pointer

	// move this sector block to the head of the list.
	if (pSearchPtr != pRoot->pBlockHead)
	{
		// make the previous block link to the next block
		pSearchPtr->pPrevBlock->pNextBlock = pSearchPtr->pNextBlock;

		// if it is the tail, move it to the head and set a new tail
		if (pSearchPtr == pRoot->pBlockTail)

			// move the tail to the previous sector block
			pRoot->pBlockTail = pSearchPtr->pPrevBlock;


		else

			// this block is in the middle of the list
			// link the two on either side together
			pSearchPtr->pNextBlock->pPrevBlock = pSearchPtr->pPrevBlock;



		// mark this as the head of the list
		pSearchPtr->pPrevBlock = NULL;

		// make the old head, the second one in line
		pSearchPtr->pNextBlock = pRoot->pBlockHead;
		pRoot->pBlockHead->pPrevBlock = pSearchPtr;

		// make this one the current head
		pRoot->pBlockHead = pSearchPtr;

	}

	return pSearchPtr;


}

//
// GetLRU will use the tail pointer to return a pointer to the Least Recently
// Used Sector Block structure.
//
PSECTOR_BLOCK PASCAL GetLRU(PBUFFER_CONTROL pRoot, INT iNumSectors)
{

	// validate parameters
	ASSERT(pRoot);
	ASSERT(iNumSectors == 1);

	// we should modify our routine to be able to return as many sectors
	// as requested (up to the number of buffers) but since we are only
	// accessing one sector at a time, we will return the current tail pointer

	// if the sector is currently being used, remove it from the sector list
	// before using again and mark it unused
	if (pRoot->pBlockTail->iDrive != -1)
	{
	 	pRoot->pBlockTail->iDrive = -1;
		pRoot->pBlockTail->ulSectorNumber = 0;
		SectorListRemove(pRoot, pRoot->pBlockTail);
	}

	return pRoot->pBlockTail;

}

//
// InsertBlock will insert the specified Sector Block structure into the
// list at the head position (to mark it as most recently used) and it will
// also insert the block into the appropriate location in the sector number
// list
//
VOID PASCAL InsertBlock(PBUFFER_CONTROL pRoot, PSECTOR_BLOCK pSectorData,
                               INT iNumSectors)
{
	PSECTOR_BLOCK pTemp;
//	COUNT i;

	// validate parameters
	ASSERT(pRoot);
	ASSERT(pSectorData);
	ASSERT(iNumSectors == 1);

	// save the current tail;
	pTemp = pRoot->pBlockTail;

	// set new tail;
	pRoot->pBlockTail = pSectorData->pPrevBlock;
	pRoot->pBlockTail->pNextBlock = NULL;

	// tack end of blocks onto the head of the list
	pTemp->pNextBlock = pRoot->pBlockHead;
	pRoot->pBlockHead->pPrevBlock = pTemp;

	// make first one in the blocks the new head
	pRoot->pBlockHead = pSectorData;
	pSectorData->pPrevBlock = NULL;


	// if multiple sector read were supported the following code should be used
	//
	//		// insert the sector(s) into the sector number list
	//		pTemp = pSectorData;
	//		for (i=0; i < iNumSectors; i++)
	//		{
	//			SectorListAdd(pRoot, pTemp);
	//			pTemp = pTemp->pNextBlock;
	//		}
	//
	// but since we only support the reading and writing of one block at a
	// time this is the code that is used:
	//
	SectorListAdd(pRoot, pSectorData);

}

//
// ResetCache will erase the contents of all of the SECTOR_BLOCKS that are
// owned by drive iDrive, remove them from the Sector Number list and move
// them to the most recently used position
//
VOID PASCAL ResetCache(PBUFFER_CONTROL pRoot,  COUNT iDrive)
{
	PSECTOR_BLOCK pTemp, pTempNext;

	// validate parameters
	ASSERT(pRoot);
//	ASSERT(iDrive >= 0 && iDrive < MAXDRIVES);
	ASSERT(iDrive >= 0 && iDrive < numDrives);

	// get pointer to start of list
	pTemp = pRoot->pSectorListHead;

	// while not at end of the list do
	while (pTemp)
	{
		// save pointer to next one in the list
		pTempNext = pTemp->pNextSector;

		// if this block is being used by specified drive, we need to "erase"
		// this block
		if (pTemp->iDrive == iDrive)
			// move to end of list
			MakeLRU(pRoot, pTemp);

		// do next block
		pTemp = pTempNext;
	}
}

// MakeLRU will reset the information contained in the specified sector block
// and move it to the end of the list.
//
// Resetting the information means:
//	setting iDrive to -1
//	setting ulSectorNumber to 0
//	moving the pointers so that this block is not in the sector number list
// 	moving the pointers so that this block is at the end of the list
VOID PASCAL MakeLRU(PBUFFER_CONTROL pRoot, PSECTOR_BLOCK pNewTail)
{

	// validate parameters
	ASSERT(pRoot);
	ASSERT(pNewTail);

	// clear drive valu -- make it unused
	pNewTail->iDrive = -1;

	// remove it from the Sector Number list
	SectorListRemove(pRoot, pNewTail);

	// clear sector number
	pNewTail->ulSectorNumber = 0;

	// if the sector is not already the tail
	if (pRoot->pBlockTail != pNewTail)
	{
		// move the prev ptr of the next block to point to the previous block
		pNewTail->pNextBlock->pPrevBlock = pNewTail->pPrevBlock;

		// if there was a prev block, move it's next ptr to this block's next ptr
		if (pNewTail->pPrevBlock)
			pNewTail->pPrevBlock->pNextBlock = pNewTail->pNextBlock;

		// if it is the head of the list
		if (pRoot->pBlockHead == pNewTail)
			// move the head of the list to one past the current block
			pRoot->pBlockHead = pNewTail->pNextBlock;

		// make current tail point to our new tail
		pRoot->pBlockTail->pNextBlock = pNewTail;

		// the old tail is now pNewTail's prior block
		pNewTail->pPrevBlock = pRoot->pBlockTail;

		// no block follows the new tail
		pNewTail->pNextBlock = NULL;

		// ht new block list tail is pNewTail
		pRoot->pBlockTail = pNewTail;
	}

}

//
// SectorListRemove will add the specified sector to the list of
// sectors being cached.
//
VOID PASCAL SectorListAdd(PBUFFER_CONTROL pRoot, PSECTOR_BLOCK pSector)
{
	// defines for the insertion point variable
	#define BEFORE -1
	#define NO_INSERT 0
	#define AFTER 1

	PSECTOR_BLOCK pTemp;
	PSECTOR_BLOCK pPrev;
	PSECTOR_BLOCK pNext;
	INT           insertPoint;		// contains a specific value as to where
											// the new record is to be inserted in
											// relation to the comparison record pointed
											// to by pTemp
											//
											//   -1  insert before pTemp
											//    0  do not insert
											//    1  insert after pTemp
											//
											// This should never be set to 0


	// validate parameters
	ASSERT(pRoot);
	ASSERT(pSector);

	// initialize list pointers
	pPrev = NULL;
	pNext = NULL;
	pTemp = pRoot->pSectorListHead;

	// if the list is empty, store the first one as the head and exit
	if (!pTemp)
	{
		// insert at head
		pRoot->pSectorListHead = pSector;
		pSector->pNextSector = NULL;

	}

	// the list is not empty
	else
	{
		pNext = pTemp->pNextSector;

		// search through this drive's cached sectors for the insertion point
		while (pNext && (pSector->iDrive         >  pTemp->iDrive ||
							 (pSector->iDrive         == pTemp->iDrive &&
							  pSector->ulSectorNumber >  pTemp->ulSectorNumber)))
		{
			pPrev = pTemp;
			pTemp = pNext;
			pNext = pNext->pNextSector;
		}

		// where do we insert?
		if (pSector->iDrive == pTemp->iDrive)
		{
			if (pSector->ulSectorNumber == pTemp->ulSectorNumber)

				// this should never occur, but we need to test for it anyhow
				insertPoint = NO_INSERT;


			else if (pSector->ulSectorNumber > pTemp->ulSectorNumber)

				// insert after the current record
				insertPoint = AFTER;


			else

				// insert before the current record
				insertPoint = BEFORE;
		}

		else if (pSector->iDrive > pTemp->iDrive)
			// new drive is greater than current, insert after current record
			insertPoint = AFTER;

		else
			// new drive is less than current, insert before current record
			insertPoint = BEFORE;


		if (insertPoint == AFTER)
		{
			// insert between pTemp and pNext
			pSector->pNextSector = pNext;
			pTemp->pNextSector = pSector;
		}

		else if (insertPoint == BEFORE)
		{
			pSector->pNextSector = pTemp;

			// insert between Head and pTemp
			if (pTemp == pRoot->pSectorListHead)
				pRoot->pSectorListHead = pSector;

			// insert between pPrev and pTemp
			else
			{
				// make sure that pPrev has a real value
				ASSERT(pPrev);

				pPrev->pNextSector = pSector;
			}
		}

		// this should be at the end of the else
		// make sure that the record was not already in the list
		ASSERT(insertPoint != NO_INSERT);
	}


}


//
// SectorListRemove will remove the specified sector from the list of
// sectors being cached.
//
VOID PASCAL SectorListRemove(PBUFFER_CONTROL pRoot, PSECTOR_BLOCK pSector)
{
	PSECTOR_BLOCK pPrev;

	// validate parameters
	ASSERT(pRoot);
	ASSERT(pSector);


	// is it the head of the Sector Number List?
	if (pRoot->pSectorListHead == pSector)

		// make head the sector following pNewTail;
		pRoot->pSectorListHead = pSector->pNextSector;


	// it is not the head, but it must be in the list
	else
	{
		// get a pointer to the head of the list
		pPrev = pRoot->pSectorListHead;

		// while we do not find the block
		while (pPrev && pPrev->pNextSector != pSector)
			pPrev = pPrev->pNextSector;

		// make sure the data is valid
		ASSERT(pPrev);

		// move link to one past the one we want to delete
		pPrev->pNextSector = pSector->pNextSector;

	}
	// remove pointer to next sector block
	pSector->pNextSector = NULL;

}

//
// Normalize will take the specified far pointer and massage it
// so that it can access all of the data that it has allocated
// to it.
//
void far * PASCAL Normalize(void far * lpToNormalize)
{
	ULONG ulNewValue;

	// validate parameters
	ASSERT(lpToNormalize);

	// get the segment and move it down 12 bit positions
	ulNewValue = ((ULONG)lpToNormalize & 0XFFFF0000L) >> 12;

	// add in the offset
	ulNewValue += (unsigned)lpToNormalize;

	// if the normalized address is TOO big, return null
	if (ulNewValue > 0xFFFFFL)
		return NULL;

	ulNewValue = (((ulNewValue & 0xFFFF0L) >> 4) << 16) | (ulNewValue & 0x0000FL);
	return (void far *)ulNewValue;
}

