//
//   $Id: CDDEV.C 1.5 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  CDDEV.C  	created mm/dd/yy  by Roy Sherrill
 *				updated 06/24/96	by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 * CDDev module provides the interface with the actual CD-ROM Device
 * Driver.  It include routines to Initialize the Driver, Read, and
 * perform status checks.
 *
 * This module is ported from the CDFS PORTME.C module.  The routine
 * cd_device_init has been added.
 *
 * revision history
 *
 * 11/19/96 jmb	Added new logic to control when we call the
 *					cdrom status routine
 * 12/04/96 jmb	Created CDChanged routine.  Removed CDStatus
 *					routine.
 * 08/27/97 jmb	Added debugging code to CDDeviceInit to
 *					verify that we were indeed supporting
 *					multiple cdroms with one driver.
 *********************************************************/

#define __FILENUM__ 6

#include <io.h>
#include <dos.h>
#include <string.h>
#include <stdio.h>
//#include <stdlib.h>

#include "datalght.h"
bool PASCAL dlOpenFile( char * szName, uchar ucOpenMode, int * piHandle );
bool PASCAL dlCloseFile(int handle);

#include "cdrom.h"
#include "cddos.h"
#include "cdstruct.h"

//
// prototype include files
//
#include "cdutil.pt"
#include "cddev.pt"
#include "util.pt"
#include "cdmemory.pt"


#define DOOR_OPEN 0x01
#define DISK_IN_DRIVE 0x08

#define MAXDELAY 36l										// Must check device at least
																	// every 2 seconds (36 ticks)

#if DEBUG
long cdStatCount;
long cdReadCount;
#endif


extern CDDRIVE *drive_core;
extern int numDrives;

struct REQHDR reqHdr;

unsigned long lastTime = 0;
unsigned long far *lpBiosTime = MK_FP(0x40, 0x6C);// Ptr to BIOS Timer Tick count!

//void far *lpDriverStrategy;								// Pointer to Device Driver
//																	// Strategy Routine
//
//void far *lpDriverInterrupt;								// Pointer to Device Driver
//																	// Interrupt Routine

struct CD_ROM_DEVHDR far *lpCDROMDevice;				// Pointer to CD-ROM Device Driver



//
// Procedure to call DOS to handle input/output control for devices.
//
unsigned PASCAL DeviceIOCTL(int handle, char func, unsigned argCX, void *argDX)
{
	unsigned retValue;

	asm	mov	ax, 4400H
	asm	mov	al, func
	asm	mov	bx, handle
	asm	mov	cx, argCX
	asm	mov	dx, argDX
	asm	int	21H
	asm	mov	retValue, ax
	return retValue;
}

//
// call the device driver strategy routine and the interrupt routine
//
void PASCAL CallCDRomDriver(void *pRequestHeader, void far *lpDriverStrategy,
											void far *lpDriverInterrupt)
{
	asm	push	ds
	asm	pop	es
	asm	mov	bx, pRequestHeader
	asm	call dword ptr [lpDriverStrategy]
	asm	call dword ptr [lpDriverInterrupt]
	CheckStackOverFlow();
}


// CDChanged will return status as to whether or not the CDROM in the CDDrive
// has been changed.
//
//    RETURNS:
//		NO		if the disk has definitely not been changed
//		YES	if the disk might have been changed.
//

BOOL PASCAL CDChanged(COUNT drive)
{
	static int diskChanged = YES;		// Init to Drive Unavailable
	struct REQHDR *rh = &reqHdr;
	char szMediaChange[2];

	// have we checked twice within last two seconds?  Return if we have
	if ((*lpBiosTime - lastTime) < MAXDELAY && diskChanged != YES)
		return diskChanged;

	lastTime = *lpBiosTime;

	//
	// use the IOCTL Input request 9 -- Media Changed
	// This call will tell us the following:
	//
	//		szMediaChange[1] will be set to
	//
	//				 1		Media not changed
	//				 0		Don't know if changed
	//				-1		Media changed
	//
	szMediaChange[0] = 9;
	szMediaChange[1] = 0xff;

	reqHdr.length = 0x1a;
	reqHdr.unit = drive_core[drive].unitNumber;
	reqHdr.cmd = 3;
	reqHdr.stat = 0;
	reqHdr.buf = szMediaChange;

	CallCDRomDriver(rh, drive_core[drive].lpDriverStrategy,
							drive_core[drive].lpDriverInterrupt);

	diskChanged = YES;
	if(szMediaChange[1] == 1 )
	{
		diskChanged = NO;
	}

	return (diskChanged);
}


BOOL PASCAL CDMountable(COUNT drive)                                        /*__fn__*/
{
	struct REQHDR *rh = &reqHdr;
	char szMediaChange[5];

	//
	// use IOCTL Input request 6 -- Device Status to get the status
	// of the drive, i.e., is the door closed and a disk ready?
	//
	szMediaChange[0] = 6;

	reqHdr.length = 0x1a;
	reqHdr.unit = drive_core[drive].unitNumber;
	reqHdr.cmd = 3;
	reqHdr.stat = 0;
	reqHdr.buf = szMediaChange;

	CallCDRomDriver(rh, drive_core[drive].lpDriverStrategy,
							drive_core[drive].lpDriverInterrupt);

	//
	// if the door is closed and there is a disk in the drive, tell the
	// user we can mount the drive, otherwise tell the user we cannot mount
	// the drive.
	//
	if(!(szMediaChange[1] & DOOR_OPEN) && !(szMediaChange[2] & DISK_IN_DRIVE))
		return TRUE;

	return (NO);
}



/////////////////////////////////////////////////////////////////////////
// CD_CDREAD -    Read n logical sectors from the indicated CD-ROM device.
//        This is a non-portable function which relies on Microsoft
//        Extensions function 08HEX, 'absolute disk read 2048 byte
//        logical block size'
//        This function returns a boolean to indicate read failure
//        and does not attempt to indicate the mode of failure.
//
// Summary -
//        #include <cdstruct.h>
//        #include <dos.h>
//        BOOL cd_cdread(drive,n,sector,buf)
//        COUNT drive;         <-drive number to read from
//        int n;               <-# logical sectors to read
//        long int sector;     <-first logical sector to read
//        unsigned char *buf;  <-buffer area to read into
//
// Description -    Read n logical sectors from a CD-ROM device
//
// Returns -      NO if read unsuccessful
//        YES if read successful
//

#pragma argsused
BOOL PASCAL CD_CDRead(COUNT drive, int count, ULONG sector,  BUF_TYPE *pBuf) /*__fn__*/
{
	struct REQHDR *pLocalReqHdr = &reqHdr;
	int tryCount = 0;

#if DEBUG
	++cdReadCount;					// Debug Code
	printf("CD_CDRead %u Sectors starting at %x\n", count, sector);
#endif

	cd_memfill(&reqHdr, sizeof(reqHdr), 0);
	reqHdr.length = 0x1d;
	reqHdr.unit   = drive_core[drive].unitNumber;
	reqHdr.cmd    = 128;
	reqHdr.buf    = pBuf;
	reqHdr.secCount = count;
	reqHdr.startSec = sector;
	do
	{
		reqHdr.stat   = 0;
//
// 		asm	push	ds
//		asm	pop	es
//		asm	mov	bx, pLocalReqHdr
//		asm	call dword ptr [lpDriverStrategy]
//		asm	call dword ptr [lpDriverInterrupt]
		CallCDRomDriver(pLocalReqHdr, drive_core[drive].lpDriverStrategy,
								drive_core[drive].lpDriverInterrupt);
	}
	while(((reqHdr.stat & 0x8000) != 0)  && tryCount++ < 3);


#if DEBUG
	if((tryCount >= 3) || (reqHdr.stat & 0x8000))
		printf("Read of CD-ROM Failed! - Abort!  stat = %xH, retry count is %u\n", reqHdr.stat, tryCount);
#endif

	return !(reqHdr.stat & 0x1F);
}


/**************************************************************************
	CD_GetVolSize will return the size of the cdrom in sectors.  If there
	was an error, the value of 0 will be returned.
 **************************************************************************/
#pragma argsused
ULONG CD_GetVolSize(COUNT drive)
{
	struct REQHDR *pLocalReqHdr = &reqHdr;

	struct GET_VOL_SIZE
	{
		unsigned char command;
		ulong volumeSize;
	} getVolumeSize;

	getVolumeSize.command = 8;
	getVolumeSize.volumeSize = 0;

	cd_memfill(&reqHdr, sizeof(reqHdr), 0);
	reqHdr.length = 0x1a;
	reqHdr.cmd = 3;
	reqHdr.unit = drive_core[drive].unitNumber;
	reqHdr.stat = 0;
	reqHdr.buf = &getVolumeSize;

	CallCDRomDriver(pLocalReqHdr, drive_core[drive].lpDriverStrategy,
							drive_core[drive].lpDriverInterrupt);

	if ((reqHdr.stat & 0x8000) != 0)
		return 0;
	else
		return getVolumeSize.volumeSize;
}



/////////////////////////////////////////////////////////////////////////
// CD_DeviceInit is called once to initialize the CD-ROM Drive Device
// Driver.  The device driver is found to be available, determine
// how many drives.  The device name is
//
//int PASCAL CDDeviceInit(char *device, void far *lpDriverStrat,
//											      void far *lpDriverInt)
int PASCAL CDDeviceInit(char *device)
{
	int handle;
	bool fSuccessful;


	struct GetDeviceHeader
	{
		unsigned char Cmd;
		void far *HdrPtr;
	} GDH;

//	// make sure the pointers are not null
//	ASSERT(lpDriverStrat == NULL);
//	ASSERT(lpDriverInt == NULL);

	fSuccessful = dlOpenFile( device, 0, &handle );
	if( !fSuccessful )
	{
		return 0;
	}

	GDH.Cmd = 0;						// Read the Address of Device Driver
	GDH.HdrPtr = 0;
	DeviceIOCTL(handle, 2, sizeof(GDH), &GDH);

	dlCloseFile(handle);					//

	lpCDROMDevice = GDH.HdrPtr;	// Save address

	if(lpCDROMDevice == NULL)
		return 0;

//	lpDriverStrat = MK_FP(FP_SEG(lpCDROMDevice), lpCDROMDevice->dhStrat);
//	lpDriverInt   = MK_FP(FP_SEG(lpCDROMDevice),  lpCDROMDevice->dhInter);
#if DEBUG
	printf("%s supports number of drives = %u\n", device, lpCDROMDevice->dhNumUnits);
#endif
	// return the number of units this device supports
	return lpCDROMDevice->dhNumUnits;
}


/////////////////////////////////////////////////////////////////////////
//
// cd_SetDeviceLtr - Saves the First Device Ltr in the Device Driver
// header
//
//void PASCAL CDSetDeviceLtr(char driveNumber)
//{
//	// Add 1 to the Drive Number to make it 1's based
//
//	lpCDROMDevice->dhCDDriveLtr = driveNumber+1;
//}



//  Check for the stack overwriting the data segment
//
void CheckStackOverFlow( void )
{
	extern unsigned STACKCHECK[];

	ASSERT( STACKCHECK[0] == 0xffff );
	ASSERT( STACKCHECK[1] == 0x5555 );
	ASSERT( STACKCHECK[2] == 0xaaaa );
}

//  Check for the stack overwriting the data segment
//
void SetStackOverFlowCheck( void )
{
	extern unsigned STACKCHECK[];

	STACKCHECK[0] = 0xffff;
	STACKCHECK[1] = 0x5555;
	STACKCHECK[2] = 0xaaaa;
}

