//
//   $Id: CDROM.H 1.7 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  cdrom.h	created mm/dd/yy  by Roy Sherrill
 *				updated 06/14/96	by John Burnham
 *
 *  Copyright (c) 1989-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *	CD-ROM Main Header file
 *
 *  revision history
 *  11/14/96 jmb Added CDREAD_ERROR define, and new Assert
 *					and AssertLP macros.
 *********************************************************/

#define FALSE 	0
#define TRUE 	!FALSE

//  Determine whether to cache file data (from a file read).  This is to
//  prevent flushing directory entries from the internal buffers to cache
//  file data which is less frequently accessed.
//
#define CACHE_DIR_ONLY		TRUE

#define DEBUG FALSE
#define CDEX_VERSION  0x217
#define UNSUPPORTED FALSE

#define PASCAL pascal

#define OPENFILE_FILENO 0
#define FINDFIRST_FILENO 1

#define CARRY 1

#define DEFAULT_DEVICE "MSCD002"

#define CRIT_DRIVE_NOT_READY 2
#define CRIT_READ_ERR		0xB
#define CRIT_GENERAL_FALT	0xC

#define CDEX_SIGNATURE 0XBAAB
#define CDEX_FILEHANDLE_MAGIC_NUMBER 0xC0DE
#define SFT_SIGNATURE 0XFEED

#define FIRSTVTOC 0x10

#define RETURN_FROM_INT2FH	0
#define CHAIN_PRIOR_INT2FH	1

#define CDROM_READY	0
#define CDROM_NOTREADY 1

#define CDREAD_ERROR -1


#define OPENFILE 0
#define FINDFIRST 1

#define ONE_KILOBYTE 1024

//
// critical error status bits to sent to the INT 24h handler
//
#define DISKIO_ERR			0x0000
#define BLOCKDEVICE_ERR 	0x8000
#define IGNORE_FLAG			0x2000
#define RETRY_FLAG			0x1000
#define FAIL_FLAG			0x0800
#define DATA_AREA_ERROR	0x0600
#define ROOT_DIR_ERROR		0x0400
#define FAT_AREA_ERROR		0x0200
#define DOS_AREA_ERROR		0x0000
#define WRITE_ERROR			0x0100

//
// critical error return values
//
#define IGNORE_ERROR		0x00			// ignore error and continue processing
#define RETRY_OPERATION	0x01			// retry the operation
#define TERMINATE_PROGRAM	0x02			// terminate the program through INT 23h
#define FAIL_IN_PROGRESS	0x03			// allow the function to terminate with error

//
// DOS extended open action defines
//
#define FILE_NOT_EXIST_ACTION_FAIL		0x00
#define FILE_NOT_EXIST_ACTION_CREATE	0x10

#define FILE_EXIST_ACTION_FAIL 			0x00
#define FILE_EXIST_ACTION_OPEN 			0x01
#define FILE_EXIST_ACTION_TRUNCATE 	0x02


#define INVALID_FUNCTION	0x01
#define INVALID_DRIVE     15
#define NET_ACCESS_DENIED 65
#define ERR_ACCESS_DENIED 5

#define SIZE_ENAME 12
#define CDROM_INFOFILE_SIZE 38
#define FF_CDLABEL 0x81

#if DEBUG
#define ASSERT(expr) if(!(expr)) AssertFail(__FILENUM__, __LINE__)

#define Assert(expr) if (!(expr)) AssertFail(__FILENUM__, __LINE__)

#define AssertLP(expr) if (expr == NULL) AssertFail(__FILENUM__, __LINE__)

#define ABORT()      AssertFail(__FILENUM__, __LINE__)
#else
#define ASSERT(expr)

#define Assert(expr)

#define AssertLP(expr)

#define ABORT()
#endif

#define	cdrDeviceNotFound			0
#define	cdrNoDriveLetters			1
#define	cdrInstalledAsDrive		2
#define	cdrInvalidOption			3
#define	cdrInternalError			4
#define	cdrLastDriveError			5
#define	cdrCDEXAlreadyInst		6
#define	cdrReqDOS5					7
#define	cdrOnly1DriveSupported	8
#define	cdrCriticalError			9
#define	cdrDriverAlreadyInst		10
#define	cdrCDEXUnloaded			11
#define	cdrCDEXNotLoaded			12
#define	cdrInvalidMOptNum 		13
#define cdrNotYetImplemented		14
#define cdrMSCDEXLoaded			15
#define cdrNoXMSLoaded          16
#define cdrInsufficientXMS 		17
#define cdrInsufficientMEM		18

#define	cdexMagicNumber	0xADAD			// value returned from INT 2F,
														// function 150Bh.

#define	driveNotSupported	0					// tells the caller of INT 2Fh,
#define	driveSupported		1					// function 150Bh if the specified
														// drive is supported.  0 if no,
														// 1 if yes.



