//
//   $Id: CDSTRUCT.H 1.4 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
* Filename = cdstruct.h
*
*
*
*  Copyright (c) 1989-2002 Datalight, Inc.
*  All Rights Reserved.
*  Datalight Confidential and Proprietary
*
* Data structure definitions for CDROM interface
*
*
*
*
* revision history:
*
* 12/16/96 jmb	Removed unneeded fields to save memory
* 12/04/96 jmb	Added new field to DSTAT to say whether or not this
*						particular find first/next is taking place on the
*						root directory.
* 01/20/97	jmb	Added a new cdrom error value CDE_OKAY
*
***************************************************************************/

// Pseudo types to  eliminate confusion about the sizes of native types

#define ULONG unsigned long int        // 32 BIT unsigned
#define LONG long int                  // 32 BIT signed
#define UCOUNT unsigned short int      // 16 BIT unsigned
#define COUNT short int                // 16 BIT signed
#define VOID void                      // void if you have it
#define BOOL short int                 // native int
#define INT  short int                 // native int
#define UTINY unsigned char            // unsigned 8 bit
#define TINY  char                     // signed 8 bit
#define TEXT  char                     // char
#define BUF_TYPE unsigned char far	    // Buffers may be in FAR memory RLS
#define IMPORT extern
#define GLOBAL extern
#define LOCAL  static
#define FAST register

// Some other defines for frequently used constants and symbols

#define FAIL 0                         // OS exit error status
#define YES 1
#define NO 0
#define HSG 1
#define ISO9660 2
#define LS_SIZE 2048                   // Logical sector size
#define UNBUFFERED 0
#define BUFFERED 1

/* Define initial and default buffer allocations and constants. These
   affect size and performance tuning */

#define NLSBUF 4            // # logical sectors in buffer pool, 2K ea.
#define NLSPATH 2           // default # logical sectors read for path reads
#define NLSDIR 2            // default # logical sectors read for directory reads
#define NLSFILE 2           // default # logical sectors read for file reads
#define MAXDIRSIZE 2048     // In memory directory buffer in bytes
#define MAXPATHSIZE 4096    // Max path table size in bytes
#define MAXDRIVES 1         // Five drives defined, A..F
#define MAXFILES 2          // Maximum open CD-ROM files through this system
#define MAXUSERS 1          // Maximum concurrent users through this system

// Define some file access and drive mount flag bits

#define CD_EXCLUSIVE   0x01 // drive is mounted for exclusive access
#define FF_EXISTENCE   0x01 // existence of file need not be made known
#define FF_DIRECTORY   0x02 // directory record identifies a directory
#define FF_ASSOCIATED  0x04 // file is an associated file
#define FF_RECORD      0x08 // record format is defined in XAR
#define FF_PROTECTION  0x10 // protection is specified in XAR
#define FF_MULTIEXTENT 0x80 // not the final directory record for file

// Error codes, returned in cd_errno if a routine failed
#define CDE_OKAY    0		  // everything is okay
#define CDE_CHANGE  1       // drive status error (media change)
#define CDE_DRIVENO 2       // drive number invalid or not mounted
#define CDE_EINVAL  3       // not a directory
#define CDE_EMFILE  4       // file block init unsuccessful
#define CDE_ENOENT  5       // path entry not found
#define CDE_EVOL    6       // read past end of volume detected
#define CDE_INDEX   7       // illegal file index
#define CDE_INVAL   8       // invalid path format
#define CDE_IOERR   9       // read error
#define CDE_VTOC    10      // VTOC not present or not recognized

// Define char constants
#define BACKSLASH '\\'
#define COLON ':'

// Define arguments to SEEK

#define PSEEK_SET    0     // offset from beginning of file
#define PSEEK_CUR    1     // offset from current file position
#define PSEEK_END    2     // offset from end of file

// Define LITTLE_ENDIAN to specify low byte first order where needed.
#define LITTLE_ENDIAN

/* Define virtual int and long int for CD ROM data structures which
   include both byte orders with low byte first as the first structure */

typedef struct vlong
{
    #ifdef LITTLE_ENDIAN
        ULONG data;
        ULONG dummy;
    #else
        ULONG dummy;
        ULONG data;
    #endif
} VLONG;

typedef struct vint
{
    #ifdef LITTLE_ENDIAN
        UCOUNT data;
        UCOUNT dummy;
    #else
        UCOUNT dummy;
        UCOUNT data;
    #endif
} VINT;

// Define date data format as recorded on CD ROM VTOC's,etc.

// text strings are not nul terminated

typedef struct cd_date
{
    TEXT year[4];
    TEXT month[2];
    TEXT day[2];
    TEXT hour[2];
    TEXT min[2];
    TEXT sec[2];
    TEXT hnd[2];
    UTINY offset;
} CD_DATE;

// HSG format, text strings not nul terminated
typedef struct hsg_date
{
    TEXT year[4];
    TEXT month[2];
    TEXT day[2];
    TEXT hour[2];
    TEXT min[2];
    TEXT sec[2];
    TEXT hnd[2];
} HSG_DATE;

// ISO9660 format, text strings not nul terminated
typedef struct iso_date
{
    TEXT year[4];
    TEXT month[2];
    TEXT day[2];
    TEXT hour[2];
    TEXT min[2];
    TEXT sec[2];
    TEXT hnd[2];
    UTINY offset;
} ISO_DATE;

// Define date data format as recorded on CD ROM directories

typedef struct dir_date
{
    UTINY year;             // year since 1900
    UTINY month;            // 1-12
    UTINY day;              // 1-31
    UTINY hour;             // 0-23
    UTINY min;              // 0-59
    UTINY sec;              // 0-59
    UTINY offset;
} DIR_DATE;

// Define buffer formats for logical sector inputs from driver.

typedef struct raw_buf       // block buffer for CD ROM logical sector reads
{
    UTINY buf[LS_SIZE];
} RAW_BUF;

typedef struct sector_block
{
	COUNT iDrive;                          //(00h) drive that owns this sector
	ULONG ulSectorNumber;                  //(02h) sector in buffer

	// this union will allow us to handle both XMS offsets and pointers
	// to blocks for sector caching purposes.
	union
	{
		ULONG ulXMSBlockOffset;             //(06h) offset of this block in XMS memory
		RAW_BUF far *lpBlock;	            //(06h) ptr to block in conventional memory
	} blockType;

	struct sector_block *pNextBlock;			//(0Ah) pointer to next block
	struct sector_block *pPrevBlock; 		//(0Ch) pointer to previous block
	struct sector_block *pNextSector;		//(0Eh) pointer to next sequential sector

} SECTOR_BLOCK;
typedef SECTOR_BLOCK *PSECTOR_BLOCK;

typedef struct buffer_control
{
  COUNT nlsbuf;                          // # of blocks allocated to this control block
	PSECTOR_BLOCK pBlockHead;
	PSECTOR_BLOCK pBlockTail;
	PSECTOR_BLOCK pSectorListHead;
} BUFFER_CONTROL;
typedef BUFFER_CONTROL *PBUFFER_CONTROL;

//typedef struct bufmng        // block buffer management array structure
//{
//    COUNT drive;             // which drive uses this block (-1 = buffer empty)
//    ULONG lsn;               // current LSN in buffer
//    COUNT score;             // block score for allocation algorithm
//    COUNT users;             // number of users interested in this block
//    struct bufmng *next_blk; // pointer to next sequential contig blk
//    struct bufmng *next_lru; // pointer to next most recently used blk
//    RAW_BUF *block;          // pointer to the memory block this record controls
//    COUNT even;              // To keep it 4 byte alligned
//} BUFMNG;

//typedef struct bufctl        // block buffer control for logical sector reads
//{
//    COUNT nlsbuf;            // # of blocks allocated to this control block
//    BUFMNG *head;            // pointer to most recently used block control
//    BUFMNG *control;         // pointer to the first sequential block control
//    RAW_BUF *block;          // pointer to buffer blocks
//    COUNT even;              // To keep it 4 byte alligned
//} BUFCTL;

typedef struct hsg_test      // when testing for HSG volume descriptors
{
    UTINY dummy[8];
    UTINY desc_type;
    TEXT std_id[5];
} HSG_TEST;

typedef struct iso_test      // when testing for ISO9660 volume descriptors
{
    UTINY desc_type;
    TEXT std_id[5];
} ISO_TEST;

/* Define generic standard and supplementary VTOC data.This 1024 byte area
*  holds the decoded VTOC information for either volume structure standard
*  (HSG of ISO9660), and may contain either standard or supplementary VTOC
*  data. The standard id and VTOC type are defined within the buffer.
*/

typedef struct vtoc              // generic vtoc descriptor
{
    UTINY desc_type;             // 1 = standard, 2 = supplementary
    UTINY std_id;                // 0 = undefined, 1 = HSG, 2 = ISO9660
//    UTINY std_version;
//    UTINY v_flags;             // defined for supplementary descriptor
//    TEXT system_id[33];        // name with nul terminator
    TEXT volume_id[33];          // name with nul terminator
    ULONG vol_size;              // in logical blocks
//    UTINY esc_seq[32];         // defined for supplementary descriptor
//    INT vol_set_size;          // # volumes in this set (1,2,3...)
//    INT vol_set_seq;           // sequence # of this volume (1,2,3...)
    INT lb_size;                 // logical block size in bytes (must be 2**n)
    LONG path_t_size;            // longest path table size in bytes
    LONG path_lbn[4];            // lbn of path tables in defined byte seq.
    UTINY root_dir[34];          // root directory descriptor
//    TEXT volume_set_id[129];   // volume set identifier with nul terminator
//    TEXT publisher_id[129];    // publisher id with nul terminator
//    TEXT data_prep_id[129];    // data preparer with nul terminator
//    TEXT application_id[129];  // application id with nul terminator
    TEXT copy_file[38];          // copyright file name with nul terminator
    TEXT abstr_file[38];         // abstract file name with nul terminator
    TEXT biblio_file[38];        // bibliography file name with nul terminator
//    CD_DATE creation_date;     // volume creation date
//    CD_DATE modification_date; // latest modification date
//    CD_DATE expiration_date;   // volume expiration date
//    CD_DATE effective_date;    // volume effective date
//    UTINY file_version;        // file version 1
} VTOC;


typedef struct hsg_vtoc          // HSG format std and coded char vtoc descriptor
{
    VLONG dummy;                 // unused LBN
    UTINY desc_type;             // 0=boot,1=std,2=coded,3=unspecified,255=term
    TEXT std_id[5];              // "CDROM"
    UTINY std_version;
    UTINY v_flags;               // defined for coded char set descriptor
    TEXT system_id[32];          // name without nul terminator
    TEXT volume_id[32];          // name without nul terminator
    VLONG dummy1;                // reserved
    VLONG vol_size;              // in logical blocks
    UTINY esc_seq[32];           // defined for coded char set descriptor
    VINT vol_set_size;           // # volumes in this set (1,2,3...)
    VINT vol_set_seq;            // sequence # of this volume (1,2,3...)
    VINT lb_size;                // logical block size in bytes (must be 2**n)
    VLONG path_t_size;           // longest path table size in bytes
    #ifdef LITTLE_ENDIAN
        LONG path_lbn[4];        // lbn of tables with low first order
        LONG dummy2[4];
    #else
        LONG dummy2[4];
        LONG path_lbn[4];        // lbn of tables with high first order
    #endif
    UTINY root_dir[34];          // root directory descriptor
    TEXT volume_set_id[128];     // volume set id without nul terminator
    TEXT publisher_id[128];      // publisher id without nul terminator
    TEXT data_prep_id[128];      // data preparer without nul terminator
    TEXT application_id[128];    // application id without nul terminator
    TEXT copy_file[32];          // copyright file name without nul terminator
    TEXT abstr_file[32];         // abstract file name without nul terminator
    HSG_DATE creation_date;      // volume creation date
    HSG_DATE modification_date;  // latest modification date
    HSG_DATE expiration_date;    // volume expiration date
    HSG_DATE effective_date;     // volume effective date
    UTINY file_version;          // file version 1
} HSG_VTOC;

typedef struct iso_vtoc          // ISO9660 format std and sup vtoc descriptor
{
    UTINY desc_type;             // 0=boot,1=std,2=coded,3=unspecified,255=term
    TEXT std_id[5];              // "CD001"
    UTINY std_version;
    UTINY v_flags;               // defined for supplementary descriptor
    TEXT system_id[32];          // name without nul terminator
    TEXT volume_id[32];          // name without nul terminator
    VLONG dummy1;                // reserved
    VLONG vol_size;              // in logical blocks
    UTINY esc_seq[32];           // defined for coded char set descriptor
    VINT vol_set_size;           // # volumes in this set (1,2,3...)
    VINT vol_set_seq;            // sequence # of this volume (1,2,3...)
    VINT lb_size;                // logical block sizg ytes (must be 2**n)
    VLONG path_t_size;           // longest path table size in bytes
    #ifdef LITTLE_ENDIAN
        LONG path_lbn[2];        // lbn of tables with low first order
        LONG dummy2[2];
    #else
        LONG dummy2[2];
        LONG path_lbn[2];        // lbn of tables with high first order
    #endif
    UTINY root_dir[34];          // root directory descriptor
    TEXT volume_set_id[128];     // volume set id without nul terminator
    TEXT publisher_id[128];      // publisher id without nul terminator
    TEXT data_prep_id[128];      // data preparer without nul terminator
    TEXT application_id[128];    // application id without nul terminator
    TEXT copy_file[37];          // copyright file name without nul terminator
    TEXT abstr_file[37];         // abstract file name without nul terminator
    TEXT biblio_file[37];        // bibliog file name without nul terminator
    ISO_DATE creation_date;      // volume creation date
    ISO_DATE modification_date;  // latest modification date
    ISO_DATE expiration_date;    // volume expiration date
    ISO_DATE effective_date;     // volume effective date
    UTINY file_version;          // file version 1
} ISO_VTOC;

/* Define generic directory record format. Directory record formats differ
*  slightly between the ISO9660 and HSG formats. This generic structure
*  removes any differences between the standards as well as interprets
*  some of the data in the records. The generic structure converts any
*  variable length fields to nul terminated fixed length fields with
*  a valid data byte count field. These fixed length fields are sized to
*  hold the maximum length variable field defined by each standard.
*/

typedef struct dir_rec           // generic directory record
{
    UTINY XAR_len;               // length of XAR in LBN's
    ULONG loc_extent;            // LBN of file start
    ULONG data_len;              // length of file in bytes (not incl XAR)
    DIR_DATE record_time;        // date/time of file creation
    UTINY file_flags;            // file flag bits
    UTINY il_size;               // interleave size in LBN's
    UTINY il_skip;               // interleave gap in LBN's
    INT VSSN;                    // volume set sequence number of file
    UTINY len_fi;                // length of file name not incl nuls
    TEXT file_id[38];            // file name, nul terminated, w/o version
    INT file_version;            // version of file number
//    UTINY len_su;              // length of valid system area in bytes
//    UTINY su_data[222];        // system use buffer
} DIR_REC;

typedef struct hsg_rec           // hsg format directory record
{
    UTINY dir_len;               // length of this directory record
    UTINY XAR_len;               // length of XAR in LBN's
    VLONG loc_extent;            // LBN of file start
    VLONG data_len;              // length of file in bytes (not incl XAR)
    UTINY record_time[6];        // date/time of file creation
    UTINY file_flags;            // file flag bits
    UTINY dummy;
    UTINY il_size;               // interleave size in LBN's
    UTINY il_skip;               // interleave gap in LBN's
    VINT VSSN;                   // volume set sequence number of file
    UTINY len_fi;                // length of file identifier
    UTINY buffer[223];           // variable length use buffer
} HSG_REC;

typedef struct iso_rec           // iso9660 format directory record
{
    UTINY dir_len;               // length of this directory record
    UTINY XAR_len;               // length of XAR in LBN's
    VLONG loc_extent;            // LBN of file start
    VLONG data_len;              // length of file in bytes (not incl XAR)
    UTINY record_time[7];        // date/time of file creation
    UTINY file_flags;            // file flag bits
    UTINY il_size;               // interleave size in LBN's
    UTINY il_skip;               // interleave gap in LBN's
    VINT VSSN;                   // volume set sequence number of file
    UTINY len_fi;                // length of file identifier
    UTINY buffer[223];           // variable length use buffer
} ISO_REC;

/* Define generic path table record format. Path record formats differ
*  slightly between ISO9660 and HSG formats. This generic structure
*  removes any differences between the standards as well as converts any
*  variable length fields to null terminated fixed length fields. Both
*  standards define directory id as a variable length field. This field may
*  not exceed 32 bytes in length. Also, if the specified length of the
*  directory id is odd, the path record is padded by one nul byte */

typedef struct gen_path          // generic path record
{
    UCOUNT cur_rec;              // path record number of this record
    ULONG loc_extent;            // LBN of dir extent start
    UTINY XAR_len;               // length of XAR in LBN's
    UTINY len_id;                // length of dir id not including nulls
    UCOUNT parent_rec;           // path record of parent directory
    TEXT dir_id[33];             // directory id, null terminated
} GEN_PATH;


typedef struct hsg_path          // HSG format path record
{
    ULONG loc_extent;            // LBN of dir extent start
    UTINY XAR_len;               // length of XAR in LBN's
    UTINY len_id;                // length of dir id not including padding
    UCOUNT parent_rec;           // path record of parent directory
    TEXT dir_id[33];             // directory id, max length
} HSG_PATH;


typedef struct iso_path          // ISO9660 format path record
{
    UTINY len_id;                // length of dir id not including padding
    UTINY XAR_len;               // length of XAR in LBN's
    ULONG loc_extent;            // LBN of dir extent start
    UCOUNT parent_rec;           // path record of parent directory
    TEXT dir_id[33];             // directory id, null terminated
} ISO_PATH;


/* Define in memory file control data structure. This structure is defined
   and allocated for each file opened through this system, including the
   path and current directory files.
*/
struct cdfile
{
        COUNT file;                  // file index # for file
        struct cdfile *next_file;    // next open file in this list
        struct cddrive *owner_drive; // which drive record owns us
//        BUFCTL *cur_buf;             // block buffer used by this file
        PBUFFER_CONTROL cur_buf;    // block buffer used by this file
        INT nls;                     // number of logical sector read ahead
        ULONG pos;                   // file position pointer
        DIR_REC dir;                 // directory record for this file
};

/* Define in memory drive control data structure. This structure is defined
   and allocated for each unique user/drive combination. This structure
   provides necessary pointers to form lists of users per drive. Presumeably,
   a user will mount a drive with other than the default user ID when a non-
   standard VTOC is required or some alternate buffer use or buffer
   allocation is desired.
*/


struct cddrive
{
        COUNT drive;                    // drive index of this drive
        COUNT userid;                   // owner id of mount. 0=standard options
        UTINY flags;                    // mount flags for drive
        COUNT users;                    // number of users with this id mounted
        struct cddrive *next_drive;     // next mount of this drive
//        BUFCTL *cur_buf;                // block buffer used by this user
        PBUFFER_CONTROL cur_buf;       // block buffer used by this user
        VTOC *cur_vtoc;                 // pointer to vtoc buffer for this user
        struct cdfile *first_file;      // first file control block opened
                                        //   by this user
        struct cdfile path_file_core;   // Space for path table file
        struct cdfile *path_file;       // file control block for current
                                        //   path table file  points to path_file_core
        GEN_PATH cur_path;              // current path table record, in
                                        //   generic record format
        struct cdfile dir_file_core;    // Space for current directory file.
        struct cdfile *dir_file;        // file control  block for current
                                        //   directory file. Pointes to dir_file_core
        DIR_REC cur_dir;                // current directory record, in
                                        //   generic record format
        DIR_REC dflt_dir;               // default directory record, in
                                        //   generic record format

			//
			// ** added 1/30/97
			//
			// MSDOS driver specific needs
			//
			BOOL cdOpened;							// flag to say whether or not
														//      this drive is mounted
			COUNT physDriveNumber;				// physical drive number in system
			COUNT unitNumber;						// unit number of cddrive under the driver
			void far *lpCDROMDevice;			// pointer to device driver header
			void far *lpDriverStrategy;		// Pointer to Device Driver Strategy Routine
			void far *lpDriverInterrupt;		// Pointer to Device Driver Interrupt Routine
			char szDevice[9];						// device name (include null)
			char dummy[3];							// paragraph alignment

};

typedef struct cdfile CDFILE;
typedef struct cddrive CDDRIVE;
typedef CDDRIVE *PCDDRIVE;

// pointer types to the structures
typedef ISO_PATH *PISO_PATH ;
typedef GEN_PATH *PGEN_PATH ;
typedef HSG_PATH *PHSG_PATH ;
typedef ISO_REC *PISO_REC ;
typedef DIR_REC *PDIR_REC ;
typedef HSG_REC *PHSG_REC ;


// Define file handle

#define CDFD COUNT

/* Define structure for gfirst,gnext operations. This structure is a
   combination of a file handle and a directory entry */

typedef struct dstat
{
    CDFD fileno;
    DIR_REC dir;
	  BOOL rootDirectorySearch;
} DSTAT;


#ifndef __GLOBALS__
//IMPORT BUFCTL *broot;            // Block buffer control
//IMPORT CDDRIVE *drv_array[];     // max # drives supported
IMPORT PBUFFER_CONTROL broot;   // Block buffer control
IMPORT PCDDRIVE  *drv_array; // max # drives supported
IMPORT CDFILE *file_array[];     // max # files supported
IMPORT COUNT dfltdrv[];          // default drives number per user
IMPORT  COUNT cd_errno;
#endif

