//
//   $Id: LOWL.C 1.9 2002/10/23 15:09:16Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  lowl.c		created mm/dd/yy  by Roy Sherrill
 *				updated 06/14/96	by John Burnham
 *
 *
 *  Copyright (c) 1989-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *	The low level routines to access the CDROM data
 *	as files, directories, Volume Desctriptors, mount
 *	and unmount the cdrom, etc.
 *
 *  revision history
 *  12/16/96 jmb	Removed unneeded fields to save memory
 *  11/14/96 jmb	Added new routine, ReadCDData, to read
 *					data from the CDROM in smarter fashion.
 *					Replaced all calls to cd_read_an with calls
 *					to ReadCDData.
 *  12/04/96 jmb Cleaned up CD_CDMOUNT to return no if the
 *					requested VTOC is not found, rather than YES!
 *  01/20/97 jmb	Added code to check the return pointer from
 *					GetSectorsFromCache
 *  02/13/97 jmb Changed loop in get_match_dir from using
 *					the length and cmpbuf to find a matching entry
 *					to just using NameMatch to find the correct
 *					file entry in the specified directory entry.
 *  10/23/97 jmb	Fixed a bug with an earlier change that was
 *					causing the copyright field to be modified
 *					when it wasn't supposed to be modified.  This
 *					latest fix removed the code making the modification.
 *					It was null terminating a field that is
 *					no longer used, so the code that was doing
 *					the null terminating was removed.
 *********************************************************/
#define __FILENUM__ 8

#include "datalght.h"

#include "cdstruct.h"
#include <stdio.h>
#include <stdlib.h>
#include <mem.h>
#include "cdrom.h"

//
// prototype includes
//
#include "lowl.pt"
#include "cddev.pt"
#include "cdutil.pt"
#include "api.pt"
#include "block.pt"
#include "cdmemory.pt"
#include "util.pt"

//
// external data
//
extern int firstCDROMDriveNumber;
extern char firstCDROMDriveLetter;
extern unsigned openFileFlag;
extern int numBuffers;
extern int numDrives;

#if CACHE_DIR_ONLY

extern int fIsFileData;

#endif

//
// external procedures
//
extern BOOL PASCAL NameMatch(char far *lpDOSName, char far *lpCDName, int iMaxSrchlen);


/*
**************************************************************************
 CD_HSG_PATH -    Initialize a generic path record from a raw data
        block. The block is assumed to be in HSG format.

 Summary -
        #include <cdstruct.h>

        COUNT cd_hsg_path(gbuf,rbuf)
         GEN_PATH *gbuf;
         HSG_PATH *rbuf;
         {
 Description -    Initialize HSG format path record.

 Returns -    Length of the HSG path record.

**************************************************************************
*/

COUNT PASCAL cd_hsg_path( GEN_PATH *pgen,  HSG_PATH *phsg)                  /*__fn__*/
{
//	GEN_PATH *pgen;        					// generic path record pointer
//	HSG_PATH *phsg;       					// HSG format raw data buffer


	ASSERT((pgen != NULL) && (phsg != NULL));

//	pgen=(GEN_PATH*)gbuf;
//	phsg=(HSG_PATH*)rbuf;

	//
	// move data to generic buffer, one structure/field at a time
	//
	pgen->loc_extent=phsg->loc_extent;
	pgen->XAR_len=phsg->XAR_len;
	pgen->len_id=phsg->len_id;
	pgen->parent_rec=phsg->parent_rec;
	copybuff(pgen->dir_id, phsg->dir_id, pgen->len_id, TRUE);

	//
	// calc record length. if the dir id length is odd the record
	// will be padded with 1 nul
	//
	if(phsg->len_id&1)    					// odd length, add padding
	  return(pgen->len_id+9);

	else                   					// even length, no padding
	  return(pgen->len_id+8);
}

/*
**************************************************************************
 CD_ISO_PATH -    Initialize a generic path record from a raw data
        block. The block is assumed to be in ISO9660 format.

 Summary -
        #include <cdstruct.h>

        COUNT cd_iso_path(gbuf,rbuf)
         GEN_PATH *gbuf;
         ISO_PATH *rbuf;
         {
 Description -    Initialize ISO format path record.

 Returns -    Length of the ISO path record.

**************************************************************************
*/


COUNT PASCAL cd_iso_path(GEN_PATH *pgen, ISO_PATH *piso)                   /*__fn__*/
{
//	GEN_PATH *pgen;        					// generic path record pointer
//	ISO_PATH *piso;       					// ISO format raw data buffer

	ASSERT((pgen != NULL) && (piso != NULL));

//	pgen=(GEN_PATH*)gbuf;
//	piso=(ISO_PATH*)rbuf;

	//
	// move data to generic buffer, one structure/field at a time
	//
	pgen->loc_extent=piso->loc_extent;
	pgen->XAR_len=piso->XAR_len;
	pgen->len_id=piso->len_id;
	pgen->parent_rec=piso->parent_rec;
	copybuff(pgen->dir_id , piso->dir_id ,pgen->len_id, TRUE);

	//
	// calc record length. if the dir id length is odd the record
	// will be padded with 1 nul
	//
	if((piso->len_id)&1)    				// odd length, add padding
	  return(pgen->len_id+9);

	else                   					// even length, no padding
	  return(pgen->len_id+8);

}


/*
**************************************************************************
 CDDOPEN -     Attempt to open and initialize a file control data structure
        for the lowest level directory specified in the path argument,
        and return a file index number into the global drive array.
        Two open routines are included, one for the default user (0)
        and one for a specified user.
        If the routine fails a negative index is returned. The
        path specifier passed in as an argument defines the directory
        to be opened. If a drive is not specified, the default drive
        for the user. If the path does not include a root specifier,
        the path is considered relative to the default directory for
        the drive.


 Summary -
        #include <cdstruct.h>

         COUNT cd_dopen(path)
         TEXT *path; <-------path specifier
         {

        or...

         COUNT cd_u_dopen(userid,path)
         UCOUNT userid; <-------user id
         TEXT *path;   <-------path specifier
         {


 Returns -
        Index number to file control structure in global file
        array if open successful or -1.
        If open failed cd_errno will contain one of the following
            CDE_EMFILE  - if file block init unsuccessful
            CDE_DRIVENO - if drive number invalid or not mounted
            CDE_ENOENT  - if path entry not found
            CDE_INVAL   - if invalid path format
        Index number to file control structure in global file
        array if open successful

**************************************************************************
*/

/* open specified dir for specified user. */

COUNT PASCAL cd_u_dopen( COUNT userid, TEXT *path)                          /*__fn__*/
{
	TEXT buf[257];
	TEXT *curpath;
	TEXT *newpath;
	TEXT term;
	COUNT length,i;
	LONG n;
	COUNT driveno;
	CDDRIVE *drive;
	UCOUNT recno;

	ASSERT(path != NULL);					// validate parameter

  cd_errno = 0;
  // remove any leading spaces and convert path to upper case

  curpath = &buf[0];
  str2upper(curpath,path);
  curpath = cd_dblank(curpath);

	//
  // begin parsing path. If a drive is not specified, use the default
  // drive for the user. Check that the drive is valid and mounted for
  // the user.
	//
  newpath=cd_parsep(curpath,&length,&term);
  if (term == COLON)
  {
     driveno = GetInternalDriveNum(*curpath - 'A');
     curpath=newpath;
     newpath=cd_parsep(curpath,&length,&term);
  }

  else
  {
		driveno = cd_u_getdefltdrvno(userid);
		ASSERT(FALSE);
	}

  drive=cd_u_drvno2drv(driveno,userid);
	if (drive == NULL)
  {
		cd_errno = CDE_DRIVENO;
		goto ret_error;
	}

	//
  // initialize current path record to the default directory in case
  // no "\" are encountered.
	//
  drive->cur_path.loc_extent=drive->dflt_dir.loc_extent;
  drive->cur_path.XAR_len=drive->dflt_dir.XAR_len;
  drive->cur_path.len_id=drive->dflt_dir.len_fi;
  for (i=0;i<33;i++)
  	drive->cur_path.dir_id[i]=drive->dflt_dir.file_id[i];

	//
  // position path table pointer for first access if current parse
  // field is "\" terminated. If length is 0, use root. Otherwise use
  // default directory.
	//
  if (term == BACKSLASH)
  {
  	if (length == 0)
  	{
  		n = get_first_path(drive->path_file,&drive->cur_path);
  		if (n <= 0)
  		{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}
  		curpath=newpath;
  		newpath=cd_parsep(curpath,&length,&term);
  	}
  	else
  	{
  		n = get_match_path(&drive->dflt_dir,drive->path_file, &drive->cur_path);
  		if (n <= 0)
  		{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}
  	}
  }

	//
  // continue parsing until all "\" terminated path fields are
  // exhausted
	//
  while (term == BACKSLASH)
  {
  	if ((length == 1) && (curpath[0] == '.'))
  	    n = drive->cur_path.cur_rec;

  	else if ((curpath[0] == '.') && (curpath[1] == '.'))
  	{
  		n = get_abs_path(drive->cur_path.parent_rec,drive->path_file,&drive->cur_path);
  		if (n <= 0)
  		{
				cd_errno = CDE_ENOENT;
				goto ret_error;
			}
  	}

  	else
  	{
  		recno=drive->cur_path.cur_rec;
  		do
  		{
  			n = get_next_path(drive->path_file,&drive->cur_path);
  			if (n <= 0)
  			{
					cd_errno = CDE_ENOENT;
					goto ret_error;
				}
  		}
  		while ( (drive->cur_path.parent_rec != recno) ||
  		        (drive->cur_path.len_id != (UTINY) length) ||
//  		       !(cmpbuf((UTINY *)&drive->cur_path.dir_id[0],(UTINY *)curpath,length)));
  		       !(cmpbuf(&drive->cur_path.dir_id[0], curpath, length)));
  	}
  	curpath=newpath;
  	newpath=cd_parsep(curpath,&length,&term);
  }

	//
  // The only valid entry at this point is a null terminated file
  // specification. Open the directory pointed to
  // by the current path record search for the file spec.
	//
  if (term)
  {
		cd_errno = CDE_EINVAL;
		goto ret_error;
	}

  drive->dir_file->dir.XAR_len=drive->cur_path.XAR_len;
  drive->dir_file->dir.loc_extent=drive->cur_path.loc_extent;
  drive->dir_file->dir.data_len=256;
  drive->dir_file->dir.file_flags=2;
  n=get_first_dir(drive->dir_file,&drive->cur_dir);

  if(n <= 0)
  {
		cd_errno = CDE_ENOENT;
		goto ret_error;
	}

//  Don't use N_SCOPY
//  drive->dir_file->dir = drive->cur_dir;
  memcpy( &drive->dir_file->dir, &drive->cur_dir, sizeof(drive->cur_dir) );
  return(cd_fopen_cur(drive));

ret_error:
    return(-1);
}

////
//// open specified dir for default user (0).
////
//COUNT cd_dopen(TEXT *path)                                            /*__fn__*/
//{
//    return (cd_u_dopen(0,path));
//}
//

/*
**************************************************************************
 CD_HSG_DIR -    Initialize a generic directory record from a raw data
        block. The block is assumed to be in HSG format.

 Summary -
        #include <cdstruct.h>

        VOID cd_hsg_dir(dbuf,rbuf)
         DIR_REC *dbuf;
         HSG_REC *rbuf;
         {
 Description -    Initialize HSG format directory descriptor.

 Returns -    nothing

**************************************************************************
*/

VOID PASCAL cd_hsg_dir(DIR_REC *pgen, HSG_REC *phsg)                         /*__fn__*/
{
//	DIR_REC *pgen;        					// generic directory pointer
//	HSG_REC *phsg;       					// HSG format raw data buffer
	UTINY uti;

	ASSERT((pgen != NULL) && (phsg != NULL));		// validate parameters

//	pgen=(DIR_REC*)dbuf;
//	phsg=(HSG_REC*)rbuf;

	//
	// move data to generic buffer, one structure/field at a time
	//

	pgen->XAR_len=phsg->XAR_len;
	pgen->loc_extent=phsg->loc_extent.data;
	pgen->data_len=phsg->data_len.data;
	copybuff(&pgen->record_time.year, &phsg->record_time[0], 6, TRUE);
	pgen->file_flags=phsg->file_flags;
	pgen->il_size=phsg->il_size;
	pgen->il_skip=phsg->il_skip;
	pgen->VSSN=phsg->VSSN.data;


	//
	// calc system use buffer length and pointers and move to generic
	// buffer. note: if the file id length is even the file id will
	// be padded with 1 nul
	//

	if((phsg->len_fi)&1)    				// odd length
		uti=phsg->len_fi;
	else                   					// even length
		uti=(UTINY) (phsg->len_fi+1);

//	pgen->len_su=(UTINY) (phsg->dir_len-33-uti);
//	copybuff((UTINY *)pgen->su_data,&(phsg->buffer[uti]),(COUNT)pgen->len_su, FALSE);


	//
	// copy file id to generic buffer, null terminate. strip ';' char
	// and calc version number if present.
	//

//	phsg->buffer[phsg->len_fi]=0; /* assure ver # text is terminated */
	pgen->file_version=0;          /* version will be 0 if not present */
	for (uti=0;uti<phsg->len_fi;uti++)
	{
//		if(cmpbuf(&(phsg->buffer[uti]),(UTINY *)";",1))
		if(cmpbuf(&(phsg->buffer[uti]), ";", 1))
		{
			pgen->file_version=atoi((TINY *)(phsg->buffer[uti+1]));
			break;
		}
		pgen->file_id[uti]=phsg->buffer[uti];
	}
	pgen->file_id[uti]=0;    				// null terminate file name
	pgen->len_fi=uti;        				//	file name size
}

/*
**************************************************************************
 CD_ISO_DIR -    Initialize a generic directory record from a raw data
        block. The block is assumed to be in ISO format.

 Summary -
        #include <cdstruct.h>

        VOID cd_iso_dir(dbuf,rbuf)
         DIR_REC *dbuf;
         HSG_REC *rbuf;
         {
 Description -    Initialize ISO format directory descriptor.

 Returns -    nothing

**************************************************************************
*/

VOID PASCAL cd_iso_dir(DIR_REC *pgen,ISO_REC *piso)                          /*__fn__*/
{
//	DIR_REC *pgen;        					// generic directory pointer
//	ISO_REC *piso;       					// ISO format raw data buffer
	UTINY uti;


	ASSERT((pgen != NULL) && (piso != NULL));		// validate parameters

//	pgen=(DIR_REC*)dbuf;
//	piso=(ISO_REC*)rbuf;


	//
	// move data to generic buffer, one structure/field at a time
	//

	pgen->XAR_len=piso->XAR_len;
	pgen->loc_extent=piso->loc_extent.data;
	pgen->data_len=piso->data_len.data;
	copybuff(&(pgen->record_time.year),&(piso->record_time[0]),6, TRUE);
	pgen->file_flags=piso->file_flags;
	pgen->il_size=piso->il_size;
	pgen->il_skip=piso->il_skip;
	pgen->VSSN=piso->VSSN.data;


	//
	// calc system use buffer length and pointers and move to generic
	// buffer. note: if the file id length is even the file id will
	// be padded with 1 nul
	//

	if((piso->len_fi)&1)    				// odd length
		uti=piso->len_fi;

	else                   					// even length
		uti=(UTINY) (piso->len_fi+1);

//	pgen->len_su=(UTINY) (piso->dir_len-33-uti);
//	copybuff(pgen->su_data,&(piso->buffer[uti]),(COUNT)pgen->len_su, FALSE);


	//
	// copy file id to generic buffer, null terminate. strip ';' char
	// and calc version number if present.
	//

//	piso->buffer[piso->len_fi]=0; 		// assure ver # text is terminated
	pgen->file_version=0;          		// version will be 0 if not present

	for (uti=0;uti<piso->len_fi;uti++)
	{
//		if(cmpbuf(&(piso->buffer[uti]),(UTINY *)";",1))
		if(cmpbuf(&(piso->buffer[uti]), ";", 1))
		{
			pgen->file_version=atoi((TINY *)(piso->buffer[uti+1]));
			break;
		}
		pgen->file_id[uti]=piso->buffer[uti];
	}
	pgen->file_id[uti]=0;    				// null terminate file name
	pgen->len_fi=uti;        				// file name size
}


/*
**************************************************************************
 CD_MOUNT -     Attempt to mount the indicated CD-ROM device for this user.
        If the drive has already been mounted for that user, check
        the VTOC type specified. If there is no change, return
        a success code. If a VTOC change is specified, read a new
        VTOC as described below. If the drive is already mounted by
        another user that has gained exclusive drive access, return
        a fail code. Otherwise, check the drive validity and attempt
        to mount the drive for the indicated user. To mount a drive,
        a cddrive structure is allocated from free memory and
        initialized and pointers within the drive lists are
        maintained. A generic VTOC descriptor is allocated and
        read from the device. Only VTOC types 1 (standard) or
        2 (coded char set) will be read. If the specified VTOC
        cannot be found or is not type 1 or 2 the pointer to the
        VTOC buffer is returned NULL.
        Since this is the first opportunity to use the block buffer
        pool, the existence of the block buffer pool is checked and
        will be created.


 Summary -
        #include <cdstruct.h>
         BOOL cd_cdmount(drive,vtoc_type,user)
         COUNT drive;
         UTINY vtoc_type;
         COUNT user;
         {
 Description -    Allocate and initialize drive control structure for a user

 Returns -    NO if mount unsuccessful
        YES if mount successful

**************************************************************************
*/

BOOL PASCAL cd_cdmount(COUNT drive, UTINY vtoc_type, COUNT user)            /*__fn__*/
{
	CDDRIVE *temp, *prior;
	RAW_BUF far *temp_buf;
	ULONG lsn;
   ULONG ulVolSize;
	INT i;
	UTINY temp_type;


	//
	// Check if block buffer pool is initialized. If not, do so here.
	//
	if (broot == NULL)
	{
		broot=cd_blkinit();
		if (broot == NULL)
		{
			return(NO);
		}
	}

	//
	// check drive number
	//
	if ((drive < 0) || (drive >= numDrives))
	    return(NO);

	//
	// Scan the user list for this drive. Test if already open by this
	// user or open exclusive. If already open by this user, test the
	// VTOC descriptor. If changed, read a new VTOC
	//
	prior = drv_array[drive];
	temp  = drv_array[drive];
	while (temp)
	{
		if (temp->flags & CD_EXCLUSIVE)
		{
			return(NO);
		}

		if (temp->userid == user)
		{
			if (!temp->cur_vtoc)
				break;

			if (temp->cur_vtoc->desc_type == vtoc_type)
			{
				temp->users++;
				return(YES);
			}
			else
				break;
		}
		prior=temp;
		temp=temp->next_drive;
	}

	if (temp == NULL)
	{
		//
		// if we got this far, a new cddrive structure must be defined
		//
		temp = cd_mem_drive(NULL, drive, user);

		if (!temp)
		    return(NO);

		temp->drive=drive;
		temp->userid=user;
		temp->flags=0;
		temp->users=1;
		temp->next_drive=NULL;
		temp->cur_buf=broot;

		if ((vtoc_type == 1) || (vtoc_type == 2))
		    temp->cur_vtoc=cd_mem_vtoc(NULL, drive, vtoc_type);

		temp->first_file=NULL;
		temp->path_file=NULL;
		temp->dir_file=NULL;

		if (prior == NULL)
			drv_array[drive]=temp;

		else
			prior->next_drive=temp;
	}


	//
	// allocate a VTOC descriptor if necessary
	//
	if (temp->cur_vtoc == NULL)
	{
		if ((vtoc_type == 1) || (vtoc_type == 2))
		{
			temp->cur_vtoc=cd_mem_vtoc(NULL, drive, vtoc_type);
		}

		if (!(temp->cur_vtoc))
			return(YES);
	}


	//
	// if the vtoc type is not 1 or 2, return
	//
	if ((vtoc_type != 1) && (vtoc_type != 2))
	{
		temp->cur_vtoc=NULL;
		return(YES);
	}

	//
	// begin a loop looking for the specified VTOC type. Break out when
	// the vtoc is found, the vtoc terminator is found, or an undefined
	// cd standard is found. VTOC's are defined as sequential 2048 byte
	// logical sectors beginning with lsn 16 for the CD-ROM format
	// standards supported by this package. Put an artificial end
	// point on this loop just in case something runs away internally.
	// note that this loop limit is not defined by either standard and
	// may be eliminated. If the specified VTOC is found, define the
	// default and current directories to the root directory entry
	// defined in the VTOC. Also after the VTOC is found, open a
	// directory file and path file.
	//
	lsn=16;
	while (lsn <= 128)
	{
		temp_buf	= GetSectorsFromCache(temp->cur_buf,drive,lsn++, 1);

		if (temp_buf ==  NULL)
			break;

		i=cd_getstd((UTINY far *)temp_buf);

		temp_type=cd_getvtype(i,(UTINY far *)temp_buf);

		if (temp_type == vtoc_type)
		{
			if (i == HSG)
			{
				cd_hsg_vtoc(temp->cur_vtoc,(HSG_VTOC far *)temp_buf);
				cd_hsg_dir(&temp->dflt_dir,(HSG_REC *)&temp->cur_vtoc->root_dir[0]);
				cd_hsg_dir(&temp->cur_dir,(HSG_REC *)&temp->cur_vtoc->root_dir[0]);
			}


			if (i == ISO9660)
			{
				cd_iso_vtoc(temp->cur_vtoc,(ISO_VTOC far *)temp_buf);
				cd_iso_dir(&temp->dflt_dir,(ISO_REC*)&temp->cur_vtoc->root_dir[0]);
				cd_iso_dir(&temp->cur_dir,(ISO_REC*)&temp->cur_vtoc->root_dir[0]);
			}


			//
			// open the root directory as the current directory file -
			// Send down the address of the core to use. This simplifies
			// file memory management by have the drive structures manage their
			// own dir and path core
			//
			temp->dir_file=cd_finit(0,temp, &(temp->dir_file_core));
			temp->dir_file->nls=NLSDIR;

			//
			// dummy up a path file directory entry and open the path file
			// as if it were a normal file
			//
			temp->cur_dir.XAR_len=0;
			temp->cur_dir.loc_extent=temp->cur_vtoc->path_lbn[0];
			temp->cur_dir.data_len=temp->cur_vtoc->path_t_size;
			temp->cur_dir.il_size=0;
			temp->cur_dir.il_skip=0;
			temp->path_file=cd_finit(0,temp, &(temp->path_file_core));
			temp->path_file->nls=NLSPATH;

			// get the volume size so that we don't accidentally try to access
			// a sector not recorded on the CD
         //
         // However, only make the assignment if it comes back non-zero
         //
         if( ulVolSize = CD_GetVolSize(drive) )
         {
            temp->cur_vtoc->vol_size = ulVolSize;
         }

			return(YES);
		}


		if (temp_type == 255)
			break;
	}

	//
	// drop here if we broke out of the loop or completed the loop.
	// The requested VTOC was not found.
	//

	//
	// Free vtoc
	//
	cd_mem_vtoc(temp->cur_vtoc, 0, 0);
	temp->cur_vtoc=NULL;
	return(NO);
}

/*
**************************************************************************
 GETPATH -      Path file record retrieval functions. Get first path
        record, get next path record, get absolute path record,
        get path record matching directory record.
        Each function returns the length of the path record requested
        as well as an updated path record. If the path record is not
        found, a path record length of 0 is returned. If an error
        occurs during the search, the error status is returned.

 Summary -
        #include <cdstruct.h>


        LONG get_first_path(file,gen_rec)
        CDFILE *file;      <--path file descriptor
        GEN_PATH *gen_rec; <--generic path record, initialized
        {

        or...

        LONG get_next_path(file,gen_rec)
        CDFILE *file;      <--path file descriptor
        GEN_PATH *gen_rec; <--generic path record, updated
        {

        or...

        LONG get_abs_path(rec,file,gen_rec)
        LONG rec;       <--path record to fetch
        CDFILE *file;      <--path file descriptor
        GEN_PATH *gen_rec; <--generic path record, initialized
        {

        or...

        LONG get_match_path(dir,file,gen_rec)
        DIR_REC *dir;       <--directory record to match
        CDFILE *file;      <--path file descriptor
        GEN_PATH *gen_rec; <--generic path record, initialized
        {


 Returns -  -1 if error , or length of path record if read successful and
            decoded path record in argument gen_rec. If -1 is returned
            cd_errno will contain one of the following.

            CDE_IOERR   - if read error
            CDE_CHANGE  - if drive status error (media change)
            CDE_VTOC    - if VTOC not present or not recognized
            CDE_EVOL    - if read past end of volume detected

**************************************************************************
*/


/* read and decode first variable length record in the path table,
   updating the file position pointer. */

LONG PASCAL get_first_path(CDFILE *file, GEN_PATH *gen_rec)                /*__fn__*/
{
	COUNT j;
	LONG l = 41;
	LONG n,n1,i;
	UTINY buf[41];


	ASSERT((file != NULL) && (gen_rec != NULL));			// validate params

	//
	// read at least 8 bytes from path table to determine actual
	// path record length. Path records are variable length
	//
	cd_errno = 0;
	n1=0;
	n=ReadCDData(file,&buf[0],l,(ULONG)n1);
	if (n < 0)
		return(n);

	if (n < 8)
	{
		n1=ReadCDData(file,&buf[(INT)n],l-n,(ULONG)n);
		if (n1 < 0) return(n1);
		n=n+n1;
		if (n < 8) return(0L);
	}

	//
	// Decode the path record to generic format and calculate length.
	// if more data must be read, do it.
	//
	if (file->owner_drive->cur_vtoc)
	{
		if (file->owner_drive->cur_vtoc->std_id == HSG)
		{
			j=cd_hsg_path(gen_rec,(HSG_PATH*)buf);
			i=j;
			if (n < i)
			{
				n1=ReadCDData(file,&buf[(INT)n],i-n,(ULONG)n);
				if (n1 < 0)
					return(n1);

				if (n1 < (i-n))
					return(0L);

				j=cd_hsg_path(gen_rec,(HSG_PATH*)&buf[0]);
				i=j;
			}
			file->pos=i;
			gen_rec->cur_rec=1;
			return(i);
		}

		else if (file->owner_drive->cur_vtoc->std_id == ISO9660)
		{

			j=cd_iso_path(gen_rec,(ISO_PATH*)&buf[0]);
			i=j;
			if (n < i)
			{
				n1=ReadCDData(file,&buf[(INT)n],i-n,(ULONG)n);
				if (n1 < 0)
					return(n1);

				if (n1 < (i-n))
					return(0L);

				j=cd_iso_path(gen_rec,(ISO_PATH*)&buf[0]);
				i=j;
			}
			file->pos=i;
			gen_rec->cur_rec=1;
			return(i);
		}
	}
	cd_errno = CDE_VTOC;
	return(-1L);
}

//
// read and decode a variable length path record starting at the current
// file position pointer and update the pointer.
//
LONG PASCAL get_next_path( CDFILE *file, GEN_PATH *gen_rec)                  /*__fn__*/
{
	COUNT j;
	LONG l = 41;
	LONG n,n1,n2,i;
	UTINY buf[41];


	ASSERT((file != NULL) && (gen_rec != NULL));			// validate parameters

	//
	// read at least 8 bytes from path table to determine actual
	// path record length. Path records are variable length
	//

	cd_errno = 0;
	n2=file->pos;
	n=ReadCDData(file,&buf[0],l,(ULONG)n2);
	if (n < 0)
		return(n);

	if (n < 8)
	{
		n1=ReadCDData(file,&buf[(INT)n],l-n,(ULONG)(n+n2));
		if (n1 < 0)
			return(n1);

		n=n+n1;
		if (n < 8)
			return(0L);
	}

	//
	// Decode the path record to generic format and calculate length.
	// if more data must be read, do it.
	//

	if (file->owner_drive->cur_vtoc)
	{
		if (file->owner_drive->cur_vtoc->std_id == HSG)
		{
			j=cd_hsg_path(gen_rec,(HSG_PATH*)&buf[0]);
			i=j;
			if (n < i)
			{
				n1=ReadCDData(file,&buf[(INT)n],i-n,(ULONG) (n+n2));
				if (n1 < 0)
					return(n1);

				if (n1 < (i-n))
					return(0L);

				j=cd_hsg_path(gen_rec,(HSG_PATH*)&buf[0]);
				i=j;
			}
			file->pos=i+n2;
			gen_rec->cur_rec +=1;
			return(i);
		}
		if (file->owner_drive->cur_vtoc->std_id == ISO9660)
		{
			j=cd_iso_path(gen_rec,(ISO_PATH*)&buf[0]);
			i=j;
			if (n < i)
			{
				n1=ReadCDData(file,&buf[(INT)n],i-n,(ULONG) (n+n2));
				if (n1 < 0)
					return(n1);

				if (n1 < (i-n))
					return(0L);

				j=cd_iso_path(gen_rec,(ISO_PATH*)&buf[0]);
				i=j;
			}
			file->pos=i+n2;
			gen_rec->cur_rec += 1;
			return(i);

		}
	}
	cd_errno = CDE_VTOC;
	return(-1L);
}


//
// read and decode a specified variable length path record and update the
// path record position pointer.
//


LONG PASCAL get_abs_path( UCOUNT rec, CDFILE *file, GEN_PATH *gen_rec)         /*__fn__*/
{
    LONG n;

    n=get_first_path(file,gen_rec);

    while ((n > 0) && (gen_rec->cur_rec != rec))
     n=get_next_path(file,gen_rec);

    return(n);
}

//
// read and decode a variable length path record that matches the specified
// directory record and update the path record position pointer.
//

LONG PASCAL get_match_path(DIR_REC *dir, CDFILE *file, GEN_PATH *gen_rec)    /*__fn__*/
{
LONG n;

	ASSERT((dir != NULL) && (file != NULL) && (gen_rec != NULL));

	n=get_first_path(file,gen_rec);

	while ((n > 0) && !(gen_rec->loc_extent == dir->loc_extent))
		n=get_next_path(file,gen_rec);

	return(n);
}

/*
**************************************************************************
 GETDIR -       Directory file record retrieval functions. Get first
        directory record, get next directory record,get directory
        record that matches a string.
        Each function returns the length of the directory record
        requested as well as an updated directory record. If the
        directory record is not found, a record length of 0 is
        returned. If an error occurs during the search, the error
        status is returned. Directory records are variable length
        with the following restrictions. The first directory record
        in a logical sector begins at the first byte position of
        that logical sector. A directory record must end in the
        logical sector it began in. Unused bytes following the last
        directory record to the end of a logical sector will be 0.


 Summary -
        #include <cdstruct.h>


        LONG get_first_dir(file,gen_rec)
        CDFILE *file;      <--directory file descriptor
        DIR_REC *gen_rec; <--generic directory record
        {

        or...

        LONG get_next_dir(file,gen_rec)
        CDFILE *file;      <--directory file descriptor
        DIR_REC *gen_rec; <--generic directory record
        {

        or...

        LONG get_match_dir(name,length,file,gen_rec)
        TEXT *name;       <--directory entry to match
        COUNT length;      <--length of directory entry
        CDFILE *file;     <--directory file descriptor
        DIR_REC *gen_rec; <--generic directory record
        {

 Returns -
        length of directory record if read successful and decoded
        directory record in argument gen_rec or -1 and cd_errno contains:

        CDE_IOERR   - if read error
        CDE_CHANGE  - if drive status error (media change)
        CDE_VTOC    - if VTOC not present or not recognized
        CDE_EVOL    - if read past end of volume detected
        CDE_EINVAL  - if not a directory

**************************************************************************
*/


/* read and decode first variable length record in the directory file,
   updating the file position pointer. */

LONG PASCAL get_first_dir(CDFILE *file, DIR_REC *gen_rec)                    /*__fn__*/
{
	LONG l = 256;
	LONG n1 = 0;
	LONG n;
	UTINY buf[256];

	ASSERT((file != NULL) && (gen_rec != NULL));

	cd_errno = 0;
	if (!(file->dir.file_flags&FF_DIRECTORY))
	{
		cd_errno = CDE_EINVAL;
		return(-1);
	}

	if (!file->owner_drive->cur_vtoc)
	{
		cd_errno = CDE_VTOC;
		return(-1);
	}

	//
	// read 256 bytes from directory to determine actual record length.
	// Directory records are variable length less than or equal to 256
	// bytes long. Based on the returned record length, update the
	// file pointer.
	//

	n=ReadCDData(file,&buf[0],l,(ULONG)n1);
	if (n < 1)
		return(n);

	n=buf[0];
	if (n < 1)
		return(n);

	if (file->owner_drive->cur_vtoc->std_id == HSG)
	{
		cd_hsg_dir(gen_rec,(HSG_REC*)&buf[0]);
		file->pos=n;
		return(n);
	}

	if (file->owner_drive->cur_vtoc->std_id == ISO9660)
	{
		cd_iso_dir(gen_rec,(ISO_REC*)&buf[0]);
		file->pos=n;
		return(n);
	}
	cd_errno = CDE_VTOC;
	return(-1);
}

//
// read and decode a variable length directory record starting at the current
// file position pointer and update the pointer.
//
LONG PASCAL get_next_dir(CDFILE *file, DIR_REC *gen_rec)                     /*__fn__*/
{
	LONG l = 256;
	LONG n,n1;
	UTINY buf[256];

	ASSERT((file != NULL) && (gen_rec != NULL));			// validate params

	cd_errno = 0;
	if (!(file->dir.file_flags&FF_DIRECTORY))
	{
	    cd_errno = CDE_EINVAL;
	    return(-1);
	}

	if (!file->owner_drive->cur_vtoc)
	{
	    cd_errno = CDE_VTOC;
	    return(-1);
	}

	//
	// read 256 bytes from directory to determine actual record length.
	// Directory records are variable length less than or equal to 256
	// bytes long. Based on the returned record length, update the
	// file pointer or read into the next logical sector.
	//

	n1=file->pos;
	n=ReadCDData(file,&buf[0],l,(ULONG)n1);

	if (n < 1)
		return(n);


	//
	// Based on the length of the record returned, read into the next
	// logical sector. The record must fit in the logical sector it
	// starts in otherwise it will begin at the first byte position of
	// the next logical sector. Also, cd_read_an will only read up to
	// the end of a logical sector.
	//

	if (buf[0] < 1)
	{
		n1 += LS_SIZE - (n1 % LS_SIZE);
		n=ReadCDData(file,&buf[0],l,(ULONG)n1);
		if (n < 1)
			return(n);
	}
	n=buf[0];
	if (n < 1)
		return(n);

	if (file->owner_drive->cur_vtoc->std_id == HSG)
	{
		cd_hsg_dir(gen_rec,(HSG_REC*)&buf[0]);
		file->pos=n + n1;
		return(n);
	}
	if (file->owner_drive->cur_vtoc->std_id == ISO9660)
	{
		cd_iso_dir(gen_rec,(ISO_REC*)&buf[0]);
		file->pos=n + n1;
		return(n);
	}
	cd_errno = CDE_VTOC;
	return(-1);
}

//
// read and decode a variable length directory record that matches the
// specified directory entry and update the directory record position
// pointer.
//
LONG PASCAL get_match_dir(TEXT *name,COUNT length,CDFILE *file,DIR_REC *gen_rec) /*__fn__*/
{
	LONG n;

	ASSERT((name != NULL) && (file != NULL) && (gen_rec != NULL));  // validate params

	n=get_first_dir(file,gen_rec);

	while (n>0)
	{
		if (cd_pathcp(gen_rec->file_id, gen_rec->len_fi, name, length ))
		{
			return(n);
		}

		else
			n=get_next_dir(file,gen_rec);
	}
	return(n);
}

/*
**************************************************************************
 CD_GETSTD -     Determine if a CD ROM logical sector conforms to either
        the HSG or ISO9660 volume descriptor format.

 Summary -
        #include <cdstruct.h>

        INT cd_getstd(buffer)
         UTINY *buffer;
         {

 Description - Test appropriate fields of a buffer to determine if format
        conforms to standard

 Returns -     NO if not recognized format
        HSG if HSG format recognized
        ISO9660 if ISO format recognized


**************************************************************************
*/


//INT PASCAL cd_getstd(UTINY *buffer)                                         /*__fn__*/
INT PASCAL cd_getstd(UTINY far *buffer)                                         /*__fn__*/
{
	HSG_TEST far *ph;
	ISO_TEST far *pi;

	ASSERT(buffer != NULL);						// validate parameters

	ph=(HSG_TEST far *)buffer;
	pi=(ISO_TEST far *)buffer;

//	if(cmpbuf((UTINY *)&(ph->std_id[0]),(UTINY *)"CDROM",5))
	if(cmpbuf(&ph->std_id[0], "CDROM", 5))
		return(HSG);

//	if(cmpbuf((UTINY *)&(pi->std_id[0]),(UTINY *)"CD001",5))
	if(cmpbuf( &pi->std_id[0], "CD001", 5))
		return(ISO9660);

	return(NO);
}

/*
**************************************************************************
 CD_GETVTYPE -    Return volume descriptor type from a known VTOC buffer
        for the given standard.

 Summary -
        #include <cdstruct.h>

        UTINY cd_getvtype(standard,buffer)
         INT standard;
         UTINY *buffer;
         {

 Description -    Return appropriate field of a known VTOC buffer as VTOC
        descriptor type.

 Returns -      VTOC descriptor type number

**************************************************************************
*/


//UTINY PASCAL cd_getvtype(INT standard, UTINY *buffer)                         /*__fn__*/
UTINY PASCAL cd_getvtype(INT standard, UTINY far *buffer)                         /*__fn__*/
{
	HSG_TEST far *ph;
	ISO_TEST far *pi;

	ASSERT(buffer != NULL);				// validate parameter

	ph=(HSG_TEST far *)buffer;
	pi=(ISO_TEST far *)buffer;

	if(standard==HSG)
		return(ph->desc_type);

	if(standard==ISO9660)
		return(pi->desc_type);

	return(0);
}


/*
**************************************************************************
 CD_HSG_VTOC -    Initialize a generic VTOC descriptor block from a raw data
        block. The block is assumed to be in HSG format and either
        type 1 (standard descriptor) or type 2 (coded char set).
        Neither field will be checked.

 Summary -
        #include <cdstruct.h>

        VOID cd_hsg_vtoc(vbuf,rbuf)
         VTOC *vbuf;
         HSG_VTOC *rbuf;
         {
 Description -    Initialize HSG format VTOC descriptor.

 Returns -    nothing

**************************************************************************
*/

VOID PASCAL cd_hsg_vtoc(VTOC *pgen, HSG_VTOC far *phsg)                          /*__fn__*/
{
	UTINY *p;           						// initializing pointer
//	VTOC *pgen;        						// generic VTOC pointer
//	HSG_VTOC *phsg;    						// HSG format raw data buffer
	COUNT i;

	ASSERT((pgen != NULL) && (phsg != NULL));		// validate parameters

//	pgen=(VTOC*)vbuf;
//	phsg=(HSG_VTOC*)rbuf;

	//
	// initialize entire generic VTOC buffer to nulls
	//
	p = (UTINY*)pgen;
	for (i=0;i<sizeof(VTOC);i++)
		*p++ = 0;

	//
	// move data to generic buffer, one structure/field at a time
	//
	pgen->desc_type=phsg->desc_type;  	// 1 = standard, 2 = coded chars
	pgen->std_id=HSG;   						// 1 = HSG, 2 = ISO9660
//	pgen->std_version=phsg->std_version;
//	pgen->v_flags=phsg->v_flags;
//	copybuff(&(pgen->system_id[0]),&(phsg->system_id[0]),32, TRUE);
	copybuff(&(pgen->volume_id[0]),&(phsg->volume_id[0]),32, TRUE);
	pgen->vol_size=phsg->vol_size.data;
//	copybuff(&(pgen->esc_seq[0]),&(phsg->esc_seq[0]),32, FALSE);
//	pgen->vol_set_size=phsg->vol_set_size.data;
//	pgen->vol_set_seq=phsg->vol_set_seq.data;
	pgen->lb_size=phsg->lb_size.data;
	pgen->path_t_size=phsg->path_t_size.data;
	pgen->path_lbn[0]=phsg->path_lbn[0];
	pgen->path_lbn[1]=phsg->path_lbn[1];
	pgen->path_lbn[2]=phsg->path_lbn[2];													// *
	pgen->path_lbn[3]=phsg->path_lbn[3];													// *
	copybuff(&(pgen->root_dir[0]),&(phsg->root_dir[0]),34, FALSE);
//	copybuff(&(pgen->volume_set_id[0]),&(phsg->volume_set_id[0]),128, TRUE);
//	copybuff(&(pgen->publisher_id[0]),&(phsg->publisher_id[0]),128, TRUE);
//	copybuff(&(pgen->data_prep_id[0]),&(phsg->data_prep_id[0]),128, TRUE);
//	copybuff(&(pgen->application_id[0]),&(phsg->application_id[0]),128, TRUE);
	copybuff(&(pgen->copy_file[0]),&(phsg->copy_file[0]),32, TRUE);
	copybuff(&(pgen->abstr_file[0]),&(phsg->abstr_file[0]),32, TRUE);
//	copybuff(&(pgen->creation_date.year[0]),&(phsg->creation_date.year[0]),
//						16, FALSE);
//	copybuff(&(pgen->modification_date.year[0]),&(phsg->modification_date.year[0]),
//						16, TRUE);
//	copybuff(&(pgen->expiration_date.year[0]), &(phsg->expiration_date.year[0]),
//						16, TRUE);
//	copybuff(&(pgen->effective_date.year[0]),&(phsg->effective_date.year[0]),
//						16, TRUE);
//	pgen->file_version=phsg->file_version;
}


/*
**************************************************************************
 CD_ISO_VTOC -    Initialize a generic VTOC descriptor block from a raw data
        block. The block is assumed to be in ISO9660 format and
        either type 1 (standard descriptor) or type 2 (coded char
        set).
        Neither field will be checked.

 Summary -
        #include <cdstruct.h>

        VOID cd_iso_vtoc(vbuf,rbuf)
         VTOC *vbuf;
         ISO_VTOC *rbuf;
         {
 Description -    Initialize ISO9660 format VTOC descriptor.

 Returns -    nothing

**************************************************************************
*/
//VOID PASCAL cd_iso_vtoc(VTOC *pgen,ISO_VTOC *piso)                           /*__fn__*/
VOID PASCAL cd_iso_vtoc(VTOC *pgen,ISO_VTOC far *piso)                           /*__fn__*/
{
	UTINY *p;          					  	// initializing pointer
//	VTOC *pgen;        						// generic VTOC pointer
//	ISO_VTOC *piso;    						// iso format raw data buffer
	COUNT i;

	ASSERT((pgen != NULL) && (piso != NULL));

//	pgen=(VTOC*)vbuf;
//	piso=(ISO_VTOC*)rbuf;

	//
	// initialize entire generic VTOC buffer to nulls
	//
	p = (UTINY*)pgen;
	for (i=0;i<sizeof(VTOC);i++) *p++ = 0;

	//
	// move data to generic buffer, one structure/field at a time
	//
	pgen->desc_type=piso->desc_type;  	// 1 = standard, 2 = coded chars
	pgen->std_id=ISO9660;   				// 1 = HSG, 2 = ISO9660
//	pgen->std_version=piso->std_version;
//	pgen->v_flags=piso->v_flags;
//	copybuff(&(pgen->system_id[0]),&(piso->system_id[0]),32, TRUE);
	copybuff(&(pgen->volume_id[0]),&(piso->volume_id[0]),32, TRUE);
	pgen->vol_size=piso->vol_size.data;
//	copybuff(&(pgen->esc_seq[0]),&(piso->esc_seq[0]),32, FALSE);
//	pgen->vol_set_size=piso->vol_set_size.data;
//	pgen->vol_set_seq=piso->vol_set_seq.data;
	pgen->lb_size=piso->lb_size.data;
	pgen->path_t_size=piso->path_t_size.data;
	pgen->path_lbn[0]=piso->path_lbn[0];
	pgen->path_lbn[1]=piso->path_lbn[1];
	copybuff(&(pgen->root_dir[0]),&(piso->root_dir[0]),34, FALSE);
//	copybuff(&(pgen->volume_set_id[0]),&(piso->volume_set_id[0]),128, TRUE);
//	copybuff(&(pgen->publisher_id[0]),&(piso->publisher_id[0]),128, TRUE);
//	copybuff(&(pgen->data_prep_id[0]),&(piso->data_prep_id[0]),128, TRUE);
//	copybuff(&(pgen->application_id[0]),&(piso->application_id[0]),128, TRUE);
	copybuff(&(pgen->copy_file[0]),&(piso->copy_file[0]),37, TRUE);
	copybuff(&(pgen->abstr_file[0]),&(piso->abstr_file[0]),37, TRUE);
	copybuff(&(pgen->biblio_file[0]),&(piso->biblio_file[0]),37, TRUE);
//	copybuff(&(pgen->creation_date.year[0]),&(piso->creation_date.year[0]),
//						16, FALSE);
//	copybuff(&(pgen->modification_date.year[0]),&(piso->modification_date.year[0]),
//						16, TRUE);
//	copybuff(&(pgen->expiration_date.year[0]),&(piso->expiration_date.year[0]),
//						16, TRUE);
//	copybuff(&(pgen->effective_date.year[0]),&(piso->effective_date.year[0]),
//						16, TRUE);
//	pgen->file_version=piso->file_version;
}

/*
**************************************************************************
 CD_DISMOUNT -  Dismount the drive for the user. If the user count is
        still greater than 0 for the user/drive combination, the
        user count is decremented. Otherwise, the drive is dismounted
        by freeing all resources owned by the drive structure.
        Any files left open, directories, path tables and records
        and the drive structure itself are freed.

 Summary -
        #include <cdstruct.h>
         VOID cd_cddismount(drive,user)
         COUNT drive;
         COUNT user;
         {
 Description -    Dellocate any drive control resources for a user

 Returns -      nothing

**************************************************************************
*/


VOID PASCAL cd_cddismount(COUNT drive, COUNT user)                           /*__fn__*/
{
//	IMPORT BUFCTL *broot;        			// block buffer and controls
	IMPORT PBUFFER_CONTROL broot;        			// block buffer and controls
//	IMPORT CDDRIVE *drv_array[];    		// drive control blocks
	IMPORT CDFILE *file_array[];    		// file pointers
	CDDRIVE *temp,*prior;
	CDFILE *nextfile,*tempfile;

	//
	// check drive number
	//
//	if ((drive<0)||(drive>=MAXDRIVES))
	if ((drive < 0) || (drive >= numDrives))
		return;

	//
	// Scan the user list for this drive. Test if already open by this
	// user. If not, return. If already open by this user, reduce the user
	// count and test. When no more users have the drive mounted,
	// deallocate resources.
	//

	prior=NULL;
	temp=drv_array[drive];
	while (temp != NULL)
	{
		if (temp->userid == user)
			break;

		prior=temp;
		temp=temp->next_drive;
	}

	if (temp == NULL)
		return;

	if(--temp->users > 0)
		return;

	if (temp->cur_vtoc != NULL)
		//
		// Free vtoc
		//
		cd_mem_vtoc(temp->cur_vtoc, 0, 0);

	nextfile = temp->first_file;
	while (nextfile != NULL)
	{
		tempfile = nextfile;
		nextfile = tempfile->next_file;
		file_array[tempfile->file] = NULL;
		cd_mem_file(tempfile, 0);
	}
	temp->path_file = NULL;
	temp->dir_file = NULL;

	if (prior == NULL)
		drv_array[drive]=temp->next_drive;

	else
		prior->next_drive=temp->next_drive;

	if (drv_array[drive] == NULL)
//		cd_blkkilld(temp->cur_buf,drive);
		ResetCache(temp->cur_buf,drive);

	//
	// Free the drive structure
	//
	cd_mem_drive(temp, 0, 0);
}

/*
**************************************************************************
 CD_FINIT -     Attempt to initialize a file control data structure and
        return a pointer to the initialized data structure. If the
        routine fails a NULL pointer is returned.

 Summary -
 		new way:
        #include <cdstruct.h>
         CDFILE *cd_finit(file,drive,direct, core)
         COUNT file;     <-------file id number
         CDDRIVE *drive; <-------drive control structure
         CDFILE *core;   <-------Core for the file to use

		old way:
        #include <cdstruct.h>
         CDFILE *cd_finit(file,drive,direct, core)
         COUNT file;     <-------file id number
         CDDRIVE *drive; <-------drive control structure
         DIR_REC direct; <-------directory entry for file
         CDFILE *core;   <-------Core for the file to use
         {

 Description -    Allocate and initialize file control structure

 Returns -    NULL- if init unsuccessful
        pointer to file control structure if init successful

**************************************************************************
*/

CDFILE * PASCAL cd_finit(COUNT file, CDDRIVE *drive, CDFILE *core)      /*__fn__*/
{
	CDFILE *temp;

	ASSERT((drive != NULL) && (core != NULL));		// validate parameters

	temp = core;

	temp->file=file;
	temp->next_file=NULL;
	temp->owner_drive=drive;
	temp->cur_buf=drive->cur_buf;
	temp->nls=NLSFILE;
	temp->pos=0;
//  Don't use N_SCOPY
//	temp->dir=drive->cur_dir;
	memcpy( &temp->dir, &drive->cur_dir, sizeof(drive->cur_dir) );
	return(temp);
}

/*
**************************************************************************
 CD_FOPEN_CUR - Attempt to open and initialize a file control data structure
        and return a file index number into the global drive array.
        If the routine fails a negative (-1) index is returned. The
        only argument passed in is a pointer to the drive control
        structure. The file is opened for the current directory entry
        in the specified drive control structure.

 Summary -
        #include <cdstruct.h>
         COUNT cd_fopen_cur(drive)
         CDDRIVE *drive; <-------drive control structure
         {

 Description -    Open file control structure described by current directory
        record in drive control structure.

 Returns -    (-1) - if init unsuccessful
        index number to file control structure in global file
        array if init successful

**************************************************************************
*/

COUNT PASCAL cd_fopen_cur(CDDRIVE *drive)          /*__fn__*/
{
	IMPORT CDFILE *file_array[];
	CDFILE *pfile_core;
	COUNT file;

	ASSERT(drive != NULL);						// validate parameters

	/*
		openFileFlag tells us to use file_array[1] if when FINDFIRST
		has opened the file and to use file_array[0] when the file is
		being opened normally.
	*/
	file = OPENFILE_FILENO;
	if (openFileFlag == FINDFIRST)
		file = FINDFIRST_FILENO;

	/*
		clear the entry and then reallocate it.
	*/
	pfile_core = cd_mem_file(NULL, file);
	file_array[file]=cd_finit(file,drive, pfile_core);

	/*
		install it in the list
	*/
	file_array[file]->next_file=NULL;

	/*
		return the file number
	*/
	return file;
}


/*****************************************************************************

	ReadCDData will attempt to read numBytesToXfer starting at position
	startPosition into the buffer pointed to by pTransferBuffer.

		CDFILE   *pFile 				<- the file structure specifying the file
												from which the data will be read

		BUF_TYPE *pTransferBuffer	<- the buffer to which the data will be sent

		LONG     numBytesToRead		<- the number of bytes we are going to read

		ULONG    startPosition		<- the position in the file to start reading


	Things that we need to do in this routine:
		* validate the parameters
		* check the status of the cdrom
		* determine the starting sector number
		* determine the number of sectors to read
		* determine the ending length (amount of data in last sector)
		* loop until we have read all of the data that was requested
		  or until we hit end of file

	We will return -1 to indicate an error
 *****************************************************************************/


unsigned endingOffset;					// the ending offset for the read
unsigned numSectorsToRead;				// the number of sectors to read
ULONG    startingSector;				// the first sector the file occupies
LONG     firstRelByteOffset;			// offset of first byte into first
												//     relative sector
unsigned numBytesInFirstSector;		// the number of bytes in the first
												//    logical sector to read
LONG     numBytesRead = 0;				// the actual number of bytes read
ULONG    relStartLogBlk; 				// the first relative logical block
ULONG    absStartLogBlk;				// the absolute first block
ULONG    numFileUnits; 					// number of sectors a file uses

LONG PASCAL ReadCDData(CDFILE *pFile,          BUF_TYPE *lpTransferBuffer,
                       ULONG   numBytesToRead, ULONG     startPosition)
{
//	RAW_BUF  *pReadBuffer;					// a pointer to the data read from the CD
	ULONG        endingPosition;
	RAW_BUF far *lpReadBuffer;			// a pointer to the data read from the CD
	INT    	    logBlkSize;			// the size of a logical block
	ULONG        numBlkPerSector; 	// number of logical blocks per sector
#if CACHE_DIR_ONLY
	static char cTempSectorBuffer[2048];
	static char cTempSectorDrive = 0xff;
	static unsigned long ulTempSectorNumber;
	int fTempSectorHasSector;
#endif

	// validate parameters
 	AssertLP( pFile );
	AssertLP( lpTransferBuffer );

	// make sure that we aren't going to run over the end of a segment
	ASSERT( (ULONG)  ((unsigned)lpTransferBuffer + numBytesToRead) <= 65536L);


	// validate the number of bytes to transfer parametery
	if (numBytesToRead <= 0)
		return 0;

	// default to no error
	cd_errno = 0;


	// check drive status for quick return
	if (CDChanged(pFile->owner_drive->drive) == YES)
	{
		cd_errno = CDE_CHANGE;
		return CDREAD_ERROR;
	}


	// check if already past end of file for a quick return
	if (startPosition >= pFile->dir.data_len)
		return(0L);


	// check the VTOC. Only recognize type 1 (standard)
	if ( ( pFile->owner_drive->cur_vtoc            == NULL ) ||
	     ( pFile->owner_drive->cur_vtoc->desc_type != 1 ) )
	{
		cd_errno = CDE_VTOC;
		return CDREAD_ERROR;
	}

	// calculate the actual number of bytes to transfer so that we don't
	// transfer more than are in the file
	endingPosition = startPosition + (numBytesToRead - 1);
	if (endingPosition >= pFile->dir.data_len)
		endingPosition = pFile->dir.data_len - 1;




	// calculate the starting sector number and the number of sectors to read
	// this code block will exit with the following variables calculated
	//
	//            startingSector
	//            numSectors
	//            firstByteOffset

	// get the size of a logical block to a local variable
	logBlkSize = pFile->owner_drive->cur_vtoc->lb_size;

	// calculate the relative logical block number to start read
	relStartLogBlk = startPosition / logBlkSize;

	// calculate the byte offset into the first relative logical block
	firstRelByteOffset = startPosition - (relStartLogBlk * logBlkSize);

	// caluclate the number of file units if recorded in interleaved mode
	if (pFile->dir.il_size > 0)
		numFileUnits = (relStartLogBlk + pFile->dir.XAR_len) / (pFile->dir.il_size);

	else
		numFileUnits = (relStartLogBlk + pFile->dir.XAR_len);

	// calculate the absolute logical block number after interleave gaps and xar
	absStartLogBlk  = relStartLogBlk ;
	absStartLogBlk += pFile->dir.XAR_len ;
	absStartLogBlk += pFile->dir.loc_extent;
	absStartLogBlk += (numFileUnits * pFile->dir.il_skip) ;

	// the absolute block number cannot exceed the volume size
	// if it does, then exit with error
	if (absStartLogBlk > pFile->owner_drive->cur_vtoc->vol_size)
	{
		cd_errno = CDE_EVOL;
		return CDREAD_ERROR;
	}

	// calculate the number of logical blocks per logical sector
	numBlkPerSector = LS_SIZE / logBlkSize;


	// calculate the starting sector
	startingSector   = absStartLogBlk / numBlkPerSector;

	// calculate the number of bytes to read after we have taken the end
	// of file into account
	numBytesToRead  = endingPosition - startPosition + 1;

	// calculate the starting offset into the first sector
	startPosition = firstRelByteOffset +
				(logBlkSize * (absStartLogBlk - (startingSector * numBlkPerSector)));


	numSectorsToRead = CalcNumSectors(startPosition, numBytesToRead, LS_SIZE);

	// figure out how many bytes are in the last sector
	endingOffset = (unsigned) ( ( startPosition  - ( startPosition % LS_SIZE ) ) +
	                          ( numBytesToRead + ( startPosition % LS_SIZE ) ) ) % LS_SIZE;

	// make sure that we got a valid number back
	ASSERT(numSectorsToRead > 0);

	// we haven't read anything yet
	numBytesRead = 0;

	// get a temporary pointer to the destination


	//
	// if the start position is not at the beginning of a sector,
	// read the first sector seperately and transfer the read
	// data to the user space.
	//
	// Things that are done here:
	//
	//	* number of sectors to read is decremented by 1
	// * destination pointer is forward the number of bytes that were just read
	// * the starting sector number is incremented by 1
	// * the total number of bytes read is increased by the amount just read
	// * the total number of bytes to read is decreased by the amount just read
	//
	if (startPosition > 0)
	{
		// get the number of bytes we want to read in the first sector
		numBytesInFirstSector = LS_SIZE - (unsigned)startPosition;

		// we do not want to overrun the space available
		if (numBytesInFirstSector > numBytesToRead)
			numBytesInFirstSector = (unsigned)numBytesToRead;


#if !CACHE_DIR_ONLY

		// we need to read the first sector, subtract one from the number
		// of sectors to read.  We will then read full sectors until we
		// need to read the last one and then read the last sector to
		lpReadBuffer = GetSectorsFromCache(pFile->cur_buf, pFile->owner_drive->drive, startingSector, 1);

		// if the read from the cache failed, return read failed
		if (lpReadBuffer == NULL)
		{
			cd_errno = CDE_IOERR;
			return CDREAD_ERROR;
		}

		// copy first sector data to user space
		_fmemcpy(lpTransferBuffer, &lpReadBuffer->buf[(unsigned)startPosition], numBytesInFirstSector);

#else

		if( !fIsFileData )
		{
			// we need to read the first sector, subtract one from the number
			// of sectors to read.  We will then read full sectors until we
			// need to read the last one and then read the last sector to
			lpReadBuffer = GetSectorsFromCache(pFile->cur_buf, pFile->owner_drive->drive, startingSector, 1);

			// if the read from the cache failed, return read failed
			if (lpReadBuffer == NULL)
			{
				cd_errno = CDE_IOERR;
				return CDREAD_ERROR;
			}

			// copy first sector data to user space
			_fmemcpy(lpTransferBuffer, &lpReadBuffer->buf[(unsigned)startPosition], numBytesInFirstSector);
		}
		else
		{
			fTempSectorHasSector = FALSE;
			if( pFile->owner_drive->drive == cTempSectorDrive )
			{
				if( startingSector == ulTempSectorNumber )
				{
					fTempSectorHasSector = TRUE;
				}
			}

			if( !fTempSectorHasSector )
			{
				if (!CD_CDRead(pFile->owner_drive->drive, 1,
								startingSector, cTempSectorBuffer) )
				{
					   cd_errno = CDE_IOERR;
				   	return CDREAD_ERROR;
				}
				cTempSectorDrive = pFile->owner_drive->drive;
				ulTempSectorNumber = startingSector;
			}


			// copy first sector data to user space
			_fmemcpy(lpTransferBuffer, &cTempSectorBuffer[(unsigned)startPosition], numBytesInFirstSector);
		}
#endif

		// we read one sector of the total count
		numSectorsToRead--;

		// move destination pointer to next available space
		lpTransferBuffer += numBytesInFirstSector;

		// move the sector counter to the next sector
		startingSector++;

		// update the byte counters
		numBytesRead += numBytesInFirstSector;

		// update the total count to read
		numBytesToRead -= numBytesInFirstSector;

		// make sure that the starting position is zero
		startPosition = 0;

	}


	// make sure that the start position is indeed at a sector boundary
	ASSERT(startPosition == 0);

	// if there are any whole sectors to read, do them in this if loop.
	// since we will be bypassing the buffering code and reading directly to
	// user space, we can potentially read all of the sectors requested in
	// one read
	if (numSectorsToRead > 1)
	{

		// this is the number of bytes that were just read to user space
		unsigned numBytesToCopy = numSectorsToRead * LS_SIZE;


		// if all we are reading is whole sectors - read all of them at once
		// to user space
		if (endingOffset == 0)
		{

			// read all of the sectors ofdata to the user space
			if (!CD_CDRead(pFile->owner_drive->drive, numSectorsToRead,
											startingSector, lpTransferBuffer) )
			{
			    cd_errno = CDE_IOERR;
		   	 return CDREAD_ERROR;
			}

			// increment the number of bytes read
			numBytesRead += numBytesToCopy;

			// decrement the number of bytes left to copy
			numBytesToRead -= numBytesToCopy;

			// increment the sector counter
			startingSector += numSectorsToRead;

			// decrement sector count
			numSectorsToRead -= numSectorsToRead;

			// increment the destination pointer
			lpTransferBuffer += numBytesToCopy;
		}

		// the last sector is not a complete read.  Read all but that one.
		else
		{

			// read numSectorsToRead-1 of data to the user space
			numBytesToCopy = (numSectorsToRead - 1) * LS_SIZE;

			// read numSectors-1 of data,  Return an error if needed
			if (!CD_CDRead(pFile->owner_drive->drive, numSectorsToRead-1,
							startingSector, lpTransferBuffer) )
			{
			    cd_errno = CDE_IOERR;
		   	 return CDREAD_ERROR;
			}

			// increment the number of bytes read
			numBytesRead     +=   numBytesToCopy;

			// decrement the number of bytes left to copy
			numBytesToRead   -=   numBytesToCopy;

			// increment the sector counter
			startingSector   += ( numSectorsToRead - 1 );

			// decrement sector count
			numSectorsToRead -= ( numSectorsToRead - 1 );

			// increment the destination pointer
			lpTransferBuffer     +=   numBytesToCopy;

		}

	}

	if (numSectorsToRead == 1)
	{
		// there will always be a partial sector of data to read at this
		// point.  We cannot transfer it directly to the user space as
		// the read from the CDROM will only copy 1 entire sector at a
		// time.  So we need to buffer the call and copy the needed portion
		// to user space

		// make sure that there really is data left to read
		ASSERT(numBytesToRead > 0);


#if !CACHE_DIR_ONLY

		// read 1 sector of data to the buffer area and transfer
		// endingPosition+1 bytes to the user space
		lpReadBuffer = GetSectorsFromCache(pFile->cur_buf, pFile->owner_drive->drive, startingSector, 1);

		// if the read from the cache failed, return read failed
		if (lpReadBuffer == NULL)
		{
			cd_errno = CDE_IOERR;
			return CDREAD_ERROR;
		}


		// copy last sector data to user space
		_fmemcpy(lpTransferBuffer, lpReadBuffer->buf, (unsigned)numBytesToRead);

#else

		if( !fIsFileData )
		{
			// read 1 sector of data to the buffer area and transfer
			// endingPosition+1 bytes to the user space
			lpReadBuffer = GetSectorsFromCache(pFile->cur_buf, pFile->owner_drive->drive, startingSector, 1);

			// if the read from the cache failed, return read failed
			if (lpReadBuffer == NULL)
			{
				cd_errno = CDE_IOERR;
				return CDREAD_ERROR;
			}


			// copy last sector data to user space
			_fmemcpy(lpTransferBuffer, lpReadBuffer->buf, (unsigned)numBytesToRead);
		}
		else
		{
			fTempSectorHasSector = FALSE;
			if( pFile->owner_drive->drive == cTempSectorDrive )
			{
				if( startingSector == ulTempSectorNumber )
				{
					fTempSectorHasSector = TRUE;
				}
			}

			if( !fTempSectorHasSector )
			{
				if (!CD_CDRead(pFile->owner_drive->drive, 1,
								startingSector, cTempSectorBuffer) )
				{
					   cd_errno = CDE_IOERR;
				   	return CDREAD_ERROR;
				}
				cTempSectorDrive = pFile->owner_drive->drive;
				ulTempSectorNumber = startingSector;
			}


			// copy last sector data to user space
			_fmemcpy(lpTransferBuffer, cTempSectorBuffer, (unsigned)numBytesToRead);
		}

#endif

		// update the byte counters
		numBytesRead += numBytesToRead;

		// decrement the number of bytes left to copy
		numBytesToRead -= numBytesToRead;

		// move the sector counter to the next sector
		startingSector++;

		// we read one sector of the total count
		numSectorsToRead--;

		// move destination pointer to next available space
		lpTransferBuffer += (unsigned)numBytesToRead;
	}

	// make sure that some of our values really are supposed to be zero
	ASSERT(numSectorsToRead == 0);

	ASSERT(numBytesToRead == 0);

	return numBytesRead;


}



/*****************************************************************************
	CalcNumSectors will calculate the number of sectors required to read from
	the CDROM using the specifed starting position, width of read and
	sectorWidth.

	It will return the number of sectors
 *****************************************************************************/
unsigned PASCAL CalcNumSectors(ULONG startingPos, ULONG width, ULONG sectorWidth)
{
	unsigned numSectors;

	// validate width parameters
	ASSERT(width > 0);
	ASSERT(sectorWidth > 0);



	// if the starting boundary is not an even boundary, move back to
	// previous boundary and include the bytes previous to the starting
	// position in the width
	width += (startingPos % sectorWidth);
	startingPos -= (startingPos % sectorWidth);


	// get the number of sectors
	numSectors = (unsigned)(width / sectorWidth);

	// if the ending location is not an even sector boundary, then increment
	// the sector count to include these leftover bytes.
	if ((startingPos + width) % sectorWidth != 0)
		numSectors++;

	// return the number of sectors to read
	return numSectors;
}
