//
//   $Id: UTIL.C 1.6 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  Copyright (c) 1989-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *********************************************************/
#define __FILENUM__ 9

/*
**************************************************************************
 PARSE -    Various path parsing functions called within the
        CD-ROM file management package. See specific functions
        for details.

****************************************************************************
*/

#include "datalght.h"

#include "cdstruct.h"
#include "cdrom.h"

#include <stdio.h>

//
// prototype includes
//
#include "cdutil.pt"
/*
**************************************************************************

CD_DBLANK  -    Search through a NULL terminated string and return a pointer to
        the first non-space character.

//Summary    -     TEXT *cd_dblank(string)
        TEXT *string; <--string passed in to be de-blanked.

Returns -     Pointer to first character in string following a series of
        spaces.

**************************************************************************
*/

TEXT * PASCAL cd_dblank(TEXT *path)                             /*__fn__*/
{
	ASSERT(path != NULL);				// validate parameter

	while (*path)
	{
		if (*path != ' ')
			break;

		path++;
	}
	return(path);
}

/*
**************************************************************************

CD_PARSEP  -    Search through a NULL terminated string and return a pointer
        to the first character following a terminator, the length of
        the terminated field, and a the terminator. The
        terminators recognized include ':','\', and '\null'.

//Summary    -     TEXT *cd_parsep(path,length,term)
        TEXT *path; <--path string passed in to be parsed
        COUNT *length; <--length of parsed string not including term
        TEXT *term; <--terminator character

Returns -     Pointer to first character in string following terminator.

**************************************************************************
*/

TEXT * PASCAL cd_parsep( TEXT *path, COUNT *length, TEXT *term)             /*__fn__*/
{
	COUNT i;


	ASSERT((path != NULL) && (length != NULL) && (term != NULL));		// validate params

	i=0;
	while ((*path != COLON) && (*path != BACKSLASH) && *path)
	{
		i++;
		path++;
	};

	*length=i;
	*term=*path++;
	return(path);
}

/*
**************************************************************************

CD_GPARSE  -     Search through a string of a defined length and return
        the length of the string of characters leading up to but not
        including the first 'term' character. Typically this would be the
        length of a file name  or file name extension.

//Summary    -     COUNT cd_gparse(path,length,term)
            TEXT *path; <--path string passed in to be parsed
            COUNT length; <--length of string to be parsed
            TEXT *term; <--terminator character to search for

Returns -     length of string leading up to but not including 'term'

**************************************************************************
*/

COUNT PASCAL cd_gparse( TEXT *path,  COUNT length,  TEXT *term)
{
	COUNT i;

	ASSERT((path != NULL) && (term != NULL));			// validate parameters

	for (i=0; i<length; i++)
		if (path[i] == *term)
			break;

	return(i);
}


/*
**************************************************************************

CD_PATCP  -     Search through a string of a defined length and each element
        of a pattern string. If all elements match according to some
        wildcard rules, return YES. Wildcard rules include:
            * matches the rest of the characters in the string
            ? matches the current character in the string

//Summary    -     BOOL cd_patcp(s,slength,pattern,plength)
            TEXT *s;       <--string passed in to be matched
            COUNT slength; <--length of string to be matched
            TEXT *pattern; <--pattern string
            COUNT plength; <--pattern length

Returns -    YES if pattern matches string
        NO  if no match

**************************************************************************
*/

BOOL PASCAL cd_patcp(TEXT *s, COUNT slength, TEXT *pattern, COUNT plength)   /*__fn__*/
{
	COUNT i;

	ASSERT((s != NULL) && (pattern != NULL));			// validate parameters

	i=slength-plength;
	while (slength && plength)
	{
	    if (*pattern == '*')
		 	return(YES);

	    if ((*pattern != *s) && (*pattern != '?'))
			return(NO);

	    s++;
	    pattern++;
		 slength--;
		 plength--;
	}


	if (i == 0)
		return(YES);

	//
	// if we ran out of filename string, but we still have search pattern
	// left, it may be a match, and it may not.
	//
	// examples:	x.*    x.?C		x.???
	//
	if (slength == 0)
	{
		while (plength)
		{
			if (*pattern == '*')
				return (YES);

			if (*pattern != '?')
				return (NO);

			pattern++;
			plength--;
		}

		//
		// if we fall out of the while loop then we found only question marks
		// in the pattern extension.  So we should match this pattern
		//
		return (YES);
	}

	return(NO);

}

/*
**************************************************************************

CD_PATHCP  -     Search through a path file string of a defined length and
        each element to a pattern string. If all elements match
        according to some wildcard rules, return YES. Wildcard rules
        include:
            * matches the rest of the characters in the string
            ? matches the current character in the string

        Both the path file and pattern strings are parsed into a
        filename and file extension segment. To match, both segments
        must match.

//Summary    -     BOOL cd_pathcp(s,slength,p,plength)
            TEXT *s;       <--string passed in to be matched
            COUNT slength; <--length of string to be matched
            TEXT *p; <--pattern string
            COUNT plength; <--pattern length

Returns -    YES if pattern matches string
        NO  if no match

**************************************************************************
*/

BOOL PASCAL cd_pathcp(TEXT *s, COUNT slength, TEXT *p, COUNT plength)         /*__fn__*/
{
COUNT sfile,pfile,xfile;
TEXT term = '.';
    sfile=cd_gparse(s,slength,&term);
    pfile=cd_gparse(p,plength,&term);

	 xfile = sfile;
	 if( pfile == 8 )
	 {
		 if( sfile > pfile )
		 {
			sfile = pfile;
		 }
	 }

	 if ((sfile && pfile) && (cd_patcp(s,sfile,p,pfile)))
	  {
			TEXT *pFileExtension, *pPatternExtension;
			COUNT fileExtLen, patternExtLen;

			sfile = xfile;

			//
			// get file extension for source string if there is one
			//
			pFileExtension = s + sfile;
			fileExtLen = slength - sfile;
			if (sfile < slength)
			{
				pFileExtension++;
				fileExtLen--;
			}

			//
			// get file extension for pattern if there is one
			//
			pPatternExtension = p + pfile;
			patternExtLen = plength - pfile;
			if (pfile < plength)
			{
				pPatternExtension++;
				patternExtLen--;
			}

        if (cd_patcp(pFileExtension, fileExtLen, pPatternExtension, patternExtLen))
			{
	        return(YES);
			}
	  }

     return(NO);
}

/*
**************************************************************************
 CD_STRINGS -    Various string handling functions called within the
        CD-ROM file management package. See specific functions
        for details.

 //Summary -
        / Compare two buffers of length n. Return true if all
          elements are equal /

        BOOL cmpbuf(string1,string2,n)
        FAST UTINY *string1;  <-pointer to first string
        FAST UTINY *string2;  <-pointer to second string
        FAST COUNT n;         <-# of elements to compare

        / Copy n elements from one buffer to another /

        VOID copybuff(to,from,n)
        FAST UTINY *to;    <-pointer to first string
        FAST UTINY *from;  <-pointer to second string
        FAST COUNT n;      <-# of elements to copy

        / Copy n elements from one buffer to another. Pad past
          the copied string with one null /

        VOID copybuffn(to,from,n)
        FAST UTINY *to;    <-pointer to first string
        FAST UTINY *from;  <-pointer to second string
        FAST COUNT n;      <-# of elements to copy

        / Copy a string and make sure the dest is upper case /

        VOID str2upper(to,from)
        FAST UTINY *to;    <-pointer to first string
        FAST UTINY *from;  <-pointer to second string

****************************************************************************
*/


/* Compare two buffers of length n. Return true if all
  elements are equal */

//BOOL PASCAL cmpbuf(void *string1, void *string2, COUNT n)        /*__fn__*/
BOOL PASCAL cmpbuf(void far *string1, void far *string2, COUNT n)        /*__fn__*/
{
	UTINY far *pString1;
	UTINY far *pString2;

	pString1 = string1;
	pString2 = string2;

	ASSERT((string1 != NULL) && (string2 != NULL));		// validate parameters

	while (n--)
		if (*pString1++ != *pString2++)
			return(NO);

	return(YES);
}


/* Copy n elements from one buffer to another */

VOID PASCAL copybuff(void far *to, void far *from,COUNT n, BOOL nullTerminate)                         /*__fn__*/
//VOID PASCAL copybuff(void *to, void *from,COUNT n, BOOL nullTerminate)                         /*__fn__*/
{
	UTINY far *pTo;
	UTINY far *pFrom;

	ASSERT((to != NULL) && (from != NULL));				// validate parameters

	pTo = to;
	pFrom = from;

	while (n--)
		*pTo++ = *pFrom++;

	if (nullTerminate)
		*pTo = 0;
}


/* Copy n elements from one buffer to another. Pad past
   the copied string with one null */

//VOID PASCAL copybuffn( void *to, void *from, COUNT n)                      /*__fn__*/
//{
//	UTINY *pTo, *pFrom;
//
//	ASSERT((to != NULL) && (from != NULL));				// validate parameters
//	pTo = to;
//	pFrom = from;
//
//	while (n--)
//		*pTo++ = *pFrom++;
//
//	*pTo=0;
//}

/* Copy a string and make sure the dest is in Upper case */

VOID PASCAL str2upper(void *to, void *from)                               /*__fn__*/
{
	TEXT c;
	TEXT *pTo, *pFrom;

	ASSERT((to != NULL) && (from != NULL));				// validate parameters

	pTo = to;
	pFrom = from;

	while(*pFrom)
	{
		c = *pFrom++;

		if  ((c >= 'a') && (c <= 'z'))
			c = (TEXT) ('A' + c - 'a');

		*pTo++ = c;
	}
	*pTo = '\0';
}

/* uppercase a character */
char PASCAL ToUpperCase(char c)
{
	if (c >= 'a' && c <= 'z')
		c = (char)(c - 0x20);
	return c;
}



