//
//   $Id: XMS.C 1.5 2002/10/15 20:34:18Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************
 *
 *  XMS.C    created 12/17/96  by John Burnham
 *
 *  Copyright (c) 1996-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 *
 *  	XMS memory allocation, deallocation, and
 *	initialization	routines.
 *********************************************************/

#define __FILENUM__ 13

#include <dos.h>

#include "datalght.h"

#include "xms.h"
#include "cdrom.h"
#include "cdstruct.h"

#include "xms.pt"
#include "cdutil.pt"


extern int bUseExtended;
extern unsigned uXMSBufferHandle;

farfuncptr lpHMAHandler;




//
// Is an XMS driver installed?
// Doesn't know whether the HMA will be used or not, so
// don't allocate that space!
//
int HimemInstalled(void)
{
	unsigned char r_al;
	unsigned uDriverSeg, uDriverOff;

	// is an XMS driver installed?
	asm mov AX, 4300H
	asm int 2FH
	asm mov r_al, AL


	// get address of HMA driver
	if(r_al == 0x80)
	{
		asm mov	AX, 4310H
		asm int	2FH
		asm mov	uDriverOff, BX
		asm mov	uDriverSeg, ES

		if(!uDriverSeg || !uDriverOff)
			r_al = 0;
		else
			lpHMAHandler = (farfuncptr)MK_FP(uDriverSeg, uDriverOff);
	}

	return (r_al == 0x80);
}


//
// Return total XMS memory in KBytes.
//
unsigned TotalXMS(void)
{
	unsigned r_ax;

	asm mov	AX, 800H

	(*lpHMAHandler)();

	asm mov	r_ax, AX

	return r_ax;
}

//
// Returns the handle of the allocated memory or zero on exit
//
unsigned AllocateXMS(unsigned sizeK)
{
	unsigned handle;
	unsigned char r_bl;

	// allocate the XMS memory
	asm mov	DX, sizeK
	asm mov	AX, 900H
	(*lpHMAHandler)();
	asm mov	handle, DX
	asm mov	r_bl, BL

	if(r_bl)
		handle = 0;
	else
	{
		// lock the memory so it can't move
		asm mov	DX, handle
		asm mov	AX, 0C00H
		(*lpHMAHandler)();
	}
	return handle;
}

//
// Returns whether or not the memory was deallocated
//
int DeallocateXMS(unsigned uHandle)
{
	unsigned char r_bl;

	ASSERT(uHandle > 0);

	// unlock the memory so we can deallocate
	asm mov	DX, uHandle
	asm mov	AX, 0D00H
	(*lpHMAHandler)();

	// deallocate the XMS memory
	asm mov	DX, uHandle
	asm mov	AX, 0A00H
	(*lpHMAHandler)();
	asm mov	r_bl, BL

	if(r_bl)
		return 0;
	else
		return 1;
}

//
// Make a copy of memory from one place to another using the
// XMS driver.
//
int XMSMove(XMS_BLOCK * pXMSBlock)
{
	auto int iWorked;
	auto int iError;

	// copy a block
	asm mov	AX, 0x0B00
	asm mov	SI, pXMSBlock
	(*lpHMAHandler)();
	asm mov	iWorked, AX
	asm xor	BH, BH
	asm mov	iError, BX

#if DEBUG
	if(!worked)
		printf("XMSMove error %04xH\n", iError);
#endif

	// enable the HMA
	asm mov	AX, 300H
	(*lpHMAHandler)();

	return iWorked;
}


unsigned long fp_to_long(void far *fp)
{
	unsigned long l;

	l = ((unsigned long)FP_SEG(fp));
	l <<= 4;
	l += FP_OFF(fp);
	return (l);
}

int XMSRead(long offset, void far *buffer, unsigned count)
{
	return XMSMoveData(fp_to_long(buffer), offset, count, -1);
}

int XMSWrite(long offset, void far *buffer, unsigned count)
{
	return XMSMoveData(offset, fp_to_long(buffer), count, 1);
}

//
// Extended Move Data uses BIOS interrupt 15 subfunction 87 to
// move data between extended memory and real mode memory.
//
// direction determines if the move is from real to protected or
// protected to real.
//
// direction		description
//   < 0			move from protected to real  (src is protected)
//   > 0			move from real to protected  (src is real)
//   = 0  			move from real to real		  (src and dst are real)
//                 (move will fail!)
//
int XMSMoveData(unsigned long ulDstOffset, unsigned long ulSrcOffset,
                           unsigned uCount, int iDirection)
{
	static XMS_BLOCK xmsBlock;
	int iWorked = 0;

	ASSERT(uCount > 0);

	// if there is no direction, return that move failed
	if (iDirection == 0)
		return 0;

	// if somehow we got here when we are not using XMS, then we must exit
	// with the call failing.
	if(!bUseExtended)
		return 0;

	else
	{
		if(uCount & 1)
			++uCount;

		xmsBlock.ulLength = uCount;
		xmsBlock.uSrcHandle = 0;
		xmsBlock.uDstHandle = 0;

		/* set up source and destination for move */
		if(iDirection < 0)
		{
			//
			// source is protected memory, destination is real memory
			//
			xmsBlock.uSrcHandle  = uXMSBufferHandle;
			xmsBlock.ulSrcOffset = ulSrcOffset;
			xmsBlock.ulDstOffset = ((ulDstOffset&0xFFFF0L) << 12) + (ulDstOffset&0xF);
		}

		else
		{
			//
			// source is real memory, destination is protected memory
			//
			xmsBlock.ulSrcOffset = ((ulSrcOffset&0xFFFF0L) << 12) + (ulSrcOffset&0xF);
			xmsBlock.uDstHandle  = uXMSBufferHandle;
			xmsBlock.ulDstOffset = ulDstOffset;
		}


		iWorked = XMSMove(&xmsBlock);
	}

	return iWorked;
}

