;
;   $Id: WRITE_P.ASM 1.5 2003/05/28 20:03:48Z thomd Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
.386
comment &
   WRITE_P.ASM

   Copyright (c) 1994-2002 by Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary


        Written by Doug Huffman December 16, 1994.  The code below is to be
        used with the mapper.asm code and is used to implement write
        protection for certain portions of memory while in V86 mode.  This is
        accomplished by making the affected pages read only and then handling
        the page faults to cause execution to continue without actually
        writing to the protected area.
&

_TEXT   segment dword use16 public 'CODE'
assume cs:_TEXT, ds:nothing

extrn   VMM_chain:word,excep_chain:word,call_VMM:near,VMM_suc:near
align 2
old_vmm_addr    dw      0               ;original VMM_chain
old_excep_addr  dw      0               ;original excep_chain


comment &
        write_protect_mem(long prot_addr)

        prot_addr is a 32 bit 4 kilobyte aligned address which points to a
        page of memory to write protect.  This code hooks into the exception
        handler chain and also into the VMM (Virtual Machine Monitor)
        functions.  The address MUST be 4k aligned and it must be below
        110000h.  NO CHECKING IS DONE ON THESE REQUIREMENTS!  This function
        returns nothing since all calls should be successful.
&
public _write_protect_mem
_write_protect_mem:
        push    bp
        mov     bp,sp
        push    ds
        push    cs
        pop     ds
assume ds:_TEXT
        cmp     old_vmm_addr,0
        jne short write_p               ;jmp if not the first call
        mov     ax,VMM_chain
        mov     old_vmm_addr,ax
        mov     ax,excep_chain
        mov     old_excep_addr,ax
        mov     VMM_chain,offset VMM_hook
        mov     excep_chain,offset excep_hook
;VMM and exceptions are now hooked, write protect some memory
write_p:
        mov     ecx,ss:[bp+4]           ;get address to protect
        mov     al,10h                  ;function to protect memory
        call    call_VMM                ;call the Virtual Machine Monitor
;The call above is transferred to the function below in protected mode
;where the page tables can be directly accessed.
        pop     ds
        pop     bp
assume ds:nothing
        ret

comment &
        The code below is called by the VMM IN PROTECTED MODE for function
        calls which the VMM cannot handle.  This code handles the 10h call
        only.  If al is not 10h, it chains to the original VMM chain vector.
        If al = 10h, the linear 32 bit address in ecx is assumed to be a 4
        kilobyte aligned address of a page to write protect.  eax, esi, ebp
        and all selectors can be trashed.  SS is a selector with 4 gbyte limit
        pointing to zero so we can write to any memory desired with ss.
&
VMM_hook:
        cmp     al,10h
        je short VMM_10
        jmp     old_vmm_addr                    ;chain if not our call
VMM_10: ;write protect the address in ecx.
        mov     eax,ecx
        shr     eax,10                          ;convert to page table index
        add     eax,3976*1024                   ;add offset of page table
;ss:eax points to location in page table to clear the write enable bit
        and byte ptr ss:[eax],0fdh              ;clear the bit in page table
        jmp     VMM_suc


comment &
        The code below is called by the exception handler, it deals only with
        page faults and handles them by assuming the fault was caused by an
        attempt to write to a protected area.  It attempts to return execution
        to the instruction following the offending instruction.  All registers
        and flags are preserved, even in the case of instructions like rep
        movs, xchg, or add where the flags and registers would normally be
        altered by the instruction.

        When the following code receives control, the V86 stack frame exists
        followed by the exception number and ax.  THIS IS PROTECTED MODE
        CODE!!!  All registers have been preserved including ax.  SS is a
        selector pointing to zero with 4 gbyte limit.
&
excep_hook:
        cmp byte ptr ss:[esp+2],14      ;check for page fault
        je short page_f
        jmp     old_excep_addr          ;chain to next handler
page_f:         ;it was a page fault
        push    ss
        pop     ds
        push    esi
        push    ebx
        push    ecx
        push    dx
        movzx   esi,word ptr ss:[esp+22];get ip of offending instruction
        movzx   ecx,word ptr ss:[esp+26];get cs of offending instruction
        shl     ecx,4                   ;convert segment to linear address
;ds:[esi+ecx] points to the offending instruction
        xor     ax,ax                   ;zero all attributes

comment &
        The action item below is jumped to for handling prefixes which have no
        effect on disassembly of further bytes.  These are prefixes such as
        segment over rides.
&
action1 label near

comment &
        In the loop below, al will contain attributes as follows:
        bit 0   if set, addressing is 32 bit
        bit 1   if set, operand is 32 bit

        ah will contain the original byte of code.

        The loop below loops until the instruction is identified and dealt
        with.  After the loop terminates, si will be the desired ip to return
        to.

        The jump index table indexes 128 4 bit nibbles.  The 256 possible
        instructions are reduced to 128 possibilities by ignoring bit #4.
        With the exception of a few special cases, bit #4 can be ignored in
        disassembling instructions which write to memory.  The special cases
        will be dealt with outside of the jump table.
&
disasm_loop1:
        movzx   bx,byte ptr ds:[ecx+esi];load one byte of code
        inc     si                      ;increment ip
        mov     dl,bl
        mov     ah,bl
        and     dl,0e0h                 ;clear all except top 3 bits
        and     bl,0fh                  ;clear all except bottom 4 bits
        shr     dl,1                    ;move top 3 bits to the right
        or      bl,dl                   ;make a 7 bit number
;CAUTION!!!!  The instruction above clears CF for the instruction below!!!!
        rcr     bx,1                    ;save lsb in CF
        mov     bl,cs:jmp_index_table[bx]
;if the carry flag is set, we must use the high nibble in bl else use the low
        jnc short low_nibble
        ror     bl,4
low_nibble:
        and     bl,0fh                  ;zero high nibble
;Caution!!!! bh MUST = 0 at this point!
        shl     bx,1                    ;address a table of words
        jmp     cs:jmp_table[bx]

comment &
        This code handles one byte instructions which have no offsets,
        immediates or SIB bytes such as movs and stos.
&
action3 label near

exit_disasm_loop1:      ;si is the desired ip to return to.
        pop     dx
        pop     ecx
        pop     ebx
        mov     ss:[esp+12],si           ;insert new ip
        pop     esi
        pop     ax
        add     esp,6           ;pop int # and dword error code
        iretd                           ;return to instruction following fault

comment &
        Each of the action items below handles a specific type of instruction
        byte.  If the byte was a prefix, it jumps back to the loop after
        setting any required attributes in al.  If the byte was not a prefix,
        it handles the complete instruction and sets si to the desired ip
        value before jumping to the label, "exit_disasm_loop1.

        The code below handles all unknown code.  This simply jmps back to the
        fault handler chain to see if anyone else in the chain can handle it.
        This should be extremely rare because all 386 instructions should be
        handled correctly.
&
action0:
        pop     dx
        pop     ecx
        pop     ebx
        pop     esi
        jmp     old_excep_addr          ;chain to next handler

comment &
        This code handles prefixes which may affect disassembly of the
        following bytes.   Currently, this handles only 66h (operand size)
        and 67h, (address size) prefixes.
&
action2:
        cmp     ah,66h          ;check if 32 bit operand
        je short _66h           ;jmp if a 32 bit operand
;we will assume it is a 32 bit address prefix
        or      al,1            ;set operand attribute
        jmp     disasm_loop1

_66h:   or      al,2            ;set address attribute
        jmp     disasm_loop1

comment &
        Handles two byte, 1 offset code such as: mov data1,bx.  We also must
        check for the special case of C0h and C1h opcodes which have an
        immediate 8 byte data following it.
&
action5:
        or      ah,1            ;set the lsb
        cmp     ah,0C1h
        je short action6_a      ;deal with a special case

comment &
        Code below handles all coprocessor instructions which write to memory.
&
action9 label near
action5_a:
        call    eat_addr
        jmp     exit_disasm_loop1

comment &
        Code below handles instructions of the 1 byte, 1 offset variety such
        as: mov data1,ax   1010001x offset
&
action4:
        and     ax,2            ;clear everything except address attribute
        add     ax,2            ;2 bytes for 16 bit and 4 bytes for 32 bit
        add     si,ax           ;calculate new ip
        jmp     exit_disasm_loop1

comment &
        Action 6 handles two byte, offset immed code where the immed is a
        byte such as: rcl data1,10

        This code must also deal with the special case of the 1101000x
        instruction which must jmp to action 5.  1100000x must stay here.
&
action6:
        test    ah,10h          ;check bit #4
        jnz     action5_a
action6_a:
        call    eat_addr
        inc     si              ;eliminate one byte of immediate data
        jmp     exit_disasm_loop1

comment &
        The code in action 7 handles two byte, offset immed code where the
        immed is a full size word or dword such as:
        add data1,10
&
action7:
        call    eat_addr
        and     ax,1            ;0 for words,1 for dwords
        inc     ax              ;1 for words,2 for dwords
        shl     ax,1            ;2 for words, 4 for dwords
        add     si,ax           ;increment for the immediate data
        jmp     exit_disasm_loop1

comment @
        Code below handles instructions which start with 0FH.  It merely
        examines the next byte and jmps to another action item based on the
        byte following the 0FH.  The jmp based on the second byte does not
        follow the same rules as if it were based on the first byte so the jmp
        table is not useful in this decision.

0F instructions

shld,shrd       0f 1010x100 byte offset byte
shld,shrd by cl 0f 1010x101 byte offset

btc data1,10    0f 10111010 byte offset byte
btc data1,ax    0f 101xx011 byte offset

seto,setx...    0f 1001xxxx byte offset

It is a byte,byte, offset,byte if (lsb = 0) && (bit #5 != 0), otherwise, it is
        byte,byte offset
@
action8:
        mov     bl,byte ptr ds:[ecx+esi];load one byte of code
        inc     si                      ;increment ip
        test    bl,1
        jnz     action5_a
        test    bl,32
        jz      action5_a
        jmp     action6_a

comment &
        The code below "eats" the addressing bytes in an instruction with a
        mod r/m byte.  ds:[ecx+esi] must point to the mod r/m byte, al must
        contain the attributes of the instruction.  returns with si pointing
        to the end of the addressing bytes, preserves all other registers.
&
eat_addr:
        push    bx
        mov     bl,ds:[ecx+esi]
        inc     si
        mov     bh,111b         ;keep only the r/m
        and     bh,bl           ;r/m is in bh
        test    al,1            ;check if 32 bit addressing
        jnz short addr_32
;addressing is 16 bit
        cmp     bl,64
        jb short mod_0
        cmp     bl,128
        jb short mod_1
;The mod portion of the mod r/m byte is 2 or 3, with a 3, the destination
;is a register, and we know the destination was memory, so we can assume
;the mod is a 2.  With mod = 2, there is a 2 byte displacement
        add     si,2            ;blow away two more bytes

no_displacement:        ;at this point si points to the end of the addressing
        pop     bx
        ret

mod_0:  ;the mod in the mod r/m byte is a 0, no displacement unless r/m = 110b
        cmp     bh,110b ;check if 16 bit displacement
        jne     no_displacement
        inc     si                      ;eliminate 1 byte of 2 byte address
mod_1:  ;if mod = 1, we have a mod r/m byte and a 1 byte displacement
        inc     si
        jmp     no_displacement

addr_32:        ;first check for sib byte
        cmp     bh,100b
        je short sib_present
sib_re_enter:
        cmp     bl,64
        jae short mod_above_64
;if r/m = 101 (or base if sib), we have a 32 bit displacement
;otherwise, no displacement.
        cmp     bh,101b
        jne     no_displacement
dword_displacement:
        add     si,4                    ;take care of a dword displacement
        jmp     no_displacement

mod_above_64:
        cmp     bl,128
        jb      mod_1
;at this point, we will assume mod = 2 since 3 is for destination of a register
        jmp     dword_displacement

sib_present:
        mov     bh,ds:[ecx+esi]         ;load sib byte
        inc     si
        and     bh,111b                 ;clear all except base bits
        jmp     sib_re_enter

comment &
        The table below has one nibble for each of 128 entries.  This allows
        for 16 different actions to be taken based on the byte of code which
        was retrieved from memory.  Bit #4 of the code byte is removed and the
        resultant 7 bit number is used as an index into this table.  The
        nibble retrieved from the table is used as an index into the jmp
        table.  This arrangement saves space since having a word oriented jmp
        table for all 256 possible bytes would consume 512 bytes directly in
        tables.  This arrangement consumes only 64 bytes in the jmp_index
        table and 20 bytes in the jmp_table for a total savings of 428 bytes
        (minus a few bytes for added complexity of code)
&
jmp_index_table  db      55h            ;action 5 for bytes 1 and 2
                 db      0
                 db      0
                 db      0
                 db      55h            ;action 5 for bytes 8 and 9
                 db      0
                 db      0
                 db      80h            ;action 8 for byte 0Fh
                 db      55h            ;action 5 for bytes 10h and 11h
                 db      0
                 db      0
                 db      1              ;action 1 for byte 16h
                 db      55h            ;action 5 for bytes 18h and 19h
                 db      0
                 db      0
                 db      1              ;action 1 for byte 1eh
                 db      0              ;20h
                 db      0
                 db      0
                 db      0
                 db      0
                 db      0
                 db      0
                 db      0
                 db      0              ;30h
                 db      0
                 db      11h            ;action 1 for bytes 34 and 35h
                 db      22h            ;action 2 for bytes 36 and 37h
                 db      0
                 db      0
                 db      0
                 db      0
                 db      76h            ;40h
                 db      66h            ;action 6 for bytes 42 and 43h
                 db      0
                 db      55h            ;action 5 for bytes 46 and 47h
                 db      55h            ;action 5 for bytes 48 and 49h
                 db      0
                 db      05h            ;action 5 for byte 4ch
                 db      50h            ;action 5 for byte 4fh
                 db      0              ;50h
                 db      44h            ;action 4 for bytes 52 and 53h
                 db      33h            ;action 3 for bytes 54 and 55h
                 db      0
                 db      0
                 db      33h            ;action 3 for bytes 5a and 5bh
                 db      0
                 db      0
                 db      55h            ;60h
                 db      55h            ;action 5 for bytes 62h and 63h
                 db      0
                 db      76h            ;action 6 for 66h, 7 for 67h
                 db      99h            ;action 9 for coprocessor instructions
                 db      99h
                 db      99h
                 db      99h
                 db      1              ;70h
                 db      11h            ;action 1 for bytes 72h and 73h
                 db      0
                 db      55h            ;action 5 for bytes 76 and 77h
                 db      0
                 db      0
                 db      0
                 db      55h            ;action 5 for bytes 7eh and 7fh


jmp_table               dw      offset action0
                        dw      offset action1
                        dw      offset action2
                        dw      offset action3
                        dw      offset action4
                        dw      offset action5
                        dw      offset action6
                        dw      offset action7
                        dw      offset action8
                        dw      offset action9
_TEXT   ends
end

