;
;   $Id: VMM.INC 1.4 2002/10/15 20:42:12Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
COMMENT ~ /*-----------------------------------------------------------------
        $DOCFILE:VMM.INC
        $DOCMODULE:VMM

        Copyright (C) Datalight, Inc 1997-2002
        All Rights Reserved.
        Datalight Confidential and Proprietary

        This file contains general VMM related structures and constants.

        $DOCHISTORY
        06/13/97 gp     Created
        07/11/97 gp     Obsolete MSGR and MSGP and added LOGMSG.
        07/15/97 dh     Added extrn definitions such as VmmDisplayString
        07/21/97 gp     Modified CnvBin16ToHexStr to return a character count
	08/28/97 gp	Added the SPRINTF and LOGPRINTF macros
	09/11/97 gp	Updated SPRINTF to accept addresses in the form AX:BX
	10/13/97 gp	Added the DBGMSG and DBGPRINTF macros
-------------------------------------------------------------------------*/ ~

LOGBUFFSIZE     equ <256>       ; max output length for LOGPRINTF

if __FILENUM__ NE 8     ; define extrns for all files except VMMERR.ASM
  extrn VmmDisplayString:proc
  extrn VmmBin16ToDecStr:proc
  extrn VmmBin16ToHexStr:proc
  extrn VmmBin08ToHexStr:proc
  extrn VmmProcessAssert:proc
  extrn VmmRegisterDump:proc
  extrn VmmSpeakerBeep:proc
  extrn VmmTimerWait:proc
  extrn _VmmSprintf:proc

  extrn logBuff:byte
  extrn lpLogBuff:dword
  extrn bLogPrintfBusy:byte
endif

BEEP_FREQUENCY	equ <440>
BEEP_DURATION	equ <18>	; Timer tics more-or-less, but due to
				; calibration difficulties, they appear
				; to be very short timer ticks.

;---------------------------------------------------------------------
;       $DOCMACRO:ASSERTFAILED
;
;       This macro replaces the standard WinLight ASSERTFAILED macro
;       because VMM has special needs.  It is called by the ASSERT
;       macro whenever an assert fails.
;---------------------------------------------------------------------
ASSERTFAILED    MACRO   filenum,linenum,ovr
local           this
.errnb <ovr>
.errb  <filenum>

        pusher  ax,bx
        mov     ax,filenum
this:   mov     bx,offset _TEXT:this
        call    VmmProcessAssert
        popper  ax,bx

                ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:LOGMSG
;
;       This macro displays the given message via the LOG TSR.
;
;       The message may be specified in one of three forms.  These are:
;       immediate string data surrounded by single quotes, variable
;       name preceded with an '@' sign, or register based address.
;
;       Examples:
;               LOGMSG  <'Display this string'>
;               LOGMSG  @stringVar
;               LOGMSG  DS:BX
;
;       It is the programmer's responsibility to ensure that the string
;       is null terminated (except when using the immediate form).
;
;       NOTE:  When using the variable name form, the variable must be
;       declared in the current code segment (the far address is formed
;       using the current CS).  Also note that when we can figure out
;       how to code the macro so TASM will let us use a '&' instead of
;       '@' we can make things look more "C like".
;
;       NOTE:  When using the register form, only 16 bit registers are
;       allowed at this time.
;
;       This macro generates code in DEBUG only, and generates no code
;       when DEBUG=0.
;---------------------------------------------------------------------
LOGMSG        MACRO   msg,ovr
local         loclable,msgvar,plen,chr1,chr3,reg1,reg2
.errb  <msg>
.errnb <ovr>

        pusher  bx,dx

  plen SIZESTR <msg>
  chr1 SUBSTR <msg>,1,1

  ifidni chr1,<'>                       ;; is it immediate string data?
        jmp     loclable
msgvar  db      &msg,0
loclable:
        mov     dx,cs
        mov     bx,offset msgvar
        call    VmmDisplayString

  elseifidni chr1,<@>                   ;; force it to be an address?
        mov     dx,cs                   ;; always assume CS for now
    chr1 SUBSTR <msg>,2,plen-1
        mov     bx,offset cs:chr1
        call    VmmDisplayString

  else
    if plen EQ 5                        ;; args in the form "DS:BX" will be 5 chars long
      chr3 SUBSTR <msg>,3,1             ;; grab the third character
      ifidni chr3,<:>                   ;; must be a colon for a properly formed address
        reg1 SUBSTR <msg>,1,2           ;; grab the segment/selector value
        ifdifi reg1,<dx>                ;; no need to load reg if it's the same
        mov     dx,reg1
        endif
        reg2 SUBSTR <msg>,4,2           ;; grab the offset
        ifidni reg2,<dx>
          .err
          %out "ERROR: LOGMSG: DX has already been trashed"
          exitm
        endif
        ifdifi reg2,<bx>                ;; no need to load reg if it's the same
        mov     bx,reg2
        endif
        call    VmmDisplayString
      else
        .err
        %out "ERROR: LOGMSG: Bad parameter (01)"
        exitm
      endif
    else
      .err
      %out "ERROR: LOGMSG: Bad parameter (02)"
      exitm
    endif
  endif

        popper  bx,dx
ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:DBGMSG
;
;       This macro displays the given message via the LOG TSR.  It is
;	identical to the LOGMSG macro, except that it only generates
;	code when DEBUG > 0.  See LOGMSG for more documentation.
;---------------------------------------------------------------------
if DEBUG
  DBGMSG        MACRO   msg,ovr
  .errb  <msg>
  .errnb <ovr>

	LOGMSG	msg
  ENDM

else

  DBGMSG        MACRO   msg,ovr
  .errb  <msg>
  .errnb <ovr>
  ENDM

endif


;---------------------------------------------------------------------
;       $DOCMACRO:LOGPRINTF
;
;	This is a generic printf macro that dumps its output to the
;	LOG TSR.
;
;	It is callable at any time, from anywhere, provided CS is
;	the VMM real mode segment, or protected mode code selector.
;
;	It uses a static buffer "LogBuff", so it is not reentrant.  A
;	"busy" flag prevents this from being a concern.
;
;	Syntax:  LOGRINTF FormatStr,param1,param2,...param20
;
;	Please see the function "_VmmSprintf" located in VMMERR.ASM
;	for information on the formatting string.
;
;	Please see the function "VmmTestPrintf" located in VMMERR.ASM
;	for sample usages.
;
;       This macro generates code in DEBUG only, and generates no code
;       when DEBUG=0.
;---------------------------------------------------------------------
LOGPRINTF     MACRO   fmt,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
local		PrntfCleanup
.errb  <fmt>
.errnb <ovr>

	pushf

	push	ds
	push	ax
	GETDATASEG
	mov	ds,ax
	inc	bLogPrintfBusy		;; set our busy flag

	cmp	bLogPrintfBusy,1	;; did it just become 1?
	jne	PrntfCleanup		;; exit if not

	mov	ax,offset logBuff	;; build far pointer appropriate to
	sdw	lpLogBuff,ds,ax		;; the mode we're currently in
	pop	ax
	pop	ds

        SPRINTF cs:lpLogBuff,fmt,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20

        LOGMSG  @logBuff

	push	ds
	push	ax
	GETDATASEG
	mov	ds,ax

  PrntfCleanup:
	dec	bLogPrintfBusy
	pop	ax
	pop	ds

	popf
ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:DBGPRINTF
;
;       This macro displays the given message via the LOG TSR.  It is
;	identical to the LOGPRINTF macro, except that it only generates
;	code when DEBUG > 0.  See LOGPRINTF for more documentation.
;---------------------------------------------------------------------
if DEBUG
  DBGPRINTF     MACRO   fmt,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
  .errb  <fmt>
  .errnb <ovr>

	LOGPRINTF fmt,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20
  ENDM

else

  DBGPRINTF     MACRO   fmt,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
  .errb  <fmt>
  .errnb <ovr>
  ENDM

endif



;---------------------------------------------------------------------
;       $DOCMACRO:CnvBin16ToDecStr
;
;       Convert a 16 bit binary value to a decimal string.
;
;       Entry      AX - Number to convert
;               ES:DI - The end of the output buffer
;
;       Exit       CX - The number of characters output
;               ES:DI - Character before the first byte of the number
;       Flags         - Math flags may be changed
;---------------------------------------------------------------------

CnvBin16ToDecStr        MACRO   ovr
local                   next
.errnb <ovr>

        pusher  ax,bx,dx
        xor     cx,cx           ; initialize the count
        mov     bx,10           ; dividing by 10
next:   xor     dx,dx           ; clear the upper half
        div     bx
        add     dx,'0'          ; make the remainder ASCII
        mov     es:[di],dl
        dec     di              ; point to the next position
        inc     cx              ; inc the count
        or      ax,ax           ; all done?
        jnz     next
        popper  ax,bx,dx

ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:GETDATASEG
;
;       This macro loads AX with the data segment value appropriate
;       for the mode we're running in.
;
;       NOTE:   Makes some assumptions that I'm not sure are true 100%
;               of the time!
;---------------------------------------------------------------------
GETDATASEG      MACRO   ovr
local           af_real,af_ret
.errnb <ovr>

        mov     ax,cs
        cmp     ax,_TEXT_E
        jne     af_real         ; jump if in real mode

        mov     ax,_TEXT_W      ; get protected mode data selector
        jmp     af_ret

af_real:mov     ax,cs           ; CS goes for code and data in real mode
af_ret:
ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:VmmBeep
;
;	Beep the speaker.
;---------------------------------------------------------------------
VmmBeep         MACRO   ovr
.errnb <ovr>

	pusher	ax,bx

	mov	ax,BEEP_FREQUENCY
	mov	bx,BEEP_DURATION
	call	VmmSpeakerBeep

	popper	ax,bx

ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:VmmWait
;
;	Wait the specified number of timer ticks.
;---------------------------------------------------------------------
VmmWait         MACRO   tics,ovr
.errnb <ovr>
.errb <tics>

  ifdifi <ax>,<tics>
	pusher	ax
	mov	ax,tics
  endif

	call	VmmTimerWait

  ifdifi <ax>,<tics>
	popper	ax
  endif

ENDM


;---------------------------------------------------------------------
;       $DOCMACRO:SPRINTF
;
;	This is a generic "sprintf" macro.  Its primary purpose is to
;	serve as a back-end for "LOGPRINTF", however it may be used
;	independently if needed.
;
;	Syntax:  SPRINTF DestStr,FormatStr,param1,param2,...param20
;
;	Please see the function "_VmmSprintf" located in VMMERR.ASM
;	for information on the formatting string.
;
;	Please see the function "VmmTestPrintf" located in VMMERR.ASM
;	for sample usages.
;---------------------------------------------------------------------
SPRINTF         macro   dest,source,p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15,p16,p17,p18,p19,p20,ovr
local           parm_c,stack_z,chr,chr1,chr2,plen,regx,regl,regh,reglen,colon
.errb   <source>
.errnb  <ovr>

parm_c = 0			;; parameter count
stack_z =  0			;; stack size
BX_trashed = 0			;; true if BX gets trashed

	push	ebx

irp     prc,<p20,p19,p18,p17,p16,p15,p14,p13,p12,p11,p10,p9,p8,p7,p6,p5,p4,p3,p2,p1,source,dest>
local loclable,varname
ifnb <prc>

  plen SIZESTR <prc>			;; determine the param len
  chr1 SUBSTR <prc>,1,1			;; grab the first character
  colon INSTR <prc>,<:>			;; find offset of any colon

  ifidni chr1,<'>                       ;; is it string data?
    BX_trashed = 1
    stack_z = stack_z + 4

        jmp     loclable
varname db      &prc,0			;; define string in the code segment
loclable:
        push    bx
        push    ax
        mov     bx,sp
        mov     ss:[bx+2],cs
	mov	ax,offset varname
        xchg    ss:[bx],ax              ;; AX restored to original value

  elseifidni chr1,<@>                   ;; force it to be an address?
    BX_trashed = 1
    stack_z = stack_z + 4
    chr1 SUBSTR <prc>,2,plen-1
        push    bx
        push    ax
        mov     bx,sp
        mov     ss:[bx+2],cs
        lea     ax,chr1
        xchg    ss:[bx],ax              ;; AX restored to original value

  elseif (plen EQ 5) AND (colon EQ 3)	;; is it something like DS:SI?
    chr1 SUBSTR <prc>,1,2		;; get the first register
    chr2 SUBSTR <prc>,4  		;; get the second register
    stack_z = stack_z + 4
	push	chr1
	push	chr2

  elseif (.TYPE prc) and OPTYPE_CONST
    BX_trashed = 1
    chr2 SUBSTR <prc>,plen,1

    ifidni chr2,<L>                     ;; if last char is an uppercase L
      stack_z = stack_z + 4
      chr2 SUBSTR <prc>,1,plen-1
        mov     ebx,chr2
        push    ebx
    else
      stack_z = stack_z + 2
        mov     bx,prc
        push    bx
    endif

;  elseif (.TYPE prc) and OPTYPE_SSBASED
;    BX_trashed = 1
;    stack_z = stack_z + 4
;        push    ss
;        push    ax
;        mov     bx,sp
;        lea     ax,prc
;        xchg    ss:[bx],ax

  elseif (.TYPE prc) and OPTYPE_REG
    ?reglength prc,reglen			;; get the width of the register

    if reglen EQ 1			;; 8 bit register
      stack_z = stack_z + 2

      chr1 SUBSTR <prc>,1,1
      chr2 SUBSTR <prc>,2,1
      regx CATSTR chr1,<x>
      regl CATSTR chr1,<l>
      regh CATSTR chr1,<h>

      ifidni regx,<bx>
        if BX_trashed
	  .err
          %out "ERROR: SPRINTF: BX has already been trashed (03)"
          exitm
        endif
      endif
        push    regx
      ifidni chr2,<H>
        xchg    regh,regl
      endif
        xor     regh,regh
        push    si
        mov     si,sp
        xchg    ss:[si+2],regx
        pop     si

    elseif reglen EQ 2
        stack_z = stack_z + 2
        push    prc			;; 16 bit register

    elseif reglen EQ 4
        stack_z = stack_z + 4
        push    prc			;; 32 bit register

    else
      .err
      %out "ERROR: SPRINTF: Bad parameter: &prc (03)"
      exitm
    endif

  else                                  ; must be data then
    if (TYPE prc) EQ 1
      BX_trashed = 1
      stack_z = stack_z + 2
        mov     bl,prc                  ; NOTE:  This won't work right if
        xor     bh,bh                   ; a multibyte variable was specified.
        push    bx                      ; Use an & to pass the address.
    elseif (TYPE prc) EQ 2
      stack_z = stack_z + 2
        push    prc
    elseif (TYPE prc) EQ 4
      stack_z = stack_z + 4
        push    wptr prc+2
        push    wptr prc
    else
      .err
      %out "ERROR: SPRINTF: Bad parameter: &prc (05)"
      exitm
    endif
  endif

parm_c = parm_c + 1

endif
endm
        mov     bx,parm_c-2             ;; don't include src and dest in parm count
        call    _VmmSprintf
        add     sp,stack_z

	pop	ebx
ENDM

