;
;   $Id: WLASSERT.INC 1.4 2002/10/15 20:42:12Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
.xlist
COMMENT ~ /*-------------------------------------------------------------
        $DOCFILE:WLASSERT.INC

        Copyright (C) Datalight, Inc. 1996-2002
        All Rights Reserved.
        Datalight Confidential and Proprietary.

        This include contains standard macros that can be used
        by any assembly modules in WinLight.

        $DOCHISTORY
        07/29/96 gp     Created (moved from WINLIGHT.INC)
        07/30/96 gp     Added more ASRT symbols
        02/18/97 gp     Updated the header
        02/27/97 gp     Updated to not use ?PARMIN and ?PARMAX
        03/04/97 gp     Forced ASSERT call to ErrHandler to be FAR
        04/14/97 gp     Fixed a case problem
        06/13/97 gp     Added the ASSERTFAILED macro.
        06/18/97 gp     Updated Abort and ASSERTFAILED to pass the DEBUG
                        level to ErrHandler (!).
        08/26/97 gp     Moved ?PARSEEXP to DLMACROS.INC
----------------------------------------------------------------------*/ ~

; ErrHandler() flags (low byte is assert type, high byte is flags)
ASRT_STANDARD           equ 0
ASRT_SLOW               equ 1
ASRT_UNTESTED           equ 2
ASRT_UNIMPLEMENTEDONCE  equ 3
ASRT_ALERTONCE          equ 4
ASRT_ABORT              equ 5
ASRT_PAUSE              equ 6
ASRT_VERIFY             equ 7
ASRT_STUBBED            equ 8
ASRT_ALERT              equ 9
ASRT_UNIMPLEMENTED      equ 10
ASRT_TYPE_MASK          equ 000FFh      ; mask the type byte
ASRT_FLAGS_MASK         equ 0FF00h      ; mask the flags byte
ASRT_DEBUG_MASK         equ 00700h      ; mask the DEBUG level bits within the flags
ASRT_FIRSTTIME          equ 08000h      ; signifies it's the first time fired


;------------------------------------------------------------------------
;       $DOCMACRO:Abort
;
;       This is the old standard Abort macro which is being gradually
;       replaced by the new ASSERT macro.
;------------------------------------------------------------------------
Abort   MACRO   lineNum
        push    __FILENUM__
        push    lineNum
        push    ASRT_ABORT or ((DEBUG shl 8) and ASRT_DEBUG_MASK)
        push    0
        push    0
        call    ErrHandler
        ENDM

;------------------------------------------------------------------------
;       $DOCMACRO:ASSUMEFETALPOSITION
;
;       Tell any available debugger we are about to die.  Generally this
;       macro should only be used in places where a standard ASSERT or ABORT
;       is not likely to work (inside an interrupt/exception handler, etc).
;------------------------------------------------------------------------
ASSUMEFETALPOSITION     MACRO
        int 3           ;; if a debugger is loaded, activate it
        hlt             ;; otherwise, alert VMM
        ENDM


;------------------------------------------------------------------------
;       $DOCMACRO:ASSERTFAILED
;
;       This macro is called by the ASSERT macro when it fails.  It is
;       separately defined so that it may be redefined on a project or
;       module basis as needed.
;------------------------------------------------------------------------
ASSERTFAILED    MACRO   filenum,linenum,ovr
.errnb <ovr>
.errb  <linenum>

        push    eax     ; for now we don't care what is returned from ErrHandler

        push    filenum
        push    linenum
        push    ASRT_STANDARD or ((DEBUG shl 8) and ASRT_DEBUG_MASK)
        push    0
        push    0
        call    fptr ErrHandler

        pop     eax
                ENDM

;------------------------------------------------------------------------
;       $DOCMACRO:ASSERT        ; <expression>,LineNumber,[DEBUGONLY]
;
; Assert that the specified expression is TRUE.  Valid operators are listed
; below.  Registers, memory variables, and immediate values are allowed,
; though they must be the same type - bytes, words, or dwords.  Type over-
; rides (byte ptr, word ptr, dword ptr) are allowed.  If two memory variables
; are specified, the first operator will be moved into a register for the
; comparison.
;
; All comparisons are considered to be unsigned, all registers are preserved,
; and the mathematical flags may be modified (see below).  Values in segment
; registers cannot be tested at this time.
;
; Valid operators are:  EQ - equal
;                       NE - not equal
;                       GT - greater than
;                       GE - greater than or equal
;                       LT - less than
;                       LE - less than or equal
;
; The operator must be separated from each operand by at least one white space.
;
; Examples:     ASSERT  <ax NE 0>,123,DEBUGONLY ;; assert AX != 0
;               ASSERT  <bptr wordVar GT 1>,456 ;; assert lobyte(wordVar) > 1
;
; If the ASSERT fails, the ASSERTFAILED macro will be executed with the
; specified line number.  If the user chooses to continue, the carry flag
; will be set, otherwise it will be clear.
;
; If the optional DEBUGONLY flag is used, this macro will only generate code
; if DEBUG=TRUE.  If you will be checking CF after a DEBUGONLY ASSERT, you
; must manually clear CF before executing the macro, otherwise CF will be
; unitialized if the code is built with DEBUG=FALSE (the general rule of
; thumb is that if you are testing the success of an ASSERT, the ASSERT in
; question should probably not be DEBUGONLY).
;------------------------------------------------------------------------

ASSERT          macro expression,lineNum,dbug,ovr
local           asrtOp1,asrtOp2,asrtOperator,assertPassed,saveAX,assertFini,needCLC
.errnb <ovr>
.errb  <lineNum>

  ifnb <dbug>
    ifidni <dbug>,<DEBUGONLY>   ; generate only if DEBUG=TRUE?
      ife DEBUG
        exitm                   ; if no DEBUG, generate no code
      endif
    else
      .err "ASSERT: Unrecognized parameter: &dbug"
      exitm
    endif
  endif

  needCLC       = TRUE  ; assume we need a CLC at this time
  saveAX        equ <>  ; becomes non-blank if (E)AX must be popped

  asrtOp1       equ <>  ; the first operand
  asrtOp2       equ <>  ; the second operand
  asrtOperator  equ <>  ; the comparison operator

  ; break the expression into its three components
  ?PARSEEXP <expression>,asrtOp1,asrtOp2,asrtOperator
  ifb &asrtOperator
    .err "ASSERT: Invalid expression: &expression"
    exitm
  endif

  ; If neither operand is a register and the second operand is not
  ; a constant, we move the first into (E)AX
  ife ((.TYPE asrtOp1) and OPTYPE_REG)
  ife ((.TYPE asrtOp2) and OPTYPE_REG)
  ife ((.TYPE asrtOp2) and OPTYPE_CONST)
    if (TYPE asrtOp1) EQ 1              ; is it a byte comparison?
        saveAX  equ <AX>
        push    ax                      ; save original AX
        mov     al,asrtOp1
        asrtOp1 equ <AL>                ; change operand1 to be AL
    elseif (TYPE asrtOp1) EQ 2          ; is it a word comparison?
        saveAX  equ <AX>
        push    ax                      ; save original AX
        mov     ax,asrtOp1
        asrtOp1 equ <AX>                ; change operand1 to be AX
    elseif (TYPE asrtOp1) EQ 4          ; is it a dword comparison?
        saveAX  equ <EAX>
        push    eax                     ; save original EAX
        mov     eax,asrtOp1
        asrtOp1 equ <EAX>               ; change operand1 to be EAX
    else
        .err "ASSERT:  Bad operand:  &asrtOp1,  must be Byte, Word, or Dword"
    endif
  endif
  endif
  endif

  ; we could save a several bytes here by checking if op2 is 0 and op1 is
  ; a register, if so, we could use an OR instruction instead of CMP.
  ; Plus we would get the added benefit the OR implicitly clears CF so
  ; the CLC and JMP at the bottom of this macro would not be needed.

        cmp     asrtOp1,asrtOp2         ; perform the comparison

  ; validate the results according to the specified operator
  ifidni &asrtOperator,<EQ>
        needCLC = FALSE                 ; an "=" comparison will clear CF
        je short assertPassed
  elseifidni &asrtOperator,<NE>
        jne short assertPassed
  elseifidni &asrtOperator,<GT>
        ja short assertPassed
  elseifidni &asrtOperator,<GE>
        jae short assertPassed
  elseifidni &asrtOperator,<LT>
        jb short assertPassed
  elseifidni &asrtOperator,<LE>
        jbe short assertPassed
  else
        .err "ASSERT: Unrecognized operator: &asrtOperator"
        exitm
  endif

        ASSERTFAILED __FILENUM__,lineNum

        stc                     ; set CF to indicate ASSERT failed

  if needCLC
        jmp short assertFini    ; if we don't need a CLC, no need to JMP
  endif

assertPassed:
  if needCLC
        clc                     ; clear CF to indicate ASSERT passed
  endif
assertFini:

  ifnb saveAX
        pop     saveAX          ; restore (E)AX if needed
  endif

                endm





.list
