//
//   $Id: ADD2CPI.C 1.2 2002/10/15 20:43:38Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
   Copyright (c) 1992-1997 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>


#include "display.h"


extern void SwapVidFilter(void);


void ExitWithMessage(void)
{
		printf("\nUsage: ADD2CPI codepage file\n\n"
				 "Captures the current display font and either appends\n"
				 "the font to or optionally replaces the font in the\n"
				 "specified .CPI file. Note that the current .CPI format\n"
				 "specifies a maximum of %d code pages.\n",MAX_CODE_PAGES);
		exit(-1);
}

CPIHeader header;
CodePageList outCodePgs = {0}, codePgs;
void *fontPtr;
char far*BIOSFontPtr, far*FILEFontPtr;
unsigned FontBytesMoved;


int main(int argc, char **argv)
{ FILE *fio;
  int i, font = -1, codepage, appending = 0;
  long offset;
  char c;
  unsigned xypos;

	printf("ADD2CPI	Copyright (c) 1996 Datalight, Inc.\n"
			 "All rights reserved.\n\n");

	// check args
	//
	if(argc < 3)
		ExitWithMessage();

	// open input and output files
	//
	codepage = atoi(argv[1]);
	if(codepage <= 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}

	fio= fopen(argv[2],"r+b");
	if(fio < 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}
	rewind(fio);

	// read the font file header and verify the file ID
	//
	fread(&header,sizeof(CPIHeader),1,fio);
	if(strcmp(header.id,"CPIKBS"))
	{
		printf("File %s is not a valid .CPI file, aborting...\n",argv[2]);
		ExitWithMessage();
	}

	// allocate memory for the new font info
	//
	fontPtr = calloc(header.chunkSize,1);
	if(fontPtr == 0)
	{
		perror(argv[0]);
		ExitWithMessage();
	}

	// read the code page list and see if the specified font is there
	// if it is then prompt for replacing it. If not make sure we have
	// room for another.
	//
	fread(codePgs,sizeof(CodePageList),1,fio);
	for(i=0; i < header.maxCodePages; i++)
	{
		if(codePgs[i] != codepage)
			continue;
		else
		{
			c = 0;
			printf("Code Page %d already exists in file %s. ",
					  codepage, argv[2]);

			while(c != 'Y')
			{
				printf("Replace it [Y/N]?");
				c = toupper(getchar());
				if(c == 'N')
					exit(0);
			}
			font = i;
		}
		// find the location of the update in the .CPI file
		//
		offset = header.firstCodePage + (font * header.chunkSize);
		fseek(fio, offset, SEEK_SET);
	}
	if(font < 0)
	{
		font = header.maxCodePages++;
		if(font >= MAX_CODE_PAGES)
		{
			printf("No room for another font in the %s file header.\n");
			ExitWithMessage();
		}

		// we are appending a new font and need to update the file
		// header by incrementing the code page count and adding a new
		// code page to the code page directory array
		//
		appending = 1;
		codePgs[font] = codepage;
		rewind(fio);
		fwrite(&header,sizeof(CPIHeader),1,fio);
		fwrite(codePgs,sizeof(CodePageList),1,fio);
		fseek(fio,0,SEEK_END);
	}

	// get a moveable pointer to the start of the file buffer
	//
	FILEFontPtr = fontPtr;



	// Install our interrupt 10h filter to copy the user font
	// during the mode change. Have moved no font data yet.
	//
	SwapVidFilter();
	FontBytesMoved = 0;

	// Copy the 8x16 font directly to the file buffer.
	// Make sure we're in a 400 scan line mode
	//
	asm {
		mov	ah, 3						// get cursor position
		xor	bx, bx
		int	10h
		mov	xypos, dx

		mov	ax, 1202h				// set scan lines
		mov	bl, 30h
		int   10h
		mov 	ax, 83h					// force mode set w/out clearing screen
		int   10h

	}
	if(FontBytesMoved)
	{
		FILEFontPtr += FontBytesMoved;
		FontBytesMoved = 0;
	}
	else
	{
		BIOSFontPtr = (char far*)(getvect(0x1F)); // get ptr to ROM font
		for(i=0; i < 16*128; i++)
			*FILEFontPtr++ = *BIOSFontPtr++;			// copy it
	}

	// Copy the 8x14 font to the file buffer.
	// Set a 350 scan line mode to force the font load
	//
	asm {
		mov	ax, 1201h
		mov	bl, 30h
		int   10h
		mov 	ax, 83h
		int   10h
	}
	if(FontBytesMoved)
	{
		FILEFontPtr += FontBytesMoved;
		FontBytesMoved = 0;
	}
	else
	{
		BIOSFontPtr = (char far*)(getvect(0x1F)); // get ptr to ROM font
		for(i=0; i < 16*128; i++)
			*FILEFontPtr++ = *BIOSFontPtr++;			// copy it
	}

	// Copy the 8x8 font to the file buffer.
	// Set a 200 scan line mode to force the font load
	//
	asm {
		mov	ax, 1200h
		mov	bl, 30h
		int   10h
		mov 	ax, 83h
		int   10h
	}
	if(FontBytesMoved)
	{
		FILEFontPtr += FontBytesMoved;
		FontBytesMoved = 0;
	}
	else
	{
		BIOSFontPtr = (char far*)(getvect(0x1F)); // get ptr to ROM font
		for(i=0; i < 16*128; i++)
			*FILEFontPtr++ = *BIOSFontPtr++;			// copy it
	}

	SwapVidFilter();

	// back into a 400 scan line mode
	//
	asm {
		mov	ax, 1202h
		mov	bl, 30h
		int   10h
		mov 	ax, 83h
		int   10h

		mov	ah, 2							// restore cursor position
		xor	bx, bx
		mov	dx, xypos
		int 	10h
	}

	// OK. Now we have a copy of the current fonts in our file buffer.
	// The file location has already been found so write out the
	// new code page.
	//
	fwrite(fontPtr,header.chunkSize,1,fio);

	// we're done!
	//
	fseek(fio,0,SEEK_END);
	fclose(fio);

	printf("Current font used to ");
	if(appending)
		printf("add code page %d to ",codepage);
	else
		printf("replace code page %d in ",codepage);
	printf("%s.\n",argv[2]);

	return 0;
}
