//
//   $Id: FONTEDIT.CPP 1.2 2002/10/15 20:43:38Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
   Copyright (c) 1989-1997 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary
*/
/*

	FontEdit - a simple font editor

*/

#include <conio.h>
#include <ctype.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <stdlib.h>
#include <string.h>

#include "display.h"

const int maxRows = 16;
const int useByte = 255;


#define cgoto(row, col) gotoxy(col+1, row+1)

typedef struct
{
	int codePage;
	int size;
	int filler[6];
} CodePageStruct;

CodePageStruct codePageStruct;

class Font
{
public:
	Font(int whichCodePage) :
				codePage(whichCodePage)
			{
				font8x16 = new unsigned char[256 * 16];
				memset(font8x16, 0, 256*16);
				font8x14 = new unsigned char[256 * 14];
				memset(font8x14, 0, 256*14);
				font8x8 = new unsigned char[256 * 8];
				memset(font8x8, 0, 256*8);

				currentSubFont = font8x16;
			}
	~Font(void)
		{
			delete [256*16] font8x16;
			delete [256*14] font8x14;
			delete [256*8] font8x8;
		}

	int		Read(int fh);
	int		Write(int fh);

	int		Create(int fh, int basedOn);

	int		GetRows(void);

	int		NextSubFont(void)
	{
		if (currentSubFont == font8x16)
			currentSubFont = font8x14;
		else if (currentSubFont == font8x14)
			currentSubFont = font8x8;
		else
			currentSubFont = font8x16;

		return GetRows();
	}

	int	GetCodePage(void)		{return codePage;}

	unsigned char *	GetChar(int offset)
		{return currentSubFont + offset*GetRows();}

	void	Zap(int fontCode);
	void	Copy(int dest, int source);
	void	Shift(int fontCode, int nLines);
	void	Borrow8x16(int fontCode);

	void	Map(int screenCode, int fontCode);
	void	UnMap(int screenCode);

private:
	int		codePage;
	unsigned	char *	font8x16;
	unsigned	char *	font8x14;
	unsigned	char *	font8x8;

	unsigned char *	currentSubFont;
};

int	Font::GetRows(void)
{
	if (currentSubFont == font8x16)
		return 16;
	else if (currentSubFont == font8x14)
		return 14;
	else if (currentSubFont == font8x8)
		return 8;
	else
		return 0;
}

int	Font::Create(int fh, int basedOn)
{
	int retVal = 1;

	// first, read the header
	CPIHeader header;
	lseek(fh, 0L, SEEK_SET);
	read(fh, &header, sizeof(header));
	if (strcmp(header.id, "CPIKBS"))
		return 0;

	// then, read the codePage list
	CodePageList codePageList;
	read(fh, &codePageList, sizeof(CodePageList));

	// read the lower 128 first
	read(fh, font8x16, 128 * 16);
	read(fh, font8x14, 128 * 14);
	read(fh, font8x8, 128 * 8);

	// locate the basedOn font
	for (int i = 0; i < header.maxCodePages; ++i)
		if (codePageList[i] == basedOn)
			break;
	if (i >= header.maxCodePages)
		return 0;

	// read the upper 128 out of the basedOn font
	lseek(fh, header.firstCodePage + (i * header.chunkSize), SEEK_SET);
	read(fh, font8x16 + (128 * 16), 128 * 16);
	read(fh, font8x14 + (128 * 14), 128 * 14);
	read(fh, font8x8 + (128 * 8), 128 * 8);

	// find a spot for the new font
	for (i = 0; i < header.maxCodePages; ++i)
		if (!codePageList[i])
			break;
	if (i >= header.maxCodePages)
		return 0;
	codePageList[i] = codePage;

	// write out updated header and codepagelist
	lseek(fh, 0L, SEEK_SET);
	write(fh, &header, sizeof(header));
	write(fh, &codePageList, sizeof(CodePageList));

	// write upper 128 characters for all three sizes
	long offset = header.firstCodePage + (i * header.chunkSize);
	lseek(fh, offset, SEEK_SET);
	write(fh, font8x16 + (128 * 16), 128 * 16);
	write(fh, font8x14 + (128 * 14), 128 * 14);
	write(fh, font8x8 + (128 * 8), 128 * 8);

	return retVal;
}

int	Font::Read(int fh)
{
	int retVal = 1;

	// first, read the header
	CPIHeader header;
	lseek(fh, 0L, SEEK_SET);
	read(fh, &header, sizeof(header));
	if (strcmp(header.id, "CPIKBS"))
		return 0;

	// then, read the codePage list
	CodePageList codePageList;
	read(fh, &codePageList, sizeof(CodePageList));

	// read the lower 128 first
	read(fh, font8x16, 128 * 16);
	read(fh, font8x14, 128 * 14);
	read(fh, font8x8, 128 * 8);

	// locate our area of interest
	for (int i = 0; i < header.maxCodePages; ++i)
		if (codePageList[i] == codePage)
			break;
	if (codePageList[i] != codePage)
		return 0;
	long offset = header.firstCodePage + (i * header.chunkSize);
	lseek(fh, offset, SEEK_SET);

	// read upper 128 characters
	read(fh, font8x16 + (128 * 16), 128 * 16);
	read(fh, font8x14 + (128 * 14), 128 * 14);
	read(fh, font8x8 + (128 * 8), 128 * 8);

	return retVal;
}

int	Font::Write(int fh)
{
	int retVal = 1;

	// first, read the header
	CPIHeader header;
	lseek(fh, 0L, SEEK_SET);
	read(fh, &header, sizeof(header));
	if (strcmp(header.id, "CPIKBS"))
		return 0;

	// then, read the codePage list
	CodePageList codePageList;
	read(fh, &codePageList, sizeof(CodePageList));

	// write the lower 128 first
	write(fh, font8x16, 128 * 16);
	write(fh, font8x14, 128 * 14);
	write(fh, font8x8, 128 * 8);

	// locate our area of interest
	for (int i = 0; i < header.maxCodePages; ++i)
		if (codePageList[i] == codePage)
			break;
	if (codePageList[i] != codePage)
		return 0;
	long offset = header.firstCodePage + (i * header.chunkSize);
	lseek(fh, offset, SEEK_SET);

	// write upper 128 characters
	write(fh, font8x16 + (128 * 16), 128 * 16);
	write(fh, font8x14 + (128 * 14), 128 * 14);
	write(fh, font8x8 + (128 * 8), 128 * 8);

	return retVal;
}

void	Font::Zap(int fontCode)
{
	int numRows = GetRows();
	unsigned char * p = currentSubFont + fontCode*numRows;
	for (int i=0; i < numRows; ++i)
		*p++ = 0;
}

void	Font::Copy(int dest, int source)
{
	int numRows = GetRows();
	memcpy(currentSubFont + dest*numRows, currentSubFont + source*numRows, numRows);
}

void	Font::Shift(int fontCode, int nLines)
{
	int numRows = GetRows();
	unsigned char newPattern[maxRows];
	unsigned char * source = currentSubFont + fontCode*numRows;

	// create a shifted copy
	for (int i=0; i < numRows; ++i)
	{
		int delta = (i - nLines + numRows) % numRows;
		newPattern[i] = *(source+delta);
	}

	// copy back to original
	for (i=0; i < numRows; ++i)
		*(source + i) = newPattern[i];
}

void	Font::Borrow8x16(int fontCode)
{
	if (currentSubFont != font8x16)
	{
		int numRows = GetRows();

		// point to desired character
		unsigned char * source = font8x16 + fontCode * 16;
		unsigned char * dest = currentSubFont + fontCode * numRows;

		for (int row = 0; row < 16; ++row)
		{
			unsigned char x = *source++;

			if (numRows == 14)
			{
				// copy all but rows 13 and 15 into the 8x14 font
				if (row != 13 && row != 15)
					*dest++ = x;
			}
			else if (numRows == 8)
			{
				// copy rows 2, 4, 5, 7, 8, 10, 11, 13 to 8x8 font
				switch(row)
				{
					case 2:
					case 4:
					case 5:
					case 7:
					case 8:
					case 10:
					case 11:
					case 13:		*dest++ = x;	break;
				}
			}
		}

	}
}

void	Font::Map(int screenCode, int fontCode)
{
	int numRows = GetRows();
	unsigned fontSeg = FP_SEG(currentSubFont);
	unsigned fontOff = FP_OFF(currentSubFont + (fontCode * numRows));

	_AX = 0x1100;
	_BX = 0x1000;
	_CX = 1;
	_DX = screenCode;
	_ES = fontSeg;

	asm	push si;
	asm	push di;
	asm	push bp;
	_BP = fontOff;
	geninterrupt(0x10);
	asm	pop bp;
	asm	pop di;
	asm	pop si;
}

class Button
{
public:
	Button(int aRow, int aCol, char value, char * aText) :
			row(aRow), col(aCol), response(value), len(strlen(aText)+3)
	{
		text = new char[len];
		strcpy(text, aText);
		Redisplay();
	}
	~Button(void) {delete[len] text;}

	void	Redisplay(void)
	{
		cgoto(row, col);
		textattr(79);
		cprintf("[%s]", text);
	}
	int	CheckForPress(int atRow, int atCol)
				{return (atRow == row && atCol >= col && atCol < col+len) ?
					response : -1;}

private:
	int row;
	int col;
	int len;
	char response;
	char * text;
};

Button *	buttonList[20];

int PollKey(void)
{
	union REGS r;
	int result = -1;

	r.h.ah = 6;
	r.h.dl = 0xFF;
	intdos(&r,&r);
	if (r.x.flags & 0x40)		// ZF
		return result;
	if (r.h.al == 0)
	{
		r.h.ah = 6;
		r.h.dl = 0xFF;
		intdos(&r,&r);
		result = (unsigned)r.h.al + 128;
	}
	else
		result = (unsigned)r.h.al;
	return result;
}

int	GetKey(void)
{
	int c = getch();
	if (c == 0)
		c = getch() + 128;
	return c;
}

static union REGS   inregs,outregs;

int	check_mouse(void)
{
   inregs.x.ax = 0;
   int86(0x33,&inregs,&outregs);
   if (!outregs.x.ax)                 /* AX=0 ->no, AX=-1 ->yes         */
      return 0;
   return(outregs.x.bx);              /* BX holds the number of buttons */
}

void	show_mouse(void)
{
   inregs.x.ax = 1;
   int86(0x33,&inregs,&outregs);
}

void	hide_mouse(void)
{
   inregs.x.ax = 2;
   int86(0x33,&inregs,&outregs);
}

int	get_mouse_position(int * p_x, int * p_y)
{
   inregs.x.ax = 3;
   int86(0x33,&inregs,&outregs);
   *p_x = outregs.x.cx;
   *p_y = outregs.x.dx;
   return(outregs.x.bx);
}

int	get_mouse_action(int * p_x, int * p_y)
{
	static int old_buttons;
	int new_buttons;

	new_buttons = get_mouse_position(p_x, p_y);
	if (new_buttons == old_buttons)
		return -1;
	old_buttons = new_buttons;
	return new_buttons;
}

void	Toggle(unsigned char * pattern, int row, int col)
{
	unsigned char thisRow = *(pattern + row);
	unsigned oldValue = (thisRow >> (7-col)) & 1;
	unsigned newValue = (oldValue) ? 0 : 1;
	thisRow &= ~(1 << (7-col));
	thisRow |= newValue << (7-col);
	*(pattern + row) = thisRow;
}

void	Set(unsigned char * pattern, int row, int col)
{
	unsigned char thisRow = *(pattern + row);
	thisRow |= 1 << (7-col);
	*(pattern + row) = thisRow;
}

void	Show(unsigned char * pattern, int numRows)
{
	int zone1 = 2;
	int zone2 = 5;
	int zone3 = 12;
	int zone4 = 14;

	if (numRows == 14)
	{
		zone3 = 11;
		zone4 = 13;
	}
	else if (numRows == 8)
	{
		zone1 = 0;
		zone2 = 2;
		zone3 = 7;
		zone4 = 8;
	}

	for (int row=0; row < numRows; ++row)
	{
		unsigned char thisRow = *(pattern+row);
		cgoto(row, 0);

		if (row < zone1 || row >= zone4)
			textattr(47);
		else if (row < zone2 || row >= zone3)
			textattr(63);
		else
			textattr(31);
		for (int col = 7; col >= 0; --col)
		{
			unsigned char c = ((thisRow >> col) & 1) ? 0xB2 : ' ';
			cprintf("%c", c);
			cprintf("%c", c);
		}
		textattr(7);

//		cprintf("%02X ", thisRow);
	}
}

int	GetMouseClick(int * row, int * col)
{
	int x;
	int y;

	int result = get_mouse_action(&x, &y);
	if (result == -1)
		return -1;

	x /= 8;
	y /= 8;

	*row = y;
	*col = x;

	return result;
}

int GetFrom(char * prompt)
{
	char buffer[80];
	*buffer = 10;		// maxlen
	*(buffer+1) = 0;	// len
	*(buffer+2) = 0;	// text
	cgoto(24, 0);
	textattr(7);
	cprintf("%s what character code?", prompt);
	cgets(buffer);
	int value = atoi(buffer+2);
	cgoto(24,0);
	clreol();
	return value;
}

void ShowChar(int row, int col, unsigned char showByte, unsigned char showChar)
{
	textattr(112);
	cgoto(row, col);
	cprintf("         ");
	cgoto(row+1, col);
	cprintf(" ");
	textattr(7);
	cprintf("%3u", showByte);
	textattr(112);
	cprintf(" ");
	textattr(7);
	cprintf("%c", 255);
	textattr(112);
	cprintf(" ");
	textattr(7);
	cprintf("%c", showChar);
	textattr(112);
	cprintf(" ");
	cgoto(row+2, col);
	cprintf("         ");
}

void	Example(unsigned showByte)
{
	unsigned char showChar = showByte;
	if (showChar == 13 || showChar == 8 || showChar == 10 || showChar == 7)
		showChar = ' ';

	ShowChar(6, 18, showByte, showChar);
	ShowChar(17, 3, showByte, showChar);
}

typedef struct
{
	int	rows;
	int	romFontSubfunction;
	int	scanLinesSubfunction;
} ScreenInfo;

ScreenInfo siTable[] =
{
	{16, 4, 2},
	{14, 1, 1},
	{8, 2, 0},
	{0}
};

void	SetScreenFont(int rows)
{
	union REGS r;

	for (int i=0; siTable[i].rows; ++i)
		if (siTable[i].rows == rows)
			break;

	if (!siTable[i].rows)
		return;

	// set scan lines
	r.h.ah = 0x12;
	r.h.al = siTable[i].scanLinesSubfunction;
	r.h.bl = 0x30;
	int86(0x10, &r, &r);

	// set mode to make sure everything takes
	r.h.ah = 0;
	r.h.al = 3;
	int86(0x10, &r, &r);

	// set font
	r.h.ah = 0x11;
	r.h.al = siTable[i].romFontSubfunction;
	r.h.bl = 0;
	int86(0x10, &r, &r);
}

void	ShowStats(char * fileName, int codePage, int rows, int mouse)
{
	cgoto(0, 60);
 	cprintf("Filename : %s", fileName);
 	cgoto(1, 60);
 	cprintf("Code Page: %u", codePage);
 	cgoto(2, 60);
 	cprintf("Subfont  : 8 x %u", rows);
	cgoto(3, 60);
	cprintf("Mouse    : %s", mouse ? "Yes" : "No");
}

void	RefreshScreen(char * fileName, Font * font, int mouse)
{
	textattr(7);
	clrscr();

	cgoto(0, 30);
	cprintf("FontEdit 1.1");

	cgoto(22, 20);
	cprintf("To push a button, enter the capitalized letter");
	cgoto(23, 20);
	cprintf("The space bar toggles a pixel");

	ShowStats(fileName, font->GetCodePage(), font->GetRows(), mouse);

	for (int i=0; buttonList[i]; ++i)
		buttonList[i]->Redisplay();

	if (mouse)
		show_mouse();
}

void main(int argc, char * argv[])
{
	char fileName[80];

	if(argc > 2)
	{
		strncpy(fileName, argv[1], sizeof(fileName)-1);
		fileName[sizeof(fileName)-1] = '\0';
	}
	else
		strcpy(fileName, FILENAME);



	int codePage = 437;
	if (argc >= 3)
		codePage = atoi(argv[2]);

	// see if the user wants help
	if (argc == 1 || codePage == 0)
	{
		cprintf("Usage: FONTEDIT file codepage [basecodepage]  \r\n");
		cprintf("   Specify the code page you wish to edit\r\n");
		cprintf("   If codepage doesn't exist, you will be\r\n");
		cprintf("     asked if you wish to create it,     \r\n");
		cprintf("     using basecodepage as a pattern     \r\n");
		cprintf("                                         \r\n");
		exit(0);
	}

	Font * font = new Font(codePage);

	int fh = open(fileName, O_RDWR|O_BINARY);
	if (fh == -1)
	{
		cprintf("Error: Can't open %s\r\n", fileName);
		exit(1);
	}

	int result = font->Read(fh);
	if (!result)
	{
		int basedOn = 437;
		if (argc > 3)
			basedOn = atoi(argv[3]);

		cprintf("Can't read CodePage %u\r\n", codePage);
		cprintf("Add CodePage %u (using %u as a base) to %s? (Y/N)",
				codePage, basedOn, fileName);
		int c = getche();
		if (toupper(c) != 'Y')
			exit(1);
		font->Create(fh, basedOn);
	}
	close(fh);

	int numRows = font->GetRows();
	int showByte = 0;

	SetScreenFont(numRows);

	int mouse = 0;
	if (check_mouse())
	{
		mouse = 1;
	}

	int numButtons = 0;

	buttonList[numButtons++] = new Button(5, 30, 'n', "Next char");
	buttonList[numButtons++] = new Button(7, 30, 'p', "Prev char");
	buttonList[numButtons++] = new Button(9, 30, 'j', "Jump to");
	buttonList[numButtons++] = new Button(11, 30, 'g', "Grab from");
	buttonList[numButtons++] = new Button(13, 30, 'z', "Zap char");
	buttonList[numButtons++] = new Button(15, 30, 'e', "Exit");

	buttonList[numButtons++] = new Button(5, 45, 'u', "shift Up");
	buttonList[numButtons++] = new Button(7, 45, 'd', "shift Down");

	buttonList[numButtons++] = new Button(9, 45, 's', "Subfont");
	buttonList[numButtons++] = new Button(11, 45, 'b', "Borrow 8x16");

	RefreshScreen(fileName, font, mouse);

	int cRow = 0;
	int cCol = 0;

	unsigned char * pattern;

	font->Map(255, showByte);
	pattern = font->GetChar(showByte);
	Show(pattern, numRows);
	Example(showByte);
	cgoto(cRow, cCol*2);

	while(1)
	{
		int key = -1;

		int changed = 0;
		if (mouse)
		{
			// mouse stuff
			int buttons;
			int mRow;
			int mCol;
			buttons = GetMouseClick(&mRow, &mCol);

			key = -1;
			if (buttons == 1)
			{
				for (int i=0; key == -1 && i < numButtons; ++i)
					key = buttonList[i]->CheckForPress(mRow, mCol);
			}
			// any click in font area
			if (buttons != -1 && mRow < numRows && mCol < 16)
			{
				cRow = mRow;
				cCol = mCol/2;
				if (buttons == 1)
				{
					Toggle(pattern, cRow, cCol);
					changed = 1;
				}
				else if (buttons == 2)
				{
					Set(pattern, cRow, cCol);
					changed = 1;
				}
			}
		}

		// keyboard stuff
		if (key == -1)
			key = PollKey();

		if (key == 27 || key == 'e' || key == 'E')
			break;
		else switch(key)
		{
			case 'b':
			case 'B':
				font->Borrow8x16(showByte);
				changed = 1;
				break;
			case 's':
			case 'S':
				numRows = font->NextSubFont();
				SetScreenFont(font->GetRows());
				RefreshScreen(fileName, font, mouse);
				changed = 1;
				break;
			case 'n':
			case 'N':
				++showByte;
				changed = 1;
				break;
			case 'p':
			case 'P':
				--showByte;
				changed = 1;
				break;
			case 'z':
			case 'Z':
				{
					int c;
					cgoto(24,0);
					textattr(7);
					cprintf("Are you sure you want to ZAP this character? (Y/N)");
					c = toupper(GetKey());
					if (c == 'Y')
						font->Zap(showByte);
					changed = 1;
					cgoto(24,0);
					clreol();
				}
				break;
			case 'g':
			case 'G':
				{
					int from = GetFrom("Grab from");
					if (from > 0 && from < 256)
						font->Copy(showByte, from);
					changed = 1;
				}
				break;
			case 'j':
			case 'J':
				{
					int from = GetFrom("Jump to");
					if (from >= 0 && from < 256)
						showByte = from;
					changed = 1;
				}
				break;
			case 'u':
			case 'U':
				font->Shift(showByte, -1);
				changed = 1;
				break;
			case 'd':
			case 'D':
				font->Shift(showByte, 1);
				changed = 1;
				break;
			case 200:
				--cRow;
				changed = 1;
				break;
			case 208:
				++cRow;
				changed = 1;
				break;
			case 205:
				++cCol;
				changed = 1;
				break;
			case 203:
				--cCol;
				changed = 1;
				break;
			case ' ':
				Toggle(pattern, cRow, cCol);
				changed = 1;
				break;
#if 0
		else if (key != -1)
		{
			cgoto(19,0);
			cprintf("%u  ", key);
		}
#endif
		}

		// make everything legal
		showByte = (showByte+256)%256;
		cRow = (cRow + numRows)%numRows;
		cCol = (cCol + 8) % 8;

		if (changed)
		{
			hide_mouse();
			font->Map(255, showByte);
			pattern = font->GetChar(showByte);
			Show(pattern, numRows);
			Example(showByte);
			cgoto(cRow, cCol*2);
			show_mouse();
		}
	}

	font->Map(255,255);

	int c;
	do
	{
		cgoto(24,0);
		textattr(7);
		cprintf("Save changes to %s? (Y/N)%c", fileName, 7);
		c = toupper(GetKey());
		if (c == 'Y' || c == 'N')
			break;
	} while (1);
	if (c == 'Y')
	{
		int fh = open(fileName, O_RDWR|O_BINARY);
		if (fh == -1)
			cprintf("Error: Can't open %s\r\n", fileName);
		else
		{
			font->Write(fh);
			close(fh);
		}
	}

	textattr(7);
	cgoto(24,0);

	SetScreenFont(16);

	delete font;
}

