;
;   $Id: himem.asm 1.10 2003/05/20 21:06:38Z thomd Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
COMMENT ~ ------------------------------------------------

   HIMEM.ASM  created 10/05/92  by Drew Gislason

   Copyright (c) 1992-2002 by Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This device driver provides the full XMS implementation.
   All function 0-0Fh are supported.

   The ROM-DOS kernel only uses a minimum (1 and 5) to load
   itself into the High Memory Area (HMA -- the 64K-16 byte
   area starting at 0FFFF:10h).

   XMS 2.0 FUNCTIONS SUPPORTED:
   ----------------------------
    0	(get XMS version)
    1	(allocate HMA)
    2	(free HMA)
    3	(global enable A20)
    4	(global disable A20)
    5	(local enable A20)
    6	(local disable A20)
    7	(query A20)
    8	(query free memory)
    9	(allocate memory)
    0A	(free memory)
    0B	(move memory)
    0C	(lock memory)
    0D	(unlock memory)
    0E	(get handle info)
    0F	(realloc memory)


   HISTORY:
   --------
   03-26-98 (DE):
   1. (SPR 1600) added /QUIET switch to not show sign-on screen
   2. (SPR 1598) Added /BIOS switch to force use of BIOS int 15h
      fn 87h for data transfers to/from XMS.
   3. (SPR 1598) Added detection code for MS-DOS HIMEM.SYS
      /MACHINE:3 which their help file calls the "Phoenix
      Cascade BIOS". Machine type three is now a valid option.
      This is magic code (see comments on LookForPTL proc).

   01-12-98 (DE):
   1. (SPR 1549) find_next was not reporting correct free memory
      when all XMS memory was in use and a 0 length handle was
      allocated. Made find_next return xms_limit when 0 XMS is
      free. This means the startK field of all handles with 0K XMS
      now have the address of the end of XMS memory.

   11-24-97 (DE):
   1. Wasn't treating freeXMS=0 as an error when allocRequest > 0
      in allocate_memory. Now check if freeXMS <= allocRequest
      then if not and freeXMS==0 then that is an error condition.

   11-09-97 (DE):
   1. Deferred hooking int 15h until first XMS "non-version number" call.
   2. Added "/HMAMIN=n" option, n in {0..63K}, default is 0K.
   3. Added "/MAXHANDLES=n" option, n in {0..128}, default is 32 handles.

   11-06-97 (DE):
   1. Fixed A20 functions so that mixed local/global calls work
      like MS'.
   2. Fixed bug in move call (B) handler so "invalid length" error
      codes are returned
   3. Changed "revision number" to be ROM-DOS release number
      (ie. 250h for 2.50.11)
   4. Changed allocate call (9) error code logic so that 0 length
      allocation calls would succeed so long as free handles were
      available, even when there was no XMS memory left to allocate.
      MS' XMSTIME program complains if we can't do this.

   04-17-97 (DE):
   1. Turned off interrupts forever to satisfy Borland 32RTM.EXE v1.5
   2. Added stubs for int 2F 4308h/4309h and removed UMB stubs

   ...(DE)
   -- Added "real big mode" (a persistent feature of the '386+ that
      leaves segment limits unchanged when switching from protected
      to real mode and so allows addressing of 4Gbytes directly from
      real mode)
   -- fixed various bugs

   02/26/96: (RME)
   1. Added /A20, /A20+, /PS2 command line options
   2. Changed default behavior to ignore all waiting for A20 settle.
      Use /A20+ to use the NULL output command to force a wait.
   10/05/94: (DAG)
   1. Added support for PS/2 A20 line support
   05/06/94: (KBS)
   1. add 1000h to driver revision number so Windows 3.1
      will recognize it as a valid XMS 2.0 driver (it
      checks for version >= 228h) (SPR #277)
   05/05/94: (KBS)
   1. now preserve BH during move operations (SPR #274)
   2. implement true local/global A20 operations (SPR #253)
   3. make copy block size an EQU, and change from
      32k to 1k at a time for better interrupt
      processing (SPR #278)
   10/18/93: (DAG)
   1. fixed move to work w/ blocks > 128K
   2. added realloc
   9/29/93: (DAG)
   1. Added all XMS functions 0-11H
   6/27/93: (CEM)
   1. Added new 8042 command to function "_control_A20"
      which allows time for gate A20 to settle to new
      state. This eliminates the need for a memory-wrap
      test at each invocation of func 5 (or 6).
   2. Added new XMS function "local disable A20" to XMS
      client I/F, (XMS func 6).


---------------------------------------------------------- ~
NOSMART

; ENABLE_PROGRAM_PROFILE	equ	1
include profile.inc
include dosver.inc

FN0B_DEBUG		=	0	; for debugging delays in fn 0Bh

XMSVERSION	equ	200H
COPYBYTES	equ	400H	; copy block size (must be <= 8000H)
SIGNON		equ	1	; set to FALSE for no signon

DEFAULT_HANDLES_COUNT	=	32d
MAX_HANDLES_COUNT	=	128d

MAX_HMA_MIN_VALUE	=	63d
DEFAULT_HMA_MIN_VALUE	=	0

; we use the following INT 15H functions
INT15_87	equ	87H	; copy extended memory
INT15_88	equ	88H	; get extended memory size
INT15_C0	equ	0C0H	; get configuration

; The MS-DOS HIMEM.SYS /MACHINE:n types that we support, A20PLUS is our
; default.
;
; Note: A20PLUS waits for NULL command to be processed, A20AT does not.
;
UNKNOWN_AT 	equ	0
A20AT		equ	1
A20PS2		equ	2
A20PTL		equ	3	; do NOT document this switch
A20PLUS		equ	-1


;--------------------------------
;  When we are COM, use ORG 100H
;
	ORG	100H


_TEXT	SEGMENT PARA PUBLIC 'CODE'
	assume CS:_TEXT
	assume DS:nothing

.186
	; BT 6-15-99
	;
	; In order to use TLINK 7.1 we must declare an entry point.  This
	; is not an entry point.  Comparing the SYS file created using
	; TLINK 4.0 without an entry point and the SYS file created using
	; TLINK 7.1 revealed no differences.
	;
falsestart:

;/////////////////////////////////////////////////////////////////////////
;//	Device Driver Header						//
;/////////////////////////////////////////////////////////////////////////
	public	_himemx
_himemx:
	dd	-1		; link to next device
	dw	8000H		; character device
	dw	hma_strat	; strategy routine
	dw	hma_serve	; service routine
	db	"XMSXXXX0",0

	; A20 type set on the command line (1=AT, 2=PS/2, 3=A20+)
	;
		public	a20_type
	a20_type	db	UNKNOWN_AT

	; allows us to install if the a20 line is enabled and we
	; can't disable it...
	;
		public	controllingA20
	controllingA20	db	1


	; 486 caching instructions available? (/486)
	cache_486	db	0

	public	hma_alloced, hma_exists, globalA20, localA20count
	public  old_15, hooked_int_15h

	; data area (accessed through code segment)
	ptrsav	dd	?	; device request header
	old_2F	dd	?	; for chaining INT 2F
	old_15	dd	?	; for chaining INT 15
	hma_alloced db	0	; has hma been allocated?
	hma_exists  db	1	; note: should eventual detect this
	globalA20 db	0	; store global A20 state
	localA20count db 0	; nested local A20 enable count

	; Set to 1 while MS Windows is running. Forces moves to
	; XMS using int 15h vs. direct real mode acces. Avoids hang.
	;
		public	windozed
	windozed	db	0

	public	hma_strat, hma_serve, hma_init
hma_strat:
	mov	word ptr CS:[ptrsav], BX
	mov	word ptr CS:[ptrsav+2], ES
	retf

; Request Header Macros
cmdlen	=	00H	; length of this command
unit	=	01H	; sub unit specifier
cmdc	=	02H	; command code
status	=	03H	; status
endaddr	=	0EH	; end address

hma_serve:
	pushf
	push	AX
	push	BX
	push	CX
	push	DX
	push	SI
	push	DI
	push	DS
	push	ES

	; make sure this is the INIT command (the only one we can handle)
	lds	BX, CS:[ptrsav]		; get pointer to i/o packet
	cmp	byte ptr [BX.cmdc], 0	; command code
	jnz	cmderr
	jmp	hma_init

cmderr:	mov	AL, 3		; unknown cmd error
err_exit:
	mov	AH, 81H		; mark error return
	jmp	short err1

exit:	mov	AH, 01H
err1:	lds	BX, CS:[ptrsav]
	mov	word ptr [BX.status], AX	; mark as complete

	pop	ES
	pop	DS
	pop	DI
	pop	SI
	pop	DX
	pop	CX
	pop	BX
	pop	AX
	popf
	retf


;/////////////////////////////////////////////////////////////////////////
;//	Interrupt 2FH, Function 43xxH Handler			   //
;/////////////////////////////////////////////////////////////////////////
	public	int_2f_handler
int_2f_handler:

	; use "real big mode" to access eXtended memory directly
	; when no protected mode programs are running
	;
	cmp	ax, 1606h
	jne	not_windows_exit
	mov	byte ptr cs:windozed, 0

	; use BIOS to control A20 line when windows is running
	;
not_windows_exit:
	cmp	ax, 1605h
	jne	not_windows_startup
	mov	byte ptr cs:windozed, 1

not_windows_startup:
	cmp	AH, 43H
	jnz	do_old_2F

	; sub-function 0 returns 80H in AL to indicate driver installed
	;
	cmp	AL, 0
	jnz	try_2f_4310
	mov	AL, 80H
	iret

	; sub-functions 10h returns XMS driver entry point in ES:BX
	;
try_2f_4310:
	cmp	AL, 10H
	jnz	try_2f_4309
   mov   BX, CS
   mov   ES, BX
	mov	BX, offset xms_handler
	iret

	; UNSUPPORTED sub-function 9 returns handle array info
	; See PC Interrupts, 2nd ed. pg 798
	;
try_2f_4309:
	cmp	al, 09h
	jnz	try_2f_4308
	iret

	; UNSUPPORTED sub-function 8 returns /MACHINE= type info
	; See PC Interrupts, 2nd ed. pg 798
	;
try_2f_4308:
	cmp	al, 08h
	jnz	do_old_2F
	iret

do_old_2F:
	jmp	dword ptr CS:[old_2F]


;/////////////////////////////////////////////////////////////////////////
;//	Interrupt 15H, Function 88xxH Handler			   //
;/////////////////////////////////////////////////////////////////////////
	public	int_15_handler
int_15_handler:
	cmp	AH, 88H
	je	do_int15
	jmp	dword ptr CS:[old_15]

do_int15:
	xor	AX, AX		; no extended memory (we're using it all)
	iret


;/////////////////////////////////////////////////////////////////////////
;//	XMS (High Memory Area) Handler				  	//
;//									//
;//	This handler is called (a far call) from an application.  It	//
;//	returns AX=0 on failure.					//
;/////////////////////////////////////////////////////////////////////////

ORGHANDLE	equ	0A8FEH	; handle0
HANDLEINC	equ	10	; handle1 = handle0 + 12

; the internal handle structure, unique to Datalight
XMSHANDLE	STRUC
	hndl_inuse	db 0	; is this handle in use
	hndl_locked	db 0	; is this locked?
	hndl_startK	dw 0	; start in K bytes
	hndl_sizeK	dw 0	; size in K bytes
XMSHANDLE	ENDS

;  Move block structure (XMS call 0BH)
;  The user passes us this structure in DS:SI
;
;  Handle may be 0 (DOS 1 Meg area) or a previously
;  allocated handle from XMS.
;
;  Addresses are either seg:off for handle 0, or a
;  32 bit linear
;
XMSCOPYBLK	STRUC
	xblk_count	dd ?	; 00, count in bytes
	xblk_src_handle	dw ?	; 04, source handle
	xblk_src_addr	dd ?	; 06, source address
	xblk_dst_handle	dw ?	; 0A, destination handle
	xblk_dst_addr	dd ?	; 0C, destination address
XMSCOPYBLK	ENDS

GDT	STRUC
	gdt_seg_limit	dw ?		; 00, contains FE00H
	gdt_addr	db 3 dup (?)	; 02, address
	gdt_data_access	db ?		; 05, contains 93H
	gdt_res		dw ?		; 06, unknown
GDT	ENDS

BIOSCOPYBLK	STRUC
	bblk_res	db 16 dup (?)	; 00, for BIOS use
	bblk_src_gdt	GDT <?>		; 10, source descriptor
	bblk_dst_gdt	GDT <?>		; 18, destination descriptor
	bblk_res2	db 19 dup (?)	; 20, for BIOS use
BIOSCOPYBLK	ENDS

; XMS errors (in BL)
;
NO_XMS_ERR		equ  00h
ERR_BADFUNC		equ  80H
ERR_VDISK_FOUND 	equ  81h
ERR_A20_ERR     	equ  82h
ERR_GENERAL		equ  8Eh
ERR_UNRECOVERABLE 	equ  8Fh
ERR_NOHMA		equ  90H
ERR_HMAINUSE	        equ  91H
ERR_LT_HMA_MIN	        equ  92h
ERR_HMANOTINUSE	        equ  93H
ERR_A20ENABLED	        equ  94H
ERR_NOFREE	        equ 0A0H
ERR_NOHANDLES	        equ 0A1H
ERR_BADHANDLE	        equ 0A2H
ERR_BADSRCHDL	        equ 0A3H
ERR_BADSRCADDR	        equ 0A4H
ERR_BADDSTHDL	        equ 0A5H
ERR_BADDSTADDR	        equ 0A6H
ERR_BADLEN	        equ 0A7H
ERR_OVERLAP	        equ 0A8H
ERR_PARITY	        equ 0A9H
ERR_NOTLOCKED	        equ 0AAH
ERR_LOCKED	        equ 0ABH
ERR_LOCK_OVERFLOW	equ 0ACh
ERR_LOCK_FAILED	        equ 0ADh
ERR_ONLY_SMALLER_UMB	equ 0B0h
ERR_NO_UMBS		equ 0B1h
ERR_BAD_UMB_SEG		equ 0B2h


	public	xms_0, xms_1, xms_2, xms_3, xms_4, xms_5, xms_6, xms_7
	public	xms_8, xms_9, xms_A, xms_B, xms_C, xms_D, xms_E, xms_F

	public	local_enable_A20, local_disable_A20, A20_mem_test
	public	real_mode_mover, GPF_ISR, setup_descriptors

	public  xms_func_array
xms_func_array	label	word
	dw	offset xms_0
	dw	offset xms_1
	dw	offset xms_2
	dw	offset xms_3
	dw	offset xms_4
	dw	offset xms_5
	dw	offset xms_6
	dw	offset xms_7
	dw	offset xms_8
	dw	offset xms_9
	dw	offset xms_A
	dw	offset xms_B
	dw	offset xms_C
	dw	offset xms_D
	dw	offset xms_E
	dw	offset xms_F

END_OF_XMS_FUNC_ARRAY	= $
;------------------------
;  XMS Function dispatch
;------------------------
org	0CEH
XMS_HANDLER_ENTRY_POINT = $

IF (END_OF_XMS_FUNC_ARRAY GT XMS_HANDLER_ENTRY_POINT)
  .ERR
  %OUT ERROR: OVERWRITING XMS FUNCTION ARRAY WITH XMS FUNCTION DISPATCHER
ENDIF
	public	xms_handler
xms_handler:
	jmp	short skip_nops
	nop
	nop
	nop

	public skip_nops
skip_nops:
	pushf
	push	CX
	push	DI

	; Borland's 32RTM.EXE v1.5 cannot tolerate any IRQ's going
	; off while it is calling HIMEM. For any A20/HMA functions,
	; leave interrupts OFF forever or the machine will crash.
	; This is the same defensive approach taken by MS.
	;
	cli
	inc	cs:stackSemaphore
	jnz	xms_entry

	mov	cs:callers_SS, ss
	mov	cs:callers_SP, sp

	mov	sp, cs
	mov	ss, sp
	mov	sp, offset localStackTop

	; we ONLY support functions 00-0FH
	;
	public	xms_entry
xms_entry:
	cmp	ah, 0FH
   ja    _xms_not_supported
;	jbe	xms_xx
;	jmp	xms_not_supported

	public	xms_xx
xms_xx:
	xchg	AH, AL
	xor	AH, AH
	or	ax, ax
	je	xms_do_not_hook_int15h

	; XMS 2.0 spec says:
	;
	; In order to maintain compatibility with existing
	; device drivers, DOS XMS drivers must not hook INT 15h
	; until the first non-Version Number call to the control
	; function is made.
	;
;TD
IF 0
	cmp	byte ptr cs:hooked_int_15h, 0
   jne   xms_do_not_hook_int15h
ENDIF

	call	HookInt15h

xms_do_not_hook_int15h:
	shl	AX, 1		; each funcptr is a word
	xchg	DI, AX
	cld			; make sure direction is forward
	jmp	CS:[DI+xms_func_array]

_xms_not_supported:
   jmp   xms_not_supported

stackSemaphore  db	-1

EVEN
callers_SS	dw	?
callers_SP	dw	?
localStack	db	1024 dup(?)
localStackTop	label	word

IF FN0B_DEBUG

BPUTCH  macro	char
	push	ax
	mov	al, char
	call	bputc
	pop	ax
endm

NEWLINE	macro
	push	ax
	mov	al, 0Dh
	call	bputc
	mov	al, 0Ah
	call	bputc
	pop	ax
endm

bputc	proc	near
	push	bx
	push	bp

	mov	ah, 0Eh
	mov	bh, 0
	int	10h

	pop	bp
	pop	bx
	retn
bputc	endp

ELSE

BPUTCH	macro
endm
NEWLINE macro
endm

ENDIF

	;------------------------------------
	; XMS function 00H,  get XMS version
xms_0:	sti				; ints should be OK here
	mov	AX, XMSVERSION
	mov	BX, VERSIONVAL
	xor	DX, DX
	mov	DL, CS:[hma_exists]
	jmp	xms_done

	;------------------------------------
	; XMS function 01H,  allocate HMA
xms_1:	cmp	cs:[hma_exists], 0
	jnz	xms1a
	mov	bl, ERR_NOHMA
	jmp	xms_error
xms1a:	cmp	cs:[hma_alloced], 0
	jz	xms1b
	mov	bl, ERR_HMAINUSE
	jmp	xms_error
xms1b:	cmp	dx, cs:hma_min
	jae	xms1c
	mov	bl, ERR_LT_HMA_MIN
	jmp	xms_error
xms1c:	inc	cs:[hma_alloced]
	jmp	xms_success

	;------------------------------------
	; XMS function 02H,  free HMA
xms_2:	cmp	CS:[hma_alloced], 0
	jnz	xms2a
	mov	BL, ERR_HMANOTINUSE
	jmp	xms_error
xms2a:	mov	CS:[hma_alloced], 0
	jmp	xms_success

	;------------------------------------
	; XMS function 03H, global enable A20
	; code derived from "DOS Internals" p. 389
xms_3:	cmp	CS:[globalA20], 1	; if already global enabled,
	jz	xms_3b			; we're done
xms_3a:
	mov	CS:[globalA20], 1
	jmp short xms_5			; do a local enable
xms_3b:
	jmp	xms_success

	;------------------------------------
	; XMS function 04H, global disable A20
	; code derived from "DOS Internals" p. 389
xms_4:	cmp	CS:[globalA20], 0	; if already global disabled,
	jz	xms_4b			; we're done

   xms_4a:
   	call	local_disable_A20
	or	bl, bl
	jz	xms_4ok
	jmp	xms_error

   xms_4ok:
	mov	CS:[globalA20], 0

   xms_4b:
	jmp	xms_success

	;------------------------------------
	; XMS function 05H, local enable A20
	; code derived from "DOS Internals" p. 389
	;
local_enable_A20	proc	near

	cli
	call	enable_a20
	inc	CS:[localA20count]
	retn
local_enable_A20	endp


xms_5:	call	local_enable_A20
	jmp	xms_success

	;------------------------------------
	; XMS function 06H, local disable A20
	; code derived from "DOS Internals" p. 389
local_disable_A20	proc	near
	cli
	cmp	CS:[localA20count], 1	; check if nested local enable
	jae	xms_6b
	call	disable_a20		; always disable if count==0
	mov	bl, ERR_A20_ERR
	jmp	short xms_6e

    xms_6b:
    	ja	xms_6c			; if >1 nests, leave enabled
	call	disable_a20		; if last nested, disable
	jmp	short xms_6d

    xms_6c:
	call	enable_a20		; nested local enable->reenable

    xms_6d:
        dec	CS:[localA20count]	; decrement nesting count
	mov	bl, 0

    xms_6e:
        retn
local_disable_A20	endp


xms_6:	call	local_disable_A20
	or	bl, bl
	jz	xms_6ok
	jmp	xms_error

    xms_6ok:
	jmp	xms_success

; enable the HMA so FFFF:10 is different than 0:0
;
enable_a20:
	cli
	mov	ax, 1

; Set a20 gate as indicated in AX: 1 = turn on, 0 = turn off.
;
do_a20:
   ENTER_FUNCTION

;  BT 3-29-02
;
;  This was added as a performance optimization.  The goal is to determine
;  the current A20 state before calling to explicitly set the state and if
;  the state is already correct then the control A20 call has been serviced.
;
IF 1
	; Determine if the state is already set to the appropriate state
	;
	push	cx
	push	ax
	call	is_a20_enabled
	mov	cx, ax
	pop	ax
	cmp	al, cl
	pop	cx
	je	do_a20_exit
ENDIF

	cmp	byte ptr cs:a20_type,A20PS2
	jl	normal_a20
	je	PS2_a20

	; the signed value of the a20_type variable is less GT 2,
	; if it is NOT ATPTL, we have a a config problem...
	;
	cmp	byte ptr cs:a20_type,A20PTL
	jne	normal_a20			; should we ASSERT here???

	; A20PTL (3)
	; This code jumps blindly into the top of the BIOS ROM. The
	; entry point address is hard coded in the MS HIMEM.SYS code and there
	; is no known documentation on this method of A20 control.
	;
	; AX=801 turn on, AX= 800 turn off, AX=802 check status
	;
PTL_a20:
	push	ax
	mov	ah, 8
	db	09Ah				; far call immed op code
	dw	0FF82h				; into magic hard coded
	dw	0F000h				; entry pt used by MS HIMEM
	pop	ax
	jmp	short flush_cache_486

	; A20PS2 (2)
	;
PS2_a20:
	push	ax
	call	_control_PS2_A20
	pop	ax
	jmp	short flush_cache_486

	; A20AT (1) or A20PLUS (-1)  (0 is unknown type, which is an invalid
	; value for the a20_type by the time this code executes but will, like
	; all other a20_type values resolve to A20AT incidentally)
	;
normal_a20:
	push	ax
	call	_control_A20
	pop	ax

flush_cache_486:
	cmp	byte ptr cs:cache_486, 0	; Flush 486+ write
	jz	do_a20_exit

	db	0fh, 09h			; 486 instruction
						; WBINVD = 0F 09
do_a20_exit:
   EXIT_FUNCTION
	ret


; disable the HMA so FFFF:10 is the same as 0:0
;
disable_a20:
	cli
	xor	ax, ax
	cmp	cs:controllingA20, al
	je	do_a20_exit
	jmp	short do_a20


	;------------------------------------
	; XMS function 07H, query A20
xms_7:	call	is_a20_enabled
	or	AX, AX
	mov	BL, 0		; no error, just not enabled
	jnz	xms_7a		; AX = 0, HMA disabled
	jmp	xms_done
xms_7a:
	jmp	xms_success


; Returns AX=1 (enabled), AX=0 (disabled)
;
is_a20_enabled:
   ENTER_FUNCTION

	cmp	byte ptr cs:a20_type,A20PTL
	jne	test_mem_contents

	; A20PTL (3)
	; This code jumps blindly into the top of the BIOS ROM. The
	; entry point address is hard coded in the MS HIMEM.SYS code and there
	; is no known documentation on this method of A20 control.
	;
	; AX=801 turn on, AX= 800 turn off, AX=802 check status
	;
	mov	ax, 802h
	db	09Ah				; far call immed op code
	dw	0FF82h				; into magic hard coded
	dw	0F000h				; entry pt used by MS HIMEM
	xchg	al, ah				; status returned in ah
   EXIT_FUNCTION
	ret					; 1= on, 0= off

    test_mem_contents:
	cli
	push	SI
	push	DS
	push	ES
	xor	SI, SI
	mov	DS, SI		; Compare 0:0H
	mov	DI, 0FFFFH
	mov	ES, DI
	mov	DI, 10H		; with FFFF:10H
	mov	CX, 10H
	cld
	repz	cmpsb		; if the same, HMA disabled
	pop	ES
	pop	DS
	pop	SI
	jnz	A20_enabled	; not the same, HMA enabled
	xor	AX, AX		; A20 disabled
   EXIT_FUNCTION
	ret

A20_enabled:
	mov	AX, 1		; A20 enabled
   EXIT_FUNCTION
	ret

	;------------------------------------
	; XMS function 08H, query free memory
xms_8:	sti			; ints should be OK here
	xor	DI, DI		; no largest block
	xor	DX, DX		; no free space
	call	first_free
xms8a:	or	CX, CX
	jz	xms8c
	add	DX, CX		; tally all free space
	cmp	DI, CX
	jae	xms8b
	mov	DI, CX		; we found a bigger free block
xms8b:	call	next_free
	jmp	xms8a

	; we've found all the free space we're gonna find
xms8c:	or	DX, DX
	jnz	xms8d
	mov	BL, ERR_NOFREE
	jmp	xms_error
xms8d:	xchg	DI, AX		; move size of largest block into AX
	mov	BL, 0
	jmp	xms_done

;----------------------
;  Find a free block of memory.
;
;  On Exit
;	AX = start (in K from 1 Meg + 64K) of 1st free block
;	CX = 0 on error, else size (in K) of 1st free block
	public	first_free
first_free	proc	near

	; first free cannot change any registers but AX, CX
	push	DS
	push	DI
	push	BX
	push	DX

	; DS = CS so we don't have to segment override everything
	push	CS
	pop	DS

	; find all active handles
	xor	CX, CX		; counter up to cs:totalHandles
	xor	AX, AX		; # of active handles found
	mov	DI, cs:pXMSHandles
	mov	BX, offset CS:xms_index
ff1:	cmp	DS:[DI.hndl_inuse], 0
	jz	ff2
	inc	AX			; one more ACTIVE handle
	mov	DS:[BX], DI		; add handle to index list
	inc	BX
	inc	BX
ff2:	add	DI, size XMSHANDLE
	inc	CX			; on to next handle
	cmp	CL, byte ptr cs:totalHandles
	jb	ff1
	mov	DS:[xms_num_index], AX	; # of active handles

	; sort active handles into order
	xor	CX, CX			; i = 0

	; for(i=0; i<xms_num_index; ++i)
ff4:	cmp	CX, DS:[xms_num_index]
	jae	ffdone			; done sorting
	mov	DX, CX			; j = i+1
	inc	DX

	; for(j=i+1; j<xms_num_index; ++j)
ff5:	cmp	DX, DS:[xms_num_index]
	jae	ff6

	; if(xms_index[j]->hndl_startK < xms_index[i]->hndl_startK)
	mov	BX, CX
	shl	BX, 1
	mov	BX, word ptr DS:[BX+xms_index]
	mov	AX, word ptr DS:[BX.hndl_startK]	; startK[i]
	mov	BX, DX
	shl	BX, 1
	mov	BX, word ptr DS:[BX+xms_index]
	cmp	AX, word ptr DS:[BX.hndl_startK]	; startK[j]
	jbe	noswap

	; swap index entries
	mov	BX, CX
	shl	BX, 1
	mov	AX, word ptr DS:[BX+xms_index]
	mov	BX, DX
	shl	BX, 1
	xchg	AX, word ptr DS:[BX+xms_index]

	mov	BX, CX
	shl	BX, 1
	mov	word ptr DS:[BX+xms_index], AX

noswap:	inc	DX	; ++j
	jmp	ff5
ff6:	inc	CX	; ++i
	jmp	ff4

	; done sorting, restore registers
ffdone:	pop	DX
	pop	BX
	pop	DI
	pop	DS

	; find the free space
ff10:	mov	CS:[xms_current_index], -1
	call	next_free
	ret
first_free	endp


;---------------------------------
;  Find the next free block of memory.
;
;  On Exit
;	AX = start (in K from 1 Meg + 64K) of free block
;	CX = 0 on error, else size (in K) of free block
	public	next_free
next_free	proc	near
   ;
   ;	{
   ;	   struct XMSHANDLE * hp;
   ;	   int cur_start, prev_end;
   ;
	push	DS
	push	SI
	push	DI
	push	DX
	push	BX

	; DS = CS so we don't have to segment override everything
	push	CS
	pop	DS

   ;	   // look for first available free space
   ;	   while(xms_current_index < xms_num_index)
   ;	   {
	jmp	short @1@242	; while loop
@1@50:
   ;	      // find end of previous item
   ;	      if(xms_current_index == -1)
   ;	         prev_end = 0;
   ;
	cmp	word ptr DS:xms_current_index, -1
	jnz	@1@98
	xor	DI, DI			; DI = prev_end
	jmp	short @1@122
@1@98:
   ;	      else
   ;	      {
   ;	         hp = xms_index[xms_current_index];
   ;	         prev_end = hp->hndl_startK+hp->hndl_sizeK;
   ;	      }
   ;
	mov	BX, DS:xms_current_index
	shl	BX, 1
	mov	SI, DS:xms_index[bx]
	mov	AX, word ptr DS:[SI+2]
	add	AX, word ptr DS:[SI+4]
	mov	DI, AX			; DI = prev_end
@1@122:
   ;	      ++xms_current_index;
   ;
	inc	word ptr DS:xms_current_index
   ;
   ;	      // find current start
   ;	      if(xms_current_index >= xms_num_index)
   ;	         cur_start = xms_limit;
   ;
	mov	AX, word ptr DS:xms_current_index
	cmp	AX, word ptr DS:xms_num_index
	jb	@1@170
	mov	DX, DS:xms_limit		; DX = cur_start
	jmp	short @1@194
@1@170:
   ;	      else
   ;	      {
   ;	         hp = xms_index[xms_current_index];
   ;	         cur_start = hp->hndl_startK;
   ;	      }
   ;
	mov	BX, word ptr DS:xms_current_index
	shl	BX, 1
	mov	SI, word ptr DS:xms_index[bx]
	mov	DX, word ptr [SI.hndl_startK]	; DX = cur_start
@1@194:
   ;	      // found some free space!
   ;	      if(cur_start > prev_end)
   ;	      {
   ;	         start = prev_end;
   ;	         return (cur_start-prev_end);
   ;	      }
   ;
	cmp	DX, DI
	jbe	@1@242
	mov	AX, DI		; return start in AX
	mov	CX, DX		; and length in CX
	sub	CX, DI
	jmp	short @1@290
@1@242:
	mov	AX, DS:xms_current_index
	cmp	AX, 0FFFFH
	je	@1@50
	cmp	AX, DS:xms_num_index
	jb	@1@50
   ;
   ;	   }
   ;	   return 0;		// size 0, not found
   ;
	; Must locate 0 length blocks at the top of XMS memory for
	; the find_next code to work when empty handles are allocated
	;
	mov	ax, ds:xms_limit
	xor	CX, CX
@1@290:
   ;	}
   ;
	pop	BX
	pop	DX
	pop	DI
	pop	SI
	pop	DS
	ret
next_free	endp

;----------------------------------------------
;  Do we have free memory at a specific place?
;  On Entry
;	word on stack = start of free block we want
;  On Exit
;	CX = size (if found), 0 if not found
;
	public	match_free
match_free	proc	near
	push	BP
	mov	BP, SP
	call	first_free
mf0:	or	CX, CX
	jz	mf9		; no more free areas, give up
	cmp	AX, word ptr [BP+4]
	je	mf9		; found!
	ja	mf8		; we've passed it, not free
	call	next_free
	jmp	short mf0
mf8:	xor	CX, CX		; failed to find it
mf9:	pop	BP
	ret
match_free	endp

;----------------------------------------
;  On Entry
;	DX    = Kbytes we want to allocate
;	CS:DI = ptr to handle
;  On Exit
;	CX=size of free blk, AX=start, CX=0 if not enough memory
;
	public	allocate_memory
allocate_memory	proc	near

	; get some freespace
	call	first_free	; returns size in CX, start in AX

	; is free space big enough?
	;
am1:	cmp	CX, DX
	jae	am2		; yes (allow allocating 0 bytes to a handle)

	or	CX, CX		; any more free space at all?
	jz	am9		; nope! quit looking.

	call	next_free	; returns size in CX, start in AX
	jmp	am1		; no, try again

	; allocate the memory
am2:	inc	byte ptr CS:[DI.hndl_inuse]
	mov	word ptr CS:[DI.hndl_startK], AX  ; start of freespace
	mov	word ptr CS:[DI.hndl_sizeK], DX   ; size requested
am9:	ret
allocate_memory	endp


	;------------------------------------
	; XMS function 09H, allocate memory
xms_9:
	sti				; ints should be OK here

	; look for a free handle
	xor	CX, CX
	mov	DI, cs:pXMSHandles
	jmp short xms9a1

xms9a:	cmp	byte ptr CS:[DI.hndl_inuse], 0
	jz	xms9b
	inc	CL
	add	DI, size XMSHANDLE
xms9a1:
	cmp	CL, byte ptr cs:totalHandles
	jb	xms9a
	mov	BL, ERR_NOHANDLES
xms9err:
   ; error condition, null the returned handle also
   ;
   xor   DX, DX
	jmp	xms_error

	; got free handle, make "magic" number
xms9b:	mov	AL, CL
	mov	AH, HANDLEINC
	mul	AH
	add	AX, ORGHANDLE
	push	AX		; save "magic" handle #

	; find enough free space
	call	allocate_memory
	cmp	cx, dx		; if alloc is big enough, that's OK, 0 or not
	jae	xms9e

	pop	AX		; pop "magic" handle off stack
	mov	BL, ERR_NOFREE	; largest free (CX) < Kbytes requested (DX)
   jmp short xms9err

	; got the free space, return handle
xms9e:	pop	DX		; place "magic" handle # in DX
	jmp	xms_success


	;------------------------------------
	; XMS function 0AH, free memory
xms_A:
	sti				; ints should be OK here
	call	checkhandle
	cmp	byte ptr CS:[DI.hndl_locked], 0
	jz	xmsA0
	mov	BL, ERR_LOCKED	; handle is locked, can't free
	jmp	xms_error
xmsA0:	mov	byte ptr CS:[DI.hndl_inuse], 0	; free the handle
	mov	BH, 0A8H	; kludge: FOXPRO needs this
	jmp	xms_success


comment ~

States corresponding to BIOS VIDEO debugging output:

CR,LF,0	entry into XMS function 0Bh, "copy memory block"
1	preliminary set up done, about to call the "copy a block routine"

!	are not in protected
b	using BIOS for move (if '!' and 'b' then think windows is running)
	you should never see 'b' with any combination of {['r'-'z'],['f'-'h']}

r	using "real big mode" for move
s	have entered real mode move code, calling local enable A20
t	verify A20 is on now loop
u	A20 verified on, ready to do the move
v	installing our GPF handler
w	beginning actual data move

f	GP fault/ IRQ 5 handler entered
g	cs:ip on stack is from HIMEM mem move code
h	segment limits re-established, back in real mode, returning to fault

x	actual data move complete
y	calling local disable of A20
z	returning from real mode move code

A	returned from the actual "copy a block proc"
B	errors evaluated, original A20 state restored, exiting xms function

end_comment ~

	;------------------------------------
	; XMS function 0BH, copy memory
xms_B:
	sti				; ints should be OK here
   ENTER_FUNCTION
	push	DS
	push	ES
	push	SI
	push	DX
	push	BX

	NEWLINE
	BPUTCH	'0'

	; save state of HMA
	call	is_a20_enabled
	push	AX		; save A20 state on the stack

	; make local copy of XMS Block
   sub   SP, size XMSCOPYBLK
   mov   CX, (size XMSCOPYBLK) / 2
	mov	AX, SS
	mov	ES, AX
	mov	DI, SP
	cld
	rep	movsw
	mov	AX, SS
	mov	DS, AX
	mov	SI, SP		; now DS:SI points to our local copy

	; odd lengths not allowed
	test	word ptr DS:[SI.xblk_count], 1
   jnz short BadMoveLength

;TD
IF 1
   ; first check the SOURCE address for various validities.  Register Usage:
   ;
   ;  DS:SI      --> XMSCOPYBLK structure
   ;  CS:DI != 0 --> XMS mode block
   ;     DI == 0 --> real mode flag
   ;  DS:BX      --> XMS handle block
   ;  DX:AX      --> limit (hndl_sizeK or 1Meg + 64K - 10h)
   ;
   lea   BX, [SI.xblk_src_handle]
   mov   DX, word ptr DS:[BX]
   or    DX, DX               ;if zero, this is a real-mode address
   jz    xmsB1

   ; XMS source address - we need to check the handle and load the limits
   ;
	BPUTCH	'X'
   call  handle2ptr
   or    DI, DI
   jz short BadHandle

   ; load the limit (DX:AX) from the structure
   ;
   mov   AX, word ptr CS:[DI.hndl_sizeK]
   mov   DX, 1024
   mul   DX                   ; DX:AX now contains the limit
   jmp short xmsB2

xmsB1:
   ; real source address - we need to check the offset for validity
   ;
	BPUTCH	'R'
   xor   AX, AX
   mov   DI, AX                  ; set flag for "real mode"

   mov   AX, word ptr DS:[BX+4]  ; segment
   mov   CL, 4
   call  _shlu2l                 ; returned in CX:AX
   add   AX, word ptr DS:[BX+2]  ; offset
   adc   CX, 0
   mov   DS:[BX+2], AX
   mov   DS:[BX+4], CX

   ; initialize the limit (DX:AX)
   ;
   mov   DX, 10h
   mov   AX, 0FFF0h              ; 1 Meg + 64k - 10h

xmsB2:
   ; now check the length of the copy versus the limit in DX:AX
   ;
	BPUTCH	':'
   cmp   DX, word ptr DS:[BX+4]
   ja short xmsB3
   jb short BadMoveLength
   cmp   AX, word ptr DS:[BX+2]
   jb short BadMoveLength

xmsB3:
   ; finally, we must convert offset to linear, only for XMS addresses
   ;
	BPUTCH	'.'
   or    DI, DI
   jz short xmsB4

   mov   AX, word ptr CS:[DI.hndl_startK]
   push  BX
   call  extended_address     ; returns address in DX:BX
   xchg  BX, AX               ; now address is in DX:AX
   pop   BX
   add   word ptr DS:[BX+2], AX
   adc   word ptr DS:[BX+4], DX
   jmp short xmsB4

BadMoveLength:
	BPUTCH	'M'
   mov   BL, ERR_BADLEN
   jmp short failed_copy

BadHandle:
   pop   BX                      ; clean up the stack
	BPUTCH	'H'
   mov   BL, ERR_BADHANDLE
   jmp short failed_copy


xmsB4:
   ; next check the DESTINATION address for various validities.
   ; Register Usage as before.
   ;
	BPUTCH	'-'
   lea   BX, [SI.xblk_dst_handle]
   mov   DX, word ptr DS:[BX]
   or    DX, DX               ;if zero, this is a real-mode address
   jz    xmsB5

   ; XMS destination address - we need to check the handle and load the limits
   ;
	BPUTCH	'X'
   call  handle2ptr
   or    DI, DI
   jz short BadHandle

   ; load the limit (DX:AX) from the structure
   ;
   mov   AX, CS:[DI.hndl_sizeK]
   mov   DX, 1024
   mul   DX                   ; DX:AX now contains the limit
   jmp short xmsB6

xmsB5:
   ; real destination address - we need to check the offset for validity
   ;
	BPUTCH	'R'
   xor   AX, AX
   mov   DI, AX                  ; set flag for "real mode"

   mov   AX, word ptr DS:[BX+4]  ; segment
   mov   CL, 4
   call  _shlu2l                 ; returned in CX:AX
   add   AX, word ptr DS:[BX+2]  ; offset
   adc   CX, 0
   mov   DS:[BX+2], AX
   mov   DS:[BX+4], CX

   ; initialize the limit (DX:AX)
   ;
   mov   DX, 10h
   mov   AX, 0FFF0h              ; 1 Meg + 64k - 10h

xmsB6:
   ; now check the length of the copy versus the limit in DX:AX
   ;
	BPUTCH	':'
   cmp   DX, word ptr DS:[BX+4]
   ja short xmsB7
   jb short BadMoveLength
   cmp   AX, word ptr DS:[BX+2]
   jb short BadMoveLength

xmsB7:
   ; finally, we must convert offset to linear, only for XMS addresses
   ;
	BPUTCH	'.'
   or    DI, DI
   jz short xmsB8

   mov   AX, word ptr CS:[DI.hndl_startK]
   push  BX
   call  extended_address     ; returns address in DX:BX
   xchg  BX, AX               ; now address is in DX:AX
   pop   BX
   add   word ptr DS:[BX+2], AX
   adc   word ptr DS:[BX+4], DX

xmsB8:
   ; copy_all_data() requires only that DS:SI point to the XMSCOPYBLK
   ;

ELSE
	; check offset to make sure it's in range
	lea	BX, [SI.xblk_src_handle]
	call	check_offset	; check SOURCE
	or	AL, AL		; is everything OK?
	jz	xmsB2
xmsB1:	mov	BL, AL		; error (bad handle or address)
	add	BL, ERR_BADSRCHDL-1
	jmp short failed_copy

BadMoveLength:
   mov   BL, ERR_BADLEN
   jmp short failed_copy

xmsB2:	lea	BX, [SI.xblk_dst_handle]
	call	check_offset	; check DESTINATION
	or	AL, AL
	jz	xmsB3
	add	AL, 2
	jmp	short xmsB1	; error (bad handle or address)
xmsB3:

	; check length to make sure it's in range
	lea	BX, [SI.xblk_src_handle]
	mov	AX, word ptr DS:[SI.xblk_count]
	mov	CX, word ptr DS:[SI.xblk_count+2]
	call	check_length
	or	AL, AL
	jz	xmsB5
xmsB4:	mov	bl, al
	jmp short failed_copy	; error!
xmsB5:
   lea	BX, [SI.xblk_dst_handle]
	mov	AX, word ptr DS:[SI.xblk_count]
	mov	CX, word ptr DS:[SI.xblk_count+2]
	call	check_length
	or	AL, AL
	jnz	xmsB4		; error!

	; convert offsets to "linear" addresses
	lea	BX, [SI.xblk_src_handle]
	call	offset2linear
	lea	BX, [SI.xblk_dst_handle]
	call	offset2linear
ENDIF

	BPUTCH	'1'

	; actually copy the stuff
	call	copy_all_data

	BPUTCH	'A'

	or	AX, AX
	jnz	done_w_copy

failed_copy:
	mov	AX, offset CS:xms_error
	jmp	short xmsB9

done_w_copy:
	xor	bl, bl
	mov	AX, offset CS:xms_success_in_ax_only

	; done, get out (either error or worked depending on AX)
xmsB9:	add	SP, 10H
	; restore HMA state
	pop	SI	; previous HMA state
	push	AX	; save "jump to" routine
	call	is_a20_enabled
	cmp	AX, SI
	jz	same	; they are the same, no need to change
	or	SI, SI
	jnz	reenable
	call	disable_a20
	jmp short same
reenable:
	call	enable_a20
same:	pop	AX

	pop	DX		; Recover old BX value
	mov	BH, DH		; restore BH
	or	bl, bl		; if BL has an error code
	jnz	not_same	; then return it
	mov	bl, dl		; else return original value in BL

not_same:
	pop	DX
	pop	SI
	pop	ES
	pop	DS

	BPUTCH	'B'

   EXIT_FUNCTION
	jmp	AX

;TD
IF 0
;---------------------------
;  offset2linear
;  On Entry
;	DS:BX = handle/offset pair
;  On Exit
;	offset converted to linear address
	public	offset2linear
offset2linear	proc	near
   ENTER_FUNCTION
	mov	DX, word ptr DS:[BX]
	or	DX, DX		; no need to convert REAL mode address
	jz	o2l_done
	push	BX
	call	handle2ptr
	mov	AX, word ptr CS:[DI.hndl_startK]
	call	extended_address	; returns address in DX:BX
	xchg	BX, AX			; DX:AX now contains address
	pop	BX
	add	word ptr DS:[BX+2], AX
	adc	word ptr DS:[BX+4], DX	; add start to offset
o2l_done:
   EXIT_FUNCTION
	ret
offset2linear	endp

;---------------------------
;  check_length
;  On Entry
;	DS:BX = handle/offset pair
;	CX:AX = length of copy
;  On Exit
;	AL=0(success), 1(bad handle), 2(bad offset)
;
	public	check_length
check_length	proc	near
   ENTER_FUNCTION

	; add offset to length
	add	AX, word ptr [BX+2]
	adc	CX, word ptr [BX+4]	; CX:AX = offset+length

	; get limit
	mov	DX, word ptr [BX]
	or	DX, DX
	jz	chln1
	push	AX		; preserve low word of length
	call	handle2ptr
	mov	AX, CS:[DI.hndl_sizeK]
	mov	DX, 1024
	mul	DX
	mov	BX, AX		; DX:BX contains limit
	pop	AX		; restore low word of length
	jmp	short chln2

chln1:	; get real mode limit
	mov	DX, 10H
	mov	BX, 0FFF0H	; 1Meg + 64K-10H

	; compare offset+length(CX:AX) to limit(DX:BX)
chln2:	cmp	DX, CX
	ja	len_in_range		; OK, within limits
	jb	len_not_in_range	; failed!
	cmp	BX, AX
	jae	in_range		; in range
len_not_in_range:
	mov	AL, ERR_BADLEN		; error! bad length
	jmp	short chln3
len_in_range:
	xor	AX, AX			; success!
chln3:
   EXIT_FUNCTION
   ret
check_length	endp

;---------------------------
;  check_offset
;  On Entry
;	DS:BX = handle/offset pair
;  On Exit
;	AL=0(success), 1(bad handle), 2(bad offset)
	public	check_offset
check_offset	proc near
   ENTER_FUNCTION

	; if handle==0 then it's a real mode address
	; convert seg:off to linear offset
	mov	DX, word ptr DS:[BX]
	or	DX, DX
	jnz	chof1
	mov	AX, word ptr DS:[BX+4]	; segment
	mov	CL, 4
	call	_shlu2l			; returned in CX:AX
	add	AX, word ptr DS:[BX+2]	; offset
	adc	CX, 0
	mov	DS:[BX+2], AX
	mov	DS:[BX+4], CX		; CX:AX = linear offset
	jmp	short chof2

	; check handle (in DX) for validity
chof1:	call	handle2ptr	; returns ptr in CS:DI
	or	DI, DI
	jnz	chof2		; handle OK
	mov	AL, 1		; bad handle
	jmp short chof3		; exit

	; get limit in bytes
chof2:	cmp	word ptr DS:[BX],0
	jz	in_range		; no limit for REAL mode
	mov	AX, CS:[DI.hndl_sizeK]	; limit in K bytes
	mov	DX, 1024
	mul	DX			; DX:AX now contains limit

	; make sure offset is below limit
	cmp	DX, word ptr DS:[BX+4]
	ja	in_range		; OK, within limits
	jb	not_in_range		; failed!
	cmp	AX, word ptr DS:[BX+2]
	ja	in_range		; in range
not_in_range:
	mov	AL, 2			; bad address
	jmp short chof3
in_range:
	xor	AX, AX
chof3:
   EXIT_FUNCTION
   ret
check_offset	endp

ENDIF

; Set to <>0 by the /BIOS command line argument.
; Forces use of int 15h, fn 87h for data movements to/from
; extended memory.
;
use_BIOS_moves	db	0


.286P

;---------------------------
;  copy_data
;  On Entry
;	DS:SI points to XMSCOPYBLK
;  On Exit
;	AX=1 (success, BL = 0), AX=0 (failure, BL = XMS error code)
;
copy_all_data	proc	near
   ENTER_FUNCTION

	cmp	byte ptr use_BIOS_moves, 0
	jne	CAD_next_copy

	smsw	ax
	shr	ax, 1
	jc	CAD_next_copy		; must use "BIOS" in protected mode

	BPUTCH	'!'

	cmp	byte ptr windozed, 0    ; use "BIOS" when Windows runs
	jne	CAD_next_copy

	BPUTCH	'r'

	call	near ptr real_mode_mover ; else move data directly
	jmp	short CAD_exit

	; do we have any bytes left to copy?
    CAD_next_copy:

	BPUTCH	'b'

	mov	AX, word ptr DS:[SI.xblk_count]
	or	AX, word ptr DS:[SI.xblk_count+2]
	jz	CAD_worked

	; copy a block (up to COPYBYTES per chunk)
	call	copy_block
	or	AX, DX
	jz	CAD_fail

	; on to next block
	sub	word ptr DS:[SI.xblk_count], AX
	sbb	word ptr DS:[SI.xblk_count+2], DX
	add	word ptr DS:[SI.xblk_src_addr], AX
	adc	word ptr DS:[SI.xblk_src_addr+2], DX
	add	word ptr DS:[SI.xblk_dst_addr], AX
	adc	word ptr DS:[SI.xblk_dst_addr+2], DX
	jmp	short CAD_next_copy

    CAD_fail:
	xor	AX, AX
   EXIT_FUNCTION
	ret
    CAD_worked:
	mov	AX, 1

    CAD_exit:
   EXIT_FUNCTION
	ret
copy_all_data	endp

.8086

;---------------------------
;  char copy_block(void)
;
;  This moves a block of data to/from extended memory using
;  the BIOS int 15h fn 87h.
;
;  On Entry
;	DS:SI points to "modified" user block
;	All limits have been checked.  The "modified" block
;	has all	addresses converted to "linear" addresses.
;
;  On Exit
;	DX:AX = # of bytes copied
;	BL    = XMS error code
;
	public	copy_block
copy_block	proc	near
   ENTER_FUNCTION
	push	ES
	push	DS
	push	SI

	; zero the BIOS COPY BLOCK structure
	;
	push	CS
	pop	ES
	mov	DI, offset CS:bios_copy_blk
	mov	CX, size BIOSCOPYBLK
	mov	AL, 0
	cld
	rep	stosb

	; fill in source fields
	mov	CX, word ptr DS:[SI.xblk_src_addr+2]
	mov	AX, word ptr DS:[SI.xblk_src_addr]
	mov	BX, OFFSET CS:bios_copy_blk.bblk_src_gdt
	call	setdescriptor

	; fill in destination fields
	mov	CX, word ptr DS:[SI.xblk_dst_addr+2]
	mov	AX, word ptr DS:[SI.xblk_dst_addr]
	mov	BX, offset CS:bios_copy_blk.bblk_dst_gdt
	call	setdescriptor

	; get count (we set a maximum at COPYBYTES)
	mov	AX, word ptr DS:[SI.xblk_count+2]	; hi word of length
	or	AX, AX
	jnz	minlen
	mov	CX, word ptr DS:[SI.xblk_count]		; lo word of length
	cmp	CX, COPYBYTES
	jae	minlen
	shr	CX, 1			; CX now contains len in words
	jmp	short len_ok
minlen:	mov	CX, (COPYBYTES/2)	; words = bytes/2
len_ok:

	; call BIOS to do the transfer
	mov	SI, CS
	mov	ES, SI
	mov	DS, SI
	mov	SI, offset CS:bios_copy_blk
	mov	AH, 87H
	push	CX		; save word count (in case BIOS doesn't)
	int	15H
	pop	CX		; restore word count
	mov 	bx, 0		; assume no error
	jnc	CB_done

	mov	cx, bx		; no bytes moved
	mov	bl, ERR_BADFUNC ; assume no BIOS support (ah=80h or ah=86h)
	cmp	ah, 080h	; exit if BIOS lacking
	jz	CB_done
	cmp	ah, 086h
	jz	CB_done

	mov	bl, ERR_BADLEN
	cmp	ah, 3
	ja	CB_done

	mov	bl, ERR_PARITY
	dec	ah     		; if parity error exit
	jz	CB_done		; else ah=2 (int err?) or ah=3 (A20 gate err)

	mov	bl, ERR_A20_ERR ; blame problem on A20 gate

	; return size in DX:AX
    CB_done:
	xor	DX, DX
	mov	AX, CX
	shl	AX, 1		; convert words back to bytes
	pop	SI
	pop	DS
	pop	ES
   EXIT_FUNCTION
	ret
copy_block	endp


;---------------------------
;  On Entry
;	CX:AX = (long) starting linear address
;	CS:BX = ptr to Global Descriptor
;  On Exit
;	The Descriptor has been filled in
setdescriptor	proc	near
   ENTER_FUNCTION
	; store info into descriptor
	mov	CS:[BX.gdt_seg_limit], 0FE00H
	mov	CS:[BX.gdt_data_access], 93H
	mov	word ptr CS:[BX.gdt_addr], AX	; CX:AX = addr
	mov	byte ptr CS:[BX.gdt_addr+2], CL	; only 3 byte addr
   EXIT_FUNCTION
	ret
setdescriptor	endp


;-------------------
;  Shift left an unsigned
;  On Entry
;	AX = number
;	CL = bits to shift
;  On Exit
;	CX:AX = shifted number
	public	_shlu2l
_shlu2l	proc	near
   ENTER_FUNCTION
	push	DX
	xor	DX, DX
shlx1:	cmp	CL, 0
	jz	shlx2
	sal	AX, 1
	rcl	DX, 1
	dec	CL
	jmp	short shlx1
shlx2:	mov	CX, DX
	pop	DX
   EXIT_FUNCTION
	ret
_shlu2l	endp

;---------------------------------
;  Convert a handle into a pointer to the structure.
;  Checks to make sure handle is in use.
;
;  On Entry
;	DX = handle
;  On Exit
;	CS:DI = ptr to handle structure
;	DI = 0 on error (bad handle)
;
	public	handle2ptr
handle2ptr	proc	near
   ENTER_FUNCTION
	push	DX
	push	CX

	; is handle within range of possible handles
	mov	AX, DX
	cmp	AX, ORGHANDLE
	jb	bad_handle
	sub	AX, ORGHANDLE
	xor	DX, DX
	mov	CX, HANDLEINC
	div	CX
	or	DX, DX		; there had better be no remainder
	jnz	bad_handle
	cmp	AX, cs:totalHandles ; is it within range of 0..totalHandles-1?
	jae	bad_handle

	; convert index # into pointer
	mov	AH, size XMSHANDLE
	mul	AH
	add	AX, cs:pXMSHandles
	xchg	DI, AX

	; make sure handle is in use
	cmp	byte ptr CS:[DI.hndl_inuse], 0
   jz    bad_handle

handle_ok:
	pop	CX
	pop	DX
   EXIT_FUNCTION
	ret

bad_handle:
   xor   DI, DI
   jmp short handle_ok

handle2ptr	endp


;-------------------------------------
;  Check to see if handle in DX is valid.
;
;  Returns handle ptr in CS:[DI]
;  Doesn't return on error (jumps directly to error!)
;
	public	checkhandle
checkhandle	proc	near
	call	handle2ptr
	or	DI, DI
	jnz	chkh1

	; error! bad handle, jump to error!
	pop	DI		; pop return address
	mov	BL, ERR_BADHANDLE
	jmp	xms_error
chkh1:	ret
checkhandle	endp

	; XMS function 0CH, lock memory
xms_C:
	sti				; ints should be OK here
	call	checkhandle
	cmp	byte ptr CS:[DI.hndl_locked], 0FFH
	jnz	xmsC1
	mov	BL, ERR_LOCK_OVERFLOW
	jmp	xms_error
xmsC1:	inc	byte ptr CS:[DI.hndl_locked]
	mov	AX, word ptr CS:[DI.hndl_startK]
	call	extended_address	; returns address in DX:BX
	jmp	xms_success

;--------------------------------
;  Change K bytes into an extended address
;
;  On Entry	AX = # of K
;  On Exit	DX:BX = address
;
	public	extended_address
extended_address	proc	near
   ENTER_FUNCTION
	push	AX
	mov	DX, 1024
	mul	DX
	add	DX, 11H
	xchg	BX, AX
	pop	AX
   EXIT_FUNCTION
	ret
extended_address	endp


	;------------------------------------
	; XMS function 0DH, unlock memory
xms_D:
	sti				; ints should be OK here
	call	checkhandle
	cmp	CS:[DI.hndl_locked], 0
	jnz	xmsD1
	mov	BL, ERR_NOTLOCKED
	jmp	xms_error
xmsD1:	dec	CS:[DI.hndl_locked]
	jmp	xms_success

	;------------------------------------
	; XMS function 0EH, get handle info
xms_E:
	sti				; ints should be OK here
	call	checkhandle
	mov	BH, CS:[DI.hndl_locked]
	mov	DX, CS:[DI.hndl_sizeK]

	; count free handles (in BL)
	xor	BL, BL
	xor	CX, CX
	mov	DI, cs:pXMSHandles
	jmp short xmsE3

xmsE1:	cmp	byte ptr CS:[DI.hndl_inuse], 0
	jnz	xmsE2
	inc	BL
xmsE2:	inc	CL
	add	DI, size XMSHANDLE
xmsE3:
	cmp	CL, byte ptr cs:totalHandles
	jb	xmsE1
	mov	AX, 1		; return SUCCESS
	jmp	xms_done

	;------------------------------------
	; XMS function 0FH, realloc memory
xms_F:
	sti				; ints should be OK here
	call	checkhandle
	cmp	byte ptr CS:[DI.hndl_locked], 0
	jz	xmsF0
	mov	BL, ERR_LOCKED	; handle is locked, can't free
	jmp	xms_error

	; we can always shrink a block
xmsF0:	xchg	DX, BX		; keep requested "new" size in DX
	cmp	DX, word ptr CS:[DI.hndl_sizeK]
	ja	not_smaller
xmsF1:	mov	word ptr CS:[DI.hndl_sizeK], DX
	or	dx, dx
	jnz	xmsF2

	; Must locate 0 length blocks at the top of XMS memory for
	; the find_next code to work when empty handles are allocated
	;
	push	dx
	mov	dx, cs:xms_limit
	mov	cs:[di].hndl_startK, dx
	pop	dx
xmsF2:
	jmp short xms_success

	; when growing, see if we can do it in place
not_smaller:
	mov	AX, word ptr CS:[DI.hndl_startK]
	add	AX, word ptr CS:[DI.hndl_sizeK]
	push	AX
	call	match_free
	add	SP, 2
	or	CX, CX
	jz	must_move	; no free memory following this one
	add	CX, word ptr CS:[DI.hndl_sizeK]
	cmp	CX, DX		; is current blk + free blk big enough?
	jae	xmsF1		; we CAN reallocate in place

	; we must allocate new memory and copy old to new
must_move:
	; step 1, create a copy block from original handle
	push	SI
	push	DS
	sub	SP, size XMSCOPYBLK	; get space for copyblk
	mov	AX, SS
	mov	DS, AX
	mov	SI, SP		; now DS:SI points to our local copy

	; step 1a, record start and length of existing data
	mov	AX, word ptr CS:[DI.hndl_startK]
	mov	word ptr [SI.xblk_src_addr], AX		; startK
	mov	AX, word ptr CS:[DI.hndl_sizeK]
	mov	word ptr [SI.xblk_count], AX		; sizeK

	; step 2, free the old memory
	mov	byte ptr CS:[DI.hndl_inuse], 0

	; step 3, get the new memory
	call	allocate_memory
	or	CX, CX
	jz	failed_realloc

	; step 4, copy the old data to new memory
	mov	AX, word ptr [SI.xblk_count]
	mov	DX, 1024
	mul	DX
	mov	word ptr [SI.xblk_count], AX
	mov	word ptr [SI.xblk_count+2], DX
	mov	AX, word ptr [SI.xblk_src_addr]
	call	extended_address	; returns address in DX:BX
	mov	word ptr [SI.xblk_src_addr], BX
	mov	word ptr [SI.xblk_src_addr+2], DX
	mov	AX, word ptr CS:[DI.hndl_startK]
	call	extended_address	; returns address in DX:BX
	mov	word ptr [SI.xblk_dst_addr], BX
	mov	word ptr [SI.xblk_dst_addr+2], DX
	call	copy_all_data

realloc_worked:
	mov	AX, offset CS:xms_success
	jmp	short xmsF9

	; restore original values
failed_realloc:
	inc	byte ptr CS:[DI.hndl_inuse]
	mov	BL, ERR_NOFREE
	mov	AX, offset CS:xms_error

xmsF9:	add	SP, size XMSCOPYBLK
	pop	DS
	pop	SI
	jmp	AX

	public	xms_success, xms_error, xms_success_in_ax_only
xms_success:
	mov	BL, 0		; put success code in bl

xms_success_in_ax_only:
	mov	AX, 1		; return SUCCESS code ax

	public	xms_done
xms_done:
	cli
	cmp	cs:stackSemaphore, 0
	jne	xms_exit_OK

	mov	ss, cs:callers_SS
	mov	sp, cs:callers_SP

	public	xms_exit_OK
xms_exit_OK:
	dec	cs:stackSemaphore
	pop	DI
	pop	CX
	popf			; CEM- restore IF & DF
	clc
	retf

xms_not_supported:
	mov	BL, 80h	; CEM - added "func not implemented" ERRcode

	public	xms_error
xms_error:
	cli
	cmp	cs:stackSemaphore, 0
	jne	xms_err_exit

	mov	ss, cs:callers_SS
	mov	sp, cs:callers_SP

	public	xms_err_exit
xms_err_exit:
	dec	cs:stackSemaphore
	pop	DI
	pop	CX
	xor	AX, AX		; return FAILURE code
	popf			; CEM - restore IF & DF
	stc
	retf

;/////////////////////////////////////////////////////////////////////////
;//
;//     Procedure name: control_A20
;//     Purpose:
;//	     Used to gate the A20 address signal on or off.
;//     C.McCall
;//
;//     Inputs:
;//	     Word on stack: 0=OFF, Non-zero=ON
;//     Returned:
;//	     AX: 0=no 8042 communications faults
;//		 Non-zero= 8042 comms fault detected (timeout)
;//			   (assume the action failed)
;//
;/////////////////////////////////////////////////////////////////////////

while_8042buf_full      MACRO
	local   notmt
	xor     cx,cx		; timeout based on target's speed-
 notmt:				; and time of last keyboard transaction.
	in      al,64h
	and     al,02h		; if 8042 input buffer is full
	loopnz  notmt		; then continue polling
			ENDM	; else (empty), quit


	PUBLIC  _control_A20
_control_A20     PROC NEAR
	push    bp
	mov     bp,sp
	pushf
	cli				;disable interrupts

	while_8042buf_full		;if 8042 input buffer not ready
	jnz     ca20_exit		;  then (fault), quit

	mov     al,0d1h			;issue "write discrete port" cmd
	out     64h,al			;  to 8042 control port

	while_8042buf_full		;if 8042 fails to process the command
	jnz     ca20_exit	       	;  then (fault), quit

	mov     ah,[bp+4]		;Check to see if we are gating ON or
	or      ah,ah			;  OFF
	jz      ca20_off

	mov     al,0dfh			; gate A20 ON..
	jmp     short ca20_wrdata

 ca20_off:
	mov     al,0ddh			; gate A20 OFF..

	;	Gate the A20 with command AL
 ca20_wrdata:
	out     60h,al			; write 8042 data port
	while_8042buf_full		; if 8042 fails to process the data
	jnz     ca20_exit		;   then (fault), quit

	;	For now, we're using the "fast" method of A20 settling.
	;	Just allow the overhead of XMS call to be the settling
	;	time.  This is faster and more reliable than MS' HIMEM.
	;
	cmp	byte ptr [a20_type], A20PLUS	; Check for A20+ method
	jne	ca20_exit

	; Use the A20+ method of gating a NULL command
	; to the instruction port of the 8042.  Wait until
	; the NULL command is done processing and we'll
	; KNOW that the gate A20 is done.
	;
	mov     al,0FFh			; Gate NULL command
	out     64h,al			; Write 8042 discrete port
	while_8042buf_full		; Wait for 8042 to process.

 ca20_exit:
	xor     ah,ah		   	;clear AH so 'C' clients can test-
					; the return-code as a WORD value.
	popf
	pop     bp			; entry.
	ret
_control_A20     ENDP


;/////////////////////////////////////////////////////////////////////////
;//
;//     Procedure name: control_PS2_A20
;//     Purpose:
;//	   Used to gate the A20 address signal on or off
;//	   for PS2 style hardware.
;//
;//     Inputs:
;//	     Word on stack: 0=OFF, Non-zero=ON
;//     Returned:
;//	     AX: 0 = no errors, else AX = error #.
;//
;/////////////////////////////////////////////////////////////////////////

	PUBLIC  _control_PS2_A20
_control_PS2_A20  proc near

	push	bp
	mov	bp, sp
	pushf
	cli

	mov	ax, [bp+4]
	or	ax, ax
    	mov     ah, 2
	jnz	PS2_turn_on_A20

    PS2_turn_off_A20:
    	xor     ah, ah

    PS2_turn_on_A20:
    	in      al, 092h
    	jmp     $+2
    	jmp     $+2
    	and     al, 0FDh
    	or      al, ah
    	out     92h, al
    	jmp     $+2
    	xor     cx, cx

    PS2_A20_verify_loop:
    	jmp     $+2
    	in      al, 92h
    	and     al, 2
    	cmp     al, ah
    	loopnz  PS2_A20_verify_loop

    PS2_A20_done:
	xor	ax, ax
	popf
	pop     bp
	ret
_control_PS2_A20  endp


; The interrupt vector table and base segment when using
; linear addresses
;
ZSEG	segment	at 0
	org	00h
div_0_vector	dd	?
	org	0Dh*4
GP_fault_vector	dd	?
ZSEG	ends

; The HMA, div_0_image is an alias for div_0_vector when the A20 line
; is disabled.
;
HMASEG	segment at 0FFFFh
	org	10h
div_0_image	dd	?
HMASEG	ends

; Protected mode data structures and the Global Descriptor Table used
; to establish the 4Gbyte limit on real mode segments.
;
segment_descriptor	struc
	limit		dw	?
	base_low_16	dw	?
	base_middle_8	db	?
	access          db	?
	extra           db      ?
	base_high_8	db      ?
segment_descriptor	ends

table_descriptor	struc
	limit		dw	?
	base		dd	?
table_descriptor	ends


ourGDT		segment_descriptor	<>
data_descriptor	segment_descriptor	<0FFFFh,0,0,93h,0CFh,0>
; don't need a code_descriptor segment_descriptor <0FFFFh,0,0,9Fh,0,0>

ourGDT_ends	label	byte
GDT_LEN	=	(ourGDT_ends - ourGDT)

GDTdescriptor	table_descriptor	<GDT_LEN,0>


.386

; Set to 1 if we want to be sure the A20 line is OFF. A 32 single bit
; comparison is only a (highly) probable indication. In case of a
; match we must change one of the dwords and still have all 32 bits
; be equal to know that the A20 line is off.
;
A20_PARANOIA	=	1

; Determines if the A20 line is actually activated by comparing the
; 4 bytes at 0000:0000 (call them X) against those at FFFF:0010
; (call them Y). If the A20 line is NOT active, these bytes will
; be the same (Y will be an alias of X) due to segment wrap.
; Otherwise, X and Y will be distinct though not neccessarily unequal.
;
; ->	void
; <-	ZF 	set if A20 line is disabled
; regs:	flags
;
A20_mem_test	proc	near
	push	eax
	push	ds
	push	es

	xor	ax, ax
	mov	ds, ax
	dec	ax
	mov	es, ax

	assume	ds:ZSEG, es:HMASEG

	; X-Y == 0  IFF  X == Y
	;
	mov	eax, dword ptr div_0_vector
	cmp	eax, dword ptr div_0_image	; if X != Y, A20 is on

IF A20_PARANOIA
	jnz short A20MT_done

	; Yet if all 32 bits of X == Y coincidentally, A20 could still be on
	; and we wouldn't know. But X-Y == 0 == 1+X^~Y  IFF  X <-> Y
	;
	pushf
	cli
	not	word ptr div_0_vector
	xor	ax, word ptr div_0_image
	not	word ptr div_0_vector
	popf
	inc	ax	    			; A20 off fer sure if ZF set
ENDIF
	assume	ds:nothing, es:nothing

    A20MT_done:
	pop	es
	pop	ds
	pop	eax
	retn
A20_mem_test	endp

.386P

; This is our General Protection Fault handler. It "automatically"
; makes sure we can address extended memory from real mode by
; setting up 4Gbyte segment limits whenever real_mode_mover()'s
; string move ops fault.
;
	public	old_GPF_ISR
old_GPF_ISR	dd	?

	public	GPF_ISR
GPF_ISR	proc	far
   ENTER_FUNCTION
	push	bp			; keep ints off!
	mov	bp, sp
	push	eax

	BPUTCH	'f'

	; make sure we aren't answering IRQ 5 instead of a GPF
	;
	mov	al, 0Bh			; get pending IRQs bit mask from PIC
	out	20h, al
	jmp	$+2
	in	al, 20h
	test	al, 20h			; if IRQ 5 active then not a GPF
	jnz short GPF_not_our_problem

	; Peek at the stack to see if the GPF came from where we
	; expected and if not, punt it away.
	;
	mov	ax, cs
	cmp	ax, [bp+4]
	jne short GPF_not_our_problem
	cmp	word ptr [bp+2], offset RMM_first_fault
	je short GPF_we_done_wrong
	cmp	word ptr [bp+2], offset RMM_second_fault
	jne short GPF_not_our_problem


	; We don't have a 4Gbyte limit established for DS and ES.
	; Take care of that now.
	;
    GPF_we_done_wrong:

	BPUTCH	'g'

	push	bx
	lgdt	fword ptr cs:GDTdescriptor	; load our GDT

	; switch to protected mode and flush the prefetch queue
	;
	cli
	mov	eax, cr0
	or	al, 1			; ints off still, except NMI :-(
	mov	cr0, eax

	jmp	short GPF_protected_flush

	; Select segments with 4Gbyte limits for DS and ES and then
	; switch back to real mode with out resetting them, flushing
	; the prefetch queue again
	;
    GPF_protected_flush:
	mov 	bx, data_descriptor - ourGDT	; bx = our 4Gbyte selector
	mov	ds, bx
	mov	es, bx

	and	al, 0FEh
	mov	cr0, eax

	jmp	short GPF_real_flush

	; restore real_mode_mover()'s register values and restart it
	;
    GPF_real_flush:

	BPUTCH	'h'

	xor	ax, ax
	mov	ds, ax
	mov	es, ax
	pop	bx
	pop	eax
	pop	bp
   EXIT_FUNCTION
	iret


	; chain_intr()
	;
    GPF_not_our_problem:
	pop	eax
	pop	bp
   EXIT_FUNCTION
	jmp	dword ptr cs:[old_GPF_ISR]

GPF_ISR	endp

.386

; Initializes our protected mode data structures used in establishing
; 4Gbyte real mode segment limits
;
	public	setup_descriptors
setup_descriptors	proc	near
	push	eax
	push	ebx

	; Get the Global Descriptor Table (GDT) linear address
	; into the GDT descriptor.
	;
	lea	bx, ourGDT
	movzx	ebx, bx
	mov	ax,  cs
	movzx	eax, ax
	shl	eax, 4
	add	ebx, eax
	mov	GDTdescriptor.base, ebx

	pop	ebx
	pop	eax
	retn
setup_descriptors	endp


.386P

; This is a helper routine for copy_all_data() above. It handles
; moving data to extended memory without leaving real mode using
; segment descriptors with a 4GByte limit. If the segment limits
; have not been initialized or if they have subsequently been
; reset to a 64K limit then a General Protection Fault (GPF) will
; occur when we try to move the data to/from extended memory. The
; GPF_ISR we (de)install here will set up the segment limits and
; then return to the instruction causing the fault. This way we
; set up the segment limits when and only when we need to.
;
; ->	DS:SI 	points to "modified" user block
;		All limits have been checked.  The "modified" block
;		has all	addresses converted to "linear" addresses.
;       	The move count is specified in bytes.
;
; <-	AX	status (1 if success, 0 if failure)
;	BL 	XMS error code (0 if success)
;
; uses:	eax, bx, ecx, flags
;
	public	real_mode_mover
real_mode_mover	proc	near
   ENTER_FUNCTION
	push	esi
	push	edi
	push	ds
	push	es

	BPUTCH	's'

	call	local_enable_A20	; enable A20 gate

	BPUTCH	't'

;BT 3-29-02
;
IF 0
	mov	cx, 0


    RMM_verify_A20_on:
	call	A20_mem_test 		; must be sure it is actually set
	loopz	RMM_verify_A20_on	; or we won't know what we do
	jnz     RMM_do_the_move

	xor	ax, ax			; A20 gate never turned on
	mov	bl, ERR_A20_ERR
	jmp	short RMM_exit
ENDIF

	; Load 32 bit count and source/destination addresses.
	; Note that we depend on BL remaining 0 for successful
	; return status.
	;
    RMM_do_the_move:
	BPUTCH	'u'

	mov	ecx, dword ptr DS:[SI.xblk_count]
	mov	edi, dword ptr DS:[SI.xblk_dst_addr]
	mov	esi, dword ptr DS:[SI.xblk_src_addr]
	xor	bx, bx
	mov	es, bx
	mov	ds, bx			; keep BL == 0

	assume	ds:ZSEG, es:ZSEG

	mov	ax, cs
	shl	eax, 10h
	mov	ax, offset GPF_ISR	; EAX=address of our GP fault handler

	BPUTCH	'v'

	pushf
	cli
   xchg  eax, GP_fault_vector  ; EAX=now previous GP fault handler
   mov   cs:old_GPF_ISR, eax   ; Must save fector so GP handler can chain unserviced calls
	popf

	cld				; go forward

; If either SI or DI are not dword aligned, the processor will generate
; a lot of overhead doing those moves. So do WORD aligned moves
; if necessary.
;
   test  si,0002              ; check for non-dword aligned SI
   jnz short RMM_align_word

   test  di,0002              ; check for non-dword aligned DI
   jz short RMM_align_dword      ; both are perfectly aligned, go for it!

RMM_align_word:
	shr	ecx, 1                  ; count of dwords to move
	BPUTCH	'w'
   jmp short RMM_second_fault

RMM_align_dword:
	shr	ecx, 2                  ; count of dwords to move
	BPUTCH	'l'

    RMM_first_fault:
	rep	movs dword ptr [esi], dword ptr es:[edi] ; move 'em
	rcl	cx, 1					 ; get odd word count

    RMM_second_fault:
	rep	movs word ptr [esi], word ptr es:[edi]  ; movsw if it exists

	BPUTCH	'x'

	pushf
	cli
   mov   GP_fault_vector, eax  ; restore the original GPF handler (EAX)
	popf

	BPUTCH	'y'

	call	local_disable_A20	; restore A20, no need to verify

	assume	ds:nothing, es:nothing

	mov	ax, 1			; return success, BL zeroed above

    RMM_exit:

	BPUTCH	'z'

	pop	es
	pop	ds
	pop	edi
	pop	esi
   EXIT_FUNCTION
	retn
real_mode_mover	endp

.8086

comment ~
The XMS 2.0 spec says:

	XMS drivers are required to hook INT 15h and watch for
	calls to functions 87h (Block Move) and 88h (Extended
	Memory Available). The INT 15h Block Move function must
	be hooked so that the state of the A20 line is preserved
	across the call. The INT 15h Extended Memory Available
	function must be hooked to return 0h to protect the HMA.

but it also says:

	In order to maintain compatibility with existing device
	drivers, DOS XMS drivers must not hook INT 15h until the
	first non-Version Number call to the control function is
	made.

This routine is called by the XMS function dispatcher for
all XMS calls with function numbers > 0.

end_comment ~
hooked_int_15h db 0	; flag to say: 1=hooked 0=not hooked yet
	public HookInt15h
HookInt15h	proc near

	; have we hooked int 15h already? if so just return
	;
	cmp	byte ptr cs:hooked_int_15h, 0
	jne	HI15h_done

	; hook the int 15h vector chain and set our "done it already" flag
	;
	push	bx
	push	ds
	push	es
	xor	bx, bx
	mov	ds, bx
	cli
	les	bx, ds:[15H*4]
	mov	word ptr ds:[15H*4], offset int_15_handler
	mov	word ptr ds:[15H*4+2], cs
	sti
	mov	word ptr cs:[old_15], bx
	mov	word ptr cs:[old_15+2], es
	pop	es
	pop	ds
	pop	bx
	inc	byte ptr cs:hooked_int_15h

    HI15h_done:
	retn
HookInt15h	endp

;------------------------
; XMS DATA AREA
;------------------------

	public	xms_limit
	public	xms_num_index, xms_current_index
	public	bios_copy_blk
EVEN
xms_limit	dw	?	; size (in K) of extended memory
xms_num_index	dw	?
xms_current_index dw	?
bios_copy_blk	BIOSCOPYBLK <>

	public	hma_min, totalHandles, pXMSHandles, lastByteToKeep, xms_index
EVEN

hma_min		dw	DEFAULT_HMA_MIN_VALUE
totalHandles	dw	MAX_HANDLES_COUNT
pXMSHandles	dw	 ((offset xms_index) + MAX_HANDLES_COUNT*2)
lastByteToKeep	dw	(((offset xms_index) + MAX_HANDLES_COUNT*2)+\
			 (MAX_HANDLES_COUNT * (SIZE XMSHANDLE)))
xms_index	dw	MAX_HANDLES_COUNT dup (0)
		XMSHANDLE MAX_HANDLES_COUNT dup(<>)

;/////////////////////////////////////////////////////////////////
;//								//
;//	Device Driver Initialization				//
;//								//
;//	Note: ALL code after this point is thrown away		//
;//	after initialization.  Do NOT place any code here	//
;//	that is needed after the device driver returns		//
;//	from init!!						//
;//								//
;/////////////////////////////////////////////////////////////////
	public hma_init
hma_init:

	; debugging! just so we can stop at the beginning!
	mov	AH, 30H
	int	21h

	; set DS to CS, so we don't have as many seg overrides
	push	CS
	pop	DS

	; parse cmdline for switch
	call	parse_cmdline

	; do PS2 or standard HMA support?
	;
	call	set_a20_type

	; initialize our protected mode data structures
	;
	call	setup_descriptors

IF SIGNON
	; print a sign-on message
	call	do_signon
ENDIF

	; do NOT install if a driver is already installed!
	mov	AX, 4300H
	int	2FH
	cmp	AL, 80H
	jnz	no_prev_xms
IF SIGNON
	mov	DX, offset CS:already_installed_msg
ENDIF
init_failed:
IF SIGNON
	push	DX	; save error message
	push	CS
	pop	DS
	mov	DX, offset CS:himem_err_msg
	mov	AH, 9
	int	21H
	pop	DX
	mov	AH, 9
	int	21H
ENDIF
	jmp	cmderr
no_prev_xms:
	; do NOT install if no A20 support available
	; enable A20 so FFFF:10 does not "wrap"
	call	enable_a20	; turn A20 on so HMA is available
	call    is_a20_enabled
	or	AX, AX
	jz	failed_a20

	; if we cannot turn the a20 line off, this switch lets us install
	; anyway and run with the a20 line constantly enabled...
	;
a20_came_on:
	cmp	byte ptr controllingA20, 0
	je	worked_a20

	; disable A20 so it should wrap
	call	disable_a20	; turn A20 line back off
	call    is_a20_enabled
	or	AX, AX
	jz	worked_a20	; should NOT be enabled

failed_a20:
IF SIGNON
	mov	DX, offset CS:no_mem_msg
ENDIF
	jmp	init_failed

worked_a20:
	; do NOT install if not enough extended memory
	mov	AH, 88H
	int	15H
	sub	AX, 64		; leave space for HMA
	jc	failed_a20
	mov	[xms_limit], AX

	; trap INT 2FH to support call 43xxH
	; int 15h not hooked until first XMS "non-version number" call
	;
	xor	BX, BX
	mov	DS, BX
	cli
	les	BX, DS:[2FH*4]
	mov	word ptr DS:[2FH*4], offset int_2f_handler
	mov	word ptr DS:[2FH*4+2], CS
	sti
	mov	word ptr CS:[old_2F], BX
	mov	word ptr CS:[old_2F+2], ES


	;  BT 12-20-01
	;
	;  This is a temporary call added to emulate MS himem.  This
	;  is the start for making our himem use some of the HMA
	;  for its handle tables but only if DOS=HIGH.  Also this
	;  code fragment will need to be moved to another location
	;  in the source where the handle tables are being allocated.
	;
	;  Following this call, it should be determined if there is
	;  enough HMA memory to accomidate the handle table and if
	;  so allocate some of the HMA using int 2fh function 4a02h.
	;
	mov	ax, 4a01h
	int	2fh


	; determine size of this driver
	lds	BX, CS:[ptrsav]
	mov	ax, cs:lastByteToKeep
	mov	word ptr DS:[BX.endaddr], ax
	mov	word ptr DS:[BX.endaddr+2], CS
	jmp	exit

; fields in the request header
COUNT	equ	12H	; count for read/write
START	equ	14H	; start for read/write

a20control_str	db     12,"/CONTROLA20:"
on_str		db	2,"ON"
off_str		db	3,"OFF"
bios_str	db	5,"/BIOS"	; use BIOS int 15h fn 87h to move data
quiet_str	db	6,"/QUIET"	; do NOT display sign on message

machine_str	db	9,"/MACHINE:"	; pascal style str, len+str
proc_486_str	db	4,"/486"	; same
machine_a20	db	4,"/A20"	; value 1, pascal str (A20AT)
machine_ps2	db	4,"/PS2"	; value 2, pascal str (A20PS2)

	public hma_min_str, num_handles_str
comment ~
The XMS 2.0 spec says:

	HIMEM.SYS supports the optional "/HMAMIN=" parameter.
	The valid values are decimal numbers between 0 and 63.
	If the HMA is currently unused, the caller's size
	parameter is compared to the /HMAMIN= parameter on the
	driver's command line. If the value passed by the caller
	is greater than or equal to the amount specified by the
	driver's parameter, the request succeeds.  This provides
	the ability to ensure that programs which use the HMA
	efficiently have priority over those which do not.
end_comment ~
hma_min_str	db	8,"/HMAMIN="

comment ~
The XMS 2.0 spec says:

	By default, HIMEM.SYS has 32 extended memory handles
	available for use. This number may be adjusted with the
	"/NUMHANDLES=" parameter. The maximum value for this
	parameter is 128 and the minimum is 0. Each handle
	currently requires 6 bytes of resident space.

end_comment ~
num_handles_str db 	12,"/NUMHANDLES="

; converts a sequence of up to 3 ASCII digits to a numeric value in
; the range of 0..DH
;
; ->:	DS:AX	ptr to the sequence of digits
;	DL	default value to use if digit sequence in valid
;	DH	maximum allowable value
; <-:	AX	value of the digits sequence or passed DL value
;		if specification missing or invalid
;
	public 	atoi
atoi	proc near

	push	bx
	push	cx
	push 	si

	or	ax, ax
	jz	ATOI_use_default

	mov	si, ax
	mov	cl, 10d
	mov	bx, (('9' shl 8) or '0')

	cld
	lodsb
	cmp	al, bh
	ja	ATOI_use_default
	sub	al, bl
	jc	ATOI_use_default
	mov	ch, al

	cld
	lodsb
	cmp	al, bh
	ja	ATOI_done
	sub	al, bl
	jc	ATOI_done
	xchg	al, ch
	mul	cl
	add	ch, al

	cld
	lodsb
	cmp	al, bh
	ja	ATOI_done
	sub	al, bl
	jc	ATOI_done
	xchg	al, ch
	mul	cl
	add	ch, al

    ATOI_done:
	mov	al, ch
	cmp	al, dh
	jbe	ATOI_exit

    ATOI_use_default:
	mov	al, dl

    ATOI_exit:
	xor	ah, ah
    	pop	si
	pop	cx
	pop	bx
	retn
atoi	endp

; Adjusts the values of the totalHandles, pXMSHandles and
; lastByteToKeep variables in order to allocate data structures
; for the specified number of handles.
;
; ->:	AX	number of XMS handles to allocate
; <-:	void
;
	public	AllocateHandles
AllocateHandles	proc near

	mov	cs:totalHandles, ax	; will be 0 <= ax <= 128
	mov	bx, offset cs:xms_index
	add	bx, ax
	add	bx, ax			; *2 for word index variables
	mov	cs:pXMSHandles, bx      ; handles strucs start after indices
	mov	cl, SIZE XMSHANDLE
	mul	cl
	add	ax, bx
	mov	cs:lastByteToKeep, ax	; handle strucs are last thing to keep
	retn
AllocateHandles endp

;------------------------------------------------
;  parse_cmdline
;
;  Parse the Command Line
;
	public	parse_cmdline
parse_cmdline	proc near
	push	DS

	; get cmdline string in DS:SI
	lds	SI, [ptrsav]
	lds	SI, DS:[SI+COUNT]

	; make cmdline UPPERCASE
	push	SI
	call	strupr
	pop	SI

find_handle_count:
	push	cs
	pop	es
	mov	di, offset num_handles_str ; "/NUMHANDLES=n", n in {0..128}
	call	search4str
	mov	dx, ((MAX_HANDLES_COUNT shl 8) or DEFAULT_HANDLES_COUNT)
	call	atoi
	call	AllocateHandles		   ; adjust "end of driver" address

find_hma_min:
	push	cs
	pop	es
	mov	di, offset hma_min_str	   ; "/HMAMIN=n", n in {0..63}
	call	search4str
	mov	dx, ((MAX_HMA_MIN_VALUE shl 8) or DEFAULT_HMA_MIN_VALUE)
	call	atoi
	mov	cs:hma_min, ax

find_bios_str:
	push	cs
	pop	es
	mov	di, offset bios_str
	call	search4str
	or	al, ah
	mov	cs:use_BIOS_moves, al

find_quiet_str:
	push	cs
	pop	es
	mov	di, offset quiet_str
	call	search4str
	or	al, ah
	mov	byte ptr cs:be_quiet, al

find_a20_control:
	push	cs
	pop	es
	mov	di, offset a20control_str  	; "/CONTROLA20:[ON|OFF]"
	call	search4str
	or	ax, ax
	jz	find_486

	mov	di, ax
	cmp	byte ptr [di], 'O'		; ignore bogus switch
	jne	find_486

	inc	di
	cmp	byte ptr [di], 'F'		; accept 'OF' as 'OFF'
	jne	find_486

	mov	byte ptr controllingA20, 0	; and set boolean to FALSE

	; Look for /486 switch
find_486:

	push	CS
	pop	ES
	mov	DI, offset proc_486_str
	call	search4str
	or	AX, AX
	jz	find_machine
	mov	es:[cache_486], 1

	; Find machine dependant switch on command line,
	; and then stop processing.  Only one machine type
	; on command line.

	; find /MACHINE: switch (if any)
find_machine:
	push	CS
	pop	ES
	mov	DI, offset machine_str
	call	search4str
	or	AX, AX
	jz	find_a20

	; we found the machine string, see if it's legal
	push	SI
	mov	SI, AX
	mov	AL, byte ptr DS:[SI]
	pop	SI
	sub	AL, "0"
	cmp	AL, A20PTL		; Only allow AT, PS2 and PTL
	ja	find_a20		; Look for explicit switches /A20 etc
	mov	es:[a20_type], AL
	jmp short done_parse		; Found a machine, we're done!

	;All variables are in the CODESEG, because there is
	;no separate DATASEG.

	; find /A20 switch (if any)
find_a20:
	push	CS
	pop	ES
	mov	DI, offset machine_a20
	call	search4str
	or	AX, AX
	jz	find_ps2

	; we found the /A20 string, see if it's the /A20+ string
	push	SI
	mov	SI, AX
	mov	AL, byte ptr DS:[SI]	; Load the next character
	pop	SI
	cmp	AL, "+"			; See if it's a '+'
	je	found_a20plus		; Yes, go store it
	mov	es:[a20_type], A20AT	; No, store normal A20
	jmp short done_parse		; Found a machine, we're done!

found_a20plus:
	mov	es:[a20_type], A20PLUS	; Mark machine A20+
	jmp short done_parse		; Found a machine, we're done!

	; find /PS2 switch (if any)
find_ps2:
	push	CS
	pop	ES
	mov	DI, offset machine_ps2
	call	search4str
	or	AX, AX
	jz	done_parse		; No machine found, all done!
	mov	es:[a20_type], A20PS2
	jmp short done_parse		; Found a machine, we're done!

done_parse:
	pop	DS
	ret
parse_cmdline	endp

;------------------------------------------------
;  search4str	Search for a string in the cmdline
;  On Entry
;	DS:SI = cmdline
;	ES:DI = string to find
;
	public	search4str
search4str	proc	near
	push	SI
	push	DI

	; get length of string
	xor	CX, CX
	mov	CL, ES:[DI]
	inc	DI
	jcxz	failedsrch

	; find a potential start of the string
chkagain:
	lodsb
	cmp	AL, 9		; tabs are OK
	jz	chkchar
	cmp	AL, ' '
	jb	failedsrch
chkchar:
	cmp	AL, ES:[DI]
	jnz	chkagain
	xor	AX, AX
	push	CX		; length
	push	SI
	push	DI
	dec	SI
	repe	cmpsb
	jnz	notfound
	mov	AX, SI		; AX points to end of string
notfound:
	pop	DI
	pop	SI
	pop	CX		; length
	test	AX, AX
	jz	chkagain
	jmp	short found
failedsrch:
	xor	AX, AX
found:
	pop	DI
	pop	SI
	ret
search4str	endp

;------------------------------------------------
;  strupr	Convert cmdline to uppercase
;  On Entry
;	DS:SI = str (terminated by anything less than space)
;
	public	strupr
strupr	proc	near
againstr:
	lodsb
	cmp	AL, 9		; tabs are OK
	jz	againstr
	cmp	AL, ' '
	jb	donestr
	cmp	AL, 'a'
	jb	againstr
	cmp	AL, 'z'
	ja	againstr
	sub	byte ptr DS:[SI-1], 'a'-'A'
	jmp	short againstr
donestr:
	ret
strupr	endp


; Black magic code obtained by inspection of MS HIMEM.SYS to auto-detect
; their /MACHINE:3 machine type. The MS-DOS v6.22 help file referst to
; this code as the "Phoenix Cascade BIOS". No known documentation exists
; on this method of controlling the A20 line by this method. The BIOS
; code is executed by a call to the entry point in the BIOS which is
; hardcoded in MS-DOS v6.22 HIMEM.SYS as 0F000:0FF82. The following
; parameters are considered valid:
;
; 	AX = 0800 (Disable A20 gate -- appears to return nothing)
;	AX = 0801 (Enable A20 gate -- appears to return nothing)
;	AX = 0802 (Querry A20 state -- returns AH=1 if on, AH=0 if off)
;
; The "Phoenix Cascade BIOS" code exists in (at least) the Phoenix BIOS
; for the ELAN boards.
;
; =>:	ES:BX	a valid far ptr to the buffer returned by int 15h fn 0C0h.
; <=:	AX	1 if the PTL BIOS found else 0
;
LookForPTL	proc
	push	bx

	xor	ax, ax			; assume failure
	add	bx, es:[bx]		; skip over machine info
	add	bx, 6			; skip over 6 more magic bytes

	cmp	word ptr es:[bx], 'TP'	; look for the 3 byte string "PTL"
    	jne	LFPTL_done
	cmp	byte ptr es:[bx][2], 'L'
    	jne	LFPTL_done

	test	byte ptr es:[bx][3], 80h ; make next sure next byte has B7 set
	jz	LFPTL_done

	inc	ax

    LFPTL_done:
	pop	bx
	retn
LookForPTL	endp


;------------------------------------------------
;  set_a20_type
;
;  Sets the A20 control type to PS/2 or AT.
;
	public	set_a20_type
set_a20_type	proc near

	; if set on cmdline, use that value else get info from BIOS
	;
	mov	AH, [a20_type]
	cmp	AH, UNKNOWN_AT
	jz	SAT_get_BIOS_info	; if unknown, try to auto-detect

	cmp	AH, A20PS2		; If PS2 has been specified set the
	je      SAT_got_PS2		; use_PS2_A20 boolean to TRUE
	jmp	short SAT_done		; else use AT or AT+ as specified

    SAT_get_BIOS_info:
	mov	AX, -1
	mov	ES, AX
	mov	AX, 0C000H
	int	15H 			; should return ptr in ES:BX
	mov	AX, ES
	cmp	AX, -1			; if C0H call not supported by BIOS
	jz	SAT_no_config_info	; then not a true PS2

	call	LookForPTL		; See if "Phoenix Cascade BIOS"
	or	ax, ax
	jz	SAT_see_if_PS2

	mov 	[a20_type], A20PTL	; if so set /MACHINE:3 (A20PTL)
	jmp	short SAT_done

    SAT_see_if_PS2:
	mov	AX, ES:[BX+2]		; check all PS2 models & submodels
	cmp	AX, 4FCH
	jz	SAT_got_PS2
	cmp	AX, 5FCH
	jz	SAT_got_PS2
	cmp	AX, 9FCH
	jz	SAT_got_PS2
	cmp	AL, 0FAH
	jz	SAT_got_PS2
	cmp	AL, 0F8H
	jz	SAT_got_PS2
	jmp	short SAT_no_config_info

    SAT_got_PS2:
	mov 	[a20_type], A20PS2
	jmp	short SAT_done

    SAT_no_config_info:

	; Some Megatel and WinSystems boards do not support the NULL
	; command in their keyboard controllers.
	;
;	mov	[a20_type], A20AT	; assume 8042 NULL cmd not supported

	; assume that the 8042 NULL command is supported
	mov	[a20_type], A20PLUS

IF (1 EQ 0)
	; 		???
	;
	; Undoc PC says commands 0DDh and 0DFh for dis/enabling A20
	; (which we use) are "not supported on most systems" (pg 270-273,311)
	;
	; Should we use int 15h fn 24xxh to control the A20 line if
	; the BIOS supports it (IBM 486 PS1, Undoc PC, pg 626-627)?
	;
	; How can we tell if the NULL command is NOT supported by the 8042
	; without crashing the machine (as Reily says).
	;
	mov	[a20_type], A20PLUS
ENDIF

    SAT_done:
	ret
set_a20_type	endp



be_quiet	db	0

IF SIGNON
signon_msg	label byte
		db	13,10,"HIMEM ",FULLVERSION,13,10
		db	COPYRIGHT,13,10,'$'

using_msg 	db      "Using ",'$'
ptl_msg		db	"PTL",'$'
ps2_msg		db	"PS2",'$'
at_msg		db	"AT", '$'
at_plus_msg	db	"+", '$'
control_msg	db	" A20 Control ",'$'
on_msg 		db	"(ON)",'$'
off_msg		db	"(OFF)",'$'
values_msg	db	" XMS handles available.",13,10
		db	"Minimum HMA usage is ",'$'
kbytes_msg	db	"K."
term_msg	db 	13,10,'$'
using_BIOS_moves_msg db "Using BIOS to move data to/from extended memory.",'$'


himem_err_msg		db	"HIMEM Error: ",'$'
already_installed_msg	db	"Already Installed!",13,10,'$'
no_mem_msg		db	"No Extended Memory!",13,10,'$'
numberString		db	4 dup(' ')

; Prints the byte in AL using printf %d format via int 21h fn 9.
;
	public	PrintNumber
PrintNumber	proc near
	push	bx
	push	cx
	push	dx
	push	di
	push	es

	push	cs
	pop	es

	mov	di, offset cs:numberString
	mov	dx, di
	mov	bx, (('$' shl 8) or '0')
	mov	cx, ((100d shl 8) or 10d)

	mov	ah, al
	cmp	al, cl
	jb	PN_1s_digit
	cmp	al, ch
	jb	PN_10s_digit

	xor	ah, ah
	div	ch
	add	al, bl
	cld
	stosb
	mov	al, ah

    PN_10s_digit:
	cbw
	div	cl
	add	al, bl
	cld
	stosb

    PN_1s_digit:
	add	bl, ah
	mov	es:[di], bx

    PN_print_it:
	mov	ah, 9
	int	21h

	pop	es
	pop	di
	pop	dx
	pop	cx
	pop	bx
	retn
PrintNumber	endp

do_signon	proc near

	push	CS
	pop	DS

	cmp	cs:be_quiet, 0
	jz	DS_doit
	jmp	near ptr DS_done

    DS_doit:
	mov	DX, offset CS:signon_msg
	mov	AH, 9
	int	21H			; print signon

	mov	dx, offset cs:using_msg ; print "Using "
	mov	ah, 9
	int	21h

	mov	dx, offset cs:ps2_msg	; assume "PS2"
	cmp	[a20_type], A20PS2
	jnz	DS_not_PS2
	mov	AH, 9			; print "PS2"
	int	21H
	jmp	short DS_A20_msg

    DS_not_PS2:
	mov	dx, offset cs:ptl_msg	; assume "PTL"
	cmp	[a20_type], A20PTL
	jnz	DS_not_PTL
	mov	AH, 9			; print "PTL"
	int	21H
	jmp	short DS_A20_msg

    DS_not_PTL:
	mov	dx, offset cs:at_msg	; not PS2, print "AT"
	mov	ah, 9
	int	21h

	cmp	[a20_type], A20PLUS	; print "+" if needed
	jne	DS_A20_msg

	mov	dx, offset cs:at_plus_msg
	mov	ah, 9
	int	21h

    DS_A20_msg:
	mov	dx, offset cs:control_msg ; print " A20 Control "
	mov	ah, 9
	int	21h

	mov	dx, offset cs:on_msg
	cmp	byte ptr controllingA20, 0
	jne    	DS_a20_control_onoroff

	mov	dx, offset cs:off_msg

    DS_a20_control_onoroff:
	mov	ah, 9      		; print "(ON)" or "(OFF)"
	int	21h

	mov	dx, offset cs:term_msg	; print "\n"
	mov	ah, 9
	int	21h

    DS_show_if_using_BIOS:
	cmp	cs:use_BIOS_moves, 0
	je	short DS_show_handles_free_and_hmamin

	mov	dx, offset cs:using_BIOS_moves_msg
	mov	ah, 9
	int	21h

	mov	dx, offset cs:term_msg	; print "\n"
	mov	ah, 9
	int	21h

    DS_show_handles_free_and_hmamin:
	mov	ax, cs:totalHandles
	call	PrintNumber

	mov	dx, offset cs:values_msg
	mov	ah, 9
	int	21h

	mov	ax, cs:hma_min
	call	PrintNumber

	mov	dx, offset cs:kbytes_msg	; print "\n"
	mov	ah, 9
	int	21h

    DS_done:
	ret
do_signon	endp
ENDIF

_TEXT	ENDS

	; BT 6-15-99
	;
	; In order to use TLINK 7.1 we must declare an entry point.  This
	; is not an entry point.  Comparing the SYS file created using
	; TLINK 4.0 without an entry point and the SYS file created using
	; TLINK 7.1 revealed no differences.
	;

	END falsestart
