//
//   $Id: KEYB.C 1.10 2002/10/15 20:48:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	KEYB.C - International keyboard driver

   Copyright (c) 1993-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary


	Usage:
		KEYB							(reports status of resident copy)
		KEYB cc						Set mapping to Country CC, default codepage/table
      KEYB cc,cp					Set mapping to Country CC, CodePage CP
		KEYB cc cp				   Set mapping to Country CC, CodePage CP
		KEYB cc cp filespec		Specify country,codpage,location of KEYBOARD.SYS
		KEYB cc,cp,filespec		Specify country,codpage,location of KEYBOARD.SYS
      KEYB cc,,filespec		   Specify country and translation table,
      								use default code page for CC
      KEYB ,,filespec		   Use current country, and code page, specify
      								translation table (1)
      KEYB ,cc       		   Use current country, specify code page (1)


      (1) KEYB must be already installed to obtain current country info.



*/

#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"

// does no thing now
//
#define  ASSERT(x)

#define 	NUL			 			'\0'	// ASCIIZ string termination char
#define 	MAX_COUNTRY_CODE_LEN		2	// max alphas in a country code string
#define 	MAX_FILENM_ARG_CHARS	 128  // max alphas in a file name argument


#include "dosver.h"
#include "keyb.h"
#include "oemmovea.pt"
#include "keyb3.pt"

#define TSR 0x4000	// return this to Terminate and Stay Resident

#define RDONLY 0
#define SEEK_CUR 1

unsigned euroCapsLock;
bool fDisableEuroCapsLock = FALSE;
unsigned char canPrefixList[MAX_CAN_PREFIX];
Prefix * prefixes;
NormalKey * normalKeys;
NormalKey * letters;
NormalKey * numbers;

typedef unsigned (* fnptr)(unsigned);

extern void Install16(fnptr);
extern void Uninstall16(void);
extern void Install2F(void);
extern void Uninstall2F(void);

extern void * AllocateBytes(unsigned bytes);

typedef unsigned char BYTE;
typedef unsigned WORD;
typedef unsigned long DWORD;

typedef struct
{
	DWORD	oldInt9;
	DWORD	oldInt2F;
	BYTE	unknown1[6];
	WORD	flags;
	BYTE	unknown2;
	BYTE	unknown3;
	BYTE	unknown4[4];
	BYTE	countryID[2];
	WORD	currentCodePage;

	// DOS 4.01+
	BYTE	unknown5[2];
	WORD	codePageTableListStart;
	WORD	codePageTableListUnknown;
	BYTE	unknown6[2];
	WORD	translationData;
	WORD	codePageTableListLast;
	BYTE	unknown7[9];

	// ROM-DOS
	DWORD	canPrefixList;
	DWORD	prefixes;
	DWORD	normalKeys;
	DWORD	letters;
	DWORD numbers;
	WORD	mapping;
	WORD	euroCapsLock;
} InternalKEYBData;

#define KEYB 1
#include "dosmsg.c"

InternalKEYBData internalData;
GlobalHeader globalHeader;

// Assume non-enhanced keyboard
unsigned getKey = NORMGET;
unsigned pollKey = NORMCHECK;
unsigned shiftState = NORMSHIFT;

int	LoadMapping(int fd, char * country, int codePage)
{
	int result;
	int i;

	result = 0;

	// if we still haven't found it, give up
	if (fd == -1)
		return result;

   if (codePage==0)
   {
      // Select the Default codepage the same way MS does
      // (adding default DV=437 and default RU=437)
      //
      if( strstr("FR,GR,IT,NL,SV,UK,US,DV,RU",strupr(country)) )
      {
         codePage = 437;
      }
      else if( !stricmp("CF",country) )
      {
         codePage = 863;
      }
      else
      {
         codePage = 850;
      }
   }

	dos_read(fd, (char far *)&globalHeader, sizeof(globalHeader));
	for (i = 0; i < globalHeader.numMappings; ++i)
	{
		Header hdr;
		dos_read(fd, (char far *)&hdr, sizeof(hdr));
		if (!stricmp(country, hdr.country) &&
				(codePage == hdr.codePage) )
		{
			// allocate data space if it hasn't already been
			if (!prefixes)
			{
				prefixes = AllocateBytes(globalHeader.maxPrefixes * sizeof(Prefix));
				normalKeys = AllocateBytes(globalHeader.maxNormalKeys * sizeof(NormalKey));
				letters = AllocateBytes(globalHeader.maxLetters * sizeof(NormalKey));
				numbers = AllocateBytes(globalHeader.maxNumbers * sizeof(NormalKey));
			}
			if (prefixes && normalKeys && letters)
			{
				// now read the data
				dos_read(fd, (char far *)canPrefixList, MAX_CAN_PREFIX);
				dos_read(fd, (char far *)prefixes, hdr.numPrefixes * sizeof(Prefix));
				dos_read(fd, (char far *)normalKeys, hdr.numNormalKeys * sizeof(NormalKey));
				dos_read(fd, (char far *)letters, hdr.numLetters * sizeof(NormalKey));
				dos_read(fd, (char far *)numbers, hdr.numNumbers * sizeof(NormalKey));

				// this is ugly, but we need to pass it back somehow...
            if (fDisableEuroCapsLock)
            {
               euroCapsLock = 0;
            }
            else
            {
   				euroCapsLock = hdr.euroCapsLock;
            }

				result = hdr.codePage;
			}
			break;
		}
		else
		{
			dos_lseek(fd, (long)hdr.totalLen, SEEK_CUR);
		}
	}

	dos_close(fd);
	return result;
}

void	Chirp(void)
{
	sound(1000);
	delay(50);
	nosound();
}

unsigned MakeKey(unsigned char scanCode, unsigned char ascii)
{
	// programs will throw the E0h char code away if the scanCode != 0
	//
	if((ascii == 0xE0) && (scanCode == 0x23))
		scanCode = 0;
	return (scanCode << 8) | ascii;
}

unsigned GetAscii(unsigned key)
{
	return key & 0xFF;
}

unsigned GetScanCode(unsigned key)
{
	return (key >> 8) & 0xFF;
}

unsigned GetShift(unsigned shift)
{
	if (shift & 0x0800)									// alt-Gr
		shift = ALTGR;
	else if ((shift & 0x08) && (shift & 0x04))	// Ctrl-Alt
		shift = ALTGR;
	else if (shift & 0x08)								// Alt
		shift = ALT;
	else if (shift & 0x04)								// Ctrl
		shift = CTRL;
	else if (shift & 0x03)								// Shift
		shift = SHIFT;
	else
		shift = NORMAL;

	return shift;
}

unsigned GetCaps(unsigned shift)
{
	unsigned caps;

	caps = GetShift(shift);

	if (shift & 0x40)										// CAPS
		if (caps == NORMAL)
			caps = SHIFT;
		else if (caps == SHIFT)
			caps = NORMAL;

	return caps;
}

unsigned GetNumLock(unsigned shift)
{
	unsigned numLock;

	numLock = GetShift(shift);

	if (shift & 0x20)										// NUMLOCK
		if (numLock == NORMAL)
			numLock = SHIFT;
		else if (numLock == SHIFT)
			numLock = NORMAL;

	return numLock;
}


// a caps lock key that is unlatched by pressing a shift key
//
int	IsEuroCaps(void)
{
	return internalData.euroCapsLock;
}



/*
	This is the heart of the program. Its called for any x0h (read) or
	x1h (check for) calls from int 16h keyboard bios interface. The value
	of the passed argument is the actual function number. The return value
	is the actual <scan code | char code> combination that would be returned
	from int 16h.

	The actual keyboard translations are done here according to the
	data in the ccxxx.h files. In addition to the normal stuff done
	by the BIOS, the following things need to be accomplished:

	1) Since various countries use other than US kbd layouts, we
	   need to change the characters associated with a scan code
	   so that they match the national convention.

	2) Expand the meaning of the right hand Alt shift key to that
	   of an AltGr key and have it produce bottom right hand symbol
	   on the key (the key must have three or more symbols).
		If there are four symbols on a key <Shift><AltGr><key> gets
		you the fourth one.

   3) In order to allow lots of different accented forms of characters
      IBM/MS allowed prefix chars ("normal" keys that when typed and
      followed immediately by a "prefixable" key, are discarded and
      used to alter the graphical character associated with the second
      key -- sort of meta-shift keys). The prefixPending and charWaiting[]
      variables are used primarily for this purpose.

*/
unsigned		GetTranslatedKey(unsigned function)
{
	int i;
	unsigned char scanCode;
	unsigned char ascii;
	unsigned shift;
	unsigned returnKey;
	unsigned rawShift;
	unsigned rawKey;
	static int prefixPending = -1;
	static unsigned charWaiting[2] = {NO_KEY, NO_KEY};

//	return Call16(function);

#if 0
	// This function should only be called for
	// functions 0, 1, 0x10, and 0x11
	if (function == NORMGET || function == EXTGET)
		function = getKey;
	else if (function == NORMCHECK || function == EXTCHECK)
		function = pollKey;
#endif

#if DEBUG
if (function == NORMGET || function == EXTGET)
	printf("Get %X...", function);
#endif
Top:
	returnKey = NO_KEY;

	if (charWaiting[0] != NO_KEY)
	{
		returnKey = charWaiting[0];
		if (function == NORMGET || function == EXTGET)
		{
			charWaiting[0] = charWaiting[1];
			charWaiting[1] = NO_KEY;
		}
#if DEBUG
printf("charWaiting...");
#endif
		goto Exit;
	}

	rawKey = Call16(function);
	rawShift = Call16(shiftState);

	// if there's no key waiting, say so
	if (rawKey == NO_KEY)
	{
		// check for mapping toggle keys
		// keybrd2.sys has different hotkeys than keyboard.sys. The hot
		// key types are identified in the global header.
		//
		if(globalHeader.hotkeytype)
		{
			if((rawShift & ALL_RAW_SHIFTS)==RAW_ALT_LSHIFT)
			{
				internalData.mapping = 0;		// use first mapping
				Chirp();								// keybrd2.sys also beeps
			}
			else if ((rawShift & ALL_RAW_SHIFTS)==RAW_ALT_RSHIFT)
			{
				internalData.mapping = 1;		// use second mapping
				Chirp();
			}
		}
		returnKey = NO_KEY;
		goto Exit;
	}

	// the key MUST be removed from the queue as soon as we know about it
	if (function == NORMCHECK)
		rawKey = Call16(NORMGET);
	else if (function == EXTCHECK)
		rawKey = Call16(EXTGET);

	scanCode = GetScanCode(rawKey);
	ascii = GetAscii(rawKey);
	shift = GetShift(rawShift);
#if DEBUG
printf("ascii(%X),shift(%X):",ascii,shift);
#endif


	// check for mapping toggle keys
	// keybrd2.sys has a different hotkey and keyboard.sys. The hot
	// key types are identified in the global header.
	//
	if(globalHeader.hotkeytype)
	{
			if((rawShift & ALL_RAW_SHIFTS)==RAW_ALT_LSHIFT)
			{
				internalData.mapping = 0;		// use first mapping
				Chirp();								// keybrd2.sys also beeps
			}
			else if ((rawShift & ALL_RAW_SHIFTS)==RAW_ALT_RSHIFT)
			{
				internalData.mapping = 1;		// use second mapping
				Chirp();
			}
	}
	else
	{
		// <Ctrl><Alt><F1>
		if (scanCode == KEYB_F1 && (rawShift & 0x08) && (rawShift & 0x04))
			internalData.mapping = 0;
		// <Ctrl><Alt><F2>
		else if (scanCode == KEYB_F2 && (rawShift & 0x08) && (rawShift & 0x04))
			internalData.mapping = 1;
	}
	// if we're not doing anything, just return
	if (!internalData.mapping)
	{
		returnKey = rawKey;
		goto Done;
	}

#if DEBUG
printf("Processing %X...", returnKey);
#endif
	// check for prefix key
	for (i=0; prefixes[i].scanCode; ++i)
	{
#if DEBUG
printf("%X/%X %X/%X...",
		prefixes[i].scanCode, scanCode, prefixes[i].shifted, shift);
#endif
		if (prefixes[i].scanCode == scanCode &&
				prefixes[i].shifted == shift)
		{
			// this IS the new prefix, even
			// if another was already pending
			prefixPending = i;
#if DEBUG
printf("Prefix...");
#endif
			goto Done;
		}
	}

	// alt-keys, altgr-keys, punctuation (ignore caps lock)
	for (i=0; normalKeys[i].scanCode; ++i)
	{
      // Make a special exception for 0x56 -- this is the scancode for
      //  both Shift-F3 and the less than/greater than key on International
      //  keyboards.  However, only the LT/GT key returns an ascii value.
      //
      if(   normalKeys[i].scanCode == scanCode
         && normalKeys[i].shifted == shift
         && (   ( scanCode != 0x56 )
             || ((scanCode == 0x56) && (ascii != 0))
            )
        )
		{
			unsigned char newValue;
			newValue = normalKeys[i].value;
			if (shift == ALT)
         {
				returnKey = MakeKey(newValue, 0);
#if DEBUG
printf("ALTkey %X...", returnKey);
#endif
         }
			else if (shift == ALTGR)
         {
            // Per note 3 above, only if we have the proper raw shift:
            //
            if(   ((rawShift & ALL_RAW_SHIFTS)!=RAW_ALT_LSHIFT)
               && ((rawShift & ALL_RAW_SHIFTS)!=RAW_ALT_RSHIFT)
              )
            {
   			   // used to:	returnKey = MakeKey(0, newValue);
   			   // but MS always returns scancode in upper byte
   			   // the zero in the upper byte is for "alt+numpad" chars
   				//
   				returnKey = MakeKey(scanCode, newValue);
#if DEBUG
printf("ALTGRkey %X...", returnKey);
#endif
            }
         }
			else
         {
				returnKey = MakeKey(scanCode, newValue);
#if DEBUG
printf("Normalkey %X...", returnKey);
#endif
         }
			break;
		}
	}

	// letter keys (worry about caps lock)
	if (returnKey == NO_KEY)
	{
		unsigned caps;
		caps = GetCaps(rawShift);
		for (i=0; letters[i].scanCode; ++i)
		{
			if (letters[i].scanCode == scanCode &&
					letters[i].shifted == caps)
			{
				returnKey = MakeKey(scanCode, letters[i].value);
#if DEBUG
printf("Letter %X...", returnKey);
#endif
				break;
			}
		}
	}

	// number keys (worry about num lock)
	if (returnKey == NO_KEY)
	{
		unsigned numLock;
		numLock = GetNumLock(rawShift);
		if( (rawKey & 0xff) == 0xe0 )
		{
			numLock = NORMAL;
		}
		for (i=0; numbers[i].scanCode; ++i)
		{
			if (numbers[i].scanCode == scanCode &&
					numbers[i].shifted == numLock)
			{
				returnKey = MakeKey(scanCode, numbers[i].value);
#if DEBUG
printf("Number %X...", returnKey);
#endif
				break;
			}
		}
	}

	// if we get to this point, it was a normal key that we just
	// want to pass through unchanged
	if (returnKey == NO_KEY)
		returnKey = MakeKey(scanCode, ascii);

	// if there is a prefix pending, process it now,
	// after the second key has been translated
	if (prefixPending != -1)
	{
		unsigned char * p;

#if DEBUG
printf("PrefixPending...");
#endif

		p = strchr(canPrefixList, GetAscii(returnKey));
		if (p)
		{
			// if the key was valid
			int offset;
			unsigned char newValue;

			offset = p - canPrefixList;
			newValue = prefixes[prefixPending].value[offset];

			// check if the user pressed a key that can be prefixed
			if (newValue != 0)
			{
				returnKey = MakeKey(0, newValue);
				prefixPending = -1;
#if DEBUG
printf("Appending Prefix...");
#endif
				goto Done;
			}
		}

#if DEBUG
printf("Invalid key...");
#endif
		// otherwise, the key was invalid
		// bogus key after a prefix; return the prefix,
		// put the bogus key in charWaiting, and chirp
		charWaiting[0] = returnKey;
		returnKey = MakeKey(0, prefixes[prefixPending].value[0]);

		prefixPending = -1;
		Chirp();

	}

Done:

	// if we are finished, we may have to loop back and try again
	if (returnKey == NO_KEY && (function == NORMGET || function == EXTGET) )
		goto Top;

	// if we have a key, but the user doesn't want it yet, we
	// have to save it for later
	if (returnKey != NO_KEY && (function != NORMGET && function != EXTGET) )
	{
		charWaiting[1] = charWaiting[0];
		charWaiting[0] = returnKey;
	}

Exit:
#if DEBUG
if (returnKey != NO_KEY)
	printf("Returing %x\n", returnKey);
#endif
	return returnKey;
}

#if 0
#pragma argsused
int memioctl(unsigned category, char far * buffer)
{
	// return true for success; false for failure
	// can set memerr for failure
	return 0;
}

#pragma argsused
int meminit(char far * command, unsigned * lastSeg)
{
	char buffer[80];

	print_str("German keyboard; Code page 850\r\n$");

	Install16(GetTranslatedKey);

	return 1;		// go ahead and install us
}
#endif

// Skips all parameter separators (spaces,tabs and commas) in the ASCIIZ
// string point to by pStr. Returns a pointer to the first "non-separator"
// char. See the syntax examples at top of this file and the description
// of the local argv[] below.
//
char *SkipSeparators(char *pStr)
{

	ASSERT(pStr); // only non-NULL parameters are valid

	// skip over any tab and space chars
	while(DL_IS_SPACE(*pStr))
		++pStr;

	// if followed by a comma, skip that too
	if(*pStr == ',')
	{
		++pStr;

		// and then skip over any other tab or space chars
		while(DL_IS_SPACE(*pStr))
			++pStr;
	}

	// return a pointer to the next token
	return pStr;
}

// This main depends on the local (vs. ANSI C lib's) argv[]/argc.
// The argv[1] expected here is an ASCIIZ string containing all
// command line arguments but without leading whitespace or
// trailing CR-LF, etc.
//
int main(int argc, char * argv[])
{
	InternalKEYBData far * thatData;
	int fd;
	int codePage;
	int newCodePage;
	char * pStr;
	char country[MAX_COUNTRY_CODE_LEN+1];
	char fileSpec[MAX_FILENM_ARG_CHARS+1];
	int  i;

	codePage = 0;
	thatData = 0;

	// locate resident copy (if any)
	{
		unsigned off;
		unsigned seg;

		asm push	DI
		asm push	ES

		asm mov	AX, 0AD80h
		asm int	2Fh
		asm cmp	AL, 0FFh
		asm jnz	nope

// If you compile this section in BC2, these instructions look like:
// 	mov		[bp-18], DI
// 	mov		BX, ES
// 	mov		[bp-20], BX
// on the other hand, in BC52, they look like
// 	mov		[bp-18], DI
// 	mov		BX, ES
// 	mov		seg, BX
//
#if 0
		asm mov	off, DI
		asm mov	BX, ES
		asm mov	seg, BX

		thatData = MK_FP(seg, off);
#else
		asm{
			mov	word ptr thatData, di
			mov	word ptr thatData+2, es
		}
#endif

nope:
		asm pop	ES
		asm pop	DI
	}

	// if no arguments, just print status report
	if (argc < 2 || (*argv[1] == NUL))
	{
		if (thatData)
			printf(string[0],
					thatData->countryID[0], thatData->countryID[1],
					thatData->currentCodePage);
		else
			printf(string[1]);
		goto ErrExit;
	}

	// if MS-DOS KEYB is already running, we can't change it
	if (thatData && thatData->oldInt9)
	{
		printf(string[2]);
		goto ErrExit;
	}

   // Look for a "/C" somewhere in the command line, set flag and remove it.
   //
   pStr = strchr( argv[1], '/');
   if(pStr)
   {
      if( (*(pStr+1) == 'C') || (*(pStr+1) == 'c') )
      {
         fDisableEuroCapsLock = TRUE;
         *(pStr) = ' ';
         *(pStr+1) = ' ';
      }
   }

	// The country code is defined as a continuous sequence of letters.
	// Copy the user's country code spec from the command line and then
	// make the country code an ASCIIZ string. pStr will point to the
	// remaining input that starts after the country code in argv[1].
	//
	for(i = 0; i<MAX_COUNTRY_CODE_LEN; i++)
	{
		if(isalpha(argv[1][i]))
			country[i]=toupper(argv[1][i]);	// make country code upper case
		else
			break;
	}
	country[i] = NUL;
	if(*country == NUL)
	{
		// if no country ID specified then try to use the installed copy's
		// country code. If no installed copy then just abort.
		//
		if(thatData)
		{
			fmemmove((unsigned char far *)country,
						(unsigned char far *)&thatData->countryID,
						MAX_COUNTRY_CODE_LEN);
		   country[MAX_COUNTRY_CODE_LEN] = NUL;
		} else
		{
			printf(string[1]);
			goto ErrExit;
		}

	}
	pStr = &argv[1][i];

	// Skip over any separators following the country code and convert
	// the the codePage argument to a decimal value.
	// If the remaining arg string is empty or starts with a comma
	// then the codePage argument was missing, use 0 to indicate default
	// code page for the current country ID.
	//
	pStr = SkipSeparators(pStr);
	if(*pStr == NUL || *pStr == ',')
		codePage = 0;
	else
		codePage = atoi(pStr);

	// skip over the code page argument
	//
	while (isdigit(*pStr))
		++pStr;

	// now grab filespec, if any. It came from the command line so
	// will always be < 127 < MAX_FILENM_ARG_CHARS chars.
	//
	pStr = SkipSeparators(pStr);
	if (*pStr != NUL)
		strcpy(fileSpec, pStr);
	else
		strcpy(fileSpec, KEYBDATAFILE);

	// attempt to load the mapping from the data file
	fd = dos_open(fileSpec, 0);

	// if we couldn't open the file, and the user wasn't specific,
	// try to find the data file in the KEYB.COM directory
	// Use 'while()' here to provide conditional 'goto's via 'break'.
	//
	while (fd == -1 && !strchr(fileSpec,'\\') && !strchr(fileSpec,':'))
	{
		char * p;
		char temp[MAX_FILENM_ARG_CHARS];

		// make sure the current program name will fit in temp[]
		if(strlen(argv[0]) >= MAX_FILENM_ARG_CHARS)
			break;

		// make a copy of the current program name and then search
		// backward for a final path separator and, if there is one,
		// then truncate the path string at that point.
		//
		strcpy(temp, argv[0]);
		if((p = strrchr(temp, '\\')) != 0)
			*(p+1) = NUL;
		else
			break;

		// Make sure both the path and the filename will fit together
		// inside the local filename buffer. If so, concat the data file name
		// to the program's current path then try opening that new file spec
		//
		if((strlen(temp)+strlen(fileSpec)) > MAX_FILENM_ARG_CHARS)
			break;
		else
			strcat(temp, fileSpec);
		fd = dos_open(temp, 0);

		break; // make sure we exit the while() loop!
	}
	if (fd == -1)	// if still no valid data file spec, print msg and exit
	{
		printf(string[5]);
		goto ErrExit;
	}

	newCodePage = LoadMapping(fd, country, codePage);
	dos_close(fd);

	if (!newCodePage)
	{
		printf(string[3], country);
		if (codePage)
			printf(string[4], codePage);
		goto ErrExit;
	}
	codePage = newCodePage;

	// update the existing resident copy (if there is one)
	if (thatData)
	{
		unsigned char far * oldCanPrefixList;
		Prefix far * oldPrefixes;
		NormalKey far * oldNormalKeys;
		NormalKey far * oldLetters;
		NormalKey far * oldNumbers;
		unsigned char far * oldCountryID;

		oldCanPrefixList = (unsigned char far *)thatData->canPrefixList;
		oldPrefixes = (Prefix far *)thatData->prefixes;
		oldNormalKeys = (NormalKey far *)thatData->normalKeys;
		oldLetters = (NormalKey far *)thatData->letters;
		oldNumbers = (NormalKey far *)thatData->numbers;
		oldCountryID = (unsigned char far *)&thatData->countryID;

		fmemmove(oldCanPrefixList, (char far *)canPrefixList, MAX_CAN_PREFIX);
		fmemmove(oldPrefixes, (char far *)prefixes, globalHeader.maxPrefixes * sizeof(Prefix));
		fmemmove(oldNormalKeys, (char far *)normalKeys, globalHeader.maxNormalKeys * sizeof(NormalKey));
		fmemmove(oldLetters, (char far *)letters, globalHeader.maxLetters * sizeof(NormalKey));
		fmemmove(oldNumbers, (char far *)numbers, globalHeader.maxNumbers * sizeof(NormalKey));

		fmemmove(oldCountryID, (char far *)country, 2);
		thatData->currentCodePage = codePage;
      if (fDisableEuroCapsLock)
      {
   		thatData->euroCapsLock = 0;
      }
      else
      {
   		thatData->euroCapsLock = euroCapsLock;
      }
		thatData->mapping = 1;

		goto Exit;
	}

	// check enhanced or normal keyboard
	{
		if (CheckEnhanced())
		{
			getKey = EXTGET;
			pollKey = EXTCHECK;
			shiftState = EXTSHIFT;
		}
	}

	memset(&internalData, 0, sizeof(internalData));
	memcpy(internalData.countryID, country, 2);
	internalData.currentCodePage = codePage;
   if (fDisableEuroCapsLock)
   {
   	internalData.euroCapsLock = 0;
   }
   else
   {
   	internalData.euroCapsLock = euroCapsLock;
   }
	internalData.canPrefixList = (DWORD)&canPrefixList;
	internalData.prefixes = (DWORD)prefixes;
	internalData.normalKeys = (DWORD)normalKeys;
	internalData.letters = (DWORD)letters;
	internalData.numbers = (DWORD)numbers;
	internalData.mapping = 1;

	// install as TSR
	Install15();
	Install2F();
	Install16(GetTranslatedKey);

	return TSR;

Exit:
	return 0;

ErrExit:
	return 1;
}
