//
//   $Id: DL8250.H 1.5 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

	8250.H   8250 UART Defines and Structures

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

********************************************************************/


/*
	[]-------------------------------------[]
	 |													|
	 |			USER SELECTABLE DEFINES			|
	 |													|
	[]-------------------------------------[]
*/

/* The Defaults -- used only if invalid values are passed to comm_open() */
#define DEFBAUD		9600	/* 300 - 57600, 115 */
#define DEFPORT		0		/* 0 or 1 */
#define DEFIRQ			4
#define DEFPARMS		N81	/* Do NOT change this */


/*
	[]-------------------------------------[]
	 |													|
	 |				SYSTEM DEFINES					|
	 |													|
	[]-------------------------------------[]
*/

/* Common Defines */
#ifndef __DATALGHT_H
#define TRUE  1
#define FALSE 0
#endif

#define ERR  -1
#define OK    0

/* Microsoft Compiler Dependant */
#ifndef __TURBOC__

#define inportb(port) 		inp(port)
#define outportb(port,val)	outp(port,val)

#endif

/* do this in-line for speed (at 115K speed is important) */
#define comm_data_rdy() while(!(inportb(LINESTATUS) & LS_DRDY)) /* wait */

/* Communications Masks, Ports and Vectors */
#define MASK1	0x10			/* 8259 interrupt masks */
#define MASK2	0x08
#define BASE1	0x3F8			/* Base port address for 8250 UART */
#define BASE2	0x2F8
#define VECT1	0x0C			/* Interrupts that the 8250 UART generates */
#define VECT2	0x0B

/* 8259 Interrupt Controller Ports */
#define EOIPORT	0x20
#define EOIVAL		0x20
#define MASKPORT	0x21		/* each bit: 0=interrupt enabled, 1=masked off */

/* dual PIC support */
#define EOIVAL2	0xA0
#define MASKPORT2	0xA1		/* each bit: 0=interrupt enabled, 1=masked off */


/* 8250 UART Port Addresses */
#define DATAREG      (cp.base)
#define BAUDLOW      (cp.base)
#define BAUDHIGH     (cp.base+1)
#define INTENABLE    (cp.base+1)
#define INTID        (cp.base+2)		// read only IRQ info
#define FIFOCONTROL  (cp.base+2)		// write only FIFO control
#define LINECONTROL  (cp.base+3)
#define MODEMCONTROL (cp.base+4)
#define LINESTATUS   (cp.base+5)
#define MODEMSTATUS  (cp.base+6)

/*
	Stop/Data/Parity Bits for 8250 UART

	Line Control Register: Base+3
	1....... DLAB: 0 means Base+0 is Rx/Tx data port, Base+1 is IER.
	.1...... Send Break
	..xxx... Parity (00=None, 01=Odd, 11=Even, 101=always 0, 111=always 1)
	.....x.. Stop Bits (0=1, 1=2)
	......xx Data Bits (00=5, 01=6, 10=7, 11=8)
	.......1 Data Terminal Ready (DTR)
*/
#define N81     3
#define E71    26
#define N82     7
#define O81    11
#define E81    27

/* 8250 (PC) UART Baud Rate Divisors
	Divisor is written to Base+0, Base+1 when DLAB = 1
*/
#define B300  384
#define B1200  96
#define B2400  48
#define B4800  24
#define B9600  12
#define B19200  6
#define B38400  3
#define B57600  2
#define B115200 1			/* highest PC clock rate */


/*
	Line Status Register: Base+5     Modem Status Register: Base+6
	1....... Timeout Error           1....... Carrier Detect
	.1...... Trans Sh-Reg Empty      .1...... Ring Indicator
	..1..... Trans Hold-Reg Empty    ..1..... Data-Set-Ready
	...1.... Break-Detect Error      ...1.... Clear-To-Send
	....1... Framing Error           ....1... (change in above)
	.....1.. Parity Error            .....1..        ""
	......1. Overrun Error           ......1.        ""
	.......1 Data Ready              .......1        ""
*/
#define LS_DRDY 0x01
#define LS_OERR 0x02
#define LS_TRDY 0x20

#define MS_CARRIER 0x80
#define MS_RING 0x40
#define MS_DSR  0x20
#define MS_CTS  0x10

#define LC_BRK			0x40	// LCR bit that activates the BREAK signal

/*
	Modem Control Register: Base+4
	---..... N/A 8250
	...1.... Loopback
	....1... Interrupt Enable (Out 2)
	.....1.. Out 1
	......1. Request To Send (RTS)
	.......1 Data Terminal Ready (DTR)
*/
#define MC_RTS  		1
#define MC_DTR  		2
#define MC_INT  		8


/*
	Interrupt ID Register: Base+2 (Read Only)
	-----... N/A 8250
	.....xxx Interrupt Type

		  000 Modem Status Changed
		  001 No Interrupt Pending
		  010 Tx Hold Register Empty
		  100 Rx Data Available
		  110 Reciver Line Status Changed
*/
#define IIR_NoInt    1
#define IIR_RxData	2
#define IIR_TxData   4

/*
	Interrupt Enable Register: Base+1 (B7 of LCR (DLAB) = 0)
	----.... N/A 8250
	....1... Modem Status
	.....1.. Receiver Line Status
	......1. Tx Holding Register Empty
	.......1 Rx Data Available
*/
#define IER_RxInt			1
#define IER_TxInt			2
#define IER_RxStat		4
#define IER_ModemStat	8


/*** Some XModem Protocol Bytes ***/
#define SOH 0x01	/* ctrl('A') */
#define STX 0x02	/* ctrl('B') */
#define EOT 0x04	/* ctrl('D') */
#define ACK 0x06	/* ctrl('F') */
#define NAK 0x15	/* ctrl('U') */
#define CAN 0x18	/* ctrl('X') */

// Structures identifying the COMn port in use.
// If you alter these definitions then you must update the
// struc definitions in 8250A.ASM as well.
//
typedef struct
{
	int  head;
	int  tail;
	int  bufsize;
	unsigned char *buffer;
} RingBuffer;

typedef struct
{
	int  base;
	int  iIrqNumber;
	void (far interrupt *old_vect)(void);
	unsigned char old_mask;
	unsigned char old_mask2;
	RingBuffer Rx;
} comport;

/* Externals */
extern comport cp;

/* ..\8250\8250a.asm */
void far interrupt comm_int_handler(void);

/* ..\8250\8250.c */
int comm_check(void);
void comm_flush(void);
void comm_to_char(void);
int comm_getc(void);
#if !USE_ZMODEM
bool comm_open(unsigned baud, unsigned parms, int port, int iIrqNumber,
					int rxBufSize, unsigned char *rxBuffer);
int comm_setup(int baud,int parms);
void comm_close(void);
int comm_poll(void);
int comm_tgetc(int ticks);
void comm_putc(char c);
#endif



#if 0
/*
	[]-------------------------------------------[]
	 |															|
	 |					PACKET DRIVEN I/O					|
	 |															|
	 |	NOTE:	Not used in the C_thru_ROM	Kernel	|
	 |															|
	[]-------------------------------------------[]
*/

#define PAC_BEG		pacbuf
#define PAC_COUNT		(*(int *)pacbuf)
#define PAC_CMD		pacbuf[2]
#define PAC_SECTOR	(*(unsigned *)(pacbuf+3))
#define PAC_RDBUFF	(pacbuf+3)
#define PAC_WRBUFF   (pacbuf+5)
#define PAC_CHGCODE	pacbuf[3]
#define PAC_ERRCODE	pacbuf[3]
#define PAC_BPB		(*(struct bpb *)(pacbuf+3))

struct bpb {
	unsigned 		sec_size;
	unsigned char	sec_per_clus;
	unsigned			res_sec;
	unsigned char	num_fats;
	unsigned			dir_ent;
	unsigned			num_sec;
	unsigned char	media;
	unsigned			fat_size;
	unsigned			sec_per_trak;
	unsigned			sides;
	unsigned			hidden;
	};

extern unsigned char pacbuf[];

#endif

