//
//   $Id: DLEMS.H 1.4 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	DLEMSLIB.H -- Low level LIM 4.0 EMS function library header file.
	Created 10-25-97 by Dennis Edwards

   Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary



*/
#if !defined(_DLEMSLIB_H_)
#define _DLEMSLIB_H_

#include "DATALGHT.H"

/*

	NOTE 1:	The EMS function chart below was taken from an
				include file for a "no longer available from Intel"
				EMS Toolkit. The function names and arguments used
				in EMMLIBx have been changed to reflect Datalight
				coding guidelines. No one should have any trouble
				determining functional equivalence. The portions of
				the orignal Intel toolkit which may be useful in
				extending this (and/or the XMSLIBx) library's
				functionality may be found in the emslib\intel
				directory.

				The function declarations in this file have been
				grouped together by function for convenience of the
				user. Please see the LIM 4.0 EMS spec for details.

	NOTE 2:	Datalight's EMM386.EXE v2.6 emulates LIM 3.2 EMS
				hardware with LIM 4.0 EMS software:	It only has four
				mappable standard pages.


   NOTE 3:	Functions marked with an asterisk (*) below are not
				supported by any known flavor of EMM386.

				A very few hardware manufacturers used to implement
				non-volatile EMS memory. The EMM handle information
				as well as the EMS memory's data could survive when
				the machine was powered down or reset. Since the
				handle attribute functions and the "prepare for warm
				boot" function require this sort of special EMS
				hardware to function they are not supported by any
				software EMM (aka EMM386).

				DMA register sets are another hardware feature that
				would allow DMA operations to be performed into
				memory that was not mapped into the lower 1M of
				memory. Since, again, special EMS hardware is
				required this functionality no software EMMs support
				these functions. Similar functionality is provided
				in software via the Virtual DMA (VDMA) interface
				specification.

	NOTE 4: 	The test modules indicated in the comments are those
				from the Datalight EMM386.EXE test program,
				EMMTEST.EXE


*/
/*===========================================================================*/
/*                          EMM FUNCTION CLASSES                             */
/*------------------------------------+--------------------------------------*/
/*     PRESENCE, STATUS, & VERSION    |      MEMORY MOVEMENT & EXCHANGE      */
/*------------------------------------+--------------------------------------*/
/*   EMM_installed                    |   move_memory_region                 */
/*   get_EMM_status                   |   xchg_memory_region                 */
/*   get_EMM_version                  |                                      */
/*                                    |                                      */
/*------------------------------------+--------------------------------------*/
/*          MEMORY ALLOCATION         |           HANDLE MANAGEMENT          */
/*------------------------------------+--------------------------------------*/
/*   get_unalloc_page_count           |   get_handle_count                   */
/*   get_alloc_page_count             |   get_handle_pages                   */
/*   get_total_page_count             |   get_all_handles_pages              */
/*   get_unalloc_raw_page_count       | * get_handle_attrib                  */
/*   get_alloc_raw_page_count         | * set_handle_attrib                  */
/*   get_total_raw_page_count         | * get_attrib_capability              */
/*   alloc_pages                      |   get_handle_name                    */
/*   dealloc_pages                    |   set_handle_name                    */
/*   realloc_pages                    |   get_handle_dir                     */
/*   alloc_std_pages                  |   search_handle_name                 */
/*   alloc_raw_pages                  |   get_total_handles                  */
/*                                    |                                      */
/*------------------------------------+--------------------------------------*/
/*     MAPPABLE MEMORY REGION INFO    |         PROGRAM FLOW CONTROL         */
/*------------------------------------+--------------------------------------*/
/*   get_page_frame_seg               |   alter_map_jump                     */
/*   get_MAPPABLE_regions             |   alter_map_call                     */
/*   get_MAPPABLE_region_count        |   get_map_call_stack_size            */
/*                                    |                                      */
/*------------------------------------+--------------------------------------*/
/*           MEMORY MAPPING           |                OS ONLY               */
/*------------------------------------+--------------------------------------*/
/*   map_unmap_page                   |   enable_OS_fcns                     */
/*   map_unmap_pages                  |   disable_OS_fcns                    */
/*                                    |   return_OS_access_key               */
/*------------------------------------+   get_hw_info                        */
/*       MEMORY MAPPING CONTEXT       |   get_alt_reg_set                    */
/*------------------------------------+   set_alt_reg_set                    */
/*   save_context                     |   get_alt_context_size               */
/*   restore_context                  |   alloc_alt_reg_set                  */
/*   get_context                      |   dealloc_alt_reg_set                */
/*   set_context                      | * alloc_DMA_reg_set                  */
/*   get_set_context                  | * enable_DMA_reg_set                 */
/*   get_context_size                 | * disable_DMA_reg_set                */
/*   get_partial_context              | * dealloc_DMA_reg_set                */
/*   get_partial_context_size         | * prep_EMM_warmboot                  */
/*   set_partial_context              |                                      */
/*                                    |                                      */
/*====================================+======================================*/


/*
	All the EMSLIBx functions return <0 on error, 0 on success.
	Never use a value obtained by any of these functions unless
	they return EMS_SUCCESS. The individual error codes are
	listed below. Note that these error codes are just the sign
	extended error code bytes defined by the LIM 4.0 EMS spec.
	This allows for efficient return code error checking. This
	library includes a function, GetEMSErrorCodeString(), which
	allows EMS errors to be described to the user.

*/

#define EMS_SUCCESS        							   ((int)0x0000)
#define EMS_SOFTWARE_ERR_STAT						   	((int)0xFF80)
#define EMS_HANDLE_NOT_FOUND_STAT					   ((int)0xFF83)
#define EMS_UNDEFINED_FCN_STAT		  				   ((int)0xFF84)
#define EMS_INSUFFICIENT_UNALLOC_HANDLES_STAT	   ((int)0xFF85)
#define EMS_SAVE_RESTORE_CONTEXT_ERR_STAT		   	((int)0xFF86)
#define EMS_INSUFFICIENT_TOTAL_PAGES_STAT		   	((int)0xFF87)
#define EMS_INSUFFICIENT_UNALLOC_PAGES_STAT		   ((int)0xFF88)
#define EMS_ALLOCATE_ZERO_PAGES_STAT				   ((int)0xFF89)
#define EMS_LOG_PAGE_INVALID_STAT			  	 		((int)0xFF8A)
#define EMS_MAPPABLE_PAGE_INVALID_STAT		   		((int)0xFF8B)
#define EMS_HANDLE_CONTEXT_FULL_STAT		  		   ((int)0xFF8C)
#define EMS_HANDLE_CONTEXT_ALREADY_SAVED_STAT	   ((int)0xFF8D)
#define EMS_HANDLE_CONTEXT_NOT_PRESENT_STAT		   ((int)0xFF8E)
#define EMS_UNDEFINED_SUB_FCN_STAT					   ((int)0xFF8F)
#define EMS_UNDEFINED_ATTRIB_STAT					   ((int)0xFF90)
#define EMS_CAPABILITY_NOT_SUPPORTED_STAT			   ((int)0xFF91)
#define EMS_MOVE_SOURCE_DEST_OVERLAP_STAT	   		((int)0xFF92) // a warning
#define EMS_BLOCK_LEN_LESS_THAN_REGION_STAT			((int)0xFF93)
#define EMS_CONV_EXP_MEM_REGION_OVERLAP_STAT			((int)0xFF94)
#define EMS_OFFSET_GREATER_THAN_PAGE_LEN_STAT		((int)0xFF95)
#define EMS_REGION_LEN_EXCEEDS_MEGABYTE_STAT			((int)0xFF96)
#define EMS_XCHG_SOURCE_DEST_OVERLAP_STAT				((int)0xFF97)
#define EMS_MEM_TYPE_NOT_DEFINED_STAT					((int)0xFF98)
#define EMS_UNDEFINED_AMRS_STAT  						((int)0xFF9A)
#define EMS_NO_MORE_AMRS									((int)0xFF9B)
#define EMS_ALT_MAP_DMA_SET_NOT_SUPPORTED_STAT		((int)0xFF9C)
#define EMS_AMRS_NOT_ALLOCATED							((int)0xFF9D)
#define EMS_HANDLE_NAME_NOT_FOUND_STAT					((int)0xFFA0)
#define EMS_DUPLICATE_HANDLE_NAME_FOUND_STAT			((int)0xFFA1)
#define EMS_MEGABYTE_WRAP_STAT 							((int)0xFFA2)
#define EMS_ARRAY_CORRUPT_STAT	   					((int)0xFFA3)
#define EMS_FCN_ACCESS_DENIED_STAT						((int)0xFFA4)

// emsermsg.asm
// returns a ptr to an ASCIIZ string describing an EMM error.
//
char * PASCAL GetEMSErrorCodeString(int EMSErrorCode);


/*
	Detection, status, version functions
*/

// emsf00.asm (no EMM function, DOS call)
//
int PASCAL CheckEMMInstalled(void);

// emsf40.asm -- (test module: EMMINFO.C)
//
// NOTE: When referencing documents that refer to decimal EMS
//			function sequence numbers vs. hex int 67h function
//			numbers (eg function 3 is "alloc" vs. int 67h fn 43h),
//			just add/subtract 40h to convert from one to the other.
//
int PASCAL GetEMMStatus(void);

// emsf46.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSVersion(unsigned *pEMSVersion);



/*
	Memory allocation functions
*/

// emsf42.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSFreePageCount(unsigned *pPagesFree);

// emsf42a.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSAllocPageCount(unsigned *pPagesUsed);

// emsf42b.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSTotalPageCount(unsigned *pPagesTotal);

// emsf5901.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSRawPageCounts(unsigned *pFreeRawPages,
										 unsigned *pTotalRawPages);

// emsf43.asm -- (test module: EMMBASIC.C)
//
int PASCAL AllocateEMSPages(unsigned int pagesToAlloc, unsigned *pHandle);

// emsf5a00.asm -- (test module: EMMBASIC.C)
//
int PASCAL AllocateStandardEMSPages(unsigned int pagesToAlloc,
												unsigned *pHandle);

// emsf5a01.asm -- (test module: EMMBASIC.C)
//
//		NOTE:	With all known flavors of EMM386, this is just an alias
//				for AllocateStandardEMSPages().
//
int PASCAL AllocateRawEMSPages(unsigned int pagesToAlloc, unsigned *pHandle);

// emsf51.asm -- (test module: EMMBASIC.C)
//
int PASCAL ReallocEMSPages(unsigned handle, unsigned *pPageCount);

// emsf45.asm -- (test module: EMMBASIC.C)
//
int PASCAL FreeEMSHandle(unsigned handle);



/*
	Memory mapping functions
*/

// emsf41.asm -- (test module: EMMINFO.C)
//
int PASCAL GetLIM32PageFrameSeg(unsigned *pPageFrameSeg);


typedef struct
{
	unsigned mappableSegment;
	unsigned mappablePageNumber;
} EMS_MAPPABLE_PAGE_STRUCT;

// emsf5801.asm -- (test module: EMMBASIC.C)
//
int PASCAL GetEMSMappablePagesCount(unsigned *pMappablePagesCount);

// emsf5800.asm -- (test module: EMMBASIC.C)
//
int PASCAL GetEMSMappablePages(unsigned *pMappablePagesCount,EMS_MAPPABLE_PAGE_STRUCT far *lpMappablePagesStruct);


#define UNMAP                   0xFFFF
#define MAPPABLE_PAGE_MODE      0
#define SEG_MODE                1
typedef struct
{
	unsigned logicalPage;
	unsigned mappablePageOrSegment;
} EMS_MAP_STRUCT;

// emsf44.asm -- (test module: EMMT4F.C)
//
int PASCAL MapUnmapEMSPage(unsigned handle, unsigned logicalPage,
									unsigned mappablePage);
// emsf50.asm -- (test module: EMMT47.C)
//
int PASCAL MapUnmapMultipleEMSPages(unsigned handle, unsigned mode,
												unsigned pageCount,
												EMS_MAP_STRUCT far *lpPageMap);



/*
	EMS handle functions
*/

#define MAX_HANDLES             255
typedef struct
{
	unsigned handle;
	unsigned pagesOwnedByHandle;
} EMS_HANDLES_PAGES_STRUCT;

// emsf4b.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMSOpenHandleCount(unsigned *pHandleCount);

// emsf4c.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSHandlePageCount(unsigned handle, unsigned *pPagesOwned);

// emsf4d.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetAllEMSHandlePageCounts(unsigned *pOpenHandlesCount,
										EMS_HANDLES_PAGES_STRUCT far * lpAllHandlePages);

// emsf5402.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSTotalHandleCount(unsigned *pEMSTotalHandleCount);


#define MAX_HANDLE_NAME_LEN     8
typedef struct
{
	char name[MAX_HANDLE_NAME_LEN];
} EMS_HANDLE_NAME_STRUCT;

// emsf5300.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSHandleName(unsigned handle, EMS_HANDLE_NAME_STRUCT far *lpHandleName);

// emsf5301.asm -- (test module: EMMHNDL.C)
//
int PASCAL SetEMSHandleName(unsigned handle, EMS_HANDLE_NAME_STRUCT far *lpHandleName);


typedef struct
{
	unsigned handle;
	char		name[MAX_HANDLE_NAME_LEN];
} EMS_HANDLE_DIR_STRUCT;

// emsf5400.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSHandleDir(unsigned *pOpenHandleCount, EMS_HANDLE_DIR_STRUCT far *lpHandleDirectory);

// emsf5401.asm -- (test module: EMMHNDL.C)
//
int PASCAL SearchForEMSHandleName(EMS_HANDLE_NAME_STRUCT far * lpHandleName, unsigned *pHandle);


/*
	These next three funtions are not supported by any known flavor of EMM386,
	except to supply volatile handles.
*/
#define VOLATILE_HANDLES_ONLY	0
#define HANDLE_IS_VOLATILE		0
#define NONVOLATILE_HANDLES   1

// emsf5200.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSHandleAttribute(unsigned handle, unsigned char *pHandleAttribute);

// emsf5201.asm -- (test module: EMMHNDL.C)
//
int PASCAL SetEMSHandleAttribute(unsigned handle, unsigned char newHandleAttribute);

// emsf5202.asm -- (test module: EMMHNDL.C)
//
int PASCAL GetEMSHandleAttributeCapability(unsigned char *pHandleAttributeCapability);



/*
	EMS Context functions
*/

// emsf47.asm -- (test module: EMMT47.C)
//
int PASCAL SaveLIM32EMSContext(unsigned handle);

// emsf48.asm -- (test module: EMMT47.C)
//
int PASCAL RestoreLIM32EMSContext(unsigned handle);


#define MAX_CONTEXT_SIZE        255
typedef struct
{
	char reserved[MAX_CONTEXT_SIZE];
} EMS_CONTEXT_STRUCT;

// emsf4e03.asm -- (test module: EMMT4E.C)
//
int PASCAL GetEMSContextSize(unsigned *pEMSContextSize);

// emsf4e00.asm -- (test module: EMMT4E.C)
//
int PASCAL GetEMSContext(EMS_CONTEXT_STRUCT far *lpOldEMSContext);

// emsf4e01.asm -- (test module: EMMT4E.C)
//
int PASCAL SetEMSContext(EMS_CONTEXT_STRUCT far *lpNewEMSContext);

// emsf4e02.asm -- (test module: EMMT4E.C)
//
int PASCAL GetAndSetEMSContext(EMS_CONTEXT_STRUCT far *lpOldEMSContext,
								      	EMS_CONTEXT_STRUCT far *lpNewEMSContext);


#define MAX_MAPPABLE_REGIONS    64
typedef struct
{
	unsigned mappablePageCount;
	unsigned mappablePageSegments[MAX_MAPPABLE_REGIONS];
} EMS_PARTIAL_CONTEXT_LIST_STRUCT;

// emsf4f02.asm -- (test module: EMMT4F.C)
//
int PASCAL GetPartialEMSContextSize(unsigned howManyPages,
												unsigned *pPartialEMSContextSize);

// emsf4f00.asm -- (test module: EMMT4F.C)
//
int PASCAL GetPartialEMSContext(EMS_PARTIAL_CONTEXT_LIST_STRUCT far *lpWhichPages,
							    EMS_CONTEXT_STRUCT far *lpOldPartialEMSContext);

// emsf4f01.asm -- (test module: EMMT4F.C)
//
int PASCAL SetPartialEMSContext(EMS_CONTEXT_STRUCT far *lpNewPartialEMSContext);



/*
	Memory move/exchange functions
*/

#define DOS_MEM_TYPE		0
#define EMS_MEM_TYPE   	1
typedef struct
{
	unsigned long 	regionSize;
	unsigned char 	sourceMemType;
	unsigned int	sourceHandle;
	unsigned int	sourceOffset;
	unsigned int	sourcePageOrSegment;
	unsigned char	destMemType;
	unsigned int	destHandle;
	unsigned int	destOffset;
	unsigned int	destPageOrSegment;
} EMS_MOVE_XCHG_STRUCT;

// emsf5700.asm -- (test module: EMMT57.C)
//
int PASCAL MoveEMSRegion(EMS_MOVE_XCHG_STRUCT far *lpMoveXchgStruct);

// emsf5701.asm -- (test module: EMMT57.C)
//
int PASCAL ExchangeEMSRegions(EMS_MOVE_XCHG_STRUCT far *lpMoveXchgStruct);



/*
	Alter page map and jump/call functions
*/
typedef struct
{
	void (far *target_function) (void);	// far destination addr of jump
	unsigned char MapStructCount;       // page count to map before jump
	EMS_MAP_STRUCT far *lpMapStruct;    // far ptr to page map array
} EMS_MAP_JUMP_STRUCT;

// emsf55.asm -- (test modules: EMMT55.C, EMMT55A.ASM)
//
int PASCAL AlterEMSPageMapAndJump(unsigned handle, unsigned mode,
									EMS_MAP_JUMP_STRUCT far *lpMapJumpStruct);

typedef struct
{
	void (far *lpTargetFunction) (void); // far destination addr of call
	unsigned char initMapStructCount;    // page count to map before call
	EMS_MAP_STRUCT far *lpInitMapStruct; // far ptr to page map array
	unsigned char finalMapStructCount;   // page count to map on before iret
	EMS_MAP_STRUCT far *lpFinalMapStruct;// far ptr to page map array
	unsigned char reserved[8];	     		 // reserved for use by EMM
} EMS_MAP_CALL_STRUCT;

// emsf5602.asm -- (test modules: EMMT55.C, EMMT55A.ASM)
//
int PASCAL GetEMSMapAndCallStackSize(unsigned *pMapCallStackSize);

// emsf5600.asm -- (test modules: EMMT55.C, EMMT55A.ASM)
//
int PASCAL AlterEMSPageMapAndCall(unsigned handle, unsigned mode,
									EMS_MAP_CALL_STRUCT far * lpMapCallStruct);



/*
	Operating system or environment specific functions
*/

// emsf5D00.asm -- (test module: EMMTOSFN.C)
//
int PASCAL EnableOSFunctions(ulong *pAccessKey);

// emsf5D01.asm -- (test module: EMMTOSFN.C)
//
int PASCAL DisableOSFunctions(ulong *pAccessKey);

// emsf5D02.asm -- (test module: EMMTOSFN.C)
//
int PASCAL ReturnAccessKey(ulong *pAccessKey);

typedef struct
{
	unsigned rawPageParagraphs;
	unsigned AMRSCount;
	unsigned contextSize;
	unsigned DMARSCount;
	unsigned DMAChannelOperation;
} EMS_HW_INFO_STRUCT;

// emsf5900.asm -- (test module: EMMINFO.C)
//
int PASCAL GetEMMHardwareInfo(EMS_HW_INFO_STRUCT far *lpHWInfoStruct);

// emsf5B00.asm
//
int PASCAL GetAlternateMapRegisterSet(unsigned *pAMRSNumber,
													EMS_CONTEXT_STRUCT far **lpEMSContext);

// emsf5B01.asm
//
int PASCAL SetAlternateMapRegisterSet(unsigned pAMRSNumber,
													EMS_CONTEXT_STRUCT far *lpEMSContext);

// emsf5B02.asm
//
int PASCAL GetAMRSSize(unsigned *pContextSize);

// emsf5B03.asm
//
int PASCAL AllocateAlternateMapRegisterSet(unsigned *pAMRSNumber);

// emsf5B04.asm
//
int PASCAL FreeAlternateMapRegisterSet(unsigned AMRSNumber);


/*
	These last five funtions are not supported by any known
	flavor of EMM386. Funtionality similar to DMA register sets
	is defined by the Virtual DMA Specification.

*/
// emsf5C.asm -- (test module: EMMTOSFN.C)
//
int PASCAL PrepareEMMForWarmBoot(void);

// emsf5B05.asm
//
int PASCAL AllocateDMARegisterSet(unsigned *pDMARSNumber);

// emsf5B06.asm
//
int PASCAL EnableDMARegisterSet(unsigned DMARSNumber,unsigned DMAChannelNumber);

// emsf5B07.asm
//
int PASCAL DisableDMARegisterSet(unsigned DMARSNumber);

// emsf5B08.asm
//
int PASCAL FreeDMARegisterSet(unsigned DMARSNumber);



#endif // #if !defined(_DLEMSLIB_H_)

