//
//   $Id: DLPATH.H 1.9 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLPATH.H  created 04/24/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library

	This modules contains general interfaces/functions dealing
	with paths and disks.  Also includes cmdline argument handling.

	Revision History
	03/16/99	jmb	Added ExpandPath

********************************************************************/

#ifndef __DLPATH_H
#define __DLPATH_H

// path manipulation functions
bool AddName(char *pPath, const char *pName);
bool RemoveName(char *pPath);
char LastChar(char *pStr);

char * FindExt(char szFileName[]);
bool IsWild(const char *pStr);
bool SetName(char szFullPath[], const char szName[]);
void SetFileExt(char szFileName[], const char szExt[]);
void CopyNamePart(char szName[], char szFullPath[]);
bool PartialToFullPath(char *pFullPath, char *pPartialPath);
bool ExpandPath(char *pFullPath, char *pPartialPath);


// lastpart.c
//
//  Return the address of the file or directory name in a given name
//  and/or path.
//
char * LastPart( char * szFileName );


//
//		File Functions
//
bool SetFileDate(unsigned handle, unsigned date, unsigned time);
bool GetFileDate(unsigned handle, unsigned * pDate, unsigned * pTime);
bool Exist( char * szFileName );

//
//		Disk Functions
//
bool SetVolumeLabel(int drive, char szVolumeLabel[]);
bool IsValidDrive(int drive);

//
//		Directory Functions
//
bool GetCurrentDir(char *pCwd, int drive);
bool SetCurrentDir(const char *pCwd);
bool MakeDirTree(char szPath[]);

//
//		Directory Attribute Bits
//
#define A_NORMAL		0x00
#define A_READONLY	0x01
#define A_HIDDEN		0x02
#define A_SYSTEM		0x04
#define A_VOLUME		0x08
#define A_SUBDIR		0x10
#define A_ARCHIVE		0x20

#define A_ANYTHING 		   0x3F
#define A_ANYFILE       	(A_ANYTHING & ~A_SUBDIR & ~A_VOLUME)

// This function uses the DOS function 5D00h (server back door)
// to do a wild card delete. To delete all files in a directory
// use the attr. The lpPath argument must be a fully canonical
// path and file mask for the files to delete (eg c:\temp\*.bak).
// Use the WILDDEL_ATTR (0x27) attribute value to delete all files
// (including Hidden, System and Read-Only directory entries).
//
// Returns the DOS error code from the 5D00h call which is
// ERR_FILE_NOT_FOUND (2) on success.
//
int PASCAL WildUnlink(char far*lpPath, uchar attr);

// other directory defines
#define DIR_ENT_SIZE	32


//
//		International Stuff (not yet coded)
//
#define IntlStrRChr(str, c) strrchr(str, c)
#define IntlStrChr(str, c) strchr(str, c)
#define IntlStrUpr(str) strupr(str)
int PASCAL IntlFileNameSplit(char *pStr, char *pDrive, char *pDir,
													char *pName, char *pExt);
#endif
