//
//   $Id: DLRECURS.H 1.5 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLRECURS.H  created 04/24/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library

Interface:
	typedef bool (*ACTUATOR)( PFIND pFind, char szPath[], int level);
	bool Recurse(char szPath[], uchar attr, int mode, ACTUATOR actuator);

Parameters:
	szPath		the path to the root of the tree to process and the
					matching file mask.
	attr			the directory attributes desired
	mode			specifies the type of search to perform on the
					directory tree. Allowable search modes are:
							BREADTH_FIRST
							DEPTH_FIRST
							AS_IS
	actuator		the address of a user supplied ACTUATOR function
					to be called as appropriate for the specified
					search mode. The actuator function must return
					TRUE to continue processing the directory tree or
					FALSE to exit and backtrack to the previous level
					in the tree. The parameters for the actuator
					function are:
					pFind			the BCC RTL ffblk for the current
									file/directory
					szPath		the path to the above file/directory
					level			the current recursion depth


Return Value:
	TRUE if all entries in the current level were processed else
	FALSE

Example:
	#include "..\include\datalght.h"
	#include "..\include\dlrecurs.h"

	bool sample( PFIND pFind, char szPath[], int level)
	{
		printf("level %d: %s\\%s",level,szPath,pFind->ff_name);
		return TRUE;
	}

	Recurse("..\\*.*",0,BREADTH_FIRST,sample);

Notes:
	none

Description:
	Recurse will process a directory tree in the prescribed order
	and call the actuator function for any directories
	encountered and for any files that match the mask specified
	in the szPath argument. The actuator will not be called for
	the "." and ".." entries in any subdirectory.

	Given the following directory tree (where d? are directories
	and f? are files):

	LEVEL:			0		1

				d1---
		  				d2---
								f1
				 				f2
		  				d3---
		  		 				f1
		  				d4
		  				f1
		  				f2

	The order in which the actuator function is presented with
	the directory entries is illustrated below for each
	allowable search mode:

	BREADTH_FIRST			DEPTH_FIRST			AS_IS

	D1\D2                D1\D2\F1          D1\D2
	D1\D3                D1\D2\F2          D1\D2\F1
	D1\D4                D1\D2             D1\D2\F2
	D1\F1                D1\D3\F1          D1\D3
	D1\F2                D1\D3             D1\D3\F1
	D1\D2\F1             D1\D4             D1\D4
	D1\D2\F2             D1\F1             D1\F1
	D1\D3\F1             D1\F2             D1\F2

	The actuator function may perform any action required on the
	files or directories it is passed, including deletion.


********************************************************************/

#ifndef __DLRECURS_H
#define __DLRECURS_H

// make sure the includes are there

#ifndef __DATALGHT_H
#include "..\include\datalght.h"
#endif
#ifndef __DLPATH_H
#include "..\include\dlpath.h"
#endif

//
//		Recursion Types/Routines
//
typedef bool (*ACTUATOR)( PFIND pFind, char szPath[], int level);
bool Recurse(char szPath[], uchar attr, int mode, ACTUATOR actuator);

//
// recurse modes
//
#define BREADTH_FIRST	1	// search breadth first
#define DEPTH_FIRST		2	// search depth first
#define AS_IS				3	// search in-order

#endif  // #ifndef __DATALGHT_H

