;  $Id: DOSTRUCT.INC 1.14 2003/08/28 14:45:04Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1994-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
;----------------------------------------------------------
; $DOCMODULE:DOSTRUCT.INC
;
; Copyright (c) 1994-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary
;
; This file contains all structures used within ROM-DOS assembly
; language modules, most of these structs are DOS 5.0
; compatible. The following list of texts can describe some of
; these structures in more detail.
;
;   1. DOS Internals
;      by Geoff Chappell, Addison-Wesley, 1994
;
;   2. Undocumented DOS, 2nd Edition
;      by Andrew Schulman, Addison-Wesley, 1993
;
;   3. MS-DOS Programmer's Reference 6.0
;      Microsoft Press, 1993
;
;   4. DOS Programmer's Reference, 3rd Edition
;      by Terry R. Dettmann,  Que Corporation 1992
;
;   5. Microsoft Developer Network On-line documentation.
;
;
;   Conditionals that can affect definitions within this file and
;   hence the files in which it is included:
;
;   USING_BIG_DISKS:
;		if >0 then EBIOS LBA disk accesses are supported
;
;   USING_FAT32:
;		if >0 then FAT32 data structures are defined otherwise
;		all data structures are the FAT12/FAT16 subset. We
;		assume all FAT32 media have the HARD_MEDIA type.
;
;   USING_SWAPPABLE_HARD_DISKS:
;
;		CURRENTLY UNAVAILABLE!!!
;
;		if >0 then support for hot swappable HARD_MEDIA media
;		types (hard disks) is available.
;
;
; $DOCHISTORY
;   02/25/99 Synchronized file with DOSTRUCT.H
;   01/04/99 created by Dennis Edwards
;
;---------------------------------------------------------- ~


; All ROMDOS assembly language modules can be compiled with a /DBIGDISKS=1
; and/or a /DFAT32=1 command line option. These options enable support for
; disks >= 8G and for FAT32 file system support, respectively. If you enable
; FAT32 you get BIGDISKS by neccessity.
;
IFDEF BIGDISKS
USING_BIG_DISKS	=	1       ; enables EBIOS fixed disk extensions
ELSE
USING_BIG_DISKS	=	0
ENDIF

IFDEF FAT32
USING_FAT32	= 	1	; enables FAT32 support
USING_BIG_DISKS	=	1       ; enables EBIOS fixed disk extensions
ELSE
USING_FAT32	= 	0
ENDIF

; DON'T CHANGE THIS!!!!
;
; This is a primarily a mechanism for disabling hard disk code in
; change_of_disk() without removing it form the source. EBIOS
; support for removable media will be required for the option to
; be available but those functions are not yet supported by any
; version of ROM-DOS.
;
USING_SWAPPABLE_HARD_DISKS	=  0


IF USING_BIG_DISKS
    IF USING_FAT32
	%OUT Assembling for FAT32.
    ELSE
	%OUT Assembling for BIG DISKS.
    ENDIF
ELSE
    %OUT Assembling for FAT16.
ENDIF

;		Common Defines
;
FALSE		= 0
TRUE		= 1		; This is !FALSE vs. NOT FALSE == ~FALSE in C



;	Absolute Read/Write Info
;	Description: Int 25 and Int 26H (Absolute Read / Write)
;	use this structure if CX == -1
;
;	Reference: DOS Ref, p 753
;
;	Note: This same structure is used by int 21h fn 7305h to read/write
;	sectors on FAT32 drives. 4G*sectors * 512bytes/sector = 2TBytes, the
;	limit in FAT32 drive size.
;
;	AX	=	7305h
;	DS:BX	=	offset of AbsRdWr data struc as defined below.
;	CX	=	-1
;	DL	=  drive number: 0=default, 1=A, 2=B...
;	SI	=  "type info"
;		B14	B13  	B1
;		x	x	?	1 = write, 0 = read
;		0	0	x	Other/unknown
;		0	1	x	FAT  data
;		1	0	x	DIR  data
;		1	1	x	FILE data
;
;		all other bits are reserved and must be 0
;
;	source: MSDN online documentation
;
;
AbsoluteRdWr Struc
    ABS_ulStartSector	dd	?	; starting sector
    ABS_uCount		dw	?	; # of sectors to read/write
    ABS_lpBuffer	dd	?	; buffer pointing to sector
AbsoluteRdWr ends


IF 0

;  Critical Error
;
;  Description: Upon INT 24H, a control word is passed to
;  the handler in AX.  AH describes the error, AL is the
;  disk on which the critical error occurred (0=A, 1=B).
;
;  Control Byte (AH)
;  1xxx xxxx   Char Device (0=disk)
;  xx1x xxxx   Ignore allowed
;  xxx1 xxxx   Retry allowed
;  xxxx 1xxx   Fail allowed
;  xxxx x11x   Disk Area (00=boot, 01=fat, 10=dir, 11=data)
;  xxxx xxx1   Write Attempt (0=read)
;



;  DEVICE HEADER
;  Description: The beginning of a valid device driver
;
;  Reference: Adv MS-DOS, p 264  -or-  DOS Ref, p 340
;
;  Device Attribute Char:
;  1xxx xxxx xxxx xxxx    Character Device (1=char)
;  x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
;  xx1x xxxx xxxx xxxx    Output Until Busy
;  xxxx 1xxx xxxx xxxx    Open/Close
;  xxxx xxxx 1xxx xxxx    Query IOCTL (4410H)
;  xxxx xxxx x1xx xxxx    Generic IOCTL (440CH)
;  xxxx xxxx xxx1 xxxx    FAST I/O (INT 29H)
;  xxxx xxxx xxxx 1xxx    Clock Device
;  xxxx xxxx xxxx x1xx    NUL Device
;  xxxx xxxx xxxx xx1x    Stdout Device
;  xxxx xxxx xxxx xxx1    Stdin Device
;
;  Device Attribute Disk:
;  0xxx xxxx xxxx xxxx    Disk Device (0=disk)
;  x1xx xxxx xxxx xxxx    IOCTL (4402H-4405H)
;  xx1x xxxx xxxx xxxx    IBM block format (0=FAT, 1=scratch?)
;  xxx1 xxxx xxxx xxxx    Network device??
;  xxxx 1xxx xxxx xxxx    Removable Media (4408H)
;  xxxx xxxx 1xxx xxxx    Query IOCTL (4411H)
;  xxxx xxxx x1xx xxxx    Generic IOCTL (440DH)
;  xxxx xxxx xxxx xx1x    Driver Supports 32 bit sectors
;
DeviceHeader struc
    DEV_next		dd	?	; 00 ptr to next device
    DEV_attr		dw	?	; 04 attribute of device
    DEV_strategy	dw	?	; 06 ptr to strat routine
    DEV_service		dw	?	; 08 ptr to serv routine
    DEV_name		db    8 dup (?)	; 0A name of char device
DeviceHeader ends


;  DEVICE REQUEST HEADER
;  Description: Passed to a device (a device call)
;
;  Device Commands
;     0   Init
;     1   Media Check
;     2   Build BPB
;     3   IOCTL Read        (bit 14 required)
;     4   Read
;     5   Nondestructive Read
;     6   Input Status
;     7   Input Flush
;     8   Write
;     9   Write w/ Verify
;    0A   Output Status
;    0B   Output Flush
;    0C   IOCTL Write       (bit 14 required)
;    0D   Device Open       (bit 11 required)
;    0E   Device Close      (bit 11 required)
;    0F   Disk Removable?   (bit 11 required)
;    10   Output Until Busy (bit 13 required)
;    13   Generic IOCTL     (bit  6 required)
;
;  Device Status
;  15           9 8 7             0
;  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
;  |E|         |B|D|               |
;  |r|         |u|o|               |
;  |r|         |s|n|               |
;  | |         |y|e|               |
;  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
;
;  Reference: Adv MS-DOS, p 265  -or-  DOS Ref, p 345
;
DeviceRequestHeader     struc
    RQHDR_length	db	?	; 00 length of the command
    RQHDR_unit		db	?	; 01 unit # for disk access
    RQHDR_command	db	?	; 02 the command itself
    RQHDR_status   	dw	?	; 03 returned status
    RQHDR_reserved 	db   8 dup (?)	; 05
    RQHDR_media		db	?	; 0D media ID for disks
    RQHDR_trans		dd	?	; 0E data transfer buffer
    RQHDR_count		dw	?	; 12 length of data
    RQHDR_start		dw	?	; 14 starting sector for data
    RQHDR_drive		db	?	; 16 starting drive #
    RQHDR_res2		db   3 dup (?)	; 17
    RQHDR_start32	dd	?	; 1A long starting sector
DeviceRequestHeader	ends

ENDIF ;; if 0


;   PROGRAM SEGMENT PREFIX (PSP)
;   Description: DOS builds 100H bytes of special data
;   before every loaded program (COM or EXE).
;
;   Reference: Adv MS-DOS, p 23  -or-  DOS Ref, p --
;
ProgramSegmentPrefix struc
   PSP_int_20		dw	?	; 00 INT 20H instruction
   PSP_top_of_memory	dw	?	; 02 the size of this MCB
   PSP_res1		db	?	; 04 unused
   PSP_dos_long_call	db 5 dup(?)	; 05 CPM style DOS call
   PSP_term_adr		dd	?	; 0A prev INT 22H handler
   PSP_ctrl_brk_adr	dd	?	; 0E prev INT 23H handler
   PSP_critical_error_adr dd	?	; 12 prev INT 24H handler
   PSP_last_psp		dw	?	; 16 prev psp segment
   PSP_hptr		db 20 dup(?)	; 18 the handle table
   PSP_env_seg		dw	?	; 2C environment segment
   PSP_user_stack	dd	?	; 2E ptr to stack on terminate
   PSP_hcount		dw	?	; 32 (20 handles default)
   PSP_htbl_addr	dd	?	; 34 ptr to handle table
   PSP_share_psp	dd	?	; 38 share only, always -1
   PSP_res2		db 4 dup(?)	; 3C unused
   PSP_dosver		dw	?	; 40 DOS version
   PSP_res3		db 0eh dup(?)	; 42 unused
   PSP_doscall		db 3 dup(?)	; 50 not supported
   PSP_res4		db 9 dup(?)	; 53 unused
   PSP_fcb1		db 16 dup(?)	; 5C a parsed FCB
   PSP_fcb2		db 16 dup(?)	; 6C a 2nd parsed FCB
   PSP_res5		db 4 dup(?)	; 7C unused
   PSP_count		db	?	; 80 size of cmdline
   PSP_command_line	db 7fh dup(?)	; 81 command line
ProgramSegmentPrefix ends


;
;  DRIVE PARAMETER BLOCK (DPB)
;
;  Description: Each FAT drive has one of these and one
;  Current Directory Structure.  The DPB keeps track of
;  the various LOGICAL areas on the disk.  It does not
;  understand anything PHYSICAL about the disk.
;
;  Reference: DOS Ref, p 637
;
DriveParameterBlock struc
    DPB_drive			db	?	; 00 0=A, 1=B, 2=C, etc
    DPB_unit			db	?	; 01 unit # to pass to device
    DPB_bytes_per_sec      	dw	?	; 02 bytes per sector (128,256,512,1024)
    DPB_ms_spc			db	?	; 04 sectors per cluster-1
    DPB_shift			db	?	; 05 shift factor to get spc
    DPB_fat_start		dw	?	; 06 # of "reserved" sectors
    DPB_num_fats		db	?	; 08 copies of fat
    DPB_dir_ent			dw	?	; 09 # of root dir entries (FAT32 drive=0)
    DPB_data_start		dw	?	; 0B (low word) of sector # for start of data
    DPB_hiclus			dw	?	; 0D highest cluster # (FAT32 drive=0)
    DPB_fat_len			dw	?	; 0F sectors per fat (FAT32 drive=0)
    DPB_dir_start		dw	?	; 11 SECTOR where root dir starts (FAT32 drive=-1)
    DPB_device			dd	?	; 13 pointer to the device driver
    DPB_media			db	?	; 17 disk ID byte (360K = 0xFD, h.d. = 0xF8)
    DPB_rebuild			db	?	; 18 0xFF means rebuild drive, 0 means no need, 1 means FSINFO updated
    DPB_next			dd	?	; 19 pointer to next DPB
    DPB_newclus			dw	?	; 1D start here when allocating (FAT32 undefined)
    DPB_freeclus		dw	?	; 1F (low word) of free clusters, -1 = unknown

IF USING_FAT32
    DPB_uFreeClustersHigh	dw	?	; 21 FAT32 high word of free clusters (-1 if not FAT32 drive, see FSINFO)
    DPB_uExtendedFlags		dw	?	; 23 FAT32 extended flags from BPB (undefined? if not FAT32 drive)
    DPB_uFSInfoSector		dw	?	; 25 FAT32 sector # of FSINFO struc  (-1 if none, undefined? if not FAT32 drive)
    DPB_uBackUpBootSector	dw	?	; 27 FAT32 sector # of backup boot sector (-1 if none, undefined? if not FAT32 drive)
    DPB_ulFirstDataSector	dd	?	; 29 FAT32 sector # of first cluster (low word also stored in data_start, if not FAT32 drive then = data_start)
    DPB_ulMaxCluster		dd	?	; 2D FAT32 count of drive clusters + 1 (if not FAT32 drive then = hiclus)
    DPB_ulSectorsPerFAT		dd	?	; 31 FAT32 count of sectors per FAT (if not FAT32 drive then = fat_len)
    DPB_ulRootDirSTartingCluster dd	?	; 35 FAT32, may move (undefined? if not FAT32 drive)
    DPB_ulNextFreeCluster	dd	?	; 39 FAT32 newclus (low word also stored in newclus if FAT32 drive else this field is 0, see FSINFO)
ENDIF
DriveParameterBlock ends			; 3D



;  BUFFER LIST STRUCTURE, DOS 5
;
;  Description: Simple disk cache mechanism through sector
;  oriented buffering scheme.  Based on most recently used
;  method.
;
;  Reference: (DOS Ref, p 965)
;
;  Bit mapped flags:
;  x1xx xxxx   Buffer is Dirty
;  xx1x xxxx   Buffer is Valid
;  xxxx 1xxx   Buffer contains Data sector
;  xxxx x1xx   Buffer contains Root Dir sector
;  xxxx xx1x   Buffer contains FAT sector
;  xxxx xxx1   Buffer contains Boot sector
;
BUFLIST struc
    BUF_prev		dw	?	; 00 least recently used
    BUF_next		dw	?	; 02 most recently used
    BUF_drive		db	?	; 04 drive, 0xFF = invalid
    BUF_flags		db	?	; 05 bit mapped flags
    BUF_sector		dd	?	; 06 what sector
    BUF_copies		db	?	; 0A copies of this sector
IF USING_FAT32
    BUF_ulDistance	dd	?	; 0B distance to next copy
ELSE
    BUF_distance	dw	?	; 0B distance to next copy
ENDIF
    BUF_lpDpb		dd	?	; 0D (0F) pointer to DPB
    BUF_store_drive	db	?	; 11 (13) "remember" during crit errors
    BUF_store_flags	db	?	; 12 (14) "remember" during crit errors
    BUF_unknown		db	?	; 13 (15)
IF USING_FAT32
    BUF_unknown2	db	?	;    (16)
    BUF_unknown3	db	?	;    (17)
ENDIF
    BUF_buffer		db	?	; 14 (18) the actual buffer
BUFLIST ends



;  CURRENT DIRECTORY STRUCTURE (CDS)
;
;  Description: Each drive has one of these.  All FAT
;  drives have in addition a Drive Parameter Block.  The
;  CDS contains the current directory and status for the
;  drive.
;
;  Reference: (DOS Ref, p 966)
;
;  Status bit map:
;  1xxx xxxx xxxx xxxx    Network drive
;  x1xx xxxx xxxx xxxx    Valid drive (must be set)
;  xx1x xxxx xxxx xxxx    JOINed drive
;  xxx1 xxxx xxxx xxxx    SUBST drive
;  xxxx xxx1 xxxx xxxx    Drive is Remote
;
CurrentDirectoryStructure struc
    CDS_cwd 			db 67 dup(?)	; 00 as in "C:\" or "Network.node\"
    CDS_status			dw	?	; 43 bit mapped status, see above
    CDS_lpDpb			dd	?	; 45 network drives have no DPB
IF USING_FAT32
    CDS_ulSubDirStartingCluster	dd	?	; 49 subdir starting cluster, never 0
ELSE
    CDS_subdir			dw	?	; 49 subdir cluster # (0000=root)
    CDS_unknown1		dw	?	; 4B
ENDIF
    CDS_unknown2		dw	?	; 4D
    CDS_skip			dw	?	; 4F chars to skip to get past root
    CDS_unknown3		db	?	; 51
    CDS_ifsptr			dd	?	; 52 ???
    CDS_unknown4		dw	?	; 56
CurrentDirectoryStructure ends



;  DIRECTORY ENTRY
;  Description: Each item in the directory looks like this
;
;  Directory attribute:
;  xxxx xxxx xxxx xxx1    Read only
;  xxxx xxxx xxxx xx1x    Hidden
;  xxxx xxxx xxxx x1xx    System
;  xxxx xxxx xxxx 1xxx    Volume label
;  xxxx xxxx xxx1 xxxx    Subdirectory
;  xxxx xxxx xx1x xxxx    Archive
;
;  Reference: Adv MS-DOS, p 184  -or-  DOS Ref, p 248
;
DirectoryEntry struc
    DIR_name			db 8 dup(?)	; 00 file name
    DIR_ext			db 3 dup(?)	; 08
    DIR_attr			db	?	; 0B file attribute
IF USING_FAT32
    DIR_reserved		db	?	; 0Ch ???
    DIR_bCentiSeconds		db	?	; 0Dh centiseconds of creation time
    DIR_uCreationTime		dw	?	; 0Eh file creation date/time
    DIR_uCreationDate		dw	?	; 10h
    DIR_uLastAccessDate		dw	?	; 12h date file was last accessed (opened)
    DIR_uStartingClusterHi	dw	?	; high word of the starting cluster
ELSE
    DIR_reserved		db 10 dup(?)	; 0C
ENDIF
    DIR_time			dw	?	; 16 file date
    DIR_date			dw	?	; 18 file time
    DIR_start			dw	?	; 1A starting cluster of file
    DIR_size			dd	?	; 1C file size in bytes
DirectoryEntry ends


;
;  ROM-DOS USE ONLY!!
;
;  This structure not used by that (we won't mention names)
;  other DOS.  It keeps track of a directory entry.
;  Note: this limits sectors to 4K in size maximum (80H
;  directory entries, 20H bytes each = 4K)
;
DIRENTRY struc
   DE_dirsec	dd	?	; sector # for this sector
   DE_dirent	db	?	; 0-7FH
   DE_eps	db	?	; entries per sector (10H at 512 byte sectors)
DIRENTRY ends




;   BIOS PARAMETER BLOCK (BPB)
;   Description: This structure is stored on disks in
;   sector 0, AT OFFSET 11d.
;
;   Reference: Adv MS-DOS, p 269  -or-  DOS Ref, p 222
;				  MSDN online documentation.
;
;	Note:	The FAT32 (DOS v7.1) implementation of this data structue
;			has additional fields.
;
;
; The BPB for FAT32 drives is an extended version of the FAT16/FAT12
; BPB. It contains identical information to a standard BPB, but also
; includes several extra fields for FAT32 specific information. This
; structure is implemented in Windows95 OEM Service Release 2 (OSR2)
; and later.
;
; A typical Win 95 OSR2 boot sector is shown below:
;
; 00 - EB 58 90 4D 53 57 49 4E-34 2E 31 00 02 08 20 00   .X.MSWIN4.1... .
; 10 - 02 00 00 00 00 F8 00 00-3F 00 80 00 3F 00 00 00   ........?...?...
; 20 - 41 97 60 00 20 18 00 00-00 00 00 00 02 00 00 00   A.`. ...........
; 30 - 01 00 06 00 00 00 00 00-00 00 00 00 00 00 00 00   ................
; 40 - 80 00 29 E7 12 49 25 46-41 54 33 32 2D 33 47 20   ..)..I%FAT32-3G
; 50 - 20 20 46 41 54 33 32 20-20 20 FA 33 C9 8E D1 BC     FAT32   .3....
; 60 - F8 7B 8E C1 BD 78 00 C5-76 00 1E 56 16 55 BF 22   .{...x..v..V.U."
; 70 - 05 89 7E 00 89 4E 02 B1-0B FC F3 A4 8E D9 BD 00   ..~..N..........
;
;
; The BPB fields that are of specific interest in FAT32 are:
;
; A_BF_BPB_RootEntries
;	This field is ignored on FAT32 drives. It is always 0.
; A_BF_BPB_SectorsPerFAT
;	This number of sectors per FAT field will always be zero in
; 	a FAT32 BPB. Use the values from A_BF_BPB_BigSectorsPerFat
;	and A_BF_BPB_BigSectorsPerFatHi for FAT32 media.
; A_BF_BPB_BigSectorsPerFat
;	The number of sectors per FAT on the FAT32 drive. This is the
;	first additional field in a BPB32.
; A_BF_BPB_BigSectorsPerFatHi
;	The high word of the FAT32 sectors per FAT value.
; A_BF_BPBExtFlags
;	Flags describing the drive. Bit 7 of this field indicates
;	whether or not information written to the active FAT will
;	be written to all copies of the FAT. The low 4 bits of this
;	field contain the 0-based FAT number of the Active FAT, but
;	are only meaningful if bit 8 is set. This field can contain
;	a combination of the following values:
;
;	BGBPB_F_ActiveFATMsk (000Fh) 	Mask for low four bits.
;	BGBPB_F_NoFATMirror (0080h)  	Mask indicating FAT mirroring
;					state. If set, FAT mirroring
;					is disabled. If clear, FAT
;					mirroring is enabled.
;	Bits 4-6 and 8-15 are reserved.
;
; A_BF_BPB_FS_Version
;	The file system version number of the FAT32 drive.
;	The high byte represents the major version, and the
;	low byte represents the minor version.
; A_BF_BPB_RootDirStrtClus
;	The cluster number of the first cluster in the FAT32
;	drive's root directory.
; A_BF_BPB_RootDirStrtClusHi
;	The high word of the FAT32 starting cluster number.
; A_BF_BPB_FSInfoSec
;	The sector number of the file system information sector.
;	The file system info sector contains a BIGFATBOOTFSINFO
;	structure. This field is set to 0FFFFh if there is no FSINFO
;	sector. Otherwise, this value must be non-zero and less than
;	the reserved sector count.
; A_BF_BPB_BkUpBootSec
;	The sector number of the backup boot sector. This field is set
;	to 0FFFFh if there is no backup boot sector. Otherwise, this
;	value must be non-zero and less than the reserved sector count.
; A_BF_BPB_Reserved
;	Reserved field. Do not use.
;
BiosParameterBlock struc

    BPB_sec_size		dw	?	; 00 bytes per sector
    BPB_sec_per_clus		db	?	; 02 sectors per cluster
    BPB_res_sec			dw	?	; 03 reserved sector count (before FAT)
    BPB_num_fats		db	?	; 05 count of FATs
IF USING_FAT32
    BPB_uDirEntries           	dw	?	; 06 count of root DIR entries (0 on FAT32)
ELSE
    BPB_dir_ent			dw	?	; 06 count of root DIR entries (FAT16 name)
ENDIF

    BPB_num_sec			dw	?	; 08 count of drive sectors (may be 0)
    BPB_media			db	?	; 0A media ID byte
IF USING_FAT32
    BPB_uFatSectors          	dw	?	; 0B count of sectors per fat (0 on FAT32)
ELSE
    BPB_fat_size		dw	?	; 0B count of sectors per fat (FAT16 name)
ENDIF
    BPB_sec_per_trak		dw	?	; 0D count of sectors per track
    BPB_sides			dw	?	; 0F count of disk heads (sides)
    BPB_hidden			dd	?	; 11 count of hidden sectors (before boot sector)
    BPB_lnumsec			dd	?	; 15 count of sectors (if num_sec==0)
IF USING_FAT32
    BPB_ulSectorsPerFat		dd	?	; 19 FAT32 sectors per FAT
    BPB_uExtendedFlags		dw	?	; 1D FAT32 extended flags (defined below)
    BPB_uFileSystemVersion	dw	?	; 1F FAT32 File system version number
    BPB_ulRootDirStartCluster 	dd	?	; 21 FAT32 Root dir start (can move)
    BPB_uFSInfoSector		dw	?	; 25 FAT32 sector # of FSINFO struc  (-1 if none)
    BPB_uBackUpBootSector	dw	?	; 27 FAT32 sector # of backup boot sector (-1 if none)
    BPB_Reserved		dw 6 dup(?)	; 29 FAT32 ???
ENDIF
BiosParameterBlock	ends

IF USING_FAT32
FAT16_BPB_SIZE		=	BPB_ulSectorsPerFat
ELSE
FAT16_BPB_SIZE		=	SIZE BiosParameterBlock
ENDIF


; Definitions for the FAT32 BPB uExtendedFlags field.
; B4-B6 and B8-B15 are reserved and zero, B0-B3 are only significant if
; B7 is set. FAT mirroring may be turned off when the 1st (or any subsequent
; mirror FAT) is damaged. FAT mirroring could conceivably be turned off when
; resizing a partition (or whatever) as well.
;
BPB_NO_FAT_MIRRORING 	=	80h   	; B7 set if NOT mirroring FAT
BPB_ACTIVE_FAT_MASK	=	0Fh	; B0-B3: active FAT index



; The boot sector on a FAT32 drive is layed out the same as for FAT16.
; The main difference is that, since the BPB32 is larger than the one
; for FAT16, the offsets of the subsequent fields are larger. Offsets
; for fields in both boot sectors are shown in comments for the
; BootInformationXX struc fields.
;
;BOOT_SECTOR struc
;   JumpInstruction		db	3 dup 	?	; 00
;   FileSystemOEMName		db	8 dup	?       ; 03
;   BootSectorBPB		BPB	<>              ; 0B
;   PhysicalDriveNumber		dw	80h             ; 24 (40)
;   ExtendedBootRecordSig	db	29h             ; 26 (42)
;   VolumeSerialNumber		dd	12345678	; 27 (43)
;   VolumeLabel			db	"No Label   "   ; 2B (47)
;   FileSystemID		db	"FAT32   "      ; 36 (52)
;   BootCodeStarts		db	?		; 3E (5A)
;BOOT_SECTOR ends
;
BOOT_VOL_LEN		=	11
BOOT_FATID_LEN		=	8
BOOT_SERIAL_LEN		=	4
BOOT_ID_TOTAL_LEN	=	(BOOT_SERIAL_LEN + BOOT_VOL_LEN + BOOT_FATID_LEN)
BOOT_SECTOR_SIGNATURE	=	0AA55h

IF USING_FAT32

BootInformation32 Struc
    BootSectorBPB32		db SIZE BiosParameterBlock dup(?)    	; 00B
    PhysicalDriveNumber32	dw	?                          			; 040
    ExtendedBootRecordSig32	db	?                          		 ; 042
    VolumeSerialNumber32	dd	?         	              ; 043
    VolumeLabel32		db	BOOT_VOL_LEN dup(?)           ; 047
    FileSystemID32		db	BOOT_FATID_LEN dup(?)         ; 052
    BootCodeStarts32		db (510-(11+BootCodeStarts32)) dup(?) ; 05A
BootInformation32 ends

BootInformation16 Struc
    BootSectorBPB16		db BPB_ulSectorsPerFat dup(?) 	 	   	; 00B
    PhysicalDriveNumber16	dw	?                   	      ; 024
    ExtendedBootRecordSig16	db	?               	      ; 026
    VolumeSerialNumber16	dd	?       		      ; 027
    VolumeLabel16		db	BOOT_VOL_LEN dup(?)           ; 02B
    FileSystemID16		db	BOOT_FATID_LEN dup(?)         ; 036
    BootCodeStarts16		db (510-(11+BootCodeStarts16)) dup(?) ; 03E
BootInformation16 ends

BootInformation union
    BootInfo16			BootInformation16 <>
    BootInfo32			BootInformation32 <>
BootInformation ends

BootSector struc
    BS_JumpInstruction		db	3 dup (?)	       	; 000
    BS_FileSystemOEMName	db	8 dup (?)               ; 003
    BS_BootSectorInfo		BootInformation <>		; 00B
    BS_BootSectorSignature	dw	BOOT_SECTOR_SIGNATURE	; 1FE
BootSector ends

BOOT_BPB	equ	0BH

ELSE	;; USING_FAT32

; Offsets in Boot Sector
;
BOOT_BPB	equ	0BH
BOOT_SERIAL	equ	27H
BOOT_VOL_STR	equ	2BH
BOOT_FATID_STR	equ	36H


ENDIF	;; USING_FAT32





IF USING_FAT32
.386

; File System Information Structure (aka BIGFATBOOTFSINFO (FAT32) by MS).
;
; Contains information about the file system on a FAT32 volume. This
; structure is implemented in Windows OEM Service Release 2 and later.
;
; ulSignature
; 	The signature of the file system information sector. The value
; 	in this field is FSINFOSIG (0x61417272L).
; ulFreeClusters
; 	The count of free clusters on the drive. Set to -1 when the
; 	count is unknown.
; ulNextFreeCluster
; 	The cluster number of the cluster that was most recently
	; allocated.
; Reserved.
; 	Reserved field. Do not use.
; uSignature2
;  	End of "boot sector" marker (0x55AA)
;
FSINFO_SECTOR_SIGNATURE	=	041615252h	; "RRaA"
FSINFO_STRUC_OFFSET	=	001E4h
FSINFO_SIGNATURE_VALUE	=       061417272h	; "rrAa"

FileSystemInfo struc
    FSINFO_ulSignature          dd    	FSINFO_SIGNATURE_VALUE	; 00
    FSINFO_ulFreeClusters	dd    	?			; 04
    FSINFO_ulNextFreeCluster    dd    	?			; 08
    FSINFO_ulReserved           dd    	3 DUP (?)		; 0C
    FSINFO_uReserved2		dw	?                       ; 18
    FSINFO_uSignature2		dw	BOOT_SECTOR_SIGNATURE	; 1A
FileSystemInfo ends

.8086
ENDIF ;; USING_FAT32



; PARTSTART
;
; Used only inside the DDT below, this structure
; contains DOS 4+ information for hard drives only.
; Floppy drives have a ULONG record of the last tick
; they were accessed on.
;
PARTSTART struc
    PS_ishd	dw	?	; set to 1 for Hard Disks
    PS_strak	dw	?	; Starting trak of this partition
PARTSTART ends



; This struc is used for drives < 8G and the numbers come from int 13h
; function 8h.
;
BiosLogicalParameters 	struc
    BLP_Tracks	 	DW	?	; 00h # of traks
    BLP_SectorsPerTrack	DB	?	; 02h # of sectors per trak
    BLP_Heads	 	DB	?	; 03h # of heads
    BLP_TotalSectors 	DD	?	; 04h total # of sectors on this drive
BiosLogicalParameters 	ends		; 08h


IF USING_BIG_DISKS
.386
LBA_BLOCK_SIZE  =	512

;  5-4-01 BT
;
;  This constant is not used.  I can not see how this value was
;  derived.
;
;MAX_CHS_SECTORS	=	0EC4000h	; 15,482,880 sectors MAX with CHS

; This structure is returned by EBIOS int 13h fn 48h, Get Drive Parameters.
; The call will only return meaningful information if int 13h fn 41h, Check
; Extensions Present, returns successfully. The information in this structure
; is defined in the Phoenix BIOS Enhanced Disk Drive Specification v3.0.
;
; NOTES:
;	(1) For logical sector information use int 13h fn 8h which reflects
;	    CHS value translation if required.
;	(2) If this value is > MAX_CHS_SECTORS, B1 of EBDPT_uInfoFlags is
;	    clear and the int 13h interface for CHS geometry is insufficient
;	    to describe the disk.
;	(3) Not currently supported.
;
EBiosDeviceParameterTable 	struc
    EBDPT_uBufferSize		dw SIZE EBiosDeviceParameterTable
    EBDPT_uInfoFlags		dw	?	; 02h -	see equates below
    EBDPT_ulPhysicalCylinders	dd	?	; 04h - (1) sector count
    EBDPT_ulPhysicalHeads	dd	?	; 08h - (1) head count
    EBDPT_ulPhysicalSPT		dd	?	; 0Ch - (1) sectors per track
    EBDPT_uqBigPhysicalSectors	dq	?	; 10h - (2) sector count
    EBDPT_uBytesPerSector	dw	?	; 18h - BPS count
IF 0
    EBDPT_lpDPTExtension	dd	?	; 1Ah - (3)
    EBDPT_uDevicePathKey	dw	?	; 1Eh - (3)
    EBDPT_cDevicePathLenth	db	?	; 20h - (3)
    EBDPT_cReserved1		db	?	; 21h - (3)
    EBDPT_uReserved2		dw	?	; 22h - (3)
    EBDPT_szHostBusType	        db 4 dup(?)	; 24h - (3)
    EBDPT_szInterfaceType	db 8 dup(?)	; 28h - (3)
    EBDPT_uqInterfacePath	dq	?	; 30h - (3)
    EBDPT_uqDevicePath		dq	?	; 38h - (3)
    EBDPT_cReserved3		db	?	; 40h - (3)
    EBDPT_cDevicePathChkSum	db	?	; 41h - (3)
ENDIF ;; 0
EBiosDeviceParameterTable 	ends

; EBIOS flags returned by int 13h fn 41h in CX. These are stored in the
; DDT_EBIOSAttributes byte of the DDT for the specified drive. Each BIT is
; set to one if the capability is supplied by the BIOS.
;
EBIOS_HAS_LBA_DISK_SUPPORT	=	01h
EBIOS_HAS_LOCK_EJECT_SUPPORT	=	02h
EBIOS_HAS_ENHANCED_SUPPORT	=	04h


; Bit values for flags in the EBDPT_uInfoFlags field above. B7-B15 are
; reserved and 0.
;
EBDPT_INFO_HANDLES_DMA_ERRORS	=	01h
EBDPT_INFO_CHS_GEOMETRY_VALID	=	02h
EBDPT_INFO_CAN_REMOVE_MEDIA	=	04h
EBDPT_INFO_SUPPORTS_VERIFY	=	08h
EBDPT_INFO_REPORTS_MEDIA_CHANGE	=	10h
EBDPT_INFO_CAN_LOCK_MEDIA	=	20h
EBDPT_INFO_NO_MEDIA_PRESENT	=	40h	; geometry set to max


; This is parameter packet passed to the EBIOS disk functions with
; int 13 fns 42h, 430xh, 44h and 47h. It uses large Linear Block Address
; sector numbers to overcome the int 13h interface limits for CHS geometries.
;
DeviceAddressPacket	struc
    DAP_ucPacketSize		db	SIZE DeviceAddressPacket
    DAP_ucReserved1		db	0
    DAP_ucBlockCount		db	?	; number of sectors to xfer (<=127)
    DAP_ucReserved2		db	0
    DAP_lpBuffer		dd	?	; Seg:Offs of transfer buffer (-1 = use flat address)
    DAP_ulStartingLBALow	dd	?	; Starting Linear Block Address (low)
    DAP_ulStartingLBAHigh	dd	0	; high 32-bits of starting LBA
    ; TD 11/14/2001 (IR 3854)
    ; These are defined in a Bios specification, but only having to do
    ; with flat model and only valid if lpBuffer = FFFF:FFFF.
    ; They also bumped the PacketSize to 0x18, which caused problems
    ;
    ; DAP_qpBufferLow     	dd	0	; Quad ptr to buffer (low 32-bits)
    ; DAP_qpBufferHigh		dd      0	; high 32-bits of 64-bit linear addr
DeviceAddressPacket	ends


.8086
ENDIF	;; USING_BIG_DISKS


BiosDriveParameters	struc
    BDP_LogicalParameters 	db SIZE BiosLogicalParameters dup(?)     ; 00h
IF USING_BIG_DISKS
    BDP_EBiosParameters		db SIZE EBiosDeviceParameterTable dup(?) ; 08h
    BDP_uEBIOSAttributes	dw	?			         ; 22h
ENDIF
BiosDriveParameters	ends



;  DRIVE DATA TABLE (DDT)
;
;  Description: One of these is allocated to each internal
;  drive unit (floppy and hard disks) during startup.
;  This is used by smartdrv, to determine what the floppies
;  and harddisks really are.
;
;  Reference: DOS Internals, p 609 and Undocumented DOS 2nd Ed, p 765
;
DATTR_NOTREMOVABLE			equ 0001h
DATTR_SUPPORTS_CHANGELINE		equ 0002h
DATTR_DISK_BPB_LOCKED			equ 0004h
DATTR_SECTORS_EQUALLY_SIZED		equ 0008h
DATTR_SHARES_PHYSICAL_DEVICE		equ 0010h
DATTR_ACTIVE_UNIT_FOR_SHARED_DEVICE	equ 0020h
DATTR_DISK_CHANGE_DETECTED		equ 0040h
DATTR_SET_DASD_BEFORE_FORMAT     	equ 0080h
DATTR_DISK_REFORMAT_DETECTED    	equ 0100h
DATTR_INVALID_MEDIA			equ 0200h ; Read/Write disabled
DATTR_LBA_REQUIRED			equ 0400h

DriveDataTable struc
   DDT_nextptr		dd	?	; 00H  next one in the list
   DDT_bios_drive	db	?	; 04H  BIOS drive #
   DDT_dos_drive	db	?	; 05H  dos_drive #
   DDT_bpb		db SIZE BiosParameterBlock dup(?) ; 06H the "current media" BPB (720K floppy)
   DDT_FatType		db	?	; 1FH (3Bh) 0=12 bit, 40H=16 bit, 20h=32 bit
   DDT_opencnt		dw	?	; 20H (3Ch) open count
   DDT_dtype		db	?	; 22H (3Eh) 0=360,1=1.2,2=720,5=hard,7=1.44
   DDT_dattr		dw	?	; 23H (3Fh) 1=not-removable, 2=chgline status 5=? 9=? 10=?
   DDT_dcyl		dw	?	; 25H (41h) # of cylinders
   DDT_defbpb		db SIZE BiosParameterBlock dup(?) ; 27H (43h) BPB for drive (1.44M drive)
IFE USING_FAT32
   DDT_unknown    	db 6 dup(?)	; 40H        ???
ENDIF
   DDT_curtrack        	db	?	; 46H (78h)  for floppy HST adjustment
   DDT_ddt_info		dd	?	; 47H (79h)  PARTSTART struc for this partition (last accessed tick count for floppies)
   DDT_volume		db 12 dup(?)	; 4BH (7Dh)  Volume Label from boot sector
   DDT_serial		dd	?	; 57H (89h)  Serial # from boot sector
   DDT_fatstr		db 9 dup(?)	; 5BH (8Dh)  "FAT12" from boot sector
IF USING_BIG_DISKS
   DDT_uEBiosDriveInfoFlags  dw	?	; 64h (96h)
   DDT_uEBiosDriveAttributes dw	?       ; 66h (98h)
ENDIF
DriveDataTable ends

; Bits used to identify the FAT type on a given drive in the DDT_FatType
; field.
;
DDT_FAT_TYPE_16	=	40h
DDT_FAT_TYPE_32	=	20h


; union aliases for ticks field
;
DDT_sttrak equ <word ptr [DDT_ddt_info+2]>	; starting track for HDs
DDT_ishd   equ <word ptr [DDT_ddt_info]>	; set to 1 if HD (BIOS drive # >= 80H)


PartitionTableEntry struc
    PART_BootFlag        	DB   	?	; 00
    PART_FirstHead      	DB   	?       ; 01
    PART_FirstSector    	DB   	?       ; 02
    PART_FirstTrack     	DB   	?       ; 03
    PART_FileSystemID   	DB   	?       ; 04
    PART_LastHead       	DB   	?       ; 05
    PART_LastSector     	DB   	?       ; 06
    PART_LastTrack      	DB   	?       ; 07
    PART_HiddenSectors    	DD   	?       ; 08
    PART_SectorCount     	DD   	?       ; 0c
PartitionTableEntry ends


; Hard Disk Partition IDs
;
DOS12       equ   01h
DOS16       equ   04h
DOSEXT      equ   05h
DOSBIG      equ   06h
DOS32       equ   0Bh
DOS32_LBA   equ   0Ch
DOSBIG_LBA  equ   0Eh
DOSEXT_LBA  equ   0Fh
RELIANCE    equ  0CCh

; Bootable partition flag (DOS)
;
BOOTABLE_PART	equ	80h

PARTOFF		EQU	1BEH	; offset of the partition information


; Format/Verify structure (440DH/842H, 862H)
;
FVBLOCK	struc
    FV_SpecFunc	db	 ?	; see FVSPECFUNC_xxx
    FV_Head	dw	 ?	; head #
    FV_Cylinder	dw	 ?	; track #
    FV_Tracks	dw	 ?	; # of tracks at one time
FVBLOCK	ends

; Format/Verify special functions
FVSPECFUNC_FMT_1TRACK	equ 0
FVSPECFUNC_RESET	equ 1
FVSPECFUNC_FMT_TRACKS	equ 2


SDPDFormatStruc	struc
    SetDPB_Size        DW    ?
    SetDPB_Level       DW    ?
    SetDPB_Function    DD    ?
    SetDPB_Value1      DD    ?
    SetDPB_Value2      DD    ?
    SetDPB_Value3      DD    ?
    SetDPB_Value4      DD    ?
SDPDFormatStruc	ends

; Set/Get Device Parameters (440DH/840H, 860H)
DEVICEPARMS struc
    DP_SpecFunc		db ?		; 00h see DPSPECFUNC_xxx
    DP_DevType		db ?		; 01h device type
    DP_DevAttr		dw ?		; 02h device attr
    DP_Cylinders	dw ?		; 04h # of cylinders (tracks)
    DP_MediaType	db ?		; 06h media type
    DP_Bpb		db SIZE BiosParameterBlock dup(?) ; 07H current BPB
DEVICEPARMS ends

IF USING_FAT32
MAX_SECTORS_PER_TRACK	= 3fh

A_SECTORTABLE struc
    AST_uSectorNumber        dw     ?
    AST_uSectorSize          dw     ?
A_SECTORTABLE ends

EA_DEVICEPARAMETERS struc
    EDP_SpecialFunction      db     ?
    EDP_DeviceType           db     ?
    EDP_DeviceAttributes     dw     ?
    EDP_Cylinders            dw     ?
    EDP_MediaType            db     ?
    EDP_Bpb                  db     SIZE BiosParameterBlock DUP (?)
    EDP_Reserved1            db     32 DUP (?)
    EDP_TrackTableEntries    dw     ?
    EDP_SectorTable          db     MAX_SECTORS_PER_TRACK * SIZE A_SECTORTABLE DUP (?)
EA_DEVICEPARAMETERS ends
ENDIF

; bit mask of special functions
;
DPSPECFUNC_DEFBPB	equ 1	; 0="disk" BPB, 1="drive" BPB
DPSPECFUNC_LAYOUT	equ 2	; 1 = tracks/spt only
DPSPECFUNC_SAMESIZE	equ 4	; always set


IF USING_FAT32

;	Extended Get Free Space Structure
;
;	This is structure used for getting the free space for a
;	high capacity drive when using windows 95 OSR2 or above
;	using int 21h fn 7303h.
;
GetDiskFreeSpace32 	struc
    GDF_uSize		    		dw	?	;
    GDF_uLevel				dw	?	;
    GDF_ulSectorsPerCluster		dd	?	;
    GDF_ulBytesPerSector		dd	?	;
    GDF_ulAvailableClusters		dd	?	;
    GDF_ulTotalClusters			dd	?	;
    GDF_ulAvailablePhysicalSectors	dd	?	;
    GDF_ulTotalPhysicalSectors		dd	?	;
    GDF_ulAvailableAllocationUnits  	dd	?      	;
    GDF_ulTotalAllocationUnits		dd	?	;
    GDF_ulReserved	    		dd    2 dup(?)	;
GetDiskFreeSpace32 	ends

ENDIF	;; USING_FAT32

;	List of Lists (LOL) Structure,  DOS 5
;
;	Description: Also called SysVars this structure provides
;	access to DOS internal structures. It is mainly static,
;	and is initialized at boot time.  See INT 21H, function
;	52H.
;
;	Reference: (DOS Ref, p 964)
;
ListOfLists	struc
	LOL_unknown	db  1Eh dup (?)
	LOL_firstbuff	dd	?	; 1EH  current buffer in chain
	LOL_buffoff	dw	?	; 22H  offset within that buffer
	LOL_firstmcb	dw	?	; 24H  1st Memory Control Block
	LOL_dpb_ptr	dd	?	; 26H  Ptr to 1st Drive Parm Blk
	LOL_sft_ptr	dd	?	; 2AH  Ptr to 1st System File Table
	LOL_clk_dev 	dd	?	; 2EH  Ptr to Clock Device
	LOL_con_dev 	dd	?	; 32H  Ptr to Console Device
	LOL_max_sec	dw	?	; 36H  Maximum Sector Size
	LOL_buf_ptr	dd	?	; 38H  Ptr to info for BUFFER structures
	LOL_cds_ptr	dd	?	; 3CH  Ptr to the CDS table
	LOL_fcb_ptr	dd	?	; 40H  Ptr to FCB Table (SFTs)
	LOL_fcb_siz	dw	?	; 44H  # of SFTs in FCB Table
	LOL_num_drives  db	?	; 46H  # of Drives (DPB structures)
	LOL_last_drive  db	?	; 47H  # of CDS structures
ListOfLists	ends
