//
//   $Id: FATLIB.H 1.5 2002/10/15 21:41:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  FATLIB.H
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This module contains the functions in the FAT library.  These functions
//  determine the FAT type all call the private functions specific to that
//  FAT type.  All special values above HICLUS are "unified", to be the
//  same as what would be returned from a FAT32 drive.  These "unified"
//  values are stripped away when a value is being set to the FAT.  They
//  are "unified" for all returned cluster values.
//
//
//  This module was designed to utilize a buffering scheme.  However, a
//  buffering scheme is NOT required in order to utilize this module so
//  long as you provide the functions that the library calls.  Listed
//  below are the required functions and there descriptions.
//
//
//  Locate a specific sector for a drive in the buffers.
//  Sets the ppBufferHandle variable with a handle used
//  by the other buffer functions. This function returns
//  TRUE if the function was successful, FALSE otherwise.
//
bool PASCAL ReadSectorBuffer( uchar ucDOSDriveNumber,
		ulong ulAbsSector, void ** ppBufferHandle );
//
//  Retrieve an address in conventional memory where
//  the contents of the buffer are located.  Use the
//  buffer handle returned by the ReadSectorBuffer
//  function.  This function returns TRUE if successful,
//  FALSE otherwise.
//
bool PASCAL GetBufferDataAddress( void * pBufferHandle,
		void far * * plpBufferAddress );
//
//  Mark a buffer as dirty (modified).  Use the buffer
//  handle returned by the ReadSectorBuffer function.
//  This function returns TRUE if successful, FALSE
//  otherwise.
//
bool PASCAL MarkBufferDirty( void * pBufferHandle );

//  Change the number of copies of this sector and
//  specify the distance between the copies.  Use the
//  buffer handle returned by the ReadSectorBuffer
//  function.  The number of copies of this sector is
//  set to 1 by ReadSectorBuffer by default.
//
bool PASCAL SetBufferCopiesAndDistance( void * pBufferHandle,
		unsigned uCopiesOfThisSector, ulong ulDistanceToNextSector );
//
//
//  History:
//
//  04-06-1999 bt  Created
//


//  Retrieve a cluster value for the specified cluster number.  You are
//  required to pass it a far pointer to the DPB so that it can determine
//  the FAT type.  This function returns TRUE if it was successful.  If
//  there is an error a FALSE value will be returned.  You can determine
//  the nature of the error be examining the error code returned by the
//  abs error code (error number for absread/abswrite functions).  If
//  an invalid cluster number is passed to this function it will assert.
//  Invalid cluster numbers are less than 2 and greater than
//  lpDPB->hiclus (lpDPB->ulMaxCluster).
//
//  Note: The value being set at the specified cluster is NOT range checked.
//
bool PASCAL GetFatEntry( LPDPB lpDPB, ulong ulClusterNumber,
	ulong * pulClusterValue );


//  Set a cluster value for the specified cluster number and return the
//  value which was previously at that cluster number.  You are required
//  to pass it a far pointer to the DPB so that it can determine the FAT
//  type.  This function returns TRUE if it was successful.  If there is
//  an error a FALSE value will be returned.  You can determine the nature
//  of the error be examining the error code returned by the abs error code
//  (error number for absread/abswrite functions).  If an invalid cluster
//  number is passed to this function it will assert.  Invalid cluster
//  numbers are less than 2 and greater than lpDPB->hiclus
//  (lpDPB->ulMaxCluster).
//
//  Note: The value being set at the specified cluster is NOT range checked.
//
bool PASCAL PutFatEntry( LPDPB lpDPB, ulong ulClusterNumber,
	ulong ulNewClusterValue, ulong * pulOldClusterValue );
