;  $Id: DOSASM.ASM 1.18 2003/05/20 14:04:54Z thomd Exp $
;
;  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;  Copyright (c) 1990-2003 Datalight, Inc.
;  All Rights Reserved
;
;  Datalight, Inc. is a Washington corporation located at:
;
;  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
;  Bothell, WA  98021                    Fax: +1.425.951.8094
;  USA                                   Http://www.datalight.com
;
;  This software, including without limitation all source code and
;  documentation, is the confidential, trade secret property of
;  Datalight, Inc., and is protected under the copyright laws of
;  the United States and other jurisdictions.  Portions of the
;  software may also be subject to one or more the following US
;  patents: US#5860082, US#6260156.
;
;  In addition to civil penalties for infringement of copyright
;  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;  penalties for violation of (a) the restrictions on circumvention
;  of copyright protection systems found in 17 U.S.C. 1201 and
;  (b) the protections for the integrity of copyright management
;  information found in 17 U.S.C. 1202.
;
;  U.S. Government Restricted Rights.  Use, duplication,
;  reproduction, or transfer of this commercial product and
;  accompanying documentation is restricted in accordance with
;  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
;  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
;  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;  THIS SOFTWARE IN ERROR.
;
COMMENT ~
-----------------------------------------------------------

   DOSASM.ASM  created 11/30/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file contains the following items:

   1.  Segment Ordering Definitions
   2.  List Of Lists and Swappable Data Area
   3.  Startup Code
   4.  Segment Location Functions

   This file contains all segment definitions and must be
   the very first file in the link command line to enforce
   proper segment ordering.

   This file also contains some data that cannot be moved
   or rearranged.  This data appears in the form of the List
   Of Lists (LOL) and the Swappable Data Area (SDA).  These
   two data structures allow access to the most important DOS
   structures and are used extensively by Networks, TSRs and
   other low level software.

   The segment ordering must be:
   ----------------------------
   BIOSEXT	for BIOS extension booting
   TEXT		the code of ROM-DOS
   DEVDATA	device headers
   DATA		the initialized data of ROM-DOS
   BSS		the uninitialized data of ROM-DOS
   HEAP		the beginning of the memory pool

   $DOCHISTORY
   09/15/1998 dg  Moved Super-Boot variables in from DEVFDHDI.ASM
                  so that the hard disk driver isn't included in all
                  versions. DOSCNTRY.C references these variables.

----------------------------------------------------------- ~
include macros.inc
include dosdef.inc


;  As soon as ROMDOS gets control from the BIOS, it must
;  switch stacks to a temporary stack so that ROMDOS knows
;  where and how much stack it has.  ROMDOS uses the
;  temporary stack up until it switches to the boot stack.
;  Below is the size in bytes of the temporary stack.
;
TEMP_STACK_BYTES	equ	80h


COMMENT ~
----------------------------------------------------------------

	SECTION 1.  Segment ordering definitions

	The following is Segment Definition and Ordering

---------------------------------------------------------------- ~


COMMENT ~ if we are booting via BIOS extension ~
_BIOSEXT SEGMENT PARA PUBLIC 'BIOSEXT'
	public	begbiosext
	begbiosext	label byte
_BIOSEXT ends

comment ~ the actual code of ROM-DOS ~
_BEGTEXT segment para public 'CODE'
_BEGTEXT ends
cBegCode
	public	beg_text
	beg_text label byte
cEndCode

comment ~ the actual code of ROM-DOS LFNs~
_LFN_TEXT segment para public 'CODE'
	public	beg_lfn_text
	beg_lfn_text label byte
_LFN_TEXT ends

comment ~ ROM-DOS internal device headers ~
; for devdata to begin on a paragraph boundary (usually 70:0)
_BEGDEVDATA segment para public 'DEVDATA'
_BEGDEVDATA ends
_DEVDATA segment byte public 'DEVDATA'
	public beg_devdata
	beg_devdata label byte

	; a word MUST be here so no devhdr can end up at DEVDATA:0.
	public	romdos_cs
	romdos_cs	dw	?
_DEVDATA ends

; segment for Microsoft C
_FAR_DATA segment para public 'FAR_DATA'
_FAR_DATA ends

comment ~ initialized data, SDA and LOL ~
; force DATA to begin on a paragraph boundary (usually 8B:0)
_BEGDATA segment para public 'DATA'
_BEGDATA ends

COMMENT ~ initialized data goes here ~
_DATA segment byte public 'DATA'
	public	beg_data
	beg_data label byte
_DATA ends

; segment for Microsoft C
_CONST segment byte public 'CONST'
_CONST ends

; segment for Microsoft C
_MSG segment byte public 'MSG'
_MSG ends

COMMENT ~ uninitialized data, must be zeroed ~
_BSS segment byte public 'BSS'
	public	beg_bss
	beg_bss label byte
_BSS ends

; put all data segments in DGROUP
DGROUP	GROUP	_DATA,_CONST,_MSG,_BSS


COMMENT ~ start of the DOS memory pool ~
_HEAP	segment	para public 'HEAP'
	public	beg_heap
	beg_heap label byte
_HEAP	ends


COMMENT ~
----------------------------------------------------------------

	SECTION 2.  List Of Lists and Swappable Data Area

	The following code and data must be in a specific
	order and place in memory.  The List of Lists is
	the very first DATA item in DOS, and so is placed
	here.

---------------------------------------------------------------- ~

cBegData
	;---------------------------------------------------
	;
	;  List Of Lists (LOL)
	;  The Seg:Off of "dos_vars" is returned in ES:BX
	;  by Int 21H(52H).  Some programs "assume" BX is
	;  26H, so we are forced to put it at this exact
	;  location.
	;
	;---------------------------------------------------

	PUBLIC	_lol
_lol	LABEL BYTE

	DB	04H DUP (0)	; 00H ???
dos5	DB	0		; 04H will be set to 1 by init_sda
	DB	09H DUP (0)	; 05H ???
netbios_id DW	0		; 0EH net bios identification
	DB	0AH DUP (0)	; 10H ???

	public	_share_retry, _share_delay
_share_retry DW	3	; 1AH  sharing retry
_share_delay DW	1	; 1CH  sharing delay before retry
buf_0	DD	?	; 1EH  current buffer in chain
bufoff	DW	?	; 22H  offset within that buffer
_mcb1	DW	?	; 24H  1st Memory Control Block

	PUBLIC	_dos_vars
_dos_vars	LABEL WORD	; address returned by INT 21H, func 52H

dpb_ptr	DD	?	;  26H  Ptr to 1st Drive Parm Blk
sft_ptr	DD	?	;  2AH  Ptr to 1st System File Table
clk_dev DD	?	;  2EH  Ptr to Clock Device
con_dev DD	?	;  32H  Ptr to Console Device
max_sec	DW	?	;  36H  Maximum Sector Size
buf_ptr	DD	?	;  38H  Ptr to info for BUFFER structures
cds_ptr	DD	?	;  3CH  Ptr to current directory structures
fcb_ptr	DD	?	;  40H  Ptr to FCB Table (SFTs)
fcb_siz	DW	?	;  44H  # of SFTs in FCB Table
NUM_DRIVES DB	?	;  46H  # of Drives (DPB structures)
LAST_DRIVE DB	?	;  47H  # of CDS structures

	;---------------------------------------------------
	;
	;  The NUL device driver is part of the List Of
	;  Lists, so here it is...
	;
	;---------------------------------------------------

	ASSUME CS:DGROUP, DS:DGROUP
	EXTRN	nuld_strat:near, nuld_serv:near
	PUBLIC	_nulx
_nulx:
	dw 	-1,-1	;  48H  Header for NUL device
	dw	8004H
	dw	offset nuld_strat
	dw	offset nuld_serv
	db	"NUL     "

	public	bootdrive, extmemsize

nulbyte    DB	?	;  5AH  Terminating char for nul device
specprogs  DW	?	;  5BH  ???
utilfuncs  DD	?	;  5DH  ???
ifschain   DD	?	;  61H  ???
numbufs    DW	?	;  65H  Total BUFFERS
lookahead  DW	?	;  67H  ???
bootdrive  DB	?	;  69H  A=1, B=2, etc
nulbyte2   DB	1	;  6AH  emulated disk??
extmemsize DW	?	;  6BH  Extended memory size

	   DD	?		;  6DH  unknown ptr???

	PUBLIC _mem_top, _retfar
		DB ?		; ???
_mem_top	DW ?		; 72H,  A000H paragraphs = 640K
_retfar:			; 74H  rd specific:
	clc
	retf

	public	_ptrsav, _ROM_DS, _ROM_DEVSEG, _dos_code_seg
	_ptrsav		DD ?	; 76H  rd specific: ptr to device hdr
	_ROM_DS		DW ?	; 7AH  rd specific: segment of DGROUP
	_ROM_DEVSEG	DW ?	; 7CH  rd specific: segment of devices
	_dos_code_seg	DW ?	; 7EH  rd specific: tossing init code
	DB	9 dup (?)	; 80H	unknown data

	public	_umb_link, _umb_arena
_umb_link	DB ?		; 89H   UMBs hooked(1)/unhooked(0)
		DW ?		; 8AH	??unknown data
_umb_arena	DW -1		; 8CH	arena, UMB link?
		DW ?		; 8EH	??unknown data

	;---------------------------------------------------
	;
	;  This table's purpose is unknown, but believed to
	;  be part of the SHARE file protection process
	;  The table MUST begin at offset 90H in DOS' data
	;  area.
		public	_share_table
	_share_table	dd	15 dup (?)	; 90H

	;---------------------------------------------------
	;
	;  Offset CCH.  We have some "dead space" here that
	;  can be used for data.
	;
	;---------------------------------------------------

	public	_bufinfo
	_bufinfo	DB	14H dup (?)

	; the floppy/hard disk driver set this on a critical error
	public	_bios_err
	_bios_err	DB	?

	public	_con_buf,_con_bufpos,_con_col
	_con_buf	DB	130 DUP (?)
	_con_bufpos	DB	?
	_con_col	DB	?


	;-------------------------------------------------
	;  Networking support?
	;  The Machine Name MUST begin at offset 306H in
	;  the DOS Data area.  Do NOT move this piece.
	;

	; machine name
	ORG	301H
	_switch_char	DB	'/'			; 301H
	_mem_fit	DW	?			; 302H
	_5E_flag	DB	?			; 304H
	_machine_name	DB	16 dup (?)		; 305H
	_windows_patch	label word
		DW	4 dup (offset _patch_byte)	; 315H
		DW	0				; 31DH

	;-------------------------------------------------
	; Swappable Data Area (SDA)
	;
	; This area MUST be in the exact following order!!
	; The relative location of these variables is
	; critical to the operations of IFS calls (CD ROM
	; drives, Networks).  We would have placed these
	; items in a C file, but unfortunately C compilers
	; tend to rearrange the data.
	;
	;-------------------------------------------------
	ORG	320H
					; rd=ROM-DOS specific
_CritErr	db	?		; 320H
_InDOS		db	?		; 321H
_crit_drive	db	?		; 322H
_err_locus	db	?		; 323H
_err_extended	dw	?		; 324H
_err_action	db	?		; 326H
_err_class	db	?		; 327H
_err_esdi	dd	?		; 328H
_DTA		dd	?		; 32CH
_current_psp	dw	?		; 330H
_int28_sp	dw	?		; 332H
_return_code	dw	?		; 334H
_CUR_DRIVE	db	?		; 336H
_BREAK_FLAG	db	?		; 337H
_error_no	dw	?		; 338H  rd
_int21_ax	label	word		; 33AH
_int21_al	db	?		; 33AH
_int21_ah	db	?		; 33BH
_err_psp 	dw	?		; 33CH  cur PSP at time of last error
_machine_id	dw	?		; 33EH  for windows and networks
_stacksNum  dw DEFAULTSTACKSNUM  ; 340h rd the n of STACKS=n,s
_stacksSize dw DEFAULTSTACKSSIZE ; 342h rd the s of STACKS=n,s
_paras_total	dw	0		; 344h  value*64 from 1st int 12h call
_pDiskStack	dw GOFFSET _disk_stackbase ; 346h  rd (STACKS in config.sys)
_pCharStack	dw GOFFSET _char_stackbase ; 348h  rd (STACKS in config.sys)
_CritIgnoreOK	db	?		; 34Ah  rd: Allow Ignore (not a subdir)
_crit_locus	db	?		; 34Bh	crit locus set by MS net
_pDiskStackTop	dw GOFFSET _disk_stack 	; 34Ch  rd (STACKS in config.sys)
_pCharStackTop	dw GOFFSET _char_stack 	; 34Eh  rd (STACKS in config.sys)
_day		db	?		; 350H  1-31
_month		db	?		; 351H  1-12
_year		dw	?		; 352H  from 1980
_day_count	dw	?		; 354H  from 1-1-1980
_day_of_week	db	?		; 356H  0=sunday
_verify_flag	db	?		; 357H	rd
_int28_OK	db	?		; 358H  1=ok to int 28 (0<fn#<0Dh)
_prn_as_well	db	?		; 359H  rd
_sda_reqhdr	db	30 dup (?)	; 35AH  struct dev_req_hdr
_sda_devptr	dd	?		; 378H  rd
_crit_stack	dd	?		; 37CH  rd
_err_already_set db 	?		; 380H  rd err
_hitbrk		db	?		; 381H  rd
_special_ret	dw	?		; 382H  rd
_sda_rdwrbuff	dd	?		; 384H  rd
sda_dispatch	dw	?		; 388H  rd
_timecheck	db	?		; 38AH	rd
_old_reqhdr	db	30 dup (?)	; 38BH  rd struct dev_req_hdr
_lock_buff	db	8 dup (?)	; 3A9H  int 2F fn 110Ah (un)lock data
_sda_no24	dw	?		; 3B1H	rd
_sda_1st_blank	dw	?		; 3B3H  rd
_sda_5a_strlen	db	?		; 3B5H	rd
_clock_txd	db	6 dup (?)	; 3B6H
_use_stdout	dw	?		; 3BCH  rd
_sda_fname1	db	128 dup (?)	; 3BEH
_sda_fname2	db	90 dup (?)	; 43EH
; warning! 2nd rename string and sda_fcb overlap
; there was no other place in the SDA to put "sda_fcb"
; Also see description of SDB in DOSSTRUCT.H for sda_fcb
; use by find first/next FCB calls
_sda_fcb	db	38 dup (?)	; 498H	rd struct FCB
_sda_sdb	db	21 dup (?)	; 4BEH	struct SDB
_sda_ffb	db	32 dup (?)	; 4D3H	struct DIR
_sda_ifs_cds	db	88 dup (?)	; 4F3H	struct CDS
_sda_ename1	db	12 dup (?)	; 54BH
_sda_ename2	db	12 dup (?)	; 557H
unknown5	db	10 dup (?)	; 563H  ??
_sda_srchattr	db	?		; 56DH
_sda_openmode	db	?		; 56EH
unknown5a	db	?		; 56FH  ??
		db	?		; 570H  ??
_sda_splice	db	0FFH		; 571H  ?? splice flag?
_sda_invoke	db	?		; 572H  ?? invoke flag
unknown5b	db	04H		; 573H  ??
_in_ifs_crit_err db	0		; 574H  rd: inside int 2Fh, fn 1206h
unknown5c	db	?		; 575h  ??
_SuperID    	db 0			; 576H  rd: 0 means not installed
_SuperOrder 	db 0			; 577H  rd: first, middle, last
_SuperBoot  	db 0			; 578H  rd: boot from SB drive?
_SuperBootDrive db 0FFH			; 579H  rd: booted from this drive
		db	6 dup (?)	; 57AH  ??
_sda_dpbptr	dd	?		; 580H
_userregs	dd	?		; 584H
_int23_sp	dw	?		; 588H	rd
_user_ds	dw	?		; 58AH  rd
_sda_cur_buf	dd	?		; 58CH	rd
_minute		db	?		; 590H  rd
_hour		db	?		; 591H  rd
_hund		db	?		; 592H  rd
_seconds	db	?		; 593H  rd
newByte		db	?		; 594H  rd
comport		db	?		; 595H  rd
retries 	db	?		; 596H  rd
comtemp 	label byte
cominit		db	?		; 597H  rd
byteQueued	db	?		; 598H  rd
con_que		db	?		; 599H  rd
con_byte 	db	?		; 59AH	rd
unknown6	db	3 dup (?)	; 59BH  ??
_sda_sftptr	dd	?		; 59EH
_sda_cdsptr	dd	?		; 5A2H
_sda_fcbptr	dd	?		; 5A6H  rd
_sda_sft_h	db	?		; 5AAH  MS-DOS word: SFT# (for IFS)
_sda_sft_h2	db	?		; 5ABH  NT client writes words
daylight	db	?		; 5ACH  rd
rollover	db	?		; 5ADH  rd
prnport		db	?		; 5AEH  rd
prninit		db	?		; 5AFH  rd
_sda_in_5D	db	?		; 5B0H  rd
unknown7	db	?		; 5B1H
_fname1ptr	dw	?		; 5B2H
_fname2ptr	dw	?		; 5B4H
unknown7a	db	4 dup (?)	; 5B6H  ??
_oldhandle	dw	?		; 5BAH  rd
unknown7b	dd	?		; 5BCH  ??
_old_sfth	db	?		; 5C0H  rd
_trydoshi	db	?		; 5C1H  rd
_using_boot_stack db	?               ; 5C2H  rd no shell running
unknown7c	db	11 dup (?)	; 5C3H  ??
_sda_rdwroff	dd	?		; 5CEH
_sda_retry	db	?		; 5D2H  rd
_sda_fcbnext	db	?		; 5D3H  rd
_sda_numfcbs	db	?		; 5D4H  rd  (used instead of lol.numfcbs)
_sda_numsfts	db	?		; 5D5H  rd
unknown7d       db      ?               ; 5D6H  ???
_sda_extfcb	db	?		; 5D7H  rd
_sda_fcbattr	db	?		; 5D8H  rd
_sda_chardev	db	?		; 5D9H  rd
_sda_callflags	dw	?		; 5DAH  rd
_sda_handle	dw	?		; 5DCH  rd
_sda_rdwrlen	dw	?		; 5DEH  rd
_sda_rdwrfcb	dw	?		; 5E0H  rd
_sda_buffptr	dd	?		; 5E2H  rd
_sda_drive	db	?		; 5E6H  rd
_sda_isnet	db	?		; 5E7H  rd
_sda_isdevpath	db	?		; 5E8H  rd
unknown8	db	?		; 5E9H  ??
_sda_bx		dw	?		; 5EAH
_sda_ds		dw	?		; 5ECH
unknown8a	db	2 dup (?)	; 5EEH  ??
_olduserregs	dd	?		; 5F0H nested userregs (words reversed)
_sopen_action   dw	?		; 5F4h NOVEL: action code for ext. open
unknown8b	db      7 dup (?)	; 5F6H  ??
_sopen_act	dw	?		; 5FDH
_sopen_attr	dw	?		; 5FFH
_sopen_mode	dw	?		; 601H
		dw	?		; 603H  ??

_def_volume	DB "NO NAME    "	; 605h  rd see DOS call 440DH (x66H)
_def_fatid	DB "FAT12   " 		; 610h	rd

unknown9	db	08 dup (?)	; 618h  ??

;unknown9	db	27 dup (?)	; 605H  ??


_ren_sdb	db	21 dup (?)	; 620H	struct SDB
_ren_ffb	db	32 dup (?)	; 635H	struct DIR
		db	?		; 655H  ??
_disk_stackbase	db	DEFAULTSTACK dup (?)	; 656H
_disk_stack	label	word		; 8C2H
_char_stackbase	db	DEFAULTSTACK dup (?)	;
_char_stack	label	word		; B2EH

	; International Support
		 DB	?		; B2EH	??
_uppercase_table DB	82H dup (?)	; B2FH	uchar uppercase_table[0x82];
_term_table	 DW	16H		; BB1H	uchar term_table[0x18];
	DB 01H,00H,0FFH,00H,00H,20H,02H,0EH
	DB '.','"','/','\','[',']',':','|'  ;	/* "/\[]:|<>+=;, */
	DB '<','>','+','=',';',',', 0	    ;	NULL terminated char list
		 DB	 17H dup (?)	; BCAH	??
_collating_table DB	102H dup (?)	; BE1H	uchar collating_table[0x102];
_lead_byte_start DW	?		; CE3H
_lead_byte_table DB	 10H dup (?)	; CE5H	uchar lead_byte_table[0x10];
_casemap:				; CF5H	case map function
	cmp	AL, 80H
	jb	casedone
	push	BX
	mov	BX, GOFFSET _uppercase_table - 7EH
	db	2EH	; CS: prefix
	xlat
	pop	BX
casedone:
	retf

	; Leap year table inside of DOS
	org 0D14h
	public _days_per_year
_days_per_year		label byte	; D14h
	db	0C8h,0A6h
	db  0C8h,0A5h
	db  0C8h,0A5h
	db  0C8h,0A5h
	; Month list inside of DOS
	org	0D1CH
	public	_days_per_month
_days_per_month		label byte	; D1CH
	DB	31 	; January
	DB	28 	; February
	DB	31 	; March
	DB	30 	; April
	DB	31 	; May
	DB	30 	; June
	DB	31 	; July
	DB	31 	; August
	DB	30 	; September
	DB	31 	; October
	DB	30 	; November
	DB	31	; December

	; /* Filename Uppercase Table */
	; uchar patch_uppercase[0x24]	/* patch starts at offset 1 */
_patch_uppercase	label byte	; D28H rd only
	DB	    9AH,45H,41H,8EH,41H,8FH,80H
	DB	45H,45H,45H,49H,49H,49H,8EH,8FH
	DB	90H,92H,92H,4FH,99H,4FH,55H,55H
	DB	59H,99H,9AH,9BH,9CH,9DH,9EH,9FH
	DB	41H,49H,4FH,55H,0A5H

	; /* Country Collating Table */
	; uchar patch_6506[0x30] 	/* patch starts at offset 0x80 */
_patch_6506	label byte		; D4CH rd only
	DB	43H,55H,45H,41H,41H,41H,41H,43H
	DB	45H,45H,45H,49H,49H,49H,41H,41H
	DB	45H,41H,41H,4FH,4FH,4FH,55H,55H
	DB	59H,4FH,55H,24H,24H,24H,24H,24H
	DB	41H,49H,4FH,55H,4EH,4EH,0A6H,0A7H
	DB	3FH,0A9H,0AAH,0ABH,0ACH,21H,22H,22H

_shell	DB	128 dup (?)		; D7CH rd only

; NOTE: _cfg_seg and _cfg_size must follow one another so that our
;       int 21h fn 30DBh interface to find the size and location of
; 	the CONFIG.SYS buffer works.
;
_cfg_seg	DW	0		; DFCH rd only
_cfg_size	DW	?		; DFEH rd only
_cfg_position	DW	?		; E00H rd only
_cfg_linenum	DW	?		; E02H rd only

MAX_INCLUDE_DEPTH equ 3		; see also DOSCFG.C
_cfg_inc_pos	DW	MAX_INCLUDE_DEPTH DUP (?)	; E04H rd only
_cfg_inc_line	DW	MAX_INCLUDE_DEPTH DUP (?)	; E0AH rd only
_cfg_inc_level	DW	?		; E10H rd only
_cfg_ask	DW	?		; E12H rd only
_cfg_ask_line	DW	?		; E14H rd only
_cfg_default	DW	?		; E16H rd only
_cfg_num_menus	DW	?		; E18H rd only
_cfg_menuptr	DW	?		; E1AH rd only
_cfg_menupos	DW	?		; E1CH rd only
_cfg_nobypass	DW	?		; E1EH rd only

_int2f_ax	DW	?		; E20H rd only
_int2f_bx	DW	?		; E22H rd only
_int2f_cx	DW	?		; E24H rd only
_int2f_dx	DW	?		; E26H rd only
_int2f_di	DW	?		; E28H rd only

_dos_psp	DW	?		; E2AH rd only
_root_psp	DW	?		; E2CH rd only
__xSS		DW	?		; E2EH rd only
__xSP		DW	?		; E30H rd only
__xCS		DW	?		; E32H rd only
__xIP		DW	?		; E34H rd only
__xDS		DW	?		; E36H rd only
__xIsEXE	DW	?		; E38H rd only

_boot_prog	DW	?		; E3AH rd only
_boot_parms	DB 16H DUP (?)		; E3CH rd only
_boot_env	DW	?		; E52H rd only
_boot_env_size	DW	?		; E54H rd only
_temp_cds	DW	?		; E56H rd only
_lo_buffer	DD	?		; E58H rd only
_booting	DB	?		; E5CH rd only
_FAT_type   DB ?  ; E5Dh rd only

	public	_tmp_name
_tmp_name	DB 67 dup (?)		; E5EH rd only
					; EA0H
               ; EC5h !!!

; BT 1-19-00 ETR 2457
;
; the end of our sda is larger than MSDOS because we don't
; only use their variable data space but we also have our
; own romdos variables.  Network software like (MSDOS) NET
; will copy the entire sda and execute a call then restore
; the sda so long as we are not in a critical section.
; Therefor, we need mark the end of the (MSDOS) sda plus our
; variables.
;
_end_sda	label	word

	; store version at EBBH for Windows
	org	0EBBH
_dos_win_ver	dw	?		; EBBH

	; some DOS behavior is modified if we are in Windows
	; in_windows tells DOS to generate 2AH calls and NOT
	; to zero machine id
	public	_in_windows
_in_windows	label	byte
_patch_byte	db	10 dup (?)	; EBDH

	public	_config_str, _country_buff
_config_str	db	128 dup (?)	; EE5H rd specific
_country_buff	db	66 dup (?)	; F65H rd specific
					; FA7H

	org	1022H
		public	windows_data
windows_data	label	word
	DW	5		; DOS "version"			1022H
	DW	offset _sda_ds	; DS stored on entry to DOS	1024H
	DW	offset _sda_bx	; BX stored on entry to DOS	1026H
	DW	offset _InDOS	; indos flag			1028H
	DW	offset _machine_id			;	102AH
	DW	offset _windows_patch			;	102CH
	DW	offset _umb_arena			;	102EH

	;---------------------------------------------------
	;
	;  NOTE: Despite DOS calls that serve the purpose,
	;  some programs STUPIDLY look for patterns of bytes
	;  within DOS to find the internal data structures.
	;  We have to fake out these programs with static
	;  data.  Do not delete, move, arrange or otherwise
	;  modify the bytes below.
	;
	;---------------------------------------------------

	ASSUME	CS:NOTHING			; KLUDGE for METRO and
	TEST	BYTE PTR SS:[_CritErr],0FFH	; SIDEKICK to find CritErr
	DB	75H,0CH
	PUSH	SS:[304H]
	INT	28H
	ASSUME	CS:_TEXT

	DW	3E80H			; KLUDGE for South Mountain Software
	DW	GOFFSET _CritErr	; to find InDOS flag
	DB	3 dup (0)
	DB	0BCH
	DW	GOFFSET _disk_stack

	;-----------------------------------------------------
	; used once on init to "fixup" any data pointers that
	; must point to DGROUP.
_data_ptr_array	label	word
	DW	GOFFSET dbcs_ptr+2
	DW	GOFFSET _country_info+1BH
	DW	0

	;--------------------------------------------------------
	; Microsoft LanManager needs these variables in
	; EXACTLY the proper place to run.  Otherwise it crashes
	; and dies.  The software shouldn't need to access these
	; directly as ALL are available through DOCUMENTED calls,
	; but you know those great programmers at Microsoft...

	org	128BH
dbcs_ptr	label word
	DW	offset _lead_byte_start	; 128BH  ptr to lead byte table
	DW	?			; segment
_country_info	label	word		; 128FH  struct COUNTRY_INFO
	DB	01H,26H,00H,01H,00H,0B5H,01H,00H
	DB	00H,24H,00H,00H,00H,00H,2CH,00H
	DB	2EH,00H,2DH,00H,3AH,00H,00H,02H
	DB	00H
	DW	offset _casemap
	DW	?
	DB	2CH
	DB	11 dup (?)

	; This is an RD specific scratch sector buffer used by the kernel
	; and the floppy/hard disk device driver.
	;
	; WARNING:	If any SECTOR SIZE is greater than the size of this
	;		array then you will crash (see MAXSECSIZE in DEVFDHD.INC).
	;
_DMA_buff      	label byte
	db	1024 dup (?)

	; make all of the SDA public
	public	prnport,prninit
	public	newByte,comport,retries,comtemp,cominit,byteQueued
	public	con_que,con_byte
	public	daylight,rollover
	public	_dos_psp,_root_psp,__xSS,__xSP,__xCS
	public	__xIP,__xDS,__xIsEXE,_boot_prog,_boot_parms
	public	_boot_env,_boot_env_size,_temp_cds,_booting,_lo_buffer
	public	_int2f_ax, _int2f_bx, _int2f_cx, _int2f_dx, _int2f_di
	public	_shell, _cfg_size, _cfg_position, _cfg_seg
	public	_cfg_linenum, _cfg_inc_pos, _cfg_inc_line
	public	_cfg_inc_level, _cfg_ask, _cfg_ask_line
	public	_cfg_default, _cfg_num_menus, _cfg_menuptr
	public	_cfg_menupos, _cfg_nobypass
	public	_patch_uppercase, _patch_6506
	public	dbcs_ptr, _country_info, _data_ptr_array
	public	_uppercase_table, _collating_table, _term_table
	public	_lead_byte_table, _casemap
	PUBLIC	_switch_char, _machine_name
	PUBLIC	_CritErr,_InDOS,_crit_drive,_err_locus,_err_extended
	PUBLIC	_error_no,_err_action,_err_class,_err_esdi,_err_psp
	PUBLIC	_DTA,_current_psp,_int28_sp,_return_code
	PUBLIC	_CUR_DRIVE,_BREAK_FLAG,_int21_ax
	PUBLIC	_int21_al,_int21_ah
	PUBLIC	_err_already_set
	PUBLIC	_sda_cur_buf
	PUBLIC  _stacksNum, _stacksSize
	PUBLIC	_pDiskStack, _pCharStack, _pDiskStackTop, _pCharStackTop
	PUBLIC	_minute,_hour,_hund,_seconds
	PUBLIC	_day,_month,_year,_day_count,_day_of_week
	PUBLIC	_mem_fit,_verify_flag,_int28_OK,_prn_as_well
	PUBLIC	_sda_reqhdr,_sda_devptr,_crit_stack,_CritIgnoreOK,_crit_locus
	PUBLIC	_hitbrk,_special_ret,_sda_rdwrbuff
	PUBLIC	sda_dispatch, _sda_no24
	PUBLIC	_sda_1st_blank
	PUBLIC	_timecheck, _sda_5a_strlen
	PUBLIC	_machine_id, _paras_total, _sda_ds, _sda_bx
	PUBLIC	_int23_sp, _user_ds
	PUBLIC	_old_reqhdr,_lock_buff
	PUBLIC	_clock_txd,_use_stdout,_sda_fname1,_sda_fname2
	PUBLIC	_sda_ifs_cds
	PUBLIC	_sda_sdb,_sda_ffb,_sda_ename1,_sda_ename2
	PUBLIC	_sda_retry,_sda_fcbnext,_sda_numfcbs
	PUBLIC	_sda_numsfts,_sda_sft_h,_sda_sft_h2,unknown5
	PUBLIC	_sda_srchattr,_sda_openmode
	PUBLIC	unknown5a,_sda_splice,_sda_invoke,unknown5b,_in_ifs_crit_err
	PUBLIC	unknown5c,_SuperID,_SuperOrder,_SuperBoot, _SuperBootDrive
	PUBLIC	_sda_extfcb,_sda_fcbattr,_sda_chardev
	PUBLIC	_sda_isnet, _sda_isdevpath, _sda_drive,_sda_callflags
	PUBLIC	_sda_handle,_sda_rdwrlen,_sda_rdwrfcb,_sda_buffptr
	PUBLIC	_sda_dpbptr,_userregs,unknown6
	PUBLIC	_trydoshi,_using_boot_stack
	PUBLIC	_sda_sftptr,_sda_cdsptr,_sda_fcbptr,unknown7
	PUBLIC	_fname1ptr,_fname2ptr
	PUBLIC	_sda_rdwroff
	PUBLIC	_oldhandle,_olduserregs,_old_sfth
	PUBLIC	_sda_fcb,unknown8
	PUBLIC	_sopen_action,_sopen_act,_sopen_attr,_sopen_mode
	PUBLIC  _def_volume, _def_fatid, unknown9
	PUBLIC	_ren_sdb,_ren_ffb
	PUBLIC	_disk_stackbase,_disk_stack,_char_stackbase,_char_stack
	PUBLIC	_dos_win_ver
	PUBLIC	_sda_in_5D
	PUBLIC	_end_sda
	public	_DMA_buff
   public   _FAT_type

cEndData


COMMENT ~
----------------------------------------------------------------

	SECTION 3.  Startup Code

	This code must bring ROM-DOS to a functioning
	state.  It involves copying code and data around,
	and initializing a C like environment.

---------------------------------------------------------------- ~
cBegCode

	c_extrn	main,near
	c_extrn	init_dev_vars,near
	c_extrn	fixup_int_handlers,near
	c_extrn	getmemtop,near
	c_extrn	GoAskRPL,near
	c_extrn	get_codeparas,near
	c_extrn	move_dos,near
	c_extrn	getDEVDATA,near
	c_extrn	is_DOS_in_ROM,near
	c_extrn	decompress,near
	c_extrn	mem_combine,near
	c_extrn	realloc_env,near
	c_extrn	configsys,near
	c_extrn	set_bootdev,near
	extrn	BIOS_PUTC:near
	extrn	MEMMOVE:near

	c_public	_astart, _acrtused

	; _astart MUST be at _TEXT:0 for LOADER to work
COMMENT ~ ;prototype;
void _astart(void)
{}
~
__astart:
__acrtused:	; a label for Microsoft C

	; enable interrupts in case BIOS left them off
	sti

	; Display to the BIOS the fact that we have control
	; REMEMBER: we are still on the BIOS stack
	mov	AL,'0'
	push	AX
	call	BIOS_PUTC	; PASCAL (not C) style function

	; set boot device if user asked us to...
	call	set_bootdev

	; Set up a temporary STACK, above devdata and dosdata
	;
	; REMEMBER: we are still on the BIOS stack up to this point
	;
	; The previous code had another problem where it was attempting
	; to set the stack below devdata with the segment relative to
	; zero.  If the segment for devdata was greater than or equal to
	; 1000h the previous code block would fail.  This problem was
	; resolved by making the temp stack SS relative to the segment
	; of devdata.
	;
	; ETR 3723 BT 4-12-01
	;
	; The previous code block below used to set a temporary stack
	; just below devdata.  In BUILD.EXE, the user can enter the
	; bottom of DOS memory (default is 70:0) where it should be
	; safe to assume that nothing is used below this memory address.
	; This code was changed so that the stack is now above devdata
	; and dosdata plus the size of the temp stack.
	;
	mov	dx, offset _DEVDATA:beg_heap
	add	dx, TEMP_STACK_BYTES
	call	getDEVDATA

	;  BT 12-13-01 ETR 3890
	;
	;  When ROM -DOS is in ROM and is copied to RAM, it will end
	;  up here with the CS = DEVDATA.  The following code will
	;  adjust the stack location for this case.
	;
	mov	bx, cs
	cmp	ax, bx
	jne	DoTempStack
	mov	cl, 4
	mov	ax, offset _TEXT:beg_lfn_text + 15
	shr	ax, CL
	add	bx, ax
	mov	ax, offset _LFN_TEXT:beg_bss+15
	shr	ax, CL
	add	ax, bx

    DoTempStack:
	cli
	mov	SS, ax
	mov	SP, dx
	sti

	; Determine real start of DATA (segment location of DEVDATA)
	; This MUST be done before CODE is moved
	;
	mov	AX, offset _TEXT:beg_lfn_text
	mov	CL, 4
	shr	AX, CL
	push	ax
	mov	AX, offset _LFN_TEXT:beg_devdata
	shr	AX, CL
	pop	cx	; lfn code size in para
	add	ax, cx
	mov	CX, CS
	add	AX, CX
	mov	DS, AX

	; move ROM-DOS code to top of memory (only if in RAM)
	; REMEMBER: we are still on the temporary stack
	;
	call	is_DOS_in_ROM	; if CS is above top of memory, it is
	or	AX, AX
	jnz	in_rom

	call	getmemtop	; get top of mem seg from BIOS int 12h call
	push	ax
	call	GoAskRPL	; if RPL in RAM, adjust top seg value down
	pop	dx
	push	AX		; AX is end of usable RAM (typically A000h)
	call	get_codeparas	; how big is RD CODE, in paras? (ret in AX)
	pop	DX		; DX contains top of memory
	sub	DX, AX
	push	DX		; push "move to" segment
	call	move_dos
	pop	AX

	; Copy initialized data from ROM to RAM
	; DS already contains the current DEVDATA segment location
	;
	; REMEMBER: we are still on the temporary stack
	;
in_rom:	call	getDEVDATA
	mov	ES, AX
	mov	CX, offset _DEVDATA:beg_bss
	xor	SI, SI
	xor	DI, DI
	cld
	rep	MOVSB		; copy it to RAM
	mov	word ptr ES:[0], CS	; store ROM address at 70:0

	; Decompress initialized data "DGROUP" if needed
	;
	; REMEMBER: we are still on the temporary stack
	;
	call	getDGROUP
in_ram:	call	decompress

	; Determine the segment of the lfn code
	;
	mov	AX, offset _TEXT:beg_lfn_text
	mov	CL, 4
	shr	AX, CL
	mov	bx, cs
	mov	_uFarDOSCallSegment, bx ; save the dos segment
	add	bx, ax
	mov	_uFarLFNCallSegment, bx

	; set up the REAL stack
	mov	AX, DS
	cli
	mov	SS, AX
	mov	SP, GOFFSET _char_stack
	sti
	mov	byte ptr _using_boot_stack,1

	; zero uninitialized data (class BSS & STACK)
	mov	CX, GOFFSET beg_heap
	mov	DI, GOFFSET beg_bss
	sub	CX, DI
	xor	AL,AL
	cld
	rep	stosb

	; add some overflow protection, chkstk function uses this!!
STACKWORD	EQU	5678H
L1:	mov	word ptr _disk_stackbase, STACKWORD
	mov	word ptr _char_stackbase, STACKWORD

	; set up some stuff for internal devices
	call	getmemtop
	mov	[_mem_top], AX		; this changes to fool devices
	mov	[_paras_total], AX	; so we need this variable too
	call	fixup_int_handlers
	call	init_dev_vars

	call	main		; never to return...

COMMENT ~
----------------------------------------------------------

   The following routines are provided so that ROM-DOS
   can get by without ANY fixups.  This allows the ROM
   code to be totally relocatable.

---------------------------------------------------------- ~

; WARNING! This function changes DS.  For assembly use only!!
; (hence, no leading underscore like all C functions have)
;
	PUBLIC getDGROUP
getDGROUP proc near
	push	AX
	call	getDS		; get DGROUP into DS
	mov	DS,AX		; affects no other registers
	pop	AX
	ret
getDGROUP endp

COMMENT ~ ;prototype;
unsigned getCS(void)
{}
~
c_func	getCS
	mov	AX, CS
	ret
c_endp	getCS

comment ~ ;prototype;
/*
   Return DGROUP.
*/
unsigned getDS(void)
{}
~
c_func	getDS
	push	CX
	mov	AX, offset _DEVDATA:beg_data
	mov	CL, 4
	shr	AX, CL
	xchg	AX, CX
	call	getDEVDATA
	add	AX, CX
	pop	CX
	ret
c_endp	getDS


;
;   Returns the segment of the start of the heap.
;   This is used by the memory management of DOS.
;
COMMENT ~ ;prototype;
unsigned getHEAP(void)
{}
~
c_func	getHEAP
	push	CX
	mov	AX, offset _DEVDATA:beg_heap
	mov	CL, 4
	shr	AX, CL
	xchg	AX, CX
	call	getDEVDATA
	add	AX, CX
	pop	CX
	ret
c_endp	getHEAP

cEndCode

cBegData
	extrn	_warn_shell:byte	; warning messages
	extrn	_warn_bad:byte
	extrn	_warn_again:byte
	extrn	_dos5_flags:word

	extrn	_uFarLFNCallSegment:word
	extrn	_uFarDOSCallSegment:word

	newline_str	db	10,0	; "\n"
cEndData

cBegCode
	extrn	_printf:near

;--------------------------------------------------------
;   This function is used both upon bootup AND if the
;   root program terminates.  If the root program ever
;   quits, this routine will allow the user to type in
;   a new root program.
;
COMMENT ;
void finish_boot(void)
{}
;
c_func	finish_boot

	; make sure that the UMB link is off.
	; The ESHELL part of Boundless' EMUL program demands it.
	;
	mov	ax, 5803h
	xor	bx, bx
	int	21h

	; throw away all initialization code (to save RAM space)
	call	toss_init
	call	mem_combine
	call	realloc_env

boot_again:
	; print a line
	mov	AX, GOFFSET newline_str
	push	AX
	call	_printf
	pop	AX

	; make the "user input" string appropriate for EXEC
	call	setup_for_exec

	; point to name and parms
	mov	DX, [_boot_prog]
	push	DS			; DS:DX = name
	pop	ES
	mov	BX, GOFFSET _boot_parms	; ES:BX = parms

	; allow use of both int 21h stacks
	;
	mov	byte ptr [_using_boot_stack], 0

	; don't allow critical errors for this spawn
	mov	[_sda_no24], 1

	; actually do the spawn

	sti
	mov	AX, 4B00H
	int	21H

	; at this point, assume everything, including stack is trashed
	call	getDGROUP	; get our data group
	mov	byte ptr [_using_boot_stack], 1	; again, 2nd stack for user stuff

	; we failed to SHELL. let user know
	; (root prog can never terminate).
   push  DS
	push	word ptr [_boot_prog]
	mov	AX, GOFFSET _warn_shell
	push	AX
	mov	AX, GOFFSET _warn_bad
	push	AX
   ;
   ;    printf(warn_bad,warn_tsr,DS:boot_prog);
   ;
	call	_printf
	add	sp, 8

	; provide DOS "A>" prompt
	mov	AH, 19H
	int	21H			; get current drive
	add	AL, 'A'
	mov	byte ptr [_warn_again], AL
	mov	AX, GOFFSET _warn_again
	push	AX
	call	_printf
	pop	AX

	; place CR at end of line so F3 works like it should
	call	setup_for_string

	; get a new string and try it again
	mov	DX, GOFFSET _shell
	mov	AX, 0C0AH
	int	21H

	; remember we cannot return from this function
	jmp	boot_again

c_endp	finish_boot

;-----------------------------------------------------
;  Given a string in the form of "command.com /p",13
;  make it palatable for EXEC.
;
;  Sets up "boot_prog" and "boot_parms".
;
	public	setup_for_exec
setup_for_exec proc near
	push	SI
	push	DI

	; set up boot program name
	mov	DI, GOFFSET _shell+2
	mov	[_boot_prog], DI

	; null terminate the "program name"
sfe_0:	cmp	byte ptr [DI], ' '
	jb	sfe_1			; no parms
	je	sfe_2
	inc	DI
	jmp	sfe_0
sfe_1:	mov	byte ptr [DI+1], 0	; no "parms", provide a terminator
sfe_2:	mov	byte ptr [DI], 0	; asciiz the "progname"

	; set up arguments
	inc	DI			; DI points to args
	mov	BX, DI
	xor	SI, SI
sfe_5:	cmp	byte ptr [BX+SI], ' '	; SI is length, BX is base
	jb	sfe_6
	inc	SI
	jmp	short sfe_5

	; memmove(void *dst, void *src, unsigned len)
sfe_6:
	inc	BX
	push	BX			; dst
	dec	BX
	push	BX			; src
	push	SI			; length
	call	MEMMOVE			; shift it over 1 byte so

	mov	BX, DI			; save ptr to args
	mov	AX, SI
	mov	byte ptr [DI], AL	; put length byte in 1st position
	inc	DI			; skip length byte
	add	DI, SI			; add length
	mov	word ptr [DI], 13	; end w/ CR followed by NULL

	; boot_parms.e.command_line = (char far *)boot_args;
	mov	word ptr [_boot_parms+2], BX
	mov	word ptr [_boot_parms+2+2], DS

	;  ETR <none>  BT 3-28-01
	;
	;  drive number (1=A,2=B) must be in the first byte
	;  of FCB1 in the psp used to exec command.  This is
	;  for MS command.com.  Trap INT 21h fn 19h and walk
	;  the next couple of lines watching the AL register
	;  to see.  Then trap the next int 21h fn 3dh open
	;  where the drive number becomes the drive letter of
	;  the path for autoexec.bat.  8^)
	;
	;  ds:069h (SDA) bootdrive
	;  ds:336h (SDA) _CUR_DRIVE
	;
	mov	word ptr [_boot_parms+6], 69h
	mov	word ptr [_boot_parms+8], ds

	pop	DI
	pop	SI
	ret
setup_for_exec endp

setup_for_string	proc near

	; replace first NULL with ' '
	mov	BX, GOFFSET _shell + 2
sfs_0:	cmp	byte ptr [BX], 0
	jz	sfs_1
	inc	BX
	jmp	short sfs_0
sfs_1:	mov	byte ptr [BX], ' '

	; move arguments back to where they were
	inc	BX
	xor	CX, CX
	mov	CL, byte ptr [BX]
	inc	CX
	push	BX		; dst (args)
	inc	BX
	push	BX		; src (args+1)
	push	CX		; length
	call	MEMMOVE

	ret
setup_for_string	endp

;--------------------------------------------
;  Throw away initialization code to save space in RAM
	public	toss_init
toss_init proc near

	; is DOS low (not in ROM or HMA)?
	test	[_dos5_flags], 1800H
	jnz	no_toss
IF 0
	;  Enable this jump, which will not drop the init code, so that
	;  you can trace through the LFN module code with softice.  Otherwise,
	;  the symbols will not line up correctly because the LFN module is
	;  linked in after the init code.
	;
	jmp	no_toss
ENDIF

	; Save the registers
	;

	push	ds
	push	es
	push	si
	push	di

	; The lfn code follows the init code.  So, if we
	; want to toss the init code we must move the LFN
	; code over the top of the init code and then
	; resize the memory
	;

	; Determine the new segment address for the LFN code
	;
	mov	BX, offset _TEXT:_main + 15
	mov	CL, 4
	shr	BX, CL ; save the (PARA) size of the near romdos code (BX)

	;
	; The LFN code follows the init code.  We need to move the
	; LFN code over the init code.
	;

	; Set the destination of the move
	;
	mov	cx, cs
	add	cx, bx
	mov	es, cx
	xor	di, di

	; Get the LFN segment and save the new address of the LFN segment
	; and save the segment where dos is at.
	;
	call	getDGROUP
	mov	si, _uFarLFNCallSegment
	mov	_uFarLFNCallSegment, cx

	; Set the address where the LFN code is now
	;
	mov	ds, si
	xor	si, si

	; Determine the size of the LFN code
	;
	mov	ax, offset _LFN_TEXT:beg_devdata
	push	ax

	; Calculate the new size of the code in paragraphs
	;
	add	ax, 15
	mov	cl, 4
	shr	ax, cl
	add	bx, ax ; add the size of the lfn code to the romdos near code

	; Set the size of the move
	;
	pop	cx

	; Move the LFN code
	;
	rep	movsb

	; reallocate the memory if low
	;
	; bx = code size in paragraphs
	;

	call	getDGROUP
	mov	ES, [_dos_code_seg]
	mov	AH, 4AH
	int	21H

	; Restore the registers
	;

	pop	di
	pop	si
	pop	es
	pop	ds

no_toss:
	ret

toss_init endp

cEndCode

	END
