;
;   $Id: DOSCRIT.ASM 1.8 2002/10/15 21:46:18Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
COMMENT ~
----------------------------------------------------------

   DOSCRIT.ASM  created  05/21/90  by Roy Sherrill

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This is the Default Critical Error Handler (INT 24H).
   When ROM-DOS generates a critical error because of
   some failure of a device (disk or char), this handler
   gets control.  Unlike MS-DOS, this handler is available
   whether COMMAND.COM is installed or not.

   Upon Entry:
   AH    = error bit map
   AL    = drive (0=A, 1=B)
   DI    = error code
   BP:SI = pointer to device

   Upon Exit:
   AL    = action (0=ignore, 1=retry, 2=abort, 3=fail)

   AH (error bit map)
   ------------------
   1xxx xxxx	Error on device (0=disk)
   xx1x xxxx	Ignore allowed
   xxx1 xxxx	Retry allowed
   xxxx 1xxx	Fail allowed
   xxxx x11x	disk: 00=boot, 01=FAT, 10=DIR, 11=DATA area
   xxxx xxx1	Error during write (0=read)

   DI (error code)
   ------------------
   0	Write protect
   1	Unknown unit
   2	Not ready
   3	Unknown command
   4	CRC
   5	Bad struct
   6	Seek
   7	Unknown media
   8	Sector not found
   9	No paper
   A	Write fault
   B	Read fault
   C	General failure
   D	Sharing violation
   E	Lock violation
   F	Invalid disk change
  10	FCB Unavailable
  11	Sharing buffer overflow
  12	Code page mismatch
  13   	Out of input
  14    Insufficient disk space
  20	Sharing violation (Same as 0Dh but from MS Networks)

----------------------------------------------------------- ~

include profile.inc

MAX_ERR_CODE	=	14h	; Invalid disk change

include macros.inc
include devhdr.inc		; device error code constants

; IFSNAME	equ	1


Fail_allowed	equ	08H
Retry_allowed	equ	10H
Ignore_allowed	equ	20H

cBegData
	c_extrn	crit_bad_msg,byte
	c_extrn	crit_msgs,byte
	c_extrn	crit_error,byte
	c_extrn	crit_read,byte
	c_extrn	crit_write,byte
	c_extrn	crit_drv_msg,byte
	c_extrn	crit_dev_msg,byte
	c_extrn	crit_quest_mark,byte
	c_extrn	crit_cr_lf,byte
	c_extrn	crit_abort,byte
	c_extrn	crit_retry,byte
	c_extrn	crit_ignore,byte
	c_extrn	crit_fail,byte
	c_extrn	crit_volume,byte
	c_extrn     crit_serial,byte

 	extrn	_err_extended:word
	extrn	_err_esdi:dword

cEndData

cBegCode
	extrn	getDGROUP:near

;
;   This is the ROM-DOS Default Critical Error Handler.
;
COMMENT ; prototype
void	int_24(void)
{}
;
c_func	int_24

	ENTER_FUNCTION

	; save all the registers
	push	DI
	push	BP
	push	SI
	push	DS
	push	ES
	push	DX
	push	AX

	; DS should point to our strings
	mov	BP, SP
	call	getDGROUP

	; see if an IFS wants to take care of the error name
	;
	call	IFS_print_err_msg	; try to print IFS message
	jnc	L0_A			; CF clear if msg printed

	mov	DX, GOFFSET crit_cr_lf	; else print our message
	call	print_string

	; make sure DI is valid. if not, use standard error message
	cmp	di, 20h
	jne	not_MS_net_sharing_violation

	mov	di, 0Dh

   not_MS_net_sharing_violation:
	mov	BX, DI
	cmp	BL, MAX_ERR_CODE
	jbe	ok_err
	mov	DX, GOFFSET crit_bad_msg
	jmp	short L0

	; display type of error (DI) in message form
	;
ok_err:	xor	BH,BH
	push	bx
	shl	BX, 1
	add	BX, GOFFSET crit_msgs
	mov	DX, [BX]
	pop	bx
L0:	call	print_string	; print critical error message
L0_A:

	; print "error"
	mov	DX, GOFFSET crit_error
	cmp	bx, 7
	jb	SayError
	add	dx, 6
SayError:
	call	print_string

	; print "reading" or "writing"
	mov	DX, GOFFSET crit_read
	test 	byte ptr [BP+1], 1
	jz	L1
	mov	DX, GOFFSET crit_write
L1:	call	print_string

	; print "drive X:" or "device NNN"
	test	byte ptr [BP+1], 80H
	jnz	chardev
	mov	DX, GOFFSET crit_drv_msg
	call	print_string
	mov	AL, [BP]
	add	AL, 'A'
	call	print_char
	jmp	nextline

	; print "device NNN"
chardev:
	mov	DX, GOFFSET crit_dev_msg
	call	print_string

	; print all 8 characters of device name
	mov	CX, 8
	mov	ES, [BP+10]	; ES:SI now points to device header
	add	SI, 10		; seek to name part of device header
L3:	mov	AL, ES:[SI]
	call	print_char
	inc	SI
	loop	L3

	; print CR/LF, next line
nextline:
 	mov	DX, GOFFSET crit_cr_lf
	call	print_string

	; If this is error code 0Fh (Invalid Disk Change) then we have
	; to display more info.
	;
	cmp	di, ERR_INVALIDDISKCHANGE
	jne	again

	; Print "Please insert volume "
	;
	mov	dx, GOFFSET crit_volume
	call 	print_string

	push	ds
	pop	es		; es->DGROUP
	lds	si, _err_esdi	; ds:si-> ddt_volume
	mov	cx, 11		; chars in volume label
	cld

	; print volume name
	;
    volume_name_loop:
	lodsb
	call	print_char
	loop	volume_name_loop
	inc	si		; skip NUL char
	push	ds		; ds:si-> serial number
	push	es
	pop	ds		; ds->DGROUP

	; print " serial "
	;
	mov 	dx, GOFFSET crit_serial
	call 	print_string

	; print serial number
	;
	pop	ds		; ds-> device data
	call	print_serial_number
	push	es
	pop	ds              ; ds->DGROUP

	; new line
	;
 	mov	DX, GOFFSET crit_cr_lf
	call	print_string

	; get user response (Abort, Retry....)
	; print "abort"
again:	mov	DX, GOFFSET crit_abort		; abort is always allowed
	call	print_string

	; print ", retry"
	test	byte ptr [BP+1], Retry_allowed	; see if retry allowed
	jz	no_retry
	mov	DX, GOFFSET crit_retry
	call	print_string
no_retry:

	; print ", ignore"
	test	byte ptr [BP+1], Ignore_allowed	; see if ignore allowed
	jz	no_ignore
	mov	DX, GOFFSET crit_ignore
	call	print_string
no_ignore:

	; print ", fail"
	test	byte ptr [BP+1], Fail_allowed	; see if fail allowed
	jz	no_fail
	mov	DX, GOFFSET crit_fail
	call	print_string
no_fail:
	; print "?"
	mov	DX, offset crit_quest_mark
	call	print_string

	; Get input (1 character)
	mov	ax, 0C01h		; purge buffer first
	int	21H

	; convert response to uppercase
	cmp	AL,'a'
	jl	upper
	cmp	AL,'z'
	jg	upper
	add	AL,0E0H
upper:	push	AX

	; next line
	mov	DX, offset crit_cr_lf
	call	print_string

	pop	BX
	xor	AX, AX

	; Check for abort
	cmp	BL, 'A'
	jz	do_abort

	; Check for Retry
	test 	byte ptr [BP+1], Retry_allowed
	jz	no_retry2
	cmp	BL, 'R'
	jz	do_retry
no_retry2:

	; Check for Ignore
	test	byte ptr [BP+1], Ignore_allowed
	jz	no_ignore2
	cmp	BL, 'I'
	jz	do_ignore
no_ignore2:

	; Check for Fail
	test	byte ptr [BP+1], Fail_allowed
	jz	no_fail2
	cmp	BL, 'F'
	jz	do_fail
no_fail2:

	; user didn't pick a valid response, try again
	jmp	again

	; set AX to proper response, 0=ignore, 1=retry, 2=abort, 3=fail
do_fail:	inc	AX
do_abort:	inc	AX
do_retry:	inc	AX
do_ignore:

	; restore all registers and get out
	pop	DX			; clean old AX off stack
	pop	DX
	pop	ES
	pop	DS
	pop	SI
	pop	BP
	pop	DI

	EXIT_FUNCTION

	iret
c_endp	int_24


; Trys to get a descriptive string from the IFS for the current extended
; error code and print it on STDOUT.
;
; on entry: The SDA _err_extended word is set to the extended error code
;	(which should have already been taken care of by the IFS before
;	 it initiated the int 2Fh, fn 1206h call.)
;
; returns CF on error (no IFS installed or IFS unable to translate err code)
; else CF clear and error IFS error description printed on STDOUT.
;
; uses es, dx, bx, ax
;
c_func IFS_print_err_msg

	ENTER_FUNCTION

	push	ds
	push	di			; save error code

	; determine if an IFS has a handler installed for crit err messages
	;
	mov	ax, 0500h		; polling function
	int	2Fh			; see if an IFS installed
	inc	al			; will return 0FFh in AL if so
	stc				; set CF in case of error
	jnz	IFSPEM_exit		; exit if no IFS msg handler installed

	; have the IFS try to translate our extended error code into a message
	;
	mov	bx, [_err_extended]
	mov	ax, 0501h		; make the translation request
	int	2Fh
	jc	IFSPEM_exit		; return CF set if request failed

	; actually print the message on STDOUT
	;
	mov	ax, es
	mov	ds, ax
	mov	dx, di			; get str ptr from es:di into ds:dx

	call	print_string		; print it
	clc				; no error, so make sure CF is clear

   IFSPEM_exit:
	pop	di			; restore device error code
	pop	ds

	EXIT_FUNCTION

	ret

c_endp IFS_print_err_msg


; Print an NUL terminated string to DOS
;
; on entry:
;	ds:dx	-> ASCIIZ string to print
;
; returns nothing
; uses dx, ax
;
c_func print_string

	ENTER_FUNCTION

	push	si		; save caller's si
	mov	si, dx		; get message ptr offset in si

	lodsb			; get first char
	or	al, al		; if NUL then
	jz	PAZS_done	; done. nothing to print.

    PAZS_loop:
	call	print_char	; print the char on STDOUT
	lodsb			; get the next char
	or	al, al		; if NUL
	jnz    	PAZS_loop	; then done, else print next char

    PAZS_done:
    	pop	si		; restore caller's si and return

	EXIT_FUNCTION

    	retn

c_endp print_string


;  Print a character to DOS
;
;  On Entry character must be in AL
c_func	print_char

	ENTER_FUNCTION

	mov	AH, 2
	mov	DL, AL
	int	21H

	EXIT_FUNCTION

	ret

c_endp	print_char


; Converts a binary number in al into a ASCII hex digit in al
;
; ->	al	binary number
; <-	al	ASCII hex digit
; uses:	ax
;
c_func	bin_to_hex

	ENTER_FUNCTION

	mov	ah, '0'		; '0' + {0,1,...9} = ASCII digit
	cmp	al, 9
	jbe	convert
	mov	ah, '7'		; '7' + {10,11,..15}  = {'A','B',...'F'}

    convert:
	add	al, ah

	EXIT_FUNCTION

	ret

c_endp	bin_to_hex


; Prints a disk serial number (1234ABCD) as "CDAB-3421"
;
; ->	ds:si	points to the disk serial number
; <-	void
; uses:	ax,bx,cx
;
c_func  print_serial_number

	ENTER_FUNCTION

	push	si

	mov	bx, 4
	mov	cx, bx
	cld

	; get "back-words" bytes in English reading order on stack
	;
    stack_chars:
	lodsb
	mov	ah, al
	shr	ah, cl
	and	ax, 0F0Fh
	push	ax
	mov	al, ah
	push	ax
	dec	bx
	jnz	stack_chars

	; print high word
	;
    print_high:
	pop	ax
	call	bin_to_hex
	call	print_char
	loop	print_high

	; print '-'
	;
	mov	al, '-'
	call	print_char
	mov	cx, 4

	; print low word
	;
    print_low:
	pop	ax
	call	bin_to_hex
	call	print_char
	loop	print_low

	pop	si

	EXIT_FUNCTION

	ret

c_endp	print_serial_number

cEndCode

	END
