//
//   $Id: DOSDEV.C 1.14 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DOSDEV.C  created 12/01/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains all direct device operation.  This
	includes calling the devices, critical errors, console
	I/O and ctrl-C checking.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 4


/* reading or writing the console uses this buffer */
extern uchar con_buf[130];
extern uchar con_bufpos;
extern uchar con_col;	/* Console column.  use for tabs, esc, etc... */

void dev_openclose(bool open)
{
	extern struct dev_req_hdr sda_reqhdr;
	extern LPDEV_HDR sda_devptr;

	ENTER_FUNCTION();

	/* if the device supports it, do an IOCTL open */
	if(sda_devptr->attr & DEVA_OPNCLS)
	{
		sda_reqhdr.command = open ? DEV_OPEN : DEV_CLOSE;
		sda_reqhdr.unit = 0;
		dev_crit_call(&sda_reqhdr);
	}

	EXIT_FUNCTION();
}

/*
	Read a single byte.  This byte may come from a FAT or
	IFS file, or a device.
*/
uchar	read1byte(void)
{
	extern uchar sda_chardev;
	extern uchar powersave;
	extern uchar hitbrk;

	uchar c;

	ENTER_FUNCTION();

	c = ctrl('C');			// Initialize value in case ^Break is hit

	if(sda_chardev)
	{
		if((sda_chardev & DEVA_CON) && !powersave)
			console_wait_for_key();

		if(hitbrk == 0)
			dev_read((void far *)&c, 1);

		/* do int 28H at least once per key */
		if(powersave)
			do_int28();
	}

	else if(!internal_read((void far *)&c, 1))
		c = 0;

	EXIT_FUNCTION();

	return c;
}

void do_int28(void)
{
	extern uchar InDOS;
	extern uchar CritErr;
	extern uchar timecheck;
	extern uchar clk_use_rtc;
	extern unsigned sda_handle;
	extern char in_windows;
	extern char int28_OK;

	int handle;

	ENTER_FUNCTION();

	// whether generating windows crit section signals or int 28s
	// DOS must still be available
	//
	if(InDOS==1 && CritErr==0)
	{
		handle = sda_handle;

		if(in_windows && R_AH > 0xC)
		{
			// wait a while for windows to do its thing
			//
			// We need this to make Windows happy but MS-NET for DOS
			// patches the critical section flags to 50h when it starts
			// up. Our int 2Fh fn.1607h filter sets in_windows to (1)
			// we can test for that value explicitly and not face the
			// hit from this NOP loop under MS-NET while still running
			// windoze with this "fix".
			//
			if(in_windows == 1)
			{ unsigned wait_a_bit, x;

				for(wait_a_bit=0; wait_a_bit<0xF000; ++wait_a_bit)
					x = wait_a_bit + 1;
				wait_a_bit = x;
			}
		}

		// if fn call #1..0Ch
		if(int28_OK)
		{
			generate_int28();
		}

		get_call_flags();
		handle_to_sda(handle);

		/*
			check for midnight rollover every 256 keys.  note:
			also allows redirectors to get date/time directly
			from DOS data area.
		*/
		if(timecheck == 0)
			touch_time(TM_READ);	/* read current time */
		++timecheck;
	}

	EXIT_FUNCTION();
}

/*
	Loop around waiting for a character to be pressed.
	Until that happens, generate INT 28s (DOS idle) so
	that a TSR program can DO something.
*/
void	console_wait_for_key(void)
{
	extern uchar hitbrk;

	int c = -1;

	ENTER_FUNCTION();

	while(!hitbrk && c < 0)
	{
		c = dev_ndread();
		if(c < 0)
			do_int28();
	}

	EXIT_FUNCTION();
}


/*
	Read 1 byte.  Check for ctrl-c, ctrl-p, and ctrl-s.
*/
uchar	readchk(void)
{
	uchar c;

	ENTER_FUNCTION();

	do
	{
		// get the character
		c = read1byte();

		// do NOT check the 2nd byte on Alt-chars (ie Alt-Q)
		if(c)
			ctrlc_check(c);

		/* we must get a different character */
	} while(c == ctrl('S') || c == ctrl('P'));

	EXIT_FUNCTION();

	return c;
}


/*
	Write a single byte.  This byte may be going to a
	file, or a device.  If the device is the console,
	this function must keep track of the console column.
*/
void	PASCAL write1byte(uchar c)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar prn_as_well;
	extern uchar CritErr;
	extern uchar sda_chardev;
	extern uchar con_col;
	extern uchar far * sda_rdwrbuff;
	extern LPDEV_HDR sda_devptr;
	extern unsigned sda_handle;
	extern unsigned sda_callflags;
	extern unsigned doscallflag[];
	extern unsigned sda_rdwrlen;


	/*
		[ locals ]
	*/
	int handle;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* write the character out quickly */
	if(sda_chardev & DEVA_FAST)
		generate_int29(c);

	/* otherwise write to file/device */
	else
	{
		/* let subordinate routines know what kind of call this is */
		sda_callflags = callf_flagsof(0x40);

		/* call the approriate IFS, FAT or device */
		sda_rdwrbuff = (void far *)&c;
		sda_rdwrlen = 1;
		if(!ifs_call())
		{
			if(sda_chardev)
				dev_write((void far *)&c, 1);
			else
				file_write();
		}
		error_no = 0;
	}

	/* if this is the console, keep track of the column */
	if((sda_chardev & DEVA_CON) || sda_handle <= STDERR)
	{
		/* echo to stdprn if requested */
		if(prn_as_well)
		{
			if(CritErr)
				prn_as_well = 0;
			else
			{
				/* switch to context of printer */
				handle = sda_handle;
				if(handle_to_sda(STDPRN))
					write1byte(c);

				/* switch back to original context */
				handle_to_sda(handle);
			}
		}

		if(c == CR)
			con_col = 0;
		else if(c == BS)
		{
			if(con_col)
				--con_col;
		}
		else if(c != LF)
			++con_col;
		if(con_col > 80)
			con_col = 0;
	}

	EXIT_FUNCTION();
}


/*
	Write with tabs.  This will write a single character
	to a FAT or IFS file, or a device.  If the device
	happens to be the console, it will expand tabs into
	many spaces.
*/
void	PASCAL write_w_tabs(uchar c)
{
	extern uchar sda_chardev;
	extern uchar con_col;
	extern uchar CritErr;
	extern LPDEV_HDR sda_devptr;
	extern unsigned sda_handle;
	extern unsigned sda_callflags;
	extern int use_stdout;

	unsigned oldflags, oldhandle;

	ENTER_FUNCTION();

	/* use STDOUT instead of the current SDA values */
	if(use_stdout)
	{
		/* save old context */
		oldhandle = sda_handle;
		oldflags  = sda_callflags;

		sda_callflags = 0;
		handle_to_sda(CritErr ? STDERR : STDOUT);
	}

	/* if this is the console, and it's a TAB character, display many spaces */
	if(c == TAB && ((sda_chardev & DEVA_CON) || sda_handle<=STDERR))
	{
		do { write1byte(' '); } while(con_col % 8);
	}
	else
		write1byte(c);

	/* done, set SDA back to what it should be */
	if(use_stdout)
	{
		sda_callflags = oldflags;
		handle_to_sda(oldhandle);
	}

	EXIT_FUNCTION();
}


/*
	Check to see if a character is available.  This will
	work on a FAT or IFS file as well as a device.

	Returns -1 if no characters waiting.  Returns the
	character itself if available.
*/
int	ndread1byte(void)
{
	extern uchar sda_chardev;

	uchar c;

	ENTER_FUNCTION();

	/* character device go directly to the device */
	if(sda_chardev)
	{
		do_int28();
		EXIT_FUNCTION();
		return dev_ndread();
	}
	/* otherwise get the character (from FAT or IFS) */
	else
	{
		/* the read failed, no more characters in the file */
		if(!internal_read((void far *)&c, 1))
		{
			EXIT_FUNCTION();
			return -1;
		}
		internal_seek(-1L, SK_REL);
	}

	EXIT_FUNCTION();

	return (int)c;
}

/*
	Check for ^C only once every 65 characters.  This
	speeds up console I/O considerably.
*/
void	PASCAL con_ctrlc_check(void)
{
	static int charcount;

	ENTER_FUNCTION();

	if(charcount >= 65)
	{
		ctrlc_check(0);
		charcount = 0;
	}
	++charcount;

	EXIT_FUNCTION();
}

/*
	Check the console for any impending ctrl-c, ctrl-s or
	ctrl-p characters.
*/
void	PASCAL ctrlc_check(uchar givenc)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar prn_as_well;
	extern uchar hitbrk;
	extern unsigned sda_callflags;
	extern unsigned sda_handle;
	extern uchar sda_chardev;
	extern uchar no_ctrl_c;
	extern struct LOL lol;
	extern LPDEV_HDR sda_devptr;

	/*
		[ locals ]
	*/
	static char ctrlc[] = "^C\r\n";
	int i;
	int handle;
	uchar c;
	unsigned oldflags;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* save old context */
	handle = sda_handle;
	oldflags = sda_callflags;

	/* switch to the STDIN context */
	sda_callflags = 0;
	if(!handle_to_sda(STDIN))
		goto done;


	//  Only check CON for control break
	//
	if( lol.condev != sda_devptr )
	{
		goto done;
	}


	/* someone hit ctrl-break */
	if(hitbrk)
	{
		hitbrk = FALSE;
		givenc = ctrl('C');
	}

	/* don't read keys if getting from a file.  LapLink might be
	   transferring itself over CTTY via COM1.  A value 03H is legal
	   from a file. */
	else if(!sda_chardev)
		goto done;

//again:
	c = givenc;
	if(!c)
	{
		/* check if a ^C has happened */
		i = ndread1byte();

		/* no characters waiting, we are through */
		if(i < 0)
			goto done;

		c = i;
	}

	/* stop and wait for a keypress */
	if(c == ctrl('S'))
	{
		if(!givenc)
			read1byte();
		c = givenc = read1byte();
	}

	/* send output to printer as well */
	if(c == ctrl('P'))
	{
		if(!givenc)
			read1byte();
		prn_as_well = !prn_as_well;
	}

	/* we have a ^C, break out */
	if(c == ctrl('C'))
	{
		// there will always be 2 NULLs when doing a ctrl-break
		if(!givenc)
			read1byte();

		// don't generate ctrl-C until we have booted
		if(!no_ctrl_c)
		{
			if(handle_to_sda(STDOUT))
				dev_write((void far *)ctrlc, sizeof(ctrlc));
			generate_int23();
		}
	}

done:

	/* switch context back to where it came from */
	sda_callflags = oldflags;
	handle_to_sda(handle);
	error_no = FALSE;

	EXIT_FUNCTION();
}


/*
	Flush the current FAT or IFS file or device.  This is
	only really used by DOS call 0CH.
*/
void	flushbytes(void)
{
	extern uchar sda_chardev;

	ENTER_FUNCTION();

	if(!sda_chardev)
	{
		EXIT_FUNCTION();
		return;
	}

	while(ndread1byte() >= 0)
		read1byte();

	EXIT_FUNCTION();
}


/*
	-------------------------------------------------------

							CONSOLE INPUT

	-------------------------------------------------------
*/


#define con_buflen con_buf[1]

/*
	Used by any file read call that happens to be reading
	from the console.  Uses the string input function to get
	the bytes.

	Returns the length of the bytes read.
*/
unsigned	PASCAL dev_con_read(uchar far *buff, unsigned len)
{
	ENTER_FUNCTION();

	// validate parameters
	ASSERT(buff);
	ASSERT(len);

	/* we have no string, get one */
	if(con_bufpos >= con_buflen)
	{
		con_buf[0] = 128;
		string_input((uchar far *)con_buf);
		con_buflen += 2;
		con_buf[con_buflen + 1] = '\n';
		con_bufpos = 0;
	}

	/* move the substring into user buffer area */
	if(con_buflen-con_bufpos <= len)
	{
		len = con_buflen - con_bufpos;
		write1byte(LF);
	}
	fmemcpy(buff, (void far *)(con_buf + 2 + con_bufpos), len);
	con_bufpos += len;

	EXIT_FUNCTION();

	return len;
}


/*
	Used by any file write call that happens to be writing
	to the console.  Will stop writing bytes if a ctrl-z is
	encountered.

	Returns the # of bytes written.
*/
unsigned PASCAL dev_con_write(uchar far *buff, unsigned len)
{
	uchar c;
	unsigned bytes = 0;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(buff);

	while(bytes < len)
	{
		/* if the character is a ctrl-z, stop printing */
		c = *buff++;
		if(c == ctrl('Z'))
			break;

		/* write out the character */
		con_ctrlc_check();
		write_w_tabs(c);
		++bytes;
	}

	EXIT_FUNCTION();

	return bytes;
}



/*
	This function, used by DOS calls 0AH and 3FH (only when
	reading from the console) is the function that handles
	F1,F3,Ins,Del,etc...  This can input a string of up to
	127 characters.

	Supports:
		F1			get next key			RIGHT		get next key
		F3			fill out string		LEFT		del previous key
		F6			same as ^Z				INS		enter insert mode
		BS			del previous key		DEL		del char
*/
void	PASCAL string_input(uchar far *buffer)
{
	/*
		[ locals ]
	*/
	uchar c;
	uchar key;
	uchar maxlen;
	uchar conlen;
	uchar orgcol;				/* original column on the screen */
	uchar curpos;
	uchar i;
	bool	insmode = FALSE;	/* are we in insert mode? */
	static uchar tempstr[130];


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(buffer);

	/* if an ESC is pressed, it must go back to this column */
	orgcol = con_col;

	/* get the maximum length of input, including return */
	maxlen = *buffer++;
	if(maxlen==0)
	{
		EXIT_FUNCTION();
		return;
	}
	--maxlen;

	/* get the current length of the string, test if it's valid */
	conlen = *buffer++;
	conlen = min(maxlen+1,conlen);
	if( *(buffer+conlen) == CR )
		fmemcpy((void far *)tempstr, buffer, conlen);
	else
		conlen = 0;

	curpos = 0;
	while(1)
	{

		/* get a key, checking for ^P, ^S and ^C */
		c = readchk();

		/* some keys (like Del) generate scan codes */
		if( c == 0 )
		{
			// do NOT check for ^S, ^P, etc.. on 2nd character
			key = read1byte();	// was readchk() -- DG

			switch(key)
			{
				case LEFT:
					c = BS;
					break;

				case RIGHT:
				case F1:
					if(curpos < conlen)
						c = tempstr[curpos];
					break;

				case F3:
					while(curpos < conlen)
					{
						string_write_w_tabs(tempstr[curpos]);
						buffer[curpos] = tempstr[curpos];
						++curpos;
					}
					break;

				case F6:
					c = ctrl('Z');
					break;

				case INS:
					insmode = !insmode;
					break;

				case DEL:
					if(curpos < conlen)
					{
						memmove(tempstr+curpos, tempstr+curpos+1, conlen-curpos);
						--conlen;
					}
					break;
			}

			if(key != INS)
				insmode = FALSE;

			/* go get another key */
			if(!c)
				continue;
		}


		/* Handle normal characters */
		switch(c)
		{
			/* start over with string */
			case ESC:
				write_w_tabs('\\');
				write_w_tabs(CR);
				write_w_tabs(LF);
				for(i=0; i<orgcol; ++i)
					write_w_tabs(' ');
				con_col = orgcol;
				curpos = 0;
				break;

			/* backspaces blot out a character */
			case BS:
				if(curpos > 0)
				{
					--curpos;

					/* find out previous character */
					c = buffer[curpos];

					/* tabs can do 1-8 spaces */
					if(c == '\t')
						i = con_col - column_of(orgcol, buffer, curpos);

					/* ctrl characters do 2 spaces */
					else if(c < ' ')
						i = 2;

					/* normal characters take 1 space */
					else
						i = 1;

					while(i && con_col)
					{
						write_w_tabs(BS);
						write_w_tabs(' ');
						write_w_tabs(BS);
						--i;
					}
				}
				break;

			/* normal characters get displayed */
			default:

				/* too many characters will beep */
				if(curpos == maxlen)
				{
					write_w_tabs(BELL);
					break;
				}

				/* insert character into the internal console buffer */
				if(insmode)
				{
					if(conlen < curpos)
						conlen = curpos;
					memmove(tempstr+curpos+1, tempstr+curpos, (conlen-curpos)+1);
					tempstr[curpos] = c;
					if(conlen < maxlen)
						++conlen;
				}

				/* treat ctrl characters specially */
				string_write_w_tabs(c);
				buffer[curpos++] = c;
				break;

			/* we are done, record string */
			case CR:
				write_w_tabs(CR);
				buffer[curpos] = CR;
				*(buffer-1) = curpos;
				EXIT_FUNCTION();
				return;

			/* linefeeds do nothing */
			case LF:
				write_w_tabs(CR);
				write_w_tabs(LF);
				break;
		}

	} /* end of while loop */

	EXIT_FUNCTION();
}

void PASCAL string_write_w_tabs(uchar c)
{
	ENTER_FUNCTION();

	if(c < ' ' && c != ctrl('I'))
	{
		write_w_tabs('^');
		write_w_tabs(c + '@');
	}
	else
		write_w_tabs(c);

	EXIT_FUNCTION();
}

/*
	We know there will be no CR or LFs in the string.
	Returns the column at this string position.
*/
int column_of(int start_col, uchar far *string, int len)
{
	register int col = start_col;
	char c;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(string);

	while(len--)
	{
		++col;
		c = *string;
		if(c == '\t')
		{
			while(col%8)
				++col;
		}
		else if(c < ' ')
			++col;
		++string;
	}

	EXIT_FUNCTION();

	return col;
}


/*
	-------------------------------------------------------

							RAW DEVICE I/O

	-------------------------------------------------------
*/

/*
	Is a character waiting in this device?

	Returns the character or -1 if none.
*/
int	dev_ndread(void)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern struct dev_req_hdr sda_reqhdr;
	extern LPDEV_HDR sda_devptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set up the request header */
	sda_reqhdr.command = DEV_NDREAD;
	dev_crit_call(&sda_reqhdr);

	/* if busy flag set, no characters waiting */
	if(sda_reqhdr.status & DEVE_BUSY)
	{
		EXIT_FUNCTION();
		return -1;
	}

	EXIT_FUNCTION();

	/* return the character */
	return sda_reqhdr.media;
}


/*
	Write characters to a device.  This is simply a raw
	write.  No other processing is done.
*/
void	PASCAL dev_write(uchar far *buffer, unsigned len)
{
	/*
		[ externals ]
	*/
	extern uchar sda_chardev;
	extern struct dev_req_hdr sda_reqhdr;
	extern unsigned sda_rdwrlen;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(buffer);
	ASSERT(len);

	/* this function cannot write to files */
	if(!sda_chardev)
	{
		EXIT_FUNCTION();
		return;
	}

	while(len)
	{
		sda_reqhdr.command = DEV_WRITE;
	   sda_reqhdr.trans = buffer;

		// Hyuandai's POS_MCR mag card reader char device driver
		// demands we send it 1 char at a time.
		//
		if(sda_chardev & IO_RAW)
			sda_reqhdr.count = len;
		else
			sda_reqhdr.count = 1;

		/* kludge: if trying to write a ^Z to cooked device, don't */
		/* note: we need to handle COOKED devices correctly.
		   set the mode to EOF?
		*/
		if(!(sda_chardev & IO_RAW) && *buffer == ctrl('Z'))
			break;

		dev_crit_call(&sda_reqhdr);

		/* update the counters */
		len -= sda_reqhdr.count;
		buffer += sda_reqhdr.count;
	}

	/* say how many bytes we really wrote */
	sda_rdwrlen -= len;

	EXIT_FUNCTION();
}


/*
	Read some characters from a device.  For all but the
	console device, this routine will keep trying until it
	has read all the requested characters.  For the console
	it may read less than requested.

	Returns the length of the read in "sda_rdwrlen".
*/
void	PASCAL dev_read(uchar far *buffer,unsigned len)
{
	/*
		[ externals ]
	*/
	extern struct dev_req_hdr sda_reqhdr;
	extern LPSFT sda_sftptr;
	extern unsigned sda_rdwrlen;


	/*
		[ locals ]
	*/
	bool raw;
	unsigned gotlen = 0;
	unsigned ioctl = sda_sftptr->ioctl;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(buffer);
	ASSERT(len);

	/* at end of file, stop reading */
	if((ioctl&DEVA_CON) == 0)
		raw = (ioctl & IO_RAW);
	else
		raw = 1;		// always raw for console
	if(!raw && (ioctl&IO_NOTEOF)==0)
	{
		*buffer = 0;
		goto done;
	}

	/* loop to get all the data */
	while(gotlen < len)
	{
		/* set up the request header */
		sda_reqhdr.command = DEV_READ;
	   sda_reqhdr.trans = buffer;
	   sda_reqhdr.count = raw ? len : 1;

		dev_crit_call(&sda_reqhdr);
		gotlen += sda_reqhdr.count;

		/* are we done with call? */
		if(raw || sda_reqhdr.count != 1)
			break;

		/* cooked mode stops at a CR */
		if((ioctl & DEVA_CON) == 0)
		{
			if(*buffer == '\r')
				break;

			/* cooked mode ends for good at a ^Z */
			if(*buffer == ctrl('Z'))
			{
				sda_sftptr->ioctl &= (~IO_NOTEOF);
				break;
			}
		}

		/* update the counters */
		buffer += sda_reqhdr.count;
	}

done:
	sda_rdwrlen = gotlen;

	EXIT_FUNCTION();
}


/*
	-------------------------------------------------------

								IOCTL

	-------------------------------------------------------
*/

/*
	Make the device call and check for an error.  If an
	error occurs, generate a critical error.

	Returns FALSE if the requested command is not supported
	by the device.  Returns TRUE if it is.
*/
bool	PASCAL dev_crit_call(struct dev_req_hdr *hdr)
{
	/*
		[ externals ]
	*/
	extern int sda_no24;
	extern int error_no;
	extern uchar CritErr;
	extern uchar sda_retry;
	extern LPDEV_HDR sda_devptr;


	/*
		[ locals ]
	*/
	/* note: figure out imperically the lengths of these commands */
	int i;
	struct dev_req_hdr htemp;
	LPDEV_HDR devptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(hdr);

	/* remember the reqhdr, because device call may change fields */
	memmove(&htemp, hdr, sizeof(struct dev_req_hdr));
	devptr = sda_devptr;

again:
	/* if call fails, retry it again */
	for(i = 0; i < sda_retry; ++i)
	{
		// The device just called may have re-entered DOS on another
		// stack and called a different device. We must restore the
		// devptr right away in case we return to a routine that
		// calls this function in a loop (assuming devptr constant).
		//
		sda_devptr = devptr;

		/* each (re)try must begin fresh
		*/
		memmove(hdr, &htemp, sizeof(struct dev_req_hdr));

		/* call the device */
		dev_call(hdr, sda_devptr);

		// The device just called may have re-entered DOS on another
		// stack and called a different device. We must restore the
		// devptr right away in case we return to a routine that
		// calls this function in a loop (assuming devptr constant).
		//
		sda_devptr = devptr;

		// if there was NOT an error, don't retry
		//
		if(!(hdr->status & DEVE_ERROR))
			break;

		if((uchar)(hdr->status) == DEVE_UNKNOWN)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// The block device handles its own retries.
		//
		if(!(devptr->attr & DEVA_CHAR))
		{
			i = sda_retry; // signal that we want a crit err generated
			break;
		}
	}

	/* if not already in a critical error, generate one */
	if(i == sda_retry)
	{
		if(sda_no24)
		{
			error_no = ERR_NOTPATH;
		}
		else if(!CritErr)
		{
			if(critical_error(hdr->command, (uchar)hdr->status))
			{
				goto again;
			}
			else
			{
				// SPR 1638 (de) must continue on a crit error with Ignore.
				// Ignoring a single device error should not terminate a
				// sequence of writes.
				//
				i = 0;
			}
		}
	}

	EXIT_FUNCTION();

	/* returns TRUE if worked, FALSE if failed */
	return (i < sda_retry);
}


/*
	Generate critical error.  This will set up all the
	proper registers to generate a critical error (INT 24H).
	The handler itself is written in assembly, and is found
	in the file "DEVCRIT.ASM".

   AH    = error bit map
   AL    = drive (0=A, 1=B)
   DI    = error code
   BP:SI = pointer to device

   1xxx xxxx	Char Device (0=disk)
   xx1x xxxx	Ignore allowed
   xxx1 xxxx	Retry allowed
   xxxx 1xxx	Fail allowed
   xxxx x11x	Area of error (00=BOOT, 01=FAT, 10=ROOT DIR, 11=DATA/SUB-DIR)
   xxxx xxx1	Error during write (0=read)

	"cmd" may be either DEV_READ or DEV_WRITE

	Returns TRUE if caller should retry the operation.

*/
int PASCAL critical_error(uchar cmd, uchar error)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_drive;
	extern uchar crit_drive;
	extern uchar crit_locus;
	extern LPDPB sda_dpbptr;
	extern LPDEV_HDR sda_devptr;
	extern LPSFT sda_sftptr;

	/*
		[ internals ]
	*/

// N/A if buffers aren't tossed before a write (if optimistic commit)
//	bool tossed=FALSE;
	bool retry=FALSE;
	LPSFT cur_sftptr;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set extended errors */
	error_no = error + ERR_LASTFILE + 1;
	default_error_codes();

	if(sda_devptr->attr & DEVA_CHAR)
	{
		// SPR 1638 (de) For character devices, set critical
		// error options here since they will always be the same...
		//
		crit_locus = (CRIT_CHARDEV | CRIT_IGNORE | CRIT_FAIL | CRIT_RETRY);

		if(cmd == DEV_WRITE)
			crit_locus |= CRIT_WRITE;
	}
	else
	{
	   // SPR 1638 (de) For any block device (disk). The crit_locus
		// is filled in by abs_disk_rdwr() (DOSDISK.C) so that the
		// disk location and the allowable action codes can be
		// correctly specified
		//
		// crit_locus |= (CRIT_DATA | CRIT_FAIL | CRIT_RETRY | CRIT_IGNORE);

		// assume we no longer "know" anything about this disk
		//
#if USING_FAT32
		if( IsFat32Drive(sda_dpbptr) )
		{
			sda_dpbptr->uFreeClustersHigh = 0xFFFF;
		}
		sda_dpbptr->freeclus          = 0xFFFF;
#else
		sda_dpbptr->freeclus = 0xFFFF;
#endif

		//	SPR 1638 (de) MS-DOS is optimistic in its commits while
		// we used to be pessimistic. For compatibility reasons,
		// we can't toss these buffers or the data will not show
		//	up on the disk, as it would under MS-DOS when a disk
		//	error occures and the int 24 handler returns with
		//	Ignore or if it just returns to the caller directly
		//	rather than returning back to DOS (as STACKER does.)
		//	See i24test.
		//
		//	buf_toss_new_dirty();
		//	tossed = TRUE;
	}

	// SPR 1638 (de) The critical error can occure on something other
	// than _curdrive when buffers are being LRU'd and flushed. In this
	// case quick_drive_setup() will have set the sda_dpbptr to the
	// drive being accessed but _curdrive will not point to where the
	// error is.
	//
	crit_drive = sda_dpbptr->drive;

	// actually do the critical error, remembering the current SFT
	// so that we don't end up read/writing to CON after a crit err
	// on a disk where the int 24h handler uses DOS' console I/O
	//
	cur_sftptr = sda_sftptr;
	retry = fat_crit_err((unsigned)error);
	sda_sftptr = cur_sftptr;

	// N/A if buffers aren't tossed before writing (if optimistic commit).
	//
	//	if( retry && tossed )
	//		restore_dirty_buffers();

	EXIT_FUNCTION();

	return retry;
}


/*
	-------------------------------------------------------

				INTERNAL OPEN/READ/CLOSE ROUTINES

	-------------------------------------------------------
*/


/*
	For Internal Use Only!!  This specialized function is
	used by EXEC and CONFIG to read in files, programs and
	devices.  It will operate both on FATs and the IFS.

	1) This function does NOT use handles,
	2) but DOES allocate an SFT
	3) Pass a name of NULL to use name already in SDA

	Returns 0xFF if failed, the SFT # if worked.
*/
uchar PASCAL internal_open(char far *name, bool dos2)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_sft_h;
	extern unsigned sda_callflags;
	extern unsigned doscallflag[];
	extern LPSFT sda_sftptr;
	extern uchar int21_ah;


	/*
		[ locals ]
	*/
	uchar call;
	uchar mode = O_RDWR;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* Exec must be read only for networks */
	if(int21_ah == 0x4B)
		mode = O_RDONLY;

	/* let subordinate routines know what kind of call this is */
	call = dos2 ? 0x3D: 0x0F;
	sda_callflags = callf_flagsof(call);

	/* parse the name into the SDA if needed */
	if(name && !path_to_sda(name,TRUE))
	{
		error_no = ERR_NOTFOUND;
	}

	/* make sure we have an available SFT */
	else if(!open_to_sda(mode))
		error_no = ERR_NOHANDLE;

	/* OK, actually DO the open */
	else
	{
		/* do the IFS/FAT call */
		if(!ifs_call())
		{
			/* go out to the disk and verify the path */
			validate_sda();
			if(!error_no)
				file_open();
		}
	}

	if(error_no)
	{
		if( error_no != ERR_DENIED )
		{
			error_no = 0;
		}
		sda_sft_h = 0xFF;
	}
	else
	{
		sda_sftptr->handles = 1;
	}

	EXIT_FUNCTION();

	return sda_sft_h;
}


/*
	For Internal Use Only!!  This function is used by the
	exec, terminate and config.  It is also called during
	the regs_to_sda process.  Assumes only that sda_sftptr
	points to a valid SFT.  It can close both FCBs and
	DOS 2 handle style SFTs.
*/
void	PASCAL internal_close(bool dos2)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar int21_ah;
	extern LPSFT sda_sftptr;
	extern unsigned sda_callflags;
	extern unsigned doscallflag[];
	extern unsigned sda_handle;


	/*
		[ locals ]
	*/
	uchar call;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* let subordinate routines know what kind of call this is */
	call = dos2 ? 0x3E : 0x10;
	sda_callflags = callf_flagsof(call);

	/* do the IFS/FAT call */
	if(!ifs_call())
	{
		file_close();

		/* free the SFT (for either DOS1 or DOS2 style) */
		/* note: don't decrement for IFS, as Novel VLMs do it */
		decr_handle_count();
	}

	/* free the handle if DOS2 call */
	if(dos2)
		set_handle_val(sda_handle, 0xFF);

	sda_callflags = callf_flagsof(int21_ah);
	error_no = 0;

	EXIT_FUNCTION();
}


/*
	For Internal Use Only!!  This function is used by EXEC
	and CONFIG.  It reads characters from the current FAT
	File, IFS or DEVice.

	Returns the # of bytes read.
*/
unsigned	PASCAL internal_read(uchar far *buffer, unsigned len)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar far * sda_rdwrbuff;
	extern LPSFT sda_sftptr;
	extern unsigned sda_rdwrlen;
	extern unsigned sda_callflags;

	/*
		[ locals ]
	*/
	unsigned oldflags;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(buffer);

	/* Make sure that the file we are reading is open!!! */
	ASSERT( sda_sftptr->handles );

	/* let subordinate routines know what kind of call this is */
	oldflags = sda_callflags;
	sda_callflags = callf_flagsof(0x3F);

	/* set the location and length of the read */
	sda_rdwrbuff = buffer;
	sda_rdwrlen = len;

	/* do the IFS/FAT call */
	if(!ifs_call())
		file_read();

	/* cleanup */
	sda_callflags = oldflags;
	if(error_no)
	{
		sda_rdwrlen = 0;
		error_no = 0;
	}

	EXIT_FUNCTION();

	return sda_rdwrlen;
}
