//
//   $Id: DOSDISK.C 1.39 2003/06/12 21:05:44Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**************************************************************************
   $DOCFILE:DOSDISK.C

	Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains FAT style disk I/O.  This includes
	functions for maniplating the FAT, functions to read
	and write sectors, and the DOS disk cache (buffers).

	$DOCHISTORY
	03/01/1999 de  FSInfo update into update free count.
	02/29/1999 de  Added USING_REMOVABLE_HARD_DISKS conditional to
						change_of_disk(). This code really needs rewritten.
	02/20/1999 dg  Added Fat32 AbsRdWr and FSINFO
	01/25/1999 bt  SPR 1882
	11/29/1990 dg  Created by Drew Gislason
**********************************************************/

#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 6


// BUF_LOOP actions
#define FLUSH_ALL 0
#define INVALIDATE 1
#define TOSS_NEW_DIRTY 2
#define RESTORE_DIRTY 3
#define OLD_DIRTY 4

/*
	-------------------------------------------------------

							FAT Management

	-------------------------------------------------------
*/


/*
	Convert a logical cluster within the Data Area of a disk
	into a physical cluster.  It is passed the FAT entry to
	start with, and the # of logical clusters.  The "lclus"
	passed should be 0 on exit.  If it is not, then the
	eofat was reached and "lclus" becomes the # of clusters
	to allocate.

	Returns the physical cluster relating to this logical
	cluster.
*/
CLUSTER PASCAL lclus_to_pclus(CLUSTER start, NCLUSTER *lclus)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPSFT sda_sftptr;

   extern unsigned sda_callflags;

	/*
		[ locals ]
	*/
	CLUSTER next;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate starting fat entry, and pointer to lclus
	ASSERT(start >= 2);
	ASSERT(lclus);

	/* search the FAT chain for the physical cluster */
	while(*lclus)
	{
      ASSERT_ValidCluster( start );
		next = get_fat_entry(start);

#if USING_FAT32 && LSEEK_ENHANCEMENT
		if( callf_handle() )
      {
         // if the next "cluster" is not contiguous with the current one,
         // we need to reset the uRDlseekHigh & uRDlseekLow fields
         //
         if( (start+1L) != next )
         {
            sda_sftptr->uRDlseekHigh = HIWORD(next);
            sda_sftptr->uRDlseekLow = LOWORD(next);
         }

      }
#endif  // #if USING_FAT32 && LSEEK_ENHANCEMENT

		if(next >= END_OF_FAT_CHECK || next < 2)
			break;

		/* generate critical error on this problem */
#if USING_FAT32
		if( IsFat32Drive(sda_dpbptr) )
		{
			if(next > sda_dpbptr->ulMaxCluster)
			{
				critical_error(0, 8);
			}
		}
		else
		{
			if(next > sda_dpbptr->hiclus)
			{
				critical_error(0, 8);
			}
		}
#else
		if(next > sda_dpbptr->hiclus)
			critical_error(0, 8);
#endif

		start = next;
		--(*lclus);
	}

	EXIT_FUNCTION();

	/* return the physical cluster */
	return start;
}

#if USING_FAT32
void CountFreeSpaceOnDrive(void)
{
	CLUSTER	ulNextFreeCluster;
	CLUSTER	ulFreeClusters;
	CLUSTER	ulMaxCluster;
	CLUSTER	ulCluster;

	ENTER_FUNCTION();

	ulMaxCluster       = sda_dpbptr->ulMaxCluster;
	ulFreeClusters     = 0;
	ulNextFreeCluster  = 0;

	for(ulCluster = 2; ulCluster <= ulMaxCluster; ++ulCluster)
	{
      ASSERT_ValidCluster( ulCluster );
		if(!get_fat_entry(ulCluster))
		{
			if(!ulNextFreeCluster)
				ulNextFreeCluster = ulCluster;
			++ulFreeClusters;
		}
	}
	sda_dpbptr->freeclus          = LOWORD(ulFreeClusters);
	sda_dpbptr->uFreeClustersHigh = HIWORD(ulFreeClusters);
	sda_dpbptr->ulNextFreeCluster = ulNextFreeCluster;

	//  Set the flag to indicate that the FSINFO sector needs updating
	//  on exit
	//
	sda_dpbptr->rebuild = 1;

	EXIT_FUNCTION();
}

/*
	Get the FS Info Fields from the disk. If they are invalid on the
	disk, then walk the FAT to determine values.
*/
void GetFSInfoFields( void )
{

	LPFSINFO lpFSInfo;

	ENTER_FUNCTION();

	// MUST be a FAT32 drive
	ASSERT(IsFat32Drive(sda_dpbptr));

	// get a pointer to the FS Info
	buf_getsector(sda_dpbptr->uFSInfoSector);
	lpFSInfo = (LPFSINFO)((char far *)sda_cur_buf->buffer + FSINFO_SIG_OFFSET);

	// bad FSInfo, create our own. Read through the entire FAT to find
	// the free clusters and first free
	if(   (lpFSInfo->ulSignature != FSINFO_SIG2)
      || (sda_dpbptr->rebuild == 0xFF)
      || (lpFSInfo->ulFreeClusters > sda_dpbptr->ulMaxCluster)
	  )
	{
		//  Rrebuild the free cluster information.
		//
		CountFreeSpaceOnDrive();

   	lpFSInfo->ulFreeClusters = sda_dpbptr->freeclus |
				((ulong)sda_dpbptr->uFreeClustersHigh << 16);
 	   lpFSInfo->ulFreeClusters = sda_dpbptr->ulNextFreeCluster;
	}
	else
	{
	   sda_dpbptr->freeclus          = LOWORD(lpFSInfo->ulFreeClusters);
	   sda_dpbptr->uFreeClustersHigh = HIWORD(lpFSInfo->ulFreeClusters);
		sda_dpbptr->ulNextFreeCluster = lpFSInfo->ulNextFreeCluster;
	}

	EXIT_FUNCTION();
}

/*
	Set the FSInfo fields on disk. Uses the current DPB. Assumes a
	GetFSInfoFields() was done previously, so the field should be valid.
*/
void SetFSInfoFields(void)
{
	LPFSINFO lpFSInfo;

	ENTER_FUNCTION();

	if( (sda_dpbptr->rebuild == 1) && (Fat32FreeClusters() != 0xffffffffL) )
	{
		// nothing to do for FAT16/12 drives
		if(!IsFat32Drive(sda_dpbptr))
		{
			EXIT_FUNCTION();
			return;
		}

		// get a pointer to the FS Info
		buf_getsector(sda_dpbptr->uFSInfoSector);
		lpFSInfo = (LPFSINFO)((char far *)sda_cur_buf->buffer + FSINFO_SIG_OFFSET);
		if( lpFSInfo->ulSignature != FSINFO_SIG2 )
		{
			// set up the sector with 0s, and other data
			fmemset(sda_cur_buf->buffer, 0, lol.bufsize);
			*(unsigned far *)(sda_cur_buf->buffer + SECTOR_SIG_OFFSET) = SECTOR_SIG;
			*(ulong far *)(sda_cur_buf->buffer) = FSINFO_SIG1;
		}

		// verify that
		ASSERT(lpFSInfo->ulSignature == FSINFO_SIG2);

		// set the fields from the DPB
		lpFSInfo->ulFreeClusters    = Fat32FreeClusters();
		lpFSInfo->ulNextFreeCluster = sda_dpbptr->ulNextFreeCluster;

		// tell buffering scheme this sector was changed
		buf_set_modified(sda_cur_buf);
		buf_flush(sda_cur_buf);
		sda_dpbptr->rebuild = 0;
	}

	EXIT_FUNCTION();
}
#endif // USING_FAT32

/*
	Allocate 1 or more clusters from the free space on a disk. This
	routine does not work for new entries (that is, if the staring
	cluster is 0).

	Returns the actual # of clusters allocated (may be 0).
*/
CLUSTER PASCAL get_new_clusters(CLUSTER *start, NCLUSTER num)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;

	/*
		[ locals ]
	*/
	NCLUSTER got;
	CLUSTER next;
	CLUSTER orgstart, orgval;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// must have a starting cluster for this chain
	ASSERT(*start >= 2);

	// must be allocating at least 1 cluster
	ASSERT(num);

	/* get as many as we need (or can) */
	orgstart = *start;

   ASSERT_ValidCluster( *start );
	orgval   = get_fat_entry(*start);
	for(got=0; got<num; ++got)
	{
		/* find the next free cluster */
		next = alloc_next_free();

		/* failed to get all the clusters, free them up, say we got none */
		if( (next == END_OF_FAT) || (next == *start) )
		{
         ASSERT_ValidCluster( *start );

         // Free only the clusters we allocated, starting just past start
         //
         next = get_fat_entry( orgstart );
         free_the_fat( next );

         ASSERT_ValidCluster( *start );

			set_fat_entry(orgstart,orgval);
			EXIT_FUNCTION();
			return 0;
		}

      ASSERT_ValidCluster( *start );
		set_fat_entry(*start,next);
		*start = next;
	}

	set_fat_entry(*start,END_OF_FAT);

	EXIT_FUNCTION();

	/* return how many clusters we got */
	return got;
}


#if USING_FAT32
/*
	helper routine to add and free clusters in the DPB.
*/
void AddToFreeCount(NCLUSTER count)
{
	ulong	ulFreeClusters;

	ENTER_FUNCTION();

	if( count )
	{
		// FAT32 drives have a more complicated add
		if(IsFat32Drive(sda_dpbptr))
		{
			if( Fat32FreeClusters() == 0xffffffffL )
			{
				GetFSInfoFields();
			}

			// update free clusters (one less)
			ulFreeClusters = Fat32FreeClusters();
			ulFreeClusters += count;
			sda_dpbptr->uFreeClustersHigh = HIWORD(ulFreeClusters);
			sda_dpbptr->freeclus          = LOWORD(ulFreeClusters);


			//  Update the FAT32 file system info block if this is a
			//  FAT32 drive
			//
			sda_dpbptr->rebuild = 1;
		}

		// update the free clusters for FAT16 drive
		else
		{
			if(sda_dpbptr->freeclus != Negative1)
			{
				sda_dpbptr->freeclus += (unsigned)count;
			}
		}
	}

	EXIT_FUNCTION();
}
#endif // USING_FAT32

/*
	Allocates a single cluster from the free pool.  Sets
	the contents of that cluster to "eofat." Updates the DPB.

	Returns the allocated cluster # of END_OF_FAT on error.
*/
CLUSTER alloc_next_free(void)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;


	/*
		[ locals ]
	*/
	CLUSTER start;
	CLUSTER hiclus;
	CLUSTER next;
	CLUSTER end;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();


#if USING_FAT32
	if( IsFat32Drive(sda_dpbptr) )
	{
      // This is a quick exit condition if we are using FAT32 fields.
      // First we must determine if the "free clusters" value is valid,
      // and update it if it is not.
      //
      // The side effect of not using this code is that the "free clusters"
      // value ends up deceremented to 0xffffffffL, and underflowing.
      //
      if( Fat32FreeClusters() == 0xffffffffL )
      {
         GetFSInfoFields();
      }
      // ... and if the free clusters is zero, quick exit
      //
      if( Fat32FreeClusters() == 0L )
      {
			EXIT_FUNCTION();
         return END_OF_FAT;
      }

		hiclus = sda_dpbptr->ulMaxCluster;
		start  = sda_dpbptr->ulNextFreeCluster;
	}
	else
	{
		hiclus = sda_dpbptr->hiclus;
		start  = sda_dpbptr->newclus;
	}
#else
	hiclus = sda_dpbptr->hiclus;
	start  = sda_dpbptr->newclus;
#endif

	/* define start and ending ranges */
	if(start<2 || start>=hiclus)
		start = hiclus;
	end = start;

	do
	{
		++start;

		/* check for wrap around */
		if((start < 2) || (start > hiclus))
			start = 2;

      ASSERT_ValidCluster( start );
		next = get_fat_entry(start);

		/* failure, we found no free clustor */
		if( (next) && (start == end) )
		{
			EXIT_FUNCTION();
			return END_OF_FAT;
		}

	} while(next);

	// FAT32 system
#if USING_FAT32

	// update freespace in FSINFOStructure on disk
	AddToFreeCount((ulong)(-1L));

	// update free clusters and next cluster
	if(IsFat32Drive(sda_dpbptr))
	{
		// update next free cluster
		sda_dpbptr->ulNextFreeCluster = start;
		sda_dpbptr->newclus           = LOWORD(start);
	}

	else
	{
		// update the next free cluster
		sda_dpbptr->newclus = LOWORD(start);
	}

	// FAT16/12 only system
#else

	// update the free clusters
	if(sda_dpbptr->freeclus != Negative1)
		--sda_dpbptr->freeclus;

	// update next free cluster
	sda_dpbptr->newclus = start;
#endif

   ASSERT_ValidCluster( start );

	EXIT_FUNCTION();

	return start;
}

/*
	Free a fat chain of a file or directory starting at a
	given point in the chain.
*/
void	PASCAL free_the_fat(CLUSTER start)
{
	CLUSTER	next;
	NCLUSTER	count;

	ENTER_FUNCTION();

	// validate starting cluster.  2 or more is OK.
	//
	/* zero out each one of these */
	count = 0;
	while(start && start < END_OF_FAT_CHECK)
	{
      ASSERT_ValidCluster( start );
		next  = get_fat_entry(start);
		set_fat_entry(start,0);
		start = next;
		++count;
	}

	// adjust freespace in DPB
#if USING_FAT32
	AddToFreeCount(count);
#else
	if(sda_dpbptr->freeclus != Negative1)
		sda_dpbptr->freeclus += count;
#endif

	EXIT_FUNCTION();
}

/*
	Convert a cluster # into a sector #.
*/
long PASCAL clus_to_sec(CLUSTER clus)
{
	extern LPDPB sda_dpbptr;

	ulong sector;

	ENTER_FUNCTION();

	// validate cluster number
	ASSERT(clus >= 2);

	sector = clus-2;
#if USING_FAT32
	if( IsFat32Drive(sda_dpbptr) )
	{
		sector = _lshlu(sector,sda_dpbptr->shift) + sda_dpbptr->ulFirstDataSector;
	}
	else
	{
		sector = _lshlu(sector,sda_dpbptr->shift) + sda_dpbptr->data_start;
	}
#else
	sector = _lshlu(sector,sda_dpbptr->shift) + sda_dpbptr->data_start;
#endif

	EXIT_FUNCTION();

	return sector;
}


/*
	Count contiguous clusters.  We are given the starting
	cluster and the maximum number to count.  Returns the
	number that are physically one right after the other on
	the disk (this will at least be 1).

	Limited to 64K clusters. Used during file operations which
	are limited to 64K bytes.
*/
unsigned PASCAL count_cont_clus(CLUSTER start, unsigned maxn)
{
	CLUSTER  next;
	unsigned cont = 1;

	ENTER_FUNCTION();

	// validate starting cluster
	while(maxn > 1)
	{
      ASSERT_ValidCluster( start );
		next = get_fat_entry(start);
		if(next != start+1)
			break;
		start = next;
		++cont;
		--maxn;
	}

	EXIT_FUNCTION();

	return cont;
}


/*
	Get the value of FAT entry (cluster #) n. Knows about 12, 16 and 32
	bit FATs. The END_OF_FAT marker is the same for any fat type.

	Returns the FAT entry, which may be 0 (free) or END_OF_FAT, or some
	value between 2 and max clusters.
*/
CLUSTER PASCAL get_fat_entry(CLUSTER n)
{
	extern LPDPB sda_dpbptr;
   extern char FAT_type;

	CLUSTER	value;
	unsigned	uFatStart;
	unsigned	uBytesPerSector;

	ENTER_FUNCTION();

	// determine some required info for get fat routines
	uFatStart       = sda_dpbptr->fat_start;
	uBytesPerSector = sda_dpbptr->bytes_per_sec;

	switch(FAT_type)
	{
		case FS_FAT12:
			ASSERT((n >= 2) && (n <= sda_dpbptr->hiclus));
			value = GetFat12Entry((unsigned)n, uFatStart, uBytesPerSector);

			/* sign extend END_OF_FAT */
			if(value > HICLUS12)
#if USING_FAT32
				value |= 0x0FFFF000L;
#else
				value |= 0xF000;
#endif
			break;

		case FS_FAT16:
			ASSERT((n >= 2) && (n <= sda_dpbptr->hiclus));
			value = GetFat16Entry((unsigned)n, uFatStart, uBytesPerSector);

#if USING_FAT32
			/* sign extend END_OF_FAT */
			if(value > HICLUS16)
				value |= 0x0FFF0000L;
#endif
			break;

#if USING_FAT32
		case FS_FAT32:
			ASSERT((n >= 2) && (n <= sda_dpbptr->ulMaxCluster));
			value = GetFat32Entry(n, uFatStart, uBytesPerSector);
			break;
#endif
	}

	EXIT_FUNCTION();

	/* return the value at that cluster */
	return value;
}

/*
	Set the value of FAT entry (cluster #) n. Knows about 12, 16 and 32
	bit FATs. It is OK to pass an END_OF_FAT value for any type FAT.
*/
void	PASCAL set_fat_entry(CLUSTER n, CLUSTER value)
{
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;
   extern char FAT_type;

	unsigned	uFatStart;
	unsigned	uBytesPerSector;

	ENTER_FUNCTION();

	// determine some required info for get fat routines
	uFatStart       = sda_dpbptr->fat_start;
	uBytesPerSector = sda_dpbptr->bytes_per_sec;

	switch(FAT_type)
	{
		case FS_FAT12:
			ASSERT((n >= 2) && (n <= sda_dpbptr->hiclus));
			SetFat12Entry((unsigned)n, (unsigned)value, uFatStart, uBytesPerSector);
			break;

		case FS_FAT16:
			ASSERT((n >= 2) && (n <= sda_dpbptr->hiclus));
			SetFat16Entry((unsigned)n, (unsigned)value, uFatStart, uBytesPerSector);
			break;

#if USING_FAT32
		case FS_FAT32:
			ASSERT((n >= 2) && (n <= sda_dpbptr->ulMaxCluster));
			SetFat32Entry(n, value, uFatStart, uBytesPerSector);
			break;
#endif
	}

	/* tell buffering scheme that sector was changed */
	buf_set_modified(sda_cur_buf);

	EXIT_FUNCTION();
}


/*
	-------------------------------------------------------

					Sector Level Disk I/O

	-------------------------------------------------------
*/

/*
	For removable disks, check to see if they have changed
	or not.  A floppy disk might be removed, so we may need
	to reread the BPB.

	Returns TRUE if we're on a new disk, FALSE if not.

	NOTE:	This function is badly out of control and needs to be redesigned!

			It combines rebuilding the DPB with the changing of
			physical media. The latter implies the former but the
			former may need to be done even if the disk has not
			changed. Also the mechanics of updating the media ID
			fields on a disk are something that should be done by a
			device not by DOS.

*/
bool change_of_disk(void)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern struct dev_req_hdr sda_reqhdr;
	extern LPDEV_HDR sda_devptr;
	extern LPBUFLIST sda_cur_buf;
	extern LPSFT sda_sftptr;
	extern uchar sda_drive;
	extern uchar crit_drive;
	extern uchar crit_locus;
	extern ulong err_esdi;
	extern uchar DMA_buff[];
	extern uchar CritErr;
	extern uchar def_volume[];
	extern uchar def_fatid[];

	/*
		[ locals ]
	*/
	ulong  serNum0;
	ulong  far * lpDDTSerialNumber;
	ulong  serNumN;
	uchar  changed;
	LPBUFLIST tmp_buf;
	LPSFT  cur_sftptr;
	uchar  ucOldCritErr;
	bool	 fReturnValue;
#if USING_FAT32
	pBOOTSECTOR pBootSector;
#else
	ulong  far * lpDiskSerialNumber;
#endif


	/*
		[ code ]
	*/


	ENTER_FUNCTION();

	// SPR 1817 (DE) -- We cannot allow this routine to cause a
	// critical error! It gets called on disk access regardless of
	// the buffer states. For instance, COMMAND.COM may force
	// this to get called it after a file COPY Aborts from crit
	// error when it tries to close the file handles. in such a
	// case, generating a critical error in this routine would
	// cause an endless critical error loop.
	//
	ucOldCritErr = CritErr;
	CritErr = TRUE;
	do
	{

		/* do the get media call */
		sda_reqhdr.command = DEV_MEDIA;
		set_media_unit();
		if(!dev_crit_call(&sda_reqhdr))
			changed = CHG_NO;
		else
			changed = (uchar)sda_reqhdr.trans;

		// Get serial number of disk that is "supposed" to be in the drive.
		// DEV_MEDIA returns a ptr to the 12 char ddt_volume field at offset
		// (trans+1=0x0F) in the device request header. This address is also
		// stored in the SDA err_esdi field to be returned by int 21h fn 59h
		// (get extended error info).
		//
		// (BT 01-25-1999  SPR 1882)
		// The lpDDTSerialPointer was being made a far pointer then a near
		// pointer and then a far pointer again.  This meant that any pointer
		// that was outside of dos's data segment caused a write into dos's
		// data segment at a random location.  Observe the asm output when
		// making modifications to these pointers.
		//
		err_esdi           = * (ulong *)(((char *)(&sda_reqhdr))+0x0F);
		lpDDTSerialNumber  = (ulong far *)((char far *)err_esdi+12);
		serNum0            = *lpDDTSerialNumber;
#if USING_FAT32
		pBootSector        = (pBOOTSECTOR)(&DMA_buff[0]);
#else
   	lpDiskSerialNumber =	((ulong *)((char far *)(&DMA_buff)+BOOT_SERIAL));
#endif


		/* note: above call MAY have changed the devptr and unit, ala stacker */
		sda_devptr = sda_dpbptr->device;
		set_media_unit();

		/* if rebuild has been set the disk has definitely changed */
		if(sda_dpbptr->rebuild == 0xFF)
			changed = CHG_YES;

		/* don't DO anything if disk not changed */
		if(changed == CHG_NO)
		{
			fReturnValue = FALSE;
			break;
		}

		// if the disk may have changed, check the buffers
		// if dirty act like not changed, if not dirty then rebuild drive
		// See MSDOS Tech Ref for dev dvr fn 1.
		//
		if(changed == CHG_MAYBE)
		{
			tmp_buf = dirty_buffer(sda_cur_buf);
			if((ulong)(tmp_buf) != 0L)
			{
				fReturnValue = FALSE;
				break;
			}
		}

		// We have to build a BPB for the newly changed disk (DISK_n) but
		// the disk that was in the drive before (DISK_m) may have dirty
		// buffers waiting to be written out. In this case we have to generate
		// a critical error prompting the user to replace DISK_m so that
		// the buffered data isn't lost.
		//
		while(changed == CHG_YES)
		{
			// if there are NOT any dirty buffers for this disk then
			// we don't care what disk is in the drive
			//
			tmp_buf = dirty_buffer(sda_cur_buf);
			if((ulong)(tmp_buf) == 0L)
				break;

#if (!USING_SWAPPABLE_HARD_DISKS)
			//
			// if this is a hard disk, just rebuild the DPB and leave...
			//
			if(sda_dpbptr->media == HARD_MEDIA)
				break;
#endif   // USING_SWAPPABLE_HARD_DISKS


			do
			{
#if USING_FAT32

				// get serial number of DISK_n from the boot sector
				//
				abs_disk_rdwr(1,0,1,(uchar far *)pBootSector);

#if USING_SWAPPABLE_HARD_DISKS
				//  Check if this is a FAT32 drive and retrieve the serial
				//  number field for the appropriate FAT type.
				//
				if( pBootSector->BootSectorInfo.BootInfo32.BootSectorBPB.uFatSectors
						== 0 )
				{
					serNumN =
						pBootSector->BootSectorInfo.BootInfo32.ulVolumeSerialNumber;
				}
				else
				{
					serNumN =
						pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber;
				}
#else   	// USING_SWAPPABLE_HARD_DISKS

				serNumN = pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber;

#endif   // USING_SWAPPABLE_HARD_DISKS
#else		// #if USING_FAT32

				// get serial number of DISK_n from the boot sector
				//
				abs_disk_rdwr(1,0,1,DMA_buff);
				serNumN = *lpDiskSerialNumber;
#endif	// #if USING_FAT32

				// If the serial numbers match then flush all dirty buffers
				//
				if(serNum0 == serNumN)
					break;

				// Generate an "Illegal Disk Change" critical error,
				// prompting for the Volume and Serial # of the original
				// disk.  On Fail or Abort we want to throw away any data
				// bufferd for DISK_m, but allow it to be committed on
				// Retry.
				//
				invalidate_dirty_buffers();
				cur_sftptr = sda_sftptr;
				crit_locus = (CRIT_DATA | CRIT_FAIL | CRIT_RETRY | CRIT_WRITE);
				crit_drive = sda_drive;

				fat_crit_err(0x0F);	// will only return on Retry or Ignore

				sda_sftptr = cur_sftptr;
				restore_dirty_buffers();

			} while(1);

			// ok, NOW we can actually write out the changed FAT/data
			buf_flush_all();

			// no need to throw away anything, as we have the correct
			// disk in the drive now.
			fReturnValue = FALSE;
			goto DONE;
		}

		/* throw away all buffers for this drive */
		buf_invalidate();

		/* if IBMBLK bit is clear get FAT sector, if set get scratch */
		if((sda_devptr->attr & DEVA_IBMBLK) == 0)
			buf_getsector((long)sda_dpbptr->fat_start);
		else
			buf_get_lru();

		/* make a request to read a new BPB */
		sda_reqhdr.command = DEV_BPB;
		set_media_unit();
		sda_reqhdr.trans = (void far *)sda_cur_buf->buffer;

		/* if the call worked, translate BPB into DPB */
		if(dev_crit_call(&sda_reqhdr))
			bpb_to_dpb(MK_FP(sda_reqhdr.start,sda_reqhdr.count), sda_dpbptr);

#if (!USING_SWAPPABLE_HARD_DISKS)
		//
		// don't try to autoserialize disks that can't be swapped even if the
		// boot sector serial number is hosed...
		//
		if(sda_dpbptr->media == HARD_MEDIA)
			goto DISK_SERIALIZED;
#endif   // USING_SWAPPABLE_HARD_DISKS

#if USING_FAT32
		// get serial number of disk from the boot sector
		//
		abs_disk_rdwr(1,0,1,(uchar far *)pBootSector);

#if USING_SWAPPABLE_HARD_DISKS
		//  Check if this is a FAT32 drive and retrieve the serial
		//  number field for the appropriate FAT type.
		//
		if( pBootSector->BootSectorInfo.BootInfo32.BootSectorBPB.uFatSectors
				== 0 )
		{
			serNumN =
					pBootSector->BootSectorInfo.BootInfo32.ulVolumeSerialNumber;
		}
		else
		{
			serNumN =
					pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber;
		}
#else    // USING_SWAPPABLE_HARD_DISKS
		serNumN = pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber;
#endif   // USING_SWAPPABLE_HARD_DISKS
#else		// #if USING_FAT32

		// get serial number of disk from the boot sector
		//
		abs_disk_rdwr(1,0,1,DMA_buff);
		serNumN = *lpDiskSerialNumber;

#endif	// #if USING_FAT32

		if(serNumN == 0)
		{ unsigned uCallersCX, uCallersDX, uTimeCX, uTimeDX;

			// SPR 1765 (de) Some manufacturers of pre-formatted
			// diskettes set all the serial numbers to zero. This makes
			// it impossible to distinguish between one diskette and
			// another. On a change of disk we now stamp the new disk
			// with a valid serial number if its current serial number
			// is zero. If the disk is write protected then we just have
			// to accept that we might lose data on a subsequent disk
			// change to another zero numbered diskette.
			//
			uCallersCX = R_CX;		// save caller's CX, DX
			uCallersDX = R_DX;

			// Create a new serial # based on current date/time.
			// The date and times are added together.
			//
			get_time();
			uTimeCX = R_CX;
			uTimeDX = R_DX;
			get_date();
			serNumN = (ulong)(MK_FP(R_DX+uTimeDX,R_CX+uTimeCX));

#if USING_FAT32
			// Update the serial number on the disk and in the DDT.
			// Also update the ID fields since they are probably bogus too.
			//
#if USING_SWAPPABLE_HARD_DISKS
			//  Check if this is a FAT32 drive and address the serial
			//  number, etc. for the appropriate FAT type.
			//
			//	NOTE:	Assumes no FAT16 removable media
			//
			if( pBootSector->BootSectorInfo.BootInfo32.BootSectorBPB.uFatSectors
					== 0 )
			{
				pBootSector->BootSectorInfo.BootInfo32.ulVolumeSerialNumber =
						serNumN;
				pBootSector->BootSectorInfo.BootInfo32.ucExtendedBootRecordSig =
						XBOOT_REC_SIG;
				//  Copy the volume label and FAT ID string
				//
				memmove(	&(pBootSector->BootSectorInfo.BootInfo32.rgucVolumeLabel),
						def_volume, (BOOT_VOL_LEN+BOOT_FATID_LEN));
				pBootSector->BootSectorInfo.BootInfo32.rgucFileSystemID[BOOT_FATID_DIGITS_NDX]='3'
				pBootSector->BootSectorInfo.BootInfo32.rgucFileSystemID[1+BOOT_FATID_DIGITS_NDX]='2'
			}
			else
			{
				pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber =
						serNumN;
				pBootSector->BootSectorInfo.BootInfo16.ucExtendedBootRecordSig =
						XBOOT_REC_SIG;
				//  Copy the volume label and FAT ID string
				//
				memmove(	&(pBootSector->BootSectorInfo.BootInfo16.rgucVolumeLabel),
						def_volume, (BOOT_VOL_LEN+BOOT_FATID_LEN));
			}

#else    // USING_SWAPPABLE_HARD_DISKS

			//
			//	NOTE:	Assumes all removable media are FAT12
			//
			pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber =	serNumN;
			pBootSector->BootSectorInfo.BootInfo16.ucExtendedBootRecordSig =	XBOOT_REC_SIG;
			//  Copy the volume label and FAT ID string
			//
			memmove(	&(pBootSector->BootSectorInfo.BootInfo16.rgucVolumeLabel),
					def_volume, (BOOT_VOL_LEN+BOOT_FATID_LEN));

#endif   // USING_SWAPPABLE_HARD_DISKS

			//  Write the serial number to disk
			//
			abs_disk_rdwr(0,0,1,(uchar far *)pBootSector);

#else    // USING_FAT32

			// Update the serial number on the disk and in the DDT.
			// Also update the ID fields since they are probably bogus too.
			//
			//	NOTE:	Assumes all removable media are FAT12
			//
			DMA_buff[BOOT_XBOOTSIG] = XBOOT_REC_SIG;
			*lpDiskSerialNumber = serNumN;
			memmove(&DMA_buff[BOOT_VOL_STR],
					  def_volume, (BOOT_VOL_LEN+BOOT_FATID_LEN));
			abs_disk_rdwr(0,0,1,DMA_buff);

#endif   // USING_FAT32

			*lpDDTSerialNumber  = serNumN;

			R_CX = uCallersCX;		// restore caller's CX, DX
			R_DX = uCallersDX;
		}

DISK_SERIALIZED:
		buf_set_invalid(sda_cur_buf);

		/* we no longer need to rebuild the DPB */
		sda_dpbptr->rebuild = 0;

		fReturnValue = TRUE;

	} while(0);

DONE:
	CritErr = ucOldCritErr;

	EXIT_FUNCTION();

	return fReturnValue;
}


#if USING_FAT32
/*
	Extended Absolute Read/Write function. Nearly identical to the
	Int 25/26h handling.
*/
void ExtendedAbsReadWrite(void)
{
	bool		bRead;
	unsigned	uRetCode;

	ENTER_FUNCTION();

	// make sure the drive is valid
	if(!defdrive_to_sda(R_DL) || (sda_cdsptr->status&C_NETWORK))
	{
		error_no = ERR_ABSRDWR;
		EXIT_FUNCTION();
		return;
	}

	// verify CX == -1
	if(R_CX != 0xFFFF)
	{
		error_no = ERR_BADLEVEL;
		EXIT_FUNCTION();
		return;
	}

	// call the function to read/write the sector
	bRead = !(R_SI & 1);
	uRetCode = PerformAbsRdWr(bRead);

	// did an error occur?
	if(uRetCode != ABS_NOERR)
		error_no = uRetCode;

	EXIT_FUNCTION();
}
#endif // USING_FAT32

/*
	This function is the meat of INT 25H and INT 26H.

	Returns error code (if any) in R_AX.
*/
void abs_rdwr(void)
{
	bool		bRead;
	unsigned uRetCode = 0;
	ulong		ulDriveSectors;

	ENTER_FUNCTION();

	// make sure the drive is valid
	if(!drive_to_sda(R_AL) || (sda_cdsptr->status&C_NETWORK))
	{
		uRetCode = ERR_ABSRDWR;
		goto ReportError;
	}

#if USING_FAT32
	// don't allow access to FAT32 drives with Int 25/26h
	if(IsFat32Drive(sda_dpbptr))
	{
		uRetCode = ERR_ABSRDWR;
		goto ReportError;
	}
#endif

	// for large disks, must use new BIG DISK method (cx = -1)
	ulDriveSectors = sda_dpbptr->data_start +
		_lshlu(sda_dpbptr->hiclus-1, sda_dpbptr->shift);
	if(HIWORD(ulDriveSectors) && R_CX != 0xFFFF)
	{
		uRetCode = ERR_ABSRDWR;
		goto ReportError;
	}

	// is this a read or write call?
	bRead = R_AH;
	uRetCode = PerformAbsRdWr(bRead);

ReportError:

	// send return code back to user
	R_AX = uRetCode;

	EXIT_FUNCTION();
}

/*
	PerformAbsRdWr

	This common routine is used both by the FAT32 version and the
	FAT16 version of abs rdwr.
*/
int PASCAL PerformAbsRdWr(bool bRead)
{
	uchar oldcrit;
	ulong ulStartSector;
	unsigned uCount;
	unsigned uRetCode;
	uchar far * lpDataBuffer;
	LPABS_RDWR lpAbsRdWr;

	ENTER_FUNCTION();

	// determine the start and size
	ulStartSector = R_DX;
	uCount        = R_CX;
	lpDataBuffer  = MK_FP(R_DS,R_BX);

	/* using BIG DISK format */
	if(uCount == 0xFFFF)
	{
		lpAbsRdWr     = (LPABS_RDWR)lpDataBuffer;
		ulStartSector = lpAbsRdWr->ulStartSector;
		uCount        = lpAbsRdWr->uCount;
		lpDataBuffer  = lpAbsRdWr->lpBuffer;
	}

	/* don't allow critical errors during INT 25H, 26H */
	oldcrit = CritErr;
	CritErr = TRUE;

#if 0
	/* hmm: what about the case where an Int 25h is the first
		access to this disk (eg: a floppy). Shouldn't a change of
		disk happen sometimes in Int 25h or 26h? - dg
	*/

	/*
		Once MS-DOS establishes what it thinks is a correct DPB, it
		does not worry about it being wrong until it detects a disk
		change. This test is not one that MS-DOS performs. In order
		to work with Performance Technology's PowerLan remote boot
		program, We need to be like MS-DOS and not force a "disk change
		event" as result of writes to the boot sector. - de
	*/

	/* if we have changed the disk, reread the bpb */
	if(changed)
	{
		change_of_disk();
		changed = 0;
	}

	/* if we are writing to the boot sector, we are "changing" the disk */
	if(!read && !start)
	{
		sda_dpbptr->rebuild = 0xFF;
		changed = 1;
	}

#endif

	/* flush all buffers for this drive to disk */
	buf_flush_all();

	/* do the actual read/write */
	abs_disk_rdwr(bRead,ulStartSector,uCount,lpDataBuffer);

	/* if writing to the drive, invalidate all the buffers */
	if(!bRead)
		buf_invalidate();

	/* restore critical error condition */
	CritErr = oldcrit;

	// return the error code, or ABS_NOERR (0x100) if worked
	if(sda_reqhdr.status & DEVE_ERROR)
		uRetCode = ((unsigned)bios_err << 8) | (uchar)sda_reqhdr.status;
	else
		uRetCode = ABS_NOERR;

	EXIT_FUNCTION();

	return uRetCode;
}

/*
	code size optimization - set media and unit for the
	drive referenced in this DOS call.
*/
void set_media_unit(void)
{
	extern LPDPB sda_dpbptr;
	extern struct dev_req_hdr sda_reqhdr;

	ENTER_FUNCTION();

	sda_reqhdr.media = sda_dpbptr->media;
	sda_reqhdr.unit  = sda_dpbptr->unit;

	EXIT_FUNCTION();
}

/*
	The absolute disk read/write routine.  This routine can
	read/write multiple sectors to/from a given buffer.  This may
	bypass (but is called by) the DOS disk buffering scheme.

	Returns nothing.
*/
bool PASCAL abs_disk_rdwr(bool rd, ulong start, unsigned count, uchar far *buff)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern struct dev_req_hdr sda_reqhdr;
	extern LPDEV_HDR sda_devptr;
	extern uchar  crit_locus;
	extern uchar  CritIgnoreOK;

#if USING_FAT32
	/*
		[ locals ]
	*/
	ulong ulDPBDataStart;
#endif

	ENTER_FUNCTION();

	// make sure pointer is ok
	ASSERT(buff);
	ASSERT(count);

	set_media_unit();
	sda_reqhdr.command = rd ? DEV_READ : DEV_WRITE;
	sda_reqhdr.trans   = buff;
	sda_reqhdr.count   = count;
	if(sda_devptr->attr & DEVA_32BIT)
	{
		sda_reqhdr.start = 0xFFFF;
		sda_reqhdr.start32 = start;
	}
	else
		sda_reqhdr.start = (unsigned)start;

#if USING_FAT32

	// SPR 1638 (DE) We used to claim all disk critical errors
	// occured in the "data" (files/sub-dirs) area of the disk.
	// Since all disk read/writes eventually come through this
	// routine. It is convenient to set up the crit_locus variable
	// here since the sda_dpbptr is already set up for the
	// appropriate drive and since the the starting sector number
	// is directly available.
	//
	if( IsFat32Drive(sda_dpbptr) )
	{
		ulDPBDataStart = sda_dpbptr->ulFirstDataSector;
	}
	else
	{
		ulDPBDataStart = (ulong)(sda_dpbptr->data_start);
	}

	if(start < ulDPBDataStart )
	{
		if(start < (ulong)(sda_dpbptr->fat_start))				// boot area
		{
			crit_locus = (CRIT_BOOT | CRIT_RETRY | CRIT_FAIL);
		}
		else if(start < (ulong)(sda_dpbptr->dir_start) )		// FAT areas
		{
			crit_locus = (CRIT_FAT | CRIT_RETRY | CRIT_FAIL);
		}
		else																	// root dir area
		{
			crit_locus = (CRIT_ROOTDIR | CRIT_RETRY | CRIT_FAIL);
		}
	}
	else																		// data area
	{
		// SPR 1638 (de)
		// although the Ignore bit is allowed for files only, the "affected
		// data area flags" are the same (xxxxx11xb = data area) for both
		// files and subdirectories under MS-DOS v6.22. See I24DIR.C.
		//
		crit_locus = (CRIT_DATA | CRIT_RETRY | CRIT_FAIL) | CritIgnoreOK;
	}
	if(!rd)
	{
		crit_locus |= CRIT_WRITE;
	}

#else // #if USING_FAT32

	// SPR 1638 (DE) We used to claim all disk critical errors
	// occured in the "data" (files/sub-dirs) area of the disk.
	// Since all disk read/writes eventually come through this
	// routine. It is convenient to set up the crit_locus variable
	// here since the sda_dpbptr is already set up for the
	// appropriate drive and since the the starting sector number
	// is directly available.
	//
	if(start < (ulong)(sda_dpbptr->data_start))
	{
		if(start < (ulong)(sda_dpbptr->fat_start))				// boot area
		{
			crit_locus = (CRIT_BOOT | CRIT_RETRY | CRIT_FAIL);
		}
		else if(start < (ulong)(sda_dpbptr->dir_start))			// FAT areas
		{
			crit_locus = (CRIT_FAT | CRIT_RETRY | CRIT_FAIL);
		}
		else																	// root dir area
		{
			crit_locus = (CRIT_ROOTDIR | CRIT_RETRY | CRIT_FAIL);
		}
	}
	else																		// data area
	{
		// SPR 1638 (de)
		// although the Ignore bit is allowed for files only, the "affected
		// data area flags" are the same (xxxxx11xb = data area) for both
		// files and subdirectories under MS-DOS v6.22. See I24DIR.C.
		//
		crit_locus = (CRIT_DATA | CRIT_RETRY | CRIT_FAIL) | CritIgnoreOK;
	}
	if(!rd)
	{
		crit_locus |= CRIT_WRITE;
	}

#endif  // #if USING_FAT32

	EXIT_FUNCTION();


	//  (BT 01-25-1999 SPR 1882)
	//  Returns a boolean value indicating if the write was successful.
	//  Previously, we were only using the global variable error_no to
	//  detect an error.  Now we can tell if there was an error on this
	//  write specifically.
	//
	return dev_crit_call(&sda_reqhdr);
}



/*
	-------------------------------------------------------

					DOS Disk Buffering Scheme

	-------------------------------------------------------
*/


#if NOT_IN_DOSPROF
/*
	Given a drive and sector, check to see if the buffer in
	question exists or not.  The sector is passed, the drive
	is in the SDA.  If found, it moves the buffer to the
	head of the list and returns TRUE.

	Returns FALSE if not found.
*/
bool	PASCAL buf_exist(ulong sector)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;
	extern struct LOL lol;


	/*
		[ locals ]
	*/
	LPBUFLIST buf;
	LPBUFLIST next;
	LPBUFLIST prev;
	void * head;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* look through whole list */
	buf = sda_cur_buf;
	head = (void *)FP_OFF(buf);
	do
	{
		/* check each buffer for the right drive/sector combo */
		if( buf_is_valid(buf) && buf->sector==sector &&
			buf->drive==sda_dpbptr->drive )
		{
			/* FOUND! if it's not already at the head, move it there */
			if((void *)FP_OFF(buf) != head)
			{
				/* extract the buffer from current position in list */
				next = MK_FP(FP_SEG(buf),buf->next);
				prev = MK_FP(FP_SEG(buf),buf->prev);
				prev->next = (BUFLIST *)FP_OFF(next);
				next->prev = (BUFLIST *)FP_OFF(prev);

				/* add it to the head */
				next = sda_cur_buf;
				prev = MK_FP(FP_SEG(next),next->prev);
				buf->next = (BUFLIST *)FP_OFF(next);
				buf->prev = (BUFLIST *)FP_OFF(prev);
				prev->next= (BUFLIST *)FP_OFF(buf);
				next->prev= (BUFLIST *)FP_OFF(buf);
				sda_cur_buf = lol.cur_buf = buf;
			}

			EXIT_FUNCTION();

			/* we found it */
			return TRUE;
		}

		/* check next buffer */
		buf = MK_FP(FP_SEG(buf),buf->next);
	} while((void *)FP_OFF(buf) != head);

	EXIT_FUNCTION();

	/* it wasn't found */
	return FALSE;
}
#endif

/*
	Get a "scratch" buffer.  Make sure it's invalid.
*/
void far * buf_get_scratch(void)
{
	extern LPBUFLIST sda_cur_buf;

	ENTER_FUNCTION();

	buf_set_invalid(sda_cur_buf);

	EXIT_FUNCTION();

	return buf_dataptr(sda_cur_buf);
}

/*
	Get the least recently used buffer and move it to the
	front of the linked list.  Flush the buffer if needed.
*/
void	PASCAL buf_get_lru(void)
{
	extern struct LOL lol;
	extern LPBUFLIST sda_cur_buf;

	ENTER_FUNCTION();

	/* search through list to find the last buffer (LRU) */
	sda_cur_buf = lol.cur_buf = MK_FP(FP_SEG(sda_cur_buf),sda_cur_buf->prev);

	/* if this buffer is valid, write it out to disk */
	if(buf_is_dirty(sda_cur_buf))
		buf_flush(sda_cur_buf);

	EXIT_FUNCTION();
}


/*
	Optimization for write!

	Given a drive (in the SDA) and a sector #, get that
	sector into the current buffer (LOL).  This will
	never read from the disk.

	This returns TRUE if a valid copy of the specified sector is
	already buffered.

*/
bool PASCAL buf_getnoread(ulong sector)
{
	/*
		[ externals ]
	*/
	extern LPBUFLIST sda_cur_buf;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// does the buffer already exist?  if so, we are done.
	//
	if(buf_exist(sector))
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	// find the least recently used buffer.
	//
	buf_get_lru();

	// SPR 1638 (de) fill in all the buffer control data
	// structure info prior to the read so that the buffer address
	// points to the correct place on the correct disk then mark
	// buffer invalid in case the read fails.
	//
	fill_in_buflist(sector);

	EXIT_FUNCTION();

	return FALSE;
}


/*
	We have the buffer we want, fill in the BUFLIST
	information in the current buffer.
*/
void	PASCAL fill_in_buflist(ulong sector)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	uchar flags;
#if USING_FAT32
	ulong ulDPBDataStart;
#endif

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

#if USING_FAT32

	// SPR 1638 (de) changed typecasts from (unsigned) to (ulong)
	// so that we don't have loss of information in sector location.
	//
	// determine the flags for this sector (affected data area)
	//
	if( IsFat32Drive(sda_dpbptr) )
	{
		ulDPBDataStart = sda_dpbptr->ulFirstDataSector;
	}
	else
	{
		ulDPBDataStart = (ulong)(sda_dpbptr->data_start);
	}

	if(sector < ulDPBDataStart)
	{
		if(sector < (ulong)(sda_dpbptr->fat_start))				/* boot area */
			flags = B_VALID | B_BOOT;
      else
      {
			if( IsFat32Drive(sda_dpbptr) )
			{
				flags = B_VALID | B_FAT;
			}
			else
			{
				if(sector < (ulong)(sda_dpbptr->dir_start))        /* FAT area */
   				flags = B_VALID | B_FAT;
				else                                                                    /* dir area */
   				flags = B_VALID | B_DIR;
			}
		}
	}
	else
		flags = B_VALID | B_DATA;

#else // #if USING_FAT32

	// SPR 1638 (de) changed typecasts from (unsigned) to (ulong)
	// so that we don't have loss of information in sector location.
	//
	// determine the flags for this sector (affected data area)
	//
	if(sector < (ulong)(sda_dpbptr->data_start))
	{
		if(sector < (ulong)(sda_dpbptr->fat_start))				/* boot area */
			flags = B_VALID | B_BOOT;
		else if(sector < (ulong)(sda_dpbptr->dir_start))		/* FAT area */
			flags = B_VALID | B_FAT;
		else																	/* dir area */
			flags = B_VALID | B_DIR;
	}
	else
		flags = B_VALID | B_DATA;

#endif // #if USING_FAT32

	/* set up the buffer */
	sda_cur_buf->drive  = sda_dpbptr->drive;
	sda_cur_buf->flags  = flags;
	sda_cur_buf->sector = sector;
	sda_cur_buf->copies = (flags & B_FAT) ? sda_dpbptr->num_fats : 1;
#if USING_FAT32
	if( IsFat32Drive(sda_dpbptr) )
	{
		sda_cur_buf->ulDistance = (flags & B_FAT) ? sda_dpbptr->ulSectorsPerFAT : 0;
	}
	else
	{
		sda_cur_buf->ulDistance = (flags & B_FAT) ? sda_dpbptr->fat_len : 0;
	}
#else
	sda_cur_buf->distance = (flags & B_FAT) ? sda_dpbptr->fat_len : 0;
#endif
	sda_cur_buf->lpDPB       = sda_dpbptr;
	sda_cur_buf->store_flags = 0;
	sda_cur_buf->store_drive = 0;

	EXIT_FUNCTION();
}


// give a ptr to a valid buffer, marks the buffer as newly modified (dirty)
//
void PASCAL buf_set_modified(LPBUFLIST buf)
{
	ENTER_FUNCTION();

	// SPR 1638 (de) only mark valid buffers as newly modified...
	//
	if(buf_is_valid(buf))
		buf_set_new_dirty(buf);

	EXIT_FUNCTION();
}

// SPR 1638 (de)
// Don't allow Ignore as an Int 24h option if we are reading/writing
// to/from a subdirectory. IGNORE is only OK for files. See I24DIR.C.
//
void CheckBufIgnoreOK(LPBUFLIST lpBuf)
{
	/*
		[ externals ]
	*/
	extern uchar  CritIgnoreOK;

	ENTER_FUNCTION();

	// assume lpBuf points to a buffered subdir/fat/root dir sector
	//
	CritIgnoreOK = FALSE;
	if(buf_is_in_data_area(lpBuf) && !buf_holds_a_dir(lpBuf))
	{
			CritIgnoreOK = CRIT_IGNORE;	// allow ignore on a file
	}

	EXIT_FUNCTION();
}


/*
	Given a drive (in the SDA) and a sector #, get that
	sector into the current buffer (LOL).  This will do
	a read from the disk if the buffer is not already in
	memory.

   This is the ONLY routine that actually reads buffers
   from disk.

	SPR 1638 (de) factored this code out of buf_getsector.
*/
void PASCAL buf_getread(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern void far * lo_buffer;
	extern LPBUFLIST sda_cur_buf;
	extern struct LOL lol;

	/*
		[ LOCALS ]
	*/
	uchar far * buffer;
	bool  bSuccessful;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// SPR 1638 (de) Ignore not allowed for I24 on subdirs, just files...
	// Invalidate the buffer and try to read the data into memory
	// from disk. If we don't return form abs_disk_rdwr because
	// of a critical error (or if we return with error_no != 0)
	// then the buffer will remain invalid.
	//
	CheckBufIgnoreOK(sda_cur_buf);
	buf_set_invalid(sda_cur_buf);
	buffer =
		(FP_SEG(sda_cur_buf) == 0xFFFF ? lo_buffer : buf_dataptr(sda_cur_buf));
	bSuccessful = abs_disk_rdwr(TRUE, sda_cur_buf->sector, 1, buffer);
	if(bSuccessful)
	{
		// If we are successful, update the buffer info from the DMA buffer
		// if we must and then  mark the buffer as valid again...
		//
		if(FP_SEG(sda_cur_buf) == 0xFFFF)
			fmemcpy(buf_dataptr(sda_cur_buf), buffer, lol.bufsize);
		buf_set_valid(sda_cur_buf);
	}

	EXIT_FUNCTION();
}

/*
	Given a drive (in the SDA) and a sector #, get that sector into the
	current buffer (LOL). This will return with sda_cur_buf pointing to
	the buffer which is holding a copy of the specified sector.

	SPR 1638 (de) changed this code to call factored subroutines
*/
LPBUFLIST PASCAL buf_getsector(ulong sector)
{
	extern LPBUFLIST sda_cur_buf;

	ENTER_FUNCTION();

	// If the specified sector is already in the buffers, just return
	//
	if(!buf_getnoread(sector))
	{
		// otherwise try to read the directory info off of the disk.
		//
		buf_getread();
	}

	EXIT_FUNCTION();

	// return a pointer to the data
	return sda_cur_buf;
}


/*
	Given a drive (in the SDA) and a sector #, get that sector
	into the current buffer (LOL). This will return with
	sda_cur_buf pointing to the buffer which is holding a copy of
	the specified sector.

	SPR 1638 (de) This is routine is similar to buf_getsector()
	above.  The ONLY difference is that this routine is ONLY
	called by the DIR management functions in DOSDIR.C. The
	reason for this routine is to mark the DIR buffers (in the
	data area of the disk with the B_DIR flag. In this way we can
	tell if an LRU'd buffer contains a DIR sector, in which case
	the Ignore option is not allowed on a critical error.

*/
void	PASCAL buf_get_dir_sector(ulong sector)
{
	/*
		[ externals ]
	*/
	extern LPBUFLIST sda_cur_buf;
	extern uchar  CritIgnoreOK;

	ENTER_FUNCTION();

	// If the specified sector is already in the buffers, just return
	//
	if(!buf_getnoread(sector))
	{
		// otherwise try to read the directory info off of the disk.
		// Mark it as a buffer containing a subdirectory if successful.
		//
		buf_getread();
		if(buf_is_valid(sda_cur_buf))
			buf_mark_as_dir(sda_cur_buf);
	}

	EXIT_FUNCTION();
}


/*
	Flush a buffer out to the disk.

	This is the ONLY routine that actually writes buffers
	to disk.
*/
void	PASCAL buf_flush(LPBUFLIST buf)
{
	extern LPDPB sda_dpbptr;
	extern void far * lo_buffer;
	extern struct LOL lol;
	extern uchar sda_drive;

	uchar  copies;
	bool   diff_drive;
	ulong  sector;
	uchar  far * buffer;
	bool	 bSuccessful;

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(buf);

	/* only write dirty buffers */
	if(buf_is_dirty(buf))
	{
		diff_drive = (sda_dpbptr->drive != buf->drive);

		/* set up the SDA for this drive */
		if(diff_drive)
			quick_drive_setup(buf->drive);

		/* write out all copies of this buffer */
		copies = buf->copies;
		sector = buf->sector;

		/*
			buffers in the HMA must be copied to conventional
			RAM before they can be written by the BIOS.
		*/
		if(FP_SEG(buf) == 0xFFFF)
		{
			fmemcpy(lo_buffer, buf_dataptr(buf), lol.bufsize);
			buffer = lo_buffer;
		}
		else
			buffer = buf_dataptr(buf);

		// spr 1638 (de) mark the buffer as clean and invalid in case
		// the device write fails/aborts. If the write works, then go
		// ahead and mark the buffer as valid again. This stops us from
		// constantly generating critical errors on the same buffered
		// bad sector.
		//
		buf_not_dirty(buf);
		while(copies)
		{
			// spr1638 (de) Ignore not allowed for I24 on subdirs
			//
			CheckBufIgnoreOK(buf);

			buf_set_invalid(buf);

			//  (BT 01-25-1999 SPR 1882)
			//  Were previously using the variable error_no to check
			//  if there was an error writing to disk. But if error_no
			//  was already set then the buffer would never be marked
			//  valid again even if this rd/wr was successful.
			//
			bSuccessful = abs_disk_rdwr(FALSE,sector,1,buffer);
			if(!bSuccessful)
			{
				break;
			}
			else
			{
				buf_set_valid(buf);
			}
#if USING_FAT32
			sector += buf->ulDistance;
#else // #if USING_FAT32
			sector += buf->distance;
#endif // #if USING_FAT32
			--copies;
		}

		/* restore the SDA to previous state */
		if(diff_drive)
			quick_drive_setup(sda_drive);
	}

	EXIT_FUNCTION();
}

/*
	Flush buffers for the current drive.  This will only
	write out buffers that are "dirty".
*/
void	PASCAL buf_flush_all(void)
{
	/*
		[ externals ]
	*/
	extern LPDEV_HDR sda_devptr;
	extern uchar sda_drive;

	/*
		[ locals ]
	*/
	LPDEV_HDR old_devptr;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set current device to this drive */
	old_devptr = sda_devptr;
	quick_drive_setup(sda_drive);

	/* go through all buffers and flush all that match */
   buf_loop(FLUSH_ALL);

	/* restore the original device pointer */
	sda_devptr = old_devptr;

	EXIT_FUNCTION();
}

/*
	Invalidate all buffers for this drive.  This does no
	disk access.
*/

void buf_invalidate(void)
{
	ENTER_FUNCTION();
	buf_loop(INVALIDATE);
	EXIT_FUNCTION();
}

/*
	For all the sectors in this range, and on the current
	drive, flush and invalidate them.
*/
void	PASCAL buf_flush_and_invalidate(ulong startsec, unsigned range)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	LPBUFLIST buf;
	void * head;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(range);

	/* go through all buffers and disable all */
	buf = sda_cur_buf;
	head = (void *)FP_OFF(buf);
	do
	{
		/* must be on this drive and valid */
		if(buf_is_valid(buf) && buf->drive == sda_dpbptr->drive)
		{
			if(buf->sector >= startsec && buf->sector < startsec+range)
			{
				buf_flush(buf);
				buf->flags = B_UNUSED;
				buf->drive = B_INVDRV;
			}
		}
		buf = MK_FP(FP_SEG(buf),buf->next);
	} while((void *)FP_OFF(buf) != head);

	EXIT_FUNCTION();
}

/*
	This function help reduce the size of ROM-DOS by
*/
void buf_loop(int func)
{
	extern LPBUFLIST sda_cur_buf;
	extern LPDPB sda_dpbptr;

	LPBUFLIST buf;
	void * head;

	ENTER_FUNCTION();

	buf = MK_FP(FP_SEG(sda_cur_buf), sda_cur_buf->prev);
	head = (void *)FP_OFF(buf);

	do
	{
      switch (func)
      {
         case FLUSH_ALL:
         	/* if this buffer is on this disk, flush it */
         	if(buf->drive == sda_dpbptr->drive && buf_is_dirty(buf))
   	      	buf_flush(buf);
            break;

         case INVALIDATE:
         	if (buf->drive == sda_dpbptr->drive)
         	{
         		buf_set_invalid(buf);
         		buf->drive = B_INVDRV;
            }
            break;

         case TOSS_NEW_DIRTY:
				if(buf_is_new_dirty(buf))
				{
					buf->store_flags = buf->flags;
					buf->store_drive = buf->drive;
					buf->flags = B_UNUSED;
					buf->drive = B_INVDRV;
				}
            break;

         case RESTORE_DIRTY:
         	if(buf->drive == B_INVDRV)
         	{
               buf->flags = buf->store_flags;
         		buf->drive = buf->store_drive;
         		buf->store_flags = B_UNUSED;
         		buf->store_drive = B_INVDRV;
         	}
            break;

         case OLD_DIRTY:
         	if(buf_is_new_dirty(buf))
               buf->flags &= ~B_NEWDIRTY;
            break;

         default:
            EXIT_FUNCTION();
            return;
      }
      buf = MK_FP(FP_SEG(buf),buf->prev);
   } while((void *)FP_OFF(buf) != head);


#if USING_FAT32
	if( func == FLUSH_ALL )
	{
		//  BT 6-10-02
		//
		//  Commit the FSINFO sector only when the FAT sectors are
		//  being flushed.
		//
		SetFSInfoFields();
	}
#endif // #if USING_FAT32


	EXIT_FUNCTION();
}

/*
	Toss all buffers that were made "dirty" during this
	DOS call.  These buffers are considered to be "new"
	dirt.  This happens on a critical error.
*/

void buf_toss_new_dirty(void)
{
	ENTER_FUNCTION();

	buf_loop(TOSS_NEW_DIRTY);

	EXIT_FUNCTION();
}

/*
	Restore storage flags to their proper place.
	This happens on a retry from a critical error on a FAT
	block device.
*/

void restore_dirty_buffers(void)
{
	ENTER_FUNCTION();

	buf_loop(RESTORE_DIRTY);

	EXIT_FUNCTION();
}

LPBUFLIST dirty_buffer(LPBUFLIST buf)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	void * head;

	ENTER_FUNCTION();

	if((ulong)(buf) == 0L)
	{
		EXIT_FUNCTION();
		return buf;
	}

	head = (void *)FP_OFF(buf);
	do
	{	// must be on this drive and be valid
		//
		if(buf_is_dirty(buf) && buf->drive == sda_dpbptr->drive)
		{
			EXIT_FUNCTION();

			// found one, return it
			//
			return buf;
		}
		buf = MK_FP(FP_SEG(buf),buf->next);

	} while((void *)FP_OFF(buf) != head);


	EXIT_FUNCTION();

	// none found, return NULL
	//
	return (LPBUFLIST)(0L);
}

void invalidate_dirty_buffers(void)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	LPBUFLIST buf;

	ENTER_FUNCTION();

	// go through all buffers and disable all dirty buffers for
	// the current drive
	//
	buf = sda_cur_buf;
	do
	{
      // find first/next dirty buffer for this drive
		//
		buf = dirty_buffer(buf);

		// if buf is NULL then there are no (more) dirty buffers
		//
		if((ulong)(buf) == 0L)
			break;

		// back up the buffer data and invalidate it
		//
		buf->store_flags = buf->flags;
		buf->store_drive = buf->drive;
		buf->flags = B_UNUSED;
		buf->drive = B_INVDRV;

	} while(1);

	EXIT_FUNCTION();
}

/*
	Mark any dirty buffers as "old" dirty buffers, that is,
	not dirtied by this call.
*/


void buf_old_dirty(void)
{
	ENTER_FUNCTION();

	buf_loop(OLD_DIRTY);

	EXIT_FUNCTION();
}
