//
//   $Id: DOSFCB.C 1.10 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSFCB.C  created 07/15/92  by Drew Gislason

   Copyright (c) 1992-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file contains the DOS FCB handling.  DOS contains
   a list of SFTs for use exclusively by FCBs.  In this
   way, DOS can keep track of "open" FCBs and use the same
   file system for both handles and FCBs.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 17

/*
	Get a local copy of the user FCB, and fill in the SDA
	variables that deal with FCBs directly.
*/
void	fcb_to_sda(void)
{
	/*
		[ externals ]
	*/
	extern uchar far * sda_fcbptr;
	extern uchar sda_extfcb;
	extern uchar sda_fcbattr;
	extern uchar sda_srchattr;
	extern FCB sda_fcb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	sda_fcbptr = MK_FP(R_DS,R_DX);

	/* deal with extended FCBs */
	sda_extfcb = sda_srchattr = sda_fcbattr = 0;
	if(*sda_fcbptr == 0xFF)
	{
		sda_extfcb = 0xFF;
		sda_srchattr = sda_fcbattr = *(sda_fcbptr+6);
		sda_fcbptr += 7;
	}

	/* get a local copy of the FCB */
	fmemcpy((void far *)&sda_fcb, sda_fcbptr, sizeof(FCB));

	EXIT_FUNCTION();
}


/*
	We have a valid SFT (open or create).  Now, fill in the
	fields of the FCB based on this SFT.
*/
void	sft_to_fcb(void)
{
	extern uchar sda_sft_h;
	extern uchar sda_drive;
	extern FCB sda_fcb;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	// move the SFT variables into the FCB
	sda_fcb.drive  = sda_drive+1;
	sda_fcb.file_size = sda_sftptr->size;
	sda_fcb.date   = sda_sftptr->date;
	sda_fcb.time   = sda_sftptr->time;
	sda_fcb.sft    = sda_sft_h;

	// for network FCBs, set a few fields so we can verify
	// the FCB is valid when it is later used to read/write.
	if(sda_sftptr->ioctl & IO_NETWORK)
	{
		sda_fcb.net_ioctl = ((uchar)sda_sftptr->ioctl) | 0x80;
#if USING_FAT32
		sda_fcb.ulStartingCluster = sda_sftptr->ulStartingCluster;
#else
		sda_fcb.net_stclus = sda_sftptr->start;
#endif
	}

	EXIT_FUNCTION();
}

/*
	Is this FCB both valid and open?
	Returns TRUE if it is, FALSE if not.  Sets up the "sda_sftptr".
*/
bool  is_valid_open_fcb(void)
{
	extern uchar sda_drive;
	extern FCB sda_fcb;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	// does the FCB even point to a valid SFT #?
	if(!h_to_sda(sda_fcb.sft))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// network FCBs use different fields for verification
	if(sda_sftptr->ioctl & IO_NETWORK)
	{
		// starting cluster MUST be the same
		// ioctl byte MUST say network
		// and the drives from the FCB and SFT MUST agree
		// for this FCB to be verified OK
#if USING_FAT32
		if(sda_fcb.ulStartingCluster != sda_sftptr->ulStartingCluster ||
#else
		if(sda_fcb.net_stclus != sda_sftptr->start ||
#endif
		   (sda_fcb.net_ioctl & 0x80) == 0 ||
			(sda_fcb.net_ioctl & 0x1F) != ((uchar)sda_sftptr->ioctl & 0x1F))
         {
			EXIT_FUNCTION();
			return FALSE;
         }
	}

	// must be a local FCB (non-network)
	else
	{
		char szName[16];

		// name in FCB may be lowercase, convert to UPPERCASE for comparison
		memmove(szName, sda_fcb.name, ENAMESIZE);
		szName[ENAMESIZE] = 0;
		strupr(szName);

		// make sure the drive and name in the FCB and the SFT agree
		if((fmemcmp((char far *)szName, sda_sftptr->name, 11) != 0) ||
				!((sda_fcb.drive == 0) || (sda_fcb.drive == sda_drive+1)))
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}

	EXIT_FUNCTION();

	// passed all tests, looks like the FCB points to this SFT
	return TRUE;
}

/*
	Does this FCB refer to a valid SFT?  If so, use it.  If
	not, open a new one (which MAY need to close the least
	recently used SFT if none available).

	Returns TRUE if it worked, FALSE if the file no longer
	exists.
*/
bool	fcb_exist(void)
{
	extern unsigned sda_callflags;

	char szName[16];
	unsigned oldfl;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* check if the FCB is already open */
	if(!is_valid_open_fcb())
	{
		/* save the flags for later restoration */
		oldfl = sda_callflags;

		/* try to open it again */
		fcb_to_dotname(szName, TRUE);
		if(internal_open( (char far *)szName, FALSE) == 0xFF)
		{
			sda_callflags = oldfl;
			EXIT_FUNCTION();
			return FALSE;
		}

		/* restore the flags */
		sda_callflags = oldfl;
		sft_to_fcb();
	}

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Set up the SDA to point to a free SFT (for an FCB call).
	This will close an existing FCB if need be.  The only
	way that this call can fail is if all FCBs are open and
	all are locked.

	Returns FALSE if all SFTs are locked.  Returns TRUE
	otherwise.
*/
bool	fcb_get_free_sft(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_numfcbs;
	extern uchar sda_fcbnext;
	extern uchar sda_sft_h;
	extern LPSFT sda_sftptr;
	extern unsigned sda_callflags;


	/*
		[ locals ]
	*/
	unsigned oldfl;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* if there's already a free one, we are done */
	if(get_new_sft())
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	/* pick the next one */
	++sda_fcbnext;
	if(sda_fcbnext >= sda_numfcbs)
		sda_fcbnext = 0;

	/* point to it in the SDA */
	sda_sft_h  = sda_fcbnext;
	sda_sftptr = h_to_sft(sda_sft_h);

	/* close it */
	oldfl = sda_callflags;
	internal_close(FALSE);
	sda_callflags = oldfl;

	/* guarantee this fcb is free */
	sda_sftptr->handles = 0;

	/* zero the SFT */
	get_new_sft();

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Convert the local FCB into a pathname form "A:FILE.EXT".
	This function understands both the normal name in an FCB
	and the one that DOS1 RENAME uses (when "first" is
	FALSE).
*/
void	PASCAL fcb_to_dotname(char *name, bool first)
{
	/*
		[ externals ]
	*/
	extern uchar CUR_DRIVE;
	extern uchar sda_srchattr;
	extern FCB sda_fcb;
	extern uchar int21_ah;


	/*
		[ locals ]
	*/
	uchar drive;
	char *fcbname;
	int i;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate pointer
	ASSERT(name);

	/* place the drive letter in the name */
	drive = sda_fcb.drive&0x7F;
	drive = drive ? drive-1 : CUR_DRIVE;
	*name++ = drive + 'A';
	*name++ = ':';

	/* volume labels are always in the root */
	if(sda_srchattr == A_VOLUME)
		*name++ = '\\';

	/* use normal or DOS1 rename position to get name from */
	fcbname = first ? (char *)sda_fcb.name : (char *)(&sda_fcb) + 0x11;
	for(i=0; i<11; ++i)
	{
		fcbname[i] = toupper(fcbname[i]);
	}
	ename_to_dot(name, fcbname);

	/*
	  Special case for FCB Delete/Rename Volume
	  Label: the name may have spaces, and so will not
	  "verify" as a valid filename.  Use a temporary one.

	  This code used to be executed for Create calls as well,
	  but DOS shouldn't allow "Create Volume Label *.*"
	  This code was removed: (	|| int21_ah==0x16 )
	*/
	if(sda_srchattr==A_VOLUME && (int21_ah==0x13 || int21_ah==0x17))
	{
		*name++ = 'X';		/* ANY valid name will do, so just use X */
		*name = 0;
	}

	EXIT_FUNCTION();
}

// is this one of the FCB read functions?
bool IsFcbRead(void)
{
	extern uchar int21_ah;

	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return (int21_ah == 0x14 || int21_ah == 0x21 || int21_ah == 0x27);
}
