//
//   $Id: DOSFILE.C 1.37 2003/08/21 14:32:18Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSFILE.C  created 12/01/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Tab = 3 chars.

	This file contains the FAT file operations for ROM-DOS.
	In here you will find file read, write, open and close,
	both for DOS1 (FCB) style calls and DOS2 (handle) style
	calls.  Anything dealing with a file must use a System
	File Table (SFT).

	Installable File System (IFS, ie network, CD-ROM drive)
	file operations are handled in DOSIFS.C.

**********************************************************/

#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 8


/*
	-------------------------------------------------------

			FAT File Create/Open/Close Type Things

	-------------------------------------------------------
*/

/*
	DOS call 6CH

	Open/Create a file with a variety of options.

	Openmode
	x1xx xxxx xxxx xxxx	All writes happen immediately
	xx1x xxxx xxxx xxxx	No critical error
	xxxx xxxx 1xxx xxxx  Child does not inherit handle

	Actions
	xxxx xxxx xxx1 xxxx	Create if file does not exist
	xxxx xxxx xxxx xx1x	Replace if file exists
	xxxx xxxx xxxx xxx1	Open if file exists
	legal actions: 1,2,10H,11H,12H
*/
void	PASCAL file_special_open(void)
{
	extern int error_no;
	extern uchar sda_chardev;		/* is this a char device or file? */
	extern uchar sda_srchattr;
	extern uchar sda_openmode;		/* the attr to create file with */
	extern unsigned sopen_act;		/* the special open action (DX) */
	extern unsigned sopen_attr;	/* the special open attribute (CX) */
	extern unsigned sopen_mode;	/* the special open mode (BX) */
	extern LPSFT sda_sftptr;		/* current SFT */
	extern SDB sda_sdb;

	bool exist;		/* does the file exist? */
	int action;		/* what action was taken? */

	ENTER_FUNCTION();

	/* no wildcards/spaces allowed in the name */
	if(!name_ok_to_create())
	{
		error_no = ERR_NOTFOUND;
		EXIT_FUNCTION();
		return;
	}

	/* determine if the file/device exists or not */
	if(sda_chardev)
		exist = (sopen_act&3) ? TRUE : FALSE;
	else
	{
		sda_sdb.srchattr = A_FILECHK;
		exist = dir_findfirst();
	}

	/* if file exists */
	if(exist)
	{
		/* Open if exists */
		if(sopen_act & ACT_EXIST_OPEN)
		{
			sda_openmode = (sopen_mode&3);
			file_open();
			action = 1;		/* file existed, was opened */
		}

		/* Create if exists */
		else if(sopen_act & ACT_EXIST_REPLACE)
		{
			sda_openmode = sopen_attr;
			int_file_create( exist );
			action = 3;		/* file existed, was replaced */
		}

		/* Fail if exists */
		else
			error_no = ERR_NOTNEW;
	}

	/* if file does not exists */
	else
	{
		/* Create if does not exist */
		if(sopen_act & ACT_NEXIST_CREATE)
		{
			sda_openmode = sopen_attr;
			int_file_create( exist );
			action = 2;		/* file did not exists, was created */
		}
		else
			error_no = ERR_NOTFOUND;
	}

	/* if there was no error, update the SFT with appropriate info */
	if(!error_no)
	{
		// MSDOS does not record the "don't inherit" bit in the rwmode
		// field but puts it in B12 of the ioctl field of the SFT.
		// All other file access bits are stored in rwmode.
		//
		R_CX = action;
		sda_sftptr->rwmode = (sopen_mode & ~O_NOCHILD);
		if(sopen_mode & O_NOCHILD)
			sda_sftptr->ioctl |= IO_NOCHILD;
	}

	EXIT_FUNCTION();
}

/*
	DOS calls 16H and 3CH

	Create a file or device.  The  path  and  filename  have
	already been verified and set up in the  SDA.   We  need
	only create an SFT and a handle.  Does NOT  modify  user
	registers.

	Sets "error_no" on failure.
*/
void	file_create(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;		/* is this a char device or file? */
	extern uchar sda_openmode;		/* the attr to create file with */
	extern SDB sda_sdb;		/* search data block */


	/*
		[ locals ]
	*/
	bool fExists;


	/*
		[ code ]
	*/


	ENTER_FUNCTION();


	/* this is a physical disk entry (not a char device) */
	if(!sda_chardev)
	{
		/* Only look for volume labels in the root directory */
		if(sda_openmode & A_VOLUME)
			prepare_volume_search();
		else
			sda_sdb.srchattr = A_DIRCHK;

		/* no wildcards allowed in the name */
		if(!name_ok_to_create())
		{
			error_no = ERR_NOTFOUND;
			EXIT_FUNCTION();
			return;
		}

		//  Perform a check to see if the file exists before
		//  calling the internal file create function
		//
		fExists = dir_findfirst();
	}


	//  The fExists variable is uninitialized when sda_chardev is
	//  nonzero but it is unused in int_file_create() in that case
	//
	int_file_create( fExists );


	EXIT_FUNCTION();
}


//  This is an internal create
//
void	int_file_create( bool fExists )
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;		/* is this a char device or file? */
	extern uchar sda_openmode;		/* the attr to create file with */
	extern DIR sda_ffb;		/* the found entry */
	extern SDB sda_sdb;		/* search data block */


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	//  BT 8-9-00
	//
	//  ROMDOS test suite test I21f3c.exe under MS-DOS verifies that you can
	//  not create volume labels
	//
	if(sda_openmode & A_VOLUME)
	{
		if( (R_AH == 0x6c) || (R_AH == 0x3c) )
		{
			error_no = ERR_DENIED;
			EXIT_FUNCTION();
			return;
		}
	}

	/* this is a physical disk entry (not a char device) */
	if(!sda_chardev)
	{
		/* if already exists, delete the old one */
		if( fExists )
		{
			/* do not delete directories or read only files */
			if((sda_openmode & A_NOMAKE) || (sda_ffb.attr & A_NODEL))
			{
denied:
				error_no = ERR_DENIED;
				EXIT_FUNCTION();
				return;
			}
#if USING_FAT32
			free_the_fat(Fat32Start(sda_ffb));
#else
			free_the_fat(sda_ffb.start);
#endif

			//  BT 6-12-02
			//
			//  Don't flush the buffers.  The FAT change and FSINFO changes will
			//  be committed after the file is closed.
			//
			//  buf_flush_all();
		}

		/* try to find an unused entry */
		else if(!dir_findblank_create())
			goto denied;

		/* put the entry on the disk */
		create_dir_entry(0);

		/* create, but do not open volume labels */
		if(sda_openmode & A_VOLUME)
		{
			EXIT_FUNCTION();
			return;
		}
	}

	/* open the newly created file */
	sda_openmode = O_RDWR;
	file_open();

	EXIT_FUNCTION();
}


/*
	DOS calls 0FH and 3DH

	Open a file or device.  Because  this  function  may  be
	used internally, no user registers are modified.  A  PSP
	is required, however.  The path and filename  must  have
	already been verified.  All we have to do is  set  up  a
	proper SFT and allocate a handle.  Does NOT modify  user
	registers.

	Sets error_no if failed.
*/
void	file_open(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;
	extern uchar sda_openmode;
	extern uchar sda_sft_h;
	extern DIR sda_ffb;
	extern LPSFT sda_sftptr;
	extern LPDEV_HDR sda_devptr;
	extern unsigned current_psp;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* check for valid open mode */
	if((sda_openmode&0x03) > O_RDWR)
	{
		error_no = ERR_BADCODE;
		EXIT_FUNCTION();
		return;
	}

	// Initialize the SFT.
	// MSDOS does not record the "don't inherit" bit in the rwmode
	// field but puts it in B12 of the ioctl field of the SFT.
	// All other file access bits are stored in rwmode.
	//
	sda_sftptr->rwmode = (sda_openmode & ~O_NOCHILD);
	sda_sftptr->device = sda_devptr;

	/* convert the dir entry or device hdr into an SFT */
	if(sda_chardev)
	{
		dev_to_sft();
		dev_openclose(TRUE);
	}
	else
	{
		/* do not allow open of directories or vol labels */
		if((sda_ffb.attr&A_NOMAKE))
		{
			error_no = ERR_DENIED;
			EXIT_FUNCTION();
			return;
		}

		dir_to_sft();
	}

	if(sda_openmode & O_NOCHILD)
		sda_sftptr->ioctl |= IO_NOCHILD;

	EXIT_FUNCTION();
}



/*
	Close a file  or  device.   Because  this  may  be  used
	internally, the user registers are not changed.  The SFT
	for the file is pointed to by "sda_sftptr" in  the  SDA.
	Does NOT decrement the handle count or free the handle.

	No errors other than a critical error  are  possible  at
	this point.
*/
void	file_close(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern unsigned sda_handle;
	extern uchar sda_chardev;
	extern LPSFT sda_sftptr;
	extern LPDEV_HDR sda_devptr;
	extern struct LOL lol;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	LPDIR dir;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* must be a device */
	if(sda_chardev)
	{
		dev_openclose(FALSE);
		EXIT_FUNCTION();
		return;
	}

	/* update the directory entry if needed */
	if((sda_sftptr->ioctl & IO_NOTWRITTEN) == 0)
	{
		/* detect change of disk here for STACKER support */
		/*
			note: what stacker support?  Commented out to support
			Geoworks - GEOS depends upon close NOT doing an INT 13H
			unless the file was modified - so for read only, INT 13H
			does not happen on close
		*/
		/* check for illegal change of disk... */
		/* if we've changed disks, we can't do anything more... */
		if(change_of_disk())
		{
			EXIT_FUNCTION();
			return;
		}

		/* get the sector contain directory entry */
		buf_getsector(sda_sftptr->dirsec);

		/* find directory entry */
		dir = (LPDIR)sda_cur_buf->buffer;
		dir += sda_sftptr->dirent;

		/* if entry has been deleted, delete the FAT tree */
		if(fmemcmp(sda_sftptr->name, dir->name, 11) != 0)
		{
			/* FAT tree is gonzo */
#if USING_FAT32
			free_the_fat(sda_sftptr->ulStartingCluster);
			sda_sftptr->ulStartingCluster = sda_sftptr->ulAbsCluster = 0;
#else
			free_the_fat(sda_sftptr->start);
			sda_sftptr->start = sda_sftptr->absclus = 0;
#endif
			sda_sftptr->size = 0;

			/* close handle */
			set_handle_val(sda_handle, 0xFF);
			error_no = ERR_BADHANDLE;
		}

		/* entry exists */
		else
		{
			/* convert the "current" SFT into a DIR structure */
			sft_to_dir();

			/* save and flush (dir entry, FAT and data buffs) to disk */
			save_dir_entry(sda_sftptr->dirent);
		}

		sda_sftptr->ioctl |= IO_NOTWRITTEN;
	}

	/* flush all areas of the current disk */
	buf_flush_all();

	EXIT_FUNCTION();
}


/*
	DOS calls 17H and 56H

	Rename one or many files.  Only the DOS1 call (17H)  can
	rename more than one file at a time.  For DOS call  56H,
	the "regs_to_sda" function will make sure  the  name  is
	found and has no wildcards before it ever gets here.
	DOS call 56H can rename a directory, but cannot move the
	directory.
*/
void	file_rename(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar int21_ah;
	extern uchar sda_srchattr;
	extern SDB sda_sdb;		/* the rename "from" location */
	extern FCB sda_fcb;
	extern DIR sda_ffb;
	extern SDB ren_sdb;		/* the rename "to" location */
	extern DIR ren_ffb;
	extern char sda_ename2[];
	extern uchar sda_in_5D;


	/*
		[ locals ]
	*/
	SDB sdb;
	int renamed = 0;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set search attribute */
	if(int21_ah == 0x56)
		sda_sdb.srchattr = A_DIRCHK;
	else
	{
		memmove(sda_sdb.emask, sda_fcb.name, 11);
		sda_sdb.srchattr = sda_srchattr;

		// MS-DOS does NOT allow hidden or system files to be renamed using
		// function 17h.
		//
		sda_sdb.srchattr &= ~(A_HIDDEN | A_SYSTEM);
	}

	/* wildcards ok.  DOS 2 wildcards already masked */
	/* if no match found, error */
	if(!dir_findfirst())
	{
		error_no = ERR_NOTFOUND;
		EXIT_FUNCTION();
		return;
	}

	/* rename all files that match */
	do
	{
		// (DE) MS-DOS allows any FILE (RSH) to be renamed using fn 56h.
		// Allow volume label renames, if DOS1 and search allows for it
		//
		if((sda_ffb.attr & A_VOLUME) && (sda_sdb.srchattr!=A_VOLUME))
		{
			error_no = ERR_DENIED;
			break;
		}

		/* don't rename "." or ".." */
#if USING_FAT32
		if(sda_sdb.ulStartingCluster && sda_sdb.dirnum<2)
#else
		if(sda_sdb.subdir && sda_sdb.dirnum<2)
#endif
			continue;

		/* modify the "from" name into a "to" name */
		name_change(sda_ffb.name, sda_ename2);

      // special character E5 handling
      //
      if((uchar)sda_ffb.name[0] == (uchar)0xE5)
      {
         // Special translation to allow 0xE5 character in directory entry
         //
         sda_ffb.name[0] = 0x05;
      }

		/* save the original File Found and Search Data Blocks */
		memmove(&ren_ffb, &sda_ffb, sizeof(DIR));
		memmove(&sdb, &sda_sdb, sizeof(SDB));

		/* search destination for name.  if it exists, fail */
#if USING_FAT32
		sda_sdb.ulStartingCluster = ren_sdb.ulStartingCluster;
#else
		sda_sdb.subdir = ren_sdb.subdir;
#endif
		memmove(sda_sdb.emask, ren_ffb.name, 11);
		sda_sdb.srchattr = A_DIRCHK;
		if(dir_findfirst())
		{
			error_no = ERR_DENIED;
			break;
		}

		/* if renaming to a new destination directory */
#if USING_FAT32
		if(sda_sdb.ulStartingCluster != sdb.ulStartingCluster)
#else
		if(sda_sdb.subdir != sdb.subdir)
#endif
		{
			/* do NOT allow a directory to be moved */
			if(ren_ffb.attr & A_SUBDIR)
			{
				error_no = ERR_DENIED;
				break;
			}

			/* we must have an avaible entry in destination */
			if(!dir_findblank())
			{
				error_no = ERR_DENIED;
				break;
			}

			/* write out the new entry */
			memmove(&sda_ffb, &ren_ffb, sizeof(DIR));
			update_dir_entry(FALSE);	/* preserve date */

			/* remove the old entry */
			memmove(&sda_sdb, &sdb, sizeof(SDB));
			(uchar)(*sda_ffb.name) = DELETED;
			update_dir_entry(FALSE);
		}

		/* the file is in the same directory, just change the entry */
		else
		{
			/* restore the original File Found and Search Data Blocks */
			memmove(&sda_sdb, &sdb, sizeof(SDB));
			memmove(&sda_ffb, &ren_ffb, sizeof(DIR));

			/* write the new entry out */
			update_dir_entry(FALSE);	/* preserve date */
		}

		/* renamed THAT file */
		++renamed;

		//  Only one file can be renamed with function 56h.  Wildcards
		//  are not allowed
		//
		if( (!sda_in_5D) && (int21_ah == 0x56) )
		{
			break;
		}

	} while(dir_findnext());
	buf_flush_all();

	/* if DOS1 and no files renamed, error */
	if(int21_ah == 0x17 && !renamed)
		error_no = ERR_DENIED;

	EXIT_FUNCTION();
}

/*
	DOS calls 13H and 41H

	Delete one or many files.  Only the DOS1 call (13H) can
	delete more than one file at a time.  The "regs_to_sda"
	function will make sure the name is found  and  has  no
	wildcards (for DOS2 call) before it ever gets here.

	Sets error_no on error.
*/
void	file_delete(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_srchattr;
	extern uchar int21_ah;
	extern DIR sda_ffb;
	extern SDB sda_sdb;
	extern uchar sda_in_5D;

	int count;
	int hit_nodel;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* what kind of file are we deleting? */
	sda_sdb.srchattr = (int21_ah == 0x13) ? sda_srchattr : A_FILECHK;
	if(sda_sdb.srchattr == A_VOLUME)
		prepare_volume_search();

	/* if it's not found, give up */
	if(!dir_findfirst())
	{
		error_no = ERR_NOTFOUND;
      EXIT_FUNCTION();
		return;
	}

	/* delete all files that match */
	count = hit_nodel = 0;
	do
	{
		/* we cannot delete a read only file or directory */
		if(sda_ffb.attr & A_NODEL)
			++hit_nodel;
		else
		{
			/* mark the entry as deleted on the disk */
			remove_dir_entry();
			++count;

			//  Only FCB delete can find more than one (hence wildcards)
			//
			if( (!sda_in_5D) && (int21_ah == 0x41) )
			{
				break;
			}
		}
	} while(dir_findnext());

	/* the only files were not deletable */
	if(hit_nodel && !count)
		error_no = ERR_DENIED;

	/* flush all changed buffers to disk */
	buf_flush_all();

	EXIT_FUNCTION();
}

/*
	DOS call 43H

	Set the file attribute.  Volume labels cannot be changed
	with this function, but directories can.
*/
void	PASCAL file_setattr(uchar newattr)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern DIR sda_ffb;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* make sure this is a valid attribute to change */
	if((sda_ffb.attr & A_VOLUME))
	{
		error_no = ERR_DENIED;
		EXIT_FUNCTION();
		return;
	}

	/* set the new attribute */
	if(sda_ffb.attr & A_SUBDIR)
		newattr |= A_SUBDIR;
	sda_ffb.attr = newattr;

	/* and write it to disk */
	update_dir_entry(FALSE);

	EXIT_FUNCTION();
}


/*
	DOS call 57H

	Set the file's time and date.  Until this handle is
	closed, the date/time is not actually written to the
	disk.
*/
void	PASCAL file_setdate(unsigned time, unsigned date)
{
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	/* set the time and date */
	sda_sftptr->time = time;
	sda_sftptr->date = date;
	sda_sftptr->ioctl |= IO_SETDATE;
	sda_sftptr->ioctl &= ~IO_NOTWRITTEN;

	EXIT_FUNCTION();
}

/*
	Using the time functions, create a very odd name,
	which (hopefully) will be unique.
*/
void create_unique_name(char *name)
{
	extern unsigned minute;
	extern unsigned hund;

	ENTER_FUNCTION();

	// string must exist
	ASSERT(name);

	touch_time(TM_READ);
	xitoa(name,hund,16);
	xitoa(name+4,minute,16);

	EXIT_FUNCTION();
}

/*
	DOS call 5AH

	This function will create a totally unique filename in
	the specified path.  The path must exist, however.

	Sets "error_no" on failure.
*/
void	file_create_unique(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char sda_ename1[];
	extern char tmp_name[];
	extern uchar sda_5a_strlen;
	extern SDB sda_sdb;


	/*
		[ locals ]
	*/
	char * name;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* the unique "name" is always 8 characters, see DOSSDA, 5AH */
	name = tmp_name + strlen(tmp_name) - 8;

	/* a "sample" name is already in place, use a better one */
	while(1)
	{
		/* devise a unique name based on the current time */
		create_unique_name(name);
		if(!path_to_sda((void far *)tmp_name,TRUE))
		{
			error_no = ERR_NOTPATH;
			break;
		}

		/* try to create the file */
		memmove(sda_sdb.emask,sda_ename1,11);
		file_create_new();

		/* if it's not new, try again. if another error occurred, stop! */
		if(error_no != ERR_NOTNEW)
			break;
		error_no = 0;
	}

	EXIT_FUNCTION();
}


/*
	DOS call 5BH

	This routine will create the file in question only if it
	has not already been created.

	Sets "error_no" on failure.
*/
void	file_create_new(void)
{
	extern int error_no;
	extern SDB sda_sdb;

	ENTER_FUNCTION();

	/* Make sure that we look for ALL valid directory entries */
	sda_sdb.srchattr = A_DIRCHK;

	/* Make sure this doesn't exist */
	if(dir_findfirst())
		error_no = ERR_NOTNEW;
	else
		int_file_create( FALSE );

	EXIT_FUNCTION();
}


/*
	Returns TRUE if the current filename contains a wildcard
	Returns FALSE if not.  Assumes the name has been parsed
	into the SDA and that "sda_ename1" is asciiz.
*/
bool wild_name(void)
{
	extern char sda_ename1[];

	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return (bool)strchr(sda_ename1,'?');
}

/*
	Is this name OK to create?  Assumes the name has been
	parsed into the SDA.
*/
bool name_ok_to_create(void)
{
	extern char sda_ename1[];

	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return !(wild_name() || (sda_ename1[0]==' ' && sda_ename1[8]==' '));
}


/*
	-------------------------------------------------------

				FAT File Read/Write Routines

	-------------------------------------------------------
*/


/*
	Write to a file or device.
*/
void	file_write(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_chardev;
	extern uchar far * sda_rdwrbuff;
	extern unsigned sda_rdwrlen;
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* for devices, do a device write */
	if(sda_chardev)
	{
		/* if it's the console, treat it specially */
		if(sda_chardev & DEVA_CON)
			sda_rdwrlen = dev_con_write(sda_rdwrbuff, sda_rdwrlen);
		else
			dev_write(sda_rdwrbuff, sda_rdwrlen);
	}
	else
	{
		/* if the write worked, there is some more work to do */
		if(file_rdwr(TRUE))
		{
			/* only say "written to" if it actually is */
			sda_sftptr->ioctl &= (~IO_NOTWRITTEN);

			/* write out the file immediately, if that's what the user wants */
			if(sda_sftptr->rwmode & O_WRITENOW)
				buf_flush_all();
		}
	}

	EXIT_FUNCTION();
}


/*
	Read from a file or device.
*/
void	file_read(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_chardev;
	extern uchar far * sda_rdwrbuff;
	extern LPSFT sda_sftptr;
	extern unsigned sda_rdwrlen;


	/*
		[ locals ]
	*/
	ulong size;
	ulong offset;


	/*
		[ code ]
	*/


	ENTER_FUNCTION();

	/* for devices, do a device read */
	if(sda_chardev)
	{
		/* if it's the console in "cooked" mode, do string input */
		if((sda_chardev & (DEVA_CON | IO_RAW)) == DEVA_CON)
			sda_rdwrlen = dev_con_read(sda_rdwrbuff, sda_rdwrlen);
		else
			dev_read(sda_rdwrbuff, sda_rdwrlen);
		EXIT_FUNCTION();
		return;
	}

	/* cannot read past the end of the file */
	offset = sda_sftptr->offset;
	size   = sda_sftptr->size;
	if(offset >= size)
	{
		sda_rdwrlen = 0;
		EXIT_FUNCTION();
		return;
	}

	/* if part of read is past end of file, shorten it */
	if(offset + sda_rdwrlen  > size)
		sda_rdwrlen = (unsigned)(size - offset);

	file_rdwr(FALSE);

	EXIT_FUNCTION();
}

// 2*1024^3 exceeds the capacity of a 32-bit signed number. We want
// the number of of sectors in a file that is no larger than 2Gbytes-1,
// the largest byte value that can be contained in a signed number.
// use (2*1024^3/512) = 2*1024^2*1024/512= 2*1024^2*2=4*1024^2 instead.
// In order to be able to store the same file on all systems, we limit
// the file size to 2G-MAX_SPC. The file limit is based on the significance
// of the signed 32 bit argument to seek() (i21h fn 42h).
//
#define MAX_SPC						(64L)
#define MAX_FILE_SECTORS			(4L * 1024 * 1024 - MAX_SPC)
#define MAX_FILE_CLUSTERS( SPC ) (MAX_FILE_SECTORS / (SPC))
#define MAX_FILE_SIZE( SPC )		(MAX_FILE_CLUSTERS(SPC) * 512 * (SPC))
#define FILE_2G						(2L*1024*1024*1024)

//  BT 4-17-02
//
//  This code was removed for the since it relied on knowing the
//  freespace on the disk which is not necessary since the cluster
//  allocations will fail if there is no more free space.
//
#if 0
// Given size in bytes, return the maximimum free space value, such that
// lCurrentFileSize + FreeSpace <= MAX_FILE_SIZE for the current SPC value.
// Assume all sectors are 512 bytes if free space > 2G matters.
//
long GetMaxFreeSpaceForFile(long lCurrentFileSize, ulong bpc_mask )
{
	long lFreeSectors, lFileSectors;
	ulong ulUnusedFileBytes;


	ENTER_FUNCTION();

	//  Determine the number of free sectors
	//
#if USING_FAT32
	if(IsFat32Drive(sda_dpbptr))
	{
		lFreeSectors = Fat32FreeClusters();
	}
	else
	{
		lFreeSectors = sda_dpbptr->freeclus;
	}
#else
	lFreeSectors = sda_dpbptr->freeclus;
#endif
	lFreeSectors = _lshlu( lFreeSectors, sda_dpbptr->shift);

	//  Determine the number of unused bytes allocated to the file
	//
	ulUnusedFileBytes = lCurrentFileSize & bpc_mask;
	if( ulUnusedFileBytes )
	{
		//  Only round up if necessary
		//
		ulUnusedFileBytes = (bpc_mask+1) - ulUnusedFileBytes;
	}

	//  Determine the number of sectors allocated to the file
	//
	lCurrentFileSize += (long)ulUnusedFileBytes;
	lFileSectors = _ldivu( lCurrentFileSize, sda_dpbptr->bytes_per_sec);

	//  Determine if we exceeded the maximum file size and if so
	//  clip the result
	//
	if(lFreeSectors+lFileSectors > MAX_FILE_SECTORS)
	{
		lFreeSectors = MAX_FILE_SECTORS - lFileSectors;
	}

	EXIT_FUNCTION();

	//  Return the number of bytes which can be written to the file
	//
	return (_lmulu(lFreeSectors, sda_dpbptr->bytes_per_sec) +
			ulUnusedFileBytes);
}
#endif


/*
	Do the read/write loop until we are done.  Uses the disk
	buffering scheme.

	Sets sda_rdwrlen to actual length read/written.  Returns
	TRUE if worked, FALSE if failed.
*/
int PASCAL file_rdwr(bool wr)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar far * sda_rdwrbuff;
	extern unsigned sda_rdwrlen;
	extern struct LOL lol;
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;
	extern LPBUFLIST sda_cur_buf;
	extern uchar  CritIgnoreOK;

	/*
		[ locals ]
	*/
	unsigned bytes;		/* bytes read/written */
	unsigned modulo;
	unsigned gotlen;
	long    sector;
	unsigned bps;			/* bytes per sector */
	unsigned shiftCount; // sda_dpbptr->shift
	unsigned leftover;	/* lsec % spc */
	unsigned maxcont;		/* maximum contigous sectors */
	unsigned rdwrlen;
	uchar far * rdwrbuff;
	bool     fPastOldEndOfFile;	// True if expanding the file
	ulong		bpc_mask;
	unsigned tmp;
	uchar	   bps_shift;
	long		lFileSize;
	long		lFileOffset;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* locals for optimization and easy typing */
	bps        = sda_dpbptr->bytes_per_sec;
	shiftCount = sda_dpbptr->shift;

	/* wonders of exponents, revisited.

		offset &  (bps -1)		==  offset % bps
		offset >> bps_shift		==  offset / bps
		secs	 << bps_shift 		==  sec	  * bps

		offset &  cluster_mask  ==  offset % bpc
		offset >> cluster_shift ==  offset / bpc
		clus	 << cluster_shift ==  clus   * bpc

	*/
	bpc_mask   = bps;
	bpc_mask   =_lshlu(bpc_mask, shiftCount);
	bpc_mask--;

	tmp       = bps >> 1;
	bps_shift = 0;
	do
	{
		bps_shift++;
	} while( tmp >>= 1 );


// (de) bpc result is too big to use, causes intermediate overflow in later
// calculations so have to split the multiplication up and divide in between...
//	bpc = bps << sda_dpbptr->shift;


	lFileSize			= sda_sftptr->size;
	lFileOffset       = sda_sftptr->offset;
//  BT 4-17-02
//
//  This code was removed for the since it relied on knowing the
//  freespace on the disk which is not necessary since the cluster
//  allocations will fail if there is no more free space.
//
//  The #else clause was added to cap the file size at 2Gb minus a
//  cluster which was also handled by GetMaxFreeSpaceForFile()
//
#if 0
	lFreeSpaceForFile = GetMaxFreeSpaceForFile(lFileSize,bpc_mask);

	lExtraBytes       = lFileOffset - (lFileSize + lFreeSpaceForFile);
	if(lExtraBytes > 0)
	{
		sda_rdwrlen = 0;
		EXIT_FUNCTION();
		return FALSE;		// file was not "dirtied" by this write
	}

	// lExtraByte = (rdwr_len + file_offset) - (file_size + free_space)
	// if lExtraBytes is positive then it represents the bytes by which
	// the rd/wr exceeds the size of the disk (as limited by the free space
	// and max file size). So... reduce the rd/wr amount by that much.
	// (Note: this computation covers, allocation, disk extent and file
	// size limits simulataneously).
	//
	lExtraBytes += sda_rdwrlen;
	if(lExtraBytes > 0)
		sda_rdwrlen -= (unsigned)lExtraBytes;

#else
	if( (FILE_2G - (bpc_mask+1) - sda_rdwrlen) < lFileOffset )
	{
		//  The maximum file size has been exceeded so reduce the
		//  read/write size
		//
		sda_rdwrlen = (unsigned)(lFileOffset - (FILE_2G -
				(bpc_mask+1) - sda_rdwrlen));
	}
#endif


	// if we are writing, we may need to get extra clusters
	// Try to allocate enough clusters for this write, if we
	// need to expand the file. If we can't allocate enough space,
	// sda_rdwrlen may have been shortened due to lack of disk space.
	// We may also truncate or expand the file if rdwrlen is 0.
	//
	if(wr)
	{
		sda_rdwrlen = AdjustFileSize();
	}

	/*
		These MUST be local (not in SDA) in case of INT 24H Retry
		on INT 24H, the user variables are reread (ie sda_rdwrlen)
	   and those values may have been changed internally
	*/
	rdwrlen  = sda_rdwrlen;
	rdwrbuff = sda_rdwrbuff;

	// Always allow Ignore option when I24 generated by file I/O.
	//
	CritIgnoreOK = CRIT_IGNORE;

	/* loop as many times as we need to read/write the bytes */
	bytes = 0;
	while(!error_no && bytes < rdwrlen)
	{
		// SPR 1638 (de) If we are writing past the current file extent
		// then there is no need to read in sectors off the disk before
		// writing to them.
		//
		fPastOldEndOfFile = (lFileOffset >= lFileSize);

		/* determine physical sector.  may get extra sectors from disk */
		gotlen = rdwrlen - bytes;
		sector = where_in_file(sda_sftptr->offset);

		/* there is an error (FAT does not agree with file size) */
		if(sector == -1L)
			break;

		/* calculate offset within that sector */
		modulo = (unsigned)sda_sftptr->offset & (bps - 1);

		/* read small portions through the buffering scheme */
		if(modulo || gotlen < bps)
		{
			/* optimization: don't get a new sector if we don't have to */
			if(wr && !modulo && fPastOldEndOfFile)
				buf_getnoread(sector);
			else
				buf_getsector(sector);

			// if there is a critical error reading the above sector
			// and the user Fails the call, we'll exit now. DON'T want
			// to mark as modified a buffer that contains trash.
			//
			if(error_no)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

			gotlen = min(gotlen, bps-modulo);
			if(wr)
			{
				fmemcpy(sda_cur_buf->buffer+modulo, rdwrbuff, gotlen);
				buf_set_modified(sda_cur_buf);

				// spr 1638 (de) make sure data held ONLY in buffers is flushed.
				// re: i24test
				//
				sda_sftptr->ioctl &= ~IO_NOTWRITTEN;
			}
			else
				fmemcpy(rdwrbuff, sda_cur_buf->buffer+modulo, gotlen);
		}

		/* read/write large portions with no buffering */
		else
		{
			/* determine logical sector % sectors/cluster (leftover) */
			/* logical cluster */

			leftover = (unsigned)_lshru((sda_sftptr->offset & bpc_mask), bps_shift);

			/* count the contiguous sectors at this offset */
#if USING_FAT32
			maxcont = contiguous_sectors(sda_sftptr->ulAbsCluster,
							leftover, (rdwrlen-bytes) >> bps_shift);
#else
			maxcont = contiguous_sectors(sda_sftptr->absclus,
							leftover, (rdwrlen-bytes) >> bps_shift);
#endif

			/* get buffering scheme out of the loop */
			buf_flush_and_invalidate(sector, maxcont);

			/* read/write as many as we need/can */
			abs_disk_rdwr(!wr, sector, maxcont, rdwrbuff);
			gotlen = maxcont << bps_shift;
		}

		/* go around again */
		sda_sftptr->offset += gotlen;
		bytes  += gotlen;
		rdwrbuff += gotlen;
	}

// SPR 1638 (de) if were to again do pessimistic commits, we would only
// update the file sizes and offsets when all device calls had worked.
//
//	if((sda_sftptr->offset > sda_sftptr->size) && (bytes > 0))
//		sda_sftptr->size = sda_sftptr->offset;

	// return how many bytes we read/wrote
	//
	sda_rdwrlen = bytes;

	EXIT_FUNCTION();

	return TRUE;
}


/*
	This function does 2 jobs.  I know, terrible cohesion.

	1) Make sure there are enough clusters for the write
	2) If user wants to truncate, free up the file FAT list
      (in this special case, len will be 0).

	Returns the # of bytes possible to write (may be 0).
*/
unsigned	PASCAL AdjustFileSize(void)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;
	extern unsigned sda_rdwrlen;


	/*
		[ locals ]
	*/
	CLUSTER  start;	 							/* starting cluster of file */
	CLUSTER  cluster;
#if LSEEK_ENHANCEMENT
	CLUSTER  previous_cluster;
#endif  // #if LSEEK_ENHANCEMENT
	NCLUSTER clusleft, nClustersWanted;
	unsigned bps;									/* bytes per sector */
	long offset;
	bool gotstart = FALSE;
	unsigned	len;
	NCLUSTER LogicalCluster;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	len 	  = sda_rdwrlen;

	/* calculate some items for clarity */
#if USING_FAT32
	start  = sda_sftptr->ulStartingCluster;
#else
	start  = sda_sftptr->start;
#endif
	offset = sda_sftptr->offset + len;
	bps	 = sda_dpbptr->bytes_per_sec;

// (de) bpc result is too big to use, causes intermediate overflow in later
// calculations so have to split the multiplication up and divide in between...
//	bpc    = bps << sda_dpbptr->shift


	// are we inside the allocated length of the file?
	//
	if(offset - (long)sda_sftptr->size <= 0)
	{
		//	yes, if (sda_rdwrlen+sda_sftptr->offset - sda_sftptr->size  <= 0)
		// (len+offset) is within the allocated extend of the the file.
		// A write of 0 bytes is a special case which sets the file length
		// the the current file position. Truncate the file IFF there is
		// already something to deallocate.
		//
		if(len == 0)
		{
			//  if offset - size = 0 then the file is already truncated to
			//  its current length (which may be zero)
			//
			if(offset == (long)sda_sftptr->size)
			{
				EXIT_FUNCTION();
				return 0;
			}

			if(offset > 0)
			{
				where_in_file(offset - 1);

#if USING_FAT32
				start = get_fat_entry(sda_sftptr->ulAbsCluster);
//				if( start != END_OF_FAT )
				{
					set_fat_entry(sda_sftptr->ulAbsCluster, END_OF_FAT);
				}
#else
				start = get_fat_entry(sda_sftptr->absclus);
//				if( start != END_OF_FAT )
				{
					set_fat_entry(sda_sftptr->absclus, END_OF_FAT);
				}
#endif
			}
			else
			{
#if USING_FAT32
				sda_sftptr->ulStartingCluster = 0;
#else
				sda_sftptr->start = 0;
#endif
			}
			sda_sftptr->size = sda_sftptr->offset;
//			if( start != END_OF_FAT )
			{
				free_the_fat(start);
			}


			//  BT 6-12-02
			//
			//  MS does not commit the FAT for file truncation until
			//  a file close
			//
			//  buf_flush_all();


			EXIT_FUNCTION();
			return len;
		}
		else
		{
			EXIT_FUNCTION();
			return len; // not truncating or expanding, just updating
		}
	}
	else 	//	sda_rdwrlen+sda_sftptr->offset - sda_sftptr->size  > 0
	{
		/* calculate logical/physical cluster for the end of the read/write */
		clusleft = 0;
		if(offset)
		{
			/* since we are writing, file must have at least 1 cluster */
			if(start == 0)
			{
				start = alloc_next_free();
				if(start == END_OF_FAT)
				{
					EXIT_FUNCTION();
					return 0;
				}

				set_fat_entry( start, END_OF_FAT );

#if USING_FAT32
				sda_sftptr->ulStartingCluster = start;
				sda_sftptr->uRelClusterLow    = 0;
				sda_sftptr->uRelClusterHigh   = 0;
				sda_sftptr->ulAbsCluster      = start;
#if LSEEK_ENHANCEMENT
            sda_sftptr->uRDlseekHigh      = HIWORD(start);
            sda_sftptr->uRDlseekLow       = LOWORD(start);
#endif  // #if LSEEK_ENHANCEMENT
#else
				sda_sftptr->start   				= start;
				sda_sftptr->relclus 				= 0;
				sda_sftptr->absclus 				= start;
#endif
				sda_sftptr->ioctl	&= (~IO_NOTWRITTEN);
				gotstart           = TRUE;
			}

			//  Assert that the starting cluster is allocated
			//
			ASSERT(get_fat_entry(start) != 0);

			clusleft = (NCLUSTER)_lshru(_ldivu(offset-1,bps),sda_dpbptr->shift);
		}

		// determine cluster number from relative cluster number
		// in the FAT chain
		//
		LogicalCluster = clusleft;
		cluster = file_l2pclus(&clusleft);
		LogicalCluster -= clusleft;

#if USING_FAT32
		sda_sftptr->uRelClusterLow    = LOWORD(LogicalCluster);
		sda_sftptr->uRelClusterHigh   = HIWORD(LogicalCluster);
		sda_sftptr->ulAbsCluster      = cluster;
#else
		sda_sftptr->relclus 				= LogicalCluster;
		sda_sftptr->absclus 				= cluster;
#endif
		nClustersWanted = clusleft;

#if LSEEK_ENHANCEMENT
      previous_cluster = cluster;
#endif  // #if LSEEK_ENHANCEMENT
		if(clusleft && (get_new_clusters(&cluster, clusleft) != clusleft))
		{
			// for new files, allocate all or nothing, free the whole chain
			// if we didn't get all we wanted...
			//
			if(gotstart)
			{
#if USING_FAT32
				free_the_fat(sda_sftptr->ulStartingCluster);
				sda_sftptr->ulStartingCluster = 0;
#else
				free_the_fat(sda_sftptr->start);
				sda_sftptr->start 	= 0;
#endif
				sda_sftptr->size 		= 0;

#if 0
            // Don't reset the file pointer on a failed write...
            //
				sda_sftptr->offset	= 0;
#endif

				len						= 0;
			}
			else
			{
				// The write will fill the disk and the file will end on a
				// cluster boundary. Not all bytes will be written to disk.
				//
				nClustersWanted  -= clusleft;		// clusters we can write
				clusleft = (NCLUSTER)(_lmulu(_lshlu(1,sda_dpbptr->shift),bps)-1);
				if(sda_sftptr->size & clusleft)	// is there a partial cluster?
					++nClustersWanted;
				sda_sftptr->size &= ~clusleft;	// mask off partial, add full clusters
				sda_sftptr->size += _lshlu(_lmulu((nClustersWanted),bps),sda_dpbptr->shift);
			}
			EXIT_FUNCTION();
			return len;
		}
#if 0
      // TD 4152 4/17/03
      //
      // This code was setting the LSEEK variables after get_new_clusters()
      // returned.  Two problems with this:
      //  1) the SFT pointers were not moved, and the LSEEK variables
      //     are tied to those pointers (sda_sftptr, in this case)
      //  2) the value of the returned cluster was used, even if the
      //     get_new_clusters() function allocated multiple clusters
      // I don't believe it is appropriate to adjust the LSEEK variables
      // if the file pointers don't move.  They will be updated on
      // future reads or writes.
      //
#if USING_FAT32 && LSEEK_ENHANCEMENT
      else
      {
         // New cluster allocation needs to update the LSEEK
         // fields, but only if there's a gap.
         // Use previous_cluster to see if there was a gap or a wrap.
         //
         if( (previous_cluster+1L) != cluster )
         {
            sda_sftptr->uRDlseekHigh = HIWORD(cluster);
            sda_sftptr->uRDlseekLow  = LOWORD(cluster);
         }
      }
#endif  // #if USING_FAT32 && LSEEK_ENHANCEMENT
#endif  // #if 0

		// SPR 1638 (de) Optimistically record the expanded file size
		// like MS-DOS does. See I24TEST.
		//
  		sda_sftptr->size = offset;
		EXIT_FUNCTION();
		return len;
	}

	EXIT_FUNCTION();
}


/*
	This function does 2 jobs:  Aggh! another bad cohesion
   example.

	1) Determine physical sector of this offset
	2) Update "relclus" and "absclus" fields in SFT.

	Returns the sector #.  Can only fail if FAT and file
	size do not agree.  Returns -1L on failure.
*/
long PASCAL where_in_file(ulong offset)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;

	/*
		[ locals ]
	*/
	NCLUSTER lclus;
	CLUSTER  cluster;
	NCLUSTER clusleft;
	unsigned bps;
	unsigned leftover;
	long sector;
	unsigned tmp;
	uchar bpc_shift, bps_shift;
#if USING_FAT32 && LSEEK_ENHANCEMENT
   NCLUSTER prev_lclus;
   CLUSTER contiguous_start;
   CLUSTER potential_cluster;
#endif  // #if USING_FAT32 && LSEEK_ENHANCEMENT


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* determine bytes per sector/cluster */
	bps = sda_dpbptr->bytes_per_sec;

	tmp       = bps >> 1;
	bps_shift = 0;
	do
	{
		bps_shift++;
	} while( tmp >>= 1 );
	bpc_shift = bps_shift + sda_dpbptr->shift;

// (de) bpc result is too big to use, causes intermediate overflow in later
// calculations so have to split the multiplication up and divide in between...
//	bpc = bps << sda_dpbptr->shift;

	/* determine logical and physical cluster for this offset */
	lclus = clusleft = (NCLUSTER)_lshru(offset, bpc_shift);

#if USING_FAT32 && LSEEK_ENHANCEMENT
   prev_lclus = MK_ULONG(sda_sftptr->uRelClusterHigh, sda_sftptr->uRelClusterLow);

   // This IF fails if we are not doing an LSEEK at all
   //
   if( lclus <= prev_lclus )
   {
      // Store the value of the dword subtraction in potential_lclus
      //
      potential_cluster = prev_lclus - lclus;

      // This IF fails if the difference in logical clusters is large
      // and the Absolute cluster is small -- a "wrap" situation
      //
      if( potential_cluster < sda_sftptr->ulAbsCluster )
      {
         // since we passed all of those, calculate the potential cluster
         // and the start of the contiguous region, for the next compare
         //
         potential_cluster = sda_sftptr->ulAbsCluster - potential_cluster;
         contiguous_start = MK_ULONG(sda_sftptr->uRDlseekHigh, sda_sftptr->uRDlseekLow);
      }
      else
      {
         potential_cluster = 0;
      }
   }
   else
   {
      potential_cluster = 0;
   }

// Assuming we are in a contiguous region, we've calculated the new
// potential_cluster using simple math.  If conditions are correct,
// we'll use the value, otherwise we make a call to file_l2pclus().
//
   if( (potential_cluster != 0) && (potential_cluster >= contiguous_start) )
   {
      cluster = potential_cluster;
      clusleft = 0;
   }
   else
#endif  // #if USING_FAT32 && LSEEK_ENHANCEMENT
   {
   	cluster = file_l2pclus(&clusleft);		/* may allocate clusters */
   }

	/* the FAT and file size do not agree - return error */
	if(!cluster || clusleft)
	{
		EXIT_FUNCTION();
		return -1L;
	}

	/* convert absolute cluster into absolute sector */
	offset  -= _lshlu(lclus, bpc_shift);
	leftover = (unsigned)_lshru(offset, bps_shift);

	sector = clus_to_sec(cluster) + leftover;

	/* optimization: update SFT "last access" variables */
#if USING_FAT32
	sda_sftptr->uRelClusterLow  = LOWORD(lclus);
	sda_sftptr->uRelClusterHigh = HIWORD(lclus);
	sda_sftptr->ulAbsCluster    = cluster;
#else
	sda_sftptr->relclus = lclus;
	sda_sftptr->absclus = cluster;
#endif

	EXIT_FUNCTION();

	return sector;
}


/*
	Given a logical cluster, return a physical cluster.
	This function operates the same as "lclus_to_pclus",
	except that is understands the current SFT and the
	shortcuts that provides.

	Returns the physical cluster.  (and clusters left if
	the end of the file is reached).
*/
CLUSTER PASCAL file_l2pclus(NCLUSTER *clusleft)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;

	/*
		[ locals ]
	*/
	CLUSTER  start;
	CLUSTER  cluster;
#if USING_FAT32
	NCLUSTER relclus;
#endif

	ENTER_FUNCTION();

	// make sure pointer is OK
	ASSERT(clusleft);

	/* Fat32 file system */
#if USING_FAT32
	start = sda_sftptr->ulStartingCluster;

#if LSEEK_ENHANCEMENT
   if( !*clusleft )
   {
      sda_sftptr->uRDlseekHigh = HIWORD(start);
      sda_sftptr->uRDlseekLow  = LOWORD(start);
   }
#endif  // #if LSEEK_ENHANCEMENT

	relclus = MK_ULONG(sda_sftptr->uRelClusterHigh, sda_sftptr->uRelClusterLow);

   /* use "last access" variables to speed things up if possible */
   if( sda_sftptr->ulAbsCluster && ( relclus <= *clusleft ) )
   {
      *clusleft -= relclus;
      start = sda_sftptr->ulAbsCluster;
   }

	/* Fat16 only access */
#else
	start = sda_sftptr->start;

	/* use "last access" variables to speed things up if possible */
	if(sda_sftptr->absclus && sda_sftptr->relclus <= *clusleft)
	{
		*clusleft -= sda_sftptr->relclus;
		start = sda_sftptr->absclus;
	}
#endif

	/* convert logical cluster to physical cluster */
	cluster = lclus_to_pclus(start, clusleft);

	EXIT_FUNCTION();

	return cluster;
}


/*
	Given the maximum contiguous sectors desired, the
	starting cluster, and starting sector within that
	cluster, return the # of sectors that are actually
	contiguous.

	Always returns at least 1.
*/
unsigned PASCAL contiguous_sectors(CLUSTER start, unsigned leftover,
																	unsigned maxcontsec)
{
	/*
		[ externals ]
	*/
	extern LPDPB sda_dpbptr;

	/*
		[ locals ]
	*/
	unsigned contclus;	/* file operation <64K so unsigned # clusters OK */
	unsigned contsec;
	unsigned spc;


	ENTER_FUNCTION();

	/* we must have a starting cluster */
	ASSERT(start);

	/* calculate maximum contiguous clusters */
	spc = 1 << sda_dpbptr->shift;
	contclus = (maxcontsec + leftover + spc - 1) / spc;
	contclus = count_cont_clus(start, contclus);

	/* calculate maximum contiguous sectors */
	contsec = (contclus << sda_dpbptr->shift) - leftover;
	if(contsec > maxcontsec)
		contsec = maxcontsec;

	EXIT_FUNCTION();

	return contsec;
}


/*
	Seek within a file or device.  This merely sets the
	file position in the SFT.  No disk I/O can happen.
*/
ulong	PASCAL file_seek(ulong pos, uchar mode)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern int error_no;
	extern ulong remote_seek(ulong pos);

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	switch(mode)
	{
		case SK_BEG:
			sda_sftptr->offset  = pos;
			break;
		case SK_REL:
			sda_sftptr->offset += pos;
			break;
		case SK_END:
			if(sda_sftptr->ioctl & IO_NETWORK)
			{
				// Call IFS: a remote writer process could have changed
				// the file length w/out our SFT being updated.
				//
				// NOTE: we call this more often than MSDOS but if we
				//       only call this when the file is READ_ONLY+DENY_READ
				//       (as appeared to be correct) then MSDOS calls this
				//       function when we don't.
				//
				sda_sftptr->offset = remote_seek(pos);
				break;
			}
			sda_sftptr->offset = sda_sftptr->size+pos;
			break;
		default:
			error_no = ERR_BADFUNC;
			break;
	}

	EXIT_FUNCTION();

	return sda_sftptr->offset;
}


/*
	-------------------------------------------------------

							Handle Management

	-------------------------------------------------------
*/

/*
	DOS call 45H

	Duplicate an existing handle.  Pass the value to be
	placed in that duplicate.

	Sets "error_no" on error.  Returns the handle otherwise.
*/
unsigned PASCAL dup_handle(uchar val)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;
	extern LPSFT sda_sftptr;

	unsigned handle;

	ENTER_FUNCTION();

	handle = get_new_handle();
	if(handle == 0xFFFF)
		error_no = ERR_NOHANDLE;
	else
	{
		set_handle_val(handle, val);
		++(sda_sftptr->handles);
		if(sda_chardev)
			dev_openclose(TRUE);
	}

	EXIT_FUNCTION();

	return handle;
}


/*
	DOS call 46H

	Given a handle, and a duplicate, make the duplicate
	point to the original handle.  We already know that the
	handle is valid, but the duplicate may not be a legal
	handle (it's ok if the duplicate is closed).

	Sets error_no on error.
*/
void	PASCAL file_force_dup(unsigned handle,unsigned dup)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;
	extern LPSFT sda_sftptr;


	/*
		[ locals ]
	*/
	unsigned hval;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* check for invalid dup handle */
	hval = get_handle_val(dup);
	if(dup == handle || hval==0xFFFF)
	{
		error_no = ERR_BADHANDLE;
		EXIT_FUNCTION();
		return;
	}

	/* if dup handle was open, close it */
	if(hval != 0xFF && handle_to_sda(dup))
	{
		file_close();
		decr_handle_count();
	}

	/* point dup handle to original */
	hval = get_handle_val(handle);
	set_handle_val((uchar)dup,(uchar)hval);

	/* and increment the # of handles on the SFT */
	handle_to_sda(handle);
	++(sda_sftptr->handles);
	if(sda_chardev)
		dev_openclose(TRUE);

	EXIT_FUNCTION();
}


/*
	Returns the SFT # stored in the handle.  This value can
	be anywhere between 0x00 - 0xFF.  If the handle is past
	the end of the handle table, this function returns
	0xFFFF to indicate an error.

	Returns SFT # (handle val) or 0xFFFF if invalid.
*/
unsigned	PASCAL get_handle_val(unsigned handle)
{
	/*
		[ externals ]
	*/
	extern unsigned current_psp;


	/*
		[ locals ]
	*/
	unsigned hcount;
	uchar far * htbl;
	struct PSP far * psp;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* the PSP tells us where we can find the handle array */
	psp = MK_FP(current_psp,0);
	htbl = psp->htbl_addr;
	hcount = psp->hcount;

	/* make sure the handle is less than the handle count */
	if(handle >= hcount)
   {
      EXIT_FUNCTION();
		return 0xFFFF;
   }

	EXIT_FUNCTION();

	/* return the SFT # from the handle table */
	return (unsigned)(htbl[handle]);
}


/*
	Given a handle # and an SFT #, set the contents of the
	handle to point to the given SFT.

	Returns the handle # if worked, 0xFFFF if failed.
*/
unsigned PASCAL set_handle_val(uchar handle, uchar val)
{
	/*
		[ externals ]
	*/
	extern unsigned current_psp;


	/*
		[ locals ]
	*/
	unsigned hcount;
	uchar far * htbl;
	struct PSP far *psp;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* the PSP tells us where we can find the handle array */
	psp    = MK_FP(current_psp,0);
	htbl   = psp->htbl_addr;
	hcount = psp->hcount;

	/* check for invalid handle */
	if(handle >= hcount)
	{
		EXIT_FUNCTION();
		return 0xFFFF;
	}

	/* set the value and return */
	htbl[handle] = val;

	EXIT_FUNCTION();

	return (unsigned)handle;
}


/*
	Look for an unused handle.  Searches the handle table.

	Returns 0xFFFF if failed.  Returns the handle # if
	worked.
*/
unsigned get_new_handle(void)
{
	/*
		[ locals ]
	*/
	uchar handle;
	unsigned val;

	ENTER_FUNCTION();

	handle = 0;
	while(1)
	{
		val = get_handle_val(handle);
		if(val == 0xFFFF)
			break;

		/* the handle is free */
		if(val == 0xFF)
		{
			EXIT_FUNCTION();
			return handle;
		}

		++handle;
	}

	EXIT_FUNCTION();

	/* return error */
	return 0xFFFF;
}


/*
	Look for an unused SFT.  Place the pointer to it into
	the SDA.  This can be used by both "handle" and "FCB"
	routines, depending on the value of the "sda_callflags".

	Returns TRUE and sets "sda_sft_h" if worked.  Also sets
	"sda_sftptr."  Returns FALSE if failed.
*/
bool	get_new_sft(void)
{
	/*
		[ externals ]
	*/
	extern uchar sda_sft_h;
	extern LPSFT sda_sftptr;

	ENTER_FUNCTION();

	/* check each sft to see if it's unused */
	sda_sft_h = 0;
	while(1)
	{
		sda_sftptr = h_to_sft(sda_sft_h);
		if((void *)sda_sftptr == NULLF)
			break;

		/* we have found an available SFT, zero it out */
		if(sda_sftptr->handles == 0)
		{
			fmemset(sda_sftptr,0,sizeof(SFT));
			EXIT_FUNCTION();
			return TRUE;
		}

		/* try the next SFT */
		++sda_sft_h;
	}

	EXIT_FUNCTION();

	/* none found that are free */
	return FALSE;
}


/*
	DOS call 67H

	Make a new, larger, handle table.  Preserve the
	currently open handles.

	Sets error_no if there is an error (ERR_NOMEM).
*/
void	PASCAL set_handle_count(unsigned count)
{
	/*
		[ externals ]
	*/
	extern unsigned current_psp;
	extern int error_no;
	extern unsigned char booting;

	/*
		[ locals ]
	*/
	struct PSP far * psp;
	unsigned hcount;
	unsigned seg;

	ENTER_FUNCTION();

	/* user better NOT ask for too many handles */
	ASSERT(count <= 0xFFF0);

	/* get the old set of handles */
	psp = MK_FP(current_psp,0);
	hcount = psp->hcount;

	/* we already have enough handles, we are done */
	if(count <= hcount)
	{
		EXIT_FUNCTION();
		return;
	}

	/* get the needed memory */
	seg = mem_alloc((count+15)/16);
	if(error_no)
	{
		EXIT_FUNCTION();
		return;
	}

	/* copy the old handles... */
	fmemset(MK_FP(seg,0),0xFF,count);
	fmemcpy(MK_FP(seg,0),psp->htbl_addr,hcount);

	/* ...and fill in the PSP */
	psp->htbl_addr = MK_FP(seg,0);
	psp->hcount = count;

	EXIT_FUNCTION();
}
