//
//   $Id: DOSIOCTL.C 1.11 2003/06/05 21:57:46Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DOSIOCTL.C  created 09/25/91  by Drew Gislason

   Copyright (c) 1991-2002 Datalight, Inc.
	All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains IOCTL calls.  These calls are too
	varied to categorize, but they all talk to devices.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 9

/*
	DOS call 44H

	Device Driver Control.  This DOS call contains many
	subfunctions which allow the user to manipulate both
	character and disk devices in a variety of ways.
*/
void ioctl(uchar al, unsigned cx)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;
	extern unsigned sda_handle;
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;
	extern LPDEV_HDR sda_devptr;
	extern struct dev_req_hdr sda_reqhdr;
	extern LPCDS sda_cdsptr;
	extern int share_retry;
	extern int share_delay;
	extern uchar far * DTA;
	extern int int2f_cx;
   extern char FAT_type;

	/*
		[ locals ]
	*/
	static char diskcalls[] = { 4,5,8,9,0xD,0xE,0xF,0x11, 0 };
	unsigned ioctl;
	unsigned * up;
	unsigned busy;
	bool ok2call;

	ENTER_FUNCTION();

	/* set up the request header */
	memset(&sda_reqhdr, 0, sizeof(sda_reqhdr));

	/* is this a block device function? */
	if(al && strchr(diskcalls,al))
	{
		/* is this a valid drive? */
		if(!defdrive_to_sda(R_BL))
			error_no = ERR_NOTDRIVE;
		if(sda_devptr)
		{
			ioctl = sda_devptr->attr;
			sda_reqhdr.unit = sda_dpbptr->unit;
		}
		else
			ioctl = 0;
	}

	/* is this a character device? */
	else if(al != 0x0B)
	{
		if(!handle_to_sda(R_BX))
			error_no = ERR_BADHANDLE;
		ioctl = sda_sftptr->ioctl;
		sda_reqhdr.unit = 0;			/* Sony CD ROM drive wants 0 here */
	}

	/* we've encounter a problem without even calling the device */
	if(error_no)
	{
		EXIT_FUNCTION();
		return;
	}

	/* perform the ioctl subfunction */
	switch(al)
	{
		/* (Char) Get Device Information */
		case 0x00:
			if(ioctl & IO_DEVICE)
				ioctl |= DEVA_CHAR;	// 8000H
			else
				//	only B6..B0 returned for file (MSDOS prog ref pg 286)
				ioctl &= 0x7F;
			R_AX = R_DX = ioctl;
			break;

		/* (Char) Set Device Information */
		case 0x01:
			if(ioctl&0x80)
			{
				ioctl &= 0xFF80;
				ioctl |= (R_DL & 0x7F);
				sda_sftptr->ioctl = ioctl;
			}
			else
				error_no = ERR_BADFUNC;
			break;

		/* Character IOCTL read/write */
		case 0x02:
		case 0x03:
			ioctl = sda_devptr->attr;
			/* fall thru... */

		/* Block IOCTL read/write */
		case 0x04:
		case 0x05:

			/* check if this device is capable of IOCTL */
			if(!(ioctl & DEVA_IOCTL))
			{
				error_no = ERR_BADFUNC;
				break;
			}

			/* set up the request header */
			sda_reqhdr.command = (al&1) ? DEV_IOWRITE : DEV_IOREAD;
			sda_reqhdr.count   = cx;
			sda_reqhdr.trans   = MK_FP(R_DS,R_DX);

			/* call the device */
			dev_call(&sda_reqhdr, sda_devptr);

			/* get the returned status */
			if(sda_reqhdr.status & DEVE_ERROR)
				error_no = ERR_BADFUNC;

			R_AX = sda_reqhdr.count;
			break;

		/* Get Input Status */
		case 0x06:

			// by default...
			// return AL= File/device "busy" and AH = pending char = ^Z
			//
			R_AX = 0x1A00;

			// if file/device is remote...
			//
			if(sda_sftptr->ioctl & IO_NETWORK)
			{
				// and not write only then do a non-destructive read
				// of the remote file/device
				//
				if(!(sda_sftptr->rwmode & O_WRONLY))
				{ ulong old_offset;
				  uchar far * old_DTA, buffer[1];

					old_offset = sda_sftptr->offset; // save file ptr
					old_DTA = DTA;							// and user's DTA
					DTA = (uchar far *)(buffer);		// set up scratch DTA

					// actually read 1 byte from remote
					// (STACK & DS:SI are superfluous arguments to 0x1108)
					generate_int2F(0x08,0,1,sda_sftptr,sda_sftptr);

					DTA = old_DTA;							// restore DTA and
					sda_sftptr->offset = old_offset;	// file ptr

					if(error_no) 			// return IFS error code if any
					{
						EXIT_FUNCTION();
						return;
					}

					if(int2f_cx != 0)		// else if any char was read
					{
						R_AX = ((*buffer << 8) | 0xFF);// return it in ah and
						break;								 // !busy stat in al
					}
				}
				break;
			}

			// if a FAT drive
			//
			if(!(sda_sftptr->ioctl & IO_DEVICE))
			{
				// if not at EOF then read status is OK (but char unknown)
				//
				if(sda_sftptr->size > sda_sftptr->offset)
				{
					R_AX = 0x00FF;
				}
				break;
			}

			// else local device, see if it's busy for reading (nothing pending)
			//
			busy = CheckLocalCharDevBusy(ioctl,DEV_NDREAD);

			// busy (at EOF, Not Ready), !busy (not at EOF, Ready)
			// NOTE: MSDOS v 6.22 returns AH=0 for busy char devices
			//
			if(busy)
			{
				R_AX = 0;	// can't read
			}
			else
			{
				// return "OK to read" in AL and the pending char in AH
				//
				R_AX = (((unsigned char)(sda_reqhdr.status) << 8) | 0xFF);
			}
			break;


		/* Get Output Status */
		case 0x07:

			//	NOTE: MSDOS V 6.22 always returns AH=1 for this call
			// and anything but a busy local char device is reported
			// as OK to write to (even if the disk is full, etc..)
			//
			R_AX = 0x1FF;

			// so only if this is a local device
			//
			if( !(sda_sftptr->ioctl & IO_NETWORK) &&
			     (sda_sftptr->ioctl & IO_DEVICE)     )
			{
				// see if it's busy for writing
				//
				busy = CheckLocalCharDevBusy(ioctl,DEV_OUTSTAT);

				// busy (at EOF, Not Ready), !busy (not at EOF, Ready)
				//
				if(busy)
				{
					R_AL = 0x00;	// can't write
				}
			}
			break;


		/* Block Device Removeable? */
		case 0x08:
			if(ioctl & DEVA_REMOVE)
			{
				sda_reqhdr.command = DEV_REMOVE;
				dev_call(&sda_reqhdr, sda_devptr);
				R_AX = (sda_reqhdr.status & DEVE_BUSY) ? 1 : 0;
			}
			else
				error_no = ERR_BADFUNC;
			break;

		/* (Disk) Block Device Remote? */
		case 0x09:
			//  BT  2-6-01
			//
			//  MS-DOS returns carry and invalid drive for drives that
			//  are not valid.
			//
			if( sda_cdsptr->status & C_VALID )
			{
				if(sda_cdsptr->status & C_NETWORK)
				{
					R_DX = 0x1000;
					break;
				}


				//  BT  2-6-01
				//
				//  MS-DOS changes this bit in the device status for
				//  substed drives.
				//
				if(sda_cdsptr->status & C_SUBST)
				{
					ioctl |= 0x8000;
				}
			}
			else
			{
				error_no = ERR_NOTDRIVE;
			}

			/* fall thru... */

		/* (Char) Handle Remote? */
		case 0x0A:
			R_DX = ioctl;
			break;

		/* Set Sharing Retry Count */
		case 0x0B:
			share_delay = R_CX;
			share_retry = R_DX;
			break;

		/* Generic IOCTL for Handles */
		case 0x0C:
		case 0x10:
			ioctl = sda_devptr->attr;
			/* fall thru... */

		/* Generic IOCTL for Block Devices */
		case 0x0D:
		case 0x11:

#if USING_FAT32
         // Special handling for a LOCK call
         //
         if( ( cx == 0x084A ) || ( cx == 0x084B ) )
         {
            // Only pass the LOCK call through if the media is FAT32
            //
            if( FAT_type != FS_FAT32 )
            {
               break;
            }
         }
#endif
			/* determine if this driver supports this call */
			ok2call = FALSE;
			if(al&0x10)				/* funcs 10H and 11H */
			{
				if(ioctl & DEVA_QUERY)
					ok2call = TRUE;
			}
			else if(ioctl & DEVA_GENIO)
				ok2call = TRUE;

			/* if driver supports it, make the call */
			if(ok2call)
			{
				/* set up the request header */
				sda_reqhdr.command = (al&0x10) ? DEV_QUERY : DEV_GENIO;

				/* the DS:DX buffer is pointed to in a different place */
				up = (unsigned *)(&sda_reqhdr.media);
				*up = (cx >> 8) | (cx << 8);	/* 0DH */
				*(up+1) = R_SI;					/* 0FH */
				*(up+2) = R_DI;					/* 11H */
				*(up+3) = R_DX;					/* 13H */
				*(up+4) = R_DS;					/* 15H */

				/* call the device */
				dev_call(&sda_reqhdr, sda_devptr);

				// Clean up after successful device call
				if(!(sda_reqhdr.status & DEVE_ERROR))
				{
					// Rebuild the DPB after set_dev_parms call
					// because BPB may have changed
#if USING_FAT32
               if((cx == 0x0840) || (cx == 0x4840))
#else
					if( cx == 0x0840 )
#endif
					{
						// Any non-zero value forces a DPB rebuild
						sda_dpbptr->rebuild = 0xFF;
					}

					// zero AX
					R_AX = 0;
					break;
				}
			}

			// Device doesn't support this call, error out
			// 02/06/96 RME - There may be a case where MS passes
			// an error code of 0Fh back through this interface.
			// No verifiable evidence could be found.
			set_extended_error( ERR_DENIED, ERR_WPROTECT+(uchar)sda_reqhdr.status,
				0x0B04, 2 );
			break;

		/* Get Logical Drive */
		case 0x0E:

		/* Set Logical Drive */
		case 0x0F:
			if(ioctl & DEVA_REMOVE)
			{
				sda_reqhdr.command = (al==0x0E ? DEV_GETDEV : DEV_SETDEV);
				dev_call(&sda_reqhdr, sda_devptr);
				if(!(sda_reqhdr.status & DEVE_ERROR))
				{
					R_AL = sda_reqhdr.unit;
					break;
				}
			}
			error_no = ERR_BADFUNC;
			break;

		default:
			error_no = ERR_BADFUNC;
			break;
	}

	EXIT_FUNCTION();
}

unsigned CheckLocalCharDevBusy(unsigned sft_ioctl, uchar dev_command)
{

	//	[ externals ]
	//
	extern LPDEV_HDR sda_devptr;
	extern struct dev_req_hdr sda_reqhdr;


	//	[ locals ]
	//
	unsigned busy;


	ENTER_FUNCTION();

	// set request header command
	//
	sda_reqhdr.command = dev_command;

	// determine busy status of device
	//
	busy = !(sft_ioctl & IO_NOTEOF);
	dev_call(&sda_reqhdr, sda_devptr);
	if(!(sda_reqhdr.status & DEVE_ERROR))
		busy = (sda_reqhdr.status & DEVE_BUSY);

	EXIT_FUNCTION();

	return busy;
}
