//
//   $Id: DOSMAIN.C 1.17 2002/11/13 17:31:06Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSMAIN.C  created 11/29/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module contains the initialization code for
	ROM-DOS.  Any code that is strictly for initialization
	goes here.  Some day we may figure out a way to elimate
	this code after it has done its job, so that the
	ROM/RAM space it uses can be freed up.

	History:

	02-02-1999 de  SPR#1893

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"

#include "farlfn.h"

#define __FILENUM__ 11

extern char * boot_prog;
extern union EXEPARM boot_parms;
extern unsigned boot_env;
extern unsigned boot_env_size;

// default to version 6
#ifndef VDOS
#define VDOS 6
#endif


/*
	Run Only On Bootup:

	Used only internally, allocate a number of bytes from
	the memory pool.  Also, zero the allocated memory.
	"bytes" may be 0 to create a 0 length MCB.

	Returns the segment of this memory.  If the memory
	cannot be allocated, this halts the machine with an
	error message.
*/
unsigned	PASCAL mem_z_alloc(unsigned bytes, char id)
{
	extern int error_no;
	unsigned seg,paras;

	/* convert bytes to paragraphs */
	paras = bytes/16 + ((bytes & 0xF)?1:0);

	/* allocate the memory */
	error_no = 0;
	seg = mem_alloc(paras);
	PERMASSERT(!error_no);

	/* zero the memory */
	fmemset(MK_FP(seg,0),0,paras*16);

	/* assign the ID */
	((struct MCB far *)MK_FP(seg-1, 0))->init_id = id;

	return seg;
}


/*
	Run Only On Bootup:

	Move the boot environment from the high part of memory
	to low conventional RAM.
*/
void realloc_env(void)
{
	unsigned tmp;
	tmp = mem_z_alloc(boot_env_size, 0);
	fmemcpy(MK_FP(tmp,0), MK_FP(boot_env,0), boot_env_size);
	mem_free(boot_env);
	boot_parms.e.envseg = boot_env = tmp;
}


/*
	Run Only On Bootup:

	Combine all the DOS memory blocks into one.  This keeps
	the MCB list short and memory allocations fast.

	See comments for FirstMCB() (DOSMEM.C).

*/
void	mem_combine(void)
{
	extern struct LOL lol;

	unsigned seg1, seg2, parasInSDBlock;
	struct MCB far * mcb;
	struct MCB far * mcb2;
	struct PSP far * psp;

	// Let the FirstMCB routine know that we have finished with the
	// second MCB chain (the one inside the System Data memory block).
	// Do this by setting the lol.first_mcb.init_id != 0.
	//
	mcb = (struct MCB far *)MK_FP(lol.first_mcb,0);
   mcb->init_id++;

	// Point to the first free block or the last allocated block
	// in the second MCB chain. MS documentation says that TSRs
	// INSTALLed in CONFIG have no environement block but they
	// really have a "pseudo-environment" consisting simply of the
	// program name (argv[0] in C). Despite the warnings in the MS
	// documentation, some programs try to free their environment
	// anyway.  Our old mem_combine scheme assumed that the first
	// free block we find is the start of available memory, which
	// will work so long as we account for the special case of TSRs
	// freeeing the "environment blocks they don't have"
	//
	mcb = MK_FP(lol.first_mcb+1,0);
	do {
		mcb = MK_FP(FP_SEG(mcb)+mcb->para_cnt+1,0);

		// skip over free environment blocks belonging to TSRs
		// that were INSTALL'ed from CONFIG.SYS
		//
		if(mcb->owner == FREEMEM)
		{
			// get actual seg of possible environment block, and assume
			// that it is followed by a PSP (it will be if this is an
			// INSTALL'ed TSR).
			//
			seg1 = 1 + FP_SEG(mcb);
			mcb2 = MK_FP(FP_SEG(mcb)+mcb->para_cnt+1,0);
			psp  = MK_FP(FP_SEG(mcb)+mcb->para_cnt+2,0);

			// make sure that we really are looking at a PSP and
			// that the PSP environment segment is the free block
			// that the mcb ptr points to, if it is then skip to
			// the program block that holds the program's PSP (plus
			// whatever else)
			//
			//  ETR 3919 BT 1-24-02
			//
			//  The code below would not work if the install command
			//  was used to load mscdex.  This TSR frees its environment.
			//  Then ROM-DOS allocated a few paragraphs before executing
			//  command.com.  This meant that the psp->env_seg did not
			//  equal the seg1 computation and thus caused all the memory
			//  past this point to be overwritten.
			//
			//  if((psp->int_20 == 0x20CD) && (psp->env_seg == seg1))
			//
			if( (psp->int_20 == 0x20CD) && (mcb2->owner != FREEMEM) )
			{
				mcb = MK_FP(FP_SEG(mcb)+mcb->para_cnt+1,0);
			}

			// "use" this variable to keep the stupid compiler happy...
			psp = psp;
		}
	} while((mcb->id != MEMSIGZ) && (mcb->owner != FREEMEM));

	// If the block is free, then the we will shrink the SD block
	// down to this block otherwise we must include this block in
	// the size of the SD block
	//
	if(mcb->owner == FREEMEM)
		seg2 = FP_SEG(mcb);
	else
		seg2 = FP_SEG(mcb) + mcb->para_cnt + 1;
	parasInSDBlock = seg2-lol.first_mcb-1;
	seg2--;

	// shrink the SD block to what was actually allocated during CONFIG
	// processing
	//
	seg1 = mem_resize(lol.first_mcb+1, parasInSDBlock);
	PERMASSERT(seg1 == lol.first_mcb+1);

	// update the owner and name fields of lol.first_mcb
	//
	*(unsigned far *)MK_FP(lol.first_mcb,1) = 8;	// system is owner of
	*(long far *)MK_FP(lol.first_mcb,8) = 'SD';	// system data

	// finally, make new IDs for each of the MCBs in the second MCB chain
	//
	while(seg1 < seg2)
	{
		mcb = MK_FP(seg1, 0);
		mcb->id = mcb->init_id;
		mcb->owner = FP_SEG(mcb) + 1;
		seg1 += mcb->para_cnt + 1;
	}

}


/*
	Run Only On Bootup:

	Initialize ROM-DOS to the point of executing the first
	program (COMMAND.COM)
*/
void	main(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char * boot_prog;
	extern char trydoshi;
	extern uchar init_break;
	extern int	init_files;
	extern int	init_buffers;
	extern int	init_fcbs;
	extern char init_lastdrive;
	extern char init_shell[];
	extern char fatal_disk[];
	extern char fatal_notfound[];
#if VDOS >= 5
	extern char warn_nohigh[];
#endif
	extern uchar sda_chardev;
	extern uchar InDOS;
	extern uchar mem_fit;
	extern uchar booting;
	extern uchar CUR_DRIVE;
	extern uchar BREAK_FLAG;
	extern uchar sda_numfcbs;
	extern uchar sda_numsfts;
	extern struct LOL lol;
	extern unsigned sda_callflags;
	extern unsigned int21_ax;
	extern unsigned dos5_flags;
	extern void PASCAL (*prfunc)(uchar c);
	extern unsigned dos_code_seg;
	extern char cfg_name[];
	extern unsigned hma_used;
	extern unsigned codeparas;
	extern unsigned bufparas;
	extern void far * fpFarLFNCall;


	/*
		[ locals ]
	*/
	unsigned codeseg;
	unsigned LFNcodeseg;

	/*
		[ code ]
	*/

	/* let user know we reached main */
	bios_putc('1');

	/* initialize the SDA */
	InDOS = booting = 1;
	init_sda();

	/* create country tables */
	init_country_info();

	/* create the DOS memory pool (the heap) */
	mem_init();

	//	 BT 10-17-01 ETR 3801
	//
	//  It was discovered while adding support to emm386 to relocate
	//  the EBDA and adjust the top of memory to include the
	//  old EBDA memory space that we should not be saving and
	//  restoring the top of memory.
	//
	//  // Remember the Int 12h top of memory, because we will
	//  // be faking it during CONFIG.SYS processing
	//  real_memtop = *(unsigned far *)MK_FP( 0x40, 0x13 );

	/*
		If ROM-DOS is in ROM, record the fact.  If ROM-DOS
		is in RAM, it is at the top of conventional memory.
		Allocate the memory it occupies.
	*/
	if(is_DOS_in_ROM())
		dos5_flags |= DOS5_ROM;
	else
	{
		mem_fit = MEM_LAST;
		codeparas = get_codeparas();
		codeseg = mem_alloc(codeparas);
		ASSERT(!error_no);
		mem_fit = MEM_FIRST;
	}

	/* set defaults in List of Lists from SYSGEN.ASM */
	BREAK_FLAG  = init_break;
	lol.numbufs = MINBUF;
	lol.lastdrive = toupper(init_lastdrive)-'A'+1;
	do_shell( (char far *) init_shell );
	lol.bufsize = BUFSIZE;

	/* create minimal internal structures */
	env_create();							/* original environment */
	buf_create(lol.numbufs,TRUE);		/* temporary buffers at top of DOS mem */
	sft_create(MINSFT);
	sda_callflags = CALLF_FCB;
	sft_create(init_fcbs);				/* SFTs and FCBs are permanent */
	sda_callflags = 0;

	/* install all interrupts */
	bios_putc('2');
	interrupt_init();
	prfunc = generate_int29;		/* for internal printfs */

	/* initialize all internal devices */
	bios_putc('3');
	init_internal_devs();

	/* no disk drives, time to die!! */
	if(lol.numdrives == 0)
		panic(fatal_disk);

	/* allocate an original PSP */
	bios_putc('4');
	build_first_psp();

	/* determine the default drive */

	bios_putc('5');
	CUR_DRIVE = get_default_drive();
	if(CUR_DRIVE >= lol.numdrives)
		CUR_DRIVE = 0;
	lol.bootdrive = CUR_DRIVE+1;		/* 1=A, 2=B, etc... */
	error_no = 0;

	/* process CONFIG.SYS (pass 1 and 2) */
	bios_putc('6');
	open_standard_handles();

   load_and_process_config( (char far *)cfg_name, 0 );

	configsys(PASS1);

	UseSTACKSStacks();

	// SPR 1893 (DE) allow DOS and STACKS to be the only First pass commands
	// so FILES, FCBS and BUFFERS can be reset in NEWFILEs while still
	// allocating STACKS before any device drivers and allocating the HMA
	// just after HIMEM has been run.
	//
	configsys(PASS2);

	/* determine if buffers will fit (along with code) into HMA */
	bufparas = (lol.numbufs * sizeof(BUFLIST) + 0xF)/0x10 +
		lol.numbufs * (lol.bufsize/0x10);
	if(codeparas  + bufparas >= 0xFF8)		/* buffers wont fit */
		bufparas = 0;

	/* calculate FILES and BUFFERS, provided user didn't set them */
	if(!init_files)
	{
		sda_numsfts = smart_calc(MINSFT,SMART_FILES);
	}
	else
	{
		sda_numsfts = init_files;
	}
	if(!init_buffers)
	{
		lol.numbufs = smart_calc(MINBUF,SMART_BUFS);
	}
	else
	{
		lol.numbufs = init_buffers;
	}
	lol.numfcbs = sda_numfcbs;					/* create more FCBs */

	/* create the rest of DOS' structures */
	bios_putc('7');
	if(lol.numfcbs > init_fcbs)
	{
		sda_callflags = CALLF_FCB;
		sft_create(sda_numfcbs - init_fcbs);
	}
	if(sda_numsfts > MINSFT)					/* create more SFTs */
	{
		sda_callflags = 0;
		sft_create(sda_numsfts - MINSFT);
	}
	if(lol.lastdrive < lol.numdrives)
		lol.lastdrive = lol.numdrives;
	cds_create(lol.lastdrive, FALSE);		/* create the CDSs */

	// create searches for use by Long Filenames
	// (if long filenames not supported, does nothing)
	SEARCHES_CREATE();

	/* handle things like INSTALL= */
	configsys(PASS3);

	//	 BT 10-17-01 ETR 3801
	//
	//  It was discovered while adding support to emm386 to relocate
	//  the EBDA and adjust the top of memory to include the
	//  old EBDA memory space that we should not be saving and
	//  restoring the top of memory.
	//
	//	// Reset real top-of-memory for Int 12h
	//	*(unsigned far *)MK_FP( 0x40, 0x13 ) = real_memtop;

	/* load ROM-DOS in HMA or low memory? (only if in RAM) */
	bios_putc('8');
#if VDOS >= 5
	if(trydoshi && !(dos5_flags&DOS5_ROM))
	{
	 	if(hma_used == 0)
		{
			warning_nohigh();
			dos5_flags &= ~DOS5_HMA;
		}
	}
#endif

	/* do NOT create the buffers until AFTER HMA has been allocated */
	buf_create(lol.numbufs, FALSE);

	/* move code to final destination */
	if( !(dos5_flags & DOS5_ROM) )
	{
		mem_free(codeseg);

		/* final destination may be the HMA */
#if VDOS >= 5
		if(dos5_flags & DOS5_HMA)
		{
			codeseg = 0xFFFF;
			LFNcodeseg = mem_alloc(get_LFNcodeparas());
			enable_hma();
			move_dos_hma(codeseg,LFNcodeseg);
		}

		/* or final destination may conventional memory */
		else
#endif
		{
			codeseg = mem_alloc(codeparas);
			move_dos(codeseg);
		}

		fixup_int_handlers();
		dos_code_seg = codeseg;
	}

	/* last but not least */
	bios_putc('9');

	/* open the standard handles */
	open_standard_handles();

	booting = 0;

	/* finish config, execute the root program, etc... */
	finish_boot();
}

/*
	Run Only On Bootup:

	This function is a slight misnomer.  It really
	initializes anything in the DOS Data Area, not just
	the SDA.
*/
void init_sda(void)
{
	/*
		[ externals ]
	*/
	extern char sda_fname1[];
	extern char sda_fname2[];
	extern char * fname1ptr;
	extern char * fname2ptr;
	extern char machine_name[];
	extern uchar sda_retry;
	extern void far * share_table[];
	extern uchar retfar;
	extern struct LOL lol;

	/*
		[ locals ]
	*/
	int i;


	/*
		[ code ]
	*/
	fname1ptr = sda_fname1;
	fname2ptr = sda_fname2;
	sda_retry = RETRY;
	lol.dos5  = 1;					// let programs know structures are DOS 5
	memset(machine_name, ' ', 0xF);	// for LanMan

	/* set up the share table */
	for(i=0; i<15; ++i)
		share_table[i] = MK_FP(getDS(), &retfar);

	lol.extmemsize = get_extmemsize();
}


/*
	Run Only On Bootup:

	Initialize interrupt vectors 20H-3FH to point into the
	DOS handlers.
*/
void	interrupt_init(void)
{
	struct INT_ITEM
	{
		uchar intnum;
		void  *func;
	};

	/*
		[ externals ]
	*/
	extern struct INT_ITEM int_array[];
	extern struct INT_ITEM int_deb_array[];
	extern struct INT_ITEM end_int_array[];
	extern uchar hma_iret;
	extern uchar drive_code;

	/*
		[ locals ]
	*/
	struct INT_ITEM * ip;
	int i;
	char far * vec;


	/*
		[ code ]
	*/

	/* check INT 21H for a drive code (communication from the boot sector) */
	drive_code = 0xFF;
	vec = get_vector(0x21);
	if(FP_SEG(vec) == 0xFFFF)
		drive_code = FP_OFF(vec);

	/* adust generic interrupts handlers to point to our code segment */
	fixup_int_handlers();

	/* default all DOS interrupts to point to an IRET instruction */
	for(i=0x20; i<=0x3F; ++i)
		set_vector(i,MK_FP(getDS(),&hma_iret));
	set_vector(0xDB,MK_FP(getDS(),&hma_iret));

	/* point DOS vectors to our handlers */
#if REMOTEDEB
	ip = int_deb_array;
#else
	ip = int_array;
#endif
	while(ip < end_int_array)
	{
		set_vector(ip->intnum,MK_FP(getDS(),ip->func));
		++ip;
	}
}

extern bool temp_cds;

/*
	Run Only On Bootup:

	Create the needed current working directories.  This
	process is done only once.
*/
void	PASCAL cds_create(int n, bool topmem)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar mem_fit;
	extern struct LOL lol;


	/*
		[ locals ]
	*/
	unsigned seg;
	uchar i;
	uchar oldmem;


	/*
		[ code ]
	*/

	// validate parameter
	ASSERT(n);

	/* let outside world know if we're using temp or "real" CDSs */
	temp_cds = topmem;

	/* if buffers have already been allocated, free them */
	if(FP_SEG(lol.first_cds))
		mem_free(FP_SEG(lol.first_cds));

	/* possibly allocate buffers at top of memory */
	oldmem = mem_fit;
	if(topmem)
		mem_fit = MEM_LAST;

	/* get some memory and zero it */
	seg = mem_z_alloc(n * sizeof(CDS), 'L');

	/* initialize each CDS */
	lol.first_cds = MK_FP(seg,0);
	for(i=0; i<n; ++i)
		cds_init(i);

	/* restore old memory allocation scheme */
	mem_fit = oldmem;
}


/*
	Run Only On Bootup:

	Initialize a single CDS structure.
*/
void	PASCAL cds_init(uchar drive)
{
	extern struct LOL lol;
	setup_cds(lol.first_cds + drive, drive);
}


/*
	Run Only On Bootup:

	Create some SFTs for use by the system.
*/
void	PASCAL sft_create(int n)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern unsigned sda_callflags;
	extern uchar sda_numsfts;
	extern uchar sda_numfcbs;


	/*
		[ locals ]
	*/
	unsigned seg;
	PLPDCB first;
	LPDCB dcb;
	LPDCB dcb2;


	/*
		[ code ]
	*/

	// validate parameter
	ASSERT(n);

	/* record the # of these things, only needed on init */
	if(callf_fcb())
		sda_numfcbs += n;
	else
		sda_numsfts += n;

	seg = mem_z_alloc(sizeof(DCB) + n * sizeof(SFT), callf_fcb() ? 'X' : 'F');
	dcb = MK_FP(seg,0);

	/* handle fcb and DOS2 allocate */
	first = callf_fcb() ? &lol.first_fcb : &lol.first_dcb;
	if(!FP_SEG(*first))
		*first = dcb;

	/* we've already allocated some, search through list */
	else
	{
		dcb2 = *first;
		while((void *)dcb2->next != (void *)0xFFFF)
			dcb2 = dcb2->next;
		dcb2->next = dcb;
	}

	(void *)dcb->next = (void *)0xFFFF;
	dcb->numsft = n;
}


/*
	Run Only On Bootup:

	Create a number of drive parameter blocks.  Links them
	into the master list.
*/
void	PASCAL dpb_create(int n)
{
	extern struct LOL lol;

	int i;
	unsigned seg;
	LPDPB dpb;

	// validate parameter
	ASSERT(n);

	seg = mem_z_alloc(sizeof(DPB) * n, 'E');
	dpb = MK_FP(seg,0);

	/* link the first one into the old list (if any) */
	if(FP_SEG(lol.first_dpb))
		drv_to_dpb(lol.numdrives-1)->next = dpb;
	else
		lol.first_dpb = dpb;

	/* link each of the drive parameter blocks */
	for(i=1; i<n; ++i)
	{
		dpb->next = dpb+1;
		++dpb;
	}
	dpb->next = MK_FP(0xFFFF,0xFFFF);
}



/*
	Run Only On Bootup:

	This function will close all handles and open the
	standard 5 handles: CON x3, AUX, PRN.
*/
void	open_standard_handles(void)
{
	/*
		[ externals ]
	*/
	extern unsigned sda_callflags;
	extern struct LOL lol;
	extern uchar sda_drive;
	extern unsigned dos_psp;

	/*
		[ locals ]
	*/
	int i;
	uchar aux_h;
	uchar con_h;
	uchar prn_h;
	LPSFT sft;
	static char nulldev[] = "NUL";

	/* free all handles in PSP */
	fmemset(MK_FP(dos_psp,0x18),0xFF,20);

	/* free all (DOS2) SFTs */
	i = 0;
	sda_callflags = 0;
	while(1)
	{
		sft = h_to_sft(i);
		if((void *)sft == NULLF)
			break;
		sft->handles = 0;
		++i;
	}

	/* Remind the SDA that we are using the bootdrive */
	sda_drive = lol.bootdrive;

	/* get SFTs for the standard devices */
	aux_h = internal_open((char far *)"AUX", TRUE);
	if(aux_h == 0xFF)
	{
		aux_h = internal_open(nulldev, TRUE);
	}
	con_h = internal_open((char far *)"CON", TRUE);
	prn_h = internal_open((char far *)"PRN", TRUE);
	if(prn_h == 0xFF)
	{
		prn_h = internal_open((char far *)nulldev, TRUE);
	}

	/* assign them to handles (open them) */
	for(i=0; i<3; ++i)
		set_handle_val(i, con_h);
	set_handle_val(3, aux_h);
	set_handle_val(4, prn_h);
}

/*
	Run Only On Bootup:

	Create the initial environment.  This environment is
	at the top of memory, but will be moved down after the
	internal DOS structures have been combined into 1 MCB.
*/
void env_create(void)
{
	extern uchar mem_fit;
	extern char env_string[];
	extern unsigned env_para;

	/* allocate and copy the default environment */
	mem_fit = MEM_LAST;
	boot_env_size = env_para * 0x10;
	boot_env = mem_z_alloc(boot_env_size,0);
	mem_fit = MEM_FIRST;
	fmemcpy(MK_FP(boot_env, 0), (void far *)env_string, boot_env_size);
}

/*
	Run Only On Bootup:

	On boot up of ROM-DOS, before we have a "real" program
	running, we still need handles (which are in the PSP),
	so make a fake one.  This is really only used during
	CONFIG.SYS so that devices have the standard handles.
*/
void	build_first_psp(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern union dos_regs far * userregs;
	extern unsigned disk_stack[];
	extern unsigned dos_psp;
	extern unsigned boot_env;

	/*
		[ code ]
	*/

	/* allocate the memory for this PSP */
	dos_psp = mem_z_alloc(0x100, 'E');

	/* fill out the PSP */
	userregs = (void far *)(disk_stack+5);
	create_new_psp(dos_psp,TRUE,TRUE);
	((struct PSP far *)MK_FP(dos_psp,0))->env_seg = boot_env;

	/* create standard handles CON x3, AUX, PRN */
	/* open_standard_handles(); */
	/* DO NOT CREATE the handles until we know a valid
		drive letter! (Boot step 5 in main()) */
}


/*
	Run Only On Bootup:

	Initializes (and links) all built-in devices.  That
	includes both character and disk devices.
*/
void	init_internal_devs(void)
{
	extern PDEV_HDR built_in[];
	extern unsigned ROM_DEVSEG;

	int i;
	LPDEV_HDR device;

	/* initialize all built-in devices, accept NUL (NUL is not in DEVSEG) */
	for(i=1; built_in[i]; ++i)
	{
#ifdef DEBUG_DEVICE_INIT
      // Added debug reporting for customer who was experiencing
      // "divide by zero" during device init
      //
     	bios_putc('a'+i-1);
#endif

		device = MK_FP(ROM_DEVSEG, built_in[i]);
		dev_init(device, (char far *)"");
	}
}


/*
	Run Only On Bootup:

	Adjust the size of the buffer structures to accomodate
	larger sectors (1K or more).
*/
void new_buf_size(int size)
{
	extern struct LOL lol;

	// size must be valid
	ASSERT(size);

	/* no need to do anything if buffers are already big enough */
	if(size <= lol.bufsize)
		return;

	/* set up the List of Lists to reflect the new size */
	lol.bufsize = size;

	/* create temporary buffers of larger size */
	buf_create(lol.numbufs, TRUE);
}


/*
	Run Only On Bootup:

	Initialize the device driver, whether it be character
	or disk, internal or external. Internal are those that
	are hard-coded into ROM-DOS, external are those
	installed via CONFIG.SYS.  This function:

		1) Calls the device INIT routine
		2) Reallocates memory to appropriate size (or frees
		   it on init error)

	lpdevHeader is a pointer to the (initial) device header for
	a previously loaded (but not yet initialized) device.
	pargStr is the parameter list from the DEVICE= line in CONFIG.SYS.

	Returns FALSE if it failed, TRUE if worked.
*/
void	PASCAL dev_init(LPDEV_HDR lpDevHeader, char far *pArgStr)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern struct dev_req_hdr sda_reqhdr;
	extern unsigned mem_top;
	extern unsigned ROM_DEVSEG;
#if USING_FAT32
	extern LPDPB sda_dpbptr;
	extern LPDEV_HDR sda_devptr;
#endif
   extern char FAT_type;

	// These next three items needed for 386MAX support.
	// Its a cheap hack, but it works...
	//
	extern char sda_ename1[];

	/*
		[ locals ]
	*/
	int i;
	unsigned seg;				// starting mem seg
	unsigned seg2;				// ending mem seg
	unsigned segtmp;
	unsigned top_of_memory;
	unsigned far * uptr;
	unsigned attr;
	LPDEV_HDR lpCurDevHeader;
	LPDPB dpb;
	struct MCB far * lpMCB;
	uchar units;
	int memsized = FALSE;	/* memory has not been sized yet */
	int devAlloced;		  // individual driver's allocation request in para
	unsigned devStart;     // absolute para address of a given device header
	int partialParagraph = 0; // used to round up segment addresses
	void far *lastAlloc;  // last value left in Trans field by device

	/*
		[ code ]
	*/

	// validate pointers
	ASSERT(lpDevHeader);
	ASSERT(pArgStr);

	/* check for memory damage */
	mem_chk();

	/* get starting and ending memory points for device */
	seg2 = seg = FP_SEG(lpDevHeader);

	/* determine "top" of memory */
	top_of_memory = (seg == ROM_DEVSEG ? mem_top : next_mcb_seg(seg));

	//	 BT 10-17-01 ETR 3801
	//
	//  It was discovered while adding support to emm386 to relocate
	//  the EBDA and adjust the top of memory to include the
	//  old EBDA memory space that we should not be saving and
	//  restoring the top of memory.  I believe that the case below
	//  was evidence that the top of memory should have never been
	//  saved and restored.  Thus, this code would have been a fix
	//  for a bad fix.
	//
	//	// Fake the Int 12h top-of-memory value to be the end of this
	//	// MCB. Convert from segment to KByte value. This is needed for
	//	// some devices (specifics not recorded...). However, 386MAX will
	//	// NOT work with this so we make it a special case.
	//	//
	//	if(fmemcmp( sda_ename1, _386MAX_name, 11) != 0)
	//		*(unsigned far *)MK_FP( 0x40, 0x13 ) = (top_of_memory >> 6 );

	/*

		A single file may have more than one device driver in
		it, so loop through each device in the file. We set the
		initial reqhdr.trans field value to the top of memory,
		there-after it gets restored to the last value returned by
		the device. Assuming no errors, we will allocate the max
		possible value returned by any of the devices in a given file.

	*/
	lastAlloc = MK_FP(top_of_memory,0);
	while(1)
	{
		/*
			Set up request header for an INIT call to the
			device; zeroing header also sets the init command
			(DEV_INIT=0)
		*/
		memset(&sda_reqhdr, 0, sizeof(struct dev_req_hdr));
		sda_reqhdr.trans = lastAlloc;

		sda_reqhdr.count = FP_OFF(pArgStr);
		sda_reqhdr.start = FP_SEG(pArgStr);

		sda_reqhdr.drive = lol.numdrives;

		/* call the device (to initialize it) */
		dev_call(&sda_reqhdr, lpDevHeader);

		lastAlloc = sda_reqhdr.trans;


		//  BT 10-17-01 ETR 3801
		//
		//  The top of memory needs to be updated in case a device
		//  changes the top of memory, such as emm386 when relocating
		//  the EBDA.
		//
		mem_top = getmemtop();
		top_of_memory = (seg == ROM_DEVSEG ? mem_top : next_mcb_seg(seg));

		/* if we are initializing an BUILT-IN device... */
		if(seg == ROM_DEVSEG)
		{
			if(sda_reqhdr.status & DEVE_ERROR)
				return;
			memsized = TRUE;	/* do NOT realloc memory for internal devices */
		}

		// only keep as much memory as any single EXTERNAL device needs
		//	but, in the case of multiple devices in a single file, track
		//	the max requested amount and allocate that (if it's available).
		//
		else
		{

			// MS PROTOMAN v2.x changes the paragraph value of the outer
			// memory block in the lol.first_mcb MCB. In such cases we
			// have to adjust the size of the device's block downward
			// since PROTOMAN does not.
			//
			segtmp = seg - lol.first_mcb;
			segtmp = 1 + *(unsigned far *)(MK_FP(lol.first_mcb,3)) - segtmp;
			lpMCB = (struct MCB far *)(MK_FP(seg-1,0));
			if(lpMCB->para_cnt > segtmp)
				lpMCB->para_cnt = segtmp;

			// Get absolute paragraph address of the device's endpoint.
			// The device may request nnnn:FFFFh as an ending address so
			// be careful not to overflow the offset before dividing.
			// If any partial paragraphs then round up.
			//
			partialParagraph = (FP_OFF(sda_reqhdr.trans) & 0x000F) ? 1 : 0;
			segtmp        = (FP_OFF(sda_reqhdr.trans) >> 4) + partialParagraph;
			segtmp       +=  FP_SEG(sda_reqhdr.trans);

			// if this is the highest valid water mark in this file, record it
			//
			if(segtmp > seg2 && segtmp < top_of_memory)
				seg2 = segtmp;

			// determine the allocation request for the individual device
			// in paragraphs. MSDOS views each allocation request as
			// relative to the start of a device's header, even if the
			// header is some positive offset from the start of the
			// device's exec load block (MSDOS Prog Ref pg 397.)
			//
			devStart = FP_SEG(lpDevHeader) + (FP_OFF(lpDevHeader)+15) / 16;
			devAlloced = segtmp - devStart;

			// Regardless of whether the external device reported an
			// init failure or not, if it said to free all its memory
			// then do that without adding it to the device chain.
			//	If it said to keep its memory then we do that too, and
			//	chain it in, regardless of whether the driver reported
			//	an init failure.
			//
			// Note: SystemSoft drivers (eg CSALLOC v3.01) return
			//       (SEG(dev_hdr)+1):0 when they want to terminate.
			//
			//			Checking that the strategy/service offset is
			//			in bounds of the allocation request for a
			//			given device does not work when executables
			//			contain multiple devices that share common
			//			strategy/service routines (eg SoftICE).
			//
			if(devAlloced <= 1)
				break;

		}

		/* character device, check for clock and console */
		attr = lpDevHeader->attr;
		if(attr & DEVA_CHAR)
		{
			if(attr & DEVA_CLOCK)
				lol.clkdev = lpDevHeader;
			else if(attr & DEVA_CON)
				lol.condev = lpDevHeader;
		}

		/* disk device, make the DPBs */
		else
		{
			/* we MUST size memory before we can create DPBs */
			if(!memsized)
			{
				if(seg2 <= seg || seg2 >= top_of_memory)
				{
					mem_free(seg);
					return;
				}
				mem_resize(seg, seg2-seg);
				memsized = TRUE;
			}

			/* get the # of units and ptrs to BPBs */
			units = sda_reqhdr.media;
			uptr = MK_FP(sda_reqhdr.start,sda_reqhdr.count);
			dpb_create(units);

			/* set up the drive parameter block for each device */
			dpb = drv_to_dpb(lol.numdrives);
			for(i=0; i<units; ++i)
			{
				dpb->drive  = lol.numdrives;
				dpb->unit   = i;
				dpb->device = lpDevHeader;

#if USING_FAT32
				//  These need to be set for FAT32 support so that
				//  GetFSInfoFields can use the romdos buffering
				//  scheme.
				//
				sda_dpbptr = dpb;
				sda_devptr = lpDevHeader;
#endif

				/* disk is bad, don't use it! */
				if(!bpb_to_dpb(MK_FP(FP_SEG(uptr),*uptr++), dpb))
					goto freeit;

				/* update buffers based on size */
				new_buf_size(dpb->bytes_per_sec);

            //  Initialize the FAT_type for get_fat_entry() and
				//  set_fat_entry()
            //
            if( sda_dpbptr->hiclus == 0 )
            {
               FAT_type = FS_FAT32;
            }
            else if( sda_dpbptr->hiclus <= HICLUS12 )
            {
               FAT_type = FS_FAT12;
            }
            else
            {
               FAT_type = FS_FAT16;
            }

				++lol.numdrives;
				++dpb;
			}

			/* create temporary CDS so drives can be used on init */
			cds_create(lol.numdrives, TRUE);
		}

		/* get pointer to segment of next device */
		uptr = (unsigned far *)(&lpDevHeader->next) + 1;

		/* temporarily update the segment of this device driver's
		   next pointer to point to the next device in the file (if any)
		*/
		*uptr = seg;

		// MSDOS links each device into the device driver chain as
		// it is initialized. We need to remember the value of the
		// next pointer field in the current device header while we
		// link the current driver to the last installed device and
		// then point the NUL devices next ptr to the current
		// device. The NUL device's device header is actually a
		// fixed part of the List of Lists.
		//
		lpCurDevHeader = lpDevHeader->next;
		lpDevHeader->next = lol.first_dev.next;
		lol.first_dev.next = lpDevHeader;

		// point to the next device header in this file (if any)
		// if there are no more devices in the file then the offset
		// (and possibly the segment) values of the next ptr will be -1.
		//
		lpDevHeader = lpCurDevHeader;

		// if there are no more devices in this file then break
		//	Don't worry about the segment values, MSDOS doesn't.
		//
		if(!(void *)lpDevHeader ||
		    (void *)lpDevHeader == (void *)0xFFFF)
			break;

		// if an external device could not be initialized then
		// don't process any other devices that might be in the file.
		//
		if(sda_reqhdr.status & DEVE_ERROR)
			break;

	}

	// if we haven't already resized the exec block then do that now
	// seg2 will have been set to the maximum segment value requested
	// by any of the successfully initialized drivers in the file.
	//
	if(!memsized)
	{
		if(seg2 <= seg || seg2 >= top_of_memory)
		{
freeit:
			mem_free(seg);
			return;
		}


		//  BT 10-17-01 ETR 3801
		//
		//  Added a check of the MCB chain in case a device attempts
		//  to modify it but instead corrupts it.
		//
		/* check for memory damage */
		mem_chk();


		mem_resize(seg, seg2-seg);
		mem_new_owner(seg, seg);		/* so parent PSP doesn't own device */
	}
}


/*
	Run Only On Bootup:

	Search the Drive Parameter Block list for this built-in
	device.  Return its drive #.
*/
uchar dev_to_drv(void * builtin_dev)
{
	extern struct LOL lol;

	LPDPB dpb;
	uchar drive;

	// validate pointer
	ASSERT(builtin_dev);

	/* set drive to invalid (in case device is not found) */
	drive = 0xFF;

	dpb = lol.first_dpb;
	while(FP_OFF(dpb) != 0xFFFF)
	{
		if(FP_OFF(dpb->device) == (unsigned)builtin_dev)
		{
			drive = dpb->drive;
			break;
		}
		dpb = dpb->next;
	}
	return drive;
}

// some BIOSes cannot read data from disk into the HMA, so
// we must have a buffer down low to do it.
extern void far * lo_buffer;

/*
	Run Only On Bootup:

	Create 1 or more buffers.  Called only internally by
	ROM-DOS to make a buffers list.  Can create the buffers
	in the top of memory (upper part of 640K), just after
	DOS or in the HMA.

	All DOS buffers must be together, since the pointers
	to next and previous are near  pointers.
*/
void	PASCAL buf_create(int n, bool topmem)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar mem_fit;
	extern struct LOL lol;
	extern LPBUFLIST bufinfo;
	extern unsigned dos5_flags;
	extern LPBUFLIST sda_cur_buf;

	/*
		[ locals ]
	*/
	unsigned seg;
	unsigned size = (sizeof(BUFLIST) - 1 + lol.bufsize);
	unsigned offset;
	LPBUFLIST buf;

	/*
		[ code ]
	*/

	// validate parameter
	ASSERT(n);

	/* if buffers have already been allocated, free them */
	if(FP_SEG(lol.bufinfo))
	{
		seg = FP_SEG(*lol.bufinfo);
		mem_free(seg);
		error_no = 0;
	}

	/* possibly allocate buffers at top of memory */
	if(topmem)
		mem_fit = MEM_LAST;

	/*
		If ROM-DOS is loaded in the HMA, load BUFFERS there too.
		Note that the BUFFERS may not fit.  Check ROM-DOS code
		size, use the rest for BUFFERS.  If the BUFFERS exceed
		that size, allocate them low.
	*/
	offset = get_DOScodeparas() << 4;
	if(!topmem && (dos5_flags & DOS5_HMA) && (offset + n*size) > offset)
	{
		/*
			Many BIOSes cannot read sectors up high, so we must
			have a sector down low where we can do the actual
			read and writes.
		*/
		seg = mem_z_alloc(size, 'B');
		lo_buffer = MK_FP(seg, 0);

		/* seg/off of buffers is in HMA */
		seg = 0xFFFF;
	}

	/* use conventional memory for buffers */
	else
	{
		seg = mem_z_alloc(n * size, 'B');
		offset = 0;
	}

	lol.bufinfo = (void far *)(&bufinfo);
	buf = bufinfo = sda_cur_buf = MK_FP(seg, offset);

	/* initialize each buffer */
	buf->prev = (void *)((n-1) * size + offset);
	while(n--)
	{
		/* set buffer to invalid (flags are already 0 from z_alloc) */
		buf->drive = B_INVDRV;
		buf->flags = B_UNUSED;

		/* link in the buffer */
		buf->next = (n==0 ? (void *)offset : (void *)(FP_OFF(buf) + size));
		if(FP_OFF(buf) > offset)
			buf->prev = (void *)(FP_OFF(buf) - size);

		/* on to next buffer */
		buf = MK_FP(seg,buf->next);
	}

	/* reset memory allocation scheme */
	mem_fit = MEM_FIRST;
}

/*
	Allocate some "near" memory. The memory will be zeroed.
	Can fail if too much near memory is allocated (64K total
	ROM-DOS data).

	Used by internal routines to allocate memory. Cannot free
	the memory. Any memory allocated in this manner MUST be
	allocated prior to processing DEVICE= statements in CONFIG.SYS.
*/
void * mem_alloc_ptr(unsigned bytes, char id)
{
	unsigned seg, ptr;
	extern unsigned ROM_DS;

	/* allocate the memory */
	seg = mem_z_alloc(bytes, id);

	/* determine the address */
	seg -= ROM_DS;
	ASSERT(seg + (bytes>>4) < 0x1000);	/* must be less than 64K */
	ptr = seg * 0x10;

	/* return the pointer */
	return (void *)(ptr);
}

/*
	Run Only On Bootup:

	Memory initializiation entails determining the start and
	ending of physical memory and setting up an MCB that
	reflects this. It also requires catering to a variety
	of programs that ugly up the works, such concessions include

	 	1) setting up two MCB chains to satisfy PROTMAN
		2) allocating a memory block to contain RPL code copied from ROM

	See comments for FirstMCB() (DOSMEM.C) and GoAskRPL (SYSGEN.ASM).

*/
void	mem_init(void)
{
	extern struct LOL lol;
	extern unsigned current_psp;
	extern unsigned mem_top;
	extern uchar mem_fit;

	struct MCB far *mcb;
	uchar cur_fit, far*pMCB;
	unsigned RPLseg, RPLsize, i;

	/* create the 1st MCB */
	lol.first_mcb = getHEAP();
	mcb = MK_FP(lol.first_mcb, 0);
	mcb->owner = FREEMEM;
	mcb->para_cnt = mem_top - lol.first_mcb - 1;
	mcb->id = MEMSIGZ;

	// this flag tells the initial FirstMCB() routine to use OUTER
	// MCB chain if non-zero
	mcb->init_id = TRUE;			// anything non-zero

	/* set owner of any allocated PSPs to SYSTEM */
	current_psp = SYSTEM_OWNER;

	// Create a 0 length MCB for use with combine, this will force
	// the start of the second MCB chain immediately following the
	// lol.first_mcb
	// "ID" of 0 (FALSE) tells FirstMCB() routine to user INNER
	// MCB chain.  When we are done loading device drivers, this
	// "ID" will be set to TRUE again.
	//
	// FROM NOW ON, mem_z_alloc()s and mem_alloc()s will be from the
	// INNER MCB chain.
	//
	mem_z_alloc(0, 0);

	// if top of mem from BIOS int 12h is reduced by the
	// existence of RPL code at the top of RAM (see SYSGEN.ASM)
	//
	RPLseg = getmemtop();
	if(RPLseg != GoAskRPL(RPLseg))
	{
		// gotta alloc a block at top of RAM to contain the RPL code
		// that is already there (copied from ROM on int 18h or int 19h,
		// depending on the ROM). The seg of the RPL code block is
		// returned by int 2Fh fn 4A06h (ref: DOS Internals, pg 140).
		//
		cur_fit = mem_fit;				// save allocation strategy
		mem_fit = MEM_LAST;
		RPLsize = RPLseg - GoAskRPL(RPLseg);

		RPLseg  = mem_alloc(RPLsize); 	// allocate the block to DOS (ha ha)
		RPLseg--;

		pMCB    = MK_FP(RPLseg,8);    	// ID is "RPL"
		*pMCB++ = 'R'; *pMCB++ = 'P'; *pMCB++ = 'L';
		for(i=0; i < 5; i++)
			*pMCB++ = '\0';

		mem_fit = cur_fit;			// restore allocation strategy
	}

	// Set the size of the first MCB in the outer chain (mem_alloc() takes
	// care of adjusting the size of block in the inner chain). Unless
	// there is RPL code hiding out at the top of RAM, also say that this
	// is the last block. We now have two distinct and nested MCB chains
	// controlling the same memory region.
	//
	mcb->para_cnt = RPLseg - lol.first_mcb - 1;
	if(mem_top == RPLseg)
		mcb->id = MEMSIGZ;
}


/*
	Run Only On Bootup:

	Allocate all of memory, and load a device driver into
	that memory space.  This is only used by CONFIG.SYS to
	load device drivers.  This does not create a PSP nor
	make a copy of the environment.

	Returns the starting segment of the device, or 0xFFFF
	on error.
*/
unsigned PASCAL load_device(char far *name)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern uchar sda_chardev;
	extern int sda_no24;


	/*
		[ locals ]
	*/
	unsigned avail;
	unsigned seg;

	/*
		[ code ]
	*/

	/* validate parameter */
	ASSERT(name);

	/*
		Don't allow Int 24Hs during bootup.  Other device drivers
		may turn off this flag (through 21 6CH), so make sure Int
		24Hs are disabled.
	 */
	sda_no24 = TRUE;

	/* get the filename into the sda */
	if(internal_open(name, FALSE) == 0xFF || sda_chardev)
		goto error;
	internal_close(FALSE);

	/* allocate all of memory, but leave that memory "free" */
	avail = mem_alloc_best(1, 0xFFFF, &seg);
	if( error_no )
		goto error;

	/* actually load the file into memory */
	if(!load_overlay(avail,seg,seg))
	{
error:
		error_no = 0;
		return 0xFFFF;
	}

	/* now that we know the file was loaded, give proper ownership */
	mem_new_owner(seg, seg);

	/* return the starting segment of this file */
	return seg;
}
