//
//   $Id: DOSMEM.C 1.12 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*****************************************************************

	DOSMEM.C     Memory Allocation and Freeing

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This file handles the memory functions.  This includes
   memory management by built-in (resident) devices, and
   DOS calls for memory management.

******************************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 12

#if NOT_IN_INIT
/*
	Based on the amount of memory, choose a "reasonable"
	number.  Used for FILES and BUFFERS.
*/
int PASCAL smart_calc(int minval, int maxval)
{
	extern unsigned mem_top;

	unsigned scale = 0xA000 / maxval;
	int num = mem_top / scale;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(minval > 0 && maxval > 0);
	ASSERT(minval < maxval);

	if(num < minval)
		num = minval;

	EXIT_FUNCTION();

	return num;
}
#endif


/*
   Returns a far pointer to the initial MCB in the current MCB chain.

	If the init_id value of the lol.first_mcb is zero then we are
	in the boot process. While in the boot process we need to
	allocate all available memory to the System Data (SD) MCB;
	this initial MCB will contain all subsequent allocations made
	during CONFIG.SYS processing and a second MCB chain is
	maintained inside of the SD MCB for these allocations. The
	first MCB in this second MCB chain is located immediately
	following the lol.first_mcb MCB (or, as below, at mcb+1).

	If the initial MCB has an init_id value of 1 then we have
	completed CONFIG.SYS processing and the SD MCB will have been
	shrunk to just contain the allocations made during CONFIG
	processing.

	Note that when the SD MCB's init_id field is 0 then the block size
	recorded in the MCB is the top of memory as far as our allocation
	routines are concerned. Once mem_combine (DOSMAIN.C) has been
	called and the init_id field is 1 then top of memory will be the
	same as that indicated in the SDA.

	These dual MCB lists and top of memory indicators are required by
	MicroSoft PROTMAN v2.x which allocates a block of memory that needs
	to persist only until the end of CONFIG.SYS processing.

*/
struct MCB far *FirstMCB(void)
{
  extern struct LOL lol;
  struct MCB far *mcb;

	ENTER_FUNCTION();

	mcb = MK_FP(lol.first_mcb,0);

	// special case during booting, use secondary MCB chain
	if(mcb->init_id == 0)
		mcb = MK_FP(lol.first_mcb+1,0); // MCBs always seg:0

	EXIT_FUNCTION();

	return mcb;
}

/*
	Return the next segment after this MCB.
*/
unsigned next_mcb_seg(unsigned seg)
{
	struct MCB far *mcb;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(seg);

	mcb = MK_FP(seg-1, 0);

	EXIT_FUNCTION();

	return seg + mcb->para_cnt;
}

/*
	This Routine checks the memory control blocks for a
	consistent state.  Produces a FATAL error message if
	the MCBs aren't valid.
*/
void mem_chk(void)
{
	/*
		[ externals ]
	*/
	extern char fatal_mem[];
	extern char fatal_mem2[];
	extern struct LOL lol;


	/*
		[ locals ]
	*/
	struct MCB far *mcb;
	unsigned seg,nextseg;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	mcb = FirstMCB();
	while(1)
	{
		/* check if the MCB is valid */
		seg = FP_SEG(mcb);
		nextseg =  seg + mcb->para_cnt + 1;
		if((mcb->id != MEMSIG && mcb->id != MEMSIGZ) || nextseg <= seg)
			panic(fatal_mem,seg);

		/* we're done */
		if(last_mcb(mcb))
			break;

		mcb = next_mcb(mcb);
	}

	EXIT_FUNCTION();
}


/*
	Allocate the best sized memory block and return the size
	of the block.  Leaves the allocated memory "free" in case of
	critical error during disk access.  Used only internally
	by things like EXEC.

	Error codes are passed backward from mem_alloc through to
	mem_alloc_best's caller.  If error_no is set on exit, there
	was an error.

	Returns the paragraphs allocated (and sets segment of the block).
*/
unsigned PASCAL mem_alloc_best(unsigned least, unsigned most, unsigned *seg)
{
	/*
		[ externals ]
	*/
	extern uchar mem_fit;			/* 0=first, 1=best, 2=last */
	extern int error_no;

	/*
		[ locals ]
	*/
	unsigned paras;
	bool umb_first;
	uchar old_memfit;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(least);
	ASSERT(most);
	ASSERT(seg);

	/* Because we may or may not set error_no on exit,
		we need to be sure that it isn't already set. */
	ASSERT( !error_no );

	/* look in Upper Memory Blocks first? */
	umb_first = 0;
	old_memfit = mem_fit;
	if( mem_fit & MEM_UMB1ST )
	{
		++umb_first;

		/* Turn off UMB1ST bit and turn on UMBONLY bit.
			We can't just allocate UMB1ST because we'd like
			to try and get at least the LeastParas allocated
			in UMBs even if MostParas won't fit. */
		mem_fit = (mem_fit & ~MEM_UMB1ST) | MEM_UMBONLY;
	}
try_again:

	/* Try to get as much memory as the program "would like."
		Since alloc returns either size *or* segment, I chose
		to assign the value to paras and then if it worked,
		move the value into *seg and set paras to the right value. */
	paras = mem_alloc(most);
	if( error_no )
	{
		/* Not enough memory.  Can we get at least as much as
			the program needs?  If not, error out. */
		if( paras < least )
		{
			/* For UMBFIRST, try the whole thing again in lower memory */
			if( umb_first )
			{
				umb_first = 0;
				error_no = 0;
				mem_fit = old_memfit;
				goto try_again;
			}
			EXIT_FUNCTION();
			return 0;
		}
		else
		{
			error_no = 0;
			*seg = mem_alloc( paras );

			ASSERT( !error_no );
		}
	}
	else
	{
		*seg = paras;
		paras = most;
	}
	mem_new_owner(*seg, FREEMEM);


	//  BT 1-8-02 ETR 3849
	//
	//  If mem_fit was set to 80h on entry to this function then it would have
	//  been set to 40h so that the best fit strategy would work correctly.
	//  However, if it was successful at finding a memory block, the
	//  allocation strategy was never restored.  Below we are now restoring
	//  the allocation strategy.
	//
	mem_fit = old_memfit;

	EXIT_FUNCTION();

	return paras;
}

/*
	Allocate a number of paragraphs and return the
	segment of this memory.  This routine can allocate
	in one of 9 ways: first fit, best fit, last fit.

	bit 80H means UMBs first
	bit 40H means UMBs only

	Returns the segment of the allocated memory, or sets
	error_no and returns the largest free block on failure.
*/
unsigned PASCAL mem_alloc(unsigned para_req)
{
	/*
		[ externals ]
	*/
	extern uchar mem_fit;			/* 0=first, 1=best, 2=last */
	extern int error_no;
	extern unsigned current_psp;
	extern struct LOL lol;
	extern uchar umb_link;
	extern unsigned umb_arena;
	extern uchar booting;


	/*
		[ locals ]
	*/
	unsigned avail;
	unsigned count;
	struct MCB far *mcb;
	struct MCB far *mcb2;
	struct MCB far *best_mcb;
	unsigned best;
	bool last, umb_first;
	uchar fit;



	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* make sure contiguous free blocks are concatinated */
	fix_mem_list();

	/* look in Upper Memory Blocks first? */
	umb_first = 0;
	if(umb_link && (mem_fit & (MEM_UMB1ST | MEM_UMBONLY)))
		++umb_first;

	/* only pay attention to first/best/last fit for most operations */
	fit = (mem_fit & 0x3);

try_again:

	/* start searching at the 1st Memory Control Block */
	mcb = FirstMCB();
	best_mcb = NULL;
	avail = best = 0;

	while(1)
	{
		/* is the MCB the last MCB? */
		last = last_mcb(mcb);

		/* is this memory block free? */
		if(mcb->owner == FREEMEM && (!umb_first || FP_SEG(mcb) >= umb_arena))
		{

			/* get some data about this block */
			count = mcb->para_cnt;

			/* record greatest mem available */
			if(count > avail)
				avail = count;

			/* only if the MCB is big enough */
			if(count >= para_req)
			{

				/* Last Fit */
				if(fit >= MEM_LAST)
				{
					best_mcb = mcb;
					best = count;
				}

				/* Best Fit */
				else if(fit)
				{

					/* have we found a better fit? */
					if(!best || count < best)
					{
						best_mcb = mcb;
						best = count;
					}
				}

				/* First Fit */
				else
				{

do_the_alloc:
					/* allocate the memory */
					mcb->owner = current_psp;
					mcb->para_cnt = para_req;

					/* make the "Master Memory List" pretty */
					if(booting)
						mcb->init_id = 'D';

					/* we have leftover memory, build a new MCB */
					if(para_req < count)
					{

						/* make a new memory control block */
						mcb2 = next_mcb(mcb);
						mcb2->owner = FREEMEM;
						mcb2->para_cnt = count - para_req - 1;
						mcb2->id = MEMSIG;

						/* the MCB we allocated from wuzza the last MCB, eh? */
						if(last)
						{
							mcb->id  = MEMSIG;
							mcb2->id = MEMSIGZ;
						}
					}

					EXIT_FUNCTION();
					return FP_SEG(mcb) + 1;
				}
			}
		} /* end of if free */

		/* we've checked the last memory block */
		if(last)
			break;

		mcb = next_mcb(mcb);
	} /* end of while loop */


	/* we have gone through the whole list for best/last fit */
	if(fit && best >= para_req)
	{

		/* return to the best/last found MCB */
		mcb   = best_mcb;
		count = best;
		last  = last_mcb(mcb);

		/* Best (or exact) Fit */
		if(fit == MEM_BEST || best == para_req)
			goto do_the_alloc;

		/* Last Fit - allocate the memory */
		mcb->para_cnt = count - para_req - 1;

		/* Last Fit - we have leftover memory, build a new MCB */
		mcb2 = next_mcb(mcb);
		mcb2->owner = current_psp;
		mcb2->para_cnt = para_req;
		mcb2->id = MEMSIG;

		/* the block we allocated from wuzza the last MCB, hey? */
		if(last)
		{
			mcb->id  = MEMSIG;
			mcb2->id = MEMSIGZ;
		}

		EXIT_FUNCTION();
		return (FP_SEG(mcb2) + 1);
	}

	/* if memory was not found, try again.  don't try again if UMBONLY */
	if(umb_first && (mem_fit & (MEM_UMB1ST | MEM_UMBONLY)) == MEM_UMB1ST)
	{
		umb_first = 0;
		goto try_again;
	}

	/* not enough memory */
	error_no = ERR_NOMEM;

	EXIT_FUNCTION();

	return avail;
}


/*
	Find the MCB pointed to by seg.  If not found,
	return 0.

	Note: ETR 1985 CardTalk device drivers go out and munipulate the
			MCB chain on their own.  MCBs for calls such as resize will
			not be found during config.sys processing unless we search
			both MCB chains. (BT)
*/
unsigned PASCAL find_mcb(unsigned seg)
{
	extern struct LOL lol;
	unsigned uFirstMCBSeg;
	unsigned uMCBChains;
	struct MCB far * mcb;

	ENTER_FUNCTION();

	// validate segment
	ASSERT(seg);

	//  Get the first MCB in the chain
	//
	uFirstMCBSeg = lol.first_mcb;
	mcb = MK_FP(uFirstMCBSeg,0);
	uMCBChains = 1;

	// special case during booting, use secondary MCB chain
	//
	if(mcb->init_id == 0)
	{
		mcb = MK_FP(uFirstMCBSeg+1,0); // MCBs always seg:0
		uMCBChains++;
	}

	do{

		/* does this MCB exist? */
		while(TRUE)
		{
			if(FP_SEG(mcb)+1 == seg)
			{
				EXIT_FUNCTION();

				/* Found it */
				return seg;
			}

			if(last_mcb(mcb))
				break;

			mcb = next_mcb(mcb);
		}

		uMCBChains--;
		mcb = MK_FP(uFirstMCBSeg,0); // MCBs always seg:0

	}while( uMCBChains );

	EXIT_FUNCTION();

	return 0;
}


/*
	Free a certain memory control block.  It will not free
	the memory unless it is a valid control block.
	Returns segment of MCB, or sets error_no on error.
*/
unsigned PASCAL mem_free(unsigned seg)
{
	extern int error_no;

	struct MCB far *mcb;

	ENTER_FUNCTION();

	// validate segment
	ASSERT(seg);

	/* the only error possible is invalid MCB */
	mcb = MK_FP((seg-1), 0);
	if(mcb->id != MEMSIGZ && mcb->id != MEMSIG)
		error_no = ERR_BADMEMSEG;
	else
		mem_new_owner(seg, FREEMEM);

	EXIT_FUNCTION();

	return seg-1;
}


/*
	Change the size of a memory control block.  It will always
	work if changing to a smaller size, but unless free memory
	is above this MCB, it will fail when enlarging.

	Returns the segment of the memory that is resized if it
	worked, or the maximum size if failed.
*/
unsigned	PASCAL mem_resize(unsigned seg, unsigned para_req)
{
	extern struct LOL lol;
	extern unsigned current_psp;
	extern int error_no;

	unsigned avail;
	unsigned avail2;
	struct MCB far *mcb;
	struct MCB far *mcb2;
	uchar last;
	uchar last2;

	ENTER_FUNCTION();

	// validate segment
	ASSERT(seg);

	/* concatinate free blocks */
	fix_mem_list();

	/* if not found, let 'em know */
	if(!find_mcb(seg))
	{
		error_no = ERR_BADMEMBLK;
		EXIT_FUNCTION();
		return 0;
	}

	/* figure out the maximum available to this MCB */
	mcb = MK_FP((seg-1), 0);
	last = last_mcb(mcb);
	if(!last)
	{
		mcb2 = next_mcb(mcb);
		last2= last_mcb(mcb2);
	}
	avail2 = 0;
	if(!last && mcb2->owner == FREEMEM)
	{
		avail2 = mcb2->para_cnt + 1;
	}
	avail = mcb->para_cnt + avail2;

	/* check if it's possible to get that much memory */
	if(para_req <= avail)
	{
		/* we have the technology, we can rebuild the memory */
		mcb->owner = current_psp;
		mcb->para_cnt = para_req;

		/* less than the avail, we must build the next MCB */
		if(para_req < avail)
		{

			mcb->id = MEMSIG;
			mcb2 = next_mcb(mcb);

			mcb2->id       = MEMSIG;
			mcb2->owner    = FREEMEM;
			mcb2->para_cnt = avail-para_req-1;
			if(last || (last2 && avail2))
				mcb2->id = MEMSIGZ;
		}

		/* same size, should only change sig */
		else if(avail2 && last2)
		{
			mcb->id = MEMSIGZ;
		}

		EXIT_FUNCTION();
		return seg;
	}

	error_no = ERR_NOMEM;

	EXIT_FUNCTION();

	return avail;
}


/*
	This routine removes sequential free memory blocks to
	make one contiguous free memory block.
*/
void	fix_mem_list(void)
{
	extern struct LOL lol;

	struct MCB far *mcb, far *mcb2;
	int found;

	ENTER_FUNCTION();

	/* go through all MCBs, looking for contiguous free blocks */
	do
	{
		/* start at the beginning again */
		mcb = FirstMCB();
		found = FALSE;

		while(1)
		{

			if(last_mcb(mcb))
				break;

			/* look for a free block */
			if(mcb->owner == FREEMEM)
			{
				mcb2 = next_mcb(mcb);

				/* combine adjacent free blocks */
				while(mcb2->owner == FREEMEM)
				{

					/* build the new combined MCB */
					mcb->id = mcb2->id;

					mcb->para_cnt += mcb2->para_cnt+1;
					found = TRUE;

					/* this was the last MCB, we are done */
					if(last_mcb(mcb))
					{
						EXIT_FUNCTION();
						return;
					}

					mcb2 = next_mcb(mcb);
				}
			} /* end of if */

			mcb = next_mcb(mcb);
		} /* end of while */

	} while(found);

	EXIT_FUNCTION();
}



/*
	Free all memory blocks "owned" by this process.  Then
	concatinate all free MCBs.
*/
void	mem_term_free(void)
{
	extern struct LOL lol;
	extern unsigned current_psp;

	struct MCB far *mcb;

	ENTER_FUNCTION();

	mcb = FirstMCB();
	while(1)
	{

		/* if we own it, free it */
		if(mcb->owner == current_psp)
			mem_free(FP_SEG(mcb)+1);

		if(last_mcb(mcb))
			break;

		mcb = next_mcb(mcb);
	}

	/* concatinate all free memory blocks */
	fix_mem_list();

	EXIT_FUNCTION();
}


/*
	Change the ownership of a memory block.  No checking
	is performed for a valid memory block.
*/
void	PASCAL mem_new_owner(unsigned seg, unsigned owner)
{
	ENTER_FUNCTION();

	// validate segment
	ASSERT(seg);

	((struct MCB far *)MK_FP(seg-1, 0))->owner = owner;

	EXIT_FUNCTION();
}
