//
//   $Id: DOSSDA.C 1.19 2003/08/22 23:02:48Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DOSSDA.C  created 11/29/90  by Drew Gislason

   Copyright (c) 1990-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This file contains all the hodge-podge routines for
	entering and leaving a DOS call.  This mainly concerns
	setting variables in the DOS SDA and returning those
	variables back to the user space.

   Glossary   Meaning
   -------------------------
   BPB        Bios Paramenter Block
   CDS        Current Directory Structure
   DCB        Device Control Block (tbl of SFTs)
   DPB        Drive Parameter Block
   LOL        List Of Lists
   MCB        Memory Control Block
   PSP        Program Segment Prefix
   SDB        Search Data Block
   SFT        System File Table

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"

#define __FILENUM__ 15


/*
	These flags allow the regs_to_sda() to validate a
	variety of things common to many functions.  For
	instance, open, create, delete, rename all have a
	drive, path and name that can be parsed into the SDA.

   The flags are 16 bits, the upper 8 bits deal with an
   IFS call, the lower 8 bits are defined below:

	Category		Meaning						Error on failure
	---------	----------------			-----------------
	xxxx xxx1	Item has drive/path		(3) Path not found
	xxxx xx1x	Item may be a device		(5) Access denied
   xxxx x1xx	Item name must exist		(2) File not found
	xxxx 1xxx	Item has FCB
	xxx1 xxxx	File has handle			(6) Bad handle
	xx1x xxxx	File does char I/O
	x1xx xxxx	Item is a rd/wr call
	1xxx xxxx	No wildcards in name		(2) File not found

	Upper byte (non-zero means possible IFS call)

	The callf_xxx() group of functions allows access to
	these bit fields.
*/

unsigned doscallflag[MAXCALL+1] =
{
/* 00H */	0x0000, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0020,
				0x0020, 0x0020, 0x0020, 0x0020, 0x0020, 0x0000, 0x0000, 0x161F,
/* 10H */	0x0618, 0x1B0B, 0x1C08, 0x1309, 0x0858, 0x0958, 0x171B, 0x1109,
				0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/* 20H */	0x0000, 0x0858, 0x0958, 0x0F0D, 0x0008, 0x0000, 0x0000, 0x0858,
				0x0958, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/* 30H */	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0C00, 0x0000,
				0x0000, 0x0301, 0x0105, 0x0581, 0x1713, 0x1617, 0x0610, 0x0850,
/* 40H */	0x0950, 0x1305, 0x0010, 0x0F05, 0x0000, 0x0010, 0x0010, 0x0000,
				0x0000, 0x0000, 0x0000, 0x0005, 0x0000, 0x0000, 0x1B03, 0x1C00,
/* 50H */	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x1105, 0x0010,
				0x0000, 0x0000, 0x1711, 0x1791, 0x0A10, 0x0000, 0x0000, 0x1E00,
/* 60H */	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
/* 68H */	0x0710, 0x0000, 0x0000, 0x0000, 0x2E13
};



/*
	-------------------------------------------------------

			CONVERT USER REGISTERS INTO SDA VARIABLES

	-------------------------------------------------------
*/



/*
	DOS calls 01H through 0CH all use a handle to perform
	their I/O.  This means they can be redirected.
*/
uchar iohandle[0xC] = { 0,1,3,3, 4,0,0,0, 1,0,0,0 };


/*
	Convert an INT21H call into status FLAGS and the
	IFS call.
*/
unsigned PASCAL callf_flagsof(uchar call)
{
	extern unsigned doscallflag[];

	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return (call <= MAXCALL ? doscallflag[call] : 0);
}

/*
	Based on the current user registers (AX in particular),
	set the internal callflags.
*/
void	PASCAL get_call_flags(void)
{
	extern uchar int21_ah;
	extern unsigned int21_ax;
	extern unsigned sda_callflags;

	ENTER_FUNCTION();

	int21_ax     = R_AX;
	sda_callflags= callf_flagsof(int21_ah);

	EXIT_FUNCTION();
}


/*
	This function sets up the SDA upon the start of a DOS
	call.  It never calls to the disk, but merely performs
	memory manipulations on strings and structures.  The fact
	that it does no disk work allows this code to be used on
	DOS call that go to the IFS (network).

	Sets error_no if failed.
*/
void	regs_to_sda(void)
{
	/*
		[ externals ]
	*/
	extern char sda_isnet;
	extern int error_no;
	extern int sda_no24;
	extern char tmp_name[];
	extern uchar CritErr;
	extern uchar sda_drive;
	extern uchar int21_ah;
	extern uchar int21_al;
	extern uchar sda_srchattr;
	extern uchar sda_extfcb;
	extern uchar sda_fcbattr;
	extern uchar sda_chardev;
	extern uchar sda_openmode;
	extern uchar sda_sft_h;
	extern uchar far * sda_fcbptr;
	extern uchar far * DTA;
	extern uchar far * sda_rdwrbuff;
	extern uchar sda_5a_strlen;
	extern struct LOL lol;
	extern SDB sda_sdb;
	extern LPSFT sda_sftptr;
	extern LPCDS sda_cdsptr;
	extern FCB sda_fcb;
	extern DIR sda_ffb;
	extern unsigned sopen_act;
	extern unsigned sopen_action;
	extern unsigned sopen_attr;
	extern unsigned sopen_mode;
	extern unsigned sda_callflags;
	extern unsigned sda_rdwrlen;
	extern unsigned sda_rdwrfcb;
	extern unsigned sda_handle;
	extern uchar sda_isdevpath;
	extern unsigned current_psp;
	extern unsigned err_psp;
	extern uchar sda_in_5D;

	/*
		[ locals ]
	*/
	char far * path;
	uchar drive;
	ulong offset;
	uchar tmp;
	unsigned seg,off;
	int len;
	char far * lpTemp;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// The SDA variables that are ALWAYS set.
	// DO NOT clear sda_srchattr, because NWCDEX.EXE requires it
	// to be left alone between IFS FindFirst and IFS FindNext
	get_call_flags();
	sda_extfcb   = sda_chardev = sda_ffb.attr = 0;
	err_psp = current_psp;

	/* character I/O functions need the sda_sftptr */
	if(callf_chario())
	{
		tmp = iohandle[int21_ah - 1];
		if(int21_ah == 6 && R_DL!=0xFF)
			tmp = STDOUT;
		if(CritErr && tmp<=STDOUT)
			tmp = STDERR;
		is_bad_handle(tmp);
		EXIT_FUNCTION();
		return;
	}

	/* copy the user FCB to our local one */
	if(callf_fcb())
	{
		// copy the FCB into our own space so we can access
		// it with near pointers
		fcb_to_sda();

#if USING_FAT32
		if(!defdrive_to_sda(sda_fcb.drive))
		{
			error_no = ERR_NOTPATH;
			EXIT_FUNCTION();
			return;
		}
#endif
	}

	/* this DOS call has a drive/path/entry, place it into the SDA */
	if(callf_path())
	{
		/* get a pointer to the path */
		if(int21_ah == 0x6C)
			path = MK_FP(R_DS, R_SI);
		else
			path = MK_FP(R_DS,R_DX);
		if(callf_fcb())
		{
			fcb_to_dotname(tmp_name, TRUE);
			path = (char far *)tmp_name;
		}

		/* create new file takes a special path */
		if(int21_ah == 0x5A)
		{
			fmemcpy((void far *)tmp_name, path, 66);

			/* if it does not end in a \, make it so */
			sda_5a_strlen = len = strlen(tmp_name);
			if(!len || lastchar(tmp_name) != '\\')
			{
				tmp_name[len++] = '\\';
			}

			/* copy a "sample" filename to this string */
			create_unique_name(tmp_name+len);

			//  Copy generated name to user path!
			//
			fmemcpy( path + sda_5a_strlen, tmp_name + sda_5a_strlen, 9 );

			path = (char far *)tmp_name;
		}

		/* make sure sda_srchattr is set for all calls */
		if(int21_ah == 0x3C)
		{
			if(R_CL & A_SUBDIR)
			{
				error_no = ERR_DENIED;
				EXIT_FUNCTION();
				return;
			}
		}

		//  BT 1-31-2000 ETR 3272
		//
		//  Filenames (even devices) cannot contain a colon.
		//  Ex: a:\dir\name:
		//
		lpTemp = path;
		if( lpTemp[1] == ':' )
		{
			lpTemp += 2;
		}
		while( *lpTemp )
		{
			if( *lpTemp == ':' )
			{
				if( (int21_ah == 0x6c) || (int21_ah == 0x3c) )
				{
					error_no = ERR_DENIED;
				}
				if( int21_ah == 0x4e )
				{
					error_no = ERR_NOTPATH;
				}
				else
				{
					error_no = ERR_NOTFOUND;
				}

				EXIT_FUNCTION();
				return;
			}
			lpTemp++;
		}

		/* the path string must look reasonable */
		if(!path_to_sda(path,TRUE))
		{
			error_no = ERR_NOTPATH;
			EXIT_FUNCTION();
			return;
		}

      // do not allow wildcards if NAME must exist
      //  IsLegalDOSChar() and the nonLFN equivalent don't catch wildcards
      //  so, an additional check is necessary at this time
      //
		if(!sda_isnet && callf_nowild() && wild_name())
		{
			// Allow wildcard handle delete and rename calls
			// when inside server callbacks
			//
			if((sda_in_5D) && ((int21_ah == 0x41) || (int21_ah == 0x56)))
				{;}
			else
			{
				error_no = ERR_NOTFOUND;
				EXIT_FUNCTION();
				return;
			}
		}

		/* Create and Open have a path and return a handle */
		if(callf_handle())
		{
			/* get open mode */
			if(int21_ah == 0x3D)
			{
				sda_openmode = int21_al;
				if(int21_al & O_NOCHILD)

				if((sda_openmode&O_RESERVED) || share_mode(sda_openmode)>4)
				{
					error_no = ERR_BADCODE;
					EXIT_FUNCTION();
					return;
				}
			}

			/*
				Open mode for FCB Open is rdwr. For 6CH we don't care
				because it's handled by later logic. DON'T use CX for
				openmode with 6CH (like is done for 3CH) because CX may
				be ANYTHING.
			*/
			else if(int21_ah == 0x6C)
			{
				sda_openmode = R_BX;
			}
			else if(int21_ah == 0x0F)
			{
				sda_openmode = O_RDWR;
			}
			else
			{
				/* must be a create (3CH or 16H) */

				if(callf_fcb())
					sda_openmode = sda_srchattr;
				else
					sda_openmode = R_CL;
				sda_openmode |= A_ARCHIVE;

				// don't allow creates of subdirectories
				if(sda_openmode & A_SUBDIR)
				{
					error_no = ERR_DENIED;
					EXIT_FUNCTION();
					return;
				}
			}

			/*
				Allocate an SFT for use. (but not for Volume labels,
				unless an FCB call)
			*/
			if(!(sda_openmode & A_VOLUME) || int21_ah <= DOS1CALL)
			{
				if(!open_to_sda(sda_openmode))
				{
					error_no = ERR_NOHANDLE;
					EXIT_FUNCTION();
					return;
				}
			}
		}
	} /* end of if path */

	/* if this function uses an SFT */
	else if(callf_handle())
	{
		/* FCB style handle */
		if(callf_fcb())
		{
			if(!fcb_exist())
			{
				error_no = 1;
				EXIT_FUNCTION();
				return;
			}
		}

		/* DOS2 style handle call */
		else if(is_bad_handle(R_BX))
		{
			EXIT_FUNCTION();
			return;
		}
	}

	/* those functions that read and write files */
	if(callf_rdwr())
	{
		if(rdwr_denied())
		{
			error_no = ERR_DENIED;
			EXIT_FUNCTION();
			return;
		}

		/* determine read/write length and buffer location */
		if(callf_fcb())
		{
			if(!sda_fcb.record_size)
			{
				error_no = 1;		// can't have record size of 0
				EXIT_FUNCTION();
				return;
			}

			/* get # of records */
			sda_rdwrlen = int21_ah >= 0x27 ? R_CX : 1;
			sda_rdwrlen *= sda_fcb.record_size;
			sda_rdwrfcb = sda_rdwrlen;
			sda_rdwrbuff = DTA;

			/* don't allow writing over the program */
			if(DTA == MK_FP(current_psp, 0x80) && sda_rdwrlen > 0x80)
			{
				error_no = 2;	/* special FCB error */
				EXIT_FUNCTION();
				return;
			}

			/* seek to the appropriate place in the file */
			if(int21_ah <= 0x15)
				offset = (ulong)(sda_fcb.cur_blk*128 + sda_fcb.cur_record);
			else
				offset = sda_fcb.rand_record;
			offset = _lmulu(offset&0xFFFFFFL, sda_fcb.record_size);
			internal_seek(offset, SK_BEG);
		}

		/* determine read/write length and buffer location */
		else
		{
			sda_rdwrlen = R_CX;
			sda_rdwrbuff = MK_FP(R_DS,R_DX);
		}

		/* normalize the rdwr buffer */
		seg = FP_SEG(sda_rdwrbuff) + (FP_OFF(sda_rdwrbuff)>>4);
		off = FP_OFF(sda_rdwrbuff) & 0xF;
		sda_rdwrbuff = MK_FP(seg,off);

		/* don't let length wrap */
		if(off + sda_rdwrlen < sda_rdwrlen)
			sda_rdwrlen = 0 - off;
	}

	/* some SDA setup does not fall into above categories */
	switch(int21_ah)
	{
		/* Close File */
		case 0x3E:
			// SPR 1766 (de) free the handle for DOS2 calls (even IFS)
			// do this BEFORE the actual close in case of crit error
			//
			set_handle_val(sda_handle, 0xFF);
			break;

		/* Special Open/Create */
		case 0x6C:
			//
			// NOVEL expects the extended open action code to be in the SDA
			// at FAR_DATA:5F4h
			//
			sopen_act  = R_DL;
			sopen_action = sopen_act;
			sopen_attr = R_CX;
			sopen_mode = R_BX;
			/* if action is not valid, return bad function */
			if(sopen_act==0 || (sopen_act&ACT_BADFUNC) || (sopen_act&3)==3)
			{
				error_no = ERR_BADFUNC;
				EXIT_FUNCTION();
				return;
			}
			sda_no24 = (sopen_mode & O_NO24);
			if( (sopen_act & ACT_EXIST_OPEN) == 0 )
			{
				if( sopen_attr & A_SUBDIR )
				{
					if( (sopen_attr & A_VOLUME) == 0 )
					{
						error_no = ERR_DENIED;
						EXIT_FUNCTION();
						return;
					}
				}
			}
			break;

		/* Rename File(s) (both DOS1 & DOS2) */
		case 0x17:
		case 0x56:

			/* save the drive of the first name */
			drive = sda_drive;

			/* make sure 2nd path name is legal (does not check if it exists) */
			path = MK_FP(R_ES,R_DI);
			if(callf_fcb())
			{
				fcb_to_dotname(tmp_name, FALSE);
				path = (char far *)tmp_name;
			}
			if(!path_to_sda(path, FALSE))
			{
				error_no = ERR_NOTPATH;
				EXIT_FUNCTION();
				return;
			}

			/* must be same drive for a rename to work */
			if(sda_drive != drive)
			{
				error_no = ERR_NOT_SAME_DV;
				EXIT_FUNCTION();
				return;
			}
			break;

		/* drive functions where DL=drive: 0=def, 1=A, 2=B... */
		case 0x1B:
		case 0x1C:
		case 0x32:
		case 0x36:

			/* make sure this is a valid drive */
			if(!defdrive_to_sda(R_DL))
			{
				error_no = ERR_NOTPATH;
				EXIT_FUNCTION();
				return;
			}
			/* if this is not a network call, it can't happen on a netdrive */
			if(!callf_ifs() && (sda_cdsptr->status & C_NETWORK))
			{
				error_no = ERR_NOTPATH;
				EXIT_FUNCTION();
				return;
			}
			break;

		/* Set File Attribute 4301H */
		case 0x43:

			/* do not allow invalid attributes */
			if(int21_al == 1 && (R_CX & 0xFFD8))
			{
				error_no = ERR_DENIED;
				EXIT_FUNCTION();
				return;
			}
			break;

		/* Find First */
		case 0x4E:
			sda_fcbattr = sda_srchattr = R_CL;
			break;

		/* Find Next */
		case 0x4F:
			fmemcpy((void far *)&sda_sdb, DTA, sizeof(SDB));
			goto checkdrive;

		/* DOS1 (FCB) Find Next */
		case 0x12:
			/* get SDB from user's FCB */
			memmove(&sda_sdb, &sda_fcb, sizeof(SDB));

			// MSDOS 6.22 stores IFS drive number following SDB
			// in user's FCB (see DOSSTRUCT.H)
			//
			if( *((unsigned char*)(&sda_fcb)+0x15) & SDB_NETDRIVE )
				sda_sdb.drive = *((unsigned char*)(&sda_fcb)+0x15);

			/* make sure drive is still valid */
checkdrive:
			if((sda_sdb.drive&0x80) == 0 && !defdrive_to_sda(sda_sdb.drive))
			{
				error_no = ERR_LASTFILE;
				EXIT_FUNCTION();
				return;
			}
			break;
	}


	//  Determine if this is an FCB call.
	//
	if(callf_fcb())
	{
#if USING_FAT32
		//
		// MS-DOS v7.1 denies certain FCB functions access to FAT32 drives.
		//
		if( !sda_isdevpath )
		{
			//  Volume label calls are excepted, other fail that are not
			//  for devices
			//
			if( sda_fcbattr != 0x08 )
			{
				if( IsFat32Drive(sda_dpbptr))
				{
					switch(R_AH)
					{
					case 0x0f:		// Func 0F FCB Open
					case 0x10:     // Func 10h FCB Close
					case 0x14:     // Func 14 FCB Read Sequential
					case 0x15:     // Func 15 FCB Write Sequential
					case 0x16:     // Func 16 Fcb Create File
					case 0x21:     // Func 21 FCB Random Read
					case 0x22:     // Func 22 FCB Random Write
					case 0x24:     // Func 24 FCB Set Random Record
					case 0x27:     // Func 27 FCB Random Read
					case 0x28:     // Func 28 FCB Random Write
						error_no = ERR_DENIED;
						EXIT_FUNCTION();
						return;
					default:
						break;
					}
				}
			}
		}
#endif
	}

	EXIT_FUNCTION();
}


/*
	Return TRUE if read or write access is denied.
*/
bool	rdwr_denied(void)
{
	/*
		[ externals ]
	*/
	extern LPSFT sda_sftptr;
	extern uchar int21_ah;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* if this is a read call, and opened for write only, denied */
	if(IsFcbRead() || int21_ah == 0x3F)
	{
		if(sda_sftptr->rwmode == O_WRONLY)
		{
			EXIT_FUNCTION();
			return TRUE;
		}
	}

	/* if write call and opened for read only, denied */
	else if((sda_sftptr->rwmode&3) == 0)
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	EXIT_FUNCTION();

	/* access is NOT denied */
	return FALSE;
}



/*
	-------------------------------------------------------

			RETURN SDA INFORMATION BACK TO USER

	-------------------------------------------------------
*/

/*
	This function will take care of all common conversions
	after the "meat" of a DOS call has already been handled.
	It does no disk access, but merely moves or translates
	structures in memory.
*/
void	sda_to_regs(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char sda_fname1[];
	extern uchar sda_sft_h;
	extern uchar int21_ah;
	extern uchar far * sda_fcbptr;
	extern SDB sda_sdb;
	extern FCB sda_fcb;
	extern LPCDS sda_cdsptr;
	extern LPSFT sda_sftptr;
	extern DIR sda_ffb;
	extern unsigned sda_handle;
	extern unsigned sda_callflags;
	extern unsigned sda_rdwrlen;
	extern unsigned sda_rdwrfcb;
	extern uchar sda_5a_strlen;
	extern uchar far * sda_rdwrbuff;
	extern unsigned sda_rdwrlen;
	extern unsigned sda_rdwrfcb;
	extern int sda_no24;


	/*
		[ locals ]
	*/
	unsigned tmp;
	char far * dstname;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set call flags in case they have changed */
	sda_callflags = callf_flagsof(int21_ah);

	switch(int21_ah)
	{
		/* create unique filename, copy filename to user space */
		case 0x5A:
			dstname = MK_FP(R_DS,R_DX+sda_5a_strlen);
			if(sda_5a_strlen==0 || *(dstname-1)!= '\\')
				*dstname++ = '\\';
			/* 8 for filename, 1 for null */
			fmemcpy(dstname,(void far *)(sda_fname1+strlen(sda_fname1)-8),9);
		break;

		/* Change Directory */
		case 0x3B:
			fmemcpy(sda_cdsptr->cwd,(void far *)sda_fname1,strlen(sda_fname1)+1);
			break;

		case 0x4E:
		case 0x4F:
			R_AX = 0;
			dos2_ffb_to_dta();
			break;

			// SPR 1638 (de) With MS-DOS v6.22, the "no int 24"
			//	option ONLY affects the open/create. Subsequent
			//	writes, commits, etc. all generate crit errors.
			// regardless of whether the SFT rwmode O_NO24 bit
			// is set or not.
			//
		case 0x6C:
			sda_no24 = FALSE;
			break;
	}

	/* open/create must allocate the SFT and set up handles */
	if(callf_path() && callf_handle() && !(sda_ffb.attr&A_VOLUME))
	{
		/* actually allocate the SFT */
		sda_sftptr->handles = 1;

		if(callf_fcb())
		{
			memset(&sda_fcb.cur_blk, 0, 0x20-0x0C);
			sda_fcb.record_size = 0x80;
			sft_to_fcb();
		}

		/* set up the handle in the current PSP */
		else
		{
			set_handle_val(sda_handle, sda_sft_h);
			R_AX = sda_handle;
		}
	}

	/* for all read/write calls, return the length */
	if(callf_rdwr() && !callf_chario())
	{
		/* update the record #s in the FCB */
		if(callf_fcb())
		{
			/* update sequential record */
			if(int21_ah <= 0x15)
			{
				++sda_fcb.cur_record;
				if(sda_fcb.cur_record >= 128)
				{
					sda_fcb.cur_record = 0;
					++sda_fcb.cur_blk;
				}
			}

			/* update random record */
			else
			{
				if(int21_ah >= 0x27)
				{
					tmp = sda_rdwrlen / sda_fcb.record_size;
					sda_fcb.rand_record += tmp;
					R_CX = tmp;
				}
				sda_fcb.cur_blk    = (unsigned)_lshru(sda_fcb.rand_record, 7);
				sda_fcb.cur_record = (unsigned)(sda_fcb.rand_record & 0x7F);
			}

			/* make sure file size is correct */
			sda_fcb.file_size = sda_sftptr->size;

			/* partial or incomplete read/write */
			if(sda_rdwrlen != sda_rdwrfcb)
			{
				// no read FCB error
				if(!sda_rdwrlen || sda_fcb.record_size == 1)
					error_no = 1;

				// partial read FCB error
				else
				{
					// partial read, fill in rest of record with 0s
					if(IsFcbRead())
						fmemset(sda_rdwrbuff + sda_rdwrlen, 0, sda_rdwrfcb-sda_rdwrlen);

					// FCB partial read error code
					error_no = 3;
				}
			}
		}

		/* DOS 2 style call */
		else
		{
			R_AX = sda_rdwrlen;
		}
	}


	/* DOS1 FindFirst/Next */
	if(int21_ah == 0x11 || int21_ah == 0x12)
	{
		// fill in the FCB with everything but the drive
		// sets up the FCB for a findnext
		memmove((char *)(&sda_fcb)+1,
				(char *)(&sda_sdb)+1, sizeof(SDB)-1);

		// If "local" move in the drive too. MSDOS 6.22 stores
		// IFS drive number following SDB in user's FCB
		// (see DOSSTRUCT.H), zero it out for local drives.
		//
		if( !(sda_sdb.drive & SDB_NETDRIVE) )
		{
			sda_fcb.drive = sda_sdb.drive;
			*((unsigned char*)(&sda_fcb)+0x15) = 0;
		}
		else
			*((unsigned char*)(&sda_fcb)+0x15) = sda_sdb.drive;
	}

	/*
		IMPORTANT!! some programs depend on the fact that
		DOS fills the user FCB BEFORE it fills the user DTA.
	*/
	if(callf_fcb())
	{
		/* move the local FCB into user space */
		fmemcpy(sda_fcbptr, (void far *)&sda_fcb, sizeof(FCB));
	}

	/* Find First/Next - DTA must be filled in AFTER the FCB */
	if(int21_ah == 0x11 || int21_ah == 0x12)
		dos1_ffb_to_dta();

	EXIT_FUNCTION();
}


/*
	Decrement the handle count for the current SFT.
	Set to FFFF, not 0 if last count.
*/
void	decr_handle_count(void)
{
	extern LPSFT sda_sftptr;
	extern unsigned current_psp;

	int handles;

	ENTER_FUNCTION();

	/* free the SFT itself */
	handles = sda_sftptr->handles;
	if(handles > 1 || (handles > 0 && sda_sftptr->owner == current_psp))
		--sda_sftptr->handles;

	EXIT_FUNCTION();
}

/*
	Get prepared for an open or a create.  This allocates an
	SFT.  It understands the difference between an FCB style
	call and a DOS2 style call, and sets the current "mode"
	and "owner."

	Returns FALSE if failed, TRUE if worked.
*/
bool	PASCAL open_to_sda(uchar mode)
{
	/*
		[ externals ]
	*/
	extern uchar sda_openmode;
	extern LPSFT sda_sftptr;
	extern unsigned sda_handle;
	extern unsigned current_psp;
	extern unsigned machine_id;
	extern unsigned sda_callflags;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* set create/open mode */
	sda_openmode = mode;

	/* make sure an SFT is available */
	if(callf_fcb())
	{
		if(!fcb_get_free_sft())
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}

	/* make sure a handle is available (but don't allocate it) */
	else
	{
		/* find an unused SFT */
		if(!get_new_sft())
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		sda_handle = get_new_handle();
		if(sda_handle == 0xFFFF)
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}

	/* set up the SFT ownership */
	sda_sftptr->owner = current_psp;
	sda_sftptr->shareid = machine_id;

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Is this handle OK, or invalid?  If it is invalid,
	free the handle.
*/
bool	PASCAL is_bad_handle(unsigned handle)
{
	extern int error_no;

	ENTER_FUNCTION();

	if(!handle_to_sda(handle))
	{
		set_handle_val(handle, 0xFF);
		error_no = ERR_BADHANDLE;
		EXIT_FUNCTION();
		return TRUE;
	}

	EXIT_FUNCTION();

	return FALSE;
}

/*
	Make sure that handle exists, and that it points to a
	valid SFT.  Points the internal SDA structures to that
	SFT.

	Returns FALSE if failed, TRUE if worked.
*/
bool	PASCAL handle_to_sda(unsigned handle)
{
	/*
		[ externals ]
	*/
	extern uchar sda_sft_h;
	extern unsigned sda_handle;

	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* if handle is invalid, or is free (not used), there is an error */
	sda_handle = handle;

	/* is the handle itself "open"? */
	sda_sft_h  = get_handle_val(handle);
	if(sda_sft_h == 0xFF)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	EXIT_FUNCTION();

	/* does this handle point to a valid SFT? */
	return h_to_sda(sda_sft_h);
}


/*
	Make sure that SFT is valid.  Points the internal SDA
	structures to that SFT.

	Returns FALSE if failed, TRUE if worked.
*/
bool	PASCAL h_to_sda(uchar h)
{
	/*
		[ externals ]
	*/
	extern int sda_no24;
	extern uchar sda_sft_h;
	extern uchar sda_chardev;
	extern LPSFT sda_sftptr;
	extern LPDPB sda_dpbptr;
	extern LPCDS sda_cdsptr;
	extern LPDEV_HDR sda_devptr;


	/*
		[ locals ]
	*/
	unsigned ioctl;


	/*
		[ code ]
	*/

   ENTER_FUNCTION();

	/* make sure the SFT exists */
	sda_sftptr = h_to_sft(h);
	if((void *)sda_sftptr == NULLF)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/* make sure the SFT is valid */
	if(sda_sftptr->handles == 0)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/* record which SFT structure it is */
	sda_sft_h = h;

	// SPR 1638 (de) With MS-DOS v6.22, the "no int 24"
	//	option ONLY affects the open/create. Subsequent
	//	writes, commits, etc. all generate crit errors
	// regardless of whether the SFT rwmode O_NO24 bit
	// is set or not.
	//
	// /* some handles won't generate a critical error */
	//	sda_no24 = sda_sftptr->rwmode & O_NO24;

	/* fill out the rest of the SDA variables */
	ioctl = sda_sftptr->ioctl;
	sda_devptr = sda_sftptr->device;
	sda_chardev = (ioctl & IO_DEVICE) ? (0x80 | (uchar)ioctl) : 0;

	/* set the CDS if a FAT drive (not IFS or CHAR) */
	if( !(ioctl & (IO_NETWORK | IO_DEVICE)) )
		drive_to_sda(ioctl&0x1F);

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Is this path a "Device Path", that is "CON" and "\DEV\CON" mean
	this is just a device.  The device name may be ANY character
	device.  A "Device Path" means no disk access will occur.
*/
bool IsDevPath(char far *lpPath, char *pFileName, char *pExpName)
{
	extern unsigned sda_callflags;
	extern uchar sda_isdevpath;

	char far *lpChar;

	ENTER_FUNCTION();

	// assume this is NOT a device path
	sda_isdevpath = FALSE;

	// if this call can't handle a device, then it's NOT a
	// device path
	if( !callf_dev() )
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// First, if path begins with "\DEV\", remove it.
	if( isslash(lpPath[4]) && toupper(lpPath[1]) == 'D' &&
			toupper(lpPath[2]) == 'E' &&
			toupper(lpPath[3]) == 'V' &&
			isslash(lpPath[0]) )
	{
		// skip the "\DEV\"
		lpPath += 5;

		// skip any trailing slashes
		while(isslash(*lpPath))
			++lpPath;
	}

	// next, must be no more backslashes.  That is, "CON" or
	// "\DEV\CON" will be found, but "\DEV\JUNK\..\CON" is
	// NOT a device path.
	lpChar = lpPath;
	while(*lpChar)
	{
		if(isslash(*lpChar))
		{
			EXIT_FUNCTION();
			return FALSE;
		}
		++lpChar;
	}

	// parse into sda paths
	dot_to_ename(pExpName, lpPath, FALSE);
	ename_to_dot(pFileName, pExpName);

	// does the device even exist?
	if( !find_device(pExpName) )
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// all tests passed. This MUST be a device!
	sda_isdevpath = TRUE;

	EXIT_FUNCTION();

	return TRUE;
}

/*
	This function will convert a path into SDA variables.
	It checks to make sure the drive is valid, but does no
	validation on directories and such.

	Returns FALSE if failed, TRUE if worked.
*/
bool PASCAL path_to_sda(char far * orgpath, int first)
{
	/*
		[ externals ]
	*/
	extern uchar sda_drive;
	extern char sda_isnet;
	extern int error_no;
	extern char sda_fname1[];
	extern char sda_ename1[];
	extern char sda_fname2[];
	extern char sda_ename2[];
	extern uchar CUR_DRIVE;
	extern uchar sda_chardev;
	extern uchar sda_srchattr;
	extern FCB sda_fcb;
	extern LPDEV_HDR sda_devptr;
	extern LPCDS sda_cdsptr;
	extern unsigned sda_callflags;
	extern uchar sda_in_5D;

	/*
		[ locals ]
	*/
	uchar drive;
	char *fname;		// ptr to the asciiz filename (ie C:\TEMP\NAME.EXT)
	char *ename;		// ptr to the expande name (ie "NAME    EXT")
	char far * path;
	uchar skip;
	char * p;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(orgpath);

	// determine WHICH place in the SDA to put the filename
	// rename uses the 2nd name, all others use the 1st
	fname = first ? sda_fname1 : sda_fname2;
	ename = first ? sda_ename1 : sda_ename2;

	/* see if the IFS wants to validate the path */
	if(!sda_in_5D)
	{
		// invalidate ename, so it's not left over from previous call
		// invalidate fname too
		memset( ename, ' ', 11 );
		*fname = 0;

		// Invalidate the CDS ptr. This is required for NT clients
		// IFF the specified path is absolute.
		//
		(void *)sda_cdsptr = NULLF;

		// ask any IFS drivers that might be listening if it want's
		// to handle the call
		generate_int2F(0x23, 0, 0, (void far *)fname, orgpath);

		// If there was no error then the IFS handled resolving the
		// pathname and it will have filled in the sda_cdsptr appropriately
		//
		if(!error_no && *fname)
		{
			// let other routines know the filename is now special,
			// and should be left alone (set sda_isnet TRUE).
			if(callf_xlatifs())
				++sda_isnet;
			goto PathValidated;
		}

		// If the path name begins with "\\" (or "\\\\" in C)
		// then it is an absolute path for which there is no directly
		// corresponding CDS and so NT servers will fail all 2f-1123 calls
		// with an absolute path. The client software also insists that the
		// offset value for the sda_cdsptr be -1. On the other hand,
		// assuming O: is network drive, if the path was "O:\path",
		// "O:path" or "path" (and O: was the default drive) then
		// these would require a CDS -- even though any of the
		// latter may be equivalent to "\\path", that's just the way
		// things work.) MS-DOS will happilly pass garbage absolute paths to
		// the IFS and return its inevitable error code.
		// - DE, SPR1435, 9/30/97
		//
		else if((*orgpath) && (*(unsigned far*)(orgpath) == 0x5C5C))
		{
			if(callf_xlatifs())
				++sda_isnet;
			error_no = 0;

			// copy upto MAXNAMELEN chars from the user specified
			// absolute IFS path into the sda_fname1 var.
			//
			skip = fstrlen(orgpath);
			if(skip > MAXNAMELEN)
				skip = MAXNAMELEN;
			fmemcpy(fname, orgpath, skip);
			fname[skip] = 0;

			goto PathValidated;
		}

		// IFS may have returned error (I couldn't parse path)
		// that is not really an error, we will parse the path
		error_no = 0;

	}

	/* check if the path has a drive part to it */
	path = orgpath;
	if(*path && *(path+1)==':')
	{
		drive = toupper(*path)-'A';
		path += 2;
	}
	else
		drive = CUR_DRIVE;

	/*
		special case: "\DEV\CON" will be found as a device,
		even if drive is not valid.  For example "Z:\DEV\CON"
		will be found, even if no drive "Z:" or path "\DEV\".

		We've got to check for EXACTLY "\DEV\devname".  Nothing
		else will do.  The "devname" MUST be a valid device name
		for this special case to occur.
	*/
	if( IsDevPath(path, fname, ename) )
		goto PathValidated;

	/* is the drive a valid one? */
	if( !drive_to_sda(drive) )
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// validate the path by converting it into an absolute path
	// note: no drive part is in the path yet
	if(!rel_to_abspath(fname, path, ename))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/*
		special case: volume labels MUST begin at root.  If the
		drive is SUBST it may have a "root" of "C:\TEMP".  We
		want just the drive letter, however.  This only works on
		local (FAT) drives, not network drives.
	*/
	if( sda_srchattr == A_VOLUME && callf_fcb() &&
			!(sda_cdsptr->status & C_NETWORK) )
	{
		// skip size is 2, because we are doing drive: "D:"
		skip = 2;

		// move the path part over to accomodate drive
		memmove(fname+skip, fname, strlen(fname)+1);

		// put in drive letter
		fname[0] = sda_cdsptr->cwd[0];
		fname[1] = ':';
	}

	// network call? leave name part alone
	// just inform the world it's a network call
	else if(*(unsigned *)(fname) == 0x5C5C)
		++sda_isnet;

	// normal case, get root from CDS
	// the final path will look something like "C:\SUBDIR\FILE.EXT"
	else
	{
		// move path part over to accomodate drive "root" string
		skip = sda_cdsptr->skip;
		memmove(fname+skip, fname, strlen(fname)+1);

		// copy drive into the given path
		fmemcpy(fname, sda_cdsptr->cwd, skip);
	}

	/*
		Special case: When we must parse the path, and are at
		the root, remove backslash if "Novell Netware" style, but
		do NOT remove backslash if "normal" drive style root (eg
		Sicom's redirector).  Only affects root directory.

		Novell's should be: "\\D.\A."   <--- note: no backslash
		Sicom's should be:  "D:\"       <--- note: backslash
	*/
	if(callf_ifs())
	{
		// remove backslash if a "\\network" style root
		// don't do anything if a "D:" style root
		p = fname + skip;
		if(*p == '\\' && *(p+1) == 0 && *(p-1) != ':')
			*p = 0;
	}

PathValidated:

	/* if the call allows for it, this path might point to a device */
	sda_chardev = 0;
	if(!sda_isnet && callf_dev() && find_device(sda_ename1))
		sda_chardev = 0x80 | (sda_devptr->attr & 0x1F);

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Default Drive to SDA.  This function is the same as
	drive_to_sda(), but drive 0 is converted to the default,
	drive 1 is A, 2=B, and so on.
*/
bool	PASCAL defdrive_to_sda(uchar drive)
{
	extern uchar CUR_DRIVE;

	ENTER_FUNCTION();

	if(!drive)
		drive = CUR_DRIVE;
	else
		--drive;

	EXIT_FUNCTION();

	return drive_to_sda(drive);
}


/*
	Make sure that the drive in question is valid for this
	DOS call.  It understands both FAT and NETWORK type
	drives.  Drive 0=A, 1=B, 2=C, etc...

	Returns TRUE if drive is valid, FALSE if not.
*/
bool	PASCAL drive_to_sda(uchar drive)
{
	/*
		[ externals ]
	*/
	extern unsigned sda_callflags;
	extern uchar sda_drive;
	extern struct LOL lol;
	extern LPCDS sda_cdsptr;
	extern LPDPB sda_dpbptr;
	extern LPDEV_HDR sda_devptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* check if the drive is valid */
	sda_cdsptr = drv_to_cds(drive);
	if(FP_OFF(sda_cdsptr) == 0xFFFF)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	/* set up the SDA */
	sda_drive  = drive;

	/* IFS drives do NOT have devices or DPBs */
	if(sda_cdsptr->status & C_NETWORK)
		(void far *)sda_dpbptr = (void far *)sda_devptr = 0;

	/* set up the DPB and device ptrs */
	else
		quick_drive_setup(drive);

	EXIT_FUNCTION();

	return TRUE;
}


/*
	Set up the minimum for doing abs_read and writes.
	Assumes that the drive is valid.
*/
void	PASCAL quick_drive_setup(uchar drive)
{
	extern LPCDS sda_cdsptr;
	extern LPDPB sda_dpbptr;
	extern LPDEV_HDR sda_devptr;
   extern char FAT_type;

	ENTER_FUNCTION();

	sda_cdsptr = drv_to_cds(drive);
	sda_dpbptr = sda_cdsptr->lpDPB;

   //  Initialize the FAT_type for get_fat_entry() and
	//  set_fat_entry()
   //
   if( sda_dpbptr->hiclus == 0 )
   {
      FAT_type = FS_FAT32;
   }
   else if( sda_dpbptr->hiclus <= HICLUS12 )
   {
      FAT_type = FS_FAT12;
   }
   else
   {
      FAT_type = FS_FAT16;
   }

	sda_devptr = sda_dpbptr->device;

	EXIT_FUNCTION();
}


/*
	Based on an expanded name, check if a device exists.
	Sets up the SDA to point to the device.

	Returns TRUE if found, FALSE if not.
*/
bool	PASCAL find_device(char *ename)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern LPDEV_HDR sda_devptr;


	/*
		[ locals ]
	*/
	LPDEV_HDR devptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	// validate ename
	ASSERT(ename);

	devptr = (void far *)(&lol.first_dev);
	do
	{
		/* is this device a character device, do the names match? */
		if((devptr->attr&DEVA_CHAR) &&
			fmemcmp(devptr->name,(void far *)ename,8) == 0)
		{
			sda_devptr = devptr;
			EXIT_FUNCTION();
			return TRUE;
		}
		devptr = devptr->next;
	} while((void *)devptr != (void *)0xFFFF);

	EXIT_FUNCTION();

	/* not found */
	return FALSE;
}


/*
	-------------------------------------------------------

			FUNCTIONS THAT GO OUT TO A FAT DRIVE

	-------------------------------------------------------
*/

/*
	After regs_to_sda() has been called, and the call is
	determined to be a FAT system call, this function
	will be called to validate some fields in the SDA.
*/
void	validate_sda(void)
{
	/*
		[ externals ]
	*/
	extern int error_no;
	extern char sda_fname1[];
	extern char sda_ename1[];
	extern char sda_fname2[];
	extern uchar sda_openmode;
	extern uchar int21_ah;
	extern unsigned int21_ax;
	extern uchar sda_chardev;
	extern uchar sda_srchattr;
	extern DIR sda_ffb;
	extern SDB ren_sdb;
	extern SDB sda_sdb;
	extern unsigned sda_callflags;
	extern LPCDS sda_cdsptr;
	extern LPDEV_HDR sda_devptr;
	extern uchar sda_drive;
	extern uchar sda_isdevpath;
	extern uchar sda_in_5D;

	/* locals */
	uchar skip;
	LPDEV_HDR org_devptr;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* don't validate "network" drives */
	if(sda_cdsptr->status & C_NETWORK)
	{
		if( !(sda_sdb.drive & SDB_NETDRIVE) )
		{
			if(int21_ah == 0x4F)
				error_no = ERR_LASTFILE;
		}
		EXIT_FUNCTION();
		return;
	}
	/*
		We MUST check the path when opening/creating, etc..
		a device.  It should NOT pass on invalid paths.

		A:\THERE\CON should always be found
		A:\NOT-THERE\CON should NEVER be found
		A:\DEV\CON should always be found (no disk access)
	*/
	if(callf_path() && !sda_isdevpath)
	{
		/*
			At this point, if it's a device, the "sda_devptr"
			points to the DEVICE not the drive.  Since we will
			be doing disk access, that needs to temporarily change.
		*/
		if(sda_chardev)
		{
			/* remember original device pointer */
			org_devptr = sda_devptr;

			/* set up the device pointer to point to the disk */
			quick_drive_setup(sda_drive);
		}

		/*
			Since we got here, it must be a DOS device (not
			IFS), so the path will be "normal".  That means the
			form of "C:\PATH\NAME".  We only want to skip the
			drive "C:" part in case this drive letter is SUBST.
			The "root" might be "C:\TEMP".
		*/
		skip = SKIPVAL;

		/* validate 2nd path for Rename */
		if(int21_ah == 0x17 || int21_ah == 0x56)
		{
			/* dir_valid requires no starting "\" */
			if(!dir_valid(sda_fname2 + skip + 1))
				goto not_path;
			else
				memmove(&ren_sdb, &sda_sdb, sizeof(SDB));
		}

		/* validate path.  does not check the name part */
		if(dir_valid(sda_fname1 + skip + 1))
		{
			/* special case: get file attr of "C:\" is OK */
#if USING_FAT32
			if(int21_ax==0x4300 && !sda_sdb.ulStartingCluster &&
					!name_ok_to_create())
			{
				sda_ffb.attr = A_SUBDIR;
				EXIT_FUNCTION();
				return;
			}
#else
			if(int21_ax==0x4300 && !sda_sdb.subdir && !name_ok_to_create())
			{
				sda_ffb.attr = A_SUBDIR;
				EXIT_FUNCTION();
				return;
			}
#endif
		}

		/* path was not valid, return error */
		else
		{
not_path:
			error_no = ERR_NOTPATH;
			EXIT_FUNCTION();
			return;
		}

		/*
			Restore device pointer if it pointed to a DEVICE,
			not a file.  We had to temporarily point "sda_devptr"
			to the disk device.  Now restore it back to the
			character device.
		*/
		if(sda_chardev)
			sda_devptr = org_devptr;
	}

	/* for some calls, the entry MUST exist */
	if(callf_name() && !sda_chardev)
	{
		// do not allow wildcards in a filename
		// unless we are deleting files from inside
		// a server callback
		//
		if( wild_name() )
		{
		 	if( sda_in_5D )
			{
			}
			else
			{
				if( int21_ah == 0x41 )
				{
					error_no = ERR_NOTFOUND;
					EXIT_FUNCTION();
					return;
				}
			}
		}

		/* if entry "found" was device, deny access */
		else if(find_device(sda_ename1))
		{
			if(int21_ax == 0x4300)
			{
				error_no = ERR_NOTFOUND;
				EXIT_FUNCTION();
				return;
			}
			error_no = ERR_DENIED;
			EXIT_FUNCTION();
			return;
		}

		/* if the file entry isn't on disk, give up */
		else
		{
			if(!dir_findfirst())
			{
				// we didn't find the specified file/subdir within the
				// specified subdir.
				//
				// If rmdir() and the specified subdir is the root...
#if USING_FAT32
				if(int21_ah==0x3A && !sda_sdb.ulStartingCluster)
#else
				if(int21_ah==0x3A && !sda_sdb.subdir)
#endif
					// if the root is the current dir and the root
					// was specified as the dir to delete
#if USING_FAT32
				   if((sda_sdb.ulStartingCluster == sda_cdsptr->ulSubDirStartingCluster) &&
#else
				   if((sda_sdb.subdir == sda_cdsptr->subdir) &&
#endif
						((strlen(sda_fname1)-1) == sda_cdsptr->skip))
							error_no = ERR_CURDIR;
					else
						// can't find subdir to delete
						error_no = ERR_NOTPATH;
				else
					error_no = ERR_NOTFOUND;
				EXIT_FUNCTION();
				return;
			}

			/* don't allow executes of a DIRECTORY */
			if(int21_ah == 0x4B && (sda_ffb.attr&A_NOMAKE))
			{
				error_no = ERR_DENIED;
				EXIT_FUNCTION();
				return;
			}

			/* open can't allow writes to read only files */
			if((int21_ah == 0x3D || int21_ah == 0x0F) &&
				(sda_ffb.attr&A_READONLY) &&
				sda_openmode>=O_WRONLY && sda_openmode<=O_RDWR)
			{
				error_no = ERR_DENIED;
				EXIT_FUNCTION();
				return;
			}
		}
	} /* end of if name MUST exist */
}

int lastchar(char *str)
{
	register char * p = str;
	unsigned char c = 0;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(str);

	while(*p)
	{
		c = *p;
		if(is_lead_byte(*p))
			++p;
		++p;
	}

	EXIT_FUNCTION();

	return c;
}
