//
//   $Id: DOSTIME.C 1.8 2002/10/15 21:44:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DOSTIME.C    Date and Time Functions

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	All access to date and time within ROM-DOS goes through
	these functions, which go through the clock device.

**********************************************************/
#include "dostruct.h"
#include "dosdef.h"
#include "dosproto.h"
#include "profile.h"


#define __FILENUM__ 16


#define FEB 1		// months get numbered from 0 in our array

extern uchar minute, hour;		/* the current system time */
extern uchar hund, seconds;
extern char days_per_month[];


/*
	Return whether this is a leapyear or not.  This is the same
	as the DOS formula (where every 4th year is leapyear).

   Note: the REAL leap year formula is every fourth year, but
   not centuries, unless they are divisible by 4 (ie 2000 is a
   leapyear, but 2100 is not).
*/
bool leapyear(unsigned year)
{
	ENTER_FUNCTION();

	// set days_per_month in February
	days_per_month[FEB] = 28;

	// leap years have 29 days in February
	if((year % 4) == 0)
	{
		++days_per_month[FEB];
      EXIT_FUNCTION();
		return TRUE;
	}


	EXIT_FUNCTION();

	// not a leap year
	return FALSE;
}


/*
	Get the current time into user registers.  This is
	really DOS Function 2CH.
*/
void	get_time(void)
{
	ENTER_FUNCTION();

	touch_time(TM_READ);					/* Read the time into globals */
	R_CX = (hour << 8) | minute;
	R_DX = (seconds << 8) | hund;

	EXIT_FUNCTION();
}


/*
	Set the current time.  This is really DOS Function 2DH.

	Returns 0 if it worked, 0FFH if it failed.
*/
int	PASCAL set_time(unsigned cx, unsigned dx)
{
	unsigned hr,mi,se,hu;

	ENTER_FUNCTION();

	/* so that the daycount is correct */
	touch_time(TM_READ);

	/* get time into locals */
	hr = cx>>8;
	mi = cx&0xFF;
	se = dx>>8;
	hu = dx&0xFF;

	/* check for limits */
	if( hr > 23 )
		goto error;
	if( mi > 59 )
		goto error;
	if( se > 59 )
		goto error;
	if( hu > 99)
		goto error;

	/* set gobals */
	hour    = hr;
	minute  = mi;
	seconds = se;
	hund    = hu;

	/* let device driver know (which will set the BIOS) */
	touch_time(TM_WRITE);

	EXIT_FUNCTION();
	return (0);

error:
	EXIT_FUNCTION();
	return (0xFF);				/* Bad Time */
}


/*
	Get the current date.

	Returns the day of the week, also sets user CX and DX.
*/
uchar	get_date(void)
{
	extern uchar month;
	extern uchar day;
	extern unsigned year;
	extern unsigned day_count;
	extern uchar day_of_week;

	ENTER_FUNCTION();

	touch_time(TM_READ);
	day_to_mdy();
	R_CX = year;
	R_DX = (month << 8) | day;
	day_of_week = (day_count+2) % 7;		/* day of the week */

	EXIT_FUNCTION();

	return (day_of_week);
}


/*
	Set the current date.

	Returns 0 if it worked, 0FFH if it failed.
*/
uchar	PASCAL set_date(unsigned cx, unsigned dx)
{
	unsigned mo,da;

	ENTER_FUNCTION();

	/* so that the time is correct */
	touch_time(TM_READ);

	/* determine month and day */
	mo = dx >> 8;
	da = dx & 0xFF;

	/* check for limits */
	if(cx < 1980 || cx > 2107)						/* is year reasonable? */
		goto error;
	if(mo < 1 || mo > 12)							/* is month reasonable? */
		goto error;

	/* leapyear will set days_per_month correct for year in cx */
	leapyear(cx);
	if(da < 1 || da > days_per_month[mo-1])	/* is day reasonable? */
		goto error;

	/* set day_count based on mdy */
	mdy_to_day(cx, mo, da);

	touch_time(TM_WRITE);

	EXIT_FUNCTION();
	return (0x00);

error:
	EXIT_FUNCTION();
	return (0xFF);
}


/*
	Change a month/day/year into the number of days since
	1-1-1980.
*/
void	mdy_to_day(int yr, int mo, int da)
{
	extern uchar month;
	extern unsigned year;
	extern unsigned day_count;

	ENTER_FUNCTION();

	day_count = 0;

	// validate parameters
	ASSERT(yr >= 1980);
	ASSERT(mo >= 1 && mo <= 12);
	ASSERT(da >= 1);

	/* count up the years */
	for(year=1980; year < yr; ++year)
	{
		day_count += 365;
		if(leapyear(year))
			++day_count;
	}

	/* count up the months */
	leapyear(year);
	for(month=1; month < mo; ++month)
		day_count += days_per_month[month-1];

	/* count up the days */
	day_count += da-1;

	EXIT_FUNCTION();
}


/*
	Change a number of days since 1-1-1980 into a month, day
	and year.

	Input
		"day_count" in the SDA.
	Output
		"day", "month" and "year"
*/
void	day_to_mdy(void)
{
	extern uchar day;
	extern uchar month;
	extern unsigned year;
	extern unsigned day_count;

	unsigned i=0, days_per_year, days_in_month;

	ENTER_FUNCTION();

	year = 1980;
	month = day = 1;

	while(TRUE)
	{
		/* determine days_per_year for THIS year */
		days_per_year = 365;
		if( leapyear(year) )
			++days_per_year;

		/* count up the years */
		if(day_count-i >= days_per_year)
		{
			++year;
			i += days_per_year;
			continue;
		}

		/* count up months */
		days_in_month = days_per_month[month-1];
		if(day_count-i >= days_in_month)
		{
			++month;
			i += days_in_month;
			continue;
		}

		day = day_count-i+1;
		EXIT_FUNCTION();
		return;
	}

	EXIT_FUNCTION();
}


/*
	This is the only routine which accesses the CLOCK
	device.  It sets up the global TIME variables upon
	reads, and sets the CLOCK based on these variables on
	writes.
*/
void	PASCAL touch_time(bool wr)
{
	/*
		[ externals ]
	*/
	extern struct LOL lol;
	extern uchar  clock_txd[6];
	extern struct dev_req_hdr sda_reqhdr;
	extern unsigned day_count;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	if(wr)
	{
		sda_reqhdr.command = DEV_WRITE;
		*(unsigned *)clock_txd = day_count;
		clock_txd[2] = minute;
		clock_txd[3] = hour;
		clock_txd[4] = hund;
		clock_txd[5] = seconds;
	}
	else
		sda_reqhdr.command = DEV_READ;

	/* call the clock device to get/set the time */
	sda_reqhdr.trans = MK_FP(getDS(), clock_txd);
	dev_call(&sda_reqhdr, lol.clkdev);

	if(!wr)
	{
		day_count= *(unsigned *)clock_txd;
		minute   = clock_txd[2];
		hour     = clock_txd[3];
		hund     = clock_txd[4];
		seconds  = clock_txd[5];
	}

	EXIT_FUNCTION();
}


/*
	This routine builds a date/time like appears in a
	directory entry.  It fills in the "sda_ffb" DIR
	structure with the current date and time.

	Returns nothing.
*/
void	get_fdate(void)
{
	/*
		[ externals ]
	*/
	extern DIR sda_ffb;
	extern uchar day;
	extern uchar month;
	extern unsigned year;


	/*
		[ code ]
	*/

	ENTER_FUNCTION();

	/* get the datest date and time */
	touch_time(TM_READ);
	day_to_mdy();

	/* fill in the SDA */
	sda_ffb.date = (year-1980) << 9;
	sda_ffb.date |= month << 5;
	sda_ffb.date |= day;
	sda_ffb.time = hour << 11;
	sda_ffb.time |= minute << 5;
	sda_ffb.time |= seconds/2;

	EXIT_FUNCTION();
}


/*
	This routine returns the current date/time as
	an unsigned long.
*/
ulong	get_filedate(void)
{
	extern DIR sda_ffb;


	ENTER_FUNCTION();

	get_fdate();

	EXIT_FUNCTION();

	return (ulong)MK_FP(sda_ffb.time, sda_ffb.date);
}


/*
	-----------------------------------------------------

				Routines Called from Assembly

	-----------------------------------------------------
*/

/*
	Returns the real value of a binary coded decimal.  Used
	by the clock driver.
*/
int	bcd(int val)
{
	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return (((val)>>4)*10 + ((val)&0xF));
}


/*
	Converts a normal # into a bcd #.
*/
int	to_bcd(unsigned val)
{
	unsigned bcd=0;
	int power = 10000;	/* check all 5 digits */

	ENTER_FUNCTION();

	while(power)
	{
		if(bcd || val / power)
		{
			bcd = bcd * 16 + (val / power);
			val %= power;
		}
		power /= 10;
	}

	EXIT_FUNCTION();

	return bcd;
}


/*
	The parameters "moday" and "year" are BCD values.
	Convert them to a day count.

	Returns the # of days since 1-1-1980.
*/
unsigned bcd2days(unsigned moday,unsigned bcdyr)
{
	extern unsigned day_count;

	unsigned month = bcd(moday >> 8);
	unsigned day   = bcd(moday&0xFF);
	unsigned year  = bcd(bcdyr >> 8) * 100 + bcd(bcdyr&0xFF);

	ENTER_FUNCTION();

	mdy_to_day(year,month,day);

	EXIT_FUNCTION();

	return day_count;
}

/*
	Returns hours,minutes,seconds and hundreds in
	"normal" time form (not BCD).
*/
ulong bcd2time(unsigned bcdhr,unsigned bcdsec)
{
	unsigned sechun,hrmin;

	ENTER_FUNCTION();

	sechun = bcd(bcdsec >> 8) << 8;
	hrmin  = ((unsigned)bcd(bcdhr >> 8) << 8) | bcd(bcdhr & 0xFF);

	EXIT_FUNCTION();

	return (((ulong)sechun << 16) | hrmin);
}

/*
	Returns month/day and the bcd year in a long form.
*/
ulong	days2bcd(unsigned days)
{
	extern uchar day;
	extern uchar month;
	extern unsigned year;
	extern unsigned day_count;

	unsigned bcdyr,monthday;

	ENTER_FUNCTION();

	day_count = days;
	day_to_mdy();

	bcdyr = to_bcd(year);
	monthday = (to_bcd(month) << 8) | to_bcd(day);

	EXIT_FUNCTION();
	return (((ulong)monthday << 16) | bcdyr);
}

/*
	Returns hours,minutes,seconds and hundreds in BCD
	form a long.
*/
ulong	time2bcd(unsigned sechun,unsigned hrmin)
{
	unsigned bcdhr,bcdsec;

	ENTER_FUNCTION();

	bcdsec = to_bcd(sechun>>8) << 8;
	bcdhr = (to_bcd(hrmin>>8) << 8) | to_bcd(hrmin&0xFF);

	EXIT_FUNCTION();

	return (((ulong)bcdsec << 16) | bcdhr);
}
