//
//   $Id: LONGNAME.C 1.40 2003/09/09 16:30:24Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   LONGNAME.C  created 12/18/96  by Drew Gislason
   Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This module contains the long filename support for ROM-DOS.
	Long filename support is identical to that found in
	Windows 95, and may be included or not based on user choice
	in BUILD.

	History
	06/19/98 - fixed LFN rename to refuse to rename to full dir/disk
	05/19/98 - added support for DOS1, DOS2 functions
 	03/12/97 - search optimization, record dir position
	03/05/97 - added Int 21h 71A6h Get File Info by Handle
				  added 5704-5707h Get File Date/Time by Handle
	02/26/97 - fixed cluster size bug - dg
	           added 71A6H File Info by Handle
	           fixed dir depth problem
	           added as option into build (libs, dummy obj)
	01/17/97 - all functions implemented - dg
				  added SEARCHES= to config.sys
				  allow renaming while file is open
	01/12/97 - initial cut at long filenames in RD kernel - dg
	01/10/97 - full unit test code in place - dg
	12/31/96 - completed findfirst/next - dg
	12/19/96 - added MakeAlias - dg
**********************************************************/
#define __FILENUM__ 20

#if UNIT_TEST
#include <dos.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

// functions
#define fmemcpy(dst,src,size) _fmemcpy(dst,src,size)
#define fstrlen(str) _fstrlen(str)
#define fmemset(str, c, len) _fmemset(str, c, len)
#define xitoa(str, num, radix) itoa(num, str, radix)
#define _lshlu(l, u) ((l) << (u))

#endif

#include "dostruct.h"
#include "dosdef.h"
#include "dl64kern.h"
#include "profile.h"

#include "fardos.h"
extern void far * fpFarDOSCall;

#if VDOS >= 7  // {  comment out the whole file unless DOS 7

#include "longname.h"

// prototypes
#include "longasm.pt"
#include "longname.pt"
#if UNIT_TEST
#include "longunit.pt"
#else
#include "dosproto.h"
#endif

// eventually handled by CONFIG.SYS options
LP_WIN32_SDB lplfnSearchArray;
int lfnMaxSearches = DEF_SEARCHES;
int lfnHandlesLeft;

// used for storing long names before they go to user space
bool fShortNameFunctions;
char szDos1Name[16];				// enough room for "a:filename.ext"
char szDos1Name2[16];			// used with DOS1 rename
char szSrcLongName[MAXPATH];
char szDstLongName[MAXPATH];
char szTempName[MAXPATH];
#define szDstShortName sda_fname1
char szProposedAlias[DOTNAMESIZE];
bool fAliasHelper;

// this is a temporary find, before it goes to user space
struct WIN32_FIND_DATA win32FindData;
struct WIN32_MORE_FIND_DATA win32MoreFoundData;
struct WIN32_SDB win32SDB;			// for internal searches (like del, ren)
struct WIN32_SDB win32OrgSDB;		// for internal searches (like del, ren)


int PASCAL is_lead_byte_pass_through(uchar c)
{
	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return IS_LEAD_BYTE(c);
}


/************************************

		Alias Functions

*************************************/

// note: not international aware
#define IsLower(c) ((c)>='a' && (c)<='z')

// is this a legal DOS character
bool IsLegalDOSChar(char c)
{
	extern char * terminators;

	ENTER_FUNCTION();

	// all illegal characters get converted to an underscore
	if(c <= ' ' || STRCHR(terminators, c))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	EXIT_FUNCTION();

	return TRUE;
}

// Return an underscore if character is not legal.  Return
// toupper of character otherwise.
char MakeLegalDOSChar(char c)
{
	ENTER_FUNCTION();

	// all illegal characters get converted to an underscore
	if(!IsLegalDOSChar(c))
	{
		EXIT_FUNCTION();
		return '_';
	}

	EXIT_FUNCTION();

	// just fine as is, but UPPERCASED
	return TOUPPER(c);
}

// helper function for MakeAlias
bool IsSpaceDot(char c)
{
	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return ((c==' ') || (c=='.'));
}

/*
	Returns TRUE if the filename looks like the standard 8.3
	kind of name.  Returns FALSE otherwise.
*/
bool IsStandardDOSName(lpchar lpLongFilename)
{
	bool	isDosName = TRUE;		// assume normal DOS name
	bool	onExt = FALSE;
	int	charsInName, charsInExt;

	ENTER_FUNCTION();

	// does this look like an 8.3 name?
	charsInName = charsInExt = 0;
	while(isDosName && *lpLongFilename)
	{
		// no illegal characters allowed
		if(!IsLegalDOSChar(TOUPPER(*lpLongFilename)))
			isDosName = FALSE;

		// only allowed to have 1 dot in the name
		if(*lpLongFilename == '.')
		{
			if(onExt)
				isDosName = FALSE;
			onExt = TRUE;
		}

		// count characters in name and extension
		else if(onExt)
			++charsInExt;
		else
			++charsInName;

		// try next character
		++lpLongFilename;
	}

	// not a standard DOS name
	if(charsInName > 8 || charsInExt > 3)
		isDosName = FALSE;

	EXIT_FUNCTION();

	// was this a normal DOS name?
	return isDosName;
}

/*
	Helper function for MakeAlias.  Copies characters up to a given
	maximum.  Used to copy filename and extension
*/
void MakePartOfAlias(int maxChars, lpchar lpLongFilename, lpchar lpEndStr,
								char *pAlias)
{
	int numChars;

	ENTER_FUNCTION();

	// skip preceding spaces and dots
	while(IsSpaceDot(*lpLongFilename))
		++lpLongFilename;

	// alias starts at END of string
	pAlias += STRLEN(pAlias);

	// record up to "n" characters, or until next dot (endstr)
	numChars = 0;
	while((numChars < maxChars) && (lpLongFilename != lpEndStr))
	{
		// don't include spaces or dots in alias name
		if(!IsSpaceDot(*lpLongFilename))
		{
			*pAlias = MakeLegalDOSChar(*lpLongFilename);
			++pAlias;
			++numChars;
		}

		// on to next character
		++lpLongFilename;
	}

	// null terminate the string
	*pAlias = 0;

	EXIT_FUNCTION();
}

/*
	Add a ~1 to end of alias string.  The "num" is the number
	to use (may not be 1, may be 12345).
*/
void AliasAppend(char *pAlias, unsigned num)
{
	int	lenAlias;
	char	szNumStr[7];	// maximum of 5 digits, 1 tilde, plus NULL
	int	lenNumStr;

	ENTER_FUNCTION();

	// get the number in string form: ~1
	*szNumStr = '~';
	XITOA(szNumStr+1, num, 10);
	lenNumStr = STRLEN(szNumStr);

	// append it, and null to end of string.  do NOT exceed
	// 8 characters
	lenAlias = STRLEN(pAlias);
	if(lenAlias + lenNumStr > 8)
		lenAlias = 8 - lenNumStr;
	MEMMOVE(pAlias+lenAlias, szNumStr, lenNumStr+1);

	EXIT_FUNCTION();
}

/*
	There already is an alias, increment the number.
*/
void IncrementAlias(char *pAlias)
{
	char *pTilde;
	char cSave;
	int num;

	ENTER_FUNCTION();

	pTilde = IntlStrRChr(pAlias, '~');
	if(*pTilde == 0)
		AliasAppend(pAlias, 1);
	else
	{
		num = ATOI(pTilde+1);
		cSave = pAlias[8];
		*pTilde = 0;
		AliasAppend(pAlias, num+1);
		pAlias[8] = cSave;
	}

	EXIT_FUNCTION();
}

/*
	This function will make an alias for a long filename.

	Alias:        LongName:
	AB~1          "a b"
	RDSRC    ZIP  "RdSrc.zip"
	THISIS~1 DOC  "This is a very long filename, isn't it.doc"
	ANOTHE~1 DOC  "Another filename!.doc"
	______~1      "+,;=[]"
	LONG~1   NAM  ".long.name"
	_~1           "+"
	HELLOT~1.TES  "hello. this. is. a. test"

	Inputs:
		lpLongFilename		Long Filename to make alias from
		useNum				Put the ~1 on end of string?
		hint					What digits are to be used for ~1
	Outputs:
		pAlias
*/
void MakeAlias(lpchar lpLongFilename, char *pAlias, bool useNum, unsigned hint)
{
	lpchar	lpLastDot;

	ENTER_FUNCTION();

	// find the last dot, that is, the file extension
	// if none, lpLastDot points to the NULL
	lpLastDot = IntlFStrRChr(lpLongFilename, '.');

	//  We ignore '.' (and ' ' but irrelevent) for the first character
	//
	if( lpLastDot == lpLongFilename )
	{
		lpLastDot = 0;
	}

	// alias must ALWAYS be a null terminated string
	*pAlias = 0;

	// make the first part of the alias name
	MakePartOfAlias(8, lpLongFilename, lpLastDot, pAlias);

	// if long names are not traditional 8.3, then use the ~1
	// scheme.
	if(useNum && !IsStandardDOSName(lpLongFilename))
		AliasAppend(pAlias, hint);

	// we have an extension!
	if((lpLastDot) && (*lpLastDot))
	{
		// will definitely have an extension
		STRCAT(pAlias, ".");

		// set up pointers to use the last bit of long filename
		lpLongFilename = lpLastDot;						// last dot
		lpLastDot = IntlFStrRChr(lpLastDot, 0);		// end of string

		// copy up to 3 characters for file extension
		MakePartOfAlias(3, lpLongFilename, lpLastDot, pAlias);
	}

	EXIT_FUNCTION();
}


/************************************

		FindFirst/Next Functions

*************************************/

/*
	Search for the end of this chain of clusters.
	Returns the cluster # of the end of chain
*/
CLUSTER EndOfChain(CLUSTER start)
{
	CLUSTER clus;

	ENTER_FUNCTION();

	while(1)
	{
#if USING_FAT32
#if DEBUG
			ASSERT( start >= 2 );
			if( IsFat32Drive(sda_dpbptr) )
			{
				ASSERT( start <= sda_dpbptr->ulMaxCluster );
			}
			else
			{
				ASSERT( start <= sda_dpbptr->hiclus );
			}
#endif
#else
#if DEBUG
			ASSERT( start >= 2 );
			ASSERT( start <= sda_dpbptr->hiclus );
#endif
#endif
		clus = GET_FAT_ENTRY(start);
		if(clus >= END_OF_FAT_CHECK)
			break;
		start = clus;
	}

	EXIT_FUNCTION();

	// return the end of the chain
	return start;
}

/*
	Create more entries in a subdir.  Will always fail in the root
	(can't expand the root).

	Returns TRUE if it got the directory entries (and ptr to DIR).
	Returns FALSE if at end of root dir, or disk is full.
*/
#pragma argsused
bool CreateMoreDirEntries(int numEntries, CLUSTER subdir)
{
	CLUSTER	start;
	CLUSTER	getnew_start;
	CLUSTER	newClus;

	ENTER_FUNCTION();

	start = subdir;
	// can't extend the root directory! (subdir 0)
#if USING_FAT32
	//  if this is a FAT32 drive we can expand the directory but if
	//  is not a FAT32 drive we can't expand the root.
	//
	if( IsFat32Drive(sda_dpbptr) && (start==0))
	{
		start = sda_dpbptr->ulRootDirStartCluster;
	}
	else if(start == 0)
	{
		EXIT_FUNCTION();
		return FALSE;
	}
#else
	if(!start)
	{
		EXIT_FUNCTION();
		return FALSE;
	}
#endif

	// follow chain until we get to EOF
	// we want the last cluster # as our starting point
	start = EndOfChain(start);
   getnew_start = start;

	// determine # of clusters needed, based on number of
	if(GET_NEW_CLUSTERS(&getnew_start, 1) != 1)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// zero out the new directory cluster
	newClus = EndOfChain(start);
	ZERO_CLUSTER(newClus);

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Helper routine to convert subdir/dirnum pair (cluster, logical #)
	to a dirsec/dirent pair (sector, index entry in that sector).

	Returns FALSE if "dirnum" is past end of directory.
	Returns TRUE and sets up "dirsec" and "dirent" if ok.
*/
bool DirnumToDirsec(CLUSTER subdir, int dirnum, int eps,
										ulong *pDirsec, uchar *pDirent)
{
	int logicalSector;	// logical sector of "dirnum" within directory chain
	CLUSTER	physicalCluster;
	NCLUSTER	logicalCluster;
#if USING_FAT32
	CLUSTER  ulSubDirectoryClusterNumber;
#endif


	ENTER_FUNCTION();

#if USING_FAT32

	//  If this is the root of a FAT32 drive then we will pretend that
	//  it is a subdirectory for the moment.
	//
	if( (IsFat32Drive(sda_dpbptr)) && (subdir == 0) )
	{
		ulSubDirectoryClusterNumber = sda_dpbptr->ulRootDirStartCluster;
	}
	else
	{
		ulSubDirectoryClusterNumber = subdir;
	}

	// root directory, easy to calculate where the direntry is
	if(ulSubDirectoryClusterNumber == 0)
	{
		// don't go past end of root
		if(dirnum >= sda_dpbptr->dir_ent)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// get sector / entry
		*pDirsec = sda_dpbptr->dir_start + (dirnum / eps);
	}

	// sector is coming out of a subdirectory
	else
	{
		// determine number of clusters into this directory
		logicalSector = dirnum / eps;

		// determine logical cluster
		logicalCluster = logicalSector >> sda_dpbptr->shift;

		// note: (optimization) normally, we'll be parsing forward
		// each time, so why recalculate? we should remember last
		// logical cluster.

		// determine physical cluster
		physicalCluster = LCLUS_TO_PCLUS( ulSubDirectoryClusterNumber,
				&logicalCluster );

		// don't go past end of subdir
		if(logicalCluster)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// determine physical sector.  This is based on physical
		// cluster, and sector within that cluster.
		// Example: a cluster has 8 sectors per cluster, "dirnum" is 33,
		// this starts on sector 2 within the 1st cluster.
		*pDirsec = CLUS_TO_SEC(physicalCluster) +
			(logicalSector % (sda_dpbptr->ms_spc + 1));
	}

#else

	// root directory, easy to calculate where the direntry is
	if(subdir == 0)
	{
		//  todo: dir_entry stuff
		//

		// don't go past end of root
		if(dirnum >= sda_dpbptr->dir_ent)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// get sector / entry
		*pDirsec = sda_dpbptr->dir_start + (dirnum / eps);
	}

	// sector is coming out of a subdirectory
	else
	{
		// determine number of clusters into this directory
		logicalSector = dirnum / eps;

		// determine logical cluster
		logicalCluster = logicalSector >> sda_dpbptr->shift;

		// note: (optimization) normally, we'll be parsing forward
		// each time, so why recalculate? we should remember last
		// logical cluster.

		// determine physical cluster
		physicalCluster = LCLUS_TO_PCLUS(subdir, &logicalCluster);

		// don't go past end of subdir
		if(logicalCluster)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// determine physical sector.  This is based on physical
		// cluster, and sector within that cluster.
		// Example: a cluster has 8 sectors per cluster, "dirnum" is 33,
		// this starts on sector 2 within the 1st cluster.
		*pDirsec = CLUS_TO_SEC(physicalCluster) +
			(logicalSector % (sda_dpbptr->ms_spc + 1));
	}

#endif

	// now, figure out dirent within that sector
	*pDirent = (dirnum % eps);

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Get a pointer to the directory entry, given a sector
	# and a directory entry # within that sector.
*/
LPGENDIR GetRawDirEntry(ulong sector, uchar dirent)
{
	void far * lpData;

	ENTER_FUNCTION();

	BUF_GETSECTOR(sector);
	lpData = buf_dataptr(sda_cur_buf);

	EXIT_FUNCTION();

	// get pointer to DIR in question
	return (((LPGENDIR)lpData) + dirent);
}

/*
	New subdirectory.  Logical sector speedup is invalid.
*/
void SetSubdir(LP_WIN32_SDB lpWin32SDB, CLUSTER subdir)
{
	ENTER_FUNCTION();

	lpWin32SDB->subdir = subdir;
	lpWin32SDB->lastLogicalSector = SPEEDUP_INVALID;

	EXIT_FUNCTION();
}

/*
	Get the directory entry specified by the Win32 SDB.  Does not
	care whether the DIR entry contains a valid file, just gets
	the raw entries.

	lpWin32SDB		(input) what drive/subdir/entry are we on?
	plpDir			(output) returned ptr to DIR entry

	Returns TRUE and a pointer to the DIR structure if worked.

	Returns FALSE if no more entries in this directory (will NOT
	extend subdirectories).
*/
bool GetDirEntry(LP_WIN32_SDB lpWin32SDB, PLPGENDIR plpDir)
{
	ulong sector;
	uchar	dirent;
	int	logicalSector;

	ENTER_FUNCTION();

	// this Win32 SDB has never been accessed, set it up
	if(!lpWin32SDB->eps)
	{
		lpWin32SDB->eps = sda_dpbptr->bytes_per_sec / sizeof(DIR);

		// make these items invalid so we always read them
		lpWin32SDB->dirsec = (ulong)SPEEDUP_INVALID;
		lpWin32SDB->lastLogicalSector = SPEEDUP_INVALID;
	}

	// is logical sector the same as before?
	// then just reuse it!
	logicalSector = lpWin32SDB->dirnum / lpWin32SDB->eps;
	if(logicalSector == lpWin32SDB->lastLogicalSector)
	{
		sector = lpWin32SDB->dirsec;
		dirent = lpWin32SDB->dirnum % lpWin32SDB->eps;
	}

	// sector is different, go determine right one
	else
	{
		// convert from subdir/dirnum to dirsec/dirent
		if(!DirnumToDirsec(lpWin32SDB->subdir, lpWin32SDB->dirnum,
				lpWin32SDB->eps, &sector, &dirent))
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// get sector from disk or disk cache scheme
		lpWin32SDB->dirsec = sector;

		// remember last logical sector we accessed
		lpWin32SDB->lastLogicalSector = logicalSector;
	}

	// get a pointer to the actual directory
	*plpDir = GetRawDirEntry(sector, dirent);

	EXIT_FUNCTION();

	// got the DIR entry
	return TRUE;
}

/*
	If a long filename were 13 characters, it would require exactly
	1 DIR.  If it were 14 chars, it would need 2 DIRs.

	Return # of DIRs required to hold this string.
*/
int EntriesForLongname(char *pLongname)
{
	ENTER_FUNCTION();
	EXIT_FUNCTION();

	return ((STRLEN(pLongname) + CHARS_IN_LFN_DIR - 1) / CHARS_IN_LFN_DIR);
}

/*
	Get unicode characters into a destination string.  Return
	the count of characters found before a terminating NULL
	or FFFF.  These are the unicode characters found in the DIR
	structure on disk (Long Filenames).
*/
int GetUniChars(char * pDstStr, LPUNICHAR lpUniChar, int maxChars)
{
	int n = 0;

	ENTER_FUNCTION();

	ASSERT(pDstStr);
	ASSERT(lpUniChar);

	// loop through getting unicode characters
	while(n < maxChars && *lpUniChar && *lpUniChar != UNIFILL)
	{
		*pDstStr++ = *lpUniChar++;
		++n;
	}

	EXIT_FUNCTION();

	// return # of characters found
	return n;
}

/*
	Helper function for GetNextLFNEntry.

	Gather the long filename part of a name.  Will stop gathering
	if it encounters an error.

	Inputs:
		lpWin32SDB			drive / dir info
		entriesInName		# of DIR entries that compose longname
	Output:
		pLongFilename		resulting longname in asciiz form

	Returns TRUE if worked, FALSE if failed.
*/
bool GatherLongFilename(LP_WIN32_SDB lpWin32SDB, int entriesInName,
																	char * pLongFilename)
{
	LPLFNDIR	lplfnDir;
	uchar		checksum;			// checksum, for validity check
	int		entry;				// loop variable
	char		sDirChars[13];		// up to 13 chars per entry
	int		n, numDirChars;	// # chars this DIR entry
	char *	p;

	ENTER_FUNCTION();

	// loop through all entries, prepending them to string
	for(entry=entriesInName; entry>0; --entry)
	{
		// get the DIR entry, where the sequence bytes begin
		if(!GetDirEntry(lpWin32SDB, (PLPGENDIR)(&lplfnDir)))
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// attribute must be correct!
		if(lplfnDir->attr != A_LONGNAME)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// sequence must be correct!
		if((lplfnDir->sequence & SEQUENCE) != entry)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// checksum must agree in ALL entries
		if(entry == entriesInName)
			checksum = lplfnDir->checksum;
		else if(lplfnDir->checksum != checksum)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// get characters from the 3 separate fields in LFN DIR entry
		n = numDirChars = GetUniChars(sDirChars, lplfnDir->fileName1, NUM_NAME1);

		// still more?
		if(n == NUM_NAME1)
		{
			n = GetUniChars(sDirChars+numDirChars,
										lplfnDir->fileName2, NUM_NAME2);
			numDirChars += n;

			// still more?
			if(n == NUM_NAME2)
				numDirChars += GetUniChars(sDirChars+numDirChars,
										lplfnDir->fileName3, NUM_NAME3);
		}

		// all but first LFN dir entry should be filled!
		if(entry < entriesInName && numDirChars < CHARS_IN_LFN_DIR)
		{
			EXIT_FUNCTION();
			return FALSE;
		}


		// Now, put characters in proper slot in the long filename
		// The DIR entries are backwards, so first entry is last
		// part of long filename.
		p = pLongFilename + (entry-1) * CHARS_IN_LFN_DIR;
		MEMMOVE(p, sDirChars, numDirChars);

		// null terminate the long filename string
		if(entry == entriesInName)
			*(p+numDirChars) = 0;

		// on to next directory entry
		++lpWin32SDB->dirnum;
	}

	EXIT_FUNCTION();

	// everything checked out!
	return TRUE;
}

/*
	Helper function for GetNextLFNEntry.

	Put DOS file date into the Win32 style find structure.
*/
void PutFindTime(struct FILETIME * pFileTime, unsigned time, unsigned date)
{
	ENTER_FUNCTION();

	pFileTime->lowDateTime  = (((ulong)date << 16) | time);
	pFileTime->highDateTime = 0;

	EXIT_FUNCTION();
}

/*
	Get the next entry.  Does NO pattern or attribute matching,
	but simply gets the next valid entry.  Skips deleted files
	and invalid entries.  Stops when there are no more entries
	(a 0 is encountered for name and attribute).

	Outputs:
		*pDirnum				The directory entry where it starts
		*pEntries         Returns # of entries occupied by LFN
		*pSubdir          Returns subdir cluster, 0=root
		pWin32FindData		The next found valid entry

	If the entry is a long filename, will grab both the long
	filename and the short alias. Otherwise, the filename is
	filled in with the short name.
*/
bool GetNextLFNEntry(LP_WIN32_SDB lpWin32SDB, P_WIN32_FIND_DATA pWin32FindData,
								unsigned *pDirnum, int *pEntries, CLUSTER *pSubdir)
{
	LPGENDIR	lpDir;
	char *	pName;
	char		ename[ENAMESIZE];
	uchar		checksum;

	ENTER_FUNCTION();

	// get directory entries
	while(1)
	{
		// get the next entry
		if(!GetDirEntry(lpWin32SDB, &lpDir))
			break;

		// at end of list, no more entries found
		if(IsEndOfDirList(lpDir))
		{
			if(sda_1st_blank == 0xFFFF)
				sda_1st_blank = lpWin32SDB->dirnum;
			break;
		}

		// is the entry deleted? skip it
		if(IsDeleted((uchar)(*lpDir->dir.name)))
		{
			if(sda_1st_blank == 0xFFFF)
				sda_1st_blank = lpWin32SDB->dirnum;

			++lpWin32SDB->dirnum;
			continue;
		}

		// record staring dirnum for a found entry
		*pEntries = 1;
		*pDirnum  = lpWin32SDB->dirnum;

		// start out with no names
		*pWin32FindData->szFilename =
		*pWin32FindData->szAliasFilename = 0;

		// is this a long filename entry?
		if(lpDir->lfnDir.attr == A_LONGNAME &&
			(lpDir->lfnDir.sequence & SEQUENCE_ID))
		{
			// remember checksum, so we can verify that it's OK
			checksum = lpDir->lfnDir.checksum;

			// gather longname, will stop gathering if invalid
			if(!GatherLongFilename(lpWin32SDB,
					(lpDir->lfnDir.sequence & SEQUENCE),
					pWin32FindData->szFilename))
			{
   			++lpWin32SDB->dirnum;
				continue;
			}

			// add in # of DIRs for longname part
			*pEntries += EntriesForLongname(pWin32FindData->szFilename);

			// get alias (or short) name. a long filename
			// without an alias is bogus
			if(!GetDirEntry(lpWin32SDB, &lpDir))
				break;
		}

		// shortname alias has bad attr, ignore this entry
		if(lpDir->winDir.attr == A_LONGNAME)
		{
   		++lpWin32SDB->dirnum;
			continue;
		}

		// shortname is deleted (probably on DOS 6.22), so ignore it
		if(IsDeleted((uchar)(*lpDir->winDir.name)))
		{
   		++lpWin32SDB->dirnum;
			continue;
		}

		// get the expanded name so we can examine it
		FMEMCPY(ename, lpDir->winDir.name, ENAMESIZE);
#if 1
      if( (uchar)ename[0] == (uchar)0x05 )
      {
         // translate this character back to an 0xE5 for result
         //
         ename[0] = 0xE5;
      }
#endif  // #if 1

		// If a long filename was already retrieved above, ename
		// will go in alias filename.  But, if checksum is invalid,
		// revert to short filename.
		pName = pWin32FindData->szFilename;
		if(*pName && checksum == DetermineChecksum(ename))
			pName = pWin32FindData->szAliasFilename;

		// invalid longname, or shortname only, so dirnum starts here
		else
			*pDirnum = lpWin32SDB->dirnum;

		// get the ename name in "dot" form in the Find Data structure
		ENAME_TO_DOT(pName, ename);

		// make sure alias is unique.
		// this is an optimization for Create() and Rename() (both files
		// and directories) so they don't have to parse the directory
		// again, looking for duplicate aliases.
		if(fAliasHelper && *szProposedAlias)
		{
			if( ( strcmp( pWin32FindData->szAliasFilename,
					szProposedAlias) == 0) || ( strcmp( pName,
					szProposedAlias) == 0) )
			{
				IncrementAlias(szProposedAlias);
			}
		}

		// get other fields from this Win32 DIR entry
		// standard DOS ones
		pWin32FindData->attr = lpDir->winDir.attr;
		pWin32FindData->fileSize = lpDir->winDir.fileSize;
		PutFindTime(&pWin32FindData->lastWriteTime,
			lpDir->winDir.lastModTime,lpDir->winDir.lastModDate);

		// LFN extensions to DIR entry
		PutFindTime(&pWin32FindData->creationTime,
			lpDir->winDir.creationTime,lpDir->winDir.creationDate);

		PutFindTime(&pWin32FindData->lastAccessDate,
			0, lpDir->winDir.lastAccessDate);

		// record subdirectory this points to
#if USING_FAT32
		*pSubdir = Fat32Start(lpDir->dir);
#else
		*pSubdir = lpDir->dir.start;
#endif

		// we found this entry, set up Win32 SDB for findnext
		++lpWin32SDB->dirnum;

		EXIT_FUNCTION();

		// done, return the fact that we got an entry!
		return TRUE;
	}

	EXIT_FUNCTION();

	// no more entry names
	return FALSE;
}

/*
	Match a single string string, given a mask.  The mask
	may contain wildcards, and it will search

	Examples:
	*			always matches
	*x*		matches if string contains an "x"
	a* 		matches if string begins with "a"
	*ing		matches if string ends in "ing"
	a*b*c		begins with "a", has a "b" in it, ends with "c"
*/
bool MatchMask(lpchar lpMask, char *pStr)
{
	bool fMatches = TRUE;
	bool fImpliedDotUsed = FALSE;
	bool fSavesValid = FALSE;
	unsigned uIndex;
	lpchar lpSavelpMask = lpMask;
	char * pSavepStr = pStr;

   ENTER_FUNCTION();

	if(*pStr == 0)
	{
		EXIT_FUNCTION();
		return FALSE;
	}


	//  Determine if the filename has an extension
	//
	for( uIndex = 0; pStr[uIndex]; uIndex++ )
	{
		if( pStr[uIndex] == '.' )
		{
			break;
		}
	}
	if( pStr[uIndex] == '.' )
	{
		fImpliedDotUsed = TRUE;
	}


	while( TRUE )
	{
		if(*lpMask == 0)
		{
			if( *pStr != 0 )
			{
				if( fSavesValid )
				{
					lpMask = lpSavelpMask;
					pStr = pSavepStr;
					pStr++;
					fSavesValid = FALSE;
				}
				else
				{
					break;
				}
			}
			else
			{
				break;
			}

		}


		//  Determine if this is a wildcard
		//
		if( *lpMask == '*' )
		{
			//  Filter any back to back wildcards
			//
			while( TRUE )
			{
				if( lpMask[1] == '*' )
				{
					lpMask++;
				}
				else if( lpMask[1] == '?' )
				{
					lpMask++;
				}
				else
				{
					break;
				}
			}


			lpSavelpMask = lpMask;


			//  Go to the character following the wildcard in the mask
			//
			lpMask++;


			//  Wildcards automatically match all characters up
			//  to the next character in the mask
			//
			while( *pStr )
			{
				if( TOUPPER(*pStr) == TOUPPER(*lpMask) )
				{
					pSavepStr = pStr;
					fSavesValid = TRUE;
					break;
				}

				pStr++;
			}


			if( !fImpliedDotUsed )
			{
				if( (*lpMask == '.') && (*pStr == 0) )
				{
					lpMask++;
					fImpliedDotUsed = TRUE;
				}
			}
		}
		else if( *lpMask == '?' )
		{
			if( *pStr != 0 )
			{
				pStr++;
			}
			lpMask++;
		}
		else
		{
         //
         // If we haven't used the ImpliedDot and we have a
         // left-over . or .* we need to get rid of them.
         //
         if( *lpMask == '.' )
         {
            if( (*pStr == 0) && !fImpliedDotUsed )
            {
               lpMask++;

               if( *lpMask == '*' )
               {
                  lpMask++;
               }

 					break;
            }
         }

			if( TOUPPER(*pStr) == TOUPPER(*lpMask) )
			{
				lpMask++;
				pStr++;
			}
			else
			{
				if( fSavesValid )
				{
					lpMask = lpSavelpMask;
					pStr = pSavepStr;
					pStr++;
					fSavesValid = FALSE;
				}
				else
				{
					fMatches = FALSE;
					break;
				}
			}
		}
	}


	if( (*pStr != 0) || (*lpMask != 0) )
	{
		fMatches = FALSE;
	}


	EXIT_FUNCTION();

	//  Return whether the strings matched
	//
	return fMatches;
}


/*
	See if one or both of the filenames match the file mask.
	Ignores case.  Will search both longname and shortname.

	Returns TRUE if they match, FALSE if not.
*/
bool MatchLongFilenames(lpchar lpLongMask, char *pLongName, char *pShortName)
{
	ENTER_FUNCTION();

	// shortname functions use old-style name matching
	if(fShortNameFunctions)
	{
		// no longname associated with alias. only name is in pLongName
		if(*pShortName == 0)
			pShortName = pLongName;

		// convert names to enames
		DOT_TO_ENAME(sda_ffb.name, pShortName, FALSE);
		DOT_TO_ENAME(sda_sdb.emask, lpLongMask, FALSE);

		EXIT_FUNCTION();

		// return if it worked.
		return NAME_MATCH(sda_ffb.name, sda_sdb.emask);
	}

	// longname functions use different search criteria
	else
	{
		// check short name first
		if(MatchMask(lpLongMask, pShortName))
		{
			EXIT_FUNCTION();
			return TRUE;
		}

		// then check longname
		if(MatchMask(lpLongMask, pLongName))
		{
			EXIT_FUNCTION();
			return TRUE;
		}
	}

	EXIT_FUNCTION();

	// didn't match
	return FALSE;
}

/*
	Get a free long filename handle.
	Returns handle # and ptr to Win32SDB.
	Returns 0 if none available.
*/
unsigned GetFreeLFNHandle(PLP_WIN32_SDB plpWin32SDB)
{
	int i;
	unsigned lfnHandle = 0;
	LP_WIN32_SDB lpWin32SDB;

	ENTER_FUNCTION();

	lpWin32SDB = lplfnSearchArray;
	for(i=0; !lfnHandle && i<lfnMaxSearches; ++i)
	{
		// not in use? then use it!
		if(!lpWin32SDB->inUse)
		{
			// mark in use
			lpWin32SDB->inUse = TRUE;

			// mark as "unaccessed"
			lpWin32SDB->eps = 0;

			// remember process that opened this search
			lpWin32SDB->process = current_psp;

			// return handle and pointer
			lfnHandle = START_HANDLE + i;
			*plpWin32SDB = lpWin32SDB;

			// debugging, how many handles do we have?
			--lfnHandlesLeft;
		}

		// try next handle
		++lpWin32SDB;
	}

	EXIT_FUNCTION();

	// return handle to Win32SDB (0 if none available)
	return lfnHandle;
}

// free all searches that belong to this process
void FreeProcessSearches(void)
{
	int i;
	LP_WIN32_SDB lpWin32SDB;

	ENTER_FUNCTION();

	// walk through array, freeing every LFN search handle (SDB)
	// that belongs to this process
	lpWin32SDB = lplfnSearchArray;
	for(i=0; i<lfnMaxSearches; ++i)
	{
		if(lpWin32SDB->process == current_psp)
			lpWin32SDB->inUse = FALSE;
		++lpWin32SDB;
	}

	EXIT_FUNCTION();
}

// Returns TRUE and the ptr to the Win32SDB for this handle
// if valid.  Returns FALSE if not.
bool IsValidHandle(unsigned lfnHandle, PLP_WIN32_SDB plpWin32SDB)
{
	LP_WIN32_SDB lpWin32SDB;

	ENTER_FUNCTION();

	// check to make sure handle is in range
	// we use the start handle, convert to index
	lfnHandle -= START_HANDLE;
	if(lfnHandle > lfnMaxSearches)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// Win32 SDB not open! can't be valid
	lpWin32SDB = lplfnSearchArray + lfnHandle;
	if(!lpWin32SDB->inUse)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// handle IS valid, set ptr
	*plpWin32SDB = lpWin32SDB;

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Done with a search, free the handle.
	Returns TRUE if worked, FALSE if invalid handle.
*/
bool FindClose(unsigned lfnHandle)
{
	LP_WIN32_SDB lpWin32SDB;

	ENTER_FUNCTION();

	// not a valid handle
	if(!IsValidHandle(lfnHandle, &lpWin32SDB))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

#if DEBUG
	fmemset(lpWin32SDB, 0x55, sizeof(struct WIN32_SDB));
#endif

	// search closed
	lpWin32SDB->inUse = FALSE;
	lpWin32SDB->eps   = 0;

	// debugging - keep track
	++lfnHandlesLeft;

	EXIT_FUNCTION();
	return TRUE;
}

//  Converts DOS date and time values to a 64-bit file time. A 64-bit
//  file time is the value specifying the number of 100ns intervals that
//  have elapsed since 12:00 A.M. January 1, 1901.
//
unsigned	DosTimeToFileTime( unsigned uDosTime, unsigned uDosDate,
	uchar ucMilliSecond, lpFILETIME lpFILETIME )
{
	// months get numbered from 0 in our array
	extern char days_per_month[];

	uchar    ucHour;
	uchar    ucMinute;
	uchar    ucSecond;
	uchar    ucDay;
	uchar    ucMonth;
	unsigned uYear;

	unsigned uTemp;
	ulong ulTotalSeconds = 0;

	//  The same as Win98 INT 21h fn 71a7h
	//
	FILETIME FTAtStartOf1980;
	uquad uqFileTime;

	ENTER_FUNCTION();

	FTAtStartOf1980.lowDateTime  = 0x8e345800L;
	FTAtStartOf1980.highDateTime = 0x01a8e7daL;

	//  get time/date into locals
	//
	ucSecond = (uchar)((uDosTime & 0x001f) << 1);
	ucMinute = (uchar)((uDosTime & 0x07e0) >> 5);
	ucHour   = (uchar)((uDosTime & 0xf800) >> 11);

	ucDay	  = (uchar)(uDosDate  & 0x001f);
	ucMonth = (uchar)((uDosDate & 0x01e0) >> 5);
	uYear   = (((uDosDate & 0xfe00) >> 9) + 1980);

	//  Range check all of the date/time values.  MSDOS does not range
	//  check for values less than 1 (ie ucDay < 1).
	//
	if(( ucSecond > 59 ) || ( ucMinute > 59 ) || ( ucHour > 23 ) ||
			( ucDay > 31 ) || (ucMonth > 12) || (ucMilliSecond > 199) )
	{
		EXIT_FUNCTION();

		//  Return an error. This is the value of the ax register for
		//  a range error.  Looks an awful lot like what the registers
		//  would be if it was an unsupported function.
		//
		return ERR_W95BADFUNC;
	}

	// set Febuary to 29 days (will be changed by the LEAP_YEAR if year > 1980)
	//
	days_per_month[1] = 29;

	//  Determine the total number of seconds in all the years
	//  since 1980
	//
	uTemp = 1980;
	while( uTemp < uYear )
	{
		// compute the total seconds in this year
		// (assume this is not a leap year)
		// 365 * 24 * 60 * 60 = 31536000
		//
		ulTotalSeconds += 31536000L;

		// Add in a days worth of seconds if this is not a
		// leap year
		// 24 * 60 * 60 = 86400
		//
		if(LEAP_YEAR(uTemp))
		{
			ulTotalSeconds += 86400L;
		}
		uTemp++;
	}

	//  Add to our previous calculation all the seconds in the days
	//  of the month
	//
	uTemp = 1;
	while( uTemp < ucMonth )
	{
		// months get numbered from 0 in our array
		// 24 * 60 * 60 = 86400
		//
		ulTotalSeconds += _LMULU( 86400L, (unsigned)days_per_month[uTemp-1] );
		uTemp++;
	}

	//  Add to our previous calculation all the seconds in the day
	// 24 * 60 * 60 = 86400
	//
	if( ucDay > 1  )
	{
		ulTotalSeconds += _LMULU( 86400L, ucDay-1 );
	}

	//  Add to our previous calculations all the seconds in the hours
	//  and minutes
	//
	if( (ucHour > 0) || (ucMinute > 0) )
	{
		ulTotalSeconds += _LMULU( 60, ((unsigned)ucHour * 60) + ucMinute );
	}

	//  Add to our previous calculation all the seconds
	//
	ulTotalSeconds += ucSecond;

	//  All the previous calculations have given us the total number of
	//  seconds in the DOSTIME and DOSDATE.  Now we need to convert
	//  this into a 64-bit value specifying the number of 100ns intervals
	//  which have elapsed since 12:00 AM, January 1, 1980.
	//
	QuadMultiply( ulTotalSeconds, 10000000L, &uqFileTime );
	AddQuadLong( &uqFileTime, _LMULU( 100000L, ucMilliSecond ) );

	//  Add in the previous value to a 64-bit value specifying the number
	//  of 100ns intervals which have elapsed between 1200 AM, January 1,
	//  1901 and 12:00 AM, January 1, 1980.
	//
	AddQuad( &uqFileTime, (uquad *)&FTAtStartOf1980 );

	//  Set the computed value in the return variable
	//
	lpFILETIME->lowDateTime  = uqFileTime.ulLo;
	lpFILETIME->highDateTime = uqFileTime.ulHi;

	EXIT_FUNCTION();

	// no errors!
	return 0;
}

/*
	Find Next, assuming Handle is OK. The work-horse for findnext.
	See FindNext for details.

	Returns 0 if everthing worked, ERR_xxx if anything failed.
*/
int FindNextLongFilenameHOK(LP_WIN32_SDB lpWin32SDB, bool dosDateMode,
					LP_WIN32_FIND_DATA lpWin32FindData)
{
	unsigned dirnum;				// starting dirnum for found entry
	int		entries;				// how many DIRs it takes up
	CLUSTER	newSubDir;			// the new subdirectory (start cluster)

	ENTER_FUNCTION();

	// validate parameters
	ASSERTLP(lpWin32FindData);
	ASSERTLP(lpWin32SDB);

	//  Some MSDOS utilities expect support for 64-bit file time
	//  so we must support it
	//
	if((dosDateMode != DOSDATE) && (dosDateMode != WIN32DATE))
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}

	// optimization for create
	sda_1st_blank = 0xFFFF;

	if( lpWin32SDB->subdir == END_OF_FAT )
	{
		EXIT_FUNCTION();
		return ERR_LASTFILE;
	}

	// loop until we find one that looks right
	while(1)
	{
		// no dir entries left? we're done
		if(!GetNextLFNEntry(lpWin32SDB, &win32FindData,
									&dirnum, &entries, &newSubDir))
		{
			EXIT_FUNCTION();
			return ERR_LASTFILE;
		}

		//  BT 2-1-2000 ETR 3271
		//
		//  Short name find first calls should not see LFN entries
		//
		if( (win32FindData.attr&0x1F) == 0x0f )
		{
			continue;
		}

		// volume labels follow special rules
		if(lpWin32SDB->searchAttr == A_VOLUME &&
			(win32FindData.attr&0x1F) != A_VOLUME)
		{
			continue;
		}

		// does this dir entry match attributes?
		// if not, continue
		else if(!
			((((lpWin32SDB->mustMatchAttr & ~win32FindData.attr) & 0x3F) == 0)
			&& (((~lpWin32SDB->searchAttr & win32FindData.attr) & 0x1E) == 0)) )
			continue;

		// if name matches mask, we've found one!
		if(MatchLongFilenames(lpWin32SDB->szFileMask,
				win32FindData.szFilename, win32FindData.szAliasFilename))
		{
			lpWin32SDB->startDirNum = dirnum;
			lpWin32SDB->numEntries  = entries;
			break;
		}
	}

	//  If they expect a 64-bit date/time format then we need to
	//  translate it.
	//
	if(dosDateMode == WIN32DATE)
	{
		DosTimeToFileTime( (unsigned)win32FindData.creationTime.lowDateTime,
				(unsigned)(win32FindData.creationTime.lowDateTime >> 16), 0,
				 &win32FindData.creationTime );
	}

	// record new subdirectory
	lpWin32SDB->newSubDir = newSubDir;

	// worked! we've found an entry that matches
	// copy the data to user space.
	FMEMCPY(lpWin32FindData, &win32FindData, FIRST_PART_W32FD);

	// only copy to the NULL on strings.
	IntlFStrCpy(lpWin32FindData->szFilename, win32FindData.szFilename);
	IntlFStrCpy(lpWin32FindData->szAliasFilename, win32FindData.szAliasFilename);

	EXIT_FUNCTION();

	// no errors!
	return 0;
}

/*
	Find the next entry.  The Win32 Search Data Block has already
	been set up by findfirst, provided the given handle is valid.

	Always uses dateMode 1 (DOS date/time)

	Inputs:
		lfnHandle        a handle created by a previous findfirst
		dosDateMode      ignored, we ALWAYS use DOS date mode
	Outputs:
		lpWin32FindData  A filled out find first data structure

	Returns 0 if everything worked.  Returns ERR_xxx if failed.
*/
int FindNextLongFilename(unsigned lfnHandle, bool dosDateMode,
					LP_WIN32_FIND_DATA lpWin32FindData)
{
	LP_WIN32_SDB lpWin32SDB;	// search data block

   ENTER_FUNCTION();

	// make sure handle is valid, get ptr to Win32 SDB
	if(!IsValidHandle(lfnHandle, &lpWin32SDB))
	{
		EXIT_FUNCTION();
		return ERR_BADHANDLE;
	}

	// set up drive variables, drive must still be valid!
	if(!DRIVE_TO_SDA(lpWin32SDB->drive))
	{
		EXIT_FUNCTION();
		return ERR_BADHANDLE;
	}

	EXIT_FUNCTION();

	// pass on through to real work-horse
	return FindNextLongFilenameHOK(lpWin32SDB, dosDateMode, lpWin32FindData);
}

/*
	Standard findfirst, but for long filenames.  Has a few more
	options in the search attributes, and the "*" wildcard works
	at the start, middle, and end.

	Inputs:
		lpPath				The file name/mask used for search
		searchAttrs			Search attributes. Includes MUST match attrs
		dosDateMode			Only DOS date mode allowed
		lpWin32FindData	Ptr to user memory space
	Outputs:
		lpWin32FindData	A filled out find first data structure
		pLfnHandle			A handle used for FindNext and FindClose

	Returns error_no. If error_no is zero, also returns handle to the
	findfirst (for use by findnext).
*/
unsigned FindFirstLongFilename(lpchar lpPath, unsigned searchAttr,
			bool dosDateMode, LP_WIN32_FIND_DATA lpWin32FindData,
			unsigned *pLfnHandle)
{
	unsigned lfnHandle;
	CLUSTER	subdir;
	LP_WIN32_SDB lpWin32SDB;
	char *	pFilename;
	bool fIsDevice = FALSE;
	ulong		ulFileDateTime;


	ENTER_FUNCTION();

	// validate parameters
	ASSERT(FP_SEG(lpWin32FindData));
//	ASSERT(FP_SEG(lpLongFilename));

	//  Some MSDOS utilities expect support for 64-bit file time
	//  so we must support it
	//
	if((dosDateMode != DOSDATE) && (dosDateMode != WIN32DATE))
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}

	// path must be reasonable, wildcards OK
	if(!ResolveLongPathnameCWD(lpPath, szSrcLongName, TRUE, TRUE))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// determine filename
	pFilename = IntlStrRChr(szSrcLongName, '\\') + 1;
	if(!*pFilename)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  Determine if this is a device
	//
	if(IsStandardDOSName( pFilename ))
	{
		DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
		if( FIND_DEVICE(sda_ename1) )
		{
			fIsDevice = TRUE;
		}
	}

	//  Make sure that the path is valid for files and devices
	//
	if( fIsDevice && (szSrcLongName == &lpPath[5]) &&
			(isslash(lpPath[0])) &&
			(TOUPPER(lpPath[1]) == 'D') &&
			(TOUPPER(lpPath[2]) == 'E') &&
			(TOUPPER(lpPath[3]) == 'V') &&
			(isslash(lpPath[4])) )
	{
		//  Device path is valid
	}
	else
	{
		// path must exist for files or device
		//
		if(!DoesPathExist(&subdir))
		{
			EXIT_FUNCTION();
			return ERR_NOTPATH;
		}
	}

	// make sure there is a search handle available
	lfnHandle = GetFreeLFNHandle(&lpWin32SDB);
	if(!lfnHandle)
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	if(!fIsDevice)
	{
		// set up the Win32 Search Data Block
		lpWin32SDB->drive  = sda_drive;
	   lpWin32SDB->dirnum = 0;
	   SetSubdir(lpWin32SDB, subdir);
		lpWin32SDB->mustMatchAttr = (uchar)(searchAttr >> 8);
		lpWin32SDB->searchAttr    = (uchar)searchAttr;

		// copy mask into SDB
		IntlFStrCpy(lpWin32SDB->szFileMask, pFilename);

		// do the findfirst (the very first "next")
		if(FindNextLongFilenameHOK(lpWin32SDB, dosDateMode, lpWin32FindData) != 0)
		{
			// failed to find anything that matches, close down the search
			FindClose(lfnHandle);
			EXIT_FUNCTION();
			return ERR_LASTFILE;
		}
	}
	else
	{
		//  Set the values in the internal data structure
		//
		lpWin32SDB->subdir = END_OF_FAT;

		// determine current date/time, to store in file entry
		ulFileDateTime = GET_FILEDATE();

		//  Set the values in the return data structure
		//
		_FMEMSET(lpWin32FindData, 0, 0x2c);
		lpWin32FindData->attr = A_SYSTEM;
		IntlFStrCpy( lpWin32FindData->szFilename, pFilename );
		lpWin32FindData->szAliasFilename[0] = 0;
		if( dosDateMode == DOSDATE )
		{
			*(ulong far *)(&lpWin32FindData->lastWriteTime) = ulFileDateTime;
		}
		else
		{
			DosTimeToFileTime( FP_SEG(ulFileDateTime),
					(unsigned)ulFileDateTime, 0,
					&lpWin32FindData->lastWriteTime );
		}
	}

	*pLfnHandle = lfnHandle;

	EXIT_FUNCTION();

	// everything worked!
	return 0;
}

/************************************

		Directory Functions

*************************************/

// helper function to ResolveLongPathname
lpchar SkipSlashes(lpchar lpPath)
{
	ENTER_FUNCTION();

	while(isslash(*lpPath))
		++lpPath;

	EXIT_FUNCTION();

	return lpPath;
}

/*
	Resolve a long filename path to an absolute path.  Leaves the
	long names as long names.  Does no disk access.  Drive has
	already been validated and factored out.

	Rules
	1) name can end in a backlash
	2) last name can have wildcards
	3) names stay in long filename form

	lpRoot 		CDS cwd, examples: "\" or "\TEMP"
	lpSrcPath	relative src path, examples: ".", "\temp", "long filename\x"
	pDstPath		absolute path in destination

	Returns TRUE if ok, FALSE if illegal characters in name,
	or name resolves to longer than 255 characters.
*/
bool ResolveLongPathname(lpchar lpCWD, lpchar lpSrcPath, char * pDstPath,
			bool wildCardsOK)
{
	int 		n;					// all purpose length and count
	char * 	pCurDstPtr;		// current position in dst path
	char *	pEndDst;			// don't go past here!
	lpchar 	lpCurSrcPtr;	// current position in src path
	char		c;
	int		dotsFound;		// how many dots where found this part?

	ENTER_FUNCTION();

	// validate arguments
	ASSERTLP(lpCWD);
	ASSERTLP(lpSrcPath);
	ASSERT(pDstPath);

	// empty source path is not legal
	if(*lpSrcPath == 0)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// does source path start at root?
	if(isslash(*lpSrcPath))
	{
		// skip all preceding slashes in source name
		lpSrcPath = SkipSlashes(lpSrcPath);

		// dest starts at root
		*pDstPath = '\\';

		// current position starts at root
		n = 1;
	}

	// or should we use cwd
	else
	{
		// copy the root into destination
		n = FSTRLEN(lpCWD);
		FMEMCPY(pDstPath, lpCWD, n+1);	// get the NULL too

		// make sure dst ends in a backslash
		if(!isslash(LASTCHAR(pDstPath)))
		{
			pDstPath[n++] = '\\';
			pDstPath[n] = 0;
		}
	}

	// set up a current position pointer
	pCurDstPtr = pDstPath + n;
	*pCurDstPtr = 0;		// must be NULL terminated for LASTCHAR()

	// get pointer to maxixum end of string (max string size)
	pEndDst = pDstPath + (MAXLNAME - 1);

	// get all paths
	while(*lpSrcPath)
	{
		// check for special "." and ".." strings
		lpCurSrcPtr = lpSrcPath;
		n = 0;
		while(*lpCurSrcPtr == '.')
		{
			++n;
			++lpCurSrcPtr;
		}

		// assume it was not "." or ".."
		dotsFound = 0;

		// nothing but dots in name, that is "." or ".."
		if(*lpCurSrcPtr == 0 || isslash(*lpCurSrcPtr))
		{
			// record that we found dots directories
			dotsFound = n;

			// skip the "." or the ".."
			lpSrcPath += dotsFound;

			// path was ".", simply stay in current dir.
			// path was "..", go to parent
			// path was "...", go to parents parent.
			// etc...
			//
			if(dotsFound)
			{
				//  Do nothing for the first dot
				//
				dotsFound--;

				//  For each remaining dot back up one directory
				//
				while(dotsFound)
				{
					// don't backup past "root"
					if(pCurDstPtr == pDstPath+1)
					{
						EXIT_FUNCTION();
						return FALSE;
					}

					// erase most recent backlash
					--pCurDstPtr;
					ASSERT(*pCurDstPtr == '\\');
					*pCurDstPtr = 0;

					// backup to backslash BEFORE the one above
					pCurDstPtr = IntlStrRChr(pDstPath, '\\');

					// should ALWAYS find at least the root backslash
					ASSERT(*pCurDstPtr != 0);

					// point to just AFTER the slash
					++pCurDstPtr;

					//  We have processed this dot
					//
					dotsFound--;
				}

				//  Restore the value of the varaible
				//
				dotsFound = n;
			}
		}

		// must be a name, copy each character
		else
		{
			bool		gotWildCards;

			// assume no wildcards
			gotWildCards = FALSE;

			// get all the characters from the name
			while(pCurDstPtr < pEndDst)
			{
				// done when we get to the slash
				c = *lpSrcPath;			// "c" is code size/speed optimization
				if( !c )
					break;

				if( isslash(c) )
				{
					while( *(pCurDstPtr-1) == ' ' )
					{
						pCurDstPtr--;
					}
					break;
				}

				// does this part of the path contain wildcards?
				if(IsWildCard(c))
					gotWildCards = TRUE;

				// handle lead bytes (they are two bytes long)
				if(IS_LEAD_BYTE(c))
					*pCurDstPtr++ = *lpSrcPath++;

				// on to next character
				*pCurDstPtr++ = *lpSrcPath++;
			}

			// string too long!
			if(pCurDstPtr >= pEndDst)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

			// wildcards NOT OK, even on name part
			if(!wildCardsOK && gotWildCards)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

			// is there more to the source path string than this name?
			if(isslash(*lpSrcPath))
			{
				// wildcards not allowed in path part of name
				if(gotWildCards)
				{
					EXIT_FUNCTION();
					return FALSE;
				}

				// append backslash
				*pCurDstPtr++ = '\\';
			}
		}

		// skip all slashes
		lpSrcPath = SkipSlashes(lpSrcPath);

		// make sure dest string is NULL terminated
		*pCurDstPtr = 0;
	}

	// if we just processed a "." or "..", don't have a trailing
	// backslash in subdirs.  don't back up if we're at root.
	if(dotsFound && pCurDstPtr != pDstPath + 1)
	{
		--pCurDstPtr;
		ASSERT(*pCurDstPtr == '\\');
		*pCurDstPtr = 0;
	}


	//  Remove any white space from the end of the string
	//
	pCurDstPtr = pDstPath + STRLEN(pDstPath) - 1;
	if( *pCurDstPtr == ' ' )
	{
		*pCurDstPtr = 0;
		pCurDstPtr--;
	}

	EXIT_FUNCTION();

	// everything worked
	return TRUE;
}

/*
	Handy function, already knows about the CDS and the current
	working directory.
*/
bool ResolveLongPathnameCWD(lpchar lpSrcPath, char * pDstPath,
		bool wildCardsOK, bool fExpandSubst )
{
#if 0
	return ResolveLongPathname(sda_cdsptr->cwd+sda_cdsptr->skip,
		lpSrcPath, pDstPath, wildCardsOK);
#else
	bool fSuccessful;
	unsigned uDstLength;

	ENTER_FUNCTION();

	fSuccessful = ResolveLongPathname(sda_cdsptr->cwd+sda_cdsptr->skip,
		lpSrcPath, pDstPath, wildCardsOK);


	//  Determine if we need to expand the substed drive path
	//
	if( (fSuccessful) && (fExpandSubst) )
	{
		if( sda_cdsptr->status & C_SUBST )
		{
			uDstLength = STRLEN( pDstPath );
			if( uDstLength == 1 )
			{
				pDstPath[0] = 0;
				uDstLength = 0;
			}
			MEMMOVE( pDstPath + ((sda_cdsptr->skip) - 2), pDstPath,
					uDstLength + 1 );
			FMEMCPY( pDstPath, sda_cdsptr->cwd + 2, ((sda_cdsptr->skip) - 2) );
			DRIVE_TO_SDA( sda_cdsptr->cwd[0] - 'A' );
		}
	}

	EXIT_FUNCTION();

	return fSuccessful;
#endif
}

/*
	Get enough free directory entries.  If a subdirectory is full,
	will try to expand that directory.  Cannot expand root dir.

	Returns FALSE if not enough free entries.
	Returns TRUE and the starting DIR number if worked.
*/
bool DeliverFreeDirEntries(LP_WIN32_SDB lpWin32SDB, int n, int *pDirNum)
{
	int		numEmpty;
	int		startDir;
	LPGENDIR lpDir;

	ENTER_FUNCTION();

	// validate parameters
	ASSERT(n);			// asking for at least 1 free entry
	ASSERT(pDirNum);

	// the search is optimized if FindFirst already found a blank entry
	if(sda_1st_blank == 0xFFFF)
		startDir = 0;
	else
		startDir = sda_1st_blank;
	lpWin32SDB->dirnum = startDir;

	numEmpty = 0;
	while(numEmpty < n)
	{
		// at end of directory, try to create more entries!
		while(!GetDirEntry(lpWin32SDB, &lpDir))
		{
			if(!CreateMoreDirEntries(n - numEmpty, lpWin32SDB->subdir))
			{
				EXIT_FUNCTION();
				return FALSE;
			}
		}

		// we found an empty/deleted entry!
		if(IsEndOfDirList(lpDir) || IsDeleted(lpDir->lfnDir.sequence))
		{
			// remember start of consecutive empty DIRs
			if(numEmpty == 0)
				startDir = lpWin32SDB->dirnum;
			++numEmpty;
		}

		else
			numEmpty = 0;			// no empties yet!

		// on to next directory
		++lpWin32SDB->dirnum;
	}

	*pDirNum = startDir;

	EXIT_FUNCTION();

	// found them!
	return TRUE;
}

/*
	Get a single name. "\Hello World\My thing"
	It expects that pSrc points to just after "\"
*/
char * GetName(lpchar lpDstName, char * pSrcName)
{
	ENTER_FUNCTION();

	// skip any preceding slash
	if(*pSrcName == '\\')
		++pSrcName;

	// loop through, copying characters
	while(*pSrcName && *pSrcName != '\\')
		*lpDstName++ = *pSrcName++;

	// null terminate the string
	*lpDstName = 0;

	EXIT_FUNCTION();

	return pSrcName;
}

void GetRoot(char *pRoot)
{
	ENTER_FUNCTION();

	// get the root part of the current CDS (eg "B:" or "A:")
#if 0
	FMEMCPY(pRoot, sda_cdsptr->cwd, sda_cdsptr->skip);
	pRoot[sda_cdsptr->skip] = 0;
#else
	pRoot[0] = sda_drive + 'A';
	pRoot[1] = ':';
	pRoot[2] = 0;
#endif

	EXIT_FUNCTION();
}

/*
	Validate a long path. This will check the disk to make sure the
	entire path exists. Does NOT care if the filename part exists.

	On Entry
		pSrcPath		an absolute path already parsed by ResolveLongPathname()
	On Exit
		Returns TRUE if worked, and win32SDB is set up.
		Returns FALSE if failed (invalid path).
*/
bool ValidateLongPath(char *pSrcPath)
{
	char *pSrcPathPtr;
	char *pDstShortPtr;
	char *pDstLongPtr;
	uchar ucTempSdaDrive;

	ENTER_FUNCTION();

	// just in case the disk has changed
	CHANGE_OF_DISK();

	//  Set the drive and starting directory.  Handle subst drives
	//  specially.
	//
	win32SDB.drive  = sda_drive;
	pSrcPathPtr = pSrcPath;
	if( sda_cdsptr->status & C_SUBST )
	{
		ucTempSdaDrive = sda_drive;
		IntlFStrCpy( szTempName, sda_cdsptr->cwd + 2 );
      //
      // The CDS path for a substed drive contains the original root,
      //    subst path and CWD.  Skip points just past the subst path.
      // Zero szTempName at that point -- we only need to verify the
      //    subst path in this recursion.
      //
      szTempName[(sda_cdsptr->skip)-2] = 0x00;
		DRIVE_TO_SDA( sda_cdsptr->cwd[0] - 'A' );
		ValidateLongPath( szTempName );
		DRIVE_TO_SDA( ucTempSdaDrive );
		sda_drive =	ucTempSdaDrive;
	}
	else
	{
		win32SDB.newSubDir = 0;
	}

	// copy in CWD into dest
	GetRoot(szDstShortName);
	GetRoot(szDstLongName);
	pDstShortPtr = szDstShortName + STRLEN(szDstShortName);
	pDstLongPtr  = szDstLongName  + STRLEN(szDstLongName);

	// assume nothing is found at this point!
	win32SDB.found = FALSE;

	// initial found
	win32MoreFoundData.subdir = 0;
	win32MoreFoundData.onFilename = FALSE;

	// start out, no names either
	*win32FindData.szFilename =
	*win32FindData.szAliasFilename = 0;

	// start at root, go through each DIR getting long filename
	while(*pSrcPathPtr)
	{
		// record the new subdirectory we are about to enter
	   SetSubdir(&win32SDB, win32SDB.newSubDir);

		// keep the "more" found data up to date as well
		win32MoreFoundData.subdir = win32SDB.subdir;

		// copy in the trailing backslash
		if(*pSrcPathPtr == '\\')
		{
			*pDstShortPtr = *pDstLongPtr = *pSrcPathPtr;
			++pDstShortPtr;
			++pDstLongPtr;
			++pSrcPathPtr;

			// null terminate the string in case we're done
			*pDstShortPtr = *pDstLongPtr = 0;
		}

		// get the entry name from the path string
		pSrcPathPtr = GetName(win32SDB.szFileMask, pSrcPathPtr);
		if(*win32SDB.szFileMask == 0)
			break;

		// if there are more directory entires, we're looking for a DIR
		win32SDB.searchAttr = A_DIRCHK;
		if(*pSrcPathPtr)
			win32SDB.mustMatchAttr = A_SUBDIR;

		// otherwise looking for either a file or subdir
		else
		{
			// we are on the filename part, get some help determining
			// a unique alias. Used in the GetNextLFNEntry() function.
			fAliasHelper = TRUE;

			// can match file or directory.
			win32SDB.mustMatchAttr = 0;
			win32MoreFoundData.onFilename = TRUE;
		}

		// start at beginning of directory
		win32SDB.dirnum = 0;

		// find failed!
		if(FindNextLongFilenameHOK(&win32SDB, DOSDATE, &win32FindData) != 0)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// no longer want the alias helper.
		fAliasHelper = FALSE;

		// now something is found!
		win32SDB.found = TRUE;

		// copy in the longname
		IntlStrCpy(pDstLongPtr,  win32FindData.szFilename);

		// is there an alias? use it
		if(*win32FindData.szAliasFilename)
			IntlStrCpy(pDstShortPtr, win32FindData.szAliasFilename);

		// no alias, just the one 8.3 name
		else
			IntlStrCpy(pDstShortPtr, win32FindData.szFilename);

		// skip to end of strings
		pDstShortPtr += STRLEN(pDstShortPtr);
		pDstLongPtr  += STRLEN(pDstLongPtr);
	}

	EXIT_FUNCTION();

	// everything worked!
	return TRUE;
}

/*
	Helper function. Deletes the DIR entries (marks with E5).
*/
bool DeleteEntries(int numToDel, CLUSTER subdir, int dirnum)
{
	LPGENDIR			lpDir;

   ENTER_FUNCTION();

	// set up fields for a
	win32SDB.drive  = sda_drive;
   SetSubdir(&win32SDB, subdir);
	win32SDB.dirnum = dirnum;

	// delete them all
	while(numToDel > 0)
	{
		// delete each entry
		if(!GetDirEntry(&win32SDB, &lpDir))
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		// delete the entry
		lpDir->lfnDir.sequence = DELETED;
		BUF_SET_MODIFIED(sda_cur_buf);

		// on to next entry
		++win32SDB.dirnum;
		--numToDel;
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Assumes name has been parsed into "szSrcLongName".

	Checks to make sure the directory exists, and returns its
	starting cluster number.

	Returns TRUE if path exists, FALSE if not.
*/
bool DoesPathExist(CLUSTER *pSubdir)
{
	char *pFilename;
	bool	pathOK;
	bool	dirExists;

   ENTER_FUNCTION();

	// look for the filename part
	pFilename = IntlStrRChr(szSrcLongName, '\\');

	// the root dir always exists, just it pass back
	if(pFilename == szSrcLongName)
	{
		*pSubdir = 0;

		EXIT_FUNCTION();

		return TRUE;
	}

	// make sure just the DIR part exists
	*pFilename = 0;
	dirExists = DoesEntryExist(&pathOK);
	*pFilename = '\\';

	// directory does NOT exist
	if(!dirExists)
	{
		EXIT_FUNCTION();

		return FALSE;
	}

	// must be a directory we're looking at
	if((win32FindData.attr & A_SUBDIR) == 0)
	{
		EXIT_FUNCTION();

		return FALSE;
	}

	// return starting cluster
	*pSubdir	= win32MoreFoundData.startClus;

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Expects the path to be in "szSrcLongName"

	Returns TRUE if the entry exists, FALSE if not.

	If the entry does exist, fills in a number of "extended"
	SDA structures: The "moreWin32FindData" structure, the
	"szSrc/DstPath" structures.
*/
bool DoesEntryExist(bool *pPathOK)
{
	ENTER_FUNCTION();

	// and entry must be found
	if(!ValidateLongPath(szSrcLongName))
	{
		// let user know whether path was fine, just name not found
		*pPathOK = win32MoreFoundData.onFilename;

		EXIT_FUNCTION();

		// everything failed
		return FALSE;
	}

	// the root by itself doesn't exist!
	if(*win32FindData.szFilename == 0)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// set up the More Found fields
	win32MoreFoundData.subdir      = win32SDB.subdir;
	win32MoreFoundData.startClus   = win32SDB.newSubDir;
	win32MoreFoundData.startDirNum = win32SDB.startDirNum;
	win32MoreFoundData.numEntries  = win32SDB.numEntries;

	// path was obviously fine
	*pPathOK = TRUE;

	EXIT_FUNCTION();

	return TRUE;
}

/*
	copy unicode characters into lfndir
*/
void CopyUniChars(struct LFN_DIR *pLfnDir, char * pSrcStr)
{
	int		i;
	bool		endOfName = FALSE;
	uniChar *pUniStr;

	ENTER_FUNCTION();

	pUniStr = pLfnDir->fileName1;
	for(i=0; i<NUM_NAME1; ++i)
	{
		if(endOfName)
			*pUniStr = 0xFFFF;
		else
		{
			*pUniStr = *pSrcStr;
			if(*pSrcStr == 0)			// just copied last char in name
				endOfName = TRUE;
			++pSrcStr;
		}

		++pUniStr;
	}

	pUniStr = pLfnDir->fileName2;
	for(i=0; i<NUM_NAME2; ++i)
	{
		if(endOfName)
			*pUniStr = 0xFFFF;
		else
		{
			*pUniStr = *pSrcStr;
			if(*pSrcStr == 0)			// just copied last char in name
				endOfName = TRUE;
			++pSrcStr;
		}

		++pUniStr;
	}

	pUniStr = pLfnDir->fileName3;
	for(i=0; i<NUM_NAME3; ++i)
	{
		if(endOfName)
			*pUniStr = 0xFFFF;
		else
		{
			*pUniStr = *pSrcStr;
			if(*pSrcStr == 0)			// just copied last char in name
				endOfName = TRUE;
			++pSrcStr;
		}

		++pUniStr;
	}

	EXIT_FUNCTION();
}

/*
	Helper function to Create File

	Writes the longname part of a series of DIRs to disk.
	Doesn't flush.
*/
void WriteLongNameToDisk(struct WIN32_SDB *pWin32SDB, char *pFilename,
			int numEntries, uchar checksum)
{
	LPGENDIR			lpDir;
	bool				firstTime = TRUE;
	struct LFN_DIR	lfnDir;

	ENTER_FUNCTION();

	while(numEntries)
	{
		// we've already verified we can get these entries
		if(!GetDirEntry(pWin32SDB, &lpDir))
			ABORT();

		// fill with 0's
		MEMSET(&lfnDir, 0, sizeof(lfnDir));

		// set the sequence #
		lfnDir.sequence = numEntries;
		if(firstTime)
		{
			lfnDir.sequence |= 0x40;
			firstTime = FALSE;
		}

		// set attribute
		lfnDir.attr     = A_LONGNAME;
		lfnDir.checksum = checksum;

		// copy Unicode characters
		CopyUniChars(&lfnDir, pFilename + (numEntries-1) * CHARS_IN_LFN_DIR);

		// write DIR to actual disk buffer
		FMEMCPY(lpDir, &lfnDir, sizeof(lfnDir));
		BUF_SET_MODIFIED(sda_cur_buf);

		// on to next entry
		--numEntries;
		++pWin32SDB->dirnum;
	}

	EXIT_FUNCTION();
}

/*
	Helper function to Create File

	Put the normal DOS 8.3 entry on disk
*/
void CreateWin32DirOnDisk(P_WIN32_SDB pWin32SDB, char *ename,
													CLUSTER startCluster, int attr)
{
	LPGENDIR			lpDir;
	struct WIN_DIR	winDir;
	ulong				fileDateTime;

	ENTER_FUNCTION();

	// determine current date/time, to store in file entry
	fileDateTime = GET_FILEDATE();

	// we've already verified we can get these entries
	if(!GetDirEntry(pWin32SDB, &lpDir))
		ABORT();

	// initialize everything to 0
	MEMSET(&winDir, 0, sizeof(winDir));

	// set up name
	MEMMOVE(winDir.name, ename, ENAMESIZE);
#if 1
   if( (uchar)winDir.name[0] == (uchar)0xE5 )
   {
      // translate this character to a 0x05 for storage on the disk
      //
      winDir.name[0] = 0x05;
   }
#endif  // #if 1

	// attribute
	winDir.attr  = attr;
#if USING_FAT32
	winDir.uStartingClusterLo = LOWORD(startCluster);
	winDir.uStartingClusterHi = HIWORD(startCluster);
#else
	winDir.start = startCluster;
#endif

	// and dates
	winDir.creationTime = winDir.lastModTime =
		(unsigned)(fileDateTime >> 16);
	winDir.creationDate = winDir.lastModDate = winDir.lastAccessDate =
		(unsigned)(fileDateTime);

	// write DIR to actual disk buffer
	FMEMCPY(lpDir, &winDir, sizeof(winDir));
	BUF_SET_MODIFIED(sda_cur_buf);

	EXIT_FUNCTION();
}

/*
	Helper to Create and Rename.  Given the path, parse just the
	filename part into win32FindData, and put set alias (if any)
	correctly.

	Returns # of DIR entries needed to hold entire longname + alias.
*/
int FillInFindDataName(char *pPath)
{
	char *pLongFilename;
	int	numEntries;

	ENTER_FUNCTION();

	// find the name at the end of the path
	pLongFilename = IntlStrRChr(pPath, '\\') + 1;

	// copy in the long filename part
	IntlStrCpy(win32FindData.szFilename, pLongFilename);

	// the same? then no alias needed
	if(strcmp(pLongFilename, szProposedAlias) == 0)
		*win32FindData.szAliasFilename = 0;
	else
		strcpy(win32FindData.szAliasFilename,szProposedAlias);

	numEntries = 1;
	if(*win32FindData.szAliasFilename)
		numEntries += EntriesForLongname(pLongFilename);

	EXIT_FUNCTION();

	return numEntries;
}

/*
	Helper to create file/directory and rename.  Assumes that
	the name has already been Resolved to an abs path and
	Validated as to whether the directory exists.

	Returns TRUE if worked, and the dirnum and numentries.
	Returns FALSE if failed (not enough entries)
*/
bool CreateGenericEntry(uchar attr, CLUSTER startCluster,
										int *pDirnum, int *pNumEntries)
{
	int		dirnum;
	int		numEntries;
	char *	shortName;
	uchar		checksum;
	char		ename[ENAMESIZE];

	ENTER_FUNCTION();

	// determine how many entries we'll need
	numEntries = FillInFindDataName(szSrcLongName);

	// get enough free entries
	win32SDB.drive  = sda_drive;
   SetSubdir(&win32SDB, win32MoreFoundData.subdir);
	win32SDB.eps    = 0;									// ignore cache
	if(!DeliverFreeDirEntries(&win32SDB, numEntries, &dirnum))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// convert alias to ename
	if(*win32FindData.szAliasFilename)
		shortName = win32FindData.szAliasFilename;
	else
		shortName = win32FindData.szFilename;
	DOT_TO_ENAME(ename, shortName, TRUE);

	// checksum the ename
	checksum = DetermineChecksum(ename);

	// write the lfn entries to disk
	win32SDB.dirnum = dirnum;
	if(*win32FindData.szAliasFilename)
	{
		WriteLongNameToDisk(&win32SDB, win32FindData.szFilename,
			numEntries-1, checksum);
	}

	// write the DOS 8.3 entry to disk
	CreateWin32DirOnDisk(&win32SDB, ename, startCluster, attr);

	// return start/size back to caller
	*pDirnum = dirnum;
	*pNumEntries = numEntries;

	EXIT_FUNCTION();

	// everything worked!
	return TRUE;
}

/*
	Assumes drive has already been handled.
	Returns TRUE if worked, FALSE if failed (bad path).
*/
bool ChangeLongCWD(lpchar lpSrcName)
{
	CLUSTER	subdir;
	char * pTemp;

   ENTER_FUNCTION();

	// resolve into an absolute path
	if(!ResolveLongPathnameCWD(lpSrcName, szSrcLongName, FALSE, FALSE))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// make sure it's valid
	if(!ValidateLongPath(szSrcLongName))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// must have been root. OK
	if(!win32SDB.found)
		subdir = 0;

	// subdirectory of some kind
	else
	{
		// remember new subdirectory
		subdir = win32SDB.newSubDir;

		// make sure final entry actual WAS a directory
		if((win32FindData.attr & A_SUBDIR) == 0)
		{
			EXIT_FUNCTION();
			return FALSE;
		}
	}


	//  Remove extra trailing backslash before it is copied into
	//  the CDS but not if it is just c:\...
	//
	pTemp = IntlStrRChr( szDstShortName, '\\' );
	if( pTemp )
	{
		pTemp--;
		if( pTemp[2] == 0 )
		{
			if( pTemp[0] != ':' )
			{
				//  Remove the trailing backslash
				//
				pTemp[1] = 0;
			}
		}
	}


	// change directory worked. copy in short name
	//
#if USING_FAT32
	sda_cdsptr->ulSubDirStartingCluster = subdir;
#else
	sda_cdsptr->subdir = subdir;
#endif
	if( sda_cdsptr->status & C_SUBST )
	{
		if( szDstShortName[3] == 0 )
		{
			IntlFStrCpy(sda_cdsptr->cwd + sda_cdsptr->skip,
					szDstShortName + 3 );
		}
		else
		{
			IntlFStrCpy(sda_cdsptr->cwd + sda_cdsptr->skip,
					szDstShortName + 2 );
		}
	}
	else
	{
		IntlFStrCpy(sda_cdsptr->cwd, szDstShortName);
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	The only way this can fail is with a bad drive, which has
	already been parsed, or if the CWD is messed up somehow.

	Returns TRUE if worked, and fill in LongPath
*/
bool GetLongCWD(lpchar lpLongPath)
{
	int skip;

	ENTER_FUNCTION();

	ASSERTLP(lpLongPath);

	// convert relative path to absolute path, wildcards NOT OK
	if(!ResolveLongPathnameCWD(".", szSrcLongName, FALSE, FALSE))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// actually get the long path in string form
	// if it fails, move back to the root
	if(!ValidateLongPath(szSrcLongName))
	{
		CHANGE_TO_ROOT();
		*szDstLongName= 0;
	}

   // got it! return it to user, but without preceding backslash
   if(*szDstLongName == 0)
   {
      *lpLongPath = 0;
   }
   else
   {
      if( sda_cdsptr->status & C_SUBST )
      {
         *lpLongPath = 0;

         if( szDstLongName[3] == 0 )
         {
            IntlFStrCpy( lpLongPath, szDstLongName + 3 );
         }
         else
         {
            IntlFStrCpy( lpLongPath, szDstLongName + 2 );
         }
      }
      else
      {
         skip = sda_cdsptr->skip + 1;
         IntlFStrCpy(lpLongPath, szDstLongName+skip);
      }
   }

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Is the name a wildcard name?
*/
bool IsWildName(char *szName)
{
   ENTER_FUNCTION();

	if(STRCHR(szName, '?'))
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	if(STRCHR(szName, '*'))
	{
		EXIT_FUNCTION();
		return TRUE;
	}

	EXIT_FUNCTION();

	return FALSE;
}

/*
	Delete a file or directory.  Takes care of freeing up the FAT,
	and all other details. Works for files and directories.

	On Entry
	    lpPath		pointer to a relative path
	    fFlags		DEL_DIR | DEL_WILDOK | DEL_CXATTR
	    isDir    TRUE=delete dir, FALSE=delete files. Mutually exclusive.
	    wildOK   TRUE=allow wildcards.

	Returns 0 if everything worked (file or dir is deleted)
	Returns ERR_xxx if failed (not found, dir not empty, etc..)
*/
int DeleteLongFileOrDirectory(lpchar lpPath, int fFlags)
{
	int		numToDel;					// number of entries to delete
	bool		entryExists, pathOK;
	bool		wildOK = (fFlags & DEL_WILDOK);
	bool		hasWild = FALSE;
	int		numFiles;
	unsigned uReadOnlyMask;
	char *	pFilename;


	ENTER_FUNCTION();

	// number of files and/or directories that match
	numFiles = 0;

	// the path, of course, must be valid, no wildcards
	if(!ResolveLongPathnameCWD(lpPath, szSrcLongName, wildOK, FALSE))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// does the file/dir exist?
	entryExists = DoesEntryExist(&pathOK);
	if(!entryExists)
	{
		//
		//  ETR ? BT 08-16-00 / 08-24-00
		//
		//  We should return a file not found error code when the file
		//  is not found.  The code above validated the base path and
		//  returned invalid path if the base path was not valid.
		//
		//		return ERR_NOTPATH;
		//
		if( R_AX == 0x7141 )
		{
			EXIT_FUNCTION();
			return ERR_NOTFOUND;
		}
		else
		{
			EXIT_FUNCTION();
			return ERR_NOTPATH;
		}
	}


	//  BT 1-26-2000 ETR 3263
	//
	//  You cannot delete devices
	//
	pFilename = IntlStrRChr(szSrcLongName, '\\') + 1;
	if(!*pFilename)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  Determine if this is a device
	//
	if(IsStandardDOSName( pFilename ))
	{
		DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
		if( FIND_DEVICE(sda_ename1) )
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
	}

	// don't allow deleting of root dir
	if(*win32FindData.szFilename == 0)
	{
		EXIT_FUNCTION();
		return ((fFlags & DEL_DIR) ? ERR_DENIED : ERR_NOTFOUND);
	}

	// for deleting wildcards, we want to know if we have wildcards
	if(wildOK)
		hasWild = IsWildName(szSrcLongName);

	// may loop if deleting with wildcards
MoreFiles:

	// long filenames have more than one entry to erase
	numToDel = 1;
	if(*win32FindData.szAliasFilename)
		numToDel += EntriesForLongname(win32FindData.szFilename);

	// if we're deleting a DIR, make sure it IS a DIR
	if(fFlags & DEL_DIR)
	{
		// don't delete if it's a file or volume
		if((win32FindData.attr & A_SUBDIR) == 0 || (win32FindData.attr & A_VOLUME))
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// make sure SUBDIR we are deleting is empty
		win32SDB.drive  = sda_drive;
	   SetSubdir(&win32SDB, win32MoreFoundData.startClus);
		win32SDB.dirnum = 2;			// skip the "." and ".." entries
		win32SDB.searchAttr = A_DIRCHK;
		win32SDB.mustMatchAttr = 0;
		IntlStrCpy(win32SDB.szFileMask, (fShortNameFunctions ? "*.*" : "*"));

		// findnext had better not find anything. If it does, don't
		// delete the directory.
		if(FindNextLongFilenameHOK(&win32SDB, DOSDATE, &win32FindData) == 0)
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
	}

	// file only stuff
	else
	{
		// volume labels are deleted with DOS1 call, no wild cards
		// subdirectories can't be deleted either
//
//    We should not delete read only files for the function 41 calls
//    but should delete them for a 7141
//
//		if(win32FindData.attr & (A_SUBDIR | A_VOLUME | A_READONLY))
//
		if( R_AX == 0x7141 )
		{
			uReadOnlyMask = 0;
		}
		else
		{
			uReadOnlyMask = A_READONLY;
		}

		if(win32FindData.attr & (A_SUBDIR | A_VOLUME | uReadOnlyMask))
		{
			// skip directory entries if deleting wildcards
			if(hasWild)
				goto Continue;
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// search attributes must match for LFN delete
		else if(fFlags & DEL_CXATTR)
		{
			if(!
				((((R_CH & ~win32FindData.attr) & 0x3F) == 0)
					&& (((~R_CL & win32FindData.attr) & 0x1F) == 0)) )
			{
				if(hasWild)
					goto Continue;
				EXIT_FUNCTION();
				return ERR_DENIED;
			}
		}

		// DOS1 style attribute check
		else
		{
			// don't delete read-only files
			if(win32FindData.attr & A_READONLY)
			{
				if(hasWild)
					goto Continue;
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			// don't find it unless hidden bits OK
			if(!ATTR_MATCH(win32FindData.attr, sda_srchattr))
			{
				if(hasWild)
					goto Continue;
				EXIT_FUNCTION();
				return ERR_NOTFOUND;
			}
		}
	}


	//  Don't remove the current directory
	//
	if( R_AX == 0x713a )
	{
#if USING_FAT32
		if( sda_cdsptr->ulSubDirStartingCluster ==
				win32MoreFoundData.startClus )
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
#else
		if( sda_cdsptr->subdir == win32MoreFoundData.startClus )
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
#endif
	}


	// found a file (or dir) to delete, remember how many
	++numFiles;

	// OK, now let's go to work
	// first, free the fat
	if(win32MoreFoundData.startClus)
	{
		FREE_THE_FAT(win32MoreFoundData.startClus);
		win32SDB.eps = 0;					// ignore cache
	}

	// then remove all entires
	DeleteEntries(numToDel,
		win32MoreFoundData.subdir,
		win32MoreFoundData.startDirNum);

	// may need to search again
	if(wildOK)
	{
Continue:
		// look for another entry
		if(FindNextLongFilenameHOK(&win32SDB, DOSDATE, &win32FindData) == 0)
		{
			// set up the More Found fields
			win32MoreFoundData.subdir      = win32SDB.subdir;
			win32MoreFoundData.startClus   = win32SDB.newSubDir;
			win32MoreFoundData.startDirNum = win32SDB.startDirNum;
			win32MoreFoundData.numEntries  = win32SDB.numEntries;
			goto MoreFiles;
		}
	}

	// flush everthing to disk
	BUF_FLUSH_ALL();

	// why did we delete no files? because a DIR matched, but no files!
	if(wildOK && numFiles==0)
	{
		EXIT_FUNCTION();
		return ERR_NOTFOUND;
	}

	EXIT_FUNCTION();

	// everything worked!
	return 0;
}

/*
	Create a long filename directory.  Assumes drive has already
	been parsed.

	Returns 0 if worked.
	Returns ERR_xxx if failed.
*/
int CreateLongDir(lpchar lpPath)
{
	bool		pathOK;
	int		dirnum, numEntries;
	CLUSTER	dirStart;
	CLUSTER	parentDir;
	char * pFilename;

	ENTER_FUNCTION();

	// Resolves the path and makes the initial alias
	if(!MakeInitialAlias(lpPath, 1))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// if it already exists, don't recreate it
	if(DoesEntryExist(&pathOK))
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// invalid path listed
	if(!pathOK)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  BT 1-27-2000 ETR 3264
	//
	//  You cannot create directories with the name of devices
	//
	pFilename = IntlStrRChr(szSrcLongName, '\\') + 1;
	if( *pFilename )
	{
		//  Determine if this is a device
		//
		if(IsStandardDOSName( pFilename ))
		{
			DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
			if( FIND_DEVICE(sda_ename1) )
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}
		}
	}

	//  bt 1-25-00 ETR 3199
	//
	//  Make sure that the short name directory does not exceed
	//  the DOS 6.22 limits
	//
	ValidateLongPath(szSrcLongName);
	if( (STRLEN(szDstShortName) + STRLEN(szProposedAlias)) > PATHLEN )
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}


	//  ETR 3502  BT 6-26-00
	//
	//  This is a temporary fix.  Global variables set by the previous
	//  call to MakeInitialAlias are being destroyed (changed) by the
	//  the two previous ETR (3264,3199) fixes.  Calling MakeInitialAlias
	//  will set the globals correctly again.
	//
	MakeInitialAlias(lpPath, 1);


	//  ETR ? BT 12-06-01
	//
	//  The GLOBAL for creating an alias (fAliasHelper) was set by
	//  a lower level call in ValidateLongPath.  DoesEntryExist will
	//  call ValidateLongPath as well which will once again mess up
	//  or alias.  We must clear this boolean to prevent our alias
	//  from getting a ~1 added to it.
	//
	fAliasHelper = FALSE;


	//  ETR ? BT 8-24-00
	//
	//  The above temporary fix destroyed our alias when a particular
	//  alias is already in use.  Get back the correct alias
	//
	DoesEntryExist(&pathOK);


	// allocate one cluster for the subdir, zero it
	dirStart = ALLOC_NEXT_FREE();
	if(dirStart == END_OF_FAT)
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}
	SET_FAT_ENTRY( dirStart, END_OF_FAT );
	ZERO_CLUSTER(dirStart);

	// if failed to create entry, free up fat
	if(!CreateGenericEntry(A_SUBDIR, dirStart, &dirnum, &numEntries))
	{
		// free the fat entry
#if USING_FAT32
#if DEBUG
			ASSERT( dirStart >= 2 );
			if( IsFat32Drive(sda_dpbptr) )
			{
				ASSERT( dirStart <= sda_dpbptr->ulMaxCluster );
			}
			else
			{
				ASSERT( dirStart <= sda_dpbptr->hiclus );
			}
#endif
#else
#if DEBUG
			ASSERT( dirStart >= 2 );
			ASSERT( dirStart <= sda_dpbptr->hiclus );
#endif
#endif
		SET_FAT_ENTRY(dirStart, 0);
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// remember the subdir where we made the entry
	parentDir = win32SDB.subdir;

	// create the "." entry in the new subdir
	MEMSET(sda_ename1, ' ', ENAMESIZE);
	sda_ename1[0] = '.';
   SetSubdir(&win32SDB, dirStart);
	win32SDB.dirnum = 0;
	CreateWin32DirOnDisk(&win32SDB, sda_ename1, dirStart, A_SUBDIR);

	// create the ".." entry in the new subdir
	sda_ename1[1] = '.';
   SetSubdir(&win32SDB, dirStart);
	win32SDB.dirnum = 1;
	CreateWin32DirOnDisk(&win32SDB, sda_ename1, parentDir, A_SUBDIR);

	// flush all that data to disk
	BUF_FLUSH_ALL();

	EXIT_FUNCTION();

	return 0;
}

/************************************

		File Functions

*************************************/

/*
	Gets the drive data into the SDA.  Uses a path as input.
	Returns TRUE and ptr to new path if worked.
	Returns FALSE if failed.
*/
bool GetDrive(lpchar * plpPath)
{
	uchar drive;

	ENTER_FUNCTION();

	// determine drive #
	if(**plpPath && *(*plpPath+1)==':')
	{
		drive = TOUPPER(**plpPath)-'A';
		*plpPath += 2;
	}
	else
		drive = CUR_DRIVE;

	// set up the drive, may fail if drive is invalid
	if(!DRIVE_TO_SDA(drive))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	EXIT_FUNCTION();

	// everything worked
	return TRUE;
}

/*
	Gets a DIR ptr to the current short entry, as described
	by the "win32MoreFoundData"
*/
LPGENDIR GetCurrentShortDirEntry(void)
{
	LPGENDIR	lpDir;

	ENTER_FUNCTION();

	// read in the sector containing the 8.3 DIR
	win32SDB.drive  = sda_drive;
	win32SDB.subdir = win32MoreFoundData.subdir;
	win32SDB.dirnum = win32MoreFoundData.startDirNum +
								win32MoreFoundData.numEntries - 1;
	win32SDB.eps    = 0;		// ignore cache

	// it HAS to work, we were just looking at this entry!
	if(!GetDirEntry(&win32SDB, &lpDir))
		ABORT();

	EXIT_FUNCTION();

	return lpDir;
}

#if UPDATE_LFN_LAST_ACCESSED_DATE
/*
	Set the last access date for the entry currently
	in the longname SDA.
*/
void UpdateLastAccessDate(void)
{
	LPGENDIR	lpDir;
	ulong		fileDateTime;	// hi part is time, lo is date

	ENTER_FUNCTION();

	fileDateTime = GET_FILEDATE();
	lpDir = GetCurrentShortDirEntry();
	lpDir->winDir.lastAccessDate = (unsigned)fileDateTime;
	BUF_SET_MODIFIED(sda_cur_buf);

	EXIT_FUNCTION();
}
#endif  // #if UPDATE_LFN_LAST_ACCESSED_DATE

/*
	Determine which will be the new DOS handle.  Doesn't
	allocate it yet (in case there's a disk critical error,
	like write-protect).

	Returns FALSE if none available.
*/
bool DetermineNewDosHandle(void)
{
	ENTER_FUNCTION();

	// get an SFT
	if(!GET_NEW_SFT())
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// get a handle from user space
	sda_handle = GET_NEW_HANDLE();
	if(sda_handle == 0xFFFF)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// set up the SFT ownership
	sda_sftptr->owner   = current_psp;
	sda_sftptr->shareid = machine_id;

	EXIT_FUNCTION();

	// everything worked
	return TRUE;
}

/*
	Assumes the above routine has been called!
*/
void AllocateNewDosHandle(void)
{
	ENTER_FUNCTION();

	sda_sftptr->handles = 1;
	SET_HANDLE_VAL(sda_handle, sda_sft_h);

	EXIT_FUNCTION();
}

/*
	Make the initial alias.
	On Entry:
		lpPath				input name
		hint					what number to start with for alias
	On Exit:
		szSrcLongName	   long name
		szProposedAlias	filled with alias, or 0 if no alias
*/
bool MakeInitialAlias(lpchar lpPath, unsigned hint)
{
	char *pLongFilename;

	ENTER_FUNCTION();

	// resolve path absolute name, in source name
	if(!ResolveLongPathnameCWD(lpPath, szSrcLongName, FALSE, FALSE))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// find just the filename
	pLongFilename = IntlStrRChr(szSrcLongName, '\\') + 1;

	//  duh..., like, ya' know, don't do anything when you
	//  don't have a name.
	//
	if( !*pLongFilename )
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// make an alias from it
	// (what happens here if there is an error?)
	//
	MakeAlias(pLongFilename, szProposedAlias, TRUE, hint);

	EXIT_FUNCTION();

	return TRUE;
}

/*
	At this point, everthing about the file is known, just
	fill in the SFT.

	Returns 0 if worked.
	Returns ERR_xxx if failed.
*/
int OpenLongFile(void)
{
	LPGENDIR lpDir;

	ENTER_FUNCTION();

	// set up various fields in SFT
	sda_sftptr->rwmode = (sopen_mode & 3);
	sda_sftptr->ioctl = (sda_drive | IO_NOTWRITTEN);

	//  ???  why is this sda_dpbptr instead of sda_devptr  ???
	sda_sftptr->device = (void far *)sda_dpbptr;

	// get current entry
	lpDir = GetCurrentShortDirEntry();

	// set up physical fields in SFT
	// eps now set by GetDirEntry
	sda_sftptr->dirsec = sda_cur_buf->sector;
   sda_sftptr->dirent = (win32SDB.dirnum % win32SDB.eps);

	// all these come from the DIR entry
	sda_sftptr->attr  = lpDir->dir.attr;
#if USING_FAT32
	sda_sftptr->ulStartingCluster = Fat32Start(lpDir->dir);
#else
	sda_sftptr->start = lpDir->dir.start;
#endif
	sda_sftptr->time  = lpDir->dir.time;
	sda_sftptr->date  = lpDir->dir.date;
	sda_sftptr->size  = lpDir->dir.size;
	FMEMCPY(sda_sftptr->name, lpDir->dir.name, ENAMESIZE);

	// bogus, can't open a directory or volume label!
	if(sda_sftptr->attr & A_NOMAKE)
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// OK, actually say we got it
	AllocateNewDosHandle();

	EXIT_FUNCTION();

	return 0;
}

/*
	Truncate a file already parsed and found in the Long SDA.

	Returns 0 if worked.
	Returns ERR_xxx if failed.
*/
int TruncateLongFile(void)
{
	LPGENDIR		lpDir;

	ENTER_FUNCTION();

	// free the fat
	FREE_THE_FAT(win32MoreFoundData.startClus);
	win32MoreFoundData.startClus = 0;

	// update the directory entry
	lpDir = GetCurrentShortDirEntry();
#if USING_FAT32
	lpDir->dir.uStartingClusterLo = 0;
	lpDir->dir.uStartingClusterHi = 0;
#else
	lpDir->dir.start = 0;
#endif
	lpDir->dir.size  = 0;
	BUF_SET_MODIFIED(sda_cur_buf);

	// don't flush if will be done later by "access date"
	// access date isn't adjusted on normal DOS filenames (ie no alias)
	if(*win32FindData.szAliasFilename == 0)
		BUF_FLUSH_ALL();

	EXIT_FUNCTION();

	return OpenLongFile();
}

/*
	Create a file already parsed in the Long SDA.  Will used
	the proposed alias, if there is an alias.

	Returns 0 if worked.
	Returns ERR_xxx if failed.
*/
int CreateLongFile(void)
{
	int	dirnum, numEntries;


	ENTER_FUNCTION();

	//  Can we make a file with these attributes?
	//
	if( ((sopen_attr & A_SUBDIR) != 0) || ((sopen_attr & (!A_MASK)) != 0) )
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// create the entry on disk
	if(!CreateGenericEntry(sopen_attr, 0, &dirnum, &numEntries))
	{
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// flush all that data to disk
	BUF_FLUSH_ALL();

	// update variables so Open() knows where it's at
	win32MoreFoundData.startDirNum = dirnum;
	win32MoreFoundData.numEntries  = numEntries;
	win32MoreFoundData.startClus   = 0;

	EXIT_FUNCTION();

	// OK, it's now on disk, open it
	return OpenLongFile();
}

/*
	Returns TRUE if the SFT matches the drive/dirSec/dirEnt.
	Also sets a pointer to the SFT that was found.

	drive		is the drive # of the SFT
	dirsec	is the sector # of the (short name) entry
	dirent	is the directory entry within that sector

	Returns FALSE if they don't match, or at the end of the list.
	Returns *plpSFT as NULLF if at end of list.
*/
bool IsSameSFT(int h, uchar drive, ulong dirsec, uchar dirent, PLPSFT plpSFT)
{
	LPSFT lpSFT;

	ENTER_FUNCTION();

	// get pointer to SFT. also pass it back to caller.
	lpSFT = *plpSFT = H_TO_SFT(h);

	// at end of list. of course they don't match.
	if(EndOfSFTList(lpSFT))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// unused SFT, no need to check
	if(!lpSFT->handles)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// cannot be a network file or a device
	if(lpSFT->ioctl & (IO_NETWORK | IO_DEVICE))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// must be same drive
	if(IoctlDrive(lpSFT->ioctl) != drive)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// must be same directory entry
	if(lpSFT->dirsec != dirsec || lpSFT->dirent != dirent)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	EXIT_FUNCTION();

	// passes all criteria. must be the same
	return TRUE;
}

/*
	Rename any SFTs that refer to this file.
*/
void RenameSFTs(uchar drive, CLUSTER oldSubdir, int oldDirnum,
		int eps, char *ename, CLUSTER newSubdir, int newDirnum)
{
	LPSFT lpSFT;
	int	h;
	ulong	oldDirsec;
	uchar	oldDirent;
	ulong	newDirsec;
	uchar	newDirent;

	ENTER_FUNCTION();

	// convert subdir/dirnum combos into dirsec/dirent
	DirnumToDirsec(oldSubdir, oldDirnum, eps, &oldDirsec, &oldDirent);
	DirnumToDirsec(newSubdir, newDirnum, eps, &newDirsec, &newDirent);

	// go through all SFTs looking for files that match the old
	// location. change them to new name and location.
	h = 0;
	do
	{
		// are they the same? change the entry!
		if(IsSameSFT(h, drive, oldDirsec, oldDirent, &lpSFT))
		{
			ASSERTLP(lpSFT);

			// fill in new entry position and name
			lpSFT->dirsec = newDirsec;
			lpSFT->dirent = newDirent;
			FMEMCPY(lpSFT->name, ename, ENAMESIZE);
		}

		// try next one until end of list
		++h;
	} while(!EndOfSFTList(lpSFT));

	EXIT_FUNCTION();
}

/*
	Rename Files or Directories.

	On Entry:
		lpSrcPath	What to rename from
		lpDstPath	What to rename to
		fFlags		REN_SHORTONLY = shortname (DOS2) rules
						REN_WILDOK = wildcards allowed, but not dirs
	Rules:
	1) Source and Destination MUST be on the same disk
	2) Source and Dest MUST be the same type (either file or dir)
	3) Files can move, dirs can only move with longname function
	4) Wildcard rename MUST be in the same dir

	Returns 0 if everything worked. ERR_xxx if failed.
*/
int RenameLongFilename(lpchar lpSrcPath, lpchar lpDstPath, int fFlags)
{
	uchar 	srcDrive;
	bool		entryExists, pathOK;
	struct 	WIN32_MORE_FIND_DATA srcData;
	struct 	WIN_DIR					winDir;
	LPGENDIR	lpDir;
	int		numEntries;
	int		dirnum;
	char *	pShortName;
	uchar		checksum;
	bool		shortOnly = (fFlags & REN_SHORTONLY);
	bool		wildOK    = (fFlags & REN_WILDOK);
	bool		moved = FALSE;
	int		numFiles  = 0;
	ulong		sector;
	CLUSTER	parent;
	char * pTempSource;
	char * pTempDestination;
	bool		fSourceIsADir;
	char * pFilename;


	ENTER_FUNCTION();

	// destination MUST be on same drive
	srcDrive = sda_drive;
	if(!GetDrive(&lpDstPath) || sda_drive != srcDrive)
	{
		EXIT_FUNCTION();
		return ERR_NOT_SAME_DV;
	}

	// the path, of course, must be valid (wildcards OK for DOS1 call)
	if(!ResolveLongPathnameCWD(lpSrcPath, szSrcLongName, wildOK, FALSE))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// make sure source file/directory exists
	if(!DoesEntryExist(&pathOK))
	{
		EXIT_FUNCTION();
		return ERR_NOTFOUND;
	}

	IntlStrCpy( szDstLongName, szSrcLongName );

	//  BT 1-26-2000 ETR 3265
	//
	//  Devices cannot be the source or destination of a rename call
	//
	pFilename = IntlStrRChr(szSrcLongName, '\\') + 1;
	if(!*pFilename)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  Determine if this is a device
	//
	if(IsStandardDOSName( pFilename ))
	{
		DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
		if( FIND_DEVICE(sda_ename1) )
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
	}

	// may loop if renaming wildcards
MoreFiles:

	// remember original SDB so we can find again
	if(wildOK)
		MEMMOVE(&win32OrgSDB, &win32SDB, sizeof(win32SDB));

	// remember WHERE source is
	MEMMOVE(&srcData, &win32MoreFoundData, sizeof(srcData));

	// get source original DIR, and everything that goes with it!
	lpDir = GetCurrentShortDirEntry();
	fSourceIsADir = ((lpDir->dir.attr & A_SUBDIR) == A_SUBDIR);

	FMEMCPY(&winDir, lpDir, sizeof(winDir));

	// matching wildcards, that means short-names only
	if(wildOK)
	{
		// directories not allowed for DOS1 (wildcard) rename
		if(winDir.attr & A_SUBDIR)
			goto Continue;

		// convert the name we found into the new name
		NAME_CHANGE(winDir.name, (char *)(&sda_fcb)+0x11);

		// that new name better not exist!
		pShortName = IntlStrRChr(szSrcLongName, '\\') + 1;
		ENAME_TO_DOT(pShortName, winDir.name);
		*szProposedAlias = 0;
	}
	// not wildcards, use normal name methods
	else
	{
		// place an absolute path for destination into "szSrcLongName".
		// also make an alias in "szProposedAlias".
		if(!MakeInitialAlias(lpDstPath, 1))
		{
			//  BT 1-28-2000 ETR 3266
			//
			//  INT 21h fn 56/7156 Rename destinations do not need
			//  to contain the name of the file so append the name
			//  now.
			//
			pFilename = IntlStrRChr(szDstLongName, '\\') + 1;
			STRCAT( szSrcLongName, pFilename );
			IntlStrCpy( szTempName, szSrcLongName );

			if(!MakeInitialAlias(szTempName, 1))
			{
				EXIT_FUNCTION();
				return ERR_NOTPATH;
			}
		}

		// for short names, make them upper case, no alias
		if(shortOnly)
		{
			*szProposedAlias = 0;
			STRUPR(szSrcLongName);
		}
	}

	//  BT 1-26-2000 ETR 3257
	//
	//  Make sure that the destination is not inside the source
	//  as ren source is .\ and destination is .\newname
	//
	//  Note: the pointers below will help the confusion over the
	//        reuse of szSrcLongName by both the source and
	//        destination.
	//
	if( fSourceIsADir )
	{
		pTempSource = szDstLongName;
		pTempDestination = szSrcLongName;
		if( pTempSource[0] != 0 )
		{
			if( pTempSource[1] == ':' )
			{
				pTempSource += 2;
			}
		}

		if( pTempDestination[0] != 0 )
		{
			if( pTempDestination[1] == ':' )
			{
				pTempDestination += 2;
			}
		}

		while( TOUPPER(*pTempSource) == TOUPPER(*pTempDestination) )
		{
			pTempSource++;
			pTempDestination++;
		}

		if((*pTempSource == 0) && (*pTempDestination == '\\'))
		{
			EXIT_FUNCTION();

			//  The destination contains the source
			//
			return ERR_NOTPATH;
		}
	}

	//  BT 1-26-2000 ETR 3265
	//
	//  Devices cannot be the source or destination of a rename call
	//
	pFilename = IntlStrRChr(szSrcLongName, '\\') + 1;
	if(!*pFilename)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  Determine if this is a device
	//
	if(IsStandardDOSName( pFilename ))
	{
		DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
		if( FIND_DEVICE(sda_ename1) )
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}
	}

	// destination must NOT exist
	entryExists = DoesEntryExist(&pathOK);

   // If the LongFileNames don't match exactly, the entry does not exist.
   // (This will allow renames from upper to lower case, for example)
   //
	if(entryExists)
   {
      pTempDestination = IntlStrRChr(szSrcLongName, '\\') + 1;

      // We already think the entry exists, reset if it doesn't match
      //
      if( strcmp( win32FindData.szFilename, pTempDestination ) )
      {
         entryExists = FALSE;
      }
   }

	if(entryExists)
	{
		if(numFiles)
			BUF_FLUSH_ALL();
		EXIT_FUNCTION();
		return ERR_DENIED;
	}

	// destination path must be OK, however
	if(!pathOK)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// determine number of entries required for destination
	numEntries = FillInFindDataName(szSrcLongName);

	// destination MAY be in a different directory.
	// if so, remove original name
	if(srcData.subdir != win32MoreFoundData.subdir)
	{
		// used if moving directories
		parent = win32MoreFoundData.subdir;

		// don't allow a directory to move on shortname functions.
		// no moving files or directories with wildcards.
		if((shortOnly && (winDir.attr & A_SUBDIR)) || wildOK)
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// first make SURE there is space for new entries
		win32SDB.drive  = sda_drive;
	   SetSubdir(&win32SDB, win32MoreFoundData.subdir);
		win32SDB.eps    = 0;
		if(!DeliverFreeDirEntries(&win32SDB, numEntries, &dirnum))
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// remove from original directory
		if(!DeleteEntries(srcData.numEntries, srcData.subdir,
				srcData.startDirNum))
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// set up for the destination directory again
		win32SDB.drive  = sda_drive;
	   SetSubdir(&win32SDB, win32MoreFoundData.subdir);
		win32SDB.eps    = 0;

		// indicate that we moved
		moved = TRUE;
	}

	// src and dst directories are the same
	else
	{
		// can we rename in place? (is new name same size or smaller?)
		if(numEntries <= srcData.numEntries)
		{
			// erase the first few entries, if not needed
			if(!DeleteEntries(srcData.numEntries, srcData.subdir,
																srcData.startDirNum))
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			// now go put new name in place
			dirnum = srcData.startDirNum;
		}

		// must get new entries
		else
		{
			// find space for new entries
			win32SDB.drive  = sda_drive;
		   SetSubdir(&win32SDB, win32MoreFoundData.subdir);
			win32SDB.eps    = 0;
			if(!DeliverFreeDirEntries(&win32SDB, numEntries, &dirnum))
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			// delete the old entry(s)
			if(!DeleteEntries(srcData.numEntries, srcData.subdir,
					srcData.startDirNum))
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}
		}
	}

	// name already converted for wildcard renames
	if(!wildOK)
	{
		// now, determine new alias name
		if(*win32FindData.szAliasFilename)
			pShortName = win32FindData.szAliasFilename;
		else
			pShortName = win32FindData.szFilename;
		DOT_TO_ENAME(winDir.name, pShortName, TRUE);
	}

	// one more file renamed
	++numFiles;

	// checksum the ename
	checksum = DetermineChecksum(winDir.name);

	// write the lfn entries to disk
	win32SDB.dirnum = dirnum;
	if(*win32FindData.szAliasFilename)
	{
		WriteLongNameToDisk(&win32SDB, win32FindData.szFilename,
			numEntries-1, checksum);
	}

	// write the DOS 8.3 entry to disk
	if(!GetDirEntry(&win32SDB, &lpDir))
		ABORT();
	FMEMCPY(lpDir, &winDir, sizeof(winDir));
	BUF_SET_MODIFIED(sda_cur_buf);

	// if a directory moved, update the parent DIR entry
	if(moved && (winDir.attr & A_SUBDIR))
	{
		// get the cluster into memory
#if USING_FAT32
		sector = CLUS_TO_SEC(Fat32Start(winDir));
#else
		sector = CLUS_TO_SEC(winDir.start);
#endif
		BUF_GETSECTOR(sector);
		lpDir = (LPGENDIR)(buf_dataptr(sda_cur_buf));

		// modify it with the new parent
		++lpDir;					// get to the ".." entry
#if USING_FAT32
		lpDir->winDir.uStartingClusterHi = HIWORD(parent);
		lpDir->winDir.uStartingClusterLo = LOWORD(parent);
#else
		lpDir->winDir.start = parent;
#endif
		BUF_SET_MODIFIED(sda_cur_buf);
	}

	// search through SFTs for this original file
	// rename (may have moved directories) any that are found
	RenameSFTs(sda_drive,
		srcData.subdir,
		srcData.startDirNum+srcData.numEntries-1,
		win32SDB.eps,
		winDir.name,
		win32SDB.subdir,
		win32SDB.dirnum);		// subdir/dirnum has the new 8.3 name

	// search again when renaming wildcards...
	if(wildOK)
	{
Continue:
		// copy back in the original Search Data Block
		MEMMOVE(&win32SDB, &win32OrgSDB, sizeof(win32SDB));

		// look for another entry
		if(FindNextLongFilenameHOK(&win32SDB, DOSDATE, &win32FindData) == 0)
		{
			// set up the More Found fields
			win32MoreFoundData.subdir      = win32SDB.subdir;
			win32MoreFoundData.startClus   = win32SDB.newSubDir;
			win32MoreFoundData.startDirNum = win32SDB.startDirNum;
			win32MoreFoundData.numEntries  = win32SDB.numEntries;
			goto MoreFiles;
		}
	}

	// flush all that data to disk
	BUF_FLUSH_ALL();

	// no files found, only directories
	if(wildOK && numFiles == 0)
	{
		EXIT_FUNCTION();
		return ERR_NOTFOUND;
	}

	EXIT_FUNCTION();

	return 0;
}

/************************************

		Main Entry Point

*************************************/

/*
	File Open/Create.  So far, the only thing processed
	has been the drive from the long filename (it will
	be \FILE, not C:\FILE)

	DS:SI		Long Filename
	BX			mode
	CX			attributes
	DX			action
	DI			Alias hint

	Returns 0 if everything worked, (and sets AX to handle).
	Returns ERR_xxx if failed.
*/
int Int21f716CH(lpchar lpPath)
{
	bool		fileExists, pathOK;
	int		actionTaken;
	int		err;
	unsigned	hint;
	char * pFilename;


	ENTER_FUNCTION();

	// validate parameters
//	ASSERTLP(lpLongFilename);

	// store mode, attr, action in SDA
	sopen_mode = R_BX;
	sopen_attr = (R_CX & A_MASK) | A_ARCHIVE;
	sopen_act  = R_DX;

	// can't create a volume label or directory!
	if(sopen_attr & A_NOMAKE)
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}

	// don't allow invalid actions
	if((sopen_act & ACT_BADFUNC) != 0 ||
		(sopen_act & ACT_INCOMPAT) == ACT_INCOMPAT)
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}

	// RD primatives like H_TO_SFT() use callflags to distinquish
	// between DOS2 and DOS1 lists of SFTs
	sda_callflags = 0;

	// determine if any SFTs and handle slots in user PSP are free
	if(!DetermineNewDosHandle())
	{
		EXIT_FUNCTION();
		return ERR_NOHANDLE;
	}

	// what is the hint going to start out as?
	hint = 1;
	if(sopen_mode & O_ALIAS_HINT)
		hint = R_DI;

	// must make proposed alias here (may have to be changed if duplicates)
	if(!MakeInitialAlias(lpPath, hint))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// first, find the file
	fileExists = DoesEntryExist(&pathOK);

	// the total path wasn't found (subdir name invalid!)
	if(!fileExists && !pathOK)
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	//  BT 1-27-2000 ETR 3267
	//
	//  We need to be able to open/create devices
	//
	if(!fileExists)
	{
		// determine filename
		pFilename = IntlStrRChr(szSrcLongName, '\\');
		if( *pFilename )
		{
			pFilename++;

			//  Determine if this is a device
			//
			if(IsStandardDOSName( pFilename ))
			{
				DOT_TO_ENAME( sda_ename1, pFilename, FALSE );
				if( FIND_DEVICE(sda_ename1) )
				{
					sda_sftptr->device = (void far *)sda_devptr;
					sda_sftptr->rwmode = (sopen_mode & ~O_NOCHILD);
					if(sopen_mode & O_NOCHILD)
					{
						sda_sftptr->ioctl |= IO_NOCHILD;
					}
					DEV_TO_SFT();
					DEV_OPENCLOSE(TRUE);
					AllocateNewDosHandle();
					R_CX = ACTION_OPENED;
					R_AX = sda_handle;

					EXIT_FUNCTION();
					return 0;
				}
			}
		}
	}

	// file DOES exist, see what user picked
	if(fileExists)
	{
		// don't open volume labels or directories
		if(win32FindData.attr & A_NOMAKE)
		{
			EXIT_FUNCTION();
			return ERR_DENIED;
		}

		// user said Open if exists
		if(sopen_act & ACT_EXIST_OPEN)
		{
			err = OpenLongFile();

			// an error happend during open, give up!
			if(err)
			{
				EXIT_FUNCTION();
				return err;
			}

			actionTaken = ACTION_OPENED;
		}

		// user said Truncate if exists
		else if(sopen_act & ACT_EXIST_REPLACE)
		{
			// don't truncate read-only files
			if(win32FindData.attr & A_READONLY)
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			err = TruncateLongFile();

			// an error happend during truncate, give up!
			if(err)
			{
				EXIT_FUNCTION();
				return err;
			}

			actionTaken = ACTION_TRUNCATED;
		}

		// user said Fail if exists
		else
		{
			EXIT_FUNCTION();
			return ERR_NOTNEW;
		}

#if UPDATE_LFN_LAST_ACCESSED_DATE
		// set last accessed date, only if it's a long filename AND
		// user says OK
		if(*win32FindData.szAliasFilename &&
			!(sopen_mode & OPEN_ACCESS_NOMODLASTACCESS))
		{
			UpdateLastAccessDate();
			BUF_FLUSH_ALL();
		}
#endif  // #if UPDATE_LFN_LAST_ACCESSED_DATE
	}

	// file does NOT exist, see what user has picked
	else
	{
		// user said Create if does not exist
		if(sopen_act & ACT_NEXIST_CREATE)
		{
			// don't allow invalid attributes
			if(sopen_attr & A_NOMAKE)
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			// create long filename (with hint for alias)
			err = CreateLongFile();

			// an error happend during create, give up!
			if(err)
			{
				EXIT_FUNCTION();
				return err;
			}

			actionTaken = ACTION_CREATED;
		}

		// user said Fail if does not exist
		else
		{
			EXIT_FUNCTION();
			return ERR_NOTFOUND;
		}
	}

	// tell user about the newly opened file!
	R_CX = actionTaken;
	R_AX = sda_handle;

	EXIT_FUNCTION();

	return 0;
}

/*
	Get Short/Long Pathname, Int 21H 7160H

	Inputs:
	DS:SI		lpSrcPath
	CL			subfunction (0-ala 60, 1 short name, 2 long name)
	CH			80=use SUBST drive, 0=use drive SUBST points to

	Outputs:
	ES:DI		Normalized Path

	Returns 0 if all worked
	Returns ERR_xxx if failed
*/
int Int21f7160H(lpchar lpSrcPath)
{
	lpchar	lpDstPath;
	char *	pDst;

	ENTER_FUNCTION();

	// check for illegal options
	if(R_CL > 2 || !(R_CH==0 || R_CH==0x80))
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}

	// handle SUBST drives (does user want subst drive, or regular drive

	// set up pointer to dst string
	lpDstPath = MK_FP(R_ES,R_DI);

	// convert relative path to absolute path, wildcards OK
	if(R_CL == 0)
	{
		// set up the "D:"
		GetRoot(szDstLongName);
		pDst = szDstLongName + STRLEN(szDstLongName);

		// resolve the path into an absolute path, wildcards OK
		if(!ResolveLongPathnameCWD(lpSrcPath, pDst, TRUE,
				((R_CH & 0x80) == 0) ))
		{
			EXIT_FUNCTION();
			return ERR_NOTPATH;
		}


		//  Update the drive letter in the path if we are
		//  showing the TRUENAME for the subst
		//
		if( ((R_CH & 0x80) == 0) && (szDstLongName[1] == ':' ) )
		{
			szDstLongName[0] = sda_drive + 'A';
		}


		// copy it into user space
		IntlFStrCpy(lpDstPath, szDstLongName);
	}
	else
	{
		// must be validate short/long name

		if(!ResolveLongPathnameCWD(lpSrcPath, szSrcLongName, FALSE,
				((R_CH & 0x80) == 0) ))
		{
			EXIT_FUNCTION();
			return ERR_NOTPATH;
		}


		//  Update the drive letter in the path if we are
		//  showing the TRUENAME for the subst
		//
		if( ((R_CH & 0x80) == 0) && (szSrcLongName[1] == ':' ) )
		{
			szSrcLongName[0] = sda_drive + 'A';
		}


		// validate the absolute path we just got!
		if(!ValidateLongPath(szSrcLongName))
		{
			EXIT_FUNCTION();
			return ERR_NOTPATH;
		}

		// copy shortname
		if(R_CL == 1)
			IntlFStrCpy(lpDstPath, szDstShortName);

		// copy longname
		else
			IntlFStrCpy(lpDstPath, szDstLongName);
	}

	EXIT_FUNCTION();

	// everything worked
	return 0;
}

/*
	Count the # of links that are associated with this SFT
	entry (defined by drive/dirsec/dirent).
*/
int CountLinks(uchar drive, ulong dirsec, uchar dirent)
{
	LPSFT lpSFT;
	int h;
	int numLinks = 0;

	ENTER_FUNCTION();

	h = 0;
	do
	{
		// if this matches, add in # of links
		if(IsSameSFT(h, drive, dirsec, dirent, &lpSFT))
			numLinks += lpSFT->handles;

		// try next one until end of list
		++h;
	} while(!EndOfSFTList(lpSFT));

	EXIT_FUNCTION();

	return numLinks;
}

/*
	Get a directory entry based on the given handle.  Also sets
	up the SDA for a handle.

	Returns TRUE if this is a valid handle.
	Returns FALSE not.
*/
bool GetDirByHandle(int handle, PLPGENDIR plpDir)
{
	ENTER_FUNCTION();

	// is this a valid handle?
	if(!HANDLE_TO_SDA(handle))
   {
      EXIT_FUNCTION();
		return FALSE;
   }

	// get the actual DIR sector into BUFFERS
	*plpDir = GetRawDirEntry(sda_sftptr->dirsec, sda_sftptr->dirent);

	EXIT_FUNCTION();

	// got ptr to the DIR entry
	return TRUE;
}

/*
	Ask current disk device for volume information. Assumes that
	the variables have been parsed into the SDA, and that
	sda_dpbptr and sda_devptr are up to date and accurate.

	Returns TRUE if disk device supports the call, FALSE if not.
*/
bool GetVolumeInfo(struct MEDIA_ID * pMediaId)
{
	unsigned * up;

	ENTER_FUNCTION();

	// validate parameter
	ASSERT(pMediaId);

	// reset request header
	MEMSET(&sda_reqhdr, 0, sizeof(sda_reqhdr));

	// call to do generic IOCTL
	sda_reqhdr.command = DEV_GENIO;
	sda_reqhdr.unit = sda_dpbptr->unit;

	/* the DS:DX buffer is pointed to in a different place */
	up = (unsigned *)(&sda_reqhdr.media);
	*up = 0x6608;						/* 0DH  function 866h */
	*(up+3) = FP_OFF(pMediaId);	/* 13H */
	*(up+4) = FP_SEG(pMediaId);	/* 15H DS:DX points to buffer */

	/* call the device */
	DEV_CALL(&sda_reqhdr, sda_devptr);

	EXIT_FUNCTION();

	// return TRUE if worked, FALSE if failed (!error)
	return !(sda_reqhdr.status & DEVE_ERROR);
}

/*
	Get File Info By Handle.

	Note Yet Implemented:
		Creation/LastAccess/LastWrite date/times
		VolumeSerialNumber

	Returns TRUE if worked, and fills in the lpInfo.
	Returns FALSE if failed.
*/
bool Int21f71A6H(unsigned handle, LP_BY_HANDLE_FILE_INFORMATION lpInfo)
{
	struct BY_HANDLE_FILE_INFORMATION info;
	struct MEDIA_ID mediaId;
	LPGENDIR lpDir;

	ENTER_FUNCTION();

	// get the directory entry based on handle
	if(!GetDirByHandle(handle, &lpDir))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// start out with all zeros
	MEMSET(&info, 0, sizeof(info));

	// file size and attr
	info.attr     = sda_sftptr->attr;
   info.fileSize = sda_sftptr->size;

	// build unique identifyer. basically, drive/dirsec/dirent uniquely
	// identifies ANY file.  multiple SFTs may be open on the same file.
	info.fileIndex = sda_sftptr->dirsec;
	info.fileIndexHigh = (sda_sftptr->dirent << 8) | IoctlDrive(sda_sftptr->ioctl);

	// count # of links
	info.numLinks = CountLinks(IoctlDrive(sda_sftptr->ioctl),
										sda_sftptr->dirsec, sda_sftptr->dirent);

	// fill in dates/times
	PutFindTime(&info.creationTime,
		lpDir->winDir.creationTime, lpDir->winDir.creationDate);
	PutFindTime(&info.lastAccessDate, 0, lpDir->winDir.lastAccessDate);
	PutFindTime(&info.lastWriteTime,
		lpDir->winDir.lastModTime, lpDir->winDir.lastModDate);

	// volume serial #
	if(GetVolumeInfo(&mediaId))
		info.volumeSerialNumber = mediaId.serialNum;

	// copy info to user space
	FMEMCPY(lpInfo, &info, sizeof(info));

	EXIT_FUNCTION();

	// it worked. found the file handle.
	return TRUE;
}

//  ETR ? BT 9-20-00
//
//  Removed support for int 21h fn 71a8h (Generate short name) because
//  windows appeared to have bugs when returning an 8.3 zero terminated
//  string.  Support for this function may be added at a later time when
//  a customer requests it.  Windows was not zero terminating the short
//  name in 8.3 form, instead it was badding the entire 13 bytes with
//  0x5f (an underscore).
//
#if 0
/*
	Make Alias, Int 21H 71A8H
	DS:SI		Long Filename
	ES:DI		Alias (output)
	DH			Shortname format (1=8.3 or 0=ename)
	DL			Char set (0=WANSI, 1=DOS, 2=UNICODE) - we only support DOS
*/
int Int21f71A8H(lpchar lpPath)
{
	char 		ename[ENAMESIZE+1];		// need space for NULL
	char		szAlias[DOTNAMESIZE];
	char *	pName;

   ENTER_FUNCTION();

	// shortname format must be 0 or 1
	// we only support DOS_CHARS character format
	if(R_DH > 1 || R_DL != DOS_CHARS)
   {
      EXIT_FUNCTION();
		return ERR_BADFUNC;
   }

	// make the alias name, no "~1" numeric ending
	MakeAlias(lpPath, szAlias, FALSE, 0);

	// user wants name expanded. do that.
	if(R_DH == 0)
	{
		DOT_TO_ENAME(ename, szAlias, TRUE);
		ename[sizeof(ename)-1] = 0;	// null terminate ename
		pName = ename;
	}
	else
		pName = szAlias;

	// copy name to user space
	IntlFStrCpy(MK_FP(R_ES,R_DI), pName);

	// everything worked
   EXIT_FUNCTION();
	return 0;
}
#endif


/*
	Get/Set Creation/LastAccessed Date/Time.
	Say what?  Handle Int 21H, 57xxH where xx is 04-07.
	Does no checking on the set time, just plops it in.

	Returns TRUE if worked.
	Returns FALSE if failed (bad handle).
*/
bool Int21f57xxh(uchar subfunc, int handle)
{
	LPGENDIR lpDir;
	unsigned date, time;

	ENTER_FUNCTION();

	// validate subfunc range
	ASSERT(subfunc >= 4 && subfunc <= 7);

	// invalid handle
	if(!GetDirByHandle(handle, &lpDir))
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// return the appropriate date/time
	switch(subfunc)
	{
		// get last access date
		case 4:
			date = lpDir->winDir.lastAccessDate;
			time = 0;
			break;

		// set last access date
		case 5:
			lpDir->winDir.lastAccessDate = R_DX;
			break;

		// get creation date/time
		case 6:
			date = lpDir->winDir.creationDate;
			time = lpDir->winDir.creationTime;
			R_SI = lpDir->winDir.milli;
			break;

		// set creation date/time
		case 7:
			lpDir->winDir.creationDate = R_DX;
			lpDir->winDir.creationTime = R_CX;
			lpDir->winDir.milli        = (uchar)R_SI;
			break;
	}

	// the odd ones are write functions, so write out the
	// dirty sector
	if(subfunc & 1)
	{
		BUF_SET_MODIFIED(sda_cur_buf);
		BUF_FLUSH_ALL();
	}

	// reading the date or time
	else
	{
		R_CX = time;
		R_DX = date;
	}

	EXIT_FUNCTION();

	return TRUE;
}

/*
	Get or Set attributes or dates for long filenames.
	Basically, int 21H 7143H.
*/
int GetSetLongnameAttributes(lpchar lpPath, uchar subfunc,
										unsigned attr, unsigned milli, unsigned date)
{
	bool			pathOK;
	LPGENDIR		lpDir;
	ulong			size;
	unsigned		clusterSize;
	unsigned		time;

	ENTER_FUNCTION();

	// make sure function is in range
	if(subfunc > 8)
	{
		EXIT_FUNCTION();
		return ERR_BADFUNC;
	}


	// path must be reasonable
	if(!ResolveLongPathnameCWD(lpPath, szSrcLongName, FALSE, FALSE))
	{
		EXIT_FUNCTION();
		return ERR_NOTPATH;
	}

	// file must exist
	if(!DoesEntryExist(&pathOK))
	{
		EXIT_FUNCTION();
		return ERR_NOTFOUND;
	}

	// get pointer to current entry
	lpDir = GetCurrentShortDirEntry();

	// time and attr are two names for the same register (CX)
	time = attr;

	// actually do the work
	switch(subfunc)
	{
		// get attributes
		case 0:
			R_AX = R_CX = lpDir->winDir.attr;
			break;

		// set attributes
		case 1:

			// can't change a file to a directory or volume label!
			if(attr & A_NOMAKE)
			{
				EXIT_FUNCTION();
				return ERR_DENIED;
			}

			// set attribute
			lpDir->winDir.attr = (lpDir->winDir.attr & A_NOMAKE) | (attr & 0x3F);
			break;

		// get physical size
		case 2:

			// note: won't handle clustersSize above 32K
			clusterSize = sda_dpbptr->bytes_per_sec << sda_dpbptr->shift;
			size = _LDIVU(lpDir->winDir.fileSize + clusterSize-1, clusterSize);
			size = _LMULU(size, clusterSize);

			// return size in DX:AX
			R_DX = FP_SEG(size);
			R_AX = FP_OFF(size);
			break;

		// set last write date/time
		case 3:
			lpDir->winDir.lastModDate = date;
			lpDir->winDir.lastModTime = time;
			break;

		// get last write date/time
		case 4:
			date = lpDir->winDir.lastModDate;
			time = lpDir->winDir.lastModTime;
			milli = 0;
			break;

		// set last access date
		case 5:
			lpDir->winDir.lastAccessDate = date;
			break;

		// get last access date
		case 6:
			date  = lpDir->winDir.lastAccessDate;
			time  = 0;
			milli = 0;
			break;

		// set creation date/time
		case 7:
			lpDir->winDir.creationDate = date;
			lpDir->winDir.creationTime = time;
			lpDir->winDir.milli        = milli;
			break;

		// get creation date/time
		case 8:
			date = lpDir->winDir.creationDate;
			time = lpDir->winDir.creationTime;
			milli = lpDir->winDir.milli;
			break;
	}

	// all the get date functions
	if(subfunc >= 4 && !(subfunc & 1))
	{
		R_CX = time;
		R_DI = date;
		R_SI = milli;
	}

	// flush changes to date/attr to disk
	else if(subfunc & 1)
	{
		BUF_SET_MODIFIED(sda_cur_buf);
		BUF_FLUSH_ALL();
	}

	// (most) all function calls modify AX
	if(subfunc && subfunc!=2)
		R_AX = (0x4300 | subfunc);

	EXIT_FUNCTION();

	// no errors
	return 0;
}

/*
	Is the item in the list?
*/
bool InList(uchar *pList, uchar item)
{
	ENTER_FUNCTION();

	while(*pList)
	{
		if(*pList == item)
		{
			EXIT_FUNCTION();
			return TRUE;
		}
		++pList;
	}

	EXIT_FUNCTION();

	return FALSE;
}

/*
	Run Only On Bootup:

	Used to create the "SEARCH" structures. The number of searches
	defaults to 3, but is set in CONFIG.SYS. The LFN search functions
	return a handle to one of the SEARCH structures when opening a
	findfirst.
*/
void SearchesCreate(void)
{
	unsigned seg;

	ENTER_FUNCTION();

	// set up number of handles, if not already
	if(!lfnMaxSearches)
		lfnMaxSearches = DEF_SEARCHES;

	// get memory for searches
	seg = MEM_Z_ALLOC(lfnMaxSearches * sizeof(struct WIN32_SDB), 'E');
	if(!seg)
		lfnMaxSearches = 0;
	else
		lplfnSearchArray = MK_FP(seg, 0);

	// know how many handles we've got
	lfnHandlesLeft = lfnMaxSearches;

	EXIT_FUNCTION();
}

/*
	Convert from FCB style name to normal name
*/
void MakeDos1Name(char *pDstName, uchar drive, char *pSrcFcbName)
{
	ENTER_FUNCTION();

	/* may have a drive letter */
	if(drive)
	{
		*pDstName++ = drive+'A'-1;		// 1=A, 2=B, etc..
		*pDstName++ = ':';
	}
	ENAME_TO_DOT(pDstName, pSrcFcbName);

	EXIT_FUNCTION();
}

/*
	Handles all the Int 21H, function 71xxH calls.  As input,
	uses the SDA variable user_regs  (uses REGS macros).
	As output, uses the same, and may call disk subfunctions.
	Sets "error_no" on errors.

	Returns TRUE if the call was for long filenames.
	Returns FALSE if not, and normal Kernel should handle call.
*/
bool LongnameFunctions(void)
{
	unsigned	lfnHandle;
	lpchar	lpPath;
	static	uchar dsdxNames[] = { 0x39, 0x3A, 0x3B, 0x41, 0x43, 0x4E, 0x56, 0xA0, 0 };
	static	uchar dssiNames[] = { 0x60, 0x6C, 0xA8, 0 };
	bool		inDriveList;
	lpchar	lpPath2;			// for rename
	uchar		rAH;
	uchar		bFnCall;			// function call #
	bool		isDOS1Call = FALSE;
	int		fFlags;
	unsigned uTempSearchAttributes;
	char far * lpFilename;
	char far * lpColon;


	ENTER_FUNCTION();

	// get DOS call flags
	GET_CALL_FLAGS();

	// assume it's not a shortname function
	fAliasHelper = fShortNameFunctions = FALSE;
	sda_srchattr = A_DIRCHK;		// default to looking for all files/dirs

	// handle date/time functions, 57xxh
	if(R_AH==0x57 && R_AL>=4 && R_AL<=7)
	{
		// if this fails, then BX must not be a valid handle
		if(!Int21f57xxh(R_AL, R_BX))
			error_no = ERR_BADHANDLE;

		EXIT_FUNCTION();

		// the call was for us
		return TRUE;
	}

	// record the function call
	rAH     = R_AH;
	bFnCall = R_AL;

	// convert DOS1 delete into DOS7 style delete
	// convert DOS1 rename into DOS7 style rename
	if(rAH == 0x13 || rAH == 0x17)
	{
		// this is a DOS1 call
		isDOS1Call = fShortNameFunctions = TRUE;

		// determine DOS2 equivalent function # for delete or rename
		// delete (13h) is similar to 41h, rename (17h) is similar to 56h
		bFnCall = (rAH == 0x13 ? 0x41 : 0x56);

		// move FCB to the SDA, and convert names
		FCB_TO_SDA();

		// don't handle volume labels, let normal DOS call do it.
		// a volume label can never be a long file name.
		if(sda_srchattr == A_VOLUME)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		MakeDos1Name(szDos1Name, sda_fcb.drive, (char *)sda_fcb.name);
		if(bFnCall == 0x56)
		{
			STRUPR((char *)(&sda_fcb)+0x11);
			MakeDos1Name(szDos1Name2, sda_fcb.drive, (char *)(&sda_fcb)+0x11);
		}
	}

	// convert DOS2 remove dir (3A) into DOS7 style remove dir
	// convert DOS2 delete (41) into DOS7 style delete
	// convert DOS2 rename (56) into DOS7 style rename
	// handle exception case for change dir (3b) allowing trailing backslash
	else if(rAH == 0x3A || rAH == 0x41 || rAH == 0x56 || rAH == 0x3b )
	{
		// don't handle the wildcard 5D rename call
		if(rAH == 0x56 && sda_in_5D)
		{
			EXIT_FUNCTION();
			return FALSE;
		}

		bFnCall = rAH;
		fShortNameFunctions = TRUE;
	}

	// aside from exceptions above, only handle 71H functions
	else if(rAH != 0x71)
	{
		EXIT_FUNCTION();
		return FALSE;
	}

	// always start with no error, and some basic SDA stuff settled
	error_no = 0;
	int21_ax = R_AX;

	// no proposed Alias, not available across DOS calls
	*szProposedAlias = 0;

	// reset interal SDB so it starts from scratch between DOS calls.
	// multiplex the Entries Per Sector field as a valid indicator.
	win32SDB.eps = 0;

	// some calls use a drive letter, some don't. assume don't.
	inDriveList = FALSE;

	// DOS1 calls are treated special
	if(isDOS1Call)
	{
		inDriveList = TRUE;
		lpPath = szDos1Name;
		lpPath2 = szDos1Name2;
	}

	// is this a DS:DX kind of name?
	else if(InList(dsdxNames, bFnCall))
	{
		inDriveList = TRUE;
		lpPath = MK_FP(R_DS,R_DX);
		if(bFnCall == 0x56)					// rename takes 2 names
			lpPath2 = MK_FP(R_ES, R_DI);
	}

	// is this a DS:SI kind of name?
	else if(InList(dssiNames, R_AL))
	{
		inDriveList = TRUE;
		lpPath = MK_FP(R_DS,R_SI);
	}

	// make sure drive is valid, set up drive variables
	if(inDriveList)
	{
		// make sure drive is valid.  setup drive variables
		if(!GetDrive(&lpPath))
		{
			error_no = ERR_NOTDRIVE;
			EXIT_FUNCTION();
			return TRUE;
		}

		// but don't deal with Network drives (yet)
		if( sda_cdsptr->status & C_NETWORK )
		{
			if(fShortNameFunctions)
			{
				EXIT_FUNCTION();
				return FALSE;
			}

			//  Return that this function is invalid for IFS calls
			//
			//  error_no = ERR_DENIED;
			//
			error_no = ERR_W95BADFUNC;
			EXIT_FUNCTION();
			return TRUE;
		}

		//  BT 1-27-2000 ETR 3268
		//
		//  Names cannot have a ':' unless they are a device
		//
		lpFilename = IntlFStrRChr(lpPath, '\\');
		if( *lpFilename == 0 )
		{
			lpFilename = lpPath;
		}
		else
		{
			lpFilename++;
		}

		lpColon = IntlFStrRChr(lpFilename, ':');
		if((*lpColon) && (lpColon != (lpPath+1)) )
		{
			if( *(lpColon+1) == 0 )
			{
				*lpColon = 0;

				//  Determine if this is a device
				//
				if(IsStandardDOSName( lpFilename ))
				{
					DOT_TO_ENAME( sda_ename1, lpFilename, FALSE );
					if( FIND_DEVICE(sda_ename1) )
					{
						*lpColon = ':';
						error_no = ERR_NOTFOUND;
						EXIT_FUNCTION();
						return TRUE;
					}
				}

				*lpColon = ':';
			}

			error_no = ERR_DENIED;
			EXIT_FUNCTION();
			return TRUE;
		}

#if 0
		if(bFnCall == 0x56)					// rename takes 2 names
		{
		}
#endif

	}

	switch(bFnCall)
	{
		// Create Longname Directory
		case 0x39:
			error_no = CreateLongDir(lpPath);
			break;

		// Delete Longname Directory
		case 0x3A:
			error_no = DeleteLongFileOrDirectory(lpPath, DEL_DIR);
			break;

		// Change Longname Directory
		case 0x3B:
			// actually change into that directory
			if(!ChangeLongCWD(lpPath))
				error_no = ERR_NOTPATH;
			break;

		// Delete Longname File
		case 0x41:
			fFlags = 0;

			/* this is a DOS7 LFN call */
			if(rAH == 0x71)
			{
				// DOS7 delete looks at CX for attribute matching
				fFlags |= DEL_CXATTR;

				// si determines if wildcard characters are allowed
				// 0=no wildcards, 1=wildcards OK */
				if(R_SI == 1)
					fFlags |= DEL_WILDOK;		// allowed
				else if(R_SI != 0)
				{
					error_no = ERR_DENIED;		// bad si value returns denied
					EXIT_FUNCTION();
					return TRUE;
				}
			}

			// DOS1 style deletes, and backdoor deletes allow wildcards
			if(isDOS1Call || sda_in_5D)
				fFlags |= DEL_WILDOK;

			// delete the file(s)
			error_no = DeleteLongFileOrDirectory(lpPath, fFlags);
			break;

		// Get/Set Longname Attributes
		case 0x43:
			error_no = GetSetLongnameAttributes(lpPath, R_BL, R_CX, R_SI, R_DI);
			break;

		// Get Current Longname Directory
		case 0x47:

			// is the drive OK?
			if(!DEFDRIVE_TO_SDA(R_DL))
				error_no = ERR_NOTDRIVE;

			// is the path OK?
			else if(!GetLongCWD(MK_FP(R_DS,R_SI)))
				error_no = ERR_NOTDRIVE;
			break;

		// Find First Longname File/Directory
		case 0x4E:

			if(rAH == 0x71)
			{
				uTempSearchAttributes = R_CX;
			}
			else
			{
//  BT  2-1-2000
//
//  I didn't believe that this was correct so I modified it
//  to copy the cl value into the ch value.  For now it does
//  not matter because int 21h fn 4eh does not get processed
//  by the longname module, but in case it ever does.
//
#if 0
				//  The non-lfn find first only has the attributes
				//  set in cl
				//
				uTempSearchAttributes = (R_CX << 8) & R_CX;
#else
				//  The non-lfn find first only has the attributes
				//  set in cl
				//
				uTempSearchAttributes = (R_CX & 0x3e);
				uTempSearchAttributes <<= 8;
				uTempSearchAttributes |= ((uchar)R_CX & 0x3e);
				fShortNameFunctions = TRUE;
#endif
			}

			// do the findfirst
			error_no = FindFirstLongFilename(lpPath, uTempSearchAttributes,
				R_SI,	MK_FP(R_ES, R_DI), &lfnHandle);

			// worked, got a handle, return it in ax
			if(!error_no)
				R_AX = lfnHandle;
			break;

		// Find Next Longname File/Directory
		case 0x4F:
			error_no = FindNextLongFilename(R_BX, R_SI, MK_FP(R_ES, R_DI));
			break;

		// Rename Longname File
		case 0x56:

			// set up flags for rename
			fFlags = 0;
			if(isDOS1Call)
				fFlags |= REN_WILDOK;
			if(fShortNameFunctions)
				fFlags |= REN_SHORTONLY;

			error_no = RenameLongFilename(lpPath, lpPath2, fFlags);
			break;

		// Get Longname Path
		case 0x60:

			//  To call expand with a path like "a:" is perfectly valid
			//  in an LFN environment.  However, the drive would have been
			//  stripped from the lpPath string.  So check to see if the
			//  path is now null but that it was not passed in that way.
			//  If this is the case, then we need to pass a "." or the
			//  resolve path code will error out.
			//
			if( (lpPath[0] == 0) && (lpPath != MK_FP(R_DS,R_SI)) )
			{
				error_no = Int21f7160H(".");
			}
			else
			{
				// already processed drive (from path), now do the rest.
				// uses REGS, so no need to pass most parameters.
				error_no = Int21f7160H(lpPath);
			}

			break;

		// Extended Longname Open / Create
		case 0x6C:

			// already processed drive (from path), now do the rest
			// uses REGS, so no need to pass most parameters
			error_no = Int21f716CH(lpPath);

			break;

		// Get Longname Volume Information
		case 0xA0:
		{
			static char fatName[] = "FAT";

			// enough room to copy name? do it!
			if(R_CX >= sizeof(fatName))
				IntlFStrCpy(MK_FP(R_ES,R_DI), fatName);

			// copy flags and lengths
			R_BX = FS_FLAGS;
			R_CX = MAXLNAME - 1;		// max longname length, minus null
			R_DX = MAXPATH;
			R_AX = 0x200;				// ID
			break;
		}

		// Close Longname FindFirst/Next
		case 0xA1:
			if(!FindClose(R_BX))
				error_no = ERR_BADHANDLE;
			break;

		// Get Longname File Information
		case 0xA6:
			if(!Int21f71A6H(R_BX, MK_FP(R_DS, R_DX)))
				error_no = ERR_BADHANDLE;
			break;

		// DOS time to FILETIME (64-bit WIN95 time/date)
		//
		case 0xA7:
			if( (uchar)R_BX == 1 )
			{
				error_no = DosTimeToFileTime( R_CX, R_DX,
						((R_BX & 0xff00) >> 8), (lpFILETIME)MK_FP(R_ES,R_DI) );
			}
			else
			{
				error_no = ERR_W95BADFUNC;
			}
			break;

//  ETR ? BT 9-20-00
//
//  Removed support for int 21h fn 71a8h (Generate short name) because
//  windows appeared to have bugs when returning an 8.3 zero terminated
//  string.  Support for this function may be added at a later time when
//  a customer requests it.  Windows was not zero terminating the short
//  name in 8.3 form, instead it was badding the entire 13 bytes with
//  0x5f (an underscore).
//
#if 0
		// Get Longname Alias
		case 0xA8:
			error_no = Int21f71A8H(lpPath);
			break;
#endif

		default:
			//  Unsupported function
			//
			error_no = R_AX & 0xff00;
			break;
	}

	EXIT_FUNCTION();

	// return whether the call was for us or not
	return TRUE;
}

// VDOS must be less than 7, make stub functions
#else

// no LongName support if less than DOS7 required, stub
bool LongnameFunctions(void)
{
	return FALSE;
}

// no LongName support if less than DOS7 required, stub
void SearchesCreate(void)
{
}

void FreeProcessSearches(void)
{
}

#endif // } VDOS >= 7
