//
//   $Id: BOOTLTR.C 1.4 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  $DOCMODULE:BOOTLTR.C
//  Created 02-05-1999 by Brandon Thomas
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This module is an interface to retrieve the boot drive letter.  This
//  is part of the Datalight Library.
//
//  bool PASCAL dlGetBootDriveLetter( uchar * pucDriveLetter )
//
//  $DOCHISTORY
//  02/05/1999 bt  Created.
//

//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"

//  $DOCPROC: dlGetBootDriveLetter
//
//  bool PASCAL dlGetBootDriveLetter( uchar * pucDriveLetter )
//
//  This function first calls dlGetDosVersion and determines if the the
//  dos version supports a call to get the drive number.  It then calls
//  dlGetBootDriveNumber and changes this into a drive letter. This
//  function returns FALSE if it fails and TRUE otherwise.  The parameters
//  are described below:
//
//  ON ENTRY:
//    none
//  ON EXIT:
//		pucDriveLetter - pointer to the drive letter
//
bool PASCAL dlGetBootDriveLetter( uchar * pucDriveLetter )
{
	//  [LOCALS]
	//
	bool     bSuccessful;
	uchar    ucDriveNumber;
	unsigned uDosVersion;


	//  [CODE]
	//

	//  Check the dos version number and make sure it is at least version
	//  2.0.
	//
	bSuccessful = dlGetDosVersion( &uDosVersion );
	if( bSuccessful )
	{
		if( uDosVersion >= 0x0200 )
		{
			//  Change the drive number into a drive letter
			//
			bSuccessful = dlGetBootDriveNumber( &ucDriveNumber );
			if( bSuccessful )
			{
				*pucDriveLetter = (ucDriveNumber - 1) + 'A';
			}
		}
		else
		{
			return FALSE;
		}
	}

	//  Return whether the function was successful
	//
	return bSuccessful;
}
