//
//   $Id: DLDFRE32.C 1.7 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*****************************************************************************

   DLDFRE32.C
   Created 02/27/99 by Dennis Edwards

   Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This module is part of the DLDOSIO library, it contains:

		bool PASCAL SmartGetFreeSpace( int iDrive, PGDF32 pDiskSpace );

	HISTORY:
   02/27/99   de	created module from DM's code that was in DLDFREE.C

*****************************************************************************/
#define  USING_FAT32		1

#include "dostruct.h"
#include "dosdef.h"
#include "datalght.h"
#include "dldosio.h"
#include "fatworld.h"


// We allocate this goofy string so the user doesn't have to. It takes
// up 4 bytes and is passed to int21 f7303 (FAT32 Get drive free space).
//
static char szDriveString[4] = "A:\\";


/*
	$DOCAPI:SmartGetFreeSpace

	This is a 'smart' version of get free space. This function will check first
	to see if the disk is a FAT32 disk, and if so, will call the FAT32 get disk
	free space, otherwise this function will build the FAT12/16 free space info
	into the FAT32 structure so that only one structure is used to retrieve the
	free space for a drive. This function returns TRUE upon success and FALSE
	upon failure.

	iDrive = (0=default,1=A,2=B,...)!!!!

*/
bool PASCAL SmartGetFreeSpace( int iDrive, PGDF32 pDiskSpace )
{
	static	fBeenCalledBefore = FALSE;
	static	fIsFat32World;
	bool		fResult;


	if(iDrive == 0)
	{
		dlGetCurrentDrive(&iDrive);
	}
	else
	{
		iDrive--;
	}

	if( !fBeenCalledBefore )
	{
		fBeenCalledBefore = TRUE;
		fIsFat32World = dlIsFat32World();
	}

	// Check to see if this is a FAT32 Drive
	if( fIsFat32World )
	{
		// The FAT32 Get Free Space function requires a drive string
		// instead of a number, so we must first convert this drive number
		//
		*szDriveString = iDrive + 'A';

		// This member must specifically be initialized to ZERO always
		//
		pDiskSpace->uLevel = 0;

		// Get the drive space structure filled out
		//
		fResult = dlGetDriveFreeSpace32(szDriveString, pDiskSpace);
	}
	else
	{
	unsigned	uFreeClusters;			// Used to calculate FAT16 disk values
	unsigned	uTotalClusters;		// Used to calculate FAT16 disk values
	unsigned	uBytesPerSector;		// Used to calculate FAT16 disk values
	unsigned	uSectorsPerCluster;	// Used to calculate FAT16 disk values

		// Get some info on this drive so we can calculate
		fResult = __dlGetFreeSpaceOnDrive(iDrive+1, &uSectorsPerCluster, &uFreeClusters,
						&uBytesPerSector, &uTotalClusters);

		if(fResult)
		{
			// Initialize as many elements of this structure as possible
			pDiskSpace->uLevel                     = 0;
			pDiskSpace->uSize                      = sizeof(GDF32);
			pDiskSpace->ulSectorsPerCluster        = (ulong) uSectorsPerCluster;
			pDiskSpace->ulBytesPerSector           = (ulong) uBytesPerSector;

			pDiskSpace->ulAvailableClusters        = (ulong) uFreeClusters;
			pDiskSpace->ulAvailableAllocationUnits = (ulong) uFreeClusters;
			pDiskSpace->ulTotalClusters            = (ulong) uTotalClusters;
			pDiskSpace->ulTotalAllocationUnits     = (ulong) uTotalClusters;

			pDiskSpace->ulAvailablePhysicalSectors = (ulong) uFreeClusters *
					uSectorsPerCluster;
			pDiskSpace->ulTotalPhysicalSectors     = (ulong) uTotalClusters *
					uSectorsPerCluster;
		}
	}
	return fResult;
}
