//
//   $Id: FMTIOCTL.C 1.5 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  $DOCMODULE:FMTIOCT.C
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This module is an interface to int-21 fn 440dh subfunction 42h which
//  is format and verify track.  This is part of the Datalight Library.
//
//  bool dlIOCTLFormatAndVerifyTrack( uchar ucDriveNumber,
//			unsigned uHead, unsigned uTrack, uchar ucSpecialFunction,
//			bool fIsFat32World );
//
//  $DOCHISTORY
//  03/14/1999 bt  Created.
//

//
//  [CONSTANTS]
//

//  We want to compile using the FAT32 structures
//
#define USING_FAT32		1


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"

//
//  [ CONSTANTS ]
//

#define DEFAULT_CATEGORY_CODE						0x08
#define FAT32_AWARE_CATEGORY_CODE				0x48

#define FORMAT_AND_VERIFY_TRACK					0x42

//
//  [ STRUCTURES ]
//

struct ParameterBlock{
	uchar		ucSpecialFunction;
	unsigned uHeadNumber;
	unsigned uTrackNumber;
	unsigned uUnknown;
	unsigned uUnknown2;
};


//  $DOCPROC: dlIOCTLFormatAndVerifyTrack
//
//  bool dlIOCTLFormatAndVerifyTrack( uchar ucDriveNumber,
//			unsigned uHead, unsigned uTrack, uchar ucSpecialFunction,
//			bool fIsFat32World );
//
//  This function is an interface to int-21h fn 440dh subfunction 42h which
//  is format and verify track.  The drive numbers start with A=1. This
//  function returns FALSE if it fails and TRUE otherwise.  This is part
//  of the Datalight Library.  The parameters are described below:
//
//  ON ENTRY:
//
//  ucDriveNumber = The drive number (0=Default, 1=A, 2=B,...)
//  uHead			= The head number of the track to be formatted
//  uTrack			= The track number to be formatted
//  ucSpecialFunction = MSDOS uses 1 and 2 for the special function number
//                      on a verify
//  fIsFat32World = Set to TRUE if in DOS 7.1/Windows enviornment
//
//  ON EXIT:
//
//  returns TRUE if the function call was successful, FALSE otherwise.
//
bool dlIOCTLFormatAndVerifyTrack( uchar ucDriveNumber,
		unsigned uHead, unsigned uTrack, uchar ucSpecialFunction,
		bool fIsFat32World )
{
	//  [LOCALS]
	//
	bool bSuccessful = TRUE;
	uchar ucCategoryCode;
	struct ParameterBlock stParameterBlock;


	//  [CODE]
	//

	stParameterBlock.uHeadNumber  	  = uHead;
	stParameterBlock.uTrackNumber 	  = uTrack;

	if( fIsFat32World )
	{
		ucCategoryCode = FAT32_AWARE_CATEGORY_CODE;
	}
	else
	{
		ucCategoryCode = DEFAULT_CATEGORY_CODE;
	}

	stParameterBlock.uUnknown2 = 0;

	if( ucSpecialFunction == 2 )
	{
		stParameterBlock.uUnknown = 2;
	}
	else
	{
		stParameterBlock.uUnknown = 1;
	}

	if( ucSpecialFunction == 0 )
	{
		//  Use the generic IOCTL function to format and verify a track
		//
		stParameterBlock.uUnknown = 1;
		stParameterBlock.ucSpecialFunction = ucSpecialFunction;
		bSuccessful = dlGenericIOCTLForBlockDevices( ucDriveNumber,
				ucCategoryCode, FORMAT_AND_VERIFY_TRACK,
				&stParameterBlock );
		ucSpecialFunction++;
	}

	//  Check if there was an error
	//
	if( bSuccessful )
	{
		//  Use the generic IOCTL function to verify a track
		//
		stParameterBlock.ucSpecialFunction = ucSpecialFunction;
		bSuccessful = dlGenericIOCTLForBlockDevices( ucDriveNumber,
				ucCategoryCode, FORMAT_AND_VERIFY_TRACK,
				&stParameterBlock );
	}

	//  Return whether the function was successful
	//
	return bSuccessful;
}
