//
//   $Id: GTDRVDPB.C 1.6 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  $DOCMODULE:GTDRVDPB.C
//  Created 02-23-1999 by Brandon Thomas
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This module is an interface to retrieve the DPB pointer for a specific
//  drive.  This function does not make sure that the drive number is
//  valid.  This is part of the Datalight Library.
//
//  bool PASCAL dlGetDPBForDrive( uchar ucDriveNumber, LPDPB * plpDPB );
//
//  $DOCHISTORY
//  02/23/1999 bt  Created.
//

//
//  [CONSTANTS]
//

//  We want to compile using the FAT32 DPB structures
//
#define USING_FAT32		1


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"


//  $DOCPROC: dlGetDPBForDrive
//
//  bool PASCAL dlGetDPBForDrive( uchar ucDriveNumber, LPDBP * plpDPB );
//
//  This function retrieves the DPB pointer for a specific drive number.
//  This function does not make sure that the drive number is valid.  The
//  drive numbers start with A=0.  This is part of the Datalight Library.
//  This function returns FALSE if it fails and TRUE otherwise.  The
//  parameters are described below:
//
//  ON ENTRY:
//    ucDriveNumber  - Drive number(A=0,B=1) to retrieve a DPB for
//
//  ON EXIT:
//		plpDPB - pointer to the DPB for the drive number
//
bool PASCAL dlGetDPBForDrive( uchar ucDriveNumber, LPDPB * plpDPB )
{
	//  [LOCALS]
	//
	bool     bSuccessful;
	LPLOL		lpLOL;
	LPDPB		lpDPB;

	//  [CODE]
	//

	//  The pointer to the DPB linked list is in the LOL so we
	//  need to retrieve the LOL pointer.
	//
	bSuccessful = dlGetLOLPointer( &lpLOL );
	if( bSuccessful )
	{
		//  Set the address of the first DPB in the linked list
		//
		lpDPB = lpLOL->first_dpb;


		//  Search the DPB (single)linked list for the DPB which
		//  coresponds to this drive.
		//
		while( lpDPB->drive != ucDriveNumber )
		{
			//  Point to the next DPB
			//
			lpDPB = lpDPB->next;

			//  Check if this is the end of the DPB list
			//
			if( ((ulong)lpDPB == 0L) || ((unsigned)lpDPB == 0xffff))
			{
				bSuccessful = FALSE;
				break;
			}
		}
	}

	//  If we found the DPB for this drive set it in the callers
	//  DPB pointer variable.
	//
	if( bSuccessful )
	{
		*(LPDPB *)plpDPB = lpDPB;
	}

	//  Return whether the function was successful
	//
	return bSuccessful;
}
