//
//   $Id: SETPARM.C 1.5 2002/10/15 22:49:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  $DOCMODULE:SETPARM.C
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This module is an interface to int-21h fn 440dh subfunction 40h which
//  is set parameters for block device.  This is part of the Datalight
//  Library.
//
//  bool dlSetParametersForBlockDevice( uchar ucDriveNumber,
//			void * pParameterBlock );
//
//  $DOCHISTORY
//  03/11/1999 bt  Created.
//

//
//  [CONSTANTS]
//

//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"

//
//  [ CONSTANTS ]
//

#define DEFAULT_CATEGORY_CODE						0x08
#define FAT32_AWARE_CATEGORY_CODE				0x48

#define SET_PARAMETERS_FOR_BLOCK_DEVICE		0x40


//  $DOCPROC: dlSetParametersForBlockDevice
//
//  bool dlSetParametersForBlockDevice( uchar ucDriveNumber,
//			void * pParameterBlock );
//
//  This function is an interface to int-21h fn 440dh subfunction 40h which
//  is set parameters for block device.  The drive numbers start with A=1.
//  This function returns FALSE if it fails and TRUE otherwise.  This is
//  part of the Datalight Library.  The parameters are described below:
//
//  ON ENTRY:
//
//  ucDriveNumber = The drive number (0=Default, 1=A, 2=B,...)
//  pParameterBlock  = The parameter block for this call
//
//  ON EXIT:
//
//  returns TRUE if the function call was successful, FALSE otherwise.
//
bool dlSetParametersForBlockDevice( uchar ucDriveNumber,
		void * pParameterBlock )
{
	//  [LOCALS]
	//
	bool bSuccessful;


	//  [CODE]
	//

	//  Use the generic IOCTL function to set the device parameters
	//
	bSuccessful = dlGenericIOCTLForBlockDevices( ucDriveNumber,
			DEFAULT_CATEGORY_CODE, SET_PARAMETERS_FOR_BLOCK_DEVICE,
			pParameterBlock );

	//  Return whether the function was successful
	//
	return bSuccessful;
}
