;
;   $Id: LFNATTR.ASM 1.4 2002/10/15 22:55:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:LFNATTR.ASM
; Created 07-07-1998 by Drew Gislason
;
; Copyright (c) 1998-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary.
;
; This module is part of the Long Filename Support for the Datalight
; Library. It contains the extended attribute function.
;
; bool PASCAL LFNExtendedGetSetAttr(int iAction, FileTime *pTime,
;	int *pAttr, char szLongName[], FileDate *pDate, int *pMilli,
;	long *pFileSize);
;
; $DOCHISTORY
; 07/07/1998 dg  Created.
;--------------------------------------------------------------------
include datalght.inc
InitModel %MEMMODEL, pascal, use16

include _dllfn.inc

.code

;--------------------------------------------------------------------
; $DOCPROC:LFNExtendedGetSetAttr
;
; bool PASCAL LFNExtendedGetSetAttr(int iAction, FileTime *pTime,
;	int *pAttr, char szLongName[], FileDate *pDate, int *pMilli,
;	long *pFileSize);
;
; Extended get or set file attributes, including date and time fields.
; This function is somewhat complicated, involving 9 possible actions:
;
;	0 - get file attributes
;	1 - set file attributes
;	2 - get physical file size
;	3 - set last write date/time
;	4 - get last write date/time
;	5 - set last access date
;	6 - get last access date
;	7 - set creation date/time
;	8 - get creation date/time
;
; On Entry
;	bl (iAction)		action
;	cx (pTime)		time, for actions 3,7
;	cx (pAttr)		attr, for action 1
;	ds:dx (szLongName)	filename
;	si (pMilli)		0-199 milliseconds, for actions 3,7
;	di (pDate)		date, for actions 3,5,7
; On Exit
;	dx:ax (pFileSize)	actual file size, action 2
;	cx (pTime)		time, for actions 4,8
;	cx (pAttr)		attr, for action 0
;	si (pMilli)		0-199 milliseconds, for actions 4,8
;	di (pDate)		date, for actions 4,6,8
;
; Returns TRUE if worked. Returns FALSE if failed.
;--------------------------------------------------------------------
ProcBegin	LFNExtendedGetSetAttr, PUBLIC
ProcArgs	@@iAction:word, @@pTime:PTR, @@pAttr:PTR, @@szLongName:PTR, @@pDate:PTR, @@pMilli:PTR, @@pFileSize:PTR
ProcUses	ds, si, di
ProcSetup
	; for set attribute, cx is used as attribute
	cmp	@@iAction, 1	; set attribute
	jnz	NotSetAttr
	lptr_ds	bx, @@pAttr
	mov	cx, [bx]
	jmp	short NotTime
    NotSetAttr:

	; otherwise, cx used as time
	lptr_ds	bx, @@pTime
	mov	cx, [bx]
	lptr_ds	bx, @@pDate
	mov	di, [bx]
	lptr_ds	bx, @@pMilli
	mov	si, [bx]
    NotTime:

	; finally, get the name into ds:dx
	lptr_ds	dx, @@szLongName

	; call LFN to get or set extended attrs
	mov	bx, @@iAction
	mov	ax, 7143h
	stc
	int	21h
	jc	@@SetPassFail

	; get attribute (action 0)
	cmp	@@iAction, 0
	jnz	NotGetAttr
	lptr_ds	bx, @@pAttr
	mov	[bx], cx
	jmp	short @@Done
    NotGetAttr:

	; get physical file size (action 2)
	cmp	@@iAction, 2
	jnz	NotFileSize
	lptr_ds	bx, @@pFileSize
	mov	[bx], ax
	mov	[bx+2], dx
	jmp	short @@Done
    NotFileSize:

	; is this a get date/time type action?
	cmp	@@iAction, 4	; get last write time/date
	jz	GetTime
	cmp	@@iAction, 8	; get creation date/time
	jz	GetTime
	cmp	@@iAction, 6	; get last access date
	jz	GetDate
	jmp	short @@Done

	; actions 4,8 return time
    GetTime:
	lptr_ds	bx, @@pTime
	mov	[bx], cx
	lptr_ds	bx, @@pMilli
	mov	[bx], si

	; actions 4,6,8 return date
    GetDate:
	lptr_ds	bx, @@pDate
	mov	[bx], di

	; we're successful, but have messed up the flags, so clear carry
    @@Done:
	clc

	; set pass/failed here (ax=TRUE if worked, FALSE if failed)
    @@SetPassFail:
	LFN_SET_DOS_ERROR
ProcEnd

	END
