//
//   $Id: LFNSMART.C 1.11 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:LFNSMART.C

	Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	These smart functions work whether long filenames are present or
	not. With long filenames present, the functions use the LFN
	equivalents. With no long filename support, the older, DOS2 style
	functions are used.

	The calling app need only make sure that all path strings conform
	to the longname string limits "LFN_MAX_PATH".

	$DOCHISTORY
	02/16/1999 dg  Added findfirst/next and file functions
	01-30-1999 bt  Added dostruct.h include
	07/17/1998 dg  Created by Drew Gislason
*********************************************************************/
#include <string.h>	// used for strcpy()
#include <ctype.h>   // used for toupper()

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlerrcod.h"

#include "_find.h"

/*
	[]---------------------------------------------------------------------[]
    |                           Find Functions                            |
	[]---------------------------------------------------------------------[]
*/

static struct WIN32_FIND_DATA lfnFindData;
static int iNumberOfOpenSmartFindHandles;
static int iOpenHandles[MAXHANDLES];

// Internal use only! record this find handle.
static bool PASCAL RecordFindHandle(int iHandle)
{
	int	i;
	int * ipOpenHandles;

	// close any handles that are open (non-zero)
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		// is this handle entry available?
		if(*ipOpenHandles == 0)
		{
			// another handle open
			*ipOpenHandles = iHandle;
			++iNumberOfOpenSmartFindHandles;

			return TRUE;
		}

		// try next handle
		++ipOpenHandles;
	}

	// not enough handle entries!
	return FALSE;
}

/*
	Remove this handle from the list
*/
static void PASCAL RemoveRecordedFindHandle(int iHandle)
{
	int	i;
	int * ipOpenHandles;

	// close the handle that refers to this one
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		// is this handle entry available?
		if(*ipOpenHandles == iHandle)
		{
			*ipOpenHandles = 0;
			--iNumberOfOpenSmartFindHandles;
		}
		++ipOpenHandles;
	}
}

// convert from lfnFindData to SMARTFIND structure
void ConvertToFFBLK(PSMARTFIND pSmartFind)
{
	char	*pszName;

	pSmartFind->attr			= lfnFindData.attr;
	pSmartFind->time			= *(unsigned *)(&lfnFindData.lastWriteTime.dos.time);
	pSmartFind->date			= *(unsigned *)(&lfnFindData.lastWriteTime.dos.date);
	pSmartFind->uLastAccessDate = *(unsigned *)(&lfnFindData.lastAccessDate.dos.date);
	pSmartFind->size			= lfnFindData.fileSize;
	if(*lfnFindData.szAliasFileName != 0)
		pszName = lfnFindData.szAliasFileName;
	else
		pszName = lfnFindData.szFileName;
	strcpy(pSmartFind->name, pszName);
}

/*
	$DOCPROC:SmartFindFirst

	Find the first file or directory that matches the given attrs. If
	LFNs available, the mask can include LFN-style wildcards.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if any files/dirs found.
*/
bool PASCAL SmartFindFirst(int nAttrs, char szPathMask[],
					PFIND pFindData)
{
	int	iConvertCode;
	int	iHandle;
	PSMARTFIND pSmartFind = (PSMARTFIND)pFindData;
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szPathMask[1] == ':' )
	{
		ucDrive = toupper(szPathMask[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNFindFirstFile(nAttrs, szPathMask, DOSDATE,
				&lfnFindData, &iHandle, &iConvertCode);
		if( fSuccessful )
		{
			// debugging aid. used to verify that SmartFindClose() is called for
			// every open. See AreSmartFindAllClosed()
			if(!RecordFindHandle(iHandle))
			{
				LFNFindClose(iHandle);
				return FALSE;
			}

			// mark this ffblk as a long findfirst
			pSmartFind->uSignature	= SMARTFIND_SIG;
			pSmartFind->dirnum		= 0xFFFF;
			pSmartFind->iHandle		= iHandle;

			// found, convert to ffblk format
			ConvertToFFBLK(pSmartFind);
		}
		else
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				dlSetDTA(pFindData);
				fSuccessful = dlFindFirstFile(nAttrs, szPathMask);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		dlSetDTA(pFindData);
		fSuccessful = dlFindFirstFile(nAttrs, szPathMask);
		if( fSuccessful )
		{
			strcpy( lfnFindData.szFileName, pFindData->name );
		}
	}

	return fSuccessful;
}


//  Return the address of the LFN in the GLOBAL LFN find
//  structure.  Be advised that this variable goes stale
//  with subsequent find first/next calls.  This was
//  placed here as a temporary optimization to for
//  retrieving the long filename from our NOT so smart
//  find first/next functions.
//
void PASCAL GetSmartFindLFNAddress( char ** pszLFN )
{
	*pszLFN = &lfnFindData.szFileName;
}


/*
	$DOCPROC:SmartFindNext

	Find the next occurence of a previously found file/dir. The find data
	must be exactly what was returned previously.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if another item found.
*/
bool PASCAL SmartFindNext( PFIND pFindData)
{
	PSMARTFIND pSmartFind = (PSMARTFIND)pFindData;
	int	iConvertCode;
	int	iHandle;
	bool	fSuccessful;

	// was it a longname findfirst?
	if(IsSmartFind(pFindData))
	{
		// do the findnext
		iHandle = SmartFindHandle(pFindData);
		fSuccessful = LFNFindNextFile(iHandle, DOSDATE, &lfnFindData, &iConvertCode);

		// found, convert to ffblk format
		if(fSuccessful)
			ConvertToFFBLK(pSmartFind);

		return fSuccessful;
	}
	else
	{
		dlSetDTA(pFindData);
		fSuccessful = dlFindNextFile();
		if( fSuccessful )
		{
			strcpy( lfnFindData.szFileName, pFindData->name );
		}
	}

	// LFNs not present, use old style function
	return fSuccessful;
}

/*
	$DOCPROC:SmartFindClose

	Close a find opened by SmartFindFirst. Required, or RD will quickly
	run out of find handles.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if close worked.
*/
bool PASCAL SmartFindClose( PFIND pFindData)
{
	int	iHandle;

	if(IsSmartFind(pFindData))
	{
		iHandle = SmartFindHandle(pFindData);

		if(!LFNFindClose(iHandle))
			return FALSE;

		// remove that handle from our list
		RemoveRecordedFindHandle(iHandle);
	}
	return TRUE;
}

/*
	Debugging aid! verify that the number of open files handles is 0.
*/
bool PASCAL SmartFindAreAllClosed(void)
{
	return (iNumberOfOpenSmartFindHandles == 0);
}

/*
	Close all handles allocated from SmartFindFirst().
*/
void PASCAL SmartFindCloseAll(void)
{
	int	i;
	int	*ipOpenHandles;

	// close any handles that are open (non-zero)
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		if(*ipOpenHandles)
		{
			LFNFindClose(*ipOpenHandles);
			*ipOpenHandles = 0;
		}
		++ipOpenHandles;
	}

	// all handles closed
	iNumberOfOpenSmartFindHandles = 0;
}

#if UNIT_TEST

void TestFindFirst(void)
{
	FIND ff;
	unsigned uDate;

	SmartFindFirst(0, "*.*", &ff);
	if(SmartFindAreAllClosed())
		printf("error. should be 1 open!\n");

	uDate = SmartGetLastAccessDate(&ff);

	SmartFindNext(&ff);

	printf("last access date %x\n",uDate);

	SmartFindCloseAll();
	if(!SmartFindAreAllClosed())
		printf("error. should be all closed!\n");
}

void TestWildcardDelete(void)
{
	if(!SmartWildcardDelete("*Bugga*"))
		printf("Cannot delete *Bugga*");
	else
		printf("deleted *Bugga*");
}

void TestGetAttr(void)
{
	static char szName[] = "This is a longname.txt";
	unsigned uAttributes;

	if(!SmartGetFileAttributes(szName, &uAttributes))
		printf("Get failed on %s\n",szName);
	else
		printf("attr %X\n",uAttributes);
}

void main(void)
{
//	TestFindFirst();

//	TestWildcardDelete();

	TestGetAttr();
}
#endif
