//
//   $Id: SMTFREE.C 1.7 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:SMTFREE.C

	Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Returns the free space on the disk, all in ulongs.

	$DOCHISTORY
	03/04/1999 dg  Created by Drew Gislason
*********************************************************************/
#include <string.h>	// used for strcpy()

#define  USING_FAT32 1

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dl64.h"
#include "dlerrcod.h"
#include "fatworld.h"


/*
	Get drive free space, using FAT32 call if possible.
	Returns TRUE if worked, FALSE if failed (usually invalid drive)
*/
bool PASCAL SmartGetDriveFreeSpace(int drive,
				ulong *pulSectorsPerCluster,
				ulong *pulFreeClusters,
				ulong *pulBytesPerSector,
				ulong *pulTotalClusters)
{
	static char szDrive[] = "C:\\";
	GDF32		sFreeSpace;
	unsigned uSectorsPerCluster;
	unsigned uFreeClusters;
	unsigned uBytesPerSector;
	unsigned uTotalClusters;

	// 0 = default drive, convert to 0=A:, 1=B:, etc..
	if(!drive && !dlGetCurrentDrive(&drive))
		return FALSE;
	else
		--drive;

	// try first to get it using FAT32 method
	*szDrive = 'A' + drive;
	sFreeSpace.uLevel = 0;

	if(dlIsFat32World())
	{
		if(!dlGetDriveFreeSpace32(szDrive, &sFreeSpace))
			return FALSE;

		// return 32 bit values
		*pulSectorsPerCluster = sFreeSpace.ulSectorsPerCluster;
		*pulFreeClusters      = sFreeSpace.ulAvailableClusters;
		*pulBytesPerSector    = sFreeSpace.ulBytesPerSector;
		*pulTotalClusters     = sFreeSpace.ulTotalClusters;
	}
	else
	{
		// return old DOS style free space
		if(!__dlGetFreeSpaceOnDrive(drive+1, &uSectorsPerCluster,
						&uFreeClusters, &uBytesPerSector, &uTotalClusters))
			return FALSE;

		// return results in long form
		*pulSectorsPerCluster = uSectorsPerCluster;
		*pulFreeClusters      = uFreeClusters;
		*pulBytesPerSector    = uBytesPerSector;
		*pulTotalClusters     = uTotalClusters;
	}

	// everything worked
	return TRUE;
}

#if UNIT_TEST
int main(void)
{
	ulong ulSectorsPerCluster;
	ulong ulFreeClusters;
	ulong ulBytesPerSector;
	ulong ulTotalClusters;

	ulong ulBytesPerCluster;
	uquad	uqFreeBytes;
	uquad uqTotalBytes;

	if(!SmartGetDriveFreeSpace(0, &ulSectorsPerCluster, &ulFreeClusters,
			&ulBytesPerSector, &ulTotalClusters))
	{
		printf("Failed to get free space!\n");
	}
	else
	{
		// display results of call
		printf("spc %lu, free %lu, bps %lu, total %lu\n",ulSectorsPerCluster,
			ulFreeClusters, ulBytesPerSector, ulTotalClusters);

		ulBytesPerCluster = ulSectorsPerCluster * ulBytesPerSector;
		uqFreeBytes  = QuadMultiply(ulFreeClusters, ulBytesPerCluster);
		uqTotalBytes = QuadMultiply(ulTotalClusters, ulBytesPerCluster);
		printf("FreeBytes %,qu, TotalBytes %,qu\n",uqFreeBytes, uqTotalBytes);
	}

	// everything worked
	return 0;
}
#endif
