//
//   $Id: SMTXPAND.C 1.12 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  SMTXPAND.C
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary.
//
//  This function returns the expanded path name using an LFN call (7160h)
//  if LFN calls are available otherwise it uses an INT 21 60h.
//
//  History:
//
//  10-19-1999  bt  Created
//

#include <ctype.h>
#include <dir.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlerrcod.h"


//  Return the absolute path for a give path
//
//  On Entry:
//
//  iLFNExpandFlags = Expansion flags to use if using an LFN call
//  szPath = name/path combination to which expand
//
//  On Exit:
//
//  szExpandedPath = expanded name/path combination
//
//  Note: This function will also clean up the long paths
//  returned from a network or subst'd drive, replacing the
//  appropriate portion with a drive letter instead.   (TD 10/19/00)
//
bool PASCAL SmartExpandPath( int iLFNExpandFlags, char * szPath,
		char * szExpandedPath )
{
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;
   char RootDrive[4];                  // Contains the "rootdrive:\"
   char RootPathName[LFN_MAX_PATH];    // Store the Root path for compare
   int PathStart;                      // Start of actual subdirectory path
   int SubdirPathLen = 0;              // How many characters to copy
   int currentdrive;                   // Assumed to be *szExpandedPath root
	bool fIsNetworkPath;


	if( szPath[1] == ':' )
	{
		ucDrive = toupper(szPath[0]) - 'A' + 1;
      RootDrive[0] = szPath[0];
	}
	else
	{
		ucDrive = 0;
      dlGetCurrentDrive( &currentdrive );
      RootDrive[0] = currentdrive + 'A';
	}
   RootDrive[1] = ':';
   RootDrive[2] = '\\';
   RootDrive[3] = 0;

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
      LFNGetFullPath( iLFNExpandFlags, RootDrive, RootPathName );
      fSuccessful = LFNGetFullPath( iLFNExpandFlags, szPath,
				szExpandedPath );

		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
            dlExpandPath( RootDrive, RootPathName );
				fSuccessful = dlExpandPath( szPath, szExpandedPath );
			}
		}
	}
	else
	{
		//  Expand of 'a:' without a backslash will fail in an
		//  nonLFN environment.
		//
		if( (szPath[0] == 0) || ((szPath[1] == ':') && (szPath[2] == 0)) )
		{
			//  An expand of nothing or of "a:" should be equal to
			//  or root path string of "a:\", however it fails in
			//  DOS but we will handle it in the libraries
			//
			RootDrive[2] = '.';
			fSuccessful = dlExpandPath( RootDrive, RootPathName );
			strcpy( szExpandedPath, RootPathName );
			RootDrive[2] = '\\';
			RootDrive[3] = 0;
			dlExpandPath( RootDrive, RootPathName );
		}
		else
		{
			// LFNs not present, use old style function
			//
	      dlExpandPath( RootDrive, RootPathName );
			fSuccessful = dlExpandPath( szPath, szExpandedPath );
		}
	}


	// Now check for network or substed path
   //
   if( (fSuccessful) && ( strlen(RootPathName) > 3 ) )
   {
		//  Determine if this is a network path
		//
		if( (RootPathName[0] == '\\') && (RootPathName[1] == '\\') )
		{
			fIsNetworkPath = TRUE;
		}
		else
		{
			fIsNetworkPath = FALSE;
		}


		//  Only perform this expansion code if we don't want to
		//  see the network path form "\\" or don't want to expand
		//  substed drives
		//
		if( (fIsNetworkPath) || (iLFNExpandFlags & FULLPATH_SUBST) )
		{
			SubdirPathLen = strlen( szExpandedPath ) - strlen( RootPathName );

	      //  \\DL\SYS3\USERS\KEITH\SUBDIR
	      //  \\DL\SYS3\USERS
	      //                 ^
	      //
	      PathStart = strspn( RootPathName, szExpandedPath );


			// Now we have the data, stuff the path with it!
	      //
	      szExpandedPath[0] = RootDrive[0];
	      szExpandedPath[1] = ':';
	      memcpy( szExpandedPath+2, szExpandedPath+PathStart,
					SubdirPathLen+1 );

			if( (szExpandedPath[2] == 0) && (szExpandedPath[1] == ':') )
			{
				szExpandedPath[2] = '\\';
				szExpandedPath[3] = 0;
			}
		}
	}


	return fSuccessful;
}
