//
//   $Id: UNITTEST.C 1.7 2002/10/15 22:54:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	UNITTEST.C

	Copyright (c) 1998-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	Unit tests for the Long filename support.

	Interesting facts:
	- Got "access denied" when trying to change create date on rdonly
     file using LFNExtendedGetSetAttr().

	History:

	01-30-1999 bt  Added dostruct.h include

*********************************************************************/
#include <dos.h>
#include <dir.h>
#include <stdio.h>
#include <string.h>
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dllfn.h"

unsigned PASCAL lastDOSError;

char szLongName[LFN_MAX_PATH];

/*
	helper function because BC2 won't let us assign 0xFFFF to a word
	(just because the word has bitfields)
*/
void SetWord(void *pPtr, word val)
{
	word *puPtr = pPtr;
	*puPtr = val;
}

/* simple function so each test header has same look-and-feel */
void Testing(char szFuncName[])
{
	printf("Testing %s()...\n",szFuncName);
}

/* just so unit test doesn't rely on the rest of the DL library */
void DosClose(int nHandle)
{
	_BX = nHandle;
	_AX = 0x3E00;
	geninterrupt(0x21);
}

/* just create a file and close it */
bool LfnCreate(char szLongName[])
{
	int	nHandle;
	int	iActionTaken;

	if(!LFNCreateOpenFile(OPEN_ACCESS_READWRITE, A_NORMAL,
			FILE_CREATE | FILE_TRUNCATE, szLongName, 1, &nHandle, &iActionTaken))
		return FALSE;
	DosClose(nHandle);
	return TRUE;
}

/*
	UnitTest for LFNDATE.ASM

	bool PASCAL LFNGetLastAccessDate(int nHandle, FileDate *pLastAccessDate);
	bool PASCAL LFNSetLastAccessDate(int nHandle, word lastAccessDate);
	bool PASCAL LFNGetCreateTimeDate(int nHandle, FileTime *pCreateTime,
									FileDate *pCreateDate, int *pCreateMilli);
	bool PASCAL LFNSetCreateTimeDate(int nHandle, word createTime,
									word createDate, int createMilli);
*/
bool TestLfnDate(void)
{
	FILE * fp;
	int	nHandle;
	FileDate	fileDate, newFileDate;
	FileTime	fileTime, newFileTime;
	int	nMilli, nNewMilli;

	fp = fopen("X.Y","wb");
	if(!fp)
	{
		printf("Cannot open x.y\n");
		return FALSE;
	}
	nHandle = fp->fd;

	/* set the file date */
	Testing("LFNSetLastAccessDate");
	fileDate.day    = 31;
	fileDate.month  = 12;
	fileDate.year   = 20;
	if(!LFNSetLastAccessDate(nHandle, WORD(fileDate)))
	{
		printf("Can't Set LastAccessDate!\n");
		return FALSE;
	}

	/* get the file date, should be the same */
	Testing("LFNGetLastAccessDate");
	SetWord(&newFileDate, 0xFFFF);
	if(!LFNGetLastAccessDate(nHandle, &newFileDate))
	{
		printf("Can't Get LastAccessDate!\n");
		return FALSE;
	}
	if(memcmp(&newFileDate,&fileDate,sizeof(fileDate)) != 0)
	{
		printf("LastAccessDate doesn't match!\n");
		return FALSE;
	}

	/* set the create date/time */
	fileTime.second = 29;
	fileTime.minute = 59;
	fileTime.hour   = 23;
	fileDate.day    = 1;
	fileDate.month  = 2;
	fileDate.year   = 3;
	nMilli          = 5;
	Testing("LFNSetCreateTimeDate");
	if(!LFNSetCreateTimeDate(nHandle, WORD(fileTime), WORD(fileDate), nMilli))
	{
		printf("Can't Set CreateTimeDate!\n");
		return FALSE;
	}

	/* get the file create date/time */
	SetWord(&newFileDate, 0xFFFF);
	SetWord(&newFileTime, 0xFFFF);
	nNewMilli = -1;
	Testing("LFNGetCreateTimeDate");
	if(!LFNGetCreateTimeDate(nHandle, &newFileTime, &newFileDate, &nNewMilli))
	{
		printf("Can't Get CreateTimeDate!\n");
		return FALSE;
	}
	if(memcmp(&fileDate, &newFileDate, sizeof(fileDate)) != 0 ||
		memcmp(&fileTime, &newFileTime, sizeof(fileTime)) != 0 ||
		nMilli != nNewMilli)
	{
		printf("CreateTimeDate doesn't match!\n");
		return FALSE;
	}

	/* clean up */
	fclose(fp);
	unlink("x.y");

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNDIR.ASM
	bool PASCAL LFNMakeDirectory(char szLongName[]);
	bool PASCAL LFNRemoveDirectory(char szLongName[]);
	bool PASCAL LFNChangeDirectory(char szLongName[]);
	bool PASCAL LFNGetCurrentDirectory(int nDrive, char szLongName[]);
*/
bool TestLfnDir(void)
{
	char *pszTestDir = "Test Directory...Long Form";
	int	len;

	/* create the directory */
	Testing("LFNMakeDirectory");
	if(!LFNMakeDirectory(pszTestDir))
		return FALSE;

	/* change into the directory */
	Testing("LFNChangeDirectory");
	if(!LFNChangeDirectory(pszTestDir))
		return FALSE;

	/* verify directory exists by getting the cwd */
	Testing("LFNGetCurrentDirectory");
	if(!LFNGetCurrentDirectory(0, szLongName))
		return FALSE;
	len = strlen(szLongName);
	if(len < strlen(pszTestDir) ||
			strcmp(pszTestDir, szLongName + (len - strlen(pszTestDir))) != 0)
	{
		printf("Get Directory name didn't compare!\n");
		return FALSE;
	}

	/* delete the directory */
	if(!LFNChangeDirectory(".."))
		return FALSE;
	if(!LFNRemoveDirectory(pszTestDir))
		return FALSE;

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNCREAT.ASM
	bool PASCAL LFNCreateOpenFile(int iModeFlags, int iAttr, int iAction,
					char szLongName[], int iAliasHint,
					int *pHandle, int *pActionTaken);
	bool PASCAL LFNDeleteFiles(int iAttrs, char szLongMask[], bool bWildOK);
*/
bool TestLfnCreat(void)
{
	char *pszTestName  = "A [not-so-very-long] filename";
	char *pszTestName2 = "not me";
	char *pszTestMask  = "*not*";
	int	nHandle;
	int	iActionTaken;

	/* test creating a file */
	Testing("LFNCreateOpenFile");
	if(!LFNCreateOpenFile(OPEN_ACCESS_READWRITE, A_HIDDEN, FILE_CREATE,
				pszTestName, 5, &nHandle, &iActionTaken))
	{
		printf("Couldn't create %s\n",pszTestName);
		return FALSE;
	}
	if(iActionTaken != ACTION_CREATED_OPENED)
	{
		printf("action should have been created!\n");
		return FALSE;
	}
	if(LFNCreateOpenFile(OPEN_ACCESS_READWRITE, A_HIDDEN, FILE_CREATE,
				pszTestName, 5, &nHandle, &iActionTaken))
	{
		printf("Already created %s. Shouldn't be able to again\n",pszTestName);
		return FALSE;
	}
	DosClose(nHandle);		/* done, close the file handle */

	/* test deleting some files */
	Testing("LFNDeleteFiles");
	if(!LFNCreateOpenFile(OPEN_ACCESS_READWRITE, A_NORMAL, FILE_CREATE,
				pszTestName2, 0, &nHandle, &iActionTaken))
		return FALSE;
	DosClose(nHandle);
	if(!LFNDeleteFiles(A_HIDDEN, pszTestMask, DEL_WILD))
		return FALSE;

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNATTR.ASM
	bool PASCAL LFNExtendedGetSetAttr(int iAction, FileTime *pTime,
					int *pAttr, char szLongName[], FileDate *pDate, int *pMilli,
					long *pFileSize);
*/
bool TestLfnAttr(void)
{
	char *pszTestName = "x.y (for Attributes Tests)";
	FileTime	time, time2;
	FileDate	date, date2;
	int		iMilli, iMilli2;
	int		iAttr;
	long		lFileSize;

	Testing("LFNExtendedGetSetAttr");

	/* test setting/getting file attrs */
	if(!LfnCreate(pszTestName))
		return FALSE;
	iAttr = A_READONLY | A_SYSTEM;
	SetWord(&time, 0x1111);
	SetWord(&date, 0x2222);
	iMilli = 0x3333;
	lFileSize = 0x44445555L;
	if(!LFNExtendedGetSetAttr(EXT_ACT_SET_ATTR, &time,
					&iAttr, pszTestName, &date, &iMilli, &lFileSize))
	{
		printf("Couldn't set attr\n");
		return FALSE;
	}
	iAttr = 888;
	if(!LFNExtendedGetSetAttr(EXT_ACT_GET_ATTR, &time, &iAttr,
			pszTestName, &date, &iMilli, &lFileSize))
	{
		printf("Couldn't get attr\n");
		return FALSE;
	}
	if(iAttr != (A_READONLY | A_SYSTEM))
	{
		printf("attr didn't match!\n");
		return FALSE;
	}
	/* set back to normal */
	iAttr = 0;
	if(!LFNExtendedGetSetAttr(EXT_ACT_SET_ATTR, &time, &iAttr, pszTestName,
				&date, &iMilli, &lFileSize))
		return FALSE;

	/* test setting/getting file date/time */
	iAttr = 0xFFFF;
	time.hour   = 1;
	time.minute = 2;
	time.second = 3;
	date.day   = 1;
	date.month = 2;
	date.year  = FileYear(2099);
	iMilli = 0;
	if(!LFNExtendedGetSetAttr(EXT_ACT_SET_CREATION_DATE_TIME, &time,
					&iAttr, pszTestName, &date, &iMilli, &lFileSize))
		return FALSE;
	/* just in case they don't get set: make em invalid */
	SetWord(&time2, 0xFFFF);
	SetWord(&date2, 0xFFFF);
	iMilli2 = ~iMilli;
	if(!LFNExtendedGetSetAttr(EXT_ACT_GET_CREATION_DATE_TIME, &time2,
					&iAttr, pszTestName, &date2, &iMilli2, &lFileSize))
		return FALSE;
	if(memcmp(&time,&time2,sizeof(time)) != 0 ||
		memcmp(&date,&date2,sizeof(date)) != 0 ||
		iMilli != iMilli2)
	{
		printf("dates don't match!\n");
		return FALSE;
	}

	/* test getting file size */
	lFileSize = 55;
	if(!LFNExtendedGetSetAttr(EXT_ACT_GET_PHYSICAL_SIZE, &time,
					&iAttr, pszTestName, &date, &iMilli, &lFileSize))
		return FALSE;
	if(lFileSize != 0)
		return FALSE;

	/* delete the file. */
	if(!LFNDeleteFiles((A_READONLY | A_SYSTEM), pszTestName, DEL_NOWILD))
	{
		printf("Couldn't cleanup\n");
		return FALSE;
	}

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNFIND.ASM
	bool PASCAL LFNFindFirstFile(int nAttrs, char szLongMask[], int nDateFmt,
			struct WIN32_FIND_DATA *pFindData, int *pHandle, int *pConvertCode);
	bool PASCAL LFNFindNextFile(int nHandle, int nDateFmt,
			struct WIN32_FIND_DATA *pFindData, int *pConvertCode);
	bool PASCAL LFNFindClose(int nHandle);
*/
bool TestLfnFind(void)
{
	char *pszTestName  = "A [not-so-very-long] filename";
	char *pszTestName2 = "not me";
	char *pszTestMask  = "*not*";
	static struct WIN32_FIND_DATA w32FindData;
	static struct WIN32_FIND_DATA w32FindData2;
	int	nHandle = -1;
	int	iConvertCode = -1;

	/* test findfirst */
	Testing("LFNFindFirstFile");
	if(!LfnCreate(pszTestName))
		return FALSE;
	if(!LfnCreate(pszTestName2))
		return FALSE;
	if(!LFNFindFirstFile(A_NORMAL, pszTestMask, DOSDATE,
			&w32FindData, &nHandle, &iConvertCode))
	{
		printf("FindFirst didn't find anything\n");
		return FALSE;
	}
	if(nHandle == -1 || iConvertCode != CONVERT_OK)
	{
		printf("Handle or convert code bogus\n");
		return FALSE;
	}
	if(strstr(w32FindData.szFileName, "not") == 0)
	{
		printf("didn't return right file!\n");
		return FALSE;
	}

	/* test findnext */
	Testing("LFNFindNextFile");
	if(!LFNFindNextFile(nHandle, DOSDATE, &w32FindData2, &iConvertCode))
	{
		printf("FindNext didn't find anything\n");
		return FALSE;
	}
	if(strstr(w32FindData2.szFileName, "not") == 0)
	{
		printf("didn't return right file2\n");
		return FALSE;
	}
	if(strcmp(w32FindData.szFileName,w32FindData2.szFileName) == 0)
	{
		printf("findnext was same as findfirst file!\n");
		return FALSE;
	}

	/* close the find */
	if(!LFNFindClose(nHandle))
	{
		printf("failed to close handle\n");
		return FALSE;
	}
	if(LFNFindClose(nHandle))
	{
		printf("can't close twice!\n");
		return FALSE;
	}

	/* clean up */
	if(!LFNDeleteFiles(A_NORMAL, pszTestMask, DEL_WILD))
		return FALSE;

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNREN.ASM
	bool PASCAL LFNRenameFile(char szOldLongName[], char szNewLongName[]);
*/
bool TestLfnRen(void)
{
	char *pszRenTest  = "Rename Test.file";
	char *pszRenTest2 = "x.y";
	FIND ffblk;

	/* rename a file */
	Testing("LFNRenameFile");
	if(!LfnCreate(pszRenTest))
		return FALSE;
	if(!LFNRenameFile(pszRenTest, pszRenTest2))
		return FALSE;

	/* must be able to find the renamed file */
	if(findfirst("x.y", &ffblk, 0) != 0)
	{
		printf("File didn't rename to expected shortname x.y\n");
		return FALSE;
	}
	unlink(pszRenTest2);

	/* can't rename with wildcards */
	if(LFNRenameFile("*.OBJ", "*.XBJ"))
	{
		printf("Wildcards should NOT be allowed\n");
		return FALSE;
	}

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNPATH.ASM
	bool PASCAL LFNGetFullPath(uchar iAction, uchar iSubst,
					char szSrcLongName[], char szDstLongName[]);
*/
bool TestLfnPath(void)
{
	char *pszSubdir = "This is a subdirectory";
	char *pszShort  = "THISIS~1";
	int	len;

	Testing("LFNGetFullPath");

	/* create a subdirectory of known name */
	if(!LFNMakeDirectory(pszSubdir))
		return FALSE;
	if(!LFNChangeDirectory(pszSubdir))
		return FALSE;

	/* try long path */
	*szLongName = 0;
	if(!LFNGetFullPath(FULLPATH_LONGNAME, ".", szLongName))
	{
		printf("Cannot get longname\n");
		return FALSE;
	}
	len = strlen(szLongName);
	if(len < strlen(pszSubdir) ||
		strcmp(pszSubdir, szLongName+ (len-strlen(pszSubdir))) != 0)
	{
		printf("Wasn't long form\n");
		return FALSE;
	}

	/* try short path */
	*szLongName = 0;
	if(!LFNGetFullPath(FULLPATH_SHORTNAME, ".", szLongName))
	{
		printf("Cannot get shortname\n");
		return FALSE;
	}
	len = strlen(szLongName);
	if(len < strlen(pszShort) ||
		strcmp(pszShort, szLongName + (len-strlen(pszShort))) != 0)
	{
		printf("Wasn't short form\n");
		return FALSE;
	}

	/* clean up */
	if(!LFNChangeDirectory(".."))
		return FALSE;
	if(!LFNRemoveDirectory(pszSubdir))
		return FALSE;

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNINFO.ASM
	bool PASCAL LFNGetVolumeInformation(int iBufSize, char szRootName[],
					char szFileSystemName[],
					int *pFlags, int *pMaxName, int *pMaxPath);
	bool PASCAL LFNGetFileInfoByHandle(int nHandle,
					struct BY_HANDLE_FILE_INFORMATION *pFileInfo);

*/
bool TestLfnInfo(void)
{
	char szSystem[30];
	char *pszRoot = "C:\\";
	char *pszTestName = "x.y";
	int	iFlags, iMaxName, iMaxPath;
	int	nHandle;
	int	iActionTaken;
	struct BY_HANDLE_FILE_INFORMATION fileInfo;

	Testing("LFNGetVolumeInformation");
	*szSystem = 0;
	if(!LFNGetVolumeInformation(sizeof(szSystem), pszRoot,
					szSystem, &iFlags, &iMaxName, &iMaxPath))
	{
		printf("Couldn't get Vol Info\n");
		return FALSE;
	}
	if(*szSystem == 0 || iMaxName != LFN_MAXNAME || iMaxPath != LFN_MAX_PATH)
	{
		printf("Invalid GetVol results\n");
		return FALSE;
	}

	Testing("LFNGetFileInfoByHandle");
	if(!LFNCreateOpenFile(OPEN_ACCESS_READWRITE, A_NORMAL,
			FILE_CREATE | FILE_TRUNCATE, pszTestName, 1, &nHandle, &iActionTaken))
		return FALSE;
	if(!LFNGetFileInfoByHandle(nHandle, &fileInfo))
	{
		printf("GetInfo failed\n");
		return FALSE;
	}
	DosClose(nHandle);

	/* everything worked */
	return TRUE;
}

/*
	UnitTest for LFNMISC.ASM
	bool PASCAL LFNConvertFileTime(int iAction, int *pMilli,
												union FILETIME *pFileTime);
	bool PASCAL LFNGenerateShortName(int iCharSet, int iNameType,
									char szLongName[], char szShortName[]);
*/
bool TestLfnMisc(void)
{
	char *pszTestName = "This is a long filename";
	char *pszTestShortName = "THISISAL   ";
	static char szShortName[DOTNAMESIZE];

	Testing("LFNGenerateShortName");
	if(!LFNGenerateShortName(BCS_OEM, SNFMT_ENAME, pszTestName, szShortName))
	{
		printf("Can't alias\n");
		return FALSE;
	}
	if(strcmp(pszTestShortName, szShortName) != 0)
	{
		printf("Aliases didn't compare\n");
		return FALSE;
	}

	/* everything worked */
	return TRUE;
}

int main(void)
{
	printf("Testing LFN Functions\n");

	/* test the date functions */
	if(!TestLfnDate())
	{
		printf("Failed!\n");
		return 1;
	}

	/* test the dir functions */
	if(!TestLfnDir())
	{
		printf("Failed!\n");
		return 2;
	}

	/* test create/del */
	if(!TestLfnCreat())
	{
		printf("Failed!\n");
		return 3;
	}

	/* test the attr functions */
	if(!TestLfnAttr())
	{
		printf("Failed!\n");
		return 4;
	}

	/* test the find functions */
	if(!TestLfnFind())
	{
		printf("Failed!\n");
		return 5;
	}

	/* test the rename function */
	if(!TestLfnRen())
	{
		printf("Failed!\n");
		return 6;
	}

	/* test the path functions */
	if(!TestLfnPath())
	{
		printf("Failed!\n");
		return 7;
	}

	/* test the info functions */
	if(!TestLfnInfo())
	{
		printf("Failed!\n");
		return 8;
	}

	/* test the misc functions */
	if(!TestLfnMisc())
	{
		printf("Failed!\n");
		return 9;
	}

	printf("All Tests Passed!\n");
	return 0;
}
