//
//   $Id: DL64MUL.C 1.7 2003/03/10 17:38:12Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************************
   $DOCMODULE:DL64MUL.C

   Copyright (c) 1999-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

   uquad PASCAL QuadMultiply(ulong ulValue1, ulong ulValue2);

   $DOCHISTORY
   02/23/1999 dg  Created by Drew Gislason
***************************************************************************/
#include "datalght.h"
#include "dl64.h"

// multiple 2 unsigned longs to get a quad
uquad PASCAL QuadMultiply(ulong ulValue1, ulong ulValue2)
{
   uquad uqReturn;
   uquad * puqReturn;
   uchar ucTempResult[12];

#if 0
   // This routine does mutliplication by adding the products of
   // successive multiplications of bytes, using unsigned long math.
   // While this sounds like it couldn't overflow, it can -- but only
   // in cases approaching MAX unsigned.
   //
   // The routine that replaces this one (below) does multiplication
   // by QUAD adding the products of successive multiplications of
   // words (unsigned ints).
   //
   uchar * pucMul1;
   uchar * pucMul2;
   unsigned uResultStartIndex = 0;
   unsigned uResultIndex;
   unsigned uMul1Index;
   unsigned uMul2Index;

   puqReturn = &uqReturn;

   pucMul1 = (uchar *)&ulValue1;
   pucMul2 = (uchar *)&ulValue2;

   //  Initialize the return value to zero
   //
   *(ulong *)&ucTempResult[0] = 0;
   *(ulong *)&ucTempResult[4] = 0;
   *(ulong *)&ucTempResult[8] = 0;

   //  Perform the multiplication of two unsigned longs
   //
   uMul2Index = 0;
   while(uMul2Index < 4)
   {
      uMul1Index = 0;
      uResultIndex = uResultStartIndex;
      while(uMul1Index < 4)
      {
         *((ulong *)&ucTempResult[uResultIndex]) +=
               (unsigned)(pucMul1[uMul1Index]) * pucMul2[uMul2Index];
         uMul1Index++;
         uResultIndex++;
      }
      uMul2Index++;
      uResultStartIndex++;
   }
#else
   ulong ulFactor1A, ulFactor1B, ulFactor2A, ulFactor2B;

   //             {1A}{1B}
   //           * {2A}{2B}
   //          -----------
   //             2B * 1B
   //        2B * 1A << 2
   //        2A * 1B << 2
   //   2A * 1A      << 4
   //

   puqReturn = &uqReturn;
   ulFactor1A = (ulValue1 & 0xFFFF0000L) >> 16;
   ulFactor1B = ulValue1 & 0xFFFFL;

   ulFactor2A = (ulValue2 & 0xFFFF0000L) >> 16;
   ulFactor2B = ulValue2 & 0xFFFFL;

   //  Initialize the return value to zero
   //
   *(ulong *)&ucTempResult[0] = 0;
   *(ulong *)&ucTempResult[4] = 0;
   *(ulong *)&ucTempResult[8] = 0;

   //  Perform the multiplication of two unsigned longs
   //  (the quad word addition is required to catch overflows!)
   //
   AddQuadLong( ((uquad *)&ucTempResult[0]), ulFactor2B * ulFactor1B );
   AddQuadLong( ((uquad *)&ucTempResult[2]), ulFactor2B * ulFactor1A );
   AddQuadLong( ((uquad *)&ucTempResult[2]), ulFactor2A * ulFactor1B );
   AddQuadLong( ((uquad *)&ucTempResult[4]), ulFactor2A * ulFactor1A );

#endif

   //  Copy the result to the return variable
   //
   puqReturn->ulLo = *(ulong *)&ucTempResult[0];
   puqReturn->ulHi = *(ulong *)&ucTempResult[4];

   // done
   return uqReturn;
}

#if UNIT_TEST
void main(void)
{
   ulong ulValue1 = 0xFF000001L;
   ulong ulValue2 = 0xE7BL;
   uquad q;

   // comprehensive test :-)
   q = QuadMultiply(ulValue1, ulValue2);
   printf("%lX * %lX = %lX%08lX\n",ulValue1, ulValue2, q.ulHi, q.ulLo);
}
#endif
