//
//   $Id: DL8250.C 1.4 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

	8250.C  created 12/18/87  by Drew Gislason

   Copyright (c) 1987-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Low level interrupt driven serial communications for the
	8250 UART.  Uses no DOS calls nor BIOS calls.  Assumes a
	completely "dumb" machine.

	UART Specific Initialization Routines
		comm_open(baud,parms,port,iIrqNumber,rxBufSize,*rxBuffer)
											- initialize comm port
		comm_poll()						- check for and return a received character
		comm_putc(c)					- put a character
		comm_close()					- disable interrupt i/o

	UART Specific Character I/O Routines
		interrupt comm_rec_char_int()		- character i/o isr

	General Character Routines
		comm_getc()     	- wait indefinitely for a character
		comm_flush()    	- flush the input buffer

********************************************************************/
#include <time.h>
#include "datalght.h"
#include "dlbiosio.h"
#include "dl8250.h"


// uncomment this line and comment out 8250a.asm in the MAKEFILE to
// have the ISR code in C instead of .ASM
//
//#define ISR_IN_C	1


// Borland C
#ifdef __TURBOC__
#include <dos.h>

// Microsoft C
#else
#define MK_FP(seg,ofs)	((void far *) \
			   (((unsigned long)(seg) << 16) | (unsigned)(ofs)))
#define enable()		_enable()
#define disable()		_disable()
#endif

/* Globals */
comport cp = {0,0,0,0,0,{0,0,0,0}};


/* Prototypes */
void comm_to_char();
static void setportvect(void (interrupt far * routine)());
static int makebaud(unsigned baud);


/*
  []--------------------------------------------------[]
	|																	|
	|			CHARACTER DRIVEN I/O ROUTINES 				|
	|																	|
  []--------------------------------------------------[]
*/


#if 0


// This code has been replaced by comm_int_handler() in 8250a.ASM.
// Don't add it to the .PT file. You still have the option to use .C, though.
//
// Note there are two version of the C ISR, this one and another immediately
// below. This version of for the Datalight debugger. The other may be
// enabled by setting ISR_IN_C to TRUE (at the top of this file).
//

/*
 *  ISR (Interrupt Service Routine)
 *
 *	 Interrupt driven CHARACTER receive.  This routine stuffs
 *  characters received into a ring buffer.
 *
 *  NOTE: Do NOT have any LOCALS in this ISR or ctrl-break will
 *        not work.
 */
#define CAN 0x18
// void far interrupt comm_rec_char_int(void)
{
	extern int in_go_state;
	static unsigned char c;

	if((inportb(INTID) & (IIR_RxData | IIR_TxData | IIR_NoInt)) != IIR_NoInt)
	{
		/* get the data */
		c = inportb(DATAREG);

		/* check for RDEB ctrl-break request */
		if(in_go_state && c==CAN)
		{
			outportb(EOIPORT, 0x20);	/* EOI to 8259 */
			generate_nmi();				/* found in KERNINT.ASM */
		}

		/* put data in buffer and check for wrap around */
		cp.buffer[cp.head++] = c;
		if(cp.head >= cp.bufsize)
			cp.head = 0;

		/* send an end-of-interrupt to the 8259 */
		outportb(EOIPORT, 0x20);
  }
  else
  {
		(cp.old_vect)();
  }
}
#endif



#if ISR_IN_C
void far interrupt comm_int_handler(void)
{ static unsigned int bufferHead;
  static unsigned char interruptID;
  static unsigned char *cpBuffer;

	do {
		// find out what kind of serial interrupt we have
		//
		interruptID = inportb(INTID);

		// If Rx data, drain the FIFO (if there is one) into our
		// ring buffer. Just over-write old data.
		//
		if(interruptID & IIR_RxData)
		{
			do {
				bufferHead = cp.Rx.head;
				cpBuffer   = cp.Rx.buffer+bufferHead;
				*cpBuffer  = inportb(DATAREG);
				if(++bufferHead == cp.Rx.bufsize)
					bufferHead = 0;
				cp.Rx.head = bufferHead;

			} while(inportb(LINESTATUS) & LS_DRDY);
		}
		else
		{
			// for any IRQ type we don't support -- disable that int type
			//
			outportb(INTENABLE, IER_RxInt);
		}
	} while(!(interruptID & IIR_NoInt));

	// rearm the PIC, done.
	//
	outportb(PIC_EOIPort,PIC_EOIVal); // non-specific EOI
}
#endif


/*
	Enable the 8250 UART for interrupt driven receive and
	polling transmit.  Initialize a structure defining the
	port,	and set up the interrupt vector for receive.

	baud		(300,1200,2400..38400,57600,115)
	parms		(3=N81, 26=7E1, see COMM.H)
	port		(0 to 3)
	bufSize  (140...32767), size of receive buffer
*/
bool comm_open(unsigned baud, unsigned parms, int port, int iIrqNumber,
			      int rxBufSize, unsigned char *rxBuffer)
{
	unsigned char ucMaskPic1;
	unsigned char ucMaskPic2;
	int iInterruptNumber;
	unsigned char val;
	bool bSuccessful = TRUE;

	/* Check for defaults */
	if(port > 3) port = DEFPORT;		/* if port is invalid, use default */
	if(port < 0) port = DEFPORT;		/* if port is invalid, use default */
	if( ( iIrqNumber < 3 ) || ( iIrqNumber > 15 ) )
	{
		if( ( port == 0 ) || ( port == 2 ) )
		{
			iIrqNumber = 4;
		}
		else
		{
			iIrqNumber = 3;
		}
	}
	if(!parms)  parms = DEFPARMS;
	if(!baud)   baud  = DEFBAUD;

#if PACKET
	/* Special case for 115K baud (because processor may be too slow) */
	if(baud == 115)
		PacDelay = 100;
	else
		PacDelay = 0;
#endif

	// The COM port base addresses are retrieved, using the COM port number
	// as an index, from a table at the start of the BIOS data area.
	// ASSUME that COM1/COM3 and COM2/COM4 share IRQs 4 and 3, respectively
	//
	cp.base = GetCommPortBaseAddress(port);
	if(cp.base == 0)								// invalid COMn?
		return FALSE;

	cp.iIrqNumber = iIrqNumber;

	if( iIrqNumber < 8 )
	{
		ucMaskPic1 = (1 << (unsigned char)iIrqNumber);
		ucMaskPic2 = 0;
	}
	else
	{
		ucMaskPic1 = 2;
		ucMaskPic2 = (1 << ((unsigned char)iIrqNumber - 8));
	}

	iInterruptNumber = (iIrqNumber<8) ? (iIrqNumber+8) : (0x70+(iIrqNumber-8));
	baud    = makebaud(baud);			/* change baud from # to divisor */

	/* save initial port and vector */
	disable();
	cp.old_vect = (void far interrupt *)( *((long far *)
			(MK_FP(0,iInterruptNumber*4))) );

	cp.old_mask = val = inportb(MASKPORT);
	if( iIrqNumber < 8 )
	{
		/* disable interupts for this port (8259 controller) */
		val |= ucMaskPic1;					/* set the bit to 1 to disable */
		outportb(MASKPORT,val);
	}

	/* disable interupts for other PIC controller */
	cp.old_mask2 = val = inportb(MASKPORT2);
	val |= ucMaskPic2;					/* set the bit to 1 to disable */
	outportb(MASKPORT2,val);

	enable();

	// initialize the receive buffer
	//
	cp.Rx.head    = (cp.Rx.tail = 0);
	cp.Rx.bufsize = rxBufSize;
	cp.Rx.buffer  = rxBuffer;

	// initialize the serial port
	//
	outportb(LINECONTROL,0x80);			/* set high bit, so we can set baud */
	outportb(BAUDLOW,baud & 0xFF);		/* baud rate */
	outportb(BAUDHIGH,baud >> 8);
	outportb(LINECONTROL,parms);			/* stop/data/parity bits */

	outportb(FIFOCONTROL,1);		// try to enable 16650 FIFO's
	if(inportb(INTID)&0xC0)			// if they don't come on its an 8250
	{
		// Set 16550 Rx FIFO trigger depth to 1 and clear all FIFOs
		//
		outportb(FIFOCONTROL,7);
	}

   // clear all pending data and interrupts (FIFO check clears Tx int)
	//
   inportb(MODEMSTATUS);    					//	clear an delta MSR ints
   while(inportb(LINESTATUS) & LS_DRDY)	// clear any LSR ints
	{
		inportb(DATAREG);							// drain all Rx data
	}


#if !POLLING
	// Put our ISR's address in the interrupt vector table and then
   // enable interrupts at the 8259 Programmable Interrupt Controller (PIC)
	//
   comm_to_char();       				// point vector to character ISR
   val  = inportb(MASKPORT);
   val &= ~ucMaskPic1;              // set bit=0 in PIC to enable IRQs
   outportb(MASKPORT,val);				// program the PIC

   val  = inportb(MASKPORT2);
   val &= ~ucMaskPic2;					// set bit=0 in PIC to enable IRQs
   outportb(MASKPORT2,val);			// program the PIC

   // Finally, enable RxData interrupts and the IRQ line at the UART.
	// Turn on DTR and RTS.
	//
   outportb(INTENABLE, IER_RxInt);
   outportb(MODEMCONTROL, (MC_INT | MC_RTS | MC_DTR));
#endif

	return bSuccessful;
}

/*
 *  Disable interrupts for this serial port.
 */
void comm_close(void)
{
	int iInterruptNumber;

	if(cp.base == 0)	// just return if port never opened
		return;

	iInterruptNumber =
			(cp.iIrqNumber<8) ? (cp.iIrqNumber+8) : (0x70+(cp.iIrqNumber-8));
	cp.base = 0;

	disable();

	/* restore interrupt state */
	outportb(MASKPORT,cp.old_mask);

	/* restore interrupt state */
	outportb(MASKPORT2,cp.old_mask2);

	/* restore interrupt vector */
	*(long far *)MK_FP(0,iInterruptNumber*4) = (long)cp.old_vect;

	enable();
}

int comm_setup(int baud,int parms)
{

	if(cp.base == 0)		// just return if port never opened
		return ERR;

	// initialize the serial port
	//
	outportb(LINECONTROL,0x80);			/* set high bit, so we can set baud */
	outportb(BAUDLOW,baud & 0xFF);		/* baud rate */
	outportb(BAUDHIGH,baud >> 8);
	outportb(LINECONTROL,parms);			/* stop/data/parity bits */
	return OK;
}

#if POLLING

/*
 *  Check UART for a received character.  Return it.
 */
int comm_poll(void)
{
	unsigned char val;

	val=inp(LINESTATUS);
	if(val&0x01)
		return (unsigned char)inportb(DATAREG);
	return ERR;
}


#else

/*
 *  Grab a character out of the receiver ring buffer.
 */
int comm_poll(void)
{ unsigned c;

	if(cp.Rx.head != cp.Rx.tail)
	{
		c = cp.Rx.buffer[cp.Rx.tail++];

		if(cp.Rx.tail >= cp.Rx.bufsize)
			cp.Rx.tail=0;

		return (c & 0xff);
	}
	return ERR;
}

#endif


/*
 *  Send a character out the UART.  This is a "polled"
 *  transmit, not interrupt driven.
 */
void comm_putc(char c)
{
	unsigned char val;
#if PACKET
	int delay;
#endif

	while(TRUE)
	{
		val=inportb(LINESTATUS);

		/* if trans buffer is empty, we can send it */
		if(val & 0x20)
		{
#if PACKET
			if(PacDelay)
				for(delay=0; delay<PacDelay; ++delay)
					/* waste time at 115K baud */;
#endif
			outportb(DATAREG,c);
			return;
		}
	}
}

int comm_check(void)
{
	int c;

	if(cp.Rx.head != cp.Rx.tail)
	{
		c = cp.Rx.buffer[cp.Rx.tail];
		return c;
	}
	return ERR;
}

/*
  []--------------------------------------------------------[]
	|																			|
	|			GENERIC ROUTINES TO ALL UART/CONTROLLERS			|
	|																			|
  []--------------------------------------------------------[]
*/

/*
 *  Throw out characters until there are no more (either
 *  in the ring buffer or waiting at the port).
 */
void comm_flush(void)
{
	while(comm_poll()!=ERR)
		/* toss the character */;
}

/*
 *  Wait indefinitely for a character from the port.
 */
int comm_getc(void)
{
	int c;
	while(TRUE)
	{
		if((c=comm_poll())!=ERR)
			break;
	}
	return c;
}


int comm_tgetc(int ticks)
{
	int  c;
	long timeout;

	if((c=comm_poll())!=ERR)
		return c;

	timeout = clock() + ticks;
	do
	{
		if((c=comm_poll())!=ERR)
			return c;

	} while(timeout > clock());

	return ERR;
}


/*
 *  Set the interrupt driven receive VECTOR to point
 *  to the character driven ISR.
 */
void comm_to_char(void)
{
	cp.Rx.head= cp.Rx.tail= 0;
	setportvect(comm_int_handler);
}


/*
 *  Set a vector to point to a routine.
 */
static void setportvect(void (interrupt far * routine)(void))
{
	int iInterruptNumber;

	disable();

	iInterruptNumber =
			(cp.iIrqNumber<8) ? (cp.iIrqNumber+8) : (0x70+(cp.iIrqNumber-8));
	*(long far *)MK_FP(0,iInterruptNumber*4) =	(long)routine;

	enable();
}


/*
 *  From an unsigned value, make a baud rate divisor
 *  for the 8250 on a PC.
 */
static int makebaud(unsigned baud)
{
	switch(baud)
	{
		case   300 : baud = B300;    break;
		case  1200 : baud = B1200;   break;
		case  2400 : baud = B2400;   break;
		case  4800 : baud = B4800;   break;
		case  9600 : baud = B9600;   break;
		case 19200 : baud = B19200;  break;
		case 38400L: baud = B38400;  break;
		case 57600L: baud = B57600;  break;
		case 115   : baud = B115200; break;
		default    : baud = B9600;   break;
	}
	return baud;
}
