//
//   $Id: DLADDNAM.C 1.9 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLADDNAM.C  created 04/25/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library:

	bool AddName(char *pPath, const char *pName);

	These functions are used to add and remove names from an
	absolute path string.

********************************************************************/
#define __FILENUM__ 506

// includes
#include <dir.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dlpath.h"
#include "dosdef.h"
#include "dllfn.h"


/*
	Add the name to the end of a path.
	Return FALSE if strings together are too long.
*/
bool AddName(char *pPath, const char *pName)
{
	unsigned uBackSlash = 0;
	int maxPathLen;

	if( LFNPresent )
	{
		maxPathLen = LFN_MAX_PATH;
	}
	else
	{
		maxPathLen = MAXPATH;
	}

	// validate parameters
	ASSERT(pPath);
	ASSERT(pName);

	// tack on the backslash
	if(LastChar(pPath) != '\\')
		uBackSlash++;

	// don't allow strings too long
	if((strlen(pPath) + strlen(pName) + 2 + uBackSlash) >= maxPathLen)
		return FALSE;

	// tack on the backslash
	if( uBackSlash )
		strcat(pPath, "\\");

	// add on the name
	strcat(pPath, pName);

	return TRUE;
}

#if UNIT_TEST	// {
#include <stdio.h>
#include <stdlib.h>

void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT file %u, line %u\n",fileNum,lineNum);
	exit(99);
}

int main(void)
{
	char szPath[MAXPATH];

	// test Add/Remove Name
	strcpy(szPath, "C:\\");
	ASSERT(AddName(szPath, "FILENAME.EXT"));
	ASSERT(strcmp(szPath, "C:\\FILENAME.EXT") == 0);
	ASSERT(AddName(szPath, "A"));
	ASSERT(strcmp(szPath, "C:\\FILENAME.EXT\\A") == 0);
	ASSERT(RemoveName(szPath));
	ASSERT(strcmp(szPath, "C:\\FILENAME.EXT") == 0);
	ASSERT(RemoveName(szPath));
	ASSERT(strcmp(szPath, "C:\\") == 0);
	ASSERT(!RemoveName(szPath));

	printf("passed");
	return 0;
}

#endif	// } UNIT_TEST
