//
//   $Id: DLBAT.C 1.4 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//	 DLBAT.C
//  Created by D. Edwards  9/25/96
//
//	 Determines if the code is executing as part of a batch process.
//
//  Copyright (c) 1996-2002 Datalight, Inc.
//  All rights reserved.
//  Datalight Confidential and Proprietary
//
//  The method employed here is the same one used by MSDOS v6.22a
//  XCOPY.EXE. It scans (what it assumes to be) COMMAND.COM for the
//  string "cox" and inspects the succeeding byte which is TRUE if
//	 a .BAT file is currently executing.
//
//  XCOPY defaults to "we're in a .BAT file" if it can't find "cox".
//
#include <mem.h>
#include <dos.h>
#include "dostruct.h"
#include "datalght.h"
#include "dlpath.h"


// The target string is the 3 bytes below
//
static char cox_chars[] = "cox";


// How many bytes in the resident COMMAND.COM to search through,
// must be less than 65280 bytes.
//
#define  MAX_BYTES_TO_SEARCH   0x1000

// first two bytes of a PSP
//
#define	INT20_OPCODES			 0x20CD

bool AreInBatch(void)
{
  void far * pShell;		// ptr to COMMAND.COM (presumably)
  void far * p2C; 		// ptr to 'c' within COMMAND
  void far * p2OX;		// ptr to possible "ox" within COMMAND
  char   batStat = 1;	// return status, default == 1, like MS XCOPY
  int 	searchLen;		// max chars to search
  ulong  searchLimit;	// max address to search
  struct PSP far *pPSP; // a pointer to a far away PSP

	pPSP = MK_FP(_psp,0);
	while(pPSP->last_psp != FP_SEG(pPSP))
	{
		if(pPSP->int_20 != INT20_OPCODES)
			goto	SHELL_NOT_FOUND;

		pPSP = MK_FP(pPSP->last_psp,0);
	}

	// Parent shell found, skip over its PSP
	//
	pShell = (void far *)((char far *)(pPSP)+0x100);

	// get last address to search, and max search length
	//
	searchLimit = (ulong)(pShell) + (ulong)(MAX_BYTES_TO_SEARCH);
	searchLen   = MAX_BYTES_TO_SEARCH;

	// look for "cox"
	do
	{
		// look for 'c', if not there, exit
		//
		p2C = _fmemchr(pShell, *cox_chars, (size_t)(searchLen));
		if(p2C == 0)
			break;


		// found a 'c', check for the "ox"
		//
		p2OX = (char far *)(p2C)+1;
		if(*((unsigned far *)(p2OX)) == *((unsigned *)(&cox_chars[1])))
		{
			// "cox" found, return the value of the succeeding byte
			//
			batStat = *((char far *)(p2C)+3);
			break;
		}


		// found a 'c' but not the "ox", adjust searchLen by bytes searched
		// if no more bytes to search, give up
		//
		searchLen     = (int)(searchLimit - (ulong)(p2C));
		pShell	     = p2OX;

	}while(searchLen > 0);

SHELL_NOT_FOUND:
	return batStat;
}



#if	UNIT_TEST

#include <stdio.h>


void	main(void)
{
	if(AreInBatch())
		printf("Executing from a .BAT file.\n");
	else
		printf("Executing from the command line.\n");
}

#endif
