//
//   $Id: DLEXITPK.C 1.5 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	DLEXITPK.C  Created by D. Edwards  04-28-97

   Copyright (c) 1996-2002 Datalight, Inc.
   All rights reserved.
   Datalight Confidential and Proprietary

	Module to handle C RTL signals, int 23h (^C) and int 24h
	(critical error).

	The main purpose of this module is to trap the various DOS
	abort "signals" and to call exit() in such cases so that all
	atexit() functions are processed. To use these interrupt and
	signal handlers, simply call:

		void InstallExitPackage(void);

	See the comments for SignalHandler() below for information on
	how the RTL signals are handled.	See the comments for
	SetCtrlBrkHandler() and SetCriticalErrorHandler() for
	information on over-riding the default actions for these
	interrupts.

*/
#define __FILENUM__	515

#include <signal.h>
#include <float.h>
#include <stdlib.h>

#include "dostruct.h"
#include "datalght.h"
#include "dlapp.h"
#include "dlerrcod.h"

#if !defined(PASCAL)

// BC2 - BC4x compatibility defines
//
#if	(__BORLANDC__ < 0x300)
	#define	_USERENTRY
	#define  PASCAL   pascal
#else
	#define	PASCAL	_pascal
#endif

#endif // #if !defined(PASCAL)


// the int handlers are private to this routine and are found in
// EXITPKGA.ASM
//
void PASCAL InstallDOSErrorHandlers(void);
void RemoveDOSErrorHandlers(void);


// These pointers are called by the actual int23h and int 24h interrupt
// handlers when they are non-NULL
//
CtrlBrkHandlerType userInt23hHandler;
CritErrHandlerType userInt24hHandler;

// These variables are accessed by the int 24h handler
//
unsigned int24ax, int24ip, int24cs, int24flags;

// These are the addresses of the previous int 23h and int 24h
// interrupt handlers.
//
ulong oldInt23hHandler, oldInt24hHandler;



// Handles various C signals by ignoring all errors.
// This signal handler is defined by the BCC runtime library (See
// the BCC DOS Reference).
//
void _USERENTRY SignalHandler(int sig)
{

  	signal(sig, SIG_ERR);           // Tell the RTL to fail the operation
  	signal(sig, SignalHandler);     // and then reset the signal handler

	switch(sig)
	{
	case SIGINT  :						  // SIGINT handled by our int 23h handler
	case SIGABRT :						  // SIGABRT handled by our int 24h handler
		break;

	case SIGTERM :
	case SIGSEGV :
	case SIGILL  :
	case SIGFPE  :
	case FPE_EXPLICITGEN:
		SHIPASSERT(0==1);
	}
}

// Installs all of the error handlers. Each signal function called
// is defined by the BCC runtime library (see BCC DOS Reference).
//
// The BCC RTL defines the following signal values:
//
//  Constant     Meaning
//
//  SIGABRT      Abnormal termination. The default action terminates
//	  the calling program with exit code 3. (Real and protected mode.)
//
//  SIGBREAK     CTRL+BREAK signal. The default action terminates the
//	  calling program. (Protected mode only.)
//
//  SIGFPE       Floating-point error, such as overflow, division by
//	  zero, or invalid operation. The default action
//	  terminates the calling program. (Real and protected
//	  mode.)
//
//  SIGILL       Illegal instruction. This signal is not generated by
//	  DOS or OS/2, but is supported for ANSI compatibility.
//	  The default action terminates the calling program.
//	  (Real and protected mode.)
//
//  SIGINT       CTRL+C interrupt. The default action issues INT 23H.
//	  (Real and protected mode.)
//
//  SIGSEGV      Illegal storage access. This signal is not generated
//	  by DOS or OS/2, but is supported for ANSI
//	  compatibility. The default action terminates the
//	  calling program. (Real and protected mode.)
//
//  SIGTERM      Termination request sent to the program. This signal
//	  is not generated by DOS or OS/2, but is supported for
//	  ANSI compatibility. The default action terminates the
//	  calling program. (Real and protected mode.)
//
//  FPE_EXPLICITGEN what-ever a floating point lib might decide...
//
void InstallExitPackage(void)
{
	// SIGINT is handled by our own int 23h handler.
	// The RTL should never see it, we set it for completeness.
	//
	signal(SIGINT  ,SignalHandler);	// user abort...

	// SIGABRT is handled by our own int 24h handler.
	// The RTL should never see it, we set it for completeness.
	//
	signal(SIGABRT ,SignalHandler);


	signal(SIGTERM ,SignalHandler);
	signal(SIGSEGV ,SignalHandler);
	signal(SIGILL  ,SignalHandler);

	signal(SIGFPE  ,SignalHandler);	// internal error...
	signal(FPE_EXPLICITGEN,SignalHandler);

	InstallDOSErrorHandlers();
	atexit(RemoveDOSErrorHandlers);
}

/*
	This sets up a user routine to be called when an int 23h
	occurs. It returns a pointer to the previously installed
	handler. Int 23h is generated by DOS when it detects a ^C,
	^Brk or an <Alt><Num3> typed on the keyboard.

	When the user routine is called DS will be set to DGROUP and
	the ax,bx,cx,dx,ds and es registers will have been saved on
	the stack. The user handler must return a value >=0 to
	continue the program otherwise the program is terminated
	using the low byte of the handlers return code as the
	argument to exit().

	If no user routine is installed all int 23h are ignored.

	NOTE:	It may not always be prudent to abort the program in
			response to an asynchronous event such as an int 23h or
			int 24h. In such cases the user should install an int
			23h and int 24h handler and set a flag that tells the
			program that it should abort as soon as it is safe to
			do so. In such cases the routines that respond to the
			"we need to abort flag" must be sure to call exit()
			rather than _exit() or abort() so that all atexit()
			routines are processed before the program returns to
			DOS: Failure to follow this practice can crash the
			system.

*/
CtrlBrkHandlerType SetCtrlBrkHandler(CtrlBrkHandlerType handler)
{ CtrlBrkHandlerType oldHandler;

	oldHandler = userInt23hHandler;
	userInt23hHandler = handler;
	return oldHandler;
}


/*
	This sets up a user routine to handle user input in response
	to an int 24h critical error interrupt. It returns a pointer
	to the previously installed handler.

	When the user routine is called DS will be set to DGROUP and
	the ax,bx,cx,dx,ds and es registers will have been saved on
	the stack.  The user routine is passed the int 24h AX, DI and
	BP:SI register values. The user routine must return a valid
	DOS int 24h return value. See the MS-DOS Prog Ref pgs 137-139
	for more information.

	On return from the user routine the int 24h handler will
	compare the user routine return value to the actions allowed
	by DOS. If the returned action code is not allowed, it will
	be promoted in accordance with DOS' rules. If the code is
	(ultimately promoted to) Abort, the int 24h interrupt handler
	will call exit(ERR_USER_ABORT). Under all other conditions
	the action code return by the user routine will be returned
	to DOS.

	If no user routine is installed all int 24h are passed on to
	the previous int 24h handler. On return from the resident int 24h
	handler, the int 24h interrupt handler will respond to the
	suggested action just as if it had been determined by a user
	routine.

	NOTE:	It may not always be prudent to abort the program in
			response to an asynchronous event such as an int 23h or
			int 24h. In such cases the user should install an int
			23h and int 24h handler and set a flag that tells the
			program that it should abort as soon as it is safe to
			do so. In such cases the routines that respond to the
			"we need to abort flag" must be sure to call exit()
			rather than _exit() or abort() so that all atexit()
			routines are processed before the program returns to
			DOS: Failure to follow this practice can crash the
			system.

*/
CritErrHandlerType SetCritErrHandler(CritErrHandlerType handler)
{ CritErrHandlerType oldHandler;

	oldHandler = userInt24hHandler;
	userInt24hHandler = handler;
	return oldHandler;
}


#if UNIT_TEST

#include <stdio.h>
#include <float.h>
#include <io.h>

void InternalError(int fileNum, int lineNum)
{
	printf("\nInternalError %u%u\n",fileNum,lineNum);
	exit(ERR_INTERNAL);
}

int far CtrlBrkHandler(void)
{
	return IGNORE_CTRL_C;
}

#pragma argsused
int far CritErrHandler(unsigned ax, unsigned di, struct dev_hdr far *bpsi)
{
	return INT24_FAIL;
}

int main(void)
{ int a,b,c, fp;
  char x;

	InstallExitPackage();

	printf("Testing default handlers...\n");
	printf("press ^C ");
	x=getchar();

	printf("\nopen the drive door for drive A: then press any key...");
	x=getchar();
	fp = open("A:\TEST.TXT",0);
	fp = fp;

	printf("\nDefault handlers tested OK.\n");
	printf("Installing user handlers.\n");

	SetCtrlBrkHandler(CtrlBrkHandler);
	SetCritErrHandler(CritErrHandler);

	printf("Testing user handlers.\n");

	printf("press ^C ");
	x=getchar();

	printf("\nopen the drive door for drive A: then press any key...");
	x=getchar();
	fp = open("A:\TEST.TXT",0);

	printf("\nTesting integer divide by zero, press any key...");
	x=getchar();
	a = 1, b = 0;
	c = a/b;
	c = c;

	x=x;
	return ERR_NONE;
}

#endif
