//
//   $Id: DLEXT15.C 1.4 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	DLEXT15.C  created 06/25/1998 by Drew Gislason

	Copyright (c) 1998-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	This module contains routines that extend the Int 15h memory
	interface (fns 87h and 88h) to include the VDISK allocation
	extensions. It also provides a convenient C interface to copying
	memory around (rather than having to set up a Global Descriptor
	Table.

	$DOCHISTORY
	06/25/1998 dg  Created.
*********************************************************************/
#define __FILENUM__ 519

// includes
#if DEBUG
#include <stdio.h>
#endif
#include <dos.h>
#include <string.h>
#include "datalght.h"
#include "dlmem.h"

/* standard addresses */
#define  ONE_MEG		0x100000L
#define  KUSED_ADDR	(ONE_MEG+0x1E)
#define  SIG_ADDR		(ONE_MEG)

/* standard signature */
char szVDiskID[] = "\000\000\000VDISK3.3";
char szVDiskID2[]= "\353\050\220DLRDISK";

static char szTestBuff[10];		/* 10 bytes to check signature */


#define	ACCESS	0x93		/* rd/wr data segment */
#define	LIMIT		0xFE00	/* limit of 63.5K bytes */

/*
	Standard 386+ Descriptor
*/
typedef struct
{
	unsigned limit;			/* segment limit */
	unsigned addrLo;			/* bits  0-15 of address */
	uchar		addrMid;			/* bits 16-23 of address */
	uchar		access;
	uchar		reserved;
	uchar		addrHi;			/* bits 24-31 of address */
} Descriptor;

/* helper functions */
ulong Int15StartAddr(unsigned kBytes);
void SetDescriptorAddr(void *pDescriptor, ulong lAddr);
ulong LinearAddress(void far *lpPtr);
bool Int15CopyWords(unsigned nWords, void *pGDT);


/*
	A very simple GDT.
*/
typedef struct
{
	unsigned char zeros[8];
	unsigned char bios_stuff[8];
	Descriptor	src;
	Descriptor	dst;
	unsigned char bios_stuff3[20];
} GlobalDescriptorTable;

/*
	$DOCPROC:CopyInt15Data

	Helper routine for CopyTo and CopyFrom Int15h.
*/
bool CopyInt15Data(ulong lDstAddr, ulong lSrcAddr, unsigned nWords)
{
	static GlobalDescriptorTable GDT;

	if(nWords > LIMIT/2)
		return FALSE;

	/* set access and limits */
	GDT.src.access = GDT.dst.access = ACCESS;
	GDT.src.limit  = GDT.dst.limit  = LIMIT;

	/* convert source address to Descriptor format */
	SetDescriptorAddr(&GDT.src, lSrcAddr);
	SetDescriptorAddr(&GDT.dst, lDstAddr);

	/* now do the copy */
	return Int15CopyWords(nWords, &GDT);
}

/*
	$DOCPROC:CopyToInt15
*/
bool CopyToInt15(ulong lDstAddr, void far *pSrc, unsigned nWords)
{
	ulong lSrcAddr = LinearAddress(pSrc);
	return CopyInt15Data(lDstAddr, lSrcAddr, nWords);
}

/*
	$DOCPROC:CopyFromInt15
*/
bool CopyFromInt15(void far *pDst, ulong lSrcAddr, unsigned nWords)
{
	ulong lDstAddr = LinearAddress(pDst);
	return CopyInt15Data(lDstAddr, lSrcAddr, nWords);
}

/*
	$DOCPROC:IsFirstVDisk

	Is this the first vdisk? Checks for VDISK3.3 at 1 Meg+3.
*/
bool IsFirstVDisk(void)
{
	bool first = TRUE;
	int i;

	/* look for VDISK signature */
	CopyFromInt15(szTestBuff, SIG_ADDR, 5);

	/* check first signature */
	for(i=0; i<10 && szVDiskID[i]==szTestBuff[i]; ++i)
		;
	if(i==10)
		first = FALSE;

	/* check for Datalight specific signature */
	for(i=0; i<10 && szVDiskID2[i]==szTestBuff[i]; ++i)
		;
	if(i==10)
		first = FALSE;

	return first;
}

/*
	$DOCPROC:AvailInt15Memory
*/
unsigned Used15Memory(void)
{
	unsigned kUsed;

	/* not used? then 0 is OK */
	if(IsFirstVDisk())
		return 0;

	/* get the # of bytes currently used */
	CopyFromInt15(&kUsed, KUSED_ADDR, 1);

	/* MS-DOS loaded high sets bit 400h.. why? */
	if(kUsed == 0x440)
		kUsed = 0x40;

	return kUsed;
}

/*
	$DOCPROC:AvailInt15Memory
*/
unsigned AvailInt15Memory(void)
{
	/* how much is currently avail. How much is used? */
	return Int15MemSize() - Used15Memory();
}

static unsigned kUsed = 0;

/*
	$DOCPROC:AllocateInt15Memory
*/
bool AllocateInt15Memory(unsigned sizeK, ulong *pStartAddr)
{
	/* don't allocate if not enough room */
	if(sizeK > AvailInt15Memory())
		return FALSE;

	/* put VDISK signature there */
	if(IsFirstVDisk())
		CopyToInt15(SIG_ADDR, szVDiskID, 5);

	/* get current usedK */
	else
		kUsed = Used15Memory();

	/* determine start address (use function so no long math: MEMDISK) */
	*pStartAddr = Int15StartAddr(kUsed);

	/* update used */
	kUsed += sizeK;
	CopyToInt15(KUSED_ADDR, &kUsed, 1);

	return TRUE;
}

/*
	$DOCPROC:RestoreInt15Used

	When a RAM Disk is intialized, the boot sector is usually
	overwritten with BPB, 0s and such. While the signature is still
	valid, the "kUsed" field at 1Meg:1Eh has been trashed. This
	function restores that value (assuming allocation was made
	through these routines).
*/
void RestoreInt15Used(void)
{
	CopyToInt15(KUSED_ADDR, &kUsed, 1);
}
