//
//   $Id: DLPATH7.C 1.4 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   DLPATH7.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Contains the routines needed to get a generic path.

	Datalight ROM-DOS Tools Library:
	bool PartialToFullPath(char *pFullPath, char *pPartialPath);

	Todo:
	Make this internationally aware.

	Revision History
	03/16/99	jmb	Added ExandPath
	03-06-01  bt  Broke up dlpath.c for library granularity
	03-06-01  bt  Changed calls to borland libraries to DL libs

**********************************************************/
#define __FILENUM__	507

// includes
#include <dir.h>
//#include <dos.h>
//#include <stdio.h>
//#include <stdlib.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlpath.h"
#include "dllfn.h"
//#include "dlcntry.h"

/*
	Converts a partial path into drive:\path\name(s).  Will check
	to make sure the path exists (although the files might not).

	Examples (assuming current DIR is C:\TEMP):
	.						C:\TEMP\*.*
	C:						C:\TEMP\*.*
	C:*.C					C:\TEMP\*.C
	*.*  					C:\TEMP\*.*
	D:\TEMP				D:\TEMP\*.*
	\DEV\ROMDOS\*.c	C:\DEV\ROMDOS\*.C

	returns TRUE if worked, FALSE if failed (bad path).
*/
bool PartialToFullPath(char *pFullPath, char *pPartialPath)
{
	int iTemp;
	int flags;
	static char szOrgCwd[DL_MAXPATH];		// original Current Working Dir for drive
	static char szDrive[3];		// eg "C:"
	static char szDir[DL_MAXPATH];			// eg "\TEMP"
	static char szName[DL_MAXPATH];		// eg "FILENAME"
	static char szExt[4];			// eg "EXT"

	// validate parameters
	ASSERT(pFullPath);
	ASSERT(pPartialPath);

	// split up the path into managable parts
	flags = IntlFileNameSplit(pPartialPath, szDrive, szDir, szName, szExt);

	// if name is similar to "CON:", remove the colon
	if(LastChar(szName) == ':')
		*(szName+strlen(szName)-1) = 0;

	// determine drive
	if(flags & DRIVE)
		IntlStrUpr(szDrive);
	else
	{
		strcpy(szDrive, "A:");
		dlGetCurrentDrive( &iTemp );
		*szDrive = iTemp + 'A';
	}

	// get current working DIR for the drive
	// returns FALSE if invalid drive
	if(!dlGetCurrentDirectory( *szDrive - 'A', szOrgCwd ))
		return FALSE;

	// no directory part, no need to change directories
	if( !(flags & (DIRECTORY|FILENAME|EXTENSION)) )
		/* don't need to chdir */;

	// directory part may exist, try changing into new directory
	else if(dlSetCurrentDirectory(pPartialPath))
	{
		*szName = 0;
		*szExt = 0;
	}

	// full thing doesn't work, just try DIR part
	else if(*szDir)
	{
		// create the full path string
		strcpy(pFullPath, szDrive);
		strcat(pFullPath, szDir);

		// no trailing '\\' (unless it's the root)
		if(LastChar(pFullPath) == '\\' && strlen(pFullPath) > 3)
			pFullPath[strlen(pFullPath)-1] = 0;

		// if we can't change into this, it's not a valid directory
		if(!dlSetCurrentDirectory(pFullPath))
			return FALSE;
	}

	// convert directory to an absolute one
	dlGetCurrentDirectory(*szDrive - 'A',pFullPath);

	// copy in the REAL directory found, skipping drive "A:" part
	strcpy(szDir, pFullPath+2);

	// if no name or extention, assume wildcards
	if(*szName == 0)
		strcpy(szName, "*");
	if(*szExt == 0)
		strcpy(szExt, ".*");

	// build full path string
	strcpy(pFullPath, szDrive);
	strcat(pFullPath, szDir);
	if(LastChar(szDir) != '\\')
		strcat(pFullPath, "\\");
	strcat(pFullPath, szName);
	strcat(pFullPath, szExt);

	// restore original CWD
	dlSetCurrentDirectory(szOrgCwd);

	return TRUE;
}
