//
//   $Id: DLRECURS.C 1.10 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLRECURS.C  created 04/25/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Datalight ROM-DOS Tools Library

	This module provides a generalized directory tree processing
	mechanism.

********************************************************************/
#define __FILENUM__	514

#define PASCAL pascal

#include <dir.h>
#include <dos.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dlpath.h"
#include "dosdef.h"
#include "dlrecurs.h"
#include "dlerrcod.h"
#include "dllfn.h"

/*
	Recurse into a subdirectory tree.  Uses the path and mask
	to do the find firsts. The recursion is performed in (at most)
	two loops per level, depending on the search mode:

							LOOP #1						LOOP #2

	BREADTH_FIRST     call actuator for			recurse into subdirs
							each file/subdir
							as encountered

	DEPTH_FIRST			recurse into subdirs		call actuator for each
							calling the actuator		file
							on return

	AS_IS					call actuator for			none
							each file/subdir and
							recursing into
							subdirs as encountered

	Returns TRUE if it processed the entire directory.

*/
bool Recurse(char szPath[], uchar attr, int mode, ACTUATOR actuator)
{
	uchar searchAttr;						// The current search attribute
	FIND found;  				// BCC RTL file find block
	bool wasFound, status;				// findfirst() and function results
	static int level = -1;				// recursion depth
	static char szName[LFN_MAX_PATH];  // local path name
	static char szAll[] = "*.*";		// the classic find all string
	unsigned uAttributes;
	unsigned uLength;

	// validate parameters
	//
	ASSERT(szPath);
	ASSERT(!(attr & 0xC0));
	ASSERT(mode==BREADTH_FIRST || mode==DEPTH_FIRST || mode==AS_IS);
	ASSERT(actuator);


	// If just starting, get the fully cannonical path name of the
	// (possibly relative) szPath argument. Store the file name
	// part of the path string in szName.
	//
	if(++level == 0)
	{
		// convert to a full path
		//
		strcpy(szName, szPath);
		if(!SmartExpandPath( 0,szPath, szName))
		{
			status = FALSE;	// that path ain't there
			goto DONE;
		}

		//  DOS 6.22 expand wil place wildcards onto the name part
		//  if it does not already contain a file name
		//
		if( SmartGetFileAttributes( szPath, &uAttributes ) )
		{
			if( (uAttributes & A_SUBDIR ) == A_SUBDIR )
			{
				uLength = strlen( szPath );
				if( szPath[uLength] != '\\' )
				{
					szPath[uLength] = '\\';
					uLength++;
					szPath[uLength] = 0;
				}
				AddName( szPath, szAll );
			}
		}
      else
      {
         // For some reason, we failed to get information, so bail
         //
			status = FALSE;
			goto DONE;
      }


		// remember original name part
		//
		CopyNamePart(szName, szPath);
	}

	// If search is DEPTH_FIRST, all we want to see on the first loop
	// are subdirs. Otherwise we want to see whatever was specified
	// along with subdirs.
	//
	if(mode == DEPTH_FIRST)
		searchAttr = A_SUBDIR;
	else
		searchAttr = attr | A_SUBDIR;

	// couldn't set the name, return false
	//
	if(!SetName(szPath, (mode == BREADTH_FIRST ? szName : szAll)))
	{
		status = FALSE;
		goto DONE;
	}

	// First Loop:
	//
	wasFound = SmartFindFirst( searchAttr, szPath, &found );
	while(wasFound)
	{
		// skip "." and ".."
		//
		if(found.attr & A_SUBDIR && *found.name == '.')
		{
			wasFound = SmartFindNext(&found);
			continue;
		}

		// set name to found name
		//
		if(!SetName(szPath, found.name))
		{
			status = FALSE;
			goto DONE;
		}

		// Except if DEPTH_FIRST, process each node in this level
		// as encountered.
		//
		if(mode != DEPTH_FIRST)
		{
			// Call the actuator for this file or directory.
			// An actuator return value of FALSE means get
			// out of this level.
			//
			if(!(*actuator)(&found, szPath, level))
			{
				status = FALSE;
				goto DONE;
			}
		}

		// Except if BREADTH_FIRST search, recurse subdirectories
		// "." or ".." eliminated above
		//
		if(mode != BREADTH_FIRST)
		{
			if(found.attr & A_SUBDIR)
			{
				AddName(szPath, szName);
				Recurse(szPath, attr, mode, actuator);

				if(mode == DEPTH_FIRST)
				{
					// The entire subtree has been visited, visit the root.
					//
					if(!(*actuator)(&found, szPath, level))
					{
						status = FALSE;
						goto DONE;
					}
				}
			}
		}

		// Go on to the next entry in this directory
		//
		wasFound = SmartFindNext(&found);
	}

	SmartFindClose(&found);

	// If processing in order, then all entries have been processed, return
	//
	if(mode == AS_IS)
	{
		status = TRUE;
		goto DONE;
	}

	// If search is BREADTH_FIRST then we have already processed
	// all entries at this level of the directory tree: Now we
	// need to start over and descend into any subdirs that may
	// exist.
	//
	// If search is BREADTH_FIRST then we have already recursed
	// into any subdirs that may exist at this level of the
	// directory tree and we will have processed each of those
	// subdirs as we backtracked. Now all we want to do is process
	// any of the specified files in this directory.
	//
	if(mode == BREADTH_FIRST)
		searchAttr = A_SUBDIR;
	else
		searchAttr = attr;

	// If can't set the name, return false
	//
	if(!SetName(szPath, (mode == DEPTH_FIRST ? szName : szAll)))
	{
		status = FALSE;
		goto DONE;
	}

	// Second Loop:
	//
	wasFound = SmartFindFirst( searchAttr, szPath, &found );
	while(wasFound)
	{
		// skip "." and ".."
		//
		if(found.attr & A_SUBDIR && *found.name == '.')
		{
			wasFound = SmartFindNext(&found);
			continue;
		}

		// set name to found name
		//
		if(!SetName(szPath, found.name))
		{
			status = FALSE;
			goto DONE;
		}

		// If DEPTH_FIRST then do what we do with this file.
		// If BREADTH_FIRST then its already been done, all we want
		// to do now is descend. AS_IS searches will have already
		// returned.
		//
		if(mode == DEPTH_FIRST)
		{
			// Call the actuator to process the file.
			//
			if(!(*actuator)(&found, szPath, level))
			{
				status = FALSE;
				goto DONE;
			}
		}

		// If BREADTH_FIRST recurse subdirectories.
		// "." or ".." eliminated above
		// AS_IS searches will have already returned.
		//
		else if(found.attr & A_SUBDIR)
		{
			AddName(szPath, szName);
			Recurse(szPath, attr, mode, actuator);
		}

		// Go on to next entry in this directory
		//
		wasFound = SmartFindNext(&found);
	}

	SmartFindClose(&found);

	// All done OK.
	//
	status = TRUE;


DONE:

	RemoveName(szPath);
	--level;
	return status;
}


#if UNIT_TEST
#include <stdio.h>
#include <stdlib.h>

// returns next char input from STDIN, no echo
//
int GetKey(void)
{ int c;

	asm {
		mov	ah, 8							// read a key
		int	21h
		and	ax, 0FFh
		mov	c, ax
	}
	if(c==0)								// if it's an extended scan code
		c = GetKey();						// skip scan code, get char code

	return c;								// return char code
}

void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT %u.%u\n",fileNum,lineNum);
	exit(ERR_INVALID_PARAMETER);
}

bool Actuator( PFIND pFind, char szPath[], int level)
{
	printf("level %u attr %02X path %s\n", level, pFind->attr, szPath);
	return TRUE;
}

int main(void)
{
	char szPath[MAXPATH];

	strcpy(szPath, "d1\\*.*");
	printf("Processing directories breadth-first.\n");
	GetKey();
	Recurse(szPath, 0, BREADTH_FIRST, Actuator);

	printf("\nProcessing directories depth-first.\n");
	GetKey();
	strcpy(szPath, "d1\\*.*");
	Recurse(szPath, 0, DEPTH_FIRST, Actuator);

	printf("\nProcessing directories in-order.\n");
	GetKey();
	strcpy(szPath, "d1\\*.*");
	Recurse(szPath, 0, AS_IS, Actuator);

	printf("passed");
	return 0;
}

#endif
