//
//   $Id: LOGFILE.C 1.7 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/************************************************************************
 *	LOGFILE.C
 *
 *	Copyright (c) 1989-2002 Datalight, Inc.
 *	All Rights Reserved
 *	Datalight Confidential and Proprietary
 *
 *	Revision History:
 *
 *    05-22-98 de allowed for compilation in different memory models
 *		04/04/97 jb revised comments for clarity
 *		03/21/97 dg original
 *
 * Description:
 *
 *		This file provides the tools and a template needed to
 *		create a ROM-DOS test program.  The program should
 *		follow all of the following rules.
 *
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <time.h>
#include <string.h>

#include "logfile.h"

#ifdef __TURBOC__
	#if !(defined( __LARGE__ ))
   	#if (sizeof(void *) == sizeof(long))
    		#define __BIGBCC__	1
		#else
			#define __BIGBCC__	0
		#endif
	#else
		#if( __LARGE__ || __HUGE__ || __COMPACT__ )
			#define __BIGBCC__	1
		#else
			#define __BIGBCC__	0
		#endif
	#endif
#else
	#define __BIGBCC__			0
#endif

#if __BIGBCC__
	#define PTR_SIZE	4
#else
	#define PTR_SIZE	2
#endif

/*
	log_printf

	This function will display the output to both the
	screen and the file (if any) specified by the "TEST"
	variable in the environment.

*/
void	log_printf(char *pszFormat, ...)
{

	static char szName[260];
	static char aBuffer[MAXBUFSIZE];		// storage buffer
	unsigned *puFormat; 						// pointer to format string

	FILE *fpOutput;							// file handle
	char far * lpszFName;					// environment information pointers


	/* copy the output into a string */
	puFormat = (unsigned *)(((char*)&pszFormat)+PTR_SIZE); 	// pointer to format string
	vsprintf(aBuffer, pszFormat, puFormat);

	/* write to the console */
	fpOutput = fopen("CON","w");
	if(fpOutput)
	{
		fputs(aBuffer, fpOutput);
		fclose(fpOutput);
	}

	/* write to the testlog file */
	lpszFName = getenv("TEST");
	if(lpszFName)
	{
		_fstrcpy( szName, lpszFName );
		fpOutput = fopen(szName,"a");
		if(fpOutput)
		{
			fputs(aBuffer,fpOutput);
			fclose(fpOutput);
		}
	}
}


/*
	log_error

	This routine will log an error-return value
*/
void log_error(int val)
{
	int i;					// ubiquitous counter

	// close all handles except the ones used by the system (0-4)
	for(i=5; i<20; ++i)
		close(i);

	log_printf("***** Failed! (error code %u) *****\n",val);
	exit(val);
}

/*
	log_error_line

	This routine will log an error-return value, along with the line
   number passed in
*/
void log_error_line(int val, int line)
{
	int i;					// ubiquitous counter

	// close all handles except the ones used by the system (0-4)
	for(i=5; i<20; ++i)
		close(i);

	log_printf("***** Failed! (error code %u, line: %u) *****\n",val, line);
	exit(val);
}

/*
	IsNet

	This routine checks for the environment variable NET, which
   is set to 1 when testing ROM-DOS over a network
*/
int IsNet( void )
{
	char *net;  // pointer to environment variable space

	// get the pointer to the NET environment variable
	net = getenv("NET");

	// if NET is not initialized, return 0; otherwise 1
	if( net == NULL )
		return(0);
	else
		return(1);
}

/*
	log_na

	This function logs an "N/A" message for tests which are not
   applicable for certain configurations
*/
void log_na( void )
{
	log_printf("       N/A\n");

	exit(0);
}



/*
	log_time

	This procedure logs the current date and time message for any tests
	that request it

*/
void	log_time( void )
{
	time_t	tCurrentTime;


	// get the current time
	tCurrentTime = time(NULL);

	// print it out
	log_printf("TIMESTAMP: %s\n",
					ctime(&tCurrentTime)
					);
}
