//
//   $Id: dlatime4.c 1.4 2002/10/15 22:57:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   DLATIME.C  created 04/30/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Convert file date/time to ascii versions, using foreign
	country date/time separators.

	Datalight ROM-DOS Tools Library:
	void InitCountryInfo(void);
	void TimeToAscii(unsigned time, char szStr[]);
	void DateToAscii(unsigned date, char szStr[]);
	bool AsciiToDate(unsigned *puDate, char szStr[]);

	Revision History
	03/16/99	jmb	added AsciiToDate

********************************************************************/
#define __FILENUM__ 511

// includes
#include <dos.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "dostruct.h"
#include "datalght.h"
#include "dldate.h"

extern struct COUNTRY_INFO countryInfo;

#define JANUARY 1
#define FEBRUARY 2
#define MARCH 3
#define APRIL 4
#define MAY 5
#define JUNE 6
#define JULY 7
#define AUGUST 8
#define SEPTEMBER 9
#define OCTOBER 10
#define NOVEMBER 11
#define DECEMBER 12

bool AsciiToDate(unsigned *puDate, char szStr[])
{
	unsigned uYear, uMonth, uDay;
	char *p1, *p2;

	// validate parameters
	ASSERT(puDate);
	ASSERT(szStr);

	// set pointers to each part of the date after the valid seperators:
	//	                                                   '.', '-', '/'
	p1 = strchr(szStr, '-');
	if (p1)
	{
		p1++;
		p2 = strchr(p1, '-');
	}
	else
	{
		p1 = strchr(szStr, '/');
		if (p1)
		{
			p1++;
			p2 = strchr(p1, '/');
		}
		else
		{
			p1 = strchr(szStr, '.');
			if (p1 == NULL)
				return FALSE;					// unknown seperator
			else
			{
				p1++;
				p2 = strchr(p1, '.');
			}
		}
	}

	// check for the second seperator, if not there, return FALSE
	if (p2 == NULL)
		return FALSE;
	p2++;

	// szStr points to first value
	// p1 points to second value
	// p2 points to third value

	// now translate the values as needed
	switch(countryInfo.datefmt)
	{

		/* dd-mm-yy */
		case 1:
			uMonth = atoi(p1);
			uDay   = atoi(szStr);
			uYear  = atoi(p2);
			break;

		/* yy-mm-dd*/
		case 2:
			uMonth = atoi(p1);
			uDay   = atoi(p2);
			uYear  = atoi(szStr);
			break;

		/* mm-dd-yy */
		default:  /* (case 0) */
			uMonth = atoi(szStr);
			uDay   = atoi(p1);
			uYear  = atoi(p2);
			break;
	}

#if DEBUG
	printf("-- %u.%u.%u --", uMonth, uDay, uYear);
#endif

	if (uYear < 80)
		uYear += 2000;

	if (uYear <= 100)
		uYear += 1900;

	if (uYear < 1980)
		return FALSE;

	// validate the year
	if (uYear < 1980 || uYear > 2107)
		return FALSE;

	uYear -= 1980;

	if (uMonth > 12 || uMonth == 0)
		return FALSE;									// invalid month

	if (uDay == 0)
		return FALSE;									// invalid day

	// check day based on month
	if (uMonth == JANUARY || uMonth == MARCH   || uMonth == MAY ||
	    uMonth == JULY    || uMonth == AUGUST  || uMonth == OCTOBER ||
		 uMonth == DECEMBER)
		if (uDay > 31)
			return FALSE;


	if (uMonth == APRIL     || uMonth == JUNE ||
	    uMonth == SEPTEMBER || uMonth == NOVEMBER)
		if (uDay > 30)
			return FALSE;

	// check for leap year
	if (uMonth == FEBRUARY)
	{
#if DEBUG
		printf("uYear=%u -- %u\n", uYear, uYear%4);
#endif
		if ((uYear%4) == 0)
		{
			if (uDay > 29)
				return FALSE;
		}
		else
		{
			if (uDay > 28)
				return FALSE;
		}
	}

	// build date
	*puDate = (uYear << 9) | (uMonth << 5) | uDay;

	// make sure date is okay
	ASSERT(*puDate > 0);

	return TRUE;
}
