//
//   $Id: SIMBUF.C 1.5 2002/10/15 23:00:22Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  SIMBUF.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//
//  Simple Buffering Scheme
//
//
//  HISTORY:
//
//  06-07-1999  bt  Created
//


//  Borland libraries
//
#include "stdio.h"
#include "stdlib.h"
#include "mem.h"


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "fatworld.h"


//  Status flags
//
#define STATUS_VALID			0x01
#define STATUS_DIRTY			0x02


//
//
#define MAX_SECTOR_SIZE		1024


typedef struct BufInfo BUF_INFO;
struct BufInfo {
	ulong	ulSectorNumber;
	uchar	ucStatusFlags;
};
BUF_INFO MRUConvMemBuffer;

ulong ulFATSectorNumberDelta = 0;

extern bool fIsFat32World;


//  Return the address of the variable which contains the number of FATs
//
unsigned * GetFATCopiesVar( void )
{
	static unsigned uFATCopies = 2;
	return &uFATCopies;
}


//  Return the number of copies of the FAT.  This is set to something
//  other than what is in the DPB if the FAT copies are not identicle
//  and the user does not want to fix it.
//
unsigned GetFATCopies( void )
{
	unsigned * puFATCopies;

	puFATCopies = GetFATCopiesVar();
	return *puFATCopies;
}


//  Set the number of copies of the FAT.  This is set to something
//  other than what is in the DPB if the FAT copies are not identicle
//  and the user does not want to fix it.
//
void SetFATCopies( unsigned uFATCopies )
{
	unsigned * puFATCopies;

	puFATCopies = GetFATCopiesVar();
	*puFATCopies = uFATCopies;
}


//  Return the address of a temporary sector buffer
//
char * GetTempSectorBuffer( void )
{
	static char cTempSectorBuffer[MAX_SECTOR_SIZE];
	return cTempSectorBuffer;
}


//  Get the address of the variable which contains the working drive
//
uchar * GetAddressWorkingDriveVariable( void )
{
	static uchar ucWorkingDrive;
	return &ucWorkingDrive;
}


//  Save the drive number which we will be checking
//
void SetWorkingDriveNumber( uchar ucDOSDriveNumber )
{
	uchar * pucWorkingDrive;

	pucWorkingDrive = GetAddressWorkingDriveVariable();
	*pucWorkingDrive = ucDOSDriveNumber;
}


//  Restore the users current drive
//
uchar GetWorkingDriveDriveNumber( void )
{
	uchar * pucWorkingDrive;

	pucWorkingDrive = GetAddressWorkingDriveVariable();
	return *pucWorkingDrive;
}


//  Get the address of our local copy of the DPB
//
bool PASCAL GetDPB( LPDPB * plpDPB )
{
	static bool fDPBRetrieved = FALSE;
	static bool fSuccessful = FALSE;
	static char pBuffer[sizeof(DPB)+2];
	static PDPB pDPB;
	uchar ucDrive;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Determine if we have retrieved the DPB already.  If we have
	//  not retrieved the DPB ...
	//
	if( !fDPBRetrieved )
	{
		//  Set the address of the DPB;
		//
		pDPB = (DPB*)(pBuffer+2);


		//  Determine if we are in DOS 7.1.  If we are in DOS 7.1 ...
		//
		if( fIsFat32World )
		{
			//  Get the DPB for this drive
			//
			fSuccessful = dlGetExtDPB( ucDrive+1, sizeof(DPB)+2, pBuffer );
		}


		//  If we were not successful at getting the DPB in DOS 7.1
		//  (an occasional weird windows fluke) or if we are not in
		//  a 7.1 world ...
		//
		if(!fSuccessful)
		{
			//  Get the DPB for this drive
			//
			fSuccessful = dlGetDriveDPB( ucDrive+1, plpDPB );


			//  If we were successful at retrieving the DPB ...
			//
			if( fSuccessful )
			{
				//  Copy the DPB into our local copy
				//
				_fmemcpy( pDPB, *plpDPB, sizeof(DPB) );
			}
		}
	}


	//  Set the address of our local copy of the DPB in the return variable
	//
	if( fSuccessful )
	{
		*plpDPB = pDPB;
	}


	//  Set a flag indicating that we already got (or attempted to get)
	//  the DPB
	//
	fDPBRetrieved = TRUE;


	//  Return whether the function was successful
	//
	return fSuccessful;
}


//  Initialize the buffering scheme
//
void InitializeBuffer( void )
{
	MRUConvMemBuffer.ulSectorNumber = 0;
	MRUConvMemBuffer.ucStatusFlags = 0;
}


//  Flush the contents of the buffer
//
void FlushBuffer( void )
{
	uchar ucDrive;
	uchar ucCopies;
	ulong ulFatLength;
	ulong ulFatEnd;
	ulong ulSectorNumber;
	bool fSuccessful;
	LPDPB lpDPB;
	char * pTempSectorBuffer;


	//  Determine if this buffer is valid.  If this buffer
	//  is valid ...
	//
	if( MRUConvMemBuffer.ucStatusFlags & STATUS_VALID )
	{
		//  Determine if this buffer is dirty.  If this buffer
		//  is dirty ...
		//
		if( MRUConvMemBuffer.ucStatusFlags & STATUS_DIRTY )
		{
			//  As soon as we commit this sector it will not be
			//  dirty anymore.  Go ahead and set this buffer
			//  not dirty
			//
			MRUConvMemBuffer.ucStatusFlags &= ~STATUS_DIRTY;


			//  Get the drive number
			//
			ucDrive = GetWorkingDriveDriveNumber();


			//  Get the DPB for the drive so that we can have the
			//  sector size
			//
			GetDPB( &lpDPB );


			//  Determine the distance between copies
			//
			if( IsFat32Drive(lpDPB) )
			{
				ulFatLength = lpDPB->ulSectorsPerFAT;
			}
			else
			{
				ulFatLength = lpDPB->fat_len;
			}


			//  Determine the sector number where the FAT ends
			//
			ulFatEnd = lpDPB->fat_start + ulFatLength;


			//  Get the sector number
			//
			ulSectorNumber = MRUConvMemBuffer.ulSectorNumber;


			//  Determine the number of copies of this
			//  sector
			//
			{
				//  Determine if this is a FAT sector.  If this
				//  is a FAT sector ...
				//
				if( (ulSectorNumber >= lpDPB->fat_start) &&
						(ulSectorNumber < ulFatEnd) )
				{
					ucCopies = GetFATCopies();
					ulSectorNumber += ulFATSectorNumberDelta;
				}
				else
				{
					ucCopies = 1;
				}
			}


			//  Get the address of the temporary sector buffer
			//
			pTempSectorBuffer = GetTempSectorBuffer();


			//  Write each copy of this sector
			//
			while( ucCopies )
			{
				//  Determine which abswrite call to use and perform
				//  the write
				//
				if( fIsFat32World )
				{
					//  Write the sectors for this copy
					//  (drive 0=default,1=A,...)
					//
					fSuccessful = dlAbsWrite32( ucDrive+1, 1,
							ulSectorNumber, pTempSectorBuffer,
							ABS32WRITE_UNKNOWN );
				}
				else
				{
					//  Write the sectors for this copy (drive 0=A,1=B,...)
					//
					fSuccessful = dlAbsDiskWrite( ucDrive, 1,
							ulSectorNumber, pTempSectorBuffer );
				}


				//  Determine if there was an error
				//
				if(!fSuccessful)
				{
					//  ???
					//
					printf("Failed to write changes!\n");
					exit(255);
				}


				//  Mark that we wrote this copy
				//
				ucCopies--;


				//  Determine the starting sector for the next copy
				//
				ulSectorNumber += ulFatLength;
			}
		}
	}
}


//  Return a handle for the specified sector
//
#pragma argsused
//bool PASCAL ReadSectorBuffer( uchar ucDOSDriveNumber,
bool PASCAL ReadSectorBuffer( uchar ucDrive,
		ulong ulAbsSector, void ** ppBufferHandle )
{
	bool fSuccessful;
	char * pTempSectorBuffer;


	//  Determine if this sector is in the	conventional memory buffers.
	//  If this sector is in the conventional memory buffers, ...
	//
	if( (MRUConvMemBuffer.ulSectorNumber == ulAbsSector) &&
			(MRUConvMemBuffer.ucStatusFlags & STATUS_VALID) )
	{
		//  The MRU conventional memory buffer contains the
		//  requested sector
		//
		return TRUE;
	}
	else
	{
		//  Flush the contents of the buffer
		//
		FlushBuffer();


		//  Get the address of the temporary sector buffer
		//
		pTempSectorBuffer = GetTempSectorBuffer();


		//  Read the sector from disk into the LRU conventional
		//  memory buffer
		//
		if( fIsFat32World )
		{
			fSuccessful = dlAbsRead32( ucDrive+1, 1, ulAbsSector,
					pTempSectorBuffer );
		}
		else
		{
			fSuccessful = dlAbsDiskRead( ucDrive, 1, ulAbsSector,
					pTempSectorBuffer );
		}
		if(!fSuccessful)
		{
			//  ???
			//
			printf("Error reading sector!\n");
//			printf( "ulAbsSector: %lxh\n", ulAbsSector );
			exit(255);
		}


		//  Set the sector number in the buffer management structure
		//
		MRUConvMemBuffer.ulSectorNumber = ulAbsSector;


		//  Mark the MRU buffer as valid and not dirty
		//
		MRUConvMemBuffer.ucStatusFlags = STATUS_VALID;
	}


	//  The MRU conventional memory buffer contains the
	//  requested sector
	//
	return TRUE;
}


//  Return the address of the buffer for the specified handle
//
#pragma argsused
bool PASCAL GetBufferDataAddress( void * pBufferHandle,
		void far * * plpBufferAddress )
{
	char * pTempSectorBuffer;


	//
	//  Presently we are not using handles.  How do you know if
	//  handle is stale?  So instead it is assumed that the
	//  address you are seeking is for what was just set as
	//  the MRU sector when you first requested the sector.
	//


	//  Get the address of the temporary sector buffer
	//
	pTempSectorBuffer = GetTempSectorBuffer();


	//  Set the address of the request sector in the return variable
	//
	*plpBufferAddress = pTempSectorBuffer;

	return TRUE;
}


//  Get the address of a sector in conventional memory
//
void * GetSectorAddress( ulong ulSectorNumber )
{
	void * pBuffer;
	unsigned * puBufferHandle;
	uchar ucDrive;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Get the sector
	//
	ReadSectorBuffer( ucDrive,	ulSectorNumber, (void **)&puBufferHandle );


	//  Get the address
	//
	GetBufferDataAddress( puBufferHandle, &pBuffer );


	//  Return the address
	//
	return pBuffer;
}


//  Required by the FAT libraries but not used
//
#pragma argsused
bool PASCAL SetBufferCopiesAndDistance( void * pBufferHandle,
		unsigned uCopiesOfThisSector, ulong ulDistanceToNextSector )
{
	//
	//  This feature of the FAT libraries is not being used
	//  in this implementation because we might
	//  not be writing multiple FAT copies.
	//

	return TRUE;
}


//  Mark the MRU buffer as modified
//
#pragma argsused
bool PASCAL MarkBufferDirty( void * pBufferHandle )
{
	//
	//  Presently we are not using handles.  How do you know if
	//  handle is stale?  So instead it is assumed that the
	//  address you are seeking is for what was just set as
	//  the MRU sector when you first requested the sector.
	//


	//  Set the MRU buffer as modified
	//
	MRUConvMemBuffer.ucStatusFlags |= STATUS_DIRTY;

	return TRUE;
}


//  Get the address of the directory for this entry index and sector
//
PDIR GetDirEntryAtSector( unsigned uEntryIndex, ulong ulSectorNumber )
{
	char * pBuffer;


	//  Get the address of this sector in conventional memory
	//
	pBuffer = GetSectorAddress( ulSectorNumber );


	//  Return the address of the request directory entry
	//
	return (PDIR)(pBuffer + ( uEntryIndex * sizeof(DIR) ));
}


//  Lock the logical disk
//
void LockDrive( void )
{
	uchar ucDrive;
	bool fSuccessful;

	ucDrive = GetWorkingDriveDriveNumber();
	fSuccessful = dlLockLogicalVolume( ucDrive+1, 0, 0 );
	if( !fSuccessful )
	{
	}
}


//  Unlock the logical disk
//
void UnlockDrive( void )
{
	uchar ucDrive;

	ucDrive = GetWorkingDriveDriveNumber();
	dlUnlockLogicalVolume( ucDrive+1 );
	dlUnlockLogicalVolume( ucDrive+1 );
}


//  Special: Adjust sector number delta
//
void SetFATSectorNumberDelta( ulong ulValue )
{
	ulFATSectorNumberDelta = ulValue;
}
