//
//   $Id: DDTDISK.C 1.5 2002/10/15 23:46:40Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DDTDISK.C  created 01/27/91 by Roy Sherrill/Drew Gislason

   Copyright (c) 1991-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This is a debugging/testing tool based on the old VDISK.
	It DOES NOT work with the new MEMDISK/VDISK code.

	Syntax:
	DEVICE=VDISK.SYS [disk_sizeK [sec_size [root_entries]]] [/e]

	This driver is a RAM disk, that can allocate its memory
	from the DOS (640K) pool, Upper Memory (if loaded high),
	the XMS memory pool (HIMEM.SYS), or using the BIOS Int 15H
	extended memory interface.

	This VDISK is just one example of Datalight's Generic Disk
	Driver, MEMDISK.  Other implementations include MEMPAGED,
	MEMFIXED, and MEMROM.  See MEMDISK.ASM and the ROM-DOS
	Developer's Guide for details on building your own Generic
	Disk Driver.

**********************************************************/

#ifndef BUILTIN
#define BUILTIN 0
#endif

#if !(defined(TSRDEV))
#define TSRDEV 0			/* allow installing device as TSR? */
#endif
#define STATIC 0			/* =1 enables STATIC disk checking/usage */
#define TESTING 0			// =1 enables main() for testing

#include <dos.h>
#if !BUILTIN
#include <mem.h>
#endif
#include "compiler.h"
#include "dosver.h"
#include "ddtdisk.h"
//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );


// do nothing for now, until we implement an assert
#define ASSERT(x)


/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |                GLOBALS and STRUCTS                 |
	 |                                                    |
	[]----------------------------------------------------[]
*/

/* DEBUGGING ONLY! */
#define TWODRIVES	0		/* DEBUGGING ONLY! */

#if SUITE
void logcmd(char cmd);
#endif

#define DEF_DIR_ENTRIES 64
#if BUILTIN
#define DEF_SIZEK 32
#else
#define DEF_SIZEK 64
#endif

/*
	note: if under MS or IBM DOS always use 2 FATs, as they
	ignore this field and assume it to be 2!  Bozos.
*/
struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
#if (SUITE >= 2)
char std_volume[] = "DDTDISK    ";
#else
char std_volume[] = "VDISK V6.22";
#endif
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

#define LABEL_TIMEOFF 0x16
#define LABEL_DATEOFF 0x18
char label[32] =
{
#if (SUITE >= 2)
	'D','D','T','D','I','S','K',' ',		/* File Name */
	' ',' ',' ',								/* Extension */
#else
	'V','D','I','S','K',' ','v','6',		/* File Name */
	'.','2','2',								/* Extension */
#endif
	0x08,											/* Attribute */
	0,0,0,0,0,0,0,0,0,0,						/* Reserved  */
	0xC0, 0x32,									/* Time      */
	0xB5, 0x22									/* Date      */
};

char Disk_Intro[11] = { 0xEB,0x28,0x90, 'D','L','R','D','I','S','K' };
unsigned Boot_ID = 0xAA55;
char Init_FAT[4] = { FAT_ID, 0xFF, 0xFF, 0 };

#if BUILTIN
char disk_image[(unsigned)DEF_SIZEK*1024+16];
#endif

#define XMS_SIGNATURE 0xF0000000L

int using_xms;			/* is an XMS driver present? */
int xms_handle;		/* the allocated memory */
int size_adjusted;	/* was the disk size adjusted */
int extended;
unsigned total_extended;	/* total # of K bytes extended */
unsigned total_conventional;
#if STATIC
int static_ok;
#endif
int opt_unload;
int opt_showhelp;
unsigned disk_seg;						/* for conventional RAM use only */
unsigned long ext_adr = 0x100020L;	/* for extended (/E) use only w/ no XMS */
unsigned args[4] = {DEF_SIZEK, DEF_SEC_SIZE, DEF_DIR_ENTRIES};
unsigned arg_count;

bool fIsFat32World;

/* prototypes */
void showhelp(void);
#if STATIC
int is_static_disk(void);
#endif
void do_args(char far *cmdargs);
char far *getcount(char far * str, unsigned *count);
void adjust_bpb(unsigned sizeK, unsigned sec_size, unsigned dir_entries);
void zero_bytes(long offset, unsigned count);
void zero_sector(unsigned sector);
void itoa(unsigned numb, char *buffer, unsigned radix);
void far_memcpy(char far *dst, char far *src, unsigned count);
void bputs(char *s);
void bputc(char c);
void bputi(unsigned val);
void bputh(unsigned val);
unsigned allocate_extended_memory(unsigned sizeK);
int first_vdisk(void);
unsigned get_used(void);
void write_used(unsigned size);
unsigned long fp_to_long(void far *fp);
int ext_memread(long offset, void far *buffer, unsigned count);
int ext_memwrite(long offset, void far *buffer, unsigned count);
int ext_move_data(unsigned long dst, unsigned long src, unsigned count);
int get_ext_mem_size(void);
int himem_installed(void);
unsigned total_xms(void);
unsigned allocate_xms(unsigned sizeK);
int xms_move(struct xms_block * xp);


/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |                THE CODE SECTION                    |
	 |                                                    |
	[]----------------------------------------------------[]
*/


#if TESTING
unsigned memtop;
unsigned char memerr=0;
unsigned getmemtop(void)
{	asm {
		int	12h
		mov 	cl, 6
		shl	ax, cl
		mov	memtop, ax
	}
	return memtop;
}
#endif


/*
   Initialize the device.  This is called only once.  It
   fills in the Bios Parameter Block structure, and returns
   the amount of memory used by the device.

   bpb       The BPB must be filled in for any device.
   cmdline   This is the command line passed to a device
             that is installed via CONFIG.SYS.  This
             parameter is meaningless for built-in devices.
   *endseg   This unsigned is the paragraph just beyond
             the code and data of the device.  Update this
             only if the device requires more RAM from DOS
             (ala a RAM disk in conventional memory).

   Returns # of drives (0 means failure)
*/
#ifdef __TURBOC__
#pragma argsused
#endif
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
#if !BUILTIN
#if !TESTING
	extern char copyright[];
	extern unsigned memtop;
#endif
#endif

	int i;
#if STATIC
	int staticdisk;
#endif
	int sectors;
	long start;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
#if (SUITE >= 2)
	bputs("\nDDTDISK ");
#else
	bputs("\nVDISK ");
#endif
	if( fIsFat32World )
	{
 		bputs(SHORTVERSION);
 		bputs(" ");
 		bputs(BUILDVERSION);
	}
	else
	{
 		bputs(SHORTVERSION622);
 		bputs(" ");
 		bputs(BUILDVERSION);
	}
	bputs("\n");

#if !BUILTIN
	/* and copyright */
#if !TESTING
	bputs(copyright);
#else
	getmemtop();
#endif
	bputs("\n");

	/* determine switches, etc... */
	do_args(cmdline);
#endif

#if TSRDEV

	/* set up a unique ID for us */
	tsr_setidstr("VDISK5.0!@#$");

	if(opt_showhelp)
	{
		showhelp();
		return 0;		/* failed to init */
	}

	/* unload the old driver */
	if(opt_unload)
	{
		bputs("\n\nUninstalling...");
		tsr_uninstall();		/* should not return... */

		bputs("No installed driver found.\n");
		return 0;		/* failed to init */
	}

	if(memdev_already_loaded())
	{
		bputs("\n\nAlready loaded.\n");
		return 0;		/* failed to init */
	}
#endif

#if STATIC
	// by default, static disks are OK but, for VDISK, are generally
	// only practical in extended memory (see also MEMFIXED.C)
	//
	static_ok = TRUE;
#endif

	/* set start of disk */
#if BUILTIN
	ASM	mov disk_seg, DS
	disk_seg += ((unsigned)disk_image + 0xF) >> 4;
#else
	disk_seg = *endseg;
#endif

	/* if no memory, quit */
	if(extended)
	{
#if STATIC
		adjust_bpb(args[0], args[1], args[2]);
		staticdisk = is_static_disk();
		if(!staticdisk)
#endif
		{
			args[0] = allocate_extended_memory(args[0]);
		}
	}
	else
	{

		total_conventional = (memtop - *endseg) / 0x40;
		if(args[0] > total_conventional)		// not enough memory
			args[0] = 0;
	}

	if(args[0] == 0)
	{
		if(extended)
			bputs("Error allocating memory!\n");
		else
		{
			bputs("Do not request more than the ");
			bputi(total_conventional);
			bputs("K available ");
			bputs("conventional");
			bputs(" memory.\n");
		}
		bputs("VDISK not installed.\n");
		return 0;
	}

	adjust_bpb(args[0], args[1], args[2]);

	/* if the disk is not EXACTLY the same, initialize it */
#if STATIC
	if(!extended)
		staticdisk = is_static_disk();
	if(!staticdisk)
#endif
	{
		/* zero the root DIR and FAT */
		sectors = std_bpb.Reserved + std_bpb.NumbFats * std_bpb.SecPerFAT;
		sectors += (std_bpb.MaxDirs * 0x20) / std_bpb.BytesPerSector;
		for(i=0; i<sectors; ++i)
			zero_sector(i);

		/* initialize boot sector Remember: don't overwrite */
		/* "extended memory used" at offset 0x3fL */
		memwrite(0x00L, sizeof(Disk_Intro), (void far *)Disk_Intro);
		memwrite(0x0bL, 18, (void far *)(&std_bpb));
		memwrite(0x27L, 4, (void far *)(&std_serial));
 		memwrite(0x2bL, 11, (void far *)(&std_volume));
 		memwrite(0x36L, 9, (void far *)(&std_fattype));
		memwrite((long)(std_bpb.BytesPerSector-2), 2, (void far *)(&Boot_ID));

		/* initialize FAT(s) */
		sectors = std_bpb.Reserved;
		for(i=0; i<std_bpb.NumbFats; ++i)
		{
#if BUILTIN
			start = _lmulu(sectors, std_bpb.BytesPerSector);
#else
			start = (long)sectors * std_bpb.BytesPerSector;
#endif
			memwrite(start, sizeof(Init_FAT), (void far *)Init_FAT);
			sectors += std_bpb.SecPerFAT;
		}

		/* write volume label entry in the root*/
#if BUILTIN
		start = _lmulu(sectors, std_bpb.BytesPerSector);
#else
		start = (long)sectors * std_bpb.BytesPerSector;
#endif
		memwrite(start, sizeof(label), (void far *)(&label));
	}

	/* Report the VDISK size, etc */
	bputs("\n   Installed as Drive:  ");
	bputc((char)(drv + 'A'));
	bputc(':');
#if STATIC
	if(staticdisk) bputs(" (static VDISK)");
#endif
	bputs("\n   Disk Size:           "); bputi(args[0]); bputs(" KB");
	if(extended)
		bputs(using_xms ? " in XMS memory" : " in Extended memory");
	if(size_adjusted) bputs(" (adjusted)");
	bputs("\n   Sector Size:         "); bputi(std_bpb.BytesPerSector);
	bputs("\n   Directory Entries:   "); bputi(std_bpb.MaxDirs);
	bputs("\n");

 	/* specify the conventional memory used in paras */
#if !BUILTIN
	if(!extended)
		*endseg = disk_seg + (std_bpb.BytesPerSector/16) * std_bpb.TotalSecs;
#endif

	/* tell DOS about the BPB */
	memmove(bpb, &std_bpb, sizeof(std_bpb));

#if TWODRIVES
	memmove(bpb+1, &std_bpb, sizeof(std_bpb));
	return 2;
#else
	return 1;
#endif
}


/*
   This routine determines if the disk has changed.  Update
   the Bios Parameter Block structure pointed to by "bpb"
   if the disk may have changed.  The definition for BPB
   may be found in MEMDISK.H.

   Returns -1 if Device HAS changed
   Returns  0 if Device MAY have changed
   Returns  1 if Device has NOT changed
*/
char ischanged = CHANGED_NO;	// allow other functions to set changed
#ifdef __TURBOC__
#pragma argsused
#endif
int memchanged(struct BPB *bpb)
{

#if TWODRIVES
	// "TWODRIVES" makes 2 drive letters point to the same
	// physical RAM disk.
	// Keep the 2 identical drives in sync.
	ischanged = CHANGED_MAYBE;
#endif

	return ischanged;
}


/*
   This function reads data (usually 1 or more sectors)
   from the ROM/RAM/Flash disk into the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            read should start.
   len      The length of the read in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memread(long offset, unsigned len, char far *buffer)
{
	unsigned seg, off, ret = 1;

	if(extended)
		ret = ext_memread(offset, buffer, len);
	else
	{
#if BUILTIN
		seg = (unsigned)_lshru(offset, 4) + disk_seg;
#else
		seg = (unsigned)(offset >> 4) + disk_seg;
#endif
		off = (unsigned)offset & 0xF;
		far_memcpy(buffer, (char far *)MK_FP(seg,off),  len);
	}
	return (ret);
}


/*
   This function writes data (usually 1 or more sectors)
   to the ROM/RAM/Flash disk from the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            write should start.
   len      The length of the write in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memwrite(long offset, unsigned len, char far *buffer)
{
	unsigned seg, off, ret = 1;

	if(extended)
		ret = ext_memwrite(offset, buffer, len);
	else
	{
#if BUILTIN
		seg = (unsigned)_lshru(offset, 4) + disk_seg;
#else
		seg = (unsigned)(offset >> 4) + disk_seg;
#endif
		off = (unsigned)offset & 0xF;
		far_memcpy((char far *)MK_FP(seg, off), buffer, len);
	}
	return (ret);
}

#if STATIC
int is_static_disk(void)
{
	struct BPB tmp_bpb;
	char tmp_intro[12];
	char tmp_fat[4];
	int cmp1, cmp2, cmp3, cmp4;
	int found_static = TRUE;

	/* don't allow static disks */
	if(!static_ok)
		return FALSE;

	if(extended)
	{
		/* get some parts off a possible "static" disk */
		ext_memread(0L, (void far *)tmp_intro, sizeof(Disk_Intro));
		ext_memread(11L, (void far *)(&tmp_bpb), 18);
		ext_memread((long)(std_bpb.Reserved * std_bpb.BytesPerSector),
			(void far *)tmp_fat, 3);
	}
	else
	{
		/* get some parts off a possible "static" disk */
		memread(0L, sizeof(Disk_Intro), (void far *)tmp_intro);
		memread(11L, 18, (void far *)(&tmp_bpb));
		memread((long)(std_bpb.Reserved * std_bpb.BytesPerSector), 3,
			(void far *)tmp_fat);
	}
#if BUILTIN
	/* compare them to what user now wants... */
	/* note: built-in compare is a TRUE/FALSE not -1,0,1 */
	cmp1 = !memcmp((char *)(&tmp_bpb),  (char *)(&std_bpb),   18);
	cmp2 = !memcmp(tmp_intro, Disk_Intro, sizeof(Disk_Intro));
	cmp3 = !memcmp(tmp_fat,   Init_FAT,   3);
#else
	/* compare them to what user now wants... */
	cmp1 = memcmp((char *)(&tmp_bpb), (char *)(&std_bpb), 18);
	cmp2 = memcmp(tmp_intro, Disk_Intro, sizeof(Disk_Intro));
	cmp3 = memcmp(tmp_fat, Init_FAT, 3);
#endif

#if DEBUG
	bputs(" cmp1 ");  bputc(cmp1 ? '1' : '0');
	bputs(" cmp2 ");  bputc(cmp2 ? '1' : '0');
	bputs(" cmp3 ");  bputc(cmp3 ? '1' : '0');
#endif

	/* were they the same? no, not a static disk! */
	if(cmp1 || cmp2 || cmp3)
		found_static = FALSE;

	return found_static;
}
#endif

/*
	Parse the command line arguments.  The CONFIG.SYS line
	may look the line below.  The cmdargs will be just the
	arguments portion:

	DEVICE = VDISK.SYS [disk_sizeK [sec_size [root_entries]]] [/e] [/s]
                      ^
	The cmdargs begin after the name "VDISK.SYS".
*/
void do_args(char far *cmdargs)
{
	char c;

	/* look for command line arguments */
	while(*cmdargs>=' ' || *cmdargs == '\t')
	{
		/* switches */
		if(*cmdargs == '/')
		{
			++cmdargs;
			c = *cmdargs;
			if(c>='a' && c<='z')		/* toupper */
				c = c-'a'+'A';
			switch(c)
			{
#if TSRDEV
				/* display help */
				case '?':
					opt_showhelp = TRUE;
					break;

				/* unload the driver */
				case 'U':
					opt_unload = TRUE;
					break;
#endif
				/* RAM disk to be placed in extended memory */
				case 'E':
					extended = TRUE;
					break;

#if STATIC
				/* RAM disk should remain static, if possible */
				case 'S':
					static_ok = TRUE;
					if(cmdargs[1] == '-')
						static_ok = FALSE;
					break;
#endif

				default:
					bputs("\nInvalid Option: /");
					bputc(*cmdargs);
					bputc('\n');
			}

			/* skip to next white space or switch */
			while(*cmdargs>' ' && *cmdargs != '/')
				++cmdargs;
		}

		/* skip white space */
		else if(*cmdargs==' ' || *cmdargs=='\t')
			++cmdargs;

		/* generic argument (a number, like disk size, etc...) */
		else
		{
			if(arg_count < 3)
				cmdargs = getcount(cmdargs, args + arg_count++);
			else
			{
				bputs("Too many arguments ");
				++cmdargs;
			}
		}
	} /* end of while */
}

/*
	Get a number from the command line.
*/
char far *getcount(char far * str, unsigned *count)
{
	*count = 0;
	while(*str)
	{
		if(*str < '0' || *str > '9')
			break;
		*count= (*count * 10) + *str - '0';
		++str;
	}
	return str;
}

/*
 * Adjust bpb uses user Size Sector Size and # of Dir Entries
 * to adjust required fields in the BPB.
 */
void adjust_bpb(unsigned sizeK, unsigned sec_size, unsigned dir_entries)
{
#if !TESTING
	extern unsigned memtop;
#endif
	unsigned DirPerSec, BytesPerFAT;

	/* determine sector size */
	if(sec_size!=128 && sec_size!=256 && sec_size!=512 && sec_size!=1024)
		sec_size = DEF_SEC_SIZE;
	std_bpb.BytesPerSector = sec_size;

	if(sizeK)
	{
		if(sizeK < 16)
		{
			size_adjusted = 1;	/* disk size was adjusted */
			sizeK = 16;
		}
		std_bpb.TotalSecs = sizeK * (1024/std_bpb.BytesPerSector);

		/* we are using a 12-bit FAT, make sure disk will fit */
		while(std_bpb.TotalSecs/std_bpb.SecPerCluster > 0xFF0)
			std_bpb.SecPerCluster *= 2;
	}

	/* determine # of dir entries, and Root Dir size */
	if(dir_entries)
	{
		if(dir_entries < 2 || dir_entries > 1024)
			dir_entries = DEF_DIR_ENTRIES;
		DirPerSec = std_bpb.BytesPerSector / 32;
		std_bpb.MaxDirs = ((dir_entries+DirPerSec-1) / DirPerSec) * DirPerSec;
	}

	/* determine the FAT size */
	BytesPerFAT = ((std_bpb.TotalSecs/std_bpb.SecPerCluster) * 3) / 2;
	std_bpb.SecPerFAT = (BytesPerFAT + std_bpb.BytesPerSector -1) / std_bpb.BytesPerSector;
	if(std_bpb.SecPerFAT == 0)
		++std_bpb.SecPerFAT;
}

/*
	Zero bytes on the disk.  Used only on bootup
*/
void zero_bytes(long offset, unsigned count)
{
	char zero[16];
	unsigned i;

	for(i=0; i<sizeof(zero); ++i)
		zero[i] = 0;

   for(i = 0; i < count; i += sizeof(zero))
		memwrite(offset+i, sizeof(zero), (void far *)(&zero));
}

/*
	Zero the numbered sector.
*/
void zero_sector(unsigned sector)
{
#if BUILTIN
	// can't start past 64K with built-in
	zero_bytes((long)(sector*std_bpb.BytesPerSector), std_bpb.BytesPerSector);
#else
	zero_bytes((long)sector*std_bpb.BytesPerSector, std_bpb.BytesPerSector);
#endif
}

/*
 * Build a char string for a given number for any radix
 */
void itoa(unsigned numb, char *buffer, unsigned radix)
{
	static char hex[] = "0123456789ABCDEF";
	char local[5];
	unsigned remainder, i=0;

	radix %= 16;	/* don't allow a radix > 16 */
	do
	{
		remainder = numb % radix;
		numb /= radix;
		local[i++] = hex[remainder];
	} while(numb);

	while(i)
		*buffer++ = local[--i];
	*buffer = '\0';
}

/*
	We've defined our own far memory copy.
*/
void far_memcpy(char far *dst, char far *src, unsigned count)
{
#if 0
	/* the C way */
	unsigned i;
	for(i=0; i<count; ++i)
		dst[i] = src[i];

	/* the fast way */
#else
	ASM	push	DS
	ASM	push	SI
	ASM	push	DI
	ASM	mov	cx, count
	ASM	lds	si, src
	ASM	les	di, dst
	ASM	cld
	ASM	rep	movsb
	ASM	pop	DI
	ASM	pop	SI
	ASM	pop	DS
#endif
}

/*
	Display routine.
*/
void bputs(char *s)
{
	while(*s)
		bputc(*s++);
}

/*
	Display a character.  Is smart about the C newline.
*/
void bputc(char c)
{
	if(c == '\n')
		bputc('\r');
	ASM mov AL, c
	ASM mov AH, 0EH
	ASM mov BX, 0
	ASM int 10H
}

/*
 *  Displays a 16-bit decimal number.
 */
void bputi(unsigned val)
{
	char buffer[15];
	itoa(val, buffer, 10);
	bputs(buffer);
}

/*
	Displays a 16-bit hex number in 0000 form.
*/
void bputh(unsigned val)
{
	char buff[5];
	int i;
	char c;

	for(i=3; i>=0; --i)
	{
		c = val % 0x10;
		c = (c >= 10 ? c-10+'A' : c+'0');
		buff[i] = c;
		val /= 0x10;
	}
	buff[4] = 0;
	bputs(buff);
}

/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |            EXTENDED MEMORY ROUTINES                |
	 |                                                    |
	[]----------------------------------------------------[]
*/

unsigned extended_used;
uchar jmpfar[5] = { 0xEA };
unsigned far * ptr19;


/*
 *  Allocates the specified # of K bytes of extended memory
 *  either through XMS or directly.
 *
 *  Returns the size found or 0 on failure.
 */
#ifdef __TURBOC__
#pragma argsused
#endif
unsigned allocate_extended_memory(unsigned sizeK)
{
	int first;
	unsigned long tmp;

	/* use the XMS driver if possible */
	if(himem_installed())
	{
		using_xms = TRUE;
		total_extended = total_xms();
		ext_adr = XMS_SIGNATURE;		/* mark as XMS */
	}
	else
		total_extended = get_ext_mem_size();

#if DEBUG
	bputs("XMS "); bputh(using_xms);
	bputs("Total "); bputh(total_extended); bputc('\n');
#endif

	if(total_extended <= 1)
	{
nomem:
		return(0);
	}

	/* if XMS is not doing it, we must do our OWN memory management */
	if(!using_xms)
	{
		first = first_vdisk();
		if(first)
		{
			// We have to access the vdisk header @ 10:0000 to record allocs.
			// ext_adr is the global base address used by ext_memread/write
			// and may be another value.
			//
			tmp = ext_adr;
			ext_adr = 0x100000UL;
			zero_bytes(0L, 32);  // we will use these 2 paras to track allocs
			ext_adr = tmp;
			total_extended--;		// we use up on Kbyte to save the header
		}
		extended_used = get_used();
		if(extended_used >= total_extended)
			goto nomem;
		total_extended -= extended_used;

#if DEBUG
	bputs("Used1 "); bputh(extended_used); bputc(' ');
#endif
	}

	/* if not enough memory, get what we can */
	if(total_extended < sizeK)
	{
		sizeK = total_extended;
		size_adjusted = 1;		/* mark disk size as adjusted */
	}

	/* allocate the memory */
	if(using_xms)
	{
		xms_handle = allocate_xms(sizeK);
		if(!xms_handle)
			goto nomem;
	}

	/* if XMS is not doing it, we must do our OWN memory management */
	else
	{
		write_used(sizeK + extended_used);
#if BUILTIN
		ext_adr += _lmulu((long)extended_used, 1024);
#else
		ext_adr += (unsigned long) extended_used * 1024L;
#endif

#if DEBUG
		bputs("Used2 "); bputh(get_used()); bputc(' ');
#endif
	}

#if DEBUG
	bputs("ext_adr 0x");
	bputh((unsigned)(ext_adr>>16));
	bputh((unsigned)ext_adr);
	bputs("L\n");
#endif

	return(sizeK);
}


/*
 *  Determine if this is the first VDISK to be loaded.
 *  This is important for determining how much extended
 *  memory is used.
 */
int first_vdisk(void)
{

	unsigned seg;
	char far * vdisk;
	int i;
	int first = 0;

	/* get old INT 19H vector */
	ptr19 = (unsigned far *)MK_FP(0,0x19*4);
	seg = *(ptr19+1);

	/* determine if we are first VDISK, look for old one on INT 19H */
	if(seg > 0x8000)
		first = 1;
	else
	{
		// verify the signature
		vdisk = MK_FP(seg, (unsigned)(Disk_Intro+3));
		for(i=0; i<5; ++i)
			if(vdisk[i] != Disk_Intro[i+3])
				first = 1;
	}

	// set INT 19H to point to us
	if(first)
	{
		ASM	cli
		*(unsigned long *)(jmpfar+1) = *(unsigned long far *)(ptr19);
		*(unsigned long far *)(ptr19) = (long)((void far *)jmpfar);
		ASM	sti
	}

#if DEBUG
	bputs("first "); bputc('0'+first);
#endif

	return first;
}


/*
 * Extended memory Read/Write
 * Performed using BIOS Interrupt 15 subfunction 88
 */
struct gdt
{
	unsigned seg_limit;
	unsigned char addr[3];
	unsigned char data_access;
	unsigned reserved;
};

struct bios_ext_move
{
	unsigned char zeros[8];
	unsigned char bios_stuff[8];
	struct gdt src_gdt;
	struct gdt dst_gdt;
	unsigned char bios_stuff3[19];
};

struct xms_block
{
	ulong    length;
	unsigned src_handle;
	ulong    src_offset;
	unsigned dst_handle;
	ulong    dst_offset;
};

#define USED_OFF 0x3fL

/*
 * Get used gets the word at 10001E.  This word is the
 * number of K bytes of extended memory used.
 * The bit (0x400) is set by MS-DOS 3.30 VDISK, unknown why!
 */
unsigned get_used(void)
{
	unsigned size;
	unsigned long tmp;


	// we have to access the vdisk header @ 10:0000.
	// ext_adr is the global base address used by ext_memread/write
	// and may be another value
	//
	tmp = ext_adr;
	ext_adr = 0x100000UL;
	ext_memread(USED_OFF, &size, 2);
	ext_adr = tmp;

	/*	size = size & ~0x400; */
	return(size);
}

/*
 * Write used sets the word at 10001E.  This word is the
 * number of K bytes of extended memory used.
 * The bit (0x400) is set by MS-DOS 3.30 VDISK, unknown why!
 */
void write_used(unsigned size)
{
	unsigned long tmp;

	/*	size |= 0x400;	*/

	// we have to access the vdisk header @ 10:0000.
	// ext_adr is the global base address used by ext_memread/write
	// and may be another value
	//
	tmp = ext_adr;
	ext_adr = 0x100000UL;
	ext_memwrite(USED_OFF, &size, 2);
	ext_adr = tmp;
}

/*
	Convert a far pointer into a long.
*/
unsigned long fp_to_long(void far *fp)
{
	unsigned long l;
#if BUILTIN
	l = _lshlu((unsigned long)FP_SEG(fp),4);
#else
	l = ((unsigned long)FP_SEG(fp)) << 4L;
#endif
	l += FP_OFF(fp);
	return (l);
}

int ext_memread(long offset, void far *buffer, unsigned count)
{
	return ext_move_data(fp_to_long(buffer), ext_adr+offset, count);
}

int ext_memwrite(long offset, void far *buffer, unsigned count)
{
	return ext_move_data(ext_adr+offset, fp_to_long(buffer), count);
}

/*
 * Extended Move Data uses BIOS interrupt 15 subfunction 87 to
 * move data between extended memory and real mode memory.
 */
#ifdef __TURBOC__
#pragma argsused
#endif
int ext_move_data(unsigned long dst, unsigned long src, unsigned count)
{
#if BUILTIN
	return 0;
#else
	static struct bios_ext_move mv_blk;
	static struct xms_block x_blk;

	unsigned char r_ah;
	unsigned cx_tmp;
	int worked = 0;
	unsigned blk_addr;


	if(using_xms)
	{
		// XMS transfers only words
		if(count & 1)
			++count;
		x_blk.length = count;
		x_blk.src_handle = x_blk.dst_handle = 0;

		/* set up source for move */
		if(src & XMS_SIGNATURE)
		{
			x_blk.src_handle = xms_handle;
			x_blk.src_offset = src & (~XMS_SIGNATURE);
		}
		else
			x_blk.src_offset = ((src&0xFFFF0L) << 12) + (src&0xF);

		/* set up destination */
		if(dst & XMS_SIGNATURE)
		{
			x_blk.dst_handle = xms_handle;
			x_blk.dst_offset = dst & (~XMS_SIGNATURE);
		}
		else
			x_blk.dst_offset = ((dst&0xFFFF0L) << 12) + (dst&0xF);

		worked = xms_move(&x_blk);
	}
	else
	{
		/* Zero the Global Descriptor Table */
		memset(&mv_blk, 0, sizeof(mv_blk));

		/* Build a Global Descriptor Table */
		*(unsigned *)(&mv_blk.src_gdt.addr[0]) = (unsigned)src;
		mv_blk.src_gdt.addr[2] = src >> 16;
		mv_blk.src_gdt.data_access = 0x93;
		mv_blk.src_gdt.seg_limit = 0xFE00;

		*(unsigned *) (&mv_blk.dst_gdt.addr[0]) = (unsigned)dst;
		mv_blk.dst_gdt.addr[2] = dst >> 16;
		mv_blk.dst_gdt.data_access = 0x93;
		mv_blk.dst_gdt.seg_limit = 0xFE00;

		/* we can only move a number of words */
		cx_tmp = (count+1) / 2;
		blk_addr = (unsigned)(&mv_blk);

		ASM	push	ES
		ASM	mov	AX, DS
		ASM	mov	ES, AX
		ASM	mov	CX, cx_tmp
		ASM	mov	SI, blk_addr
		ASM	mov	AX, 8700H
		ASM	int	15H
		ASM	pop	ES
		ASM	mov	r_ah, AH
		if(r_ah == 0)
			worked = 1;
	}

	if(worked)
		return 1;
	else
	{
#if DEBUG
		bputs(" EXT MEM Error!\n");
#endif
		memerr = DEVERR_GENERALFAIL;
		return 0;
	}
#endif
}


/*
 * get Extended Memory Size returns the number of K bytes
 * of extended memory.
 */
int get_ext_mem_size(void)
{
	unsigned r_ax, flags;

	ASM mov	AX, 8800H
	ASM int 15H
	ASM mov r_ax, AX
	ASM pushf
	ASM pop flags
	if(flags & 1)
		return (0);
	return (r_ax);
}

/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |                 XMS ROUTINES                       |
	 |                                                    |
	[]----------------------------------------------------[]
*/

typedef void (far * farfuncptr)(void);
farfuncptr hma_handler;

/*
	Is an XMS driver installed?
	Doesn't know whether the HMA will be used or not, so
	don't allocate that space!
*/
int himem_installed(void)
{
	unsigned char r_al;
	unsigned r_seg, r_off;

	/* is an XMS driver installed? */
	ASM mov AX, 4300H
	ASM int 2FH
	ASM mov r_al, AL

	/* get address of HMA driver */
	if(r_al == 0x80)
	{
		ASM mov 	AX, 4310H
		ASM int 	2FH
		ASM mov	r_off, BX
		ASM mov	r_seg, ES
		if(!r_seg || !r_off)
			r_al = 0;
		else
			hma_handler = (farfuncptr)MK_FP(r_seg, r_off);
#if DEBUG
		bputs("XMS handler "); bputh(r_seg); bputc(':'); bputh(r_off);
#endif
	}

	return (r_al == 0x80);
}

/*
	Return total XMS memory, in K bytes.
*/
unsigned total_xms(void)
{
	unsigned r_ax;

	ASM mov AX, 800H
	(*hma_handler)();
	ASM mov r_ax, AX

	return r_ax;
}

/*
	Returns the handle, or 0 if failed.
*/
unsigned allocate_xms(unsigned sizeK)
{
	unsigned handle;
	unsigned char r_bl;

	/* allocate the XMS memory */
	ASM mov DX, sizeK
	ASM mov AX, 900H
	(*hma_handler)();
	ASM mov handle, DX
	ASM mov r_bl, BL
	if(r_bl)
		handle = 0;
	else
	{
		/* lock the memory so it can't move */
		ASM mov DX, handle
		ASM mov AX, 0C00H
		(*hma_handler)();
	}
	return handle;
}

void free_xms(unsigned handle)
{
	// unlock the locked block
	asm mov ax, 0D00h
	asm mov dx, handle
	(*hma_handler)();

	// now free the unlocked block
	asm mov ah, 0Ah
	asm mov dx, handle
	(*hma_handler)();
}

/*
	Make a copy of memory from one place to another using the
	XMS driver.
*/
int xms_move(struct xms_block * xp)
{
	auto int worked;
	auto int err;

	/* copy a block */
	ASM mov AX, 0B00H
	ASM mov SI, xp
	(*hma_handler)();
	ASM mov worked, AX
	ASM xor BH, BH
	ASM mov err, BX

	if(!worked)
	{
		bputs("XMSmv_err ");
		bputh(err);
	}

	// enable the HMA
	ASM	mov	AX, 300H
	(*hma_handler)();

	return worked;
}

/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |             	IOCTL SECTION               			|
	 |                                                    |
	 |  This optional function is only needed if you must |
	 |  work with the DOS Format, or require some other   |
	 |  application/device communication.                 |
	 |                                                    |
	[]----------------------------------------------------[]
*/

#if SUITE
#define SIZELOGBUFF	(4*1024)
char logbuff[SIZELOGBUFF];
int logpos;
#endif

#if IOCTL

struct DeviceParams // see MS-DOS Programmers Reference page 38
{
	uchar 	dpSpecFunc;
	uchar 	dpDevType;
	unsigned dpDevAttr;
	unsigned dpCylinders;
	uchar 	dpMediaType;
	struct   BPB bpb;
};

struct SpecialFunc	// see MS-DOS Programmers Reference page 58
{
	uchar sfFunction;
	uchar sfAccessFlg;
};
#define sfFUNC_NUM0	0
#define sfACCESS_OK	1

/*
   (Optional)

	Used to determine what IOCTL calls are supported
	Returns TRUE if the category is supported by memioctl
	FALSE, if not.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
bool memqueryioctl(unsigned category)
{
	switch(category)
	{
 #if SUITE
		case 0x8D0:
		case 0x8D1:
 #endif

		case 0x840:	// set device parmaters
		case 0x842:	// format track
		case 0x846:	// set media ID
		case 0x860: // get device parameters
		case 0x867:
		case 0x868: // Sense media type (act like MS RAMDISK (v6.22): fail)
			return TRUE;

		default:
			return FALSE;
	}
}


#if (SUITE >= 2)
lpFR lpFailureRecord = (lpFR)0L;

/*
	This function is used primarily for debugging the Kernel.
	It allows an App to call into the device to specify that
	a specific DOS call to the device will fail.

	Returns -1 if the device should handle the call normally.

	It returns the device error code {0,1,2,...,14h} that the
	test app wants returned if the current call is one that the
	test app wants to fail. This code must also modify the device
	request header if any fields need to be changed there.

*/
int PASCAL FakingError(char far *lpDevHdr)
{
	int iStatus = -1;
	int iDevCmd, iCmdCode;
	char cErrCode;

	// if no current Failure Record, just do the normal device things
	//
	if(lpFailureRecord == (lpFR)0L)
		return iStatus;

	// If we are told to fail a write, we will also fail any reads to
	// the same region of the disk. Other than that, if this is
	// not the command that the test app wants to have fail, just
	// handle it in the normal way...
	//
	iDevCmd  = *(lpDevHdr+REQHDR_CMD);
	iCmdCode = lpFailureRecord->cCmdCode;
	if(iDevCmd != lpFailureRecord->cCmdCode)
	{
   	do
		{
			if(iCmdCode == DEV_WRITE)
			{
				if((iDevCmd == DEV_READ)   ||
			   	(iDevCmd == DEV_NDREAD))
				{
					cErrCode = DEVERR_READFAULT;
					break;
				}
			   if((iDevCmd == DEV_VWRITE) ||
					(iDevCmd == DEV_WRITE))
				{
					cErrCode = DEVERR_WRITEFAULT;
					break;
				}
			}
	 		return iStatus;
		} while(0);
	}
	else
	{
		cErrCode = lpFailureRecord->cErrCode;
	}

	// This is were we handle special cases...
	// If the function doesn't qualify (writing to FAT when looking for
	// a file write to fail, for example) then just return -1 and the
	// driver will handle the call in the normal way. Break to force
	// the device call to fail after doing any qualifying and additional
	// data manipulation you may need to do...
	//
	switch(*(lpDevHdr+REQHDR_CMD))
	{

	// Qualify the write to a specific part of the disk and force the
	// specified failure if we are there.
	//
	case DEV_READ:
	case DEV_NDREAD:
	case DEV_WRITE:
	case DEV_VWRITE:
		if((starting_sector(lpDevHdr) < first_sector(lpFailureRecord)) ||
			(starting_sector(lpDevHdr)+sector_count(lpDevHdr)-1 > last_sector(lpFailureRecord)) ||
			(error_count(lpFailureRecord) == 0)
		  )
			return iStatus;

		sector_count(lpDevHdr) = sectors_done(lpFailureRecord);
		break;

	default:
		return iStatus;
	}

	// These two steps will be done any time that we want to fake an error.
	// We set the lpFailureRecord to NULL so that we know we've already
	// done what the test app asked us to.
	//
	iStatus = cErrCode;
	if(--error_count(lpFailureRecord) == 0)
		lpFailureRecord = (lpFR)0L;

	return iStatus;
}

#endif


/*
   (Optional)

	This function supports the DOS IOCTL interface.  This
	allows the VDISK to be formatted with the standard DOS
	FORMAT program.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
bool memioctl(unsigned category, char far *buffer)
{

 #if SUITE

	// allow user to retrieve the log buffer
	if(category == 0x8D0)
	{
		// pass far ptr to buffer to user program
		*(unsigned far *)(buffer)   = logpos;
		*(unsigned far *)(buffer+2) = (unsigned)logbuff;
		*(unsigned far *)(buffer+4) = _DS;

		// reset the buffer
		logpos = 0;
		return 1;	// worked
	}

	// allow user program to "change" the disk at will
	else if(category == 0x8D1)
	{
		ischanged = CHANGED_YES;
		return 1;	// worked
	}
 #endif

 #if (SUITE >= 2)
	// Allow a test app to specify that a specific type of call will fail.
	// We'll figure out what exactly to do in FakingError() when it's called.
	//
	else if(category == 0x8D2)
	{
		lpFailureRecord = (lpFR)buffer;
		return 1;
	}
 #endif


	switch(category)
	{
		// These functions are stubbed since there is no need to do anything
		//	else when formatting a RAM disk. Add additional functionality here
		// as might be required.
		//
		case 0x840:	// set device parmaters
		case 0x842:	// format track
		case 0x846:	// set media ID
			return(TRUE);

		case 0x860: // get device parameters
		{
			struct DeviceParams far* lpDevParams =
				(struct DeviceParams far*)(buffer);

			// FORMAT will take care of all neccessary re-initializing of the
			// VDISK by writing directly to the FAT and ROOT DIR sectors.
			// So we just act like a non-removable floppy...
			//
			lpDevParams->dpDevType = 7; // insist this is a floppy
			lpDevParams->dpDevAttr = 1; // insist this is not removable

			// set # of cylinders to a small number, as the format
			// call for each cylinder does nothing in VDISK
			// FORMAT does not appear to use this
			lpDevParams->dpCylinders = 1;

			// user also wants a copy of the BPB
			far_memcpy((char far *)(&(lpDevParams->bpb)),
				(char far *)(&std_bpb), sizeof(std_bpb));
			return(TRUE);
		}

		// get access flag
		// sub-sub-function 0 returns 0 if disk is not formatted else 1 if it is
		// VDISK is always formatted.
		//
		case 0x867:
		{
			struct SpecialFunc far * lpSpecFuncs =
				(struct SpecialFunc far*)(buffer);

			// documentation says special function should ALWAYS be 0
			// see pg 38 DOS Prog Ref
			ASSERT(lpSpecFuncs->sfFunction == sfFUNC__NUM0);

			lpSpecFuncs->sfAccessFlg = sfACCESS_OK;
			return(TRUE);
		}

		case 0x868: // Sense media type (act like MS RAMDISK (v6.22): fail)
		default:
			break;
	}

	memerr = DEVERR_BADCMD;
	return 0;	// failed
}


#if SUITE
struct dev_req_hdr /*()*/
{
   uchar      length;        /* 00 length of the command    */
   uchar      unit;          /* 01 unit # for disk access   */
   uchar      command;       /* 02 the command itself       */
   unsigned   status;        /* 03 returned status          */
   uchar      reserved[8];   /* 05 */
   uchar      media;         /* 0D media ID for disks       */
   void far * trans;         /* 0E data transfer buffer     */
   unsigned   count;         /* 12 length of data           */
   unsigned   start;         /* 14 starting sector for data */
   uchar      drive;         /* 16 starting drive #         */
   uchar      res2[3];       /* 17 */
   ulong      start32;       /* 1A long starting sector     */
};

void logcmd(char cmd)
{
	extern struct dev_req_hdr far * ptrsav;
	int len;
	char far * ptr;

	// put the cmd in character form
	if(logpos < SIZELOGBUFF)
		logbuff[logpos++] = cmd;

	// put the entire request header in the buffer
	len = ptrsav->length;
	ptr = (char far *)ptrsav;

	if(!len)
	{
		if(logpos < SIZELOGBUFF)
			logbuff[logpos++] = 0;
	}

	else while(len && logpos < SIZELOGBUFF)
	{
		logbuff[logpos++] = *ptr++;
		--len;
	}
}
#endif	// SUITE

#endif	// IOCTL


/*
	[]----------------------------------------------------[]
	 |                                                    |
	 |             TSR-STYLE DEVICE SECTION               |
	 |                                                    |
	 |  These functions are only required if the MEMDISK  |
	 |  driver is set up to install VDISK from the        |
	 |  command-line.  See the description of MEMDISK in  |
	 |  the ROM-DOS Developer's Kit Manual.               |
	 |                                                    |
	[]----------------------------------------------------[]

*/

#if TSRDEV
void showhelp(void)
{
	bputs("\n\nDEVICE = VDISK.EXE [/?] [/u] [disk_sizeK [sec_size [root_entries]]] [/e] [/s]\n");
}

/*
   (Optional)

	If VDISK is loaded as a TSR, it can also be unloaded
 	thereby freeing the memory its drive(s) occupied.
*/
void memuninit(void)
{
	if(using_xms)
	{
		free_xms(xms_handle);		// free XMS if we used any
	}
	else if(extended)					// else if using VDISK allocations
	{
		// and if we are at the front of the int 19h chain
		//
		if(*(unsigned long far *)(ptr19) == (long)((void far *)jmpfar))
		{
			// then we can deallocate our VDISK and uninstall our int 19h
			// handler
			//
			write_used(extended_used - args[0]);
			ASM	cli
			*(unsigned long far *)(ptr19) = *(unsigned long *)(jmpfar+1);
			ASM	sti
		}
		else 							// we cannot safely unload!
		{;}
	}

	bputs("complete.\n");
}
#endif

#if TESTING

#include <string.h>

#define	NUM_TESTS	5

#pragma argsused
void main(int argc, char **argv)
{
	struct BPB bpb;
	unsigned char buffer[1025]={0};
	unsigned i,j,k, nSectors;
	unsigned long tmp;
	unsigned endseg;


	// This unit test is only for eXtended Memory reads/writes.
	// add any others as needed.
	//
	for(i=1; i < argc; i++)
		strcat(buffer, argv[i]);

	meminit(&bpb, buffer, &endseg, 3);

	bputs("\n\rPress any key to start test...");
	asm {
		xor	ax, ax
		int	16h
	}

	tmp =  (unsigned long)(args[0]) * 1024L;
	nSectors = (unsigned)(tmp / 512L);

	for(k=0; k < NUM_TESTS; k++)
	{
		bputs("\n\r");
		for(i = 0; i < nSectors; i++)
		{
			for(j = 0; j < 512; j++)
				buffer[j] = (unsigned char)(i & 0xFF);
			memwrite((unsigned long)(i)*512L, 512, buffer);
			bputs(".");
		}
		for(i = 0; i < nSectors; i++)
		{ int ok = 1;

			memread((unsigned long)(i)*512L, 512, buffer);
			for(j = 0; j < 512; j++)
			{
				if(buffer[j] != (unsigned char)(i & 0xFF))
				{
					bputs("Buffers failed to compare! Test: "); bputi(k);
					bputs("  Sector: ");								  bputi(i);
					bputs("  Byte: ");								  bputi(j);
					bputs("  Read: ");                          bputi(buffer[j]);
					bputs("\n\r");
					ok = 0;
					break;
				}
			}
			if(ok) bputs("=");
		}
	}
}
#endif
