//
//   $Id: DDTDISK.H 1.5 2002/10/17 14:59:24Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	DDTDISK.H  created  10/15/93 by Drew Gislason

   Copyright (c) 1993-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This is a debugging/testing tool based on the old VDISK.
	It DOES NOT work with the new MEMDISK/VDISK code.

	Include this in any implementations of a generic disk
	driver.  An assembly version of this file is called
	DDTDISK.INC.

	While DDTDISK implies this only works with memory
	disk drivers, it works fine as a base for ANY kind of
	disk driver.

	05-21-98 (DE) Added additional device error codes and some
				device request hearder field offsets. The latter is
				used by the FakingError() debugging function.

**********************************************************/

/* used to call pascal style functions */
#define PASCAL pascal

/* new types */
typedef int bool;
#define TRUE  1
#define FALSE 0
#define true  TRUE
#define false FALSE

typedef unsigned char uchar;
typedef unsigned long ulong;


/*
   BIOS PARAMETER BLOCK (BPB)

   Description: Every disk has this structure in the boot
   sector (sector 0).  The BPB describes the disk to DOS.
   DOS gets the BPB information both upon initialization,
   and any time the disk may have been removed and replaced
   with some other disk.

	Any decent DOS book describes the BPB.
*/
struct BPB /*()*/
{
	unsigned  BytesPerSector;	/* 00H calculated */
	uchar     SecPerCluster;	/* 02H calculated */
	unsigned  Reserved;			/* 03H */
	uchar     NumbFats;			/* 05H */
	unsigned  MaxDirs;			/* 06H calculated */
	unsigned  TotalSecs;			/* 08H calculated */
	uchar	    MediaDescrip;		/* 0AH */
	unsigned  SecPerFAT;			/* 0BH calculated */
	unsigned  SecPerTrak;		/* 0DH */
	unsigned  Sides;				/* 0FH */
   ulong     Hidden;				/* 11H */
   ulong     LNumSec;			/* 15H */
};
#define START_BPB		11

typedef void (far *far_func_ptr)();

/*
	TSR Definition

	Description: Memory Disks can be configured to install
	as either a device driver or a TSR (which can then be
	unloaded if required).

	This structure allows a program to query a previously
	installed TSR Generic Disk.
*/
struct tsr_def /*()*/
{
	far_func_ptr tsr_uninit_addr;		/* a TSR "uninstall" routine */
	uchar tsr_start_drive;				/* starting drive of TSR */
	uchar tsr_num_drives;				/* # of drives in TSR */
	uchar tsr_hooked;						/* are the drives presently "active" */
	uchar tsr_id_len;						/* length of ID string */
	char *tsr_id_ptr;						/* the unique ID string itself */
};


/* has the disk changed? for mem_changed */
#define CHANGED_YES		-1
#define CHANGED_MAYBE	0
#define CHANGED_NO		1

#define DEF_SEC_SIZE		512		/* default sector size */
#define FAT_ID				0xF0		/* default FAT id byte */

/* categories for IOCTL calls */
#define BLOCK_IOCTL_READ  0xFF03
#define BLOCK_IOCTL_WRITE 0xFF0C

/* valid ROM-DISK identification */
#define ROMDISK_ID_OFF  3
#define ROMDISK_ID "DLRDISK"
#define SECTOR_ID   0xAA55		// sector ID offset is determined by BPB

/*
	Device Error Codes

	If the device must return an error, the following
	defines the possible error conditions:

  NOTE: These are the same errors identified by MSDOS v6.22 COMMAND.COM.
        Microsoft only documents errors 0-C in their prog. ref.
        See PC Interrupts (2nd Ed.) Pg. 524.

*/
#define DEVERR_WRITEPROT	0x00	/* write protected */
#define DEVERR_BADUNIT		0x01	/* invalid unit # */
#define DEVERR_NOTREADY		0x02	/* disk not ready (drive door open) */
#define DEVERR_BADCMD		0x03
#define DEVERR_CRC			0x04
#define DEVERR_BADSTRUCT	0x05
#define DEVERR_SEEK			0x06
#define DEVERR_BADMEDIA		0x07
#define DEVERR_SECNOTFOUND	0x08	/* sector not found */
#define DEVERR_NOPAPER		0x09
#define DEVERR_WRITEFAULT	0x0A
#define DEVERR_READFAULT	0x0B
#define DEVERR_GENERALFAIL	0x0C
#define DEVERR_SHARING  	0x0D	// Sharing violation
#define DEVERR_LOCKING  	0x0E	// Lock violation
#define DEVERR_DISKCHANGE  0x0F	// Invalid disk change
#define DEVERR_NOFCB  		0x10	// FCB Unavailable
#define DEVERR_SHAREBUFF	0x11	// Sharing buffer overflow
#define DEVERR_BADCODEPG	0x12	// Code page mismatch
#define DEVERR_NOINPUT  	0x13  // Out of input
#define DEVERR_NODISKSPACE	0x14  // Insufficient disk space


#if (SUITE >= 2)
#define DDTDISK 1
#include "device.h"
#endif



// some handy macros
#if BUILTIN
#define HI_BYTE32(x) ((unsigned)_lshru((ulong)(x), 24))
#define HI_WORD(x)   ((unsigned)_lshru((ulong)(x), 16))
#define LO_WORD(x)   ((unsigned)(x))
#else
#define HI_BYTE32(x) ((unsigned)((ulong)(x) >> 24))
#define HI_WORD(x)   ((unsigned)((ulong)(x) >> 16))
#define LO_WORD(x)   ((unsigned)(x))
#endif

/* useful data items */
extern uchar memerr;
extern uchar memunit;
extern unsigned memtop;

#if BUILTIN
/* prototypes for built-in drivers */
ulong PASCAL _lshru(ulong l, unsigned u);
ulong PASCAL _lshlu(ulong l, unsigned u);
ulong PASCAL _lmulu(ulong l, unsigned u);
unsigned PASCAL _ldivu(ulong ul, unsigned u) ;
int PASCAL strlen(char *s);
int PASCAL fstrlen(char far * s);
void PASCAL memmove(void *to,void *from,unsigned len) ;
void PASCAL fmemmove(void far *dst,void far *src,unsigned len);
void PASCAL memset(void *dst,uchar val,unsigned len) ;
void PASCAL fmemset(void far *dst,uchar val,unsigned len) ;

/*
	WARNING! in the ROM-DOS kernel (that is, for all built-in
	drivers) memory compares are TRUE/FALSE functions, not
	-/0/+ like in "normal" C.
*/
bool PASCAL memcmp(char *str1,char *str2,unsigned len) ;
bool PASCAL fmemcmp(char far *str1,char far *str2,unsigned len) ;

#endif

#if TSRDEV
void tsr_setidstr(char *id);
int memdev_already_loaded(void);
int tsr_uninstall(void);
#endif

/* prototypes */
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv);
void memuninit(void);
int memchanged(struct BPB *bpb);
bool memread(long offset, unsigned len, char far * buffer);
bool memwrite(long offset, unsigned len, char far * buffer);
bool memioctl(unsigned category, char far * buffer);
