;
;   $Id: DOSMATH.ASM 1.4 2002/10/15 23:47:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
COMMENT ~
-----------------------------------------------------------

   DOSMATH.ASM  created 11/30/89  by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   The long math routines, and fast string manipulation
   routines are all contained in this assembly file.
   ROM-DOS uses no routines from the compiler's library.

----------------------------------------------------------- ~
include macros.inc

cBegCode


COMMENT ; prototype
ulong PASCAL _lshru(ulong l, unsigned u)
{}
;
COMMENT ; prototype
ulong PASCAL _lshlu(ulong l, unsigned u)
{}
;
COMMENT ; prototype
ulong PASCAL _lmulu(ulong l, unsigned u)
{}
;
COMMENT ; prototype
ulong PASCAL _ldivu(ulong ul, unsigned u)
{}
;
COMMENT ; prototype
void PASCAL memmove(void *to,void *from,unsigned len)
{}
;
COMMENT ; prototype
void PASCAL fmemcpy(void far *dst,void far *src,unsigned len)
{}
;

; Enable 386 instructions if so defined
;
IFDEF USE_386
.386
ENDIF


;---------------------------------------------------------
;
;		MATH OPERATIONS
;
;---------------------------------------------------------


;-------------------------------------------------------
;
; ulong PASCAL _lshru(ulong l, unsigned u)
;
; This function performs a shift right on a dword by an
; word.  It does not actually shift by a word but instead
; shifts by the low byte.
;
; NOTE: uses SHR vs. SAR (sign of result not maintained)
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       cx -> destroyed (used for shift count)
;
;
	public _LSHRU
_LSHRU:
	IFDEF USE_386
	;
	;  This is the function using 386 instructions
	;

		enter	0,0		; instead of push bp
					; instead of mov bp, sp

		; save high word of edx
		;
		push	edx

		; load edx with l
		;
		mov	edx, p+2[bp]

		; load cl with u (yes cl is all we look at)
		;
		mov	cl, p[bp]	; cx  = u

		; perform the shift
		;
		shr	edx, cl

		; restore the high word of edx and set the result in
		; edx into dx:ax
		;
		mov	ax,  dx
		rol	edx, 16
		mov	cx,  dx
		pop	edx
		mov	dx,  cx

		leave			; instead of pop bp
		ret	6

	ELSE
	;
	;  This is the function using 8086 instructions
	;
		push	bp
		mov	bp,sp

		mov	ax,p+2[bp]
		mov	dx,p+4[bp]	; dx:ax = l
		mov	cx,p[bp]	; cx    = u
		jcxz	SHRX

	SHR1:	shr	dx,1
		rcr	ax,1
		loop	SHR1

	SHRX:	pop	bp
		ret	6


	ENDIF


;-------------------------------------------------------
;
; ulong PASCAL _lshlu(ulong l, unsigned u)
;
; This function performs a shift left on a dword by an
; word.  It does not actually shift by a word but instead
; shifts by the low byte.
;
; NOTE: uses SHL vs. SAL (sign of result not maintained)
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       cx -> destroyed (used for shift count)
;
;
	public	_LSHLU
_LSHLU:
	IFDEF USE_386
	;
	;  This is the function using 386 instructions
	;
		enter	0,0		; instead of push bp
					; instead of mov bp, sp

		; save high word of edx
		;
		push	edx

		; load edx with l
		;
		mov	edx, p+2[bp]

		; load cl with u (yes cl is all we look at)
		;
		mov	cl, p[bp]	; cx  = u

		; perform the shift
		;
		shl	edx, cl

		; restore the high word of edx and set the result in
		; edx into dx:ax
		;
		mov	ax,  dx
		rol	edx, 16
		mov	cx,  dx
		pop	edx
		mov	dx,  cx

		leave			; instead of pop bp
		ret	6

	ELSE
	;
	;  This is the function using 8086 instructions
	;
		push	bp
		mov	bp,sp

		mov	ax,p+2[bp]
		mov	dx,p+4[bp]	; dx:ax = l
		mov	cx,p[bp]	; cx    = u
		jcxz	SHLDONE

	SHL1:	shl	ax,1
		rcl	dx,1
		loop	SHL1

	SHLDONE:
		pop	bp
		ret	6


	ENDIF



;-------------------------------------------------------
;
; ulong PASCAL _lmulu(ulong l, unsigned u)
;
; This function multiplies a dword by a word and
; returns a dword.
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       bx -> destroyed (used for computation)
;       cx -> destroyed (used for computation)
;
; Note: This could be a signed MUL function by XORing the Sign BITs
;     	of the multiplier and multiplicand and then ORing the resultant
;	BIT with the Sign BIT of the product.
;
;
	public	_LMULU
_LMULU:
	push	bp
	mov	bp,sp

	; multiply lower 16 bits
	mov	ax,p[bp]	; unsigned
	mov	dx,p+2[bp]	; low  16 bits of ulong
	mul	dx
	mov	cx,dx		; save high part
	mov	bx,ax		; save low order part

	; multiply upper 16 bits
	mov	ax,p[bp]	; unsigned
	mov	dx,p+4[bp]	; hi 16 bites of ulong
	mul	dx		; high order part
	add	ax,cx		; add in overflow from prev multiply

	mov	dx,ax		; high 16 bits
	mov	ax,bx		; low  16 bits

	pop	bp
	ret	6



;-------------------------------------------------------
;
; ulong PASCAL _ldivu(ulong ul, unsigned u)
;
; This function divides a dword by a word and returns a
; dword.  This routine will never generate a div by zero
; exception.
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       bx -> destroyed (used for computation)
;       cx -> destroyed (used for computation)
;
; Note:	This could be a signed DIV function by XORing the Sign BITs
;      	of the dividend and divisor and ORing the resultant BIT with
;	the Sign BIT of the quotient. Remainders are always positive.
;
;
	public	_LDIVU
_LDIVU:
	PUSH	BP
	MOV	BP,SP

	MOV	AX,P+2[BP]
	MOV	DX,P+4[BP]	; DX:AX = ulong
	MOV	CX,P[BP]	; CX = unsigned

	xor	bx, bx
	or	cx, cx
	jnz	non_zero_divisor
	mov	ax, bx
	jmp	short ldivu_done

   non_zero_divisor:
	cmp	dx, cx
	jae	quotient_overflow

	or	dx, dx
	jnz     regular_div_ok

	cmp	ax, cx
	jae	regular_div_ok

    quotient_underflow:
	mov	ax, bx
	jmp	short ldivu_done

    quotient_overflow:
    	push	ax
	mov	ax, dx
	mov	dx, bx
	div	cx
	mov	bx, ax
	pop	ax

    regular_div_ok:
	div	cx

    ldivu_done:
	mov	dx, bx

    ldivu_exit:
	POP	BP
	RET	6




IF 0	; the following block of code is commented out




;-------------------------------------------------------
;
; unsigned _lmodu(ulong ul, unsigned u)
;
; This function returns the modulo of a dword divided
; by a word.
;
; This function does not preserve the following registers:
;
;	ax -> return value
;       cx -> destroyed (used for computation)
;       dx -> destroyed (used for computation)
;
	public _LMODU
c_func	_lmodu
	PUSH	BP
	MOV	BP,SP

	MOV	AX,P[BP]
	MOV	DX,P+2[BP]	; DX:AX = ulong
	MOV	CX,P+4[BP]	; CX = unsigned
	DIV	CX
	MOV	AX,DX

	POP	BP
	RET
c_endp	_lmodu ;


;-------------------------------------------------------
;
; ulong _ldiv(ulong l1, ulong l2)
;
; This function divides two dwords and returns a dword.
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       bx -> destroyed (used for computation)
;       cx -> destroyed (used for computation)
;
	public _LDIV
c_func	_ldiv
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI

	MOV	AX,P+2[BP]	; high 16 bits of l1
	MOV	BX,P+0[BP]	; low  16 bits of l1
	MOV	CX,P+6[BP]	; high 16 bits of l2
	MOV	DX,P+4[BP]	; low  16 bits of l2

	CALL	LXXX

	MOV	DX,AX		; high order bits of divide
	MOV	AX,BX		; low order bits of divide

	POP	DI
	POP	SI
	POP	BP
	RET
c_endp	_ldiv ;


;-------------------------------------------------------
;
; ulong _lmod(ulong l1, ulong l2)
;
; This function divides two dwords and returns the modulo.
;
; This function does not preserve the following registers:
;
;	ax -> low word of return value
;       dx -> high word of return value
;       bx -> destroyed (used for computation)
;       cx -> destroyed (used for computation)
;
c_func _lmod
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI

	MOV	AX,P+2[BP]	; high 16 bits of l1
	MOV	BX,P+0[BP]	; low  16 bits of l1
	MOV	CX,P+6[BP]	; high 16 bits of l2
	MOV	DX,P+4[BP]	; low  16 bits of l2

	CALL	LXXX

	MOV	AX,DX		; low order bits of remainder
	MOV	DX,CX		; high order bits of remainder

	POP	DI
	POP	SI
	POP	BP
	RET
c_endp _lmod


;-------------------------------------------
; Unsigned Long Divide
;
; Input:
;	[AX,BX],[CX,DX]
; Output:
;	[AX,BX] = [AX,BX] / [CX,DX]
;	[CX,DX] = [AX,BX] % [CX,DX]
;	SI,DI	destroyed

LXXX	PROC	NEAR
	MOV	SI,CX
	MOV	DI,DX			;[SI,DI] = divisor
	OR	CX,DX			;test for divide by 0
	JZ	div0			;yes

	;left justify [SI,DI] and leave count of shifts + 1 in BP

	PUSH	BP			;need a temporary
	MOV	BP,1			;at least 1 shift
L6:	TEST	SI,SI			;left justified?
	JS	L1			;yes
	JNZ	L2			;if non-zero
	XCHG	SI,DI			;SI = DI, DI = 0
	MOV	BP,17			;at least 17 shifts
	JMP	L6

L2:	INC	BP			;another shift
	SHL	DI,1
	RCL	SI,1			;[SI,DI] <<= 1
	TEST	SI,SI			;affect sign flag
	JNS	L2			;not left justified yet

L1:	MOV	CX,AX
	MOV	DX,BX			;[CX,DX] = [AX,BX]
	XOR	AX,AX
	XOR	BX,BX			;[AX,BX] = 0
L4:	CMP	SI,CX			;is [CX,DX] > [SI,DI]?
	JA	L3
	JB	L5			;definitely less than
	CMP	DI,DX			;check low order word
	JA	L3
L5:	SUB	DX,DI
	SBB	CX,SI			;[CX,DX] -= [SI,DI]
	STC				;rotate in a 1
L3:	RCL	BX,1
	RCL	AX,1			;[AX,BX] = ([AX,BX] << 1) + C
	SHR	SI,1
	RCR	DI,1			;[SI,DI] >>= 1
	DEC	BP			;control count
	JNE	L4
	POP	BP
	RET

div0:	MOV	AX,-1
	MOV	BX,AX			;quotient is -1
	XOR	CX,CX
	XOR	DX,DX			;remainder is 0
	RET

LXXX	ENDP




ENDIF	; end of the commented out block




;---------------------------------------------------------
;
;		STRING OPERATIONS
;
;---------------------------------------------------------


;-------------------------------------------------------
;
; void PASCAL memmove(void *to,void *from,unsigned len)
;
; This function moves a specified number of bytes from
; a source to a destination.
;
; This function does not preserve the following registers:
;
;	ax -> destroyed
;	bx -> destroyed
;	cx -> destroyed
;	dx -> destroyed
;	es -> destroyed
;
; NOTE: This move does all these moves correctly:
;
;		memmove(p, p+1, 5);
;		memmove(p, p+2, 5);
;		memmove(p, p+3, 5);
;		memmove(p, p+4, 5);
;		memmove(p+1, p, 5);
;		memmove(p+2, p, 5);
;		memmove(p+3, p, 5);
;		memmove(p+4, p, 5);
;
;
	public	MEMMOVE
MEMMOVE:
	IFDEF USE_386
	;
	;  This is the function using 386 instructions
	;

		push	bp
		mov	bp,sp
		push	si
		push	di

		; near move, es must equal ds
		;
		mov	ax, ds
		mov	es, ax

		mov	cx, p[bp]
		mov	si, p+2[bp]
		mov	di, p+4[bp]
		cmp	si, di

		; move forward if "from" is greater (no jump)
		; move backward if "to" is greater  (jump)
		;
		jbe	BACKWARD_MOVE

		; save flag indicating # of bytes to
		; pop off return stack (pascal func)
		;
		xor	ax, ax

	FORWARD_MOVE:

		; es:di = dest
		; ds:si = src
		; cx    = length
		;

		; Test if count is odd
		;
		mov	bx, cx
		and	cx, 03h
		jcxz	FWD_MOVSB_DONE

	FWD_MOVSB:
		; Perform the this part of the move with a
		; byte move
		;
		rep	movsb

	FWD_MOVSB_DONE:
		; Restore the count in cx
		;
		mov	cx, bx

	FWD_MOVSW:
		; Perform the this part of the move with a
		; dword move
		;
		shr	cx, 2
		jcxz	FOWARD_MOVE_DONE
		rep	movsd

	FOWARD_MOVE_DONE:

		; flag indicating size of pascal return
		; 0 means near ptrs, 1 means far
		;
		or	ax, ax
		jnz	FARMOV

		; Return from the memmove call
		;
		pop	di
		pop	si
		pop	bp
		ret	6	; 2 near ptrs + len

	FARMOV:
		; Return from the fmemcpy call
		;
		pop	ds
		pop	di
		pop	si
		pop	bp
		ret	10	; 2 far ptrs + len


	BACKWARD_MOVE:

		std

		add	si, cx
		dec	si
		add	di, cx
		dec	di

		; es:di = dest
		; ds:si = src
		; cx    = length
		;

		; Test if count is odd
		;
		mov	bx, cx
		and	cx, 03h
		jcxz	BWD_MOVSB_DONE

	BWD_MOVSB:
		; Perform the this part of the move with a
		; byte move
		;
		rep	movsb

	BWD_MOVSB_DONE:

		; Restore the count in cx
		;
		mov	cx, bx

	BWD_MOVSW:
		; Perform the this part of the move with a
		; dword move
		;
		sub	si, 3
		sub	di, 3
		shr	cx, 2
		jcxz	BACKWARD_MOVE_DONE
		rep	movsd

	BACKWARD_MOVE_DONE:

		; now cld so it is forward like normal
		;
		cld

		pop	di
		pop	si
		pop	bp
		ret	6	; 2 near ptrs + len


	ELSE
	;
	;  This is the function using 8086 instructions
	;

		push	bp
		mov	bp,sp
		push	si
		push	di
		push	ds
		xor	ax, ax		; save flag indicating # of bytes to
		push	ax		; pop off return stack (pascal func)

		; near move, es must equal ds
		push	ds
		pop	es
		mov	ax,p+2[bp]	; move forward if "from" is greater
		cmp	ax,p+4[bp]	; move backward if "to" is greater
		jnb	M1
		std
		mov	ax,1
		jmp	M2
	M1:	cld
		xor	ax,ax
	M2:	mov	si,p+2[bp]
		mov	di,p+4[bp]
		mov	cx,p[bp]

		; es:di = dest
		; ds:si = src
		; cx    = length
		; ax    = direction (0=forward, 1=backward)
	DO_MOVE:
		or	ax,ax
		je	M3
		add	si,cx
		dec	si
		add	di,cx
		dec	di
	M3:	test	di,1
		je	M4
		jcxz	M5
		movsb
		dec	cx
	M4:	sub	si,ax
		sub	di,ax
		shr	cx,1
		; do not cld here, we may be going the other way
		rep	movsw
		jnb	M5
		add	si,ax
		add	di,ax
		movsb
		; now cld so it is forward like normal

	M5:	cld			; now cld so it is forward like normal
		pop	ax		; flag indicating size of pascal return
		pop	ds
		pop	di
		pop	si
		pop	bp
		or	ax, ax		; 0 means near ptrs, 1 means far
		jnz	FARMOV
		ret	6	; 2 near ptrs + len
	FARMOV:	ret	10	; 2 far ptrs + len


	ENDIF


;-------------------------------------------------------
;
; void PASCAL fmemcpy(void far *dst,void far *src,unsigned len)
;
; This function moves a specified number of bytes from
; a source to a destination.
;
; This function does not preserve the following registers:
;
;	ax -> destroyed
;	bx -> destroyed
;	cx -> destroyed
;	dx -> destroyed
;	es -> destroyed
;
;   NOTE:   unlike memmove(), this routine does no
;   overlap checking at all.
;
	public	FMEMCPY
FMEMCPY:
	IFDEF USE_386
	;
	;  This is the function using 386 instructions
	;
		push	bp
		mov	bp, sp
		push	si
		push	di
		push	ds

		; far move is always forward
		;
		cld

		; save flag indicating # of bytes to
		; pop off return stack (pascal func)
		;
		xor	ax, ax
		inc	ax

		mov	cx, p[bp]	; len
		mov	di, p+6[bp]	; dst
		mov	es, p+8[bp]
		mov	si, p+2[bp]	; src
		mov	ds, p+4[bp]

		jmp	short FORWARD_MOVE

	ELSE
	;
	;  This is the function using 8086 instructions
	;

		push	bp
		mov	bp,sp
		push	si
		push	di
		push	ds
		mov	ax, 1		; save flag indicating # of bytes to
		push	ax		; pop off return stack (pascal func)

		cld			; far move is always forward
		xor	ax,ax
		mov	di,p+6[bp]	; dst
		mov	es,p+8[bp]
		mov	si,p+2[bp]	; src
		mov	ds,p+4[bp]
		mov	cx,p[bp]	; len

		jmp	DO_MOVE

	ENDIF

COMMENT ; prototype
void PASCAL memset(void *dst,uchar val,unsigned len)
{}
;
	public	MEMSET
MEMSET:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI
	XOR	AX, AX		; flag indicating # of bytes to
	PUSH	AX		; pop off stack on pascal return

	; get the parameters
	PUSH	DS
	POP	ES
	MOV	DI,P+4[BP]
	MOV	AL,P+2[BP]
	MOV	CX,P[BP]

	; fill the bytes
memset1:CLD
	REP	STOSB

	POP	AX		; flag indicating size of pascal return
	POP	DI
	POP	SI
	POP	BP
	OR	AX, AX
	JNZ	farset
	RET	6		; near ptr + val + len
farset:	RET	8		; far ptr + val + len

COMMENT ; prototype
void PASCAL fmemset(void far *dst,uchar val,unsigned len)
{}
;
	public	FMEMSET
FMEMSET:
	PUSH	BP
	MOV	BP, SP
	PUSH	SI
	PUSH	DI
	MOV	AX, 1		; flag indicating # of bytes to
	PUSH	AX		; pop off stack on pascal return

	; get the parameters
	MOV	CX, P[BP]	; len
	MOV	AL, P+2[BP]	; val
	MOV	DI, P+4[BP]
	MOV	ES, P+6[BP]	; far * dst

	; go do the move
	JMP	memset1


COMMENT ; prototype
/*
   Standard C memcmp.
   Returns -1 if str1 < str2, 0 if equal, 1 if str1 > str2
*/
int PASCAL memcmp(char *str1,char *str2,unsigned len)
{}
;
	public	MEMCMP
MEMCMP:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI
	PUSH	DS
	XOR	AX, AX		; flag indicating # of bytes to
	PUSH	AX		; pop off stack on pascal return

	; make sure ES = DS
	PUSH	DS
	POP	ES

	; get parameters from stack
	MOV	SI,P+4[BP]	; str1
	MOV	DI,P+2[BP]	; str2
	JMP	cmpnearfar


COMMENT ; prototype
/*
   Standard C fmemcmp.
   Returns -1 if str1 < str2, 0 if equal, 1 if str1 > str2
*/
int PASCAL fmemcmp(char far *str1,char far *str2,unsigned len)
{}
;
	public	FMEMCMP
FMEMCMP:
	push	bp
	mov	bp, sp
	push	si
	push	di
	push	ds

	mov	ax, 1		; flag indicating # of bytes to
	push	ax		; pop off stack on pascal return

	; get parameters from stack
	mov	ds, P+8[bp]
	mov	si, P+6[bp]	; str1
	mov	es, P+4[bp]
	mov	di, P+2[bp]	; str2

    cmpnearfar:
	MOV	CX, P[BP]	; len

    cmpagain:
	cmpsb
	jnz	cmpfail
	loop	cmpagain

	; strings are identical
	mov	ax, 0		; str1 == str2
	jmp	short cmpdone

    cmpfail:
	jc	cmpless
	mov	ax, 1		; str1 > str2
	jmp	short cmpdone
    cmpless:
	mov	ax, -1		; str1 < str2

    cmpdone:
	pop	bx		; flag for pascal return size
	pop	ds
	pop	di
	pop	si
	pop	bp
	or	bx, bx		; 1 means far, 0 means near
	jnz	cmpfar
	ret	6		; 2 near ptrs + len
cmpfar:	ret	10		; 2 far ptrs + len


COMMENT ; prototype
int PASCAL strlen(char *s)
{}
;
	public	STRLEN
STRLEN:
	push	BP
	mov	BP, SP
	push	DS
	pop	ES		; ES must point to segment
	call	strlenx		; requires far pointer
	pop	BP
	ret	2

COMMENT ; prototype
int PASCAL fstrlen(char far * s)
{}
;
	public	FSTRLEN
FSTRLEN:
	push	BP
	mov	BP, SP
	mov	ES, P+2[BP]	; ES must point to segment
	call	strlenx
	pop	BP
	ret	4

;-------------------------------------
;  internal use by strlen/fstrlen
;  Returns string length in AX
strlenx	proc	near
	push	DI
	mov	DI, P[BP]	; offset

	; get maximum possible length (no wrapping past FFFF)
	mov	CX, DI
	not	CX
	xor	AX, AX
	cld
	repnz	scasb

	; got endptr in DI, convert to length in AX
	xchg	DI, AX
	sub	AX, P[BP]
	dec	AX

	; restore regs and return
	pop	DI
	ret
strlenx	endp

cEndCode
	end
