;
;   $Id: MEMMATH.ASM 1.4 2002/10/15 23:47:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:MEMMATH.ASM
;
; Copyright (c) 1991-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary
;
; Long math functions used by the Generic Memory Disk Drivers.
;--------------------------------------------------------------------
include macros.inc

cBegCode

;--------------------------------------------------------------------
; $DOCPROC:_ldiv
;
; ulong _ldiv(ulong lVal, ulong lDivideBy)
;
; Divide a long lVal by another long lDivideBy. Returns a long result.
;--------------------------------------------------------------------
c_func	_ldiv
	push	bp
	mov	bp,sp
	push	si
	push	di

	mov	ax,p+2[bp]	; high 16 bits of l1
	mov	bx,p+0[bp]	; low  16 bits of l1
	mov	cx,p+6[bp]	; high 16 bits of l2
	mov	dx,p+4[bp]	; low  16 bits of l2

	call	LXXX

	mov	dx,ax		; high order bits of divide
	mov	ax,bx		; low order bits of divide

	pop	di
	pop	si
	pop	bp
	ret
c_endp	_ldiv


;--------------------------------------------------------------------
; $DOCPROC:_ldiv
;
; ulong _lmod(ulong lVal, ulong lModBy)
;
; Return the modulo (remainder). lVal is divided by lModBy, the
; remainder (a long) is returned.
;--------------------------------------------------------------------
c_func _lmod
	push	bp
	mov	bp,sp
	push	si
	push	di

	mov	ax,p+2[bp]	; high 16 bits of l1
	mov	bx,p+0[bp]	; low  16 bits of l1
	mov	cx,p+6[bp]	; high 16 bits of l2
	mov	dx,p+4[bp]	; low  16 bits of l2

	call	LXXX

	mov	ax,dx		; low order bits of remainder
	mov	dx,cx		; high order bits of remainder

	pop	di
	pop	si
	pop	bp
	ret
c_endp _lmod


;--------------------------------------------------------------------
; $DOCHELPER:LXXX
;
; Divide a long by another long. Returns both modulo and
; input:
;	[ax,bx],[cx,dx]
; output:
;	[ax,bx] = [ax,bx] / [cx,dx]
;	[cx,dx] = [ax,bx] % [cx,dx]
;	si,di	destroyed
;--------------------------------------------------------------------
LXXX	proc	near
	mov	si,cx
	mov	di,dx			;[si,di] = divisor
	or	cx,dx			;test for divide by 0
	jz	div0			;yes

	;left justify [si,di] and leave count of shifts + 1 in bp
	push	bp			;need a temporary
	mov	bp,1			;at least 1 shift
l6:	test	si,si			;left justified?
	js	l1			;yes
	jnz	l2			;if non-zero
	xchg	si,di			;si = di, di = 0
	mov	bp,17			;at least 17 shifts
	jmp	l6

l2:	inc	bp			;another shift
	shl	di,1
	rcl	si,1			;[si,di] <<= 1
	test	si,si			;affect sign flag
	jns	l2			;not left justified yet

l1:	mov	cx,ax
	mov	dx,bx			;[cx,dx] = [ax,bx]
	xor	ax,ax
	xor	bx,bx			;[ax,bx] = 0
l4:	cmp	si,cx			;is [cx,dx] > [si,di]?
	ja	l3
	jb	l5			;definitely less than
	cmp	di,dx			;check low order word
	ja	l3
l5:	sub	dx,di
	sbb	cx,si			;[cx,dx] -= [si,di]
	stc				;rotate in a 1
l3:	rcl	bx,1
	rcl	ax,1			;[ax,bx] = ([ax,bx] << 1) + c
	shr	si,1
	rcr	di,1			;[si,di] >>= 1
	dec	bp			;control count
	jne	l4
	pop	bp
	ret

div0:	mov	ax,-1
	mov	bx,ax			;quotient is -1
	xor	cx,cx
	xor	dx,dx			;remainder is 0
	ret

LXXX	endp

cEndCode

	end
