//  $Id: MEMPAGED.C 1.7 2003/03/20 16:58:16Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1993-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:MEMPAGED.C

	Copyright (c) 1993-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Paged memory RAM/ROM disk driver.

	The Datalight Generic Memory Disk Driver, MemPaged, provides all
	the code to make a bootable, Paged Disk Driver. It requires a
	paging mechanism to do the actual paging.

	Paged Disk Examples:
	MEMDISK+MEMPAGED+SC400PAG    = Paged Elan SC400 ROM disk
	MEMDISK+MEMPAGED+INT15PAG    = Paged Int15 RAM disk

	MemPaged relies on these paging mechanism routines/data:
	bool init_page(void);
	bool set_page(int iPage);
	void uninit_page(void);
	char szPagedDiskIDString[];
	long lPagedDiskWindowSize;
	uchar far *lpPagedDiskWindow;

	See INT15PAG.C or SC400PAG.C for examples of the paging mechanism.

	$DOCHISTORY
	07/23/1998 dg  Updated to Solid Code
	03/01/1993 dg  Created by Drew Gislason
*******************************************************************/
#define __FILENUM__	204
#include "dosver.h"
#include "compiler.h"
#include "memdisk.h"

bool	bStatic;								/* /S option  */

//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );

/* messages */
char szSignOn[]    = "\nMEMPAGED %s %s %s\n%s\n\n";
char szErrInit[]   = "Failed to initialize!\n";
char szErrNoDisk[] = "Failed to find disk!\n";
char szInstalled[] = "Installed %dKB paged disk as drive %c:\n";

#define DEF_DIRS  64			/* default # of root directory entries */
#define DEF_SIZEK 64			/* default size of VDISK in KBytes */
#define DEF_DIR_ENTRIES 	DEF_DIRS
struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
char std_volume[] = "VDISK V6.22";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	char * szSignOn = "\nDatalight Dynamic Driver Loader %s %s\n%s\n";

	if( fIsFat32World )
	{
		printf(szSignOn,szPagedDiskIDString,SHORTVERSION,BUILDVERSION,
				COPYRIGHT);
	}
	else
	{
		printf(szSignOn,szPagedDiskIDString,SHORTVERSION622,BUILDVERSION,
				COPYRIGHT);
	}
}

/*
	$DOCPROC:meminit

	Initialize the VDISK. Checks for static disks, and decides
	what memory to use.

	On Entry
		cmdline	the cmdline from device = in CONFIG.SYS
		*endseg	ignored
		drv		drive # assigned from DOS (0=A, 1=B, etc..)
	On Exit
		bpb		updated with accurate BPB for paged disk

	Returns 1 if worked, 0 if failed.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	int		nSecPerKB;
	unsigned nDiskSizeK;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	/* validate parameter */
	ASSERT(bpb);

	/* initialize the paging mechanism (formats paged RAM disks). */
	if(!init_page())
	{
		printf(szErrInit);
		return 0;
	}

	/* look for disk image */
	if(!IsValidDisk())
	{
		printf(szErrNoDisk);
		return 0;
	}
	GetDiskBPB(bpb);

	/* set total # of sectors */
	nSecPerKB = 1024 / bpb->BytesPerSector;
	nDiskSizeK = bpb->TotalSecs / nSecPerKB;

	/* show installed message */
	printf(szInstalled, nDiskSizeK, drv+'A');

	/* one disk */
	return 1;
}

/*
	$DOCPROC:RdWrPagedMemory

	Read from or write to the paged memory. Used by both memread() and
	memwrite(). Note: lPagedDiskWindowSize can range from 2 bytes to
	64K bytes.

   Returns TRUE if worked, FALSE if failed.
*/
int PASCAL RdWrPagedMemory(long offset, unsigned len, char far *buffer,
																					bool bRead)
{
	unsigned nPage;
	unsigned nSize;
	unsigned nPageOffset;
	uchar far *lpPtr;

	/* read or write all the bytes */
	while(len)
	{
		/* determine page # and offset */
		nPage = (unsigned)_ldiv(offset, lPagedDiskWindowSize);
		nPageOffset = (unsigned)_lmod(offset, lPagedDiskWindowSize);

		/* write as many bytes as we can */
		if(len < lPagedDiskWindowSize - nPageOffset)
			nSize = len;
		else
			nSize = (unsigned)(lPagedDiskWindowSize - nPageOffset);
		lpPtr = lpPagedDiskWindow + nPageOffset;

		#if DEBUG
			printf("nSize %x, lpPtr %lp\n", nSize, lpPtr);
		#endif

		/* Set paged memory hardware to correct page */
		if(!set_page(nPage))
		{
			memerr = DEVERR_GENERALFAIL;
			return FALSE;
		}

		/* actually read or write the bytes */
		if(bRead)
			fmemcpy(buffer, lpPtr, nSize);
		else
			fmemcpy(lpPtr, buffer, nSize);

		/* on to next part of buffer */
		offset += nSize;
		buffer += nSize;
		len    -= nSize;
	}

	/* Worked! */
	return TRUE;
}

/*
   This function reads data (usually 1 or more sectors)
   from the ROM/RAM/Flash disk into the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            read should start.
   len      The length of the read in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked.
   Returns FALSE(0) and sets memerr if failed.
*/
int memread(long offset, unsigned len, char far *buffer)
{
	#if DEBUG > 1
		printf("memread(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(offset < 0x2000000L);	/* disks are always <= 32Meg */
	ASSERT(lpPagedDiskWindow);
	ASSERT(lPagedDiskWindowSize);

	/* read from the disk */
	return RdWrPagedMemory(offset, len, buffer, TRUE);
}

#if !ROMDISK
/*
   This function writes data (usually 1 or more sectors)
   to the ROM/RAM/Flash disk from the buffer specified
   by DOS.

   offset   The offset (in bytes) into the disk where the
            write should start.
   len      The length of the write in bytes.  The length
            will not exceed 62K (62*1024 bytes).
   buffer   The buffer, in segment:offset form.

   Returns TRUE(non-zero) if worked.
   Returns FALSE(0) and sets memerr if failed.
*/
int memwrite(long offset, unsigned len, char far *buffer)
{
	#if DEBUG > 1
		printf("memwrite(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(offset < 0x2000000L);	/* disks are always <= 32Meg */
	ASSERT(lpPagedDiskWindow);
	ASSERT(lPagedDiskWindowSize);

	/* write to memory */
	return RdWrPagedMemory(offset, len, buffer, FALSE);
}
#endif /* !ROMDISK */

#if DEBUG
/*
	$DOCPROC:AssertFail

	When an assertion fires, this function is called. Only defined if
	DEBUG is defined.
*/
void PASCAL AssertFail(int nFile, int nLine)
{
	panic("MEMPAGED Internal Error %d.%d!\n", nFile, nLine);
}
#endif	/* DEBUG */


#if UNIT_TEST

#if ROMDISK
#include <dos.h>

char	buffer[512];
char	buffer2[512];
struct BPB	testBPB;
unsigned		endseg = 0x1000;
unsigned		memtop = 0xA000;
uchar memerr;
uchar memunit;

/*
	$DOCPROC:DosPutc

	Send output of memdisk printf() through DOS (rather than BIOS) so
	it can be piped to a file.
*/
void PASCAL DosPutc(uchar c)
{
	bdos(2, c, 0);
}

/*
	$DOCPROC:RomDiskUnitTest

	Simply initialize a ROM disk. Used to test ROM disks.
*/
bool RomDiskUnitTest(char szCmdline[])
{
	extern void (PASCAL *prfunc)(uchar c);

	/* send output through STDOUT so it can be piped into a file */
	prfunc = DosPutc;

	/* initialize the RAM disk */
	if(!meminit(&testBPB, szCmdline, &endseg, 4))
	{
		printf("Failed to initialize the disk!\n");
		return FALSE;
	}

	/* BPB must be valid */
	if(testBPB.TotalSecs == 0)
	{
		printf("Invalid BPB!\n");
		return FALSE;
	}

	printf("Unit Tests passed.\n");
	return TRUE;
}
#endif	/* ROMDISK */

/*
	UnitTest starts here.
*/
int main(void)
{
#if ROMDISK
	/* ROM disk UnitTest */
	if(!RomDiskUnitTest(""))
		return 1;
#else
	/* RAM disk UnitTest */
	if(!RamDiskUnitTest(""))
		return 1;
#endif

	/* no errors */
	return 0;
}
#endif	/* UNIT_TEST */
