;
;   $Id: MEMPTR.ASM 1.4 2002/10/15 23:47:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:MEMPTR.ASM
;
; Copyright (c) 1998-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary
;
; Routines to convert between pointers and linear addresses. Used by
; the Generic Memory Disks (these COULD be done in C, but then they
; would rely on intrinsic (compiler) math functions, which would in
; turn make the C code so it couldn't be built-in to the ROM-DOS
; Kernel. The RD Kernel cannot use compiler library routines.
;--------------------------------------------------------------------
include macros.inc

cBegCode

;--------------------------------------------------------------------
; $DOCPROC:SegToLinear
;
; ulong PASCAL SegToLinear(unsigned seg)
;
; Return a linear address from a real-mode segment.
;--------------------------------------------------------------------
	public	SEGTOLINEAR
SEGTOLINEAR:
	push	bp
	mov	bp, sp

	; convert real-mode segment to 20-bit linear address
	; (basically shift left by 4)
	xor	dx, dx
	mov	ax, P[bp]	; dx:ax = segment
	mov	cx, 4
    ShiftLoop:			; shift by 4 bits (20 bit address)
	shl	ax,1
	rcl	dx,1
	loop	ShiftLoop

	pop	bp
	ret	2

;--------------------------------------------------------------------
; $DOCPROC:AddPtr
;
; void far * PASCAL AddPtr(unsigned seg, ulong lAddr);
;
; Add a linear address to an x86 segment to return a far ptr.
; Example: seg = 2000h, lAddr = 12345h, the ptr is 3234:5h
;--------------------------------------------------------------------
	public	ADDPTR
ADDPTR:
	push	bp
	mov	bp, sp

	; convert segment to linear address
	push	P+4[bp]
	call	SEGTOLINEAR

	; add in linear address
	add	ax, P[BP]	; low word of lAddr
	adc	dx, P+2[BP]

	; convert to pointer
	push	dx
	push	ax
	call	LINEAR2PTR

	pop	bp
	ret	6

;--------------------------------------------------------------------
; $DOCPROC:LinearAddress
;
; ulong PASCAL LinearAddress(void *pPtr);
;
; Returns the linear address represented by this real-mode pointer.
;--------------------------------------------------------------------
	public	LINEARADDRESS
LINEARADDRESS:
	push	bp
	mov	bp, sp

	; convert segment (ds) to linear address
	mov	ax, ds
	push	ax
	call	SEGTOLINEAR

	; add in offset
	add	ax, P[bp]
	adc	dx, 0		; dx:ax = linear address
	pop	bp
	ret	2

;--------------------------------------------------------------------
; $DOCPROC:Linear2Ptr
;
; void far * PASCAL Linear2Ptr(ulong lAddress);
;
; Converts a linear address to a ptr.
;--------------------------------------------------------------------
	public	LINEAR2PTR
LINEAR2PTR:
	push	bp
	mov	bp, sp

	push	ax		; save low word
	mov	cx, 4
    PtrLoop:			; get upper 16-bits of 20-bit addr
	shr	dx, 1
	rcr	ax, 1
	loop	PtrLoop
	mov	dx, ax		; dx = segment
	pop	ax
	and	ax, 0Fh

	pop	bp
	ret	4

;--------------------------------------------------------------------
; $DOCPROC:AddNormalize
;
; void far * PASCAL AddNormalize(void far *lpPtr, ulong lOffset);
;
; Returns a normalized ptr after adding the offset.
; Example: AddNormalize(1234:5678, AB) would return 17A6:3.
;--------------------------------------------------------------------
	public	ADDNORMALIZE
ADDNORMALIZE:
	push	bp
	mov	bp, sp

	; convert far ptr to linear address
	push	P+6[bp]		; seg
	call	SEGTOLINEAR
	add	ax, P+4[bp]	; offset
	adc	dx, 0

	; add in long
	add	ax, P[bp]	; add lo words
	adc	dx, P+2[bp]	; add hi words

	; convert back to ptr
	push	dx		; hi word
	push	ax		; low word
	call	LINEAR2PTR

	pop	bp
	ret	8

cEndCode
	END
