//  $Id: MEMROM.C 1.6 2003/03/20 16:57:24Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1995-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:MEMROM.C

	Copyright (c) 1995-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Fixed location extended memory ROM disk.

	Syntax:
	DEVICE=MEMROM.SYS [/L#]

	/L#  Linear address for start of disk (example 1000000 is 16 Megs)

	The fixed location ROM disk can be anywhere in memory (just
	provide a linear address). Does no searching for the ROM disk, so
	the disk MUST exist at that fixed address. Create the ROM disk
	using ROMDISK.EXE.

	MEMROM uses Int15h to transfer the memory and requires that Int15h
	support addresses up to the end of the ROM disk.

	$DOCHISTORY
	07/20/1998 dg  Updated to Solid Code
	03/01/1995 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__	202
#include "dosver.h"
#include "compiler.h"
#include "memdisk.h"


//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );

/* linear address start of ROM-DISK. */
#define _1MEG 0x100000L
ulong lStartAddr = (_1MEG*32);	/* default location 32 Meg */

/* messages */
char szSignOn[]    = "\nMEMROM %s %s\n%s\n\n";
char szErrNoDisk[] = "Failed to find ROM disk at %p%ph!\n";
char szErrBadOpt[] = "Use only the /L switch";
char szInstalled[] = "Installed %dKB ROM disk as drive %c: at %p%ph\n";

bool fIsFat32World;

#define DEF_DIRS  64			/* default # of root directory entries */
#define DEF_SIZEK 64			/* default size of VDISK in KBytes */
#define DEF_DIR_ENTRIES 	DEF_DIRS
struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
char std_volume[] = "VDISK V6.22";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

/* default cmdline for built-in drivers */
char szBuiltinCmdline[] = "/LD0000";

/* prototypes */
bool ParseCmdline(char far *lpCmdline);

void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	$DOCPROC:meminit

	Initialize the MEMROM ROM disk.

	On Entry
		cmdline	the cmdline from device= in CONFIG.SYS
		*endseg	paragraph of code+data. adusted for conventional disk.
		drv		drive # assigned from DOS (0=A, 1=B, etc..)
	On Exit
		bpb		updated with accurate BPB for RAM disk
		*endseg	updated if conventional memory RAM disk

	Returns 1 if worked, 0 if failed.
*/
#ifdef __TURBOC__
#pragma argsused
#endif
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	int		nSecPerKB;
	unsigned nDiskSizeK;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	/* validate parameter */
	ASSERT(bpb);

	/* determine switches, and sizes suggested by user */
	#if BUILTIN
		cmdline = szBuiltinCmdline;
	#endif
	if(!ParseCmdline(cmdline))
		return 0;

	/* look for disk image */
	if(!IsValidDisk())
	{
		printf(szErrNoDisk, HIWORD(lStartAddr), LOWORD(lStartAddr));
		return 0;
	}
	GetDiskBPB(bpb);

	/* set total # of sectors */
	nSecPerKB = 1024 / bpb->BytesPerSector;
	nDiskSizeK = bpb->TotalSecs / nSecPerKB;

	/* show installed message */
	printf(szInstalled, nDiskSizeK, drv+'A',
					HIWORD(lStartAddr), LOWORD(lStartAddr));

	/* one disk */
	return 1;
}

/*
	$DOCPROC:memread

   Read one or more sectors from the RAM disk into a buffer specified
	by DOS.

	On Entry
	   offset   The offset (in bytes) into the disk where the read
					starts.
	   len      The length of the read in bytes. The length cannot
					exceed 64KB - 1 sector.
	   buffer   Pointer to the buffer to read into.

	On Exit
	   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memread(long offset, unsigned len, char far *buffer)
{
	#if DEBUG > 1
		printf("memread(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(offset < 0x2000000L);		/* disks are always <= 32Meg */
	ASSERT(lStartAddr);

	/* copy from the extended memory */
	if(!CopyFromInt15(buffer, lStartAddr+offset, len/2))
		return FALSE;

	/* worked! */
	return TRUE;
}

/*
	$DOCPROC:ParseCmdline

	Parse the command line arguments. Parses both switches and
	arguments as follows:

	Syntax:
	DEVICE=MEMROM.SYS [/L#]
                    ^
*/
bool ParseCmdline(char far *lpCmdline)
{
	char	c;

	/* validate parameter */
	ASSERT(lpCmdline);

	/* look for command line arguments */
	while(1)
	{
		/* skip whitespace to next switch */
		lpCmdline = MemSkipWhite(lpCmdline);
		c = *lpCmdline;

		/* at end of line? */
		if(c < ' ')
			break;

		/* cmdline options (switches) */
		if(c == '/')
		{
			/* get switch char */
			++lpCmdline;						/* skip '/' */
			c = MemToUpper(*lpCmdline++);

			switch(c)
			{
				/* RAM disk to be placed in extended memory */
				case 'L':
					lStartAddr = 0;
					while(MemIsXDigit(*lpCmdline))
					{
						lStartAddr = _lmulu(lStartAddr, 0x10) +
														MemXDigitVal(*lpCmdline);
						++lpCmdline;
					}
					break;

				default:
					printf(szErrBadOpt);
					return FALSE;
			}
		}

		/* must be a cmdline argument */
		else
		{
			printf(szErrBadOpt);
			return FALSE;
		}

	} /* end of while */

	/* everything worked */
	return TRUE;
}

#if DEBUG
/*
	$DOCPROC:AssertFail

	When an assertion fires, this function is called. Only defined if
	DEBUG is defined.
*/
void PASCAL AssertFail(int nFile, int nLine)
{
	panic("MEMROM Internal Error %d.%d!\n",nFile,nLine);
}
#endif	/* DEBUG */


#if UNIT_TEST
#include <dos.h>

char	buffer[512];
char	buffer2[512];
struct BPB	testBPB;
unsigned		endseg = 0x1000;
unsigned		memtop = 0xA000;

/*
	$DOCPROC:DosPutc

	Send output of memdisk printf() through DOS (rather than BIOS) so
	it can be piped to a file.
*/
void PASCAL DosPutc(uchar c)
{
	bdos(2, c, 0);
}

/*
	$DOCPROC:RomDiskUnitTest

	Simply initialize a ROM disk. Used to test ROM disks.
*/
bool RomDiskUnitTest(char szCmdline[])
{
	extern void (PASCAL *prfunc)(uchar c);

	/* send output through STDOUT so it can be piped into a file */
	prfunc = DosPutc;

	/* initialize the RAM disk */
	if(!meminit(&testBPB, szCmdline, &endseg, 4))
	{
		printf("Failed to initialize the disk!\n");
		return FALSE;
	}

	/* BPB must be valid */
	if(testBPB.TotalSecs == 0)
	{
		printf("Invalid BPB!\n");
		return FALSE;
	}

	printf("Unit Tests passed.\n");
	return TRUE;
}

int main(void)
{
	static char szUnitTestCmdline[] = "/LD0000";

	/* do the unit test */
	if(!RomDiskUnitTest(szUnitTestCmdline))
		return 1;

	/* no errors */
	return 0;
}

#endif	/* UNIT_TEST */
