//  $Id: VDISK.C 1.8 2003/03/20 16:52:20Z thomd Exp $
//
//  *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//  Copyright (c) 1997-2003 Datalight, Inc.
//  All Rights Reserved
//
//  Datalight, Inc. is a Washington corporation located at:
//
//  21520 30th Dr SE, Suite 110,          Tel: +1.425.951.8086
//  Bothell, WA  98021                    Fax: +1.425.951.8094
//  USA                                   Http://www.datalight.com
//
//  This software, including without limitation all source code and
//  documentation, is the confidential, trade secret property of
//  Datalight, Inc., and is protected under the copyright laws of
//  the United States and other jurisdictions.  Portions of the
//  software may also be subject to one or more the following US
//  patents: US#5860082, US#6260156.
//
//  In addition to civil penalties for infringement of copyright
//  under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//  penalties for violation of (a) the restrictions on circumvention
//  of copyright protection systems found in 17 U.S.C. 1201 and
//  (b) the protections for the integrity of copyright management
//  information found in 17 U.S.C. 1202.
//
//  U.S. Government Restricted Rights.  Use, duplication,
//  reproduction, or transfer of this commercial product and
//  accompanying documentation is restricted in accordance with
//  FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//  IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//  PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//  AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//  INC. AND THE LICENSEE ("BINDING AGREEMENT").  YOUR RIGHT, IF ANY,
//  TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//  TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//  USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//  TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT.  CONTACT
//  DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//  THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:VDISK.C

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary.

	Conventional, Extended, XMS and EMS RAM disk.

	Syntax:
	DEVICE=VDISK.SYS [DiskSize [SectorSize [NumEntries]]] [/E] [/A] [/S]

	DiskSize		Size of disk in Kilobytes. Valid range is 4 - 32767.
	SectorSize	Size of each disk sector. Valid values are 128, 256 and 512.
	NumEntries	Number of root directory entries.

	/E				Use extended memory (use XMS if available)
	/A				Use expanded (EMS) memory
	/S				Check for static disk. Preserved across reboots.

	This driver is a RAM disk, that can allocate its memory from the
	DOS 640K memory pool, Extended memory, XMS memory or Expanded
	(EMS) memory. It can also use the Int 15H extended memory
	interface even if HIMEM is not loaded.

	VDISK is just one example of Datalight's Generic Disk Driver,
	MEMDISK.  Other implementations include MEMPAGED, MEMFIXED, and
	MEMROM.  See MEMDISK.ASM and the ROM-DOS Developer's Guide for
	details on building your own Generic Disk Driver.

	Compile Time Options:
	-DBUILTIN=1         Allows VDISK to be built-in to ROM-DOS
	-DDEBUG=1           Displays run-time debugging
	-DUNIT_TEST=1       For Compiling stand-alone unit test

	$DOCHISTORY
	07/22/1998 dg  Updated to Solid Code standards
	11/12/1997 dg  Created by Drew Gislason
*********************************************************************/
#define __FILENUM__	200
#include "dosver.h"
#include "compiler.h"
#include "memdisk.h"
#include "vdisk.h"

//#include "fatworld.h"
bool PASCAL dlIsFat32World( void );

/* cmdline options and arguments */
bool	bExtended = FALSE;					/* /E option */
bool	bUseEMS   = FALSE;					/* /A option */
bool	bStatic   = FALSE;					/* /S option */
bool	bVerbose  = FALSE;					/* /V option */

/* disk parameters */
#define DEF_DIR_ENTRIES 	DEF_DIRS
unsigned nTotalSecs;
unsigned nDiskSizeK = DEF_SIZEK;
int	nBytesPerSector = DEF_SEC_SIZE;
int	nRootEntries = DEF_DIRS;

/* start of disk for Conventional, Int 15, XMS and EMS */
unsigned nStartSeg;				/* if conventional RAM disk */
ulong	lStartAddr = 0x100000L;	/* if Int 15h memory disk   */
int	hXMS;							/* handle if XMS disk   */
unsigned hEMS;						/* handle if EMS disk   */
unsigned nEMSFrameSeg;			/* frame segment */

#define EMS_PAGE	0
#define EMS_PAGESIZE	(16384)	/* 16K pages */

/* used to determine RAM disk type */
enum DiskType nDiskType;

bool fIsFat32World;

/* messages */
char szSignOn[]    = "\nVDISK %s %s\n%s\n\n";
char szErrHeader[] = "VDISK Error: %s!\n";
char szErrFormat[] = "Format failed";
char szErrNoMem[]  = "Not enough memory";
char szErrAlloc[]  = "Failed to allocate memory";
char szErrNoExt[]  = "No Extended memory";
char szErrNoEMS[]  = "No EMS memory";
char szErrBadOpt[] = "Invalid option \"/%s\"!\n";
char szErrBadArgs[]= "Too many arguments!\n";
char szInstalled[] = "Installed %dKB %s %sRAM disk as drive %c:\n";
char szConventional[]= "Conventional";
char szExtended[]  = "Extended";
char szXMS[]       = "XMS";
char szEMS[]       = "EMS";
char szStatic[]    = "(static) ";

/* default built-in disk is a 1 Meg extended disk */
#if BUILTIN
char szBuiltinCmdline[] = "1024 /E";
#endif

struct BPB std_bpb =
{
	DEF_SEC_SIZE,1,1,2,DEF_DIR_ENTRIES,DEF_SIZEK*2,FAT_ID,6,9,2,0
};
char std_volume[] = "VDISK V6.22";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

#if UNIT_TEST

/* set DISK=1 through 4 in the environment. DISK must be set to compile. */
#if 	DISK==1
	char szUnitTestCmdline[] = "4KB";
#elif DISK==2
	char szUnitTestCmdline[] = "4096 /E";
#elif DISK==3
	char szUnitTestCmdline[] = "2048 /E /S";
#elif DISK==3
	char szUnitTestCmdline[] = "1024 /A";
#endif	; DISK == ???

#endif	; #if UNIT_TEST


void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	$DOCPROC:meminit

	Initialize the VDISK. Checks for static disks, and decides
	what memory to use.

	On Entry
		cmdline	the cmdline from device= in CONFIG.SYS
		*endseg	paragraph of code+data. adusted for conventional disk.
		drv		drive # assigned from DOS (0=A, 1=B, etc..)
	On Exit
		bpb		updated with accurate BPB for VDISK
		*endseg	updated if conventional memory VDISK

	Returns 1 if worked, 0 if failed.
*/
int meminit(struct BPB bpb[], char far *cmdline, unsigned *endseg, int drv)
{
	int		nSecPerKB;
	unsigned kBytesFree;
	unsigned nPages;
	bool		bFailed;
	char		*pszDiskType;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* display sign-on */
	DisplaySignOnMessage();

	/* validate parameter */
	ASSERT(bpb);

	/* determine switches, and sizes suggested by user */
	#if BUILTIN
		cmdline = szBuiltinCmdline;
	#endif
	if(!ParseCmdline(cmdline))
		return 0;

	/* user wants EMS, but is EMM loaded? */
	if(bUseEMS)
	{
		/* no EMM or Frame Seg? give up! */
		if(CheckEMMInstalled() != EMS_SUCCESS ||
			GetLIM32PageFrameSeg(&nEMSFrameSeg) != EMS_SUCCESS)
		{
			printf(szErrHeader, szErrNoEMS);
			return 0;
		}

		/* get free pages (16K chunks) and convert to K bytes */
		if(GetEMSFreePageCount(&kBytesFree) != EMS_SUCCESS)
			kBytesFree = 0;
		if(kBytesFree > 2048)		/* VDISK can only handle 32 Meg disks */
			kBytesFree = 32768U;
		else
			kBytesFree *= 16;			/* convert 16K pages to K bytes */
		#if DEBUG
			printf("Available EMS memory %d\n",kBytesFree);
		#endif

		/* disk type is EMS memory */
		nDiskType = dtEMS;
		pszDiskType = szEMS;
	}

	/* user wants an extended memory disk */
	else if(bExtended)
	{
		/* try XMS first */
		if(IsHimemInstalled())
		{
			kBytesFree = TotalXMS();
			#if DEBUG
				printf("Available XMS memory %d\n",kBytesFree);
			#endif
			nDiskType = dtXMS;
			pszDiskType = szXMS;
		}

		/* no XMS, use Int 15h directly */
		else
		{
			/* set nDiskType before calling IsValidDisk(). memread()
				needs to know the type.
			*/
			nDiskType = dtInt15;
			pszDiskType = szExtended;

//  ETR 3180
//
//  Moved all the static disk detection into FormatRAMDisk();
//
#if 0
			/* does a VDisk already exist? Use it! */
			if(bStatic && !IsFirstVDisk() && IsValidDisk())
			{
				/* determine disk size in kBytes */
				GetDiskBPB(bpb);
				nSecPerKB = 1024 / bpb->BytesPerSector;
				nDiskSizeK = bpb->TotalSecs / nSecPerKB;

				goto ShowDisk;
			}
#endif

			/* if no extended memory, inform user */
			kBytesFree = AvailInt15Memory();
			#if DEBUG
				printf("Available extended memory %d\n",kBytesFree);
			#endif
			if(kBytesFree == 0)
			{
				printf(szErrHeader, szErrNoExt);
				return 0;
			}

//  ETR 3180
//
//  Moved all the static disk detection into FormatRAMDisk();
//
#if 0
			/* if we don't want a static disk, wipe it out */
			if(!bStatic && IsFirstVDisk())
				OnInt19WipeVDisk();
#endif
		}
	}

	/* conventional memory */
	else
	{
		/* determine free size in K bytes (convert from paras) */
		nStartSeg = *endseg;
		kBytesFree = (memtop - nStartSeg) / (1024 / 16);
		nDiskType = dtConventional;
		pszDiskType = szConventional;
	}

	/* not enough free memory, inform user */
	if(kBytesFree < 4)
	{
		printf(szErrHeader,szErrNoMem);
		return 0;
	}

//  ETR 1920 BT 12-17-99
//
//  We want to error if there is not the amount of memory
//  specified by the user
//
//	/* reduce disk size if not enough memory */
//	if(nDiskSizeK > kBytesFree)
//		nDiskSizeK = kBytesFree;
//
	if(nDiskSizeK > kBytesFree)
	{
		printf(szErrHeader,szErrNoMem);
		return 0;
	}

	/* allow up to 32 Meg disks */
	if(nDiskSizeK > 32768U)
		nDiskSizeK = 32768U;

	/* limit 8Meg for 128-byte sectors, 16Meg for 256 byte sectors */
	if(nDiskSizeK >= 8*1024 && nBytesPerSector == 128)
		nBytesPerSector = 256;
	if(nDiskSizeK >= 16*1024 && nBytesPerSector == 256)
		nBytesPerSector = 512;

	/* set total # of sectors */
	nSecPerKB = 1024 / nBytesPerSector;
	nTotalSecs = nDiskSizeK * nSecPerKB;

	/* allocate the memory */
	bFailed = FALSE;
	switch(nDiskType)
	{
		/* allocate the conventional memory */
		case dtConventional:
			*endseg += nDiskSizeK * (1024/16);
			#if DEBUG
				printf("Conventional: start seg %p\n",nStartSeg);
			#endif
			break;

		/* allocate the int 15h memory */
		case dtInt15:
			if(!AllocateInt15Memory(nDiskSizeK, &lStartAddr))
				bFailed = TRUE;
			#if DEBUG
				printf("Extended: start addr %lp\n",lStartAddr);
			#endif
			break;

		/* allocate XMS memory */
		case dtXMS:
			hXMS = AllocateXMS(nDiskSizeK);
			if(!hXMS)
				bFailed = TRUE;
			#if DEBUG
				printf("XMS: handle %p\n",hXMS);
			#endif
			break;

		/* allocate EMS memory */
		case dtEMS:
			nPages = (nDiskSizeK+15) / 16;
			if(AllocateEMSPages(nPages, &hEMS) != EMS_SUCCESS)
				bFailed = TRUE;
			#if DEBUG
				printf("EMS: handle %p\n",hEMS);
			#endif
			break;
	}

	/* failed to allocate the memory! */
	if(bFailed)
	{
		printf(szErrHeader, szErrAlloc);
		return 0;
	}

	/* actually format the media */
	if(!FormatRAMDisk(nBytesPerSector, nRootEntries, nTotalSecs))
	{
		printf(szErrHeader, szErrFormat);
		return 0;
	}

	/* the format has trashed the "used" field at 1Meg:1Eh restore it */
	if(nDiskType == dtInt15)
		RestoreInt15Used();

	/* show installed message */
ShowDisk:
	printf(szInstalled, nDiskSizeK, pszDiskType,
						(bStatic ? szStatic : ""), drv+'A');

	if(bVerbose)
	{
		if(nDiskType == dtInt15)
			printf("Start %lp\n",lStartAddr);
	}

	/* tell memdisk about this BPB */
	GetDiskBPB(bpb);
	memmove(&std_bpb, bpb, sizeof(std_bpb));

	/* one disk */
	return 1;
}

/*
	$DOCPROC:memread

   Read one or more sectors from the RAM disk into a buffer specified
	by DOS.

	On Entry
	   offset   The offset (in bytes) into the disk where the read
					starts.
	   len      The length of the read in bytes. The length cannot
					exceed 64KB - 1 sector.
	   buffer   Pointer to the buffer to read into.

	On Exit
	   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memread(long offset, unsigned len, char far *buffer)
{
	bool worked = FALSE;

	#if DEBUG > 1
		printf("memread(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);
	ASSERT(offset < 0x2000000L);	/* disks are always <= 32Meg */

	switch(nDiskType)
	{
		case dtConventional:
			ASSERT(nStartSeg);
			fmemcpy(buffer, AddPtr(nStartSeg, offset), len);
			worked = TRUE;
			break;
		case dtInt15:
			ASSERT(lStartAddr);
			worked = CopyFromInt15(buffer, lStartAddr+offset, len/2);
			break;
		case dtXMS:
			ASSERT(hXMS);
			CopyFromXMS(buffer, hXMS, offset, len);
			worked = TRUE;
			break;
		case dtEMS:
			ASSERT(hEMS);
			worked = RdWrEMS(offset, len, buffer, TRUE);
			break;
		default:
			ABORT();
			break;
	}

	/* did this read work? */
	return worked;
}


/*
	$DOCPROC:memwrite

   Write one or more sectors to the RAM disk from a buffer specified
	by DOS.

	On Entry
	   offset   The offset (in bytes) into the disk where the write
					starts.
	   len      The length of the write in bytes. The length cannot
					exceed 64KB - 1 sector.
	   buffer   Pointer to the buffer to write from.

	On Exit
	   Returns TRUE(non-zero) if worked, FALSE(zero) if failed.
*/
bool memwrite(long offset, unsigned len, char far *buffer)
{
	bool worked = FALSE;

	/* display write address in DEBUG mode */
	#if DEBUG > 1
		printf("memwrite(%lx, %x, %lp)\n",offset,len,buffer);
	#endif

	/* validate parameters */
	ASSERT(len);
	ASSERT(buffer);

	switch(nDiskType)
	{
		case dtConventional:
			ASSERT(nStartSeg);
			fmemcpy(AddPtr(nStartSeg, offset), buffer, len);
			worked = TRUE;
			break;
		case dtInt15:
			ASSERT(lStartAddr);
			worked = CopyToInt15(lStartAddr+offset, buffer, len/2);
			break;
		case dtXMS:
			ASSERT(hXMS);
			CopyToXMS(hXMS, offset, buffer, len);
			worked = TRUE;
			break;
		case dtEMS:
			ASSERT(hEMS);
			worked = RdWrEMS(offset, len, buffer, FALSE);
			break;
		default:
			ABORT();
			break;
	}

	/* did this write work? */
	return worked;
}

// see MS-DOS Programmers Reference page 38
struct DeviceParams
{
	uchar 	dpSpecFunc;
	uchar 	dpDevType;
	unsigned dpDevAttr;
	unsigned dpCylinders;
	uchar 	dpMediaType;
	struct   BPB bpb;
};

struct SpecialFunc	// see MS-DOS Programmers Reference page 58
{
	uchar sfFunction;
	uchar sfAccessFlg;
};
#define sfFUNC_NUM0	0
#define sfACCESS_OK	1

/*
	$DOCPROC:memioctl

	Allows the VDISK to be formatted with the standard DOS FORMAT
	program.
*/
bool memioctl(unsigned category, char far *buffer)
{
	switch(category)
	{
		// These functions are stubbed since there is no need to do anything
		//	else when formatting a RAM disk. Add additional functionality here
		// as might be required.
		//
		case 0x840:	// set device parmaters
		case 0x842:	// format track
		case 0x846:	// set media ID
			return(TRUE);

		case 0x860: // get device parameters
		{
			struct DeviceParams far* lpDevParams =
				(struct DeviceParams far*)(buffer);

			// FORMAT will take care of all neccessary re-initializing of the
			// VDISK by writing directly to the FAT and ROOT DIR sectors.
			// So we just act like a non-removable floppy...
			//
			lpDevParams->dpDevType = 7; // insist this is a floppy
			lpDevParams->dpDevAttr = 1; // insist this is not removable

			// set # of cylinders to a small number, as the format
			// call for each cylinder does nothing in VDISK
			// FORMAT does not appear to use this
			lpDevParams->dpCylinders = 1;

			// user also wants a copy of the BPB
			fmemcpy((char far *)(&(lpDevParams->bpb)),
				(char far *)(&std_bpb), sizeof(std_bpb));
			return(TRUE);
		}

		// get access flag
		// sub-sub-function 0 returns 0 if disk is not formatted else 1 if it is
		// VDISK is always formatted.
		//
		case 0x867:
		{
			struct SpecialFunc far * lpSpecFuncs =
				(struct SpecialFunc far*)(buffer);

			// documentation says special function should ALWAYS be 0
			// see pg 38 DOS Prog Ref
			ASSERT(lpSpecFuncs->sfFunction == sfFUNC_NUM0);

			lpSpecFuncs->sfAccessFlg = sfACCESS_OK;
			return(TRUE);
		}

		case 0x868: // Sense media type (act like MS RAMDISK (v6.22): fail)
		default:
			break;
	}

	/* not a supported IOCTL call */
	memerr = DEVERR_BADCMD;
	return FALSE;
}

/*
	$DOCPROC:ParseCmdline

	Parse the command line arguments. Parses both switches and
	arguments as follows:

	Syntax:
	DEVICE=VDISK.SYS [DiskSize [SectorSize [NumEntries]]] [/E] [/A] [/S]
                   ^
	The lpCmdline begins after the name "VDISK.SYS".
*/
bool ParseCmdline(char far *lpCmdline)
{
	char	c;
	int	nArgCount = 0;

	/* validate parameter */
	ASSERT(lpCmdline);

	/* look for command line arguments */
	while(1)
	{
		/* skip whitespace to next switch */
		lpCmdline = MemSkipWhite(lpCmdline);
		c = *lpCmdline;

		/* at end of line? */
		if(c < ' ')
			break;

		/* cmdline options (switches) */
		if(c == '/')
		{
			/* get switch char */
			++lpCmdline;					/* skip '/' */
			c = MemToUpper(*lpCmdline++);

			switch(c)
			{
				/* use expanded (EMS) memory */
				case 'A':
					bUseEMS = TRUE;
					break;

				/* Use extended memory (or XMS if available) */
				case 'E':
					bExtended = TRUE;
					break;

				/* RAM disk should remain static, if possible */
				case 'S':
					bStatic = TRUE;
					break;

				/* verbose option */
				case 'V':
					bVerbose = TRUE;
					break;

				default:
					printf(szErrBadOpt, c);
					return FALSE;
			}
		}

		/* must be a cmdline argument */
		else
		{
			switch(nArgCount)
			{
				/* arg0 is disk sizeK */
				case 0:
					nDiskSizeK = MinMax(lpCmdline, 4, 32767);
					break;

				/* arg1 is bytes per sector */
				case 1:
					nBytesPerSector = MinMax(lpCmdline, 128, 512);
					if(nBytesPerSector != 128 && nBytesPerSector != 256)
						nBytesPerSector = 512;
					break;

				/* arg2 is # root directory entires */
				case 2:
					nRootEntries = MinMax(lpCmdline, 4, 512);
					break;

				/* too many arguments */
				case 3:
					printf(szErrBadArgs);
					return FALSE;
			}

			/* skip the argument */
			lpCmdline = MemSkipArg(lpCmdline);

			/* one more argument */
			++nArgCount;
		}

	} /* end of while */

	/* everything worked */
	return TRUE;
}

/*
	$DOCPROC:RdWrEMS

	Read from or write to the hEMS handle. Break read/write into 16K
	pages.

	Returns TRUE if worked, FALSE if failed.
*/
bool RdWrEMS(long lOffset, unsigned nLen, char far *lpBuffer, bool bRead)
{
	unsigned nLogicalPage;
	unsigned nPageOffset;
	unsigned nThisLen;

	while(nLen)
	{
		/* determine logical page and offset */
		nLogicalPage = (unsigned)_ldivu(lOffset, EMS_PAGESIZE);
		nPageOffset  = (unsigned)_lmod(lOffset, EMS_PAGESIZE);

		/* determine how much we can read from this page */
		nThisLen = EMS_PAGESIZE - nPageOffset;
		if(nThisLen > nLen)
			nThisLen = nLen;

		/* read from the EMS memory */
		if(MapUnmapEMSPage(hEMS, nLogicalPage, EMS_PAGE) != EMS_SUCCESS)
			return FALSE;

		/* read or write the memory */
		if(bRead)
			fmemcpy(lpBuffer, AddPtr(nEMSFrameSeg, nPageOffset), nThisLen);
		else
			fmemcpy(AddPtr(nEMSFrameSeg, nPageOffset), lpBuffer, nThisLen);

		/* on to next read/write */
		nLen -= nThisLen;
		lOffset += nThisLen;
		lpBuffer += nThisLen;
	}
	return TRUE;
}


#if DEBUG
/*
	$DOCPROC:AssertFail

	When an assertion fires, this function is called. Only defined if
	DEBUG is defined.
*/
void PASCAL AssertFail(int nFile, int nLine)
{
	panic("VDISK Internal Error %d.%d!\n", nFile, nLine);
}
#endif	/* DEBUG */


#if UNIT_TEST

/* for testing a 4K conventional disk */
char  aConventionalDisk[4096+16];

/*
	$DOCPROC:GetDiskSeg

	Used to get the segment of the Unit Test conventional disk.
*/
unsigned GetDiskSeg(void)
{
	ulong	lAddr = LinearAddress(aConventionalDisk);
	return (unsigned)(lAddr>>4) + 1;
}

int main(void)
{
	extern unsigned endseg;

	/* get a segment to a small conventional disk area */
	endseg = GetDiskSeg();

	/* return error code if failed. messages are displayed by test. */
	if(!RamDiskUnitTest(szUnitTestCmdline))
		return 1;

	/* no errors */
	return 0;
}
#endif	/* UNIT_TEST */
