;
;   $Id: INT16.ASM 1.4 2002/10/15 23:49:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
;  INT16.ASM
;
;  Copyright (c) 1994-2002 Datalight, Inc.
;  All Rights Reserved
;  Datalight Confidential and Proprietary
;
;  INT 16 - Keyboard BIOS Interrupt
;       Functions Supported
;               0  - Get a Key
;               1  - Check if Key available
;               2  - Check Keyboard Status
;               5  - Push a key into Buffer
;               10 - Get a Key
;               11 - Check if Key available

_TEXT Segment byte public 'CODE'

        assume cs:_TEXT, ds:_DATA

;
; BIOS INT 16 Interrupt Service Routine
;
        public _INT16_isr
_INT16_isr:
        push    ds
        push    bp
        mov     bp, sp
        push    bx

        mov     bx, Seg _DATA           ; Assume no error
        mov     ds, bx
        test    ah,0efh                 ; check if 0 or 10h
        jnz     not_0
;ah is 0 or 10h, wait for a character from the console
        mov     bx,_Int16BufTail
        jmp short GC1
GC2:
        sti                     ; Turn Interrupt On if no chars available
GC1:    cmp     bx, _Int16BufHead       ;  While( Char Not Avail)
        je      GC2                     ;  Do Nothing, until a char
                                        ;  is entered into the ring buffer
        mov     ax, Word Ptr [bx]       ;  Get Char
;increment the tail
        add     bx,2
        cmp     bx, Offset _DATA:_Int16Buffer + 32
        jne     IT1
        mov     bx, Offset _Int16Buffer
IT1:    mov     _Int16BufTail,bx
        jmp     Done

not_0:
        cmp     ah, 1                   ; Else if CharAvail Function
        je      is_1
        cmp     ah, 11H
        jnz     not_1
is_1:   mov     bx, _Int16BufTail
        cmp     bx, _Int16BufHead     ;set ZF according to availability of char
        mov     ax,[bx]               ;SCAN code value for return
        pushf
        pop     bx      ; Get the flags register to check the zero flag bit
        and     byte ptr [bp+8],NOT 40h ; Clear the Z flag on the stack
        and     bl,40h                  ; Leave only the Z flag bit
        or      byte Ptr [bp+8],bl      ; Put Flags onto stack for return
        jmp     Done

not_1:
        cmp     ah, 2                   ; Else if Get Status Flag Function
        jne     not_2
        mov     al, kb_stat             ; Get Status Flags
        jmp     Done

not_2:
        cmp     ah, 5                   ; Else if Int16PutChar Function
;There is no BIOS standard for reporting non supported calls in int 16
        jnz     done                    ; jmp if non supported call
        mov     ax, cx                  ; CH=Scan Code, CL=ASCII Character
        call    _Int16PutChar           ; returns CF = 1 if error
        mov     ax,500h                 ; indicate success
        jnc     done
;There was an error, return al = 1
        inc     al
Done:
        pop     bx
        pop     bp
        pop     ds
        iret

;
; Int 16 Put Char is a utility routine to put a char in to the int 16
; buffer.  Returns CF = 1 if buffer is full.
; Assumes that DS = _DATA
;
        public _Int16PutChar
_Int16PutChar:

        mov     bx, _Int16BufTail       ; If((Tail-Head)== 2) Then FULL
        sub     bx, _Int16BufHead
        cmp     bx, 2
        je      PC1

        mov     bx, _Int16BufTail       ; If(Tail == Start, Head == end) Then FUll
        cmp     bx, Offset _Int16Buffer
        mov     bx, _Int16BufHead
        jne     NotFull
        cmp     bx, Offset _Int16Buffer + 30
        je      PC1                     ; jmp if full
NotFull:
        mov     [bx], ax                ; put a character in the buffer
        add     bx,2
        cmp     bx, Offset _Int16Buffer + 32
        jne     IH1
        mov     bx, Offset _Int16Buffer
IH1:
        mov     _Int16BufHead,bx
        clc
        ret

PC1:    stc
        ret

_TEXT Ends

_DATA Segment public byte 'DATA'

        extrn _Int16BufHead:word
        extrn _Int16BufTail:word
        extrn _Int16Buffer:word
        extrn kb_stat:byte

_DATA Ends

        End
