;
;   $Id: MINIBIOS.INC 1.4 2002/10/15 23:49:14Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
comment ^--------------------------------------------------------------------
        MINIBIOS.INC rev 3.1

        Copyright (c) 1994-2002 by Datalight, Inc.
        All rights reserved.
	Datalight Proprietary and Confidential.

        Generic configuration to be customized for the specific target
        platform.

        This file contains general user-definable options for the
        miniBIOS.

        CHECK THAT ALL OPTIONS ARE DEFINED PROPERLY FOR THE DESIRED
        BIOS CONFIGURATION

 ===========================================================================^

;---------------------------------------------
; System Configuration Table Specification
;
;
;  The system model and submodel below are returned by int 15h function C0H.
;  If 386 or above, this must indicate which style A20 control system the
;  platform has.  The PS/2 style A20 control uses I/O port 92h to enable and
;  disable the A20 line.  The AT syle A20 control scheme uses the 8042 keyboard
;  controller.  For compatibility with himem.sys and emm386.exe, this must be
;  accurate.  Many modern systems have both the AT and PS/2 style A20 line
;  controllers.  If the system has both, set it up for the PS/2 style since it
;  is faster and then disable the 8042 A20 enable.
SystemModel     equ     0FCH
SystemSubModel  equ     04      ; Indicate PS/2 style A20 enable/disable
;SystemSubModel  equ     00      ; Indicate AT style A20 enable/disable
Features        equ     0       ;Place the correct features byte here.

;---------------------------------------------
; A20 gate support PS/2 style I/O Port 92h
;       or  AT style 8042 Keyboard Interface
;       set to 0 if AT compatible.
PS_2_Port92 equ 1   ; Yes, write to Port 92H to control A20

;---------------------------------------------
; EQUIPMENT LIST DEFINITION
;| 15-14  | 13   | 12   | 11-9    | 8    | 7-6     | 5-4 | 3-2   | 1    | 0       |
;| prntrs | modm | game | ser pts | rsvr | flp dsk | vid | mouse | 8087 | fd inst |
; Default value 0030H = No lpt's,no serial port,no floppy drives,
;       no disp, no 8087 co-proc.
;       Int14.asm will fill in the number of serial ports detected.
OriginalEquipmentList equ 0030h

;---------------------------------------------
; ROM_SCAN_START is the first segment that the miniBIOS
; searches for BIOS Extension.
;
ROM_SCAN_START  equ     0C000h

;---------------------------------------------
; ROM_SCAN_END is the last segment that the miniBIOS
; searches for BIOS Extension.
;
ROM_SCAN_END    equ     0FF00h

;------------- RAM TEST OPTIONS --------------
TEST_LIMIT      equ     0F000h ; Specify the RAM test ceiling SEGMENT
;TEST_LIMIT      equ     0A000h ; RAM test ceiling for PC compatibility

;Define below is required only if 386 or above with extended memory.
EXT_TEST_LIMIT  equ     4000000h        ;Extended RAM test ceiling in bytes

comment &
        DEFAULT_A20_ENABLE controls the initial state of the A20 line, 1
        indicates enabled, 0 indicates disabled.  Some early PC 8088
        applications relied on the wrap at the 1 megabyte boundary so most
        true PC compatible computers provide a means of disabling the A20 line
        to emulate the wrap at 1 megabyte.  This can slow down access to
        extended memory because any application accessing extended memory has
        to first enable the A20 line, then disable it when it is finished.

        Set this define to 1 unless it causes problems with some of those
        applications which rely on the wrap at 1 megabyte.

        This define is required only if the platform is a 386 or above.
&
DEFAULT_A20_ENABLE      equ     1

comment &
RAM_TEST controls the type of memory test done by the minibios.

        0 eliminates all RAM testing except the minimum required to determine
        the top of RAM.  This is the fastest RAM test.

        1 performs a more thorough test of RAM, checking for stuck bits and
        stuck address lines but it is the slowest option.

        Options 0 and 1 both cause a non-destructive RAM test to be performed
        which preserves any information which may have been in RAM.  This is
        useful if portions of memory are battery backed up SRAM.

        2 causes a thorough RAM test, like option 1 but without preserving the
        original information.  This test concludes by filling the RAM with 0s.
        This test is much faster than option 1 but slower than option 0.

&
RAM_TEST        equ     2

;---------------------------------------------
; APM_SUPPORTED is a flag that defines if the miniBIOS APM
; code is linked in: 1 = include APM; 0 = DoNot include APM
;
APM_SUPPORTED   equ     0

;---------------------------------------------
; ReportNonSupportedCalls is a flag that enables or disables miniBIOS
; reporting of non supported interrupts.  NonSupported calls show the
; address of the call and the registers at the time the non Suppported call
; occurred.
;  0 = disable NonSupported Call display
;  1 = enable NonSupported Call display
;
ReportNonSupportedCalls  equ     1

;---------------------------------------------
; Define Hardware Interrupts 08-0F and 70h-77h supported during
; INT 15h Function 87H memory moves for the extended memory
; VDISK support or extended memory ROM-DISK support.
; HWD_INTS_Supported is a flag that enables or disables
; interrupts when the processor is in protected mode for extended memory
; moves.  Adds 496 bytes to the size of the miniBIOS.
;  0 = disable Hardware Interrupts during protected mode INT 15h FUNC 87H
;  1 = enable Hardware Interrupts  during protected mode INT 15h FUNC 87H
;  For strict PC-AT compatibility, specify 0
;  This define only required if 386 or above.
HWD_INTS  equ     1

comment &
---------------------------------------------
 COM_PORT_DIVIDEND used by console_init
        The com ports can be configured to use an external clock to get
        standard 8250 UART divisors or they can be configured to use the
        processor clock.  For a PC compatible system, this should be 115200.
        Read comments regarding Baud_Rate_Gen_COMCLK.
&
COM_PORT_DIVIDEND       equ     115200

comment &
---------------------------------------------
 Console_Baud_Rate used by console_init
&
Console_Baud_Rate equ   19200

;---------------------------------------------
; Parms  - The Parms  parameter passed to console_init
;       Sets stop, data and parity bits
;       Possible values are 8N1 (value = 3) or E71 (value = 26)
Console_Parms equ 3

;---------------------------------------------
; Console_Port  - 1 - 4 specifies COM1 - COM4, to be used as the console port.
; This COMM port will not be available through int 14h services and will not
; be counted on the equipment list.
Console_Port equ 1              ;COM1

;---------------------------------------------
; SupportPostCodes  - Flag to turn on and off Post Code Support

SupportPostCodes equ    1

;---------------------------------------------
; CPU Specification
;       86  = 8088, 8086, V20, V30 CPU
;       186 = V25, V35, 80188, 80186
;       386 = 386 or above CPU

Target_CPU equ  386

; end of minibios.inc

