//
//   $Id: BACKUP.C 1.8 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

   BACKUP.C  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	This program will backup files to any DOS drive.  If the
	destination is floppies, it will prompt for the next disk.

	Look for "todo:" in the code to find unimplemented features.

********************************************************************/
#define __FILENUM__ 1

// includes
#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dosver.h"
#include "backup.h"
#include "bacvol.h"
#include "datalght.h"
#include "dostruct.h"
#include "dldosio.h"
#include "fatworld.h"
#include "abortlfn.h"

// messages
#define BACKUP 1		// indicate we want the backup messages
#include "dosmsg.c"

// prototypes
#include "backup.pt"
#include "bacstore.pt"
#include "bacvol.pt"

// cmdline switches
bool	swAppend;			// append to existing backup set (/A)
bool	swHidden;			// backup hidden files? (/H)
bool	swLogFile;			// make a log file? (/L)
bool	swArchive;			// only backup those with archive bit set? (/M)
bool	swSubdir;			// recurse into subdirecties? (/S)
#if DEBUG
int	swVolSize;		// set the volume size in Kbytes (debugging only, /V)
int	swMemSize;		// set memory size, in K bytes (debugging only, /X)
#endif
bool	swPrompt = TRUE;	// normally, wait for user to insert disks (/Y)

// the log file
char	szLogFile[MAXLONGPATH];
char	szLogExt[] = "LOG";
FILE *fpLogFile;

// Source info
char	szSourcePath[MAXLONGPATH];			// example C:\DEV\ROMDOS\*.*
char	szSourceMask[MAXBAKNAME];			// example *.*
int	rootLength;								// length of root string

// Destination Volume info
int	backupDrive;							// destination drive for backup
char	szBackupBaseName[MAXBAKNAME];		// example DL970306

// record totals (# of files, and bytes)
ulong numFiles;
ulong fileBytes;

bool fIsFat32World;

void DisplaySignOnMessage( void )
{
	char * szSignOn = "BACKUP %s %s\n%s\n";


	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	Backup program.  See usage for how it works.
*/

int main(int argc, char *argv[])
{
	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Abort if there are LFNs on the command line
	//
	AbortForLFNArgs();

	// display signon
	DisplaySignOnMessage();

	// process command-line arguments into global switches
	// and backup name.  Will error and show usage if problem
	// in cmdline args.
	ProcessCmdline(argc, argv);

	// initialize the country info
	InitCountryInfo();

	// initialize file storage system.  can only fail if
	// there is not enough memory.
	if(!InitFileStore())
		Error(errMsgNoMem, ERR_NOMEM);

	// start from scratch for file statistics
	numFiles = fileBytes = 0L;

	// create initial volume name, no vol # yet
	InitVolume();

	// now, go through each file/directory and back up the stuff
	Backup(szSourcePath);

	// mark that this is the last volume
	MarkLastVolume();

	// done, flush all remaining files to disk
	FlushFileStore();

	// done, free the memory
	KillFileStore();

	// list as complete
	SetFileExt(szVolumeName, "");
	printf(msgComplete, szVolumeName);
	printf(msgStatistics, numFiles, fileBytes, curVolume);

	// close the .LOG file, complete with statistics
	if(fpLogFile)
	{
		// output final stats
		sprintf(szLogFile, msgComplete, szVolumeName);
		fputs(szLogFile, fpLogFile);
		sprintf(szLogFile, msgStatistics, numFiles, fileBytes,
			curVolume);
		fputs(szLogFile, fpLogFile);

		// done with .LOG file
		fclose(fpLogFile);
	}

	// everything worked
	return 0;
}

/*
	Scan for switches, and determine source and destination
	names.  Will show help if an errors encountered.
*/
void ProcessCmdline(int argc, char *argv[])
{
	int	i, switchLen;
	char	*pSwitch;
	char	*pLogFile;
	bool	srcFound, dstFound, moreFound;
	char	szTempName[MAXBAKNAME];

	// no arguments, just show help
	if(argc == 1)
		Usage();

	// loop through, looking for switches
	for(i=1; i<argc; ++i)
	{
		// any switches in this cmdline argument?
		pSwitch = argv[i];
		while(pSwitch)
		{
			// look for next switch
			pSwitch = strchr(pSwitch, '/');
			if(!pSwitch)
				break;

			// default switch length (example: /s)
			switchLen = 2;

			// got a switch?
			switch(toupper(pSwitch[1]))
			{
				// append to existing file
				case 'A':
					swAppend = TRUE;
					break;

				// backup hidden files too
				case 'H':
					swHidden = TRUE;
					break;

				// recurse into subdirectories
				case 'C':
				case 'Y':
					swPrompt = FALSE;
					break;

				// recurse into subdirectories
				case 'S':
					swSubdir = TRUE;
					break;

				// record log file
				case 'L':

					// we'll be recording a log file
					swLogFile = TRUE;

					// determine log file name
					pLogFile = pSwitch + 2;
					if(*pLogFile == ':')
					{
						// determine full log file name
						if(!dlExpandPath(pLogFile+1, szLogFile))
							Error(errMsgCreateLog, ERR_LOGFILE);

						// determine switch length
						switchLen = strlen(pSwitch);
					}
					break;

				// only back up those with archive bit set
				case 'M':
					swArchive = TRUE;
					break;

#if DEBUG
				// set hard-coded volume size
				case 'V':
					swVolSize = atoi(pSwitch+2);
					switchLen = 2 + strlen(pSwitch+2);
					break;

				// set hard-coded memory size
				case 'X':
					swMemSize = atoi(pSwitch+2);
					switchLen = 2 + strlen(pSwitch+2);
					break;
#endif

				// does user want help?
				case '?':
					Usage();
					break;

				default:
					printf(errMsgBadOpt, pSwitch[1]);
					printf(msgUsage);
					exit(ERR_BADOPT);
			}

			// remove switch from the cmdline, remember to copy NUL
			memmove(pSwitch, pSwitch+switchLen, strlen(pSwitch+switchLen)+1);

		}	// end of while (pSwitch)

	}	// end of for(i<argc)

	// find source path and destination drive
	srcFound = dstFound = moreFound = FALSE;
	for(i=1; i<argc; ++i)
	{
		// may skip arg if switches were first
		if(*argv[i] == 0)
			continue;

		// getting source name?
		if(!srcFound)
		{
			// make sure path exists
			if(!dlExpandPath(argv[i], szSourcePath))
				Error(errMsgBadSource, ERR_BADPATH);

			// find just the "name" part
			CopyNamePart(szSourceMask, szSourcePath);

			// determine length of the "root" part of the name
			rootLength = strlen(szSourcePath) - strlen(szSourceMask);

			// let routine know we found the source name
			srcFound = TRUE;
		}

		// getting destination drive?
		else if(!dstFound)
		{
			// get drive letter
			backupDrive = toupper(*argv[i]) - 'A';

			// make sure drive is valid
			if(!IsValidDrive(backupDrive))
				Error(errMsgBadDest, ERR_BADPATH);

			// all backup files for today will have the same name
			// example "DL970328"
			GetBackupBaseName(szBackupBaseName);

			dstFound = TRUE;
		}

		// too many arguments found
		else
			moreFound = TRUE;
	}

	// wrong # of args (too many or too few)
	if(!srcFound || !dstFound || moreFound)
	{
		printf(errMsgWrongNumber);
		printf(msgUsage);
		exit(ERR_BADOPT);
	}

	// sending output to log file
	if(swLogFile)
	{
		// no user supplied name? make our own
		if(*szLogFile == 0)
		{
			// create a name for the .LOG file
			GetBackupBaseName(szTempName);
			SetFileExt(szTempName, szLogExt);

			// make the name absolute
			if(!dlExpandPath(szTempName, szLogFile))
				Error(errMsgCreateLog, ERR_LOGFILE);
		}

		// don't allow .LOG file on same drive as backup!
		ASSERT(*szLogFile && *(szLogFile+1) == ':');
		if(*szLogFile - 'A' == backupDrive)
			Error(errMsgCreateLog, ERR_LOGFILE);

		// open the .LOG file
		fpLogFile = fopen(szLogFile, "w");
		if(!fpLogFile)
			Error(errMsgCreateLog, ERR_LOGFILE);
	}

}

/*
	Display simple error message and exit with error code.
*/
void Error(const char *pMsg, int errCode)
{
	// display error message
	printf(errMsg);
	printf(pMsg);

	ErrorExit(errCode);
}

void ErrorExit(int errCode)
{
	// free memory handles, in case we have XMS
	KillFileStore();

	// exit with code
	exit(errCode);
}

/*
	Display usage.
*/
void Usage(void)
{
	printf(msgUsage);
	exit(ERR_USAGE);
}

/*
	Get just the base name part of the backup name.  Uses
	the Datalight signature, plus current date.

	Example: "DL970306"
*/
void GetBackupBaseName(char *pName)
{
	struct date curDate;

	// get current date and put it into the name in the form
	getdate(&curDate);
	sprintf(pName, "DL%02d%02d%02d", curDate.da_year%100,
		curDate.da_mon, curDate.da_day);
}

/*
	Backup from the source to the dest.  Keep track of the
	findfirst/next.  This function is recursive, and will
	modify the "pSourcePath" parameter as it recurses into
	the directory tree.
*/
void Backup(char *pSourcePath)
{
	struct ffblk ff;		// file find block
	int err;
	static int level = 0;

	// validate parameter
	ASSERT(pSourcePath);

	// backup files that match the spec
	err = findfirst(pSourcePath, &ff, (swHidden ? A_ANYFILE : A_NORMAL));
	while(!err)
	{
		// set name to found name
		if(!SetName(pSourcePath, ff.ff_name))
			goto TooDeep;

		// store the file
		StoreFile(pSourcePath, ff.ff_attrib, ff.ff_fdate, ff.ff_ftime);

		// go on to next file in this directory
		err = findnext(&ff);
	}

	// only handle subdirectories if user says so
	if(!swSubdir)
		return;

	// look for all directories
	if(!SetName(pSourcePath, "*.*"))
		goto TooDeep;

	// any subdirectories to recurse into?
	err = findfirst(pSourcePath, &ff, A_SUBDIR);
	while(!err)
	{
		// is this a directory? (only recurse into subdirs)
		// ignore the "." and ".." directory entries.  To do this,
		// we cheat and just see if the dir begins with '.'
		if((ff.ff_attrib & A_SUBDIR) && *ff.ff_name!='.')
		{
			// don't go too deep into directories
			if(level+1 >= MAXDEEP)
			{
TooDeep:
				Error(errMsgPathTooDeep, ERR_TOODEEP);
				return;
			}

			// one more level deep
			++level;

			// add that subdirectory to pathname
			if(!SetName(pSourcePath, ff.ff_name))
				goto TooDeep;

			// make sure user wildcard name is still tacked on
			if(!AddName(pSourcePath, szSourceMask))
				goto TooDeep;

			// now backup THAT directory
			Backup(pSourcePath);

			// back out again
			if(!RemoveName(pSourcePath))
				ABORT();

			// one level up
			--level;
			ASSERT(level >= 0);
		}

		// on to next directory
		err = findnext(&ff);
	}
}
