//
//   $Id: BACKUP.H 1.7 2002/10/15 23:51:46Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

   BACKUP.H  created 03/07/97  by Drew Gislason

	Copyright (c) 1997-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Include file for the BACKUP and RESTORE programs.

**********************************************************/

#include "..\include\datalght.h"
#include "..\include\dlpath.h"
#include "..\include\dlmem.h"
#include "..\include\dldate.h"

//
// structures
//

// this structure is recorded for each file
struct FileRecord
{
	// note: some routines count on "next" field being first!
	ulong		next;					// 00h next FileRecord on this volume
	unsigned	sig;					// 04h signature
	uchar		moreVolumes;		// 06h more volumes in backup
	uchar		attr;					// 07h file attr
	unsigned	date;					// 08h date file was last modified
	unsigned	time;					// 0Ah time file was last modified
	ulong		size;					// 0Ch total size of file
	ulong		startOffset;		// 10h starting offset in file
	int		nameSize;			// 14h size of filename after header
	int		startVol;			// 16h starting volume for this file
	uchar		reserved[8];		// 18h for future use
	char		szFileName[1];		// 20h name of file, complete with path
};
#define FILE_RECORD_SIG		0xBABA	// BABA ROM-DOS!
#define MIN_RECORD_SIZE		(2*1024)
#define MIN_OUT_SIZE		512
#define MORE_DATA			0x11
#define MORE_VOLUMES		0x22
#define LAST_VOLUME		0x33

// size of backup filename "A:DL970306.001";
#define MAXBAKNAME	16
#define MAXLONGPATH	80		// eventually 260 characters
#define ROOTLEN		 3		// length of "C:\"
#define DATESIZE		10		// size of date or time string

/*
   BIOS PARAMETER BLOCK (BPB)
   Description: This structure is stored on disks in
   sector 0, AT OFFSET 11d.

   Reference: Adv MS-DOS, p 269  -or-  DOS Ref, p 222
*/
struct BPB /*()*/
{
   unsigned   sec_size;      /* 00 */
   uchar      sec_per_clus;  /* 02 */
   unsigned   res_sec;       /* 03 */
   uchar      num_fats;      /* 05 */
   unsigned   dir_ent;       /* 06 */
   unsigned   num_sec;       /* 08 */
   uchar      media;         /* 0A */
   unsigned   fat_size;      /* 0B */
   unsigned   sec_per_trak;  /* 0D  Floppy/Harddisk extensions */
   unsigned   sides;         /* 0F       ignored by DOS        */
   long       hidden;        /* 11    for device use only      */
   long       lnumsec;       /* 15 */
};

// amount of memory required for disk I/O storage
#define DISKIO_BUFFSIZE ((unsigned)32*1024)

// the dir entry is not a file if it has ANY of these bits set
// 11xx xxxx   not defined
// xxx1 xxxx   directory entry
// xxxx 1xxx   volume label
#define NOTFILE 0xD8

// skip the "A:" part of a full path
#define SKIPDRIVE 2

// how many levels deep into dirs can we go?
#define MAXDEEP	25

// error codes
#define ERR_USAGE			1
#define ERR_BADOPT		2
#define ERR_NOMEM			3
#define ERR_TOODEEP		4
#define ERR_BADPATH		5
#define ERR_CORRUPT		6
#define ERR_LOGFILE		7
#define ERR_WRITEPROT	8		/* fixed disk: volume write protected */
#define ERR_NOTFOUND		9		/* fixed disk: last vol not found for append */
#define ERR_USERABORT	19
#define ERR_ASSERT		99

// attribute for overwrite prompt
#define A_OVERWRITE (A_READONLY | A_HIDDEN | A_SYSTEM | A_ARCHIVE)

// assert error messages
extern char errMsgInternal[];
extern char errMsgPleaseCall[];

// stuff out of backup
extern bool	swArchive;
extern bool swAppend;
extern int	rootLength;

// variables shared between modules
extern bool	swPrompt;
extern int	backupDrive;
extern char	szBackupBaseName[MAXBAKNAME];
#if DEBUG
extern int swVolSize;		// set the volume size, used for debugging
extern int swMemSize;
#endif

// messages used in other modules besides BACKUP.C
#if BACKUP
extern char szVolumeSignature[];
extern char msgInsertDisk[];
extern char msgInsertLastVol[];
extern char msgReturn[];
extern char msgOtherReturn[];
extern char errMsgFormatFailed[];
extern char errMsgFromThisSet[];
extern char errMsgWriteProtect[];
extern char errMsgNoFormat[];
extern char errMsgNotLastVol[];

// for .LOG file
extern char msgLogLine[];
extern char msgContinued[];
extern char msgDirectoryOf[];
extern char szLogFile[];
extern FILE *fpLogFile;
extern ulong numFiles;
extern ulong fileBytes;

#endif
