//
//   $Id: PATH.C 1.5 2002/10/15 23:53:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	PATH.C

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
	Datalight Confidential and Proprietary

	This file contains the path routines used by programs
	such as ATTRIB and TREE.

**********************************************************/
#include <dir.h>
#include <dos.h>
#include <stdio.h>
#include <string.h>


#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlpath.h"


char P_oldcwd[MAXDIR];
char P_drive[MAXDRIVE];
char P_dir[MAXDIR];
char P_name[MAXFILE+MAXEXT];

int path_rel_to_abs(char *relname, char *absname);
void add_path(char *path, char *sub_dir);
void del_path(char *path);


#if 0
char absname[67];
char relname[67];

void main(void)
{
	while(1)
	{
		printf("Enter relative name: ");
		gets(relname);
		if(!path_rel_to_abs(relname, absname))
			printf("failed!\n");
		else
			printf("Changed '%s' to '%s'\n", relname, absname);
	}
}
#endif

/*
	Return the last character in this string.
*/
char lastchar(char *s)
{
	int len = strlen(s);
	return len ? s[len-1] : 0;
}

/*
	Get the current working directory for this drive.
	Returns a string in the form of "F:\SUBDIR"
	Returns TRUE if worked, FALSE if failed.
*/
int INT_get_cwd(char * cwd, int drive)
{
	extern int errno;
	bool fSuccessful;

	errno = 0;
	fSuccessful = dlGetCurrentDirectory( drive+1, &cwd[3] );
	if( !fSuccessful )
	{
		errno = dlCheckDOSError();
   	return FALSE;
   }

	/* set up the drive part of the path */
	*cwd++ = drive+'A';
	*cwd++ = ':';
	*cwd   = '\\';

	return TRUE;
}

/*
	Parse a path into drive/path/filename.  Puts these
	pieces into the "P_drive", "P_dir", "P_name" global
	strings.  Changes directory into the new dir.  Saves
	the old directory in "P_oldcwd".

	P_oldcwd is guaranteed to be on the same drive as
	P_drive which may or may not be the current drive.

	Returns TRUE if worked, FALSE if failed.
*/
int determine_dir(char *path)
{
	extern char P_drive[];
	extern char P_dir[];
	extern char P_name[];
	extern char P_oldcwd[];
	int iTemp;

	int flags;
	char newpath[MAXPATH];
	char ext[MAXEXT];

	/* split up the path */
	flags = IntlFileNameSplit(path, P_drive, P_dir, P_name, ext);

	/* if name is similar to "CON:", remove the colon */
	if(lastchar(P_name) == ':')
		*(P_name+strlen(P_name)-1) = 0;

	/* determine drive */
	if(flags&DRIVE)
		strupr(P_drive);
	else
	{
		strcpy(P_drive, "A:");
		dlGetCurrentDrive( &iTemp );
		*P_drive = (uchar)(iTemp + 'A');
	}

	/* remember the old CWD for the given drive */
	if(!INT_get_cwd(P_oldcwd, *P_drive - 'A'))
		return FALSE;

	/* no directory part */
	if(!(flags&(DIRECTORY|FILENAME|EXTENSION)))
		/* don't need to chdir */;

	/* change into new directory */
	else if(dlSetCurrentDirectory(path))
	{
		*P_name = 0;
		*ext = 0;
	}

	/* remove name part and try again */
	else if(*P_dir)
	{
		/* make a new string */
		strcpy(newpath, P_drive);
		strcat(newpath, P_dir);

		/* no trailing '\\' unless it's the root */
		if(strlen(newpath) > 3 && lastchar(newpath) == '\\')
			newpath[strlen(newpath)-1] = 0;

		/* if we can't change into this, it's not a valid directory */
		if(!dlSetCurrentDirectory(newpath))
			return FALSE;
	}

	/* convert directory to an absolute one */
	INT_get_cwd(newpath, *P_drive - 'A');
	strcpy(P_dir, newpath+2);

	/* if no name or extention, assume they are all */
	strcat(P_name, ext);
	if(*P_name == 0)
		strcpy(P_name, "*.*");

	return TRUE;
}

/*
	Given the "current" drive/path as defined by P_drive
	and P_path, and the names that's passed, create an
	absolute pathname.

*/
void make_path(char *dst, char *name)
{
	extern char P_drive[];
	extern char P_dir[];

	strcpy(dst, P_drive);
	strcat(dst, P_dir);
	if(lastchar(P_dir) != '\\')
		strcat(dst, "\\");
	strcat(dst, name);
}

/*
	Given a name, get the full name.  Returns FALSE if
	the path or filename does not exist.
*/
int path_rel_to_abs(char *name, char *absname)
{
	extern char P_drive[];
	extern char P_dir[];
	extern char P_name[];
	extern char P_oldcwd[];

	/* get the absolute name from this relative name */
	if(!determine_dir(name))
		return FALSE;
	dlSetCurrentDirectory(P_oldcwd);

	/* return the absolute name */
	make_path(absname, P_name);
	return TRUE;
}


void add_path(char *path, char *sub_dir)
{
	if(lastchar(path) != '\\')
		strcat(path, "\\");
	strcat(path, sub_dir);
}

void del_path(char *path)
{
	int i;
	for(i=strlen(path); i > 0; --i)
	{
		if(path[i] == '\\')
		{
			path[i] = '\0';
			return;
		}
	}
#if 0
	if(path[1] == ':')
		path[2] = '\0';
	/* If not "\" then don't bother */
#endif
}

