//
//   $Id: ROMDISK.C 1.5 2002/10/15 23:54:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/***************************************************************
	$DOCMODULE:ROMDISK.C

	Copyright (c) 1988-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Usage = ROMDISK <filespec> [outfile] {options}

	Options
		/D<Seg> Define RXE DataSegName (default="DATA")
		/E      Do not place extended address in Intel HEX files
		/F##    Set hexadecimal fill byte (default=FFH)
		/H[#]   Output to a hex file ROM-DISK.HEX (# = Starting Address)
		/I[#]   Output an image file ROM-DISK.IMG (# = Starting Address)
	   /N      Do not run RXEOPTIM to optimize the RXE file
		/O      Omit timestamp from volume label
      /R##    Chose a hexadecimal interrupt # for RXEs (default=90)
		/S      Recurse into subdirectories
		/T      Test mode only - don't build ROM disk
		/V"str" Volume Label (default="ROMDISK", /V- means none)
		/Z#     Set decimal Sector SiZe (default = 128 bytes)

	Example
		C>ROMDISK ..\romdir disk2 /s/iE000

	$DOCHISTORY
	11/01/1988 dg  Created by Drew Gislason
****************************************************************/
#include <assert.h>
#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dosver.h"
#include "datalght.h"
#include "fatworld.h"
#include "abortlfn.h"

#define Assert(x) assert(x)

#define MAX_SEG_NAME_LEN 32

#define MAXENTRIES	512			/* maximum total entries */
unsigned char FILLBYTE = 0xFF;	/* dead space is filled with this byte */
#define SECID 0xAA55

char signon[] = "ROMDISK %s %s\n%s\n\n";
char USAGE[] =
	"Usage = ROMDISK <filespec> [outfile] {options}\n"
	"\n"
	"Options\n"
#if 0
	// compressed disk driver removed because it's too slow
	// and buggy.
	"\t/C\tCompress the disk image (requires special driver)\n"
#endif
	"\t/D<Seg> Define RXE DataSegName(default \"DATA\")\n"
	"\t/E\tDo not place extended address in Intel HEX files\n"
	"\t/F#\tSet fill byte (default=FFH)\n"
	"\t/H[#]\tOutput to a hex file ROMDISK.HEX (# = Starting Address)\n"
	"\t/I[#]\tOutput an image file ROMDISK.IMG (# = Starting Address)\n"
	"\t/N\tDo not run RXEOPTIM to optimize the RXE file\n"
	"\t/O\tOmit timestamp from volume label\n"
	"\t/R#\tChose a hexadecimal interrupt # for RXEs (default=90H)\n"
	"\t/S\tRecurse into subdirectories\n"
	"\t/T\tTest mode only - don't build ROM disk\n"
	"\t/V\"str\"\tVolume Label (default \"ROMDISK\", /V- means none)\n"
	"\t/Z#\tSet decimal Sector SiZe (default = 128 bytes)\n"
	"\n"
	"Example\n"
	"\tC>ROMDISK ..\\romdir disk2 /s/iE000\n";

bool fIsFat32World;
bool fCallRXEOPTIM = FALSE;

/*
	[]-------------------------[]
	 |		General Defines		|
	[]-------------------------[]
*/

#define PATHSIZE	82	/* enough to hold a complete path/filename */

#define	OMIT_NONE	0					// Include all timestamps
#define	OMIT_VOL		1					// Omit timestamp on volume label

/*
	[]-------------------------[]
	 |			New Types			|
	[]-------------------------[]
*/

#define F_RXE		0x80
#define F_SUBDIR	0x10
#define A_VOLUME	0x08

/*
	The list of files to add to the ROM disk.  Note, the
	lower 8 bits are the file "attribute" as seen by a
	directory.

	flags
	xxxx xxxx 1xxx xxxx   RXE File
	nnnn nnnn xxxx xxxx   RXE interrupt # (only on RXEs)
	xxxx xxxx xxx1 xxxx   Subdirectory
	xxxx xxxx xxxx 1xxx   Volume Label
*/
struct FLIST
{
	char name[PATHSIZE];	/* full pathname of file */
	ulong size;				/* size of the file - for determining disk size */
   unsigned time;			/* file date */
   unsigned date;			/* file time */
   unsigned start;		/* starting cluster of file */
	unsigned flags;		/* see above definitions */
	int level;				/* 0-n levels deep */
};

/*
   DIRECTORY ENTRY
   Description: Each item in the directory looks like this

   Directory attribute:
   xxxx xxxx xxxx xxx1    Read only
   xxxx xxxx xxxx xx1x    Hidden
   xxxx xxxx xxxx x1xx    System
   xxxx xxxx xxxx 1xxx    Volume label
   xxxx xxxx xxx1 xxxx    Subdirectory
   xxxx xxxx xx1x xxxx    Archive

   Reference: Adv MS-DOS, p 184  -or-  DOS Ref, p 248
*/
struct DIR
{
   char     name[8];         /* 00 file name                */
   char     ext[3];          /* 08 */
   uchar    attr;            /* 0B file attribute           */
   uchar    reserved[10];    /* 0C */
   unsigned time;            /* 16 file date                */
   unsigned date;            /* 18 file time                */
   unsigned start;           /* 1A starting cluster of file */
   ulong    size;            /* 1C file size in bytes       */
};

/*
   BIOS PARAMETER BLOCK (BPB)
   Description: This structure is stored on disks in
   sector 0, AT OFFSET 11H.

   Reference: Adv MS-DOS, p 269  -or-  DOS Ref, p 222
*/
struct BPB
{
   unsigned   sec_size;      /* 00 */
   uchar      sec_per_clus;  /* 02 */
   unsigned   res_sec;       /* 03 */
   uchar      num_fats;      /* 05 */
   unsigned   dir_ent;       /* 06 */
   unsigned   num_sec;       /* 08 */
   uchar      media;         /* 0A */
   unsigned   fat_size;      /* 0B */
   unsigned   sec_per_trak;  /* 0D  Floppy/Harddisk extensions */
   unsigned   sides;         /* 0F       ignored by DOS        */
   long       hidden;        /* 11    for device use only      */
   long       lnumsec;       /* 15 */
};

/*
	[]----------------[]
	 |		Globals		|
	[]----------------[]
*/

int sw_image=TRUE;			/* switch, image or HEX */
int sw_subdir;					/* switch, recurse into subdirs */
int sw_testmode;				/* test mode - don't output anything */
int sw_extended=TRUE;		/* extended addresses */
int sw_omit=OMIT_NONE;		/* omit timestamps? */
int sw_volume=TRUE;			/* normally, we want a volume label */
unsigned sec_size = 128;	/* user can set sector size */
#if COMPRESS
int sw_compress;				/* compress the disk image */
#endif

#define SEC_PER_CLUS	1
#define CLUS_SIZE		(sec_size*SEC_PER_CLUS)
#define DIR_PER_SEC	(sec_size/sizeof(struct DIR))
#define FAT_PER_SEC	(sec_size*2/3)
#define SEC_FORMULA	(1L+dir_secs+fat_secs+(start-2)*SEC_PER_CLUS)
#define SIZE_FORMULA	SEC_FORMULA * sec_size

char sysid[] = "DLRDISK";  /* System ID */

struct BPB boot_rec =
{
	128,				/* Bytes Per Sector						(computed) */
	1,					/* Sectors per Cluster                         */
	1,					/* Boot Sector (# of sectors)                  */
	1,					/* Number of Copies of FAT                     */
	112,				/* Number of Root Directory entries	(computed) */
	720,				/* Total Sectors on Disk				(computed) */
	0xF8,				/* Format ID                                   */
	2,					/* Number of Sectors per FAT			(computed) */
	0xF000,			/* Number of Sectors per Track                 */
	1					/* Number of sides                             */
};

char Volume[12]= "ROM-DISK   ";		/* default Volume Label */
char DEFNAME[]	= "ROM-DISK";			/* default name for ROMDISK */
char RXENAME[] = "RXE_CVT.EXE";		/* name of EXE to RXE converter */
char RXEOPTIM[] = "RXEOPTIM.EXE";	/* name of EXE to RXE optimizing program */
#if COMPRESS
char CMPNAME[] = "COMPDISK.EXE";		/* name of IMG to CMP converter */
#endif
int rxe_int = 0x90;						/* interrupt of RXE program */
int got_rxe;
struct FLIST *filelist;					/* list of files (MAXENTRIES of 'em) */
char filespec[PATHSIZE];				/* where to get files from */
char outfile[PATHSIZE];					/* output file */
unsigned numfiles;						/* total # of files in list */
unsigned start_addr = 0xC000;			/* address for start of ROMDISK */
char * fatbuff;							/* the FAT, retained in memory */
char szDataSeg[MAX_SEG_NAME_LEN] = "DATA";	/* RXE Data Seg Name */


char P_oldcwd[MAXDIR];	/* in a form so CHDIR can work */
char P_drive[MAXDRIVE];	/* A:, B: */
char P_dir[MAXDIR];		/* \ABC\ */
char P_name[16];			/* TEMP*.C */

unsigned root_sectors;		/* # of sectors for root entries */
unsigned fat_sectors;		/* # of FAT sectors */
unsigned data_sectors;		/* # of data sectors on disk */
unsigned subdir_sectors;	/* total sectors in subdirs */
unsigned total_sectors;		/* total sectors on disk */

#define MAX12BITFAT	0xFF0		/* maximum clusters for 12 bit FAT */

#define DTOTAL  0
#define DCOMP	 1
#define DBOOT   2
#define DFAT    3
#define DROOT   4
#define DDIR    5
#define DRXE    6
#define DUSER   7
#define DAVAIL  8
#define DSECTOR 9
#define MAXSTAT 10

struct stat
{
	long bytes;
	unsigned number;
};
struct stat stat[MAXSTAT];

/*
	[]-------------------[]
	 |		Prototypes		|
	[]-------------------[]
*/
#include "romdisk.pt"

/*
	[]-------------------[]
	 |		The Code			|
	[]-------------------[]
*/

/*
	Usage = ROMDISK <filespec> [outfile] {options}
*/
void main(int argc,char *argv[])
{
	char pathname[PATHSIZE];

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Abort if there are LFNs on the command line
	//
	AbortForLFNArgs();

	/* sign-on and help */
	if(argc<2)
	{
		showhelp();
		exit(1);
	}

	/* process the command line arguments */
	process_args(argc, argv);

	/* get memory for the list of files */
	filelist = calloc(MAXENTRIES,sizeof(struct FLIST));
	if(!filelist)
	{
		printf("Error! Insufficient memory for FILELIST\n");
		exit(2);
	}
	/* get complete list of files */
	build_file_list(filespec,0);

	/* calculate sizes of subdirs, fat, etc... */
	determine_sizes();

	/* actually build the romdisk */
	if(*outfile == 0)
		strcpy(outfile, DEFNAME);
	RecommendExt(outfile, sw_image ? ".IMG" : ".HEX");
	if(!sw_testmode)
	{
		build_rom_disk(outfile);
#if COMPRESS
		if(sw_compress)
			compress(outfile);
#endif
	}

	/* determine the path we got the disk from */
	strcpy(pathname, P_drive);
	strcat(pathname, P_dir);
	if(lastchar(pathname) != '\\')
		strcat(pathname, "\\");
	strcat(pathname,P_name);

	/* display where files(s) came from and where they are going */
	printf("\nROM Disk %-8s  ", (sw_image ? "Image" : "HEX File"));
	if(sw_volume)
		printf("Volume '%s'\n", Volume);
	else
		printf("No volume label\n");
	printf("Built from         %s\n"
			 "Placed in          %s",
			pathname, outfile);
	if(got_rxe)
		printf(" at segment %04X",start_addr);
	printf("\n\n");

	/* display the results - statistics like chkdsk */
	display_stats();

	exit(0);
}

// compress removed as it didn't work so well
#if COMPRESS
/*
	Compress a ROM-DISK image.  Uses sub-program IMPLODE.EXE.
*/
void compress(char *outfile)
{
	FILE * fp;
	char infile[PATHSIZE];

	/* create source and destination filenames */
	strcpy(infile, outfile);
	RecommendExt(outfile, ".CMP");

	if(spawnlp(0,CMPNAME,"",infile,outfile,NULL))
	{
		printf("\n"
				 "Error! Failed to compress ROMDISK image!\n"
				 "Make sure %s is in the current directory or PATH.\n",CMPNAME);
		exit(12);
	}

	/* delete the old image */
	unlink(infile);

	/* set up statistics */
	fp = fopen(outfile,"rb");
	fseek(fp,0L,2);
	stat[DCOMP].bytes  = ftell(fp);
	fclose(fp);
	stat[DCOMP].number = 100 -
		((stat[DCOMP].bytes * 100L) / stat[DTOTAL].bytes);
}
#endif

/*
	Given a fullname "C:\PATH\NAME.EXT", return the
	name part "NAME.EXT"
*/
char *namepart(char *fullname)
{
	char * p;
	p = strrchr(fullname,'\\');
	if(p)
		return p+1;
	return fullname;
}

/*
	Given an entry in the file list, make a DIR
	structure out of it.
*/
void list_to_dir(struct DIR *dp, int n)
{
	char ename[12];

	/* copy just the filename (not path) in expanded form */
	if(filelist[n].flags & A_VOLUME)
		strcpy(ename, namepart(filelist[n].name));
	else
		dot_to_ename(ename, namepart(filelist[n].name));
	strcpy(dp->name, ename);
	dp->date = filelist[n].date;
	dp->time = filelist[n].time;
	dp->start = filelist[n].start;
	if(filelist[n].flags & F_SUBDIR)
		dp->size = 0L;
	else
		dp->size  = filelist[n].size;
	dp->attr = filelist[n].flags & 0x7F;	/* and off RXE flag */
	memset(dp->reserved, 0, sizeof(dp->reserved));
}

/*
	Given an entry in the file list, show the output
	path and filename.
*/
void show_outname(int n)
{
	static int skiproot = -1;
	char * p;
	static char rxestr[] = " -- RXE (uses interrupt 90H)";
	char tmpnum[4];

	if(skiproot<0)
	{
		p = strrchr(filelist[0].name,'\\');
		skiproot = (int)(p ? p-filelist[0].name : 0);
	}

	/* display the fact that a file is an RXE */
	if(filelist[n].flags & F_RXE)
	{
		p = rxestr;
		sprintf(tmpnum,"%02X",filelist[n].flags >> 8);
		memmove(rxestr+24, tmpnum, 2);
	}
	else
		p = "";
	printf("%s%s\n",filelist[n].name+skiproot,p);
}

void make_dir(struct DIR *dp, char * name, unsigned start, int n)
{
	char ename[12];

	dot_to_ename(ename, name);
	strcpy(dp->name, ename);
	dp->date = filelist[n].date;
	dp->time = filelist[n].time;
	dp->start= start;
	dp->size = 0L;
	dp->attr = F_SUBDIR;
	memset(dp->reserved, 0, sizeof(dp->reserved));
}

/*
 *  Do NOT call if entry has no clusters!
 */
unsigned getnumclus(int i)
{
	unsigned numclus = 0;
	int i2;

	/* some basic assumptions */
	Assert(filelist[i].start);
	Assert(data_sectors);

	i2 = i + 1;
	while(numclus == 0)
	{
		if(i2 >= numfiles)
			numclus = (data_sectors+2) - filelist[i].start;
		else if(filelist[i2].start > filelist[i].start)
			numclus = filelist[i2].start - filelist[i].start;
		++i2;
	}
	return numclus;
}

/*
	We now have all the information we need to make the ROM
	disk.  There are a number of other reasons why the disk
	might not be made:

	1) Cannot create the output file
	2) HEX file extends past end of 8086 memory
	3) RXE extends past end of 8086 memory
	4) Not enough memory for FAT
*/
void build_rom_disk(char *outfile)
{
	unsigned i, n;			/* index variables */
	unsigned cluster;		/* cluster # for counting */
	unsigned numclus;		/* number of clusters in this file/subdir */
	char * bootsec;		/* a buffer for the boot sector */
	char * databuff;		/* a temporary buffer for the data */
	unsigned bufsize;		/* the buffer size for data */
	FILE * fpout;			/* ROM disk output */
	FILE * fpin;			/* input file(s) */
	unsigned parent;		/* parent cluster # */
	long maxlen;			/* maximum length of this file */
	unsigned dirnum;		/* directory entry # */

	/* an identification sequence for a Datalight ROM disk */
	static char jmp[] = { 0xEB, 0x28, 0x90 };
	static struct DIR dir;

	/* allocate the memory for the FAT and boot sector */
	Assert((long)fat_sectors * sec_size < 0x10000L);
	fatbuff = malloc(fat_sectors * sec_size);
	bootsec = malloc(sec_size);
	if(!fatbuff || !bootsec)
	{
		printf("Error! Insufficient memory for FAT\n");
		exit(10);
	}
	memset(fatbuff, 0, fat_sectors * sec_size);

	/* does HEX file extend past end of memory? */
	if(!sw_image &&
		((long)start_addr*0x10 + (long)total_sectors*sec_size > 0xFFFFFL))
	{
		printf("\n\007Error! HEX file starting at %04XH extends past end of 8086 memory!\n"
				 "Use /I (Image) or create a smaller ROM disk.\n\n",start_addr);
		display_stats();
		exit(11);
	}

	/* create the output file */
	fpout = fopen(outfile,"wb");
	if(!fpout)
	{
		printf("Error! Cannot create output file '%s'\n",outfile);
		exit(13);
	}

	/* build the BOOT sector */
	memset(bootsec, FILLBYTE, sec_size);
	memmove(bootsec, jmp, 3);
	memmove(bootsec+3, sysid, sizeof(sysid));
	memmove(bootsec+11, &boot_rec, sizeof(struct BPB));
	*(unsigned *)(bootsec+sec_size-sizeof(unsigned)) = SECID;

	/* write out the BOOT sector */
	write_to_romdisk(bootsec, sec_size, fpout);
	free(bootsec);

	/* build the FAT */
	cluster = 0xFF00 + boot_rec.media;
	set_fat_entry(0, cluster);
	set_fat_entry(1, 0xFFFF);
	cluster = 2;
	for(i=0; i<numfiles; ++i)
	{
		if(filelist[i].start)
		{
			numclus = getnumclus(i);
			Assert((long)numclus + (cluster-2) <= (long)data_sectors);

			while(numclus > 1)
			{
				set_fat_entry(cluster, cluster+1);
				--numclus;
				++cluster;
			}
			set_fat_entry(cluster++, 0xFFFF);
		}
	}

	/* write out the FAT */
	write_to_romdisk(fatbuff, fat_sectors * sec_size, fpout);
	free(fatbuff);

	/* write out the root DIR */
	n = 0;
	for(i=0; i<boot_rec.dir_ent; ++i)
	{
		while(n < numfiles && filelist[n].level > 0)
			++n;

		/* found a root entry */
		if(n < numfiles)
			list_to_dir(&dir, n);

		/* past the end of the root entries - zero means unused */
		else
			memset(&dir, 0, sizeof(struct DIR));

		write_to_romdisk(&dir, sizeof(struct DIR), fpout);
		++n;
	}

	/* get space for data buffer */
	databuff = maxmalloc(&bufsize);

	/* write the DATA part */
	for(i=0; i<numfiles; ++i)
	{
		/* write out the subdirectory */
		if(filelist[i].flags & F_SUBDIR)
		{
			/* determine parent */
			if(filelist[i].level == 0)
				parent = 0;
			else
			{
				n = i;
				while(filelist[n].level >= filelist[i].level)
					--n;
				Assert(filelist[n].flags & F_SUBDIR);
				parent = filelist[n].start;
			}

			/* write out "." and ".." */
			make_dir(&dir, ".", filelist[i].start, i);
			write_to_romdisk(&dir, sizeof(struct DIR), fpout);
			make_dir(&dir, "..", parent, i);
			write_to_romdisk(&dir, sizeof(struct DIR), fpout);

			/* write out the rest of the directory */
			dirnum = 2;
			n = i+1;
			while(filelist[n].level > filelist[i].level)
			{
				/* is this an entry */
				if(filelist[n].level == filelist[i].level+1)
				{
					list_to_dir(&dir, n);
					write_to_romdisk(&dir, sizeof(struct DIR), fpout);
					++dirnum;
				}
				++n;
			}

			/* write out any "unused" entries to fill out cluster */
			memset(&dir, 0, sizeof(struct DIR));
			while((dirnum * sizeof(struct DIR)) % sec_size)
			{
				write_to_romdisk(&dir, sizeof(struct DIR), fpout);
				++dirnum;
			}
		}

		/* normal file */
		else
		{
			/* don't open volume labels */
			if(filelist[i].flags & A_VOLUME)
				fpin = NULL;

			/* read the file */
			else
			{
				show_outname(i);
				fpin = fopen(filelist[i].name,"rb");
				if(!fpin)
				{
					printf("Error! Cannot open file '%s'\n",filelist[i].name);
					exit(14);
				}
			}

			maxlen = filelist[i].size;
			while(maxlen)
			{
				n = bufsize;
				if(maxlen < bufsize)
					n = (unsigned)maxlen;
				if(fread(databuff, 1, n, fpin) != n)
				{
					printf("\nError reading from '%s'!\n",filelist[i].name);
					exit(9);
				}
				write_to_romdisk(databuff, n, fpout);
				maxlen -= n;
			}

			/* close the input file */
			if(fpin)
				fclose(fpin);

			/* write out any "unused" portion to fill out cluster */
			if(filelist[i].size % sec_size)
			{
				n = sec_size - (unsigned)(filelist[i].size % sec_size);
				memset(databuff, FILLBYTE, n);
				write_to_romdisk(databuff, n, fpout);
			}
		}
	}

	/* HEX files need a terminating record */
	if(!sw_image)
		write_hex_done(fpout);

	/* delete all converted RXE files */
	for(i=0; i<numfiles; ++i)
	{
		if(filelist[i].flags & F_RXE)
			unlink(filelist[i].name);
	}

	fclose(fpout);
}

void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

/*
	display the opening screen
*/
void showhelp(void)
{
	DisplaySignOnMessage();
	printf(USAGE);
}

void process_args(int argc, char *argv[])
{
	int i;
	char * p;
	char c;
	char * pstart;		/* start of switch */
//	int found;
	int volarg=0;		/* volume argument */
	int vollen;

	/* look through all command line arguments for switches */
	for(i=1; i<argc; ++i)
	{
		// do NOT consider the volume label to contain any
		// cmdline options.  For example, /v "Hello" should skip
		// past the "Hello" when looking for other options.
		if(i==volarg)
			continue;

		while(1)
		{
			p = argv[i];

			/* find a switch character */
			while(*p && *p != '/')
				++p;

			/* no switch found, break */
			if(*p == 0)
				break;

			/* found a switch, process it */
			pstart = p++;
			c = toupper(*p);	/* get the switch letter in UPPERCASE */
			++p;					/* move on to any switch arguments */
			switch(c)
			{
// compressed disk driver no longer implemented
#if COMPRESS
				/* /C compress disk image */
				case 'C':
					sw_compress = TRUE;
					break;
#endif

				/* /D (Define the RXE DataSegName (Default is "DATA") */
				case 'D':
					// get the new data segment name
					strncpy(szDataSeg, p, sizeof(szDataSeg));
					szDataSeg[sizeof(szDataSeg)-1] = 0;

					// don't allow invalid characters in segment name
					if(strpbrk(szDataSeg, " \\/\";[]{}:<>") != NULL)
					{
						printf("\nError: Data Segment name ('%s') is invalid.\n",
								szDataSeg);
						exit(29);
					}

					// the entire parameter is used for segment name
					p += strlen(p);
					break;

				/* /E (no extended address in HEX file) */
				case 'E':
					sw_extended = FALSE;
					break;

				/* /F (set fill byte) */
				case 'F':
					FILLBYTE = atoh(p);
					while(isxdigit(*p))
						++p;
					break;

				/* /H[#] (HEX) or /I[#] (Image) */
				case 'H':
				case 'I':
					sw_image = (c=='I');		/* I=true, H=false */
					if(*p && isxdigit(*p))
					{
						start_addr = atoh(p);
						while(isxdigit(*p))
							++p;
					}
					break;

				//
				//
				case 'N':
					fCallRXEOPTIM = TRUE;
					break;


				/* /O (Omit volume label timestamps) */
				case 'O':
					sw_omit = OMIT_VOL;
					break;

				/* /R (set RXE initial interrupt #) */
				case 'R':
					rxe_int = atoh(p);
					while(isxdigit(*p))
						++p;
					break;

				/* /S (Recurse into Subdirectories) */
				case 'S':
					sw_subdir = TRUE;
					break;

				/* /T (test mode - no output file) */
				case 'T':
					sw_testmode = TRUE;
					break;

				/* /V "label" Volume label */
				case 'V':
					// does user want NO volume label?
					if( argv[i][2] == '-' )
					{
						// we DON'T want a volume label
						sw_volume = FALSE;

						// most args are just 2 chars the "/v"
						// account for '-' so it too gets removed from args
						++p;
					}

					// argv[i] MUST be "/v" with no following characters.
					// argv[i+1] MUST be a valid argument and will be
					// the volume label, even it was intended to be a
					// switch.  Do it right, user!
					else if( argv[i][2] || ((i+1) >= argc ))
					{
						showhelp();
						printf("\nError: Use \"quotes\" with the /V option!\n");
						exit(20);
					}

					// we are turning ON the volume label
					else
					{
						sw_volume = TRUE;

						// copy the volume from the cmdline to here
						volarg = i+1;
						for(vollen=0; vollen < 11 && argv[volarg][vollen]; ++vollen)
							Volume[vollen] = argv[volarg][vollen];

						// make sure to pad it to 11 spaces
						while(vollen < 11)
							Volume[vollen++] = ' ';
						Volume[vollen] = 0;
					}

					break;

				/* /Z# (Set sector size) */
				case 'Z':
					sec_size = atoi(p);
					if((sec_size % 128) || !sec_size || sec_size>512)
					{
						printf("Invalid sector size %d, use 128, 256 or 512\n");
						exit(21);
					}
					while(isdigit(*p))
						++p;
					break;

				default:
					printf("Invalid option '%s'\n\n",argv[i]);
				case '?':
					showhelp();
					exit(22);
			} /* end of switch */

			/* remove the option from command line */
			strcpy(pstart, p);
		}
	}	/* end of for loop through argv[] */

	/* find the filelist and output filename */
	for(i=1; i<argc; ++i)
	{
		if(i==volarg)
			continue;

		/* first argument is filelist, next (if any) is output filename */
		if(*argv[i])
		{
			if(*filespec == 0)
				strcpy(filespec,argv[i]);
			else if(*outfile == 0)
				strcpy(outfile, argv[i]);
		}
	}

	/* we must have a filespec (outfile is optional) */
	if(*filespec==0)
	{
		showhelp();
		printf("\nError: You specified no files to place in ROM-DISK!\n");
		exit(23);
	}

// compression no longer implemented
#if COMPRESS
	if(sw_compress && !volarg)
		strcpy(Volume,"COMPRESSED ");

	/* cannot compress hex files (only image files) */
	if(sw_compress && !sw_image)
	{
		printf("Cannot compress Intel HEX ROM-DISKs - use /I (Image)\n");
		showhelp();
		exit(24);
	}
#endif

}

/*
	Given a filespec, build a list of files.  This function
	is recursive.
*/
void build_file_list(char *filespec, int level)
{
	struct ffblk ff;
	char * root;
	int err;
	char * p;

	root = malloc(PATHSIZE);
	Assert(root);

	if(!level)
	{
		numfiles = 0;
		err = !determine_dir(filespec);
		INT_set_cwd(P_oldcwd);
		if(err)
		{
			printf("Invalid filespec '%s'\n",filespec);
			exit(5);
		}
		strcpy(root,P_drive);
		strcat(root,P_dir);
		if(lastchar(root) != '\\')
			strcat(root,"\\");		/* always end root dirs in "\" */
		strcat(root, P_name);

	}
	else
		strcpy(root, filespec);

	/* find files in this directory first */
	err = findfirst(root,&ff,0x7);

	/* if processing root dir, add in the Volume Label */
	if(!level && sw_volume)
	{
		newname(root, Volume);
		strcpy(filelist[numfiles].name,root);
		filelist[numfiles].size = filelist[numfiles].start = 0;
		filelist[numfiles].flags = A_VOLUME;
		filelist[numfiles].level = 0;

		// If the user requested to omit time stamps on the
		// volume label, set them to 0 (1-1-1980 12:00am)
		if( sw_omit == OMIT_VOL )
		{
			filelist[numfiles].date = 0;
			filelist[numfiles].time = 0;
		}
		else
		{
			filelist[numfiles].date = get_sysdate();
			filelist[numfiles].time = get_systime();
		}

		++numfiles;
	}

	while(!err && numfiles<MAXENTRIES)
	{
		/* copy the pertinent info into our filelist */
		newname(root, ff.ff_name);
		strcpy(filelist[numfiles].name,root);
		filelist[numfiles].size = ff.ff_fsize;
		filelist[numfiles].date = ff.ff_fdate;
		filelist[numfiles].time = ff.ff_ftime;
		filelist[numfiles].flags = ff.ff_attrib;
		filelist[numfiles].level = level;
		++numfiles;

		err = findnext(&ff);
	}

	/* then look for subdirectories */
	if(sw_subdir)
	{
		newname(root, P_name);
		err = findfirst(root,&ff,0x17);
		while(!err && numfiles<MAXENTRIES)
		{
			/* if this is a subdirectory */
			if((ff.ff_attrib & F_SUBDIR) && strcmpi(ff.ff_name,".") &&
				strcmpi(ff.ff_name,".."))
			{
				/* copy the pertinent info into our filelist */
				newname(root, ff.ff_name);
				strcpy(filelist[numfiles].name,root);
				filelist[numfiles].size = 0L;
				filelist[numfiles].date = ff.ff_fdate;
				filelist[numfiles].time = ff.ff_ftime;
				filelist[numfiles].flags = ff.ff_attrib;
				filelist[numfiles].level = level;
				++numfiles;

				/* go into this subdirectory */
				strcat(root,"\\");			// add a trailing backslash */
				newname(root, P_name);		// and namespec
				build_file_list(root, level+1);
				p = strrchr(root,'\\');
				Assert(p);
				*p = 0;							// remove name and backslash
			}

			err = findnext(&ff);
		}
	} /* done processing subdirs */

	/* if we are completely done, make sure we found at least 1 file */
	if(level == 0 && !numfiles)
	{
		newname(root, P_name);
		printf("No files to copy from '%s'\n",root);
		printf("Disk not created!!\n");
		exit(6);
	}

	/* free the space for root string */
	free(root);
}

void newname(char *path, char *name)
{
	char * p = strrchr(path,'\\');
	Assert(p);
	strcpy(p+1,name);
}

int lastchar(char *s)
{
	int len = strlen(s);
	return len ? s[len-1] : 0;
}

int INT_get_cur_dr(void)
{
	return (bdos(0x19,0,0) & 0xFF);
}


void INT_set_cur_dr(unsigned int drive)
{
	bdos(0x0E,drive,0);
}

/*
	Get the current working directory for this drive.
	Returns a string in the form of "F:\SUBDIR"
	Returns TRUE if worked, FALSE if failed.
*/
int INT_get_cwd(char * cwd, int drive)
{
	/* get the directory */
	if(getcurdir(drive+1,cwd+3))
		return FALSE;

	/* set up the drive part of the path */
	*cwd++ = drive+'A';
	*cwd++ = ':';
	*cwd   = '\\';

	return TRUE;
}

/*
	Returns TRUE if worked, FALSE if failed.
*/
int INT_set_cwd(char * path)
{
#if 0
	return INT_dir(0x3B, path);
#endif
	return !(chdir(path));
}

#if 0
/*
	mkdir, rmdir, chdir

	Returns TRUE if worked, FALSE if failed.
*/
int INT_dir(unsigned func, char *path)
{
	union REGS r;

	r.h.ah = func;
	r.x.dx = (unsigned)path;
	intdos(&r,&r);
	if(r.x.cflag)
		return FALSE;
	return TRUE;
}
#endif


/*
	Parse a path into drive/path/filename.  Puts these
	pieces into the "P_drive", "P_dir", "P_name" global
	strings.  Changes directory into the new dir.  Saves
	the old directory in "P_oldcwd".

	P_oldcwd is guaranteed to be on the same drive as
	P_drive which may or may not be the current drive.

	Returns TRUE if worked, FALSE if failed.
*/
int determine_dir(char *path)
{
	extern char P_drive[];
	extern char P_dir[];
	extern char P_name[];
	extern char P_oldcwd[];

	int flags;
	char newpath[MAXPATH];
	char ext[MAXEXT];

	/* split up the path */
	flags = fnsplit(path, P_drive, P_dir, P_name, ext);

	/* if name is similar to "CON:", remove the colon */
	if(lastchar(P_name) == ':')
		*(P_name+strlen(P_name)-1) = 0;

	/* determine drive */
	if(flags&DRIVE)
		strupr(P_drive);
	else
	{
		strcpy(P_drive, "A:");
		*P_drive = INT_get_cur_dr() + 'A';
	}

	/* remember the old CWD for the given drive */
	if(!INT_get_cwd(P_oldcwd, *P_drive - 'A'))
	{
		return FALSE;		/* not valid drive */
	}

	/* no directory part */
	if(!(flags&(DIRECTORY|FILENAME|EXTENSION)))
		/* don't need to chdir */;

	/* change into new directory */
	else if(INT_set_cwd(path))
	{
		*P_name = 0;
		*ext = 0;
	}

	/* remove name part and try again */
	else if(*P_dir)
	{
		/* make a new string */
		strcpy(newpath, P_drive);
		strcat(newpath, P_dir);

		/* no trailing '\\' unless it's the root */
		if(strlen(newpath) > 3 && lastchar(newpath) == '\\')
			newpath[strlen(newpath)-1] = 0;

		/* if we can't change into this, it's not a valid directory */
		if(!INT_set_cwd(newpath))
		{
			return FALSE;	/* not a valid path on this drive */
		}
	}

	/* convert directory to an absolute one */
	INT_get_cwd(newpath, *P_drive - 'A');
	strcpy(P_dir, newpath+2);

	/* if no name or extention, assume they are all */
	strcat(P_name, ext);
	if(*P_name == 0)
		strcpy(P_name, "*.*");

	return TRUE;
}

/*
	Remove a file of the same name, but different extension
	from the file list.
	Returns -1 if not found, returns new value for "original"
	if found.
*/
int remove_from_list(int original, char *ext)
{
	int found, n;
	char name[PATHSIZE];

	/* search in this directory for the name and new extension */
	found = -1;
	strcpy(name, filelist[original].name);
	ForceExt(name, ext);
	for(n=0; n<numfiles && found<0; ++n)
	{
		/* have we found a match? Won't delete the original */
		if(original != n && strcmpi(name, filelist[n].name)==0)
			found = n;
	}

	/* remove the entry from the filelist */
	if(found >= 0)
	{
		memmove(filelist+found, filelist+found+1,
			(numfiles-found-1)*sizeof(struct FLIST));
		--numfiles;
		memset(filelist+numfiles, 0, sizeof(struct FLIST));

		/* everything's shifted down by 1, maybe this has too! */
		if(original > found)
			--original;
	}
	else
		original = -1;		/* not found */

	return original;
}

/*
	Mark all RXEs - this will also remove the associated
	.MAP file names from the file list.
*/
void mark_rxes(void)
{
	int i, found;
//	int map;

	for(i=0; i<numfiles; ++i)
	{
		/* we have found an .RXE */
		if(ext_cmp(filelist[i].name,"RXE"))
		{
			/* remove the .MAP from the file list */
			found = remove_from_list(i, ".MAP");
			if(found < 0)
			{
				printf("RXE convert failed! '%s' not found!\n",filelist[i].name);
				exit(20);
			}
			else
				i = found;

			/* remove any EXEs of the same path/name */
			found = remove_from_list(i, ".EXE");
			if(found >= 0)
				i = found;

			/* mark the file as RXE */
			filelist[i].flags |= F_RXE;
			got_rxe = TRUE;
		}
	}
}

/*
	Actually call the RXE converter - this will make
	a new name.
*/
void rxe_convert(int n, unsigned seg)
{
	char options[8];
	char segment[8];
	char outfile[PATHSIZE];
	char rxefile[PATHSIZE];
	char exefile[PATHSIZE];
	struct ffblk ff;

	/* set up command line parameters */
	sprintf(options, "/Q/I%02X", rxe_int);
	ForceExt(filelist[n].name, ".RXE");
	sprintf(segment, "%04X", seg);
	strcpy(outfile, filelist[n].name);
	ForceExt(outfile, ".EXE");

	/* call the RXE converter */
	if(spawnlp(0,RXENAME,"",options,filelist[n].name,szDataSeg,segment,outfile,NULL))
	{
		printf("\nError! Failed to convert RXE!\n"
					"Make sure RXE_CVT and RXE_INIT are in the PATH.\n");
		exit(21);
	}


	//  build the path to the EXE file
	//
	//  NOTE: extensions are reversed for the ROMDISK process
	//
	strcpy(exefile, filelist[n].name );


	//  build the path to the RXE file
	//
	//  NOTE: extensions are reversed for the ROMDISK process
	//
	strcpy(rxefile, filelist[n].name );
	ForceExt(rxefile, ".EXE");


	if( fCallRXEOPTIM )
	{
		//  Call the RXE optimizer
		//
		if(spawnlp(0,RXEOPTIM,"",exefile, rxefile, NULL) )
		{
			printf("\nError! Failed to optimize RXE!\n"
						"Make sure RXEOPTIM.EXE are in the PATH.\n");
			exit(21);
		}
	}

// We were setting the file size based on the old file size, not the
//   converted version that the comment mentions.
//
// 	if(findfirst(outfile, &ff, 0))
//
	/* set the new filesize based on resulting size */
	if(findfirst(rxefile, &ff, 0))
	{
		printf("\nError! RXE_CVT failed to create '%s'!\n",outfile);
		exit(22);
	}
	filelist[n].size = ff.ff_fsize;
}

#define MAXSIZE 0xFFF8L
void toolarge(long bytes, char *msg)
{
	unsigned suggested_size = sec_size;
	while(bytes >= MAXSIZE)
	{
		suggested_size *= 2;
		bytes /= 2;
	}
	printf("%s too large for current sector size!  Use %d or more (/Z%d)\n",
		msg,suggested_size,suggested_size);
	exit(11);
}

/*
	Determine the size of the disk.  This includes FAT size,
	root dir size, # of clusters, etc..
*/
void determine_sizes(void)
{

	extern unsigned _psp;
	extern struct BPB boot_rec;

	unsigned root_entries;		/* # of root entries (including subdir names) */
	unsigned sectors;				/* sectors in this subdir */
	unsigned entries;				/* entries in this subdir */
	unsigned cluster;
	unsigned segment;
	long bytes;
	long tmp_data_sectors;
	long lsectors;
	int i, n;

	/* mark all RXEs */
	mark_rxes();

	/* determine root DIR sectors */
	root_entries = 0;
	for(i=0; i<numfiles; ++i)
	{
		if(filelist[i].level == 0)
			++root_entries;
	}
	root_sectors = (root_entries * sizeof(struct DIR)) / sec_size;
	if((root_entries * sizeof(struct DIR)) % sec_size)
		++root_sectors;
	root_entries = (root_sectors * sec_size) / sizeof(struct DIR);

	/* determine # of sectors in subdirectories */
	subdir_sectors = 0;
	for(i=0; i<numfiles; ++i)
	{
		/* count the entries in this subdir */
		if(filelist[i].flags & F_SUBDIR)
		{
			entries = 2;	/* always at least "." and ".." */

			n = i+1;
			while(filelist[n].level > filelist[i].level)
			{
				if(filelist[n].level == filelist[i].level+1)
					++entries;
				++n;
			}

			sectors = (entries * sizeof(struct DIR)) / sec_size;
			if((entries * sizeof(struct DIR)) % sec_size)
				++sectors;
			filelist[i].size = (long)sectors;
			subdir_sectors += sectors;

			/* update statistics */
			stat[DDIR].bytes += (long)sectors * sec_size;
			++stat[DDIR].number;
		}
	}

	/* determine data sectors */
	tmp_data_sectors = 0L;
	for(i=0; i<numfiles; ++i)
	{
		/* we have a file (not a subdir) */
		if((filelist[i].flags & F_SUBDIR) == 0)
		{
			/* convert RXEs so we know the size */
			/* note: this is not the REAL conversion - used only for size */
			if(filelist[i].flags & F_RXE)
				rxe_convert(i, 0);

			lsectors = filelist[i].size / sec_size;
			if(filelist[i].size % sec_size)
				++lsectors;
			/* if a file is too big for this sector size, report it! */
			if(lsectors >= MAXSIZE)
				toolarge(lsectors, "File");

			tmp_data_sectors += lsectors;

			/* update statistics */
			if(filelist[i].flags & F_RXE)
			{
				stat[DRXE].bytes += lsectors * sec_size;
				++stat[DRXE].number;
			}
			else if((filelist[i].flags & A_VOLUME) == 0)
			{
				stat[DUSER].bytes += lsectors * sec_size;
				++stat[DUSER].number;
			}
		}
	}

	/* if the # of sectors for a disk is too big, report it! */
	tmp_data_sectors += subdir_sectors;
	if(tmp_data_sectors >= MAXSIZE)
		toolarge(tmp_data_sectors, "Data sectors");
	data_sectors = (unsigned)tmp_data_sectors;

	/* determine fat sectors */
	if(data_sectors > MAX12BITFAT)
		bytes = (long)2 * ((long)data_sectors + 2);
	else
		bytes = (((long)data_sectors+2+1) / 2) * 3;
	fat_sectors = (unsigned)(bytes / sec_size);
	if(bytes % sec_size)
		++fat_sectors;

	/*
		FATs cannot be over 64K!
		With 1spc this is about 16Meg at 512 byte sectors
	*/
	if((long)fat_sectors * sec_size >= MAXSIZE)
		toolarge((long)fat_sectors * sec_size, "FAT");

	/* determine starting cluster of each file/subdir */
	cluster = 2;
	for(i=0; i<numfiles; ++i)
	{
		/* files with no data have no starting cluster */
		if(filelist[i].size)
			filelist[i].start = cluster;

		/* compute size of subdir */
		if(filelist[i].flags & F_SUBDIR)
			cluster += (unsigned)filelist[i].size;

		/* compute size of file */
		else
		{
			/* do the RXE conversion here to get REAL size */
			if(filelist[i].flags & F_RXE)
			{
				segment = 1+fat_sectors+root_sectors+(cluster-2);
				segment = (unsigned)(((long)sec_size * segment) >> 4);
				segment += start_addr;
				rxe_convert(i, segment);
				ForceExt(filelist[i].name, ".EXE");
				filelist[i].flags |= ((unsigned)rxe_int << 8);

				/* the next RXE will use a different interrupt */
				++rxe_int;
			}

			cluster += (unsigned)(filelist[i].size / sec_size);
			if(filelist[i].size % sec_size)
				++cluster;
		}
	}

	/* determine spaced used by disk overhead */
	stat[DBOOT].bytes  = sec_size;
	stat[DBOOT].number = 1;
	stat[DFAT].bytes   = (long)fat_sectors * sec_size;
	stat[DFAT].number  = fat_sectors;
	stat[DROOT].bytes  = (long)root_sectors * sec_size;
	stat[DROOT].number = root_sectors;

	/* determine total disk size in sectors */
	total_sectors = 1 + fat_sectors + root_sectors + data_sectors;
	stat[DSECTOR].number = total_sectors;
	stat[DSECTOR].bytes = sec_size;
	stat[DTOTAL].bytes = (long)total_sectors * sec_size;
	stat[DTOTAL].number = 1;

	/* always 0 bytes available */
	stat[DAVAIL].bytes = 0;
	stat[DAVAIL].number = 1;

	/* set up the BPB */
	boot_rec.sec_size = sec_size;
	boot_rec.dir_ent  = root_entries;
	boot_rec.num_sec  = total_sectors;
	boot_rec.fat_size = fat_sectors;
}

void display_stats(void)
{
	int i;

	static char *stat_str[] =
	{
		"bytes total ROM disk size",
		"bytes AFTER compression (%u%% savings)",
		"bytes in boot sector",
		"bytes in %u FAT sectors",
		"bytes in root directory",
		"bytes in %u directories",
		"bytes in %u RXE file(s)",
		"bytes in %u user file(s)",
		"bytes available on disk",
		"bytes in each of %u sectors"
	};

	/* display warning of incomplete disk */
	if(numfiles == MAXENTRIES)
		printf("Warning!  Not all files were placed on disk!\n"
				 "Maximum entries %d\n",MAXENTRIES);

	for(i=0;i<MAXSTAT;++i)
	{
		if(stat[i].number)
		{
			printf("%9ld ",stat[i].bytes);
			printf(stat_str[i],stat[i].number);
			printf("\n");

			if(i==DAVAIL)
				printf("\n");
		}
	}
}


/*
	Write the data to the romdisk - this may go to an image
	or a HEX file.  Assumes all writes are sequential,
	immediately following the last write.
*/
void write_to_romdisk(void *buffer, unsigned size, FILE *fp)
{
	if(sw_image)
		fwrite(buffer, 1, size, fp);
	else
		write_hex(buffer, size, fp);
}


#if 0
void print_find(struct ffblk *ff,int ignore)
{
	char date[20],time[20];

	/* print the name */
	printf("  %-12s",ff->ff_name);

	/* print size */
	printf("%6ld ",ff->ff_fsize);

	/* print date and time */
	timetoa(time,ff->ff_ftime);
	datetoa(date,ff->ff_fdate);
	printf(" %s %s",time,date);
	if(ignore)
		printf("    (entry exists) -- ignored");
	printf("\n");
}


/*
		[]-------------------------[]
		 |									|
		 |		GENERIC UTILITIES		|
		 |									|
		[]-------------------------[]
*/


/*
	the buffer "s" must be 80 chars long.
*/
void get_string(char *s)
{
	int len,i;

	if(response<0)
		gets(s);
	else {
		len = read(response,s,78);
		if(len<=0)
			*s = NULL;
		else {
			for(i=0;i<len;++i)
				if(s[i] == '\r') {
					lseek(response,-1L * (len-i-2), 1);
					break;
					}
			s[i] = '\0';
			}
		printf("%s\n",s);
		}
}


/*
	Returns date in string s, which must be at
	least 14 chars.
*/
void datetoa(char *s, unsigned date)
{
	static char *month[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

	#define MONTH ((date>>5)&0xF)-1
	#define DAY   (date&0x1F)
	#define YEAR  1980+(date>>9)

	sprintf(s,"%s-%02d-%d",month[MONTH],DAY,YEAR);
}


/*
	Time  xxxxx | xxxxxx | xxxxx  (hour|min|sec/2)
	Date  xxxxxxx | xxxx | xxxxx  (year|mon|day)

	Returns time in string s, which must be at
	least 8 chars.
*/
void timetoa(char *s, unsigned time)
{
	#define HOURS   (time>>11)
	#define MINUTES (time>>5)&0x3F
	#define SECONDS (time&0x1F)<<1

	int hours=HOURS;
	sprintf(s,"%2d:%02d%s",(hours%12==0 ? 12 : hours%12), MINUTES,
		hours>=12 ? "pm" : "am");
}
#endif

/*
	Compare the extension on the filename to the one given.
	Returns TRUE if they are identical, FALSE if not.
	Ignores case.
*/
int ext_cmp(char *fullname, char *ext)
{
	char * p;

	/* find the extension on the fullname */
	p = strrchr(fullname, '\\');
	if(!p)
		p = fullname;
	p = strchr(p, '.');

	/* ignore "." in extension, if any */
	if(*ext == '.')
		++ext;

	/* if no extension or extension does not match, return FALSE */
	if(!p || strcmpi(p+1, ext)!=0)
		return FALSE;

	return TRUE;
}

/*
	We recomment the following extension.  Basically, if an
	extension already exists, use it.  If not, append the
	given extension. The ext may be either form ".EXT" or
	"EXT".
*/
void RecommendExt(char *pName, char *pExt)
{
	char *p;

	// validate parameters
	Assert(pName && *pName);
	Assert(pExt && *pExt);

	// find the start of the extension
	p = FindExt(pName);
	Assert(p);

	// only put new extension on it if there's not already one there
	if(*p == 0)
		ForceExt(pName, pExt);
}

/*
	Force the file extension on "name" to be "ext".
	The ext may be either form ".EXT" or "EXT".
*/
void ForceExt(char *pName, char *pExt)
{
	char *p;

	// validate parameters
	Assert(pName && *pName);
	Assert(pExt && *pExt);

	// find the start of the extension
	p = FindExt(pName);
	Assert(p);

	/* put new extension on it */
	if(*pExt != '.')
		*p++ = '.';
	strcpy(p, pExt);
}

/*
	Find where the extension begins.
*/
char * FindExt(char *pName)
{
	char * pExt;

	// make sure the pName points to something
	Assert(*pName);

	// determine the start of the filename part
	pExt = strrchr(pName,'\\');
	if(!pExt)
		pExt = pName;

	// determine start of extension
	pExt = strchr(pExt, '.');
	if(!pExt)
		pExt = pName + strlen(pName);

	// return the extension found
	return pExt;
}

/*
	Convert a string into a 16 bit hex number.
*/
int atoh(char *s)
{
	int h=0;
	while(*s)
	{
		if(*s>='0' && *s<='9')
			h= h*16 + *s-'0';
		else if(*s>='A' && *s<='F')
			h= h*16 + *s-'A'+10;
		else if(*s>='a' && *s<='f')
			h= h*16 + *s-'a'+10;
		else
			break;	/* not a valid hex digit, end of number */
		++s;
	}
	return h;
}

void set_fat_entry(unsigned clus,unsigned val)
{
	unsigned entry;

	/* 16 bit FAT */
	if(data_sectors > MAX12BITFAT)
		*(unsigned *)(fatbuff+clus*2) = val;

	/* 12 bit FAT */
	else
	{
		entry= *(unsigned *)(fatbuff + clus/2*3 + clus%2);
		if(clus%2 == 0)
		{
			entry&= 0xF000;
			entry|= val&0xFFF;
		}
		else
		{
			entry&= 0x000F;
			entry|= val<<4;
		}
		*(unsigned *)(fatbuff + clus/2*3 + clus%2) = entry;
	}
}

/*
	Allocate the maximum space possible, up to 62K
	This cannot fail.
*/
void * maxmalloc(unsigned *size)
{
	static char buff[512];

	void * p = 0;
	*size = 62 * 1024;

	/* get as much memory as we can */
	while(!p && *size > 6*1024)
	{
		p = malloc(*size);
		if(!p)
			*size -= 2*1024;
	}

	/* allocate a little less so we have room for opening files */
	if(p)
	{
		*size -= 4*1024;
		p = realloc(p, *size);
	}

	/* can't allocate any space, use our internal buffer */
	if(!p)
	{
		*size = sizeof(buff);
		p = buff;
	}

	return p;
}

/*
	Convert a "FILENAME.EXT" to an expanded name form;
	that is "FILE.C" to "FILE    C  ".  The ename is
	always 12 characters - 11 name/ext + 1 for NULL.
*/
void dot_to_ename(char *ename, char *dotname)
{
	int i;

	/* special cases "." and ".." */
	if(*dotname == '.' && (*(dotname+1)==0 || *(dotname+1)=='.'))
	{
		memset(ename, ' ', 11);
		*ename = '.';
		if(*(dotname+1) == '.')
			*(ename+1) = '.';
		*(ename+11) = 0;
		return;
	}

	/* expand name part */
	for(i=0;i<8;++i)
	{
		if(*dotname && *dotname != '.')
		{
			*ename++ = toupper(*dotname);
			dotname++;
		}
		else
			*ename++ = ' ';
	}

	/* expand extension part */
	if(*dotname == '.')
		++dotname;
	for(i=0;i<3;++i)
	{
		if(*dotname)
		{
			*ename++ = toupper(*dotname);
			++dotname;
		}
		else
			*ename++ = ' ';
	}
	*ename = 0;
}


/*
		[]-------------------------[]
		 |		HEX FILE STUFF			|
		[]-------------------------[]
*/

/*
	Generate an Intel Hex record given the buffer and size
	of the record
*/
void gen_hex(unsigned char *buf, int size, FILE *fp)
{
	int i;
	signed char sum;

	char *hex = "0123456789ABCDEF";
	sum = 0;							/* calculate check sum */
	fputc(':', fp);
	for(i = 0; i < size; ++i)
	{
 		fputc(hex[buf[i] >> 4], fp);
		fputc(hex[buf[i] & 0xf], fp);
		sum -= buf[i];
	}
	fputc(hex[(sum >> 4) & 0xf], fp);
	fputc(hex[sum & 0xf], fp);
	fputc('\r', fp);
	fputc('\n', fp);
}

void write_hex_type(uchar *buff, int size, int address, int type, FILE *fp)
{
	unsigned i;
	int j, nbytes;
	static unsigned char xbuf[40] = { 0x2, 0x0, 0x0, 0x0};

	for(i=0; i<size; i+=32)
	{
		nbytes = size - i;
		if(nbytes > 32)
			nbytes = 32;
		xbuf[0] = nbytes;						/* REC LEN */
		xbuf[1] = (address+i) >> 8;		/* LOAD ADDRESS (hi part) */
		xbuf[2] = (address+i) & 0xff;		/* LOAD ADDRESS (lo part) */
		xbuf[3] = type;						/* REC TYP-Data Rcd */
		for(j=0; j<nbytes; ++j)
			xbuf[4+j] = buff[i+j];

		gen_hex(xbuf, nbytes+4, fp);
	}
}

#if 0
/*
	Write out a hex record with extended address information
*/
void gen_hex_ext(unsigned short seg, FILE *fp)
{
	static char record[8] = {0x2, 0x0, 0x0, 0x2, 0x0, 0x0};
	*((unsigned *) (record + 1)) = seg;
	record[4] = seg >> 8;
	record[5] = seg & 0xff;
	gen_hex(record, 6, fp);
}
#endif


/*
	Write a Intel Extended address record to the
	specified file
*/
void write_hex_eadr(unsigned ext_adr, FILE *fp)
{
	unsigned char addr[2];

	addr[0] = ext_adr >> 8;
	addr[1] = ext_adr & 0xFF;
	write_hex_type(addr, sizeof(addr), 0, 2, fp);
}

/*
	Terminate a HEX file.
*/
void write_hex_done(FILE * fp)
{
	fputs(":00000001FF\r\n", fp);
}

/*
	All writes to a HEX file go through this routine.  This
	routine in turn calls lower-level HEX routines to do the
	actual writes.
*/
void write_hex(char *buffer, unsigned size, FILE *fp)
{
	static long start=0L;

	int len;

	while(size)
	{
		/* determine length to write out */
		len = (size > 32) ? 32 : size;

		/* write extended addressing information */
		if(sw_extended && (start & 0xFFFF) == 0)
			write_hex_eadr(start_addr + (unsigned)(start>>4), fp);

		/* write out the hex information */
		write_hex_type((unsigned char *)buffer, len, (int)start, 0, fp);

		/* update the counters */
		buffer += len;
		start += len;
		size -= len;
	}
}

/*
	Based on date from DOS, convert it into a file
	date stamp format.
*/
unsigned get_sysdate(void)
{
	unsigned date;
	union REGS r;

	r.h.ah= 0x2A;
	intdos(&r,&r);
	date = r.x.cx-1980;				/*  year */
	date = date<<9;
	date|= (int)(r.h.dh)<<5;		/* month */
	date|= r.h.dl;						/*   day */
	return date;
}

/*
	Based on a time from DOS, convert it into a file
	time stamp format.
*/
unsigned get_systime(void)
{
	unsigned time;
	union REGS r;

	r.h.ah= 0x2C;
	intdos(&r,&r);
	time = (int)(r.h.ch)<<11;		/*     hours */
	time|= (int)(r.h.cl)<<5;		/*   minutes */
	time|= (r.h.dh)>>1;				/* seconds/2 */
	return time;
}
