//
//   $Id: RSZ.C 1.4 2002/10/15 23:55:32Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

	RSZ.C  created 08/08/97  by Dennis Edwards

	Copyright (c) 1997-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	This is the main module for the RSZ ZMODEM file transfer utility.
	The RSZ error codes are defined below:

  	  0	success
  	  1	help requested
  	  2   invalid command line option
  	  3   unable to initialize port

 	 99	internal error

	236   File not found
	244  	packet header lost
	245  	exceeded expected block length
	246  	unexpected data/invalid ZDLE sequence
	247  	CRC incorrect
	248  	timed out on character receive

	250  	couldn't synchronize file positions
	251  	memory allocation error
	252  	DOS read/write error
	253  	lost carrier
	254	aborted by remote
	255   aborted by local operator

   05/31/1999 (DG)
		1. Added timezone
		2. Doesn't check for KBHITs on serial console

	08-19-98 (BT)
		1. Added support for irq command line switch

	05-15-98 (TQ)
		1. Added /v switch to force Status ouput on systems
		   redirected over serial ports or CTTY.  Normally disabled.

	01-10-98 (DE)
		1. Added /q (quiet) switch to not display file transfer status info.
		2. Defined a "file not found" error return status and message.
		3. Now display count of files transfered.

	12-09-97 (DE)
		1. Fixed local abort to work consistently and with ^X, ^C, etc.
		   as well as ESC
		2. Added status display output.
		3. Defined error codes

*******************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "dostruct.h"
#include "dosver.h"
#include "datalght.h"
#include "dlapp.h"
#include "dldosio.h"
#include "dlbiosio.h"
#include "fatworld.h"
#include "abortlfn.h"

#include "rsz.h"			// typedefs and protocol driver interface spec


#define RXD_BUFF_SIZE	8196
#define TXD_BUFF_SIZE	8196
unsigned char rxdBuffer[RXD_BUFF_SIZE+TXD_BUFF_SIZE];

bool fIsFat32World;

// disable out the /L option for now but keep it around in case
// we add status reporting of some kind to this utility...
//
char UsageMsg[]=
//"Usage: RSZ /Pn /Bn /Hn /Lfile /Fn /[R|[S file1 [file2 ...]]]\n\r"
"Usage: RSZ /Pn [/In] /Bn /Hn /Fn [/Q] [/V] /R|[S file1 [file2 ...]]\n\r"
"  /P    port number:\n\r"
"           n = {1,2,3,4} for COM1 to COM4 (default is 1 for COM1)\n\r"
"  /I    IRQ number:\n\r"
"           n = {3-15} for IRQ3-IRQ15 (default is IRQ3 for COM2 and COM4\n\r"
"               and is IRQ4 for COM1 and COM3)\n\r"
"  /B    baudrate, if not specified uses current port parameters:\n\r"
"           n = 50,110,300,...,115200. Always \"N81\" when changed.\n\r"
"  /H    handshaking options, both sides must use the same one of:\n\r"
"           0= none (default), 1= software, 2= hardware\n\r"
"  /F    file management options (all files are \"binary\"), one of:\n\r"
"           0= skip, 1= resume, 2= make duplicate, 3= replace (default)\n\r"
"  /R    receive files specified by sender using ZMODMEM protocol\n\r"
"  /S    send the specified files using ZMODEM protocol, wild-cards allowed.\n\r"
"  /Q    turn off status display\n\r"
"  /V    force status display on when console is redirected to 2nd COM Port.\n\r"
"Examples:  RSZ /r\n\r"
"           RSZ /p3 /b115200 /h2 /f0 /s a.b lmno.p z.*\n\r";

char szTemp[128];

#define DEFAULT_LOG STDOUT

int  portNumber=1;
int  irqNumber=255;
long baudRate= 0L;
int  handShakerType=NO_HANDSHAKING;
char szLogFileName[14];
int  logFile = DEFAULT_LOG;
int  fileOptions=CLOBBER_EXISTING;
bool receive=FALSE, send=FALSE, displayOutput=TRUE, outputIsFile=FALSE;
bool quiet=FALSE, loud=FALSE;

ARGDEF Args[]=	{	{INT_SWITCH,		&portNumber,		'P', 0},
						{INT_SWITCH, 		&irqNumber,			'I', 0},
						{LONG_SWITCH, 		&baudRate, 			'B', 0},
						{INT_SWITCH, 		&handShakerType,	'H', 0},
						{STRING_SWITCH,	szLogFileName,		'L',12},
						{INT_SWITCH,		&fileOptions, 		'F', 0},
						{BOOL_SWITCH, 		&receive, 			'R', 0},
						{BOOL_SWITCH, 		&send,				'S', 0},
						{BOOL_SWITCH, 		&quiet,				'Q', 0}, // force no status
						{BOOL_SWITCH, 	   &loud,				'V', 0}, // force status on
						{0,0,0,0}};


#include "mcommifc.c"

// debugging tool, create a file
bool PASCAL CreateFile(char szFileName[], int *pHandle)
{
	unsigned uAction = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
	return dlOpenOrCreateFile(O_WRONLY, A_NORMAL, &uAction, szFileName, pHandle);
}

// Say "Howdy ma'am"
void SignOn(void)
{ char s[128];
  unsigned len;
  char *szSignOn = "RSZ - ZMODEM file transfer utility.\n\r%s %s\n\r%s\n\r\n";

	if( fIsFat32World )
	{
		len = sprintf(s,szSignOn, SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		len = sprintf(s,szSignOn, SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}

	dlWriteFile(logFile,&len,s);
}

// Say "Here's the deal..."
void Usage(void)
{ unsigned len;
	len = strlen(UsageMsg);
	dlWriteFile(logFile,&len,UsageMsg);
	exit(1);
}

#if DEBUG
void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT %d.%d\n\r",fileNum,lineNum);
	exit(99);
}
#endif

// Say "Uh, 'scuse me"
void OptError(char *msg)
{ char s[128];
  unsigned len;

	len = sprintf(s,"\a\n\rError, invalid command line option: %s\n\r",msg);
	dlWriteFile(logFile,&len,s);
	exit(2);
}

// Say whats happening with the file transfer(s).
//
void StatusMessagePrinter(char *msg)
{
  unsigned len;

	if(displayOutput || *szLogFileName)
	{
		len = strlen(msg);
		dlWriteFile(logFile,&len,msg);
	}
}

// Queries STDOUT to determine if it has been redirected.
// If it has been redirected then we want to suspend any status
// output since that might be to the same port as the file transfer
// and really mess things up.
//
// Returns:
//
//	  1: CON is Video
//	  0: CON is CTTY or Int 10h goes to a serial port
//	 -1: CON is redirected to a file
//
int CheckSTDOUTRedirection(void)
{
  unsigned deviceData;

	// Have to worry about if we have been CTTY'd and if we are running
	// on a BIOS where the console is a terminal (where CON = COMn = /Pn)
	//
	dlGetDeviceAttributes(STDOUT,&deviceData);
	if(deviceData & IO_DEVICE)
	{
		unsigned char curChar, curAttribute;

		if(deviceData & IO_FAST)
		{
			// STDOUT is the CON device, is CON a video display or COMn?
			// Assume that if we read the char and attribute from the cursor
			// position on video page 0 and it comes back as anything but
			// 0x800 (a NUL char in dark gray on black) then we have an actual
			// video display. If we get back AX unchanged then do not output
			// status info to STDOUT as it may be the same COMx that we will
			// be using for our file transfers.
			//
			BIOSVidGetCurCharAndAttr(0,&curChar,&curAttribute);
			if((curChar != 0) && (curAttribute != 8))
				return 1;
			else
				return 0;
		}
		else
		{
			// STDOUT is a non-CON device. Assume that it is CTTY. Since
			// CTTY may be to the same serial port as we are using for
			// our serial transfers, don't display status info on STDOUT
			//
			return 0;
		}
	}
	else
	{
		// STDOUT is a file go ahead and display output into it.
		//
		return -1;
	}
}

// It all starts here...
int main(int argc, char **argv)
{
	int i, status;
	char fileList[128];
	unsigned oldBreakLevel;
	unsigned len;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Abort if there are LFNs on the command line
	//
	AbortForLFNArgs();

	SignOn();

	/* adjust for local time zone */
	tzset();

	// Process user input.
	//
	if((argc <= 1) ||	((i = ParseCmdline(argc-1, argv+1, Args)) < 0))
	{
		Usage();
	}
	else if(((send && receive) || !(send || receive)) ||
			  ((i == 0) && send) ||
			  ((handShakerType < NO_HANDSHAKING) || (handShakerType > HARDWARE_HANDSHAKING)) ||
			  ((fileOptions < SKIP_IT) || (fileOptions > CLOBBER_EXISTING)) ||
			  ((portNumber < 1) || (portNumber > 4)) ||
			  ((baudRate < 0L)  || (baudRate > 115200L)))
		Usage();
	else
	{
		// Combine argv[] file names into a single string with names
		// separated by spaces.
		//
		argc = i;
		i = 1;
		strcpy(fileList, argv[i++]);
		while(i <= argc)
		{
			strcat(fileList, " ");
			strcat(fileList, argv[i++]);
		}
	}

	// set up log file
	if(*szLogFileName)
	{
		if(CreateFile(szLogFileName, &logFile))
		{
			len = sprintf(szTemp, "Logging...\n");
			dlWriteFile(logFile, &len, szTemp);
		}
		else
			logFile = DEFAULT_LOG;
	}

	//  08-19-98  BT
	//
	//  Determine the IRQ if one was not specified at the command line
	//
	if(irqNumber == 255)
	{
		irqNumber = ( ( portNumber - 1 ) & 1) ? 3 : 4;
	}

	if( ( irqNumber < 3 ) || ( irqNumber > 15 ) )
	{
		Usage();
	}


	switch(CheckSTDOUTRedirection())
	{
	case 1:	//	CON is Video
		displayOutput = TRUE;
		outputIsFile  = FALSE;
		break;

	case  0: // CON is CTTY or Int 10h goes to a serial port
	 	displayOutput = loud;	// may be true or false
		break;

	case -1:	// CON is redirected to a file
		displayOutput = TRUE;
		outputIsFile  = TRUE;
		break;
	}

	// if not displaying output because of a serial console, don't
	// allow KBHITs, or Zmodem transfer will fail.
	fNoKbHit = !displayOutput;

	// remember current break level (1 = check for ^Cs) and turn
	// break checking off, set STDIN and STDOUT to raw mode.
	//
	dlGetBreakLevel((int*)&oldBreakLevel);
	dlSetBreakLevel(BREAK_OFF);

	// Set the user specified handshaking method, don't care what it was.
	//
	SetHandShakingType(handShakerType);

	// Set file management options, don't care what they were.
	//
	SetFileManagementType(fileOptions,(fileOptions==RESUME_TRANSFER));


	//  08-19-98  BT
	//
	// Try to open the specified port
	//
	status = comm_open(baudRate, "N81", portNumber-1, irqNumber);
	if(status)
	{
		len = sprintf(szTemp, "Cannot open COM%d\n", portNumber);
		dlWriteFile(logFile,&len,szTemp);
		goto DoExit;
	}

	// Send or Receive the files
	//
	if(send)
		status = SendZMODEM(fileList);
	else
		status = ReceiveZMODEM();

	// done. close everything and exit
	comm_close();
DoExit:
	dlSetBreakLevel(oldBreakLevel);
	dlCloseFile(logFile);

	return status;
}
