//
//   $Id: SHARE.C 1.4 2002/10/15 23:56:38Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

	SHARE.C  created 3/94  by Drew Gislason

	Copyright (c) 1994-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	SHARE [/L:30] [/U]

		/L  Set # of locks
		/U  Unload previous SHARE


	WARNING! This code is run while SS != DS. Never use near ptrs
	when passing parameters to other functions!

	SHARE's main job is to provide a file locking and retry
	mechanism to allow multiple networks tasks safe access to
	files.

	This SHARE works both with Datalight's ROM-DOS and other
	DOS compatibles.  It relies on some internal DOS structures
	(mainly the LOL and SFTs), as well as some SHARE defined
	ones.

	INT 20H				Terminate
	INT 21H, 5C00H		Lock Handle
	    21H, 5C01H		Unlock Handle
	    21H, 31H		TSR
	    21H, 3CH		Create
	    21H, 3DxxH		Open (possibly with sharing)
	    21H, 3EH		Close
	    21H, 3FH		Read
	    21H, 40H		Write
	    21H, 41H		Delete
	    21H, 4CH		Terminate
	    21H, 6CH		Special Open (w/ sharing)
	INT 27H				TSR

	INT 2FH, 1000H		Installation Check
	    2FH, 1011H		Uninstall

********************************************************************/
#define SHARE 1
#include <mem.h>

#include "dosmsg.c"
#include "dosver.h"
#include "dostruct.h"
#include "datalght.h"
#include "fatworld.h"
#include "share.h"
#include "shareprt.h"

int numLocks = DEFLOCKS;
ShareLock * shareLock;
LPLOL lol;		// list of lists - we need this

bool fIsFat32World;
unsigned uErrorCode;

//  Prototypes
//
int atoi(char *s);
int mode_with_deny(unsigned char open_mode, unsigned char deny_mode);
int uninstall(void);

/*
	SHARE [/L:30] [/U]
*/
int main(int argc, char *argv[])
{
	int i;
	char * p;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	// signon
	puts(SIGNON);
	if( fIsFat32World )
	{
		puts(SHORTVERSION);
		puts(" ");
		puts(BUILDVERSION);
	}
	else
	{
		puts(SHORTVERSION622);
		puts(" ");
		puts(BUILDVERSION);
	}
	puts("\n");
	puts(COPYRIGHT); puts("\n\n");

	// process arguments
	for(i=1; i<argc; ++i)
	{
		p = argv[i];
		while(*p && *p != '/')
			++p;
		if(*p == 0)
			continue;
		++p;

		// Set number of locks?
		if(*p=='?')
		{
			puts(USAGE);
			return 1;
		}

		if(*p=='l' || *p=='L')
		{
			++p;
			if(*p == ':')
				++p;
			numLocks = atoi(p);
			if(!numLocks)
				numLocks = DEFLOCKS;
			if(numLocks < MINLOCKS)
				numLocks = MINLOCKS;
			if(numLocks > MAXLOCKS)
				numLocks = MAXLOCKS;
		}

		if(*p=='u' || *p=='U')
		{
			if(!uninstall())
			{
				puts(msg_cant_uninstall);
				return 2;
			}
			puts(msg_uninstalled);
			return 0;
		}
	}

	// check for already installed
	if(already_installed())
	{
		puts(msg_already_installed);
		return 3;
	}

	// allocate the share locks
	shareLock = zmalloc(numLocks * sizeof(ShareLock));

	// get the DOS data segment
	lol = getLOL();

	// trap the interrupts
	trap_ints();

	// everything worked, OK, so stay resident
	puts(msg_installed);
	return TERM_TSR;
}

// convert a string to an integer
int atoi(char *s)
{
	int val = 0;
	while(1)
	{
		if(*s>='0' && *s<='9')
			val = val*10 + (*s-'0');
		else
			break;
		++s;
	}
	return val;
}

// Allocate and zero the requested bytes.
void * zmalloc(unsigned bytes)
{
	void * AllocateBytes(unsigned bytes);

	void * p;
	p = AllocateBytes(bytes);
	if(p)
		memset(p,0,bytes);
	return p;
}

/*
	Convert an SFT # into an SFT pointer.  An "SFT #" is
	found in the user's PSP handle table.

	Returns a pointer to the SFT or (void *)NULLF if the SFT
	doesn't exist.
*/
LPSFT h_to_sft(int h)
{
	LPDCB dcb;
	LPSFT sft;

	dcb = lol->first_dcb;
	(void *)sft = NULLF;
	while(1)
	{
		/* we have found it */
		if(h < dcb->numsft)
		{
			sft = MK_FP(FP_SEG(dcb),dcb->sft+h);
			break;
		}

		/* check if we are at the end of the list */
		if((void *)dcb->next == NULLF)
			break;

		/* check the next device control block */
		h -= dcb->numsft;
		dcb = dcb->next;
	}

	return sft;
}

/*
	From a user handle,
*/
LPSFT handle2sft(int handle)
{
	struct PSP far * psp;
	LPSFT sft;
	uchar h;

	// don't use a handle greater than handle count
	psp = MK_FP(getPSP(), 0);
	if(handle >= psp->hcount)
		return (LPSFT)NULLF;

	// find the SFT
	h = *(((uchar far *)psp->htbl_addr)+handle);
	sft = h_to_sft(h);

	// don't return unused SFTs
	if((void *)sft != NULLF)
	{
		if(sft->handles == 0)
			return (LPSFT)NULLF;
	}

	return sft;
}

// returns true if the region and lock intersect
int intersect(ulong start, ulong len, ulong lkstart, ulong lklen)
{
	ulong lkend, end;
	end = start + len - 1;
	lkend = lkstart + lklen-1;
	if( (start>=lkstart && start<=lkend) ||
		 (end>=lkstart && end<=lkend) ||
		 (start<lkstart && end>lkend) )
		return 1;
	return 0;
}

// returns true if the two SFTs point to the same file
// return false if not
int samefile(LPSFT sft, LPSFT sft2)
{
	if(sft2 == sft ||
			(sft2->dirsec==sft->dirsec &&
			sft2->dirent==sft->dirent &&
			(sft2->ioctl&0x3F)==(sft->ioctl&0x3F) &&
			sft2->device==sft->device) )
		return 1;
	return 0;
}

// can_work_together should handle the special cases of opening
// a read-only file in compatibility mode plus deny-none or
// deny-write.  currently, it doesn't recognize anything special
// about files w/ read-only attribute, and will deny those opens
//
// NOTE: See PC-Interrupts INT 21 Fn 3D for a great table that
// describes the behavior of this function
int can_work_together(unsigned char old_mode, unsigned char new_mode)
{
	unsigned char old_deny_mode = old_mode & O_DENYMASK;
	unsigned char new_deny_mode = new_mode & O_DENYMASK;

	unsigned char old_open_mode = old_mode & 3;
	unsigned char new_open_mode = new_mode & 3;

	// COMP mode is always compatible with COMP mode and
	// always incompatible with everything else
	if (old_deny_mode == 0)
	{
		if (new_deny_mode == 0)
			return 1;
		else
			return 0;
	}
	else if (new_deny_mode == 0)
		return 0;

	// at this point, both deny modes are real, so check
	// the easy case of DENYOPEN in either position
	if (old_deny_mode == O_DENYOPEN || new_deny_mode == O_DENYOPEN)
		return 0;

	if (!mode_with_deny(old_open_mode, new_deny_mode))
		return 0;
	if (!mode_with_deny(new_open_mode, old_deny_mode))
		return 0;

	// no problems encountered--we'll allow it
	return 1;
}

int mode_with_deny(unsigned char open_mode, unsigned char deny_mode)
{
	if (deny_mode == O_DENYREAD && open_mode != 1)
		return 0;
	if (deny_mode == O_DENYWRITE && open_mode != 0)
		return 0;

	// must be ok
	return 1;
}

/*
	Is this file denied for open (or read or write)?
	Returns error code if denied.
	Returns 0 if everything is ok.
*/
int is_denied(int handle)
{
	int i;
	unsigned new_mode;
	unsigned old_mode;
	LPSFT new_sft;
	LPSFT old_sft;

	// if it's an invalid handle, who cares?  I don't
	new_sft=handle2sft(handle);
	if((void *)new_sft == NULLF)
		return 0;

	// devices are never denied
	if(new_sft->ioctl&IO_DEVICE)
		return 0;

	// check through all existing SFTs for denial.
	for(i=0; ; ++i)
	{
		// get the next SFT
		old_sft = h_to_sft(i);
		if((void *)old_sft == NULLF)	// end of list, done
			break;

		// don't look at empty SFTs or the one in question
		if(old_sft->handles == 0 || old_sft == new_sft)
			continue;

		if( !samefile(new_sft, old_sft) )
			continue;

		new_mode = new_sft->rwmode;
		old_mode = old_sft->rwmode;

		// This IS the same file--check to see if we
		// can allow a second open or not
		if (can_work_together(old_mode, new_mode))
			continue;

		// we've got a denial here! (the Queen of Denial)
		return ERR_SHARE;
	}

	// no problems or sharing conflicts
	return 0;
}

/*
	Is this region locked for read/write?
	Returns error code if locked.
	Returns 0 if everything is ok.
*/
int is_locked(int handle, unsigned len)
{
	int i;
	unsigned owner;
	LPSFT sft;

	// if it's an invalid handle, who cares?  I don't
	sft=handle2sft(handle);
	if((void *)sft == NULLF)
		return 0;

	// could the handle be locked?
	owner = getPSP();
	for(i=0; i<numLocks; ++i)
	{
		// ignore locks for this owner
		if(!shareLock[i].owner || shareLock[i].owner==owner)
			continue;

		// does the portion intersect
		if( samefile(sft, shareLock[i].sft) &&
			intersect(sft->offset, (ulong)len,
			shareLock[i].start, shareLock[i].length) )
		return ERR_LOCKED;
	}
	return 0;
}

/*
	Close all locks for this process, as the process
	is terminating.
*/
void unlock_close_all(void)
{
	int i;
	unsigned owner;

	// check through all existing SFTs for denial.
	owner = getPSP();
	for(i=0; i<numLocks; ++i)
	{
		// only free up the lock if this process owns BOTH the
		// lock and the SFT
		if(shareLock[i].owner==owner && shareLock[i].sft->owner==owner)
			shareLock[i].owner = 0;
	}
}

// unlock handles automatically on File Close
// (but only if SFT will be freed up)
void unlock_close(int handle)
{
	int i;
	unsigned owner;
	LPSFT sft;

	// if it's an invalid handle, who cares?  I don't
	sft=handle2sft(handle);
	if((void *)sft == NULLF)
		return;

	// look for any SFTs that match
	owner = getPSP();
	for(i=0; i<numLocks; ++i)
	{
		// free the lock
		if(shareLock[i].owner==owner && shareLock[i].sft==sft &&
			sft->handles==1)
			shareLock[i].owner = 0;
	}
}

/*
	Returns 0 if it worked.  Error # if failed.
*/
int lock_handle(Region rg)
{
	int i;
	LPSFT sft;

	// first, convert the handle into an SFT.
	sft=handle2sft(rg.handle);
	if((void *)sft == NULLF)
		return ERR_BADHANDLE;

	// ensure this region doesn't overlap with another region
	// of the same file
	for(i=0; i<numLocks; ++i)
	{
		// no need to check "free" locks
		if(shareLock[i].owner == 0)
			continue;

		// only look at those that are of the same FILE (may or
		// may not be the same SFT)
		if( samefile(sft, shareLock[i].sft) &&
			intersect(rg.start, rg.length,
			shareLock[i].start, shareLock[i].length) )
		{
			return ERR_LOCKED;
		}
	}

	// find a free region to use
	for(i=0; i<numLocks; ++i)
	{
		if(shareLock[i].owner == 0)
			break;
	}
	if(i == numLocks)
		return ERR_NOLOCKS;

	// now fill it in with our information
	shareLock[i].owner = getPSP();
	shareLock[i].start = rg.start;
	shareLock[i].length= rg.length;
	shareLock[i].sft   = sft;

	// everything worked
	return 0;
}

/*
	Unlock a handle.
	The region must EXACTLY match a region in the list

	Returns 0 if it worked.  Error # if failed.
*/
int unlock_handle(Region rg)
{
	int i;
	unsigned owner;
	LPSFT sft;

	// first, convert the handle into an SFT.
	sft = handle2sft(rg.handle);
	if((void *)sft == NULLF)
		return ERR_BADHANDLE;

	// find this region in the list
	owner = getPSP();
	for(i=0; i<numLocks; ++i)
	{
		// only the owner of the lock can release it
		if(shareLock[i].owner == owner && shareLock[i].sft == sft &&
			shareLock[i].start==rg.start && shareLock[i].length==rg.length)
			break;
	}
	if(i == numLocks)
		return ERR_LOCKED;

	// free the lock
	shareLock[i].owner = 0;

	// everything worked
	return 0;
}

/*
	Do other SFTs point to this same file?
*/
int others_have_open(int handle)
{
	int i;
	LPSFT sft;
	LPSFT sft2;

	// if it's an invalid handle, no others can point to it
	sft = handle2sft(handle);
	if((void *)sft == NULLF)
		return 0;

	for(i=0; ; ++i)
	{
		// get the next SFT
		sft2 = h_to_sft(i);
		if((void *)sft2 == NULLF)	// end of list, done
			break;

		// don't look at empty SFTs or the one in question
		if(sft2->handles == 0 || sft2 == sft)
			continue;

		// we found another SFT pointing to the same file
		if( samefile(sft, sft2) )
			return 1;
	}

	// not found in list, so no others point to it
	return 0;
}

/*
	Syncronize this handle that was just opened.
*/
void open_sync(int handle)
{
	int i;
	LPSFT sft;
	LPSFT sft2;

	// if it's an invalid handle, something else is wrong, no sync
	sft = handle2sft(handle);
	if((void *)sft == NULLF)
		return;

	// could the handle be locked?
	// check through all existing SFTs for denial.
	for(i=0; ; ++i)
	{
		// get the next SFT
		sft2 = h_to_sft(i);
		if((void *)sft2 == NULLF)	// end of list, done
			break;

		// don't look at empty SFTs or the one in question
		if(sft2->handles == 0 || sft2 == sft)
			continue;

		// ignore different files
		if( !samefile(sft, sft2) )
			continue;

		// this is the same file, but different SFT, syncronize it
		sft->size    = sft2->size;
		if( fIsFat32World )
		{
			sft->ulStartingCluster = sft2->ulStartingCluster;
			sft->ulAbsCluster      = sft2->ulAbsCluster;
			sft->uRelClusterLow = sft2->uRelClusterLow;
			sft->uRelClusterHigh = sft2->uRelClusterHigh;
		}
		else
		{
			((LPSFT622)sft)->start   = ((LPSFT622)sft2)->start;
			((LPSFT622)sft)->absclus = ((LPSFT622)sft2)->absclus;
			((LPSFT622)sft)->relclus = ((LPSFT622)sft2)->relclus;
		}
	}
}

/*
	Syncronize handles after a write.

	A write has occurred (possibly adding to or removing
	some of a file).  make sure all SFTs that point to this
	same file remain in sync.
*/
void sync_handle(int handle)
{
	int i;
	LPSFT sft;
	LPSFT sft2;

	// if it's an invalid handle, who cares?  I don't
	sft = handle2sft(handle);
	if((void *)sft == NULLF)
		return;

	// could the handle be locked?
	// check through all existing SFTs for denial.
	for(i=0; ; ++i)
	{
		// get the next SFT
		sft2 = h_to_sft(i);
		if((void *)sft2 == NULLF)	// end of list, done
			break;

		// don't look at empty SFTs or the one in question
		if(sft2->handles == 0 || sft2 == sft)
			continue;

		// ignore different files
		if( !samefile(sft, sft2) )
			continue;

		// make sure the vitals are in sync
		sft2->size    = sft->size;
		if( fIsFat32World )
		{
			sft->ulStartingCluster = sft2->ulStartingCluster;
			sft->ulAbsCluster      = sft2->ulAbsCluster;
			sft->uRelClusterLow = sft2->uRelClusterLow;
			sft->uRelClusterHigh = sft2->uRelClusterHigh;
		}
		else
		{
			((LPSFT622)sft)->start   = ((LPSFT622)sft2)->start;
			((LPSFT622)sft)->absclus = ((LPSFT622)sft2)->absclus;
			((LPSFT622)sft)->relclus = ((LPSFT622)sft2)->relclus;
		}
	}
}
