//
//   $Id: RXE_BLK.C 1.16 2002/10/16 15:53:20Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 *  Copyright (c) 1989-2002 Datalight, Inc.
 *  All Rights Reserved.
 *  Datalight Confidential and Proprietary
 *
 * RXE converter - Converts a Block of Code from EXE format to RXE
 * format.
 *
 * 	Bugs:
 * 		1) Should not have any room in header, RXE runs out of ROM
 * 		   and there for does not have to have it's code start on
 *		   a sector boundry.
 */

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <ctype.h>
#include <assert.h>
#include <dos.h>
#include <time.h>
#include <string.h>
#include <fcntl.h>
#include <alloc.h>
#include <mem.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "datalght.h"
#include "dostruct.h"
#include "rxe.h"

#include "exehdr.h"
#include "rxe-prot.pt"

#define DATA_SEG 0x10
#define MEM_SEG  0x1000
#define RXE_INTERRUPT rxe_int
#define DATA_SEG_TABLE_SIZE 16
#define BUF_SIZE ( (unsigned) 32 * 1024)
#define NOP 0x90

struct exehdr exehdr;			/* Original EXE header */
struct exehdr new_exehdr;		/* New EXE header for XIP */

typedef unsigned long ulong;


/* struct rxehdr rxehdr; */

struct rel_item huge * hpRelItems;
struct rel_item huge * hpFixupList;
struct rxe_base * rxe_base;

char * prog_text;
unsigned * uptr;
unsigned char * byte_ptr;

int exe, rxe_fixups;
unsigned data_seg_table[DATA_SEG_TABLE_SIZE];
int data_seg_entry = 0;
long imglen, count, text_len;
unsigned data_seg, memory_seg;
unsigned rxe_base_ip;
unsigned rxe_base_paras;

extern int partial;	/* FALSE if dseg start on para boundry else TRUE */
extern do_getsize;
extern quit;

/*
 *
 * Fixup in CODE
 *	If it's a CODE fixup then it is done
 * 	if it's a DATA fixup then it is converted to a int 18, data byte
 * Fixup in  DATA
 *	If it's a CODE fixup then it is done
 * 	if it's a DATA fixup it is modified so DOS can perform the fixup
 */

void convert_exe_to_rxe( char * exe_name, unsigned dseg, unsigned mseg,
		unsigned rxe_int )
{
	unsigned long code_block_size;
	unsigned long data_block_size;
	unsigned long size;
	unsigned len;
	unsigned long ulRelItemsByteSize;
#if DEBUG
	unsigned	uIndex;
#endif

	//  init_rxe_base performs initial loading and initialization of the
	//  RXE_BASE XIP loader.  initialize RXE_INIT.EXE
	//
	init_rxe_base();


	//  Open the source (EXE) file
	//
	exe = open(exe_name, O_RDONLY | O_BINARY );
	if(exe == -1)
	{
		error(20, exe_name);
	}


	data_seg = dseg;
	memory_seg = mseg;


	// read the exe header
	//
	if(read(exe, &exehdr, sizeof(exehdr)) != sizeof(exehdr) ||
			exehdr.sig != EXESIG)
	{
		error(1, exe_name);
	}


	//  Determine if this is already an RXE file
	//
	if(is_xip(exe))
	{
		error(11, exe_name);
	}


	//  Make sure that there is at least 1 relocatable item (data)
	//
	if(exehdr.nreloc == 0)
	{
		error(8);
	}


	/* must be able to allocate the memory */
	/* Add 16 for XIP fixups */
	ulRelItemsByteSize = (unsigned long)(exehdr.nreloc + 16) *
			sizeof(struct rel_item);
	hpRelItems = new( ulRelItemsByteSize );
	if(!hpRelItems)
	{
		error(2);
	}

	/* read in all the */
	lseek(exe, (long)exehdr.reldsp, 0);

	BigRead( exe, (void far *)hpRelItems, ulRelItemsByteSize );

#if DEBUG
	for(uIndex=0; uIndex < exehdr.nreloc; uIndex++ )
	{
		printf("hpRelItems[%u] = %04X:%04x\n", uIndex,
				hpRelItems[uIndex].segment, hpRelItems[uIndex].offset);
	}
#endif


//  BT  4-26-00
//
//  ETR 3407 - This calculation is not correct.  It shortens the image
//  by 512 bytes.  Verify this with the filesize calculation:
//
//  Filesize = (exehdr.imglen * 512) + exehdr.imgmod + (exehdr.hdrlen << 4);
//
//	imglen  = (exehdr.imglen - 1) * 512L;
//
	if( exehdr.imgmod )
	{
		imglen  = (exehdr.imglen - 1) * 512L;
	}
	else
	{
		imglen  = exehdr.imglen * 512L;
	}
	imglen += exehdr.imgmod;

//	printf("exehdr.imglen = %u (%04xh)\n", exehdr.imglen, exehdr.imglen );
//	printf("exehdr.imgmod = %u (%04xh)\n", exehdr.imgmod, exehdr.imgmod );

	if((imglen >> 4) < data_seg)
	{
		error(12, (unsigned) (imglen >> 4));
	}

 	prog_text = new(BUF_SIZE + 2 + 16);			/* add 2 for Fixup on BUF_SIZE */
															/* add 16 for extra para for negative searching */

	prog_text += 16;									/* Move 16 bytes into buffer for actual code */

	lseek( exe, (long)exehdr.hdrlen * 16, 0 );      /* Read Prog Text */
	rxe_fixups = count_rxe_fixups( data_seg );
	lseek( exe, (long)exehdr.hdrlen * 16, 0 );      /* Read Prog Text */

	if(do_getsize)
	{
		FILE * f;
		assert(partial == 0);
		size = imglen + ((ulong)rxe_base_paras * 16);
		if(quit)
		{
			f = fopen("rxe_size.$$$", "w");
			fprintf(f, "%ld\n", size);
			fclose(f);
		}
		else
		{
			printf("XIP image size = %ld\n", size);
		}
		exit(0);
	}

	modify_exe_header( rxe_fixups );

	rxe_write( &new_exehdr, sizeof(exehdr) );		/* Write header */

	rxe_lseek( (long)exehdr.reldsp, 0 );

	rxe_write( hpFixupList, (unsigned long)new_exehdr.nreloc *
			sizeof(struct rel_item) );

							/* Write Rel items */

	rxe_lseek((long)exehdr.hdrlen * 16L, 0);	/* Lseek to Code */

	count = 0L;
	text_len = imglen - (exehdr.hdrlen * 16L);

	data_block_size = text_len - ((long)dseg << 4L);
	code_block_size = text_len - data_block_size;

	if(partial)
	{
		error(17);
	}

	assert(text_len > data_block_size);
	assert(text_len > code_block_size);

	rxe_base->cd_sum = FixUp1Block(code_block_size, count, dseg, rxe_int);

	rxe_base->ld_seg = mseg;
	rxe_base->int_no = rxe_int;
	rxe_base->cd_seg = mseg + exehdr.hdrlen;
	rxe_base->cd_cs = mseg + exehdr.hdrlen + exehdr.csdsp; /* BUG: Fix when smaller header */
	rxe_base->cd_ip = exehdr.ipofs;
	rxe_base->cd_len = dseg;
	rxe_base->rxe_base_len = rxe_base_paras;

//  Debug:  Location where the XIP header is being written to disk
//
#if DEBUG
	{
		extern int output;

		printf("xip header: %lxh\n",	lseek( output, 0, SEEK_CUR ) );
	}
#endif

	rxe_write( rxe_base, (ulong)rxe_base_paras * 16);

	if(partial)
	{
		/*  Move back 1 para if code and data in same para */
		lseek( exe, (long) -16L, 1 );
	}

//  Debug:  Location where the data is being written to disk
//
#if DEBUG
	{
		extern int output;

		printf("data: %lxh\n",	lseek( output, 0, SEEK_CUR ) );
	}
#endif

	FixUp1Block(data_block_size, count, dseg, rxe_int);

	do{					/* if CodeView Symbol table */
		len = (unsigned)(read(exe, prog_text, 512));
		if(len)
		{
			rxe_write(prog_text, len);
		}
	}while(len);

	update_header();

	report();
	close(exe);
}


/*
 * Fixup up a segment with in a RXE.
 * 	Code segments are fixed up at this time!
 * 	Data Segment are left in the hpRelItems list and final fixup is
 *		performed by ROM-DOS.
 */
unsigned do_fixup(unsigned seg)
{
	int ret;


	if(seg < data_seg)
	{
		ret = (seg + memory_seg + exehdr.hdrlen); 	/* Code area */
	}
	else
	{
		ret =  (seg - data_seg) + rxe_base_paras;	/* Data area */
	}


	return (ret);
}


/*
 * Delete_reloc
 * Removes a relocatable item from the relocatable item list
 */
unsigned delete_reloc( unsigned uNumber )
{
	unsigned uIndex;


	for( uIndex = uNumber; uIndex < exehdr.nreloc; uIndex++ )
	{
		hpRelItems[uIndex].segment = hpRelItems[uIndex+1].segment;
		hpRelItems[uIndex].offset  = hpRelItems[uIndex+1].offset ;
	}

	exehdr.nreloc--;

	return (uNumber - 1);
}


/*
 * Modify the EXE header to account for changes in EXE to
 * support XIP
 */
void modify_exe_header(int total_fixups)
{
	unsigned long imglen;
	memcpy(&new_exehdr, &exehdr, sizeof(struct exehdr));
	new_exehdr.nreloc = total_fixups;
	new_exehdr.ssdsp = do_fixup(exehdr.ssdsp);
	new_exehdr.hdrlen += data_seg;
	if(partial)
	{
		++new_exehdr.hdrlen;
	}
	new_exehdr.csdsp = 0;			/* By definition */
	new_exehdr.ipofs = rxe_base_ip;

	imglen  = (exehdr.imglen-1) * 512L;
	if(exehdr.imgmod == 0)
	{
		imglen += 512;
	}
	else
	{
		imglen += exehdr.imgmod;
	}

	imglen += (ulong)rxe_base_paras * 16;
	if(partial)
	{
		imglen += 16;
	}

	new_exehdr.imgmod = (imglen % 512L);
	new_exehdr.imglen = ((imglen + 511) / 512);
}


/*
 * Get_entry
 * Get entry gets a data segment entry from the entry table
 * if it exists or adds it to the table and returns the new entry.
 */
unsigned get_entry(unsigned fixup)
{
	unsigned i;
	fixup = do_fixup(fixup);
	for(i=0; i<data_seg_entry; ++i){	/* Can I find it? */
		if(fixup == data_seg_table[i])
		{
			return i;
		}
	}

	data_seg_table[data_seg_entry] = fixup;	/* Then Add it to table */
	++data_seg_entry;
	if(data_seg_entry >= DATA_SEG_TABLE_SIZE)
	{
		error(3);
	}


	return i;
}


/*
 * New
 * New is a C malloc with error checking
 */
void far * new( unsigned long ulNumBytes )
{
	void far * lp;


	lp = farmalloc( ulNumBytes );
	if(lp == NULL)
	{
		error(2);
	}

	return lp;
}


void report(void )
{
	unsigned long code_size, data_size, rxe_base_size;

	if(quit)
	{
		return;
	}

	code_size = (new_exehdr.hdrlen - exehdr.hdrlen) * 16L;

	rxe_base_size = (ulong)rxe_base_paras * 16;

	data_size  = (new_exehdr.imglen-1) * 512L;
	data_size += new_exehdr.imgmod;
	data_size -= new_exehdr.hdrlen * 16L;

	printf("\n\nRXE Conversion Statics\n\n");
	printf("\tCODE Block size is %6lu\n", code_size);
	printf("\tXIP Loader size is %6lu\n", rxe_base_size);
	printf("\tDATA Block size is %6lu\n", data_size - rxe_base_size);
	printf("\t-------------------------\n");
	printf("\tTotal RAM savings = %u%%\n\n",
			(code_size * 100)/(code_size + data_size));
}

#define MAX_FIXUPS 4000

int count_rxe_fixups( unsigned data_seg )
{
	unsigned uIndex;
	unsigned uCount = 0;
	long offset;
	unsigned header, fixup, new_fixup;

	header = exehdr.hdrlen * 16;

	hpFixupList = new(MAX_FIXUPS * sizeof(struct rel_item));

	for( uIndex = 0; uIndex < exehdr.nreloc; uIndex++ )
	{
		offset = ( (long)hpRelItems[uIndex].segment << 4 ) +
				hpRelItems[uIndex].offset;

		/* if DATA block */
		if( hpRelItems[uIndex].segment >= data_seg )
		{
			lseek(exe, header+offset, 0);
			read(exe, &fixup, 2);
			if(fixup >= data_seg)
			{
				hpFixupList[uCount].segment =
						do_fixup( hpRelItems[uIndex].segment );
				hpFixupList[uCount].offset = hpRelItems[uIndex].offset;
				new_fixup = do_fixup( fixup );

#if 0
				//  Debug code
				//
				printf("Fixup %d was %04X:%04X to (uCount %u) %04X:%04X\n",
						uIndex, hpRelItems[uIndex].segment,
						hpRelItems[uIndex].offset,	uCount,
						hpFixupList[uCount].segment, hpFixupList[uCount].offset);
				printf("Fixup is %04X and s/b %04X\n", fixup, new_fixup);
#endif

				lseek(exe, header+offset, 0);
				write(exe, &new_fixup, 2);
				assert( uCount < MAX_FIXUPS );
				++uCount;
			}
		}
	}


	return uCount;
}

#define CS_OVERRIDE 0x2E
#define DS_OVERRIDE 0x3E
#define ES_OVERRIDE 0x26
#define SS_OVERRIDE 0x36
#define IS_A_SEG_OVERRIDE(x) (((x) == CS_OVERRIDE) ||\
															((x) == DS_OVERRIDE) ||\
															((x) == ES_OVERRIDE) ||\
															((x) == SS_OVERRIDE)   \
														 )

/*
 * Do block either a Code Block or a Data Block
 * all the conversion and fixing up is performed in this routine.
 */
unsigned FixUp1Block( long text_len, long count, unsigned dseg,
		unsigned rxe_int)
{
	unsigned uIndex;
	int j, reg, complex = 99, Nop = FALSE;
	unsigned char mode, msb_offset, lsb_offset;
	unsigned long pos;
	static unsigned long text_pos = 0L;
	static int fFirstTime = TRUE;

	int intr_index;

	unsigned fixup, seg, off, read_size = 0;
	unsigned sum = 0;
	unsigned long len = 0;

	extern int output;


	while(text_len > count)
	{
		if((unsigned long) (text_len - count) > (unsigned long) BUF_SIZE)
		{
			read_size = BUF_SIZE;
		}
		else
		{
			read_size = (unsigned)(text_len - count);
		}

/*
printf("text_len = %ld count %ld (text_len-count) = %ld BUF_SIZE %ld read_size %u\n", text_len, count, text_len-count, (long) BUF_SIZE, read_size);
*/
		if(read_size == 0)
		{
			printf("Text_len = %ld  Count = %ld\n", text_len, count);
		}
		assert(read_size != 0);

#if 1
		if( fFirstTime )
		{
			lseek(output, -16L, 1);      /* Get Previous 16 bytes! */  // in RXE file
			read(output, prog_text-16, 16);
			fFirstTime = FALSE;
		}
		len = (unsigned)(read(exe, prog_text, read_size));
#else
		lseek(exe, -16L, 1);      /* Get Previous 16 bytes! */
		len = (unsigned)(read(exe, prog_text-16, read_size+16));		/* add 16 for para before ! */
		len -= 16;
#endif

		if(len != read_size)
		{
			error(7);
		}


		/* go through each item in table. */
		for( uIndex = 0; uIndex < exehdr.nreloc; uIndex++ )
		{
			seg = hpRelItems[uIndex].segment;
			off = hpRelItems[uIndex].offset;
			pos = ((long)seg << 4) + off;

			/* if it's not in the currently loaded block of
				code/data, ignore it */
			if(pos < text_pos || pos >= (text_pos + len))
			{
				continue;
			}


//  Debug code used to have a place to set a breakpoint when
//  evaluating a curtain fixup.  Just change the segment and
//  offset to match the fixup address you are looking for
//
#if 0
			if( (hpRelItems[uIndex].offset == 0x0131) &&
					(hpRelItems[uIndex].segment == 0x57ED) )
			{
				unsigned uBT;

				//  Set breakpoint here
				//
				uBT = 1;
			}
#endif

			// If a fixup on a buffer boundry, get 1 more byte for
 			// complete fixup
			if(pos +1 == text_pos + BUF_SIZE)
			{
				if(read(exe, prog_text+BUF_SIZE, 1) != 1)
					error(7);
				++read_size;
			}

			uptr = (unsigned *) ((char *)prog_text +
					(unsigned)(pos - text_pos));
/*
			if(!(uptr > (unsigned *)(prog_text + len)));
				printf("uptr = %p prog_text + len = %p\n", uptr, prog_text + len);
*/

			fixup = *uptr;

#if DEBUG
//	printf("i %d, seg:off %04X:%04X fixup %04X\n",i,seg,off,fixup);
#endif

			/* If a CODE fixup (simplest case) */
			if(fixup < data_seg)
			{
				*uptr = do_fixup(fixup);
			}

		 	/* if DATA fixup in CODE area */
			else if(seg < data_seg)
			{
				// This array contains the first byte of each of the TEST, MOV, IMUL,
				// and conditional coniditional (ADD, OR, ADC, SBB, AND, SUB, XOR
				// and CMP) which each take imm16 data and so could theoretically
				// conflict with any of the supported functions that we support.
				//
				// Put our SUPPORTED opcode first in the array so we don't test for
				// that more than once when applying our maximal match algorithm.
				//
				static unsigned char rgucConflictingOpcodes[] = {0xC7,0x69,0x81,0xF7};
				int i;
				int fToldYouItWasAmbiguous = FALSE;

  			Nop = FALSE;
				byte_ptr = (unsigned char *)uptr -1;

				/*
				   Check for Ambiguous Fixups

					 We are trying to tell if there is a single byte
					 "mov r16, imm16" opcode {B8,B9,...BF}  or a
					 "push imm16" opcode (0x68) that is embedded within
					 a "mov rm16, imm16" opcode sequence "C7 modr/m ..."
					 If there is then the C7 may be spurious data and one
					 of the other opcode sequences may be correct.

					 AND... that same abiguity in our supported single byte
					 instructions could exist in ANY multibyte opcode that takes
					 imm16 data, include the ones we don't support.

				   OR...it could be that their is data in an actual
				   "C7 modr/m..." sequence that just coindidentally has
				   the same value as a single byte opcode that we support.
				   Either way we can't be sure what we're supposed to do...

					 AND... the same ambiguity in decoding our multibyte opcodes
				   could exist with any single byte opcode that takes imm16 data,
				   including the ones that we DON'T support.

				 */
				switch(byte_ptr[0])
				{
					// If this is one of the single byte opcodes we support
					//
					case 0x68: // push imm16
					case 0xB8: // mov r16, imm16
					case 0xB9:
					case 0xBA:
					case 0xBB:
					case 0xBC:
					case 0xBD:
					case 0xBE:
					case 0xBF:

						// then for each possibly conflicting multibyte instruction
						//
						for(i=0; i<4; ++i)
						{
							// at each position where it could possibly occur
							//
							for(j=-3; j<0; ++j)
							{
								// if one of those bytes is found
								//
								if(byte_ptr[j] == rgucConflictingOpcodes[i])
								{
									// gripe about it, but only gripe once per fixup
									//
									warning(19, seg, off);
									fToldYouItWasAmbiguous = TRUE;
									goto ToldYouSo;
								}
							}
						}
						break;

					//  if this is an unsupported single byte opcode
					//
					case 0x05:	// add ax, imm16
					case 0x0D:	// or ax, imm16
					case 0x15:	// adc ax, imm16
					case 0x1D:	// sbb ax, imm16
					case 0x25:	// and ax, imm16
					case 0x2D:	// sub ax, imm16
					case 0x35:	// xor ax, imm16
					case 0x3D:	// cmp ax, imm16
					case 0xA9:	// test ax, imm16
					case 0xC2:	// retn imm16
					case 0xC8:	// enter imm16, imm8
					case 0xCA:	// retf imm16

						// then look for our supported multibyte instructions
						// at every location they might occur
						//
						for(j=-3; j<0; ++j)
						{
							if(byte_ptr[j] == 0xC7)
							{
								// if we found a conflict then gripe about it,
								// but only once per fixup
								//
								warning(19, seg, off);
								fToldYouItWasAmbiguous = TRUE;
								goto ToldYouSo;
							}
						}
						// rut roh...
						// we discovered an unsupported single byte opcode that
						// was not part of an instruction sequence we supported.
						// Abort.
						//
						error(4,seg,off);

ToldYouSo:
					default:
						break;

				}  // switch(byte_ptr[0])

				// Decode the one byte instructions first.
				// These are moves directly into registers and can't
				// have a meaningful segment override, so don't look for one.
				//
				//  None of these take meaningful segment overrides and all
				//  ambiguities in the single byte instructions have already been
				//  discovered and announced.
				//
				complex = 0;
				intr_index = 0;
				switch(*byte_ptr)
				{
					// "mov r16, SEG ..." (mov r16, imm16: B8+rw ?? ??)
					//
					case 0xB8:	/* MOV AX,? */
						reg = 0; goto mk_rxe_int;
					case 0xBB:	/* MOV BX,? */
						reg = 1; goto mk_rxe_int;
					case 0xB9:	/* MOV CX,? */
						reg = 2; goto mk_rxe_int;
					case 0xBA:	/* MOV DX,? */
						reg = 3; goto mk_rxe_int;
					case 0xBE:	/* MOV SI,? */
						reg = 4; goto mk_rxe_int;
					case 0xBF:	/* MOV DI,? */
						reg = 5; goto mk_rxe_int;
					case 0xBD:	/* MOV BP,? */
						reg = 6; goto mk_rxe_int;
					case 0xBC:	/* MOV SP,?  is illegal */
						error(18, seg, off);

					// "push SEG ..." (push imm16: 68 ?? ??)
					//
					case 0x68:
						intr_index = 1;
						reg = 0;
						goto mk_rxe_int;

					default:
						break;
				}

				// No supported single byte instructions found for this fixup.
				// Now look for "mov rm16, imm16" there are 4 possibilities:
				//
				// opcode modr/m   	...
				//  C7		11??????	FIXUP
				//  C7		00??????	FIXUP
				//  C7		01??????	Disp8		FIXUP
				//  C7		10??????	Disp16	FIXUP
				//
				// There could also be something like
				//  	mov	al, C7
				//    C7 modr/m imm8 FIXUP
				// which RXE's maximal match algorith will horribly mutilate.
				// To stifle this destructive tendency we must make sure that the
				// bytes we think we want to emit actually cause the FIXUP to end
				// up at the place where the rest of the universe expects it to be.
				// If our maximal match algorithm does not succeed in locating the
				// fix up properly then we warn the user and decrement the pattern
				// length and try again.  If we can't match the pattern at all
				// then we abort.  Otherwise we cheerfully modify the code.
				//
				for(j=-3; j <0; ++j)
				{
					// set to TRUE if we can reasonably believe we properly decoded
					// the instructions before the fixup.
					//
					int fFoundCode = FALSE;

					// byte_ptr[0] is the byte immediately in front of the FIXUP
					//
					if(byte_ptr[j] == 0xC7)
					{
						mode = byte_ptr[j+1];

						// This here case is "C7 11??????b FIXUP".  These are
						// moves directly into a register and cannot have a
						// meaningful segment override, so don't look for one.
						//
						if((mode & 0xC0) == 0xC0)
						{
 						  // If we find this 2 byte instruction at any place other than
							// at byte_ptr[-1] then we have a false match, try again.
							//
							if(j != -1)
							{
 								if(!fToldYouItWasAmbiguous)
								{
									warning(19, seg, off);
									fToldYouItWasAmbiguous = TRUE;
								}
								continue;
							}

							// All ambiguities with single byte opcodes have already
							// been detected and announced. But now we need to test
							// to see if what we think is a supported instruction
							// could really be just data inside an UNSUPPORTED instruction
							// sequence.  If turns out this is possible and we haven't
							// already complained about this fixup then do it now.
							//
							if(!fToldYouItWasAmbiguous)
							{
								// Already looked at the 0xC7 opcode to get here. If that
								// was wrong then the ambiguity would have already announced
								// so we know that it is not there.  Don't check for it.
								//
								for(i = 1; i<4; ++i)
								{
									if(byte_ptr[j-1] == rgucConflictingOpcodes[i])
									{
										warning(19, seg, off);
										fToldYouItWasAmbiguous = TRUE;
									}
								}
							}

							complex = 0;
							byte_ptr = &byte_ptr[j];
							Nop = TRUE;
							switch(mode & 7)
							{
								case 0: reg = 0; break;		// AX
								case 1: reg = 2; break;		// CX
								case 2: reg = 3; break;		// DX
								case 3: reg = 1; break;		// BX
								case 5: reg = 6; break;		// BP
								case 6: reg = 4; break;		// SI
								case 7: reg = 5; break;		// DI

								case 4: 									// SP
										error(18, seg, off);
								default:
										printf("Error: Unsupported instructions at Seg:Off = %04X:%04X Mode = %02X\n", seg, off, mode);
										assert(0);
							}
							fFoundCode = TRUE;
							goto mk_rxe_int;
						}

						switch(mode){

							// These are the 4 byte "mov r16+disp16, imm16" instructions
							//
							case 0x80:
							case 0x81:
							case 0x82:
							case 0x83:
							case 0x84:
							case 0x85:
							case 0x86:
							case 0x87:
							case 0x6:
	 						  // If we find this 4 byte instruction at any place other than
								// at byte_ptr[-3] then we have a false match, try again.
								//
								if(j != -3)
								{
	 								if(!fToldYouItWasAmbiguous)
									{
										warning(19, seg, off);
										fToldYouItWasAmbiguous = TRUE;
									}
									continue;
								}

								// If there might be a seg override here then whine
								// about it.  We can't reliably detect them.
								//
								if(IS_A_SEG_OVERRIDE(byte_ptr[j-1]))
								{
									warning(21, seg, off);
								}
								msb_offset = byte_ptr[0];
								lsb_offset = byte_ptr[-1];
								complex = 3;
								fFoundCode = TRUE;
								break;

							// These are the 3 byte "mov r16+disp8, imm16" instructions
							//
							case 0x40:
							case 0x41:
							case 0x42:
							case 0x43:
							case 0x44:
							case 0x45:
							case 0x46:
							case 0x47:
	 						  // If we find this 3 byte instruction at any place other than
								// at byte_ptr[-2] then we have a false match, try again.
								//
								if(j != -2)
								{
	 								if(!fToldYouItWasAmbiguous)
									{
										warning(19, seg, off);
										fToldYouItWasAmbiguous = TRUE;
									}
									continue;
								}

								// All ambiguities with single byte opcodes have already
								// been detected and announced. But now we need to test
								// to see if what we think is a supported instruction
								// could really be just data inside an UNSUPPORTED instruction
								// sequence.  If turns out this is possible and we haven't
								// already complained about this fixup then do it now.
								//
								if(!fToldYouItWasAmbiguous)
								{
									// Already looked at the 0xC7 opcode to get here. If that
									// was wrong then the ambiguity would have already announced
									// so we know that it is not there.  Don't check for it.
									//
									for(i = 1; i<4; ++i)
									{
										if(byte_ptr[j-1] == rgucConflictingOpcodes[i])
										{
											warning(19, seg, off);
											fToldYouItWasAmbiguous = TRUE;
										}
									}
								}

								// If there might be a seg override here then whine
								// about it.  We can't reliably detect them.
								//
								if(IS_A_SEG_OVERRIDE(byte_ptr[j-1]))
								{
									warning(21, seg, off);
								}
								lsb_offset = byte_ptr[0];
								complex = 2;
								fFoundCode = TRUE;
								break;

							// These are the 2 byte "mov r16, imm16" instructions
							// These moves use either register indirection or a straight
							// 16-bit displacement.
							//
							case 0x00:
							case 0x01:
							case 0x02:
							case 0x03:
							case 0x04:
							case 0x05:
							case 0x07:
	 						  // If we find this 2 byte instruction at any place other than
								// at byte_ptr[-1] then we have a false match, try again.
								//
								if(j != -1)
								{
	 								if(!fToldYouItWasAmbiguous)
									{
										warning(19, seg, off);
										fToldYouItWasAmbiguous = TRUE;
									}
									continue;
								}

								// All ambiguities with single byte opcodes have already
								// been detected and announced. But now we need to test
								// to see if what we think is a supported instruction
								// could really be just data inside an UNSUPPORTED instruction
								// sequence.  If turns out this is possible and we haven't
								// already complained about this fixup then do it now.
								//
								if(!fToldYouItWasAmbiguous)
								{
									// Already looked at the 0xC7 opcode to get here. If that
									// was wrong then the ambiguity would have already announced
									// so we know that it is not there.  Don't check for it.
									//
									for(i = 1; i<4; ++i)
									{
										if(byte_ptr[j-1] == rgucConflictingOpcodes[i])
										{
											warning(19, seg, off);
											fToldYouItWasAmbiguous = TRUE;
										}
									}
								}

								// If there might be a seg override here then whine
								// about it.  We can't reliably detect them.
								//
								if(IS_A_SEG_OVERRIDE(byte_ptr[j-1]))
								{
									warning(21, seg, off);
								}
								complex = 1;
								fFoundCode = TRUE;
								break;

							default:
								printf("Error: Unsupported instructions at Seg:Off = %04X:%04X Mode = %02X\n", seg, off, mode);
								assert(0);
						} // switch(mode)

						if(fFoundCode)
						{
							byte_ptr += j;
							reg = 7;
							goto mk_rxe_int;
						}
					}  // if(byte_ptr[j] == 0xC7)

				// If we fall out of this for(..) loop then we had encountered
				// a Data fixup in a Code block, the instructions for which we
				// could not decode.  That is a fatal error.
				//
				} /* end of for(j=) loop */
				error(4, seg, off);

mk_rxe_int:
				byte_ptr[0] = 0xCD;
				byte_ptr[1] = RXE_INTERRUPT + intr_index;
				byte_ptr[2] = (reg << 5) | get_entry(fixup);
				switch(complex){
					case 3:byte_ptr[5] = msb_offset;
					case 2:byte_ptr[4] = lsb_offset;
					case 1:byte_ptr[3] = mode;
					case 0:break;
				}
				if(Nop)
				{
					byte_ptr[3] = NOP;
				}

				uIndex = delete_reloc( uIndex );

			} // if (fixup code block)

			/* Data seg fixup in data block */
			else
			{
				*uptr = do_fixup(fixup);
				hpRelItems[uIndex].segment = do_fixup(hpRelItems[uIndex].segment);
			}

		} /* end of for (i<exehdr.nreleoc) loop */

		if(count < (dseg * 16L))
		{
			if(count + len < (dseg * 16L))
			{
				sum += short_sum((void far *)prog_text, len / 2);
			}
			else
			{
				sum += short_sum((void far *)prog_text, (unsigned) ((dseg * 16L)-count)/2);
			}
		}

      // Now back up 16 bytes, which is our buffer "underflow" region
      // and write that data first.
      //
		lseek( output, -16L, SEEK_CUR );
		write(output, prog_text-16, 16);

      // Then write our main buffer
      //
		len = rxe_write(prog_text, read_size);

		if(len != read_size)
		{
			error(7);
		}
		count += len;
		text_pos += len;

      // Finally, move the last 16 bytes of the buffer back to the "underflow"
      // region, in case there are fixup instructions that start there.
      //
		memmove( prog_text - 16, prog_text + read_size - 16, 16);

	}


	return sum;
}


/*
 * init_rxe_base performs initial loading and initialization of the
 * RXE_BASE XIP loader
 */
void init_rxe_base(void)
{
	struct exehdr exehdr;
	int h, ret;
	unsigned imglen;
	extern char * RXERootPath;
	char rxe_init[66], * p;

	strcpy(rxe_init, "RXE_INIT.EXE");
	h = _open(rxe_init, 0);

	if(h == -1)								/* If not then try the same place */
	{
		strcpy(rxe_init, RXERootPath);	/* as the RXE_CVT.EXE program */
		p = strrchr(rxe_init, '\\');
		assert(p != NULL);
		strcpy(p+1, "RXE_INIT.EXE");		/* Move over the "\" */
		h = _open(rxe_init, 0);
	}

	if(h == -1)
	{
		error(14);
	}

	read(h, &exehdr, sizeof(struct exehdr));

	imglen  = (exehdr.imglen-1) * 512;
	if(exehdr.imgmod == 0)
	{
		imglen += 512;
	}
	else
	{
		imglen += exehdr.imgmod;
	}

	imglen -= 512;
	rxe_base_ip = exehdr.ipofs;

	lseek(h, 512L, 0);

#if 1
	rxe_base_paras = imglen + 15;
	rxe_base = malloc(rxe_base_paras);
	memset( rxe_base, 0, rxe_base_paras );
	rxe_base_paras /= 16;
#else
	rxe_base = malloc(imglen);
#endif

	ret = read(h, rxe_base, imglen);
	if(ret != imglen)
	{
		printf("imglen = %d ret = %d\n", imglen, ret);
		assert(0);
	}
	close(h);
//	rxe_base_paras = (imglen + 15) / 16;
}

/*
 * Add a fixup to the EXE fixup table for the
 * rxe_base.rxe_fixup_table
 */
void add_fixup(int i)
{
#if 0
	hpRelItems[new_exehdr.nreloc].segment = 0;
	hpRelItems[new_exehdr.nreloc].offset = (unsigned)
		(((char *)&rxe_base->rxe_fixup_table[i])-(char *)rxe_base);
#else
	hpFixupList[new_exehdr.nreloc].segment = 0;
	hpFixupList[new_exehdr.nreloc].offset = (unsigned)
		(((char *)&rxe_base->rxe_fixup_table[i])-(char *)rxe_base);
#endif

	new_exehdr.nreloc++;
}


/*
 * After the processing is done then update the header with the new fixups,
 * changes the the exehdr, and rewrite the rxe_base structure with the
 * nessary fixups placed in it.
 */
void update_header(void)
{
	extern int output;
	unsigned uIndex;
	long lRet;

	for( uIndex=0; uIndex < data_seg_entry; uIndex++ )
	{
		add_fixup( uIndex );
	}

	lRet = lseek( output, 0L, 0 );
	if(lRet == -1)
	{
		error(7);
	}

	lRet  = write( output, &new_exehdr, sizeof(exehdr) );
	if( lRet != sizeof(exehdr) )
	{
		error(7);
	}

	lRet = lseek(output, (long)new_exehdr.reldsp, 0);
	if(lRet == -1)
	{
		error(7);
	}

	lRet = BigWrite( output, hpFixupList, (unsigned long)new_exehdr.nreloc *
			sizeof(struct rel_item) );
	if( lRet != ((unsigned long)new_exehdr.nreloc * sizeof(struct rel_item)) )
	{
		error(7);
	}

//	assert( ((unsigned long)(long)new_exehdr.hdrlen * 16L) > 65535L );
	lRet = lseek( output, (long)new_exehdr.hdrlen * 16L, 0 );
	if(lRet == -1)
	{
		error(7);
	}

	_fmemcpy( &rxe_base->rxe_fixup_table[0], data_seg_table,
			data_seg_entry * sizeof(unsigned) );

//  No longer needed
//	assert( ((unsigned long)rxe_base_paras * 16) > 65535L );
	lRet = BigWrite( output, rxe_base, (ulong)rxe_base_paras * 16 );
	if(lRet  != ((long)rxe_base_paras * 16))
	{
		error(7);
	}
}

/*
 * is_xip returns TRUE or FALSE depending if the program is a Datalight
 * XIP program.
 */
int is_xip(int exe)
{
	struct rxe_base rxe_base;

	lseek(exe, exehdr.hdrlen * 16L, 0);
	read(exe, &rxe_base, sizeof(rxe_base));
	return (rxe_base.rxe_sig == XIPSIG);
}


//
//
long BigRead( int iHandle, void huge * hpBuf, unsigned long ulLength )
{
	unsigned long ulBytesToRead;
	unsigned uReadSize;
	int iRet;


	//  Continue reading from the file in small blocks until the
	//  requested size is read
	//
	ulBytesToRead = ulLength;
	while( ulBytesToRead )
	{
		//  Determine the read size
		//
		if( ulBytesToRead > 32768L )
		{
			uReadSize = (unsigned)(32768L);
		}
		else
		{
			uReadSize = (unsigned)ulBytesToRead;
		}


		//  Read the data
		//
		iRet = read( iHandle, (void far *)hpBuf, uReadSize );
		if( iRet == -1 )
		{
			return iRet;
		}


		//  Adjust the number of bytes left to read
		//
		ulBytesToRead -= (unsigned)iRet;


		//  Adjust the destination address
		//
		(char huge *)hpBuf += (unsigned)iRet;


		//
		//
		if( iRet == 0 )
		{
			break;
		}
	}


	return (long)(ulLength - ulBytesToRead);
}



//
//
long BigWrite( int iHandle, void huge * hpBuf, unsigned long ulLength )
{
	unsigned long ulBytesToWrite;
	unsigned uWriteSize;
	int iRet;


	//  Continue writing from the file in small blocks until the
	//  requested size is write
	//
	ulBytesToWrite = ulLength;
	while( ulBytesToWrite )
	{
		//  Determine the write size
		//
		if( ulBytesToWrite > 32768L )
		{
			uWriteSize = (unsigned)(32768L);
		}
		else
		{
			uWriteSize = (unsigned)ulBytesToWrite;
		}


		//  Write the data
		//
		iRet = write( iHandle, (void far *)hpBuf, uWriteSize );
		if( iRet == -1 )
		{
			return iRet;
		}


		//  Adjust the number of bytes left to write
		//
		ulBytesToWrite -= (unsigned)iRet;


		//  Adjust the destination address
		//
		(char huge *)hpBuf += (unsigned)iRet;


		//
		//
		if( iRet == 0 )
		{
			break;
		}
	}


	return (long)(ulLength - ulBytesToWrite);
}
