;
;   $Id: MASM.INC 1.2 2002/10/16 16:01:26Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
pushcontext listing
.nolist

COMMENT ~ /*-------------------------------------------------------------
        $DOCFILE:MASM.INC

        Copyright (C) Datalight, Inc. 1997
        All Rights Reserved.
        Datalight Confidential and Proprietary.

        This include contains MASM specific Datalight standard defines
        and macros.

        This include should be included in all Datalight assembly
        language modules.  It is generally automatically included by
        DATALGHT.INC.

	For a detailed description of the MASM prologue/epilogue
	code generation process, see the MASM 6.1 Programmer's Guide
	pages 201-204.

        $DOCHISTORY
        02/17/97 gp     Created
	03/20/97 gp	First published version.  Major cleanup and docs.
	10/13/97 gp	Added the ?MasmProcSetup macro.
----------------------------------------------------------------------*/ ~

;; prologue flags used by MASM
PROLFLAGS_LANG_MASK     equ <0007h>
PROLFLAGS_LANG_UNKNOWN  equ <0000h>
PROLFLAGS_LANG_C        equ <0001h>
PROLFLAGS_LANG_SYSCALL  equ <0002h>
PROLFLAGS_LANG_STDCALL  equ <0003h>
PROLFLAGS_LANG_PASCAL   equ <0004h>
PROLFLAGS_LANG_FORTRAN  equ <0005h>
PROLFLAGS_LANG_BASIC    equ <0006h>
PROLFLAGS_RESERVED1     equ <0008h>
PROLFLAGS_CALLERCLEAN   equ <0010h>
PROLFLAGS_FARPROC       equ <0020h>
PROLFLAGS_PRIVATE       equ <0040h>
PROLFLAGS_EXPORT        equ <0080h>
PROLFLAGS_IRET          equ <0100h>
PROLFLAGS_RESERVED2     equ <FE00h>

COMMENT ~ /*-----------------------------------------------------------------
        $DOCMACRO:MASMPROLOGUE

        This macro is used internally by MASM when the PROC directive
        is used.  MASM is instructed to use this macro with the
        "OPTION PROLOGUE:xxx" directive.

        For a detailed description of this process, see the MASM 6.1
        Programmer's Guide pages 201-204.
  
        Valid rgUserParams are:
  
          LOADDS        - Causes DS to be inited on entry
          NOLOADDS      - Prevents DS from being inited on entry
          INCBP         - Causes the saved BP to be incremented for FAR procs
          NOINCBP       - Prevents the saved BP from being incremented
          FORCEFRAME    - Forces a stack frame to be created even if it is
                          otherwise not needed.
  
        These optional parameters will override any module defaults that have
        been specified with the PROC_LOADDS, PROC_INCBP, and PROC_FORCEFRAME
        symbols.
  
        Note that stack frames are always set up if there are any arguments
        or locals, or the INCBP flag is set.
  
        For EXPORT functions, the LOADDS option is required (if not specified
        as a module default) to cause the proper prologue code to be generated
        for a Windows load time fixup.
  
        For non-EXPORT functions, the LOADDS option simply causes DS to be
        loaded with DGROUP with AX getting trashed in the process.
  
        Whenever LOADDS is used, DS is automatically saved and restored so
        there is no need to save it in the USES list.

	$DOCTODO
	Update to allow 32 bit stacks.
-------------------------------------------------------------------------*/ ~

MASMPROLOGUE    MACRO szProcName, flags, cbParams, cbLocals, rgRegs, rgUserParams, ovr
LOCAL   ?cbLocals
.errnb <ovr>            ;; no parameter overflow

;; ?cbLocals may be adjusted if additional bytes are saved between the
;; pushed BP and the local variables.
?cbLocals	= cbLocals

;; init basic defaults to simplify the if/else logic below
?procDoFrame	= FALSE
?procLoadDS	= FALSE
?procIncBP	= FALSE

;;------------------------------------------------------------------
;; Create tentative values based on the module defaults defined with
;; the InitProcBehavior macro
;;------------------------------------------------------------------

;; Determine according to the module default if we want to load DS

if PROC_LOADDS EQ PROCINIT_ALWAYS
  ?procLoadDS = TRUE
elseif PROC_LOADDS EQ PROCINIT_EXPORT
  if flags AND PROLFLAGS_EXPORT
    ?procLoadDS = TRUE
  endif
elseif PROC_LOADDS EQ PROCINIT_DBG
  if DEBUG 
    ?procLoadDS = TRUE
  endif
elseif PROC_LOADDS EQ PROCINIT_NEVER
  ?procLoadDS = FALSE
else
  .err
  %out Unknown value for PROC_LOADDS = &PROC_LOADDS
endif

;; Determine according to the module default if we want to INC BP

if PROC_INCBP EQ PROCINIT_ALWAYS
  ?procIncBP = TRUE
elseif PROC_INCBP EQ PROCINIT_EXPORT
  if flags AND PROLFLAGS_EXPORT
    ?procIncBP = TRUE
  endif
elseif PROC_INCBP EQ PROCINIT_DBG
  if DEBUG 
    ?procIncBP = TRUE
  endif
elseif PROC_INCBP EQ PROCINIT_NEVER
  ?procIncBP = FALSE
else
  .err
  %out Unknown value for PROC_INCBP = &PROC_INCBP
endif

;; Determine according to the module default if we want to force a stack frame

if PROC_FORCEFRAME EQ PROCINIT_ALWAYS
  ?procDoFrame = TRUE
elseif PROC_FORCEFRAME EQ PROCINIT_EXPORT
  if flags AND PROLFLAGS_EXPORT
    ?procDoFrame = TRUE
  endif
elseif PROC_FORCEFRAME EQ PROCINIT_DBG
  if DEBUG 
    ?procDoFrame = TRUE
  endif
elseif PROC_FORCEFRAME EQ PROCINIT_NEVER
  ?procDoFrame = FALSE
else
  .err
  %out Unknown value for PROC_FORCEFRAME = &PROC_FORCEFRAME
endif

;;-------------------------------------------------------------------
;; Now that we have the system defaults determined as appropriate for
;; this PROC, look at the user specified parameters to see if any
;; are to be overridden.
;;-------------------------------------------------------------------

ifnb <rgUserParams>                     ;; parse user params if existing
  for p,<rgUserParams>                  ;; For every user param
    ifidn <p>, <LOADDS>                 ;; is it LoadDS
      ?procLoadDS = TRUE                ;; yes -- do loadds sequence
    elseifidn <p>, <NOLOADDS>           ;; don't do LoadDS?
      ?procLoadDS = FALSE               ;; yes -- clear loadds flag
    elseifidn <p>, <INCBP>              ;; is it IncBP
      if flags AND PROLFLAGS_FARPROC    ;; and far?
        ?procIncBP = TRUE               ;; yes -- do IncBP sequence
      else
        %out Warning: Ignoring INCBP for NEAR procedure
      endif
    elseifidn <p>, <NOINCBP>            ;; is it NoIncBP
      ?procIncBP = FALSE                ;; yes -- Clear the incbp flag
    elseifidn <p>, <FORCEFRAME>         ;; is it ForceFrame?
      ?procDoFrame = TRUE               ;; yes -- force out a frame
    else
      .err
      %out Unrecognized MASMPROLOGUE parameter &p
    endif
  endm                   ;; End of user parameter parsing loop
endif

;; Now determine if we really need any stack frame.  A stack frame is
;; needed if:
;;      1. ?procDoFrame is set
;;      2. ?procIncBP is set
;;      3. cbLocals NE 0
;;      4. cbParams NE 0

?procDoFrame = ?procDoFrame OR ?procIncBP OR (?cbLocals NE 0) OR (cbParams NE 0)

if ?procDoFrame EQ 0
  ;; just because we don't need a frame, we may still want DS loaded
  if ?procLoadDS
    if flags AND PROLFLAGS_EXPORT
        push    ds              ;; Put DS into AX -- we will place
        pop     ax              ;; back in DS later.  This sequence
        nop                     ;; is altered by the OS if needed
        push    ds
        mov     ds,ax
    else
        ;; not EXPORT, so do a normal DS load
        push    ds
        mov     ax,DGROUP
        mov     ds,ax
    endif
  endif

else  
  ;;--------------------------------------------
  ;; At this point we know we need a stack frame
  ;;--------------------------------------------

  if ?procLoadDS                  ;; if LOADDS and EXPORT, generate first
    if flags AND PROLFLAGS_EXPORT ;; half of DS init code.
        push    ds                ;; Put DS into AX -- we will place
        pop     ax                ;; back in DS later.  This sequence
        nop                       ;; is altered by the OS if needed
    endif
  endif

  if ?procIncBP                 ;; mark as a far procedure for stack
        inc     bp              ;; walking
  endif

        push    bp              ;; create the frame
        mov     bp,sp

  if ?procLoadDS                ;; Load up DS with the value in AX
    ?cbLocals = ?cbLocals + 2   ;; adjust internal locals count for use with EXITM
    if flags AND PROLFLAGS_EXPORT
        push    ds              ;;
        mov     ds,ax           ;; EXPORT functions just get patched in code
    else 
        push    ds              ;;
        mov     ax,DGROUP       ;; otherwise just use DGROUP
        mov     ds,ax       
    endif
  endif

  if cbLocals NE 0
        sub     sp,cbLocals     ;; make space on the stack for locals
  endif
endif


;;--------------------------------------------
;; Save any registers as specified
;;--------------------------------------------
ifnb rgRegs                     ;; There are registers to be saved
  for r,rgRegs                  ; save regs specified with USES
        PUSHX   r
  endm
endif

;;------------------------------------------------------------------
;; Get out of here, returning the number of bytes in locals, PLUS
;; any extra stuff we pushed prior to creating the space for the 
;; locals
;;------------------------------------------------------------------
        exitm <?cbLocals>

                ENDM



COMMENT ~ /*-----------------------------------------------------------------
        $DOCMACRO:MASMEPILOGUE
  
        This macro is used internally by MASM when it encounters a RET in 
        the code (should be the RET embedded in the ProcEnd macro).

	MASM is instructed to use this macro with the "OPTION EPILOGUE:xxx"
        directive.
  
        This macro gets the exact same parameters as passed to MASMPROLOGUE,
        so rather that recalculate the same things again, we rely on the
        three symbols ?procDoFrame, ?procLoadDS, and ?procIncBP to pass us
        the information we need to shut down correctly.
-------------------------------------------------------------------------*/ ~

MASMEPILOGUE    MACRO szProcName, flags, cbParams, cbLocals, rgRegs, rgUserParams, ovr
.errnb <ovr>    ;; not TOO many parameters
        
;;--------------------------------------------
;; Restore any registers as specified
;;--------------------------------------------
ifnb rgRegs                     ;; Pop off the registers -- they are in
                                ;; inverse order from the prologue call 
  for r,rgRegs                  ; pop regs specified with USES
        POPX    r
  endm
endif

if ?procDoFrame EQ 0
  if ?procLoadDS
        pop     ds              ;; restore DS when no stack frame in place
  endif
else
  if ?procLoadDS
        dec     bp
        dec     bp              ; point [bp] to the saved DS
        mov     sp,bp           ;; remove locals
        pop     ds              ;;
        pop     bp              ;; restore regs
  else
        mov     sp,bp
        pop     bp
  endif

  if ?procIncBP                 ;; Remove the increment of BP if necessary
        dec     bp               
  endif
endif

if flags AND PROLFLAGS_CALLERCLEAN ;; Caller pops stack arguments
        ret
else                            ;; Callee pops args
  if cbParams NE 0              ;; Put out the correct form of return
        ret     cbParams
  else
        ret
  endif
endif
                ENDM



COMMENT ~ /*-----------------------------------------------------------------
        $DOCMACRO:?MasmProcSetup

	This is an internally used macro that does the proc initialization
	for MASM when the high level constructs are used.
-------------------------------------------------------------------------*/ ~
?MasmProcSetup	MACRO	ovr
local           attribs, sizName
.errnb <ovr>    ;; not TOO many parameters

  ifidni $pLang,<NOLANGUAGE>
    $pLang equ <PASCAL>		;; MASM doesn't use NOLANGUAGE, so use PASCAL
  endif

  ;; build an "attribs" string appropriate for MASM
  attribs CATSTR $pDist,< >
  attribs CATSTR attribs,$pLang
  attribs CATSTR attribs,< >
  attribs CATSTR attribs,$pVis
;  attribs CATSTR attribs,< >
;  attribs CATSTR attribs,<<LOADDS>>

  ;; default name to use for MASM
  $pMasmName equ $pName

  ;; MASM adds it's own underscore, so remove the one already there, if C
  ifidni $pLang,<C>
    ifidni $pVis,<PUBLIC>
      sizName SIZESTR $pName
      $pMasmName SUBSTR $pName,2,sizName-1
    endif
  endif

  ;; generate the appropriate PROC statement, depending whether there
  ;; are any arguments or not
  ifb $pArgs
    ifb $pUses
      % $pMasmName PROC attribs
    else
      % $pMasmName PROC attribs USES $pUses
    endif
  else
    ifb $pUses
      % $pMasmName PROC attribs, $pArgs
    else
      % $pMasmName PROC attribs USES $pUses, $pArgs
    endif
  endif
 
  ;; set up any locals as needed
  ifnb $pLocals
    % LOCAL $pLocals
  endif
ENDM





popcontext      listing
.listmacro
