//
//   $Id: INI.C 1.2 2002/10/16 16:31:34Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// INI.C
// Updated @ Date 14 March 1996, 23:1 by Gerrit van Niekerk

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ini.h"
#include "commands.h"

#define MAX_LINE_LENGTH    160

//===========================================================================
// Function:     read_line()
// Arguments:    <FILE *> fp - a pointer to the file to be read from
//               <char *> bp - a pointer to the copy buffer
// Returns:      TRUE if successful FALSE otherwise
//===========================================================================
int read_line(FILE *fp, char *bp)
{
	int c = '\0';
	int i = 0;
	/* Read one line from the source file */
	while( (c = getc(fp)) != '\n' ) {
		if( c == EOF )         /* return FALSE on unexpected EOF */
			return(0);
		if (i < MAX_LINE_LENGTH - 2)
			bp[i++] = c;
	}
	bp[i] = '\0';
	repargs(bp,MAX_LINE_LENGTH);
	return(1);
}

//===========================================================================
// Function:     GetPrivateProfileInt()
// Arguments:    <char *> section - the name of the section to search for
//               <char *> entry - the name of the entry to find the value of
//               <int> def - the default value in the event of a failed read
//               <char *> file_name - the name of the .ini file to read from
// Returns:      the value located at entry
//===========================================================================
int GetPrivateProfileInt(char *section, char *entry, int def,
		char *file_name)
{
	char buff[10];
	char *ep;
	int value = 0;
	char sign = 0,hex = 0;
	int i;

	if (GetPrivateProfileString(section,entry,"",buff,
					sizeof(buff),file_name) == 0)
		return def;
	/* Copy only numbers fail on characters */

	for(i = 0,ep = buff;; i++ ) {
		char ch = ep[i] | ('a' - 'A');
		if (ch == 'x')
			hex = 1;
		else if (ch == '-')
			sign = 1;
		else if (hex && isxdigit(ch))
			value = value * 16 + (ch > '9' ? ch - 'a' + 10 : ch - '0');
		else if (isdigit(ch))
			value = value * 10 + ch - '0';
		else
			break;
	}
	return(sign ? -value : value);
}

FILE *fpProfile;

void OpenProfile(char *file_name)	// speed up by calling beforehand
{
	fpProfile = fopen(file_name,"r");
}

void CloseProfile()
{
	fclose(fpProfile);
	fpProfile = 0;
}

//===========================================================================
// Function:     GetPrivateProfileString()
// Arguments:    <char *> section - the name of the section to search for
//               <char *> entry - the name of the entry to find the value of
//               <char *> def - default string in the event of a failed read
//               <char *> buffer - a pointer to the buffer to copy into
//               <int> buffer_len - the max number of characters to copy
//               <char *> file_name - the name of the .ini file to read from
// Returns:      the number of characters copied into the supplied buffer
//===========================================================================
int GetPrivateProfileString(char *section, char *entry, char *def,
	char *buffer, int buffer_len, char *file_name)
{
	FILE *fp;
	char buff[MAX_LINE_LENGTH];
	char *ep;
	char t_section[MAX_LINE_LENGTH];

	int len = strlen(entry);
	if ((fp = fpProfile) == 0) {
		if ((fp = fopen(file_name,"r")) == 0) {
			strncpy(buffer,def,buffer_len);
			return(strlen(buffer));
		}
	}
	else
		rewind(fp);
	sprintf(t_section,"[%s]",section);    /* Format the section name */
	/*  Move through file 1 line at a time until a section is matched or EOF */
	do {
		if( !read_line(fp,buff) ) {
			if (!fpProfile)
				fclose(fp);
			strncpy(buffer,def,buffer_len);
			return(strlen(buffer));
		}
	} while( stricmp(buff,t_section) );
	/* Now that the section has been found, find the entry.
	 * Stop searching upon leaving the section's area. */
	if (!entry) {	// want list of entries
		int len = 0,clen;
		while (read_line(fp,buff) && buff[0] && buff[0] != '[') {
			if ((ep = strchr(buff,'=')) == 0 || ep == buff)
				continue;
			--ep;
			while (*ep == ' ' || *ep == '\t')
				--ep;
			++ep;
			*ep = 0;
			clen = strlen(buff);
			if (len + clen + 2 <= buffer_len) {
				strcpy(buffer + len,buff);
				len += (clen + 1);
			}
			else {
				strncpy(buffer + len,buff,buffer_len - len - 2);
				if (!fpProfile)
					fclose(fp);
				return buffer_len - 2;
			}
		}
		if (!fpProfile)
			fclose(fp);
		buffer[len + 1] = 0;
		return len;
	}
	do {
		if( !read_line(fp,buff) || buff[0] == '\0' || buff[0] == '[') {
			if (!fpProfile)
				fclose(fp);
			strncpy(buffer,def,buffer_len);
			return(strlen(buffer));
		}
	}  while(strnicmp(buff,entry,len) || !strchr(buff,'='));
	ep = strrchr(buff,'=');    /* Parse out the equal sign */
	ep++;
	while (*ep == ' ' || *ep == '\t')
		++ep;
	/* Copy up to buffer_len chars to buffer */
	strncpy(buffer,ep,buffer_len - 1);

	buffer[buffer_len - 1] = '\0';
	if (!fpProfile)
		fclose(fp);	/* Clean up and return the amount copied */
	return(strlen(buffer));
}

#ifdef WRITE_PROFILE
//===========================================================================
// Function:    WritePrivateProfileString()
// Arguments:   <char *> section - the name of the section to search for
//              <char *> entry - the name of the entry to find the value of
//              <char *> buffer - pointer to the buffer that holds the string
//              <char *> file_name - the name of the .ini file to read from
// Returns:     TRUE if successful, otherwise FALSE
//===========================================================================
int WritePrivateProfileString(char *section,
	char *entry, char *buffer, char *file_name)
{
	FILE *rfp, *wfp;
	char tmp_name[15];
	char buff[MAX_LINE_LENGTH];
	char t_section[MAX_LINE_LENGTH];
	int len = strlen(entry);
	tmpnam(tmp_name); /* Get a temporary file name to copy to */
	sprintf(t_section,"[%s]",section);/* Format the section name */
	if( !(rfp = fopen(file_name,"r")) )  /* If the .ini file doesn't exist */
	{
		if( !(wfp = fopen(file_name,"w")) ) /*  then make one */
		{
			return(0);
		}
		fprintf(wfp,"%s\n",t_section);
		fprintf(wfp,"%s=%s\n",entry,buffer);
		fclose(wfp);
		return(1);
	}
	if( !(wfp = fopen(tmp_name,"w")) )
	{
		fclose(rfp);
		return(0);
	}
	/* Move through the file one line at a time until a section is
	 * matched or until EOF. Copy to temp file as it is read. */
	do
	{
		if( !read_line(rfp,buff) )
		{   /* Failed to find section, so add one to the end */
			fprintf(wfp,"\n%s\n",t_section);
			fprintf(wfp,"%s=%s\n",entry,buffer);
			/* Clean up and rename */
			fclose(rfp);
			fclose(wfp);
			unlink(file_name);
			rename(tmp_name,file_name);
			return(1);
		}
		fprintf(wfp,"%s\n",buff);
	} while( stricmp(buff,t_section) );
	/* Now that the section has been found, find the entry. Stop searching
	 * upon leaving the section's area. Copy the file as it is read
	 * and create an entry if one is not found.  */
	while( 1 )
	{
		if( !read_line(rfp,buff) )
		{   /* EOF without an entry so make one */
			fprintf(wfp,"%s=%s\n",entry,buffer);
			/* Clean up and rename */
			fclose(rfp);
			fclose(wfp);
			unlink(file_name);
			rename(tmp_name,file_name);
			return(1);

		}
		if( !strnicmp(buff,entry,len) || buff[0] == '\0' )
			break;
		fprintf(wfp,"%s\n",buff);
	}
	if( buff[0] == '\0' )
	{
		fprintf(wfp,"%s=%s\n",entry,buffer);
		do
		{
			fprintf(wfp,"%s\n",buff);
		} while( read_line(rfp,buff) );
	}
	else
	{
		fprintf(wfp,"%s=%s\n",entry,buffer);
		while( read_line(rfp,buff) )
		{
			 fprintf(wfp,"%s\n",buff);
		}
	}
	/* Clean up and rename */
	fclose(wfp);
	fclose(rfp);
	unlink(file_name);
	rename(tmp_name,file_name);
	return(1);
}
#endif

/*
Example for use:

#ifdef WINDOWS
GetPrivateProfileString(ini_section,ini_entry, default_str,buffer,buffer_len,ini_file);
#else
	  get_private_profile_string(ini_section,ini_entry, default_str,buffer,buffer_len,ini_file);
#endif
*/
