//
//   $Id: Lfnsmart.c 1.3 2002/10/16 16:32:16Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:LFNSMART.C

	Copyright (c) 1998-2001 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	These smart functions work whether long filenames are present or
	not. With long filenames present, the functions use the LFN
	equivalents. With no long filename support, the older, DOS2 style
	functions are used.

	The calling app need only make sure that all path strings conform
	to the longname string limits "LFN_MAX_PATH".

	$DOCHISTORY
	02/16/1999 dg  Added findfirst/next and file functions
	01-30-1999 bt  Added dostruct.h include
	07/17/1998 dg  Created by Drew Gislason
*********************************************************************/
//#include <dir.h>		// used only for ffblk in SmartFindxxx() functions
#include <string.h>	// used for strcpy()
#include <ctype.h>   // used for toupper()

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlerrcod.h"

/*
	$DOCPROC:LFNPresent

	Returns TRUE if kernel supports long filenames. Returns FALSE if
	not.

	State Machine:  Only checks once and saves the result
*/
bool PASCAL LFNPresent(void)
{
	static uchar ucCallStatus = 0;

	char szRootName[] = "\\";
	char szFileSystemName[32];
	int	iFlags, iMaxName, iMaxPath;
	bool fSuccessful;

	if( ucCallStatus == 0 )
	{
		// return TRUE if long filename support is in the kernel
		fSuccessful = LFNGetVolumeInformation(sizeof(szFileSystemName),
				szRootName,	szFileSystemName, &iFlags, &iMaxName, &iMaxPath);
		if( fSuccessful )
		{
			ucCallStatus = 2;
		}
		else
		{
			ucCallStatus = 1;
		}
	}

	return (ucCallStatus == 2);
}



/*
	[]---------------------------------------------------------------------[]
    |                         Directory Functions                         |
	[]---------------------------------------------------------------------[]
*/

/*
	$DOCPROC:SmartGetCurrentDirectory

	Determine current diretory for this drive. Drive 0 means default, 1=A.
	If LFNs not installed, uses the old shortname function.

	Returns TRUE and changes into it if directory exists.
*/
bool PASCAL SmartGetCurrentDirectory(int nDrive, char szPathName[])
{
	bool	fSuccessful;
	int	iErrorCode = 0;


	if(LFNPresent() && DriveSupportsLFNs(nDrive))
	{
		fSuccessful = LFNGetCurrentDirectory(nDrive, szPathName);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlGetCurrentDirectory(nDrive, szPathName);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		//
		fSuccessful = dlGetCurrentDirectory(nDrive, szPathName);
	}

	return fSuccessful;
}

/*
	$DOCPROC:SmartChangeDirectory

	Change to the given directory. If LFNs not installed, uses the old
	shortname function.

	Returns TRUE and changes into it if directory exists.
*/
bool PASCAL SmartChangeDirectory(char * szPathName)
{
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szPathName[1] == ':' )
	{
		ucDrive = toupper(szPathName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNChangeDirectory(szPathName);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlSetCurrentDirectory(szPathName);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		//
		fSuccessful = dlSetCurrentDirectory(szPathName);
	}

	return fSuccessful;
}

/*
	$DOCPROC:SmartMakeDirectory

	Create a directory. If LFNs not installed, uses the old shortname
	function.

	Returns TRUE if directory was created.
*/
bool PASCAL SmartMakeDirectory(char szPathName[])
{
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szPathName[1] == ':' )
	{
		ucDrive = toupper(szPathName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNMakeDirectory(szPathName);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlMakeDirectory(szPathName);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		fSuccessful = dlMakeDirectory(szPathName);
	}

	return fSuccessful;
}

/*
	$DOCPROC:SmartRemoveDirectory

	Remove a directory. If LFNs not installed, uses the old shortname
	function.

	Returns TRUE if directory was made.
*/
bool PASCAL SmartRemoveDirectory(char szPathName[])
{
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szPathName[1] == ':' )
	{
		ucDrive = toupper(szPathName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNRemoveDirectory(szPathName);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlRemoveDirectory(szPathName);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		fSuccessful = dlRemoveDirectory(szPathName);
	}

	return fSuccessful;
}

/*
	[]---------------------------------------------------------------------[]
    |                           File Functions                            |
	[]---------------------------------------------------------------------[]
*/

/*
	$DOCPROC:SmartCreateOpenFile

	Open or create a file. This is DOS fn 6Ch. If LFNs not installed, uses
	old shortname function.

	Returns TRUE if file opened/created, plus handle and action.
*/
bool PASCAL SmartCreateOpenFile(int iModeFlags, int iAttr, int iAction,
					char szFileName[], int *pHandle, int *pActionTaken)
{
	unsigned uActionTaken;
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szFileName[1] == ':' )
	{
		ucDrive = toupper(szFileName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNCreateOpenFile(iModeFlags, iAttr, iAction,
					szFileName, 0, pHandle, pActionTaken);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				uActionTaken = iAction;
				fSuccessful = dlOpenOrCreateFile(iModeFlags, iAttr, &uActionTaken,
						szFileName, pHandle);
				if(fSuccessful)
				{
					// old-style function overloads action with action taken
					*pActionTaken = uActionTaken;
				}
			}
		}
	}
	else
	{
		uActionTaken = iAction;
		fSuccessful = dlOpenOrCreateFile(iModeFlags, iAttr, &uActionTaken,
				szFileName, pHandle);
		if(fSuccessful)
		{
			// old-style function overloads action with action taken
			*pActionTaken = uActionTaken;
		}
	}

	return fSuccessful;
}

/*
	[]---------------------------------------------------------------------[]
    |                           Find Functions                            |
	[]---------------------------------------------------------------------[]
*/

/*
	This structure is designed for backwards compatibility with existing
	code that uses the "C" ffblk structure.

	For short name systems, used exactly like ffblk. In long name systems,
	the SMARTFIND structure contains the LFN handle, and a signature to
	identify it.
*/
struct SmartFindFirstNext
{
	unsigned		uSignature;				// 00  will be 'SF'
	int			iHandle;					// 02  handle to long findfirst/next
	unsigned		uLastAccessDate;		// 04  last access date
   uchar      	res1[7];      			// 06  not used
   unsigned   	dirnum;      			// 0D  will be 0xFFFF for smart dir
   uchar      	res[6];      			// 0F  not used
   uchar      	attr;        			// 15  attribute found
   unsigned   	time;        			// 16  file's time
   unsigned   	date;        			// 18  file's date
   ulong      	size;        			// 1A  file's size
   char       	name[13];    			// 1E  filename in dot form
};
typedef struct SmartFindFirstNext SMARTFIND;
typedef SMARTFIND * PSMARTFIND;
typedef SMARTFIND far * LPSMARTFIND;

#define SMARTFIND_SIG	'FS'		// appears as SF in memory
#define IsSmartFind(pSF) (((PSMARTFIND)(pSF))->uSignature==SMARTFIND_SIG && \
	((PSMARTFIND)(pSF))->dirnum==0xFFFF)
#define SmartFindHandle(pSF) (((PSMARTFIND)(pSF))->iHandle)

// maximum # of handles open at once with SmartFindFirst(). Also limited
// by # available from DOS.
#define MAXHANDLES 20

static struct WIN32_FIND_DATA lfnFindData;
static int iNumberOfOpenSmartFindHandles;
static int iOpenHandles[MAXHANDLES];

// Internal use only! record this find handle.
static bool PASCAL RecordFindHandle(int iHandle)
{
	int	i;
	int * ipOpenHandles;

	// close any handles that are open (non-zero)
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		// is this handle entry available?
		if(*ipOpenHandles == 0)
		{
			// another handle open
			*ipOpenHandles = iHandle;
			++iNumberOfOpenSmartFindHandles;

			return TRUE;
		}

		// try next handle
		++ipOpenHandles;
	}

	// not enough handle entries!
	return FALSE;
}

/*
	Remove this handle from the list
*/
static void PASCAL RemoveRecordedFindHandle(int iHandle)
{
	int	i;
	int * ipOpenHandles;

	// close the handle that refers to this one
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		// is this handle entry available?
		if(*ipOpenHandles == iHandle)
		{
			*ipOpenHandles = 0;
			--iNumberOfOpenSmartFindHandles;
		}
		++ipOpenHandles;
	}
}

// convert from lfnFindData to SMARTFIND structure
void ConvertToFFBLK(PSMARTFIND pSmartFind)
{
	char	*pszName;

	pSmartFind->attr			= lfnFindData.attr;
	pSmartFind->time			= *(unsigned *)(&lfnFindData.lastWriteTime.dos.time);
	pSmartFind->date			= *(unsigned *)(&lfnFindData.lastWriteTime.dos.date);
	pSmartFind->uLastAccessDate = *(unsigned *)(&lfnFindData.lastAccessDate.dos.date);
	pSmartFind->size			= lfnFindData.fileSize;
	if(*lfnFindData.szAliasFileName != 0)
		pszName = lfnFindData.szAliasFileName;
	else
		pszName = lfnFindData.szFileName;
	strcpy(pSmartFind->name, pszName);
}

/*
	$DOCPROC:SmartFindFirst

	Find the first file or directory that matches the given attrs. If
	LFNs available, the mask can include LFN-style wildcards.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if any files/dirs found.
*/
bool PASCAL SmartFindFirst(int nAttrs, char szPathMask[],
					PFIND pFindData)
{
	int	iConvertCode;
	int	iHandle;
	PSMARTFIND pSmartFind = (PSMARTFIND)pFindData;
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szPathMask[1] == ':' )
	{
		ucDrive = toupper(szPathMask[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNFindFirstFile(nAttrs, szPathMask, DOSDATE,
				&lfnFindData, &iHandle, &iConvertCode);
		if( fSuccessful )
		{
			// debugging aid. used to verify that SmartFindClose() is called for
			// every open. See AreSmartFindAllClosed()
			if(!RecordFindHandle(iHandle))
			{
				LFNFindClose(iHandle);
				return FALSE;
			}

			// mark this ffblk as a long findfirst
			pSmartFind->uSignature	= SMARTFIND_SIG;
			pSmartFind->dirnum		= 0xFFFF;
			pSmartFind->iHandle		= iHandle;

			// found, convert to ffblk format
			ConvertToFFBLK(pSmartFind);
		}
		else
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				dlSetDTA(pFindData);
				fSuccessful = dlFindFirstFile(nAttrs, szPathMask);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		dlSetDTA(pFindData);
		fSuccessful = dlFindFirstFile(nAttrs, szPathMask);
	}

	return fSuccessful;
}

/*
	$DOCPROC:SmartFindNext

	Find the next occurence of a previously found file/dir. The find data
	must be exactly what was returned previously.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if another item found.
*/
bool PASCAL SmartFindNext(PFIND pFindData)
{
	PSMARTFIND pSmartFind = (PSMARTFIND)pFindData;
	int	iConvertCode;
	int	iHandle;
	bool	bWorked;

	// was it a longname findfirst?
	if(IsSmartFind(pFindData))
	{
		// do the findnext
		iHandle = SmartFindHandle(pFindData);
		bWorked = LFNFindNextFile(iHandle, DOSDATE, &lfnFindData, &iConvertCode);

		// found, convert to ffblk format
		if(bWorked)
			ConvertToFFBLK(pSmartFind);

		return bWorked;
	}

	// LFNs not present, use old style function
	if(!dlSetDTA(pFindData))
		return FALSE;
	return dlFindNextFile();
}

/*
	$DOCPROC:SmartFindClose

	Close a find opened by SmartFindFirst. Required, or RD will quickly
	run out of find handles.

	Warning! the "ffblk" is overloaded in the case of LFNs. See the
	SMARTFIND structure for details.

	Returns TRUE if close worked.
*/
bool PASCAL SmartFindClose(PFIND pFindData)
{
	int	iHandle;

	if(IsSmartFind(pFindData))
	{
		iHandle = SmartFindHandle(pFindData);

		if(!LFNFindClose(iHandle))
			return FALSE;

		// remove that handle from our list
		RemoveRecordedFindHandle(iHandle);
	}
	return TRUE;
}

/*
	Debugging aid! verify that the number of open files handles is 0.
*/
bool PASCAL SmartFindAreAllClosed(void)
{
	return (iNumberOfOpenSmartFindHandles == 0);
}

/*
	Close all handles allocated from SmartFindFirst().
*/
void PASCAL SmartFindCloseAll(void)
{
	int	i;
	int	*ipOpenHandles;

	// close any handles that are open (non-zero)
	ipOpenHandles = iOpenHandles;
	for(i=0; i<MAXHANDLES; ++i)
	{
		if(*ipOpenHandles)
		{
			LFNFindClose(*ipOpenHandles);
			*ipOpenHandles = 0;
		}
		++ipOpenHandles;
	}

	// all handles closed
	iNumberOfOpenSmartFindHandles = 0;
}

/*
	Get the last access date for this found file.
	Returns 0 if not an LFN entry or DOS kernel.
*/
unsigned PASCAL SmartGetLastAccessDate(struct ffblk far * lpFindData)
{
	LPSMARTFIND lpSmartFind = (LPSMARTFIND)lpFindData;

	if(IsSmartFind(lpSmartFind))
		return lpSmartFind->uLastAccessDate;
	return 0;
}

/*
	$DOCPROC:SmartGetFileAttributes

	Get file attributes. Uses longname function if possible.
*/
bool PASCAL SmartGetFileAttributes(char szName[], unsigned *pAttributes)
{
	FileTime time;
	FileDate date;
	int		iMilli;
	ulong		ulFileSize;
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szName[1] == ':' )
	{
		ucDrive = toupper(szName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNExtendedGetSetAttr(EXT_ACT_GET_ATTR, &time,
				pAttributes, szName, &date, &iMilli, &ulFileSize);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlGetFileAttributes(szName, pAttributes);
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		fSuccessful = dlGetFileAttributes(szName, pAttributes);
	}

	return fSuccessful;
}

/*
	$DOCPROC:SmartSetFileAttributes

	Set file attributes. Uses longname function if possible.
*/
bool PASCAL SmartSetFileAttributes(char szName[], unsigned uAttributes)
{
	FileTime time;
	FileDate date;
	int		iMilli;
	ulong		ulFileSize;
	bool	fSuccessful;
	int	iErrorCode = 0;
	uchar ucDrive;


	if( szName[1] == ':' )
	{
		ucDrive = toupper(szName[0]) - 'A' + 1;
	}
	else
	{
		ucDrive = 0;
	}

	if(LFNPresent() && DriveSupportsLFNs(ucDrive))
	{
		fSuccessful = LFNExtendedGetSetAttr(EXT_ACT_SET_ATTR, &time,
				&uAttributes, szName, &date, &iMilli, &ulFileSize);
		if( !fSuccessful )
		{
			iErrorCode = dlCheckDOSError();
			if( iErrorCode == ERR_INVALID_FUNCTION )
			{
				// Network drives may not support LFNs
				//
				fSuccessful = dlSetFileAttributes( uAttributes, szName );
			}
		}
	}
	else
	{
		// LFNs not present, use old style function
		fSuccessful = dlSetFileAttributes( uAttributes, szName );
	}

	return fSuccessful;
}

#if UNIT_TEST

void TestFindFirst(void)
{
	FIND ff;
	unsigned uDate;

	SmartFindFirst(0, "*.*", &ff);
	if(SmartFindAreAllClosed())
		printf("error. should be 1 open!\n");

	uDate = SmartGetLastAccessDate(&ff);

	SmartFindNext(&ff);

	printf("last access date %x\n",uDate);

	SmartFindCloseAll();
	if(!SmartFindAreAllClosed())
		printf("error. should be all closed!\n");
}

void TestWildcardDelete(void)
{
	if(!SmartWildcardDelete("*Bugga*"))
		printf("Cannot delete *Bugga*");
	else
		printf("deleted *Bugga*");
}

void TestGetAttr(void)
{
	static char szName[] = "This is a longname.txt";
	unsigned uAttributes;

	if(!SmartGetFileAttributes(szName, &uAttributes))
		printf("Get failed on %s\n",szName);
	else
		printf("attr %X\n",uAttributes);
}

void main(void)
{
//	TestFindFirst();

//	TestWildcardDelete();

	TestGetAttr();
}
#endif
