//
//   $Id: MAKEMAIL.C 1.8 2002/10/16 16:32:16Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// MAKEMAIL.C
// Copyright (c) 2001 Datalight, Inc.
// Portions copyright (c) GPvNO 2001
// All Rights Reserved
// Datalight Confidential and Proprietary
//
// Mail message formatter.
//
// Revision History
// 10/10/2002 pb  Fixed an error in the script file processing.
// 08/06/2001 pb  Fixed a problem in the command line processing.
// 07/26/2001 pb  Added Scripting support.
//                Clarified some of the error messages.
// 07/13/2001 pb  added checks for multiple occurences on the command line
//                for -b, -f, -s, and -o arguments.
// 05/09/2001 jmb fixed usage - was broken when updated to use ROMDOS
//                version info.
// 04/30/2001 jmb used ROMDOS version number and copyright info.
// 04/30/2001 jmb created header
//
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <dir.h>
#include <alloc.h>
#include <stdarg.h>
#include <process.h>
#include <time.h>
#include <io.h>
#include <sys\stat.h>
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"

#ifdef DL
#include "dosver.h"
#include "fatworld.h"
#endif

// prototypes
void check_dontwipe(void);

// error codes returned
#define PARAMETER_ERR	101
#define NO_API				102
#define ABORTED			103
#define XFER_ABORTED		104
#define CANT_WRITE		105
#define CANT_READ			106
#define FORMAT_ERR		107
#define CANT_OPEN       108

#define PROGRAM "MAKEMAIL"
#ifndef DL
#define INET "Inet"
#else
#define INET "Sockets"
#endif

static char szBuf[200] =
INET" "PROGRAM"%s version 1.00(dt)\n\0";
static char szUsage[] ="Use:\n"
PROGRAM" -nScriptFile -tToAddress [-tCCAddress1 [-tCCAddress2] ...]\n"
"-fFromAddress -sSubject -bBodyTextFile [-aAttachment1 [-aAttachment2]]\n"
"-oOutputFile\n"
;
#define BUFSZ	sizeof(szBuf)			// command buffer size
#define MAXTO	40

int iNumTo = 0;
char *ppszTo[MAXTO],*pszFrom,*pszSubject = "",*pszBody = "";
static char *pszOutFile;
FILE *fp,*fpIn;
char cVerbose;
char szMime[50];

typedef struct attachlist
{
	char szFile[255];
	struct attachlist * pNextAttachment;
};

char * szScriptFileBuffer;


struct attachlist * pFirstAttachment;
struct attachlist * pCurrentAttachment;

int AttachFiles( FILE* fp, char* szFile );
bool DoCmdLineParsing( int iArgumentCount, char * pszArguments[] );

main(int argc,char *argv[])
{
	char *inet_ntoa();
	int i;
	struct tm *time_now;
	time_t secs_now;
	static char sz[80];
	char szAtchTemp[255];
	bool fSuccessful = FALSE;

#ifdef DL
					printf("Datalight Sockets %s ", PROGRAM);
					if (dlIsFat32World())
						printf("%s ", SHORTVERSION);
					else
						printf("%s ", SHORTVERSION622);
					printf("%s\n%s\n%s\n\n",
						BUILDVERSION, COPYRIGHT, _BUILD_NUMBER);
#endif

	randomize();
	pszOutFile = "";

	pFirstAttachment = 0;
	pCurrentAttachment = 0;

	// Call the command line parsing function with argv and argc since
	// there is no script file.
	//
	fSuccessful = DoCmdLineParsing( argc - 1, &argv[1] );
	if( !fSuccessful )
	{
		exit(PARAMETER_ERR);
	}
	args_done:
#ifndef DL
	check_dontwipe();
#endif
	if (cVerbose)
		printf(szBuf,"");
	if (!pszFrom)
	{
		if (cVerbose)
			printf("Please supply From address\n");
		exit(PARAMETER_ERR);
	}
	if (!iNumTo)
	{
		if (cVerbose)
			printf("Please specify at least one recipient\n");
		exit(PARAMETER_ERR);
	}
	if (iNumTo > MAXTO)
	{
		if (cVerbose)
			printf("Too many recipients (%d maximum)\n",MAXTO);
		exit(PARAMETER_ERR);
	}
	if (!pszOutFile)
	{
		if (cVerbose)
			printf("Please supply output file\n");
		exit(PARAMETER_ERR);
	}
	if ((fp = fopen(pszOutFile,"wt")) == 0)
	{
		if (cVerbose)
			printf("Open error on '%s'\n",pszOutFile);
		exit(CANT_WRITE);
	}
	fprintf(fp,"MAIL FROM: <%s>\n",pszFrom);
	for (i = 0;i < iNumTo;++i)
		fprintf(fp,"RCPT TO: <%s>\n", ppszTo[i] );
	fprintf(fp,"DATA\n");

	fprintf(fp,"From: %s\n",pszFrom);
	fprintf(fp,"To: %s", ppszTo[0]);
	for (i = 1;i < iNumTo;++i)
		fprintf(fp,", %s", ppszTo[i] );
	fprintf(fp, "\n" );

	// get current time
	tzset();
	time(&secs_now);
	time_now = localtime(&secs_now);

	//	Date: Sat, 9 Oct 1999 22:13:16 +0200
	strftime(sz,80,"Date: %a, %d %b %Y %H:%M:%S",time_now);
	fprintf(fp,"%s %+05ld\n",sz,- _timezone / 36);

	// MIME boundary
	sprintf( szMime, "------------%04X%04X%04X%04X%04X%04X",
		random(0xFFFF), random(0xFFFF), random(0xFFFF), random(0xFFFF),
		random(0xFFFF), random(0xFFFF) );

	fprintf(fp,	"MIME-Version: 1.0\n"
					"Content-Type: multipart/mixed;\n"
					" boundary=\"%s\"\n", szMime );
	fprintf(fp,	"Subject: %s\n"
					"Priority: normal\n"
					"X-Accept-Language: en\n"
					"X-mailer: ",pszSubject);
	fprintf(fp,szBuf,"/SENDMAIL");

	fprintf(fp,	"\nThis is a multi-part message in MIME format.\n" );

	if (pszBody)
	{
		*szBuf = '.';
		fprintf(fp,	"--%s\n"
						"Content-Type: text/plain; charset=us-ascii\n"
						"Content-Transfer-Encoding: 7bit\n\n", szMime );

		if ((fpIn = fopen(pszBody,"rt")) != 0)
		{
			while (fgets(szBuf + 1,BUFSZ - 1,fpIn) != 0)
			{
				char *p = szBuf + (szBuf[1] == '.' ? 0 : 1);
				fprintf(fp,"%s",p);
			}
			fclose(fpIn);
		}
		else
		{
			printf("WARNING: Body Text File (%s) could not be found.\n",
								pszBody);
		}
	}

	// Attaching Files
	//
	pCurrentAttachment = pFirstAttachment;
	while( pCurrentAttachment )
	{
 		SmartExpandPath( _FULLPATH_SHORTNAME, pCurrentAttachment->szFile,
				szAtchTemp );
		AttachFiles( fp, szAtchTemp );
		pCurrentAttachment = pCurrentAttachment->pNextAttachment;
	}

	// Freeing Files
	//
	pCurrentAttachment = pFirstAttachment->pNextAttachment;
	while( pFirstAttachment != 0 )
	{
		free( pFirstAttachment );
		pFirstAttachment = pCurrentAttachment;
		pCurrentAttachment = pFirstAttachment->pNextAttachment;
	}

	// add the MIME termination
	fprintf(fp, "--%s--\n", szMime );

	// add a blank line with a '.' to terminate a sendmail message.
	fprintf(fp,"\n.\n");
	fclose(fp);

	// free the script buffer
	free( szScriptFileBuffer );



	return 0;
}

#define MAX_CHUNKS	15
int AttachFile( FILE* fpOutput, char *szFile )
{
	FILE*	fpInput;
	char	chInput[3];
	char	chOutput[5] = {0,0,0,0,0};
	char	chLine[(MAX_CHUNKS*4)+1];
	int	iRead, iWrite, iChunks;
	int	i;

	fpInput = fopen( szFile, "rb" );
	if( !fpInput )
		return 0;

	// output MIME header
	fprintf( fpOutput,	"--%s\n"
								"Content-Type: application/octet-stream;\n"
								" name=\"%s\"\n"
								"Content-Transfer-Encoding: x-uuencode\n"
								"Content-Disposition: attachment;\n"
								" filename=\"%s\"\n\n",
								szMime, szFile, szFile );


	// output UU header
	fprintf( fpOutput, "begin 644 %s\n", szFile );

	while( !feof( fpInput ))
	{
		// clear output buffers
		memset( chOutput, 0, sizeof(chOutput) );
		memset( chLine, 0, sizeof(chLine) );
		chLine[0] = ' ';
		iWrite = 0;

		for( iChunks=0; iChunks < MAX_CHUNKS; iChunks++ )
		{
			// read three bytes in
			memset( chInput, 0, 3 );
			iRead = fread( chInput, 1, 3, fpInput );
			iWrite += iRead;

			// prepare four bytes output
			chOutput[0] = 0x20 + ((chInput[0]>>2) & 0x3f );
			chOutput[1] = 0x20 + (((chInput[0]<<4) | ((chInput[1]>>4)&0xf)) & 0x3f);
			chOutput[2] = 0x20 + (((chInput[1]<<2) | ((chInput[2]>>6)&0x3)) & 0x3f);
			chOutput[3] = 0x20 + ((chInput[2]) & 0x3f);

			// append this chunk to the output line
			if( iRead )
				strncat( chLine, chOutput, iRead+1 );

			if( iRead < 3 )
				break;
		}

		// prepend the string length
		chLine[0] = 0x20 + iWrite;
		for( i=0; i < strlen(chLine); i++ )
		{
			if( chLine[i] == ' ' )
				fputc( '`', fpOutput );
			else
				fputc( chLine[i], fpOutput );
		}
		fprintf( fpOutput, "\n" );
	}

	// terminate the block
	fprintf( fpOutput, "`\nend\n" );

	fclose( fpInput );

	return 1;
}

int AttachFiles( FILE* fpOutput, char* szFile )
{
	struct ffblk	ff;
	unsigned			uResult;
	int				iAttached = 0;

	uResult = findfirst( szFile, &ff, 1 );
	while( uResult != 0xFFFF )
	{
		if( AttachFile( fpOutput, ff.ff_name ) )
		{
			iAttached++;
		}

		uResult = findnext( &ff );
	}

	return iAttached;
}

bool DoCmdLineParsing( int iArgumentCount, char * pszArguments[] )
{
	char szOutTemp[255];
	char szInTemp[255];
	char szScriptFileTemp[255];
	char * pszArg;
	char * pszArgument;
	int iargHandle;
	int iargAction;
	int iCount;
 	bool fFoundBodyFileSwitch = FALSE;
	bool fFoundFromSwitch = FALSE;
	bool fFoundSubjectSwitch = FALSE;
	bool fFoundOutFileSwitch = FALSE;
	bool fFoundScriptFileSwitch = FALSE;
	FILE * fpScriptFile;
	long lSizeOfFile;
	struct stat sFileStats;
	int iFileHandle;
	int iErr;
	unsigned uArgSize;
	unsigned uCount;

	uCount = 0;

	// count through the arguments
	pszArgument = pszArguments[0];

	while( pszArgument[0] != 0x00 )
	{
		// Are we done with the arguments?
		//
        if( uCount < iArgumentCount || fFoundScriptFileSwitch)
		{
			while( (pszArgument[0] == 0x09) || (pszArgument[0] == 0x20) ||
					(pszArgument[0] == 0x0d) || (pszArgument[0] == 0x0a) )
			{
				pszArgument++;
			}

			if (pszArgument[0] == '/' || pszArgument[0] == '-')
			{
				uCount++;

				if( pszArgument[2] == '=' )
				{
					// Move the string down by one element to get rid of
					// the '='. We will move the length of pszArgument[3] + 1
					// so we get the termination character also.
					//
					memmove( &pszArgument[2], &pszArgument[3],
									strlen(&pszArgument[3]) +1 );
				}


				// Check for whitespace in the argument.
				//
				if( pszArgument[2] == 0x00 )
				{
					pszArg = LFNSkipWhite(&pszArgument[0] + 3);
				}
				else
				{
					pszArg = LFNSkipWhite(&pszArgument[0] + 2);
				}

				if(*pszArg == '"' || *pszArg == '\'')
				{
					char cEndQuote = *pszArg;

					uArgSize = 3;
					// find end of quoted argument
					while(pszArgument[uArgSize] != cEndQuote &&
							pszArgument[uArgSize] != 0x0d      &&
							pszArgument[uArgSize] != 0x0A      &&
							pszArgument[uArgSize] != 0x00      )
					{
						uArgSize++;
					}

					// keep the ending quote if it exists
					if( pszArgument[uArgSize] == cEndQuote )
					{
						uArgSize++;
					}

				}
				else
				{
					// Determine the size of the argument so that we can
					// increment pszArgument by the correct amount after each
					// run through the switch statement.
					//
					for( uArgSize = 2; (pszArgument[uArgSize] != 0x20)
							&& (pszArgument[uArgSize] != 0x09)
							&& (pszArgument[uArgSize] != 0x0d)
							&& (pszArgument[uArgSize] != 0x0a)
							&& (pszArgument[uArgSize] != 0x00); uArgSize++ );
				}

				// null terminate command line argument
				pszArgument[uArgSize] = 0x00;

				// remove quotes if necessary...
				LFNStripArgQuotes(pszArgument);


				switch (toupper(pszArgument[1]))
				{
					case 'N':      // Script file name
						if( !fFoundScriptFileSwitch )
						{
							fFoundScriptFileSwitch = TRUE;

							// If the file name given is LFN then change it to
							// short name so that we can process it.
							//

			  				// change to short name if lfn.
			  				SmartExpandPath( _FULLPATH_SHORTNAME, pszArg,
									szScriptFileTemp );
			  				pszArg = szScriptFileTemp;

							// Open the file and error out if the open failed.
							//
							fpScriptFile = fopen( pszArg, "rb" );
							if( !fpScriptFile )
							{
								printf("Error:\n");
								printf("  Unable to open %s.\n", pszArg);
								exit(PARAMETER_ERR);
							}

							// Determine the size of the file and assign it to
							// a variable that will be used to allocate a buffer.
							//
							iFileHandle = fileno( fpScriptFile );
							fstat( iFileHandle, &sFileStats );
							lSizeOfFile = sFileStats.st_size;


							// Allocate a buffer the size of the file to
							// read the file into.
							//
							szScriptFileBuffer = (char *) malloc( lSizeOfFile + 1 );


							// Read the file into the buffer.
							//
							iErr = fread( szScriptFileBuffer, lSizeOfFile, 1,
										fpScriptFile );
							if( iErr != 1 )
							{
								fclose( fpScriptFile );
								free( szScriptFileBuffer );
								printf("Error:\n");
								printf("  Failed to read script file.\n");
								exit(255);
							}


							fclose( fpScriptFile );


							// We have read the file into the buffer now we must
							// process the arguments. We must also take into account
							// any arguments on the command line that might be
							// following the argument that got us into this switch
							// case.
							//
							pszArgument = szScriptFileBuffer;
							pszArgument[lSizeOfFile] = 0x00;
							pszArgument[lSizeOfFile+1] = 0x00;
							uArgSize = 0;

							break;
						}
						else
						{
							printf("Error:\n");
							printf("  Only one script file may be specified.\n");
							exit(PARAMETER_ERR);
							break;
						}
					case 'B':		// body text file name
						if( !fFoundBodyFileSwitch )
						{
							fFoundBodyFileSwitch = TRUE;
			  				// change to short name if lfn.
			  				SmartExpandPath( _FULLPATH_SHORTNAME, pszArg, szInTemp );
			  				pszArg = szInTemp;
			  				pszBody = pszArg;

			 				// Increment the pointer by the size of the
						 	// argument, including the null.
						 	//
						 	pszArgument += uArgSize+1;

							break;
						}
						else
						{
							printf("Error:\n");
							printf("  Only one body text file may be specified.\n");
							exit(PARAMETER_ERR);
							break;
						}
					case 'F':
						if( !fFoundFromSwitch )
						{
							fFoundFromSwitch = TRUE;
				  			pszFrom = pszArg;
			  				// Increment the pointer by the size of the
						 	// argument, including the null.
						 	//
						 	pszArgument += uArgSize+1;

							break;
						}
						else
						{
							printf("Error:\n");
							printf("  Multiple From addresses are not allowed.\n");
							exit(PARAMETER_ERR);
							break;
						}
					case 'S':
						if( !fFoundSubjectSwitch )
						{
							fFoundSubjectSwitch = TRUE;
				  			pszSubject = pszArg;
			 				// Increment the pointer by the size of the
						 	// argument, including the null.
						 	//
						 	pszArgument += uArgSize+1;

							break;
						}
						else
						{
							printf("Error:\n");
							printf("  Only one subject allowed.\n");
							exit(PARAMETER_ERR);
							break;
						}
					case 'T':
				  		if (iNumTo < MAXTO)
				  			ppszTo[iNumTo] = pszArg;
				  		iNumTo++;
						// Increment the pointer by the size of the
					 	// argument, including the null.
					 	//
					 	pszArgument += uArgSize+1;

						break;
					case 'O':
						if( !fFoundOutFileSwitch )
						{
							fFoundOutFileSwitch = TRUE;
			  				// change to short name if lfn.
			  				if (!SmartExpandPath( _FULLPATH_SHORTNAME, pszArg, szOutTemp ))
			  				{
			  					// file does not exist, create it...
			  					SmartCreateOpenFile(OPEN_ACCESS_READWRITE, A_NORMAL,
			  												FILE_CREATE, pszArg, &iargHandle,
			  												&iargAction);
			  					dlCloseFile(iargHandle);

								if (!SmartExpandPath( _FULLPATH_SHORTNAME, pszArg, szOutTemp ))
				  				{
				  					printf("Unable to open or create output file \"%s\".\n",
				  							pszArg);
				  					return 2;
				  				}
			  				}

			   			pszArg = szOutTemp;
			  				pszOutFile = pszArg;

			 				// Increment the pointer by the size of the
						 	// argument, including the null.
						 	//
						 	pszArgument += uArgSize+1;
							break;
						}
						else
						{
							printf("Error:\n");
							printf("  Multiple output files are not allowed.\n");
							exit(PARAMETER_ERR);
							break;
						}
					case 'V':
				  		cVerbose = 1;
			 			// Increment the pointer by the size of the
			 		 	// argument, including the null.
			 		 	//
			 		 	pszArgument += uArgSize+1;

						break;
					case 'A':
						// We will process the attachments later...
						// right now lets just get the file names out of the file
						// or off of the command line.
						//
						if( !pFirstAttachment )
						{
							pFirstAttachment =
								(struct attachlist *) malloc( sizeof(struct attachlist) );

							pCurrentAttachment = pFirstAttachment;
						}
						else
						{
							struct attachlist * pNewAttachment;
							pNewAttachment =
								(struct attachlist *) malloc( sizeof(struct attachlist) );

							pCurrentAttachment->pNextAttachment = pNewAttachment;
							pCurrentAttachment = pNewAttachment;
						}
						if(!pCurrentAttachment)
						{
							printf("MALLOC failed!  Cannot allocate node!\n");
							exit(2);
						}

						strcpy(pCurrentAttachment->szFile, pszArg);

						pCurrentAttachment->pNextAttachment = 0;

			 			// Increment the pointer by the size of the
			 		 	// argument, including the null.
			 		 	//
			 		 	pszArgument += uArgSize+1;

						break;
					default:

#ifndef DL
						printf(szBuf, "");
						printf(szBuf + strlen(szBuf) + 1);
#endif
						printf("%s", szUsage);
						return 0;
				}
			}
			// check for blank line
			else if( pszArgument[0] != 0x00)
			{
				printf("%s\n", pszArgument);
				printf("Error:\n");
				printf("  Invalid arguments specified.\n");
				exit(PARAMETER_ERR);
			}
		}
		else
		{
			// We are done with the arguments....
			//
			break;
		}
	}
	return TRUE;
}
