/*---------------------------------------------------------------------------
 * filename - GetKey.c
 *
 * Get a scancode and key in extended BIOS format
 *--------------------------------------------------------------------------*/

#define INCL_CON
#define INCL_USER
#include <ntbc.h>

#include <conio.h>
#include <stdio.h>      /* EOF definition */

extern HANDLE _hin;  /* From CRTINIT.C for console input */

/*----------------------------------------------------------------------
 * The following table maps NT virtual keycodes to PC BIOS keyboard
 * values.  For each virtual keycode there are four possible BIOS
 * values: normal, shift, Ctrl, and ALT.  Key combinations that have
 * no BIOS equivalent have a value of -1, and are ignored.
 */

#define EXT(key)    ((key) << 8)
//#define ISEXT(val)  ((val)&0x100)
//#define EXTVAL(val) ((val)&0xff)
#define VK_ENTER	0x100		// Numpad enter key

struct kbd
{
	 short keycode;            /* virtual keycode */
	 short normal;             /* BIOS keycode - normal */
	 short shift;              /* BIOS keycode - Shift- */
	 short ctrl;               /* BIOS keycode - Ctrl- */
	 short alt;                /* BIOS keycode - Alt- */
} kbdtab [] =
{
/* Virtual key		Normal	Shift		Control  Alt */

	{ VK_BACK,		0x0e08,  0x0e08,  0x0e7f,  0x0e00 },
	{ VK_TAB,      0x0f09,  0x0f00,  0x9400,  0x0f09 },
	{ VK_CLEAR,   	0x4c00,  0x4c35,  0x8f00,  -1     }, // Numpad 5
	{ VK_RETURN,   0x1c0d,  0x1c0d,  0x1c0a,  0x1c00 },
	{ VK_ESCAPE,   0x011b,  0x011b,  0x011b,  0x0100 },
	{ VK_SPACE,    0x3920,  0x3920,  0x3920,  0x2000 },
	{ '0',         0x0b30,  0x0b29,  -1,      0x8100 },
	{ '1',         0x0231,  0x0221,  -1,      0x7800 },
	{ '2',         0x0332,  0x0340,  0x0300,  0x7900 },
	{ '3',         0x0433,  0x0423,  -1,      0x7a00 },
	{ '4',         0x0534,  0x0524,  -1,      0x7b00 },
	{ '5',         0x0635,  0x0625,  -1,      0x7c00 },
	{ '6',         0x0736,  0x075e,  0x071e,  0x7d00 },
	{ '7',         0x0837,  0x0826,  -1,      0x7e00 },
	{ '8',         0x0938,  0x092a,  -1,      0x7f00 },
	{ '9',         0x0a39,  0x0a28,  -1,      0x8000 },

	{ 'A',         0x1eff,  0x1eff,  0x1eff,  0x1e00 },
	{ 'B',         0x30ff,  0x30ff,  0x30ff,  0x3000 },
	{ 'C',         0x2eff,  0x2eff,  0x2eff,  0x2e00 },
	{ 'D',         0x20ff,  0x20ff,  0x20ff,  0x2000 },
	{ 'E',         0x12ff,  0x12ff,  0x12ff,  0x1200 },
	{ 'F',         0x21ff,  0x21ff,  0x21ff,  0x2100 },
	{ 'G',         0x22ff,  0x22ff,  0x22ff,  0x2200 },
	{ 'H',         0x23ff,  0x23ff,  0x23ff,  0x2300 },
	{ 'I',         0x17ff,  0x17ff,  0x17ff,  0x1700 },
	{ 'J',         0x24ff,  0x24ff,  0x24ff,  0x2400 },
	{ 'K',         0x25ff,  0x25ff,  0x25ff,  0x2500 },
	{ 'L',         0x26ff,  0x26ff,  0x26ff,  0x2600 },
	{ 'M',         0x32ff,  0x32ff,  0x32ff,  0x3200 },
	{ 'N',         0x31ff,  0x31ff,  0x31ff,  0x3100 },
	{ 'O',         0x18ff,  0x18ff,  0x18ff,  0x1800 },
	{ 'P',         0x19ff,  0x19ff,  0x19ff,  0x1900 },
	{ 'Q',         0x10ff,  0x10ff,  0x10ff,  0x1000 },
	{ 'R',         0x13ff,  0x13ff,  0x13ff,  0x1300 },
	{ 'S',         0x1fff,  0x1fff,  0x1fff,  0x1f00 },
	{ 'T',         0x14ff,  0x14ff,  0x14ff,  0x1400 },
	{ 'U',         0x16ff,  0x16ff,  0x16ff,  0x1600 },
	{ 'V',         0x2fff,  0x2fff,  0x2fff,  0x2f00 },
	{ 'W',         0x11ff,  0x11ff,  0x11ff,  0x1100 },
	{ 'X',         0x2dff,  0x2dff,  0x2dff,  0x2d00 },
	{ 'Y',         0x15ff,  0x15ff,  0x15ff,  0x1500 },
	{ 'Z',         0x2cff,  0x2cff,  0x2cff,  0x2c00 },

	{ VK_PRIOR,    EXT(73), EXT(73), EXT(132),EXT(153) },
	{ VK_NEXT,     EXT(81), EXT(81), EXT(118),EXT(161) },
	{ VK_END,      EXT(79), EXT(79), EXT(117),EXT(159) },
	{ VK_HOME,     EXT(71), EXT(71), EXT(119),EXT(151) },
	{ VK_LEFT,     EXT(75), EXT(75), EXT(115),EXT(155) },
	{ VK_UP,       EXT(72), EXT(72), EXT(141),EXT(152) },
	{ VK_RIGHT,    EXT(77), EXT(77), EXT(116),EXT(157) },
	{ VK_DOWN,     EXT(80), EXT(80), EXT(145),EXT(160) },
	{ VK_INSERT,   EXT(82), EXT(82), EXT(146),EXT(162) },
	{ VK_DELETE,   EXT(83), EXT(83), EXT(147),EXT(163) },
	{ VK_NUMPAD0,  0x5230,  EXT(82), EXT(146),-1       },
	{ VK_NUMPAD1,  0x4f31,  EXT(79), EXT(117),-1       },
	{ VK_NUMPAD2,  0x5032,  EXT(80), EXT(145),-1       },
	{ VK_NUMPAD3,  0x5133,  EXT(81), EXT(118),-1       },
	{ VK_NUMPAD4,  0x4b34,  EXT(75), EXT(115),-1       },
	{ VK_NUMPAD5,  0x4c35,  EXT(76), EXT(143),-1       },
	{ VK_NUMPAD6,  0x4d36,  EXT(77), EXT(116),-1       },
	{ VK_NUMPAD7,  0x4737,  EXT(71), EXT(119),-1       },
	{ VK_NUMPAD8,  0x4838,  EXT(72), EXT(141),-1       },
	{ VK_NUMPAD9,  0x4939,  EXT(73), EXT(132),-1       },
	{ VK_MULTIPLY,	0x372a,  0x372a,  EXT(150),EXT(55)  },
	{ VK_ADD,      0x4e2b,  0x4e2b,  EXT(144),EXT(78)  },
	{ VK_SUBTRACT, 0x4a2d,  0x4a2d,  EXT(142),EXT(74)  },
	{ VK_DECIMAL,  0x532e,  0x532e,  EXT(83), -1			},
	{ VK_DIVIDE,   0xe02f,  0xe02f,  0x9500,	0xa400   },
	{ VK_F1,       EXT(59), EXT(84), EXT(94), EXT(104) },
	{ VK_F2,       EXT(60), EXT(85), EXT(95), EXT(105) },
	{ VK_F3,       EXT(61), EXT(86), EXT(96), EXT(106) },
	{ VK_F4,       EXT(62), EXT(87), EXT(97), EXT(107) },
	{ VK_F5,       EXT(63), EXT(88), EXT(98), EXT(108) },
	{ VK_F6,       EXT(64), EXT(89), EXT(99), EXT(109) },
	{ VK_F7,       EXT(65), EXT(90), EXT(100),EXT(110) },
	{ VK_F8,       EXT(66), EXT(91), EXT(101),EXT(111) },
	{ VK_F9,       EXT(67), EXT(92), EXT(102),EXT(112) },
	{ VK_F10,      EXT(68), EXT(93), EXT(103),EXT(113) },
	{ VK_F11,      EXT(133),EXT(135),EXT(137),EXT(139) },
	{ VK_F12,      EXT(134),EXT(136),EXT(138),EXT(140) },
	{ 0xdc,        0x2b5c,  0x2b7c,  0x2b1c,  EXT(43)  },	// \ |
	{ 0xbf,        0x352f,  0x353f,  -1,      EXT(53)  },  // / ?
	{ 0xbd,        0x0c2d,  0x0c5f,  0x0c1f,  EXT(130) },  // - _
	{ 0xbb,        0x0d3d,  0x0d2b,  -1,      EXT(131) },  // = +
	{ 0xdb,        0x1a5b,  0x1a7b,  0x1a1b,  EXT(26)  },  // [ {
	{ 0xdd,        0x1b5d,  0x1b7d,  0x1b1d,  EXT(27)  },  // ] }
	{ 0xba,        0x273b,  0x273a,  -1,      EXT(39)  }, // ; :
	{ 0xde,        0x2827,  0x2822,  -1,      EXT(40)  }, // ' "
	{ 0xbc,        0x332c,  0x333c,  -1,      EXT(51)  },	// , <
	{ 0xbe,        0x342e,  0x343e,  -1,      EXT(52)  },	// . >
	{ 0xc0,        0x2960,  0x297e,  -1,      EXT(41)  },	// ` ~

	{ VK_ENTER,		0xe00d,	0xe00d,	0xe00a,	0xa600	},
	{ -1,          -1,      -1,      -1,      -1       }  /** END **/
};

int GetKey(char *pcControl)
{
	 INPUT_RECORD inp;
	 DWORD nread;
	 DWORD kbdmode;
	 struct kbd *k;
	 int keycode, state, c;
	 char cPrint = (char)(*pcControl > 1);

	 if (!kbhit())
	 	return -1;
	 /* Get the keyboard mode, then set it to binary mode.
	  * This is done to prevent NT from handling control keys
	  * like Control-C and Control-S.
	  */
	 if (GetConsoleMode(_hin, &kbdmode) != TRUE)
		  return -1;

	 if (SetConsoleMode(_hin, 0 ) != TRUE)
		  return -1;

	 /* Get keyboard events until a recognizable one appears.
	  */
	 for (;;)
	 {
		  if (ReadConsoleInput(_hin, &inp, 1, &nread) != TRUE)
		  {
				c = -1;
				break;
		  }
		  else if ((inp.EventType==KEY_EVENT) && inp.Event.KeyEvent.bKeyDown)
		  {
				WORD ScanCode;
				CHAR asc;

				keycode = inp.Event.KeyEvent.wVirtualKeyCode;
				state   = inp.Event.KeyEvent.dwControlKeyState;
				ScanCode = inp.Event.KeyEvent.wVirtualScanCode;
				asc = inp.Event.KeyEvent.uChar.AsciiChar;

				/*
				 * Test for the Win95 anomaly where the NumericPad '.' button
				 * (with NumLock on) returns a 0x2E (VK_DELETE) when it should
				 * return a 0x6E (VK_DECIMAL) like NT does.
				 */
				*pcControl = (char)(state & (RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED) ? 1 : 0);
				if ((state & NUMLOCK_ON) && (keycode == VK_DELETE) && !(state & ENHANCED_KEY))
					 keycode = VK_DECIMAL;

				if (keycode == VK_RETURN && (state & ENHANCED_KEY)) {
					keycode = VK_ENTER;
					state &= ~ENHANCED_KEY;
				}
				/* Look up the virtual keycode in the table.  Ignore
				 * unrecognized keys.
				 */
				for (k = &kbdtab[0]; keycode != k->keycode && k->keycode != -1; k++)
					 ;
#if defined(_MBCS)
				if (k->keycode == -1 && !inp.Event.KeyEvent.uChar.AsciiChar)    /* value not in table */
					 continue;
#else
				//if (k->keycode == -1)       /* value not in table */
					// continue;
#endif

				/* Check the state of the shift keys.  ALT has highest
				 * priority, followed by Control, followed by Shift.
				 * Select the appropriate table entry based on shift state.
				 */
				if (state & (RIGHT_ALT_PRESSED | LEFT_ALT_PRESSED))
					 c = k->alt;
				else if (state & (RIGHT_CTRL_PRESSED | LEFT_CTRL_PRESSED))
					 c = k->ctrl;
				else if (state & SHIFT_PRESSED)
					 c = k->shift;
				else
					c = k->normal;
				if ((c != -1) && ((c & 0xff) == 0xff))
					c = (c & 0xff00) | asc;
				if (!(c & 0xff) && (state & ENHANCED_KEY) && keycode != VK_DIVIDE)
					c |= 0xe0;
				if (c == -1)
					break;
				c &= 0xffff;
				if (cPrint) {
					printf("%04x %02x %02x %02x %c ",c,keycode,ScanCode,
															asc,isgraph(asc) ? asc : '.');
					printf("%s%s%s%s%s%s%s%s%s\n",
								state & CAPSLOCK_ON ? "Caps " : "",
								state & ENHANCED_KEY ? "Enh " : "",
								state & LEFT_ALT_PRESSED ? "LAlt " : "",
								state & LEFT_CTRL_PRESSED ? "LCtl " : "",
								state & NUMLOCK_ON ? "Num " : "",
								state & RIGHT_ALT_PRESSED ? "RAlt " : "",
								state & RIGHT_CTRL_PRESSED ? "RCtl " : "",
								state & SCROLLLOCK_ON ? "Scroll " : "",
								state & SHIFT_PRESSED ? "Shft " : "");
				}
				break;
		  }
	 }

	 /* Change the keyboard back to its original mode.
	  */
	 if (SetConsoleMode(_hin, kbdmode) != TRUE)
		  return -1;

	 return c;
}


