//
//   $Id: LOGIN.H 1.2 2002/10/16 16:40:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 * Header file for login.c
 */

#define L_ALLOWED		1	//operation allowed
#define L_OK			0	//operation OK
#define LERR_PASS		-1	//password invalid
#define LERR_USER 	-2 //user invalid
#define LERR_PERM 	-3 //no permission for operation
#define LERR_ACC  	-4 //no access to directory
#define LERR_TYPE 	-5 //invalid type specified
#define LERR_MEM		-6 //not enough memory available for op
#define LERR_FOPEN	-7 //cannot open file

/* Initialize login and verification service
 *
 * szFilename   default permissions file. If 0 given, no defaults will be loaded
 * cUseHTAccess   flag indicating usage of HTAccess
 *
 * Return L_OK if initialize worked
 * LERR_MEM or LERR_OPEN if error occured
 */
int InitLogin(char *szFilename, char cUseHTAccess);

/*
 * Free all login information
 */
void FreeLoginInf(void);

/*
 * Verify an operation (action)
 *
 * Arguments:
 * cType - Type of operation, either RQ_CGI, RQ_GET, RQ_POST, RQ_REMCON
 * szPathname - Full path for request
 * szUser - Username
 * szPass - Password supplied by user
 *
 * Return:
 * L_ALLOWED if and only if allowed
 * LERR_PASS if password incorrect
 * LERR_USER if user not found
 * LERR_PERM if user and password exists but no permission
 * LERR_ACC  if no access to path
 * LERR_TYPE if invalid cType
 */
int Verify(char cType, char *szPathname, char *szUser, char *szPass);

/*
 * Process a path name string and put the output in szNewPath
 * If it is not an absolute directory httpdir is prepended
 * At most iMaxlen characters will be put in szNewPath
 * All ../ and ./ references are resolved and removed
 *
 *	Return the length of the szNewPath
 */
int CalcFullPath(char *szNewPath, char *szPath, int iMaxLen);
