//
//   $Id: PRINTING.H 1.4 2003/08/17 02:55:26Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// PRINTING.H Used for SPD (Socket Printing Daemon),
// LPD (Line Printer Daemon) and LPDD (direct LPD - no spooling)
// Revision history
// 2002-04-26 gvn /d=<delay count> to replace global /p=

#include <stdio.h>
#include "mbuf.h"

#define NUMPRINT 4
#define LPT	 0x17
#define INT17_FLAG 0x0100

// AH = 2 read the printer Status into (ah)
//   7   6   5   4   3   2   1   0
//   |   |   |   |   |   |   |   +-- time out
//   |   |   |   |   |   |   +------ last character sent to printer (Non-standard)
//   |   |   |   |   |   +---------- unused
//   |   |   |   |   +-------------- 1 = i/o error
//   |   |   |   +------------------ 1 = selected
//   |   |   +---------------------- 1 = out of paper
//   |   +-------------------------- 1 = acknowledge
//   +------------------------------ 1 = not busy
//
// DX =  printer to be used (0,1,2)

// printer status
#define PS_READY		0x80
#define PS_ACKED		0x40
#define PS_PAPEROUT  0x20
#define PS_SELECTED	0x10
#define PS_ERROR		0x08
#define PS_LASTCHAR	0x02
#define PS_TIMEOUT	0x01

// PS_READY	| PS_ACKED | PS_PAPEROUT | PS_SELECTED | PS_ERROR | PS_LASTCHAR | PS_TIMEOUT

typedef struct {	// printer queue
	MBUF *mbuf;		// mbuf being printed
	struct s_printcon *pPrintCon;	// currently printing connection
	TCB *ltcb;		// listen tcb
	char *pszName;	// printer name for LPD
	char mask;		// status mask
	char stat;		// status to print on
	char report;	// status bits to report on
	char status;	// last status from printer
	WORD flags;		// Various flags for busy etc.
#define PRT_SRV_BUSY	0x0001
#define LPD_SRV_BUSY	0x0002
#define LPD_FF			0x0004
#define LPD_CR			0x0008
#define LAST_CR		0x0010
#define LAST_LF		0x0020
#define CAN_INPUT		0x0040	// input supported
	WORD wDelay;	// Delay count
} PRINTQUE;

typedef struct {
	char far *name;		// Name in INET.INI of printer
	char *DataFileName;  // Name of data file being printed
	char *PrintBuffer;	// Buffer for printer data
	FILE	*fp;				// File Pointer for the opened file
	long	pos;				// Position in log file where found
	size_t cnt;				// Number of bytes remaining
	char	*pBuf;			// Pointer to buf
	WORD	flags;			// conversion flags
#define LPD_PRINT_FORMATTED	1
	char	eof;				// end of file encountered
} LPD_STAT;

// Printer connection control block. Used for LPDD/SPD

#define PR_BUFLEN		80	// buffer length
#define PR_USERLEN	2	// not yet implemented

typedef  struct s_printcon {
	// Following used by both LPDD and SPD
	struct s_printcon *psNext;	// link to next connection
	TCB 	*pTCB;					// tcb of connection
	PRINTQUE	*pPrintQue;			// pointer to printer queue
	WORD	wState;					// state of the connection
	// Following only used by LPDD
	WORD	wJobNumber;				// LPD job number
	char	pcBuf[PR_BUFLEN];		// buffer for LPD parameters
	char	pcUser[PR_USERLEN];	// user for status/remove
	int	iIndex;					// index into buffer
	DWORD	dwLen;					// length of Control or Data
} PRINTCON;

// Values of wState:
enum {
	ST_CONNECTED,
	ST_PRINT_JOBS,
	ST_RECEIVE_JOB,
	ST_STATUS_SHORT,
	ST_STATUS_LONG,
	ST_REMOVE_JOBS,
	ST_RCV_CONTROL_COUNT,
	ST_RCV_CONTROL_JOBNUM,
	ST_RCV_CONTROL_DATA,
	ST_RCV_CONTROL_END,
	ST_RCV_DATA_COUNT,
	ST_RCV_DATA_JOBNUM,
	ST_RCV_DATA_DATA,
	ST_RCV_DATA_END,
	ST_SPD
};

#define LPDCON_LEN	sizeof(PRINTCON)
#define SPDCON_LEN	(sizeof(PRINTCON *) + sizeof(TCB	*) + sizeof(WORD))

// LPD acknowledgements
#define GOOD					0	// Positive acknowledge
#define BAD_COMMAND			1	// Bad command
#define BAD_SUBCOM			2	// Bad sub-command
#define BAD_QLEN				3	// Printer Queue name too long
#define BAD_END				4	// Zero expected

// LPD Commands as sent by client:
#define LPD_START_PRINTING	0x01
#define LPD_RECEIVE_JOB		0x02
#define LPD_STATUS_SHORT	0x03
#define LPD_STATUS_LONG		0x04
#define LPD_REMOVE_JOB		0x05

// Sub-commands for LPD_RECEIVE_JOB:
#define ABORT_JOB					0x01	// Abort all received jobs
#define RECEIVE_CONTROL_FILE	0x02	// Receive Control File Parameters
#define RECEIVE_DATA_FILE     0x03 	// Receive Data File Parameters

void SendAck(PRINTCON *pPrintCon, int type);

extern char cPrintingUsed;	// set if printing is used
extern char print_flag;		// set if any printer active
extern PRINTQUE pque[];		// Status of printing ports
