//
//   $Id: RIP.H 1.2 2002/10/16 16:40:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* RIP.H
 * Routing Information Protocol (RIP)
 * Updated @ Date 19 October 1993, 23:3 by Gerrit van Niekerk
 */

#ifndef	__RIP_H
#define	__RIP_H
#ifndef	__MBUF_H
#include "mbuf.h"
#endif

#ifndef	__IFACE_H
#include "iface.h"
#endif

#ifndef __UDP_H
#include "udp.h"
#endif

#define	RIP_INFINITY	16
#define	RIP_TTL		240	/* Default time-to-live for an entry */
#define	RIP_IPFAM	2

/* UDP Port for RIP */
#define	RIP_PORT	520

/* RIP Packet Types */
#define RIPCMD_REQUEST		1	/* want info */
#define RIPCMD_RESPONSE		2	/* responding to request */
#define	RIPCMD_MAX		3

#define HOPCNT_INFINITY		16	/* per Xerox NS */
#define MAXRIPROUTES		25	/* maximum # routes per RIP pkt */

struct rip_list {
	struct rip_list *prev;
	struct rip_list *next;	/* doubly linked list */

	/* address to scream at periodically:
	 * this address must have a direct network interface route and an
	 * ARP entry for the appropriate  hardware broadcast address, if approp.
	 */
	DWORD dest;

	/* basic rate of RIP clocks on this net interface */
	DWORD interval;

	struct timer rip_time;	/* time to output next on this net. iface */

	/* the interface to transmit on  and receive from */
	IFACE *iface;

	/* described below with the mask defs */
	char	flags;
#define	RIP_SPLIT 0x1	/* Do split horizon processing */
#define	RIP_US	0x2	/* Include ourselves in the list */
};
#define	NULLRL	(struct rip_list *)0

/* Host format of a single entry in a RIP response packet */	
struct rip_route {
	WORD	addr_fam;
#define AUTHENT	0xffff
	WORD	route_tag;
#define	auth_type route_tag
	DWORD	target;
#define authent(p) (BYTE *)(p->target)
	DWORD	subnet_mask;
	DWORD	next_hop;
	DWORD	metric;
};
#define	RIPROUTE	20	/* Size of each routing entry */
#define	RIPHEADER	4	/* Size of rip header before routes */
#define	MAXRIPPACKET	RIPHEADER + (MAXRIPROUTES*RIPROUTE)

/* RIP statistics counters */
struct rip_stat {
	DWORD output;		/* Packets sent */
	DWORD rcvd;		/* Packets received */
	DWORD request;		/* Number of request packets received */
	DWORD response;		/* Number of responses received */
	DWORD unknown;		/* Number of unknown command pkts received */
	DWORD version;		/* Number of version errors */
	DWORD addr_family;	/* Number of address family errors */
	DWORD refusals;		/* Number of packets dropped from a host
					on the refuse list */
};

struct rip_refuse {
	struct rip_refuse *prev;
	struct rip_refuse *next;
	DWORD	target;
};
#define	NULLREF	(struct rip_refuse *)0

/* RIP primitives */
int rip_init (void);
void rt_timeout (void *s);
void rip_trigger (IFACE *ifp);
int rip_add (DWORD dest,DWORD interval,char flags);
int riprefadd (DWORD gateway);
int riprefdrop (DWORD gateway);
int ripreq (DWORD dest,WORD replyport);
int rip_drop (DWORD dest);
int nbits (DWORD target);
void pullentry (struct rip_route *ep,struct mbuf **bpp);
void rip_drop_iface(IFACE *ifp);

/* RIP Definition */
extern WORD Rip_trace;
extern int Rip_merge;
extern struct rip_stat Rip_stat;
extern struct rip_list *Rip_list;
extern struct rip_refuse *Rip_refuse;
extern struct udp_cb *Rip_cb;

#endif	/* __RIP_H */

