//
//   $Id: SER.H 1.3 2002/10/16 16:40:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// SER.H
// Copyright (c) GP van Niekerk Ondernemings BK 1990-2000

// I/O definitions specific to serial I/O

#define	SER_INT_MAX	5		// # of interrupt vectors allowed for serial ports
extern unsigned nser;		// Actual number of serial lines
extern unsigned *crctab;	// crc table

// Asynch controller control block
typedef struct ser {
	void interrupt (*oldvec)(void);// Original interrupt vector [cs:ip]
	char *buf;					// Ring buffer
	unsigned bufsize;			// Size of ring buffer
	char *wp;					// Write pointer
	char *rp;					// Read pointer
	unsigned short icnt;		// count of characters in buffer
	unsigned stop_cnt;		// count at which to stop flow
	unsigned start_cnt;		// count at which to start flow
	BYTE *data;					// current output pointer
	unsigned short cnt;		// byte count remaining
	struct ser *palt;			// pointer to alternate control block
	unsigned flags;			// flags

#define 	FL_HDLC			0x01	// hdlc mode
#define 	FL_OFSTOPPED	0x02	// output flow stopped
#define 	FL_IFSTOPPED	0x04	// input flow stopped
#define 	FL_MODEM			0x08	// modem type connection
#define	FL_CONNECTED	0x10	// connected in modem mode
#define	FL_IS16550A		0x20	// is 16550A chip
#define	FL_CHANB			0x40	// channel B for SCC's
#define	FL_LAPB			0x80	// LAPB used on serial link
#define	FL_EVENT			0x100	// modem event
#define 	FL_PPP			0x200	// Point to Point Protocol
#define	FL_DISABLED		0x400	// This interface is disabled
#define	FL_SEND_XON		0x800	// Send a flow-control XON
#define	FL_SEND_XOFF	0x1000// Send a flow-control XOFF
#define	FL_SUB_ISR		0x2000// Subsequent interrupt service routine

	char iflow;					// input flow control
#define IF_XON	     1
#define IF_DTR	     2
#define IF_RTS	     3
#define IF_INV	     8		// inverse of modem signal
	char oflow;					// output flow control
#define OF_XON	     1
#define OF_DCD	     2
#define OF_DSR	     3
#define OF_CTS	     4
#define OF_INV	     8		// inverse of modem signal
	char mcr;					// value of modem control register
	unsigned addr;				// Base I/O address
	unsigned vec;				// Interrupt vector
	struct ser *next;			// next serial controller using same int vector
	unsigned long speed;		// Line speed
	char datamask;				// data bits MASK (7F or FF)
	char lcr;					// value of line control register
	char oldlcr;				// old value of line control register
	char oldmask;				// old interrupt mask
	// raw interface usage
	void *iface;				// Pointer to interface
	struct timer timer;		// Character timer
	unsigned overrun;			// overrun error count
	unsigned parerr;			// parity error count
	unsigned frmerr;			// framing/CRC error count
	unsigned brkerr;			// break/abort error count
	unsigned buferr;			// insufficient buffer space
	unsigned undererr;		// xmit underrun errors
	unsigned long rcvcount;	// number of characters received
	unsigned long xmtcount;	// number of characters xmitted
	char deliver;				// set when input to deliver
#define DEL_CR	1				// cr seen in data
#define DEL_LF	2				// lf seen in data
#define DEL_TO	4				// character timeout
#define DEL_BH	8				// buffer half full
	char ostate;				// output state
#define OS_IDLE		0
#define OS_RAW			1
#define OS_FSTART		2
#define OS_DATA		3
#define OS_FRESC		4
#define OS_FREND		5
#define OS_CRC1		6
#define OS_CRC1ESC	7
#define OS_CRC1END	8
#define OS_CRC2		9
#define OS_CRC2ESC	10
#define OS_CRC2END	11
#define OS_END			12
#define OS_FHDLC		13
#define OS_HDLCDATA	14
#define OS_HDLCEND	15
#define OS_HDLCCRC	16
#define OS_HDLC_EA	17
#define OS_FCS1		18
#define OS_FCS1ESC	19
#define OS_FCS2		20
#define OS_FCS2ESC	21
#define OS_MODEM		22

	char istate;		// Input state
#define IS_RAW			0
#define IS_DATA		1
#define IS_ESC			2
#define IS_IGN			3
#define IS_DATAB		4
#define IS_ESCB		5
#define IS_IGNB		6
#define IS_HDLC		7
#define IS_PPP			8
#define IS_PPPESC		9
#define IS_PPPIGN		10
#define IS_MODEM		11
//#define IS_MODEM_RCV	9
//#define IS_MODEM_XID	10

	// slip structures
	struct mbuf *sndq;	// send queue
	struct mbuf *tbp;		// Transmit mbuf being sent
	WORD oaccum;			// CRC output accumulator
	struct mbuf *rcvq;	// receive queue
	struct mbuf *rbp;		// Pointer to mbuf currently being written
	WORD accum;			// CRC input accumulator
	WORD errors;			// Receiver input errors
	// PPP values
	int32 accm_o;			// async char control mask - output
	int32 accm_i;			// async char control mask - input
	BYTE  last_o_ch;		// last output char
	BYTE  last_i_ch;		// last input char
	// modem parameters
	MODEMPAR *pmpar;		// pointer to modem parameters
	struct slcompress *slcomp;
#ifdef V25
	void interrupt (*oldvec1)(void);// Original interrupt vector1 [cs:ip]
	BYTE volatile far *RxB;
	BYTE volatile far *TxB;
	BYTE volatile far *SCM;
	volatile BYTE far *SCE;
#endif
} SER;

extern struct ser *ser[];

// scc mode

#define SCC_HDLC 	0
#define SCC_ASYNC	1

// flow control characters

#define XON		0x11
#define XOFF	0x13

// SLIP definitions

#define	SLIP_ALLOC	40	// Receiver allocation increment

#define	FR_END		0300	// Frame End
#define	FR_ESC		0333	// Frame Escape
#define	T_FR_END		0334	// Transposed frame end
#define	T_FR_ESC		0335	// Transposed frame escape

// prototypes
// asy.c

int asy_init(IFACE *ifp,unsigned int speed);
void asy_stop(IFACE *iface);
asy_ioctl(IFACE *interface,int argc,char *argv[]);
int asy_speed(IFACE *ifp,unsigned int speed);
asy_xmit(IFACE *ifp,BYTE *buf,WORD cnt);
int asy_srecv(IFACE *ifp,char *buf,WORD cnt);
asy_modem_init(IFACE *ifp);
void setbit(unsigned port,char bits);
void clrbit(unsigned port,char bits);

// ascc.c

int scc_init(WORD dev,char mode);
int scc_stop(struct interface *iface);
scc_ioctl(struct interface *interface,int argc,char *argv[]);
int scc_speed(WORD dev,int speed);
scc_xmit(IFACE *ifp,char *buf,WORD cnt);
int scc_srecv(IFACE *ifp,char *buf,WORD cnt);

// s_iface.c

int find_mdd(IFACE *ifp,char *xid);
void release_sif(IFACE *ifp);

