//
//   $Id: TCP.H 1.4 2003/02/15 02:10:18Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// TCP.H
// TCP implementation. Follows RFC 793 as closely as possible

// 2003-02-14 gvn update for Listen/Accept

#ifndef __TCP_H
#define __TCP_H
#ifndef __NETUSER_H
#include "netfunc.h"
#endif
#ifndef __IP_H
#include "ip.h"
#endif
#ifndef __INTERNET_H
#include "internet.h"
#endif


#define	DEF_WND	2920	// Default receiver window
#define	NTCB		19		// # TCB hash table headers
#define	DEF_MSS	1460	// Default maximum segment size
#define	DEF_RTT	1000	// Initial guess at round trip time (1 sec)
#define	MSL2		30		// Guess at two maximum-segment lifetimes
// Round trip timing parameters
#define	AGAIN		8		// Average RTT gain = 1/8
#define	DGAIN		4		// Mean deviation gain = 1/4

// TCP segment header -- internal representation
// Note that this structure is NOT the actual header as it appears on the
// network (in particular, the offset field is missing).
// All that knowledge is in the functions ntohtcp() and htontcp() in tcpsubr.c

#define TCPLEN			20	// Minimum Header length, bytes
#define	TCP_MAXOPT	40	// Largest option field, bytes
struct tcp {
	WORD source;		// Source port
	WORD dest;			// Destination port
	int32 seq;			// Sequence number
	int32 ack;			// Acknowledgment number
	char flags;			// Flags, data offset
#define	URG	0x20	// URGent flag
#define	ACK	0x10	// ACKnowledgment flag
#define	PSH	0x08	// PuSH flag
#define	RST	0x04	// ReSeT flag
#define	SYN	0x02	// SYNchronize flag
#define	FIN	0x01	// FINal flag
	WORD wnd;			// Receiver flow control window
	WORD checksum;	// Checksum
	WORD up;			// Urgent pointer
	WORD mss;			// Optional max seg size
	char optlen;		// Length of options field, bytes
	char options[TCP_MAXOPT];	// Options field
};
// TCP options
#define	EOL_KIND		0
#define	NOOP_KIND	1
#define	MSS_KIND		2

#define	TCPLEN		20
#define	MSS_LENGTH	4
// Resequencing queue entry
struct reseq {
	struct reseq *next;	// Linked-list pointer
	char tos;				// Type of service
	struct tcp seg;		// TCP header
	struct mbuf *bp;		// data
	WORD length;			// data length
};
#define	NULLRESEQ	(struct reseq *)0

// TCP connection control block
#define TCB_DEF
typedef struct tcb {
	struct tcb *prev;		// Linked list pointers for hash table
	struct tcb *next;

	struct connection conn;

	char state;	// Connection state
#define	CLOSED			0	// Must be 0
#define	LISTEN			1
#define	SYN_SENT			2
#define	SYN_RECEIVED	3
#define	ESTABLISHED		4
#define	FINWAIT1			5
#define	FINWAIT2			6
#define	CLOSE_WAIT		7
#define	CLOSING			8
#define	LAST_ACK			9
#define	TIME_WAIT		10

	char reason;			// Reason for closing
#define	NORMAL_R		0	// Normal close
#define	RESET			1	// Reset by other end
#define	TIMEOUT		2	// Excessive retransmissions
#define	NETWORK		3	// Network problem (ICMP message)

	// If reason == NETWORK, the ICMP type and code values are stored here
	char type;
	char code;

	// Send sequence variables
	struct {
		int32 una;	// First unacknowledged sequence number
		int32 nxt;	// Next sequence num to be sent for the first time
		int32 ptr;	// Working transmission pointer
		WORD wnd;	// Other end's offered receive window
		int32 up;	// Send urgent pointer
		int32 wl1;	// Sequence number used for last window update
		int32 wl2;	// Ack number used for last window update
	} snd;
	int32 iss;		// Initial send sequence number
	WORD cwind;	// Congestion window
	WORD ssthresh;// Slow-start threshold
	int32 resent;	// Count of bytes retransmitted

	// Receive sequence variables
	struct {
		int32 nxt;	// Incoming sequence number expected next
		WORD wnd;	// Our window
		int32 up;	// Receive urgent pointer
	} rcv;
	int32 irs;		// Initial receive sequence number
	WORD mss;		// Maximum segment size
	int32 rerecv;	// Count of duplicate bytes received

	WORD window;	// Receiver window and send queue limit

	char backoff;	// Backoff interval
	BYTE retry;		// retry count
	void (far *r_upcall)(struct tcb *,int);		// Call when "significant" amount of data arrives
	void (far *t_upcall)(struct tcb *,int,int);	// Call when ok to send more data
	void (far *s_upcall)(struct tcb *,char,char);// Call when connection state changes
	char flags;					// Control flags
#define	FORCE			1		// We owe the other end an ACK or window update
#define	CLONE			2		// Server-type TCB, cloned on incoming SYN
#define	RETRAN		4		// A retransmission has occurred
#define	TCP_RCV		8		// Set on data arrival, reset by user
#define	TCP_XMT		0x10	// Set when ok to send more, reset by user
#define	TCP_STC		0x20	// Set on state change, reset by user
#define	SEND_URG		0x40	// Set when sending urgent data
#define	TCP_EXT		0x80	// External connection (via API)
	char tos;					// Type of service (for IP)

	struct mbuf *rcvq;	// Receive queue
	WORD rcvcnt;

	struct mbuf *sndq;	// Send queue
	WORD sndcnt;			// Number of unacknowledged sequence numbers on
								// send queue. NB: includes SYN and FIN, which don't
								// actually appear on sndq!
	struct reseq *reseq;	// Out-of-order segment queue
	struct timer timer;	// Retransmission timer
	struct timer rtt_timer;	// Round trip timer
	int32 rttseq;			// Sequence number being timed
	int32 srtt;				// Smoothed round trip time, milliseconds
	int32 mdev;				// Mean deviation, milliseconds
	char *user;				// User link
	// Illogical position for new ack strategy variables to remain compatible
	// with old external code
	TIMER	ack_timer;		// Ack timer
	WORD rcv_wndav;		// Our real available receive window
	WORD mss_rcv;			// MSS for receiving

	WORD cnt_to_urg;		// count to urgent data
	WORD urg_cnt;			// count of urgent data
	char flags1;			// some more flags
#define F1_LISTEN0	1	// was listening on 0.0.0.0:0
#define F1_KEEPALIVE	2	// keep alive is active
#define F1_KATIMING	4	// keep alive timer running
#define F1_MUSTACK	8	// Ack is outstanding
#define F1_MAXSEGRCVD	0x10	// Max segment has been received
	BYTE num_nacc;			// Number of not accepted connections
	BYTE max_nacc;			// Maximum number not accepted connections
	struct tcb *naccq;	// Not accepted connections queue
} TCB;

#define	NULLTCB	(struct tcb *)0

// TCP statistics and status variables
#define TCP_STAT_DEF
struct tcp_stat {
	WORD runt;						// Smaller than minimum size
	WORD checksum;				// TCP header checksum errors
	WORD conout;					// Outgoing connection attempts
	WORD conin;					// Incoming connection attempts
	WORD resets;					// Resets generated
	WORD bdcsts;					// Bogus broadcast packets
	WORD tcp_window;
	WORD tcp_mss;					// Maximum segment size to be sent with SYN
	WORD tcp_irtt;				// Initial guess at round trip time
	WORD tcp_retry;				// maximum transmission retries
	WORD tel_tos;					// TOS used by Telnet
	char flush_flag;				// Flag to decide when to send
#define FLUSH_UNCONDITIONAL 1	// Override Nagle
#define FLUSH_DONTSEND	    2	// Don't send now
	struct tcb *tcbs[NTCB];
	WORD tcp_smss;				// maximum send segment size
	WORD tcp_time_max;			// maximum time for retry
	char	cAckOnPush;				// send an ack when push bit is received
};

extern struct tcp_stat tcp_stat;

WORD getlport(void);
extern int32 iss(void);
struct tcb *lookup_tcb(struct connection *conn);
struct tcb *create_tcb(struct connection *conn);
void rehash_tcb(struct tcb *tcb);
void tcp_output(struct tcb *tcb);
void close_self(struct tcb *tcb,WORD how);
void dump_seg(struct tcb *tcb);
void setstate(struct tcb *tcb,char state);
void send_syn(struct tcb *tcb);
void link_tcb(struct tcb *tcb);
void unlink_tcb(struct tcb *tcb);
void tcp_timeout(struct tcb *tcb);
void reset_tcp(struct tcb *tcb);
void reset_current(struct tcb *tcb);
int tcpval(struct tcb *tcb);
void tcp_input(IFACE *ifp,struct mbuf *bp,char protocol,DWORD source,
			DWORD dest,char tos,WORD length,char rxbroadcast);

struct mbuf *htontcp(struct tcp *tcph,struct mbuf *data,struct pseudo_header *ph);

void state_tcp(struct tcb *tcb,int state),
			tcp_dump(struct mbuf **bpp,DWORD source,DWORD dest,int check);

// prototypes
struct tcb *
open_tcp(
struct socket *lsocket,	// Local socket
struct socket *fsocket,	// Remote socket
int mode,					// Active/passive/server
WORD window,				// Receive window (and send buffer) sizes
void (far *r_upcall)(),	// Function to call when data arrives
void (far *t_upcall)(),	// Function to call when ok to send more data
void (far *s_upcall)(),	// Function to call when connection state changes
char tos,
char *user);		// User linkage area
TCB *accept_tcp(TCB *tcb);
int send_tcp(struct tcb *tcb,struct mbuf *bp,char urgent);
int recv_tcp(struct tcb *tcb,struct mbuf **bp,WORD cnt,char *urgent);
#define URG_WAITING	1	// urgent data waiting
#define URG_READ		2	// urgent data contained in current read
int close_tcp(struct tcb *tcb);
int del_tcp(struct tcb *tcb);
int tprintf(struct tcb *tcb,const char *message,...);
int tcp_active(void);
void tcp_resetall(void);
char *tcp_port(WORD n);
int kick_tcp(struct tcb *tcb);
int backoff(int n);
int seq_within(int32 x,int32 low,int32 high);
int seq_lt(int32 x,int32 y);
int seq_le(int32 x,int32 y);
int seq_gt(int32 x,int32 y);
int seq_ge(int32 x,int32 y);
int ntohtcp(struct tcp *tcph,struct mbuf **bpp);

#endif
