//
//   $Id: UDP.H 1.3 2003/01/07 22:59:02Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* UDP.H
 * User Datagram Protocol definitions
2002-12-18 gvn make udp tables accessable to external program
 */

#ifndef __UDP_H
#define __UDP_H
#ifndef __IFACE_H
#include "iface.h"
#endif
#ifndef __NETUSER_H
#include "netuser.h"
#endif
#ifndef __INTERNET_H
#include "internet.h"
#endif

#define	NUDP	20

/* Structure of a UDP protocol header */
struct udp {
	WORD source;	/* Source port */
	WORD dest;	/* Destination port */
	WORD length;	/* Length of header and data */
	WORD checksum;	/* Checksum over pseudo-header, header and data */
};
#define	UDPHDR	8	/* Length of UDP header */

/* User Datagram Protocol control block
 * Each entry on the receive queue consists of the
 * remote socket structure, followed by any data
 */
struct udp_cb {
	struct udp_cb *prev;	/* Linked list pointers */
	struct udp_cb *next;
	struct socket socket;	/* Local port accepting datagrams */
	void (far *r_upcall)(struct socket *,int,struct interface *);	/* Function to call when one arrives */
	struct mbuf *rcvq;	/* Queue of pending datagrams */
	int rcvcnt;		/* Count of pending datagrams */
	int flags;		/* Flags describing control block */
#define UDPF_EXT 1		/* Externally defined control block */
};
#define	NULLUDP	(struct udp_cb *)0

/* UDP statistics counters */
struct udp_stat {
	WORD rcvd;		/* Packets received */
	WORD sent;		/* Packets sent */
	WORD cksum;	/* Checksum errors */
	WORD unknown;	/* Unknown socket */
	WORD bdcsts;	/* Incoming broadcasts */
	struct udp_cb *udps[NUDP];	/* Hash table for UDP structures */
};
extern struct udp_stat udp_stat;	/* Statistics */

/* UDP prototypes */
int open_udp(struct socket *lsocket,void (far *r_upcall)(),int ext);
int send_udp(
	struct socket *lsocket,	/* Source socket */
	struct socket *fsocket,	/* Destination socket */
	char tos,				/* Type-of-service for IP */
	char ttl,				/* Time-to-live for IP */
	struct mbuf *bp,		/* Data field, if any */
	WORD length,			/* Length of data field */
	WORD id,				/* Optional ID field for IP */
	char df,				/* Don't Fragment flag for IP */
	IFACE *ifp);				// send directlu to interface if set
int recv_udp(
	struct socket *lsocket,		/* Local socket to receive on */
	struct socket *fsocket,		/* Place to stash incoming socket */
	struct mbuf **bp);		/* Place to stash data packet */
int del_udp(struct socket *lsocket);
void udp_dump();
void udp_input(IFACE *ifp,MBUF *bp,char protocol,DWORD source,
	DWORD dest,char tos,WORD length,char rxbroadcast);
struct udp_cb *lookup_udp(struct socket *socket);
MBUF *htonudp(struct udp *udp,MBUF *bp,struct pseudo_header *ph);
void ntohudp(struct udp *udp,struct mbuf **bpp);

#endif
