//
//   $Id: ARPCMD.C 1.6 2003/07/01 19:58:06Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// ARPCMD.C
//
// Copyright (c) 2001 Datalight, Inc
// All Rights Reserved
// Datalight Confidential and Proprietary
//
// ARP commands
//
// Portions Copyright (c) GP van Niekerk Ondernemings 1990-2001
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
// Revision History
// 07/23/2001 gvn used updated allocation routines and parameters

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif


#include <stdio.h>
#include <string.h>
#include <alloc.h>
#include <ctype.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "timer.h"
#include "enet.h"
#include "arp.h"
#include "cmdparse.h"
#include "netfunc.h"

#if APL
static void dumparp(void);
#endif

typedef struct interface {
	int dummy;
} IFACE;

extern int arp_init_type;
extern char badhost[];
extern char nospace[];

int doarpadd(),doarpdrop(), doarpinit();
struct scmds far arpcmds[] = {
	{"add", doarpadd},
#if APL
	{"drop", doarpdrop},
#endif
   {"init", doarpinit},
	{"", NULLFP}
};

#if APL
char *arptypes[] = {
	"",
	"10 Mb Ethernet",
	"3 Mb Ethernet",
	"AX.25",
	"Pronet",
	"Chaos",
	"IEEE",
	"Arcnet",
	"AppleTalk",
	"IPX",
};
#endif

int doarp(int argc,char *argv[])
{
	if(argc < 2){
#if APL
		dumparp();
#endif
		return 0;
	}
	return subcmd(arpcmds,argc,argv,0);
}

static doarpadd(int argc,char *argv[])
{
	WORD hwalen,hardware,naddr;
	DWORD addr;
	BYTE *hwaddr;
	struct arp_tab *ap;
	struct arp_type *at;
	void arp_drop();

	if((addr = resolve(argv[1])) == 0){
		printf(badhost,argv[1]);
		return 1;
	}
	/* This is a kludge. It really ought to be table driven */
	switch(tolower(argv[2][0])){
	case 'e':	/* "ether" */
		hardware = ARP_ETHER;
		naddr = 1;
		break;		
/*
	case 'a':	/* "ax25" */
		hardware = ARP_AX25;
		naddr = argc - 3;
		break;
	case 'm':	/* "mac appletalk" */
		hardware = ARP_APPLETALK;
		naddr = 1;
		break;
*/
	case 'i':	/* ipx or ieee */
		hardware = tolower(argv[2][1]) == 'p' ? ARP_IPX : ARP_IEEE;
		naddr = 1;
		break;
	default:
#if APL
		printf("unknown hardware type \"%s\"\n",argv[2]);
#endif
		return -1;
	}
	/* If an entry already exists, clear it */
	if((ap = arp_lookup(hardware,addr)) != NULLARP)
		arp_drop(ap);

	at = &arp_type[hardware];
	if(at->scan == 0){
#if APL
		printf("No interface\n");
#endif
		return 1;
	}
	/* Allocate buffer for hardware address and fill with remaining args */
	hwalen = at->hwalen * naddr;
	if((hwaddr = malloc(hwalen,AllocArpcMac)) == 0){
		printf(nospace);
		return 0;
	}
	/* Destination address */
	(*at->scan)(hwaddr,&argv[3],argc - 3);
	ap = arp_add(addr,hardware,hwaddr,hwalen,0);	/* Put in table */
	free(hwaddr);					/* Clean up */
	stop_timer(&ap->timer);			/* Make entry permanent */
	ap->timer.count = ap->timer.start = 0;
	return 0;
}

static doarpinit(int argc, char *argv[])
{
	if (argc < 1)
		return -1;
	if (strcmpi(argv[1], "DHCP") == 0)
	{
		arp_init_type = ARP_TYPE_DHCP;
	}
	else if (strcmpi(argv[1], "GRAT") == 0)
	{
		arp_init_type = ARP_TYPE_GRAT;
	}
	else if (strcmpi(argv[1], "BOTH") == 0)
	{
		arp_init_type = ARP_TYPE_BOTH;
	}
	else
	{
		printf("unknown arp type '%s'\n", argv[1]);
		return 1;
	}

	return 0;

}

#if APL
/* Remove an ARP entry */
#pragma argsused
static doarpdrop(int argc,char *argv[])
{
	WORD hardware;
	int32 addr;
	struct arp_tab *ap;
	void arp_drop();

	if((addr = resolve(argv[1])) == 0){
		printf(badhost,argv[1]);
		return 1;
	}
	/* This is a kludge. It really ought to be table driven */
	switch(tolower(argv[2][0])){
	case 'e':	/* "ether" */
		hardware = ARP_ETHER;
		break;
	case 'i':	/* ipx */
		hardware = ARP_IPX;
		break;
/*
	case 'a':	/* "ax25" */
		hardware = ARP_AX25;
		break;
	case 'm':	/* "mac appletalk" */
		hardware = ARP_APPLETALK;
		break;
*/
	default:
		hardware = 0;
		break;
	}
	if((ap = arp_lookup(hardware,addr)) == NULLARP)
		return -1;
	arp_drop(ap);
	return 0;
}
/* Dump ARP table */

static void dumparp(void)
{
	register int i;
	extern struct arp_stat arp_stat;
	register struct arp_tab *ap;
	char e[128];
	extern char *arptypes[];

	printf("ARP: Received %u BadType %u BadAddress %u\n",
			 arp_stat.recv,arp_stat.badtype,arp_stat.badaddr);
	printf("     RequestsIn %u Replies %u RequestsOut %u\n",
		arp_stat.inreq,arp_stat.replies,arp_stat.outreq);
	printf("IP addr         Type           Time Q Addr\n");
	for (i = 0;i < ARPSIZE;i++) {
		for(ap = arp_tab[i];ap != (struct arp_tab *)NULL;ap = ap->next){
			printf("%-16s",inet_ntoa(ap->ip_addr));
			printf("%-15s",arptypes[ap->hardware]);
			printf("%-5ld",ap->timer.count*(long)MSPTICK/1000);
			if(ap->state == ARP_PENDING)
				printf("%-2u",len_q(ap->pending));
			else
				printf("  ");
			if(ap->state == ARP_VALID){
				if(arp_type[ap->hardware].format != 0){
					(*arp_type[ap->hardware].format)(e,ap->hw_addr);
				} else {
					e[0] = '\0';
				}
				printf("%s",e);
			} else {
				printf("[unknown]");
			}
/* print out IPX route address if applicable
			if (ap->hardware == ARP_IPX) {
				if (ap->rt_addr) {
					int i;

					for (i = 0;i < 6;++i)
						printf("%02x",ap->rt_addr[i]);
					printf(" ");
				}
				else	printf(" Unknown");
			}
*/
			if (ap->rif) {
				int i;
				for (i = 0;i < (ap->rif[0] & 0x1f);++i)
					printf(" %02x",(BYTE)ap->rif[i]);
			}
			printf("\n");
		}
	}
	return;
}
#endif
