//
//   $Id: BOOTP.C 1.8 2003/08/21 21:36:22Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	BOOTP.C

 * Center for Information Technology Integration
 *           The University of Michigan
 *                    Ann Arbor
 *
 * Dedicated to the public domain.
 * Send questions to info@citi.umich.edu
 *
 * BOOTP is documented in RFC 951 and RFC 1048
 * Updated to include DHCP as per RFC 2131 and RFC 2132
 * DHCP implementation Copyright (c) GP van Niekerk Ondernemings 1998-2002

 Revision history:
 08/21/2003 jmb get domain suffix from DHCP/BOOTP server
 07/29/2003	gvn include hostname in request if symbolic (from HOSTS)
 03/05/2003 jmb added code to discard any gateway addresses specified after the first one.
 11/19/2002 jmb make DHCP_WWW a compile time option
 2002-09-04 gvn XID constant for duration of DHCP negotiation
 */

/* Cannot do this with DHCP!
#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif
*/

#include <stdlib.h>
#include <mem.h>
#include <string.h>
#include <dos.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "netfunc.h"
#include "udp.h"
#include "iface.h"
#include "ip.h"
#include "internet.h"
#include "rip.h"
#include "enet.h"
#include "arp.h"
#include "bootp.h"
#include "timer.h"
#include "dns.h"
#include "_api.h"

#ifdef DEBUG
static char *mes_type[] = {"0","Discover","Offer","Request","Decline","Ack","Nak","Release","Inform"};
#endif

extern unsigned long master_tick;
extern BYTE hostname[];
extern BYTE *dns_suffix;
#ifdef __LARGE__
extern VARBLOCKL stats;
#else
extern VARBLOCK stats;
#endif

static void bootp_rx(struct socket *plsock,int cnt,IFACE *ifp);
static void ntoh_bootp(struct mbuf **bpp,struct bootp *bootpp);
int send_bootp(IFACE *ifp);

#define BOOTP_RETRANS	2000l		/* The interval between sendings */

BYTE SilentStartup = 0;
DWORD bootp_retrans = BOOTP_RETRANS;
WORD bootp_maxretry = 5;

// Timer function for retransmission of requests and renewing/rebinding
void dhcp_timer(char *arg)
{
	IFACE *ifp = (IFACE *)arg;
	TIMER *t = &ifp->t1;

	if (ifp->send_retry) {		// (re) sending BOOTP/DHCP request
		if (ifp->dhcp_state == D_SELECTING) {
			ifp->dhcp_state = D_INIT;
		}
		else if (ifp->dhcp_state == D_REQUESTING) {
			ifp->dhcp_state = D_SELECTING;
		}
		else if (ifp->dhcp_state == D_RENEWING) {
			if ((long)(ifp->rebind -= t->start * MSPTICK) < MSPTICK ||
									ifp->send_retry > bootp_maxretry) {
				ifp->send_retry = 1;
				goto t2_timeout;
			}
			ifp->dhcp_state = D_BOUND;
			goto t1_timeout;
		}
		else if (ifp->dhcp_state == D_REBINDING) {
			if ((long)(ifp->lease -= t->start * MSPTICK) < MSPTICK ||
										ifp->send_retry > bootp_maxretry) {
				ifp->send_retry = 1;
				goto lease_timeout;
			}
			ifp->dhcp_state = D_RENEWING;
			goto t2_timeout;
		}
		send_bootp(ifp);		// send a retry
		if (ifp->dhcp_state == D_SELECTING)
			ifp->dhcp_state = D_REQUESTING;
	}
	else {
		ifp->send_retry = 1;
		switch (ifp->dhcp_state) {
		case D_BOUND:
			t1_timeout:
			send_bootp(ifp);
			ifp->dhcp_state = D_RENEWING;
			break;
		case D_RENEWING:
			t2_timeout:
			send_bootp(ifp);
			ifp->dhcp_state = D_REBINDING;
			break;
		case D_REBINDING:
			lease_timeout:
			ifp->stat |= IF_DISABLED;
			ifp->dhcp_state = D_INIT;
			ifp->ip_addr = 0;
			send_bootp(ifp);
			break;
		}
	}
	if (ifp->dhcp_state == D_INIT)
		ifp->dhcp_state = D_SELECTING;
}

int bootp(IFACE *ifp)
{
	int retv = -1;
	int sense_key(void);
	int dhcp = ifp->dhcp_state != D_BOOTP;
	char *pprot = dhcp ? "DHCP" : "BOOTP",last_retry;
	TIMER *t = &ifp->t1;

	t->func = dhcp_timer;
	t->arg = (char *)ifp;
	ifp->send_retry = last_retry = 1;
//	randomize();
	srand(peek(0x40,0x6c));
	set_timer(t,(random(1000) + MSPTICK));		// wait a random time of up to 1000 ms
	start_timer(t);

	/* Send the bootp request packet until a response is received or time
		out */
	ifp->stat |= IF_DISABLED;
	while (1) {
		if (ifp->ip_addr) {
			retv = 1;
			break;
		}
		if (ifp->send_retry > bootp_maxretry) {
			printf("%s: timed out, values not set\n",pprot);
			break;
		}
		if (!SilentStartup && ifp->dhcp_state <= D_SELECTING &&
											ifp->send_retry != last_retry) {
			printf("Sending %s request %u...\n",pprot,ifp->send_retry - 1);
			last_retry = ifp->send_retry;
		}
		/* Give other tasks a chance to run. */
#if APL
		if (sense_key() != -1) {
			printf("%s interrupted\n",pprot);
			break;
		}
#else
		schedule();
#endif
		//if (ifp->lsock.port == 0)
		//	continue;				// nothing sent yet
	}
	return retv;
}

char mestype[] = {	0,0,DHCPDISCOVER,DHCPREQUEST,DHCPDECLINE,
					DHCPREQUEST,DHCPREQUEST,DHCPREQUEST,DHCPDECLINE,
					DHCPINFORM};


int send_bootp(IFACE *ifp)
{
	struct socket fsock;
	MBUF *bp;
	BYTE *cp,mtype,if_stat;
	DWORD target_time,retry_time;

	if (ifp->send_retry > bootp_maxretry + 1) {
		ifp->send_retry = 1;
	}
	if (ifp->lsock.port == 0) {		// start UDP if not started
		ifp->lsock.address = 0;
		ifp->lsock.port = IPPORT_BOOTPC;
		if (open_udp(&ifp->lsock,bootp_rx,0) <= 0) {
			ifp->lsock.port = 0;
			return -1;
		}
	}
	fsock.address = ifp->dhcp_state == D_BOUND ?
								ifp->dhcp_server : 0xfffffffful;
	fsock.port = IPPORT_BOOTPS;
	/* Allocate BOOTP packet and fill it in */
	if((bp = alloc_mbuf(BOOTP_SIZE,AllocBootpSnd)) == NULLBUF)
		return -1;
	if (ifp->dhcp_state == D_BOOTP || ifp->dhcp_state == D_INIT)
		ifp->xid = master_tick;
	if (ifp->send_retry == 1)
		ifp->start_time = master_tick;
	retry_time = bootp_retrans * (1 << (ifp->send_retry++ - 1));
	switch (ifp->dhcp_state) {
	case D_BOUND:
		target_time = ifp->rebind;
		break;
	case D_RENEWING:
		target_time = ifp->lease;
		break;
	default:
		target_time = retry_time;
		break;
	}
	if (target_time < retry_time)
		retry_time = target_time;
	set_timer(&ifp->t1,retry_time);
	start_timer(&ifp->t1);
	cp = bp->data;					/* names per the RFC: */
	memset(cp, 0, BOOTP_SIZE);
	*cp++ = BOOTREQUEST;			/* op */
	*cp++ = ifp->hwtype;			/* htype */
	*cp++ = ifp->hwalen;			/* hlen */
	cp++;								/* hops */
	cp = put32(cp,ifp->xid);	/* xid */
	cp = put16(cp,ifp->dhcp_state != D_DECLINING ?
		(WORD)(ifp->xid - ifp->start_time) / MSPTICK : 0);/* secs */
	cp += 2;							/* flags */
	if (ifp->dhcp_state == D_BOUND || ifp->dhcp_state == D_RENEWING ||
				ifp->dhcp_state == D_REBINDING || ifp->dhcp_state == D_INFORMING)
		put32(cp, ifp->ip_addr);/* ciaddr */
	cp += 16;						/* yiaddr */
										/* siaddr */
										/* giaddr */
	memcpy(cp, ifp->hwaddr,ifp->hwalen);
	cp += 208;						/* chaddr sname file */
	cp = put32(cp,0x63825363L);/* options, magic cookie */
	if (ifp->dhcp_state != D_BOOTP) {
		*cp++ = DHCP_MESTYPE;
		*cp++ = 1;					// length
		*cp++ = mtype = mestype[ifp->dhcp_state];
#ifdef DEBUG
		printf("Send %s(%u) to = %ld\n",mes_type[mtype],ifp->dhcp_state,retry_time);
#endif
		*cp++ = DHCP_CLIENTID;
		*cp++ = ifp->hwalen + 1;// length
		*cp++ = ifp->hwtype;
		memcpy(cp, ifp->hwaddr,ifp->hwalen);
		cp += ifp->hwalen;
		if (mtype == DHCPDISCOVER || mtype == DHCPREQUEST || mtype == DHCPINFORM) {
#ifdef DHCPWWW
			static char par_req_list[] = {
				DHCP_PARRLIST,6,BOOTP_SUBNET,BOOTP_GATEWAY,BOOTP_DNS,
				BOOTP_HOSTNAME,DHCP_WWW, BOOTP_DOMAIN_SUFFIX};
#else
			static char par_req_list[] = {
				DHCP_PARRLIST,5,BOOTP_SUBNET,BOOTP_GATEWAY,BOOTP_DNS,
				BOOTP_HOSTNAME, BOOTP_DOMAIN_SUFFIX};
#endif

			memcpy(cp,par_req_list,sizeof(par_req_list));
			cp += sizeof(par_req_list);
			{
				BYTE *psz;
				int iLen = strlen((char *)hostname);
				if (iLen) {
					psz = hostname;
					*cp++ = BOOTP_HOSTNAME;
					*cp++ = iLen;
					while (iLen--)
						*cp++ = *psz++;
				}
			}
		}
		if ((mtype == DHCPREQUEST && ifp->dhcp_state == D_SELECTING)
			|| mtype == DHCPDECLINE) {
			*cp++ = DHCP_SERVERID;
			*cp++ = 4;
			cp = put32(cp,ifp->dhcp_server);
			*cp++ = DHCP_REQIPA;
			*cp++ = 4;
			cp = put32(cp,ifp->TempIP);
		}
		*cp++ = BOOTP_END;
	}
	else
		cp += 60;
//	bp->cnt = cp - bp->data;
	bp->cnt = BOOTP_SIZE;
	if_stat = ifp->stat;
	ifp->stat &= ~IF_DISABLED;
	send_udp(&ifp->lsock, &fsock,0,0,bp,bp->cnt,0,0,ifp);
	ifp->stat |= if_stat & IF_DISABLED;
	return 0;
}

// called from ARP when duplicate IP address is detected
void dhcp_decline(IFACE *ifp)
{
	if (ifp->dhcp_state) {
		TIMER *t = &ifp->t1;

		ifp->ip_addr = 0;
		if (ifp->dhcp_state != D_BOOTP) {
			ifp->dhcp_state = D_DECLINING;
			send_bootp(ifp);
			stop_timer(t);
			ifp->dhcp_state = D_INIT;
		}
		ifp->send_retry = 1;
		set_timer(t,10000);		// wait for 10 seconds
		start_timer(t);
	}
	ifp->stat |= IF_DISABLED;	// disable the interface
}

void dhcp_inform(IFACE *ifp)
{
	TIMER *t = &ifp->t1;

	ifp->flags |= DHCPI_FLAG;
	if (ifp->ip_addr == 0) {
		return;
	}
	if (ifp->dhcp_state)
		return;
	t->func = dhcp_timer;
	t->arg = (char *)ifp;
	ifp->dhcp_state = D_INFORMING;
	ifp->send_retry = 1;
	send_bootp(ifp);
   return;
}

//void dhcp_release(IFACE *ifp)
//{
//}

static int mask2width(DWORD mask)
{
	int width,i;

	width = 0;
	for(i = 31;i >= 0;i--){
		if(!(mask & (1L << i)))
			break;
		width++;
	}
	return width;
}

/* Process BOOTP/DHCP input received from 'interface'. */
#pragma argsused
static void bootp_rx(struct socket *plsock,int cnt,IFACE *ifp)
{
	struct socket fsock;
	MBUF   	*bp,*bp_file,*bp_sname,*bpo;
	int    	count;
	DWORD 	gateway = 0;
	DWORD 	nameserver = 0;
	DWORD 	netmask,time;
	struct bootp reply;
	BYTE	ch,dmtype,opt_overload = 0,set_request = 0;
	BYTE	netmaskFlag = 0,setopt = (ifp->dhcp_state == D_BOOTP),renew = 0;
	BYTE Silent = SilentStartup;
	fsock.address = 0xfffffffful;
	fsock.port = IPPORT_BOOTPS;
#ifdef DEBUG
	static int xcount;

	if (xcount > 6) {
//		free_p(bp);			// BOOTP datagram MUST be big enough
//		printf("BOOTP size\r\n");
		return;
	}
	++xcount;
	printf("DHCP: Rx - ");
#endif
	ifp->lsock.address = 0;
	if (recv_udp(&ifp->lsock, &fsock, &bp) < 0) {
		if (net_error != WOULDBLK) {
			printf("BOOTP/DHCP: Net_error %d, no values set\n",net_error);
			ifp->ip_addr = 1;
		}
		return;
	}
	if (ifp->dhcp_state == D_BOOTP && len_mbuf(bp) < sizeof(struct bootp)) {
		free_p(bp);			// BOOTP datagram MUST be big enough
#ifdef DEBUG
		printf("BOOTP size\r\n");
#endif
		return;
	}
	dup_p(&bp_file,bp,reply.file - &reply.op,128,AllocBootpFile);
	dup_p(&bp_sname,bp,reply.sname - &reply.op,64,AllocBootpSname);
	ntoh_bootp(&bp, &reply);
	dup_p(&bpo,bp,0,len_mbuf(bp),AllocBootpMbuf);
	if(reply.op != BOOTREPLY || reply.xid != ifp->xid) {
#ifdef DEBUG
		printf("DHCP err: op %d xid %lu expected %lu\r\n",reply.op,reply.xid,ifp->xid);
#endif
		goto cleanup;
	}
	/*	use this to force time-outs on RENEWING and REBINDING
	if (ifp->dhcp_state >= D_RENEWING)
		goto cleanup;
	*/
	if (ifp->send_retry) {
		stop_timer(&ifp->t1);
		ifp->send_retry = 0;
	}
	/* now process the options block, check for cookie first. */
	if(get32(reply.vend) != 0x63825363L){
		printf("Invalid magic cookie.\n");
		goto cleanup;
	}
	while (bpo) {
		if ((ch = pullchar(&bpo)) == BOOTP_PAD)
			continue;
		else if (ch == BOOTP_END) {
			if (opt_overload & OVERLOAD_FILE) {
				opt_overload &= ~OVERLOAD_FILE;
				free_p(bpo);
				bpo = bp_file;
				bp_file = 0;
				continue;
			}
			if (opt_overload & OVERLOAD_SNAME) {
				opt_overload &= ~OVERLOAD_SNAME;
				free_p(bpo);
				bpo = bp_sname;
				bp_sname = 0;
				continue;
			}
			break;
		}
		count = pullchar(&bpo);
		switch (ch) {
		case BOOTP_SUBNET:	/* fixed length, 4 octets */
			netmask = pull32(&bpo);
			netmaskFlag = 1;
			if (setopt) {
				ifp->netbits = mask2width(netmask);
				if(!Silent)
					printf("  Subnet mask: %s\n", inet_ntoa(netmask));
			}
			break;
		case BOOTP_HOSTNAME:
			pullup(&bpo,setopt ? hostname : 0,count);
			break;
		case BOOTP_DNS:
			while (count > 0) {
				nameserver = pull32(&bpo);
				if (setopt && !renew) {
					add_dns_address(nameserver,0);
					if(!Silent)
						printf("  Nameserver: %s\n", inet_ntoa(nameserver));
				}
				count -= 4;
			}
			break;
		case BOOTP_GATEWAY:
			gateway = pull32(&bpo);
			count -= 4;
			/* Add the gateway as the default */
			if (setopt) {
				if (renew) {		// 99-12-10 for Creda Press
					if (gateway) {
						rt_drop(0,0);
						rt_add(0,0,gateway,1,ifp,0,RT_STATIC,0);
					}
				}
				else
					rt_add(0,0,gateway,1,ifp,0,RT_STATIC,0);
				if(!Silent)
					printf("  Default gateway: %s\n", inet_ntoa(gateway));
			}

			//
			// remove any remaining specified gateways as we only support one gateway and we will use
			// the first gateway received as the primary gateway.
			//
			// TODO:  a better solution might be to figure out which gateway has the quickest response time
			//        and to use that gateway.
			//
			while (count > 0)
			{
				pull32(&bpo);
				count -= 4;
			}
			break;
		case BOOTP_DOMAIN_SUFFIX:
			if (dns_suffix)
				free(dns_suffix);
			dns_suffix = malloc((count+1), AllocDnsSufx);
			if ( dns_suffix != 0)
				{
				pullup(&bpo, dns_suffix, count);
				printf("  Domain Suffix: '%s'\n", dns_suffix);
				}
			break;
#ifdef DHCPWWW
		case DHCP_WWW:		// 8.17. Default World Wide Web (WWW) Server Option
			if (setopt && !renew) {
				stats.http_addr = pull32(&bpo);
				if(!Silent)
					printf("  WWW server: %s\n", inet_ntoa(stats.http_addr));
			}
			break;
#endif
		case DHCP_SERVERID:
			if (set_request || setopt)
				ifp->dhcp_server = pull32(&bpo);
			else
				pull32(&bpo);
			break;
		case DHCP_MESTYPE:
			dmtype = pullchar(&bpo);
#ifdef DEBUG
			printf("Receive %s(%u)\n",mes_type[dmtype],ifp->dhcp_state);
#endif
			switch (ifp->dhcp_state) {
			case D_SELECTING:
				if (dmtype != DHCPOFFER)
					goto cleanup;			// ignore silently
				ifp->TempIP = reply.yiaddr;
				set_request = 1;
				break;
			case D_RENEWING:
			case D_REBINDING:
				renew = 1;
				Silent = 1; 		// don't show renewing
				// fall-thru
			case D_REQUESTING:
				if (dmtype == DHCPNAK) {
					ifp->dhcp_state = D_INIT;
					ifp->send_retry = 1;
					ifp->stat |= IF_DISABLED;
					ifp->ip_addr = 0;
					send_bootp(ifp);
					ifp->dhcp_state = D_SELECTING;
					goto cleanup;
				}
				if (dmtype != DHCPACK)
					goto cleanup;
				ifp->dhcp_state = D_BOUND;
				ifp->renew = ifp->rebind = 0;	// for Linux server
				setopt = 1;
				break;
			case D_INFORMING:
				ifp->dhcp_state = D_NONE;
				setopt = 1;
				break;
			}
			break;
		case DHCP_LEASETIME:	// 9.2. IP Address Lease Time
			time = pull32(&bpo);
			if (setopt) {
				ifp->lease = time;
			}
			break;
		case DHCP_RENEWTIME:	// 9.11. Renewal (T1) Time Value
			time = pull32(&bpo);
			if (setopt) {
				ifp->renew = time;
			}
			break;
		case DHCP_REBINDTIME:	// 9.12. Rebinding (T2) Time Value
			time = pull32(&bpo);
			if (setopt) {
				ifp->rebind = time;
			}
			break;
		case DHCP_OPTOVER:		// 9.3. Option Overload
			opt_overload = pullchar(&bpo);
			break;
		default:		/* variable field we don't know about */
			pullup(&bpo,0,count);
			break;
		}
	}
	if (set_request) {
		ifp->send_retry = 1;
		send_bootp(ifp);
		ifp->dhcp_state = D_REQUESTING;
	}
	if (setopt) {
		if (!netmaskFlag) {
			if ((ifp->ip_addr & 0x80000000l) == 0)	/* class A network */
				ifp->netbits = 8;
			else if ((ifp->ip_addr & 0xc0000000l) == 0x80000000l)/* B */
				ifp->netbits = 16;
			else if ((ifp->ip_addr & 0xe0000000l) == 0xc0000000l)/* C */
				ifp->netbits = 24;
			else
				ifp->netbits = 32;
		}
		if (ifp->ip_addr == 0) {
			ifp->ip_addr =  reply.yiaddr;
			if(!Silent)
				printf("  IP address: %s\n",inet_ntoa(ifp->ip_addr));
		}
		ifp->stat &= ~IF_DISABLED;
		arp_init(ifp,ifp->hwtype,EADDR_LEN,IP_TYPE,ARP_TYPE,ether_bdcst,
							pether,gaether);
		if (!Silent) {
			printf("Interface %s configured:\n", ifp->name);
		}
		if (!renew)
			rt_add(ifp->ip_addr & NETMASK(ifp->netbits),ifp->netbits,0,1,ifp,0,RT_STATIC,0);
		if (!ifp->rebind)
			ifp->rebind = (7 * ifp->lease) / 8;
		if (!ifp->renew)
			ifp->renew = ifp->lease / 2;
		if (ifp->dhcp_state) {		// skip if ACK on DHCPINFORM
			set_timer(&ifp->t1,ifp->renew * 1000L);
			start_timer(&ifp->t1);
			ifp->lease -= ifp->rebind;
			ifp->lease *= 1000;
			ifp->rebind -= ifp->renew;
			ifp->rebind *= 1000;
		}
		del_udp(&ifp->lsock);	// won't need this for quite some time
		ifp->lsock.port = 0;
	}
	cleanup:
	free_p(bpo);
	free_p(bp_sname);
	free_p(bp_file);
	free_p(bp);
	return;
}

static void ntoh_bootp(struct mbuf **bpp,struct bootp *bootpp)
{
	bootpp->op = pullchar(bpp);    	/* op */
	bootpp->htype = pullchar(bpp);	/* htype */
	bootpp->hlen = pullchar(bpp);		/* hlen */
	bootpp->hops = pullchar(bpp);		/* hops */
	bootpp->xid = pull32(bpp);			/* xid */
	bootpp->secs = pull16(bpp);		/* secs */
	bootpp->flags = pull16(bpp);		/* flags */
	bootpp->ciaddr = pull32(bpp);		/* ciaddr */
	bootpp->yiaddr = pull32(bpp);		/* ciaddr */
	bootpp->siaddr = pull32(bpp);		/* siaddr */
	bootpp->giaddr = pull32(bpp);		/* giaddr */
	pullup(bpp, bootpp->chaddr, 16);	/* chaddr */
	pullup(bpp, bootpp->sname, 64);	/* sname */
	pullup(bpp, bootpp->file, 128);	/* file name */
	pullup(bpp, bootpp->vend, 4);		/* vendor - only get magic cookie */
}
