//
//   $Id: DNS.C 1.6 2002/10/16 16:42:10Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// DNS.C
// Domain Name Service
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// 2002-07-08 Fix bug that only last server in list is used
// 2002-07-31 Changes for PPP DNS

#if APL==0
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "internet.h"
#include "timer.h"
#include "netfunc.h"
#include "iface.h"
#include "arp.h"
#include "ip.h"
#include "udp.h"
#include "cmdparse.h"
#include "get16.h"
#include "dns.h"

#define DNS_PORT 	53
#define DNS_LPORT	1043

int dodnssuffix(int argc,char *argv[],void *p);	// set/list dns retry
int dodnstime(int argc,char *argv[],void *p);	// set/list dns retry
int dodnsretry(int argc,char *argv[],void *p);	// set/list dns retry
int dodnsserv(int argc,char *argv[],void *p);	// set/list dns server(s)
void send_dns_req(char *name,int32 ipaddr);
add_dns_name(char *name);
int32 get_dns_response(void);
static int dn_expand(BYTE *msg,BYTE *encoded,BYTE *full,int fullen);

struct scmds far dnscmds[] = {
	{"retry",	dodnsretry},
	{"server",	dodnsserv},
	{"suffix",	dodnssuffix},
	{"time",		dodnstime},
	{"",		NULLFP},
};

extern char badhost[];

typedef struct dnsserv {	// dns server definition
	struct dnsserv *next;
	DWORD ip_addr;
} DNSSERV;

DNSSERV *dns_serv;
int dns_retry = 2;
int dns_open;
int dns_id = 1;
unsigned dns_time = 2000/MSPTICK;// time to wait for dns response in ticks
char *dns_suffix;		// suffix to append when no '.' in name

#define KEYBUFSZ	200

#ifdef INI_FILE
#include "ini.h"

void ini_domain(char *ini_file)	// initialise dns
{
	char szKey[KEYBUFSZ];
	char *pszKey;
	char name[80];
	char *argv[2];
	int len;

	dns_retry = GetPrivateProfileInt("Domain","Retry",dns_retry,ini_file);
	len = GetPrivateProfileString("Domain",NULL,"",szKey,KEYBUFSZ,ini_file);
	if (len && *szKey) for (pszKey = szKey;pszKey < szKey + len;
										pszKey += (strlen(pszKey) + 1)) {
		argv[1] = name;
		if (stricmp(pszKey,"suffix") == 0) {
			GetPrivateProfileString("Domain",pszKey,"",name,80,ini_file);
			dodnssuffix(2,argv,0);
		}
		else if (stricmp(pszKey,"time") == 0) {
			GetPrivateProfileString("Domain",pszKey,"",name,80,ini_file);
			dodnstime(2,argv,0);
		}
		else if (stricmp(pszKey,"Retry") != 0) {
			if (*pszKey == ';')
				continue;
			GetPrivateProfileString("Domain",pszKey,"",name,80,ini_file);
			add_dns_name(name);
		}
	}
}
#endif

dodomain(int argc,char *argv[])
{
	return subcmd(dnscmds,argc,argv,0);
}

#pragma argsused
int dodnsserv(int argc,char *argv[],void *p)	// set/list dns server(s)
{
	DNSSERV *pdns;

	if (argc < 2) {
		for (pdns = dns_serv;pdns;pdns = pdns->next) {
			printf("%s\n",inet_ntoa(pdns->ip_addr));
		}
		return 0;
	}
	return add_dns_name(argv[1]);
}

add_dns_name(char *name)
{
	int32 addr;

	if ((addr = resolve(name)) == 0) {
		printf(badhost,name);
		return 1;
	}
	return add_dns_address(addr,0);
}

// iPos == 1 for first, 2 for second, anything else for last
int add_dns_address(DWORD addr, int iPos)
{
	DNSSERV *pdns,*dns;

	if ((dns = calloc(1,sizeof(DNSSERV),AllocDnsServ)) == 0)
		return -1;
	dns->ip_addr = addr;
	if (iPos == 1 || !dns_serv) {
		dns->next = dns_serv;
		dns_serv = dns;
	}
	else if (iPos == 2) {
		dns->next = dns_serv->next;
		dns_serv->next = dns;
	}
	else {
		for (pdns = dns_serv;pdns->next;pdns = pdns->next);
		pdns->next = dns;
	}
	/* Start DNS agent listening at local DNS UDP port */
	if (dns_open <= 0) {
		SOCKET lsock;

		lsock.address = 0;
		lsock.port = DNS_LPORT;
		if ((dns_open = open_udp(&lsock,0 /*dns_rx*/ ,0)) <= 0)
			return -1;
	}
	return 0;
}

int drop_dns_address(DWORD addr, int iPos)
{
	DNSSERV *pdns;

	if (!dns_serv)
		return -1;
	if (iPos == 1) {
		if (dns_serv->ip_addr != addr)
			return -1;
		pdns = dns_serv;
		dns_serv = pdns->next;
	}
	else if (iPos == 2) {
		if ((pdns = dns_serv->next) == 0 || pdns->ip_addr != addr)
			return -1;
		dns_serv->next = pdns->next;
	}
	else
		return -1;
	free(pdns);
	return 0;
}

DWORD GetDnsAddress(int iNum)
{
	DNSSERV *pdns;

	if ((pdns = dns_serv) != 0) {
		if (iNum == 0 || (pdns = dns_serv->next) != 0)
			return pdns->ip_addr;
	}
	return 0;
}

#pragma argsused
int dodnsretry(int argc,char *argv[],void *p)	// set/list dns retry
{
	if(argc < 2)
		printf("%u\n",dns_retry);
	else
		dns_retry = atoi(argv[1]);
	return 0;
}

#pragma argsused
int dodnssuffix(int argc,char *argv[],void *p)	// set/list dns retry
{
	if(argc < 2)
		printf("%s\n",dns_suffix);
	else {
		if (dns_suffix)
			free(dns_suffix);
		if ((dns_suffix = malloc(strlen(argv[1]),AllocDnsSufx)) == 0)
			return -1;
		strcpy(dns_suffix,argv[1]);
	}
	return 0;
}

#pragma argsused
int dodnstime(int argc,char *argv[],void *p)	// set/list dns retry
{
	if(argc < 2)
		printf("%lu ms\n",(long)dns_time * MSPTICK);
	else
		dns_time = (unsigned)(atol(argv[1]) / MSPTICK);
	return 0;
}

// for CAPI
int get_dns_servers(char far *buf, int size)
{
	DNSSERV *pdns;
	int i;

	for (i = 0,pdns = dns_serv;pdns && i < size;i += 4) {
		*buf++ = ((char *)&pdns->ip_addr)[3];
		*buf++ = ((char *)&pdns->ip_addr)[2];
		*buf++ = ((char *)&pdns->ip_addr)[1];
		*buf++ = ((char *)&pdns->ip_addr)[0];
		pdns = pdns->next;
	}
	return i;
}

DWORD resolve_dns(char *name)
{
	DNSSERV *pdns;
	int i;
	int32 addr = 0;
	char namesuf[80];
	char *namex;

	if (dns_suffix && strchr(name,'.') == 0) {
		sprintf(namesuf,"%s.%s",name,dns_suffix);
		namex = namesuf;
	}
	else
		namex = name;
	for (i = 0;i < dns_retry;++i) {
		for (pdns = dns_serv;pdns;pdns = pdns->next) {
			send_dns_req(namex,pdns->ip_addr);
			if ((addr = get_dns_response()) != 0)
				return addr;
		}
	}
	return addr;
}

void send_dns_req(char *name,int32 ipaddr)
{
	int dlen,len;
	MBUF *bp;
	BYTE *cp,*cp1;
	struct socket lsock,fsock;

	// build request buffer
	dlen = strlen(name);
	if ((bp = alloc_mbuf(dlen + 18,AllocDnsSnd)) == 0)
		return;
	cp = bp->data;
	memset(cp,0,dlen + 18);
	put16(cp,/*++*/dns_id);
	cp[2] = cp[5] = 1;
	cp += 12;		// point to name

	while (1) {
		cp1 = (BYTE *)strchr(name,'.');
		if (cp1)
			len = cp1 - (BYTE *)name;// More to come
		else
			len = dlen;	// Last component
		*cp++ = len;
		if (len == 0)
			break;
		strncpy((char *)cp,name,len);
		cp += len;
		if (!cp1) {
			*cp++ = 0;	// Last one
			break;
		}
		name += len + 1;
		dlen -= len + 1;
	}
	cp[1] = cp[3] = 1;
	bp->cnt = cp - bp->data + 4;
	lsock.address = 0;
	lsock.port = DNS_LPORT;
	fsock.address = ipaddr;
	fsock.port = DNS_PORT;
	send_udp(&lsock,&fsock,0,0,bp,bp->cnt,0,0,0);
}

int32 get_dns_response()
{
	char name[64];
	MBUF *bp;
	SOCKET lsock,fsock;
	BYTE *msg,*cp;
	extern unsigned long master_tick;
	char pulled;
	int len,dlen;
	int32 result;
	unsigned long target_tick = master_tick + dns_time;
	int id,qdcount,ancount,type,class,i;

	lsock.address = 0;
	lsock.port = DNS_LPORT;
	while ((long)(target_tick - master_tick) > 0) {
		enable();		// better do it incase API entered with ints disabled
		schedule();
		/* receive the DNS datagram */
		if (recv_udp(&lsock,&fsock,&bp) <= 0)
			continue;
		if (bp->next == 0) {
			len = bp->cnt;
			msg = bp->data;
			pulled = 0;
		}
		else {
			len = len_mbuf(bp);
			if ((msg = malloc(len,AllocDnsMsg)) == 0) {
				free_p(bp);
				break;
			}
			pullup(&bp,msg,len);
			pulled = 1;
		}
		id = get16(msg + 0);
		if (id != dns_id)
			goto fail;
		++dns_id;
//		flags = get16(msg + 2);
		qdcount = get16(msg + 4);
		ancount = get16(msg + 6);
//		nscount = get16(msg + 8);
//		arcount = get16(msg + 10);
		cp = msg + 12;
		for (i = 0;i < qdcount;i++) {
			if ((len = dn_expand(msg,cp,(BYTE *)name,40)) < 0)
				goto fail;
			cp += len;
			type = get16(cp);
			class = get16(cp + 2);
			cp += 4;
			if (type != 1 || class != 1)
				goto fail;
		}
		for (i = 0;i < ancount;i++) {
			if ((len = dn_expand(msg,cp,(BYTE *)name,40)) < 0)
				goto fail;
			cp += len;
			type = get16(cp);
			class = get16(cp + 2);
			dlen = get16(cp + 8);
			cp += 10;
			if (type == 1 && class == 1 && dlen == 4) {
				result = get32(cp);
				if (pulled) {
					free(msg);
				}
				free_p(bp);
				return result;
			}
			cp += dlen;	// ignore this RR
		}
		fail:
		if (pulled) {
			free(msg);
		}
		free_p(bp);
	}
	return 0;
}

// Convert encoded domain name to text
static int dn_expand(
BYTE *msg,		/* Complete domain message */
BYTE *encoded,		/* Pointer to encoded name */
BYTE *full,		/* Pointer to result buffer */
int fullen)		/* Length of same */
{
	unsigned int slen;	/* Length of current segment */
	register BYTE *cp;
	int clen = 0;	/* Total length of encoded name */
	int indirect = 0;	/* Set if indirection encountered */
	int nseg = 0;		/* Total number of segments in name */

	cp = encoded;
	for(;;){
		slen = uchar(*cp++);	/* Length of this segment */
		if(!indirect)
			clen++;
		if((slen & 0xc0) == 0xc0){
			if(!indirect)
				clen++;
			indirect = 1;
			/* Follow indirection */
			cp = msg + ((slen & 0x3f)<<8) + uchar(*cp);
			slen = uchar(*cp++);
		}
		if(slen == 0)	/* zero length == all done */
			break;
		fullen -= slen + 1;
		if(fullen < 0)
			return -1;
		if(!indirect)
			clen += slen;
		while(slen-- != 0)
			*full++ = *cp++;
		*full++ = '.';
		nseg++;
	}
	if(nseg == 0){
		/* Root name; represent as single dot */
		*full++ = '.';
	}
	*full++ = '\0';
	return clen;	/* Length of encoded message */
}
