//
//   $Id: ICMP.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// ICMP.C
// Internet Control Message Protocol implementation
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include "inetdefs.h"
#include "mbuf.h"
#include "internet.h"
#include "timer.h"
#include "iface.h"
#include "ip.h"
#include "icmp.h"
#include "ping.h"

struct icmp_errors icmp_errors;
struct icmp_stats icmp_stats;

extern char dont_time;	   // set if iface timer should not be started

/* Process an incoming ICMP packet */
#pragma argsused
void icmp_input(IFACE *ifp,
MBUF *bp,		/* Pointer to ICMP message */
char protocol,	/* Should always be ICMP_PTCL */
DWORD source,	/* Sender of ICMP message */
DWORD dest,		/* Us */
char tos,		/* Type of Service */
WORD length,	/* Length of ICMP message */
char rxbroadcast)
{
	struct mbuf *htonicmp();
	struct icmp icmp;	/* ICMP header */
	struct ip ip;		/* Offending datagram header */
	WORD type;		/* Type of ICMP message */
	char stat;

	if(rxbroadcast){
		/* Broadcast ICMP packets are to be IGNORED !! */
		icmp_errors.bdcsts++;
		free_p(bp);
		return;
	}
	if(cksum(NULLHEADER,bp,length) != 0){
		/* Bad ICMP checksum; discard */
		icmp_errors.checksum++;
		free_p(bp);
		return;
	}
	ntohicmp(&icmp,&bp);

	/* Process the message. Some messages are passed up to the protocol
	 * module for handling, others are handled here.
	 */
	type = icmp.type;
	if(type < ICMP_TYPES)
		icmp_stats.input[type]++;

	switch(uchar(type)){
	case TIME_EXCEED:	/* Time-to-live Exceeded */
	case DEST_UNREACH:	/* Destination Unreachable */
	case QUENCH:		/* Source Quench */
		ntohip(&ip,&bp);	/* Extract offending IP header */
		switch(uchar(ip.protocol)){
		case TCP_PTCL:
			tcp_icmp(ip.source,ip.dest,icmp.type,icmp.code,&bp);
			break;
		}
		break;
	case ECHO:		/* Echo Request */
		/* Change type to ECHO_REPLY, recompute checksum,
		 * and return datagram.
		 */
		icmp.type = ECHO_REPLY;
#if APL
		if (icmp.args.echo.id == SILENT)
			ping_log(source,bp);
#endif
		bp = htonicmp(&icmp,bp);
		icmp_stats.output[ECHO_REPLY]++;
		dont_time = 1;
		if (ifp) {
			stat = ifp->stat;
			if (icmp.args.echo.id == QUERY)
				ifp->stat |= IF_UP;
		}
		ip_send(dest,source,ICMP_PTCL,tos,0,bp,length,0,0,0);
		if (ifp)
			ifp->stat = stat;
		dont_time = 0;
		if (icmp.args.echo.id == 1234)
			pingem_solicited(source);
		return;
	case REDIRECT:		/* Redirect */
	case PARAM_PROB:	/* Parameter Problem */
		break;
	   case ECHO_REPLY:		/* Echo Reply */
		echo_proc(ifp,source,dest,&icmp);
		break;
	case TIMESTAMP:		/* Timestamp */
	case TIME_REPLY:	/* Timestamp Reply */
	case INFO_RQST:		/* Information Request */
	case INFO_REPLY:	/* Information Reply */
		break;
	}
	free_p(bp);
}
/* Return an ICMP response to the sender of a datagram */
void icmp_output(struct ip *ip,struct mbuf *bp,char type,char code,union icmp_args *args)
//struct ip *ip;		/* Header of offending datagram */
//struct mbuf *bp;		/* Data portion of datagram */
//char type,code;		/* Codes to send */
//union icmp_args *args;
{
	struct mbuf *htonicmp();
	struct mbuf *htonip();
	struct icmp icmp;	/* ICMP protocol header */
	WORD dlen;		/* Length of data portion of offending pkt */
	WORD length;		/* Total length of reply */
	struct mbuf *bpd;

	if(ip == NULLIP)
		return;
	if(type < ICMP_TYPES)
		icmp_stats.output[type]++;

	if(uchar(ip->protocol) == ICMP_PTCL){
		/* Never send an ICMP message about another ICMP message
		   except for echo request */
		if (dup_p(&bpd,bp,0,len_mbuf(bp),AllocIcmpBuf) == 0) {
			icmp_errors.noloop++;
			return;
		}
		ntohicmp(&icmp,&bpd);
		free_p(bpd);
		if (icmp.type != ECHO) {
			icmp_errors.noloop++;
			return;
		}
	}
	/* Compute amount of original datagram to return.
	 * We return the original IP header, and up to 8 bytes past that.
	 */
	dlen = min(8,len_mbuf(bp));
	length = dlen + ICMPLEN + IPLEN + ip->optlen;
	if(bp != NULLBUF){
		/* Take excerpt from data portion */
		bp = copy_p(bp,dlen,AllocIcmpData);
	}
	/* Recreate and tack on offending IP header */
	bp = htonip(ip,bp,0);

	icmp.type = type;
	icmp.code = code;
	switch(uchar(icmp.type)){
	case PARAM_PROB:
		icmp.args.pointer = args->pointer;
		break;
	case REDIRECT:
		icmp.args.address = args->address;
		break;
	case ECHO:
	case ECHO_REPLY:
	case INFO_RQST:
	case INFO_REPLY:
	case TIMESTAMP:
	case TIME_REPLY:
		icmp.args.echo.id = args->echo.id;
		icmp.args.echo.seq = args->echo.seq;
		break;
	default:
		icmp.args.unused = 0;
		break;
	}
	bp = htonicmp(&icmp,bp);	/* Now stick on the ICMP header */

	ip_send(0,ip->source,ICMP_PTCL,ip->tos,0,bp,length,0,0,0);
}
/* Generate ICMP header in network byte order, link data, compute checksum */
struct mbuf *htonicmp(struct icmp *icmp,struct mbuf *data)
{
	struct mbuf *rval;
	register BYTE *cp;
	WORD checksum;

	rval = alloc_mbuf(ICMPLEN,AllocIcmpSnd);
	rval->cnt = ICMPLEN;
	cp = rval->data;

	*cp++ = icmp->type;
	*cp++ = icmp->code;
	cp = put16(cp,0);		/* Clear checksum */
	cp = put16(cp,icmp->args.echo.id);
	cp = put16(cp,icmp->args.echo.seq);

	/* Link in data, compute checksum, and stash result */
	rval->next = data;
	checksum = cksum(NULLHEADER,rval,len_mbuf(rval));
	cp = &rval->data[2];
	cp = put16(cp,checksum);

	return rval;
}
/* Pull off ICMP header */
void ntohicmp(struct icmp *icmp,struct mbuf **bpp)
{
	icmp->type = pullchar(bpp);
	icmp->code = pullchar(bpp);
	(void) pull16(bpp);		/* Toss checksum */
	icmp->args.echo.id = pull16(bpp);
	icmp->args.echo.seq = pull16(bpp);
}
