//
//   $Id: IGMP.C 1.4 2003/09/09 17:28:24Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// IGMP.C
// Internet Group Management Protocol version 1 implementation
// Copyright (c) GP van Niekerk Ondernemings 1999-2003
// Revision History
// 09/09/2003 gvn return proper error code in JoinGroup

#include <stdlib.h>
#include <dos.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "internet.h"
#include "timer.h"
#include "iface.h"
#include "ip.h"
#include "igmp.h"
#include "tcp.h"
#include "_capi.h"

GROUP_AD *psGroupAdFirst;	// pointer to chain of group addresses

IFACE *GetMCInterface(DWORD dwIFAddr)
{
	IFACE *psIface;

	if (dwIFAddr) {
		for (psIface = ifaces;psIface;psIface = psIface->next) {
			if (psIface->ip_addr == dwIFAddr)
				return psIface;
		}
	}
	return ifaces;
}

GROUP_AD *FindGroupStruct(DWORD dwGroupAd,IFACE *psIface,
																GROUP_AD **ppsGroupPrev)
{
	GROUP_AD *psGroupAd,*psGroupPrev = 0;

	for (psGroupAd = psGroupAdFirst;psGroupAd;psGroupAd = psGroupAd->psNext)
	{
		if (psGroupAd->dwGroupAddr == dwGroupAd &&
												psGroupAd->psIface == psIface) {
			break;
		}
		psGroupPrev = psGroupAd;
	}
	*ppsGroupPrev = psGroupPrev;
	return psGroupAd;
}

static char argv[] = {0, IOCTL_SETMULTICAST};
static char *pargv[1] = {argv};

void IgmpTimeout(char *p)
{
	GROUP_AD *psGroupAd = (GROUP_AD *)p;

	psGroupAd->bState = IGMP_IDLE;
	igmp_output(psGroupAd->dwGroupAddr,IGMP_REPORT1,0,psGroupAd->dwGroupAddr);
}

int JoinGroup(DWORD dwGroupAddr,DWORD dwIFAddr)
{
	int iRet = 0;
	IFACE *psIface;
	GROUP_AD *psGroupAd, *psGroupPrev;

	if (dwGroupAddr == 0 && dwIFAddr == 0) {	// special test to generate query
		igmp_output(ALL_SYSTEMS,IGMP_QUERY,0,0);
		return 0;
	}
	psIface = GetMCInterface(dwIFAddr);
	psGroupAd = FindGroupStruct(dwGroupAddr,psIface,&psGroupPrev);
	if (!psGroupAd) {
		if ((psGroupAd = calloc(sizeof(GROUP_AD),1,AllocIgmpGa)) == 0) {
			return NO_SPACE;
		}
		if (psGroupAdFirst)
			psGroupAd->psNext = psGroupAdFirst;
		psGroupAdFirst = psGroupAd;
		psGroupAd->dwGroupAddr = dwGroupAddr;
		psGroupAd->psIface = psIface;
		psGroupAd->sTimer.func = IgmpTimeout;
		psGroupAd->sTimer.arg = (char *)psGroupAd;
		srand((WORD)psIface->ip_addr + (WORD)peek(0x40,0x6c));
		set_timer(&psGroupAd->sTimer,random(10000));
		start_timer(&psGroupAd->sTimer);
		psGroupAd->bState = IGMP_DELAYING;
		igmp_output(dwGroupAddr,IGMP_REPORT1,0,dwGroupAddr);
		if ((*psIface->ioctl)(psIface,1,pargv) != 0)
			iRet = ErrIllegalOp;
	}
	++psGroupAd->wUseCount;
	return iRet;
}

int LeaveGroup(DWORD dwGroupAddr,DWORD dwIFAddr)
{
	int iRet = 0;
	IFACE *psIface = GetMCInterface(dwIFAddr);
	GROUP_AD *psGroupPrev;
	GROUP_AD *psGroupAd = FindGroupStruct(dwGroupAddr,psIface,&psGroupPrev);

	if (!psGroupAd) {
		return ErrIllegalOp;
	}
	if (--psGroupAd->wUseCount == 0) {
		// unlink group address structure
		if (psGroupPrev)
			psGroupPrev->psNext = psGroupAd->psNext;
		else
			psGroupAdFirst = psGroupAd->psNext;
		if (run_timer(&psGroupAd->sTimer))
			stop_timer(&psGroupAd->sTimer);
		free(psGroupAd);
		if ((*psIface->ioctl)(psIface,1,pargv) != 0)
			iRet = ErrIllegalOp;
	}
	return iRet;
}

// Process an incoming IGMP datagram
#pragma argsused
void igmp_input(IFACE *ifp,
MBUF *pmb,		// Pointer to IGMP message
BYTE protocol,	// Should always be IGMP_PTCL
DWORD source,	// Sender of IGMP message
DWORD dest,		// Us
char tos,		// Type of Service
WORD length,	// Length of IGMP message
char rxbroadcast)
{
	GROUP_AD *psGroupAd;
	BYTE bType;
//	BYTE bMaxResp;
	DWORD dwGroupAddr;

	if(length < 8 || cksum(NULLHEADER,pmb,length) != 0){
		// Bad IGMP datagram; discard
		free_p(pmb);
		return;
	}
	bType = pullchar(&pmb);
	/*bMaxResp = */pullchar(&pmb);
	pull16(&pmb);		// discard checksum
	dwGroupAddr	= pull32(&pmb);

	// Process the message
#ifdef DEBUG
	printf("IGMP: To %s type 0x%x resp %d GrpAddr ",inet_ntoa(dest),
						bType,bMaxResp);
	printf("%s\n",inet_ntoa(dwGroupAddr));
#endif
	switch (bType) {
	case IGMP_QUERY:
		if (dest != ALL_SYSTEMS)
			break;		// invalid query
		for (psGroupAd = psGroupAdFirst;psGroupAd;psGroupAd = psGroupAd->psNext)
		{
			if (psGroupAd->psIface == ifp && psGroupAd->bState == IGMP_IDLE) {
				// query received event
				srand((WORD)ifp->ip_addr + (WORD)peek(0x40,0x6c));
				set_timer(&psGroupAd->sTimer,random(10000));
				start_timer(&psGroupAd->sTimer);
				psGroupAd->bState = IGMP_DELAYING;
			}
		}
		break;
	case IGMP_REPORT1:
		if (dwGroupAddr != dest)
			break;		// invalid report
		for (psGroupAd = psGroupAdFirst;psGroupAd;psGroupAd = psGroupAd->psNext)
		{
			if (psGroupAd->psIface == ifp &&
								dwGroupAddr == psGroupAd->dwGroupAddr &&
								psGroupAd->bState != IGMP_DELAYING) {
				// report received event
				stop_timer(&psGroupAd->sTimer);
				psGroupAd->bState = IGMP_IDLE;
			}
		}
		break;
	}
	free_p(pmb);
}

// Send an IGMP datagram
void igmp_output(DWORD dwDest,BYTE bType,BYTE bMaxResp,DWORD dwGroupAddr)
{
	struct mbuf *pmb;
	BYTE *pb;
	WORD wChecksum;

	if ((pmb = alloc_mbuf(IGMPLEN,AllocIgmpSnd)) == 0)
		return;		// out of mem :(
	pmb->cnt = IGMPLEN;
	pb = pmb->data;

	*pb++ = bType;
	*pb++ = bMaxResp;
	pb = put16(pb,0);		// Clear checksum
	put32(pb,dwGroupAddr);
	wChecksum = cksum(NULLHEADER,pmb,len_mbuf(pmb));
	pb = pmb->data + 2;
	put16(pb,wChecksum);
	wNoMulticastEcho = 1;
	ip_send(0,dwDest,IGMP_PTCL,0,1,pmb,IGMPLEN,0,0,0);
	wNoMulticastEcho = 0;
}

