//
//   $Id: MODEM.C 1.3 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// modem.c
// Modem driver for INET
// Copyright (c) GP van Niekerk Ondernemings 1990-1998

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <string.h>
#include <ctype.h>
#include <alloc.h>
#include "inetdefs.h"
#include "timer.h"
#include "modem.h"

MODEMPAR *modem_par(char *filename)
{
	MODEMPAR *pmpar = 0;
	FILE *fp;
	char str[250];
	char *mp[DIALAR + NUMDIAL];
	char *pp,*ps,*pe;
	char lmem[250];
	char *plm = lmem;
	char ch,flags = 0;
	int time = 10000,retry = 0,modem_char_time = 55;
	int len,llen = 0,indx,numdial = 0,i;
	char com_char = '@';

	memset(mp,0,sizeof(mp));
	if ((fp = fopen((ps = getfilename(str,filename,1)),"rt")) == 0) {
		printf("Open error on '%s'\n",ps);
		return pmpar;
	}
	printf("Modem file: %s\n",ps);
	while (fgets(str,sizeof(str),fp) != 0) {
		str[(len = strlen(str)) - 1] = 0;
		for (ps = str;(pe = strchr(ps,'^')) != 0;ps = pe + 1) {
			if ((ch = pe[1]) == 0)
				break;
			*pe = ch == ' ' ? '^' : (ch & 0x1f);
			strcpy(pe + 1,pe + 2);
		}
		switch(str[0]) {
		case 0:
		case '#':
		case ';':		/* take as comment */
			break;
		case 'a':		// answer script
			indx = ANSWER;
			goto putpar;
		case 'b':		// break character
			com_char = str[2];
			break;
		case 'c':		/* connect command */
			indx = CONNECT;
			goto putpar;
		case 'd':		// dial command
			indx = DIAL;
			goto putpar;
		case 'i':		/* initialize */
			indx = INIT;
			goto putpar;
		case 'm':		/* modem pacing timeout value */
			modem_char_time = atoi(str + 2);
			break;
		case 'n':		/* numbers */
			if (numdial == NUMDIAL) {
				printf("Too many dial numbers:\n%s",str);
				break;
			}
			indx = DIALAR + numdial++;
			goto putpar;
		case 'p':		/* parameters */
			for (ps = str + 2;*ps;ps++) {
				switch(*ps) {
				case 'd':
					flags |= MF_DIAL;	// always dial when DCD is lost
					break;
				case 'n':
					flags |= MF_NODOD;// No dial-on-demand
					break;
				case 'r':
					flags |= MF_DISR;
					break;
				case 'x':
					flags |= MF_DISX;
					break;
				}
			}
			break;
		case 'r':		// retry value
			retry = atoi(str + 2);
			break;
		case 't':		/* timeout value */
			time = atoi(str + 2);
			break;
		case 'x':		// exchange id
			indx = XCHANGE;
			putpar:
			if (mp[indx]) {
				printf("Duplicate command:\n%s",str);
				break;
			}
			if (--len > (sizeof(lmem) - llen)) {
				printf("Too many modem commands\n");
				return 0;
			}
			else {
#ifdef __LARGE__
				*(mp + indx) = plm;
#else
				*(char near **)(mp + indx) = plm;
#endif
				strncpy(plm,str + 2,len);
				plm += len;
				llen += len;
			}
			break;
		default:
			printf("Invalid modem command:\n%s",str);
			break;
		}
	}
	if ((pp = malloc(llen,AllocModCom)) == 0) {
		printf("Out of memory\n");
		return pmpar;
	}
	memcpy(pp,lmem,llen);
	for (i = 0;i < DIALAR + NUMDIAL;++i) {
		if ((unsigned)mp[i] > 1)
			mp[i] = mp[i] - lmem + pp;
	}
	if ((pmpar = calloc(1,(len = sizeof(char *) * (numdial + 1)) +
						sizeof(MODEMPAR),AllocModPar)) == 0) {
		printf("Out of memory\n");
		return pmpar;
	}
	len += (sizeof(char *) * DIALAR);
	pmpar->com_char = com_char;
	pmpar->flags = flags;
	pmpar->time = time;
	pmpar->modem_char_time = modem_char_time;
	pmpar->retry_val = retry;
	pmpar->fdial = pmpar->dialar;
	memcpy(&pmpar->init,mp,len);
	fclose(fp);
	return pmpar;
}

#if APL
char far modem_states[] = {"IDLEINITDIALCONNANSWRCV XID"};
//							^	^	^	^	^	^	^
char far timer_states[] = {"IDLEWDCDWTO WRCVWXIDWADDPACE"};
//							^	^	^	^	^	^	^   ^

void dial_status(MODEMPAR *pm)
{
	if (!pm) {
		printf("No dial parameters\n");
		return;
	}
	printf("Dial state:%u %4.4Fs\n",pm->state,modem_states + 4 * pm->state);
	if (pm->timer.state == TIMER_RUN) {
		printf("Timer running (%ld/%ld ms)\n",
		 (long)MSPTICK * (pm->timer.start - pm->timer.count),
		 (long)MSPTICK * pm->timer.start);
		printf("Timer state:%u %4.4Fs\n",pm->tmstate,timer_states + 4 * pm->tmstate);
//		printf("Hangup time: %ld\n",(long)MSPTICK * pm->time);
	}
	printf("Flags %x %s%s%s\n",pm->flags,
				pm->flags & MF_INIT ? "Initialised " : "",
				pm->flags & MF_ACKSEEN ? "XID acked " : "",
				pm->flags & MF_DISCON ? "Disconnected" : "");
	printf("retry %u/%u\n",pm->retry,pm->retry_val);
	if (pm->dial && *pm->dial)
		printf("Current dial number %s\n",*pm->dial);
	if (pm->nextptr)
		printf("Next output: %-60.60s\n",pm->nextptr);
	if (pm->xchange)
		printf("XID: %s\n",pm->xchange);
}
#endif

/* code to allow more than 250 chars in modem definition
   adapt for small model
				if ((pp = farmalloc(llen + len)) == 0) {
					printf("Out of memory\n");
					exit(1);
				}
				_fmemcpy(pp,lmem,llen);
				_fmemcpy(pp + llen,str + 2,len);
				mp[indx] = pp + llen;
				for (i = 0;i < DIALAR + NUMDIAL;++i) {
					if (FP_SEG(mp[i]) == 0 && FP_OFF(mp[i]))
						mp[i] = mp[i] - lmem + pp;
				}
				plm = lmem;
				llen = 0;
*/
