//
//   $Id: PKTDRV.C 1.7 2003/09/09 17:19:28Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// PKTDRV.C
// Packet driver interface
// Copyright (c) GP van Niekerk Ondernemings 1990-2003

// Revision history
// 09/09/03 Fix IOCTL_SETMULTICAST to return 0
// 08/28/01 Fix pdr_sched to save BP for asy scheduling
// 07/23/01 Increase Pkt recv size by 4 for flakey Cisco PD
// 07/21/01 Allocation tracing

#if !APL
#pragma option -zCDRIVER_CODE
#endif

#include <stdio.h>
#include <alloc.h>
#include <dos.h>
#include <string.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#include "timer.h"
#include "arp.h"
#include "pktdrv.h"
#include "enet.h"
#include "ip.h"
#include "igmp.h"

#if APL
#define   TRACE   1
#include "trace.h"
#endif

typedef char far * fptr;

#define NPDRMAX 4      /* max number of packet drivers */
unsigned npdr = 0;      /* number of PACKET DRIVER adaptors */

#ifndef __LARGE__
unsigned pstack_start;
#endif
BYTE *obuf;            /* output buffer for send */
unsigned obufsize;      // size of output buffer

void far receiver_a0(),receiver_a1(),receiver_a2(),receiver_a3();

void far (*receiver[])() = {
   receiver_a0,receiver_a1,receiver_a2,receiver_a3
};

PDCB pktcb[4];

char hin_recv = 0;

#if !APL
void near pdr_sched(void)
{
   extern unsigned stack_pointer;
   extern int in_tcp;
   extern char must_schedule;
   static WORD savess,savebp;

   if (!in_tcp) {
      asm   push bp;
      in_tcp = 0x81;
      savess = _SS;
      savebp = _SP;
      _SS = _DS;
      _SP = stack_pointer;
      _BP = stack_pointer + 6;
      enable();   // OK to enable in Packet driver receive
      schedule();

      disable();
      _SS = savess;
      _SP = savebp;
      in_tcp = 0;
      asm   pop bp;
   } 
   else
   {
      hin_recv = 1;
      must_schedule = 1;
   }
}

void interrupt pdr_sched0()
{
   (*pktcb[0].oldvec)();
   pdr_sched();
}

void interrupt pdr_sched1()
{
   (*pktcb[1].oldvec)();
   pdr_sched();
}

void interrupt pdr_sched2()
{
   (*pktcb[2].oldvec)();
   pdr_sched();
}

void interrupt pdr_sched3()
{
   (*pktcb[3].oldvec)();
   pdr_sched();
}

void interrupt (*pdr_schedtab[])() = {
   pdr_sched0,pdr_sched1,pdr_sched2,pdr_sched3,
};
#endif

/* PACKET DRIVER IOCTL */
#pragma argsused
pdr_ioctl(struct interface *ifp,int argc,char *argv[])
{
   PDCB *ppdcb = ifp->dcb;
   struct REGPACK regs;
#if APL
   char buf[20];
   PDPAR far *ppdp;
   PDSTAT far *ppds;
#endif
   if (argc == 1 && argv[0][0] == 0) {   // internal commands
      switch (argv[0][1]) {
      case IOCTL_HOTSWAP:
         regs.r_bx = ppdcb->iphandle;
         regs.r_ax = GET_ADDRESS << 8;
         regs.r_cx = EADDR_LEN;
         regs.r_es = FP_SEG(ifp->hwaddr);
			regs.r_di = FP_OFF(ifp->hwaddr);
			intr(ifp->dev,&regs);
			arp_output(ifp,ifp->hwtype,ifp->ip_addr,ifp->ip_addr);   // gratuitous ARP
			return 0;
		case IOCTL_SETMULTICAST:
#ifdef MULTICAST_FILTER
			{
				int iCnt,iMode;
				GROUP_AD *psGroupAd;

				for (psGroupAd = psGroupAdFirst,iCnt = 0;psGroupAd;
											psGroupAd = psGroupAd->psNext) {
					if (psGroupAd->psIface == ifp)
						++iCnt;
				}
				for (iMode = (iCnt ? 5 : 3);iMode < (iCnt ? 7 : 4);iMode++) {
					regs.r_ax = SET_RCV_MODE << 8;
					regs.r_bx = ppdcb->iphandle;
					regs.r_cx = iMode;   // first try multicast, then all
					intr(ifp->dev,&regs);
					if (!(regs.r_flags & CF)) {
						printf("RcvMode set to %d\n",iMode);
						return 0;
					}
					printf("RcvMode %d failed with error %d!\n",iMode,regs.r_dx >> 8);
				}
				return -1;
			}
#else
			return 0;
#endif
			default:
				return -1;
		}
	}
#if APL
	pether(buf,ifp->hwaddr);
	printf("Packet driver parameters\nEthernet address %s\n",buf);
	printf("Frames:received %lu sent %lu\n",
               ppdcb->rcvdcount,ppdcb->sentcount);
   if (ppdcb->packets_dropped)
      printf("Packets dropped %lu\n",ppdcb->packets_dropped);
   regs.r_ax = GET_PARAMETERS << 8;
   intr(ifp->dev,&regs);
   if (!(regs.r_flags & CF)) {
      ppdp = MK_FP(regs.r_es,regs.r_di);
      printf("Rev %u.%02u par_len %u add_len %u mtu %u\n",
         ppdp->major_rev,ppdp->minor_rev,ppdp->length,
         ppdp->addr_len,ppdp->mtu);
      printf("Rcv_bufs %u xmt_bufs %u int_num 0x%x\n",
         ppdp->rcv_bufs,ppdp->xmt_bufs,ppdp->int_num);
   }
   regs.r_ax = GET_STATISTICS << 8;
   regs.r_bx = ppdcb->iphandle;
   intr(ifp->dev,&regs);
   if (!(regs.r_flags & CF)) {
      ppds = MK_FP(regs.r_ds,regs.r_si);
      printf("Packets:in %lu out %lu bytes:in %lu out %lu\n",
         ppds->packets_in,ppds->packets_out,ppds->bytes_in,
                     ppds->bytes_out);
      printf("Errors:in %lu out %lu packets lost %lu\n",
         ppds->errors_in,ppds->errors_out,ppds->packets_lost);
   }
   if (ifp->dhcp_state) {
      printf("DHCP_STATE %u retry %u renew %lu rebind %lu lease %lu\n",
         ifp->dhcp_state,ifp->send_retry,ifp->renew,
         ifp->rebind / 1000,ifp->lease / 1000);
   }
#endif
   return 0;
}

/* Send a raw packet on PACKET DRIVER */
void pdr_raw(struct interface *ifp,struct mbuf *bp)
{
   struct REGPACK regs;
   PDCB *ppdcb = ifp->dcb;

#ifdef TRACE
   if (ifp->trace & IF_TRACE_OUT) {
      dump(ifp,IF_TRACE_OUT,
         ifp->flags & TOKEN_FLAG ? TRACE_TOKEN : TRACE_ETHER,bp);
   }
#endif
   /* send the datagram */
   if ((regs.r_cx = len_mbuf(bp)) < RUNT && !(ifp->flags & TOKEN_FLAG))
      regs.r_cx = RUNT;
   pullup(&bp,obuf,obufsize);
   free_p(bp);
   if (!(ifp->stat & IF_DISABLED)) {
      regs.r_ax = SEND_PKT << 8;
      regs.r_si = FP_OFF(obuf);
      regs.r_ds = FP_SEG(obuf);
      deb3(23," SEND_PDR %04x %d %d",_FLAGS,regs.x.cx);
      intr(ifp->dev,&regs);
      deb2(23+DE,"X",_FLAGS);
      ++ppdcb->sentcount;
   }
}

/* Send an IP datagram on Slip PACKET DRIVER */
#pragma argsused
void slp_send(
struct mbuf *bp,      /* Buffer to send */
struct interface *ifp,   /* Pointer to interface control block */
int32 gateway,         /* IP address of next hop */
char precedence,
char delay,
char throughput,
char reliability)
{
   struct REGPACK regs;
   PDCB *ppdcb = ifp->dcb;

#ifdef TRACE
   if (ifp->trace & IF_TRACE_OUT) {
      dump(ifp,IF_TRACE_OUT,TRACE_IP,bp);
   }
#endif
   /* send the datagram */
   regs.r_cx = len_mbuf(bp);
   pullup(&bp,obuf,obufsize);
   free_p(bp);
   regs.r_ax = SEND_PKT << 8;
   regs.r_si = FP_OFF(obuf);
   regs.r_ds = FP_SEG(obuf);
   intr(ifp->dev,&regs);
   ++ppdcb->sentcount;
}

/* Process any incoming packets */
void pdr_recv(IFACE *ifp)
{
   PDCB *ppdcb = ifp->dcb;
   void arp_input();
   struct mbuf *bp;
   struct ether hdr;
   char multicast;
   int cnt;
   BYTE rif[40];

   for (cnt = 5;cnt;--cnt) {
      idisable();
      if ((bp = dequeue(&ppdcb->rcvq)) == 0) {
         popf();
         break;
      }
      --ppdcb->rcvcnt;
      popf();
      ++ppdcb->rcvdcount;
      /* Remove Ethernet header and kick packet upstairs */
      rif[0] = 0;
      if (ifp->flags & TOKEN_FLAG) {
#ifdef TRACE
         dump(ifp,IF_TRACE_IN,TRACE_TOKEN,bp);
#endif
         ntohtoken(&hdr,&bp,rif);
      }
      else {
#ifdef TRACE
         dump(ifp,IF_TRACE_IN,TRACE_ETHER,bp);
#endif
         ntohether(&hdr,&bp);
      }
      if(hdr.dest[0] & 1) {
         if (memcmp(hdr.source,ifp->hwaddr,EADDR_LEN) == 0)
            goto free_it;
         multicast = 1;
      }
      else
         multicast = 0;
      switch(hdr.type){
      case ARP_TYPE:
         arp_input(ifp,bp,rif[0] ? rif : 0);
         break;
      case IP_TYPE:
         ip_route(bp,multicast,ifp);
         break;
      default:
         free_it:
         free_p(bp);
         break;
      }
   }
   ppdcb->rcvcnt = 0;
}

/* Process any slip incoming packets */
void slp_recv(IFACE *ifp)
{
   PDCB *ppdcb = ifp->dcb;
   struct mbuf *bp;

   while (1) {
      idisable();
      if ((bp = dequeue(&ppdcb->rcvq)) == 0) {
         popf();
         break;
      }
      --ppdcb->rcvcnt;
      popf();
      ++ppdcb->rcvdcount;
#ifdef TRACE
      dump(ifp,IF_TRACE_IN,TRACE_IP,bp);
#endif
      ip_route(bp,0,ifp);
   }
   ppdcb->rcvcnt = 0;
}

void dhcp_release(IFACE *ifp);

/* Packet driver stop routine */
void pdr_stop(IFACE *ifp)
{
   struct REGPACK regs;
   PDCB *ppdcb = ifp->dcb;

//   if (ifp->dhcp_state == D_BOUND)
//      dhcp_release(ifp);
   if (ifp->flags & (ETH_FLAG | TOKEN_FLAG)) {
      regs.r_ax = RELEASE_TYPE << 8;
      regs.r_bx = ppdcb->arphandle;
      intr(ifp->dev,&regs);
   }
   regs.r_ax = RELEASE_TYPE << 8;
   regs.r_bx = ppdcb->iphandle;
   intr(ifp->dev,&regs);
#if !APL
   if (ppdcb->vec) {
      setvect(ppdcb->vec,ppdcb->oldvec);
   }
#endif
}

/* Asyncronous receiver() function */
void far *receiver_c()
{
   extern char bMinPacketSize;
   MBUF *bp;
   int len,flag;
   PDCB *ppdcb;

   flag = _AX;
   len = _CX;

#ifdef __LARGE__
   ppdcb = (PDCB *)MK_FP(_DS,_DI);
#else
   ppdcb = (PDCB *)_DI;
#endif
   if (flag == 0) {      /* request buffer */
      //deb1(24," PDR rcv 0");
      if (ppdcb->buffer) {
         ++ppdcb->packets_dropped;
         //deb1(24+DE,"N");
         return 0;
      }
      if (++ppdcb->rcvcnt < ppdcb->rcvmax) 
      {
         if (bMinPacketSize)
         {
            if (len < 64)
            {
               len = 64;
            }
         }

         if ((bp = alloc_mbuf(len,AllocPktBuf)) != 0) {
            bp->cnt = len;
            ppdcb->buffer = bp;
            //deb1(24+DE,"B");
            return (void far *)bp->data;
         }
      }
      --ppdcb->rcvcnt;
      ++ppdcb->packets_dropped;
      //deb1(24+DE,"M");
      return 0;
   }
   else {            /* buffer filled */
      //deb1(24," PDR rcv 1");
      enqueue(&ppdcb->rcvq,ppdcb->buffer);
      ppdcb->buffer = 0;
   }
   //deb1(24+DE,"D");
   return 0;
}
