//
//   $Id: PPP.C 1.11 2003/08/22 05:37:34Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
Revision history (since 2002)

2002-05-03 gvn changes to try and solve GPRS problems
2002-05-04 gvn Cleanup code, MRU request checks interface mtu, fixed some
			  potential memory leaks
2002-05-10 gvn fix state problem causing GPRS failure
2002-06-18 gvn remove NULL in PAP request
2002-07-30 gvn DNS
2002-08-09 gvn Fix ACCM changing during transmission causing bad LCP frames
2003-08-04 gvn optionally add route to peer
2003-08-21 gvn fix MRU/MTU being set to default when not negotiated

 * PPP.C
 * Copyright (C) 2001 Datalight, Inc.
 * All Rights Reserved
 * Datalight Confidential and Proprietary
 * Portions Copyright (c) GPvNO 1990-2001
 *
 *  Send and receive datagrams on serial lines with
 *	Point-to-Point Protocol
 *
 *	This implementation of PPP is declared to be in the public domain.
 *
 *	Based (in part) upon previous implementations by:
 *	1989	-- Drew Daniel Perkins		(ddp@andrew.cmu.edu)
 *		   Carnegie Mellon University
 *	09-90	-- Katie Stevens		(dkstevens@ucdavis.edu)
 *		   UC Davis, Computing Services
 *
 *	Jan 91	Bill_Simpson@um.cc.umich.edu
 *		Computer Systems Consulting Services
 *
 *	Feb 91	Glenn McGregor			(ghm@merit.edu)
 *		Testing and suggestions.
 *
 *	May 91	Bill Simpson & Glenn McGregor
 *		Update to newest LCP and IPCP draft RFCs.
 *		Add quick installation features.
 *		Add support for echo and discard message sending.
 *
 *	Jul 91	Glenn McGregor & Bill Simpson
 *		Improve PAP user interface and fix related bugs.
 *		Remove kwaits and "phase machine".
 *  Aug 98 Converted for use by INET GP van Niekerk
 * RFC 1661 - PPP
 * RFC 1332 - IPCP
 * RFC 2153 - PPP vendor extensions (for information only)
 * RFC 1877 - DNS (not implemented)
 * RFC 1994 - CHAP (implemented 2001)
 */

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dos.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
//#include "cmdparse.h"
#include "slhc.h"
#include "internet.h"
#include "ip.h"
#if APL
#define TRACE 1
#include "trace.h"
#endif
#include "ppp.h"
#include "get16.h"
#include "modem.h"
#include "ser.h"
#include "userpass.h"
#include "bootp.h"
#include "md5_locl.h"
#include "dns.h"

//#define BUF_TRACE
#ifdef BUF_TRACE
#define BUFTRACE(name,adr) trprintf("ALLOC %10s(%u) %p\n",name,((WORD *)(adr))[-1],adr)
#else
#define BUFTRACE(name,adr)
#endif

extern WORD sec_timer;
extern DWORD master_tick;
DWORD rdclock(void){return master_tick;}
extern char badhost[];
extern char *pioctl_po[];

/* Routines local to this file */
static void near fsm_reset(FSM *psFSM);
static void near ap_reset(FSM *psFSM);
static MBUF *free_pz(MBUF **bpp);
static near int ntohopt(OPTION_HDR *opt, MBUF **bpp);
static near int ntohoptIpcp(OPTION_HDR *opt, MBUF **bpp);

#define fsm_timer(psFSM) start_timer(&(psFSM->timer))

#ifdef PPP_TRACE
char *pszSkipCodes[] = {
	"deliberate cancellation",
	"checksum err",
	"missing ALL address",
	"missing UI",
	"missing upper protocol byte",
	"missing lower protocol byte",
	"not open for Compressed TCP/IP traffic",
	"Compressed TCP/IP not enabled",
	"not open for Uncompressed TCP/IP traffic",
	"Uncompressed TCP/IP not enabled",
};
extern char nospace[];
void clutter_comwin(void);
void unclutter_comwin(void);

void trace_log(char *format,...)
{
	extern char fptracef;
	va_list ap;
	char dlin [100];

	if (!PPPtrace)
		return;
	va_start(ap, format);
	vsprintf(dlin, format, ap);
	va_end(ap);
	if (!fptracef)
		unclutter_comwin();
	trprintf("%s\n",dlin);
	if (!fptracef)
		clutter_comwin();
}

/************************************************************************/
/* General log routine */
static void ppp_log(PPP *psPPP,char *comment)
{
	if (psPPP->trace)
		printf("%s PPP %s\n",psPPP->iface->name,comment);
}

/* Bad packet */
static void ppp_error(PPP *psPPP,MBUF **bpp,char *comment)
{
	free_pz(bpp);
	ppp_log( psPPP, comment );
}
/************************************************************************/
/* FSM log routine */
void near fsm_log(FSM *psFSM,char *comment)
{
	if (PPPtrace > 1)
		trace_log("%s PPP/%s %-8s; %s",
			psFSM->psPPP->iface->name,
			psFSM->pdc->name,
			fsmStates[psFSM->state],
			comment);
}

#define FsmLog(szMessage) fsm_log(psFSM,szMessage)
#define PppLog(szMessage) ppp_log(psPPP,szMessage)
#define PppError(ppb,szMessage) ppp_error(psPPP,ppb,szMessage)
#define TraceLog(szFormat,arg) trace_log(szFormat,arg)
#define TraceLog3(szFormat,a1,a2) trace_log(szFormat,a1,a2)
#define TraceLog4(szFormat,a1,a2,a3) trace_log(szFormat,a1,a2,a3)
#define TraceLog5(szFormat,a1,a2,a3,a4) trace_log(szFormat,a1,a2,a3,a4)
#define TraceLog6(szFormat,a1,a2,a3,a4,a5) trace_log(szFormat,a1,a2,a3,a4,a5)
#define TraceLog7(szFormat,a1,a2,a3,a4,a5,a6) trace_log(szFormat,a1,a2,a3,a4,a5,a6)
#else
#define FsmLog(szMessage)
#define PppLog(szMessage)
#define PppError(ppb,szMessage) free_pz(ppb)
#define TraceLog(szFormat,arg)
#define TraceLog3(szFormat,a1,a2)
#define TraceLog4(szFormat,a1,a2,a3)
#define TraceLog5(szFormat,a1,a2,a3,a4)
#define TraceLog6(szFormat,a1,a2,a3,a4,a5)
#define TraceLog7(szFormat,a1,a2,a3,a4,a5,a6)
#endif

//static int ppp_raw(IFACE *ifp, MBUF *data);

static void ppp_recv(int dev, void *p1, void *p2);

//static int ppp_discard(IFACE *ifp, MBUF **bpp);
//static int ppp_echo(IFACE *ifp, MBUF **bpp);

enum {
	SKIP_CANCEL, 	// deliberate cancellation
	SKIP_CS, 		// checksum err
	SKIP_MALL,		// missing ALL address
	SKIP_MUI,		// missing UI
	SKIP_MUPB,		// missing upper protocol byte
	SKIP_MLPB,		// missing lower protocol byte
	SKIP_NOC,		// not open for Compressed TCP/IP traffic
	SKIP_CNE,		// Compressed TCP/IP not enabled
	SKIP_NOUC,		// not open for Uncompressed TCP/IP traffic
	SKIP_UCNE		// Uncompressed TCP/IP not enabled
};

#define pppfcs(fcs, c)		((fcs >> 8) ^ crctab[(fcs ^ c) & 0x00ff])
#define SP_CHAR			0x20


/****************************************************************************/
// Some modified MBUF routines
void prepend(MBUF **bpp,WORD size)
{
	MBUF *bp = *bpp;

	if(bpp == NULL)
		return;
	if (((*bpp) = alloc_mbuf(size,AllocPppHdr)) == 0)
		return;
	BUFTRACE("Prepend",*bpp);
	(*bpp)->next = bp;
	bp = *bpp;
	bp->cnt = size;
}

MBUF *free_pz(MBUF **bpp)
{
	MBUF *bp = *bpp;
	bp = free_p(bp);
	*bpp = 0;
	return bp;
}

/* Convert PPP header in host form to network form */
static void near htonppp(PPP_HDR *ppp,MBUF **bpp,int iLen)
{
	register BYTE *cp;

	/* Prepend header onto packet data */
	prepend(bpp,iLen);

	/* Load header with proper values */
	cp = (BYTE *)((*bpp)->data);
	if (iLen == PPP_HDR_LEN_O) {
		*(DWORD *)cp = ppp->accm_o;// put the ACCM value for this buffer at the front
		cp += 4;		
	}
	*cp++ = ppp->addr;
	*cp++ = ppp->control;
	cp = put16(cp,ppp->protocol);
}

/* Unknown type input packet */
#pragma argsused
void ppp_skipped(PPP *psPPP,MBUF **bpp,int code)
{
	IPCP *psIPCP;

	if ( (psIPCP = (IPCP *)psPPP->fsm[IPcp].pdv) != NULL ) {
		slhc_toss( psIPCP->slhcp );
	}
	PppError( bpp, pszSkipCodes[code]);
}

// Set the PDV to incoming or outgoing
void SetPdv(FSM *psFSM,int incoming)
{
	int i;

	for (i = Lcp;i < fsmi_Size;i++)
		psFSM[i].pdv = incoming ? psFSM[i].pdv_in : psFSM[i].pdv_out;
}

int StartLcp(PPP *psPPP)
{
	IFACE *ifp = psPPP->iface;
	SER *serp = (SER *)ifp->dcb;
	FSM *psFSM = &(psPPP->fsm[Lcp]);

	if (!(psFSM->flags & FSM_ACTIVE))
		return -1;
	if (!(serp->flags & FL_CONNECTED)) {
		MODEMPAR *pm = serp->pmpar;
		char ioctl[2];
		char *pioctl[1];

		if (pm == 0 || (pm->flags & MF_NODOD))
			return -1;
		ioctl[0] = 0;
		ioctl[1] = IOCTL_CONNECT;
		pioctl[0] = ioctl;
		return (*ifp->ioctl)(ifp,1,pioctl);
	}
	SetPdv(psFSM,0);
	fsm_start(psFSM);
	ap_reset(psPPP->fsm + Ap);	// reset AP
	return 0;
}

/****************************************************************************/
/* Send IP datagram with Point-to-Point Protocol */
#pragma argsused
void ppp_send(
MBUF *bp,		/* Buffer to send */
IFACE *ifp,		/* Pointer to interface control block */
DWORD gateway,	/* Ignored (PPP is point-to-point) */
char precedence,char delay,char throughput,char reliability)
{
	PPP *psPPP;
	IPCP *psIPCP;
	int protocol = PPP_IP_PROTOCOL;

	if (ifp == NULL || (psPPP = (PPP *)ifp->prot_ext) == NULL) {
		free_p(bp);
		return;
	}

	if (psPPP->fsm[IPcp].state != fsmOPENED) {
		PppError( &bp, "not open for IP traffic" );
		psPPP->OutError++;
		StartLcp(psPPP);
		return;
	}

	psIPCP = (IPCP *)psPPP->fsm[IPcp].pdv;
	if (psIPCP->remote.work.negotiate & IPCP_N_COMPRESS) {
		/* Attempt IP/TCP header compression */
		switch ( slhc_compress(psIPCP->slhcp, &bp,
			psIPCP->remote.work.slot_compress) ) {
		case SL_TYPE_IP:
			protocol = PPP_IP_PROTOCOL;
			break;
		case SL_TYPE_COMPRESSED_TCP:
			protocol = PPP_COMPR_PROTOCOL;
			break;
		case SL_TYPE_UNCOMPRESSED_TCP:
			protocol = PPP_UNCOMP_PROTOCOL;
			break;
		default:
			PppError(&bp, "bad IP packet" );
			psPPP->OutError++;
			return;
		};
	}
	(*ifp->output)(ifp,NULL,NULL,protocol,bp,0);
}

/* Send a packet with PPP header */
#pragma argsused
INTVOID ppp_output(
	IFACE *ifp,		/* Pointer to interface control block */
	BYTE *dest,		/* Dest addr (ignored; PPP is point-to-point) */
	BYTE *source,	/* Source addr (ignored; PPP is point-to-point) */
	WORD protocol,	/* PPP Protocol Type field */
	MBUF *data,		/* Actual data to be sent */
	BYTE *rif)
{
	PPP *psPPP;
	PPP_HDR hdr;
	SER *serp;
	LCP *psLCP;

	if (ifp == NULL || (psPPP = (PPP *)ifp->prot_ext) == NULL) {
		free_p(data);
		return VI_ERR;
	}
	if (psPPP->phase == pppDEAD) {
		PppError( &data, "line not up" );
		psPPP->OutError++;
		return VI_ERR;
	}
	serp = (SER *)ifp->dcb;
	if (!(serp->flags & FL_CONNECTED) && serp->tbp) {
		free_p(data);
		return VI_ERR;// don't send a lot of config requests, NT doesn't like it
	}
	psLCP = (LCP *)psPPP->fsm[Lcp].pdv;
	hdr.addr = HDLC_ALL_ADDR;
	hdr.control = HDLC_UI;
	hdr.protocol = protocol;

	hdr.accm_o = LCP_ACCM_DEFAULT;
	if ((protocol != PPP_LCP_PROTOCOL)
	 && (psLCP->remote.work.negotiate & LCP_N_ACCM) ) {
		hdr.accm_o = psLCP->remote.work.accm;
	}
	htonppp(&hdr, &data, PPP_HDR_LEN_O);
	VI_RETURN (*ifp->raw)(ifp,data);
}

/* Encode a raw packet in PPP framing, put on link output queue */


void ppp_raw(IFACE *ifp,MBUF *bp)
{
	PPP *psPPP = (PPP *)ifp->prot_ext;
	LCP *psLCP = (LCP *)psPPP->fsm[Lcp].pdv;
	int full_lcp, full_ac, full_p;
	PPP_HDR ph;
	int len = PPP_HDR_LEN;
	SER *serp = (SER *)ifp->dcb;
	MBUF *nbp;
	register BYTE *cp;
	DWORD dwAccm;

	psPPP->OutTxOctetCount += len_mbuf(bp) - 4 + 2;  // count FCS bytes, subtract accm
//??	ifp->rawsndcnt++;
//??	ifp->lastsent = sec_timer;

	// Get the HDLC/PPP header, without actually pulling mbuf up
	if (bp == NULL || bp->cnt < PPP_HDR_LEN + 4) {
		PppError( &bp, "link header missing" );
		psPPP->OutError++;
		return;
	}
	dwAccm = *(DWORD *)bp->data;
	ph.addr = bp->data[4];
	ph.control = bp->data[5];
	ph.protocol = get16(&bp->data[6]);

	full_lcp = (ph.protocol == PPP_LCP_PROTOCOL);
	if ((full_p = (full_lcp
			||  !(psLCP->remote.work.negotiate & LCP_N_PFC)
			||  ph.protocol >= 0x00ff))
		== 0 ) {
		--len;
	}

	// Discard HDLC address and control fields if possible
	if ( ( full_ac = ( full_lcp
			||  !(psLCP->remote.work.negotiate & LCP_N_ACFC) ) )
		== 0 ) {
		len -= 2;
	}

	// Load header with proper values
	bp->cnt -= PPP_HDR_LEN - len;
	cp = (bp->data += PPP_HDR_LEN - len);
	*(DWORD *)cp = dwAccm;
	cp += 4;
	if ( full_ac ) {
		*cp++ = ph.addr;
		*cp++ = ph.control;
	}
	if ( full_p )
		*cp++ = (ph.protocol >> 8);
	*cp = (ph.protocol & 0x00ff);

#ifdef TRACE
	{
		MBUF *tbp;

		dup_p(&tbp,bp,4,len_mbuf(bp) - 4,AllocTraceDup1);
		if (tbp) {
			dump(ifp,IF_TRACE_OUT,TRACE_PPP,tbp);
			free_p(tbp);
		}
	}
#endif

	/* Copy packet to a new mbuf to solve problem where
	 * an ack of a segment while a retransmit is taking place, frees
	 * the mbuf data causing garbage to be sent. Could also be used to
	 * re-use used portions of the mbuf after VJ compression since we are
	 * guarenteed to not have a duplicated header with no real data, BUT
	 * the code will have to be moved to before compression.
	 */
	nbp = copy_p(bp,len_mbuf(bp),AllocPppDup);
	free_p(bp);
	if (nbp)
		bp = nbp;
	else
		return;
	enqueue(&serp->sndq,bp);
	if(serp->tbp == NULLBUF && serp->ostate == OS_IDLE) {
		if (serp->sndq == NULLBUF || (bp = serp->tbp = dequeue(&serp->sndq)) == 0) {
			return;	/* No work */
		}
		serp->accm_o = *(DWORD *)bp->data;
		bp->data += 4;
		bp->cnt -= 4;
		(*ifp->xmit)(ifp,0,1);
	}
}

/****************************************************************************/
/* Packetize PPP input from device */
#pragma argsused
void ppp_recv(int dev,void *p1,void *p2)
{
}

/* Process incoming PPP packets */
/* (called from network task) */

void ppp_proc(IFACE *ifp)
{
	PPP *psPPP;
	LCP *psLCP;
	IPCP *psIPCP;
	PPP_HDR ph;
	WORD negotiated = 0;
#ifdef TRACE
	WORD wError = 0;
#endif
	struct mbuf *bp;
	SER *serp;
	int cnt,state = IS_PPP;
	BYTE *pIn,*pOut,c;

	serp = (SER *)ifp->dcb;
	psPPP = (PPP *)ifp->prot_ext;
	/* Process any pending input */
	serp->accum = 0xffff;
	if ((bp = dequeue(&serp->rcvq)) != NULLBUF) {
#ifdef TRACE
		dump(ifp,IF_TRACE_IN,0,bp);
#endif
		BUFTRACE("ppp_proc",bp);
		cnt = bp->cnt;
		bp->cnt = 0;
		pIn = pOut = bp->data;
		while (cnt--) {
			c = *pIn++;
			switch (state) {
			case IS_PPP:
				if (c < SP_CHAR &&
#ifdef TRACE
							!(ifp->sflags & SF_PPP_TRACE) &&
#endif
							(serp->accm_i & (1L << c))) {
					break;
				}
				else if (c == HDLC_ESC_ASYNC) {
					state = IS_PPPESC;
					break;
				}
				goto storpb;
			case IS_PPPESC:
				if (c == HDLC_FLAG) {
					ppp_skipped( psPPP, &serp->rbp,SKIP_CANCEL); //"deliberate cancellation"
					psPPP->InFrame++;
					break;
				}
				else if (c < SP_CHAR &&
#ifdef TRACE
								!(ifp->sflags & SF_PPP_TRACE) &&
#endif
								(serp->accm_i & (1L << c))) {
					break;
				}
				c ^= HDLC_ESC_COMPL;
				state = IS_PPP;
				storpb:
				serp->accum = (serp->accum >> 8) ^
										crctab[(serp->accum ^ c) & 0xff];
				if (bp->cnt >= serp->bufsize) {
					serp->errors++;
					state = IS_PPPIGN;
					bp->cnt = 0;
					serp->accum = 0xffff;
				}
				else {
					*pOut++ = c;
					++bp->cnt;
				}
				break;
			case IS_PPPIGN:
				if (c == HDLC_FLAG)
					state = IS_PPP;
				break;
			}
		}
		if (bp->cnt > 2)
			bp->cnt -= 2;
		if (serp->accum != HDLC_FCS_FINAL) {
			ppp_skipped( psPPP, &serp->rbp,SKIP_CS); // checksum err
			psPPP->InChecksum++;
#ifdef TRACE
			wError = IF_TRACE_OUT;
		}
		dump(ifp,IF_TRACE_IN | wError,TRACE_PPP,bp);
		if (ifp->sflags & SF_PPP_TRACE) {
#endif
			free_p(bp);
			goto done;
		}
#ifdef TRACE
		if (wError) {
			free_p(bp);
			goto done;
		}
#endif
		psPPP->InRxOctetCount += len_mbuf(bp) + 2;   // count FCS bytes

		// Use negotiated values if LCP finished
		if (psPPP->fsm[Lcp].state == fsmOPENED) {
			psLCP = (LCP *)psPPP->fsm[Lcp].pdv;

			negotiated = psLCP->local.work.negotiate;
		}

		// HDLC address and control fields may be compressed out
		if ((BYTE)bp->data[0] != HDLC_ALL_ADDR) {
			if (!(negotiated & LCP_N_ACFC)) {
				ppp_skipped( psPPP, &bp, SKIP_MALL); //missing ALL address
				psPPP->InFrame++;
				return;
			}
		} else if ((BYTE)bp->data[1] != HDLC_UI) {
			if (!(negotiated & LCP_N_ACFC)
			 || !(negotiated & LCP_N_PFC)) {
				ppp_skipped( psPPP, &bp, SKIP_MUI);	//missing UI
				psPPP->InFrame++;
				return;
			}
		} else {
			// skip address/control fields
			pull16(&bp);
		}

		// Initialize the expected header
		ph.addr = HDLC_ALL_ADDR;
		ph.control = HDLC_UI;
		ph.protocol = pullchar(&bp);

		// First byte of PPP protocol field may be compressed out
		if ( ph.protocol & 0x01 ) {
			if (!(negotiated & LCP_N_PFC)) {
				ppp_skipped( psPPP, &bp, SKIP_MUPB);	// missing upper protocol byte
				psPPP->InFrame++;
				return;
			}
		} else {
			ph.protocol = (ph.protocol << 8) | pullchar(&bp);

			// Second byte of PPP protocol field must be odd
			if ( !(ph.protocol & 0x01) ) {
				ppp_skipped( psPPP, &bp, SKIP_MLPB);	//"missing lower protocol byte" );
				psPPP->InFrame++;
				return;
			}
		}
		switch(ph.protocol) {
		case PPP_IP_PROTOCOL:	// Regular IP
			if ( psPPP->fsm[IPcp].state != fsmOPENED ) {
				PppError( &bp, "not open for IP traffic" );
				psPPP->InError++;
				break;
			}
			ip_route(bp,0,ifp);
			break;

		case PPP_COMPR_PROTOCOL:	// Van Jacobson Compressed TCP/IP
			if ( psPPP->fsm[IPcp].state != fsmOPENED ) {
				ppp_skipped( psPPP, &bp, SKIP_NOC);	// "not open for Compressed TCP/IP traffic" );
				psPPP->InError++;
				break;
			}

			psIPCP = (IPCP *)psPPP->fsm[IPcp].pdv;
			if (!(psIPCP->local.work.negotiate & IPCP_N_COMPRESS)) {
				ppp_skipped( psPPP, &bp, SKIP_CNE);	//"Compressed TCP/IP not enabled" );
				psPPP->InError++;
				break;
			}

			if ( slhc_uncompress(psIPCP->slhcp, &bp) <= 0 ) {
				PppError( &bp, "Compressed TCP/IP packet error" );
				psPPP->InError++;
				break;
			}
			ip_route(bp,0,ifp);
			break;

		case PPP_UNCOMP_PROTOCOL:	// Van Jacobson Uncompressed TCP/IP
			if ( psPPP->fsm[IPcp].state != fsmOPENED ) {
				ppp_skipped( psPPP, &bp, SKIP_NOUC);	//"not open for Uncompressed TCP/IP traffic" );
				psPPP->InError++;
				break;
			}

			psIPCP = (IPCP *)psPPP->fsm[IPcp].pdv;
			if (!(psIPCP->local.work.negotiate & IPCP_N_COMPRESS)) {
				ppp_skipped( psPPP, &bp, SKIP_UCNE);	//"Uncompressed TCP/IP not enabled" );
				psPPP->InError++;
				break;
			}

			if ( slhc_remember(psIPCP->slhcp, &bp) <= 0 ) {
				PppError( &bp, "Uncompressed TCP/IP packet error" );
				psPPP->InError++;
				break;
			}
			ip_route(bp,0,ifp);
			break;

		case PPP_LCP_PROTOCOL:	// Link Control Protocol
			psPPP->InNCP[Lcp]++;
			fsm_proc(&(psPPP->fsm[Lcp]),&bp);
			if (psPPP->fsm[Lcp].state == fsmOPENED) {
				psLCP = (LCP *)psPPP->fsm[Lcp].pdv;
				if (psLCP->local.work.negotiate & LCP_N_ACCM) {
					serp->accm_i = psLCP->local.work.accm;
				}
				else
					serp->accm_i = LCP_ACCM_DEFAULT;
			}
			else
				serp->accm_i = LCP_ACCM_DEFAULT;
			break;

		case PPP_PAP_PROTOCOL:	// Password Authenticate Protocol
		case PPP_CHAP_PROTOCOL:	// Challenge Handshake Authenticate Protocol
			if (psPPP->phase != pppAP
			 && psPPP->phase != pppREADY) {
				PppError( &bp, "not ready for Authentication" );
				psPPP->InError++;
				break;
			}
			psPPP->InNCP[Ap]++;
			ap_proc(&(psPPP->fsm[Ap]),&bp,ph.protocol);
			break;

		case PPP_IPCP_PROTOCOL:	// IP Control Protocol
			if (psPPP->phase != pppREADY) {
				PppError( &bp, "not ready for IPCP traffic" );
				psPPP->InError++;
				break;
			}
			psPPP->InNCP[IPcp]++;
			fsm_proc(&(psPPP->fsm[IPcp]),&bp);
			break;

		default:
			TraceLog3("%s PPP Unknown packet protocol: %x;",psPPP->iface->name,	ph.protocol);
			psPPP->InUnknown++;

			// Build Protocol Reject packet:
			// put the header back on ...

			htonppp(&ph,&bp,PPP_HDR_LEN);
			// ... then pull off the address and control fields ...
			pull16(&bp);
			// ... and send it as an LCP packet
			fsm_send( &(psPPP->fsm[Lcp]), 0, PROT_REJ, 0, &bp);
			break;
		};
	}
	done:
	(*ifp->ioctl)(ifp,1,pioctl_po);
	/* Kick the transmitter if it's idle */
	(*ifp->xmit)(ifp,0,0);
	return;
}

/************************************************************************/

/* Keep track of changes in I-O status */
/* (called through iface iostatus vector) */
int ppp_iostatus(IFACE *ifp,int command,DWORD value)
{
	PPP *psPPP = (PPP *)ifp->prot_ext;
	SER *serp = (SER *)ifp->dcb;

	switch ( command ) {
	case PARAM_UP:
		if (serp->flags & FL_CONNECTED) {
			PppLog("Physical layer up");
			psPPP->phase = pppLCP;
		}
		if (value)
			StartLcp(psPPP);
		return 0;

	case PARAM_DOWN:
		if (psPPP->flags & PPP_CBVCALL)
			return 0;	// ignore when trying to call back
		PppLog("Physical layer down");
		fsm_down( &(psPPP->fsm[Lcp]) );
		psPPP->phase = pppDEAD;
		return 0;
	};
	return -1;
}

int ppp_free(IFACE *ifp)
{
	PPP *psPPP = (PPP *)ifp->prot_ext;
	int fsmi;

	for ( fsmi = Lcp; fsmi < fsmi_Size; ) {
		fsm_free( &(psPPP->fsm[fsmi++]) );
	}
	free( psPPP->peername );
	free( psPPP );
	return 0;
}

//  PPPLCP.C	-- negotiate data link options

static BYTE lcp_option_length[] = {
	 0,		/* unused */
	 4,		/* MRU */
	 6,		/* ACCM */
	 4,		/* authentication */
	 4,		/* encryption */
	 6,		/* magic number */
	 6,		/* monitor reporting period */
	 2,		/* Protocol compression */
	 2		/* Address/Control compression */
};


static void near lcp_option(MBUF **bpp,
			LCP_VALUE *psValue,
			BYTE o_type,
			BYTE o_length,
			MBUF **copy_bpp );
static void near lcp_makeoptions(MBUF **bpp,
			LCP_VALUE *psValue,
			WORD negotiating);
static MBUF * near lcp_makereq(FSM *psFSM);

static int near lcp_check(MBUF **bpp,
			LCP *psLCP,
			LCP_SIDE *psSide,
			OPTION_HDR *psOption,
			int request,IFACE *psIface);

static int near lcp_request(FSM *psFSM,
			CONFIG_HDR *config,
			MBUF **data);
static int near lcp_ack(FSM *psFSM,
			CONFIG_HDR *ackcnf,
			MBUF **data);
static int near lcp_nak(FSM *psFSM,
			CONFIG_HDR *nakcnf,
			MBUF **data);
static int near lcp_reject(FSM *psFSM,
			CONFIG_HDR *rejcnf,
			MBUF **data);

static void near lcp_reset(FSM *psFSM);
static void near lcp_starting(FSM *psFSM);
static void near lcp_stopping(FSM *psFSM);

static void near lcp_closing(FSM *psFSM);
static void near lcp_opening(FSM *psFSM);

static void near lcp_free(FSM *psFSM);

FSM_CONSTANT lcp_constants = {
	"Lcp",
	PPP_LCP_PROTOCOL,
	0x0FFE,				/* codes 1-11 recognized */

	Lcp,
	LCP_REQ_TRY,
	LCP_NAK_TRY,
	LCP_TERM_TRY,
	LCP_TIMEOUT * 1000L,

	lcp_free,

	lcp_reset,
	lcp_starting,
	lcp_opening,
	lcp_closing,
	lcp_stopping,

	lcp_makereq,
	lcp_request,
	lcp_ack,
	lcp_nak,
	lcp_reject
};

/************************************************************************/
/*			E V E N T   P R O C E S S I N G			*/
/************************************************************************/

static void near lcp_option(MBUF **bpp,LCP_VALUE *psValue,
	BYTE o_type,BYTE o_length,MBUF **copy_bpp)
{
	MBUF *bp;
	register BYTE *cp;
	register int toss = o_length - OPTION_HDR_LEN;

	if ((bp = alloc_mbuf(o_length,AllocPppLopt)) == NULL) {
		return;
	}
	BUFTRACE("lcp_option",bp);
	cp = bp->data;
	*cp++ = o_type;
	*cp++ = o_length;

	switch ( o_type ) {
	case LCP_MRU:
		put16(cp, psValue->mru);
		toss -= 2;
		TraceLog("    making MRU: %d", psValue->mru);
		break;

	case LCP_ACCM:
		put32(cp, psValue->accm);
		toss -= 4;
		TraceLog("    making ACCM: 0x%08lx", psValue->accm);
		break;

	case LCP_AUTHENT:
		put16(cp, psValue->authentication);
		cp += 2;
		toss -= 2;
		if (psValue->authentication == PPP_CHAP_PROTOCOL) {
			toss--;
			*cp++ = AUTHENT_CHAP_MD5;
		}
		TraceLog("    making Auth Protocol: 0x%04x",psValue->authentication);
		break;

	case LCP_MAGIC:
		put32(cp, psValue->magic_number);
		toss -= 4;
		TraceLog("    making Magic Number: 0x%08lx",psValue->magic_number);
		break;

	case LCP_PFC:
		TraceLog("    making Protocol compression",0);
		break;

	case LCP_ACFC:
		TraceLog("    making Addr/Ctl compression",0);
		break;

	case LCP_QUALITY:		/* not implemented */
	default:
		TraceLog("    making unimplemented type %d", o_type);
		break;
	};

	while ( toss-- > 0 ) {
		*cp++ = pullchar(copy_bpp);
	}
	bp->cnt += o_length;
	append(bpp, bp);
}


/************************************************************************/
/* Build a list of options */
static void near lcp_makeoptions(MBUF **bpp,LCP_VALUE *psValue,WORD negotiating)
{
	register int o_type;

	for ( o_type = 1; o_type <= LCP_OPTION_LIMIT; o_type++ ) {
		if (negotiating & (1 << o_type)) {
			lcp_option( bpp, psValue, o_type,
				lcp_option_length[ o_type ] +
				((o_type == LCP_AUTHENT &&
				psValue->authentication == PPP_CHAP_PROTOCOL) ? 1 : 0), NULL);
		}
	}
}


/************************************************************************/
/* Build a request to send to remote host */
static MBUF * near lcp_makereq(FSM *psFSM)
{
	LCP *psLCP = (LCP *)psFSM->pdv;
	MBUF *req_bp = NULL;

	lcp_makeoptions( &req_bp, &(psLCP->local.work),
				psLCP->local.work.negotiate );
	return(req_bp);
}

/************************************************************************/
/* Check the options, updating the working values.
 * Returns -1 if ran out of data, ACK/NAK/REJ as appropriate.
 */
static int near lcp_check(MBUF **bpp,LCP *psLCP,LCP_SIDE *psSide,
	OPTION_HDR *psOption,int request,IFACE *psIface)
{
	int toss = psOption->len - OPTION_HDR_LEN;
	int option_result = CONFIG_ACK;		/* Assume good values */
	int i;
	char auth_algorithm;

	switch(psOption->type) {
	case LCP_MRU:
		psSide->work.mru = pull16(bpp);
		toss -= 2;
		TraceLog("    checking MRU: %d", psSide->work.mru);
		/* Check if new value is appropriate */
		if (psSide->work.mru < LCP_MRU_LO) {
			psSide->work.mru = LCP_MRU_LO;
			option_result = CONFIG_NAK;
		} else if (psSide->work.mru > LCP_MRU_HI) {
			psSide->work.mru = LCP_MRU_HI;
			option_result = CONFIG_NAK;
		} else if (psSide->work.mru > psIface->mtu) {
			psSide->work.mru = psIface->mtu;
			option_result = CONFIG_NAK;
		}
		if ( request && (psSide->want.negotiate & LCP_N_MRU)
		  && psSide->work.mru > psSide->want.mru ) {
			psSide->work.mru = psSide->want.mru;
			option_result = psSide->want.mru;
		}
		break;

	case LCP_ACCM:
		psSide->work.accm = pull32(bpp);
		toss -= 4;
		TraceLog("    checking ACCM: 0x%08lx", psSide->work.accm);
		/* Remote host may ask to escape more control  */
		/* characters than we require, but must escape */
		/* at least the control chars that we require. */
		if ( (!request || (psSide->want.negotiate & LCP_N_ACCM))
		  && psSide->work.accm !=
				(psSide->work.accm | psSide->want.accm) ) {
			psSide->work.accm |= psSide->want.accm;
			option_result = CONFIG_NAK;
		}
		break;

	case LCP_AUTHENT:
		psSide->work.authentication = pull16(bpp);
		toss -= 2;
		if ((i = psOption->len - 4) > 0) {
			--i;
         --toss;
			auth_algorithm = pullchar(bpp);
		}
		for (;i > 0;--i,--toss) {
			pullchar(bpp);
		}
		TraceLog("    checking Auth Protocol: 0x%04x",psSide->work.authentication);
		/* Check if new value is appropriate */
		switch ( psSide->work.authentication ) {
		case PPP_PAP_PROTOCOL:
			if (psSide->want.authentication == PPP_CHAP_PROTOCOL)
				goto auth_nak;
			break;
		case PPP_CHAP_PROTOCOL:
			if (psSide->want.authentication != PPP_CHAP_PROTOCOL)
				goto auth_nak;
			if (auth_algorithm != AUTHENT_CHAP_MD5)
				goto auth_nak;
			break;
		default:
			auth_nak:
			psSide->work.authentication = psSide->want.authentication;
			option_result = CONFIG_NAK;
			psOption->len = psSide->work.authentication == PPP_CHAP_PROTOCOL ? 5 : 4;
			break;
		};
		break;

	case LCP_MAGIC:
		psSide->work.magic_number = pull32(bpp);
		toss -= 4;
		TraceLog("    checking Magic Number: 0x%08lx",psSide->work.magic_number);
		/* Ensure that magic numbers are different */
		if (psSide->work.magic_number == 0L
		 || psLCP->remote.work.magic_number == psLCP->local.work.magic_number) {
			psSide->work.magic_number += rdclock();
			option_result = CONFIG_NAK;
		}
		break;

	case LCP_PFC:
		TraceLog("    checking Protocol compression",0);
		break;

	case LCP_ACFC:
		TraceLog("    checking Addr/Ctl compression",0);
		break;

	case LCP_QUALITY:		/* not implemented */
	default:
		option_result = CONFIG_REJ;
		break;
	};

	if (psOption->type > LCP_OPTION_LIMIT
	 || !(psSide->will_negotiate & (1 << psOption->type))) {
		option_result = CONFIG_REJ;
	}

	if ( toss < 0 )
		return -1;

	if ( !request  &&  toss > 0 ) {
		/* toss extra bytes in option */
		while( toss-- > 0 ) {
			if ( pullchar(bpp) == -1 )
				return -1;
		}
	}

	return (option_result);
}


/************************************************************************/
/* Check Link Control options requested by the remote host */
static int near lcp_request(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	LCP *psLCP = (LCP *)psFSM->pdv;
	long signed_length = config->len;
	MBUF *reply_bp = NULL;			/* reply packet */
	int reply_result = CONFIG_ACK;	/* reply to request */
	WORD desired;					/* desired to negotiate */
	OPTION_HDR option;				/* option header storage */
	int option_result;				/* option reply */

	psLCP->remote.work.negotiate = 0;/* clear flags */

	/* Process options requested by remote host */
	while (signed_length > 0  &&  ntohopt(&option, data) != -1) {
		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			free_pz(&reply_bp);
			return -1;
		}

		if ( ( option_result = lcp_check( data, psLCP,
				&(psLCP->remote), &option, 1,psFSM->psPPP->iface ) ) == -1 ) {
			free_pz(data);
			free_pz(&reply_bp);
			return -1;
		}
		TraceLog4("LCP REQ: result %s, option %d, length %d",fsmCodes[option_result],option.type,option.len);
		if ( option_result < reply_result ) {
			continue;
		} else if ( option_result > reply_result ) {
			/* Discard current list of replies */
			free_pz(&reply_bp);
			reply_bp = NULL;
			reply_result = option_result;
		}

		/* remember that we processed option */
		if ( option_result != CONFIG_REJ
		  && option.type <= LCP_OPTION_LIMIT ) {
			psLCP->remote.work.negotiate |= (1 << option.type);
		}

		/* Add option response to the return list */
		lcp_option( &reply_bp, &(psLCP->remote.work),
			option.type, option.len, data );
	}

	/* Now check for any missing options which are desired */
	if ( psFSM->retry_nak > 0
	 &&  (desired = psLCP->remote.want.negotiate
				& ~psLCP->remote.work.negotiate) != 0 ) {
		switch ( reply_result ) {
		case CONFIG_ACK:
			free_pz(&reply_bp);
			reply_bp = NULL;
			reply_result = CONFIG_NAK;
			/* fallthru */
		case CONFIG_NAK:
			lcp_makeoptions( &reply_bp, &(psLCP->remote.want),
				desired );
			psFSM->retry_nak--;
			break;
		case CONFIG_REJ:
			/* do nothing */
			break;
		};
	} else if ( reply_result == CONFIG_NAK ) {
		/* if too many NAKs, reject instead */
		if ( psFSM->retry_nak > 0 )
			psFSM->retry_nak--;
		else
			reply_result = CONFIG_REJ;
	}

	/* Send ACK/NAK/REJ to remote host */
	fsm_send(psFSM, 0, reply_result, config->id, &reply_bp);
	free_pz(data);
	return (reply_result != CONFIG_ACK);
}


/************************************************************************/
/* Process configuration ACK sent by remote host */
static int near lcp_ack(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	MBUF *req_bp;
	int error = 0;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* Get a copy of last request we sent */
	req_bp = lcp_makereq(psFSM);

	/* Overall buffer length should match */
	if (config->len != len_mbuf(req_bp)) {
		error = 1;
	} else {
		register int req_char;
		register int ack_char;

		/* Each byte should match */
		while ((req_char = pullchar(&req_bp)) != -1) {
			if ((ack_char = pullchar(data)) == -1
			 || ack_char != req_char ) {
				error = 1;
				break;
			}
		}
	}
	free_pz(&req_bp);
	free_pz(data);

	if (error) {
		return -1;
	}

	return 0;
}


/************************************************************************/
/* Process configuration NAK sent by remote host */
static int near lcp_nak(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	LCP *psLCP = (LCP *)psFSM->pdv;
	LCP_SIDE *local_p = &(psLCP->local);
	long signed_length = config->len;
	OPTION_HDR option;
	int last_option = 0;
	int result;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* First, process in order.  Then, process extra "important" options */
	while (signed_length > 0  &&  ntohopt(&option, data) != -1) {
		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			return -1;
		}
		if ( option.type > LCP_OPTION_LIMIT ) {
		} else if ( option.type < last_option
		  || !(local_p->work.negotiate & (1 << option.type)) ) {
			if (local_p->work.negotiate & (1 << option.type)) {
				free_pz(data);
				return -1;		/* was requested */
			}
			local_p->work.negotiate |= (1 << option.type);
			last_option = LCP_OPTION_LIMIT + 1;
		} else {
			last_option = option.type;
		}
		if ( ( result = lcp_check( data, psLCP,
				local_p, &option, 0, psFSM->psPPP->iface ) ) == -1 ) {
			free_pz(data);
			return -1;
		}
		/* update the negotiation status */
		if ( result == CONFIG_REJ
		  && option.type <= LCP_OPTION_LIMIT ) {
			local_p->work.negotiate &= ~(1 << option.type);
		}
	}
	free_pz(data);
	return 0;
}


/************************************************************************/
/* Process configuration reject sent by remote host */
static int near lcp_reject(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	LCP *psLCP = (LCP *)psFSM->pdv;
	LCP_SIDE *local_p = &(psLCP->local);
	long signed_length = config->len;
	OPTION_HDR option;
	int last_option = 0;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* Process in order, checking for errors */
	while (signed_length > 0  &&  ntohopt(&option, data) != -1) {
		register int k;

		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			return -1;
		}
		if ( option.type > LCP_OPTION_LIMIT ) {
		} else if ( option.type < last_option
		 || !(local_p->work.negotiate & (1 << option.type))) {
			free_pz(data);
			return -1;
		}
		for ( k = option.len - OPTION_HDR_LEN; k-- > 0; ) {
			if ( pullchar(data) == -1 ) {
				free_pz(data);
				return -1;
			}
		}
		last_option = option.type;

		if ( option.type <= LCP_OPTION_LIMIT ) {
			local_p->work.negotiate &= ~(1 << option.type);
		}
	}
	free_pz(data);
	return 0;
}


/************************************************************************/
/*			I N I T I A L I Z A T I O N			*/
/************************************************************************/

/* Check for PPP Network-Layer Protocol Phase */
void ppp_ready(PPP *psPPP)
{
	if ( !(psPPP->flags & (PPP_AP_LOCAL | PPP_AP_REMOTE)) ) {
		/* no pending authentication */
		psPPP->phase = pppREADY;

		psPPP->upsince = sec_timer;
		fsm_start( &(psPPP->fsm[IPcp]) );
	}
}

/****************************************************************************/
/* Reset configuration options before request */
static void near LcpReset(LCP *psLCP)
{
	if ( psLCP->local.want.negotiate & LCP_N_MAGIC ) {
		psLCP->local.want.magic_number += rdclock();
	}

	psLCP->local.work = psLCP->local.want;
	psLCP->local.will_negotiate |= psLCP->local.want.negotiate;

	psLCP->remote.work.negotiate = 0;
	psLCP->remote.will_negotiate |= psLCP->remote.want.negotiate;
}

static void near lcp_reset(FSM *psFSM)
{
	PPP *psPPP;
	SER *serp;

	LcpReset((LCP *)psFSM->pdv_out);
	LcpReset((LCP *)psFSM->pdv_in);
	psPPP = psFSM->psPPP;
	serp = (SER *)psPPP->iface->dcb;
	serp->accm_i = serp->accm_o = LCP_ACCM_DEFAULT;
}

/************************************************************************/
/* Prepare to begin configuration exchange */
static void near lcp_starting(FSM *psFSM)
{
	psFSM->psPPP->phase = pppLCP;
}

/************************************************************************/
/* After termination */
static void near lcp_stopping(FSM *psFSM)
{
	void asy_disconnect(SER *serp);		// disconnect by dropping DTR for 1 second
	PPP *psPPP = psFSM->psPPP;
	IFACE *ifp = psPPP->iface;

	asy_disconnect((SER *)ifp->dcb);
	ppp_iostatus(ifp,PARAM_DOWN,0);	// stop PPP
	if (psPPP->flags & PPP_CBV) {
		psPPP->flags &= ~PPP_CBV;
		psPPP->flags |= PPP_CBVCALL;
	}
	else
		psPPP->flags &= ~PPP_CBVCALL;
}

/************************************************************************/
/* Close higher levels in preparation for link shutdown */
static void near lcp_closing(FSM *psFSM)
{
	PPP *psPPP = psFSM->psPPP;

	psPPP->phase = pppTERMINATE;

	fsm_down( &(psPPP->fsm[IPcp]) );
	ap_down( &(psPPP->fsm[Ap]) );
}

#define CHALLENGE_LENGTH 16

//#define CHALLENGE_DEBUG
#ifdef CHALLENGE_DEBUG
static char cChal[] = {
0x8b,0xd1,0x04,0xe1,0x86,0xc1,0xc8,0xd8,
0xa0,0x9a,0xbb,0x96,0x49,0xae,0xc0,0x01
};
/*
challenge:
p sent at 21:00:33.22:
PPP: len  24 CO prot CHAP Challenge id 0 len 22
0000  10 8b d1 04 e1 86 c1 c8 d8 a0 9a bb 96 49 ae c0  ..Q.a.AHX .;.I.@
0010  01 53                                            .S

expected response - user: gvn password gvn
p recv at 21:00:33.28:
PPP: len  26 CO prot CHAP Response id 0 len 24
0000  10 6e 7e 5f f9 e0 db 66 fb c5 70 f7 3c 16 58 f0  .n~_y`[f{Epw<.Xp
0010  32 67 76 6e                                      2gvn
*/
#endif

static int near ap_send(FSM *psFSM, char cCommand);

/************************************************************************/
/* configuration negotiation complete */
static void near lcp_opening(FSM *psFSM)
{
	LCP *psLCP = 	(LCP *)psFSM->pdv;
	PPP *psPPP =	psFSM->psPPP;
	FSM *psApFSM = psPPP->fsm + Ap;
	IFACE *ifp = 	psPPP->iface;

	if (ifp->mtu > psLCP->remote.work.mru) {
		/* Set new Max Transmission Unit for outgoing packets */
		ifp->mtu = psLCP->remote.work.mru;
		((SER *)ifp->dcb)->bufsize = 2 * ifp->mtu + 20;	// save on buffer size as well
		TraceLog("    Set new MTU for outgoing packets: %d",ifp->mtu);
	}

	/* check for authentication */
	psPPP->phase = pppAP;
	psPPP->flags &= ~(PPP_AP_LOCAL | PPP_AP_REMOTE);
	free(psPPP->peername);
	psPPP->peername = NULL;

	if (psLCP->local.work.negotiate & LCP_N_AUTHENT) {
		psApFSM->flags |= PPP_AP_LOCAL;
		psPPP->flags |= PPP_AP_LOCAL;
		switch (psLCP->local.work.authentication) {
		case PPP_CHAP_PROTOCOL:
			// send challenge
			psApFSM->state |= AP_CHALLENGE_SENT;
			ap_send(psApFSM,SEND_CHALLENGE);
			break;
		}
	}
	if (psLCP->remote.work.negotiate & LCP_N_AUTHENT) {
		psApFSM->flags |= PPP_AP_REMOTE;
		psPPP->flags |= PPP_AP_REMOTE;
		switch (psLCP->remote.work.authentication) {
		case PPP_PAP_PROTOCOL:
			psApFSM->state |= AP_REQUEST_SENT;
			ap_send(psApFSM, CONFIG_REQ);
			break;
		}
	}

	/* re-check for authentication */
	ppp_ready(psPPP);
}

/************************************************************************/
#pragma argsused
static void near lcp_free(FSM *psFSM)
{
	/* nothing to do */
}

//  PPPIPCP.C	-- negotiate IP parameters

static BYTE option_length[] = {
	 0,		/* unused */
	10,		/* addresses */
	 6,		/* compression */
	 6, 		// address
	 6,		// DNS1
	 6			// DNS2
};


static void near ipcp_option(MBUF **bpp,
			IPCP_VALUE *psValue,
			BYTE bOptionType,
			BYTE o_length,
			MBUF **copy_bpp);
static void near ipcp_makeoptions(MBUF **bpp,
			IPCP_VALUE *psValue,
			WORD negotiating);
static MBUF * near ipcp_makereq(FSM *psFSM);

static int near ipcp_check(MBUF **bpp,FSM *psFSM,
			IPCP *psIPCP,
			IPCP_SIDE *psSide,
			OPTION_HDR *psOption,
			int request);

static int near ipcp_request(FSM *psFSM,
			CONFIG_HDR *config,
			MBUF **data);
static int near ipcp_ack(FSM *psFSM,
			CONFIG_HDR *config,
			MBUF **data);
static int near ipcp_nak(FSM *psFSM,
			CONFIG_HDR *config,
			MBUF **data);
static int near ipcp_reject(FSM *psFSM,
			CONFIG_HDR *config,
			MBUF **data);

static void near ipcp_reset(FSM *psFSM);

static void near ipcp_starting(FSM *psFSM);
static void near ipcp_stopping(FSM *psFSM);

static void near ipcp_closing(FSM *psFSM);
static void near ipcp_opening(FSM *psFSM);

static void near ipcp_free(FSM *psFSM);
static DWORD near ipcp_lookuppeer(PPP *psPPP);


FSM_CONSTANT ipcp_constants = {
	"IPcp",
	PPP_IPCP_PROTOCOL,
	0x00FE,				/* codes 1-7 recognized */

	IPcp,
	IPCP_REQ_TRY,
	IPCP_NAK_TRY,
	IPCP_TERM_TRY,
	IPCP_TIMEOUT * 1000L,

	ipcp_free,

	ipcp_reset,
	ipcp_starting,
	ipcp_opening,
	ipcp_closing,
	ipcp_stopping,

	ipcp_makereq,
	ipcp_request,
	ipcp_ack,
	ipcp_nak,
	ipcp_reject,
};


/************************************************************************/
/*			E V E N T   P R O C E S S I N G			*/
/************************************************************************/

static void near ipcp_option(MBUF **bpp,IPCP_VALUE *psValue,BYTE o_type,
	BYTE o_length,MBUF **copy_bpp)
{
	MBUF *bp;
	register BYTE *cp;
	register int toss = o_length - OPTION_HDR_LEN;
	int i;

	if ((bp = alloc_mbuf(o_length,AllocPppIopt)) == NULL) {
		return;
	}
	BUFTRACE("ipcp_option",bp);
	cp = bp->data;
	*cp++ = o_type == IPCP_DNS1 ? IPCP_PRI_DNS :
			  o_type == IPCP_DNS2 ? IPCP_SEC_DNS : o_type;
	*cp++ = o_length;

	switch ( o_type ) {
	/*
	case IPCP_ADDRESSES:
		cp = put32(cp, psValue->dwLocalAddress);
		cp = put32(cp, psValue->dwRemoteAddress);
		toss -= 8;
		TraceLog("    making IP source address: %s",inet_ntoa(psValue->dwLocalAddress));
		TraceLog("    making IP destination address %s",inet_ntoa(psValue->dwRemoteAddress));
		break;
	*/
	case IPCP_COMPRESS:
		cp = put16(cp, psValue->compression);
		toss -= 2;
		TraceLog("    making IP compression 0x%04x",psValue->compression);
		if ( psValue->compression == PPP_COMPR_PROTOCOL ) {
			*cp++ = psValue->slots - 1;
			*cp++ = psValue->slot_compress;
			toss -= 2;
			TraceLog3("    with IP compression slots %d, flag %x",psValue->slots,psValue->slot_compress);
		}
		break;
	case IPCP_ADDRESS:
		cp = put32(cp, psValue->dwLocalAddress);
		toss -= 4;
		TraceLog("    making IP source address: %s",inet_ntoa(psValue->dwLocalAddress));
		break;

	case IPCP_DNS1:
		i = 0;
		goto make_dns;

	case IPCP_DNS2:
		i = 1;
		make_dns:
		cp = put32(cp, psValue->ardwDnsAddress[i]);
		toss -= 4;
		TraceLog3("    making DNS%d address: %s",i + 1,inet_ntoa(psValue->ardwDnsAddress[i]));
		break;

	default:
		TraceLog("    making unimplemented type %d", o_type);
		break;
	};

	while ( toss-- > 0 ) {
		*cp++ = pullchar(copy_bpp);
	}
	bp->cnt += o_length;
	append(bpp, bp);
}

/************************************************************************/
/* Build a list of options */
static void near ipcp_makeoptions(MBUF **bpp,IPCP_VALUE *psValue,WORD negotiating)
{
	int o_type;

	for ( o_type = 1; o_type <= IPCP_OPTION_LIMIT; o_type++ ) {
		if (negotiating & (1 << o_type)) {
			ipcp_option( bpp, psValue,o_type, option_length[ o_type ], NULL);
		}
	}
}


/************************************************************************/
/* Build a request to send to remote host */
static MBUF * near ipcp_makereq(FSM *psFSM)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;
	MBUF *req_bp = NULL;

	ipcp_makeoptions( &req_bp, &(psIPCP->local.work),
				psIPCP->local.work.negotiate );
	return(req_bp);
}

/************************************************************************/
/* Check the options, updating the working values.
 * Returns -1 if ran out of data, ACK/NAK/REJ as appropriate.
 */
static int near ipcp_check(MBUF **bpp,FSM *psFSM,IPCP *psIPCP,IPCP_SIDE *psSide,
								OPTION_HDR *psOption,int request)
{
	int toss = psOption->len - OPTION_HDR_LEN;
	int option_result = CONFIG_ACK;		/* Assume good values */
	int i, test;
	DWORD dwRemoteAddress;

	switch(psOption->type) {
/*
	case IPCP_ADDRESSES:
		psSide->work.address = pull32(bpp);
		psSide->work.dwRemoteAddress = pull32(bpp);
		toss -= 8;
		TraceLog("    checking IP source address: %s",inet_ntoa(psSide->work.address));
		TraceLog("    checking IP destination address %s",inet_ntoa(psSide->work.dwRemoteAddress));
		if ( !request ) {
			// override any undesirable changes
			if (psIPCP->remote.want.address != 0L) {
				psIPCP->local.work.dwRemoteAddress
					= psIPCP->remote.want.address;
			}
			if (psIPCP->local.want.address != 0L) {
				psIPCP->local.work.address
					= psIPCP->local.want.address;
			}
			break;
		}

		// Ensure that addresses match
		if (psIPCP->remote.work.address == psIPCP->remote.want.address) {
			if (psIPCP->remote.want.address == 0L) {
				// don't know address either
				option_result = CONFIG_REJ;
			}
		} else if (psIPCP->remote.want.address == 0L) {
			psIPCP->local.work.dwRemoteAddress = psIPCP->remote.work.address;
		} else {
			psIPCP->remote.work.address = psIPCP->remote.want.address;
			option_result = CONFIG_NAK;
		}

		if (psIPCP->remote.work.dwRemoteAddress == psIPCP->local.want.address) {
			if (psIPCP->local.want.address == 0L) {
				// don't know address either
				option_result = CONFIG_REJ;
			}
		} else if (psIPCP->local.want.address == 0L) {
			psIPCP->local.work.address = psIPCP->remote.work.dwRemoteAddress;
		} else {
			option_result = CONFIG_NAK;
			psIPCP->remote.work.dwRemoteAddress = psIPCP->local.want.address;
		}
		break;
*/
	case IPCP_COMPRESS:
		psSide->work.compression = pull16(bpp);
		toss -= 2;
		TraceLog("    checking IP compression 0x%04x",psSide->work.compression);
		/* Check if requested type is acceptable */
		switch ( psSide->work.compression ) {
		case PPP_COMPR_PROTOCOL:
			if ( (test = pullchar(bpp)) == -1 ) {
				return -1;
			}
			if ( (psSide->work.slots = test + 1) < IPCP_SLOT_LO) {
				psSide->work.slots = IPCP_SLOT_LO;
				option_result = CONFIG_NAK;
			} else if (psSide->work.slots > IPCP_SLOT_HI) {
				psSide->work.slots = IPCP_SLOT_HI;
				option_result = CONFIG_NAK;
			}

			if ( (test = pullchar(bpp)) == -1 ) {
				return -1;
			}
			if ( (psSide->work.slot_compress = test) > 1 ) {
				psSide->work.slot_compress = 1;
				option_result = CONFIG_NAK;
			}
			toss -= 2;
			TraceLog3("    with IP compression slots %d, flag %x",psSide->work.slots,psSide->work.slot_compress);
			break;

		default:
			if ( psSide->want.negotiate & IPCP_N_COMPRESS ) {
				psSide->work.compression = psSide->want.compression;
				psSide->work.slots = psSide->want.slots;
				psSide->work.slot_compress = psSide->want.slot_compress;
			} else {
				psSide->work.compression = PPP_COMPR_PROTOCOL;
				psSide->work.slots = IPCP_SLOT_DEFAULT;
				psSide->work.slot_compress = IPCP_SLOT_COMPRESS;
			}
			option_result = CONFIG_NAK;
			break;
		};
		break;

	case IPCP_ADDRESS:
		psSide->work.dwLocalAddress = pull32(bpp);
		toss -= 4;
		TraceLog("    checking IP source address: %s",inet_ntoa(psSide->work.dwLocalAddress));
		if ( !request ) {
			break;
		}

		// Request to assign remote address
		if ((dwRemoteAddress = psIPCP->remote.want.dwLocalAddress) == 0L)
			dwRemoteAddress = ipcp_lookuppeer(psFSM->psPPP);
		if (dwRemoteAddress) {
			// we have an address for him
			if (psSide->work.dwLocalAddress != dwRemoteAddress) {
				psSide->work.dwLocalAddress = dwRemoteAddress;
				option_result = CONFIG_NAK;
			}
		}
		else if (psSide->work.dwLocalAddress == 0)
			option_result = CONFIG_REJ;	// we don't have a valid address!
		break;

	case IPCP_DNS1:
		i = 0;
		goto check_dns;

	case IPCP_DNS2:
		i = 1;
		check_dns:
		psSide->work.ardwDnsAddress[i] = pull32(bpp);
		toss -= 4;
		TraceLog3("    checking DNS%d IP source address: %s",i,inet_ntoa(psSide->work.ardwDnsAddress[i]));
		if ( !request ) {
			break;
		}

		// Request to assign remote DNS1 address
		if ((dwRemoteAddress = psIPCP->remote.want.ardwDnsAddress[i]) == 0L)
			dwRemoteAddress = GetDnsAddress(i);
		if (dwRemoteAddress) {
			// we have an address for him
			if (psSide->work.ardwDnsAddress[i] != dwRemoteAddress) {
				psSide->work.ardwDnsAddress[i] = dwRemoteAddress;
				option_result = CONFIG_NAK;
			}
		}
		else if (psSide->work.ardwDnsAddress[i] == 0)
			option_result = CONFIG_REJ;	// we don't have a valid address!
		break;

	default:
		option_result = CONFIG_REJ;
		break;
	};

	if (psOption->type > IPCP_OPTION_LIMIT
	 || !(psSide->will_negotiate & (1 << psOption->type))) {
		option_result = CONFIG_REJ;
	}

	if ( toss < 0 )
		return -1;

	if ( !request  &&  toss > 0 ) {
		/* toss extra bytes in option */
		while( toss-- > 0 ) {
			if ( pullchar(bpp) == -1 )
				return -1;
		}
	}

	return (option_result);
}


/************************************************************************/
/* Check options requested by the remote host */
static int near ipcp_request(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;
	long signed_length = config->len;
	MBUF *reply_bp = NULL;			/* reply packet */
	int reply_result = CONFIG_ACK;/* reply to request */
	WORD desired;						/* desired to negotiate */
	OPTION_HDR option;				/* option header storage */
	int option_result;				/* option reply */

	psIPCP->remote.work.negotiate = 0;	/* clear flags */

	/* Process options requested by remote host */
	while (signed_length > 0  &&  ntohoptIpcp(&option, data) != -1) {
		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			free_pz(&reply_bp);
			return -1;
		}

		if ( ( option_result = ipcp_check( data,psFSM,psIPCP,
				&(psIPCP->remote), &option, 1 ) ) == -1 ) {
			free_pz(data);
			free_pz(&reply_bp);
			return -1;
		}

		TraceLog4("IPCP REQ: result %s, option %d, length %d",fsmCodes[option_result],option.type,option.len);
		if ( option_result < reply_result ) {
			continue;
		} else if ( option_result > reply_result ) {
			/* Discard current list of replies */
			free_pz(&reply_bp);
			reply_bp = NULL;
			reply_result = option_result;
		}

		/* remember that we processed option */
		if ( option_result != CONFIG_REJ
		 && option.type <= IPCP_OPTION_LIMIT ) {
			psIPCP->remote.work.negotiate |= (1 << option.type);
		}

		/* Add option response to the return list */
		ipcp_option( &reply_bp, &(psIPCP->remote.work),
			option.type, option.len, data );
	}

	/* Now check for any missing options which are desired */
	if ( psFSM->retry_nak > 0
	 &&  (desired = psIPCP->remote.want.negotiate
				& ~psIPCP->remote.work.negotiate) != 0 ) {
		switch ( reply_result ) {
		case CONFIG_ACK:
			free_pz(&reply_bp);
			reply_bp = NULL;
			reply_result = CONFIG_NAK;
			/* fallthru */
		case CONFIG_NAK:
			ipcp_makeoptions( &reply_bp, &(psIPCP->remote.want),desired );
			psFSM->retry_nak--;
			break;
		case CONFIG_REJ:
			/* do nothing */
			break;
		};
	} else if ( reply_result == CONFIG_NAK ) {
		/* if too many NAKs, reject instead */
		if ( psFSM->retry_nak > 0 )
			psFSM->retry_nak--;
		else
			reply_result = CONFIG_REJ;
	}

	/* Send ACK/NAK/REJ to remote host */
	fsm_send(psFSM, 0, reply_result, config->id, &reply_bp);
	free_pz(data);
	return (reply_result != CONFIG_ACK);
}


/************************************************************************/
/* Process configuration ACK sent by remote host */
static int near ipcp_ack(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	MBUF *req_bp;
	int error = 0;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* Get a copy of last request we sent */
	req_bp = ipcp_makereq(psFSM);

	/* Overall buffer length should match */
	if (config->len != len_mbuf(req_bp)) {
		error = 1;
	} else {
		register int req_char;
		register int ack_char;

		/* Each byte should match */
		while ((req_char = pullchar(&req_bp)) != -1) {
			if ((ack_char = pullchar(data)) == -1
			 || ack_char != req_char ) {
				error = 1;
				break;
			}
		}
	}
	free_pz(&req_bp);
	free_pz(data);

	if (error) {
		return -1;
	}

	return 0;
}


/************************************************************************/
/* Process configuration NAK sent by remote host */
static int near ipcp_nak(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;
	IPCP_SIDE *local_p = &(psIPCP->local);
	long signed_length = config->len;
	OPTION_HDR option;
	int last_option = 0;
	int result;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* First, process in order.  Then, process extra "important" options */
	while (signed_length > 0  &&  ntohoptIpcp(&option, data) != -1) {
		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			return -1;
		}
		if ( option.type > IPCP_OPTION_LIMIT ) {
		} else if ( option.type < last_option
		 || !(local_p->work.negotiate & (1 << option.type)) ) {
			if (local_p->work.negotiate & (1 << option.type)) {
				free_pz(data);
				return -1;		/* was requested */
			}
			local_p->work.negotiate |= (1 << option.type);
			last_option = IPCP_OPTION_LIMIT + 1;
		} else {
			last_option = option.type;
		}
		if ( ( result = ipcp_check( data,psFSM,psIPCP,
				local_p, &option, 0 ) ) == -1 ) {
			free_pz(data);
			return -1;
		}
		/* update the negotiation status */
		if ( result == CONFIG_REJ
		  && option.type <= IPCP_OPTION_LIMIT ) {
			local_p->work.negotiate &= ~(1 << option.type);
		}
	}
	free_pz(data);
	return 0;
}


/************************************************************************/
/* Process configuration reject sent by remote host */
static int near ipcp_reject(FSM *psFSM,CONFIG_HDR *config,MBUF **data)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;
	IPCP_SIDE *local_p = &(psIPCP->local);
	long signed_length = config->len;
	OPTION_HDR option;
	int last_option = 0;

	/* ID field must match last request we sent */
	if (config->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* Process in order, checking for errors */
	while (signed_length > 0  &&  ntohoptIpcp(&option, data) != -1) {
		register int k;

		if ((signed_length -= option.len) < 0) {
			free_pz(data);
			return -1;
		}
		if ( option.type > IPCP_OPTION_LIMIT ) {
		} else if (option.type < last_option
		 || !(local_p->work.negotiate & (1 << option.type))) {
			free_pz(data);
			return -1;
		}
		for ( k = option.len - OPTION_HDR_LEN; k-- > 0; ) {
			if ( pullchar(data) == -1 ) {
				free_pz(data);
				return -1;
			}
		}
		last_option = option.type;

		if ( option.type <= IPCP_OPTION_LIMIT ) {
			local_p->work.negotiate &= ~(1 << option.type);
		}
	}
	free_pz(data);
	return 0;
}


/************************************************************************/
/*			I N I T I A L I Z A T I O N			*/
/************************************************************************/

/* Reset configuration options before request */
static void near IpcpReset(IPCP *psIPCP)
{
	if (psIPCP->local.work.negotiate & IPCP_N_DNS2) {
		drop_dns_address(psIPCP->local.work.ardwDnsAddress[1], 2);
	}
	if (psIPCP->local.work.negotiate & IPCP_N_DNS1) {
		drop_dns_address(psIPCP->local.work.ardwDnsAddress[0], 1);
	}
	psIPCP->local.work = psIPCP->local.want;
	psIPCP->local.will_negotiate |= psIPCP->local.want.negotiate;

	psIPCP->remote.work.negotiate = 0;
	psIPCP->remote.will_negotiate |= psIPCP->remote.want.negotiate;
}

static void near ipcp_reset(FSM *psFSM)
{
	IPCP *psIPCP =	(IPCP *)psFSM->pdv_out;
	IFACE *psIface = psFSM->psPPP->iface;

	IpcpReset(psIPCP);
	if (psIface->sflags & SF_ASSIGNED_AD) {
		psIface->sflags &= ~SF_ASSIGNED_AD;
		psIface->ip_addr = psIPCP->local.want.dwLocalAddress;
   }
	IpcpReset((IPCP *)psFSM->pdv_in);
}


/************************************************************************/
/* After termination */
#pragma argsused
static void near ipcp_stopping(FSM *psFSM)
{
}


/************************************************************************/
/* Close IPCP */
static void near ipcp_closing(FSM *psFSM)
{
	IPCP *psIPCP = 	(IPCP *)psFSM->pdv;

	/* free old slhc configuration, if any */
	slhc_free( psIPCP->slhcp );
	psIPCP->slhcp = NULL;

	if (psIPCP->remote.want.bFlags & IPCP_ADD_RT) {
		TraceLog3("%s PPP/IPCP Drop route to peer (%s)",/*ifp->name*/"",
								inet_ntoa(psIPCP->remote.work.dwLocalAddress));
		rt_drop(psIPCP->remote.work.dwLocalAddress, (unsigned int)32);
	}
}


/************************************************************************/
/* configuration negotiation complete */
static void near ipcp_opening(FSM *psFSM)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;
	PPP *psPPP = psFSM->psPPP;
	IFACE *ifp = psPPP->iface;
	DWORD address = psIPCP->local.work.dwLocalAddress;
	int rslots = 0;
	int tslots = 0;

	if (psPPP->flags & PPP_CBV) {
		fsm_close( &(psPPP->fsm[Lcp]) );
	}
	else
		psPPP->flags &= ~PPP_CBVCALL;		// that's the end of this callback
	/* Set our IP address to reflect negotiated option */
	if (!ifp->ip_addr) {
		if (address != ifp->ip_addr) {
			/* address not the same as last time */
			if (def_ip_addr == 0L) {
				/* no global address */
				def_ip_addr = address;
			} else if ( def_ip_addr == ifp->ip_addr ) {
				/* global was same as local; must be replaced */
				/* !!! TO DO: reset tcp connections */
				def_ip_addr = address;
			}
			ifp->ip_addr = address;
			ifp->sflags |= SF_ASSIGNED_AD;
		}
		TraceLog3("%s PPP/IPCP Setting new IP address: %s",ifp->name,inet_ntoa(address));
	}

	if (psIPCP->remote.want.bFlags & IPCP_ADD_RT) {
		rt_add(psIPCP->remote.work.dwLocalAddress,32,0,1,ifp,0,RT_STATIC,0);

		TraceLog3("%s PPP/IPCP Add route to peer (%s)",ifp->name,
									inet_ntoa(psIPCP->remote.work.dwLocalAddress));
	}

	if (psIPCP->local.work.negotiate & IPCP_N_DNS1) {
		add_dns_address(psIPCP->local.work.ardwDnsAddress[0], 1);
	}
	if (psIPCP->local.work.negotiate & IPCP_N_DNS2) {
		add_dns_address(psIPCP->local.work.ardwDnsAddress[1], 2);
	}
	/* free old slhc configuration, if any */
	slhc_free( psIPCP->slhcp );
	psIPCP->slhcp = NULL;

	if (psIPCP->local.work.negotiate & IPCP_N_COMPRESS) {
		rslots = psIPCP->local.work.slots;
	}
	if (psIPCP->remote.work.negotiate & IPCP_N_COMPRESS) {
		tslots = psIPCP->remote.work.slots;
	}

	if ( rslots != 0 || tslots != 0 ) {
		psIPCP->slhcp = slhc_init( rslots, tslots );
		BUFTRACE("slhc_init",psIPCP->slhcp);
		BUFTRACE("slhc_init->tstate",psIPCP->slhcp->tstate);
		BUFTRACE("slhc_init->rstate",psIPCP->slhcp->rstate);
		TraceLog6("%s PPP/IPCP Compression enabled; Recv slots = %d, flag = %x; Xmit slots = %d, flag = %x",ifp->name,rslots,psIPCP->local.work.slot_compress,tslots,psIPCP->remote.work.slot_compress);
	}
	if (ifp->flags & DHCPI_FLAG)
		dhcp_inform(ifp);
}

/************************************************************************
// Check the address against all other assigned addresses
static int32 ipcp_addr_idle(int32 addr)
{
	IFACE *ifp;

	// Check if peer IP address is already in use on another interface
	// !!! need to look at *remote* address, not local!
	for (ifp=ifaces; ifp != NULL; ifp = ifp->next) {
		if (ifp->ip_addr == addr)
			return 0L;
	}
	return addr;
}

// ************************************************************************
// Assign the next unused address from a pool
static int32 ipcp_poolnext(IPCP *psIPCP)
{
	int32 i = 1L + psIPCP->peer_max - psIPCP->peer_min;
	int32 nextaddr = 0L;

	while ( i-- > 0  &&  nextaddr == 0L ) {
		if (++psIPCP->local.want.dwRemoteAddress < psIPCP->peer_min
		 || psIPCP->local.want.dwRemoteAddress > psIPCP->peer_max)
			psIPCP->local.want.dwRemoteAddress = psIPCP->peer_min;

		nextaddr = ipcp_addr_idle(psIPCP->local.want.dwRemoteAddress);
	}
	return(nextaddr);
}
*/

/************************************************************************/
/* Check if we have a specific IP address to assign to remote peer host */
/* !!! TO DO: subnet mask, and routing */
static DWORD near ipcp_lookuppeer(PPP *psPPP)
{
	USERPASS *psU;

	if ((psU = FindUserPass((char *)psPPP->peername,0)) != 0) {
		if ((psU->wFlags & UP_CALLBACK) && !(psPPP->flags & PPP_CBVCALL))
			psPPP->flags |= PPP_CBV;
		psPPP->flags &= ~PPP_CBVCALL;
		return psU->dwIpAddress;
	}
	return 0l;
}


/************************************************************************/
/* Prepare to begin configuration exchange */
#pragma argsused
static void near ipcp_starting(FSM *psFSM)
{
/*
	IPCP *psIPCP =		psFSM->pdv;
	IFACE *ifp;

	ifp = psFSM->psPPP->iface;
	if (!(ifp->sflags & SF_ASSIGNED_AD))
		psIPCP->local.want.dwLocalAddress = ifp->ip_addr;
*/
}


/************************************************************************/
static void near ipcp_free(FSM *psFSM)
{
	IPCP *psIPCP = (IPCP *)psFSM->pdv;

	slhc_free( psIPCP->slhcp );
}

//  PPPPAP.C	-- Password Authentication Protocol for PPP

static void near ap_free(FSM *psFSM);

FSM_CONSTANT ap_constants = {
	"Ap",
	PPP_PAP_PROTOCOL,
	0x000E,				/* codes 1-3 recognized */

	Ap,
	PAP_REQ_TRY,
	PAP_FAIL_MAX,
	0,
	PAP_TIMEOUT * 1000L,

	ap_free,

	fsm_no_action,		/* ap_reset, */
	fsm_no_action,		/* ap_starting, */
	fsm_no_action,		/* ap_opening, */
	fsm_no_action,		/* ap_closing, */
	fsm_no_action,		/* ap_stopping, */

	0,						// ap_makereq,
	fsm_no_check,		/* ap_request, */
	fsm_no_check,		/* ap_ack, */
	fsm_no_check,		/* ap_nak, */
	fsm_no_check,		/* ap_reject */
};


/****************************************************************************/
// send Authentication Protocol packet
static int near ap_send(FSM *psFSM, char cCommand)
{
	MBUF *bp;
	AP *psAP = (AP *)psFSM->pdv;
	BYTE *cp,bSaveId;
	int len;
	int *piChal,i,retv;
	static int iSeq;

	if (cCommand == CONFIG_REQ) {
		len = 2 + strlen(psAP->psApUsername) + strlen(psAP->psApPassword);
	}
	else if (cCommand == SEND_CHALLENGE) {
		len = CHALLENGE_LENGTH + 2;
	}
	else if (cCommand == CHAP_RESPONSE) {
		len = MD5_DIGEST_LENGTH + strlen(psAP->psApUsername) + 1;
	}

	fsm_timer(psFSM);
	if ((bp = alloc_mbuf(len,AllocPppAsnd)) == NULL)
		return -1;
	if (cCommand == CONFIG_REQ) {
		/* Load user id and password for authenticate packet */
		cp = bp->data;
		*cp++ = (BYTE)strlen(psAP->psApUsername);
		if ( strlen(psAP->psApUsername) > 0 )
			cp = (BYTE *)stpcpy((char *)cp, psAP->psApUsername);

		*cp++ = strlen(psAP->psApPassword);
		if ( strlen(psAP->psApPassword) > 0 )
			cp = (BYTE *)stpcpy((char *)cp, psAP->psApPassword);

		bp->cnt += len;
		psFSM->state |= AP_REQUEST_SENT;
		return fsm_send(psFSM, 0, cCommand, 0, &bp);
	}
	else if (cCommand == SEND_CHALLENGE) {
		free(psAP->pcChallenge);
		if ((piChal = (int *)psAP->pcChallenge = malloc(CHALLENGE_LENGTH,AllocPppChal)) == 0)
			return -1;
		srand(peek(0x40,0x6c) + iSeq++);
		for (i = 0;i < CHALLENGE_LENGTH / 2;++i)
			*piChal++ = rand();
		*bp->data = CHALLENGE_LENGTH;
	#ifdef CHALLENGE_DEBUG
		memcpy(psAP->pcChallenge,cChal,CHALLENGE_LENGTH); // debug!
		psFSM->psPPP->id = 0;	// debug!
	#endif
		memcpy(bp->data + 1,psAP->pcChallenge,CHALLENGE_LENGTH);
		bp->data[CHALLENGE_LENGTH + 1] = 'S';	// S for Sockets
		bp->cnt = CHALLENGE_LENGTH + 2;
		bSaveId = psFSM->lastid;
		retv = fsm_send(psFSM, PPP_CHAP_PROTOCOL, CHAP_CHALLENGE, 0, &bp);
		psAP->bChallengeId = psFSM->lastid;
		psFSM->lastid = bSaveId;
		return retv;
	}
	else if (cCommand == CHAP_RESPONSE) {
		*bp->data = MD5_DIGEST_LENGTH;
		memcpy(bp->data + 1,psAP->pcResponse + 1,MD5_DIGEST_LENGTH);
		memcpy(bp->data + 1 + MD5_DIGEST_LENGTH,psAP->psApUsername,
											strlen(psAP->psApUsername));
		bp->cnt = len;
		return fsm_send(psFSM, PPP_CHAP_PROTOCOL, CHAP_RESPONSE, *psAP->pcResponse, &bp);
	}
	return -1;
}

/****************************************************************************/

/* abandon AP attempt; shutdown LCP layer */
static void ap_shutdown(FSM *psFSM)
{
	PPP *psPPP = psFSM->psPPP;

	FsmLog( "Failed; close connection" );
	fsm_close( &(psPPP->fsm[Lcp]) );
}

static void near ap_reset(FSM *psFSM)
{
	psFSM->state = 0;
	psFSM->retry = psFSM->bChallengeRetry = psFSM->bResponseRetry = psFSM->try_req;
}


/* Configuration negotiation complete */
static void ap_opening(FSM *psFSM,int flag)
{
	register PPP *psPPP = psFSM->psPPP;

	FsmLog( "Open");

	if ( !((psFSM->flags &= ~flag) & (PPP_AP_LOCAL | PPP_AP_REMOTE)) ) {
		psFSM->state |= AP_OPENED;
	}
	psPPP->flags &= ~flag;
	ppp_ready(psPPP);
}

/****************************************************************************/
/* Check request from remote host */
static int CheckApRequest(FSM *psFSM,CONFIG_HDR *hdr,MBUF **data)
{
	MBUF *reply_bp;
	int result;
	BYTE *message;
	int mess_length;
	char *username = 0;
	int userlen;
	char *password = 0;
	int passwordlen;

	/* Extract userID/password sent by remote host */
	if ( (userlen = pullchar(data)) != -1 ) {
		register int i;
		char *cp;

		cp = username = (char *)malloc(userlen+1,AllocPppUsr);
		for ( i = userlen; i-- > 0; ) {
			*cp++ = pullchar(data);
		}
		*cp = '\0';
	}

	TraceLog("    checking user: %s", username);
	if ( (passwordlen = pullchar(data)) != -1 ) {
		register int i;
		char *cp;

		cp = password = (char *)malloc(passwordlen+1,AllocPppPass);
		for ( i = passwordlen; i-- > 0; ) {
			*cp++ = pullchar(data);
		}
		*cp = '\0';
	}

	TraceLog("    checking password: %s", password);
	if (FindUserPass(username,password) != 0) {
		free( psFSM->psPPP->peername );
		psFSM->psPPP->peername = (BYTE *)strdup((char *)username,AllocPppPeerName);
		result = CONFIG_ACK;
		message = (BYTE *)" Welcome";
	} else {
		result = CONFIG_NAK;
		message = (BYTE *)" Invalid username or password";
	}

	/* the space at the beginning of the message is crucial */
	/* it is replaced with the length of the message */
	mess_length = strlen((char *)message);
	reply_bp = qdata(message,mess_length,AllocPppMsg);
	reply_bp->data[0] = (char)(mess_length - 1);

	fsm_send(psFSM, 0, result, hdr->id, &reply_bp);

	if (result == CONFIG_NAK) {
		ap_shutdown(psFSM);
	}
	free_pz(data);
	free(username);
	free(password);
	return (result != CONFIG_ACK);
}

/* Check acknowledgement from remote host */
static int CheckApAck(FSM *psFSM,CONFIG_HDR *hdr,MBUF **data)
{
	AP *psAP = (AP *)psFSM->pdv;
	char *message;
	int mess_length;
	int full_length;
	int len;

	/* ID field must match last request we sent */
	if (hdr->id != psFSM->lastid) {
		free_pz(data);
		return -1;
	}

	/* Log ASCII message from remote host, if any */
	if ( (mess_length = pullchar(data)) != -1 ) {
		message = (char *)malloc( mess_length+1,AllocPppRMsg);
		full_length = len_mbuf(*data);
		len = pullup(data, (BYTE *)message, mess_length);
		message[len] = '\0';

		free( psAP->psApMessage);
		psAP->psApMessage = message;

		TraceLog5("%s PPP/AP %s %s: %s",psFSM->psPPP->iface->name,(len < mess_length) ? "Short" : (mess_length < full_length) ? "Long" : "Valid",(hdr->code == CONFIG_ACK) ? "Ack" : "Nak",message);
		return (len < mess_length  ||  mess_length < full_length);
	}
	free_pz(data);
	return -1;
}

/****************************************************************************/
/* Process challenge from remote host */
static int CheckApChallenge(FSM *psFSM,CONFIG_HDR *hdr,MBUF **data)
{
	BYTE *pValue = NULL;
	int len;
	MD5_CTX sMD5_CTX;
	AP *psAP = (AP *)psFSM->pdv;

	/* Extract value sent by remote host */
	if ( (len = pullchar(data)) == -1 ||
							(pValue = (BYTE *)malloc(len,AllocPppRChal)) == 0) {
		free_pz(data);
		return -1;
	}
	MD5_Init(&sMD5_CTX);
	MD5_Update(&sMD5_CTX,&hdr->id,1);
	MD5_Update(&sMD5_CTX,(BYTE *)psAP->psApPassword,strlen(psAP->psApPassword));
	pullup(data,(BYTE *)pValue,len);
	MD5_Update(&sMD5_CTX,(BYTE *)pValue,len);
	free_pz(data);
	free(pValue);
	if ((pValue = (BYTE *)malloc(MD5_DIGEST_LENGTH + 1,AllocPppMd5)) == 0)
		return -1;
	*pValue = hdr->id;
	MD5_Final(pValue + 1,&sMD5_CTX);
	free(psAP->pcResponse);
	psAP->pcResponse = (BYTE *)pValue;
	psFSM->state |= AP_RESPONSE_SENT;
	ap_send(psFSM,CHAP_RESPONSE);
	return 0;
}

int CheckApResponse(FSM *psFSM,CONFIG_HDR *hdr,MBUF **data)
{
	MBUF *bp;
	PPP *psPPP = psFSM->psPPP;
	AP *psAP = (AP *)psFSM->pdv;
	int len;
	char cResult = CHAP_FAILURE;
	char *pszMessage;
	BYTE pcDigest[MD5_DIGEST_LENGTH];
	BYTE pcLocalDigest[MD5_DIGEST_LENGTH];
	MD5_CTX sMD5_CTX;
	USERPASS *psU;

	if ( (len = pullchar(data)) < MD5_DIGEST_LENGTH || len_mbuf(*data) <= MD5_DIGEST_LENGTH) {
		pszMessage = "Protocol format error";
	}
	else {
		pullup(data,pcDigest,MD5_DIGEST_LENGTH);
		len = len_mbuf(*data);
		free(psPPP->peername);
		if ((psPPP->peername = malloc(len + 1,AllocPppPeer)) != 0) {
			if ((psFSM->state &= ~AP_CHALLENGE_SENT) == 0)
				stop_timer(&psFSM->timer);
			len = pullup(data,psPPP->peername,len);
			psPPP->peername[len] = 0;
			if ((psU = FindUserPass((char *)psPPP->peername,0)) != 0) {
				BYTE *pszPass = (BYTE *)psU->szUserPass + strlen(psU->szUserPass) + 1;
				// Calculate and compare MD5 hash value
				MD5_Init(&sMD5_CTX);
				MD5_Update(&sMD5_CTX,&hdr->id,1);
				MD5_Update(&sMD5_CTX,pszPass,strlen((char *)pszPass));
				MD5_Update(&sMD5_CTX,psAP->pcChallenge,CHALLENGE_LENGTH);
				MD5_Final(pcLocalDigest,&sMD5_CTX);
				if (memcmp(pcDigest,pcLocalDigest,MD5_DIGEST_LENGTH) == 0) {
					pszMessage = "Welcome";
					cResult = CHAP_SUCCESS;
				}
				else goto fail;
			}
			else {
				fail:
				pszMessage = "No such user";
			}
		}
	}
	free_pz(data);
	len = strlen(pszMessage);
	if ((bp = alloc_mbuf(len,AllocPppAMsg)) == 0) {
		return -1;
	}
	memcpy(bp->data,pszMessage,len);
	bp->cnt = len;
	fsm_send(psFSM, PPP_CHAP_PROTOCOL, cResult, hdr->id, &bp);
	free_p(bp);
	if (cResult == CHAP_SUCCESS)
		ap_opening(psFSM, PPP_AP_LOCAL);
	return 0;
}

/************************************************************************/
/*			E V E N T   P R O C E S S I N G			*/
/************************************************************************/

/* Process incoming packet */
static void near ap_proc(FSM *psFSM,MBUF **bpp,WORD protocol)
{
	//AP *psAP = psFSM->pdv;
	CONFIG_HDR hdr;

	if ( ntohcnf(&hdr, bpp) == -1 ) {
		FsmLog( "short authentication packet" );
	}
	TraceLog6("%s PPP/%s Recv,  option: %s, id: %d, len: %d",psFSM->psPPP->iface->name,psFSM->pdc->name,fsmCodes[hdr.code],hdr.id,hdr.len);
	hdr.len -= CONFIG_HDR_LEN;		/* Length includes envelope */
	trim_mbuf(bpp, hdr.len);		/* Trim off padding */

	if (protocol == PPP_PAP_PROTOCOL) {
		switch(hdr.code) {
		case CONFIG_REQ:
			if ( CheckApRequest(psFSM, &hdr, bpp) == 0) {
				ap_opening(psFSM, PPP_AP_LOCAL);
			}
			break;

		case CONFIG_ACK:
			if (CheckApAck(psFSM, &hdr, bpp) == 0) {
				if ((psFSM->state &= ~AP_REQUEST_SENT) == 0)
					stop_timer(&(psFSM->timer));
				ap_opening(psFSM, PPP_AP_REMOTE);
			}
			break;

		case CONFIG_NAK:
			if (CheckApAck(psFSM, &hdr, bpp) == 0) {	// set message

				/* Must have sent a bad username or password */
				ap_shutdown(psFSM);
			}
			break;

		default:
			goto default1;
		}
	}
	else /*if (protocol == PPP_CHAP_PROTOCOL)*/ {
		switch(hdr.code) {
			case CHAP_CHALLENGE:
				CheckApChallenge(psFSM, &hdr, bpp);
				break;
			case CHAP_RESPONSE:
				CheckApResponse(psFSM, &hdr, bpp);
				break;
			case CHAP_SUCCESS:
				if ((psFSM->state &= ~AP_RESPONSE_SENT) == 0)
					stop_timer(&psFSM->timer);
				ap_opening(psFSM, PPP_AP_REMOTE);
				break;
			case CHAP_FAILURE:
				ap_shutdown(psFSM);
				break;
			default:
				default1:
				TraceLog3("%s PPP/Ap Unknown packet type: %d; dropping packet",psFSM->psPPP->iface->name,hdr.code);
				free_pz(bpp);
				break;
		}
	}
}


/* Timeout while waiting for reply from remote host */
void ap_timeout(void *vp)
{
	FSM *psFSM = (FSM *)vp;
	AP *psAP = (AP *)psFSM->pdv;

	FsmLog( "Timeout" );

	if (psFSM->state & AP_RESPONSE_SENT) {
		if (--psFSM->bResponseRetry > 0)
			ap_send(psFSM,CHAP_RESPONSE);
		else
			ap_shutdown(psFSM);
		return;
	}
	if (psFSM->state & AP_CHALLENGE_SENT) {
		if (--psFSM->bChallengeRetry > 0)
			ap_send(psFSM,SEND_CHALLENGE);
		else
			ap_shutdown(psFSM);
		return;
	}
	if (psFSM->state & AP_REQUEST_SENT) {
		if (--psFSM->retry > 0) {
			free ( psAP->psApMessage );
			psAP->psApMessage = strdup("Request timeout",AllocPppApMsg);
			ap_send(psFSM,CONFIG_REQ);
		} else {
			free ( psAP->psApMessage );
			psAP->psApMessage = strdup("Request retry exceeded",AllocPppApMsg);
			FsmLog( "Request retry exceeded");
			ap_shutdown(psFSM);
		}
	}
}


/************************************************************************/
/*			I N I T I A L I Z A T I O N			*/
/************************************************************************/

static void near ap_down(FSM *psFSM)
{
	AP *psAP = (AP *)psFSM->pdv;

	if ( psAP == NULL )
		return;

	FsmLog( "Down");
	psFSM->flags = 0;

	if ( psFSM->state & (AP_REQUEST_SENT|AP_CHALLENGE_SENT|AP_RESPONSE_SENT))
		stop_timer(&(psFSM->timer));
	ap_reset(psFSM);
	psFSM->state = 0;
}


static void near ap_free(FSM *psFSM)
{
	AP *psAP = (AP *)psFSM->pdv;

	free(psAP->psApMessage);
	free(psAP->pcChallenge);
   free(psAP->pcResponse);
}

//  PPPFSM.C	-- PPP Finite State Machine

char *fsmStates[] = {
	"Closed",
	"Listen",
	"Req Sent",
	"Ack Rcvd",
	"Ack Sent",
	"Opened",
	"TermSent"
};

char *fsmCodes[] = {
	NULL,
	"Config Req",
	"Config Ack",
	"Config Nak",
	"Config Reject",
	"Termin Req",
	"Termin Ack",
	"Code Reject",
	"Protocol Reject",
	"Echo Request",
	"Echo Reply",
	"Discard Request",
};

static int fsm_sendtermreq(FSM *psFSM);
static int fsm_sendtermack(FSM *psFSM, BYTE id);

static void fsm_timeout(void *vp);

static void fsm_opening(FSM *psFSM);


/************************************************************************/
/* Convert header in host form to network form */
void htoncnf(CONFIG_HDR *cnf,MBUF **bpp)
{
	register BYTE *cp;

	/* Prepend bytes for LCP/IPCP header */
	prepend(bpp,CONFIG_HDR_LEN);

	/* Load header with proper values */
	cp = (*bpp)->data;
	*cp++ = cnf->code;
	*cp++ = cnf->id;
	put16(cp, cnf->len);
}

/* Extract header from incoming packet */
int ntohcnf(CONFIG_HDR *cnf,MBUF **bpp)
{
	BYTE cnfb[CONFIG_HDR_LEN];

	if ( cnf == NULL )
		return -1;

	if ( pullup( bpp, cnfb, CONFIG_HDR_LEN ) < CONFIG_HDR_LEN )
		return -1;

		cnf->code = cnfb[0];
		cnf->id = cnfb[1];
	cnf->len = get16(&cnfb[2]);
	return 0;
}

/***************************************/
/* Extract configuration option header */
static near int ntohopt(OPTION_HDR *opt,MBUF **bpp)
{
	BYTE optb[OPTION_HDR_LEN];

	if ( opt == NULL )
		return -1;

	if ( pullup( bpp, optb, OPTION_HDR_LEN ) < OPTION_HDR_LEN )
		return -1;

	opt->type = optb[0];
	opt->len = optb[1];
	return 0;
}

static near int ntohoptIpcp(OPTION_HDR *opt,MBUF **bpp)
{
	int iRetv;

	if ((iRetv = ntohopt(opt,bpp)) == 0) {
		if (opt->type == IPCP_PRI_DNS)
			opt->type = IPCP_DNS1;
		else if (opt->type == IPCP_SEC_DNS)
			opt->type = IPCP_DNS2;
	}
	return iRetv;
}


/************************************************************************/
#pragma argsused
static void near fsm_no_action(FSM *psFSM)
{
}

#pragma argsused
static int near fsm_no_check(FSM *psFSM,CONFIG_HDR *hdr,MBUF **bpp)
{
	return 0;
}

/************************************************************************/
/* Set a timer in case an expected event does not occur */
/*
void fsm_timer(FSM *psFSM)
{
	start_timer( &(psFSM->timer) );
}
*/

/************************************************************************/
/* Send a packet to the remote host */
static int near fsm_send(FSM *psFSM,WORD protocol, BYTE code,BYTE id,MBUF **data)
{
	PPP *psPPP = psFSM->psPPP;
	IFACE *iface = psPPP->iface;
	CONFIG_HDR hdr;
	LCP *psLCP;
	MBUF *bp;

	switch( hdr.code = code ) {
	case CONFIG_REQ:
	case TERM_REQ:
	case ECHO_REQ:
		/* Save ID field for match against replies from remote host */
		psFSM->lastid = psPPP->id;
		/* fallthru */
	case PROT_REJ:
	case DISCARD_REQ:
		/* Use a unique ID field value */
		hdr.id = psPPP->id++;
		break;

	case CONFIG_ACK:
	case CONFIG_NAK:
	case CONFIG_REJ:
	case TERM_ACK:
	case CODE_REJ:
	case ECHO_REPLY:
		/* Use ID sent by remote host */
		hdr.id = id;
		break;

	default:
		/* we're in trouble */
		TraceLog5("%s PPP/%s %-8s; Send with bogus code: %d",iface->name,psFSM->pdc->name,fsmStates[psFSM->state],code);
		free_pz(data);
		return -1;
	}

	switch( code ) {
	case ECHO_REQ:
	case ECHO_REPLY:
	case DISCARD_REQ:
		psLCP = (LCP *)psFSM->pdv;
		if ((bp = alloc_mbuf(4,AllocPppFSnd)) == 0)
			return -1;
		BUFTRACE("fsm_send",bp);
		put32(bp->data, psLCP->local.work.magic_number);
		bp->next = *data;
		*data = bp;
		break;
	}
	hdr.len = len_mbuf(*data) + CONFIG_HDR_LEN;
	if (data == 0) {
		bp = 0;
		data = &bp;
	}

	/* Prepend header to packet data */
	htoncnf(&hdr,data);
	TraceLog7("%s PPP/%s %-8s; Sending %s, id: %d, len: %d",iface->name,psFSM->pdc->name,fsmStates[psFSM->state],fsmCodes[code],hdr.id,hdr.len);
	psPPP->OutNCP[psFSM->pdc->fsmi]++;
	if (!protocol)
		protocol = psFSM->pdc->protocol;
	((*iface->output)(iface,NULL,NULL,protocol,*data,0));
	*data = 0;
	return 0;
}


/************************************************************************/
/* Send a configuration request */
static int near fsm_sendreq(FSM *psFSM)
{
	MBUF *bp;

	if ( psFSM->retry <= 0 )
		return -1;

	fsm_timer(psFSM);
	if (((SER *)psFSM->psPPP->iface->dcb)->flags & FL_CONNECTED) {
		psFSM->retry--;
	}
	//printf("%s retry=%d\n",psFSM->pdc->name,psFSM->retry);
	bp = (*psFSM->pdc->makereq)(psFSM);
	return fsm_send(psFSM, 0, CONFIG_REQ, 0, &bp);
}


/************************************************************************/
/* Send a termination request */
static int fsm_sendtermreq(FSM *psFSM)
{
	if ( psFSM->retry <= 0 )
		return -1;

	psFSM->retry--;
	fsm_timer(psFSM);
	return( fsm_send(psFSM, 0, TERM_REQ, 0, NULL) );
}


/************************************************************************/
/* Send Terminate Ack */
static int fsm_sendtermack(FSM *psFSM,BYTE id)
{
	return( fsm_send(psFSM, 0, TERM_ACK, id, NULL) );
}


/************************************************************************/
/* Reset state machine */
static void near fsm_reset(FSM *psFSM)
{
	psFSM->state = (psFSM->flags & FSM_PASSIVE) ? fsmLISTEN : fsmCLOSED;
	psFSM->retry = psFSM->try_req;
	psFSM->retry_nak = psFSM->try_nak;

	(*psFSM->pdc->reset)(psFSM);
}


/************************************************************************/
/* Configuration negotiation complete */
static void fsm_opening(FSM *psFSM)
{
	FsmLog( "Opened");
	stop_timer(&(psFSM->timer));

	psFSM->state = fsmOPENED;	// moved here for CBV
	(*psFSM->pdc->opening)(psFSM);
}


/************************************************************************/
/*			E V E N T   P R O C E S S I N G			*/
/************************************************************************/

/* Process incoming packet */
static void near fsm_proc(FSM *psFSM,MBUF **bpp)
{
	CONFIG_HDR hdr;

	if ( ntohcnf(&hdr, bpp) == -1 ) {
		FsmLog( "short configuration packet" );
	}
	TraceLog7("%s PPP/%s %-8s; Processing %s, id: %d, len: %d",psFSM->psPPP->iface->name,psFSM->pdc->name,fsmStates[psFSM->state],fsmCodes[hdr.code],hdr.id,	hdr.len);
	hdr.len -= CONFIG_HDR_LEN;		/* Length includes envelope */
	trim_mbuf(bpp, hdr.len);		/* Trim off padding */

	switch(hdr.code) {
	case CONFIG_REQ:
		switch(psFSM->state) {
		case fsmOPENED:		/* Unexpected event? */
			(*psFSM->pdc->closing)(psFSM);
			fsm_reset(psFSM);
			goto handle_config_req;
		case fsmLISTEN:
			if (psFSM == &((psFSM->psPPP)->fsm[Lcp]))
				SetPdv(psFSM,1);
			handle_config_req:
			(*psFSM->pdc->starting)(psFSM);
			fsm_sendreq(psFSM);
			/* fallthru */
		case fsmREQ_Sent:
		case fsmACK_Sent:	/* Unexpected event? */
			psFSM->state =
			((*psFSM->pdc->request)(psFSM, &hdr, bpp) == 0)
				? fsmACK_Sent : fsmREQ_Sent;
//			fsm_timer(psFSM);	// give peer time to respond
			break;

		case fsmACK_Rcvd:
			if ((*psFSM->pdc->request)(psFSM, &hdr, bpp) == 0) {
				fsm_opening(psFSM);
			} else {
				/* give peer time to respond */
				fsm_timer(psFSM);
			}
			break;

		case fsmCLOSED:
			/* Don't accept any connections */
			fsm_sendtermack(psFSM, hdr.id);
			/* fallthru */
		case fsmTERM_Sent:
			/* We are attempting to close connection; */
			/* wait for timeout to resend a Terminate Request */
			//free_pz(bpp);
			break;
		};
		break;

	case CONFIG_ACK:
		switch(psFSM->state) {
		case fsmREQ_Sent:
			if ((*psFSM->pdc->ack)(psFSM, &hdr, bpp) == 0) {
				psFSM->state = fsmACK_Rcvd;
			}
			break;

		case fsmACK_Sent:
			if ((*psFSM->pdc->ack)(psFSM, &hdr, bpp) == 0) {
				fsm_opening(psFSM);
			}
			break;

		case fsmOPENED:		/* Unexpected event? */
			(*psFSM->pdc->closing)(psFSM);
			(*psFSM->pdc->starting)(psFSM);
			fsm_reset(psFSM);
			/* fallthru */
		case fsmACK_Rcvd:	/* Unexpected event? */
			//free_pz(bpp);
			fsm_sendreq(psFSM);
			psFSM->state = fsmREQ_Sent;
			break;

		case fsmCLOSED:
		case fsmLISTEN:
			/* Out of Sync; kill the remote */
			fsm_sendtermack(psFSM, hdr.id);
			/* fallthru */
		case fsmTERM_Sent:
			/* We are attempting to close connection; */
			/* wait for timeout to resend a Terminate Request */
			//free_pz(bpp);
			break;
		};
		break;

	case CONFIG_NAK:
		switch(psFSM->state) {
		case fsmREQ_Sent:
		case fsmACK_Sent:
			/* Update our config request to reflect NAKed options */
			if ((*psFSM->pdc->nak)(psFSM, &hdr, bpp) == 0) {
				/* Send updated config request */
				fsm_sendreq(psFSM);
			}
			break;

		case fsmOPENED:		/* Unexpected event? */
			(*psFSM->pdc->closing)(psFSM);
			(*psFSM->pdc->starting)(psFSM);
			fsm_reset(psFSM);
			/* fallthru */
		case fsmACK_Rcvd:	/* Unexpected event? */
			//free_pz(bpp);
			fsm_sendreq(psFSM);
			psFSM->state = fsmREQ_Sent;
			break;

		case fsmCLOSED:
		case fsmLISTEN:
			/* Out of Sync; kill the remote */
			fsm_sendtermack(psFSM, hdr.id);
			/* fallthru */
		case fsmTERM_Sent:
			/* We are attempting to close connection; */
			/* wait for timeout to resend a Terminate Request */
			//free_pz(bpp);
			break;
		};
		break;

	case CONFIG_REJ:
		switch(psFSM->state) {
		case fsmREQ_Sent:
		case fsmACK_Sent:
			if((*psFSM->pdc->reject)(psFSM, &hdr, bpp) == 0) {
				fsm_sendreq(psFSM);
			}
			break;

		case fsmOPENED:		/* Unexpected event? */
			(*psFSM->pdc->closing)(psFSM);
			(*psFSM->pdc->starting)(psFSM);
			fsm_reset(psFSM);
			/* fallthru */
		case fsmACK_Rcvd:	/* Unexpected event? */
			//free_pz(bpp);
			fsm_sendreq(psFSM);
			psFSM->state = fsmREQ_Sent;
			break;

		case fsmCLOSED:
		case fsmLISTEN:
			/* Out of Sync; kill the remote */
			fsm_sendtermack(psFSM, hdr.id);
			/* fallthru */
		case fsmTERM_Sent:
			/* We are attempting to close connection; */
			/* wait for timeout to resend a Terminate Request */
			//free_pz(bpp);
			break;
		};
		break;

	case TERM_REQ:
		FsmLog( "Peer requested Termination");
		switch(psFSM->state) {
		case fsmOPENED:
			fsm_sendtermack(psFSM, hdr.id);
			(*psFSM->pdc->closing)(psFSM);
			(*psFSM->pdc->stopping)(psFSM);
			fsm_reset(psFSM);
			break;

		case fsmACK_Rcvd:
		case fsmACK_Sent:
			psFSM->state = fsmREQ_Sent;
			/* fallthru */
		case fsmREQ_Sent:
		case fsmTERM_Sent:
			/* waiting for timeout */
			/* fallthru */
		case fsmCLOSED:
		case fsmLISTEN:
			/* Unexpected, but make them happy */
			fsm_sendtermack(psFSM, hdr.id);
			break;
		};
		break;

	case TERM_ACK:
		switch(psFSM->state) {
		case fsmTERM_Sent:
			stop_timer(&(psFSM->timer));
			FsmLog( "Terminated");
			(*psFSM->pdc->stopping)(psFSM);
			fsm_reset(psFSM);
			break;

		case fsmOPENED:
			/* Remote host has abruptly closed connection */
			FsmLog( "Terminated unexpectly");
			(*psFSM->pdc->closing)(psFSM);
			fsm_reset(psFSM);
			if ( fsm_sendreq(psFSM) == 0 ) {
				psFSM->state = fsmREQ_Sent;
			}
			break;

		case fsmACK_Sent:
		case fsmACK_Rcvd:
			psFSM->state = fsmREQ_Sent;
			/* fallthru */
		case fsmREQ_Sent:
			/* waiting for timeout */
			/* fallthru */
		case fsmCLOSED:
		case fsmLISTEN:
			/* Unexpected, but no action needed */
			break;
		};
		break;

	case CODE_REJ:
		TraceLog3("%s PPP/%s Code Reject; indicates faulty implementation",psFSM->psPPP->iface->name,psFSM->pdc->name);
		(*psFSM->pdc->stopping)(psFSM);
		fsm_reset(psFSM);
		//free_pz(bpp);
		break;

	case PROT_REJ:
		TraceLog3("%s PPP/%s Protocol Reject; please do not use this protocol",psFSM->psPPP->iface->name,psFSM->pdc->name);
		//free_pz(bpp);
		break;

	case ECHO_REQ:
		switch(psFSM->state) {
		case fsmOPENED:
			fsm_send( psFSM, 0, ECHO_REPLY, hdr.id, bpp );
			break;

		case fsmCLOSED:
		case fsmLISTEN:
			/* Out of Sync; kill the remote */
			fsm_sendtermack(psFSM, hdr.id);
			/* fallthru */
		case fsmREQ_Sent:
		case fsmACK_Rcvd:
		case fsmACK_Sent:
		case fsmTERM_Sent:
			/* ignore */
			//free_pz(bpp);
			break;
		};
		break;

	case ECHO_REPLY:
	case DISCARD_REQ:
	case QUALITY_REPORT:
		//free_pz(bpp);
		break;

	default:
		TraceLog4("%s PPP/%s Unknown packet type: %d; Sending Code Reject",psFSM->psPPP->iface->name,psFSM->pdc->name,hdr.code);
		hdr.len += CONFIG_HDR_LEN;	/* restore length */
		htoncnf( &hdr, bpp );	/* put header back on */
		fsm_send( psFSM, 0, CODE_REJ, hdr.id, bpp );

		switch(psFSM->state) {
		case fsmREQ_Sent:
		case fsmACK_Rcvd:
		case fsmACK_Sent:
		case fsmOPENED:
			psFSM->state = (psFSM->flags & FSM_PASSIVE) ? fsmLISTEN : fsmCLOSED;
			break;

		case fsmCLOSED:
		case fsmLISTEN:
		case fsmTERM_Sent:
			/* no change */
			break;
		};
		break;
	}
	free_pz(bpp);		// catch all remaining memory leaks!
}


/************************************************************************/
/* Timeout while waiting for reply from remote host */
static void fsm_timeout(void *vp)
{
	FSM *psFSM = (FSM *)vp;

	FsmLog( "Timeout" );
	switch(psFSM->state) {
	case fsmREQ_Sent:
	case fsmACK_Rcvd:
	case fsmACK_Sent:
		if (psFSM->retry > 0) {
			fsm_sendreq(psFSM);
			if (psFSM->state == fsmACK_Rcvd)
				psFSM->state = fsmREQ_Sent;
		} else {
			FsmLog( "Request retry exceeded");
			fsm_reset(psFSM);
		}
		break;

	case fsmTERM_Sent:
		if (psFSM->retry > 0) {
			fsm_sendtermreq(psFSM);
		} else {
			FsmLog( "Terminate retry exceeded");
			(*psFSM->pdc->stopping)(psFSM);
			fsm_reset(psFSM);
		}
		break;

	case fsmCLOSED:
	case fsmLISTEN:
	case fsmOPENED:
		/* nothing to do */
		break;
	}
}


/************************************************************************/
/*			I N I T I A L I Z A T I O N			*/
/************************************************************************/

/************************************************************************/
/* Initialize the fsm for this protocol
 * Called from protocol _init
 */
void fsm_init(FSM *psFSM)
{
	TIMER *t = &(psFSM->timer);

	psFSM->try_req = psFSM->pdc->try_req;
	psFSM->try_nak = psFSM->pdc->try_nak;
	psFSM->try_terminate = psFSM->pdc->try_terminate;
	fsm_reset(psFSM);

	/* Initialize timer */
	t->func = psFSM->pdc->fsmi == Ap ? (void (*)(char *))ap_timeout :
													(void (*)(char *))fsm_timeout;
	t->arg = (char *)psFSM;
	set_timer(t, psFSM->pdc->timeout);
	fsm_timer(psFSM);
	stop_timer(t);
}

/* Start FSM (after open event, and physical line up) */
void fsm_start(FSM *psFSM)
{
	if ( psFSM->pdv == NULL )
		return;

	FsmLog( "Start");
	if ( !(psFSM->flags & (FSM_ACTIVE | FSM_PASSIVE)) )
		return;

	switch ( psFSM->state ) {
	case fsmCLOSED:
	case fsmLISTEN:
	case fsmTERM_Sent:
		(*psFSM->pdc->starting)(psFSM);
		fsm_reset(psFSM);

		if ( psFSM->flags & FSM_ACTIVE ){
			fsm_sendreq(psFSM);
			psFSM->state = fsmREQ_Sent;
		}
		break;
	default:
		/* already started */
		break;
	};
}


/************************************************************************/
/* Physical Line Down Event */
static void near fsm_down(FSM *psFSM)
{
	if ( psFSM->pdv == NULL )
		return;

	FsmLog( "Down");
	switch ( psFSM->state ) {
	case fsmREQ_Sent:
	case fsmACK_Rcvd:
	case fsmACK_Sent:
		stop_timer(&(psFSM->timer));
		fsm_reset(psFSM);
		break;

	case fsmOPENED:
		(*psFSM->pdc->closing)(psFSM);
		/* fallthru */
	case fsmTERM_Sent:
		fsm_reset(psFSM);
		break;

	case fsmCLOSED:
	case fsmLISTEN:
		/* nothing to do */
		break;
	};
}


/************************************************************************/
/* Close the connection */
void fsm_close(FSM *psFSM)
{
	if ( psFSM->pdv == NULL )
		return;

	FsmLog( "Close");
	switch ( psFSM->state ) {
	case fsmOPENED:
		(*psFSM->pdc->closing)(psFSM);
		/* fallthru */
	case fsmACK_Sent:
		psFSM->retry = psFSM->try_terminate;
		fsm_sendtermreq(psFSM);
		psFSM->state = fsmTERM_Sent;
		break;

	case fsmREQ_Sent:
	case fsmACK_Rcvd:
		/* simply wait for REQ timeout to expire */
		psFSM->retry = 0;
		psFSM->state = fsmTERM_Sent;
		break;

	case fsmLISTEN:
		psFSM->state = fsmCLOSED;
		break;

	case fsmTERM_Sent:
	case fsmCLOSED:
		/* nothing to do */
		break;
	};
}

static void near fsm_free(FSM *psFSM)
{
	if ( psFSM->pdv_out != NULL ) {
		(*psFSM->pdc->free)(psFSM);

		free( psFSM->pdv_out );
		psFSM->pdv_out = NULL;
	}
	if ( psFSM->pdv_in != NULL ) {
		(*psFSM->pdc->free)(psFSM);

		free( psFSM->pdv_in );
		psFSM->pdv_in = NULL;
	}
}


//	PPPDUMP.C

#ifdef TRACE
/* dump a PPP packet */
ncp_dump(WORD type,MBUF **bpp);
void vjcomp_dump(MBUF **bpp,int unused);

#pragma argsused
int ppp_dump(struct mbuf **bpp,int check,int trace,char *header)
{
	PPP_HDR hdr;
	BYTE *pFixAdr, bFixData;
	MBUF *tbp;

	trprintf("%sPPP: len %3u ",header, len_mbuf(*bpp));

	/* HDLC address and control fields may be compressed out */
	if ((BYTE)(*bpp)->data[0] != HDLC_ALL_ADDR) {
		trprintf("CO ");
	} else if ((BYTE)(*bpp)->data[1] != HDLC_UI) {
		trprintf("(missing UI!) ");
	} else {
		/* skip address/control fields */
		pull16(bpp);
	}

	/* Initialize the expected header */
	hdr.addr = HDLC_ALL_ADDR;
	hdr.control = HDLC_UI;
	hdr.protocol = pullchar(bpp);

	/* First byte of PPP protocol field may be compressed out */
	if ( hdr.protocol & 0x01 ) {
		trprintf("CP ");
	} else {
		hdr.protocol = (hdr.protocol << 8) | pullchar(bpp);

		/* Second byte of PPP protocol field must be odd */
		if ( !(hdr.protocol & 0x01) ) {
			trprintf( "(not odd!) " );
		}
	}

	trprintf("prot ");
	switch(hdr.protocol){
		case PPP_IP_PROTOCOL:
			trprintf("IP\n");
			ip_dump(bpp,0,0,"");
			break;
		case PPP_IPCP_PROTOCOL:
		case PPP_LCP_PROTOCOL:
		case PPP_PAP_PROTOCOL:
		case PPP_CHAP_PROTOCOL:
		case PPP_CCP_PROTOCOL:
			ncp_dump(hdr.protocol,bpp);
			break;
		case PPP_COMPR_PROTOCOL:
			trprintf("VJ Compressed TCP/IP, ");
			vjcomp_dump(bpp,0);
			break;
		case PPP_UNCOMP_PROTOCOL:
			/* Restore the bytes used with Uncompressed TCP */
			pFixAdr = (*bpp)->data + 9;
			bFixData = *pFixAdr;
			trprintf("VJ Uncompressed TCP/IP, connection %u\n",
				bFixData);
			*pFixAdr = TCP_PTCL;
			ip_dump(bpp,1,trace,"");
			*pFixAdr = bFixData;
			break;
		default:
			trprintf("unknown 0x%04x\n",hdr.protocol);
			break;
	}
	return 0;
}

static char *codes[] =
{
	"Configure-Request",
	"Configure-Ack",
	"Configure-Nak",
	"Configure-Reject",
	"Terminate-Request",
	"Terminate-Ack",
	"Code-Reject",
	"Protocol-Reject",
	"Echo-Request",
	"Echo-Reply",
	"Discard-Request"
};

static char *chap_codes[] =
{
	"Challenge",
	"Response",
	"Success",
	"Failure"
};

#define IPCP_OFFSET	200		// offset to IPCP type field
#define CCP_OFFSET	200

ncp_dump(WORD protocol,MBUF **bpp)
{
	char szError[20], *pszProtocol, *pszCode, *pszType, *pszAuth;
	BYTE code = pullchar(bpp);
	BYTE id = pullchar(bpp);
	BYTE bVal;
	WORD wVal;
	DWORD dwVal;
	BYTE type, olen, olen1, offset;

	int len = pull16(bpp);

	if (protocol != PPP_CHAP_PROTOCOL && code > 0 && code < 12)
		pszCode = codes[code - 1];
	else if (protocol == PPP_CHAP_PROTOCOL && code > 0 && code < 5)
		pszCode = chap_codes[code - 1];
	else {
		pszCode = szError;
		sprintf(szError,"Unknown code x%02x",code);
	}
	switch (protocol) {
	case PPP_CCP_PROTOCOL:
		pszProtocol = "CCP";
		offset = CCP_OFFSET;
		goto lcp_ipcp;
	case PPP_IPCP_PROTOCOL:
		pszProtocol = "IPCP";
		offset = IPCP_OFFSET;
		goto lcp_ipcp;
	case PPP_LCP_PROTOCOL:
		pszProtocol = "LCP";
		offset = 0;
		lcp_ipcp:
		trprintf("%s %s id %u len %u\n",pszProtocol,pszCode,id,len);
		len -= 4;
		if (code >= 1 && code <= 4) {	// options
			while (len > 0) {
				type = pullchar(bpp);
				olen = pullchar(bpp);
				olen1 = olen;

				switch (type + offset) {
				case LCP_MRU:				// 0x01
					wVal = pull16(bpp);
					trprintf("LCP: MRU len %u %u ",olen,wVal);
					olen -= 2;
					len -= 2;
					goto end_option;
				case LCP_ACCM:				// 0x02
					dwVal = pull32(bpp);
					trprintf("LCP: ACCM len %u x%08lx ",olen,dwVal);
					olen -= 4;
					len -= 4;
					goto end_option;
				case LCP_AUTHENT:			// 0x03
					pszType = "AUTH";
					goto set_protocol;
				case LCP_QUALITY:			// 0x04
					pszType = "QUAL";
					set_protocol:
					wVal = pull16(bpp);
					if (wVal == PPP_PAP_PROTOCOL)
						pszAuth = "PAP";
					else if (wVal == PPP_SPAP_PROTOCOL)
						pszAuth = "SPAP";
					else if (wVal == PPP_CHAP_PROTOCOL)
						pszAuth = "CHAP";
					else if (wVal == PPP_EAP_PROTOCOL)
						pszAuth = "EAP";
					else {
						pszAuth =  szError;
						sprintf(szError,"x%04x",wVal);
					}
					trprintf("LCP: %s len %u %s ",pszType,olen,pszAuth);
					olen -= 2;
					len -= 2;
					goto end_option;
				case LCP_MAGIC:			// 0x05
					pszType = "MAGIC";
					goto print_option;
				case LCP_PFC:				// 0x07
					pszType = "PFC";
					goto print_option;
				case LCP_ACFC:				// 0x08
					pszType = "ACFC";
					goto print_option;
				case IPCP_ADDRESSES + IPCP_OFFSET:	// 0x01
					pszType = "ADDRESSES";
					goto print_option;
				case IPCP_COMPRESS + IPCP_OFFSET:	// 0x02
					pszType = "COMPRESS";
					goto print_option;
				case IPCP_ADDRESS + IPCP_OFFSET:		// 0x03
					pszType = "ADDRESS";
					goto print_ipaddress;
				case IPCP_PRI_DNS + IPCP_OFFSET:
					pszType = "DNS1";
					goto print_ipaddress;
				case IPCP_SEC_DNS + IPCP_OFFSET:
					pszType = "DNS2";

					print_ipaddress:
					dwVal = pull32(bpp);
					trprintf("IPCP: %s len %u %s ",pszType,olen,inet_ntoa(dwVal));
					olen -= 4;
					len -= 4;
					goto end_option;
				default:
					sprintf(szError,"Option x%02x ",type);
					pszType = szError;
					print_option:
					trprintf("%s: %s len %u ",pszProtocol,pszType,olen1);
					end_option:
					olen -= 2;
					len -= 2;
					for (;olen && len;--olen,--len) {
						bVal = pullchar(bpp);
						trprintf("x%02x ",bVal);
					}
					trprintf("\n");
				}
			}
		}
		break;
	case PPP_PAP_PROTOCOL:
		pszProtocol = "PAP";
		trprintf("%s %s id %u len %u\n",pszProtocol,pszCode,id,len);
		break;
	case PPP_CHAP_PROTOCOL:
		pszProtocol = "CHAP";
		trprintf("%s %s id %u len %u\n",pszProtocol,pszCode,id,len);
		break;
	}
	return 0;
}

static WORD decodeint(struct mbuf **bpp);

static WORD decodeint(struct mbuf **bpp)
{
	BYTE bsTmp[2];

	pullup(bpp,bsTmp,1);
	if (bsTmp[0] == 0)
		pullup(bpp,bsTmp,2);
	else {
		bsTmp[1] = bsTmp[0];
		bsTmp[0] = 0;
	}
	return(get16(bsTmp));
}

#pragma argsused
void vjcomp_dump(MBUF **bpp,int unused)
{
	BYTE bChanges;
	BYTE bsTmp[2];

	if(bpp == NULL || *bpp == NULL)
		return;

	/* Dump compressed TCP/IP header */
	bChanges = pullchar(bpp);
	trprintf("changes 0x%02x",bChanges);
	if (bChanges & NEW_C) {
		pullup(bpp,bsTmp,1);
		trprintf(" connection %u",bsTmp[0]);
	}
	pullup(bpp,bsTmp,2);
	trprintf("\nVJC: TCP cs 0x%04x",get16(bsTmp));

	if (bChanges & TCP_PUSH_BIT)
		trprintf(" PSH");

	switch (bChanges & SPECIALS_MASK) {
	case SPECIAL_I:
		trprintf(" D-ACK/SEQ implied");
		break;

	case SPECIAL_D:
		trprintf(" D-SEQ implied");
		break;

	default:
		if (bChanges & NEW_U) {
			trprintf(" UP %d",decodeint(bpp));
		}
		if (bChanges & NEW_W)
			trprintf(" D-WINDOW %d",decodeint(bpp));
		if (bChanges & NEW_A)
			trprintf(" D-ACK %d",decodeint(bpp));
		if (bChanges & NEW_S)
			trprintf(" D-SEQ %d",decodeint(bpp));
		break;
	};
	if (bChanges & NEW_I) {
		trprintf(" D-ID %d\n",decodeint(bpp));
	} else {
		trprintf(" IncID\n");
	}
}
#endif


#ifdef ALLOC_TRACE
char *Strdup(const char *s,unsigned id)
#else
char *Strdup(const char *s)
#endif
{
	char *psz;

	if ((psz = malloc(strlen(s) + 1,id)) != 0)
		strcpy(psz,s);
	return psz;
}

