//
//   $Id: PPPCMD.C 1.5 2003/08/04 19:51:12Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*  PPPCMD.C	-- PPP related user commands
 *	This implementation of PPP is declared to be in the public domain.
 *
 *	Jan 91	Bill_Simpson@um.cc.umich.edu
 *		Computer Systems Consulting Services
 *
 *	Acknowledgements and correction history may be found in PPP.C
Revision history:
2003-08-04 gvn optionally add route to peer
2002-07-30 gvn DNS

 */

#if APL==0
#pragma option -zAINETINIT
#pragma option -zFINETINIT
#endif

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "iface.h"
#include "cmdparse.h"
#include "slhc.h"
#include "internet.h"
#include "ip.h"
#include "ppp.h"
#include "get16.h"
#include "modem.h"
#include "ser.h"

extern WORD sec_timer;
DWORD rdclock(void);
extern char badhost[];

//static int doppp_quick(int argc, char *argv[], void *p);
static int doppp_trace(int argc, char *argv[], void *p);
#if APL
#include "trace.h"
static int spot(WORD work,WORD want,WORD will,WORD mask);
static void lcpstat(FSM *psFSM);
static void apstat(FSM *psFSM);
static void ipcpstat(FSM *psFSM);
#endif
static int dotry_nak(int argc, char *argv[], void *p);
static int dotry_req(int argc, char *argv[], void *p);
static int dotry_terminate(int argc, char *argv[], void *p);
int doppp_stat(int argc,char *argv[], void *p);

/* "ppp" subcommands */
static struct scmds Pppcmds[] = {
	{"ipcp",		doppp_ipcp},
	{"ipcpin",	doppp_ipcpin},
	{"lcp",		doppp_lcp},
	{"lcpin",	doppp_lcpin},
	{"pap",		doppp_ap},
	{"ap",		doppp_ap},
	{"papin",	doppp_apin},
	{"apin",		doppp_apin},
#if APL
	{"ppp",		doppp_stat},
#endif
//	"quick",	doppp_quick,
	{"trace",	doppp_trace},
	{"",0}
};

/* "ppp <iface> <ncp> retry" subcommands */
static struct scmds PppTrycmds[] = {
	{"configure",	dotry_req},
	{"failure",		dotry_nak},
	{"terminate",	dotry_terminate},
	{"",0}
};
#if APL
static char *PPPStatus[] = {
	"Physical Line Dead",
	"Establishment Phase",
	"Authentication Phase",
	"Network Protocol Phase",
	"Termination Phase"
};

static char *NCPStatus[] = {
	"Closed",
	"Listening - waiting for request",
	"Starting configuration exchange",
	"Ack received; waiting for request",
	"Request ack'ed; waiting for ack",
	"Opened",
	"Terminate request sent"
};
#endif
int PPPtrace;
IFACE *PPPiface;  /* iface for trace */


/****************************************************************************/

int doppp_commands(int argc,char *argv[],void *p)
{
	return subcmd(Pppcmds,++argc,--argv,p);
}

#if APL
/* Close connection on PPP interface */
#pragma argsused
int doppp_close(int argc,char *argv[],void *p)
{
	register FSM *psFSM = p;

	psFSM->flags &= ~(FSM_ACTIVE | FSM_PASSIVE);

	fsm_close( psFSM );
	return 0;
}
#endif

#pragma argsused
int doppp_passive(int argc,char *argv[],void *p)
{
	register FSM *psFSM = p;

	psFSM->flags |= FSM_PASSIVE;
	if ( psFSM->state < fsmLISTEN ) {
		psFSM->state = fsmLISTEN;
	}

	return 0;
}


static int near doppp_active(void *p)
{
	register FSM *psFSM = p;

	psFSM->flags |= FSM_ACTIVE;

	return 0;
}

/*
static int doppp_quick(int argc,char *argv[],void *p)
{
	register IFACE *ifp = p;
	register PPP *psPPP = ifp->prot_ext;
	LCP *psLCP = psPPP->fsm[Lcp].pdv;
	IPCP *psIPCP = psPPP->fsm[IPcp].pdv;

	psLCP->local.want.accm = 0L;
	psLCP->local.want.negotiate |= LCP_N_ACCM;
	psLCP->local.want.magic_number += (long)&psLCP->local.want.magic_number;
	psLCP->local.want.negotiate |= LCP_N_MAGIC;
	psLCP->local.want.negotiate |= LCP_N_ACFC;
	psLCP->local.want.negotiate |= LCP_N_PFC;

	psIPCP->local.want.compression = PPP_COMPR_PROTOCOL;
	psIPCP->local.want.slots = 16;
	psIPCP->local.want.slot_compress = 1;
	psIPCP->local.want.negotiate |= IPCP_N_COMPRESS;
	doppp_active( &(psPPP->fsm[IPcp]) );

	return 0;
}
*/

/****************************************************************************/
#if APL
#pragma argsused
int doppp_stat(int argc,char *argv[], void *p)
{
	IFACE *ifp = p;
	register PPP *psPPP = ifp->prot_ext;
	FSM *psFSM = &(psPPP->fsm[Lcp]);

	printf("%s", PPPStatus[psPPP->phase]);
	if (psPPP->phase == pppREADY) {
		printf("\t(open for %u seconds)",sec_timer - psPPP->upsince);
	}
	printf(" - %s\n",psFSM->pdv == psFSM->pdv_out ? "outgoing" : "incoming");
	if (psPPP->flags & PPP_CBV)
		printf("Callback pending\n");
	if (psPPP->flags & PPP_CBVCALL)
		printf("Callback in progress\n");

	printf("%10lu In,%11lu Flags,%6u ME, %6u FE, %7u CSE,%6u other\n",
		psPPP->InRxOctetCount,
		psPPP->InOpenFlag,
		psPPP->InMemory,
		psPPP->InFrame,
		psPPP->InChecksum,
		psPPP->InError);
	printf("\t\t%9u Lcp,%8u Ap, %6u IPcp,%6u Unknown\n",
		psPPP->InNCP[Lcp],
		psPPP->InNCP[Ap],
		psPPP->InNCP[IPcp],
		psPPP->InUnknown);
	printf("%10lu Out,%10lu Flags,%6u ME, %6u Fail\n",
		psPPP->OutTxOctetCount,
		psPPP->OutOpenFlag,
		psPPP->OutMemory,
		psPPP->OutError);
	printf("\t\t%9u Lcp,%8u Ap, %6u IPcp\n",
		psPPP->OutNCP[Lcp],
		psPPP->OutNCP[Ap],
		psPPP->OutNCP[IPcp]);
	if ( psFSM->pdv )
		lcpstat(psFSM);
	if ( psPPP->fsm[Ap].pdv != NULL )
		apstat(&(psPPP->fsm[Ap]));
	if ( psPPP->fsm[IPcp].pdv != NULL )
		ipcpstat(&(psPPP->fsm[IPcp]));
	return 0;
}

static int spot(WORD work,WORD want,WORD will,WORD mask)
{
	char blot = ' ';
	int result = (work & mask);

	if ( !(will & mask) ) {
		blot = '*';
	} else if ( (want ^ work) & mask ) {
		blot = (result ? '+' : '-');
	}
	printf( "%c", blot );
	return result;
}

static void lcpstat(FSM *psFSM)
{
	LCP *psLCP = psFSM->pdv;
	LCP_VALUE *localp = &(psLCP->local.work);
	WORD  localwork = psLCP->local.work.negotiate;
	WORD  localwant = psLCP->local.want.negotiate;
	WORD  localwill = psLCP->local.will_negotiate;
	LCP_VALUE *remotep = &(psLCP->remote.work);
	WORD  remotework = psLCP->remote.work.negotiate;
	WORD  remotewant = psLCP->remote.want.negotiate;
	WORD  remotewill = psLCP->remote.will_negotiate;

	printf("LCP %s, flags %s%s\n",
		NCPStatus[psFSM->state],psFSM->flags & FSM_ACTIVE ? "active " : "",
										psFSM->flags & FSM_PASSIVE ? "passive " : "");

	printf("\t\t MRU\t  ACCM\t\t AP\t PFC  ACFC Magic\n");
	printf("\tLocal:\t");
	spot( localwork, localwant, localwill, LCP_N_MRU );
	printf( "%4d\t", localp->mru );
	spot( localwork, localwant, localwill, LCP_N_ACCM );
	printf( "0x%08lx\t", localp->accm );
	if ( spot( localwork, localwant, localwill, LCP_N_AUTHENT ) ) {
		switch ( localp->authentication ) {
		case PPP_PAP_PROTOCOL:
			printf( "Pap\t" );
			break;
		case PPP_CHAP_PROTOCOL:
			printf( "Chap\t" );
			break;
		default:
			printf( "0x%04x\t", localp->authentication);
			break;
		};
	} else {
		printf( "None\t" );
	}

	printf( spot( localwork, localwant, localwill, LCP_N_PFC )
		 ? "Yes " : "No  " );
	printf( spot( localwork, localwant, localwill, LCP_N_ACFC )
		 ? "Yes " : "No  " );

	spot( localwork, localwant, localwill, LCP_N_MAGIC );
	if ( localp->magic_number != 0L ) {
		printf( "0x%08lx\n", localp->magic_number );
	} else {
		printf( "unused\n" );
	}

	printf("\tRemote:\t");

	spot( remotework, remotewant, remotewill, LCP_N_MRU );
	printf( "%4d\t", remotep->mru );

	spot( remotework, remotewant, remotewill, LCP_N_ACCM );
	printf( "0x%08lx\t", remotep->accm );

	if ( spot( remotework, remotewant, remotewill, LCP_N_AUTHENT ) ) {
		switch ( remotep->authentication ) {
		case PPP_PAP_PROTOCOL:
			printf( "Pap\t" );
			break;
		case PPP_CHAP_PROTOCOL:
			printf( "Chap\t" );
			break;
		default:
			printf( "0x%04x\t", remotep->authentication);
			break;
		};
	} else {
		printf( "None\t" );
	}

	printf( spot( remotework, remotewant, remotewill, LCP_N_PFC )
		 ? "Yes " : "No  " );
	printf( spot( remotework, remotewant, remotewill, LCP_N_ACFC )
		 ? "Yes " : "No  " );

	spot( remotework, remotewant, remotewill, LCP_N_MAGIC );
	if ( remotep->magic_number != 0L ) {
		printf( "0x%08lx\n", remotep->magic_number );
	} else {
		printf( "unused\n" );
	}
}


static void apstat(FSM *psFSM)
{
	AP *psAP = psFSM->pdv;

	printf("AP %s%s%s%s%s\n",
				psFSM->state & AP_OPENED ? "Opened " : "",
				psFSM->state & AP_REQUEST_SENT ? "REQUEST " : "",
				psFSM->state & AP_CHALLENGE_SENT ? "CHALLENGE " : "",
				psFSM->state & AP_RESPONSE_SENT ? "RESPONSE " : "",
				psFSM->state & (AP_REQUEST_SENT|AP_CHALLENGE_SENT|AP_RESPONSE_SENT) ? "sent" : "");

	printf( "\tMessage: '%s'\n", (psAP->psApMessage == NULL) ?
		"none" : psAP->psApMessage );
}

static void ipcpstat(FSM *psFSM)
{
	IPCP *psIPCP = psFSM->pdv;
	IPCP_VALUE *localp = &(psIPCP->local.work);
	WORD  localwork = psIPCP->local.work.negotiate;
	IPCP_VALUE *remotep = &(psIPCP->remote.work);
	WORD  remotework = psIPCP->remote.work.negotiate;

	printf("IPCP %s\n",
		NCPStatus[psFSM->state]);
	printf("\tlocal IP address: %s",
		inet_ntoa(localp->dwLocalAddress));
	printf("  remote IP address: %s\n",
		inet_ntoa(remotep->dwLocalAddress));

	if (localwork & IPCP_N_COMPRESS) {
		printf("    In\tTCP header compression enabled:"
			" slots = %d, flag = 0x%02x\n",
			localp->slots,
			localp->slot_compress);
		slhc_status(psIPCP->slhcp,1);
	}

	if (remotework & IPCP_N_COMPRESS) {
		printf("    Out\tTCP header compression enabled:"
			" slots = %d, flag = 0x%02x\n",
			remotep->slots,
			remotep->slot_compress);
		slhc_status(psIPCP->slhcp,2);
	}
}
#endif

/****************************************************************************/
static int near SetIntVal(int argc,char *argv[],void *p,int size)
{
	long *plVal = p;
	BYTE *pbVal = p;

	if (argc < 2) {
		if (size == 4)
			printf("%ld\n",*plVal);
		else
			printf("%d\n",*pbVal);
	} else {
		long x = strtol( argv[1], NULL, 0 );

		if (x <= 0) {
			printf("Value %s (%ld) must be > 0\n",
				argv[1], x);
			return -1;
		} else if (size == 4) {
			*plVal = x;
		}
		else
			*pbVal = (BYTE)x;
	}
	return 0;
}


/* Set timeout interval for waiting for response from remote peer */
int doppp_timeout(int argc,char *argv[],void *p)
{
	return SetIntVal(argc,argv,&((FSM *)p)->pdc->timeout,4);
}

int doppp_try(int argc,char *argv[],void *p)
{
	return subcmd(PppTrycmds, argc, argv, p);
}

static int dotry_nak(int argc,char *argv[],void *p)
{
	return SetIntVal(argc,argv,&((FSM *)p)->try_nak,1);
}

static int dotry_req(int argc,char *argv[],void *p)
{
	return SetIntVal(argc,argv,&((FSM *)p)->try_req,1);
}

static int dotry_terminate(int argc,char *argv[],void *p)
{
	return SetIntVal(argc,argv,&((FSM *)p)->try_terminate,1);
}


static int doppp_trace(int argc,char *argv[],void *p)
{
	register IFACE *ifp = p;
	register PPP *psPPP = ifp->prot_ext;
	int tracing = psPPP->trace;
	int result = setint(&tracing,"PPP tracing",argc,argv);

	PPPtrace = psPPP->trace = tracing;
	PPPiface = ifp;
	return result;
}

/************************************************************************/
static int dolcp_local(int argc, char *argv[], void *p);
static int dolcp_open(int argc, char *argv[], void *p);
static int dolcp_listen(int argc, char *argv[], void *p);
static int dolcp_start(int argc, char *argv[], void *p);
static int dolcp_remote(int argc, char *argv[], void *p);
static int dolcp_accm(int argc, char *argv[], void *p);
static int dolcp_acfc(int argc, char *argv[], void *p);
static int dolcp_auth(int argc, char *argv[], void *p);
static int dolcp_magic(int argc, char *argv[], void *p);
static int dolcp_mru(int argc, char *argv[], void *p);
static int dolcp_pfc(int argc, char *argv[], void *p);

/* "ppp <iface> lcp" subcommands */
static struct scmds Lcpcmds[] = {
#if APL
	{"close",	doppp_close},
#endif
	{"listen",	dolcp_listen},
	{"local",	dolcp_local},
	{"open",		dolcp_open},
	{"remote",	dolcp_remote},
	{"retry",	doppp_try},
	{"start",	dolcp_start},
	{"timeout",	doppp_timeout},
	{"",0}
};

/* "ppp <iface> lcp [local | remote]" subcommands */
static struct scmds Lcpside_cmds[] = {
	{"accm",		dolcp_accm},
	{"acfc",		dolcp_acfc},
	{"authen",	dolcp_auth},
	{"magic",	dolcp_magic},
	{"mru",		dolcp_mru},
	{"pfc",		dolcp_pfc},
	{"",0}
};


static int near doppp_lcpc(int argc,char *argv[],void *p, int in)
{
	int retv;
	register IFACE *ifp = p;
	register PPP *psPPP = ifp->prot_ext;
	FSM *psFSM = &(psPPP->fsm[Lcp]);
	void *pP = psFSM->pdv;

	psFSM->pdv = in ? psFSM->pdv_in : psFSM->pdv_out;

	retv = subcmd(Lcpcmds, argc, argv, psFSM);
	psFSM->pdv = pP;
	return retv;
}

int doppp_lcp(int argc,char *argv[],void *p)
{
	return doppp_lcpc(argc,argv,p,0);
}

int doppp_lcpin(int argc,char *argv[],void *p)
{
	return doppp_lcpc(argc,argv,p,1);
}


static int dolcp_local(int argc,char *argv[],void *p)
{
	int i;

	FSM *psFSM = p;
	LCP *psLCP = psFSM->pdv;
	i = subcmd(Lcpside_cmds, argc, argv, &(psLCP->local));
   fsm_init(psFSM);
	return i;
}

static int dolcp_listen(int argc,char *argv[],void *p)
{
	FSM *psFSM = p;

	doppp_passive( argc, argv, p );
	ppp_iostatus(psFSM->psPPP->iface,PARAM_UP,0);
	return 0;
}

#pragma argsused
static int dolcp_open(int argc,char *argv[],void *p)
{
	FSM *psFSM = p;

	doppp_active( p );
	ppp_iostatus(psFSM->psPPP->iface,PARAM_UP,0);
	return 0;
}

#pragma argsused
static int dolcp_start(int argc,char *argv[],void *p)
{
	FSM *psFSM = p;

	doppp_active( p );
	ppp_iostatus(psFSM->psPPP->iface,PARAM_UP,1);
	return 0;
}

#pragma argsused
static int dolcp_remote(int argc,char *argv[],void *p)
{
	int i;
	FSM *psFSM = p;
	LCP *psLCP = psFSM->pdv;
	i = subcmd(Lcpside_cmds, argc, argv, &(psLCP->remote));
	fsm_init(psFSM);
	return i;
}

/************************************************************************/

static int dolcp_accm(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;

	if (argc < 2) {
		printf("0x%08lx\n",side_p->want.accm);
	} else if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_ACCM,
			"Allow ACCM", --argc, &argv[1] );
	} else {
		side_p->want.accm = strtoul(argv[1], NULL, 0);
		if ( side_p->want.accm != LCP_ACCM_DEFAULT )
			side_p->want.negotiate |= LCP_N_ACCM;
		else
			side_p->want.negotiate &= ~LCP_N_ACCM;
	}
	return 0;
}

static int dolcp_acfc(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;

	if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_ACFC,
			"Allow Address/Control Field Compression", --argc, &argv[1] );
	}
	return bit16cmd( &(side_p->want.negotiate), LCP_N_ACFC,
		"Address/Control Field Compression", argc, argv );
}


static int dolcp_auth(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;

	if (argc < 2) {
		if ( side_p->want.negotiate & LCP_N_AUTHENT ) {
			switch ( side_p->want.authentication ) {
			case PPP_PAP_PROTOCOL:
				printf("Pap\n");
				break;
			case PPP_CHAP_PROTOCOL:
				printf( "Chap\t" );
				break;
			default:
				printf("0x%04x\n", side_p->want.authentication);
				break;
			};
		} else {
			printf("None\n");
		}
	} else if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_AUTHENT,
			"Allow Authentication", --argc, &argv[1] );
	} else if (stricmp(argv[1],"pap") == 0) {
		side_p->want.negotiate |= LCP_N_AUTHENT;
		side_p->want.authentication = PPP_PAP_PROTOCOL;
	} else if (stricmp(argv[1],"chap") == 0) {
		side_p->want.negotiate |= LCP_N_AUTHENT;
		side_p->want.authentication = PPP_CHAP_PROTOCOL;
	} else if (stricmp(argv[1],"none") == 0) {
		side_p->want.negotiate &= ~LCP_N_AUTHENT;
	} else {
		printf("allow pap chap none\n");
		return 1;
	}
	return 0;
}


static int dolcp_magic(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;
	int result = 0;

	if (argc < 2) {
		printf("%d\n",side_p->want.magic_number);
	} else if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_MAGIC,
			"Allow Magic Number", --argc, &argv[1] );
	} else {
		register long x = strtoul(argv[1], NULL, 0);

		if ( !x ) {
			int test;

			/* Check for keyword */
			result = setbool( &test, "Magic Number", argc, argv );

			if ( test ) {
				/* Make a non-zero random number */
				x = rdclock() << (((BYTE)rdclock() & 0xf)+8);
			}
		}
		if ( x ) {
			side_p->want.negotiate |= LCP_N_MAGIC;
		} else {
			side_p->want.negotiate &= ~LCP_N_MAGIC;
		}
		side_p->want.magic_number = x;
	}
	return result;
}

static int dolcp_mru(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;

	if (argc < 2) {
		printf("%d\n",side_p->want.mru);
	} else if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_MRU,
			"Allow MRU", --argc, &argv[1] );
	} else {
		register int x = (int)strtol( argv[1], NULL, 0 );

		if (x < LCP_MRU_LO || x > LCP_MRU_HI) {
			printf("MRU %s (%d) out of range %d thru %d\n",
				argv[1], x, LCP_MRU_LO, LCP_MRU_HI);
			return -1;
		} else if ( x != LCP_MRU_DEFAULT ) {
			side_p->want.negotiate |= LCP_N_MRU;
		} else {
			side_p->want.negotiate &= ~LCP_N_MRU;
		}
		side_p->want.mru = x;
	}
	return 0;
}

static int dolcp_pfc(int argc,char *argv[],void *p)
{
	LCP_SIDE *side_p = p;

	if (stricmp(argv[1],"allow") == 0) {
		return bit16cmd(&(side_p->will_negotiate),LCP_N_PFC,
			"Allow Protocol Field Compression", --argc, &argv[1] );
	}
	return bit16cmd( &(side_p->want.negotiate), LCP_N_PFC,
		"Protocol Field Compression", argc, argv );
}

/************************************************************************/
static int doipcp_local(int argc, char *argv[], void *p);
static int doipcp_open(int argc, char *argv[], void *p);
int doipcp_pool(int argc, char *argv[], void *p);
static int doipcp_remote(int argc, char *argv[], void *p);

static int doipcp_address(int argc, char *argv[], void *p);
static int doipcp_dns1(int argc, char *argv[], void *p);
static int doipcp_dns2(int argc, char *argv[], void *p);
static int doipcp_route(int argc, char *argv[], void *p);
static int doipcp_compress(int argc, char *argv[], void *p);

/* "ppp <iface> ipcp" subcommands */
static struct scmds IPcpcmds[] = {
#if APL
	{"close",	doppp_close},
#endif
	{"listen",	doppp_passive},
	{"local",	doipcp_local},
	{"open",		doipcp_open},
//	"pool",		doipcp_pool,
	{"remote",	doipcp_remote},
	{"retry",	doppp_try},
	{"timeout",	doppp_timeout},
	{"",0}
};

/* "ppp <iface> ipcp {local | remote}" subcommands */
static struct scmds IPcpside_cmds[] = {
	{"address",	doipcp_address},
	{"compress",doipcp_compress},
	{"dns1",		doipcp_dns1},
	{"dns2",		doipcp_dns2},
	{"route",	doipcp_route},
	{"",0}
};

static int near doppp_ipcpc(int argc,char *argv[],void *p, int in)
{
	int retv;
	register IFACE *ifp = p;
	register PPP *psPPP = ifp->prot_ext;
	FSM *psFSM = &(psPPP->fsm[IPcp]);
	void *pP = psFSM->pdv;

	psFSM->pdv = in ? psFSM->pdv_in : psFSM->pdv_out;
	retv = subcmd(IPcpcmds, argc, argv, psFSM);
	psFSM->pdv = pP;
	return retv;
}

int doppp_ipcp(int argc,char *argv[],void *p)
{
	return doppp_ipcpc(argc,argv,p,0);
}

int doppp_ipcpin(int argc,char *argv[],void *p)
{
	return doppp_ipcpc(argc,argv,p,1);
}

static int doipcp_local(int argc,char *argv[],void *p)
{
	int i;

	FSM *psFSM = p;
	IPCP *psIPCP = psFSM->pdv;
	i = subcmd(IPcpside_cmds, argc, argv, &(psIPCP->local));
	fsm_init(psFSM);
	return i;
}

#pragma argsused
static int doipcp_open(int argc,char *argv[],void *p)
{
	FSM *psFSM = p;

	doppp_active( p );

	if ( psFSM->psPPP->phase == pppREADY ) {
		fsm_start( psFSM );
	}
	return 0;
}

/* Set a pool of peer addresses for PPP interface
static int doipcp_pool(int argc,char *argv[],void *p)
{
	FSM *psFSM = p;
	IPCP *psIPCP = psFSM->pdv;
	int32 pool_addr;
	int pool_cnt;

	if (argc < 2) {
		if ( psIPCP->peer_min == 0L ) {
			printf("None");
		} else {
			printf("%s thru ", inet_ntoa(psIPCP->peer_min));
			printf("%s\n", inet_ntoa(psIPCP->peer_max));
		}
		return 0;
	}

	if ((pool_addr = resolve(argv[1])) == 0L) {
		printf(badhost,argv[1]);
	}

	// May specify a consecutive range of addresses; otherwise assume 1
	if (argc < 3)
		pool_cnt = 1;
	else
		pool_cnt = (int)strtol( argv[2], NULL, 0 );

	if (pool_cnt <= 0) {
		printf("Pool count %s (%d) must be > 0\n");
		return -1;
	}

	psIPCP->peer_min = pool_addr;
	psIPCP->peer_max = pool_addr + pool_cnt - 1;
	return 0;
}
*/

static int doipcp_remote(int argc,char *argv[],void *p)
{
	int i;
	FSM *psFSM = p;
	IPCP *psIPCP = psFSM->pdv;
	i = subcmd(IPcpside_cmds, argc, argv, &(psIPCP->remote));
	fsm_init(psFSM);
	return i;
}

/************************************************************************/
/* Set addresses for PPP interface */
static int doipcp_address(int argc,char *argv[],void *p)
{
	IPCP_SIDE *side_p = p;
	int32 x32;

	if (argc < 2) {
		printf("%s\n", inet_ntoa(side_p->want.dwLocalAddress));
		return 0;
	} else if ( stricmp(argv[1],"allow") == 0 ) {
		return bit16cmd( &(side_p->will_negotiate), IPCP_N_ADDRESS,
			"Allow Address", --argc, &argv[1] );
	}
	x32 = resolve(argv[1]);
	side_p->want.dwLocalAddress = x32;
	side_p->want.negotiate |= IPCP_N_ADDRESS;
	return 0;
}

static int doipcp_dns(int i,int argc,char *argv[],void *p)
{
	IPCP_SIDE *side_p = p;
	int32 x32;
	int iMask = i ? IPCP_N_DNS2 : IPCP_N_DNS1;

	if (argc < 2) {
		printf("%s\n", inet_ntoa(side_p->want.ardwDnsAddress[i]));
		return 0;
	} else if ( stricmp(argv[1],"allow") == 0 ) {
		return bit16cmd( &(side_p->will_negotiate), iMask,
			"Allow Address", --argc, &argv[1] );
	}
	x32 = resolve(argv[1]);
	side_p->want.ardwDnsAddress[i] = x32;
	side_p->want.negotiate |= iMask;
	return 0;
}

static int doipcp_dns1(int argc,char *argv[],void *p)
{
	return doipcp_dns(0,argc,argv,p);
}

static int doipcp_dns2(int argc,char *argv[],void *p)
{
	return doipcp_dns(1,argc,argv,p);
}

static int doipcp_route(int argc,char *argv[],void *p)
{
	IPCP_SIDE *side_p = p;
	int iRoute = side_p->want.bFlags & IPCP_ADD_RT,iResult;

	iResult = setbool( &iRoute, "Add route", argc, argv );
	if (iRoute)
		side_p->want.bFlags |= IPCP_ADD_RT;
	else
		side_p->want.bFlags &= ~IPCP_ADD_RT;
	return iResult;
}

/* Set IP compression type for PPP interface */
static int doipcp_compress(int argc,char *argv[],void *p)
{
	IPCP_SIDE *side_p = p;

	if (argc < 2) {
		if ( side_p->want.negotiate & IPCP_N_COMPRESS ) {
			switch ( side_p->want.compression ) {
			case PPP_COMPR_PROTOCOL:
				printf("TCP header compression enabled; "
					"Slots = %d, slot compress = %x\n",
					side_p->want.slots,
					side_p->want.slot_compress);
				break;
			default:
				printf("0x%04x\n", side_p->want.compression);
				break;
			};
		} else {
			printf("None\n");
		}
	} else if ( stricmp(argv[1],"allow") == 0 ) {
		return bit16cmd( &(side_p->will_negotiate), IPCP_N_COMPRESS,
			"Allow Compression", --argc, &argv[1] );
	} else if ( stricmp(argv[1],"tcp") == 0
		 || stricmp(argv[1],"vj") == 0 ) {
		side_p->want.compression = PPP_COMPR_PROTOCOL;
		if ( argc >= 3 ) {
			side_p->want.slots = (WORD)strtol(argv[2],NULL,0);
			if ( side_p->want.slots < 1 || side_p->want.slots > 255 ) {
				printf( "slots must be in range 1 to 255" );
				return 1;
			}
		} else {
			side_p->want.slots = IPCP_SLOT_DEFAULT;
		}
		if ( argc >= 4 ) {
			side_p->want.slot_compress = (WORD)strtol(argv[3],NULL,0);
		} else {
			side_p->want.slot_compress = IPCP_SLOT_COMPRESS;
		}
		side_p->want.negotiate |= IPCP_N_COMPRESS;
	} else if (stricmp(argv[1],"none") == 0) {
		side_p->want.negotiate &= ~IPCP_N_COMPRESS;
	} else {
		printf("allow tcp none\n");
		return 1;
	}
	return 0;
}


/****************************************************************************/
static int doap_user(int argc, char *argv[], void *p);

/* "ppp <iface> pap" subcommands */
static struct scmds Apcmds[] = {
	{"timeout",	doppp_timeout},
	{"retry",	doppp_try},
	{"user",		doap_user},
	{"",0}
};

static int near doppp_apc(int argc,char *argv[],void *p,int in)
{
	int	retv;
	IFACE *ifp = p;
	PPP *psPPP = ifp->prot_ext;
	FSM *psFSM = &(psPPP->fsm[Ap]);
	void *pP = psFSM->pdv;

	psFSM->pdv = in ? psFSM->pdv_in : psFSM->pdv_out;
	retv = subcmd(Apcmds, argc, argv, psFSM);
	psFSM->pdv = pP;
	fsm_init(psFSM);
	return retv;
}

int doppp_ap(int argc,char *argv[],void *p)
{
	return doppp_apc(argc,argv,p,0);
}

int doppp_apin(int argc,char *argv[],void *p)
{
	return doppp_apc(argc,argv,p,1);
}


/* Set user/password */
int doap_user(int argc,char *argv[],void *p)
{
	register FSM *psFSM = p;
	register AP *psAP = psFSM->pdv;

	if (argc < 2) {
		printf("%s\n",
			(psAP->psApUsername == NULL) ? "None" : psAP->psApUsername);
		return 0;
	}
	free(psAP->psApUsername);
	psAP->psApUsername = NULL;
	free(psAP->psApPassword);
	psAP->psApPassword = NULL;

	if (stricmp(argv[1],"none") != 0) {
		psAP->psApUsername = strdup(argv[1],AllocPppcUser);
		if (argc > 2) {
			psAP->psApPassword = strdup(argv[2],AllocPppcPass);
		}
	}
	return 0;
}

void ap_timeout(void *vp);

int ppp_raw(IFACE *ifp,MBUF *bp);
extern FSM_CONSTANT lcp_constants;
/* for test purposes, accept anything we understand in the NAK */
static WORD lcp_negotiate = LCP_N_MRU | LCP_N_ACCM | LCP_N_AUTHENT
		| LCP_N_PFC | LCP_N_ACFC | LCP_N_MAGIC;

/* These defaults are defined in the PPP RFCs, and must not be changed */
static LCP_VALUE lcp_default = {
	0,		/* no need to negotiate defaults */

	LCP_MRU_DEFAULT,
	LCP_ACCM_DEFAULT,
	0,		/* no authentication */
	0,		/* no encryption */
	0L,		/* no magic number */
	0L,		/* no reporting period */
};

extern FSM_CONSTANT ap_constants;

static void near LcpInit(LCP *psLCP)
{
	// Set option parameters to first request defaults
	psLCP->local.want = lcp_default;
	psLCP->local.will_negotiate = lcp_negotiate;

	psLCP->remote.want = lcp_default;
	psLCP->remote.work = lcp_default;
	psLCP->remote.will_negotiate = lcp_negotiate;
}

extern FSM_CONSTANT ipcp_constants;
/* for test purposes, accept anything we understand */
static WORD ipcp_negotiate = IPCP_N_ADDRESS | IPCP_N_COMPRESS;
/* These defaults are defined in the PPP RFCs, and must not be changed */
static IPCP_VALUE ipcp_default = {
	0,			/* no need to negotiate defaults */

	0L,		/* no source address */
//	0L,		/* no destination address */
	0,			/* no compression protocol */
	0,			/* no slots */
	0,			/* no slot compression */
	0,			// don't add route
	{0L,0L},	// No DNS
};

static void near IpcpInit(IPCP *psIPCP)
{
	psIPCP->local.want = ipcp_default;
	psIPCP->local.will_negotiate = ipcp_negotiate;

	psIPCP->remote.want = ipcp_default;
	psIPCP->remote.work = ipcp_default;
	psIPCP->remote.will_negotiate = ipcp_negotiate;
}

/****************************************************************************/
/* Initialize PPP control structures for a Point-to-Point interface */
int ppp_init(IFACE *ifp)
{
	PPP *psPPP;
	FSM *psFSM;
	LCP *psLCP;
	IPCP *psIPCP;

	/* Setup for Point-to-Point Protocol */

	if ((psPPP = calloc(1,sizeof(PPP),AllocPppcPpp)) == 0)
		return -1;
	psPPP->iface = ifp;
	psPPP->phase = pppDEAD;
	ifp->prot_ext = psPPP;
	ifp->raw = ppp_raw;
	ifp->output = ppp_output;
	ifp->send = ppp_send;
	ifp->recv = ppp_proc;
	// initialize LCP
	psFSM = &(psPPP->fsm[Lcp]);

	psFSM->psPPP = psPPP;
	psFSM->pdc = &lcp_constants;
	if ((psFSM->pdv = psFSM->pdv_out = psLCP = calloc(1,sizeof(LCP),AllocPppcLcp)) == 0)
		return -1;
	LcpInit(psLCP);
	if ((psFSM->pdv_in = psLCP = calloc(1,sizeof(LCP),AllocPppcLcp)) == 0)
		return -1;
	LcpInit(psLCP);
	fsm_init(psFSM);

	// initialize AP
	psFSM = &(psPPP->fsm[Ap]);
	psFSM->psPPP = psPPP;
	psFSM->pdc = &ap_constants;
	if ((psFSM->pdv = psFSM->pdv_out = calloc(1,sizeof(AP),AllocPppcAp)) == 0 ||
			(psFSM->pdv_in = calloc(1,sizeof(AP),AllocPppcAp)) == 0)
		return -1;
	fsm_init(psFSM);

	// initialize IPCP
	psFSM = &(psPPP->fsm[IPcp]);
	psFSM->psPPP = psPPP;
	psFSM->pdc = &ipcp_constants;
	if ((psFSM->pdv = psFSM->pdv_out = psIPCP = calloc(1,sizeof(IPCP),AllocPppcIpcp)) == 0)
		return -1;
	IpcpInit(psIPCP);
	if ((psFSM->pdv_in = psIPCP = calloc(1,sizeof(IPCP),AllocPppcIpcp)) == 0)
		return -1;
	IpcpInit(psIPCP);
	fsm_init(psFSM);
	return 0;
}
