//
//   $Id: TCP.C 1.7 2003/01/07 22:55:10Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// TCP.C
// Copyright (C) 2001 Datalight, Inc.
// All Rights Reserved
// Datalight Confidential and Proprietary
// Portions Copyright (c) GPvNO 1990-2001
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
// Transmission Control Protocol implementation
// RFC 793 TCP Standard
// RFC 813 Window/Ack strategies - background info
// RFC1122 Requirements for Internet Hosts
//	RFC 2018 TCP Selective Acknowledgement Options - not implemented
// RFC 1191 Path MTU discovery - info
// RFC 896 Nagle Congestion Control in IP/TCP Internetworks
// RFC 2001 TCP Slow Start, Congestion Avoidance, Fast Retransmit, and Fast Recovery Algorithms - info
// RFC 1122 Requirements for Internet hosts
// RFC 2119 Key words for use in RFCs to Indicate Requirement Levels - info
// RFC 2414 Increasing TCP's Initial Window - not implemented
// RFC 2525 TCP implementation Problems - info

// Revision History
// 03/28/2002 gvn in_tcp for upcalls
// 07/23/2001 gvn Allocation tracing
// 04/13/2001 gvn Remove TCP precedence checking

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "inetdefs.h"
#include "timer.h"
#include "mbuf.h"
#include "netfunc.h"
#include "internet.h"
#include "tcp.h"
#include "ip.h"
#include "icmp.h"
#include "get16.h"
#include "log.h"

char hold_tcp;		// set to not immediately send TCP data

struct tcb *
open_tcp(lsocket,fsocket,mode,window,r_upcall,t_upcall,s_upcall,tos,user)
struct socket *lsocket;	/* Local socket */
struct socket *fsocket;	/* Remote socket */
int mode;		/* Active/passive/server */
WORD window;		/* Receive window (and send buffer) sizes */
void (far *r_upcall)();	/* Function to call when data arrives */
void (far *t_upcall)();	/* Function to call when ok to send more data */
void (far *s_upcall)();	/* Function to call when connection state changes */
char tos;
char *user;		/* User linkage area */
{
	extern char tcp_ext;
	struct connection conn;
	register struct tcb *tcb;

	if(lsocket == NULLSOCK){
		net_error = INVALID;
		return NULLTCB;
	}
	conn.local.port = lsocket->port;
	if(fsocket != NULLSOCK){
		conn.remote.address = fsocket->address;
		conn.remote.port = fsocket->port;
	} else {
		conn.remote.address = 0;
		conn.remote.port = 0;
	}
	if (!lsocket->address)
		lsocket->address = this_addr(conn.remote.address,0);
	conn.local.address = lsocket->address;
	if((tcb = lookup_tcb(&conn)) == NULLTCB){
		if((tcb = create_tcb(&conn)) == NULLTCB){
			net_error = NO_SPACE;
			return NULLTCB;
		}
	} else if (tcb->state != LISTEN) {
		net_error = CON_EXISTS;
		return NULLTCB;
	}
	else
		tcb->flags &= ~CLONE;
	tcb->user = user;
	if (window != 0)
		tcb->window = window;
	else
		tcb->window = tcp_stat.tcp_window;
	tcb->rcv.wnd = tcb->rcv_wndav = tcb->window;
	tcb->flags |= tcp_ext;
	tcp_ext = 0;
	tcb->r_upcall = r_upcall;
	tcb->t_upcall = t_upcall;
	tcb->s_upcall = s_upcall;
	tcb->tos = tos;
	switch (mode & 3) {
	case TCP_SERVER:
		tcb->flags |= CLONE;
		tcb->max_nacc = ((WORD)mode) >> 8;
	case TCP_PASSIVE:	/* Note fall-thru */
		tcb->flags1 |= F1_KEEPALIVE;	// servers all use keep alive
		setstate(tcb,LISTEN);
		break;
	case TCP_ACTIVE:
		/* Send SYN, go into SYN_SENT state */
		send_syn(tcb);
		setstate(tcb,SYN_SENT);
		tcp_output(tcb);
		tcp_stat.conout++;
		break;
	}
	return tcb;
}

// Accept cloned server connections
TCB *accept_tcp(TCB *tcb)
{
	TCB *ctcb;

	if (!(tcb->flags & CLONE)) {
		net_error = INVALID;
		return NULLTCB;
	}
	if ((ctcb = tcb->naccq) == 0 || tcb->num_nacc == 0) {
		net_error = WOULDBLK;
		return NULLTCB;
	}
	tcb->num_nacc--;
	tcb->naccq = ctcb->naccq;
	ctcb->naccq = 0;
	return ctcb;
}

/* User send routine */
int send_tcp(struct tcb *tcb,struct mbuf *bp,char urgent)
{
	WORD cnt;

	if(tcb == NULLTCB || bp == NULLBUF){
		free_p(bp);
		net_error = INVALID;
		return -1;
	}
	cnt = len_mbuf(bp);
#ifdef	TIGHT
	/* If this would overfill our send queue, reject it entirely */
	if(tcb->sndcnt + cnt > tcb->window){
		free_p(bp);
		net_error = WOULDBLK;
		return -1;
	}
#endif
	switch(tcb->state){
	case CLOSED:
		free_p(bp);
		net_error = NO_CONN;
		return -1;
	case LISTEN:	/* Change state from passive to active */
		send_syn(tcb);
		setstate(tcb,SYN_SENT);	/* Note fall-thru */
	case SYN_SENT:
	case SYN_RECEIVED:
	case ESTABLISHED:
	case CLOSE_WAIT:
		tcb->sndcnt += cnt;
		append(&tcb->sndq,bp);
		if (urgent) {
			tcb->flags |= SEND_URG;
			tcb->snd.up = tcb->snd.una + tcb->sndcnt - 1;
		}
		if (!hold_tcp)
			tcp_output(tcb);
		break;
	case FINWAIT1:
	case FINWAIT2:
	case CLOSING:
	case LAST_ACK:
	case TIME_WAIT:
		free_p(bp);
		net_error = CON_CLOS;
		return -1;
	}
	return cnt;
}
/* User receive routine */
int recv_tcp(struct tcb *tcb,struct mbuf **bp,WORD cnt,char *urgentp)
{
	char urgent = 0;

	if(tcb == NULLTCB || bp == (struct mbuf **)NULL){
		net_error = INVALID;
		return -1;
	}
	/* cnt == 0 means "I want it all" */
	if(cnt == 0)
		cnt = tcb->rcvcnt;
	/* If there's something on the queue, just return it regardless
	 * of the state we're in.
	 */
	if(tcb->rcvcnt != 0){
		if (tcb->cnt_to_urg) {
			urgent = URG_WAITING;
			if (cnt > tcb->cnt_to_urg)
				cnt = tcb->cnt_to_urg;
		}
		else if (tcb->urg_cnt) {
			urgent = URG_READ;
			if (cnt > tcb->urg_cnt)
				cnt = tcb->urg_cnt;
		}
		/* See if the user can take all of it */
		if(tcb->rcvcnt <= cnt){
			cnt = tcb->rcvcnt;
			*bp = tcb->rcvq;
			tcb->rcvq = NULLBUF;
		} else {
			if((*bp = alloc_mbuf(cnt,AllocTcpRcv)) == NULLBUF){
				net_error = NO_SPACE;
				if (urgentp)
					*urgentp = urgent;
				return -1;
			}
			pullup(&tcb->rcvq,(*bp)->data,cnt);
			(*bp)->cnt = cnt;
		}
		if (urgent & URG_WAITING)
			tcb->cnt_to_urg -= cnt;
		if (urgent & URG_READ) {
			if ((tcb->urg_cnt -= cnt) > 0)
				urgent = URG_READ | URG_WAITING;
		}
		tcb->rcvcnt -= cnt;
		tcb->rcv_wndav += cnt;
		// RFC 1122 4.2.3.3
		// Do a window update if the difference between the available window
		// and the offered window is >= min(max_window/2,mss_rcv)
		if ((tcb->rcv_wndav - tcb->rcv.wnd) >= min(tcb->mss_rcv,tcb->window / 2) &&
								tcb->state < CLOSE_WAIT) {
			tcb->rcv.wnd = tcb->rcv_wndav;
			tcb->flags |= FORCE;
			tcp_output(tcb);
		}
		if (urgentp)
			*urgentp = urgent;
		//printf("D %d\n",cnt);
		return cnt;
	} else {
		/* If there's nothing on the queue, our action depends on what state
		 * we're in (i.e., whether or not we're expecting any more data).
		 * If no more data is expected, then simply return 0; this is
		 * interpreted as "end of file".
		 */
		switch(tcb->state){
		case LISTEN:
		case SYN_SENT:
		case SYN_RECEIVED:
		case ESTABLISHED:
		case FINWAIT1:
		case FINWAIT2:
			*bp = NULLBUF;
			net_error = WOULDBLK;
			if (urgentp)
				*urgentp = urgent;
			return -1;
		case CLOSED:
		case CLOSE_WAIT:
		case CLOSING:
		case LAST_ACK:
		case TIME_WAIT:
			*bp = NULLBUF;
			if (urgentp)
				*urgentp = urgent;
			return 0;
		}
	}
	return 0;
}
/* This really means "I have no more data to send". It only closes the
 * connection in one direction, and we can continue to receive data
 * indefinitely.
 */
int close_tcp(struct tcb *tcb)
{
	if(tcb == NULLTCB){
		net_error = INVALID;
		return -1;
	}
	switch(tcb->state){
	case LISTEN:
	case SYN_SENT:
		close_self(tcb,NORMAL_R);
		return 0;
	case SYN_RECEIVED:
	case ESTABLISHED:
		tcb->sndcnt++;
		tcb->snd.nxt++;
		setstate(tcb,FINWAIT1);
		tcp_output(tcb);
		return 0;
	case CLOSE_WAIT:
		tcb->sndcnt++;
		tcb->snd.nxt++;
		setstate(tcb,LAST_ACK);
		tcp_output(tcb);
		return 0;
	case FINWAIT1:
	case FINWAIT2:
	case CLOSING:
	case LAST_ACK:
	case TIME_WAIT:
		net_error = CON_CLOS;
		return -1;
	}
	return -1;	/* "Can't happen" */
}
/* Delete TCB, free resources. The user is not notified, even if the TCB is
 * not in the CLOSED state. This function should normally be called by the
 * user only in response to a state change upcall to CLOSED state.
 */
int del_tcp(struct tcb *tcb)
{
	struct reseq *rp,*rp1;
	TCB *nq;

	if(tcb == NULLTCB){
		net_error = INVALID;
		return -1;
	}
	unlink_tcb(tcb);
	stop_timer(&tcb->timer);
	stop_timer(&tcb->rtt_timer);
	stop_timer(&tcb->ack_timer);
	for(rp = tcb->reseq;rp != NULLRESEQ;rp = rp1){
		rp1 = rp->next;
		free_p(rp->bp);
		free((char *)rp);
	}
	tcb->reseq = NULLRESEQ;
	free_p(tcb->rcvq);
	free_p(tcb->sndq);
	nq = tcb->naccq;
	free((char *)tcb);
	if (nq)
		del_tcp(nq);
	return 0;
}

#if APL
/* Do printf on a tcp connection */
/*VARARGS*/
int tprintf(struct tcb *tcb,const char *message,...)
{
	struct mbuf *bp;
	int len = 0;
	va_list ap;
	char dlin [200];

	if(tcb == NULLTCB)
		return len;

	va_start(ap, message);
	len = vsprintf(dlin,message,ap);
	va_end(ap);
	if ((bp = alloc_mbuf(len + 1,AllocTcpTprintf)) != 0) {
		strcpy(bp->data,dlin);
		bp->cnt = len;
		send_tcp(tcb,bp,0);
	}
	return len;
}
#endif
/* Return 1 if arg is a valid TCB, 0 otherwise */
int
tcpval(tcb)
struct tcb *tcb;
{
	register int i;
	register struct tcb *tcb1;

	if(tcb == NULLTCB)
		return 0;	/* Null pointer can't be valid */
	for(i=0;i<NTCB;i++){
		for(tcb1=tcp_stat.tcbs[i];tcb1 != NULLTCB;tcb1 = tcb1->next){
			if(tcb1 == tcb)
				return 1;
		}
	}
	return 0;
}

int kick_tcp(struct tcb *tcb)
{
	if(!tcpval(tcb))
		return -1;
	tcp_timeout(tcb);
	return 0;
}

int tcp_active()	/* check for active connections */
{
	register int i;
	register struct tcb *tcb;

	for (i=0;i<NTCB;i++) {
		for(tcb=tcp_stat.tcbs[i];tcb != NULLTCB;tcb = tcb->next){
			if(tcb->state > SYN_SENT && tcb->state < TIME_WAIT)
				return 1;
		}
	}
	return 0;
}

void tcp_resetall()	/* reset all active connections */
{
	register int i;
	register struct tcb *tcb,*tcbnext;

	for (i=0;i<NTCB;i++) {
		for(tcb=tcp_stat.tcbs[i];tcb != NULLTCB;tcb = tcbnext){
			tcbnext = tcb->next;
			reset_tcp(tcb);
		}
	}
}

void reset_tcp(struct tcb *tcb)
{
	if (tcb->state != LISTEN && tcb->state != SYN_SENT) {
		reset_current(tcb);
	}
	close_self(tcb,RESET);
}

void SetMustAck(TCB *tcb,char push)
{
	if (push && tcp_stat.cAckOnPush) {
		tcb->flags |= FORCE;
	}
	else {
		tcb->flags1 |= F1_MUSTACK;
		start_timer(&tcb->ack_timer);
	}
}


/* Return character string corresponding to a TCP well-known port, or
 * the decimal number if unknown.
 */
char *tcp_port(WORD n)
{
	static char buf[8];

	switch(n){
	case ECHO_PORT:
		return "echo";
	case DISCARD_PORT:
		return "discard";
	case FTPD_PORT:
		return "ftp_data";
	case FTP_PORT:
		return "ftp";
	case TELNET_PORT:
		return "telnet";
	case SMTP_PORT:
		return "smtp";
	case RLOGIN_PORT:
		return "rlogin";
	default:
		sprintf(buf,"%u",n);
		return buf;
	}
}

WORD getlport(void)
{
	extern WORD tcp_lport,tcp_lport_i;

	if (++tcp_lport == 0xffff)
		tcp_lport = tcp_lport_i;
    return tcp_lport;
}

// General TCP functions

static WORD hash_tcb(struct connection *conn);
void setstate(struct tcb *tcb,char newstate);

struct tcp_stat tcp_stat = {
	0,0,0,0,0,0,DEF_WND,DEF_MSS,DEF_RTT,6,0,0,
	{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},DEF_MSS
};

/* Lookup connection, return TCB pointer or NULLTCB if nonexistant */
struct tcb *lookup_tcb(struct connection *conn)
{
	register struct tcb *tcb;

	tcb = tcp_stat.tcbs[hash_tcb(conn)];
	while (tcb != NULLTCB) {
		if (conn->local.address == tcb->conn.local.address
		 && conn->remote.address == tcb->conn.remote.address
		 && conn->local.port == tcb->conn.local.port
		 && conn->remote.port == tcb->conn.remote.port)
			break;
		tcb = tcb->next;
	}
	return tcb;
}

/* Create a TCB, return pointer. Return pointer if TCB already exists. */
struct tcb *create_tcb(struct connection *conn)
{
	register struct tcb *tcb;
	void tcp_ack_timeout(char *arg);

	if((tcb = lookup_tcb(conn)) != NULLTCB)
		return tcb;
	if((tcb = (struct tcb *)calloc(1,sizeof (struct tcb),AllocTcpTcb)) == NULLTCB)
		return NULLTCB;
	tcb->conn = *conn;

	tcb->cwind = tcb->mss = tcp_stat.tcp_mss;
	tcb->ssthresh = 65535u;
	tcb->srtt = tcp_stat.tcp_irtt;
	/* Initialize timer intervals */
	tcb->timer.start = tcb->srtt / MSPTICK;
	tcb->timer.func = (void(*)(char *))tcp_timeout;
	tcb->timer.arg = tcb->ack_timer.arg = (char *)tcb;
	tcb->ack_timer.start = (tcb->srtt < 200 ? tcb->srtt : 200) / MSPTICK;
	tcb->ack_timer.func = (void (*)(char*))tcp_ack_timeout;
	tcb->rtt_timer.start = MAX_TIME; /* Largest possible value */

	link_tcb(tcb);
	return tcb;
}

/* Close our TCB */
void close_self(struct tcb *tcb,WORD reason)
{
	struct reseq *rp,*rp1;

	stop_timer(&tcb->timer);
	stop_timer(&tcb->rtt_timer);
   stop_timer(&tcb->ack_timer);
	tcb->reason = reason;

	/* Flush reassembly queue; nothing more can arrive */
	for(rp = tcb->reseq;rp != NULLRESEQ;rp = rp1){
		rp1 = rp->next;
		free_p(rp->bp);
		free((char *)rp);
	}
	tcb->reseq = NULLRESEQ;
	setstate(tcb,CLOSED);
}

/* Determine initial sequence number */
int32 iss()
{
	static int32 seq;

	seq += 250000l;
	return seq;
}

/* Sequence number comparisons
 * Return true if x is between low and high inclusive,
 * false otherwise
 */
int seq_within(int32 x,int32 low,int32 high)
{
	if(low <= high){
		if(low <= x && x <= high)
			return 1;
	} else {
		if(low >= x && x >= high)
			return 1;
	}
	return 0;
}

int seq_lt(int32 x,int32 y)
{
	return (long)(x-y) < 0;
}

int seq_le(int32 x,int32 y)
{
	return (long)(x-y) <= 0;
}

int seq_gt(int32 x,int32 y)
{
	return (long)(x-y) > 0;
}

int seq_ge(int32 x,int32 y)
{
	return (long)(x-y) >= 0;
}

/* Hash a connect structure into the hash chain header array */
static WORD hash_tcb(struct connection *conn)
{
	register WORD hval;

	/* Compute hash function on connection structure */
	hval = hiword(conn->remote.address);
	hval ^= loword(conn->remote.address);
	hval ^= hiword(conn->local.address);
	hval ^= loword(conn->local.address);
	hval ^= conn->remote.port;
	hval ^= conn->local.port;
	return hval % NTCB;
}
/* Insert TCB at head of proper hash chain */
void link_tcb(struct tcb *tcb)
{
	register struct tcb **tcbhead;
	WORD hash_tcb();

	tcb->prev = NULLTCB;
	idisable();
	tcbhead = &tcp_stat.tcbs[hash_tcb(&tcb->conn)];
	tcb->next = *tcbhead;
	if(tcb->next != NULLTCB){
		tcb->next->prev = tcb;
	}
	*tcbhead = tcb;
	popf();
}
/* Remove TCB from whatever hash chain it may be on */
void unlink_tcb(struct tcb *tcb)
{
	register struct tcb **tcbhead;
	WORD hash_tcb();

	idisable();
	tcbhead = &tcp_stat.tcbs[hash_tcb(&tcb->conn)];
	if(*tcbhead == tcb)
		*tcbhead = tcb->next;	/* We're the first one on the chain */
	if(tcb->prev != NULLTCB)
		tcb->prev->next = tcb->next;
	if(tcb->next != NULLTCB)
		tcb->next->prev = tcb->prev;
	popf();
}

void setstate(struct tcb *tcb,char newstate)
{
	register char oldstate;

	oldstate = tcb->state;
	tcb->state = newstate;
	tcb->flags |= TCP_STC;
	if (tcb->s_upcall) {
		if (tcb->flags & TCP_EXT) {
			extern unsigned stack_pointer;
			extern int in_tcp;

			asm push stack_pointer;
			stack_pointer = _SP - 200;
			in_tcp &= 0xf;
			(*tcb->s_upcall)(tcb,oldstate,newstate);
			in_tcp |= 0x70;
			asm pop stack_pointer;
		}
		else {
			(*tcb->s_upcall)(tcb,oldstate,newstate);
		}
	}
	/* Notify the user that he can begin sending data */
	if(newstate == ESTABLISHED){
		tcb->flags |= TCP_XMT;
		if (tcb->t_upcall) {
			if (tcb->flags & TCP_EXT) {
				extern unsigned stack_pointer;
				extern int in_tcp;

				asm push stack_pointer;
				stack_pointer = _SP - 200;
				in_tcp &= 0xf;
				(*tcb->t_upcall)(tcb,tcb->window - tcb->sndcnt,tcb->sndcnt);
				in_tcp |= 0x70;
				asm pop stack_pointer;
			}
			else {
				(*tcb->t_upcall)(tcb,tcb->window - tcb->sndcnt,tcb->sndcnt);
			}
		}
	}
}
/* Convert TCP header in host format into mbuf ready for transmission,
 * link in data (if any), and compute checksum
 */
struct mbuf *htontcp(struct tcp *tcph,struct mbuf *data,struct pseudo_header *ph)
{
	WORD hdrlen;
	struct mbuf *rval;
	BYTE *cp;
	WORD csum;

	hdrlen = (tcph->mss != 0) ? TCPLEN + MSS_LENGTH : TCPLEN;
	if ((rval = alloc_mbuf(hdrlen,AllocTcpHdr)) == 0) {
		return 0;
	}
	rval->cnt = hdrlen;

	cp = rval->data;
	cp = put16(cp,tcph->source);
	cp = put16(cp,tcph->dest);
	cp = put32(cp,tcph->seq);
	cp = put32(cp,tcph->ack);
	*cp++ = hdrlen << 2;	/* Offset field */
	*cp++ = tcph->flags;
	cp = put16(cp,tcph->wnd);
	*cp++ = 0;	/* Zero out checksum field */
	*cp++ = 0;
	cp = put16(cp,tcph->up);

	if(tcph->mss != 0){
		*cp++ = MSS_KIND;
		*cp++ = MSS_LENGTH;
		cp = put16(cp,tcph->mss);
	}
	rval->next = data;
	csum = ph ? cksum(ph,rval,ph->length) : tcph->checksum;
	cp = &rval->data[16];	/* Checksum field */
	*cp++ = csum >> 8;
	*cp = csum;
	return rval;
}
/* Pull TCP header off mbuf */
int ntohtcp(struct tcp *tcph,struct mbuf **bpp)
{
	int hdrlen,i,optlen,kind;
	BYTE hdrbuf[TCPLEN],*cp;

	i = pullup(bpp,hdrbuf,TCPLEN);
	tcph->source = get16(hdrbuf);
	tcph->dest = get16(hdrbuf + 2);
	tcph->seq = get32(hdrbuf + 4);
	tcph->ack = get32(hdrbuf + 8);
	hdrlen = (hdrbuf[12] & 0xf0) >> 2;
	tcph->flags = hdrbuf[13];
	tcph->wnd = get16(hdrbuf + 14);
	tcph->checksum = get16(hdrbuf + 16);
	tcph->up = get16(hdrbuf + 18);
	tcph->mss = 0;
	tcph->optlen = hdrlen - TCPLEN;

	/* Check for option field. Only space for one is allowed, but
	 * since there's only one TCP option (MSS) this isn't a problem
	 */
	if(i < TCPLEN || hdrlen < TCPLEN)
		return -1;	/* Header smaller than legal minimum */
	if(tcph->optlen == 0)
		return (int)hdrlen;	/* No options, all done */

	if(tcph->optlen > len_mbuf(*bpp)){
		/* Remainder too short for options length specified */
		return -1;
	}
	pullup(bpp,(BYTE *)tcph->options,tcph->optlen);	/* "Can't fail" */
	/* Process options */
	for(cp = (BYTE *)tcph->options,i=tcph->optlen; i > 0;){
		kind = *cp++;
		/* Process single-byte options */
		switch(kind){
		case EOL_KIND:
			i--;
			cp++;
			return (int)hdrlen;	/* End of options list */
		case NOOP_KIND:
			i--;
			cp++;
			continue;	/* Go look for next option */
		}
		/* All other options have a length field */
		optlen = uchar(*cp++);

		/* Process valid multi-byte options */
		switch(kind){
		case MSS_KIND:
			if(optlen == MSS_LENGTH){
				tcph->mss = get16(cp);
			}
			break;
		}
		optlen = max(2,optlen);	/* Enforce legal minimum */
		i -= optlen;
		cp += optlen - 2;
	}
	return (int)hdrlen;
}

// Process incoming TCP segments. Page number references are to ARPA RFC-793,
// the TCP specification.

#define MTUCHECK 1		/* 1 to check mtu of interface for mss */

void link_tcb();
static void reset(DWORD source,DWORD dest,char tos,int length,struct tcp *seg);
static int trim(struct tcb *tcb,struct tcp *seg,struct mbuf **bp,WORD *length);
static void proc_syn(struct tcb *tcb,char tos,struct tcp *seg);
static void update(TCB *tcb,struct tcp *seg);
static void add_reseq(struct tcb *tcb,char tos,struct tcp *seg,struct mbuf *bp,WORD length);
static void get_reseq(struct tcb *tcb,char *tos,struct tcp *seg,struct mbuf **bp,WORD *length);

/* This function is called from IP with the IP header in machine byte order,
 * along with a mbuf chain pointing to the TCP header.
 */
#pragma argsused
void tcp_input(IFACE *ifp,
struct mbuf *bp,	/* Data field, if any */
char protocol,		/* Should always be TCP_PTCL */
DWORD source,		/* Remote IP address */
DWORD dest,		    /* Our IP address */
char tos,		    /* Type of Service */
WORD length,		/* Length of data field */
char rxbroadcast)	/* Incoming broadcast - discard if true */
{
	register struct tcb *tcb;	/* TCP Protocol control block */
	struct tcp seg;			/* Local copy of segment header */
	struct connection conn;		/* Local copy of addresses */
	struct pseudo_header ph;	/* Pseudo-header for checksumming */
	int hdrlen;			/* Length of TCP header */
	MBUF *nbp;

	if(bp == NULLBUF)
		return;

	if(rxbroadcast){
		/* Any TCP packet arriving as a broadcast is
		 * to be completely IGNORED!!
		 */
		tcp_stat.bdcsts++;
		free_p(bp);
		return;
	}
	ph.source = source;
	ph.dest = dest;
	ph.protocol = protocol;
	ph.length = length;
	if(cksum(&ph,bp,length) != 0){
		/* Checksum failed, ignore segment completely */
		tcp_stat.checksum++;
		free_p(bp);
		return;
	}
	/* Form local copy of TCP header in host byte order */
	if((hdrlen = ntohtcp(&seg,&bp)) < 0){
		/* TCP header is too small */
		tcp_stat.runt++;
		free_p(bp);
		return;
	}
	length -= hdrlen;

	/* Fill in connection structure and find TCB */
	conn.local.address = dest;
	conn.local.port = seg.dest;
	conn.remote.address = source;
	conn.remote.port = seg.source;
	
	if((tcb = lookup_tcb(&conn)) == NULLTCB){
		struct tcb *ntcb,*ptcb;

		/* Check that this segment carries a SYN, and that
		 * there's a LISTEN on this socket with
		 * unspecified source address and port
		 */
		conn.remote.address = 0;
		conn.remote.port = 0;
		conn.local.address = 0;
		if (!(seg.flags & SYN) || ((tcb = lookup_tcb(&conn)) == NULLTCB &&
			(conn.local.address = dest) == 0,(tcb = lookup_tcb(&conn)) == NULLTCB)) {
			/* No unspecified LISTEN either, so reject */
			free_p(bp);
			reset(source,dest,tos,length,&seg);
			return;
		}
		/* We've found a server listen socket, so clone the TCB */
		if(tcb->flags & CLONE){
			if ((tcb->max_nacc && tcb->num_nacc >= tcb->max_nacc) ||
					(ntcb = (TCB *)malloc(sizeof(TCB),AllocTcpLTcb)) == NULLTCB) {
				free_p(bp);
				/* This may fail, but we should at least try */
				reset(source,dest,tos,length,&seg);
				return;
			}
			*ntcb = *tcb;
			ntcb->naccq = 0;
			ntcb->flags &= ~CLONE;
			if (tcb->max_nacc) {
				tcb->num_nacc++;
				for (ptcb = tcb;ptcb->naccq;ptcb = ptcb->naccq);
				ptcb->naccq = ntcb;
			}
			tcb = ntcb;
			tcb->timer.arg = tcb->ack_timer.arg = (char *)tcb;
		} else
			unlink_tcb(tcb);	/* It'll be put back on later */

		/* Stuff the foreign socket into the TCB */
		tcb->flags1 |= F1_LISTEN0;
		tcb->conn.remote.address = source;
		tcb->conn.remote.port = seg.source;
		tcb->conn.local.address = dest;

		/* NOW put on right hash chain */
		link_tcb(tcb);
	}
	//printf("R %4d %10lu %10lu %5u\n",length,seg.seq,seg.ack,seg.wnd);
	/* Do unsynchronized-state processing (p. 65-68) */
	switch(tcb->state){
	case CLOSED:
		free_p(bp);
		reset(source,dest,tos,length,&seg);
		return;
	case LISTEN:
		if(seg.flags & RST){
			free_p(bp);
			return;
		}
		if(seg.flags & ACK){
			free_p(bp);
			reset(source,dest,tos,length,&seg);
			return;
		}
		if(seg.flags & SYN){

			/* (Security check is bypassed) */
			/* page 66 */
			tcp_stat.conin++;
			proc_syn(tcb,tos,&seg);
			send_syn(tcb);
			setstate(tcb,SYN_RECEIVED);		
			if(length != 0 || seg.flags & FIN) {
				break;		/* Continue processing if there's more */
			}
			tcp_output(tcb);
		}
		free_p(bp);	/* Unlikely to get here directly */
		return;
	case SYN_SENT:
		if(seg.flags & ACK){
			if (!seq_within(seg.ack,tcb->iss+1,tcb->snd.nxt)) {
				free_p(bp);
				reset(source,dest,tos,length,&seg);
				return;
			}
		}
		if(seg.flags & RST){	/* p 67 */
			if(seg.flags & ACK){
				/* The ack must be acceptable since we just checked it.
				 * This is how the remote side refuses connect requests.
				 */
				close_self(tcb,RESET);
			}
			free_p(bp);
			return;
		}
		/* (Security check skipped here) */
		/* Check incoming precedence; it must match if there's an ACK
		if((seg.flags & ACK) && PREC(tos) != PREC(tcb->tos)){
			free_p(bp);
			reset(source,dest,tos,length,&seg);
			return;
		}
		*/
		if(seg.flags & SYN){
			proc_syn(tcb,tos,&seg);
			if(seg.flags & ACK){
				/* Our SYN has been acked, otherwise the ACK
				 * wouldn't have been valid.
				 */
				update(tcb,&seg);
				setstate(tcb,ESTABLISHED);
			} else {
				setstate(tcb,SYN_RECEIVED);
			}
			if(length != 0 || (seg.flags & FIN)) {
				break;		/* Continue processing if there's more */
			}
			tcp_output(tcb);
		} else {
			free_p(bp);	/* Ignore if neither SYN or RST is set */
		}
		return;
	}
	/* We reach this point directly in any synchronized state. Note that
	 * if we fell through from LISTEN or SYN_SENT processing because of a
	 * data-bearing SYN, window trimming and sequence testing "cannot fail".
	 */

	/* Trim segment to fit receive window. */
	if(trim(tcb,&seg,&bp,&length) == -1){
		/* Segment is unacceptable */
		if(!(seg.flags & RST)){
			tcb->flags |= FORCE;
			tcp_output(tcb);
		}
		return;
	}
	if (run_timer(&tcb->timer) && (tcb->flags1 & F1_KATIMING)) {
		start_timer(&tcb->timer);
	}
	/* If segment isn't the next one expected, and there's data
	 * or flags associated with it, put it on the resequencing
	 * queue, ACK it and return.
	 *
	 * Processing the ACK in an out-of-sequence segment without
	 * flags or data should be safe, however.
	 */
	if(seg.seq != tcb->rcv.nxt
	 && (length != 0 || (seg.flags & (SYN|FIN)) )){
		add_reseq(tcb,tos,&seg,bp,length);
		SetMustAck(tcb,seg.flags & PSH);
		tcp_output(tcb);
		return;
	}
	/* This loop first processes the current segment, and then
	 * repeats if it can process the resequencing queue.
	 */
	for(;;){
		/* We reach this point with an acceptable segment; all data and flags
		 * are in the window, and the starting sequence number equals rcv.nxt
		 * (p. 70)
		 */	
		if(seg.flags & RST){
			if(tcb->state == SYN_RECEIVED && !(tcb->flags & CLONE)){
				/* Go back to listen state only if this was
				 * not a cloned server TCB
				 */
				if (tcb->flags1 & F1_LISTEN0) {
					unlink_tcb(tcb);
					tcb->conn.remote.address = 0;
					tcb->conn.remote.port = 0;
					tcb->flags1 &= ~F1_LISTEN0;
					tcb->sndcnt = 0;
					link_tcb(tcb);
				}
				setstate(tcb,LISTEN);
			} else {
				close_self(tcb,RESET);
			}
			free_p(bp);
			return;
		}
		/* (Security check skipped here) p. 71 */
		/* Check for precedence mismatch or erroneous extra SYN */
		if(/*PREC(tos) != PREC(tcb->tos) ||*/ (seg.flags & SYN)){
			free_p(bp);
			reset(source,dest,tos,length,&seg);
			return;
		}
		/* Check ack field p. 72 */
		if(!(seg.flags & ACK)){
			free_p(bp);	/* All segments after synchronization must have ACK */
			return;
		}
		/* Process ACK */
		switch(tcb->state){
		case SYN_RECEIVED:
			if (seq_within(seg.ack,tcb->snd.una+1,tcb->snd.nxt)) {
				update(tcb,&seg);
				setstate(tcb,ESTABLISHED);
			} else {
				free_p(bp);
				reset(source,dest,tos,length,&seg);
				return;
			}
			break;
		case ESTABLISHED:
		case CLOSE_WAIT:
			update(tcb,&seg);
			break;
		case FINWAIT1:	/* p. 73 */
			update(tcb,&seg);
			if(tcb->sndcnt == 0){
				/* Our FIN is acknowledged */
				setstate(tcb,FINWAIT2);
			}
			break;
		case FINWAIT2:
			update(tcb,&seg);
			break;
		case CLOSING:
			update(tcb,&seg);
			if(tcb->sndcnt == 0){
				/* Our FIN is acknowledged */
				goto set_time_wait;
			}
			break;
		case LAST_ACK:
			update(tcb,&seg);
			if(tcb->sndcnt == 0){
				/* Our FIN is acknowledged, close connection */
				close_self(tcb,NORMAL_R);
				return;
			}
		case TIME_WAIT:
			//tcb->flags |= FORCE;	causes endless ACK loop if both sides land up in TIME_WAIT
			start_timer(&tcb->timer);
		}

		/* Process the segment text, if any, beginning at rcv.nxt (p. 74) */
		if(length != 0){
			switch(tcb->state){
			case SYN_RECEIVED:
			case ESTABLISHED:
			case FINWAIT1:
			case FINWAIT2:
				/* URGent bit processing) */
				if (seg.flags & URG) {
					if (!tcb->urg_cnt) {
						tcb->cnt_to_urg = tcb->rcvcnt;
					}
					tcb->urg_cnt += min(length,seg.up + 1);
				}
				/* Place on receive queue */
				tcb->rcvcnt += length;
				if (append(&tcb->rcvq,bp) > 4 && !tcb->urg_cnt) {
					/* combine long queue into single mbuf */
					if ((nbp = alloc_mbuf(tcb->rcvcnt,AllocTcpRCom)) != NULLBUF) {
						nbp->cnt = pullup(&tcb->rcvq,nbp->data,tcb->rcvcnt);
						tcb->rcvq = nbp;
					}
				}
				tcb->rcv.nxt += length;
				if ((tcb->rcv.wnd -= length) > 64000u)
					tcb->rcv.wnd = 0;
				if ((tcb->rcv_wndav -= length) > 64000u)
					tcb->rcv_wndav = 0;
				if (length > tcb->mss_rcv - 40) {
					if (!(tcb->flags1 & F1_MAXSEGRCVD)) {
						tcb->flags1 |= F1_MAXSEGRCVD;
						SetMustAck(tcb,seg.flags & PSH);
               }
					else {
						tcb->flags |= FORCE;
               }
				}
            else
					SetMustAck(tcb,seg.flags & PSH);
				break;
			default:
				/* Ignore segment text */
				free_p(bp);
				break;
			}
		}
		/* If the user has set up a r_upcall function and there's
		 * data to be read, notify him.
		 *
		 * This is done before sending an acknowledgement,
		 * to give the user a chance to piggyback some reply data.
		 * It's also done before processing FIN so that the state
		 * change upcall will occur after the user has had a chance
		 * to read the last of the incoming data.
		 */
		tcb->flags |= TCP_RCV;	/* set flag for API */
		if (tcb->r_upcall && tcb->rcvcnt != 0) {
			if (tcb->flags & TCP_EXT) {
				extern unsigned stack_pointer;
				extern int in_tcp;

				asm push stack_pointer;
				stack_pointer = _SP - 200;
				in_tcp &= 0xf;
				(*tcb->r_upcall)(tcb,tcb->rcvcnt);
				in_tcp |= 0x70;
				asm pop stack_pointer;
			}
			else
				(*tcb->r_upcall)(tcb,tcb->rcvcnt);
		}
		/* process FIN bit (p 75) */
		if(seg.flags & FIN){
			tcb->flags |= FORCE;	/* Always respond with an ACK */

			switch(tcb->state){
			case SYN_RECEIVED:
			case ESTABLISHED:
				tcb->rcv.nxt++;
				setstate(tcb,CLOSE_WAIT);
				break;
			case FINWAIT1:
				tcb->rcv.nxt++;
				if(tcb->sndcnt == 0){
					/* Our FIN has been acked; bypass CLOSING state */
					goto set_time_wait;
				} else {
					setstate(tcb,CLOSING);
				}
				break;
			case FINWAIT2:
				tcb->rcv.nxt++;
				set_time_wait:
				setstate(tcb,TIME_WAIT);
				tcb->timer.start = tcb->srtt > 4000 ? MSL2 * (1000/MSPTICK) :
									(tcb->srtt * 8)/MSPTICK;
				start_timer(&tcb->timer);
				break;
			case CLOSE_WAIT:
			case CLOSING:
			case LAST_ACK:
				break;		/* Ignore */
			case TIME_WAIT:	/* p 76 */
				start_timer(&tcb->timer);
				break;
			}
		}
		/* Scan the resequencing queue, looking for a segment we can handle,
		 * and freeing all those that are now obsolete.
		 */
		while(tcb->reseq != NULLRESEQ && seq_ge(tcb->rcv.nxt,tcb->reseq->seg.seq)){
			get_reseq(tcb,&tos,&seg,&bp,&length);
			if(trim(tcb,&seg,&bp,&length) == 0)
				goto gotone;
			/* Segment is an old one; trim has freed it */
		}
		break;
gotone:	;
	}
	tcp_output(tcb);	/* Send any necessary ack */
}

/* Process an incoming ICMP response */
void tcp_icmp(DWORD source,DWORD dest,char type,char code,struct mbuf **bpp)
//DWORD source;		/* Original IP datagram source (i.e. us) */
//DWORD dest;			/* Original IP datagram dest (i.e., them) */
//char type,code;		/* ICMP error codes */
//struct mbuf **bpp;	/* First 8 bytes of TCP header */
{
	struct tcp seg;
	struct connection conn;
	register struct tcb *tcb;

	/* Extract the socket info from the returned TCP header fragment
	 * Note that since this is a datagram we sent, the source fields
	 * refer to the local side.
	 */
	ntohtcp(&seg,bpp);
	conn.local.port = seg.source;
	conn.remote.port = seg.dest;
	conn.local.address = source;
	conn.remote.address = dest;
	if((tcb = lookup_tcb(&conn)) == NULLTCB)
		return;	/* Unknown connection, ignore */

	/* Verify that the sequence number in the returned segment corresponds
	 * to something currently unacknowledged. If not, it can safely
	 * be ignored.
	 */
	if (!seq_within(seg.seq,tcb->snd.una,tcb->snd.nxt))
		return;

	/* The strategy here is that Destination Unreachable and Time Exceeded
	 * messages that occur after a connection has been established are likely
	 * to be transient events, and shouldn't kill our connection (at least
	 * until after we've tried a few more times). On the other hand, if
	 * they occur on our very first attempt to send a datagram on a new
	 * connection, they're probably "for real". In any event, the info
	 * is saved.
	 */
	switch(uchar(type)){
	case DEST_UNREACH:
	case TIME_EXCEED:
		tcb->type = type;
		tcb->code = code;
		if(tcb->state == SYN_SENT || tcb->state == SYN_RECEIVED){
			close_self(tcb,NETWORK);
		}
		break;
	case QUENCH:
		/* Source quench; cut the congestion window in half,
		 * but don't let it go below one packet
		 */
		tcb->cwind /= 2;
		tcb->cwind = max(tcb->mss,tcb->cwind);
		break;
	}
}
/* Send an acceptable reset (RST) response for this segment
 * The RST reply is composed in place on the input segment
 */
static void reset(DWORD source,DWORD dest,char tos,int length,struct tcp *seg)
//DWORD source;			/* Remote IP address */
//DWORD dest;			/* Our IP address */
//char tos;			/* Type of Service */
//WORD length;			/* Length of data portion */
//register struct tcp *seg;	/* Offending TCP header */
{
	struct mbuf *hbp;
	struct pseudo_header ph;
	WORD tmp;
	char rflags;

	if(seg->flags & RST)
		return;	/* Never send an RST in response to an RST */

	tcp_stat.resets++;

	/* Compose the RST IP pseudo-header, swapping addresses */
	ph.source = dest;
	ph.dest = source;
	ph.protocol = TCP_PTCL;
	ph.length = TCPLEN;

	/* Swap port numbers */
	tmp = seg->dest;
	seg->dest = seg->source;
	seg->source = tmp;

	rflags = RST;
	if(seg->flags & ACK){
		/* This reset is being sent to clear a half-open connection.
		 * Set the sequence number of the RST to the incoming ACK
		 * so it will be acceptable.
		 */
		seg->seq = seg->ack;
		seg->ack = 0;
	} else {
		/* We're rejecting a connect request (SYN) from LISTEN state
		 * so we have to "acknowledge" their SYN.
		 */
		rflags |= ACK;
		seg->ack = seg->seq;
		seg->seq = 0;
		if(seg->flags & SYN)
			seg->ack++;
		seg->ack += length;
		if(seg->flags & FIN)
			seg->ack++;
	}
	seg->flags = rflags;
	seg->wnd = 0;
	seg->up = 0;
	seg->mss = 0;
	if ((hbp = htontcp(seg,NULLBUF,&ph)) != 0) {
		/* Ship it out (note swap of addresses) */
		ip_send(dest,source,TCP_PTCL,tos,0,hbp,ph.length,0,0,0);
	}
}

/* Send a reset (RST) as a result of local condition
 * ie user abort or "ulp" timeout
 */
void
reset_current(tcb)
register struct tcb *tcb;
{
	struct pseudo_header ph;/* Pseudo-header for checksum calcs */
	struct mbuf *hbp;	/* Header buffer pointer */
	struct tcp seg;		/* Local working copy of header */

	if(tcb == NULLTCB)
		return;

	seg.source = tcb->conn.local.port;
	seg.dest = tcb->conn.remote.port;
	seg.flags = RST;
	seg.mss = 0;
	seg.seq = tcb->snd.una;
	seg.ack = 0;
	seg.wnd = 0;
	seg.up = 0;

	/* Fill in fields of pseudo IP header */
	ph.source = tcb->conn.local.address;
	ph.dest = tcb->conn.remote.address;
	ph.protocol = TCP_PTCL;
	ph.length = TCPLEN;

	/* Generate TCP header, and compute checksum */
	if ((hbp = htontcp(&seg,NULLBUF,&ph)) != 0) {
		ip_send(tcb->conn.local.address,tcb->conn.remote.address,
				TCP_PTCL,tcb->tos,0,hbp,ph.length,0,0,0);
	}
}

/* Process an incoming acknowledgement and window indication.
 * From page 72.
 */
static void update(TCB *tcb,struct tcp *seg)
{
	WORD acked;
	WORD expand;
	WORD sndcnt;		// need this for compatibility

	acked = 0;
	if(seq_gt(seg->ack,tcb->snd.nxt)){
		tcb->flags |= FORCE;	/* Acks something not yet sent */
		return;
	}
	/* Decide if we need to do a window update.
	 * This is always checked whenever a legal ACK is received,
	 * even if it doesn't actually acknowledge anything,
	 * because it might be a spontaneous window reopening.
	 */
	if(seq_gt(seg->seq,tcb->snd.wl1) || ((seg->seq == tcb->snd.wl1)
	 && seq_ge(seg->ack,tcb->snd.wl2))){
		/* If the window had been closed, crank back the
		 * send pointer so we'll immediately resume transmission.
		 * Otherwise we'd have to wait until the next probe.
		 */
		if(tcb->snd.wnd == 0 && seg->wnd != 0)
			tcb->snd.ptr = tcb->snd.una;
		tcb->snd.wnd = seg->wnd;
		tcb->snd.wl1 = seg->seq;
		tcb->snd.wl2 = seg->ack;
	}
	/* Any ack will reset the retry counter because the other side is
	 * still alive (if not well !)
	 */
	tcb->retry = 0;
	/* See if anything new is being acknowledged */
	if(!seq_gt(seg->ack,tcb->snd.una))
		return;	/* Nothing more to do */

	/* We're here, so the ACK must have actually acked something */
	acked = (WORD)(seg->ack - tcb->snd.una);

	/* Expand congestion window if not already at limit */
	if(tcb->cwind < tcb->snd.wnd){
		if(tcb->cwind < tcb->ssthresh){
			/* Still doing slow start/CUTE, expand by amount acked */
			if ((expand = min(acked,tcb->mss)) == 1)
				expand = 0;
		} else {
			/* Steady-state test of extra path capacity */
			expand = (WORD)(((long)tcb->mss * tcb->mss) / tcb->cwind);
		}
		/* Guard against arithmetic overflow */
		if(tcb->cwind + expand < tcb->cwind)
			expand = MAXINT16 - tcb->cwind;

		/* Don't expand beyond the offered window */
		if(tcb->cwind + expand > tcb->snd.wnd)
			expand = tcb->snd.wnd - tcb->cwind;

		if(expand != 0){
			tcb->cwind += expand;
		}
	}
	/* Round trip time estimation */
	if(run_timer(&tcb->rtt_timer) && seq_ge(seg->ack,tcb->rttseq)){
		/* A timed sequence number has been acked */
		stop_timer(&tcb->rtt_timer);
		if(!(tcb->flags & RETRAN)){
			int32 rtt;	/* measured round trip time */
			int32 abserr;	/* abs(rtt - srtt) */

			/* This packet was sent only once and now
			 * it's been acked, so process the round trip time
			 */
			rtt = tcb->rtt_timer.start - tcb->rtt_timer.count;
			rtt *= MSPTICK;		/* milliseconds */

			/* If this ACKs our SYN, this is the first ACK
			 * we've received; base our entire SRTT estimate
			 * on it. Otherwise average it in with the prior
			 * history, also computing mean deviation.
			 */
			if(rtt > tcb->srtt &&
			 (tcb->state == SYN_SENT || tcb->state == SYN_RECEIVED)){
				tcb->srtt = rtt;
			} else {
				abserr = (rtt > tcb->srtt) ? rtt - tcb->srtt : tcb->srtt - rtt;
				tcb->srtt = ((AGAIN-1)*tcb->srtt + rtt) / AGAIN;
				tcb->mdev = ((DGAIN-1)*tcb->mdev + abserr) / DGAIN;
			}
			/* Reset the backoff level */
			tcb->backoff = 0;
		}
	}

	/* If we're waiting for an ack of our SYN, process it */
	switch(tcb->state){
	case SYN_SENT:
	case SYN_RECEIVED:
		acked--;
		tcb->sndcnt--;
	}
	/* Remove acknowledged bytes from the send queue and update the
	 * unacknowledged pointer. If a FIN is being acked,
	 * pullup won't be able to remove it from the queue.
	 */
	pullup(&tcb->sndq,NULLCHAR,acked);

	/* This will include the FIN if there is one */
	sndcnt = tcb->sndcnt;
	tcb->sndcnt -= acked;
	tcb->snd.una = seg->ack;
	if ((tcb->flags & SEND_URG) && seq_ge(tcb->snd.una,tcb->snd.up + 1))
		tcb->flags &= ~SEND_URG;
	/* Stop retransmission timer, but restart it if there is still
	 * unacknowledged data.
	 */	
	stop_timer(&tcb->timer);
	if(tcb->snd.una != tcb->snd.nxt) {
		start_timer(&tcb->timer);
	}
	else if (tcb->flags1 & F1_KEEPALIVE) {
		tcb->flags1 |= F1_KATIMING;
		tcb->timer.start = 60050u / MSPTICK;
		start_timer(&tcb->timer);
	}

	/* If retransmissions have been occurring, make sure the
	 * send pointer doesn't repeat ancient history
	 */
	if(seq_lt(tcb->snd.ptr,tcb->snd.una))
		tcb->snd.ptr = tcb->snd.una;

	/* Clear the retransmission flag since the oldest
	 * unacknowledged segment (the only one that is ever retransmitted)
	 * has now been acked.
	 */
	tcb->flags &= ~RETRAN;

	/* If outgoing data was acked, notify the user so he can send more
	 * unless we've already sent a FIN.
	 */
	if (acked != 0) {
		switch(tcb->state) {
	 	case ESTABLISHED:
		case CLOSE_WAIT:
			tcb->flags |= TCP_XMT;
			if (tcb->t_upcall) {
				if (tcb->flags & TCP_EXT) {
					extern unsigned stack_pointer;
					extern int in_tcp;

					asm push stack_pointer;
					stack_pointer = _SP - 200;
					in_tcp &= 0xf;
					(*tcb->t_upcall)(tcb,
						tcb->window - tcb->sndcnt,sndcnt);
					in_tcp |= 0x70;
					asm pop stack_pointer;
				}
				else {
					(*tcb->t_upcall)(tcb,
					    tcb->window - tcb->sndcnt,sndcnt);
				}
			}
		}
	}
}

/* Process an incoming SYN */
static void proc_syn(struct tcb *tcb,char tos,struct tcp *seg)
{
	WORD mtu,ip_mtu();

	tcb->flags |= FORCE;	/* Always send a response */

	/* Note: It's not specified in RFC 793, but SND.WL1 and
	 * SND.WND are initialized here since it's possible for the
	 * window update routine in update() to fail depending on the
	 * IRS if they are left uninitialized.
	 */
	/* Check incoming precedence and increase if higher */
	if(PREC(tos) > PREC(tcb->tos))
		tcb->tos = tos;
	tcb->rcv.nxt = seg->seq + 1;	/* p 68 */
	tcb->snd.wl1 = tcb->irs = seg->seq;
	tcb->snd.wnd = seg->wnd;
	if(seg->mss != 0)
		tcb->mss = seg->mss;
	/* Check the MTU of the interface we'll use to reach this guy
	 * and lower the MSS so that unnecessary fragmentation won't occur
	 */
	if((mtu = ip_mtu(tcb->conn.remote.address)) != 0){
		/* Allow space for the TCP and IP headers */
#if !MTUCHECK
		mtu = 10000;	/* to disable mtu fragmentation check */
#endif
		mtu -= TCPLEN + IPLEN;
		tcb->cwind = tcb->mss = min(mtu,tcb->mss);
	}
}

/* Generate an initial sequence number and put a SYN on the send queue */
void
send_syn(tcb)
register struct tcb *tcb;
{
	tcb->iss = iss();
	tcb->rttseq = tcb->snd.wl2 = tcb->snd.una = tcb->iss;
	tcb->snd.ptr = tcb->snd.nxt = tcb->rttseq;
	tcb->sndcnt++;
	tcb->flags |= FORCE;
}

/* Add an entry to the resequencing queue in the proper place */
static void add_reseq(struct tcb *tcb,char tos,struct tcp *seg,struct mbuf *bp,WORD length)
{
	register struct reseq *rp,*rp1;

	/* Allocate reassembly descriptor */
	if((rp = (struct reseq *)malloc(sizeof (struct reseq),AllocTcpReseq)) == NULLRESEQ){
		/* No space, toss on floor */
		free_p(bp);
		return;
	}
	rp->seg = *seg;
	rp->tos = tos;
	rp->bp = bp;
	rp->length = length;

	/* Place on reassembly list sorting by starting seq number */
	rp1 = tcb->reseq;
	if(rp1 == NULLRESEQ || seq_lt(seg->seq,rp1->seg.seq)){
		/* Either the list is empty, or we're less than all other
		 * entries; insert at beginning.
		 */
		rp->next = rp1;
		tcb->reseq = rp;
	} else {
		/* Find the last entry less than us */
		for(;;){
			if(rp1->next == NULLRESEQ || seq_lt(seg->seq,rp1->next->seg.seq)){
				/* We belong just after this one */
				rp->next = rp1->next;
				rp1->next = rp;
				break;
			}
			rp1 = rp1->next;
		}
	}
}

/* Fetch the first entry off the resequencing queue */
static void get_reseq(struct tcb *tcb,char *tos,struct tcp *seg,struct mbuf **bp,WORD *length)
{
	register struct reseq *rp;

	if((rp = tcb->reseq) == NULLRESEQ)
		return;

	tcb->reseq = rp->next;

	*tos = rp->tos;
	*seg = rp->seg;
	*bp = rp->bp;
	*length = rp->length;
	free((char *)rp);
}

/* Trim segment to fit window. Return 0 if OK, -1 if segment is
 * unacceptable.
 */
static int trim(struct tcb *tcb,struct tcp *seg,struct mbuf **bp,WORD *pwLength)
{
	struct mbuf *nbp;
	long dupcnt,excess;
	WORD len;		/* Segment length including flags */
	char accept;

	accept = 0;
	len = *pwLength;
	if(seg->flags & SYN)
		len++;
	if(seg->flags & FIN)
		len++;

	/* Acceptability tests */
	if(tcb->rcv.wnd == 0){
		/* Only in-order, zero-length segments are acceptable when our window
		 * is closed.
		 */
		if(seg->seq == tcb->rcv.nxt && len == 0){
			return 0;	/* Acceptable, no trimming needed */
		}
	} else {
		/* Some part of the segment must be in the window */
		if (len == 0 && seq_within(seg->seq,tcb->rcv.nxt,
					(int32)(tcb->rcv.nxt+tcb->rcv.wnd))) {
			return 0;
		} else if (seq_within(seg->seq,tcb->rcv.nxt,
				(int32)(tcb->rcv.nxt+tcb->rcv.wnd-1)) || // Begin is
				seq_within((int32)(seg->seq+len-1),tcb->rcv.nxt,
				 (int32)(tcb->rcv.nxt+tcb->rcv.wnd-1)) || // End is
				seq_within(tcb->rcv.nxt,seg->seq,
				 (int32)(seg->seq+len-1))) { 	// Straddles
#ifdef MSTCPBUG
				if (tcp_stat.NoMSTCPbug || seg->seq != tcb->rcv.nxt ||
					(int32)(seg->seq+len) <= (int32)(tcb->rcv.nxt+tcb->rcv.wnd))
#endif
					accept++;
		}
	}
	if (!accept) {
		free_p(*bp);
		return -1;
	}
	dupcnt = tcb->rcv.nxt - seg->seq;
	if(dupcnt > 0){
		tcb->rerecv += dupcnt;
		/* Trim off SYN if present */
		if(seg->flags & SYN){
			/* SYN is before first data byte */
			seg->flags &= ~SYN;
			seg->seq++;
			dupcnt--;
		}
		if(dupcnt > 0){
			pullup(bp,NULLCHAR,(WORD)dupcnt);
			seg->seq += dupcnt;
			*pwLength -= (WORD)dupcnt;
		}
	}
	excess = seg->seq + *pwLength - (tcb->rcv.nxt + tcb->rcv.wnd);
	if(excess > 0){
		tcb->rerecv += excess;
		/* Trim right edge */
		*pwLength -= (WORD)excess;
		nbp = copy_p(*bp,*pwLength,AllocTcpTrim);
		free_p(*bp);
		*bp = nbp;
		seg->flags &= ~FIN;	/* FIN follows last data byte */
	}
	return 0;
}


/*#define Catch if (*(int *)0) debug()*/
#define Catch

/* Send a segment on the specified connection. One gets sent only
 * if there is data to be sent or if "force" is non zero
 */
void tcp_output(struct tcb *tcb)
{
	struct pseudo_header ph;/* Pseudo-header for checksum calcs */
	struct mbuf *hbp,*dbp;	/* Header and data buffer pointers */
	WORD hsize;		/* Size of header */
	struct tcp seg;		/* Local working copy of header */
	WORD ssize;		/* Size of current segment being sent,
							* including SYN and FIN flags */
	WORD dsize;		/* Size of segment less SYN and FIN */
	WORD usable;		/* Usable window */
	WORD sent;			/* Sequence count (incl SYN/FIN) already in the pipe */

	if(tcb == NULLTCB)
		return;

	switch(tcb->state){
	case LISTEN:
	case CLOSED:
		return;	/* Don't send anything */
	}
	// check for connection started before local IP address was assigned
	// eg using PPP
	if (tcb->conn.local.address == 0) {
		DWORD address;
		if ((address = this_addr(tcb->conn.remote.address,0)) != 0) {
			unlink_tcb(tcb);
			tcb->conn.local.address = address;
			link_tcb(tcb);
		}
	}
	for (;;) {
		sent = (WORD)(tcb->snd.ptr - tcb->snd.una);

		/* There can only be one outstanding segment in this state
		 * since the other end would reject any segment without SYN
		 */
		if (tcb->state == SYN_SENT && sent != 0)
			break;
		if (tcb->snd.wnd == 0) {
			/* Allow only one closed-window probe at a time */
			if (sent != 0)
				usable = 0;
			else
				/* Force a closed-window probe */
				usable = 1;
		} else {
			/* usable window = offered window - unacked bytes in
			 * transit limited by the congestion window
			 */
			usable = min(tcb->snd.wnd,tcb->cwind) - sent;

			/* John Nagle's "single outstanding segment" rule.
			 * Allow only one segment in the pipeline unless there
			 * is enough unsent data to form at least one
			 * maximum-sized segment.
			 */
			if (!(tcp_stat.flush_flag & FLUSH_UNCONDITIONAL) &&
					sent != 0 &&
					tcb->sndcnt - sent < tcb->mss) {
				usable = 0;
			}
			if ((tcp_stat.flush_flag & FLUSH_DONTSEND) &&
					tcb->sndcnt - sent < tcb->mss) {
				usable = 0;
			}
		}
		/* Compute size of segment to send. This is either the usable
		 * window, the mss, or the amount we have on hand, whichever
		 * is less.
		 */
		ssize = min(tcb->sndcnt - sent,usable);
		ssize = min(ssize,tcb->mss);
		ssize = min(ssize,tcp_stat.tcp_smss);
		dsize = ssize;

		if (ssize == 0 && !(tcb->flags & FORCE))
			break;		/* No need to send anything */
		tcb->flags &= ~FORCE;	/* Only one forced segment! */
		if (tcb->flags1 & F1_MUSTACK) {
			stop_timer(&tcb->ack_timer);
		}
		tcb->flags1 &= ~(F1_MUSTACK | F1_MAXSEGRCVD);
		seg.source = tcb->conn.local.port;
		seg.dest = tcb->conn.remote.port;

		/* Set the SYN and ACK flags according to the state we're in.
		 * It is assumed that if this segment is associated with a
		 * state transition, then the state change will already have
		 * been made. This allows this routine to be called from a
		 * retransmission timeout with force=1.
		 * If SYN is being sent, adjust the dsize counter so we'll
		 * try to get the right amount of data off the send queue.
		 */
		seg.flags = ACK; /* Every state except SYN_SENT */
		hsize = TCPLEN;	/* Except when SYN being sent */
		seg.mss = 0;

		switch(tcb->state){
		case SYN_SENT:
			seg.flags = 0;	/* Note fall-thru */
		case SYN_RECEIVED:
			if(tcb->snd.ptr == tcb->iss){
				WORD mtu;
				seg.flags |= SYN;
				dsize--;
				/* Also send MSS */
				if((mtu = ip_mtu(tcb->conn.remote.address)) != 0){
					mtu -= 40;
					seg.mss = min(mtu,tcp_stat.tcp_mss);
				}
				else
					seg.mss = tcp_stat.tcp_mss;
				tcb->mss_rcv = seg.mss;
				hsize = TCPLEN + MSS_LENGTH;
			}
			break;
		}
		seg.seq = (usable || tcb->snd.wnd != 0) ? tcb->snd.ptr :
							  tcb->snd.una;
		seg.ack = tcb->rcv.nxt;
		// if all received data has been delivered and we have something to send,
		// then offer full window
		if (ssize && tcb->rcv_wndav == tcb->window)
			tcb->rcv.wnd = tcb->rcv_wndav;
		seg.wnd = tcb->rcv.wnd;
		if (tcb->flags & SEND_URG) {
			seg.flags |= URG;
			seg.up = (WORD)(tcb->snd.up - tcb->snd.ptr);
		}
		else
			seg.up = 0;

		/* Now try to extract some data from the send queue.
		 * Since SYN and FIN occupy sequence space and are reflected
		 * in sndcnt but don't actually sit in the send queue,
		 * dup_p will return one less than dsize if a FIN needs to be
		 * sent.
		 */
		if(dsize != 0){
			if(dup_p(&dbp,tcb->sndq,sent,dsize,AllocTcpTrimFin) != dsize){
				/* We ran past the end of the send
					queue; send a FIN */
				seg.flags |= FIN;
				dsize--;
			}
		} else
			dbp = NULLBUF;

		/* If the entire send queue will now be in the pipe, set the
		 * push flag
		 */
		if(dsize != 0 && sent + ssize == tcb->sndcnt)
			seg.flags |= PSH;

		/* If this transmission includes previously transmitted data,
		 * snd.nxt will already be past snd.ptr. In this case,
		 * compute the amount of retransmitted data and keep score
		 */
		if(tcb->snd.ptr < tcb->snd.nxt)
			tcb->resent += min(tcb->snd.nxt - tcb->snd.ptr,ssize);

		tcb->snd.ptr += ssize;
		/* If this is the first transmission of a range of sequence
		 * numbers, record it so we'll accept acknowledgments
		 * for it later
		 */
		if(seq_gt(tcb->snd.ptr,tcb->snd.nxt))
			tcb->snd.nxt = tcb->snd.ptr;

		/* Fill in fields of pseudo IP header */
		ph.source = tcb->conn.local.address;
		ph.dest = tcb->conn.remote.address;
		ph.protocol = TCP_PTCL;
		ph.length = hsize + dsize;

		/* Generate TCP header, compute checksum, and link in data */
		if ((hbp = htontcp(&seg,dbp,&ph)) == NULLBUF) {
			free_p(dbp);
		}

		/* If we're sending some data or flags, start retransmission
		 * and round trip timers if they aren't already running.
		 */
		if(ssize != 0) {
			DWORD mstime = (DWORD)backoff(tcb->backoff) *
						(2 * tcb->mdev + tcb->srtt + 2 * MSPTICK);
			if (tcp_stat.tcp_time_max && (DWORD)tcp_stat.tcp_time_max < mstime)
				mstime = (DWORD)tcp_stat.tcp_time_max;
			tcb->timer.start = mstime / MSPTICK;
			if(!run_timer(&tcb->timer) || (tcb->flags1 & F1_KATIMING)) {
				tcb->flags1 &= ~F1_KATIMING;
				start_timer(&tcb->timer);
			}
			/* If round trip timer isn't running, start it */
			if(!run_timer(&tcb->rtt_timer)){
				start_timer(&tcb->rtt_timer);
				tcb->rttseq = tcb->snd.ptr;
			}
		}
		if (hbp) {
			ip_send(tcb->conn.local.address,
				tcb->conn.remote.address,
				TCP_PTCL,tcb->tos,0,hbp,ph.length,0,0,0);
		}
	}
}

// TCP timeout functions
// Ack Timer timeout
void tcp_ack_timeout(char *arg)
{
	struct tcb *tcb = (struct tcb *)arg;

	if(tcb == NULLTCB)
		return;

	tcb->flags |= FORCE;
	tcp_output(tcb);
}

/* Timer timeout */
void tcp_timeout(struct tcb *tcb)
{
	register struct mbuf *bp;

	if(tcb == NULLTCB)
		return;

	/* Make sure the timer has stopped (we might have been kicked) */
	stop_timer(&tcb->timer);

	switch(tcb->state){
	case TIME_WAIT:		/* 2MSL timer has expired */
		close_self(tcb,NORMAL_R);
		break;
	default:		/* Retransmission timer has expired */
		if (tcb->flags1 & F1_KATIMING) {
#ifdef K
			tcb->flags |= FORCE;
			tcp_output(tcb);
#endif
			tcb->snd.ptr = tcb->snd.nxt = tcb->snd.una = tcb->snd.wl2 =
								tcb->snd.una - 1;
			if ((bp = alloc_mbuf(1,AllocTcpKa)) != 0) {
				bp->cnt = 1;
				*bp->data = 0;
				send_tcp(tcb,bp,0);
				break;
			}
			start_timer(&tcb->timer);
			break;
		}
		tcb->flags |= RETRAN;	/* Indicate > 1  transmission */
		++tcb->backoff;
		if (++tcb->retry >= tcp_stat.tcp_retry) {
#if APL
			log(tcb,"reset - too many timeouts: %d",tcb->retry);
#endif
			if(tcb->state == SYN_RECEIVED && !(tcb->flags & CLONE)){
				/* Go back to listen state only if this was
				 * not a cloned server TCB
				 */
				if (tcb->flags1 & F1_LISTEN0) {
					unlink_tcb(tcb);
					tcb->conn.remote.address = 0;
					tcb->conn.remote.port = 0;
					tcb->flags1 &= ~F1_LISTEN0;
					tcb->sndcnt = 0;
					link_tcb(tcb);
				}
				setstate(tcb,LISTEN);
				break;
			}
			if (tcb->state != LISTEN && tcb->state != SYN_SENT) {
				reset_current(tcb);
			}
			close_self(tcb,TIMEOUT);
			break;
		}
		tcb->snd.ptr = tcb->snd.una;
		/* Reduce slowstart threshold to half current window */
		tcb->ssthresh = tcb->cwind / 2;
		tcb->ssthresh = max(tcb->ssthresh,tcb->mss);
		/* Shrink congestion window to 1 packet */
		tcb->cwind = tcb->mss;
		tcp_output(tcb);
	}
}
/* Backoff function - the subject of much research */
int backoff(int n)
{
	/* Use binary exponential up to retry #4, and quadratic after that
	 * This yields the sequence
	 * 1, 2, 4, 8, 16, 25, 36, 49, 64, 81, 100 ...
	 */

	if(n <= 4)
		return 1 << n;	/* Binary exponential back off */
	else
		return n * n;	/* Quadratic back off */
}

