//
//   $Id: TIMER.C 1.4 2002/10/16 16:42:12Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// TIMER.C
// Timer functions
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
// change log
// 03/28/2002 gvn in_tcp for upcall

#include <stdio.h>
#include <dos.h>
#include "inetdefs.h"
#include "timer.h"
#include "log.h"

WORD sec_timer;		/* one second timer */

static WORD mstimer;

/* Head of running timer chain */
struct timer *timers;

void tick()
{
	struct timer *t,*tp;
	struct timer *expired = NULLTIMER;
	WORD num_expired = 0,num_run = 0,num_timers = 0;	// debug
	int i;
	static char re_enter;
	extern main(),pether();

	if (re_enter)
		return;
	++re_enter;
	if ((mstimer += MSPTICK) >= 1000) {
		++sec_timer;
		mstimer -= 1000;
	}
	/* Run through the list of running timers, decrementing each one.
	 * If one has expired, take it off the running list and put it
	 * on a singly linked list of expired timers
	 */
	idisable();
	for(t = timers,i = 0;t != NULLTIMER; t = tp,i++) {
		tp = t->next;
		++num_timers;	// debug
		if(tp == t || i > 1000) {
			popf();
			fatal("Timer loop");
		}
		if((t->state & TIMER_RUN) && --(t->count) == 0){
			stop_timer(t);
			t->state = TIMER_EXPIRE | (t->state & TIMER_EXT);
			/* Put on head of expired timer list */
			t->next = expired;
			expired = t;
			++num_expired;	// debug
		}
	}
	popf();
	/* Now go through the list of expired timers, removing each
	 * one and kicking the notify function, if there is one
	 */
	while((t = expired) != NULLTIMER){
		expired = t->next;
		if(t->func){
			if (t->state & TIMER_EXT) {
				extern unsigned stack_pointer;
				extern int in_tcp;

				asm push stack_pointer;
				stack_pointer = _SP - 200;
				in_tcp &= 0xf;
				(*t->func)(t->arg);
				in_tcp |= 0x70;
				asm pop stack_pointer;
			}
			else
				(*t->func)(t->arg);
		}
		++num_run;	// debug
	}
	--re_enter;
#if APL
	{	// debugging code
		extern char rip_advertise;
		extern TIMER rip_timer;

		if (rip_advertise && !(rip_timer.state & TIMER_RUN)) {
			start_timer(&rip_timer);
			log(0,"RIP timer lost, num=%u, exp=%u, run=%u",
					num_timers,num_expired,num_run);
		}
	}	// end debug code
#endif
}
/* Start a timer */
void start_timer(register struct timer *t)
{
	if (t == NULLTIMER)
		return;

	idisable();
	t->count = t->start ? t->start : 1;
	if (!(t->state & TIMER_RUN)) {
		t->state = TIMER_RUN | (t->state & TIMER_EXT);
		/* Put on head of active timer list */
		t->prev = NULLTIMER;
		t->next = timers;
		if(t->next != NULLTIMER)
			t->next->prev = t;
		timers = t;
	}
	popf();
}
/* Stop a timer */
void stop_timer(register struct timer *t)
{
	if(t == NULLTIMER)
		return;
	idisable();
	if (t->state & TIMER_RUN) {
		/* Delete from active timer list */
		if(timers == t)
			timers = t->next;
		if(t->next != NULLTIMER)
			t->next->prev = t->prev;
		if(t->prev != NULLTIMER)
			t->prev->next = t->next;
	}
	t->state = TIMER_STOP | (t->state & TIMER_EXT);
	popf();
}
