//
//   $Id: UDP.C 1.6 2003/01/07 22:53:02Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// UDP.C
// User Datagram Protocol implememntation
// Copyright (c) GP van Niekerk Ondernemings 1990-1998
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

//#pragma inline
#include <stdlib.h>
#include <mem.h>
#include "inetdefs.h"
#include "mbuf.h"
#include "netfunc.h"
#include "udp.h"
#include "internet.h"
#include "ip.h"
#include "iface.h"

// change history
// 2002-12-18 gvn make udp tables accessable to external program
// 2002-03-28 gvn in_tcp for upcalls

struct udp_stat udp_stat;	/* Statistics */
static WORD hash_udp(struct socket *socket);

/* Create a UDP control block for lsocket, so that we can queue
 * incoming datagrams.
 */
int open_udp(struct socket *lsocket,void (far *r_upcall)(),int ext)
{
	register struct udp_cb *up;
	WORD hval;

	if((up = lookup_udp(lsocket)) != NULLUDP)
		return 0;	/* Already exists */
	if ((up = (struct udp_cb *)calloc(1,sizeof (struct udp_cb),AllocUdpUdp)) == 0) {
		net_error = NO_SPACE;
		return -1;
	}
	up->socket.address = lsocket->address;
	up->socket.port = lsocket->port;
	up->r_upcall = r_upcall;
	up->flags = ext ? UDPF_EXT : 0;
	hval = hash_udp(lsocket);
	up->next = udp_stat.udps[hval];
	up->prev = NULLUDP;
	if(up->next != NULLUDP)
		up->next->prev = up;
	udp_stat.udps[hval] = up;
	return 1;
}

/* Send a UDP datagram */
int send_udp(
	struct socket *lsocket,	/* Source socket */
	struct socket *fsocket,	/* Destination socket */
	char tos,				/* Type-of-service for IP */
	char ttl,				/* Time-to-live for IP */
	struct mbuf *bp,		/* Data field, if any */
	WORD length,			/* Length of data field */
	WORD id,				/* Optional ID field for IP */
	char df,				/* Don't Fragment flag for IP */
	IFACE *ifp)				// send directly to interface if set
{
	struct mbuf *htonudp();
	struct pseudo_header ph;
	struct udp udp;
	extern IFACE *ifpsecurity;
	int numif = 0;
	int32 netmask;
	MBUF *tbp;

	length = UDPHDR;
	if(bp != NULLBUF)
		length += len_mbuf(bp);

	udp.source = lsocket->port;
	udp.dest = fsocket->port;
	udp.length = length;

	/* Create IP pseudo-header, compute checksum and send it */
	ph.length = length;
	ph.dest = fsocket->address;
	ph.protocol = UDP_PTCL;
	if ((ph.dest & 0x7fffffffl) == 0x7fffffffl) {	// broadcast
		if (ifpsecurity) {
			lsocket->address = ifpsecurity->ip_addr;
			ph.source = lsocket->address;
			bp = htonudp(&udp,bp,&ph);
			udp_stat.sent++;
			ip_send(lsocket->address,fsocket->address,UDP_PTCL,tos,ttl,bp,
				length,id,df,ifpsecurity);
			ifpsecurity = 0;
		}
		else {
			if (ifp) {
				numif = 1;
				goto send_to_interface;
			}
			for (ifp = ifaces,numif = 0;ifp;ifp = ifp->next) {
				if (!(ifp->sflags & (SF_RAW | SF_UNC)))
					++numif;
			}
			for (ifp = ifaces;ifp && numif;ifp = ifp->next) {
				send_to_interface:
				if (!(ifp->sflags & (SF_RAW | SF_UNC))) {
					if (--numif) {
						if ((tbp = copy_p(bp,len_mbuf(bp),AllocUdpDup)) == NULLBUF)
							continue;
					}
					else
						tbp = bp;
					if (!(ph.dest & 0x80000000l)) {
						netmask = NETMASK(ifp->netbits);
						ph.dest = (ifp->ip_addr & netmask) | ~netmask;
					}
					fsocket->address = ph.dest;
					lsocket->address = ifp->ip_addr;
					ph.source = lsocket->address;
					tbp = htonudp(&udp,tbp,&ph);
					udp_stat.sent++;
					ip_send(lsocket->address,fsocket->address,UDP_PTCL,tos,ttl,tbp,length,id,df,ifp);
				}
			}
		}
		if (numif)
			free_p(bp);
	}
	else {
		if (!lsocket->address)
			lsocket->address = this_addr(fsocket->address,&ifp);
		ph.source = lsocket->address;
		bp = htonudp(&udp,bp,&ph);
		udp_stat.sent++;
		ip_send(lsocket->address,fsocket->address,UDP_PTCL,tos,ttl,bp,length,id,df,ifp);
	}
	return length;
}

/* Accept a waiting datagram, if available. Returns length of datagram */
int recv_udp(
	struct socket *lsocket,		/* Local socket to receive on */
	struct socket *fsocket,		/* Place to stash incoming socket */
	struct mbuf **bp)		/* Place to stash data packet */
{
	register struct udp_cb *up;
	struct socket *sp;
	struct mbuf *buf;
	WORD length;

	up = lookup_udp(lsocket);
	if(up == NULLUDP){
		net_error = NO_CONN;
		return -1;
	}
	if(up->rcvcnt == 0){
		net_error = WOULDBLK;
		return -1;
	}
	buf = dequeue(&up->rcvq);
	up->rcvcnt--;

	sp = (struct socket *)buf->data;
	/* Fill in the user's foreign socket structure, if given */
	if(fsocket != NULLSOCK){
		fsocket->address = sp->address;
		fsocket->port = sp->port;
	}
	/* Strip socket header and hand data to user */
	pullup(&buf,NULLCHAR,sizeof(struct socket));
	length = len_mbuf(buf);
	if(bp != (struct mbuf **)NULL)
		*bp = buf;
	else
		free_p(buf);
	return length;
}
/* Delete a UDP control block */
int del_udp(struct socket *lsocket)
{
	register struct udp_cb *up;
	struct mbuf *bp;
	WORD hval;

	if((up = lookup_udp(lsocket)) == NULLUDP){
		net_error = INVALID;
		return -1;
	}		
	/* Get rid of any pending packets */
	while(up->rcvcnt != 0){
		bp = up->rcvq;
		up->rcvq = up->rcvq->anext;
		free_p(bp);
		up->rcvcnt--;
	}
	hval = hash_udp(&up->socket);
	if(udp_stat.udps[hval] == up){
		/* First on list */
		if ((udp_stat.udps[hval] = up->next) != 0)
			up->next->prev = 0;
	} else {
		if ((up->prev->next = up->next) != 0)
			up->next->prev = up->prev;
	}
	free((char *)up);
	return 0;
}
/* Process an incoming UDP datagram */
#pragma argsused
void udp_input(IFACE *ifp,MBUF *bp,char protocol,DWORD source,
	DWORD dest,char tos,WORD length,char rxbroadcast)
{
	struct pseudo_header ph;
	struct udp udp;
	struct udp_cb *up;
	struct socket lsocket;
	struct socket *fsocket;
	struct mbuf *sp;
	int ckfail = 0;

	if(bp == NULLBUF)
		return;

	udp_stat.rcvd++;

	/* Create pseudo-header and verify checksum */
	ph.source = source;
	ph.dest = dest;
	ph.protocol = protocol;
	ph.length = length;

	if(cksum(&ph,bp,length) != 0)
		/* Checksum apparently failed, note for later */
		ckfail++;

	/* Extract UDP header in host order */
	ntohudp(&udp,&bp);

	/* If the checksum field is zero, then ignore a checksum error.
	 * I think this is dangerously wrong, but it is in the spec.
	 */
	if(ckfail && udp.checksum != 0){
		udp_stat.cksum++;
		free_p(bp);
		return;
	}
	lsocket.port = udp.dest;
	lsocket.address = 0;	// assume a non-specific listen
	/* See if there's somebody around to read it */
	if((up = lookup_udp(&lsocket)) == NULLUDP){
		// not this one, try specific
		if(rxbroadcast){
			lsocket.address = ifp->ip_addr;
			udp_stat.bdcsts++;
		} else
			lsocket.address = dest;

		if((up = lookup_udp(&lsocket)) == NULLUDP){
			// no luck, drop it
			udp_stat.unknown++;
			free_p(bp);
			return;
		}
	}
	/* Create a buffer which will contain the foreign socket info */
	if((sp = alloc_mbuf(sizeof(struct socket),AllocUdpSock)) == NULLBUF){
		/* No space, drop whole packet */
		free_p(bp);
		return;
	}
	sp->cnt = sizeof(struct socket);

	fsocket = (struct socket *)sp->data;
	fsocket->address = source;
	fsocket->port = udp.source;

	/* Yes, remove the now redundant UDP header, chain the foreign socket
	 * info in front of it and queue it
	 */

	sp->next = bp;
	enqueue(&up->rcvq,sp);
	up->rcvcnt++;
	if(up->r_upcall) {
		if (up->flags & UDPF_EXT) {
			extern unsigned stack_pointer;
			extern int in_tcp;

			asm push stack_pointer;
			stack_pointer = _SP - 200;
			in_tcp &= 0xf;
			(*up->r_upcall)(&lsocket,up->rcvcnt,ifp);
			in_tcp |= 0x70;
			asm pop stack_pointer;
		}
		else
			(*up->r_upcall)(&lsocket,up->rcvcnt,ifp);
	}
}
/* Look up UDP socket, return control block pointer or NULLUDP if nonexistant */
struct udp_cb *lookup_udp(struct socket *socket)
{
	register struct udp_cb *up;

	up = udp_stat.udps[hash_udp(socket)];
	while(up != NULLUDP){
		if(memcmp((char *)socket,(char *)&up->socket,sizeof(struct socket)) == 0)
			break;
		up = up->next;
	}
	return up;
}

/* Hash a UDP socket (address and port) structure */
static WORD hash_udp(struct socket *socket)
{
	WORD hval;

	/* Compute hash function on socket structure */
	hval = hiword(socket->address);
	hval ^= loword(socket->address);
	hval ^= socket->port;
	return hval % (WORD)NUDP;
}
/* Convert UDP header in internal format to an mbuf in external format */
MBUF *htonudp(struct udp *udp,MBUF *bp,struct pseudo_header *ph)
{
	struct mbuf *hbp;
	BYTE *cp;
	WORD checksum;

	/* Allocate UDP protocol header and fill it in */
	if((hbp = alloc_mbuf(UDPHDR,AllocUdpHdr)) == NULLBUF){
		net_error = NO_SPACE;
		return NULLBUF;
	}
	hbp->cnt = UDPHDR;

	cp = hbp->data;
	cp = put16(cp,udp->source);	/* Source port */
	cp = put16(cp,udp->dest);	/* Destination port */
	cp = put16(cp,udp->length);	/* Length */
	*cp++ = 0;			/* Clear checksum */
	*cp-- = 0;

	/* Link in the user data */
	hbp->next = bp;

	/* All zeros and all ones is equivalent in one's complement arithmetic;
	 * the spec requires us to change zeros into ones to distinguish an
	 * all-zero checksum from no checksum at all
	 */
	if((checksum = cksum(ph,hbp,ph->length)) == 0)
		checksum = 0xffff;
	put16(cp,checksum);
	return hbp;
}
/* Convert UDP header in mbuf to internal structure */
void ntohudp(struct udp *udp,struct mbuf **bpp)
{
	udp->source = pull16(bpp);
	udp->dest = pull16(bpp);
	udp->length = pull16(bpp);
	udp->checksum = pull16(bpp);
}
