//
//   $Id: DosConsole.java 1.3 2003/01/16 23:34:22Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// HISTORY
//    1/16/2003   jmb   added support for checking for special keys that
//                      generate a zero keycode.
//
/*
  A class implementing a remote dos console.

  Developed for use with ROM-DOS RT and TSR HTTP server.
*/
/**
  Protocol: All data sent is text, except for <esc> and what follows:
  Special chars following esc:

setup:

  <m mode>   - video mode
  <s x y>   - screen size.

inline:

  <esc>      - dipslay esc char
  <p x y>   - data input position. x,y == byte values
  <c x y>   - position cursor. Same format as above
  <a atr>   - change attribute - exact dos attribute. All data that follows assumes this attr.
      atr: byte {Blink, red backg, green backg, blue backg, intensity, red foreg, green foreg, blue foreg} mono: underline = 0x1.
  <r>      - reset state machine to normal op. Same as no operation.
*/
package inet;

import java.net.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;


public class DosConsole
extends Emul

{
   //map dos chars to unicode (java) chars
   protected static final byte BLINK = (byte)0x80;
   protected static final byte INTENSE = 0x8;
   protected static final byte RED = 0x4;
   protected static final byte GREEN = 0x2;
   protected static final byte BLUE = 0x1;
   protected static Color[] dosColors = new Color[16];

//receiving
   public static final byte T_SERINFO      = (byte)'i';//receive string containing serial info
   public static final byte T_SCREENSIZE   = (byte)'o';//receive rows and cols
   public static final byte T_SCROLL       = (byte)'s';
   public static final byte T_MOVE         = (byte)'m';
   public static final byte T_CARET      = (byte)'c';
   public static final byte T_DATA         = (byte)'d';
   public static final byte T_PATTERN      = (byte)'p';
   public static final byte T_ATRPATTERN    = (byte)'a';//atr forms pattern, eg text on white&black screen.
   public static final byte LOWERCASE      = 0x20;

//sending:
   public static final byte KEYPRESS      = (byte)'k';//receives a key, {'k', scancode, ascii}
   public static final byte SHIFSTATE      = (byte)'s';//changes shiftstate {'s', shiftstate}
   public static final byte REFRESH      = (byte)'r';//refreshes whole screen {'r'}
   public static final byte REBOOT         = (byte)'b';//reboots the server

   Constants keymap;
   Constants special;

   //set up Colors
   static {
      for (int i = 0; i < 16; i++) {
         boolean intense = (i & INTENSE) != 0;
         int r = (i & RED)   != 0 ? intense ? 255 : 200 : intense ? 84 : 0;
         int g = (i & GREEN) != 0 ? intense ? 255 : 200 : intense ? 84 : 0;
         int b = (i & BLUE)  != 0 ? intense ? 255 : 200 : intense ? 84 : 0;
         dosColors[i] = new Color(r, g, b);
      }
   }



   public DosConsole(Container parent, String host, int port, Constants keymap, Constants SpecialKeys, Frame conFrame)
      throws IOException
   {
      this(parent, keymap, SpecialKeys, conFrame);
      openCon(host, port);
   }

   public DosConsole(Container parent, Socket s, Constants keymap, Constants SpecialKeys, Frame conFrame)
      throws IOException
   {
      this(parent, keymap, SpecialKeys, conFrame);
      openCon(s);
   }

   public DosConsole(Container parent, Constants keymap, Constants SpecialKeys, Frame conFrame)
   {
      super(parent, null, conFrame);
      //Debug.out("emul super created");
      this.keymap = keymap;
      this.special = SpecialKeys;
      /*
      mapKeys = new Button("Key mappings");
      saveKeys = new Button("Save mappings");
      openCon = new Button("Open connection");
      closeCon = new Button("Close connection");
      reboot = new Button("Reboot RC server");
      refresh = new Button("Refresh display");
*/
//      buttons.add(mapKeys, c);
//      buttons.add(saveKeys, c);
//      buttons.add(openCon);
//      buttons.add(openCon);
//      buttons.add(reboot);
//      buttons.add(refresh);
/*
      mapKeys.addActionListener(this);
      openCon.addActionListener(this);
      closeCon.addActionListener(this);
      saveKeys.addActionListener(this);
      reboot.addActionListener(this);
      refresh.addActionListener(this);
*/
      //System.out.println("Remote Console started");
   }

   protected void clearBufs()
   {
      for (int i = 0; i < screenbytes; i++) {
         data[i] = ' ';//(char)(i % 90 + 32);//test pattern
         atrs[i] = 7;//(char)(i % 90 + 32);//test pattern7;//white on black
      }
   }

   private void moIn(int vec)
   {
      bufP += vec;
      invalRegion (bufP - vec, bufP);
      while (bufP >= screenbytes)
         bufP -= screenbytes;
//      Debug.out("MOIN, bufP = "+bufP+", vec = "+vec);
   }

   //reads length bytes into array ar. Wraps read(byte[], int, int) of class inputstream
   private void ra(byte[] ar, int length)
      throws IOException
   {
      if (Debug.dv) {
      }
      readcount += length;
      int i;

                // wait until enough data is available
                while (is.available() < length)
                        ;

      if ((i = is.read(ar, 0, length)) != length)
         throw new IOException("Closing- couldnt read all, only "+i+" of "+length+", available:"+is.available());
   }

   //commonly used function
   private int getSize(int type)
      throws IOException
   {
      if ((type & LOWERCASE) != 0) {
         readcount++;
         return is.read();
      }
      readcount += 2;
      return (is.read() << 8) + is.read();
   }
/*      byte b[] = new byte[2];
      ra(b, 2);
      int ret = (b[0] & 0xff) << 8 | b[1] & 0xff;
      if (Debug.dv) {
         Debug.out(" "+ret);
         if (ret + bufP > screenbytes)
            Debug.out("getsize OUT OF BOUND, return "+ret+", bufP:"+bufP+", read:"+readcount);
      }
      return ret;
   }
*/
   protected long readcount = 0;
//////////////////////////////////////////////////////////////
   protected void readLoop()
   {
          //System.err.println("inside readLoop!");
      InputStream is = this.is;
      byte[] rb = new byte[screenbytes * 2];
      int i, tot;
      byte temp;
      char ctemp;
                int cmd=-1;                
      try {
         while (true) {
            while (bufP >= screenbytes)
               bufP -= screenbytes;
            int k = is.read();
                                cmd = k;
            readcount++;
            //Debug.out("read byte = "+(char)k+", bufP = "+bufP+", read:"+readcount);
            switch (k | LOWERCASE) {
            case -1:
               throw new IOException("Closed");
            case T_SCROLL:
               tot = getSize(k);
                                        byte rowtemp =(byte)is.read();
               i = rowtemp * cols;
//               Debug.out("Scroll, offset, length"+i+" "+tot);
               readcount++;
                                        //DJVN : 28/06/2001 ADDED to solve "invalid line scroll" problem.
                                        //Compiles and runs fine
                                        if (rowtemp < rows)
                                        {
                 System.arraycopy(data, bufP + i, data, bufP, tot);
                 System.arraycopy(atrs, bufP + i, atrs, bufP, tot);
                                        }
               moIn(tot);
               break;
            case T_MOVE:
               bufP += getSize(k);
               break;
            case T_CARET:
               ra(rb, 2);
               setCaret(rb[0], rb[1]);
               break;
            case T_SCREENSIZE:
               ra(rb, 2);
               setSizes(rb[0], rb[1]);
               break;
            case T_SERINFO:
               tot = getSize(k);
               ra (rb, tot);
               //do something with info like displaying it, etc.
               break;
            case T_DATA:
               tot = getSize(k) * 2;
               ra (rb, tot);
               i = 0;
               while (i < tot) {
                  data[bufP] = ASCII[rb[i++] & 0xff];
                  atrs[bufP++] = (char)rb[i++];
               }
               invalRegion(bufP - tot / 2, bufP);
               break;
            case T_PATTERN:
               i = getSize(k);
               tot = i + bufP;
               ra (rb, 2);
               ctemp = ASCII[rb[0] & 0xff];
               temp = rb[1];
               while (bufP < tot) {
                  data[bufP] = ctemp;
                  atrs[bufP++] = (char)temp;
               }
               invalRegion(bufP - i, bufP);
               break;
            case T_ATRPATTERN:
               tot = getSize(k) + 1;
               ra (rb, tot);
               temp = rb[0];
               i = 1;
               while (i < tot) {
                  data[bufP] = ASCII[rb[i++] & 0xff];
                  atrs[bufP++] = (char)temp;
               }
               invalRegion(bufP - tot, bufP);
               break;
            default://must be error
               if (Debug.dv)
                  Debug.out("ERRORBYTE in readloop: "+k+", count:"+readcount);
            }

         }//end while
      }
      catch (Exception e) {
                        // System.err.println("Exception caught: CMD("+cmd+") error: "+e.getMessage());
         if (Debug.dv) {
            Debug.out(e +" readcount "+readcount+" bufP"+bufP);
            //Debug.pst();
         }
      }
      closeCon();
   }

   //overrides emulator.setColors
   //called from painting
   protected void setColors(Graphics fg, Graphics bg, char atr)
   {
      fg.setColor(dosColors[atr & 0xf]);
      bg.setColor(dosColors[(atr >>> 4) & 7]);
//      if (Debug.dv)
//         Debug.out("set bg color to (+atr + pos)"+bg.getColor()+atr+" place "+((atr >>> 4) & 8));
      if (blinkon && (atr & BLINK) != 0) {
         Color ftemp = fg.getColor();
         fg.setColor(bg.getColor());
         bg.setColor(ftemp);
      }
   }

   protected boolean isAtrBlink(char atr)
   {
      return (atr & BLINK) != 0;
   }

        //normally called with "mod,key".
        //when a modifier alone is released, it's called with "mod,0".
   protected boolean keyPress(int mod, int key, char keyChar)
   {
          //System.err.println("inside keyPress!");
      //System.err.println("keyPress(mod="+mod+", key="+key+", keyChar="+keyChar+")");
      try {
      
         int[] i;

         // see if it is a special translation         
         i = special.getIntArray(""+keyChar+"");
         if (i == null || i.length < 2)
         {
            // not a special translation -- is it in the
            // keymap?
            i = keymap.getIntArray("" + mod + "+" + key);
            if (i == null || i.length < 2) 
            {
               // not in keymap, just send the key with 
               // 0 scancode
               i = new int[2];
               i[1] = key;
            }
         }            
         //System.err.println("Sending Key scan=" +i[0]+", ascii="+i[1]);
         //send a dos scan-code and ascii value to server
         byte[] b = {KEYPRESS, (byte)i[0],(byte)i[1]};
         os.write(b);
         return true;
      }
      catch (Exception e) {
         if (Debug.dv)
            Debug.out("sending keys:"+e);
         return false;
      }
   }

   public void actionPerformed(ActionEvent e)
   {
      Object o = e.getSource();
/*      if (o == mapKeys) {
         new KeyMapDialog(conFrame, keymappings);
      }
      else if (o == saveKeys) {
         if (keymappings.save(conFrame) == 0)
            setStat("key mappings saved");
         else
            setStat("couldnt save settings");
      }
      else if (o == openCon) {
         OpenConnectionDialog d = new OpenConnectionDialog(conFrame, keymappings);
         try {
            openCon(d.getHost(), d.getPort());
         }
         catch (IOException ioe) {
            setStat("couldnt connect"+ioe);
         }
      }
      else if (o == closeCon) {
         closeCon();
      }
      else if (o == reboot) {
         try {
            os.write(REBOOT);
         }
         catch (IOException ioe) {
            setStat("couldnt connect"+ioe);
         }
      }
      else if (o == refresh) {
         try {
            os.write(REFRESH);
         }
         catch (IOException ioe) {
            setStat("couldnt connect"+ioe);
         }
      }
      canvas.requestFocus();
      */
   }
}
