//
//   $Id: Emul.java 1.4 2003/01/22 17:28:38Z johnb Exp johnb $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// HISTORY
//    1/22/2003   jmb   Added code to default to using Courier New (or 
//                      CourierNew) if not use plain courier -- search
//                      the availble fonts for Courier New
//    1/16/2003   jmb   Moved ASCII->UNICODE translation table to it's own
//                      class to allow for translation to other languagnes.
//                      Added version check to allow TAB key to function
//                      under JRE 1.4 and newer.  Added special keystroke
//                      processing for keys that generate a zero keycode.
package inet;

import java.net.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.lang.*;

public abstract class Emul
implements Runnable, ComponentListener, AdjustmentListener, KeyListener, ActionListener, MouseMotionListener

{

	//possible directions for scrolling
	public static final int SCROLL_LEFT		= 1;
	public static final int SCROLL_RIGHT	        = 2;
	public static final int SCROLL_UP		= 3;
	public static final int SCROLL_DOWN		= 4;

	protected Container parent;
	protected Scrollbar sh, sv;
	protected int cols;
	protected int rows;
	protected int screenbytes;
	protected int colwidth, rowheight, charDescent, emulWidth, emulHeight;
	protected int transx, transy;
	protected Font font, bfont, ifont, bifont;//a regular, bold, italic and bold italic
	protected int bufP;//buffer pointer, position in buffer where we're reading
	protected int row, col;//the col and row of bufP - allways updated with bufP
	protected char[] data;
	protected char[] atrs;

	private int start, end;
	protected Thread painter;
	protected Panel buttons = new Panel();
	protected Label statLine = new Label();

	protected boolean blinkon;
	protected int blinktime = 1000;
	protected Thread blinker;
	boolean blinkonly;

	protected Thread reader;

	protected Socket socket;
	protected InputStream is;
	protected OutputStream os;

	protected EmulCanvas canvas;
	protected Frame conFrame;

	protected Thread caret_blinker;
	protected int caretPos;
	protected boolean scrollWithCaret = true;
	protected int blinkperiod = 400;
	protected int caretX, caretY;
	private int carTransX = -1, carTransY = -1;
	protected Graphics carGraph;

	private Button mapKeys, saveKeys, conOpen, conClose;

	DataBase keymappings;
	int mods;
	boolean lookForRelease = false;
	byte shift_state;

   public char[] ASCII = new char[256];
   ASCIITable ASCII_TABLE = new ASCIITable();
   Double jre_ver;

	class EmulCanvas extends Canvas
	{
		public void update(Graphics g)
		{
			paint(g);
		}

		public void paint(Graphics g)
		{
			requestFocus();
			Emul.this.paint(g, 0, screenbytes, false);
		}

		public boolean isFocusTraversable()
		{
			return true;
		}
		public Dimension getPreferredSize()
		{
			return new Dimension(cols * colwidth, rows * rowheight);
		}

	}

	public Emul(Container parent, DataBase keymap, Frame conFrame)
	{
		//Debug.out("appl start, start emul create");
		this.conFrame = conFrame;
		this.parent = parent;
		keymappings = keymap;
		
      ASCII = ASCII_TABLE.ASCII;
		canvas = new EmulCanvas();

      
      jre_ver = new Double(System.getProperty("java.specification.version"));
		if (jre_ver.doubleValue() >= 1.4)
		   canvas.setFocusTraversalKeysEnabled(false);
			
		sv = new Scrollbar(sv.VERTICAL, 1,1,1,100);
		sh = new Scrollbar(sh.HORIZONTAL, 1, 1, 1, 100);
		setSizes(80, 25);
		//NB setFont(keymap.getS(keymap.F_FACE), keymap.getI(keymap.F_SIZE));

                GraphicsEnvironment ge = 
                        GraphicsEnvironment.getLocalGraphicsEnvironment();

                String allfonts[] = ge.getAvailableFontFamilyNames();
                int usefont = 0;

                for (int i=1; i < allfonts.length; i++)
                {
//                        System.err.println("allfonts["+i+"] = "+allfonts[i]);
                        if (allfonts[i].compareToIgnoreCase("Courier New") == 0 ||
                            allfonts[i].compareToIgnoreCase("CourierNew") == 0)
                        {
//                                System.err.println("^^^^ using this font!");
                                usefont = i;
                        }
                }
                
                if (usefont > 0)
                        setFont(allfonts[usefont], 14);
                else
                        setFont("courier",14);

		parent.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.weighty = 0;
		c.weightx =  1;
		//parent.add(buttons, c);
		c.gridy = 1;
		c.fill = c.BOTH;
		c.weighty = c.weightx =  1;
		parent.add(canvas, c);
		c.gridy = 3;
		c.gridwidth = 2;
		c.weighty = 0;
		parent.add(statLine, c);
		statLine.setBackground(Color.blue);
		statLine.setForeground(Color.yellow);
		parent.invalidate();
		parent.validate();
		(blinker = new Thread(this)).start();
		(painter = new Thread(this)).start();
		(caret_blinker = new Thread(this)).start();

		sh.addAdjustmentListener(this);
		sv.addAdjustmentListener(this);
		canvas.addComponentListener(this);
		canvas.addKeyListener(this);
		canvas.addMouseMotionListener(this);
		if (Debug.dv)
			Debug.out("tracing for dos-console 15");
		setCaret(0);
		/*
		if (keymap.getB(keymap.specify_buttons)) {//user specified buttons himself
		}
		else {//we put in the default buttons
		}
		*/
		//NB crash???conFrame.setIconImage(iconImage);
		//buttons.add(mapKeys = new Button("Map keys"));
		//buttons.add(saveKeys = new Button("Save keys"));
		//buttons.add(conOpen = new Button("Open connection"));
		//conClose = new Button("Close connection");

		//Debug.out("appl start, after buttons created");

		//mapKeys.addActionListener(this);
		//saveKeys.addActionListener(this);
		//conOpen.addActionListener(this);
		//conClose.addActionListener(this);

	}

/*	//parses a string coming from the database contianing info on a button: text, command, position, etc.
	protected void parseButtonString(String s, Button b)
	{
		String s = keymap.getB(
		StringTokenizer st = new StringTokenizer(s, ",;");
		try {

		}
		catch (Exception n) {//finished or error, either case we use what we have
		}
	}
*/
	protected void setSizes(int cs, int rs)
	{
		rows = rs;
		cols = cs;
		screenbytes = rs * cs;
		data = new char[screenbytes];
		atrs = new char[screenbytes];
		emulWidth = colwidth * cols;
		emulHeight = rowheight * rows;
		clearBufs();
		parent.validate();
	}

	protected void setFont(String face, int size)
	{
   	font = new Font(face, Font.PLAIN, size);

		bfont = new Font(face, Font.BOLD, size);
		ifont = new Font(face, Font.ITALIC, size);
		bifont = new Font(face, Font.ITALIC | Font.BOLD, size);
		FontMetrics fm = canvas.getFontMetrics(font);

		rowheight = fm.getHeight();
		charDescent = rowheight - fm.getDescent();
		colwidth = fm.charWidth(' ');//NBNB ONLY for fixed-pitch font

		emulWidth = colwidth * cols;
		emulHeight = rowheight * rows;
		parent.validate();
	}


	public int getBufPos()
	{
		return bufP;
	}

	public int getRow()
	{
		return row;
	}

	public int getCol()
	{
		return col;
	}
	/**
	all position setting functions call setBufP(int col, int row) which may be overriden
	*/
	public void setRow(int row)
	{
		setBufP(col, row);
	}

	public void moveRow(int vec)
	{
		setRow(row + vec);
	}

	public void setCol(int col)
	{
		setBufP(col, row);
	}

	public void moveCol(int vec)
	{
		setCol(col + vec);
	}

	protected void moveBufP(int vec)
	{
		setBufP(bufP + vec);
	}
	protected final void setBufP(int pos)
	{
		setBufP(bufP % cols, bufP / cols);
	}

	protected void setBufP(int col, int row)
	{
		if (col < 0)
			col = 0;
		else if (col >= cols)
			col = cols - 1;
		if (row < 0)
			row = 0;
		else if (row >= rows)
			row = rows - 1;
		this.col = col;
		this.row = row;
		bufP = row * cols + col;
	}

	private void closeT(Thread t)
	{
		try {
			if (t == null || !t.isAlive())
				return;
			t.stop();
		}
		catch (Throwable e) {
			if (Debug.dv)
				Debug.out("Error closing"+t+"  "+e);
		}
	}

	public void openCon (String host, int port)
		throws IOException
	{
		openCon(new Socket(host, port));
	}

	public void openCon (Socket s)
		throws IOException
	{
		//Debug.out("in opencon, socket = "+s);
		socket = s;
		is = s.getInputStream();
		os = s.getOutputStream();
		//Debug.out("in opencon, is, os = "+is+os);
		clearBufs();
		(reader = new Thread(this)).start();
		//buttons.remove(openCon);
		//buttons.add(closeCon);
		if (buttons.isAncestorOf(conOpen)) {//remove this button and add conclose
			buttons.remove(conOpen);
			buttons.add(conClose);
		}
		setStat("Connected");
	}

	public void closeCon ()
	{
		closeT(reader);
/*		try {
			if (is != null)
				is.close();
		}
		catch (Exception e) {
			if (Debug.dv)
				Debug.out("Error closing 1"+is+e);
		}
		try {
			if (os != null)
				os.close();
		}
		catch (Exception e) {
			if (Debug.dv)
				Debug.out("Error closing 2"+os+e);
		}
*/		try {
			if (socket != null) {
				socket.close();
				socket = null;
			}
		}
		catch (Exception e) {
			if (Debug.dv)
				Debug.out("Error closing 3"+socket+e);
		}
		//buttons.remove(closeCon);
		//buttons.add(openCon);
			if (buttons.isAncestorOf(conClose)) {
				buttons.add(conOpen);
				buttons.remove(conClose);
			}
		setStat("Connection Closed");
	}

	public void finalize()
	{
		close();
	}

	public void close ()
	{
		try {
			if (keymappings != null && keymappings.isModified()) {
				if (OD.s(conFrame, keymappings, keymappings.getS(DataBase.ConNotS_title),
				keymappings.getS(DataBase.ConNotS_Mes), OD.yes | OD.no, OD.yes) == OD.yes)
					keymappings.saveAs(conFrame);
			}
		}
		catch (Exception e)
		{
			Debug.out(e);
		}
		closeCon();
		closeT(caret_blinker);
		closeT(blinker);
		closeT(painter);
	}

	public void run ()
	{
		Thread t = Thread.currentThread();
		//t.setPriority(Thread.MAX_PRIORITY);
		if (t == blinker)
			blinkLoop();
		else if (t == painter)
			paintLoop();
		else if (t == reader)
			readLoop();
		else if (t == caret_blinker)
			caretLoop();
	}

	protected void clearBufs()
	{
		setData(0, screenbytes, ' ', (char)0);
	}

	/**
	Set data and attribute bytes in buffer to specific values
	*/
	protected void setData(int from, int length, char c, char atr)
	{
		while (length-- > 0) {
			data[from] = c;
			atrs[from++] = atr;
		}
	}

	/**
	Move length chars in both data and attribute arrays from 'from' to 'to'
	*/
	protected void moveData(int from, int to, int length)
	{
		System.arraycopy(data, from, data, to, length);
		System.arraycopy(atrs, from, atrs, to, length);
	}

	protected abstract void readLoop();

	protected void setColors(Graphics fg, Graphics bg, char atr)
	{
		fg.setColor(Color.white);
		bg.setColor(Color.black);
	}

	protected boolean isAtrUnderline(char atr)
	{
		return false;
	}

	protected boolean isAtrBold(char atr)
	{
		return false;
	}

	protected boolean isAtrBlink(char atr)
	{
		return false;
	}

	protected boolean isAtrItalic(char atr)
	{
		return false;
	}

	private volatile boolean invalLock = false;//used as lock in next two functions

	//invalidates a region so that it may be painted again.
	//sets the obj-vars start and end to the widest range between it and s and e.
	protected void invalRegion(int s, int e)
	{
		if (s < 0)
			s = 0;
		if (e > screenbytes)
			e = screenbytes;
		if (s > e)
			return;
		//own lock because system call too slow because but better to sync. synchronized (painter) {
		while (invalLock)
			Thread.yield();
		invalLock = true;
		if (s < start)
			start = s;
		if (e > end)
			end = e;
		invalLock = false;
		//System.err.println("INVALidate s=" +s+", e="+e+", st="+start+", en="+end);
		//if (Debug.dv && e > 1940)
		//	Debug.out("INVALidate s=" +s+", e="+e+", st="+start+", en="+end);
		//	painter.interrupt();
		//	painter.isAlive();//VERY important for bug in MS JVM (end 98)
		painter.resume();//only has effect if painter is suspended
		//}
	}

	//used for continuous painting while reading data / editing.
	protected void paintLoop()
	{
		try {
			while (true) {
				//check if there is an invalid region.
				if (end > start) {
					int s = start;
					int e = end;
					while (invalLock)
						Thread.yield();
					invalLock = true;
					start = screenbytes;
					end = 0;
					invalLock = false;
					paint(null, s, e, false);
				}
				else
					painter.suspend();
				//If there is, get the canvas' graphics and paint it
			}
		}
		catch (Exception e) {
			if (Debug.dv)
				Debug.out("paintloop:"+e);
		}
	}

	//should be private but jview (april 99) cannot handle it
	private void paint (Graphics f, int start, int end, boolean blinkonly)
	{
		char[] dat = this.data;
		char[] video = this.atrs;
		Graphics b;

		while (invalLock)	        // while invalidating a region
			Thread.yield();		//   we need to wait

		if (f == null) {
			if ((f = canvas.getGraphics()) == null) {
				invalRegion(start, end);
				return;
			}
		}
		f.translate(transx,transy);//scrolling: set caret as well
		//determines the region in our buffer that must be painted
		//get a narrowest band that needs to be painted
		Dimension d = canvas.getSize();
		f.setClip(-transx, -transy, d.width, d.height);
		//Rectangle r = f.getClipBounds();
		int regs = -transy / rowheight * cols;//region start in buffer
		int rege = regs + cols * (2 + d.height / rowheight);//region end in buffer
		if (start < regs)
			start = regs;
		if (end > rege)
			end = rege;
		b = f.create();
		//Debug.out(f.getClipBounds() + " "+transx + " "+transy+" "+d);

		int chunkStart, chunkEnd;

		while (start < end) {

			//(1) determine chunk
			chunkStart = start;						//inclusive
			char videoTemp = video[start++];
			while (start < end && video[start] == videoTemp)
				start++;
			chunkEnd = start;						//exclusive
			//(2) draw chunk
			//2.1 sort out blinking
			if (!isAtrBlink(videoTemp) && blinkonly)
				continue;//dont paint this chunk if blink not issue
			//2.2 set colors for graphics-contexts
			setColors(f, b, videoTemp);

			f.setFont(isAtrBold(videoTemp) ? isAtrItalic(videoTemp) ? bifont : bfont :
			                  isAtrItalic(videoTemp) ? ifont : font);


			//2.3 calculate parameters for chunk
			int cs = chunkStart % cols;						//cs = column start
			int xs = cs * colwidth;				//xs = x - location start
			int rs = chunkStart / cols;						//r = row
			int ys = rs * rowheight;
			while (true) {
				int sl = cols - cs;						// string length, # of bytes to draw
				int eods = chunkStart + sl;					//end Of Draw String
				if (eods > chunkEnd)
					sl = chunkEnd - chunkStart;
				//2.4 draws to device context
				b.fillRect(xs, ys, sl * colwidth, rowheight);
				f.drawChars(dat, chunkStart, sl, xs, ys + charDescent);
				if (isAtrUnderline(videoTemp))
					f.drawLine(xs, ys + rowheight - 1, xs + sl * colwidth, ys + rowheight-1);
				if (eods > chunkEnd)
					break;
				ys += rowheight;
				xs = 0;
				chunkStart += sl;
				cs = 0;
			}//end while painting chunk
		}//end while start < end
	}

	//a blinker implementation
	protected void blinkLoop()
	{
		while (true) {
			try {
				Thread.sleep(blinktime);
				blinkon = !blinkon;
				paint(null, 0, screenbytes, true);
			}
			catch (Exception e) {
				if (Debug.dv)
					Debug.out(e);
			}
		}
	}

	//scrolls the display
	public void scroll(int direction, int units)
	{
		//out("scrolling "+units);
		switch (direction) {
		case SCROLL_UP:		sv.setValue(sv.getValue() - units * rowheight);break;
		case SCROLL_DOWN:	sv.setValue(sv.getValue() + units * rowheight);break;
		case SCROLL_LEFT:	sh.setValue(sh.getValue() - units * colwidth);break;
		case SCROLL_RIGHT:	sh.setValue(sh.getValue() + units * colwidth);break;
		}
		transx = -sh.getValue();
		transy = -sv.getValue();
		canvas.repaint();
	}

	protected void caretLoop()
	{
		try {
			while(true) {
				synchronized (caret_blinker) {
					paintCaret();
				}
				Thread.sleep(blinkperiod);
			}
		}
		catch (Exception e) {
			if (Debug.dv)
				Debug.out(e);
		}
	}

	protected void showCaret(boolean show)
	{
		if (show)
			caret_blinker.resume();
		else {
			caret_blinker.suspend();
			invalRegion	(caretPos, caretPos + 1);
		}
	}

	protected void setCaret(int pos)
	{
		setCaret(pos % cols, pos / cols);
	}

	protected void setCaret(int col, int row)
	{
		synchronized (caret_blinker) {
			if (scrollWithCaret) {
				//vx vy where caret will be when set (in pixels)
				int vx = col * colwidth + transx;
				int vy = row * rowheight + transy;
				Dimension s = canvas.getSize();
				//NB remember this is integer algebra, be careful
				if (vx < 0)
					scroll(SCROLL_LEFT, (- 1 - vx) / colwidth + 1);
				else if (vx + colwidth > s.width)
					scroll(SCROLL_RIGHT, (vx + colwidth - s.width - 1) / colwidth + 1);
				if (vy < 0)
					scroll(SCROLL_UP, (-1 - vy) / rowheight + 1);
				else if (vy + rowheight > s.height)
					scroll(SCROLL_DOWN, (vy + rowheight - s.height - 1) / rowheight + 1);
			}
			setCarGraph();
			invalRegion	(caretPos, caretPos + 1);
			caretPos = row * cols + col;
			caretX = col * colwidth;
			caretY = row * rowheight;
			paintCaret();
			setStat("Row = "+row+"   Col = "+col);
		}
	}

	protected void setCarGraph()
	{
		if (carTransX == transx && carTransY == transy)
			return;
		synchronized (caret_blinker) {
			carGraph = canvas.getGraphics();
			if (carGraph == null)
				return;
			carGraph = carGraph.create();
			carGraph.translate(carTransX = transx, carTransY = transy);
			carGraph.setXORMode(Color.black);
			carGraph.setColor(Color.white);
		}
	}

	protected void paintCaret()
	{
		if (carGraph != null)
			carGraph.fillRect(caretX, caretY, colwidth, rowheight);
//		Debug.out("painted caret");
	}

	//adjustment listener
	public void adjustmentValueChanged(AdjustmentEvent e)
	{
		//if (Debug.dv) Debug.out("adjusted, transx, transy, e.getValue, width, height:>"+transx+"+"+transy+"+"+e.getValue()+"+"+emulWidth+"+"+emulHeight);
		if (e.getSource() == sv)
			transy = -e.getValue();
		else if (e.getSource() == sh)
			transx = -e.getValue();
		canvas.repaint();
		setCarGraph();
	}

	//used by next function
	private int setSB(Scrollbar s, int dsize, int size, int unit, int trans)
	{
		//size -= dsize;//NB size becomes difference
		if (size > 0) {
			if (!s.isShowing()) {
				GridBagConstraints c = new GridBagConstraints();
				c.fill = c.BOTH;
				if (s == sh) {
					c.gridy = 2;
					c.weighty = 0;
					c.weightx = 1;
				}
				else {
					c.weighty = c.gridy = c.gridx = 1;
					c.weightx = 0;
				}
				parent.add(s, c);
			}
			s.setBlockIncrement(dsize);
			s.setUnitIncrement(unit);
			if (trans < -size)
				trans = -size;
			//setStat("Setvals: dsize:"+dsize+", size:"+size);
			s.setValues(-trans, dsize, 0, size);// + dsize);//NBBNBNBNBNBN
		}
		else {
			parent.remove(s);
			trans = 0;
		}
		return trans;
	}

	//ComponentListener methods
	public void componentResized(ComponentEvent e)
	{
		Dimension d = canvas.getSize();
		//first test for rare case were scrollbars could be removed
		if (sv.isShowing() && sh.isShowing() &&	sv.getSize().width + d.width >= emulWidth
			&& sh.getSize().height + d.height >= emulHeight) {
			parent.remove(sv);
			parent.remove(sh);
			transx = transy = 0;
		}
		else {
			transx = setSB(sh, d.width, emulWidth, colwidth, transx);
			transy = setSB(sv, d.height, emulHeight, rowheight, transy);
		}
		parent.validate();
		setCarGraph();
	}

	public void componentMoved(ComponentEvent e) {componentResized(e);}
	public void componentShown(ComponentEvent e){componentResized(e);}
	public void componentHidden(ComponentEvent e){componentResized(e);}

	//workaround for bug in some jdk's
	//may later be used as feature
	static int getKeyCode(KeyEvent k)
	{
		char keyChar = k.getKeyChar();
		return ",./;=-[\\]<>?:\"{}_+".indexOf(keyChar) >= 0 ? keyChar : k.getKeyCode();
	}

	public void keyPressed(KeyEvent k)
	{
	   int special_mods = 0;
		mods = k.getModifiers();
		int keycode = getKeyCode(k);
		char keyChar = k.getKeyChar();
//      System.err.println("keyPressed(mods="+mods+", keycode="+keycode+", keyChar="+keyChar+")");
		switch(keycode) 
		{
         case KeyEvent.VK_ALT:
         case KeyEvent.VK_CONTROL:
         case KeyEvent.VK_META:
         case KeyEvent.VK_SHIFT:
			   lookForRelease = true;
			   break;
         default:
            if (keyPress(mods, keycode, keyChar))
               k.consume();
            lookForRelease = false;
         
    
      }
   }

	public void keyReleased(KeyEvent k)
	{
		if(!lookForRelease)
			return;
		switch(k.getKeyCode()) {
		case KeyEvent.VK_ALT:
		case KeyEvent.VK_CONTROL:
		case KeyEvent.VK_META:
		case KeyEvent.VK_SHIFT:
			if (keyPress(mods, 0, (char)0))
				k.consume();
			lookForRelease = false;
		}
	}

	public void keyTyped(KeyEvent k)
	{
//      System.err.println("keytyped(mods="+k.getModifiers()+
//                         ", keycode="+getKeyCode(k)+
//                         ", keyChar="+k.getKeyChar()+")");
	}

	/**
	key presses passed to subclasses. This method enables sublclasses to map many kinds of keys,...NB
	should return true if key handled, false otherwise
	*/
	protected abstract boolean keyPress(int mod, int key, char c);

	public void actionPerformed(ActionEvent e)
	{
		Object o = e.getSource();
		if (o == mapKeys) {
			new KeyMapDialog(conFrame, keymappings);
		}
		else if (o == saveKeys) {
			if (keymappings != null && keymappings.isModified()) {
				try {
					keymappings.saveAs(conFrame);
				}
				catch (Exception x) {
					Debug.out(x);
				}
			}
		}
		else if (o == conOpen) {
			while (true) {
				try {
					OpenConnectionDialog od = new OpenConnectionDialog(conFrame, keymappings);
					openCon(od.getHost(), od.getPort());
					break;
				}
				catch (Exception ex) {
					if (OD.s(conFrame, keymappings, "Couldnt open", "Errmsg"+ex+"\r\n Do you want to retry?", OD.yes | OD.no, OD.yes) != OD.yes)
						break;
				}
			}
		}
		else if (o == conClose) {
			closeCon();
		}
	}

	void setStat(String text)
	{
		statLine.setText(text);
	}

	//MouseMotionListener
    public void mouseDragged(MouseEvent e)
	{
	}

    public void mouseMoved(MouseEvent e)
	{
		int col = e.getX() / colwidth;
		int row = e.getY() / rowheight;
		int buf = row * cols + col;
		//setStat("Position row:"+row+", col:"+col+" ,bufPos:"+buf+", data:"+Integer.toHexString((int)data[buf])+", atr:"+Integer.toHexString((int)atrs[buf]));
	}


	static Image iconImage;
	static {
		int i[] = {
			0x20000000,
			0x18000030,
			0x18000030,
			0x20000030,
			0x187c3c7c,
			0x18666630,
			0x18667e30,
			0x18666030,
			0x18666030,
			0x18666636,
			0x18663c1c,
			0x20000000,
		};
		int pixels[] = new int[i.length * 32];
		int h = i.length;
		int w = 32;
		for (int y = 0; y < i.length; y++) {
			int r = y * 255 / i.length;
			for (int x = 0; x < 32; x++) {
				int b = x * 8;
				int g = 255;// - (r+b) / 2;
				if ((i[y] & 1 << (31 - x)) != 0)
					pixels[y * 32 + x] = 0xff000000 | r << 16 | g << 8 | b;
			}
		}
		iconImage = Toolkit.getDefaultToolkit().createImage(new java.awt.image.MemoryImageSource(w,h, pixels, 0, w));
	}


   
/*
	//ascii to unicode mapping
	public static final char[] ASCII =
	{
		    32, 0x263a, 0x263b, 0x2665, 0x2666, 0x2663, 0x2660, 0x25cf,   // 0
		0x25d8,      9, 0x25d9, 0x2642, 0x2640, 0x266a, 0x266b, 0x263c,  
		0x25ba, 0x25c4, 0x2195,   0x13,   0x14,   0x15, 0x25ac, 0x21a8,   // 16
		0x2191, 0x2193, 0x2192, 0x2190,   0x1c, 0x2194, 0x25b2, 0x25bc,  
		  0x20,   0x21,   0x22,   0x23,   0x24,   0x25,   0x26,   0x27,   // 32
		  0x28,   0x29,   0x2a,   0x2b,   0x2c,   0x2d,   0x2e,   0x2f,   
		  0x30,   0x31,   0x32,   0x33,   0x34,   0x35,   0x36,   0x37,   // 48
		  0x38,   0x39,   0x3a,   0x3b,   0x3c,   0x3d,   0x3e,   0x3f,
		  0x40,   0x41,   0x42,   0x43,   0x44,   0x45,   0x46,   0x47,   // 64
		  0x48,   0x49,   0x4a,   0x4b,   0x4c,   0x4d,   0x4e,   0x4f,
		  0x50,   0x51,   0x52,   0x53,   0x54,   0x55,   0x56,   0x57,   // 80
		  0x58,   0x59,   0x5a,   0x5b,   0x5c,   0x5d,   0x5e,   0x5f,
		  0x60,   0x61,   0x62,   0x63,   0x64,   0x65,   0x66,   0x67,   // 96
		  0x68,   0x69,   0x6a,   0x6b,   0x6c,   0x6d,   0x6e,   0x6f,
		  0x70,   0x71,   0x72,   0x73,   0x74,   0x75,   0x76,   0x77,   // 112
		  0x78,   0x79,   0x7a,   0x7b,   0x7c,   0x7d,   0x7e,   0x7f,
		  0xc7,   0xfc,   0xe9,   0xe2,   0xe4,   0xe0,   0xe5,   0xe7,   // 128
		  0xea,   0xeb,   0xe8,   0xef,   0xee,   0xec,   0xc4,   0xc5,
		  0xc9,   0xe6,   0xc6,   0xf4,   0xf6,   0xf2,   0xfb,   0xf9,   // 144
		  0xff,   0xd6,   0xdc,   0xf8,   0x9c,   0xd8,   0x9e,   0x9f,   
		  0xe1,   0xed,   0xf3,   0xfa,   0xf1,   0xd1,   0xa6,   0xa7,   // 160
		  0xa8, 0x2310,   0xac,   0xbd,   0xbc,   0x69,   0xab,   0xbb,
		0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,   // 176
	   0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
		0x2514, 0x2534, 0x252b, 0x251c, 0x2500, 0x253b, 0x255e, 0x255f,   // 192
		0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256b,   0xa4,
		0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x20ac, 0x2553, 0x256b,   // 208
		0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
		0x03b1,   0xe1, 0x0393, 0x03c0, 0x03a3, 0x03c3,   0xe6, 0x03c4,   // 224
		  0xe8,   0xe9, 0x03a9, 0x03b4, 0x221e, 0x03a6, 0x03b5,   0xef,
		0x2261,   0xb1, 0x2265, 0x2264, 0x2320,   0xa7,   0xf7, 0x2248,   // 240
		  0xb0, 0x2219,   0xb7, 0x221a, 0x2229,   0xb2, 0x25d8,   0x20,
	};
	
*/
	}


