//
//   $Id: FTPASS.C 1.3 2003/02/15 02:52:34Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//   Portions Copyright (c) 1993-2002 GP van Niekerk Ondernemings
//   Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
FTPSS.C
FTP Server state machine - see RFC 959
Updated @ Date 17 July 1996, 12:12 by Gerrit van Niekerk
*/

#define	LINELEN		128	/* Length of command buffer */

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <ctype.h>
#include <alloc.h>
#include <time.h>
#include <string.h>
#include <sys\stat.h>
#include <io.h>
#include <dir.h>
#include "inetdefs.h"
#include "comiler.h"
#include "api.h"
#include "rftp.h"
#include "ftps.h"
#include "pathname.h"
#include "dirutil.h"

#define UNIXBUG	0		/* simulate UNIX early 226 */

#define DEL_DAT	100

extern char *pszHostname,version[];
extern unsigned s_tcbo;

void upcall_c(void);
static void ftpcommand(struct ftp *ftp);
static int pport(struct socket *sock,char *arg);

permcheck(struct ftp *ftp,int op,char *file);
static void ftplogin(struct ftp *ftp,char *pass);

/* Command table */
static char *commands[] = {
	"user",
#define	USER_CMD	0
	"acct",
#define	ACCT_CMD	1
	"pass",
#define	PASS_CMD	2
	"type",
#define	TYPE_CMD	3
	"list",
#define	LIST_CMD	4
	"cwd",
#define	CWD_CMD		5
	"dele",
#define	DELE_CMD	6
	"noop",
#define	NOOP_CMD	7
	"quit",
#define	QUIT_CMD	8
	"retr",
#define	RETR_CMD	9
	"stor",
#define	STOR_CMD	10
	"port",
#define	PORT_CMD	11
	"nlst",
#define	NLST_CMD	12
	"pwd",
#define	PWD_CMD		13
	"xpwd",			/* For compatibility with 4.2BSD */
#define	XPWD_CMD	14
	"mkd ",
#define	MKD_CMD		15
	"xmkd",			/* For compatibility with 4.2BSD */
#define	XMKD_CMD	16
	"xrmd",			/* For compatibility with 4.2BSD */
#define	XRMD_CMD	17
	"rmd ",
#define	RMD_CMD		18
	"stru",
#define	STRU_CMD	19
	"mode",
#define	MODE_CMD	20
	"help",
#define	HELP_CMD	21
	"stat",
#define STAT_CMD	22
	"abor",
#define ABOR_CMD	23
	"appe",
#define APPE_CMD	24
	"rnfr",
#define RNFR_CMD	25
	"rnto",
#define RNTO_CMD	26
	"syst",
#define SYST_CMD	27
	"pasv",
#define PASV_CMD	28
	"size",
#define SIZE_CMD	29
	NULL
};

/* Response messages */
static char sending[] = "150 Opening data connection for %.30s %.30s\r\n";
static char okay[] = 	"200 Ok\r\n";
static char mkdok[] = 	"257 MKD ok\r\n";
static char typeok[] = 	"200 Type OK\r\n";
static char portok[] = 	"200 Port command ok\r\n";
static char statcmd[] = "211 Waiting for command, Type = %s\r\n";
static char statsnd[] = "211 Sending, Type = %s, Count = %lu\r\n";
static char statrcv[] = "211 Receiving, Type = %s, Count = %lu\r\n";
static char helpstart[]="214-The following commands are recognized:\r\n";
static char syst[]    = "215 MSDOS %s FTP version %s\r\n";
static char helpend[] = "\r\n214 End of help\r\n";
static char banner[] = 	"220 %s FTP version %s ready, %s\r\n";
static char bye[] = 	"221 FTP server logging off\r\n";
	   char rxok[] = 	"226 File received OK\r\n";
	   char txok[] = 	"226 File sent OK\r\n";
static char abortok[] = "226 Abort successful\r\n";
static char logged[] = 	"230 Logged in\r\n";
static char deleok[] = 	"250 File deleted\r\n";
static char rntook[] =  "250 File renamed\r\n";
static char cwdmsg[] = 	"250 \"%.70s\" is current directory\r\n";
static char pwdmsg[] = 	"257 \"%.70s\" is current directory\r\n";
static char givepass[] ="331 Enter PASS command\r\n";
static char givernto[] ="350 Give new name\r\n";
static char notav[] = 	"421 Data transfer in progress\r\n";
	   char noconn[] = 	"425 Data connection reset\r\n";
static char badcmd[] = 	"500 Unknown command \"%.70s\"\r\n";
static char only8[] = 	"501 Only logical bytesize 8 supported\r\n";
static char badtype[] = "501 Unknown type \"%.70s\"\r\n";
static char badport[] = "501 Bad port syntax\r\n";
static char unimp[] = 	"502 Command not yet implemented\r\n";
static char nornfr[] =  "503 Rename from not given\r\n";
static char compar[] = 	"504 Command not implemented for parameter \"%.60s\"\r\n";
static char notlog[] =	"530 Not logged in\r\n";
static char delefail[] ="550 Delete failed\r\n";
static char cantopen[] ="550 Can't find/read file \"%.70s\"\r\n";
static char noperm[] = 	"550 Permission denied\r\n";
static char notdir[] =  "550 \"%.70s\" is not a directory\r\n";
	   char txerr[] =	"552 Read error, file send aborted\r\n";
	   char rxerr[] =	"552 Write error, file receive aborted\r\n";
static char nodir[] = 	"553 Can't read directory \"%.70s\"\r\n";
static char cantmake[] ="553 Can't create \"%.70s\"\r\n";

typedef struct scqs {			// state change queue
	WORD tcb;
	char state;
} SCQ;
#define SCQSIZE 30
SCQ scq[SCQSIZE];
SCQ *psqi = scq;
SCQ *psqo = scq;
int nscq;

/* FTP Server Control channel State change upcall handler */
#pragma argsused
void _loadds far FtpControlStateChange(WORD tcbo,char old,char new)
{
	switch (new) {
	case ESTABLISHED:
	case CLOSE_WAIT:
	case CLOSED:
		if (nscq < SCQSIZE) {
			idisable();
			psqi->tcb = tcbo;
			psqi->state = new;
			if (++psqi >= scq + SCQSIZE)
				psqi = scq;
			++nscq;
			popf();
		}
	}
}

/* FTP Server Control & Data channel State change stable time handler */
void FtpControlStateChangeHandler()
{
	struct ftp *ftp;
	TCB far *tcb;
	WORD tcbo;
	char new;
	long t;
	char *cp,*cp1;

	if (nscq == 0)
		return;
	idisable();
	new = psqo->state;
	tcbo = psqo->tcb;
	if (++psqo >= scq + SCQSIZE)
		psqo = scq;
	--nscq;
	popf();
	tcb = (TCB far *)MK_FP(tcbs,(unsigned)tcbo);
	ftp = (struct ftp *)tcb->user;
	switch (new) {
	case ESTABLISHED:
		if ((ftp = ftp_create()) == NULL) {
			/* No space, kill connection */
			CloseTcp((unsigned)tcbo);
			DelTcp((unsigned)tcbo);
			return;
		}
		ftp->control = (unsigned)tcbo;		/* Downward link */
		tcb->user = (char *)ftp;	/* Upward link */

		/* Set default data port */
		ftp->port.address = tcb->conn.remote.address;
		ftp->port.port = FTPD_PORT;
		time(&t);
		cp = ctime(&t);
		if ((cp1 = strchr(cp,'\n')) != NULL)
			*cp1 = '\0';
		Tprintf(ftp,banner,pszHostname,version,cp);
			break;
	case CLOSE_WAIT:
	case CLOSED:
	case TIME_WAIT:
		ftp_delete(ftp);
		break;
	case DEL_DAT:
		ResetTcp((unsigned)tcbo);
		break;
	}
}

// FTP Server Control channel Receiver upcall handler
void FtpControlReceiveHandler(struct ftp *ftp)
{
	char c;

	// Assemble an input line in the session buffer. Return if incomplete
	while (RecvTcp(ftp->control,&c,1) == 1) {
		switch(c) {
		case '\r':	// Strip cr's
			continue;
		case '\n':	// Complete line; process it
			ftp->buf[ftp->cnt] = '\0';
			ftpcommand(ftp);
			ftp->cnt = 0;
			break;
		default:	// Assemble line
			if (ftp->cnt != LINELEN - 1)
				ftp->buf[ftp->cnt++] = c;
			break;
		}
	}
}

/* Parse and execute ftp commands */
static void ftpcommand(struct ftp *ftp)
{
	void chartimeout();
	char *cmd,*arg,*cp,**cmdp,*pcom,*filename;
	char *mode;
	TCB far *tcb;
	char *ptype;
	char **ps;
	int i,comnum;
	char append_flag = 0;
	unsigned attrib;
	struct stat statbuf;
	extern long ip_addr;

	if (ftp->cnt == 0) {
		/* Can't be a legal FTP command */
		Tprintf(ftp,badcmd,"");
		return;
	}
	cmd = ftp->buf;

	/* Translate entire buffer to lower case */
	for (cp = cmd;*cp != '\0';cp++)
		*cp = tolower(*cp);
	/* Find command in table; if not present, return syntax error */
	for (cmdp = commands;*cmdp != NULL;cmdp++)
		if (strncmp(*cmdp,cmd,strlen(*cmdp)) == 0)
			break;
	if (*cmdp == NULL) {
		cmd[70] = 0;	// don't overrun Tprintf buffer
		Tprintf(ftp,badcmd,cmd);
		return;
	}
	comnum = cmdp - commands;
	/* Allow only USER, PASS, HELP and QUIT before logging in */
	if (ftp->cd == NULL|| ftp->path == NULL) {
		switch (comnum) {
		case USER_CMD:
		case PASS_CMD:
		case QUIT_CMD:
		case HELP_CMD:
			break;
		default:
			Tprintf(ftp,notlog);
			return;
		}
	}
	/* do not allow xfer commands unless in command state */
	if (ftp->state != COMMAND_STATE) {
		switch (comnum) {
		case LIST_CMD:
		case QUIT_CMD:
		case RETR_CMD:
		case STOR_CMD:
		case APPE_CMD:
		case PORT_CMD:
		case NLST_CMD:
			Tprintf(ftp,notav);
			return;
		}
	}
	arg = cmd + strlen(*cmdp);
	while(*arg == ' ')
		arg++;

	/* Execute specific command */
	switch (cmdp-commands) {
	case USER_CMD:
		if (ftp->username)
			free(ftp->username);
		if ((ftp->username = malloc(strlen(arg) + 1)) == NULL) {
			CloseTcp(ftp->control);
			DelTcp(ftp->control);
			break;
		}
		strcpy(ftp->username,arg);
		Tprintf(ftp,givepass);
		break;
	case TYPE_CMD:
		switch (toupper(arg[0])) {
		case 'A':	/* Ascii */
			ftp->type = ASCII_TYPE;
			Tprintf(ftp,typeok);
			break;
		case 'E':	/* ebcdic */
			Tprintf(ftp,compar,arg);
			break;
		case 'L':
			while(*arg != ' ' && *arg != '\0')
				arg++;
			if (*arg == '\0' || *++arg != '8') {
				Tprintf(ftp,only8);
				break;
			}	/* Note fall-thru */
		case 'B':	/* Binary */
		case 'I':	/* Image */
			ftp->type = IMAGE_TYPE;
			Tprintf(ftp,typeok);
			break;
		default:	/* Invalid */
			Tprintf(ftp,badtype,arg);
			break;
		}
		break;
	case QUIT_CMD:
		Tprintf(ftp,bye);
		CloseTcp(ftp->control);
		break;
	case RETR_CMD:
		if (ftp->filename)
			free(ftp->filename);
		ftp->filename = pathname(ftp,arg);
/*  for MSDOS the EOL for text files is the same as that of FTP
		if (ftp->type == IMAGE_TYPE)
			mode = binmode[READ_BINARY];
		else
			mode = "r";
*/
		mode = "rb";
		if (!permcheck(ftp,RETR_CMD,ftp->filename)) {
			Tprintf(ftp,noperm);
		}
		else if ((ftp->fp = fopen(ftp->filename,mode)) == NULL) {
			Tprintf(ftp,cantopen,ftp->filename);
		}
		else {
			doretr:
			ftp->error = 0;
			pcom = "RETR";
			goto retr_com;
		}
//		free(file);
		break;
	case APPE_CMD:
		append_flag = 1;
	case STOR_CMD:
		if (ftp->filename)
			free(ftp->filename);
		ftp->filename = pathname(ftp,arg);
		mode = "wb";
		if (append_flag)
			mode[0] = 'a';
		if (!permcheck(ftp,STOR_CMD,ftp->filename)) {
			Tprintf(ftp,noperm);
			free(ftp->filename);
			ftp->filename = 0;
			break;
		}
		else if ((ftp->fp = fopen(ftp->filename,mode)) == NULL) {
			Tprintf(ftp,cantmake,ftp->filename);
		}
		else {
			dostor:
			ftp->error = 0;
			ftp->state = RECEIVING_STATE;
			Tprintf(ftp,sending,append_flag ? "APPE" : "STOR",arg);
			if (!(ftp->flags & FF_PASV)) {
				ftp->data = OpenTcpData(ftp,1);
			}
			else
				ftp->flags &= ~FF_PASV;
		}
//		free(file);
		break;
	case PORT_CMD:
		if (pport(&ftp->port,arg) == -1) {
			Tprintf(ftp,badport);
		}
		else {
			Tprintf(ftp,portok);
		}
		break;
	case LIST_CMD:
		if (*arg == 0)
			arg = "*.*";
		if (ftp->filename)
			free(ftp->filename);
		ftp->filename = pathname(ftp,arg);
		ftp->error = 0;
		if (!permcheck(ftp,RETR_CMD,ftp->filename)) {
			Tprintf(ftp,noperm);
		}
		else if ((ftp->fp = dir(ftp->filename,1)) == NULL) {
			Tprintf(ftp,nodir,ftp->filename);
		}
		else {
			dolist:
			pcom = "LIST";
			goto retr_com;
		}
//		free(file);
		break;
	case NLST_CMD:
		if (*arg == 0)
			arg = "*.*";
		if (ftp->filename)
			free(ftp->filename);
		ftp->filename = pathname(ftp,arg);
		if (!permcheck(ftp,RETR_CMD,ftp->filename)) {
			Tprintf(ftp,noperm);
		}
		else if ((ftp->fp = dir(ftp->filename,0)) == NULL) {
			Tprintf(ftp,nodir,ftp->filename);
		}
		else {
			donlst:
			pcom = "NLST";
			retr_com:
			ftp->flags &= ~FF_SNDEOF;
			ftp->state = SENDING_STATE;
			Tprintf(ftp,sending,pcom,ftp->filename);
			if (!(ftp->flags & FF_PASV)) {
				ftp->data = OpenTcpData(ftp,2);
			}
			else {
				ftp->flags &= ~FF_PASV;
				tcb = MK_FP(tcbs,ftp->data);
				if (ftp->data && tcb->state == ESTABLISHED) {
//					ftpdt(tcb,tcb->window - tcb->sndcnt);
				}
			}
		}
//		free(file);
		break;
	case CWD_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,RETR_CMD,filename)) {
			Tprintf(ftp,noperm);
			free(filename);
		}
		else if (strcmp(filename,"\\") == 0 || (filename[1] == ':' &&
				(!filename[2] || strcmp(filename + 2,"\\") == 0)) ||
						access(filename,0) == 0) {
			/* Succeeded, record in control block */
			if(_dos_getfileattr(filename, &attrib) == 0 && !(attrib & _A_SUBDIR)) {
				Tprintf(ftp, notdir, filename);
				free(filename);
				break;
			}
			free(ftp->cd);
			ftp->cd = filename;
			Tprintf(ftp,cwdmsg,filename);
		}
		else {
			/* Failed, don't change anything */
			Tprintf(ftp,nodir,filename);
			free(filename);
		}
		break;
	case XPWD_CMD:
	case PWD_CMD:
		Tprintf(ftp,pwdmsg,ftp->cd);
		break;
	case ACCT_CMD:
		Tprintf(ftp,unimp);
		break;
	case DELE_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,DELE_CMD,filename)) {
			Tprintf(ftp,noperm);
		}
		else if (unlink(filename) == 0) {
			Tprintf(ftp,deleok);
		}
		else {
			Tprintf(ftp,delefail);
		}
		free(filename);
		break;
	case PASS_CMD:
		ftplogin(ftp,arg);
		break;
	case XMKD_CMD:
	case MKD_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,MKD_CMD,filename)) {
			Tprintf(ftp,noperm);
		}
		else if (mkdir(filename) == 0) {
			Tprintf(ftp,mkdok);
		}
		else {
			Tprintf(ftp,cantmake,filename);
		}
		free(filename);
		break;
	case XRMD_CMD:
	case RMD_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,RMD_CMD,filename)) {
			Tprintf(ftp,noperm);
		}
		else if (rmdir(filename) == 0) {
			Tprintf(ftp,deleok);
		}
		else {
			Tprintf(ftp,delefail);
		}
		free(filename);
		break;
	case STRU_CMD:
		if (tolower(arg[0]) != 'f')
			Tprintf(ftp,compar,arg);
		else
			goto sendok;
		break;
	case MODE_CMD:
		if (tolower(arg[0]) != 's')
			Tprintf(ftp,compar,arg);
		else
			goto sendok;
		break;
	case NOOP_CMD:
		sendok:
		Tprintf(ftp,okay);
		break;
	case HELP_CMD:
		Tprintf(ftp,helpstart);
		for (ps = commands,i = 0;*ps;i++,ps++) {
			Tprintf(ftp,"%s%s",
					*ps,(i & 7) != 7 ? "  " : "\r\n");
		}
		Tprintf(ftp,helpend);
		break;
	case STAT_CMD:
		tcb = (TCB far *)MK_FP(tcbs,ftp->data);
		ptype = ftp->type == IMAGE_TYPE ? "Image" : "Ascii";
		switch (ftp->state) {
		case COMMAND_STATE:	/* Awaiting user command */
			Tprintf(ftp,statcmd,ptype);
			break;
		case SENDING_STATE:	/* Sending data to user */
			Tprintf(ftp,statsnd,ptype,
				tcb ? tcb->snd.una - tcb->iss : 0);
			break;
		case RECEIVING_STATE:	/* Storing data from user */
			Tprintf(ftp,statrcv,ptype,
					tcb ? tcb->rcv.nxt - tcb->irs : 0);
			break;
		}
		break;
	case ABOR_CMD:
		if (ftp->state != COMMAND_STATE)
			Tprintf(ftp,"426 Connection closed; transfer"
						" aborted\r\n");
		if (ftp->data != 0) {
			ftp->error = -1;
			CloseTcp(ftp->data);
		}
		fclose(ftp->fp);
		ftp->fp = NULL;
		ftp->state = COMMAND_STATE;
		Tprintf(ftp,abortok);
		break;
	case RNFR_CMD:
		if (ftp->oldname) {
			free(ftp->oldname);
			ftp->oldname = 0;
		}
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,RNFR_CMD,filename)) {
			Tprintf(ftp,noperm);
			free(filename);
		}
		else if (access(filename,0) == 0) {
			Tprintf(ftp,givernto);
			ftp->oldname = filename;
		}
		else {
			Tprintf(ftp,cantopen,filename);
			free(filename);
		}
		break;
	case RNTO_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,RNTO_CMD,filename)) {
			Tprintf(ftp,noperm);
		}
		else if (!ftp->oldname) {
			Tprintf(ftp,nornfr);
		}
		else if (rename(ftp->oldname,filename) == 0) {
			Tprintf(ftp,rntook);
			free(ftp->oldname);
			ftp->oldname = 0;
		}
		else
			Tprintf(ftp,cantmake,filename);
		free(filename);
		break;
	case SYST_CMD:
		Tprintf(ftp,syst,pszHostname,version);
		break;
	case PASV_CMD:
		ftp->data = OpenTcpData(ftp,3);
		break;
	case SIZE_CMD:
		filename = pathname(ftp,arg);
		if (!permcheck(ftp,RETR_CMD,filename)) {
			Tprintf(ftp,noperm);
		}
		else {
			if (stat(filename, &statbuf) == -1)
				Tprintf(ftp,"550 %s: does not exist\r\n",filename);
			else if (statbuf.st_mode & S_IFREG)
				Tprintf(ftp,"213 %lu\r\n",statbuf.st_size);
			else
				Tprintf(ftp,"550 %s: not a plain file\r\n",filename);
		}
		free(filename);
		break;
	}
}

static int pport(struct socket *sock,char *arg)
{
	long n;
	int i;

	n = 0;
	for (i = 0;i < 4;i++) {
		n = atoi(arg) + (n << 8);
		if ((arg = strchr(arg,',')) == NULL)
			return -1;
		arg++;
	}
	sock->address = n;
	i = atoi(arg);
	if ((arg = strchr(arg,',')) == NULL)
		return -1;
	arg++;
	i = atoi(arg) + (i << 8);
	sock->port = i;
	return 0;
}

/* Attempt to log in the user whose name is in ftp->username and password
 * in pass
 */
static void ftplogin(struct ftp *ftp,char *pass)
{
	char buf[80],*cp,*cp1;
	FILE *fp;
	int anony = 0;

	if ((fp = fopen(userfile,"r")) == NULL) {
		/* Userfile doesn't exist */
		Tprintf(ftp,notlog);
		return;
	}
	while (fgets(buf,sizeof(buf),fp),!feof(fp)) {
		if (buf[0] == '#')
			continue;	/* Comment */
		if ((cp = strchr(buf,' ')) == NULL)
			/* Bogus entry */
			continue;
		*cp++ = '\0';		/* Now points to password */
		if (strcmp(ftp->username,buf) == 0)
			break;		/* Found user name */
	}
	if (feof(fp)) {
		/* User name not found in file */
		fclose(fp);
		Tprintf(ftp,notlog);
		return;
	}
	fclose(fp);
	/* Look for space after password field in file */
	if ((cp1 = strchr(cp,' ')) == NULL) {
		/* Invalid file entry */
		Tprintf(ftp,notlog);
		return;
	}
	*cp1++ = '\0';	/* Now points to path field */
	if (strcmp(cp,"*") == 0)
		anony = 1;	/* User ID is password-free */
	if (!anony && strcmp(cp,pass) != 0) {
		/* Password required, but wrong one given */
		Tprintf(ftp,notlog);
		return;
	}
	if ((cp = strchr(cp1,' ')) == NULL) {
		/* Permission field missing */
		Tprintf(ftp,notlog);
		return;
	}
	*cp++ = '\0';	/* now points to permission field */

	/* Set up current directory and path prefix */
	ftp->cd = malloc((unsigned)strlen(cp1)+1);
	ftp->path = malloc((unsigned)strlen(cp1)+1);
	strcpy(ftp->cd,cp1);
	strcpy(ftp->path,cp1);

	/* And finally set the permission bits */
	ftp->perms = 0;
	while (1) {
		char ch = tolower(*cp);
		if (ch == 'r')
			ftp->perms |= FTP_READ;	 /* Read files */
		else if (ch == 'c')
			ftp->perms |= FTP_CREATE;/* Create new files */
		else if (ch == 'w')
			ftp->perms |= FTP_WRITE; /* Overwrite or delete existing files */
		else if (ch == 'd')
			ftp->perms |= FTP_DIRECTORY;/* dir/disk access */
		else if (isdigit(ch)) {
			ftp->perms = atoi(cp);
			break;
		}
		else
			break;
		cp++;
	}
	Tprintf(ftp,logged);
}

/* Illegal characters in a DOS filename */
char badchars[] = "\"[]|<>+=;,";
/* char badchars[] = "\"[]:|<>+=;,"; */

/* Return 1 if the file operation is allowed, 0 otherwise */
permcheck(struct ftp *ftp,int op,char *file)
{
	char *cp;

	if (file == NULL|| ftp->path == NULL)
		return 0;	/* Probably hasn't logged in yet */
	/* Check for characters illegal in MS-DOS file names */
	for (cp = badchars;*cp != '\0';cp++) {
		if (strchr(file,*cp) != NULL)
			return 0;
	}
	/* The target file must be under the user's allowed search path unless
		   he has unlimited directory access */
	if (!(ftp->perms & FTP_DIRECTORY) &&
		strncmp(file,ftp->path,strlen(ftp->path)) != 0)
		return 0;
	switch(op) {
	case RETR_CMD:
		/* User must have permission to read files */
		if (ftp->perms & FTP_READ)
			return 1;
		return 0;
	case DELE_CMD:
	case RMD_CMD:
		/* User must have permission to (over)write files */
		if (ftp->perms & FTP_WRITE)
			return 1;
		return 0;
	case STOR_CMD:
	case MKD_CMD:
	case RNFR_CMD:
		/* User must have permission to (over)write files, or permission
		 * to create them if the file doesn't already exist
		 */
		if (ftp->perms & FTP_WRITE)
			return 1;
		if (access(file,2) == -1 && (ftp->perms & FTP_CREATE))
			return 1;
		return 0;
	case RNTO_CMD:
		/* User must have create permission */
		if (ftp->perms & FTP_CREATE)
			return 1;
		return 0;
	}
	return 0;
}
