//
//   $Id: FTPSS.C 1.9 2003/05/19 17:40:00Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//   Portions Copyright (c) 1993-2002 GP van Niekerk Ondernemings
//   Portions Copyright (c) 1986-1989 Phil Karn, KA9Q
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// FTPSS.C
//
// FTP Server state machine - RFC 959
//
// Revision History
// 05/15/2003 gvn Fixed NLST not working properly after ABOR
// 02/14/2002 jmb IR#3931
//                added MDTM command support as supported by the immiment
//                release of a new FTP specification.
// 01/02/2002 gvn fixed 'site' command to repond correctly
// 07/27/2001 gvn replaced must_schedule variable with call to SetSchedule
//


#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <dos.h>
#include <dir.h>
#include <ctype.h>
#include <alloc.h>
#include <time.h>
#include <string.h>
#include <sys\stat.h>
#include <share.h>
#include "dostruct.h"	//  DOS structures
#include "dosdef.h"
#include "datalght.h"
#include "dldosio.h" 	//  some int 21h (DOS)
#include "dllfn.h"  		//  LFN and LFN "smart" functions
#include "dirutil.h"
#include "inetdefs.h"
#include "compiler.h"
#include "api.h"
#include <time.h>
#include "htapic.h"
#include "httftps.h"
#include "lfnopen.h"

#define UNIXBUG	0		// simulate UNIX early 226

#define DEL_DAT	100

extern char *pszHostname,version[];
extern unsigned s_tcbo;

void UpcallC(void);

PermCheck(struct ftp *ftp,int op,char *pszPathName);
static void ftplogin(struct ftp *ftp,char *pass);
char *pathname(struct ftp *ftp,char *path);
static void ftpcommand(struct ftp *ftp);
static int pport(struct isocket *sock,char *arg);

// Command table
static char *commands[] = {
	"user",
#define	USER_CMD	0
	"acct",
#define	ACCT_CMD	1
	"pass",
#define	PASS_CMD	2
	"type",
#define	TYPE_CMD	3
	"list",
#define	LIST_CMD	4
	"cwd",
#define	CWD_CMD	5
	"dele",
#define	DELE_CMD	6
	"noop",
#define	NOOP_CMD	7
	"quit",
#define	QUIT_CMD	8
	"retr",
#define	RETR_CMD	9
	"stor",
#define	STOR_CMD	10
	"port",
#define	PORT_CMD	11
	"nlst",
#define	NLST_CMD	12
	"pwd",
#define	PWD_CMD	13
	"xpwd",			// For compatibility with 4.2BSD
#define	XPWD_CMD	14
	"mkd ",
#define	MKD_CMD	15
	"xmkd",			// For compatibility with 4.2BSD
#define	XMKD_CMD	16
	"xrmd",			// For compatibility with 4.2BSD
#define	XRMD_CMD	17
	"rmd ",
#define	RMD_CMD	18
	"stru",
#define	STRU_CMD	19
	"mode",
#define	MODE_CMD	20
	"help",
#define	HELP_CMD	21
	"stat",
#define STAT_CMD	22
	"abor",
#define ABOR_CMD	23
	"site",
#define SITE_CMD	24
	"appe",
#define APPE_CMD	25
	"rnfr",
#define RNFR_CMD	26
	"rnto",
#define RNTO_CMD	27
	"syst",
#define SYST_CMD	28
	"pasv",
#define PASV_CMD	29
	"size",
#define SIZE_CMD	30
	"cdup",
#define CDUP_CMD	31
   "mdtm",
#define MDTM_CMD	32
	NULL
};

// Response messages
static char sending[] = "150 Opening data connection for %.4s %.260s\r\n";
static char okay[] = 	"200 Ok\r\n";
static char mkdok[] = 	"257 MKD ok\r\n";
static char typeok[] = 	"200 Type OK\r\n";
static char portok[] = 	"200 Port command ok\r\n";
static char statcmd[] = "211 Waiting for command, Type = %s\r\n";
static char statsnd[] = "211 Sending, Type = %s, Count = %lu\r\n";
static char statrcv[] = "211 Receiving, Type = %s, Count = %lu\r\n";
static char helpstart[]="214-The following commands are recognized:\r\n";
static char mdtmok[]  = "213 %04d%02d%02d%02d%02d%02d\r\n";
static char syst[]    = "215 %s %s FTP %s\r\n";
static char helpend[] = "\r\n214 End of help\r\n";
static char banner[] = 	"220 %s FTP %s ready, %s\r\n";
static char bye[] = 		"221 FTP server logging off\r\n";
		 char rxok[] = 	"226 File received OK\r\n";
		 char txok[] = 	"226 File sent OK\r\n";
static char abortok[] = "226 Abort successful\r\n";
static char logged[] = 	"230 Logged in\r\n";
static char deleok[] = 	"250 File deleted\r\n";
static char rntook[] =  "250 File renamed\r\n";
static char cwdmsg[] = 	"250 \"%.260s\" is current directory\r\n";
static char pwdmsg[] = 	"257 \"%.260s\" is current directory\r\n";
static char givepass[] ="331 Enter PASS command\r\n";
static char givernto[] ="350 Give new name\r\n";
static char notav[] = 	"421 Data transfer in progress\r\n";
		 char noconn[] = 	"425 Data connection reset\r\n";
static char badcmd[] = 	"500 Unknown command \"%.70s\"\r\n";
static char only8[] = 	"501 Only logical bytesize 8 supported\r\n";
static char badtype[] = "501 Unknown type \"%.70s\"\r\n";
static char badport[] = "501 Bad port syntax\r\n";
static char unimp[] = 	"502 Command not yet implemented\r\n";
static char nornfr[] =  "503 Rename from not given\r\n";
static char compar[] = 	"504 Command not implemented for parameter \"%.60s\"\r\n";
static char notlog[] =	"530 Not logged in\r\n";
static char delefail[] ="550 Delete failed\r\n";
static char cantopen[] ="550 Can't find/read file \"%.260s\"\r\n";
static char noperm[] = 	"550 Permission denied\r\n";
static char notdir[] =  "550 \"%.260s\" is not a directory\r\n";
		 char txerr[] =	"552 Read error, file send aborted\r\n";
		 char rxerr[] =	"552 Write error, file receive aborted\r\n";
static char nodir[] = 	"553 Can't read directory \"%.260s\"\r\n";
static char cantmake[] ="553 Can't create \"%.260s\"\r\n";

typedef struct scqs {			// state change queue
	WORD tcb;
	int id;
	char state;
} SCQ;
#define SCQSIZE 30
SCQ scq[SCQSIZE];
SCQ *psqi = scq;
SCQ *psqo = scq;
int nscq;

// FTP Server Control channel State change upcall handler
#pragma argsused
void _loadds far FtpControlStateChange(WORD tcbo,char old,char new)
{
	struct ftp *ftp;
	TCB far *tcb;

	switch (new) {
	case ESTABLISHED:
	case CLOSE_WAIT:
	case CLOSED:
   	tcb = MK_FP(tcbs,tcbo);
		if (nscq < SCQSIZE) {
			disable();
			psqi->tcb = tcbo;
			if ((ftp = (struct ftp *)tcb->user) == 0)
				psqi->id = 0;
			else
				psqi->id = ftp->id;
			psqi->state = new;
			if (++psqi >= scq + SCQSIZE)
				psqi = scq;
			++nscq;
			enable();
		}
	}
#ifndef API
	UpcallC();
#endif
}

// FTP Server Control channel State change stable time handler
void FtpControlStateChangeHandler()
{
	struct ftp *ftp;
	TCB far *tcb;
	WORD tcbo;
	char new;
	long t;
	char *cp,*cp1;
	int id;

	if (nscq == 0)
		return;
	disable();
	new = psqo->state;
	tcbo = psqo->tcb;
	id = psqo->id;
	if (++psqo >= scq + SCQSIZE)
		psqo = scq;
	--nscq;
	enable();
	if (!tcbo)
		return;
	tcb = (TCB far *)MK_FP(tcbs,tcbo);
	ftp = (struct ftp *)tcb->user;
	switch (new) {
	case ESTABLISHED:
		if ((iMaxNumcon && iNumcon >= iMaxNumcon) || (ftp = ftp_create()) == NULL) {
			// No space, kill connection
			CloseTcp(tcbo);
			DelTcp(tcbo);
			return;
		}
		++iNumcon;
		ftp->control = tcbo;		// Downward link
		tcb->user = (WORD)ftp;	// Upward link

		// Set default data port
		ftp->port.dwAddress = tcb->conn.sRemote.dwAddress;
		ftp->port.wPort = FTPD_PORT;
		//log(ftp,"open FTP");
		time(&t);
		cp = ctime(&t);
		if ((cp1 = strchr(cp,'\n')) != NULL)
			*cp1 = '\0';
		Tprintf(ftp->control,banner,pszHostname,version,cp);
			break;
	case CLOSE_WAIT:
	case CLOSED:
		--iNumcon;
		ftp_delete(ftp,id);
		break;
	case DEL_DAT:
		ResetTcp(tcbo);
		break;
	}
}

// FTP Server Control channel Receiver upcall handler
#ifndef API
#pragma argsused
void _loadds far FtpControlReceive(WORD tcbo,WORD cnt)
{
 	register struct ftp *ftp;
	TCB far *tcb = (TCB far *)MK_FP(tcbs,tcbo);

	if ((ftp = (struct ftp *)tcb->user) == NULL) {
		// Unknown connection, just kill it
		if (nscq < SCQSIZE) {
			disable();
			psqi->tcb = tcbo;
			psqi->state = DEL_DAT;
			if (++psqi >= scq + SCQSIZE)
				psqi = scq;
			++nscq;
			enable();
		}
		return;
	}
	ftp->flags |= FF_CON_RC;
	UpcallC();
}
#endif

void FtpControlReceiveHandler(struct ftp *ftp)
{
	char c;

	// Assemble an input line in the session buffer. Return if incomplete
	while (RecvTcp(ftp->control,&c,1,0) == 1) {
		switch(c) {
		case '\r':	// Strip cr's
			continue;
		case '\n':	// Complete line; process it
			ftp->buf[ftp->cnt] = '\0';
			ftpcommand(ftp);
			ftp->cnt = 0;
			break;
		default:	// Assemble line
			if (ftp->cnt != LNBFLEN - 1)
				ftp->buf[ftp->cnt++] = c;
			break;
		}
	}
}

#ifndef API
// xfer is done, end off
void ftp_done(struct ftp *ftp)
{
	unsigned rcvcnt;

	if (ftp->fp != NULL)
		fclose(ftp->fp);
	ftp->fp = NULL;
	if (ftp->error) {
		if (ftp->error != -1)
			Tprintf(ftp->control,ftp->state == SENDING_STATE ?
								txerr : rxerr);
		ftp->error = 0;
	}
	else
		Tprintf(ftp->control,ftp->state == SENDING_STATE ? txok : rxok);
	ftp->state = COMMAND_STATE;
	// Kick command parser if something is waiting
	if ((rcvcnt = ((TCB far *)MK_FP(tcbs,ftp->control))->rcvcnt) != 0)
		FtpControlReceive(ftp->control,rcvcnt);
}

// FTP server data channel connection state change upcall handler
void _loadds far FtpDataStateChange(WORD tcbo,char old,char new)
{
	register struct ftp *ftp;
	extern char asc_eof;
	TCB far *tcb = (TCB far *)MK_FP(tcbs,tcbo);

	if ((ftp = (struct ftp *)tcb->user) == NULL) {
		if (tcb->state == LISTEN || tcb->state == SYN_SENT ||
													tcb->state == SYN_RECEIVED)
			return;	// ignore until ESTABLISHED
		// Unknown connection. Kill it
		goto qDelTcp;
	}
	else if (ftp->state == SENDING_STATE) {
#if UNIXBUG
		if (new == FINWAIT1) {
#else
		if (old == FINWAIT1 || old == CLOSING) {
			// We've received an ack of our FIN while sending;
			// we're done

#endif
			ftp->flags |= FF_DAT_DN;	// ftp_done(ftp);
		}
		else if (new == CLOSE_WAIT) {
			ftp->flags |= FF_DAT_CL;	// CloseTcp(tcbo);
			ftp->flags |= FF_DAT_DN;	// ftp_done(ftp);
		}
		SetSchedule();
	}
	else if (ftp->state == RECEIVING_STATE && (new == CLOSE_WAIT ||
							new == FINWAIT1)) {
		// FIN received on incoming file or we closed
		if (new == CLOSE_WAIT)
			ftp->flags |= FF_DAT_CL;	// CloseTcp(tcbo);
		ftp->flags |= FF_DAT_DN;		// ftp_done(ftp);
		SetSchedule();
	}
	if (new == CLOSED) {
		if (tcb->reason != R_NORMAL && ftp->data == FP_OFF(tcb)) {
			// Data connection was reset, complain about it
			ftp->flags |= FF_DAT_RS;
   		SetSchedule();
		}
		// Clear only if another transfer hasn't already started
		if (ftp->data == FP_OFF(tcb)) {
			ftp->flags |= FF_DAT_CLSD;
   		SetSchedule();
		}
		qDelTcp:
		if (nscq < SCQSIZE) {
			disable();
			psqi->tcb = tcbo;
			psqi->state = DEL_DAT;
			if (++psqi >= scq + SCQSIZE)
				psqi = scq;
			++nscq;
			enable();
		}
	}
#ifndef API
	UpcallC();
#endif
}

void FtpDataStateChangeHandler(struct ftp *ftp)
{
	int rcvcnt;

	if (ftp->state == RECEIVING_STATE) {
		FtpDataReceiveHandler(ftp);	// look for last segments to receive
	}
	if (ftp->flags & FF_DAT_CL) {
		ftp->flags &= ~FF_DAT_CL;
		CloseTcp(ftp->data);
		DelTcp(ftp->data);
	}
	if (ftp->flags & FF_DAT_DN) {
		ftp->flags &= ~FF_DAT_DN;
		DelTcp(ftp->data);
		ftp_done(ftp);
	}
	if (ftp->flags & FF_DAT_RS) {
		ftp->flags &= ~FF_DAT_RS;
		DelTcp(ftp->data);
		if (ftp->fp != NULL) {
			Tprintf(ftp->control,noconn);
			// And clean up
			fclose(ftp->fp);
			ftp->fp = NULL;
		}
		ftp->state = COMMAND_STATE;
		// Kick command parser if something is waiting
		if ((rcvcnt = ((TCB far *)MK_FP(tcbs,ftp->control))->rcvcnt) != 0)
			FtpControlReceive(ftp->control,rcvcnt);
	}
	if (ftp->flags & FF_DAT_CLSD) {
		DelTcp(ftp->data);
		ftp->flags &= ~FF_DAT_CLSD;
		ftp->data = 0;
	}
}
#endif

static char *getreldir(struct ftp *ftp,char *pdir)	// get relative directory
{
	char *ppath,*pdinit = pdir;

	for (ppath = ftp->path;*pdir && *ppath == *pdir;++ppath,++pdir);
	if (!*pdir) {
		if (*ppath)
			return pdinit;
		return "/";
	}
	if (*pdir != '/' && pdir != ftp->cd && pdir[1] != ':')
		--pdir;
	return pdir;
}

/* function to determine whether 'name' specifies the console */
int isconsole(char *name)
{
	char *p;
	int len;

	if (name[0] == '\\')
		++name;
	if (name[1] == ':')
		name += 2;
	if ((p = strrchr(name,'\\')) != 0)
		name = p + 1;
	if ((p = strrchr(name,'/')) != 0)
		name = p + 1;
	len = strlen(name);
	return (len == 3 || (len > 3 && (name[3] == '.' || name[3] == ':'))) &&
		!strnicmp(name,"con",3);
}

// Parse and execute ftp commands
static void ftpcommand(struct ftp *ftp)
{
	void chartimeout();
	char *cmd,*arg,**cmdp,*pszPathName,*pcom;
	char *mode;
	TCB far *tcb;
	char *ptype;
	char **ps;
	int i,comnum;
	char append_flag = 0;
	unsigned uAttributes;
	struct stat statbuf;
	extern long ip_addr;
	char szShortPath[260], *pszShortPath;
   FIND fFind;

	if (ftp->cnt == 0) {
		// Can't be a legal FTP command
		Tprintf(ftp->control,badcmd,"");
		return;
	}
	cmd = ftp->buf;

	// Find command in table; if not present, return syntax error
	for (cmdp = commands;*cmdp != NULL;cmdp++)
		if (strnicmp(*cmdp,cmd,strlen(*cmdp)) == 0)
			break;
	if (*cmdp == NULL) {
		cmd[70] = 0;	// don't overrun Tprintf buffer
		Tprintf(ftp->control,badcmd,cmd);
		return;
	}
	comnum = cmdp - commands;
	// Allow only USER, PASS, HELP and QUIT before logging in
	if (ftp->cd == NULL|| ftp->path == NULL) {
		switch (comnum) {
		case USER_CMD:
		case PASS_CMD:
		case QUIT_CMD:
		case HELP_CMD:
			break;
		default:
			Tprintf(ftp->control,notlog);
			return;
		}
	}
	// do not allow xfer commands unless in command state
	if (ftp->state != COMMAND_STATE) {
		switch (comnum) {
		case LIST_CMD:
		case QUIT_CMD:
		case RETR_CMD:
		case STOR_CMD:
		case APPE_CMD:
		case PORT_CMD:
		case NLST_CMD:
			Tprintf(ftp->control,notav);
			return;
		}
	}
	arg = cmd + strlen(*cmdp);
	while(*arg == ' ')
		arg++;

	// Execute specific command
	switch (cmdp-commands) {
	case USER_CMD:
		if (ftp->username)
			free(ftp->username);
		if ((ftp->username = malloc(strlen(arg) + 1)) == NULL) {
			CloseTcp(ftp->control);
			DelTcp(ftp->control);
			break;
		}
		strcpy(ftp->username,arg);
		Tprintf(ftp->control,givepass);
		break;
	case TYPE_CMD:
		switch (toupper(arg[0])) {
		case 'A':	// Ascii
			ftp->type = ASCII_TYPE;
			Tprintf(ftp->control,typeok);
			break;
		case 'E':	// ebcdic
			Tprintf(ftp->control,compar,arg);
			break;
		case 'L':
			while(*arg != ' ' && *arg != '\0')
				arg++;
			if (*arg == '\0' || *++arg != '8') {
				Tprintf(ftp->control,only8);
				break;
			}	// Note fall-thru
		case 'B':	// Binary
		case 'I':	// Image
			ftp->type = IMAGE_TYPE;
			Tprintf(ftp->control,typeok);
			break;
		default:	// Invalid
			Tprintf(ftp->control,badtype,arg);
			break;
		}
		break;
	case QUIT_CMD:
		Tprintf(ftp->control,bye);
		ftp_delete(ftp,ftp->id);
		break;
	case RETR_CMD:
		ftp->error = 0;
		pszPathName = pathname(ftp,arg);
		//  for MSDOS the EOL for text files is the same as that of FTP
		mode = "rb";
		if (!PermCheck(ftp,RETR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if ((ftp->fp = LFNfopen(pszPathName,mode)) == NULL) {
			Tprintf(ftp->control,cantopen,pszPathName);
		}
		else {
			//log(ftp,"RETR %.40s/%.30s",ftp->cd,arg);
			pcom = "RETR";
			goto retr_com;
		}
		free(pszPathName);
		break;
	case APPE_CMD:
		append_flag = 1;
	case STOR_CMD:
		ftp->error = 0;
		if (ftp->flags & FF_NOPATH) {
			char *p = arg,*pe;

			while (isspace(*p))
				++p;
			pe = p;
			while (!(isspace(*pe)))
				++pe;
			*pe = 0;
			pszPathName = malloc(strlen(p) + 1);
			strcpy(pszPathName,p);
		}
		else
			pszPathName = pathname(ftp,arg);
#ifdef API
		strncpy(ftp->status_line,pszPathName,99);
#endif
		//  for MSDOS the EOL for text files is the same as that of FTP
		mode = "wb";
		if (append_flag)
			mode[0] = 'a';
		if (!PermCheck(ftp,STOR_CMD,pszPathName) || isconsole(pszPathName) != 0) {
			Tprintf(ftp->control,noperm);
			free(pszPathName);
			break;
		}
		if ((ftp->fp = LFNfopen(pszPathName,mode)) == NULL) {
			Tprintf(ftp->control,cantmake,pszPathName);
		}
		else {
			//log(ftp,"%s %.40s/%.30s",append_flag ? "APPE" : "STOR",ftp->cd,arg);
			ftp->state = RECEIVING_STATE;
			Tprintf(ftp->control,sending,append_flag ? "APPE" : "STOR",arg);
			if (!(ftp->flags & FF_PASV)) {
				ftp->data = OpenTcp(ftp,1);
				ftp->flags &= ~FF_DAT_CLSD;
			}
			else
				ftp->flags &= ~FF_PASV;
		}
		free(pszPathName);
		break;
	case PORT_CMD:
		if (pport(&ftp->port,arg) == -1) {
			Tprintf(ftp->control,badport);
		}
		else {
			Tprintf(ftp->control,portok);
		}
		break;
	case LIST_CMD:
		ftp->error = 0;
		if (*arg == 0)
			arg = "*.*";
		pszPathName = pathname(ftp,arg);

		if (!PermCheck(ftp,RETR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if ((ftp->fp = dir(pszPathName,1)) == NULL) {
			Tprintf(ftp->control,nodir,pszPathName);
		}
		else {
			pcom = "LIST";
			goto retr_com;
		}
		free(pszPathName);
		break;
	case NLST_CMD:
		ftp->error = 0;
		if (*arg == 0)
			arg = "*.*";
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RETR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if ((ftp->fp = dir(pszPathName,0)) == NULL) {
			Tprintf(ftp->control,nodir,pszPathName);
		}
		else {
			pcom = "NLST";
			retr_com:
#ifdef API
			ftp->flags &= ~FF_SNDEOF;
			strncpy(ftp->status_line,pszPathName,99);
#endif
			ftp->state = SENDING_STATE;
			Tprintf(ftp->control,sending,pcom,pszPathName);
			if (!(ftp->flags & FF_PASV)) {
				ftp->data = OpenTcp(ftp,2);
				ftp->flags &= ~FF_DAT_CLSD;
			}
			else {
				ftp->flags &= ~FF_PASV;
				tcb = MK_FP(tcbs,ftp->data);
				if (ftp->data && tcb->state == ESTABLISHED) {
					ftp->flags |= FF_DAT_SN;
					SetSchedule();
				}
			}
		}
		free(pszPathName);
		break;
	case CDUP_CMD:
		arg = "..";
		// fall thru
	case CWD_CMD:
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RETR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
			free(pszPathName);
		}
		else if (strcmp(pszPathName,"/") == 0 || (pszPathName[1] == ':' &&
				(!pszPathName[2] || strcmp(pszPathName + 2,"/") == 0)) ||
						SmartGetFileAttributes(pszPathName, &uAttributes)) {
			// Succeeded, record in control block
			if (SmartGetFileAttributes(pszPathName, &uAttributes) &&
													!(uAttributes & _A_SUBDIR)) {
				Tprintf(ftp->control, notdir,getreldir(ftp,pszPathName));
				free(pszPathName);
				break;
			}
			free(ftp->cd);
			ftp->cd = pszPathName;
			Tprintf(ftp->control,cwdmsg,getreldir(ftp,pszPathName));
		}
		else {
			// Failed, don't change anything
			Tprintf(ftp->control,nodir,getreldir(ftp,pszPathName));
			free(pszPathName);
		}
		break;
	case XPWD_CMD:
	case PWD_CMD:
		Tprintf(ftp->control,pwdmsg,getreldir(ftp,ftp->cd));
		break;
	case ACCT_CMD:
		Tprintf(ftp->control,unimp);
		break;
	case DELE_CMD:
		pszPathName = pathname(ftp,arg);
		if (SmartExpandPath(FULLPATH_SHORTNAME, pszPathName, szShortPath))
			pszShortPath = szShortPath;
		else
			pszShortPath = pszPathName;
		if (!PermCheck(ftp,DELE_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if (unlink(pszShortPath) == 0) {
			Tprintf(ftp->control,deleok);
		}
		else {
			Tprintf(ftp->control,delefail);
		}
		free(pszPathName);
		break;
	case PASS_CMD:
		ftplogin(ftp,arg);
		break;
	case XMKD_CMD:
	case MKD_CMD:
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,MKD_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if (SmartMakeDirectory(pszPathName)) {
			Tprintf(ftp->control,mkdok);
		}
		else {
			Tprintf(ftp->control,cantmake,getreldir(ftp,pszPathName));
		}
		free(pszPathName);
		break;
	case XRMD_CMD:
	case RMD_CMD:
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RMD_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if (SmartRemoveDirectory(pszPathName) != 0) {
			Tprintf(ftp->control,deleok);
		}
		else {
			Tprintf(ftp->control,delefail);
		}
		free(pszPathName);
		break;
	case STRU_CMD:
		if (tolower(arg[0]) != 'f')
			Tprintf(ftp->control,compar,arg);
		else
			goto sendok;
		break;
	case MODE_CMD:
		if (tolower(arg[0]) != 's')
			Tprintf(ftp->control,compar,arg);
		else
			goto sendok;
		break;
	case NOOP_CMD:
		sendok:
		Tprintf(ftp->control,okay);
		break;
	case HELP_CMD:
		Tprintf(ftp->control,helpstart);
		for (ps = commands,i = 0;*ps;i++,ps++) {
			Tprintf(ftp->control,"%s%s",
					*ps,(i & 7) != 7 ? "  " : "\r\n");
		}
		Tprintf(ftp->control,helpend);
		break;
	case STAT_CMD:
		tcb = (TCB far *)MK_FP(tcbs,ftp->data);
		ptype = ftp->type == IMAGE_TYPE ? "Image" : "Ascii";
		switch (ftp->state) {
		case COMMAND_STATE:	// Awaiting user command
			Tprintf(ftp->control,statcmd,ptype);
			break;
		case SENDING_STATE:	// Sending data to user
			Tprintf(ftp->control,statsnd,ptype,
				tcb ? tcb->snd.una - tcb->iss : 0);
			break;
		case RECEIVING_STATE:	// Storing data from user
			Tprintf(ftp->control,statrcv,ptype,
					tcb ? tcb->rcv.nxt - tcb->irs : 0);
			break;
		}
		break;
	case ABOR_CMD:
		if (ftp->state != COMMAND_STATE)
			Tprintf(ftp->control,"426 Connection closed; transfer"
						" aborted\r\n");
		if (ftp->data != 0) {
			ftp->error = -1;
			CloseTcp(ftp->data);
		}
		fclose(ftp->fp);
		ftp->fp = NULL;
		ftp->state = COMMAND_STATE;
		Tprintf(ftp->control,abortok);
		break;
	case SITE_CMD:
		if (strnicmp(arg,"path",4) == 0) {
			ftp->flags &= ~FF_NOPATH;
			goto sendok;
		}

		if (strnicmp(arg,"nopath",6) == 0) {
			ftp->flags |= FF_NOPATH;
			goto sendok;
		}
		Tprintf(ftp->control,compar,arg);
		break;
	case RNFR_CMD:
		if (ftp->oldname) {
			free(ftp->oldname);
			ftp->oldname = 0;
		}
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RNFR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
			free(pszPathName);
		}
		if (SmartGetFileAttributes(pszPathName, &uAttributes) != 0) {
			Tprintf(ftp->control,givernto);
			ftp->oldname = pszPathName;
		}
		else {
			Tprintf(ftp->control,cantopen,pszPathName);
			free(pszPathName);
		}
		break;
	case RNTO_CMD:
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RNTO_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else if (!ftp->oldname) {
			Tprintf(ftp->control,nornfr);
		}
		else if (SmartRenameFileOrDirectory(ftp->oldname,pszPathName) != 0) {
			Tprintf(ftp->control,rntook);
			free(ftp->oldname);
			ftp->oldname = 0;
		}
		else
			Tprintf(ftp->control,cantmake,pszPathName);
		free(pszPathName);
		break;
	case SYST_CMD:
		Tprintf(ftp->control,syst,"UNIX",pszHostname,version);
		break;
	case PASV_CMD:
		ftp->data = OpenTcp(ftp,3);
		break;
	case SIZE_CMD:
		pszPathName = pathname(ftp,arg);
		if (!PermCheck(ftp,RETR_CMD,pszPathName)) {
			Tprintf(ftp->control,noperm);
		}
		else {
			if (SmartExpandPath(FULLPATH_SHORTNAME, pszPathName, szShortPath))
				pszShortPath = szShortPath;
			else
				pszShortPath = pszPathName;
			if (stat(pszShortPath, &statbuf) == -1)
				Tprintf(ftp->control,"550 %.260s: does not exist\r\n",pszPathName);
			else if (statbuf.st_mode & S_IFREG)
				Tprintf(ftp->control,"213 %lu\r\n",statbuf.st_size);
			else
				Tprintf(ftp->control,"550 %.260s: not a plain file\r\n",pszPathName);
		}
		free(pszPathName);
		break;
   case MDTM_CMD:
		pszPathName = pathname(ftp,arg);
		if (SmartExpandPath(FULLPATH_SHORTNAME, pszPathName, szShortPath))
			pszShortPath = szShortPath;
		else
			pszShortPath = pszPathName;

		if (!PermCheck(ftp,MDTM_CMD,pszPathName)) 
		{
			Tprintf(ftp->control,noperm);
		}

      else if (SmartFindFirst(A_ALLDIR, pszPathName, &fFind) != 0)
      {
         unsigned uDate = fFind.date;
         unsigned uTime = fFind.time;

         Tprintf(ftp->control, mdtmok, (uDate >> 9) + 1980, 
            (uDate >> 5) & 0x000f, uDate & 0x001f, (uTime >> 11),
            (uTime >> 5) & 0x003f, uTime & 0x001f);
         
         SmartFindClose(&fFind);
		}
		else {
			Tprintf(ftp->control,cantopen,pszPathName);
		}
		free(pszPathName);
		break;
	}
}

static int pport(struct isocket *sock,char *arg)
{
	long n;
	int i;

	n = 0;
	for (i=0;i<4;i++) {
		n = atoi(arg) + (n << 8);
		if ((arg = strchr(arg,',')) == NULL)
			return -1;
		arg++;
	}
	sock->dwAddress = n;
	n = atoi(arg);
	if ((arg = strchr(arg,',')) == NULL)
		return -1;
	arg++;
	n = atoi(arg) + (n << 8);
	sock->wPort = (unsigned)n;
	return 0;
}

// Attempt to log in the user whose name is in ftp->username and password
// in pass

static void ftplogin(struct ftp *ftp,char *pass)
{
	char buf[80],*cp,*cp1,*p;
	FILE *fp;
	int anony = 0;

	if ((fp = LFNfopen(pszUserfile,"r")) == NULL) {
		// Userfile doesn't exist
		Tprintf(ftp->control,notlog);
		return;
	}
	while (fgets(buf,sizeof(buf),fp),!feof(fp)) {
		if (buf[0] == '#')
			continue;	// Comment
		if ((cp = strchr(buf,' ')) == NULL)
			// Bogus entry
			continue;
		*cp++ = '\0';		// Now points to password
		if (strcmp(ftp->username,buf) == 0)
			break;		// Found user name
	}
	if (feof(fp)) {
		// User name not found in file
		fclose(fp);
		Tprintf(ftp->control,notlog);
		return;
	}
	fclose(fp);
	// Look for space after password field in file
	if ((cp1 = strchr(cp,' ')) == NULL) {
		// Invalid file entry
		Tprintf(ftp->control,notlog);
		return;
	}
	*cp1++ = '\0';	// Now points to path field
	if (strcmp(cp,"*") == 0)
		anony = 1;	// User ID is password-free
	if (!anony && strcmp(cp,pass) != 0) {
		// Password required, but wrong one given
		Tprintf(ftp->control,notlog);
		return;
	}
	if ((cp = strchr(cp1,' ')) == NULL) {
		// Permission field missing
		Tprintf(ftp->control,notlog);
		return;
	}
	*cp++ = '\0';	// now points to permission field

	// Set up current directory and path prefix
	ftp->cd = malloc((unsigned)strlen(cp1)+1);
	ftp->path = malloc((unsigned)strlen(cp1)+1);
	while ((p = strchr(cp1,'\\')) != 0)
		*p = '/';
	strcpy(ftp->cd,cp1);
	strcpy(ftp->path,cp1);

	// And finally set the permission bits
	ftp->perms = 0;
	while (1) {
		char ch = tolower(*cp);
		if (ch == 'r')
			ftp->perms |= FTP_READ;	 // Read files
		else if (ch == 'c')
			ftp->perms |= FTP_CREATE;// Create new files
		else if (ch == 'w')
			ftp->perms |= FTP_WRITE; // Overwrite or delete existing files
		else if (ch == 'd')
			ftp->perms |= FTP_DIRECTORY;// dir/disk access
		else if (isdigit(ch)) {
			ftp->perms = atoi(cp);
			break;
		}
		else
			break;
		cp++;
	}
	Tprintf(ftp->control,logged);
	//log(ftp,"%s logged in",anony ? pass : ftp->username);
}

// Illegal characters in a DOS filename
char badchars[] = "\"[]|<>+=;,";
// char badchars[] = "\"[]:|<>+=;,";

// Return 1 if the file operation is allowed, 0 otherwise
PermCheck(struct ftp *ftp,int op,char *pszPathName)
{
	char *cp;
	unsigned uAttributes;

	if (pszPathName == NULL|| ftp->path == NULL)
		return 0;	// Probably hasn't logged in yet
	// Check for characters illegal in MS-DOS file names
	for (cp = badchars;*cp != '\0';cp++) {
		if (strchr(pszPathName,*cp) != NULL)
			return 0;
	}
	// The target file must be under the user's allowed search path unless
	//	he has unlimited directory access
	if (!(ftp->perms & FTP_DIRECTORY) &&
		strnicmp(pszPathName,ftp->path,strlen(ftp->path)) != 0)
		return 0;
	switch(op) {
	case RETR_CMD:
   case MDTM_CMD:
		// User must have permission to read files
		if (ftp->perms & FTP_READ)
			return 1;
		return 0;
	case DELE_CMD:
	case RMD_CMD:
		// User must have permission to (over)write files
		if (ftp->perms & FTP_WRITE)
			return 1;
		return 0;
	case STOR_CMD:
	case MKD_CMD:
	case RNFR_CMD:
		// User must have permission to (over)write files, or permission
		// to create them if the file doesn't already exist

		if (ftp->perms & FTP_WRITE)
			return 1;
		if (!SmartGetFileAttributes(pszPathName, &uAttributes) &&
							(ftp->perms & FTP_CREATE))
			return 1;
		return 0;
	case RNTO_CMD:
		// User must have create permission
		if (ftp->perms & FTP_CREATE)
			return 1;
		return 0;
	}
	return 0;
}
