//
//   $Id: LOGIN.C 1.4 2003/05/19 17:38:08Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 * LOGIN.C
 * Copyright (c) GP van Niekerk Ondernemings 1990-2003
 *
 * Module containing functions for handling access to http server
 *
 * Should be used as follows:
 * At startup InitLogin(filaname, htaccess) should be called to load default perms
 * For evert action verify should be called
 * At the end FreeLoginInf should be called
 *
 * All functions work with only fully qualified pathnames

 2003-01-27 gvn Always authenticate if authentication is not available

 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <ctype.h>
#include <alloc.h>
#include "lfnopen.h"
#include "compiler.h"
#include "api.h"
#include <time.h>
#include "htapic.h"
#include "httftps.h"
#include "login.h"

typedef struct loginInf{
	//char *pcBuf;			//buffer containing al data, next 3 point to it (to reduce alloc overhead)

	//Memory for next 3 variables only allocated at szName. Other follow directly
	char *szName;		//username
	char *szPassword;	//if 0 anonymous access
	char *szDir;		//if 0 all dirs allowed

	int iDirLen;

	struct loginInf *next;	//for linked list

	int fRemcon:1;	//may use remote console
	int fCGI:1;		//may use CGI
	int fPost:1;		//may post files
	int fGet:1;		//may get files
} LOGININF;


static int ParsePasswordFile(char *szFilename, char htaccess);
static LOGININF* psUserList;	//start of list
static char fUseHTAccess;     //do we use htaccess?
static char fInitCalled;		//was init called?


/* Initialize login and verification service
 *
 * szFilename   default permissions file. If 0 given, no defaults will be loaded
 * cUseHTAccess   flag indicating usage of HTAccess
 *
 * Return L_OK if initialize worked
 * LERR_MEM or LERR_OPEN if error occured
 */
int InitLogin(char *szFilename, char cUseHTAccess)
{
	int iRet = L_OK;

	if (szFilename != 0)
		iRet = ParsePasswordFile(szFilename, 0);
	if (cUseHTAccess) {
		fUseHTAccess = cUseHTAccess;
		iRet = L_OK;
	}
	if (iRet == L_OK)
		fInitCalled = 1;
	return iRet;
}

/*
 * Free the user list up to but excluding psDontfree
 * If you want to free everything pass NULL as argument
 */
static void FreeListUntil(LOGININF *psDontfree)
{
	while (psUserList != psDontfree) {
		LOGININF *psTemp = psUserList;
		psUserList = psUserList->next;
		if (psTemp->szName)
			free(psTemp->szName);
		free(psTemp);
	}
}


/*
 * Free all login information
 */
void FreeLoginInf(void)
{
	FreeListUntil(0);
	fInitCalled = 0;
}

/*
 * Look for file HTACCESS in current directory. If not found propagate up
 * to root (or until found). If found load into linked list
 */
static void LoadOverridePerms(char *szFilename)
{
	char szNewname[260];
	int iLen;
	register char *cp;

	iLen = strlen(szFilename);
	if (iLen > 260)
		return;
	sprintf(szNewname, szFilename);
	cp = strchr(szNewname, 0);//find end of string

	if (!cp)//string not terminated!!!
		return;
	while (1) {
		while (*cp != '/' && *cp != '\\' && cp != szNewname)
			cp--;
		if (cp == szNewname)	//past root directory, didn't find file, break
			break;
		sprintf(cp + 1, "HTACCESS");
		if (ParsePasswordFile(szNewname, 1) != LERR_FOPEN) {
			//we could open file worked
			break;
		}
		*cp = 0;//go to parent directory
	}
}

/*
 * Read the password file and build linked list, starting at psUserList
 * If htaccess is turned on, no filename will be inserted
 *
 * Return L_OK if initialize worked
 * LERR_MEM or LERR_OPEN if error occured
 */
static int ParsePasswordFile(char *szFilename, char htaccess)
{
	/*
	multiple lines
	Format of line:username<space>password<space>workdir<space>permissions
	*/
	FILE *fpFile;
	char *cPW;        //points to password field in read buffer
	char *cDir;		   //points to directory field in read buffer
	char *cPerm;		//points to permissions field in read buffer
#define BUFLEN	300
	static char carBuf[BUFLEN];//in case long file names problem
	static char szFullPath[BUFLEN];//in case long file names problem
	int iLen;		//buffer length
	LOGININF *psUser;

	psUser = (LOGININF *)1;//only marks pointer for non alloc failure

	if ((fpFile = LFNfopen(szFilename,"rb")) == NULL) {
		return LERR_FOPEN;
	}

	while (fgets(carBuf, BUFLEN, fpFile)) {//???, !feof(fpFile)) {
		if (carBuf[0] == '#' || carBuf[0] == 0)
			continue;	// Comment or empty line
		//get pointers to username, pw, working directory and permissions

		if ((cPW = strchr(carBuf, ' ')) == 0 || cPW == carBuf)//username invalid
			continue;
		*cPW++ = 0;
		if ((cPerm = strchr(cPW, ' ')) == 0 || cPerm == cPW) //passw invalid
			continue;
		*cPerm++ = 0;//terminate
		if (!htaccess) {
			cDir = cPerm;
			if ((cPerm = strchr(cDir, ' ')) == 0 || cDir == cPerm) //path invalid
				continue;
			*cPerm++ = 0;//terminate
			if (strcmp(cDir, "/") == 0)	//access to whole system
				cDir = 0;
		}
		else
			cDir = 0;

		//create new LOGININF
		if ((psUser = calloc(1, sizeof(LOGININF))) == 0)
			break;

		if (cDir) {
			psUser->iDirLen = CalcFullPath(szFullPath, cDir, BUFLEN);
			iLen = psUser->iDirLen + cDir - carBuf + 1;
		}
		else
			iLen = cPerm - carBuf;

		//allocate buffer
		if ((psUser->szName = malloc(iLen)) == 0)
			break;

		strcpy(psUser->szName, carBuf);
		psUser->szPassword = psUser->szName - carBuf + cPW;//set pointer
		strcpy(psUser->szPassword, cPW);
		if (cDir) {
			psUser->szDir = psUser->szName - carBuf + cDir;//set pointer
			strcpy(psUser->szDir, szFullPath);
		}

		while (1) {
			switch(tolower(*cPerm++)) {
			case 'm':
				psUser->fRemcon = 1;
				continue;
			case 'g':
				psUser->fCGI = 1;
				continue;
			case 'p':
				psUser->fPost = 1;
				continue;
			case 'e':
				psUser->fGet = 1;
				continue;
			case '\r':
			case '\n':
			case ' ':
			case '\t':
			case 0:
			case '#':
			case ';':
				break;
			default:
				continue;//just ignore, probably ftp perm
			}
			break;
		}
		//insert at beginning of LL
		psUser->next = psUserList;
		psUserList = psUser;

	}//end while fileread

	fclose(fpFile);
	if (!psUser || !psUser->szName) {//couldnt allocate memory
		FreeLoginInf();
		return LERR_MEM;
	}
	return L_OK;
}

/*
 * Verify an operation (action)
 *
 * Arguments:
 * cType - Type of operation, either RQ_CGI, RQ_GET, RQ_POST, RQ_REMCON
 * szPathname - Full path for request
 * szUser - Username
 * szPass - Password supplied by user
 *
 * Return:
 * L_ALLOWED if and only if allowed
 * LERR_PASS if password incorrect
 * LERR_USER if user not found
 * LERR_PERM if user and password exists but no permission
 * LERR_ACC  if no access to path
 * LERR_TYPE if invalid cType
 */
int Verify(char cType, char *szPathname, char *szUser, char *szPass)
{
	LOGININF *psDefFirst = 0;
	LOGININF *psUserInfo;
	int iRet = 0;
	char fOverrides = 0;

	//if (cType)
	//	return -1;
	if (szUser == 0 || *szUser == 0)
		szUser = "*";	  //wildcard for default (undifined) user
	if (szPass == 0 || *szPass == 0)
		szPass = "*";

	if (fUseHTAccess && cType != RQ_REMCON) {//check for override file
		psDefFirst = psUserList;		  //save first default setting
		LoadOverridePerms(szPathname);
		fOverrides = 1;
	}

	//find user info
	if ((psUserInfo = psUserList) == 0) { // no authentication - accept anything
		iRet = L_ALLOWED;
		goto exit;
	}
	while (psUserInfo) {
		if (strcmp(szUser, psUserInfo->szName) == 0)
			break;
		psUserInfo = psUserInfo->next;
	}

	if (psUserInfo == 0)//user doesnt exist
		iRet = LERR_USER;
	else if (psUserInfo->szPassword != 0 &&	//password required
	 strcmp(psUserInfo->szPassword, szPass) != 0)//password doesnt match
		iRet = LERR_PASS;
	else {//check permission
		switch(cType) {
		case RQ_CGI:
			if (!psUserInfo->fCGI)
				iRet = LERR_PERM;
			break;
		case RQ_GET:
			if (!psUserInfo->fGet)
				iRet = LERR_PERM;
			break;
		case RQ_POST:
			if (!psUserInfo->fPost)
				iRet = LERR_PERM;
			break;
		case RQ_REMCON:
			if (psUserInfo->fRemcon)
				iRet = L_ALLOWED;//All OK
			else
				iRet = LERR_PERM;
			break;
		default://unknown type, throw out
			iRet = LERR_TYPE;
		}
	}
	if (iRet == 0) {//check directory
		//full szPathname should start with full user->szDir
		if(psUserInfo->szDir &&	//not HT-access
		 memicmp(szPathname, psUserInfo->szDir, psUserInfo->iDirLen) != 0)//not
			iRet = LERR_ACC;
		else
			iRet = L_ALLOWED;
	}
	exit:
	if (fOverrides)
		FreeListUntil(psDefFirst);
	return iRet;
}

/*
 * Process a path name string szPath and put the output in szNewPath
 * If it is not an absolute directory httpdir is prepended
 * At most iMaxlen characters will be put in szNewPath
 * All ../ and ./ references are removed
 *
 *	Return the length of the new path, excluidng terminating 0
 */
int CalcFullPath(char *szNewPath, char *szPath, int iMaxLen)
{
	register char *cp;
	int iLen;

	//replace all backslashes with forward slashes
	cp = szPath;
	while ((cp = strchr(cp, '\\')) != 0)
		*cp = '/';

	cp = szNewPath;

	iMaxLen -= 3;//2 for 'C:' 1 for terminating 0
	if (iMaxLen < 0)
		return 0;
	while (*szPath == '/')
		szPath++;
	/* prepend httpdir if needed */
	if (isalpha(szPath[0]) && szPath[1] == ':') {//full path
		*cp++ = *szPath++;
		*cp++ = *szPath++;
	}
	else {
		iLen = strlen(httpdir);
		iMaxLen = iMaxLen - iLen + 2;//2 for C:
		if (iMaxLen < 0)
			return 0;
		strcpy(cp, httpdir);	//prepend http dir
		cp += iLen;
	}


	/* Now start crunching the pathname argument */
	while (1) {
		/* check boundry */
		if (iMaxLen <= 0)
			break;
		/* Strip leading /'s; one will be written later */
		while (*szPath == '/')
			szPath++;
		if (*szPath == 0)
			break;		//done
		/* Look for parent directory references, either at the end
		 * of the path or imbedded in it
		 */
		if (strcmp(szPath, "..") == 0 || strncmp(szPath, "../", 3) == 0) {
			/* Hop up a level */
			while (*cp != '/' && *(cp - 1) != ':') {/* Don't back up beyond root */
				cp--;
				iMaxLen++;
			}
			*cp = 0;			/* In case there's another .. */
			szPath += 2;		/* Skip ".." */
		}

		/* Look for current directory references, either at the end
		 * of the path or imbedded in it
		 */
		else if (strcmp(szPath, ".") == 0 || strncmp(szPath, "./", 2) == 0) {
			/* "no op" */
			szPath++;			/* Skip "." */
		}

		else {
			/* Ordinary name, copy up to next '/' or end of path */
			*cp++ = '/';
			iMaxLen--;
			while (*szPath != '/' && *szPath != '\0' && iMaxLen-- > 0)
				*cp++ = *szPath++;
			*cp = 0;
		}
	}
	*cp = 0;		/* terminate */
	return cp - szNewPath;
}
