//
//   $Id: PATHNAME.C 1.4 2003/05/19 17:41:26Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
// PATHNAME.C
// Portions Copyright (c) GP van Niekerk Ondernemings 1990-2003
// Portions Copyright (c) 1986-1989 Phil Karn, KA9Q

#include <stdlib.h>
#include <stdio.h>
#include <alloc.h>
#include <string.h>
#include "compiler.h"
#include "api.h"
#include <time.h>
#include "htapic.h"
#include "httftps.h"
#include "pathname.h"

void static crunch(char *buf,char *path);

/* Given a working directory and an arbitrary pathname, resolve them into
 * an absolute pathname. Memory is allocated for the result, which
 * the caller must free
 */

char * pathname(struct ftp *ftp,char *path)
//char *cd;	/* Current working directory */
//char *path;	/* Pathname argument */
{
	register char *buf,*cp,*cd = ftp->cd;
	int len;

	if(cd == 0 || path == 0)
		return 0;
	while ((cp = strchr(path,'\\')) != 0)
		*cp = '/';
	len = strlen(cd) + strlen(path) + 10;
	if((buf = (char*)malloc(len)) == NULL ) {
		return 0;
	}
	/* Strip any leading white space on args */
	while(*cd == ' ' || *cd == '\t')
		cd++;
	buf[0] = '\0';

	/* Interpret path relative to ftp->path */
	if (path[1] == ':') {
		buf[0] = path[0];
		buf[1] = path[1];
		buf[2] = 0;
		crunch(buf + 2,path + 2);
		/* Special case: null final path means the root directory */
		if(buf[2] == '\0') {
			buf[2] = '/';
			buf[3] = '\0';
		}
	}
	else {
		if(path[0] != '/')
			crunch(buf,cd);
		else
			crunch(buf,ftp->path);
		crunch(buf,path);
		/* Special case: null final path means the root directory */
		if(buf[0] == '\0') {
			buf[0] = '/';
			buf[1] = '\0';
		}
		else if (buf[1] == ':' && buf[2] == '\0') {
			buf[2] = '/';
			buf[3] = '\0';
		}
	}
	return buf;
}

/* Process a path name string, starting with and adding to
 * the existing buffer
 */
void static crunch(char *buf,char *path)
{
	register char *cp;
	char drive = 0;

	cp = buf + strlen(buf);	/* Start write at end of current buffer */

	/* Now start crunching the pathname argument */
	for(;;) {
		/* Strip leading /'s; one will be written later */
		if (path[1] == ':') {
			drive = *path;
			path += 2;
		}
		while(*path == '/')
			path++;
		if(*path == '\0') {
			if (drive) {
				*cp++ = drive;
				*cp++ = ':';
			}
			break;		/* no more, all done */
		}
		/* Look for parent directory references, either at the end
		 * of the path or imbedded in it
		 */
		if(strcmp(path,"..") == 0 || strncmp(path,"../",3) == 0){
			/* Hop up a level */
			if((cp = strrchr(buf,'/')) == 0)
				cp = buf;	/* Don't back up beyond root */
			*cp = '\0';		/* In case there's another .. */
			path += 2;		/* Skip ".." */
			while(*path == '/')	/* Skip one or more slashes */
				path++;
		/* Look for current directory references, either at the end
		 * of the path or imbedded in it
		 */
		} else if(strcmp(path,".") == 0 || strncmp(path,"./",2) == 0){
			/* "no op" */
			path++;			/* Skip "." */
			while(*path == '/')	/* Skip one or more slashes */
				path++;
		} else {
			/* Ordinary name, copy up to next '/' or end of path */
			if (drive) {
				*cp++ = drive;
				drive = 0;
				*cp++ = ':';
			}
			*cp++ = '/';
			while(*path != '/' && *path != '\0')
				*cp++ = *path++;
		}
	}
	*cp++ = '\0';
}
