//
//   $Id: RCS.C 1.7 2003/05/19 17:39:26Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 * RCS.C
 *
 * Copyright (C) 2001 Datalight, Inc.
 * All Rights Reserved
 * Datalight Confidential and Proprietary
 * Portions Copyright (c) GP van Niekerk Ondernemings 1990-2000
 *
 * remote console server
 *
 * Revision History
 *		03/29/2003 gvn changed variable names, fixed invalid datastream when sending large blocks
 *    07/20/2001 gvn cleaned up comments
 *    07/13/2000 gvn Enable interrupts during calculation
 */

#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <process.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include "compiler.h"
#include "api.h"
#include <time.h>
#include "htapic.h"
#include "httftps.h"
#include "login.h"

#define PROBETIME	55//probe the display buffer every PROBTIME millisecs to see if data has changed
#define RCTIMERINT	TIMERINT
//#define RCTIMERINT	9

static void Login(void);

static int iDataSize, iRows = 25, iCols = 80, iNumRows, iNumCols, iSendSize;

static char *pcData;	//copy of video / attr buffer to determine what has changed
static char *pcSendBuf;
static char *pcWorkBuf;// pcWorkBuf = pcSendBuf + 500;
static int iSendCount;

//login variables
static char cLoggingIn = 1;
static char arcLogName[LINELEN];//login name
static char arcLogPass[LINELEN];//login password
static int iLogLen = 0;


#define READSIZE	300
static char arcReadBuf[READSIZE];//should never read this much data at a time
int iReadOffset;
int iSendOffset;
char far *fpcBiosVideo;
static int api_int;
static int iLoopPos;				//the position in the screen buffer where to look next
static unsigned tcbs, serv_tcbo;//segment where tcb's recide
static void interrupt (*oldTimer)(void);
void interrupt newTimer(void);
static unsigned tcbo_c;			//our one and only connection's tcbo
//static unsigned max_input;
static unsigned char curX, curY;
#define QUELEN	50
static char arcQue[QUELEN];	//ringbuffer
static int iQIn; 					//position where byte is entered
static int iQOut;					//postion where byte is taken away

//types:

//receiving:
#define T_KEY			'k'//receives a key, {'k', ascii, scancode}
#define T_SHIFSTATE	's'//changes shiftstate {'s', shiftstate}
#define T_REFRESH		'r'//refreshes whole screen {'r'}

#define CARET		'c'	//positions caret
#define ASPR		'o'	//aspect ratio
#define INFO		'i' 	//serial info
#define SCROLL		's' 	//scroll
#define MOVE   	'm' 	//move
#define ATR_DATA  'a' 	//constant attribute plus string of data
#define REPEAT_DA 'p'	//repeat data/attribute pair
#define DIS_DATA  'd'	//display data - data/attribute string
#define TOUPPER 	0xdf	// & to get uppercase (for long blocks)


//#define SANITY_CHECK
#ifdef SANITY_CHECK
int insanecount = 0;
void insane(void)
{
	insanecount++;
}
#endif

/*nice idea, but doesnt work to well from service interrupt, probably because SS != DS?.

//sometimes 0's are passed as arguments, therefore we need to supply the number of args explicitly
void AddTypeAndLen(int length, ...)
{
	va_list ap;

	va_start(ap, length);
	while (length--) {
		if (iSendCount < iSendSize)
			pcSendBuf[iSendCount++] = (char)va_arg(ap, int);
	}
	va_end(ap);
}
*/

//put control byte in sendbuffer, then length byte that it applies for.
static int AddTypeAndLen(char cType, int iLength)
{
	if (iSendCount > iSendSize - 3) {//not enough space
#ifdef SANITY_CHECK
		insane();
#endif
		return -1;
	}
	if (iLength > 0xff) {	// will need 2 bytes for length info.
		// Send upper case so we know it's longer.
		pcSendBuf[iSendCount++] = cType & TOUPPER;
		pcSendBuf[iSendCount++] = iLength >> 8;
	}
	else
		pcSendBuf[iSendCount++] = cType;

	pcSendBuf[iSendCount++] = iLength & 0xff;
	return 0;
}

//put data byte in send buffer
static int AddByte(char byte)
{
	if (iSendCount < iSendSize) {
		pcSendBuf[iSendCount++] = byte;
		return 0;
	}
#ifdef SANITY_CHECK
		insane();
#endif
	return -1;
}

/*
How it works:
Data is send in blocks
There are different 'shorthands' for defining data blocks, both in terms of
existing data and implicitly. Some ways are more powerful than others, eg.
to say the block scrolled down one line will produce a shorter telegram than
indicating all the data-byte changes.

Power is determined by several factors, minimum network traffic the most
important. Others include computational time, amount of changes (and
repaints) by client and simplicity.

We always attempt to use the strongest way possible.

  The blocks (telegrams) used now are the following:

  Defined in terms of old data:
	  Same: The data didn't change
	  Scrolled up: The data scrolled one line up.
	  Scrolled down: The data scrolled one line down.

 Defined in terms of this block:
	Full pattern - data and attribute bytes form a pattern, eg. spaces on
						black screen
	Half pattern - either data or attribute byte is constant
	Data - No way is found of reducing the volume so the data is send as is.

  Other
	Setup info: monitor mode, iRows and columns, etc.
	Serialization info: used for licensing, etc.

*/


#ifdef DEB
#define LOGSIZE 5000
static char logbuf[LOGSIZE];
static int logcount = 0;

void log(void) {
	FILE *log;

	log = fopen("1.log","wb");
	//fprintf(log, "Logging %d bytes which are send to client\n",logcount);
	if(logcount != fwrite(logbuf, 1, logcount, log))
		fprintf(log, "\n\nERROR WRITING!!! %d\n\n", logcount);
	fclose(log);
}

void logString(char *string, int length)
{
	if (logcount >= LOGSIZE)
		return;//cannot do anything
	if (length == -1)
		length = strchr(string, 0) - string;
	if (length + logcount > LOGSIZE)
		length = LOGSIZE - logcount;
	memcpy(logbuf + logcount, string, length);
	logcount += length;
}
#endif

static void CopyDat(int iStart, int iEnd)
{
	iStart <<= 1;
	memcpy(pcData + iStart, pcWorkBuf + iStart, (iEnd << 1) - iStart);
}

// A "differ block" is from differing data to the next differing data following
// a match and is at least 3 bytes long.
static void DoDifferBlock (int iStart, int iEnd)
//look for patterns in block of data
{
	int *piDifVid = (int *)pcWorkBuf;
	int i, j;
	char cFin = 0;
	int iCount = 0;
	//new state, current state, old state
	char cNewState = 0, cCurrentState = 1, cCntState = 0, cSndState = 0;
	int iBufStart = iStart << 1, iBufEnd;

	i = iStart - 1;
	piDifVid += iStart;
	iEnd--;//make end inclusive, for faster conditional in while(1)
#ifdef SANITY_CHECK
	if (iStart > iEnd || iStart < 0 || iEnd > iDataSize)
		insane();
#endif

	while (1) {
		if (++i == iEnd) {
			cFin = 1;
			iBufEnd = (iEnd + 1) << 1;
		}
		else {
			cNewState = piDifVid[0] == piDifVid[1] ? 3 :
							(piDifVid[0] >> 8) == (piDifVid[1] >> 8) ? 2 : 1;
			piDifVid++;
			if (cNewState == cCurrentState && !cCntState)
				continue;
			if (cCntState) {
#define LIMIT 3
				if(cNewState >= cCntState) {
					if(++iCount == LIMIT) {
						cSndState = cCurrentState;
						cCurrentState = cCntState;
						cCntState = 0;
						iBufEnd = (i - LIMIT) << 1;
					}
				}
				else if (cNewState > cCurrentState)
					cCntState = cNewState;
				else if (cNewState < cCurrentState) {
					cSndState = cCurrentState;
					cCurrentState = cNewState;
					iBufEnd = (i + 1) << 1;
					cCntState = 0;
				}
				else
					cCntState = 0;     //onvoldoende:
			}
			else if (cNewState > cCurrentState) {
				cCntState = cNewState;
				iCount = 0;
			}
			else  {//if (cNewState < cCurrentState)
				cSndState = cCurrentState;
				cCurrentState = cNewState;
				iBufEnd = (i + 1) << 1;
			}
		}
		if (cSndState || cFin) {
			int iSize = iBufEnd - iBufStart;
			if (iSize < 1) //nothing to send
				cSndState = 4;//just to skip switch
#ifdef SANITY_CHECK
			if (iSize < 0 || iSize > (iDataSize << 1) || iBufStart < 0)
				insane();
#endif
			if (!cSndState)
				cSndState = cCurrentState ? cCurrentState : cSndState ? cSndState : 1;
			switch (cSndState) {
			case 1://send data and atr
				//if (!);//check for space in sendbuffer
				AddTypeAndLen(DIS_DATA, iSize >> 1);
				memmove(pcSendBuf + iSendCount, pcWorkBuf + iBufStart, iSize);
				iSendCount += iSize;
				break;
			case 2://send data and const atr byte
				//if (!);//check for space in sendbuffer
				AddTypeAndLen(ATR_DATA, iSize >> 1);
				AddByte(*(pcWorkBuf + iBufStart + 1));
				for (j = iBufStart; j < iBufEnd; j += 2)
					pcSendBuf[iSendCount++] = *(pcWorkBuf + j);
				break;
			case 3://send const data + atr bytes
				//if (!);//check for space in sendbuffer
				AddTypeAndLen(REPEAT_DA, iSize >> 1);
				AddByte(*(pcWorkBuf + iBufStart));
				AddByte(*(pcWorkBuf + iBufStart + 1));
				break;
			}
			if (cFin)
				break;
			iBufStart = iBufEnd;
			cSndState = 0;
		}
	}
}


/*
To analyze: This is the isr called on a timer. It starts at the beginning of the
screen and increments i until the end is reached.  It transmits differblocks as
they are encountered by calling DoDifferBlock(). Since only changes are sent, an
algorithm must compare every current byte in video memory to the previous image.
The algorithm starts its functioning near the end of the while loop the first
time the while executes.  There it tests for patterns or differblocks.
If one of those are found, the next time the loop executes, they are processed
(near the middle of the loop).
If the entire screen is covered, i == iDataSize, and everything for this
screenupdate is completed as is evident of all the break statements found near
the beginning of this infinite loop.

Scroll operation:
	In the client, first a move buffer pointer command is received.
	The scroll command is interpreted as follows:
	First get the amount of lines to be scrolled.
	Then get the offset from the current buffer pointer from where the data must
	be copied.
	Example : >move 80,
		  >scroll 1840;start row 1:
			Apply scroll to 1840/80 =23 lines,
			[start 1 row further than "row 1" (where we just moved with "move") = 2]
		  ie scroll line 2 to line 1.

*/
static void calc2(void)
{
	int i = -1;
	int iStart = 0, iSize, iDifStart = 0;
	int *piVideo = (int *)pcWorkBuf;
	int *piData = (int *)pcData;
	int *piDataStart;
	int *piDataMax = piData + iDataSize;
	char cGotMatch = 0;
	char cInDifBlock = 1;
	int iScrollOffset = 0;

	//first copy data to pcWorkBuf
	_fmemcpy (pcWorkBuf, fpcBiosVideo, iDataSize * 2);

	piVideo--;//cluge for loop (see below)
	piData--;
	while(1) {
		piVideo++;
		piData++;
		if (++i == iDataSize) {
			//We came to the end of the screen, complete everything.
			if (cInDifBlock) {
				DoDifferBlock(iDifStart, iDataSize);
				break;
			}
			if (iStart > iDifStart)
				DoDifferBlock(iDifStart, iStart);
			iSize = iDataSize - iStart;
			if (!iScrollOffset) {
				if (iSendCount)
					AddTypeAndLen(MOVE, iSize);
				break;//else nothing new, dont send anything
			}
#ifndef NO_SCROLL
			AddTypeAndLen(SCROLL, iSize); //Add amount of lines to be scrolled, counted in chars.
			AddByte(iScrollOffset / iCols); //Add from which line to scroll from.
							//It is really the offset from the current
							//position of the pointer for the disp.buffer.
			CopyDat(iStart, i);
#endif
			break;
		} // if (++i == iDataSize)
		//Not at the end of the screen:
		if (cGotMatch) {
			if (*piVideo == *piData && piData < piDataMax)//watch for top boundary
				continue;
		}
		if (!cInDifBlock) {//indicates that we had a match but now fell out
			iSize = i - iStart;
			if (iSize > 3) {//this block is long enough to be worth sending

				//if there's a differ block waiting, do it
				if (iStart > iDifStart)
					DoDifferBlock (iDifStart, iStart);
				iDifStart = i;
#ifndef NO_SCROLL
				if (!iScrollOffset)
					AddTypeAndLen(MOVE, iSize);
				else {
					//AddTypeAndLen(MOVE,0);	// debug
					AddTypeAndLen(SCROLL, iSize);
					AddByte(iScrollOffset / iCols);
					CopyDat(iStart, i);
				}
#else
				AddTypeAndLen(MOVE, iSize);
#endif
			}//iSize > 3
			else {
				cInDifBlock = 1;
				CopyDat(iStart, i);
			}
		} //!cInDifBlock

		//look for new type : Algorithm really starts here.
		piDataStart = piData = (int*)pcData + i;
		cGotMatch = 0;
		do {
			if (*piData == *piVideo) {
				iScrollOffset = piData - piDataStart;  // words scrolled
				iStart = i;
				cGotMatch = 1;
				cInDifBlock = 0;
				break;
			}
#ifndef NO_SCROLL
			piData += iCols;
			if (piData >= piDataMax)
				piData -= iDataSize; //decrement a screensize.
#endif
		} while (piData != piDataStart);
		if (cGotMatch)
			continue;
		*piData = *piVideo;//copy data-attrib word (piData will be pointing to right place after do-while
		if (!cInDifBlock) {
			cInDifBlock = 1;
			iDifStart = i;
		}
	}//end while
}

void SendSize(void)
{
	AddByte(ASPR);
	AddByte(iNumCols);
	AddByte(iNumRows);
}

static char intimer = 0;
static int sent;
static unsigned intsp, intss, myss, stack;
//void interrupt far newTimer()
void RCtimer(void)
{
#ifdef DEB
	if(!intimer) {
#else
	if(tcbo_c && !intimer) {
#endif
		intimer = 1;

		intsp = _SP;
		intss = _SS;
		_SS = myss;
		_SP = stack;
		enable();//because this is a real interrupt interrupts are turned off at this stage
		if (iSendOffset == -1) {
			iSendCount = iSendOffset = 0;
			if (cLoggingIn)
				Login();
			else {
				//get caret position
				_BH = 0;
				_AH = 3;
				geninterrupt(0x10);
				if (curX != _DL || curY != _DH) {
					AddByte(CARET);
					AddByte(curX = _DL);
					AddByte(curY = _DH);
				}
				if (iNumRows != iRows || iNumCols != iCols) {
					iNumCols = iCols;
					iNumRows = iRows;
					SendSize();
				}
				calc2();
#ifdef DEB
				if (iSendCount) {
					logString(pcSendBuf, iSendCount);
					//logString("XXX", 3);
					iSendCount = 0;
				}
#endif
			}
		}
		if (iSendCount) {
			sent = SendTcph(tcbo_c, pcSendBuf + iSendOffset, iSendCount - iSendOffset);
			if (sent > 0) {
				iSendOffset += sent;
				if (iSendOffset >= iSendCount)
					iSendOffset = -1;//may calculate again
			}
			else if (sent < 0) {
				extern int err_code, err_count;
				err_code = -sent;
				err_count++;
			}
		}
		else
			iSendOffset = -1;//may calculate again

		disable();
		intimer = 0;
		_SS = intss;
		_SP = intsp;
	}
}

/*
writes a text message in the sendbuffer so it may be send to the client
assumes caret is correctly positioned at beginning of input field

static int WriteInBuf(char *text)
{
	int iLength = strlen(text);

	AddTypeAndLen(ATR_DATA, iLength >> 1);
	AddByte(7);						//white on black
	memcpy(pcSendBuf + iSendCount, text, iLength);
	iSendCount += iLength;
	return iLength;
}
*/
/*
sends a string and positions the cursor just behind the end of the string
*/
static void sendString(char *szString)
{
	int iLen;

	iLen = strlen(szString);
	if (!iLen)
		return;//nothing to do
	AddTypeAndLen(ATR_DATA, iLen);
	AddByte(7);//the constant attribute
	while (*szString)
		AddByte(*szString++);
	//position caret
	curX += iLen;
	while (curX >= iCols) {
		curX -= iCols;
		curY++;
	}
	AddByte(CARET);
	AddByte(curX);
	AddByte(curY);
}

/*
move ip positions and update cursor
*/
static void Move(int iP)
{
	AddTypeAndLen (MOVE, iP);
	curX += (iP % iCols);
	curY += (iP / iCols);
	while (curX >= iCols) {
		curX -= iCols;
		curY++;
	}
	while (curY >= iRows)
		curY -= iRows;
}
/*
handle login procedures by both reading input bytes and queing output bytes
*/
static void Login(void)
{
	static char cCh;
	static char sz1byte[2] = "x";
	static int tRelTicks;
	static char *pBuf;

	if (cLoggingIn == 1) {//start of process
		SendSize();
		curX = curY = 0;
		sendString("Enter login name: ");
		cLoggingIn = 2;
	}
	if (cLoggingIn == 4 && tRelTicks++ > 35) {//timeout period
		sendString("Re-enter login name: ");
		cLoggingIn = 2;
	}

	while (iQOut != iQIn) {
		cCh = arcQue[iQOut++];
		if (iQOut == QUELEN)
			iQOut = 0;	//wrap

		if (cLoggingIn == 4)
			continue;

		if (cCh == 0xd)//CR
			cCh = 0;
		if (iLogLen >= LINELEN - 1 && cCh != 0)
			continue;
		if (cLoggingIn == 2)
			pBuf = arcLogName;
		else if (cLoggingIn == 3)
			pBuf = arcLogPass;
		else
			 sendString("Internal server error");
		pBuf[iLogLen++] = cCh;

		//echo character
		if (cCh != 0) {
			if (cLoggingIn == 3)//password
				cCh = '*';//echo *
			*sz1byte = cCh;
			sendString(sz1byte);
		}
		else {
			iLogLen = 0;
			//cr, move to next line, etc.
			if (cLoggingIn == 2) {
				Move(iCols - curX);
				sendString("Enter Password: ");
				cLoggingIn = 3;
			}
			else {
				//move to end
				Move((iRows - curY) * iCols - curX);

				//clear all
				AddTypeAndLen(REPEAT_DA, iCols * iRows);
				AddByte(' ');
				AddByte(7);

				//test
				if (Verify(RQ_REMCON, 0, arcLogName, arcLogPass) == L_ALLOWED)
					cLoggingIn = 0;
				else {
					cLoggingIn = 4;
					sendString("Login failed! Please wait 2 seconds...");
					tRelTicks = 0;
				}
			}
		}
	}
}

static void reset(void) //reset the state of the connection.
{
	int *p = (int *)pcData + iDataSize;
	iSendCount = iLoopPos = 0;
	iSendOffset = -1;
	curX = curY = -1;
	iNumRows = iRows;
	iNumCols = iCols;
	while ((char *)p != pcData) //fill with white on black spaces, just as client assumes default
		*--p = 0x720; //little endian
}

static void boot(void)
{
	void (far *freset)(void);

	freset = MK_FP(0xffff,0);
	(*freset)();
}

//receive upcall
#define KEYPRESS		'k'
#define SHIFTSTATE	's'
#define REBOOT			'b'
#define REFRESH		'r'
#define INFO			'i'
#define NORM			0
#define ASCII			1
#define SCANCODE		2

static char state = NORM;
static char scancode;
#pragma argsused
static void _loadds far r_u(WORD tcbo,WORD cnt)
{
	int i = 0;
	static union REGS r;
	static char c;

	r.h.ah = API_TCPREAD;
	r.x.bx = tcbo_c;
	r.x.cx = READSIZE;// - iReadOffset;
	r.x.dx = FP_SEG(arcReadBuf);
	r.x.si = FP_OFF(arcReadBuf);// + iReadOffset;
	int86(api_int, &r, &r);
	if (r.x.ax)
		return;//error
	cnt = r.x.cx;// + iReadOffset;
	while (cnt--) {
		c = arcReadBuf[i++];
		switch (state) {
		case NORM:
			switch (c) {
			case KEYPRESS:
				state = SCANCODE;
				break;
			case SHIFTSTATE:
				state = SHIFTSTATE;
				break;
			case REFRESH:
				reset();
				break;
			case REBOOT:
				boot();
				break;
			}
			break;
		case SHIFTSTATE:
			state = NORM;
			break;
		case SCANCODE:
			scancode = c;
			state = ASCII;
			break;
		case ASCII:
			if (cLoggingIn) {
				//put in que that will be handled at timer interrupt
				if (iQIn + 1 != iQOut) {
					arcQue[iQIn++] = c;
					if (iQIn == QUELEN)
						iQIn = 0;	//wrap
				}
			}
			else {
				//easy fix for now: dont capture interrupt 16h but use int 16h funct 5
				r.h.ah = 5;
				r.h.ch = scancode;//scancode
				r.h.cl = c;//asci
				int86(0x16, &r, &r);
				//throw away if (r.h.al == 1)//buffer full
				//	goto endloop;//wait till next upcall
			}
			state = NORM;
			break;
		}
	}
//	endloop:
//	memcpy(arcReadBuf,arcReadBuf + i,cnt);
//	iReadOffset = cnt;
}

#pragma argsused
static void _loadds far s_u(WORD tcbo,char old, char new)
{
	switch (new) {
	case ESTABLISHED:
		if (!tcbo_c) {
			reset();
			tcbo_c = tcbo;
			cLoggingIn = 1;
			//create_RT((unsigned)tcbo);
		}
		else {//allow only one connection
			CloseTcp(tcbo);
			DelTcp(tcbo);
		}
		break;
	case CLOSE_WAIT:
		CloseTcp(tcbo_c);
		break;
	case TIME_WAIT:
	case CLOSED:
		DelTcp(tcbo_c);
		tcbo_c = 0;
		//delete_RT((unsigned)tcbo);
		break;
	}
}

TCP_OP_P rms = {0, 0, 0, TCP_SERVER, 0,
	r_u, 0, s_u, 0, 0};

/*
static char	debsdat[34] = {' ', 0x70,' ',0x70,0xf0,0x74,' ',0x70,' ',0x70
						,'F',0x74,'i',0x70,'l', 0x70,'e', 0x70
						,' ', 0x70,' ', 0x70,'E',0x74,'d', 0x70
						,'i', 0x70,'t', 0x70,' ', 0x70,' ',0x70};
*/
static char *pcStackBlock;
#define STACKSIZE 100
//start a remote terminal server listening on port port
int init_RTS(int apiInt, int port)
{
	union REGS r;

	api_int = apiInt;
	rms.wLocPort = port;
	//printf("Starting remote console server LISTEN on port %u\n",port);
	r.x.cx = FP_OFF(&rms);
	r.x.bx = FP_SEG(&rms);
	r.h.ah = API_TCPOPEN;
	int86(api_int,&r,&r);
	tcbs = r.x.bx;
	//printf ("RCS tcbs = %u\n", tcbs);
	serv_tcbo = r.x.cx;

	//get the correct vector to the bios video
	if (!fpcBiosVideo) {
		r.h.ah = 0xf;
		int86(0x10, &r, &r);

		if (r.h.al != 7)
			fpcBiosVideo = MK_FP(0xb800,0);//color
		else
			fpcBiosVideo = MK_FP(0xb000,0);//mono
	}
	//allocate buffers, etc.
	//printf("fpcBiosVideo = %Fp, iRows = %d, iCols = %d\n",fpcBiosVideo, iRows, iCols);
/*#ifdef DEB
fpcBiosVideo = (char far*)debsdat;
iRows = 1;
iCols = 17;
#endif
*/
	iDataSize = iRows * iCols;
	pcData = malloc(iDataSize * 2);
	pcSendBuf = malloc(iSendSize = iDataSize * 2 + 500);//for safety against buffer overrun and data overlap
	pcWorkBuf = pcSendBuf + 500;
	pcStackBlock = malloc(STACKSIZE);
	memset(pcStackBlock,0x55,STACKSIZE);	// for debugging
	if (!pcData || !pcSendBuf || !pcStackBlock) {
		printf("not enough memory to start RTS server\n");
		close_RTS();
		return -1;
	}
	myss = _DS;
	stack = (unsigned)pcStackBlock + STACKSIZE - 2;//stacks work wrong way round
	reset();
	//set timer
//	oldTimer = getvect(RCTIMERINT);
//	setvect(RCTIMERINT, newTimer);
	return 0;
}

void interrupt newTimer(void)
{
	(*oldTimer)();
	RCtimer();
}


void SetRCtimer()
{
	oldTimer = getvect(RCTIMERINT);

	setvect(RCTIMERINT, newTimer);
}

void close_RTS(void)
{
	if (oldTimer) {
		setvect(RCTIMERINT, oldTimer);
		oldTimer = 0;
	}
	if (pcData)
		free(pcData);
	if (pcSendBuf)
		free(pcSendBuf);
	if (pcStackBlock)
		free(pcStackBlock);
	pcSendBuf = pcData = pcStackBlock = 0;
	if (tcbo_c) {
		CloseTcp(tcbo_c);
		DelTcp(tcbo_c);
		tcbo_c = 0;
	}
	CloseTcp(serv_tcbo);//close server socket
	DelTcp(serv_tcbo);
	serv_tcbo = 0;
}


static unsigned hstoi(char *p)
{
	unsigned ret = 0;
	while (*p) {
		ret <<= 4;
		if (*p >= '0' && *p <= '9')
			ret += *p - '0';
		else if (*p >= 'A' && *p <= 'F')
			ret += *p - 'A' + 10;
		else if (*p >= 'a' && *p <= 'f')
			ret += *p - 'a' + 10;
		else
			return 0;
		p++;
	}
	return ret;
}

void setBiosVideo(char *val)
{
	char *p = strchr(val, ':');
	if (p)
		*p++ = 0;
	fpcBiosVideo = MK_FP(hstoi(val), p ? hstoi(p) : 0);
}

void setAspectRatio(char *val)
{
	char *p = strchr(val, ',');

	if (p) {
		*p++ = 0;
		iRows = atoi(p);
	}
	iCols = atoi(val);
}
