//
//   $Id: CAPIDIAG.C 1.12 2003/06/27 06:19:14Z johnb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
	$DOCFILE:CAPIDIAG.C

	Copyright (C) 1999-2001 Datalight, Inc.
	All Rights Reserved

	Diagnostic testing Sockets/DOS using CAPI

	$DOCHISTORY

	06/26/2003 gvn Fixes to bulk read/write tests, reliable echo server
	12/30/2002 gvn added tests for ListenAccept, Accept, GetKernelInfo
	11/30/2002 gvn cleaned up SetAsynNotification test
	09/27/2002 gvn added SetAsyncNotification return value test
	10/03/2001 jmb  added DJGPP support
	05/25/2001 jmb  updated source from GPvNO - improved Volume speed data
						 tests.
	05/02/2001 jmb  modified to use build number instead of version number
	04/30/2001 gvn  Volume data speed test
	04/30/2001 jmb  displayed Datalight copyright message
	04/20/2001 gvn  Added MSVC compatibility.  Fixed hang when the specified
						 server does not exist or does not offer ECHO services.
	03/29/2001 jmb  removed debug code to exit prematurely.
	03/20/2001 jmb  added check to see if GetKernelInformation fails
						 and to exit the program if it does.  I also changed
						 check of carry flag on return from a Sockets API call
						 to r.x.cflag from r.x.flags -- r.x.flags is not
						 available in MSVC.
	1.14 2000-04-12 Conditional compile to show GetNetInfo results
	1.13 2000-01-05 Test blocking write of 0 bytes
	1.12 1999-12-15 Changed test for Bad Request
	1.11 1999-08-13 Added test for extended register save
	1.10 1999-07-02 Sockets 1.1 release
	1.00 1999-06-07 First version

	Compile and link with CAPI.C
	Use TABS=3 spaces for properly formatted source

	CAPIDIAG requires a remote server with both a UDP and TCP echo server
	for a complete test. NT server, Unix or Inet can provide this. It can
	also be run without such a server as a stand-alone test using only the
	internal loopback capabilities of Sockets.
	Extended register test must be enabled by defining EXTREG_TEST. It has
	only been tested on BC3.1.
*/

#ifdef EXTREG_TEST
#pragma inline
#endif

#ifdef __DJGPP__
#include <dpmi.h>
#include <go32.h>
#endif


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include <process.h>
#include <stdarg.h>
#include <time.h>
#include <dos.h>
#include <ctype.h>
#include <signal.h>

#include "compiler.h"
#include "capi.h"
#include "api.h"	// for TCB definition
#include "capidiag.pt"

//
// prototypes
//
#ifdef __DJGPP__
 int D_FAR ClientHandler(int iSocketTmp, int iEventTmp, DWORD dwArgTmp);
 int D_FAR ServerHandler(int iSockTmp, int iEventTmp, DWORD dwArgTmp);
#else
 int D_FAR ClientHandler(int iSocket, int iEvent, DWORD dwArg);
 int D_FAR ServerHandler(int iSocket, int iEvent, DWORD dwArg);
#endif


#ifndef __DJGPP__
#ifndef enable
	#define enable _enable
#endif
#ifndef disable
	#define disable _disable
#endif
#endif

#ifdef NULL_POINTER_TEST
char arcNullData[30];
#endif

int iGlobalServerType;	// Local Server type in use
int iClientSock, iServerSock, iListenSock;
NET_ADDR sClientAddr = {0,7,0,0}, sServerAddr = {0,0,7,0};

int iFailAction;		// action on failure
int iFail;				// fail/pass
int iVerbose;			// display tests
WORD wTimeOut = 10;	// timeout value
WORD wTimeOutHalf = 5;
int iFailures;
int iNumberTests;
int iWriteSize, iReadSize, iReadOffset, iWriteOffset;
int iBlockError;
int iErrorAddress;
DWORD dwBufsWritten, dwBufsRead;
WORD wTime, wTarget, wTimeRead;
BYTE bWriting;
WORD wRead,wWrite;
WORD wBulkTicks = 182;
BYTE bTestingAccept;	// testing AcceptSocket()

char szData[] = "0123456789abcdefghijklmnopqrstuvwxyzABC";
#define BUFLEN		1024
char rgcReadBuf[BUFLEN];
char rgcWriteBuf[BUFLEN];
DWORD dwMyIPAddress;
DWORD dwServerIPAddress;
BYTE bDebugServer;		// debug the server
volatile BYTE bControlBreak;
char szDiagnostic[400];	// Currently running diagnostic
#define WRITE_AHEAD	3	// # of blocks to write ahead in bulk transfer tests

// Client event variables, set at interrupt time
volatile int iEventPosted, iEventError;    //client callback related

typedef struct sClientEventStruc
{
	int iEventCount;		// count of events
	int iEventError;		// last event error
} CLIENT_EVENTS;

volatile CLIENT_EVENTS arsClientEvents[MAX_AS_EVENT + 2];

char *ppszEventMessage[] =
{
	"NET_AS_ALARM",
	"NET_AS_OPEN",
	"NET_AS_RCV",
	"NET_AS_XMT",
	"NET_AS_XMT_FLUSH",
	"NET_AS_FCLOSE",
	"NET_AS_CLOSE",
	"NET_AS_ERROR",
	"NET_AS_UNKNOWN",
};

DPMI_SECTION_EXTERNS;

void ControlBreak(int iSignal)
{
	bControlBreak = iSignal;
}

int main(int argc, char *argv[])
{
	WORD wLen;
	char szNameBuf[80], szVersion[20];
	BYTE pbBuf[16];
	char *pszArg;
	int i, iVersion;
	KERNEL_CONFIG sKernelConfig;


	printf("Datalight Sockets CAPI Diagnostic version 2.03(%s)\n",__DATE__);
	printf("Copyright (C) 1999-2003 Datalight, Inc.\nAll Rights Reserved\n"
			 "Portions copyright (c) GPvNO 2003\n\n");
	while (argc > 1 && argv[1][0] == '/')
	{
		pszArg = argv[1] + ((argv[1][2] == '=') ? 3 : 2);
		switch (tolower(argv[1][1]))
		{
			case 'b':	// bulk test time in seconds
				if ((wBulkTicks = (atoi(pszArg) * 182) / 10) == 0)
					wBulkTicks = 18;
				break;
			case 'd':
				bDebugServer = 1;
				break;
			default:
				printf(	"CAPIDIAG [/d][/l=<level>][/v=<verbose>] [<server>]\n"
							"/b=<bulk_test_seconds>\n"
							"/d - Debug server\n"
							"/f=<fail_action>\n"
							"   0 print error and continue\n"
							"   1 Print error and wait for Esc\n"
							"   2 Print error and exit\n"
							"/t=<timeout_seconds>\n"
							"/v=<verbosity> add values below\n"
							"   0 only print errors\n"
							"   1 print names of all tests\n"
							"   2 print additional information on errors\n"
							"   4 print summary of events\n"
							"   8 print kernel configuration\n"
							"  16 print buffer for bulk readback failure\n");
				exit(0);
			case 'f':
				iFailAction = atoi(pszArg);
				break;
			case 't':
				wTimeOut = atoi(pszArg);
				if ((wTimeOutHalf = wTimeOut / 2) == 0)
					wTimeOutHalf = 1;
				break;
			case 'v':
				iVerbose = atoi(pszArg);
				break;
		}
		--argc;
		++argv;
	}
#ifdef NULL_POINTER_TEST
	memcpy(arcNullData,(char *)0,sizeof(arcNullData));
#endif
	//
	// lock code and data
	//
	DPMI_LOCK_SECTIONS;

	wLen = sizeof(szNameBuf);
	szNameBuf[0] = 0;
	i = GetKernelInformation(0, K_INF_HOSTNAME, 0, szNameBuf, &wLen);
	if ((iVersion = GetSocketsVersion()) < 0)
	{
		if (iNetErrNo == ERR_API_NOT_LOADED)
			sprintf(szVersion, "API not loaded");
		else
			sprintf(szVersion, "build < 205");
	}
	else
		sprintf(szVersion, "build %04u", iVersion);
	printf("CAPI Diagnostic on '%s', Sockets %s\n", szNameBuf, szVersion);
	if (iNetErrNo == ERR_API_NOT_LOADED)
		exit(1);
	signal(SIGINT, ControlBreak);
	Start("GetKernelConfig");
	if (GetKernelConfig(&sKernelConfig) < 0)
		Fail(0);
	End();
	if (iVerbose & 8)
	{
		printf("Kernel configuration:\n"
			"Number of TCP connections allowed       %5d\n"
			"Number of UDP connections allowed       %5d\n"
			"Number of IP connections allowed        %5d\n"
			"Number of RAW_NET connections allowed   %5d\n"
			"Number of TCP connections in use        %5d\n"
			"Number of UDP connections in use        %5d\n"
			"Number of IP connections in use         %5d\n"
			"Number of RAW_NET connections in use    %5d\n"
			"Number of active Dos Compatible Sockets %5d\n"
			"Number of active Sockets                %5d\n"
			"Maximum header on an attached network   %5d\n"
			"Maximum trailer on an attached network  %5d\n"
			"Size of a large packet buffer           %5d\n"
			"Number of network interfaces attached   %5d\n"
			"Milliseconds since kernel started       %5ld\n"
			"IP broadcast address in use   %s\n\n",
				sKernelConfig.bKMaxTcp,sKernelConfig.bKMaxUdp,
				sKernelConfig.bKMaxIp,sKernelConfig.bKMaxRaw,
				sKernelConfig.bKActTcp,sKernelConfig.bKActUdp,
				sKernelConfig.bKActIp,sKernelConfig.bKActRaw,
				sKernelConfig.wKActDCS,sKernelConfig.wKActSoc,
				sKernelConfig.bKMaxLnh,sKernelConfig.bKMaxLnt,
				sKernelConfig.bKLBUF_SIZE,sKernelConfig.bKNnet,
				sKernelConfig.dwKCticks,
				IPAddressString(sKernelConfig.dwKBroadcast));
	}
	Start("GetKernelInformation(K_INF_HOSTNAME)");
	if (i < 0)
		Fail(0);
	End();

	wLen = 16;
	Start("GetKernelInformation(K_INF_DNS_SERVERS)");
	if (GetKernelInformation(0, K_INF_DNS_SERVERS, 0, pbBuf, &wLen) < 0)
		Fail(0);
	else if (iVerbose & 8)
	{
		printf("DNS(%d): ",wLen);
		for (i = 0;i < (int)wLen;i += 4)
			printf("%s ",IPAddressString(*(DWORD *)(pbBuf + i)));
		printf("\n");
	}
	End();

	wLen = 6;
	Start("GetKernelInformation(K_INF_MAC_ADDR)");
	if (GetKernelInformation(0, K_INF_MAC_ADDR, 0, pbBuf, &wLen) < 0)
		Fail(0);
	else if (iVerbose & 8)
	{
		printf("MAC address(%d): ",wLen);
		for (i = 0;i < (int)wLen;i++)
			printf("%02X ",pbBuf[i]);
		printf("\n");
	}
	End();

	wLen = 4;
	Start("GetKernelInformation(K_INF_IP_ADDRESS)");
	if (GetKernelInformation(0, K_INF_IP_ADDR, 0, (char *)&dwMyIPAddress, &wLen) < 0)
		Fail(0);
	End();
	printf("Local IP address:  %s\n",IPAddressString(dwMyIPAddress));
	dwServerIPAddress = dwMyIPAddress;

	RunDiagnostics();
	if (argc > 1)
	{
		Start("ResolveName(\"%s\")",argv[1]);
		if ((dwServerIPAddress = ResolveName(argv[1], szNameBuf, sizeof(szNameBuf))) == 0)
			Fail(0);
		End();
		RunDiagnostics();
	}
	printf("Number of tests: %d, Failures: %d\n", iNumberTests, iFailures);
	return 0;
}

void RunDiagnostics(void)
{
	int i;
	char szAddress[40];
	DWORD dwAddress;

	printf("Remote IP address: %s\n", IPAddressString(dwServerIPAddress));
	sClientAddr.dwRemoteHost = dwServerIPAddress;

	sprintf(szAddress,"%s",IPAddressString(dwServerIPAddress));
	Start("ParseAddress");
	if ((dwAddress = ParseAddress(szAddress)) == 0L)
		Fail(0);
	else if (dwAddress != dwServerIPAddress)
		Fail("&returned %s", IPAddressString(dwAddress));
	End();

	// Check for bad system call returning an error
	Start("Bad request");
	if (BadRequest() != -1 || iNetErrNo != ERR_BAD_SYS_CALL)
		Fail(0);
	End();

	Start("GetVersion");
	if ((i = GetVersion()) < 0)
		Fail(0);
	else if (i != 0x214)
		Fail("& returned %#x, expected 0x214",i);
	End();

	Start("ICMPPing(10)");
	if (ICMPPing(dwServerIPAddress,10) < 0)
	{
		Fail(0);
		End();
	}
	else
	{
		End();
		Start("ICMPPing(1000)");
		if (ICMPPing(dwServerIPAddress,1000) < 0)
			Fail(0);
		End();
	}
	// flush keyboard buffer
	while (kbhit())
		getch();

	TestSocket(STREAM,0);
	if (dwServerIPAddress == dwMyIPAddress)
		TestSocket(STREAM,1);
	TestSocket(DATA_GRAM,0);
}

#ifdef EXTREG_TEST
struct sExStruct
{
	DWORD eax;
	DWORD ebx;
	DWORD ecx;
	DWORD edx;
} sExs, sExr;
#endif

DPMI_CB_REGS(client_regs);
DPMI_CB_REGS(server_regs);
DPMI_CB_INFO(client_info);
DPMI_CB_INFO(server_info);

void Exit(int iCode)
{
	AbortSocket(iClientSock);
	AbortSocket(iServerSock);
	AbortSocket(iListenSock);
	DPMIINFO_FREE(server_info);
	DPMIINFO_FREE(client_info);
	exit(iCode);
}

WORD Getch(void)
{
	WORD wKey;
	char *pszArgvl[2];	//used in spawning COMSPEC

	wKey = getch();
	if (wKey == 0)
		/* function key */
		wKey = 1000 + getch();
	switch (wKey)
	{
		case 1045:		/* alt X */
			printf("User exit\n");
			Exit(2);
			break;
		case 1035:	// Alt H
			printf( "\n"
					"Escape        Abort test\n"
					"Alt-E         Execute command.com\n"
					"Alt-H         Help\n"
					"Alt-X         eXit\n"
					"Any other key Display current test\n");
			break;
		case 1018:	// alt e
			if ((pszArgvl[0] = getenv("COMSPEC")) == 0)
				pszArgvl[0] = "\\COMMAND.COM";
			pszArgvl[1] = 0;
			spawnvp(P_WAIT, pszArgvl[0], pszArgvl);
			break;
	}
	return wKey;
}

DWORD dwClientHandlerAddr;
DWORD dwServerHandlerAddr;

void PrintReadBuffer(void)
{
	int i;
	char cCount;

	if (iVerbose & 16)
	{
		for (i = 0,cCount = *rgcReadBuf - 4;i < BUFLEN;++i,++cCount)
		{
			if ((i & 15) == 0)
				printf("\n%03x ",i);
			printf("%02x%s",(BYTE)rgcReadBuf[i],
					((i > 3) && rgcReadBuf[i] != cCount) ? "* " : "  ");
		}
		printf("\n");
	}
}

int TestSocket(int iMode, int iServerType)
{
	DWORD dwLocalIPAddress;
	NET_ADDR sPeerAddr,sAddr;
	NET_INFO sNetInfo;
	int i, iRead, iWritten, iSize, iResult;
	char *pszMode = iMode == STREAM ? (iServerType ? "STREAM LA" : "STREAM") :
																	 "DATA_GRAM";
	long lIflags = 0L, lOflags = 0L, lMask;
	BYTE bReadFailed = 0;
	char sz[40];
	void D_FAR *pfPreviousHandler;

#ifdef EXTREG_TEST
#ifdef __DJGPP__
	_ASM("movl %%eax, %0\n\t"
		  "movl %%ebx, %1\n\t"
		  "movl %%ecx, %2\n\t"
		  "movl %%edx, %3\n\t"
		  : "=g" (sExs.eax), "=g" (sExs.ebx),
			 "=g" (sExs.ecx), "=g" (sExs.edx)
		  :
		  : "memory"
		 );
#else
	// save Extended registers
	asm	.386;
	asm	mov dword ptr sExs.eax,eax;
	asm	mov dword ptr sExs.ebx,ebx;
	asm	mov dword ptr sExs.ecx,ecx;
	asm	mov dword ptr sExs.edx,edx;
	asm	.8086
#endif
#endif

	// set address for client and server handlers
	DPMI_CB_ALLOCRETF(client_info, ClientHandler, client_regs);
	if (iNetErrNo == ERR_NO_MEM)
	{
		Fail(0);
		dwClientHandlerAddr = 0l;
	}
	else
	{
#ifdef __DJGPP__
		dwClientHandlerAddr = ((DWORD)client_info.rm_segment << 16L) + client_info.rm_offset;
#else
		dwClientHandlerAddr = (DWORD)ClientHandler;
#endif
	}

	DPMI_CB_ALLOCRETF(server_info, ServerHandler, server_regs);
	if (iNetErrNo == ERR_NO_MEM)
	{
		Fail(0);
		dwServerHandlerAddr = 0l;
	}
	else
	{
#ifdef __DJGPP__
		dwServerHandlerAddr = ((DWORD)server_info.rm_segment << 16L) + server_info.rm_offset;
#else
		dwServerHandlerAddr = (DWORD)ServerHandler;
#endif
	}


	if (dwServerIPAddress == dwMyIPAddress)
		StartListen(iMode, iServerType);
	sClientAddr.wRemotePort = 7;
	sClientAddr.wLocalPort = 0;
	iEventPosted = 0;
	memset((void *)arsClientEvents,0,sizeof(arsClientEvents));

	Start("GetSocket(%s)", pszMode);
	if ((iClientSock = GetSocket()) < 0)
		Fail(0);
	End();

	// make iClientSock non-blocking
	Start("%s SetSocketOption(NET_OPT_NON_BLOCKING)", pszMode);
	if (SetSocketOption(iClientSock, 0, NET_OPT_NON_BLOCKING, 1, 1) < 0)
		Fail(0);
	End();


	// set invalid ASYNC Handler
	Start("%s SetAsyncNotification(INVALID)",pszMode);
	if (SetAsyncNotification(
						iClientSock, 20, Handler, 0l) != (int D_FAR *)-1 ||
						iNetErrNo != ERR_BAD_ARG)
		Fail(0);
	End();
	// set ASYNC Handler on OPEN
	Start("%s SetAsyncNotification() return value",pszMode);
	if ((pfPreviousHandler = SetAsyncNotification(iClientSock, NET_AS_OPEN, Handler0, 0)) == (int D_FAR *)-1)
		Fail(0);
	if (pfPreviousHandler != 0)
		Fail("Return value is %Fp, should be 0000:0000",pfPreviousHandler);
	//printf("Previous Handler %Fp\n",pfPreviousHandler);
	if ((pfPreviousHandler = SetAsyncNotification(iClientSock, NET_AS_OPEN, Handler, 0)) == (int D_FAR *)-1)
		Fail(0);
	if (pfPreviousHandler != (void D_FAR *)Handler0)
		Fail("Return value is %Fp, should be %Fp",pfPreviousHandler,(void D_FAR *)Handler0);
	//printf("Previous Handler %Fp\n",pfPreviousHandler);
	End();
	// set ASYNC Handler on RCV
	Start("%s SetAsyncNotification(RCV)",pszMode);
	if (SetAsyncNotification(iClientSock, NET_AS_RCV, Handler, 0l) == (int D_FAR *)-1)
		Fail(0);
	End();
	// set ASYNC Handler on XMT
	Start("%s SetAsyncNotification(XMT)",pszMode);
	if (SetAsyncNotification(iClientSock, NET_AS_XMT, Handler, 0l) == (int D_FAR *)-1)
		Fail(0);
	End();
	// set ASYNC Handler on CLOSE
	Start("%s SetAsyncNotification(CLOSE)",pszMode);
	if (SetAsyncNotification(iClientSock, NET_AS_CLOSE, Handler, 0l) == (int D_FAR *)-1)
		Fail(0);
	End();
	// set ASYNC Handler on FCLOSE
	Start("%s SetAsyncNotification(FCLOSE)",pszMode);
	if (SetAsyncNotification(iClientSock, NET_AS_FCLOSE, Handler, 0l) == (int D_FAR *)-1)
		Fail(0);
	End();
	// set ASYNC Handler on ERROR
	Start("%s SetAsyncNotification(ERROR)",pszMode);
	if (SetAsyncNotification(iClientSock, NET_AS_ERROR, Handler, 0l) == (int D_FAR *)-1)
		Fail(0);
	End();

	// test SetAlarm
	Start("SetAlarm");
	if (SetAlarm(iClientSock, 200, Handler, 0) < 0)
		Fail(0);
	End();
	Start("Alarm event");
	if (arsClientEvents[NET_AS_ALARM].iEventCount != 0)
		Fail("&fired too soon");
	SetTimer(wTimeOut);
	while (!iEventPosted)
	{
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
		if (arsClientEvents[NET_AS_ALARM].iEventCount != 0)
			break;
	}
	End();

	iEventPosted = 0;
	Start("ConnectSocket(%s)",pszMode);
	if (ConnectSocket(iClientSock, iMode, &sClientAddr) < 0)
		Fail(0);
	End();


	if (iMode == STREAM)
	{
		Start("%s open event",pszMode);
		SetTimer(wTimeOut*2);
		while (!iEventPosted)
		{
			if (DiagWait())
			{
				iFail = 1;
				goto EventAbort;
			}
		}
		switch (iEventPosted)
		{
		case NET_AS_ERROR:
			Fail("&NET_AS_ERROR %s", GetErrorString(iEventError));
			break;
		case NET_AS_CLOSE:
			Fail("&NET_AS_CLOSE");
			break;
		case NET_AS_OPEN:
			// This is the "good" case
			if (iVerbose & 1)
				printf("Connection established ");
			break;
		default:
			Fail("&Unexpected event %d",pszMode, iEventPosted);
			break;
		}
		EventAbort:
		End();
	}


	// Check for no receive events
	Start("%s no receive event",pszMode);
	if ((iResult = arsClientEvents[NET_AS_RCV].iEventCount) != 0)
		Fail("&count is %d, expected 0", iResult);
	End();

	// Check for no transmit events
	Start("%s no transmit event",pszMode);
	if ((iResult = arsClientEvents[NET_AS_XMT].iEventCount) != 0)
		Fail("&count is %d, expected 0", iResult);
	End();

	Start("%s GetAddress",pszMode);
	if ((dwLocalIPAddress = GetAddress(iClientSock)) == 0)
		Fail(0);
	if (dwLocalIPAddress != dwMyIPAddress)
		Fail("&returned %s", IPAddressString(dwLocalIPAddress));
	End();

	Start("%s GetPeerAddress",pszMode);
	if (GetPeerAddress(iClientSock,&sPeerAddr) < 0)
		Fail(0);
	if (sPeerAddr.dwRemoteHost != dwServerIPAddress)
		Fail("&returned wrong RemoteHost %s",IPAddressString(sPeerAddr.dwRemoteHost));
	if (sPeerAddr.wRemotePort != 7)
		Fail("&returned wrong RemotePort %d",sPeerAddr.wRemotePort);
	End();

	Start("%s GetNetInfo",pszMode);
	if (GetNetInfo(iClientSock,&sNetInfo) < 0)
		Fail(0);
	if (sNetInfo.dwIpAddress != dwMyIPAddress)
		Fail("&returned wrong IPAddress %s",IPAddressString(sNetInfo.dwIpAddress));
#ifdef SHOW_NETINFO
	printf("NetInfo():\ndwIpAddress %s\n",IPAddressString(sNetInfo.dwIpAddress));
	printf("dwIpSubnet  %s\n",IPAddressString(sNetInfo.dwIpSubnet));
	printf("iUp %d,iLanLen %d\nLanAddr %02x:%02x:%02x:%02x:%02x:%02x\n",
		sNetInfo.iUp,sNetInfo.iLanLen,
		sNetInfo.pLanAddr[0],
		sNetInfo.pLanAddr[1],
		sNetInfo.pLanAddr[2],
		sNetInfo.pLanAddr[3],
		sNetInfo.pLanAddr[4],
		sNetInfo.pLanAddr[5]);
#endif
	End();

	// Nothing should be in receive queue
	Start("%s ReadSocket(buf,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)", pszMode);
	if ((iRead = ReadSocket(iClientSock, rgcReadBuf,sizeof(rgcReadBuf),0,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)) != -1 ||
			iNetErrNo != ERR_WOULD_BLOCK)
		Fail("&unexpected count %d",iRead);
	End();
	Start("%s ReadSocket(0,NET_FLG_PEEK)",pszMode);
	if ((iRead = ReadSocket(iClientSock, 0,sizeof(rgcReadBuf),0,NET_FLG_PEEK)) != -1 ||
			iNetErrNo != ERR_WOULD_BLOCK)
		Fail("&unexpected count %d",iRead);
	End();
	Start("%s ReadSocket(0,0)",pszMode);
	if ((iRead = ReadSocket(iClientSock, rgcReadBuf,sizeof(rgcReadBuf),0,0)) != -1 ||
			iNetErrNo != ERR_WOULD_BLOCK)
		Fail("&unexpected count %d",iRead);
	End();

	Start("DisableAsyncNotification from enabled");
	if ((iResult = DisableAsyncNotification()) < 0)
		Fail(0);
	if (iResult != 1)
		Fail("&result %d, expected 1", iResult);
	End();

	Start("DisableAsyncNotification from disabled");
	if ((iResult = DisableAsyncNotification()) < 0)
		Fail(0);
	if (iResult != 0)
		Fail("&result %d, expected 0", iResult);
	End();

	// Send some data to echo server
	Start("%s WriteSocket",pszMode);
	if ((iWritten = WriteSocket(iClientSock, szData, sizeof(szData), 0)) < 0)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	Delay(10);		// give time to receive response

	// Check for no receive events
	Start("%s no receive event",pszMode);
	if ((iResult = arsClientEvents[NET_AS_RCV].iEventCount) != 0)
		Fail("&count is %d, expected 0", iResult);
	End();

	// Check for no transmit events
	Start("%s no transmit event",pszMode);
	if ((iResult = arsClientEvents[NET_AS_XMT].iEventCount) != 0)
		Fail("&count is %d, expected 0", iResult);
	End();

	Start("EnableAsyncNotification from disabled");
	if ((iResult = EnableAsyncNotification()) < 0)
		Fail(0);
	if (iResult != 0)
		Fail("&result %d, expected 0", iResult);
	End();

	Start("EnableAsyncNotification from enabled");
	if ((iResult = EnableAsyncNotification()) < 0)
		Fail(0);
	if (iResult != 1)
		Fail("&result %d, expected 1", iResult);
	End();

	if (iMode == DATA_GRAM)
	{
		// Just check datagram count, no reading should take place
		Start("%s ReadSocket(pBuf,NET_FLG_PEEK|NET_FLG_NON_BLOCKING)==1", pszMode);
		memset(rgcReadBuf,0,sizeof(rgcReadBuf));
		if ((iRead = ReadSocketCount(iClientSock,iMode,rgcReadBuf,sizeof(rgcReadBuf),
								1,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)) == 1)
		{
			if (memcmp(szData,rgcReadBuf,iRead) == 0)
				Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
		}
		End();
	}

	// Just check count, no reading should take place
	Start("%s ReadSocket(0,NET_FLG_PEEK)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, 0, sizeof(rgcReadBuf),
									sizeof(szData),NET_FLG_PEEK)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) == 0)
			Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	// Check for receive upcall having happened
	Start("%s receive event",pszMode);
	if (arsClientEvents[NET_AS_RCV].iEventCount == 0)
		Fail(0);
	End();

	// Send some more data to echo server
	Start("%s WriteSocket",pszMode);
	if ((iWritten = WriteSocket(iClientSock, szData, sizeof(szData), 0)) < 0)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	if (iMode == DATA_GRAM)
	{
		// Just check datagram count, no reading should take place
		Start("%s ReadSocket(pBuf,NET_FLG_PEEK|NET_FLG_NON_BLOCKING)==2", pszMode);
		memset(rgcReadBuf,0,sizeof(rgcReadBuf));
		if ((iRead = ReadSocketCount(iClientSock,iMode, rgcReadBuf,sizeof(rgcReadBuf),
								2,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)) == 2)
		{
			if (memcmp(szData,rgcReadBuf,iRead) == 0)
				Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
		}
		End();
	}

	// Just check count, no reading should take place
	// PCTCP seems to remember the count once it has been read and not
	// update it until it has been read. The following delay is used to
	// make sure the first read is OK
	Delay(2);		// PCTCP needs it
	iSize = iMode == STREAM ? 2 * sizeof(szData) : sizeof(szData);
	Start("%s ReadSocket(0,NET_FLG_PEEK)==%d", pszMode, iSize);
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, 0, sizeof(rgcReadBuf),
							iSize,NET_FLG_PEEK)) == iSize)
	{
		if (memcmp(szData,rgcReadBuf,iRead) == 0)
			Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	// Read back without de-queueing data
	Start("%s ReadSocket(pBuf,NET_FLG_PEEK)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, rgcReadBuf,sizeof(szData),sizeof(szData),NET_FLG_PEEK)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	if (iMode == DATA_GRAM)
	{
		// Just check datagram count, no reading should take place
		Start("%s ReadSocket(pBuf,NET_FLG_PEEK|NET_FLG_NON_BLOCKING)==2", pszMode);
		memset(rgcReadBuf,0,sizeof(rgcReadBuf));
		if ((iRead = ReadSocketCount(iClientSock,iMode,rgcReadBuf,sizeof(rgcReadBuf),
								2,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)) == 2)
		{
			if (memcmp(szData,rgcReadBuf,iRead) == 0)
				Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
		}
		End();
	}

	// Just check count, no reading should take place
	Start("%s ReadSocket(0,NET_FLG_PEEK)==%d", pszMode, iSize);
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, 0,
								sizeof(rgcReadBuf), iSize,NET_FLG_PEEK)) == iSize)
	{
		if (memcmp(szData,rgcReadBuf,iRead) == 0)
			Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	// Read data
	Start("%s ReadSocket(pBuf,0)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, rgcReadBuf,sizeof(szData),
												sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	if (iMode == DATA_GRAM)
	{
		// Just check datagram count, no reading should take place
		Start("%s ReadSocket(pBuf,NET_FLG_PEEK|NET_FLG_NON_BLOCKING)==1", pszMode);
		memset(rgcReadBuf,0,sizeof(rgcReadBuf));
		if ((iRead = ReadSocketCount(iClientSock,iMode,rgcReadBuf,sizeof(rgcReadBuf),
								1,NET_FLG_PEEK | NET_FLG_NON_BLOCKING)) == 1)
		{
			if (memcmp(szData,rgcReadBuf,iRead) == 0)
				Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
		}
		End();
	}

	// Just check count, no reading should take place
	Start("%s ReadSocket(0,NET_FLG_PEEK)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock, iMode, 0, sizeof(rgcReadBuf),
									sizeof(szData),NET_FLG_PEEK)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) == 0)
			Fail("Readback should NOT have worked:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

	// Read data
	Start("%s ReadSocket(pBuf,0)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock,iMode,rgcReadBuf,sizeof(rgcReadBuf),
												sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();

#ifdef FORCE_ERROR
	// Force an error
	Start("%s ReadSocket(pBuf,0)", pszMode);
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock,iMode, rgcReadBuf,sizeof(rgcReadBuf),sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	End();
#endif

	// next read should block
	Start("%s ReadSocket(pBuf,0)", pszMode);
	if ((iRead = ReadSocket(iClientSock, rgcReadBuf,sizeof(rgcReadBuf),0,0)) != -1)
		Fail("&unexpected data count %d", iRead);
	else if (iNetErrNo != ERR_WOULD_BLOCK)
		Fail(0);
	End();

	// Check for receive events
	Start("%s receive event",pszMode);
	if ((iResult = arsClientEvents[NET_AS_RCV].iEventCount) == 0)
		Fail("&count is 0, something expected");
	End();

	if (iMode == STREAM)
	{
		// Check for transmit events
		Start("%s transmit event",pszMode);
		if ((iResult = arsClientEvents[NET_AS_XMT].iEventCount) == 0)
			Fail("&count is 0, something expected");
		End();
	}

	if (iMode == STREAM)
	{
		iEventPosted = 0;
		Start("EofSocket()");
		if (EofSocket(iClientSock) < 0)
			Fail(0);
		End();
		Start("%s close event",pszMode);
		SetTimer(wTimeOut);
		while (!iEventPosted)
		{
			if (arsClientEvents[NET_AS_FCLOSE].iEventCount &&
								arsClientEvents[NET_AS_CLOSE].iEventCount)
				break;
			if (DiagWait())
			{
				iFail = 1;
				break;
			}
		}
		End();
	}
	else		// Datagram
	{
		sAddr.dwRemoteHost = dwServerIPAddress;
		sAddr.wRemotePort = 7;
		sAddr.wLocalPort = 0;
		Start("WriteToSocket()");
		if ((iWritten = WriteToSocket(iClientSock, szData, sizeof(szData), &sAddr,0)) < 0)
			Fail(0);
		else if (iWritten != sizeof(szData))
			Fail("&short write: %d expected %d",iWritten,sizeof(szData));
		End();
		SetTimer(wTimeOut);
		Start("ReadFromSocket()");
		while (1)
		{
			if (DiagWait())
			{
				iFail = 1;
				break;
			}
			sAddr.dwRemoteHost = 0;
			sAddr.wRemotePort = 0;
			sAddr.wLocalPort = 0;
			if ((iRead = ReadFromSocket(iClientSock,rgcReadBuf,
						sizeof(rgcReadBuf), &sAddr, NET_FLG_NON_BLOCKING)) < 0)
			{
				if (iNetErrNo != ERR_WOULD_BLOCK)
				{
					Fail(0);
					break;
				}
			}
			else {
				if (iRead != iWritten)
					Fail("Short read %d, expected %d",iRead,iWritten);
				if (sAddr.dwRemoteHost != dwServerIPAddress || sAddr.wRemotePort != 7)
				{
					strcpy(sz,IPAddressString(dwServerIPAddress));
					Fail("Received from %s:%u expected %s:7",
						IPAddressString(sAddr.dwRemoteHost),sAddr.wRemotePort,sz);
				}
				break;
			}
		}
		End();
	}

	// contrary to documentation the next test should pass
	Start("IsSocket on normal socket");
	if (IsSocket(iClientSock) < 0)
		Fail(0);
	End();

	Start("%s ReleaseSocket(client)",pszMode);
	if (ReleaseSocket(iClientSock) < 0)
		Fail(0);
	End();


	Start("IsSocket on released normal socket");
	if (IsSocket(iClientSock) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	if (dwServerIPAddress == dwMyIPAddress)
	{
		Start("%s ReleaseSocket(server)",pszMode);
		if (ReleaseSocket(iServerSock) < 0)
			Fail(0);
		End();
		if (!iServerType)
			StartListen(iMode, iServerType);
	}

	// test Dos Compatible sockets

	Start("GetDCSocket");
	if ((iClientSock = GetDCSocket()) < 0)
		Fail(0);
	End();

	Start("IsSocket on 0");
	if (IsSocket(0) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	Start("IsSocket");
	if (IsSocket(iClientSock) < 0)
		Fail(0);
	End();

	// test SelectSocket
	lMask = 1 << iClientSock;
	Start("SelectSocket");
	if (SelectSocket(32,&lIflags,&lOflags) != -1 || iNetErrNo != ERR_BAD_ARG)
		Fail(0);
	End();

	lIflags = lOflags = 0L;
	Start("SelectSocket");
	if (SelectSocket(31,&lIflags,&lOflags) < 0)
		Fail(0);
	else if (lIflags || lOflags)
		Fail("&returned lIflags %#lx lOflags %#lx (expected %#lx %#lx)",
			lIflags,lOflags,0l,0l);
	End();

	// make iClientSock non-blocking
	Start("%s SetSocketOption(NET_OPT_NON_BLOCKING)", pszMode);
	if (SetSocketOption(iClientSock, 0, NET_OPT_NON_BLOCKING, 1, 1) < 0)
		Fail(0);
	End();

	// make iClientSock wait for flush
	Start("%s SetSocketOption(NET_OPT_WAIT_FLUSH)", pszMode);
	if (SetSocketOption(iClientSock, 0, NET_OPT_WAIT_FLUSH, 1, 1) < 0)
		Fail(0);
	End();

	sClientAddr.wRemotePort = 7;
	sClientAddr.wLocalPort = 0;
	Start("ConnectSocket(%s)",pszMode);
	if (ConnectSocket(iClientSock, iMode, &sClientAddr) < 0)
		Fail(0);
	End();

	SetTimer(wTimeOut);
	lOflags = 0L;
	Start("%s SelectSocket output", pszMode);
	while (1)
	{
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
		if (SelectSocket(31,&lIflags,&lOflags) < 0)
		{
			Fail(0);
			break;
		}
		if (lOflags & lMask)
			break;
	}
	if (lIflags & lMask)
		Fail("&returned lIflags %#lx lOflags %#lx (expected %#lx %#lx)",
			lIflags,lOflags,0l,lMask);
	End();

	// Send some data to echo server
	Start("%s DC WriteSocket",pszMode);
	if ((iWritten = WriteSocket(iClientSock, szData, sizeof(szData), 0)) < 0)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	if (iMode == STREAM)
	{
		Start("Flush timeout");
		SetTimer(wTimeOutHalf);
		while (1)
		{
			if (DiagWait)
				break;
			if (SelectSocket(31,&lIflags,&lOflags) < 0)
			{
				Fail(0);
				break;
			}
			if (lIflags & lMask)
			{
				Fail(0);
				break;
			}
		}
		End();
		Start("FlushSocket");
		if (FlushSocket(iClientSock) < 0)
			Fail(0);
		End();
	}
	SetTimer(wTimeOut);
	Start("%s SelectSocket input", pszMode);
	while (1)
	{
		if (DiagWait())
		{
			iFail = 1;
			break;
		}
		if (SelectSocket(31,&lIflags,&lOflags) < 0)
		{
			Fail(0);
			break;
		}
		if (lIflags & lMask)
			break;
	}
	End();

	// a read should succeed
	Start("%s DC ReadSocket(0,NET_FLG_PEEK)",pszMode);
	if ((iRead = ReadSocket(iClientSock, 0, sizeof(rgcReadBuf),0,
											NET_FLG_PEEK)) <= 0)
		Fail(0);
	End();

	Start("%s DC ReadSocket(pBuf,0)==%d", pszMode, sizeof(szData));
	memset(rgcReadBuf,0,sizeof(rgcReadBuf));
	if ((iRead = ReadSocketCount(iClientSock,iMode, rgcReadBuf,sizeof(szData),
												sizeof(szData),0)) == sizeof(szData))
	{
		if (memcmp(szData,rgcReadBuf,iRead) != 0)
			Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
	}
	if (iRead == -1)
		bReadFailed = 1;
	End();

	Start("%s AbortSocket",pszMode);
	if (AbortSocket(iClientSock) < 0)
		Fail(0);
	End();

	if (dwServerIPAddress == dwMyIPAddress)
	{
		Start("%s ReleaseSocket(server)",pszMode);
		if (ReleaseSocket(iServerSock) < 0)
			Fail(0);
		End();
		if (!iServerType)
			StartListen(iMode, iServerType);
	}

	// make a new connection with -1 which will be blocking
	sClientAddr.wRemotePort = 7;
	sClientAddr.wLocalPort = 0;
	Start("ConnectSocket(%s)",pszMode);
	if ((iClientSock = ConnectSocket(-1, iMode, &sClientAddr)) < 0)
		Fail(0);
	End();

	// Write zero bytes
	Start("%s WriteSocket(0,blocking)",pszMode);
	if ((iWritten = WriteSocket(iClientSock, szData, 0, 0)) < 0)
		Fail(0);
	else if (iWritten != 0)
		Fail("&return: %d, expected %d",iWritten,0);
	End();


	// Send some data to echo server
	Start("%s WriteSocket(blocking)",pszMode);
	if ((iWritten = WriteSocket(iClientSock, szData, sizeof(szData), 0)) < 0)
		Fail(0);
	else if (iWritten != sizeof(szData))
		Fail("&short write: %d expected %d",iWritten,sizeof(szData));
	End();

	if (!bReadFailed)
	{
		Start("%s DC ReadSocket(blocking)==%d", pszMode, sizeof(szData));
		memset(rgcReadBuf,0,sizeof(rgcReadBuf));
		if ((iRead = ReadSocketCount(iClientSock,iMode, rgcReadBuf,sizeof(szData),
													sizeof(szData),0)) == sizeof(szData))
		{
			if (memcmp(szData,rgcReadBuf,iRead) != 0)
				Fail("Readback error:\n%.*s\n%.*s",iRead,szData,iRead,rgcReadBuf);
		}
		End();
	}
	else
		printf("%s DC ReadSocket(blocking) skipped\n", pszMode);


	Start("%s AbortDCSockets",pszMode);
	if (AbortDCSockets() < 0)
		Fail(0);
	End();


	Start("%s DC ReleaseSocket after AbortSocket",pszMode);
	if (ReleaseSocket(iClientSock) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	// test ReleaseDCSockets
	Start("GetDCSocket");
	if ((iClientSock = GetDCSocket()) < 0)
		Fail(0);
	End();

	Start("%s ReleaseDCSockets",pszMode);
	if (ReleaseDCSockets() < 0)
		Fail(0);
	End();

	Start("ReleaseSocket after ReleaseDCSockets");
	if (ReleaseSocket(iClientSock) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	// test ConvertDCSocket
	Start("GetDCSocket");
	if ((iClientSock = GetDCSocket()) < 0)
		Fail(0);
	End();

	Start("ConvertDCSocket");
	if ((iResult = ConvertDCSocket(iClientSock)) < 0)
		Fail(0);
	else if (iResult == iClientSock || iResult < 64)
		Fail("&%d converted to %d", iClientSock, iResult);
	End();

	Start("ReleaseSocket on converted from socket");
	if (ReleaseSocket(iClientSock) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	Start("ReleaseSocket on converted to socket");
	if (ReleaseSocket(iResult) < 0)
		Fail(0);
	End();

	Start("ReleaseSocket on converted from socket");
	if (ReleaseSocket(iClientSock) != -1 || iNetErrNo != ERR_NOT_NET_CONN)
		Fail(0);
	End();

	// Volume data transfers
	if (!bReadFailed)
	{
		for (i = 0;i < BUFLEN;++i)
			rgcWriteBuf[i] = i;
		Start("GetSocket(%s)", pszMode);
		if ((iClientSock = GetSocket()) < 0)
			Fail(0);
		End();
		SetupBulk(iMode,pszMode,iServerType);
		Start("Write/Read bulk, polling - %s",pszMode);
		while (bWriting || dwBufsRead != dwBufsWritten)
		{
			if (dwBufsWritten - dwBufsRead < WRITE_AHEAD && bWriting)
			{
				if (WriteBlock(iClientSock) < 0)
				{
					Fail("& write error");
					break;
				}
			}
			if ((i = ReadBlock(iClientSock)) < 0)
			{
				if (i == -2)
					Fail("& read error");
				else if (!iErrorAddress)
				{
					Fail(" Readback error,  written %lu read %lu offset %u good %lu bad %lu",
							dwBufsWritten,dwBufsRead + 1,iErrorAddress,
							dwBufsRead,*(DWORD *)rgcReadBuf);
					PrintReadBuffer();
				}
				else
				{
					Fail(" Readback error,  written %lu read %lu offset 0x%03x good 0x%02x bad 0x%02x",
							dwBufsWritten,dwBufsRead + 1,iErrorAddress,
							(BYTE)(iErrorAddress + *rgcReadBuf - 4),
							(BYTE)rgcReadBuf[iErrorAddress]);
					PrintReadBuffer();
				}
				break;
			}
			else if (i == 0)	// nothing read
			{
				if ((int)(wTimeRead - GetTicks()) <= 0)
				{
					Fail(" Read timeout, written %lu read %lu blocks",
												dwBufsWritten,dwBufsRead);
					break;
				}
			}
		}
		if ((wTime = GetTicks() - wTime) == 0)
			wTime = 1;
		sprintf(sz,"%.0f",(65536.0 * BUFLEN * dwBufsWritten) / (wTime * 3600.0));
		Seperate1000(sz);
		printf("%-9s Poll: %14s bytes/sec\n",pszMode,sz);
		End();
		Start("%s AbortSocket(client)",pszMode);
		if (AbortSocket(iClientSock) < 0)
			Fail(0);
		End();

		// Write/Read bulk, Notification driven
		Start("GetSocket(%s)", pszMode);
		if ((iClientSock = GetSocket()) < 0)
			Fail(0);
		End();
		// set ASYNC Handler on RECEIVE
		Start("%s SetAsyncNotification(NET_AS_RCV)",pszMode);
		if (SetAsyncNotification(iClientSock, NET_AS_RCV, AsyncNotificationHandler, dwClientHandlerAddr) == (int D_FAR *)-1)
		{
			Fail(0);
		}
		End();
		// set ASYNC Handler on XMIT
		Start("%s SetAsyncNotification(NET_AS_XMT)",pszMode);
		if (SetAsyncNotification(iClientSock, NET_AS_XMT, AsyncNotificationHandler, dwClientHandlerAddr) == (int D_FAR *)-1)
		{
			Fail(0);
		}
		End();
		SetupBulk(iMode,pszMode,iServerType);
		Start("Write/Read bulk, notification - %s",pszMode);
		// Start by writing first block
		if (/*iMode == DATA_GRAM && */WriteBlock(iClientSock) < 0)
		{
			Fail(" Initial write error");
			goto fail2;
		}
		IsSocket(iClientSock);
		while (bWriting || dwBufsRead != dwBufsWritten)
		{
			if (iBlockError < 0)
			{
				if (iBlockError == -1)
					Fail("& write error");
				else if (iBlockError == -2)
					Fail("& read error");
				else if (!iErrorAddress)
				{
					Fail(" Readback error,  written %lu read %lu offset %u good %lu bad %ld",
							dwBufsWritten,dwBufsRead + 1,iErrorAddress,
							dwBufsRead,*(DWORD *)rgcReadBuf);
					PrintReadBuffer();
				}
				else
				{
					Fail(" Readback error,  written %lu read %lu offset 0x%03x good 0x%02x bad 0x%02x",
							dwBufsWritten,dwBufsRead + 1,iErrorAddress,
							(BYTE)(iErrorAddress + *rgcReadBuf - 4),
							(BYTE)rgcReadBuf[iErrorAddress]);
					PrintReadBuffer();
				}
				break;
			}
			if ((int)(wTimeRead - GetTicks()) <= 0)
			{
				Fail(" Read timeout, written %lu read %lu blocks"" Notify r %u w %u",
											dwBufsWritten,dwBufsRead,wRead,wWrite);
				break;
			}
			if (kbhit())
			{
				WORD wKey = Getch();
				printf(" User interrupt, written %lu read %lu blocks\n"
						" Notify r %u w %u Target t %u Read t %u Cur t %u Writing %u\n",
						dwBufsWritten,dwBufsRead,wRead,wWrite,wTarget,wTimeRead,GetTicks(), bWriting);
				if (wKey == 0x1b) {
					Fail(0);
					break;
				}
			}
		}
		if ((wTime = GetTicks() - wTime) == 0)
			wTime = 1;
		sprintf(sz,"%.0f",(65536.0 * BUFLEN * dwBufsWritten) / (wTime * 3600.0));
		Seperate1000(sz);
		printf("%-9s Notify: %12s bytes/sec",pszMode,sz);
		printf(" Blocks %lu RN %u WN %u\n",dwBufsRead,wRead,wWrite);
		fail2:
		End();
		Start("%s AbortSocket(client)",pszMode);
		if (AbortSocket(iClientSock) < 0)
			Fail(0);
		End();

	}
	else
		printf("Volume traffic tests skipped\n");

	if (dwServerIPAddress == dwMyIPAddress)
	{
		Start("%s ReleaseSocket(server)",pszMode);
		if (ReleaseSocket(iServerSock) < 0)
			Fail(0);
		End();
		if (iServerType)
		{
			Start("%s ReleaseSocket(listen)",pszMode);
			if (ReleaseSocket(iListenSock) < 0)
				Fail(0);
			End();
		}
	}

#ifdef EXTREG_TEST
#ifdef __DJGPP__
	_ASM("movl %%eax, %0\n\t"
		  "movl %%ebx, %1\n\t"
		  "movl %%ecx, %2\n\t"
		  "movl %%edx, %3\n\t"
		  : "=g" (sExr.eax), "=g" (sExr.ebx),
			 "=g" (sExr.ecx), "=g" (sExr.edx)
		  :
		  : "memory"
		 );
#else
	// load Extended registers
	asm	.386;
	asm	mov dword ptr sExr.eax,eax;
	asm	mov dword ptr sExr.ebx,ebx;
	asm	mov dword ptr sExr.ecx,ecx;
	asm	mov dword ptr sExr.edx,edx;
	asm	.8086
#endif

	Start("Extended register save");
	iNetErrNo = 0;		// prevent confusing result
	if (((sExr.eax ^ sExs.eax) & 0xffff0000L) ||
		 ((sExr.ebx ^ sExs.ebx) & 0xffff0000L) ||
		 ((sExr.ecx ^ sExs.ecx) & 0xffff0000L) ||
		 ((sExr.edx ^ sExs.edx) & 0xffff0000L))
		Fail(0);
	End();
#endif

	// print event statistics if requested
	if (iVerbose & 4)
	{
		int i;

		printf("%s events:\n",pszMode);
		for (i = 0;i <= MAX_AS_EVENT + 1;i++)
		{
			printf("%-16.16s %4d %s\n",ppszEventMessage[i],
							arsClientEvents[i].iEventCount,
							i == NET_AS_ERROR ?
							GetErrorString(arsClientEvents[i].iEventError) : "");
		}
	}
	DPMIINFO_FREE(server_info);
	DPMIINFO_FREE(client_info);
	return 0;
}

void SetupBulk(int iMode, char *pszMode, int iServerType)
{
	if (dwServerIPAddress == dwMyIPAddress)
	{
		Start("%s ReleaseSocket(server)",pszMode);
		if (ReleaseSocket(iServerSock) < 0)
			Fail(0);
		End();
		if (!iServerType)
			StartListen(iMode, iServerType);
	}

	// Blocking connect
	sClientAddr.wRemotePort = 7;
	sClientAddr.wLocalPort = 0;
	Start("ConnectSocket(%s)",pszMode);
	if (ConnectSocket(iClientSock, iMode, &sClientAddr) < 0)
		Fail(0);
	End();
	// Write and read back data for 10 seconds
	wTime = GetTicks();
	while (wTime == GetTicks());
	wTime = GetTicks();
	wTarget = wTime + wBulkTicks;					// 10 seconds default
	wTimeRead = GetTicks() + wBulkTicks / 2;	// 5 seconds default
	dwBufsWritten = dwBufsRead = 0;
	iWriteSize = iReadSize = BUFLEN;
	iWriteOffset = iReadOffset = iBlockError = 0;
	bWriting = 1;
	wRead = wWrite = 0;
}

// Write block for bulk transfer
int WriteBlock(int iSock)
{
	int i,iWritten;
	char cCount;

	*(DWORD *)rgcWriteBuf = dwBufsWritten;
	for (i = 4, cCount = rgcWriteBuf[0];i < BUFLEN;++i,++cCount)
		rgcWriteBuf[i] = cCount;
	if ((iWritten = WriteSocket(iSock,rgcWriteBuf + iWriteOffset,
									iWriteSize, NET_FLG_NON_BLOCKING)) < 0)
	{
		if (iNetErrNo != ERR_WOULD_BLOCK)
		{
			return -1;
		}
	}
	else if (iWritten < iWriteSize)
	{
		if (iVerbose & 32)
			printf("Written %x of %x at %x, %02x - %02x\n",
				iWritten,iWriteSize,iWriteOffset,
				(BYTE)rgcWriteBuf[iWriteOffset],
				(BYTE)rgcWriteBuf[iWriteOffset + iWritten - 1]);
		iWriteOffset += iWritten;
		iWriteSize -= iWritten;
	}
	else
	{
		if (iVerbose & 32)
			printf("Written %x of %x at %x, %02x - %02x\n",
				iWritten,iWriteSize,iWriteOffset,
				(BYTE)rgcWriteBuf[iWriteOffset],
				(BYTE)rgcWriteBuf[iWriteOffset + iWritten - 1]);
		iWriteOffset = 0;
		iWriteSize = BUFLEN;
		dwBufsWritten++;
		if ((int)(wTarget - GetTicks()) <= 0)
			bWriting = 0;
	}
	return 0;
}

// Read block for bulk transfer
int ReadBlock(int iSock)
{
	int iRead, i;
	char cCount;

	if ((iRead = ReadSocket(iSock,rgcReadBuf + iReadOffset,
									iReadSize, 0, NET_FLG_NON_BLOCKING)) < 0)
	{
		if (iNetErrNo != ERR_WOULD_BLOCK)
		{
			return -2;
		}
		return 0;	// nothing read
	}
	else if (iRead < iReadSize)
	{
		if (iVerbose & 32)
			printf("Read %x of %x at %x, %02x-%02x\n",
				iRead,iReadSize,iReadOffset,
				(BYTE)rgcReadBuf[iReadOffset],
				(BYTE)rgcReadBuf[iReadOffset + iRead - 1]);
		iReadOffset += iRead;
		iReadSize -= iRead;
		wTimeRead = GetTicks() + 91;	// 5 seconds
	}
	else
	{
		if (iVerbose & 32)
			printf("Read %x of %x at %x, %02x-%02x\n",
				iRead,iReadSize,iReadOffset,
				(BYTE)rgcReadBuf[iReadOffset],
				(BYTE)rgcReadBuf[iReadOffset + iRead - 1]);
		iReadOffset = 0;
		iReadSize = BUFLEN;
		wTimeRead = GetTicks() + 91;	// 5 seconds
		if (*(DWORD *)rgcReadBuf != dwBufsRead)
		{
			iErrorAddress = 0;
			return -3;
		}
		for (i = 4,cCount = *rgcReadBuf;i < BUFLEN;++i,++cCount)
		{
			if (rgcReadBuf[i] != cCount)
			{
				iErrorAddress = i;
				return -3;
			}
		}
		dwBufsRead++;
	}
	return 1;
}


DWORD dwDummy;
// Client handler for notification driven bulk transfer
#ifdef __DJGPP__
 int D_FAR ClientHandler(int iSocketTmp, int iEventTmp, DWORD dwArgTmp)
#else
 int D_FAR ClientHandler(int iSocket, int iEvent, DWORD dwArg)
#endif
{
#ifdef __DJGPP__
	int iSocket;
	int iEvent;
	DWORD dwArg;

	iSocket = client_regs.x.bx;
	iEvent  = client_regs.x.cx;
	dwArg  = ((DWORD)client_regs.x.si << 16L) + client_regs.x.dx;
#endif

	switch (iEvent)
	{
	case NET_AS_RCV:
		++wRead;
		while ((iBlockError = ReadBlock(iSocket)) > 0)
			;
		goto write;
	case NET_AS_XMT:
		++wWrite;
		write:
		if (bWriting && dwBufsWritten - dwBufsRead < WRITE_AHEAD)
			iBlockError = WriteBlock(iSocket);
		break;
	default:
		dwDummy = dwArg;	// to suppress compiler warning
		break;
	}
	return 0;
}

// Seperate 1000's
void Seperate1000(char *psz)
{
	int i, iChars = strlen(psz);
	int iSpaces = (iChars - 1) / 3;

	psz[iChars + iSpaces] = 0;
	for (i = 0;iChars--;)
	{
		psz[iChars + iSpaces] = psz[iChars];
		if (++i == 3)
		{
			i = 0;
			--iSpaces;
			if (iChars + iSpaces > 0)
				psz[iChars + iSpaces] = ' ';
		}
	}
}

// Read expected number of bytes or until interrupted
int ReadSocketCount(int iSocket,int iMode,char *pBuf,int iBufLen,int iExpectedLen,
	WORD wFlags)
{
	int i, iRead = 0;
	char *pReadBuf = pBuf;
	NET_ADDR sAddr;
	char sz[40];

	SetTimer(wTimeOut);
	while (1)
	{
		if (DiagWait())
		{
			iFail = 1;
			if (iVerbose & 2)
				printf("Read %d\n",iRead);
			return -1;
		}
		memset(&sAddr,0,sizeof(NET_ADDR));
		if ((i = ReadSocket(iSocket,pReadBuf,iBufLen - iRead,&sAddr,wFlags)) < 0)
		{
			if (iNetErrNo != ERR_WOULD_BLOCK)
			{
				Fail(0);
				return -1;
			}
		}
		else if (i == 0 && iMode == STREAM) {
			Fail("Peer closed connection while reading %d byte(s)",
							iExpectedLen);
			return -1;
		}
		else {
			if (sAddr.dwRemoteHost != dwServerIPAddress || sAddr.wRemotePort != 7)
			{
				strcpy(sz,IPAddressString(dwServerIPAddress));
				Fail("Received from %s:%u expected %s:7",
					IPAddressString(sAddr.dwRemoteHost),sAddr.wRemotePort,sz);
				return -1;
			}
			if (wFlags & NET_FLG_PEEK)
				iRead = i;
			else
				iRead += i;
			if (iRead >= iExpectedLen)
				break;
			if (pReadBuf && !(wFlags & NET_FLG_PEEK))
				pReadBuf += i;
		}
	}
	if (iRead != iExpectedLen)
		Fail("&unexpected count %d, expected %d",iRead,iExpectedLen);
	return iRead;
}

/*
 Async notification Handler for client socket - called by Sockets.
 Do NOT use any functions making DOS calls, eg printf().
 */

#ifdef __DJGPP__
extern __dpmi_regs cb_regs;
#endif

int LOADDS D_FAR Handler(void)
{
	// count the event
#ifdef __DJGPP__

	iEventPosted = cb_regs.x.cx;
	iEventError  = cb_regs.x.dx;

#else
	_ASM
	{
		mov iEventPosted, cx;
		mov iEventError, dx;
	}
#endif

	// stay within limits of array
	if ((WORD)iEventPosted > MAX_AS_EVENT + 1)
		iEventPosted = MAX_AS_EVENT + 1;
	arsClientEvents[iEventPosted].iEventCount++;
	arsClientEvents[iEventPosted].iEventError = iEventError;

#ifdef EXTREG_TEST
#ifdef __DJGPP__
	cb_regs.d.eax += 0x10000l;
	cb_regs.d.ebx += 0x20000l;
	cb_regs.d.ecx += 0x40000l;
	cb_regs.d.edx += 0x80000l;
#else
	// damage the extended registers to test saving
	asm .386
	asm add eax,010000h;
	asm add ebx,020000h;
	asm add ecx,040000h;
	asm add edx,080000h;
	asm .8086
#endif
#endif

	return 0;
}

// Handler to test return value from SetAsyncNotification()

int LOADDS D_FAR Handler0(void)
{
	return Handler();
}

// Async notification Handler for server - called by Sockets.
// Do NOT use any functions making DOS calls, eg printf().

int iServerMode;

enum		// Server action event logging
{
	SA_NOTIFY,
	SA_READ,
	SA_WRITE,
	SA_WRITETO,
	SA_EOF,
	SA_ACCEPT,
	SA_FULL
};

char *ppszActionMessage[] =
{
	"Notify",
	"ReadSocket",
	"WriteSocket",
	"WriteToSocket",
	"EofSocket",
	"AcceptSocket",
	"Queue full",
};

typedef struct sEventQue
{
	int iSocket;
	int iAction;
	int iEvent;
	DWORD dwArg;
	int iError;
} EVENT_QUE;

#define EVENT_QUE_SIZE 40

EVENT_QUE arsEventQue[EVENT_QUE_SIZE];

int iEventQueIn;		// event queue input index
int iEventQueOut;		// event queue output index
int iEventQueCount;	// count of events

void QueEvent(int iSocket,int iAction,int iEvent,DWORD dwArg,int iError)
{
	EVENT_QUE *psEventQue;

	if (iEventQueCount == EVENT_QUE_SIZE - 1)
	{
		arsEventQue[iEventQueIn].iAction = SA_FULL;
		return;
	}
	iEventQueCount++;
	psEventQue = arsEventQue + iEventQueIn;
	psEventQue->iSocket = iSocket;
	psEventQue->iAction = iAction;
	psEventQue->iEvent = iEvent;
	psEventQue->dwArg = dwArg;
	psEventQue->iError = iError;
	if (++iEventQueIn == EVENT_QUE_SIZE)
		iEventQueIn = 0;
}

#ifdef __DJGPP__
 int D_FAR ServerHandler(int iSockTmp, int iEventTmp, DWORD dwArgTmp)
#else
 int D_FAR ServerHandler(int iSocket, int iEvent, DWORD dwArg)
#endif
{
#ifdef __DJGPP__
	int iSocket;
	int iEvent;
	DWORD dwArg;
#endif
	static char rgcBuf[2 * BUFLEN];
	static int iWriteCount,iWriteOffset;
	int iCount;
	NET_ADDR sAddr;

#ifdef __DJGPP__
	iSocket = server_regs.x.bx;
	iEvent  = server_regs.x.cx;
	dwArg  = ((DWORD)server_regs.x.si << 16L) + server_regs.x.dx;
#endif
	QueEvent(iSocket,SA_NOTIFY,iEvent,dwArg,0);
	switch (iEvent)
	{
	case NET_AS_ERROR:
	case NET_AS_CLOSE:
		break;
	case NET_AS_OPEN:
		if (bTestingAccept)
		{
			++bTestingAccept;
		}
		else if (iGlobalServerType)
		{
			if ((iServerSock = AcceptSocket(iSocket,STREAM, &sServerAddr)) < 0)
				QueEvent(iSocket,SA_ACCEPT,iEvent,dwArg,iNetErrNo);
		}
		iWriteCount = iWriteOffset = 0;
		break;
	case NET_AS_FCLOSE:		// peer closed connection
	case NET_AS_RCV:
	case NET_AS_XMT:
		//only echo data back
		if (iWriteCount)
		{
			write:
			if ((iCount = WriteSocket(iSocket, rgcBuf + iWriteOffset,iWriteCount, 0)) < 0)
			{
				QueEvent(iSocket,SA_WRITE,iEvent,dwArg,iNetErrNo);
				break;
			}
			else if (iCount < iWriteCount)
			{
				iWriteOffset += iCount;
				iWriteCount -= iCount;
				break;
			}
			iWriteOffset = iWriteCount = 0;
		}
		if ((iCount = ReadSocket(iSocket, rgcBuf, sizeof(rgcBuf), &sAddr, 0)) > 0)
		{
			if (iServerMode == STREAM)
			{
				iWriteCount = iCount;
				goto write;
			}
			else
			{
				if (WriteToSocket(iSocket, rgcBuf, iCount, &sAddr, 0) < 0)
				{
					QueEvent(iSocket,SA_WRITETO,iEvent,dwArg,iNetErrNo);
				}
			}
		}
		else if (iCount == 0 || iEvent == NET_AS_FCLOSE)
		{
			if (EofSocket(iSocket) < 0)
			{
				QueEvent(iSocket,SA_EOF,iEvent,dwArg,iNetErrNo);
			}
		}
		else
		{
			QueEvent(iSocket,SA_READ,iEvent,dwArg,iNetErrNo);
		}
		break;
	}

	return 0;
}

/*
Start a server socket listening on the port number contained in the
sServerAddr structure.
Set all the api callbacks concerning this socket to ServerHandler.

Returns:
	The newly created server socket.
*/
#define NUM_QUEUED	3		// number of Listen/Accept queued entries
int StartListen(int iMode, int iServerType)
{
	int retv = 0,i,ariClientSock[NUM_QUEUED + 1],ariServerSock[NUM_QUEUED + 1];
	WORD wLen;
	TCB sTcb;
	char *pszMode = iMode == STREAM ?
			(iServerType ? "STREAM Listen/Accept" : "STREAM Listen") : "DATA_GRAM";

	iServerMode = iMode;
	iGlobalServerType = iServerType;
	Start("%s server GetSocket",pszMode);
	if ((iServerSock = GetSocket()) < 0)
	{
		Fail(0);
		retv = -1;
	}
	End();
	// make iServerSock non-blocking
	Start("%s server SetSockOption(NET_OPT_NON_BLOCKING)",pszMode);
	if (SetSocketOption(iServerSock, 0, NET_OPT_NON_BLOCKING, 1, 1) < 0)
	{
		Fail(0);
		retv = -2;
	}
	End();


	// set ASYNC Handler on OPEN
	Start("%s server SetAsyncNotification(NET_AS_OPEN)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_OPEN, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -4;
	}
	End();
	// set ASYNC Handler on CLOSE
	Start("%s server SetAsyncNotification(NET_AS_CLOSE)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_CLOSE, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -5;
	}
	End();
	// set ASYNC Handler on FCLOSE
	Start("%s server SetAsyncNotification(NET_AS_FCLOSE)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_FCLOSE, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -6;
	}
	End();
	// set ASYNC Handler on ERROR
	Start("%s server SetAsyncNotification(NET_AS_ERROR)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_ERROR, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -7;
	}
	End();
	// set ASYNC Handler on RECEIVE
	Start("%s server SetAsyncNotification(NET_AS_RCV)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_RCV, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -8;
	}
	End();
	// set ASYNC Handler on XMIT
	Start("%s server SetAsyncNotification(NET_AS_XMT)",pszMode);
	if (SetAsyncNotification(iServerSock, NET_AS_XMT, AsyncNotificationHandler, dwServerHandlerAddr) == (int D_FAR *)-1)
	{
		Fail(0);
		retv = -9;
	}
	End();
	if (!iServerType)
	{
		Start("%s server ListenSocket",pszMode);
		sServerAddr.wRemotePort = 0;
		sServerAddr.dwRemoteHost = 0;
		sServerAddr.wLocalPort = 7;
		if (ListenSocket(iServerSock, iMode, &sServerAddr) < 0)
		{
			Fail(0);
			return -3;
		}
		End();
	}
	else
	{
		Start("%s server ListenAcceptSocket",pszMode);
		sServerAddr.wRemotePort = 0;
		sServerAddr.dwRemoteHost = 0;
		sServerAddr.wLocalPort = 7;
		if (ListenAcceptSocket(iServerSock, iMode, NUM_QUEUED, &sServerAddr) < 0)
		{
			Fail(0);
			return -3;
		}
		iListenSock = iServerSock;
		End();
		wLen = sizeof(TCB);
		Start("GetKernelInformation(Listen0,K_INF_TCP_CB)");
		if (GetKernelInformation(iListenSock, K_INF_TCP_CB, 0, (char *)&sTcb, &wLen) < 0)
			Fail(0);
		if (sTcb.state != LISTEN)
			Fail("TCP state is %u, expected %u",sTcb.state,LISTEN);
		if (sTcb.max_nacc != NUM_QUEUED)
			Fail("Max not accepted is %u, expected %u",sTcb.max_nacc,NUM_QUEUED);
		if (sTcb.num_nacc != 0)
			Fail("Number not accepted is %u, expected %u",sTcb.num_nacc,0);
		End();
		bTestingAccept = 1;
		for (i = 0;i <= NUM_QUEUED;++i)
		{
			Start("%s client %d GetSocket",pszMode,i);
			if ((ariClientSock[i] = GetSocket()) < 0)
				Fail(0);
			End();
			// set timeout
			Start("%s client %d SetSockOption(NET_OPT_TIMEOUT)",pszMode,i);
			if (SetSocketOption(ariClientSock[i], 0, NET_OPT_TIMEOUT, 200L, 4) < 0)
			{
				Fail(0);
				End();
				break;
			}
			End();
			Start("%s client %d ConnectSocket()",pszMode,i);
			sClientAddr.wLocalPort = 0;
			if (ConnectSocket(ariClientSock[i], iMode, &sClientAddr) < 0)
			{
				if (i < NUM_QUEUED || iNetErrNo != ERR_RESET)
					Fail(0);
			}
			else if (i == NUM_QUEUED)
				Fail(0);		// should have failed!
			End();
			if (i < NUM_QUEUED)
			{
				Start("%s connect %d NET_AS_OPEN event",pszMode,i);
				if (i + 2 != bTestingAccept)
					Fail("& actual %d, expected %d",bTestingAccept,i + 2);
				End();
				Start("GetKernelInformation(Listen after connect %u,K_INF_TCP_CB)",i + 1);
				if (GetKernelInformation(iListenSock, K_INF_TCP_CB, 0, (char *)&sTcb, &wLen) < 0)
					Fail(0);
				if (sTcb.state != LISTEN)
					Fail("TCP state is %u, expected %u",sTcb.state,LISTEN);
				if (sTcb.max_nacc != NUM_QUEUED)
					Fail("Max not accepted is %u, expected %u",sTcb.max_nacc,NUM_QUEUED);
				if (sTcb.num_nacc != i + 1)
					Fail("Number not accepted is %u, expected %u",sTcb.num_nacc,i + 1);
				End();
			}
			else {
				Start("%s connect %d ReleaseSocket()",pszMode,i);
				if (ReleaseSocket(ariClientSock[NUM_QUEUED]) < 0)
					Fail(0);
				End();
			}
		}
		for (i = 0;i <= NUM_QUEUED;++i)
		{
			Start("%s accept %d AcceptSocket()",pszMode,i);
			if ((ariServerSock[i] = AcceptSocket(iServerSock, iMode, &sServerAddr)) < 0)
			{
				if (i < NUM_QUEUED || iNetErrNo != ERR_WOULD_BLOCK)
					Fail(0);
			}
			else if (i == NUM_QUEUED)
			{
				Fail(0);		// should have failed!
				ReleaseSocket(ariServerSock[NUM_QUEUED]);
			}
			else
			{
				Start("GetKernelInformation(Listen after accept %u,K_INF_TCP_CB)",i + 1);
				if (GetKernelInformation(iListenSock, K_INF_TCP_CB, 0, (char *)&sTcb, &wLen) < 0)
					Fail(0);
				if (sTcb.state != LISTEN)
					Fail("TCP state is %u, expected %u",sTcb.state,LISTEN);
				if (sTcb.max_nacc != NUM_QUEUED)
					Fail("Max not accepted is %u, expected %u",sTcb.max_nacc,NUM_QUEUED);
				if (sTcb.num_nacc != NUM_QUEUED - i - 1)
					Fail("Number not accepted is %u, expected %u",sTcb.num_nacc,NUM_QUEUED - i - 1);
				End();
			}
			End();
		}
		for (i = 0;i < NUM_QUEUED;++i)
		{
			Start("%s connect %d ReleaseSocket()",pszMode,i);
			if (ReleaseSocket(ariClientSock[i]) < 0)
				Fail(0);
			End();
			Start("%s accept %d ReleaseSocket()",pszMode,i);
			if (ReleaseSocket(ariServerSock[i]) < 0)
				Fail(0);
			End();
		}
		bTestingAccept = 0;
	}
	return retv;
}

/*
Create a human understandable string from a Sockets error code.

Argument:
	uErrCode - The sockets error code.

Returns:
	A pointer to the (static) string representation of the error.
*/
char *GetErrorString(unsigned uErrCode)
{
	static char rgcUnk[30];
	static char *rgszErrs[] =
	{
		"NO_ERR",					// 0	// No error
		"ERR_IN_USE",				// 1	// A connection already exists
		"ERR_DOS",					// 2	// A DOS error occured
		"ERR_NO_MEM",				// 3	// No memory to perform function
		"ERR_NOT_NET_CONN",		// 4	// Connection does not exist
		"ERR_ILLEGAL_OP",			// 5	// Protocol or mode not supported
		"ERR_BAD_PACKET",
		"ERR_NO_HOST",				// 7	// No host address specified
		"ERR_CANT_OPEN",
		"ERR_NET_UNREACHABLE",
		"ERR_HOST_UNREACHABLE",
		"ERR_PROTOCOL_UNREACHABLE",
		"ERR_PORT_UNREACHABLE",
		"ERR_TIMEOUT",				// 13	// The function timed out
		"ERR_HOST_UNKNOWN",		// 14	// Unknown host has been specified
		"ERR_NO_SERVERS",
		"ERR_SERVER_ERR",
		"ERR_BAD_FORMAT",
		"ERR_BAD_ARG",				// 18	// Bad arguments
		"ERR_EOF",					// 19	// The connection has been closed by peer
		"ERR_RESET",				// 20	// The connection has been reset by peer
		"ERR_WOULD_BLOCK",		// 21	// Operation would block
		"ERR_UNBOUND",				// 22	// The descriptor has not been assigned yet
		"ERR_NO_SOCKET",			// 23	// No descriptor is available
		"ERR_BAD_SYS_CALL",		// 24	// Bad parameter in call
		"ERR_CANT_BROADCAST",
		"ERR_NOT_ESTAB",			// 26	// The connection has not been established
		"ERR_RE_ENTRY",			// 27	// The kernel is in use, try again later
	};

	if (uErrCode == ERR_API_NOT_LOADED)
		return "Sockets API not loaded";
	if ((uErrCode & 0xff) > ERR_RE_ENTRY)
	{
		sprintf(rgcUnk, "Unknown error 0x%04X", uErrCode);
		return rgcUnk;
	}
	return rgszErrs[uErrCode & 0xff];
}

char *IPAddressString(DWORD dwIPAddress)
{
	static char sz[20];

	sprintf(sz, "%u.%u.%u.%u",
			((BYTE *)&dwIPAddress)[0],((BYTE *)&dwIPAddress)[1],
			((BYTE *)&dwIPAddress)[2],((BYTE *)&dwIPAddress)[3]);
	return sz;
}

time_t dwTargetTime;

// Set the timer for a wait
void SetTimer(WORD wTimeOut)
{
	dwTargetTime = time(0) + wTimeOut;
}

// Wait for keyboard or timeout
int DiagWait(void)
{
	WORD wKey;

	if (bDebugServer)
	{
		while (_disable(),iEventQueCount)
		{
			EVENT_QUE *psEventQue = arsEventQue + iEventQueOut;
			int iSocket = psEventQue->iSocket;
			int iAction = psEventQue->iAction;
			int iEvent = psEventQue->iEvent;
			DWORD dwArg = psEventQue->dwArg;
			int iError = psEventQue->iError;

			--iEventQueCount;
			if (++iEventQueOut == EVENT_QUE_SIZE)
				iEventQueOut = 0;
			_enable();
			printf("Server socket %d %s %s arg %lu %s\n",iSocket,
				ppszActionMessage[iAction],
				ppszEventMessage[iEvent],dwArg,GetErrorString(iError));
		}
		_enable();
	}
	if (kbhit())
	{
		wKey = Getch();
		if (wKey == 0x1b)	// Escape
		{
			Fail("&- user abort");
			return 1;
		}
		printf("Waiting for %s\n", szDiagnostic);
	}
	if (dwTargetTime && (dwTargetTime < time(0)))
	{
		Fail("&- timeout");
		return 1;
	}
	return 0;
}

// Start a test
// Save test name for diagnostics

int Start(char *szFormat, ...)
{
	va_list ap;

	iFail = 0;
	if (bControlBreak)
	{
		printf("User abort\n");
		Exit(1);
	}
	iNumberTests++;
	va_start(ap, szFormat);
	vsprintf(szDiagnostic, szFormat, ap);
	va_end(ap);
	if (iVerbose & 1)
		printf("%s ",szDiagnostic);
	return 0;
}

// fail a test
int Fail (char *szFormat, ...)
{
	va_list ap;
	char szLine[400];

	if (iFail)
		return 0;	// fail once only
	++iFailures;
	iFail = 1;
	if (!(iVerbose & 1))
		printf("%s ",szDiagnostic);
	if (szFormat)
	{
		va_start(ap, szFormat);
		vsprintf(szLine, szFormat + (*szFormat == '&' ? 1 : 0), ap);
		va_end(ap);
		if (*szFormat == '&')
			printf("%s - failed %s\n", szLine, GetErrorString(iNetErrNo));
		else
			printf("- failed %s\n%s\n", GetErrorString(iNetErrNo),szLine);
	}
	else
		printf("- failed %s\n", GetErrorString(iNetErrNo));
	if (iFailAction)
	{
		if (iFailAction < 2)
		{
			dwTargetTime = 0;		// don't time out
			printf("Press Esc to continue\n");
			while (!DiagWait());
			iFail = 0;
		}
		else {
			Exit(1);
		}
	}
	return 0;
}

// End a test. Print message if enabled
void End(void)
{
#ifdef NULL_POINTER_TEST
	if (memcmp(arcNullData,(char *)0,sizeof(arcNullData)) != 0)
	{
		printf("Null pointer in %s\n",szDiagnostic);
		dwTargetTime = 0;		// don't time out
		printf("Press Esc to continue\n");
		while (!DiagWait());
		iFail = 0;
	}
#endif
	if ((iVerbose & 1) && !iFail)
	{
		printf("- Passed\n");
	}
}

int BadRequest(void)
{
	union REGS r;
	extern int iSocketsApiInterrupt;

	r.h.ah = 0x70;
	int86(iSocketsApiInterrupt,&r,&r);
	if (r.x.cflag & 1) {
		iNetErrNo = r.h.al;
		iNetSubErrNo = r.h.ah;
		return -1;
	}
	return r.x.ax;
}

/* Delay for iTicks clock ticks */
WORD GetTicks(void)
{
   return GET_TICKS;
}

void Delay(WORD wTicks)
{
	WORD wTarget = GetTicks() + wTicks + 1;

	while ((int)(wTarget - GetTicks()) > 0) {
		/*
		if (kbhit()) {
			printf("Target = %x, Current = %x, diff = %d\n",
					wTarget,GetTicks(),(int)(wTarget - GetTicks()));
			if (getch() == '\x1b')
				return;
		}
		*/
	}
}
