//
//   $Id: PDTEST.C 1.4 2003/07/23 17:11:55Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/* packet driver test */

/* Updated @ Date 17 February 1993, 17:2 by Gerrit van Niekerk
*/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include "inetdefs.h"

#ifdef DL
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dosver.h"
#include "fatworld.h"
#endif

#define DRIVER_INFO		1
#define ACCESS_TYPE		2
#define RELEASE_TYPE		3
#define SEND_PKT		4
#define TERMINATE		5
#define GET_ADDRESS		6
#define RESET_INTERFACE         7
#define GET_PARAMETERS          10	/* high-performance */
#define AS_SEND_PKT		11	/* high-performance */
#define SET_RCV_MODE		20	/* extended */
#define GET_RCV_MODE		21	/* extended */
#define SET_MULTICAST_LIST	22	/* extended */
#define GET_MULTICAST_LIST	23	/* extended */
#define GET_STATISTICS          24	/* extended */
#define SET_ADDRESS		25	/* extended */

/* Error codes */

#define BAD_HANDLE      1	/* Invalid handle number */
#define NO_CLASS	2	/* No interfaces of specified class found */
#define NO_TYPE		3	/* No interfaces of specified type found */
#define NO_NUMBER	4	/* No interfaces of specified number found */
#define BAD_TYPE	5	/* Bad packet type specified */
#define NO_MULTICAST	6	/* This interface does not support multicast */
#define CANT_TERMINATE	7	/* This packet driver cannot terminate */
#define BAD_MODE	8	/* An invalid receiver mode was specified */
#define NO_SPACE	9	/* Failed because of insufficient space */
#define TYPE_INUSE      10	/* The type had previously been accessed */
#define BAD_COMMAND	11	/* The command was out of range */
#define CANT_SEND	12	/* The packet couldn't be sent */
#define CANT_SET	13	/* Hardware address couldn't be changed */
#define BAD_ADDRESS	14	/* Hardware address has bad length or format */
#define CANT_RESET      15	/* Couldn't reset interface */

typedef struct pdpar {			/* Packet Driver parameters */
	unsigned char	major_rev;	/* Revision of Packet Driver spec */
	unsigned char	minor_rev;	/*  this driver conforms to. */
	unsigned char	length;         /* Length of structure in bytes */
	unsigned char	addr_len;	/* Length of a MAC-layer address */
	unsigned short	mtu;		/* MTU, including MAC headers */
	unsigned short	multicast_aval; /* Buffer size for multicast addr */
	unsigned short	rcv_bufs;	/* (# of back-to-back MTU rcvs) - 1 */
	unsigned short	xmt_bufs;	/* (# of successive xmits) - 1 */
	unsigned short	int_num;	/* Interrupt # to hook for post-EOI
					   processing, 0 == none */
} PDPAR;

typedef struct pdstat {
	unsigned long	packets_in;	/* Totals across all handles */
	unsigned long	packets_out;
	unsigned long	bytes_in;	/* Including MAC headers */
	unsigned long	bytes_out;
	unsigned long	errors_in;	/* Totals across all error types */
	unsigned long	errors_out;
	unsigned long	packets_lost;	/* No buffer from receiver(), card */
					/*  out of resources, etc. */
} PDSTAT;

typedef char far * fptr;

int test(void);
int check_error(union REGS regs,char operation);

char access;
int handle;
int drint;

main()
{
	int i;
	char far *id;
	static char pid[] = "PKT DRVR";
	char found = 0;

#ifdef DL
	printf("Datalight Sockets Packet Driver Test ");
	if (dlIsFat32World())
		printf("%s ", SHORTVERSION);
	else
		printf("%s ", SHORTVERSION622);
	printf("%s\n%s\nPortions copyright (c) GPvNO 2003\n%s\n\n",
		BUILDVERSION, COPYRIGHT, _BUILD_NUMBER);
#endif

	for (drint = 0x60;drint <= 0x80;++drint) {
		id = *(fptr far *)MK_FP(0,4 * drint);
		for (i = 0;i < 9;++i) {
			if (id[i + 3] != pid[i])
				goto notfound;
		}
		found = 1;
		printf("Packet driver found at 0x%02x\n",drint);
		test();
		notfound:;
	}
	if (!found)
		printf("No packet driver found\n");
	return 0;
}

#pragma argsused
void interrupt receiver(
	unsigned bp,
	unsigned di,unsigned si,
	unsigned ds,unsigned es,
	unsigned dx,unsigned cx,
	unsigned bx,unsigned ax)

{
	es = di = 0;		/* drop all packets */
}

int test(void)
{
	union REGS regs;
	struct SREGS segregs;
	char functionality,class;
	char far *id;
	int i,addrlen = 6;
	BYTE addr[20];
	PDPAR far *ppdp;
	PDSTAT far *ppds;
	char dum_type[2] = {0xff,0xfe};

	regs.x.ax = (DRIVER_INFO << 8) + 0xff;
	int86x(drint,&regs,&regs,&segregs);
	check_error(regs,DRIVER_INFO);
	id = MK_FP(segregs.ds,regs.x.si);
	printf("Version %d, class %d, type %d, number %d,"
		" functionality %d\nName: ",
		regs.x.bx,regs.h.ch,regs.x.dx,regs.h.cl,regs.h.al);
	functionality = regs.h.al;
	class = regs.h.ch;
	while (*id)
		printf("%c",*id++);
	printf("\n");
	regs.x.ax = ACCESS_TYPE << 8;
	regs.h.al = regs.h.ch;
	regs.x.bx = regs.x.dx;
	regs.h.dl = regs.h.cl;
	regs.x.cx = 2;			/* typelen */
	regs.x.si = FP_OFF(dum_type);
	segregs.ds = FP_SEG(dum_type);
	segregs.es = FP_SEG(receiver);
	regs.x.di = FP_OFF(receiver);
	int86x(drint,&regs,&regs,&segregs);
	check_error(regs,ACCESS_TYPE);
	access = 1;
	handle = regs.x.ax;
	if (functionality == 5 || functionality == 6) {	/* high performance */
		printf("High performance driver\n");
		regs.x.ax = GET_PARAMETERS << 8;
		int86x(drint,&regs,&regs,&segregs);
		check_error(regs,GET_PARAMETERS);
		ppdp = MK_FP(segregs.es,regs.x.di);
		printf("Rev %u.%02u par_len %u add_len %u "
			"mtu %u multicast_buf %u\n",
			ppdp->major_rev,ppdp->minor_rev,ppdp->length,
			ppdp->addr_len,ppdp->mtu,ppdp->multicast_aval);
		printf("Rcv_bufs %u xmt_bufs %u int_num 0x%x\n",
			ppdp->rcv_bufs,ppdp->xmt_bufs,ppdp->int_num);
		addrlen = ppdp->addr_len;
	}
	for (i = 0;i < addrlen;++i)
		addr[i] = 0;
	if (class != 6) {
		regs.x.ax = GET_ADDRESS << 8;
		regs.x.bx = handle;
		regs.x.cx = addrlen;
		segregs.es = FP_SEG(addr);
		regs.x.di = FP_OFF(addr);
		int86x(drint,&regs,&regs,&segregs);
		check_error(regs,GET_ADDRESS);
		printf("Address: ");
		for (i = 0;i < addrlen;++i)
			printf("%02x%s",addr[i],i == addrlen - 1 ? "\n" : ":");
	}
	if (functionality == 2 || functionality == 6) {	/* extended */
		printf("Extended driver\n");
		regs.x.ax = GET_STATISTICS << 8;
		regs.x.bx = handle;
		int86x(drint,&regs,&regs,&segregs);
		check_error(regs,GET_STATISTICS);
		ppds = MK_FP(segregs.ds,regs.x.si);
		printf("Packets:in %lu out %lu bytes:in %lu out %lu\n",
			ppds->packets_in,ppds->packets_out,ppds->bytes_in,
							ppds->bytes_out);
		printf("Errors:in %lu out %lu packets lost %lu\n",
			ppds->errors_in,ppds->errors_out,ppds->packets_lost);
	}
	regs.h.ah = RELEASE_TYPE;
	regs.x.bx = handle;
	int86(drint,&regs,&regs);
	access = 0;
	check_error(regs,RELEASE_TYPE);
	return 0;
}

char *ops[] = {
	"INVALID_0",		/* 0 */
	"DRIVER_INFO",		/* 1 */
	"ACCESS_TYPE",		/* 2 */
	"RELEASE_TYPE",		/* 3 */
	"SEND_PKT",		/* 4 */
	"TERMINATE",		/* 5 */
	"GET_ADDRESS",		/* 6 */
	"RESET_INTERFACE",	/* 7 */
	"INVALID_8",
	"INVALID_9",
	"GET_PARAMETERS",	/* 10 */
	"AS_SEND_PKT",		/* 11 */
	"INVALID_12",
	"INVALID_13",
	"INVALID_14",
	"INVALID_15",
	"INVALID_16",
	"INVALID_17",
	"INVALID_18",
	"INVALID_19",
	"SET_RCV_MODE",		/* 20 */
	"GET_RCV_MODE",		/* 21 */
	"SET_MULTICAST_LIST",	/* 22 */
	"GET_MULTICAST_LIST",	/* 23 */
	"GET_STATISTICS",	/* 24 */
	"SET_ADDRESS",		/* 25 */
	"INVALID_26","","","",
};

/* Error codes */

char *ers[] = {
	"No error",
	"Invalid handle number",
	"No interfaces of specified class found",
	"No interfaces of specified type found",
	"No interfaces of specified number found",
	"Bad packet type specified",
	"This interface does not support multicast",
	"This packet driver cannot terminate",
	"An invalid receiver mode was specified",
	"Failed because of insufficient space",
	"The type had previously been accessed",
	"The command was out of range",
	"The packet couldn't be sent",
	"Hardware address couldn't be changed",
	"Hardware address has bad length or format",
	"Couldn't reset interface",
	"","","","","","",
};

int check_error(union REGS regs,char operation)
{
	if (regs.x.flags & 1) {
		printf("Error on %s call: %s\n",ops[operation],ers[regs.h.dh]);
		if (access) {
			regs.h.ah = RELEASE_TYPE;
			regs.x.bx = handle;
			int86(drint,&regs,&regs);
		}
		exit(1);
	}
	return 0;
}
