//
//   $Id: LRUCACHE.C 1.8 2002/10/16 21:38:52Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	LRUCACHE.C
	Created 06-28-2001  by Keith Garvin

	Copyright (c) 2001-2002 Datalight, Inc.
	All rights reserved.
	Confidential and Proprietary

	LRUCACHE provides a Least Recently Used cache.  A static array
	array of entiries is used to keep track of the most recently
	read data and discarding the least recently used data.

	Least Recent +---+ (Always index zero)
	             |___|
	 Most Recent |___| (uNumberOfEntries - 1)
	             |___| (uNumberOfEntries)
	             |   |
	             +---+ (uMaxNumberOfEntries)

	Public functions provided by this module:

		InitializeLruCache()
		DeinitializeLruCache()
		GetLeastRecentLruCacheEntry()
		GetMostRecentLruCacheEntry()
		UpdateCacheEntryToMru()

*/

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "stdc.h"

#include "lrucache.h"

#include "lrucache.pt"


/* Our cache pointer */
LRU_ENTRY far * lpuCacheArray = NULL;
static unsigned uNumberOfEntries = 0;
static unsigned uMaxNumberOfEntries = 0;


/*
	Allocate the cache given the size of the cache in bytes
	to cache.  Initialize the cache to a known and invalid
	state.
*/
bool PASCAL InitializeLruCache(unsigned uNumberOfCacheEntries)
{
	unsigned uSegment;
	unsigned uSizeOfLruInParas;

	/*
		06/29/2001 pb Added a check to make sure we weren't called with
		a request for zero entries.
	*/
	if(uNumberOfCacheEntries == 0)
	{
		return FALSE;
	}

	/* Determine the number of paragraphs we need and round up */
	uSizeOfLruInParas = uNumberOfCacheEntries * sizeof(LRU_ENTRY);
	uSizeOfLruInParas = (uSizeOfLruInParas + 0x0F) / 0x10;
	if(!dlAllocMem(uSizeOfLruInParas, &uSegment))
	{
		return FALSE;
	}
	if(uSegment == 0)
	{
		return FALSE;
	}
	lpuCacheArray = (LRU_ENTRY far *) MK_FP(uSegment, 0);


	ASSERT( lpuCacheArray );
	/* Initialize the cache to a invalid state */
	_fmemset(&lpuCacheArray[0], 0,
			(uNumberOfCacheEntries * sizeof(LRU_ENTRY)) );

	uMaxNumberOfEntries = uNumberOfCacheEntries;
	uNumberOfEntries = 0;
	return TRUE;
}

/*
	Allocate the cache given the size of the cache in bytes
	to cache.  Initialize the cache to a known and invalid
	state.
*/
void PASCAL DeinitializeLruCache(void)
{
	/* Deallocate the array and clear our internal variables */
	ASSERT( lpuCacheArray );
	dlFreeMem(FP_SEG(lpuCacheArray));
	lpuCacheArray = NULL;
	uMaxNumberOfEntries = 0;
	uNumberOfEntries = 0;
}


/*
	Allow the caller to parse through the cache from
	LRU entry up to the MRU entries.
*/
bool PASCAL GetLruEntry(unsigned uLruIndex, LRU_ENTRY * plruEntry)
{
	ASSERT(uLruIndex < uMaxNumberOfEntries);

	/* Dont return out-fo-bounds data */
	if(uLruIndex >= uNumberOfEntries)
	{
		return FALSE;
	}

	*plruEntry = lpuCacheArray[uLruIndex];
	return TRUE;
}


/*
	Simply snatch up the least recent entry from the cache.
	Returns FALSE if the cache is empty
*/
bool PASCAL GetLeastRecentLruCacheEntry(LRU_ENTRY * pCacheEntry)
{
	ASSERT( pCacheEntry );

	/* Can't do it if the cache is empty */
	if(uNumberOfEntries == 0)
		return FALSE;

	*pCacheEntry = lpuCacheArray[0];
	return TRUE;
}


/*
	Get the most recently used cache entry from the list.
*/
bool PASCAL GetMostRecentLruCacheEntry(LRU_ENTRY * pCacheEntry)
{
	ASSERT( pCacheEntry );

	/* Can't do it if the cache is empty */
	if(uNumberOfEntries == 0)
		return FALSE;

	ASSERT( uNumberOfEntries <= uMaxNumberOfEntries );
	*pCacheEntry = lpuCacheArray[uNumberOfEntries - 1];
	return TRUE;
}

/*
	Find and add a given entry in the cache.  If found, it is
	updated to be the most recently used, otherwise, failure
	is returned to the caller.
*/
bool PASCAL UpdateCacheEntryToMru(LRU_ENTRY cacheEntry)
{
	unsigned uMoveLength;
	unsigned uIndex = 0;
	bool fFound;

	/* Parse the list from most recent to least recent */
	fFound = FALSE;
	uIndex = uNumberOfEntries;
	while(uIndex)
	{
		uIndex--;

		/* Search for each entry */
		if(lpuCacheArray[uIndex] == cacheEntry)
		{
			/*
				We found the entry in the list, make room at the
				end of the list to make it the new MRU.
			*/
			uMoveLength = ((uNumberOfEntries - 1) - uIndex)
					* sizeof(LRU_ENTRY);

			ASSERT( uMoveLength < (uMaxNumberOfEntries * sizeof(LRU_ENTRY)) );

			_fmemmove(&lpuCacheArray[uIndex], &lpuCacheArray[uIndex + 1],
					uMoveLength);

			/* Found the entry, uIndex points to it */
			fFound = TRUE;
			break;
		}
	}


	/* Add the entry to the end of the list */
	if(!fFound)
	{
		/*
			This case, we are adding an entry that did not exist
			previously and there is room in the cache for it.
		*/

		//  If we are trying to add an entry in the list and
		//  the list is full there is a problem.  Not the
		//
		ASSERT( uNumberOfEntries < uMaxNumberOfEntries );
		uNumberOfEntries++;
	}

	ASSERT( uNumberOfEntries );
	/* Add the entry to the last entry of the list */
	lpuCacheArray[uNumberOfEntries - 1] = cacheEntry;

	return fFound;
}



#if TESTING

#define TESTX(x) if(!(x)) { \
			printf("Failed file: %s line: %u\n", \
			__FILE__, __LINE__); return FALSE; }

bool PASCAL TestLruCache(void)
{
 #define MAGIC_NUMBER 0X0BAD

	LRU_ENTRY cacheEntry;

	/* Init the LRU */
	TESTX( InitializeLruCache(5) );

	/*
		The following two calls should fail since there is
		nothing there and the cache entry should not be changed.
	*/
	cacheEntry = MAGIC_NUMBER;
	TESTX( !GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == MAGIC_NUMBER);
	TESTX( !GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == MAGIC_NUMBER);

	/* None of the following should be in the list */
	TESTX( !UpdateCacheEntryToMru(100) );
	TESTX( !UpdateCacheEntryToMru(200) );
	TESTX( !UpdateCacheEntryToMru(300) );
	TESTX( !UpdateCacheEntryToMru(400) );
	TESTX( !UpdateCacheEntryToMru(500) );

	/* LRU -----> MRU is: 100, 200, 300, 400, 500 */
	GetLruEntry(0, &cacheEntry);
	TESTX(cacheEntry == 100);
	GetLruEntry(1, &cacheEntry);
	TESTX(cacheEntry == 200);
	GetLruEntry(2, &cacheEntry);
	TESTX(cacheEntry == 300);
	GetLruEntry(3, &cacheEntry);
	TESTX(cacheEntry == 400);
	GetLruEntry(4, &cacheEntry);
	TESTX(cacheEntry == 500);

	/* Now the cache should have 100 as the LRU and 500 as the MRU */
	TESTX( GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 100);
	TESTX( GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 500);

	/* These should be in the list */
	TESTX( UpdateCacheEntryToMru(400) );
	/* LRU -----> MRU is: 100, 200, 300, 500, 400 */

	TESTX( UpdateCacheEntryToMru(300) );
	/* LRU -----> MRU is: 100, 200, 500, 400, 300 */

	TESTX( UpdateCacheEntryToMru(200) );
	/* LRU -----> MRU is: 100, 500, 400, 300, 200 */

	TESTX( UpdateCacheEntryToMru(100) );
	/* LRU -----> MRU is: 500, 400, 300, 200, 100 */

	TESTX( UpdateCacheEntryToMru(200) );
	/* LRU -----> MRU is: 500, 400, 300, 100, 200 */

	/* Now 200 should be the MRU and 500 is the LRU */
	TESTX( GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 200);
	TESTX( GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 500);

	TESTX( UpdateCacheEntryToMru(500));
	/* LRU -----> MRU is: 400, 300, 100, 200, 500 */

	/* Now 500 should be the MRU and 400 is the LRU */
	TESTX( GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 500);
	TESTX( GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 400);


	/* Finally fill the cache and make sure we lose the previous LRU */
	TESTX( !UpdateCacheEntryToMru(600));
	/* LRU -----> MRU is: 300, 100, 200, 500, 600 */

	TESTX( GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 600);
	TESTX( GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 300);


	/* Again: fill cache & make sure we lose the previous LRU */
	TESTX( !UpdateCacheEntryToMru(700));

	/* LRU -----> MRU is: 100, 200, 500, 600, 700 */
	GetLruEntry(0, &cacheEntry);
	TESTX(cacheEntry == 100);
	GetLruEntry(1, &cacheEntry);
	TESTX(cacheEntry == 200);
	GetLruEntry(2, &cacheEntry);
	TESTX(cacheEntry == 500);
	GetLruEntry(3, &cacheEntry);
	TESTX(cacheEntry == 600);
	GetLruEntry(4, &cacheEntry);
	TESTX(cacheEntry == 700);

	TESTX( GetMostRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 700);
	TESTX( GetLeastRecentLruCacheEntry(&cacheEntry) );
	TESTX(cacheEntry == 100);

	/* Clear the cache */
	DeinitializeLruCache();

	/*
		06/29/2001 pb Added the following tests
		Now we will test more extreme conditions
	*/


	/* Initialize the Cache with zero entries. */
	TESTX( !InitializeLruCache(0) );

	/* Initialize a Cache with only one entry */
	TESTX( InitializeLruCache(1) );

	/* We should fail this call because there should be no room for this */
	TESTX( !UpdateCacheEntryToMru(100) );
	TESTX( !UpdateCacheEntryToMru(200) );

	/* Check for the last entry entered (should be the only one) */
	TESTX( GetMostRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 200 );
	TESTX( GetLeastRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 200 );

	/* Clear the cache */
	DeinitializeLruCache();

	/* Initialize the Cache with a large amount of entries */
	TESTX( InitializeLruCache(50) );

	/* These should not be in the list yet as it should be empty */
	for( cacheEntry = 101; cacheEntry <= 150; cacheEntry++ )
	{
		TESTX( !UpdateCacheEntryToMru(cacheEntry) );
	}

	/*
	   Make sure that the only entry in the list appears as both
	   the most and least recently used.
	*/
	TESTX( GetMostRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 150 );
	TESTX( GetLeastRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 101 );

	/* Add an entry and make sure that the first one drops off */
	TESTX( !UpdateCacheEntryToMru(500) );
	TESTX( GetMostRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 500 );
	TESTX( GetLeastRecentLruCacheEntry( &cacheEntry ) );
	TESTX( cacheEntry == 102 );

	/* Clear the cache */
	DeinitializeLruCache();

	/* End pb's additions */

	/* If the execution gets to this point, the tests passed */
	return TRUE;
}

#endif


