//
//   $Id: compars2.c 1.4 2003/02/18 15:57:48Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  COMPARSE.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  09-28-2001  bt  Created
//


//  Borland libraries
//
#include <stdio.h>
#include <string.h>
#include <mem.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlapp.h"
#include "windows.h"
#include "fatworld.h"
#include "dosver.h"
#include "bldver.h"


//  Local includes
//
#include "comparse.h"
#include "misery.h"


//  Prototypes
//
#include "compars2.pt"
#include "cache.pt"
void exit(int);

//  Variables
//
bool fIsFat32World;
bool fSwitches[MAX_BOOLEAN_SWITCHES];
int iInputHandle;
int iOutputHandle;
uchar ucMaxDrive;
ulong ulMaxSector;
unsigned uMaxSectorCount;


void TestSectors( bool fIsRead, uchar ucDrive, ulong ulSector,
		unsigned uSectorCount, bool fWriteCached );


//  Program starts here
//
void TestCommOutput( char * szInputFileName )
{
	int iActionTaken;
	bool fSuccessful;
	char * szInputLine;
	bool fIsRead;
	uchar ucDrive;
	ulong ulSector;
	unsigned uSectorCount;
	ulong ulLineCount = 0;


	InvalidateWriteCountArray();


	//  Open the file to be parsed
	//
	fSuccessful = SmartCreateOpenFile( OPEN_ACCESS_READONLY, A_NORMAL,
			FILE_OPEN, szInputFileName, &iInputHandle, &iActionTaken);
	if( !fSuccessful )
	{
		printf( "Unable to open file: %s\n", szInputFileName );
		exit(1);
	}
	printf("Testing COMM output file %s\n", szInputFileName );


	while( TRUE )
	{
		//  Get the next line from the input file
		//
		szInputLine = GetSourceLine();
//		printf( "%s\n", szInputLine );
		if( *szInputLine == 0 )
		{
			break;
		}


		//  Parse the line
		//
		fSuccessful = ProcessLine( szInputLine, &fIsRead, &ucDrive, &ulSector,
				&uSectorCount );
		if( !fSuccessful )
		{
			break;
		}
		ulLineCount++;
		printf("\rline: %lu", ulLineCount );


//  Debug code used to stop to set a breakpoint on a specific line number
//
#if 0
		if( (ulLineCount == 27343L) || (ulLineCount == 21582L) )
		{
			uchar ucBreakHere;
			ucBreakHere = 0;
		}
#endif


		//  Test this request
		//
		TestSectors( fIsRead, ucDrive, ulSector, uSectorCount, TRUE );

	}


	dlCloseFile( iInputHandle );


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		FlushDiskCache( ucDrive );
	}
	ValidateUncommittedSectors();


	for( ucDrive = 0; ucDrive < MAX_DRIVES; ucDrive++ )
	{
		InvalidateDiskCache( ucDrive );
	}
	InvalidateWriteCountArray();


	printf("\nTesting COMM output file %s  PASSED\n", szInputFileName );
}


bool ProcessLine( char * szLine, bool * pfIsRead, uchar * pucDrive,
		ulong * pulSector, unsigned * puSectorCount )
{
	int iResult;
	ulong ulValue;


	//  Set the read write flag
	//
	if( (szLine[0] != 'C') && (szLine[0] != 'H') )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine++;
	if( szLine[0] == 'R' )
	{
		*pfIsRead = TRUE;
	}
	else if( szLine[0] == 'W' )
	{
		*pfIsRead = FALSE;
	}
	else
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine++;
	szLine = LFNSkipWhite( szLine );


	//  Skip the drive label
	//
	iResult = memcmp( szLine, "Drv: ", 5 );
	if( iResult != 0 )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine += 5;
	szLine = GetHexValue( szLine, &ulValue );
	if( *szLine != 'h' )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine++;
	szLine = LFNSkipWhite( szLine );
	if( ulValue > 255 )
	{
		printf("Error processing line\n");
		exit(1);
	}
	*pucDrive = (uchar)ulValue;
	if( *pucDrive > ucMaxDrive )
	{
		ucMaxDrive = *pucDrive;
	}


	//  Skip the sector label
	//
	iResult = memcmp( szLine, "Sec: ", 5 );
	if( iResult != 0 )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine += 5;
	szLine = GetHexValue( szLine, &ulValue );
	if( *szLine != 'h' )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine++;
	szLine = LFNSkipWhite( szLine );
	*pulSector = ulValue;
	if( *pulSector > ulMaxSector )
	{
		ulMaxSector = *pulSector;
	}


	//  Skip the count label
	//
	iResult = memcmp( szLine, "Cnt: ", 5 );
	if( iResult != 0 )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine += 5;
	szLine = GetHexValue( szLine, &ulValue );
	if( *szLine != 'h' )
	{
		printf("Error processing line\n");
		exit(1);
	}
	szLine++;
	szLine = LFNSkipWhite( szLine );
	if( ulValue & 0xffff0000 )
	{
		printf("Error processing line\n");
		exit(1);
	}
	*puSectorCount = (unsigned)ulValue;
	if( *puSectorCount > uMaxSectorCount )
	{
		uMaxSectorCount = *puSectorCount;
	}


	if( *szLine != 0 )
	{
		printf("Error processing line\n");
		exit(1);
	}

	return TRUE;
}


char * GetHexValue( char * szLine, ulong * pulValue )
{
	unsigned uIndex;
	char cByte;
	ulong ulValue = 0;


	for( uIndex = 0; uIndex < 8; uIndex++ )
	{
		cByte = toupper( *szLine );

		if( (cByte >= '0') && (cByte <= '9') )
		{
			ulValue <<= 4;
			ulValue += cByte - '0';
		}
		else if( (cByte >= 'A') && (cByte <= 'F') )
		{
			ulValue <<= 4;
			ulValue += cByte - 'A' + 10;
		}
		else
		{
			break;
		}


		szLine++;
	}


	*pulValue = ulValue;
	return szLine;
}


char * PASCAL GetSourceLine( void )
{
	bool fSuccessful;
	unsigned uBytesRead;
	static char szInputLine[240];
	static unsigned uCharactersInBuffer = 0;
	static unsigned uCharactersInLastLine = 0;
	char * szCR;


	//  Move all the data remaining in the buffer forward
	//  past the "last" line found
	//
	uCharactersInBuffer -= uCharactersInLastLine;
	memmove( &szInputLine[0], &szInputLine[uCharactersInLastLine],
			uCharactersInBuffer );


	//  Fill up the buffer
	//
	uBytesRead = sizeof(szInputLine) - uCharactersInBuffer;
	fSuccessful = dlReadFile( iInputHandle, &uBytesRead,
			&szInputLine[uCharactersInBuffer] );
	if( !fSuccessful )
	{
		printf("Error reading input file\n" );
		exit(1);
	}
	uCharactersInBuffer += uBytesRead;


	//  Determine if there are any characters left in the buffer
	//
	uCharactersInLastLine = 0;
	if( uCharactersInBuffer == 0 )
	{
		*szInputLine = 0;
	}
	else
	{
		//  Terminate the string at the first CR
		//
		szCR = memchr( szInputLine, 0x0d, uCharactersInBuffer );
		if( szCR == NULL )
		{
			printf("Error finding the end of the string\n");
			exit(1);
		}
		*szCR = 0;
		if( szCR[1] == 0x0a )
		{
			szCR++;
		}
		*szCR = 0;
		uCharactersInLastLine = szCR - szInputLine;
		uCharactersInLastLine++;
	}


	//  Return the address of the line
	//
	return szInputLine;
}
