//
//   $Id: CHKASERT.C 1.6 2002/10/16 21:40:24Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CHKASERT.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This module contains the buffering scheme functions for chkdsk
//
//  HISTORY:
//
//  03-15-2000  bt  Created (Started over from scratch)
//  03-23-1989  Drew Gislason (created old chkdsk)
//


//  Borland libraries
//
#include <dir.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <mem.h>
#include <string.h>


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlxms.h"
#include "dlmem.h"


//  Chkdsk header files
//
#include "chkdsk.h"
#include "chkasert.h"


//  Prototypes
//
#include "chkdsk.pt"


#if DEBUG_ASSERTS


//  Variables
//
unsigned uAvailablePara    = 0;
unsigned uAllocatedPara    = 0;
unsigned uAllocatedSegment = 0;


//  Make sure that this is a valid DPB pointer
//
void IsDPBPtr( LPDPB lpDPB, unsigned uFileNumber, unsigned uLine )
{
	LPDPB lpDOSInternalDPB;
	bool fDPBValid;
	unsigned uDPBSize;
	uchar ucDrive;


	//  Get the drive number
	//
	ucDrive = GetWorkingDriveDriveNumber();


	//  Get the DOS internal DPB for comparison
	//
	if( !dlGetDPBForDrive(ucDrive, &lpDOSInternalDPB) )
	{
		CHKDSKInternalError( uFileNumber, uLine );
	}


	//  Copy the field DOS doesn't set when giving you a DPB
	//
	lpDPB->device = lpDOSInternalDPB->device;
	lpDPB->next = lpDOSInternalDPB->next;
	lpDPB->newclus = lpDOSInternalDPB->newclus;
	lpDPB->freeclus = lpDOSInternalDPB->freeclus;
	if(IsFat32Drive(lpDOSInternalDPB))
	{
		lpDPB->data_start = lpDOSInternalDPB->data_start;
		lpDPB->dir_start = lpDOSInternalDPB->dir_start;
		lpDPB->ulNextFreeCluster = lpDOSInternalDPB->ulNextFreeCluster;
	}


	//  Determine the size of the DPB
	//
	if( fIsFat32World && IsFat32Drive(lpDOSInternalDPB) )
	{
		uDPBSize = sizeof(DPB);
	}
	else
	{
//		uDPBSize = (unsigned)((&(lpDPB->uFreeClustersHigh)) - lpDPB);
		uDPBSize = 0x21;
	}


	//  Compare the DPBs.  If the DPBs did not match ...
	//
	if( _fmemcmp(lpDPB, lpDOSInternalDPB, uDPBSize) != 0 )
	{
		CHKDSKInternalError( uFileNumber, uLine );
	}
}


//  Make sure that this value is even
//
void IsEvenValue( ulong ulValue, unsigned uFileNumber, unsigned uLine )
{
	if( ulValue & 1L )
	{
		CHKDSKInternalError( uFileNumber, uLine );
	}
}


//  Make sure that this address is within the range of allocated
//  conventional memory (for conv buffers)
//
void IsInConvMemBuff(void far * lp, unsigned uFileNumber, unsigned uLine )
{
	ulong ulTemp;
	ulong ulLimit;


	ulTemp = SegmentedAddressToPhysical( lp );
	ulLimit = (ulong)uAllocatedSegment << 4;
	if( ulTemp < ulLimit )
	{
		printf("below min allocated memory range\n");
		printf("uAllocatedSegment: %xh\n", uAllocatedSegment );
		printf("uAllocatedPara: %xh\n", uAllocatedPara );
		CHKDSKInternalError( uFileNumber, uLine );
	}
	ulLimit = (ulong)(uAllocatedSegment + uAllocatedPara) << 4;
	if( ulTemp > ulLimit )
	{
		printf("exceed max allocated memory range\n");
		printf("uAllocatedSegment: %xh\n", uAllocatedSegment );
		printf("uAllocatedPara: %xh\n", uAllocatedPara );
		CHKDSKInternalError( uFileNumber, uLine );
	}
	ulTemp = SegmentedAddressToPhysical( (char far *)lp + 512 );
	if( ulTemp > ulLimit )
	{
		printf("512 bytes not accessable from pointer\n");
		printf("uAllocatedSegment: %xh\n", uAllocatedSegment );
		printf("uAllocatedPara: %xh\n", uAllocatedPara );
		CHKDSKInternalError( uFileNumber, uLine );
	}
}

#endif
