//
//   $Id: VISITLST.C 1.7 2003/06/06 17:01:54Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  VISITLST.C
//
//  Copyright (c) 2000-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This module contains the visited list interface routines for chkdsk.
//  The visited list is a table which contains the usage and other
//  information for every cluster on the drive.
//
//  HISTORY:
//
//  03-15-2000  bt  Created (Started over from scratch)
//  03-23-1989  Drew Gislason (created old chkdsk)
//


#define __FILENUMBER__		4


//  Borland libraries
//
#include "dir.h"
#include "stdio.h"
#include "stdlib.h"
#include "ctype.h"
#include "mem.h"
#include "string.h"


//  Datalight libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dlmem.h"
#include "fatlib.h"



//  Chkdsk header files
//
#include "chkdsk.h"
#include "chkasert.h"


//  Prototypes
//
#include "chkdsk.pt"


//  Examine and/or fix the FAT chain
//
uchar FollowFATChain( ulong ulStartingCluster, ulong * pulClusterCount,
		ulong * pulFirstCrossLink, ulong * pulAllocatedClusterCount,
		bool fFirstPass )
{
	uchar ucErrorCode;
	ulong ulTemp;

	ucErrorCode = _FollowFATChain( ulStartingCluster, pulClusterCount,
			pulFirstCrossLink, pulAllocatedClusterCount, fFirstPass, TRUE );
	_FollowFATChain( ulStartingCluster, &ulTemp, &ulTemp, &ulTemp,
			fFirstPass,	FALSE );

	return ucErrorCode;
}


//  Examine and/or fix the FAT chain
//
// This block of instructions gets entered four times.
//	1)  This is our first time following the chain.  We
//		are setting the cross-linked flag for cross-linked
//		clusters in the visited list this way a future
//		pass we can list which files are cross-linked
//		together.  We are also keeping track of the number
//		of clusters in the chain on this pass.  This
//		time through, the mode is set to the first pass.
//		We are setting the in-process flag in the visited
//		list so that we can detect when we are
//		circularly linked.
//	2)  This is our second time following the chain.  We
//		are only clearing the in-process flag in the
//		visited list from a previous pass so that other
//		cirularly linked chains can be detected.  No status
//		information or cluster counts are returned in
//		this pass.  This time through, the mode is set
//		to first pass.
//	3)  This is our third time following the chain.  We
//		have already been though the whole drive once.  The
//		visited list now reflects the state of the entire
//		FAT.  We are just seeing if there are any errors
//		we need to fix in this chain.  On the third pass,
//		the mode is not set to first pass.  No cluster count
//		information is returned on this pass.
//	4)  This is our fourth time following the chain.  If
//		there were problems found in the third pass, the
//		user was prompted if they wanted to fix the problems
//		but not by this code block.  So we may or may not
//		be fixing the problems that we have found.  In any
//		case, we are clearing the in-process bit in the
//		visited list so that other circular linked chains
//		can be detected.  We are also returning the cluster
//		count so that a file size entry may once again
//		be adjusted if the chain was terminated.
//
uchar _FollowFATChain( ulong ulStartingCluster, ulong * pulClusterCount,
		ulong * pulFirstCrossLink, ulong * pulAllocatedClusterCount,
		bool fFirstPass, bool fSettingInProcessBit )
{
	static uchar ucError;
	static ulong ulCurDirCluster;
	static ulong ulPrevCluster;
	static ulong ulNextCluster;
	static ulong ulNewCluster;
	static ulong ulTemp;
	static uchar ucVisitedListEntry;
	static ulong ulMaxClusterNumber = 0;
	static LPDPB lpDPB;
	bool fContainsCrosslinks = FALSE;


	//  If we have not already set up the non changing variables,
	//  set them up now
	//
	if( !ulMaxClusterNumber )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );


		//  Determine the maximum cluster number
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulMaxClusterNumber = lpDPB->ulMaxCluster;
		}
		else
		{
			ulMaxClusterNumber = lpDPB->hiclus;
		}
	}
	ASSERT_LPDPB( lpDPB );


	//  Initialize the variable which will contain the cluster
	//  number of the first cross-linked file.
	//
	*pulFirstCrossLink = 0;


	//  Initialize a cluster counter to zero
	//
	*pulClusterCount = 0;


	//  Initialize an allocated cluster counter to zero
	//
	*pulAllocatedClusterCount = 0;


	//  Initialize the status variable to zero
	//
	ucError = CHAIN_NO_ERROR;


	//
	//
	if( ulStartingCluster == 0 )
	{
		//  There are no clusters allocated in this chain
		//
	}
	else
	{
		//
		//
		ulCurDirCluster = ulStartingCluster;


		//  Save the current cluster number in the variable which
		//  contains the previous cluster number
		//
		ulPrevCluster = ulCurDirCluster;


		//
		//
		while( TRUE )
		{
			//  Get the visited list information for this cluster entry
			//
			ucVisitedListEntry = GetVisitedListClusterEntry( ulCurDirCluster );


			//  Determine if this cluster is already cross linked.  If this
			//  cluster is already cross-linked ...
			//
			if( (ucVisitedListEntry & VL_CROSS_LINKED) != 0 )
			{
				//  Record the location of the first cross-linked file
				//
				if( *pulFirstCrossLink == 0 )
				{
					*pulFirstCrossLink = ulCurDirCluster;
				}


				//  Set a flag indicating whether or not this cluster chain
				//  is already cross-linked
				//
				fContainsCrosslinks = TRUE;
			}


			//  Determine if this entry is invalid.  If this entry is
			//  invalid ...
			//
			if( (ulCurDirCluster < 2) ||
					(ulCurDirCluster > ulMaxClusterNumber) )
			{
				//  Determine if this chain contains cross-links.  If this
				//  file contains cross-links and we are running in simulation
				//  mode (no /f option), this chain would have been terminated
				//  at the bad cluster.  Since we do not require the entire
				//  FAT to be in memory this sector could have just been read
				//  in from disk and will not contain a terminated chain since
				//  changes are not being written.  So, assume that if there
				//  are cross-linked clusters in this chain that the chain
				//  would already be terminated.  Thus  Don't report that
				//  the chain is invalid.
				//
				if( fContainsCrosslinks && !DoFixErrors() )
				{
					//  See above
					//
					ucError |= CHAIN_ADJUST_SIZE;
				}
				else
				{
					//  This is not a valid cluster number.  Return the status
					//  indicating that this chain contained an invalid cluster
					//  number and was truncated.
					//
					ucError |= CHAIN_INVALID;
				}


				//  Determine if this is the first cluster in the chain.  If
				//  this is the first cluster in the chain.
				//
				if( ulCurDirCluster == ulPrevCluster )
				{
					//  The first cluster is invalid.  Set the error code to
					//  indicate that the first cluster was invalid
					//
					ucError |= CHAIN_FIRST_INVALID;
				}
				else
				{
					//  Terminate the chain at the previous cluster
					//
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
						PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
								&ulTemp );
					}
				}


				return ucError;
			}


			//  Determine if this value in this cluster entry is valid.  If
			//  the value at this cluster entry is invalid ...
			//
			GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  This is OK
			}
			else if( ulNextCluster == 0 )
			{
				//  Determine if this chain contains cross-links.  If this
				//  file contains cross-links and we are running in simulation
				//  mode (no /f option), this chain would have been terminated
				//  at the bad cluster.  Since we do not require the entire
				//  FAT to be in memory this sector could have just been read
				//  in from disk and will not contain a terminated chain since
				//  changes are not being written.  So, assume that if there
				//  are cross-linked clusters in this chain that the chain
				//  would already be terminated.  Thus  Don't report that
				//  the chain is invalid.
				//
				if( fContainsCrosslinks && !DoFixErrors() )
				{
					//  See above
					//
					ucError |= CHAIN_ADJUST_SIZE;
				}
				else
				{
					//  This is not a valid cluster (by use).  Return the status
					//  indicating that this chain contained an invalid cluster
					//  and was truncated.
					//
					ucError |= CHAIN_INVALID;
				}


				//  Determine if this is the first cluster in the chain.  If
				//  this is the first cluster in the chain.
				//
				if( ulCurDirCluster == ulPrevCluster )
				{
					//  The first cluster is invalid.  Set the error code to
					//  indicate that the first cluster was invalid
					//
					ucError |= CHAIN_FIRST_INVALID;
				}
				else
				{
					//  Terminate the chain at the previous cluster
					//
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
						PutFatEntry( lpDPB, ulCurDirCluster, END_OF_FAT,
								&ulTemp );
					}
				}
			}
			else if( (ulNextCluster < 2) ||
					(ulNextCluster > ulMaxClusterNumber) )
			{
				//  Determine if this chain contains cross-links.  If this
				//  file contains cross-links and we are running in simulation
				//  mode (no /f option), this chain would have been terminated
				//  at the bad cluster.  Since we do not require the entire
				//  FAT to be in memory this sector could have just been read
				//  in from disk and will not contain a terminated chain since
				//  changes are not being written.  So, assume that if there
				//  are cross-linked clusters in this chain that the chain
				//  would already be terminated.  Thus  Don't report that
				//  the chain is invalid.
				//
				if( fContainsCrosslinks && !DoFixErrors() )
				{
					//  See above
					//
					ucError |= CHAIN_ADJUST_SIZE;
				}
				else
				{
					//  This is not a valid cluster (by use).  Return the status
					//  indicating that this chain contained an invalid cluster
					//  and was truncated.
					//
					ucError |= CHAIN_INVALID;
				}


				//  Determine if this is the first cluster in the chain.  If
				//  this is the first cluster in the chain.
				//
				if( ulCurDirCluster == ulPrevCluster )
				{
					//  The first cluster is invalid.  Set the error code to
					//  indicate that the first cluster was invalid
					//
					ucError |= CHAIN_FIRST_INVALID;
				}
				else
				{
					//  Terminate the chain at the previous cluster
					//
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
//						PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
						PutFatEntry( lpDPB, ulCurDirCluster, END_OF_FAT,
								&ulTemp );
					}
				}

//				return ucError;
			}


			//  Determine if you are setting the in process bit.  If
			//  you are setting the in process bit
			//
			if( fSettingInProcessBit )
			{
				//  Check if the in-process bit is set in the visited list.
				//  If the in-process bit is set in the visited list...
				//
				if( ucVisitedListEntry & VL_IN_PROCESS )
				{
					//
					//  We are circularly linked in the same chain.  It is
					//  senseless to continue following the chain.
					//


					//  if the mode is set to first pass
					//
//					if( fFirstPass )
					{
						//  Determine if this chain contains cross-links.  If this
						//  file contains cross-links and we are running in
						//  simulation mode (no /f option), this chain would have
						//  been terminated at the bad cluster.  Since we do not
						//  require the entire FAT to be in memory this sector
						//  could have just been read in from disk and will not
						//  contain a terminated chain since changes are not being
						//  written.  So, assume that if there are cross-linked
						//  clusters in this chain that the chain would already be
						//  terminated.  Thus  Don't report that the chain is
						//  invalid.
						//
						if( fContainsCrosslinks && !DoFixErrors() )
						{
							//  See above
							//
//	bt						ucError |= CHAIN_ADJUST_SIZE;
							ucError |= CHAIN_CROSS_LINKED;
						}
						else
						{
							//  Set the allocation error bit in the status variable
							//
// bt						ucError |= CHAIN_INVALID;
							ucError |= CHAIN_CROSS_LINKED;
						}
					}


// bt				//  Record the location of the first cross-linked file
					//
					if( *pulFirstCrossLink == 0 )
					{
						*pulFirstCrossLink = ulCurDirCluster;
					}


					//  Terminate the chain at the previous cluster
					//
// bt
					if( !fFirstPass )
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
						PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
								&ulTemp );
					}


					//  End repeat
					//
					break;
				}
				else
				{
					//  Set the in-process bit in the visited list
					//
					ucVisitedListEntry |= VL_IN_PROCESS;
				}


				//  Increament the cluster counter
				//
				(*pulClusterCount)++;


				//  Increament the count or allocated clusters if
				//  we are not in a cross linked chain
				//
				if(!fContainsCrosslinks)
				{
					(*pulAllocatedClusterCount)++;
				}

				//  if the mode is set to first pass
				//
				if( fFirstPass )
				{
					//  Check if the used bit is set in the visited list.  If
					//  the used bit is set in the visited list...
					//
					if( ucVisitedListEntry & VL_USED )
					{
						//  Set the cross-linked bit in the status variable
						//
						ucError |= CHAIN_CROSS_LINKED;


						//  Set a flag indicating that this chain is cross-linked
						//
						fContainsCrosslinks = TRUE;


						//  Set the cross-linked bit in the visited list
						//
						ucVisitedListEntry |= VL_CROSS_LINKED;


						//  Record the location of the first cross-linked file
						//
						if( *pulFirstCrossLink == 0 )
						{
							*pulFirstCrossLink = ulCurDirCluster;
						}
					}
					else
					{
						//  Set the used bit in the visited list
						//
						ucVisitedListEntry |= VL_USED;
					}
				}
				else
				{
					//  Check if the cross-linked bit is set in the visited
					//  list.  If the cross-linked bit is set in the visited
					//  list...
					//
					if( ucVisitedListEntry & VL_CROSS_LINKED )
					{
						//  Set the cross-linked bit in the visited list
						//
						ucVisitedListEntry |= VL_CROSS_LINKED;


						//  Record the location of the first cross-linked file
						//
						if( *pulFirstCrossLink == 0 )
						{
							*pulFirstCrossLink = ulCurDirCluster;
						}


						//  Set the cross-linked bit in the status variable
						//
						ucError |= CHAIN_CROSS_LINKED;
					}
				}
			}
			else
			{
				//  Check if the in-process bit is set in the visited list.
				//  If the in-process bit is set in the visited list ...
				//
				if( ucVisitedListEntry & VL_IN_PROCESS )
				{
					//  Clear the in-process bit in the visited list
					//
					ucVisitedListEntry &= ~VL_IN_PROCESS;
				}
				else
				{
					//
					//  We are circularly linked in the same chain.  It is
					//  senseless to continue following the chain
					//

					//  If the mode is not first pass
					//
					if( !fFirstPass )
					{
						//  Terminate the chain at the previous cluster
						//
						{
							//  Determine the cluster number of the
							//  previous cluster and set an EOF value
							//  at this entry
							//
							PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
									&ulTemp );
						}
					}


					//  end repeat
					//
					break;
				}


				//  Increament the cluster counter
				//
				(*pulClusterCount)++;


				if( !fFirstPass )
				{
					//  Check if the copy bit is set in the visited list.
					//  If the copy bit is set in the visited list ...
					//
					if( ucVisitedListEntry & VL_DO_COPY )
					{
						//
						//  We need to make a copy of this cluster
						//
						{
							//  Determine the next cluster number in this chain
							//
							GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );


							//  Allocate a new cluster.  If there was an error
							//  allocating a new cluster ...
							//
							if( !AllocateCluster(&ulNewCluster,FALSE) )
							{
								//  Display a message stating that there was not
								//  enough free space to fix the problem and
								//  instruct the user to free some space and
								//  run chkdsk again.
								//
								DisplayInsufficientSpaceErrorMsg();

								//  We have done all we can do but we still need
								//  to clear our in processing flags in the visited
								//  list so we will continue.  The above error
								//  message will only be displayed once.
								//
							}
							else
							{
								//  Increament the count or allocated clusters
								//
								(*pulAllocatedClusterCount)++;


								//  Save the current visited list entry information
								//
								SetVisitedListClusterEntry( ulCurDirCluster,
										ucVisitedListEntry );


								//  Copy the contents of the current cluster
								//  to the newly allocated cluster
								//
								CopyCluster( ulCurDirCluster, ulNewCluster );


								//  Set the FAT entry information for the previous and
								//  next clusters.
								//
								PutFatEntry( lpDPB, ulPrevCluster, ulNewCluster,
										&ulTemp );
								PutFatEntry( lpDPB, ulNewCluster, ulNextCluster,
										&ulTemp );


								//  Set the new cluster as the current cluster
								//
								ulCurDirCluster = ulNewCluster;


								//  Get the visited list information for this entry
								//
								ucVisitedListEntry = GetVisitedListClusterEntry(
										ulCurDirCluster );


								//  Set the used bit in the visited list
								//
								ucVisitedListEntry |= VL_USED;
							}
						}
					}
					else
					{
						//  Check if the cross-linked bit is set in the
						//  visited list.  If the cross-linked bit is set
						//  in the visited list ...
						//
						if( ucVisitedListEntry & VL_CROSS_LINKED )
						{
							//
							//  The next chain that comes across this
							//  cross-linked cluster will copy it and
							//  adjusted its chain accordingly.
							//

							//  Set the copy bit in the visited list
							//
							ucVisitedListEntry |= VL_DO_COPY;
						}
					}
				}
			}


			//  Save the visited list information for this cluster entry
			//
			SetVisitedListClusterEntry( ulCurDirCluster, ucVisitedListEntry );


			//  Get the next cluster entry
			//
			GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  We are done processing this directory
				//
				break;
			}

			//  Save the current cluster number in the variable which
			//  contains the previous cluster number
			//
			ulPrevCluster = ulCurDirCluster;


			//  Continue on with the next cluster
			//
			ulCurDirCluster = ulNextCluster;
		}
	}


	//  Save the visited list information for this cluster entry
	//
	SetVisitedListClusterEntry( ulCurDirCluster, ucVisitedListEntry );


	//
	//
	return ucError;
}


//  Examine a lost cluster chain and possibly fix the FAT chain
//
// This block of instructions gets entered four times.
// 1) This is our first time following the lost cluster chain.
//    We are not setting the used bit or marking cross-linked
//    clusters.  Basically we are just counting the clusters
//    in the chain.  We are setting the in-process bit for
//    each cluster in the chain.  We will stop following the
//    chain when we find a bad cluster, circular linked cluster,
//    cross-linked cluster, or bad cluster number.
// 2) This is our second time following the lost cluster chain.
//    We are just clearing the in-process bit for each cluster
//    in the chain.  We will stop following the chain when we
//    find a bad cluster, circular linked cluster, cross-linked
//    cluster, or bad cluster number.
// 3) This is our third time following the lost cluster chain.
// 4) This is our fouth time following the lost cluster chain.
//
uchar _FollowLostClusterChain( ulong ulStartingCluster,
		ulong * pulClusterCount, bool fFirstPass, bool fSettingInProcessBit,
		bool fFreeLostClusters )
{
	uchar ucError;
	ulong ulCurDirCluster;
	ulong ulPrevCluster;
	ulong ulNextCluster;
	ulong ulTemp;
	uchar ucVisitedListEntry;
	static ulong ulMaxClusterNumber = 0;
	static LPDPB lpDPB;


	//  If we have not already set up the non changing variables,
	//  set them up now
	//
	if( !ulMaxClusterNumber )
	{
		//  Get the DPB for the drive
		//
		GetDPB( &lpDPB );


		//  Determine the maximum cluster number
		//
		if( IsFat32Drive(lpDPB) )
		{
			ulMaxClusterNumber = lpDPB->ulMaxCluster;
		}
		else
		{
			ulMaxClusterNumber = lpDPB->hiclus;
		}
	}
	ASSERT_LPDPB( lpDPB );


	//  Initialize a cluster counter to zero
	//
	*pulClusterCount = 0;


	//  Initialize the status variable to zero
	//
	ucError = CHAIN_NO_ERROR;


	//
	//
	if( ulStartingCluster == 0 )
	{
		//  There are no clusters allocated in this chain
		//
	}
	else
	{
		//  Set the cluster number we are currently working on
		//
		ulCurDirCluster = ulStartingCluster;


		//  Save the current cluster number in the variable which
		//  contains the previous cluster number
		//
		ulPrevCluster = ulCurDirCluster;


		//  Follow the entire cluster chain
		//
		while( TRUE )
		{
			//  Determine if this entry is invalid.  If this entry is
			//  invalid ...
			//
			if( (ulCurDirCluster < 2) ||
					(ulCurDirCluster > ulMaxClusterNumber) )
			{
				//  Determine if this is our first pass following the chain.
				//  If this is not our first pass following the chain ...
				//
				if( (!fFirstPass) && (!fFreeLostClusters) )
				{
					//  Terminate the chain at the previous cluster
					//
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
						PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
								&ulTemp );
					}
				}


				//  This is not a valid cluster number.
				//
				return ucError;
			}


			//  Determine if this value in this cluster entry is valid.  If
			//  the value at this cluster entry is invalid ...
			//
			GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  This is OK
			}
			else if( (ulNextCluster < 2) ||
					(ulNextCluster > ulMaxClusterNumber) )
			{
				//  Determine if this is our first pass following the chain.
				//  If this is not our first pass following the chain ...
				//
				if( (!fFirstPass) && (!fFreeLostClusters) )
				{
					//  Terminate the chain at the previous cluster
					//
					{
						//  Determine the cluster number of the
						//  previous cluster and set an EOF value
						//  at this entry
						//
						PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
								&ulTemp );
					}
				}


				//  This is not a valid cluster (by use).
				//
            // However, we must set the visited list information if
            // we are ever to correct the error.  So don't return.
            //
            // return ucError;
			}


			//  Get the visited list information for this cluster entry
			//
			ucVisitedListEntry = GetVisitedListClusterEntry( ulCurDirCluster );


			//  Determine if you are setting the in process bit.  If
			//  you are setting the in process bit
			//
			if( fSettingInProcessBit )
			{
				//  Check if the used bit is set in the visited list.  If
				//  the used bit is set in the visited list...
				//
				if( ucVisitedListEntry & VL_USED )
				{
					//
					//  This cluster is already in use by another chain.
					//  Like MS, we will ignore the rest of the chain
					//


					//  End repeat
					//
					break;
				}


				//  Check if the in-process bit is set in the visited list.
				//  If the in-process bit is set in the visited list...
				//
				if( ucVisitedListEntry & VL_IN_PROCESS )
				{
					//
					//  We are circularly linked in the same chain.  It is
					//  senseless to continue following the chain.
					//


					//  End repeat
					//
					break;
				}
				else
				{
					//  Set the in-process bit in the visited list
					//
					ucVisitedListEntry |= VL_IN_PROCESS;
				}


				//  Increament the cluster counter
				//
				(*pulClusterCount)++;


				//  if the mode is set to first pass
				//
				if( fFirstPass )
				{
#if 0
					//  Check if the used bit is set in the visited list.  If
					//  the used bit is set in the visited list...
					//
					if( ucVisitedListEntry & VL_USED )
					{
						//
						//  This cluster is already in use by another chain.
						//  Like MS, we will ignore the rest of the chain
						//


						//  End repeat
						//
						break;
					}
#endif
				}
				else
				{
					//  Check if the used bit is set in the visited list.  If
					//  the used bit is set in the visited list...
					//
					if( ucVisitedListEntry & VL_USED )
					{
						//  Set the cross-linked bit in the status variable
						//
						ucError |= CHAIN_CROSS_LINKED;


						//  Set the cross-linked bit in the visited list
						//
						ucVisitedListEntry |= VL_CROSS_LINKED;
					}
					else
					{
						//  Set the used bit in the visited list
						//
						ucVisitedListEntry |= VL_USED;
					}
				}
#if 0
				else
				{
					//  Check if the cross-linked bit is set in the visited
					//  list.  If the cross-linked bit is set in the visited
					//  list...
					//
					if( ucVisitedListEntry & VL_CROSS_LINKED )
					{
						//  Set the cross-linked bit in the status variable
						//
						ucError |= CHAIN_CROSS_LINKED;
					}
				}
#endif
			}
			else
			{
				//  Check if the in-process bit is set in the visited list.
				//  If the in-process bit is set in the visited list ...
				//
				if( ucVisitedListEntry == VL_IN_PROCESS )
				{
					//  Clear the in-process bit in the visited list
					//
					ucVisitedListEntry &= ~VL_IN_PROCESS;


					//  Determine if this is our first pass through the chain.
					//  if this is not our first pass through the chain ...
					//
					if( !fFirstPass )
					{
						//  Determine if we are freeing lost clusters.  If we
						//  are not freeing lost clusters
						//
						if(!fFreeLostClusters)
						{
							//  Mark this entry as used
							//
							ucVisitedListEntry |= VL_USED;
						}
					}
				}
				else
				{
					//
					//  We are circularly linked in the same chain.  It is
					//  senseless to continue following the chain
					//


					//  Determine if this is our first pass following the chain.
					//  If this is not our first pass following the chain ...
					//
					if( (!fFirstPass) && (!fFreeLostClusters) )
					{
						//  Terminate the chain at the previous cluster
						//
						{
							//  Determine the cluster number of the
							//  previous cluster and set an EOF value
							//  at this entry
							//
							PutFatEntry( lpDPB, ulPrevCluster, END_OF_FAT,
									&ulTemp );
						}
					}


					//  end repeat
					//
					break;
				}


				//  Increament the cluster counter
				//
				(*pulClusterCount)++;
			}


			//  Save the visited list information for this cluster entry
			//
			SetVisitedListClusterEntry( ulCurDirCluster, ucVisitedListEntry );


			//  Save the current cluster number in the variable which
			//  contains the previous cluster number
			//
			ulPrevCluster = ulCurDirCluster;


			//  Get the next cluster entry
			//
			GetFatEntry( lpDPB, ulCurDirCluster, &ulNextCluster );


			//  Determine if this is our first pass following the chain.
			//  If this is not our first pass following the chain and
			//  we are freeing lost clusters ...
			//
			if( (!fFirstPass) && (fFreeLostClusters) )
			{
				//  Mark the cluster we just processed as free
				//
				PutFatEntry( lpDPB, ulPrevCluster, 0, &ulTemp );
			}


			//  Check if we are done with this chain
			//
			if( ulNextCluster >= END_OF_FAT_CHECK )
			{
				//  We are done processing this directory
				//
				break;
			}


			//  Continue on with the next cluster
			//
			ulCurDirCluster = ulNextCluster;
		}
	}


	//  Save the visited list information for this cluster entry
	//
	SetVisitedListClusterEntry( ulCurDirCluster, ucVisitedListEntry );


	//
	//
	return ucError;
}


//  Get the visited list information for this cluster entry
//
uchar GetVisitedListClusterEntry( ulong ulClusterEntry )
{
	bool fIsInLowNibble;
	ulong ulBlockNumber;
	unsigned uOffsetInBlock;
	uchar * pucVisitedListSwap;
	ulong * pulVLSBlockNumber;
	uchar ucValue;


	//  Determine which block this entry exists in, the offset
	//  into that block, and whether it is the high nibble or
	//  the low nibble.
	//
	fIsInLowNibble  = ((ulClusterEntry & 1) != 0);
	ulClusterEntry >>= 1;
	ulBlockNumber   = ulClusterEntry / VISITED_LIST_SWAP_SIZE;
	uOffsetInBlock  = (unsigned)(ulClusterEntry % VISITED_LIST_SWAP_SIZE);


	//  Determine the location of the visited list swap area
	//
	pucVisitedListSwap = GetVisitedListSwapAddress();


	//  Determine if the block number we need is already in the
	//  visited list swap area.  If it is not already in the
	//  visited list swap area ...
	//
	pulVLSBlockNumber = GetVLSBlockNumberVariable();
	if( *pulVLSBlockNumber != ulBlockNumber )
	{
		//  Write the data currently in the visited list swap
		//  area into the memory store.
		//
 		{
			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();


			//  Set the next location for a write to the memory store
			//
			MemStoreSeek( *pulVLSBlockNumber * VISITED_LIST_SWAP_SIZE,
					SEEK_SET );


			//  Write the vistited list swap area into the memory store
			//
			MemStoreWrite( pucVisitedListSwap, VISITED_LIST_SWAP_SIZE );


			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();
		}


		//  Record the block number we are about to load into the
		//  visited list swap area
		//
		*pulVLSBlockNumber = ulBlockNumber;


		//  Load the visited list swap area with the visited list
		//  block we are looking for
		//
		{
			//  Set the next location for a write to the memory store
			//
			MemStoreSeek( *pulVLSBlockNumber * VISITED_LIST_SWAP_SIZE,
					SEEK_SET );


			//  Write the vistited list swap area into the memory store
			//
			MemStoreRead( pucVisitedListSwap, VISITED_LIST_SWAP_SIZE );
		}
	}


	//  Get the byte which contains the entry we are looking for
	//
	ucValue = pucVisitedListSwap[uOffsetInBlock];


	//  Get the entry information for this cluster number
	//
	if( fIsInLowNibble )
	{
		//  Mask off the other entry information
		//
		ucValue &= 0x0f;
	}
	else
	{
		//  Shift the entry information we need into the low byte
		//
		ucValue >>= 4;
	}


	//  Return the entry information for this cluster number
	//
	return ucValue;
}


//  return the address of the variable which contains the current
//  block number in the visited list swap area
//
ulong * GetVLSBlockNumberVariable( void )
{
	static ulong ulCurrentVLSBlock = 0;
	return &ulCurrentVLSBlock;
}


//  Save the visited list information for this cluster entry
//
void SetVisitedListClusterEntry( ulong ulClusterEntry, uchar ucNewValue )
{
	bool fIsInLowNibble;
	ulong ulBlockNumber;
	unsigned uOffsetInBlock;
	uchar * pucVisitedListSwap;
	ulong * pulVLSBlockNumber;
	uchar ucValue;


	//  Determine which block this entry exists in, the offset
	//  into that block, and whether it is the high nibble or
	//  the low nibble.
	//
	fIsInLowNibble  = ((ulClusterEntry & 1) != 0);
	ulClusterEntry >>= 1;
	ulBlockNumber   = ulClusterEntry / VISITED_LIST_SWAP_SIZE;
	uOffsetInBlock  = (unsigned)(ulClusterEntry % VISITED_LIST_SWAP_SIZE);


	//  Determine the location of the visited list swap area
	//
	pucVisitedListSwap = GetVisitedListSwapAddress();


	//  Determine if the block number we need is already in the
	//  visited list swap area.  If it is not already in the
	//  visited list swap area ...
	//
	pulVLSBlockNumber = GetVLSBlockNumberVariable();
	if( *pulVLSBlockNumber != ulBlockNumber )
	{
		//  Write the data currently in the visited list swap
		//  area into the memory store.
		//
 		{
			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();


			//  Set the next location for a write to the memory store
			//
			MemStoreSeek( *pulVLSBlockNumber * VISITED_LIST_SWAP_SIZE,
					SEEK_SET );


			//  Write the vistited list swap area into the memory store
			//
			MemStoreWrite( pucVisitedListSwap, VISITED_LIST_SWAP_SIZE );


			//  Make sure that the buffering scheme is OK
			//
			ASSERT_BUFFERS();
		}


		//  Record the block number we are about to load into the
		//  visited list swap area
		//
		*pulVLSBlockNumber = ulBlockNumber;


		//  Load the visited list swap area with the visited list
		//  block we are looking for
		//
		{
			//  Set the next location for a write to the memory store
			//
			MemStoreSeek( *pulVLSBlockNumber * VISITED_LIST_SWAP_SIZE,
					SEEK_SET );


			//  Write the vistited list swap area into the memory store
			//
			MemStoreRead( pucVisitedListSwap, VISITED_LIST_SWAP_SIZE );
		}
	}


	//  Get the byte which contains the entry we are looking for
	//
	ucValue = pucVisitedListSwap[uOffsetInBlock];


	//  Get the entry information for this cluster number
	//
	if( fIsInLowNibble )
	{
		//  Mask off the old entry information
		//
		ucValue &= 0xf0;


		//  Or in the new entry information
		//
		ucValue |= ucNewValue & 0x0f;
	}
	else
	{
		//  Shift the entry information we need into the low byte
		//
		ucNewValue <<= 4;


		//  Mask off the old entry information
		//
		ucValue &= 0x0f;


		//  Or in the new entry information
		//
		ucValue |= ucNewValue;
	}


	//  Get the byte which contains the entry we are looking for
	//
	pucVisitedListSwap[uOffsetInBlock] = ucValue;
}


//  Return the address of the conventional memory swap address
//  for the visited list
//
uchar * GetVisitedListSwapAddress( void )
{
	static uchar ucVisitedListSwap[VISITED_LIST_SWAP_SIZE];
	return ucVisitedListSwap;
}
