//
//   $Id: CLIAPI.C 1.8 2002/10/16 21:42:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  CLIAPI.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  This file contains all the API functions that an int 2fh IFS
//  interface is expected to perform.
//
//  History:
//
//  04-01-2001  bt  Created
//


//  Borland libraries
//
#include <string.h>
#include <mem.h>
#include <stdio.h>


//  Local libraries
//
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dlapp.h"


//  Local definitions
//
#include "client.h"
#include "cliapi.h"
#include "netapi.h"
#include "commapi.h"
#include "ifsdisp.h"


//  Prototypes
//
#include "cliapi.pt"


void TermLocation( void );


bool MapRemoteDrive( uchar ucRemoteDriveLetter,	bool fIsRemovable,
		bool fIsRemote );
void DisplayDriveMapping( void );
void InstallIFSVector( void );
bool RestoreIFSVector( void );


extern CDSMAP DriveMap[26];
extern uDriveMapSize;


char szDontMapList[26];
char szOnlyMapList[26];


//  ========================================================================
//
//  The following functions are required APIs for the INT 2fh interface.
//
//  ========================================================================




//  InitInt2fInterface
//
//
//  Initialize the INT 2fh interface
//
//
//  On entry:
//
//    argc = The number of options in the argv array
//    argv = Array of pointers to command line options
//
//  On exit:
//
//  on success,
//
//  	function returns TRUE
//
//  on error,
//
//		function returns FALSE
//
//
bool InitInt2fInterface( int argc, char *argv[] )
{
	unsigned uCountOfMappedDrives = 0;
	PREMOTELIST pList;
	unsigned uErrorCode;
	uchar ucKey;
	unsigned uIndex;
	bool fSuccessful;
	bool fInstalled;
	static ARGDEF ClientOpts1[] =
	{
		{ BOOL_SWITCH, &fSwitches[SW_HIDE_SIGNON],      '[', 0 },
		{ BOOL_SWITCH, &fSwitches[SW_UNLOAD],      		'U', 0 },
		{ BOOL_SWITCH, &fSwitches[SW_REMOVABLE],    		'R', 0 },
		{ BOOL_SWITCH, &fSwitches[SW_NETWORK],    		'N', 0 },
		{ STRING_SWITCH, szDontMapList,                 'D', 26 },
		{ STRING_SWITCH, szOnlyMapList,                 'O', 26 },
		{ 0 }
	};


	//  Process some of the boolean switches in the command line
	//
	SetParseFlags(PARSEF_MULTIPASS | PARSEF_NOARGSOK);
	argc = ParseCmdline(argc, &argv[0], ClientOpts1);


	//  Uppercase any drive letters the user may have supplied
	//
	strupr( szDontMapList );
	strupr( szOnlyMapList );


	//  Installation check
	//
	fInstalled = InstallationCheck();
	if( fInstalled )
	{
		//  Determine if the user wanted to unload
		//
		if( DoUnload() )
		{
			Int2fUnload();
			return FALSE;
		}


		//  Determine if we should display the signon message and if
		//  so display it
		//
		DisplaySignonMessage();


		//
		//
		printf("Already installed.  Current mappings ...\n");
		DisplayDriveMapping();
		return FALSE;
	}


	//  Determine if the user wanted to unload
	//
	if( DoUnload() )
	{
		printf("Client not found in memory.\n");
		return FALSE;
	}


	//  Process the remaining part of the command line and
	//  initialize the transport layer
	//
	fSuccessful = NetworkTXRXInit( argc, &argv[0] );
	if( !fSuccessful )
	{
		return FALSE;
	}


	//  Get the mappable drive list from the server
	//
	printf("Communicating with server...");
	printf("Press any key to abort.\n");
	while( TRUE )
	{
		uErrorCode = GetDriveMapList( &pList );
		if( !uErrorCode )
		{
			break;
		}

		if( GetKey(&ucKey) )
		{
			printf("User abort.\n");
			return FALSE;
		}
	}


	//  Add each of the mappable drives to our I2F interface
	//
	printf("Mapping remote drives...\n");
	for( uIndex = 0; uIndex < 26; uIndex++ )
	{
		if( (pList)[uIndex].ucRemoteDriveLetter )
		{
			fSuccessful = MapRemoteDrive( (pList)[uIndex].ucRemoteDriveLetter,
					(pList)[uIndex].fIsRemovable, (pList)[uIndex].fIsRemote );
			if( fSuccessful )
			{
				uCountOfMappedDrives++;
			}
		}
		else
		{
			break;
		}
	}


	//  If there were no drives mapped, then the initialization
	//  was not successful
	//
	if( uCountOfMappedDrives == 0 )
	{
		printf("No remote drives to map.\n");
		return FALSE;
	}


	//  Display a list of the mapped drives
	//
	DisplayDriveMapping();


	//  Enable the INT 2fh Interface
	//
	InstallIFSVector();


	//  Return that initialization was successful
	//
	return TRUE;
}


//  GetRemoteDriveInfo  INT 2fh fn 110ch
//
//
//  Get some information about the remote drive such as the
//  total drive space and free space.  Because of the limitation
//  in the size of the variables, the values returned from this
//  function may have been adjusted to fit the size of the
//  variables but still reflect the amount of space on the remote
//  drive.
//
//
//  On entry:
//
//  	ucRemoteDriveLetter	= Drive letter of drive on remote machine
//  	pucSectorsPerCluster = Variable for sectors per cluster
//  	pucMediaIdByte			= Variable for media id byte
//  	puTotalClusters		= Variable for total clusters
//  	puBytesPerSector		= Variable for bytes per sector
//  	puAvailableClusters	= Variable for available clusters
//
//  On exit:
//
//  on success,
//
//  	pucSectorsPerCluster	= is updated
//  	ucMediaIdByte			= is updated
//  	puTotalClusters		= is updated
//  	puBytesPerSector 		= is updated
//  	puAvailableClusters	= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL GetRemoteDriveInfo( uchar ucRemoteDriveLetter,
		uchar * pucSectorsPerCluster, uchar * pucMediaIdByte,
		unsigned * puTotalClusters, unsigned * puBytesPerSector,
		unsigned * puAvailableClusters )
{
	unsigned uError;
	PREQ110C  pRequest;
	PSERV110C pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x110c;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ110C) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x110C) )
	{
		return ERR_BADFUNC;
	}


	//  Update the requested information
	//
	if( !pService->uErrorCode )
	{
		*pucSectorsPerCluster = pService->ucSectorsPerCluster;
		*pucMediaIdByte       = pService->ucMediaIdByte;
		*puTotalClusters      = pService->uTotalClusters;
		*puBytesPerSector     = pService->uBytesPerSector;
		*puAvailableClusters  = pService->uAvailableClusters;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoteFindFirst  INT 2fh fn 111bh
//
//
//  Find the first entry that matches the mask given in the RemoteFindFirst
//  call on a remote machine
//
//  NOTE: lpFIND must be filled out before lpDIR.  When MS hands down the
//        call to this function, the DIR structure overlaps the FIND
//        structure.
//
//
//  On entry:
//
//		szCWD	 				= Current working directory on the remote machine
//		szMask				= name or mask including path to search for
//		uSearchAttributes	= Attributes to search for on remote machine
//		lpFIND 				= Address of the find first data block (in the DTA)
//
//  On exit:
//
//  on success,
//
//		lpFIND 			= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoteFindFirst( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szMask, unsigned uSearchAttributes,
		LPFIND lpFIND, LPDIR lpDIR )
{
	unsigned uError;
	PREQ111B pRequest;
	PSERV111B pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x111b;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szMask[0]), szMask );
	pRequest->uSearchAttr = uSearchAttributes;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ111B) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x111b) )
	{
		return ERR_BADFUNC;
	}


	//  Update the DOS structures with the results
	//
	_fmemcpy( lpFIND, &(pService->Find), sizeof(FIND) );
	if( !pService->uErrorCode )
	{
		_fmemcpy( lpDIR, &(pService->Dir), sizeof(DIR) );
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoteFindNext  INT 2fh fn 111ch
//
//
//  Find the next entry that matches the mask given in the RemoteFindFirst
//  call on a remote machine
//
//  NOTE: lpFIND must be filled out before lpDIR.  When MS hands down the
//        call to this function, the DIR structure overlaps the FIND
//        structure.
//
//
//  On entry:
//
//		szCWD	= Current working directory on the remote machine
//		lpFIND	= Address of the find first data block (in the DTA)
//
//  On exit:
//
//  on success,
//
//		lpFIND 			= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoteFindNext( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, LPFIND lpFIND, LPDIR lpDIR )
{
	unsigned uError;
	PREQ111C pRequest;
	PSERV111C pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x111c;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fmemcpy( &(pRequest->Find), lpFIND, sizeof(FIND) );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ111C) );
	if( uError != NET_SERVICED )
	{
		return ERR_BADFUNC;
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x111c) )
	{
		return ERR_BADFUNC;
	}


	//  Update the DOS structures with the results
	//
	if( !pService->uErrorCode )
	{
		_fmemcpy( lpFIND, &(pService->Find), sizeof(FIND) );
		_fmemcpy( lpDIR, &(pService->Dir), sizeof(DIR) );
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  GetRemoteFileAttributesAndSize  INT 2fh fn 110fh
//
//
//  Retrieve the attributes and size of a file on a remote machine
//
//
//  On entry:
//
//		szCWD	 			= Current working directory on the remote machine
//		szPath 			= Path for the directory on the remote machine
//		puAttributes	= Variable for file attributes of the remote file
//		pulSize			= Variable for file size of the remote file
//
//  On exit:
//
//  on success,
//
//		puAttributes	= is updated
//		pulSize			= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL GetRemoteFileAttributesAndSize( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath, unsigned * puAttributes,
		ulong * pulSize )
{
	unsigned uError;
	PREQ110F pRequest;
	PSERV110F pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x110f;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ110F) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x110f) )
	{
		return ERR_BADFUNC;
	}


	//  Update the results
	//
	if( !pService->uErrorCode )
	{
		*puAttributes = pService->uAttributes;
		*pulSize = pService->ulSize;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  SetRemoteFileAttributes  INT 2fh fn 110eh
//
//
//  Set the attributes for a file on a remote machine
//
//
//  On entry:
//
//		szCWD	 			= Current working directory on the remote machine
//		szPath 			= Path for the directory on the remote machine
//		uAttributes		= Desired attributes for the remote file
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL SetRemoteFileAttributes( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath, unsigned uAttributes )
{
	unsigned uError;
	PREQ110E pRequest;
	PSERV110E pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x110e;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );
	pRequest->uAttributes = uAttributes;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ110E) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x110e) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  CreateRemoteDirectory  INT 2fh fn 1103h
//
//
//  Create a directory on a remote machine
//
//
//  On entry:
//
//		szCWD	 = Current working directory on the remote machine
//		szPath = Path for the directory on the remote machine
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL CreateRemoteDirectory( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath )
{
	unsigned uError;
	PREQ1103  pRequest;
	PSERV1103 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1103;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]),  szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1103) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1103) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoveRemoteDirectory  INT 2fh fn 1101h
//
//
//  Remove a directory on a remote machine
//
//
//  On entry:
//
//		szCWD	 = Current working directory on the remote machine
//		szPath = Path for the directory on the remote machine
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoveRemoteDirectory( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath )
{
	unsigned uError;
	PREQ1101  pRequest;
	PSERV1101 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1101;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]),  szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1101) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1101) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  ChangeRemoteDirectory  INT 2fh fn 1105h
//
//
//  Change the current directory on a remote machine
//
//
//  On entry:
//
//		szCWD	 = Current working directory on the remote machine
//		szPath = Path for the directory on the remote machine
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL ChangeRemoteDirectory( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath )
{
	unsigned uError;
	PREQ1105  pRequest;
	PSERV1105 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1105;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]),  szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1105) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1105) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  DeleteRemoteFile  INT 2fh fn 1113h
//
//
//  Delete a file on the remote machine
//
//
//  On entry:
//
//		szCWD		= Current working directory on the remote machine
//		szPath	= Path for the file on the remote machine
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL DeleteRemoteFile( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath )
{
	unsigned uError;
	PREQ1113 pRequest;
	PSERV1113 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1113;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1113) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1113) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RenameRemoteFile  INT 2fh fn 1111h
//
//
//  Rename a file on the remote machine
//
//
//  On entry:
//
//		szCWD				= Current working directory on the remote machine
//		szCurrentName	= Current name of the file on the remote machine
//		szNewName		= Desired name of the file on the remote machine
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RenameRemoteFile( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szCurrentName, char _D_ * szNewName )
{
	unsigned uError;
	PREQ1111 pRequest;
	PSERV1111 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1111;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szCurrentName[0]), szCurrentName );
	_fstrcpy( &(pRequest->szNewName[0]), szNewName );


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1111) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1111) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoteExtendedOpenCreateFile INT 2fh fn 112Eh
//
//
//  Open, create, or truncate a file on a remote machine
//
//
//  On entry:
//
//		szCWD				= Current working directory on the remote machine
//		szPath			= Path for the directory on the remote machine
//		uOpenMode		= Desired action (open, create, or truncate)
//		uAttributes    = File attributes for the file
//    uProcessID     = Process ID on remote machine
//		puAction       = Variable for the action performed (open, create, etc)
//		puRemoteHandle	= Variable for file handle on remote machine
//    puTime         = Variable for file time in DOS packed form
//    puDate         = Variable for file date in DOS packed form
//		pulSize        = Variable for the file size
//		puAttributes   = Variable for the file attributes
//
//  On exit:
//
//  on success,
//
//		puAction       = is updated
//		puRemoteHandle	= is updated
//    puTime         = Time in DOS packed form
//    puDate         = Date in DOS packed form
//		pulSize        = is updated
//		puAttributes   = is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoteExtendedOpenCreateFile( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath, unsigned uOpenMode,
		unsigned uAttributes, unsigned uProcessID, unsigned * puAction,
		unsigned * puRemoteHandle, unsigned * puTime, unsigned * puDate,
		ulong * pulSize, unsigned * puAttributes )
{
	unsigned uError;
	PREQ112E pRequest;
	PSERV112E pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x112E;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );
	pRequest->uOpenMode = uOpenMode;
	pRequest->uAttributes = uAttributes;
	pRequest->uAction = *puAction;
	pRequest->uProcessID = uProcessID;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ112E) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x112E) )
	{
		return ERR_BADFUNC;
	}


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*puAction = pService->uRemoteAction;
		*puRemoteHandle = pService->uRemoteHandle;
		*puTime  = pService->uFileTime;
		*puDate  = pService->uFileDate;
		*pulSize = pService->ulSize;
		*puAttributes = pService->uAttributes;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoteOpenFile INT 2fh fn 1116h
//
//
//  Open, create, or truncate a file on a remote machine
//
//
//  On entry:
//
//		szCWD				= Current working directory on the remote machine
//		szPath			= Path for the directory on the remote machine
//		uOpenMode		= Desired action (open, create, or truncate)
//    uProcessID     = Process ID on remote machine
//		puRemoteHandle	= Variable for file handle on remote machine
//    puTime         = Variable for file time in DOS packed form
//    puDate         = Variable for file date in DOS packed form
//		pulSize        = Variable for the file size
//		puAttributes   = Variable for the file attributes
//
//  On exit:
//
//  on success,
//
//		puRemoteHandle	= is updated
//    puTime         = Time in DOS packed form
//    puDate         = Date in DOS packed form
//		pulSize        = is updated
//		puAttributes   = is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoteOpenFile( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath, unsigned uOpenMode,
		unsigned uProcessID, unsigned * puRemoteHandle,	unsigned * puTime,
		unsigned * puDate, ulong * pulSize, unsigned * puAttributes )
{
	unsigned uError;
	PREQ1116 pRequest;
	PSERV1116 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1116;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );
	pRequest->uOpenMode = uOpenMode;
	pRequest->uProcessID = uProcessID;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1116) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1116) )
	{
		return ERR_BADFUNC;
	}


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*puRemoteHandle = pService->uRemoteHandle;
		*puTime  = pService->uFileTime;
		*puDate  = pService->uFileDate;
		*pulSize = pService->ulSize;
		*puAttributes = pService->uAttributes;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  RemoteCreateFile INT 2fh fn 1117h
//
//
//  Create or truncate a file on a remote machine
//
//
//  On entry:
//
//		szCWD				= Current working directory on the remote machine
//		szPath			= Path for the directory on the remote machine
//		uAttributes    = File attributes for the file
//    uProcessID     = Process ID on remote machine
//		puRemoteHandle	= Variable for file handle on remote machine
//    puTime         = Variable for file time in DOS packed form
//    puDate         = Variable for file date in DOS packed form
//		pulSize        = Variable for the file size
//		puAttributes   = Variable for the file attributes
//
//  On exit:
//
//  on success,
//
//		puRemoteHandle	= is updated
//    puTime         = Time in DOS packed form
//    puDate         = Date in DOS packed form
//		pulSize        = is updated
//		puAttributes   = is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL RemoteCreateFile( uchar ucRemoteDriveLetter,
		char _D_ * szCWD, char _D_ * szPath, unsigned uAttributes,
		bool fCreateNew, unsigned uProcessID, unsigned * puRemoteHandle,
		unsigned * puTime, unsigned * puDate, ulong * pulSize,
		unsigned * puAttributes )
{
	unsigned uError;
	PREQ1117 pRequest;
	PSERV1117 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1117;
	pRequest->fService   = FALSE;
	pRequest->ucDriveLetter = ucRemoteDriveLetter;
	_fstrcpy( &(pRequest->szCWD[0]), szCWD );
	_fstrcpy( &(pRequest->szPath[0]), szPath );
	pRequest->uAttributes = uAttributes;
	pRequest->fCreateNew = fCreateNew;
	pRequest->uProcessID = uProcessID;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1117) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1117) )
	{
		return ERR_BADFUNC;
	}


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*puRemoteHandle = pService->uRemoteHandle;
		*puTime  = pService->uFileTime;
		*puDate  = pService->uFileDate;
		*pulSize = pService->ulSize;
		*puAttributes = pService->uAttributes;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  SetRemoteFilePosition  INT 2fh fn 1121h
//
//
//  Set the position of a file on a remote machine
//
//
//  On entry:
//
//		uRemoteHandle 			= File handle on remote machine
//		ulPositionFromEnd		= Desired position in file from the end
//		pulResultingPosition	= Variable for the file position after the call
//
//  On exit:
//
//  on success,
//
//		pulResultingPosition	= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL SetRemoteFilePosition( unsigned uRemoteHandle,
		ulong ulPositionFromEnd, ulong * pulResultingPosition )
{
	unsigned uError;
	PREQ1121 pRequest;
	PSERV1121 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1121;
	pRequest->fService   = FALSE;
	pRequest->uRemoteHandle = uRemoteHandle;
	pRequest->ulPositionFromEnd = ulPositionFromEnd;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1121) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1121) )
	{
		return ERR_BADFUNC;
	}


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*pulResultingPosition = pService->ulResultingPosition;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  ReadRemoteFile  INT 2fh fn 1108h
//
//
//  Read from a file on a remote machine
//
//
//  On entry:
//
//		uRemoteHandle	= File handle on remote machine
//		uNumberOfBytes	= Number of bytes to read
//    ulFilePosition = Position in the file to begin reading
//		lpBuffer			= Address of buffer
//		puByteRead		= Variable for the number of bytes actually read
//
//  On exit:
//
//  on success,
//
//		puByteRead	= is updated
//		lpBuffer		= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL ReadRemoteFile( unsigned uRemoteHandle, ulong ulFilePosition,
		unsigned uNumberOfBytes, void _D_ * lpBuffer, unsigned * puBytesRead )
{
	unsigned uError;
	PREQ1108 pRequest;
	PSERV1108 pService;
	unsigned uBytesRemaining;
	uchar far * lpDest;
	unsigned uReadSize;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Normalize the destination address
	//
	lpDest = lpBuffer;


	//  Continue requesting reads (of a smaller size) until
	//  the functions requested read size has been performed
	//
	uBytesRemaining = uNumberOfBytes;
	do
	{
		//  Build the request packet
		//
		pRequest->uFunction  = 0x1108;
		pRequest->fService   = FALSE;
		pRequest->uRemoteHandle = uRemoteHandle;
		pRequest->ulFilePosition = ulFilePosition;
		if( uBytesRemaining > MAX_RW_SIZE )
		{
			uReadSize = MAX_RW_SIZE;
		}
		else
		{
			uReadSize = uBytesRemaining;
		}
		pRequest->uNumberOfBytes = uReadSize;


		//  Send the request to the server.
		//
		uError = ServiceRequest( pRequest, sizeof(REQ1108) );
		if( uError != NET_SERVICED )
		{
			if( uError == NET_TIMEOUT )
			{
				return uError;
			}
			else
			{
				return ERR_BADFUNC;
			}
		}


		//  Make sure that the call was handled
		//
		if( (pService->fService != TRUE) || (pService->uFunction != 0x1108) )
		{
			return ERR_BADFUNC;
		}


		//  Determine if there was an error;
		//
		if( pService->uErrorCode )
		{
			break;
		}


		//  Update the number of bytes remaining to read and the
		//  file position
		//
		uBytesRemaining -= pService->uBytesRead;
		ulFilePosition += pService->uBytesRead;


		//  Copy the data
		//
		_fmemcpy( lpDest, &(pService->ucData[0]), pService->uBytesRead );
		lpDest += pService->uBytesRead;


		//  Determine if the end of file was encountered in the read
		//
		if( (uReadSize < pService->uBytesRead) || (pService->uBytesRead == 0) )
		{
			break;
		}


	}while( uBytesRemaining );


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*puBytesRead = uNumberOfBytes - uBytesRemaining;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  WriteRemoteFile  INT 2fh fn 1109h
//
//
//  Write to a file on a remote machine
//
//
//  On entry:
//
//		uRemoteHandle	= File handle on remote machine
//		uNumberOfBytes	= Number of bytes to write
//    ulFilePosition = Position in the file to begin writing
//		lpBuffer			= Address of buffer
//		puBytesWritten	= Variable for the number of bytes actually written
//
//  On exit:
//
//  on success,
//
//		puBytesWritten	= is updated
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL WriteRemoteFile( unsigned uRemoteHandle,
		ulong ulFilePosition, unsigned uNumberOfBytes, void _D_ * lpBuffer,
		unsigned * puBytesWritten )
{
	unsigned uError;
	PREQ1109 pRequest;
	PSERV1109 pService;
	unsigned uBytesRemaining;
	uchar far * lpDest;
	unsigned uWriteSize;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Normalize the destination address
	//
	lpDest = lpBuffer;


	//  Continue requesting reads (of a smaller size) until
	//  the functions requested read size has been performed
	//
	uBytesRemaining = uNumberOfBytes;
	do
	{
		//  Build the request packet
		//
		pRequest->uFunction  = 0x1109;
		pRequest->fService   = FALSE;
		pRequest->uRemoteHandle = uRemoteHandle;
		pRequest->ulFilePosition = ulFilePosition;
		if( uBytesRemaining > MAX_RW_SIZE )
		{
			uWriteSize = MAX_RW_SIZE;
		}
		else
		{
			uWriteSize = uBytesRemaining;

		}
		pRequest->uNumberOfBytes = uWriteSize;


		//  Copy the data
		//
		_fmemcpy( &(pRequest->ucData[0]), lpDest, pRequest->uNumberOfBytes );


		//  Send the request to the server.
		//
		uError = ServiceRequest( pRequest, sizeof(REQ1109) - MAX_RW_SIZE +
				pRequest->uNumberOfBytes );
		if( uError != NET_SERVICED )
		{
			if( uError == NET_TIMEOUT )
			{
				return uError;
			}
			else
			{
				return ERR_BADFUNC;
			}
		}


		//  Make sure that the call was handled
		//
		if( (pService->fService != TRUE) || (pService->uFunction != 0x1109) )
		{
			return ERR_BADFUNC;
		}


		//  Determine if there was an error;
		//
		if( pService->uErrorCode )
		{
			break;
		}


		//  Update the data address
		//
		lpDest += pService->uBytesWritten;


		//  Update the number of bytes remaining to write and the
		//  file position
		//
		uBytesRemaining -= pService->uBytesWritten;
		ulFilePosition += pService->uBytesWritten;


		//  Determine if the end of file was encountered in the read
		//
		if( (uWriteSize < pService->uBytesWritten) ||
				(pService->uBytesWritten == 0) )
		{
			break;
		}

	}while( uBytesRemaining );


	//  Update the return variables
	//
	if( !pService->uErrorCode )
	{
		*puBytesWritten = uNumberOfBytes - uBytesRemaining;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  CommitRemoteFile  INT 2fh fn 1107h
//
//
//  Commit a file on a remote machine
//
//
//  On entry:
//
//		uRemoteHandle   = File handle on remote machine
//    fUpdateDateTime = Boolean for whether to update the date and time
//    uTime           = Time in DOS packed form
//    uDate           = Date in DOS packed form
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL CommitRemoteFile( unsigned uRemoteHandle,
		bool fUpdateDateTime, unsigned uTime, unsigned uDate )
{
	unsigned uError;
	PREQ1107 pRequest;
	PSERV1107 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1107;
	pRequest->fService   = FALSE;
	pRequest->uRemoteHandle = uRemoteHandle;
	pRequest->fUpdateDateTime = fUpdateDateTime;
	pRequest->uFileTime = uTime;
	pRequest->uFileDate = uDate;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1107) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1107) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}


//  CloseRemoteFile  INT 2fh fn 1106h
//
//
//  Close a file on a remote machine
//
//
//  On entry:
//
//		uRemoteHandle = File handle on remote machine
//    fUpdateDateTime = Boolean for whether to update the date and time
//    uTime           = Time in DOS packed form
//    uDate           = Date in DOS packed form
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned PASCAL CloseRemoteFile( unsigned uRemoteHandle,
		bool fUpdateDateTime, unsigned uTime, unsigned uDate,
		bool fDuplicateHandleAndClose )
{
	unsigned uError;
	PREQ1106 pRequest;
	PSERV1106 pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1106;
	pRequest->fService   = FALSE;
	pRequest->uRemoteHandle = uRemoteHandle;
	pRequest->fUpdateDateTime = fUpdateDateTime;
	pRequest->uFileTime = uTime;
	pRequest->uFileDate = uDate;
	pRequest->fDuplicateHandleAndClose = fDuplicateHandleAndClose;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ1106) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1106) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}



//  CloseAllRemoteFiles  INT 2fh fn 111dh
//
//
//  Close all files on a remote machine for local process termination
//
//
//  On entry:
//
//		uProcessID = The identifier process that is terminating
//
//  On exit:
//
//  on success,
//
//  	function returns 0
//
//  on error,
//
//		function returns the DOS error code describing the error
//
//
unsigned CloseAllRemoteFiles( unsigned uProcessID )
{
	unsigned uError;
	PREQ111D pRequest;
	PSERV111D pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x111d;
	pRequest->fService   = FALSE;
	pRequest->uProcessID = uProcessID;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQ111D) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x111D) )
	{
		return ERR_BADFUNC;
	}


	//  Return any error code
	//
	return pService->uErrorCode;
}



//  ========================================================================
//
//  The remainder of this file is specific to the implementation.
//
//  ========================================================================




//  GetGlobalRequestPacketAddress
//
//
//  Return the address of the data space used to form packets
//
//
//  Note:  This data space must be able to hold the maximum
//         possible packet size
//
//
//  On entry:
//
//    Nothing
//
//  On exit:
//
//  	function returns = The address of the data space for a packet
//
//
void * GetGlobalRequestPacketAddress( void )
{
	static uchar ucPacket[REQ_PACKET_SIZE];
	return ucPacket;
}


#if 0

//  Debugging only
//
void PeekAtFunctionNumbers( unsigned uFunction )
{
	PREQPEEK pRequest;
	extern CALLREG CallersRegisters;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = uFunction;
	pRequest->fService   = FALSE;
	pRequest->uIP = CallersRegisters.ip;
	pRequest->uCS = CallersRegisters.cs;
	pRequest->uAX = CallersRegisters.ax;
	pRequest->uBX = CallersRegisters.bx;
	pRequest->uCX = CallersRegisters.cx;
	pRequest->uDX = CallersRegisters.dx;
	pRequest->uSI = CallersRegisters.si;
	pRequest->uDI = CallersRegisters.di;
	pRequest->uDS = CallersRegisters.ds;
	pRequest->uES = CallersRegisters.es;


	//  Send the request to the server.
	//
	ServiceRequest( pRequest, sizeof(REQSHORT) );
}

#endif  //  #if 0


//  GetDriveMapList
//
//
//  Return the address of the remote drive list array
//
//
//  On entry:
//
//    ppRemoteList = Address of a variable that will contain the
//                   address of the remote list array
//
//  On exit:
//
//		function returns the DOS error code describing the error
//
//
unsigned GetDriveMapList( PREMOTELIST * ppRemoteList )
{
	unsigned uError;
	PREQMAP  pRequest;
	PSERVMAP pService;


	//  Set the address where the request structure will be formed
	//  and serviced
	//
	pRequest = GetGlobalRequestPacketAddress();
	pService = GetGlobalRequestPacketAddress();


	//  Build the request packet
	//
	pRequest->uFunction  = 0x1100;
	pRequest->fService   = FALSE;


	//  Send the request to the server.
	//
	uError = ServiceRequest( pRequest, sizeof(REQMAP) );
	if( uError != NET_SERVICED )
	{
		if( uError == NET_TIMEOUT )
		{
			return uError;
		}
		else
		{
			return ERR_BADFUNC;
		}
	}


	//  Make sure that the call was handled
	//
	if( (pService->fService != TRUE) || (pService->uFunction != 0x1100) )
	{
		return ERR_BADFUNC;
	}


	//  Set the address of the remote list in the return variable
	//
	*ppRemoteList = pService->DriveList;


	//  Return any error code
	//
	return pService->uErrorCode;
}


//
//
bool InstallationCheck( void )
{
	unsigned uInstallCheck;
	void far * lpDriveMap;


	asm{
		push	bx
		push	si
		push	di

		mov	ax, 01131h
		mov	bx, 0bbbbh
		mov	cx, 0cccch
		mov	dx, 0ddddh
		int	2fh
		mov	uInstallCheck, ax

		pop	di
		pop	si
		pop	bx
	};


	if( uInstallCheck == 0x44ff )
	{
		asm{
			push	bx
			push	si
			push	di
			push	es

			mov	ax, 01131h
			mov	bx, 0ddddh
			mov	cx, 0cccch
			mov	dx, 0bbbbh
			int	2fh
			mov	word ptr [lpDriveMap], si
			mov	word ptr [lpDriveMap+2], es

			pop	es
			pop	di
			pop	si
			pop	bx
		};

		_fmemcpy( &DriveMap, lpDriveMap, uDriveMapSize );

		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


//
//
void Int2fUnload( void )
{
	bool fCanUninstall;
	extern _psp;


	asm{

		//  Call the previous iteration to prepare itself to be uninstalled
		//
		push	bx
		push	si
		push	di
		push	ds
		pushf
		mov	ax, 01131h
		mov	bx, 0ddddh
		mov	cx, 0ddddh
		mov	dx, 0ddddh
		int	2fh
		popf
		pop	ds

		//  Save the flag indicating whether it is safe to uninstall
		//  the previous iteration
		//
		mov	fCanUninstall, dx

		pop	di
		pop	si
		pop	bx
	};


	if( fCanUninstall )
	{
		//  Display the uninstall message
		//
		printf("Client unloaded successfully.\n");
	}
	else
	{
		//  Display the can't uninstall message
		//
		printf("Client is disabled but cannot be unloaded.\n");
	}
}


//  UnloadClient
//
//
//  Invalidate all the drives that were mapped.  Unhook the int 2fh
//  vector.  Then terminate the TSR while returning to the int 2fh caller.
//
//
//  On entry:
//
//    nothing
//
//
//  On exit:
//
//    nothing
//
//
void UnloadClient( void )
{
	bool fSuccessful;
	LPPSP lpPSP;
	LPPSP lpCurrentPSP;
	unsigned uCurrentPSP;
	unsigned far * lpCallersStack;
	extern CALLREG CallersRegisters;
	extern _psp;


	//  Invalidate all the IFS drives
	//
	UnmapDrives();


	//  Uninitialize the "network"
	//
	NetworkTXRXDeInit();


	//  Uninstall the Int 2fh handler
	//
	fSuccessful = RestoreIFSVector();
	if( fSuccessful )
	{
		//  Set flag indicating that we can be unloaded
		//
		CallersRegisters.dx = 1;


		//  Get the address of the callers stack
		//
		lpCallersStack = MK_FP( CallersRegisters.ss, CallersRegisters.sp );


		//  Get the current PSP
		//
		asm	push	bx
		asm	mov	ah, 51h
		asm	int	21h
		asm	mov	uCurrentPSP, bx
		asm	pop	bx


		//  Set the address of where to go after termination
		//
		lpPSP = MK_FP( _psp, 0 );
		lpPSP->last_psp = uCurrentPSP;
		lpPSP->term_adr = MK_FP(lpCallersStack[1],lpCallersStack[0]);


		//  Set the callers stack for termination
		//
		lpCallersStack  = &(lpCallersStack[3]);
		lpCallersStack -= 12;
		lpCallersStack[0] = CallersRegisters.ax;
		lpCallersStack[1] = CallersRegisters.bx;
		lpCallersStack[2] = CallersRegisters.cx;
		lpCallersStack[3] = CallersRegisters.dx;
		lpCallersStack[4] = CallersRegisters.si;
		lpCallersStack[5] = CallersRegisters.di;
		lpCallersStack[6] = CallersRegisters.bp;
		lpCallersStack[7] = CallersRegisters.ds;
		lpCallersStack[8] = CallersRegisters.es;
		lpCallersStack[9]  = 0;  //  Flags DOS will set
		lpCallersStack[10] = 0;  //  CS DOS will set
		lpCallersStack[11] = 0;  //  IP DOS will set


		//  Set the stack address for after termination
		//
		lpCurrentPSP = MK_FP( uCurrentPSP, 0 );
		lpCurrentPSP->user_stack = lpCallersStack;


		//  Set the current process to this iteration
		//
		asm	mov	ah, 50h
		asm	mov	bx, _psp
		asm	int	21h


		//  Terminate
		//
		asm	mov	ax, 4c00h
		asm	int	21h
	}
	else
	{
		//  Set flag indicating that we cannot be unloaded because
		//  the int 2fh vector has been changed on top of us
		//
		CallersRegisters.dx = 0;
	}
}
