//
//   $Id: DEFMAIN.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DEFMAIN.C
// Main module for the Datalight disk DEFRAGmentation utility.
// Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//
// This module contains the C _main() and the major data structure
// allocation and interface routines.
//
#define	__FILENUM__	1


// !!!
//
// what #includes do we really need in all these .C files?
//
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <alloc.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>

#include "dostruct.h"
#include "dosver.h"

#include "datalght.h"
#include "fatworld.h"
#include "dlpath.h"
#include "dlapp.h"
#include "abortf32.h"

#include "defrag.h"
#define	DEFRAG	1
#include "defmsg.h"		// contents may eventually be added to DOSMSG.C

#include "defutil.pt"
#include "defutila.pt"
#include "defintr.pt"
#include "deferr.pt"
#include "defchkdk.pt"
#include "defdirs.pt"
#include "defragf.pt"
#include "defcmpct.pt"
#include "bitsets.pt"
#include "defopts.pt"
#include "cpuinfo.pt"

// !!!
//
// what do we really need for a stack?
//
unsigned _stklen = 0xFFE0;

// !!!
//
// min free disk space we need to run, should we track FAT chain
// lengths or just WARN that we can't handle a file
//
#define	MINIMUM_FREE_CLUSTERS	3

// defdirs.c
//
extern ulong dirsSkipped;

// defragf.c
//
extern ulong filesSkipped;

// defchkdk.c
//
extern int percentFragmented;
extern ulong discontinuitiesFixed, discontinuitiesAtStart;

// defopts.c
// user specifiable options
//
extern unsigned char driveLetter;
extern ushort DOSDrive;
extern bool compactingTheDisk, rebootingWhenDone, directoriesOnly;
extern bool getUserConfirmation;

// pointers to "undocumented" DOS data structures
//
LPDPB lpDPB;


// global variables
//
bool	 isBig;								// TRUE if drive isabigun (>64K of sectors)
ushort totalClusters, 	 				// on this disk
		 maxClusterIndex, 				// max index for BackwardFAT/FAT
		 FAT16, 								// !0  if 16-bit FAT
		 freeClusters,						// on this disk
		 bytesPerSector,					// for this disk
		 sectorsPerCluster,		 		// for this disk
		 DIREntriesPerSector;			// for this disk

ushort partialSectorMask, 				// AND mask
		 partialClusterSectorsMask;	// AND mask
ulong  partialClusterBytesMask;		// AND mask

ushort bytesToSectors,		 			// shift count
 		 sectorsToClusters, 				// shift count
		 bytesToClusters,					// shift count
		 DIREntriesToSectors;         // shift count

ushort dataStart, 						// first data sector
		 FATSize, 							// sectors in each FAT
		 rootDIREntries,		 			// on this disk
		 FATStart, 							// first sector of 1st. FAT
		 rootDIRStart;		 				// first sector of root dir

ushort firstFree, 						// first free cluster
		 lastFree;		 					// last free cluster

bool useLongFileNames;	// are long file names a problem this time?

ushort huge *FAT;				// our local copy of the File Allocation Table
ushort huge *BackwardFAT; 	// our free list and Backward File Allocation Table

bool fIsFat32World;

// This routine looks at the specified disk and decides if we can defragment
// it. If there are no files the disk is defragged and we exit. If we cannot
// defragment the disk then we exit with the appropriate error code.
// If this routine returns, then all of the DOS data structures and limit
// variables will have been initialized.
//
// ->	driveLetter	DOS drive letter in upper case
// <- void	Will not return if drive we cannot defragment the drive
//
void PASCAL CheckOutDisk(char driveLetter)
{ FIND ffblk;
  short result;
  char path[16] = "Z:\\*.*";
  ushort setBITs;

	SHIPASSERT(driveLetter >= 'A');
	SHIPASSERT(driveLetter <= 'Z');

	DOSDrive  = driveLetter - 'A';				// get DOS drive number

	// check for free clusters on the drive. If 0 (full) or -1 (unknown)
	// or drive is invalid then we can't use this drive.
	//
	GetFreeSpaceOnDrive((uchar)(DOSDrive+1),&sectorsPerCluster,&freeClusters,
								&bytesPerSector, &totalClusters);
	if	(sectorsPerCluster == 0xFFFF)
	{
		FatalError(EL_BAD_DISK);
	}

	maxClusterIndex 	= totalClusters +1; // location of last cluster in FAT

	if((freeClusters == 0) || (freeClusters == 0xFFFF))
	{
		FatalError(EL_ENODISK);
	}
	isBig = (((ulong)totalClusters * (ulong)sectorsPerCluster) > 65535L) ?
				TRUE : FALSE;

	// Both sectorsPerCluster and bytesPerSector are guaranteed to be
	// a power of two. Otherwise DOS will consider the drive invalid.
	// The C '>> n' operators quickly perform (div 2^n) and the
	// n's may be added together to perform multiple divisions.
	//
	bytesToSectors = 0;
	partialSectorMask = bytesPerSector -1;
	setBITs = partialSectorMask;
	while(setBITs)
	{
		setBITs >>= 1;
		++bytesToSectors;		// get SHR count for (div bytesPerSector)
	}
	sectorsToClusters  = 0;
	partialClusterSectorsMask = sectorsPerCluster -1;
	partialClusterBytesMask = (ulong)(sectorsPerCluster)*
									  (ulong)(bytesPerSector)-1L;
	setBITs = sectorsPerCluster -1;
	while(setBITs)
	{
		setBITs >>= 1;
		++sectorsToClusters; // get SHR count for (div sectorsPerCluster)
	}
	bytesToClusters = bytesToSectors + sectorsToClusters;

	// Each GEN_DIR entry is 20h = 2^5 bytes long
	// Since they are a power of 2 we can use shift instead of div here too
	//
	DIREntriesToSectors = 0;
	DIREntriesPerSector = bytesPerSector >> DIREntriesToBytes;
	setBITs = DIREntriesPerSector -1;
	while(setBITs)
	{
		setBITs >>= 1;
		++DIREntriesToSectors;
	}

	lpDPB = GetDPBFromDOSDrive(DOSDrive);		// and DPB

	// These ptrs have to be valid...
	//
	if(lpDPB == 0L)
	{
		FatalError(EL_BAD_DISK);
	}
	lpDPB->rebuild = 0xFF;	// make sure we get up to date info

	// Trivial success: If disk is empty then it is defragmented.
	//
	path[0] = driveLetter;
	SetDTA(&ffblk);
	result = FindFirst(path,~(A_DEVICE | A_VOLUME | A_HIDDEN));
	if (!result)
	{
		Display(drive_compacted_msg);
		exit(EL_SUCCESS);
	}
	else
	{
		dataStart 			= lpDPB->data_start;
		FATStart				= lpDPB->fat_start;
		FATSize				= lpDPB->fat_len;
		rootDIRStart		= lpDPB->dir_start;
		rootDIREntries		= lpDPB->dir_ent;

		FAT16 				= (lpDPB->hiclus > HICLUS_OF_12_BIT_FAT);
	}
	return;
}


// Reads the clusterIndex'th entry from the FAT maintained in RAM. This knows
// about 12-bit FATs
//
// ->	clusterIndex index of the FAT entry to read: 2 <= clusterIndex <= 0xFFEF
// <- value of the FAT entry
//
// Note: if a 12-bit FAT value greater than 0xFF6, ReadFATInRam() returns
// 		the high nybble of the returned word set to 1s. This way,
//			bad cluster marks, FAT_EOF marks, etc. work the same for both 12
//			and 16 bit FATs.
//
ushort PASCAL ReadFATInRAM(ushort clusterIndex)
{ ushort k, value;
  char huge *pEntry;

#if DEBUG
  static ushort lastCluster = 0;
	if(clusterIndex < 2)
	{
		printf("\nLast cluster index: %X, this cluster index: %X\n",
				  lastCluster,clusterIndex);
	}
	lastCluster = clusterIndex;
#endif

	SHIPASSERT(clusterIndex >= 2);

	if(FAT16)
	{
		SHIPASSERT(clusterIndex <= 0xFFEF);

		return FAT[clusterIndex];
	}
	else
	{
		SHIPASSERT(clusterIndex <= 0xFEF);

		// byte offset = n+(n/2) = 3n/2
		//
		k  = clusterIndex >> 1;
		k += clusterIndex;
		pEntry = (char huge *)(FAT);
		pEntry += k;
		value = *(ushort huge *)(pEntry);

		if((clusterIndex & 1) == 1)
			value >>= 4;
		value &= 0xFFF;

		// return high nybble set if value is bad sector, FAT_EOF, etc.
		//
		if(value > 0xFF6)
			value |= 0xF000;

		return value;
	}
}


// Writes value to the clusterIndex'th entry to the FAT maintained in RAM.
// This knows about 12-bit FATs.
//
// ->	clusterIndex   index of the FAT entry to write:
//				2 <= clusterIndex <= 0x(F)FEF
//    value	value to write at FAT[clusterIndex], not range checked but
//       	masked off to fit in a twelve bit variable as needed.
// <-	FALSE if the clusterIndex'th cluster was free before this call AND IF
//				the sector was not previously dirty.
//    TRUE  if the FAT is significantly inconsistent with the
//				FAT on disk. Copying clusters onto previously free
//				clusters is not considered a significant inconsistency
//				since, even if we lose power, no data will be lost.
//				In the case of 12-bit FAT entries that cross sector
//				boundaries, this will return TRUE if either FAT is
//				inconsistent.
//
bool PASCAL WriteFATInRAM(ushort clusterIndex, ushort value)
{ ushort k, oldValue;
  bool wasNotFree, wasAlreadyDirty;
  char huge *pEntry;

	SHIPASSERT(clusterIndex >= 2);

	wasAlreadyDirty = FALSE;
	if(FAT16)
	{
		SHIPASSERT(clusterIndex <= 0xFFEF);

		wasNotFree = FAT[clusterIndex];
		FAT[clusterIndex]   = value;
	}
	else
	{
		SHIPASSERT(clusterIndex <= 0xFEF);

		value &= 0xFFF;

		// byte offset = n+(n/2) = 3n/2
		//
		k  = clusterIndex >> 1;
		k += clusterIndex;
		pEntry = (char huge *)(FAT);
		pEntry += k;
		oldValue = *(ushort huge *)(pEntry);

		// Update the entry, checking to see if the cluster was
		// free before this write operation.
		//
		if((clusterIndex & 1)==0)
		{
			wasNotFree  = oldValue & 0x0FFF;
			oldValue   &= 0xF000;
		}
		else
		{
			value     <<= 4;
			wasNotFree  = oldValue & 0xFFF0;
			oldValue   &= 0x000F;
		}
		value |= oldValue;
		*(ushort huge *)(pEntry) = value;
	}

	// If the cluster was not free before then the sector is dirty now.
	//
	if(wasNotFree)
		wasAlreadyDirty = FATEntryIsDirty(clusterIndex, FAT16);

	// Return TRUE if data in RAM is significantly inconsistent with the disk.
	//
	return(wasNotFree || wasAlreadyDirty);
}


// This writes any dirty FAT sectors out to disk. A bit-set is maintained
// by the '386 specific code in BITSETS.ASM and there is an "IsDirty" bit
// for each sector of the FAT.
//
// -> secondFAT: TRUE if writing to 2nd FAT else 1st FAT is updated.
//               If secondFAT is TRUE and the routine is successful
//					  then all the FAT sectors are marked clean.
// <-	TRUE if successful
//
bool PASCAL WriteDirtyFATSectors(bool secondFAT)
{ ushort i, first, count, firstFATSector;
  bool  anyDirty = FALSE;
  char huge *lpFirstFATSectorInRAM;

	for(i = 0; i < FATSize; i++)
		anyDirty |= IsThisFATSectorDirty(i);

	if(!anyDirty)
		return TRUE;

	firstFATSector = (secondFAT) ? FATStart + FATSize : FATStart;

	// Write out each dirty sector, write multiple sectors when we can.
	//
	for(i=0, first = 0xFFFF, count = 0; i <= FATSize; i++)
	{
		if(i < FATSize)
		{
			if(IsThisFATSectorDirty(i))
			{
				if(first == 0xFFFF)
				{
					first = i;
					lpFirstFATSectorInRAM  = (char huge *)(FAT);
					lpFirstFATSectorInRAM += i * (long)(bytesPerSector);
				}
				++count;
			}
			else if(count > 0)
			{
				if(!DefragAbsWrite(DOSDrive,count,(long)(firstFATSector+first),
										(void far*)(lpFirstFATSectorInRAM)))
					return FALSE;

				first = 0xFFFF;
				count = 0;
			}
		}
		else if(count > 0)
		{
			if(!DefragAbsWrite(DOSDrive,count,(long)(firstFATSector+first),
									(void far*)(lpFirstFATSectorInRAM)))
				return FALSE;

			first = 0xFFFF;
			count = 0;
		}
	}

	if(secondFAT)
		AllFATSectorsAreClean();

	return TRUE;
}


// Allocates a single cluster from the front of the cluster
// number ordered free list. The free list is kept in a second
// copy of the FAT and uses the same linked list scheme as
// the cluster allocation chains in the FAT. We also use this
// second copy of the FAT to maintain a backlinked list of each
// cluster chain in the FAT.
//
// ->	void
// <- cluster number of the new cluster.
//
ushort PASCAL GetLeastFree(void)
{ ushort newCluster;

	if(!((firstFree >= 2) && (firstFree <= 0xFFEF)))
	{
		// out of disk space
		//
		return 0;
	}

	newCluster = firstFree;
	firstFree  = BackwardFAT[firstFree];

	return newCluster;
}


// This attempts to allocate size consecutive clusters on the
// disk. If a sufficient block is found, it is removed from the
// free list but no other action is taken (the FAT and BackwardFAT are not
// otherwise updated). If a block of size clusters cannot be
// found, this returns 0 and sets size to the maxmimum size block
// of consecutive clusters available. Otherwise the cluster number
// of the first cluster in the block is returned.
//
// ->	size				number of consecutive clusters needed
// 	leastCluster   least possible cluster number in block
// <-	0					on failure (size holds max block size) else
//		cluster number of first cluster number in newly allocated block
//
ushort PASCAL GetFirstFit(ushort *size)
{ ushort blockStart, maxClusters, count, pred, succ, skipper;

	ASSERT(*size > 0);
	ASSERT(*size < (maxClusterIndex -2));

	if(!((firstFree >= 2) && (firstFree <= 0xFFEF)))
	{
		// out of disk space
		//
		*size = 0;
		return 0;
	}

	pred    		= firstFree;	// the running BackwardFAT index
	skipper 		= pred;			// the last free cluster before the block
	blockStart  = pred;			// the first free cluster in the block

	count       = 1;				// max and count are 1
	maxClusters = 1;

	// until we run out of free cluster or find a sufficient block
	//
	while((pred < lastFree) && (count < *size))
	{
		succ = BackwardFAT[pred];			// find the next free cluster

		if(succ == (pred+1))		// if pred and succ are adjacent
		{
			++count;								// count contiguous free clusters
			if(count > maxClusters)
				maxClusters = count;			// remember max
		}
		else
		{
			count 	  = 1;		// start count over
			blockStart = succ;	// next possible block start is at next free
			skipper    = pred;	// this cannot be part of the block
		}
		pred = succ;				// move up the list
	}

	if(count < *size)
		blockStart = 0;			// if didn't find a sufficient free block
	else
	{
		if(blockStart == firstFree)
			// no pred clusters in list
			//
			firstFree    = BackwardFAT[blockStart+count-1];
		else
			BackwardFAT[skipper] = BackwardFAT[blockStart+count-1];

		pred = blockStart;
		succ = pred+1;

		BackwardFAT[pred] = FAT_EOF;

		while(pred < blockStart+count-1)
		{
			BackwardFAT[succ] = pred;
			WriteFATInRAM(pred++,succ++);
		}

		WriteFATInRAM(pred,FAT_EOF);
	}

	*size = maxClusters;
	return  blockStart;
}


// Adds a cluster to the BackwardFAT free list.
//
// ->	clusterIndex	cluster number of the cluster to free.
//
void PASCAL AddToFreeList(ushort clusterIndex)
{
	SHIPASSERT(clusterIndex >= 2);
	ASSERT(clusterIndex <= 0xFFEF);

	if(clusterIndex < firstFree)
	{
		BackwardFAT[clusterIndex]  = firstFree;
		firstFree = clusterIndex;
	}
	else if(clusterIndex > lastFree)
	{
		BackwardFAT[clusterIndex]  = FAT_EOF;
		BackwardFAT[lastFree] = clusterIndex;
		lastFree = clusterIndex;
	}
	else
	{ ushort pred, succ;

		pred = firstFree;

		for(succ = BackwardFAT[pred];
			 succ <= lastFree;
			 pred = succ, succ = BackwardFAT[pred])
		{
			if(succ < clusterIndex)
				continue;

			BackwardFAT[pred]   = clusterIndex;
			BackwardFAT[clusterIndex] = succ;
			break;
		}
	}
	WriteFATInRAM(clusterIndex,FREE_CLUSTER);
}


// Makes a FAT chain free in our internal data structures.
//
// -> clusterIndex is the starting cluster of the allocated chain
//
void PASCAL FreeFATChainInRAM(ushort clusterIndex)
{ ushort pred, succ;

	ASSERT(clusterIndex >= 2);
	ASSERT(clusterIndex <= 0xFFEF);

	pred = clusterIndex;
	do
	{
		succ = ReadFATInRAM(pred);
		AddToFreeList(pred);
		pred = succ;

	} while(!IS_FAT_EOF(succ));
}


// Initialize the BackwardFAT (backward allocation table). This array also
// holds the free list.
//
void PASCAL	initBackwardFAT(void)
{ushort i, value;

	ASSERT(BackwardFAT);

	// Set the entire BackwardFAT array to -1. The first cluster in any file
	// will never change from this value. The free clusters in the FAT
	// will be formed into a list in increasing sector order.
	//
	for(i = 2; i <= maxClusterIndex; i++)
		BackwardFAT[i] = 0xFFFF;        				// no huge memset :-(

	for(i = 2; i <= maxClusterIndex; i++)
	{
		value = ReadFATInRAM(i);
		switch(value)
		{
		// free clusters
		//
		case 0:
			if(firstFree == 0)
			{
				firstFree = i;			// start free list
				lastFree = i;
			}
			BackwardFAT[lastFree] = i;		// append to free list
			BackwardFAT[i] = FAT_EOF;
			lastFree = i;
			break;

		//
		// Note: if a 12-bit FAT value is 0xFF7 or above, ReadFATInRam() returns
		// 		the high nybble of the returned word set to 1s. This way,
		//			bad clusters, FAT_EOF marks, etc. work the same for both 12
		//			and 16 bit FATs.
		//

		// bad clusters
		//
		case 0xFFF7:
			BackwardFAT[i] = value;
			break;

		// end of file or subdir
		//
		case 0xFFF8:
		case 0xFFF9:
		case 0xFFFA:
		case 0xFFFB:
		case 0xFFFC:
		case 0xFFFD:
		case 0xFFFE:
		case 0xFFFF:
			break;		// back link already recorded ignore it

		// reserved FAT values
		//
		case 0xFFF0:
		case 0xFFF1:
		case 0xFFF2:
		case 0xFFF3:
		case 0xFFF4:
		case 0xFFF5:
		case 0xFFF6:
			//
			// Fall through. MS-DOS' documentation says these aren't cluster
			// numbers. MS-DOS will try to use the above as cluster numbers
			// but Drew G reports it doesn't handle them properly.

		// some value between 2 and 0xFFEF indicating a forward link to the
		// next cluster in the file. Write the current cluster number in
		// BackwardFAT location specified by the forward link in the FAT.
		// This forms a backlink to the current cluster.
		//
		default:
			BackwardFAT[value] = i;
		}
	}
}


short oldDisk;			// current drive when we started up
char oldCWD[261];		// current dir    "   "    "     "

// restores the original default drive and directory settings "atexit"
// also track dynamic allocations and free them on exit.
//
void RestoreCWD(void)
{
	// restore the old Current Working DIR
	//
	SetDrive(oldDisk);
	ChangeDir(oldCWD);
}


void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(signon_msg,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon_msg,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

// the main program
//
short main(int argc, char **argv)
{ long BackwardFATSize;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	DisplaySignOnMessage();

	// DEFRAG uses '386 bit op instructions
	//
	//
	if(!Is32BitCPU())
	{
		FatalError(EL_BAD_CPU);
	}
	else if(InWindows())
	{
		// Cannot run inside Windows
		//
		FatalError(EL_WINDOWS_RUNNING);
	}

	// Must we worry about long file names?
	//
	useLongFileNames = SupportsLongFileNames();
	if(useLongFileNames)
	{
		FatalError(EL_LONG_NAMES);
	}

	// process user input
	//
	GetOpts(argc, argv);

	// Can't have open files on drive to be defragged.
	//
	if(OpenFiles())
	{
		// Cannot handle open files
		//
		FatalError(EL_OPEN_FILES);
	}

	// Install our ^C, DOS crit err and SIGxxx handlers.
	//
	HandleErrors();

	// make sure we can use this disk
	//
	ResetDrive(DOSDrive);
	CheckOutDisk(driveLetter);

	//  If the drive is a FAT32 (means we are on a FAT32 kernel) drive then
	//  we want to error out for not supporting FAT32 drives.
	//
	AbortProgramIfFAT32Drive( (driveLetter - 'A' + 1) );

	// remember the current DIR and setup to restore
	// the current drive/dir on program exit()
	//
	GetDrive(&(uchar)(oldDisk));
	GetDir((uchar)(oldDisk),oldCWD);
	atexit(RestoreCWD);
	yield(0);

	// change to the root of the target drive
	//
	SetDrive((uchar)(DOSDrive));
	ChangeDir("\\");
	yield(0);

	// !!!
	//
	// On this first cut we want the whole FAT in memory at all times.
	// This could take up a little less than 128K of contiguous RAM.
	// If we want to do background DEFRAGs later, we'll have to account
	// for FATs that aren't in memory and reread them as neccessary.
	//
	FAT = (ushort huge*)DefragAlloc((ulong)(FATSize * 512L), 1L);
	yield(0);

	// Read the entire FAT off the disk and into our huge FAT array.
	//
	DefragAbsRead(DOSDrive, FATSize, FATStart, (void*)FAT);
	yield(0);

	// Run the CHKDSK code on the disk but refuse to fix anything
	//
	if(ReportedDiskErrors())
	{
		// Whoa! Don't try to DEFRAG a disk with errors...
		//
		exit(EL_BAD_DISK);
	}

	if(percentFragmented < MIN_DEFRAG_PERCENT)
	{
		if(!AreInBatch())
		{
			if(getUserConfirmation)
			{
				if(!YesOrNo(defrag_anyway_msg))
				{
					rebootingWhenDone = FALSE;
					goto ALL_DONE;
				}
			}
		}
	}

	// Only ReadFATInRAM and WriteFATInRam know about 12-bit FATs and we
	// *really* want to keep it that way. If we have a 12-bit FAT on the disk
	// allocate 3/2 as many bytes as there are in the FAT to contain the
	// back-links and the free list. (3/2 * 512 == 768)
	// Total = (number of sectors in FAT) * BackwardFATSize <= 128K
	//
	BackwardFATSize = (FAT16) ? 512L : 768L;
	BackwardFAT=(ushort huge*)DefragAlloc((ulong)(FATSize*BackwardFATSize),1L);
	yield(0);

	// Build the backlinks and the freelist.
	//
	initBackwardFAT();
	yield(0);

	// We need to allocate a big buffer to do file I/O in
	//
	InitFileIO();
	yield(0);

	//
	// Optimize directories
	//
	OptimizeDirectories();
	printf(dirsoptimum_msg);

	if(directoriesOnly)
		goto ALL_DONE;

	yield(0);
	//
	// Defrag the disk's files
	//
	DoDefrag();
	printf(filesdefragged_msg);
	yield(0);


#if 0
	//
	// Compact the disk
	//
	if(compactingTheDisk)
	{
		DoCompacting();
		printf(drive_compacted_msg);
	}
#endif


ALL_DONE:
	// flush all our buffers and reset the drive so changes are sure to take
	//
	ResetDrive(DOSDrive);

#if DEBUG
	if(dirsSkipped  ||  filesSkipped  ||
	   (discontinuitiesFixed != discontinuitiesAtStart))
	{
		if(dirsSkipped)
			printf("%ld directory trees ",dirsSkipped);
		if(dirsSkipped  &&  filesSkipped)
			printf("and ");
		if(filesSkipped)
			printf("%ld files ",filesSkipped);
		if(dirsSkipped  ||  filesSkipped)
			printf("were skipped.\n");
		printf("%ld of %ld discontinuities were resolved.\n",
				discontinuitiesFixed,discontinuitiesAtStart);
	}
#endif

	if(rebootingWhenDone)
		RebootTheMachine();

	return(EL_SUCCESS);
}
