//
//   $Id: DEFUTIL.C 1.4 2002/10/16 21:45:08Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DEFUTIL.C  Created 03-13-97 Dennis Edwards
//
// Copyright (c) 1989-2002 Datalight, Inc.
//	All Rights Reserved.
//	Datalight Confidential and Proprietary
//
// Disk utility module for the Datalight disk DEFRAGmentation utility.
//
#define	__FILENUM__	2


#include <dos.h>
#include <mem.h>
#include <bios.h>
#include <assert.h>

#include "dostruct.h"

#include "defrag.h"

#include "defutil.pt"


// defopts.c
extern ushort DOSDrive;					// the drive to work on and

// DOS Globals
//
LPLOL lpLOL   			=		0L;
int lastDrive			=		-1;
LPCDS lpfirstCDS  	=		0L;


// Returns a ptr to DOS' "List of Lists"
// See Undocumented DOS, DOSTRUCT.H
//
LPLOL PASCAL GetLOLPtr(void)
{ ushort rSeg;

	// find DOS' List of Lists
	asm {
		mov AX, 5200H
		int 21H
		mov rSeg, ES		// assumes LOL starts at ES:0
	}
	return MK_FP(rSeg, 0);
}

// Retreives lastDrive variable from the List of Lists
//
int PASCAL GetLastDrive(void)
{

	ASSERT(lpLOL != 0L);

	return lpLOL->lastdrive;
}

// Retreives a ptr to the first Current Directory Structure
// from the List of Lists
//
LPCDS PASCAL GetFirstCDSPtr(void)
{

	ASSERT(lpLOL != 0L);

	return lpLOL->first_cds;
}

//	Retreive a pointer to the CDS for the specified drive.
//
//	->	drive:	drive number: {0,1,...,('Z'-'A')}
//	<- far ptr to the CDS for the specified drive or NULL (0L) on error.
//
LPCDS PASCAL GetCDSPtr(int drive)
{

	// make sure drive is in a valid range (drives A-Z)
	ASSERT(drive >= 0 && drive < 26);
	ASSERT(lpfirstCDS != 0L);

	// find the CDS for the given drive
	// if passed end of list,
	if(drive >= lastDrive)
		return 0L;

	// return a pointer to that CDS
	return lpfirstCDS + drive;
}

// For these macros p is a ptr to a CDS
//
#define isinvalid(p)		(!((p)->status & CDS_VALID))
#define isnetwork(p) 	((p)->status & CDS_NETWORK)
#define issubst(p)   	((p)->status & (CDS_JOIN | CDS_SUBST))
#define isCDROM(p)      ((p)->status & CDS_CDROM)

// For this macro, p is a ptr to a DPB
#define isfixed(p)  	((p)->status == HARDMEDIA)


#define ismemory(p)     (0)	// how do we know???
#define isREMSERV(p)    (0)	// how do we know???


//	Retreive a pointer to the DPB for the specified drive.
//
//	->	drive:	drive number: {0,1,...,('Z'-'A')}
//	<- far ptr to the CDS for the specified drive or NULL (0L) on error.
//
LPDPB PASCAL GetDPBPtr(int drive)
{ LPCDS lpCDS;

	lpCDS = GetCDSPtr(drive);
	if(lpCDS == 0L)
		return 0L;
	if(isinvalid(lpCDS) || isnetwork(lpCDS) || issubst(lpCDS))
		return 0L;

	return lpCDS->lpDPB;		// get ptr to the Drive Parameter Block
}

// !!!
//
// STUB! -- see Undoc DOS II for implementation info.
bool PASCAL IsStacked(int drive)
{
	drive = 0;
	return drive;
}
// STUB! -- see Undoc DOS II for implementation info.
bool PASCAL IsDblSpaced(int drive)
{
	drive = 0;
	return drive;
}

// Given a DOS drive number, returns the corresponding Drive
// Parameter Block (DPB).
//
// ->	DOS drive # (0 <= DOSDrive <= 25d)
// <- lpDPB for drive or 0L if not found
//
LPDPB PASCAL GetDPBFromDOSDrive( ushort DOSDrive )
{ LPCDS lpCDS;
  LPDPB lpDPB;

	// Retreive some of DOS' global variables
	//
	lpLOL 		= GetLOLPtr();
	lastDrive	= GetLastDrive();
	lpfirstCDS 	= GetFirstCDSPtr();

	// Get the drive's CDS and use the status field value to
	// disqualify several drive types
	//
	lpCDS = GetCDSPtr(DOSDrive);
	while(lpCDS)				// must have a CDS!
	{
		//
		// We only want to deal with mechanical FAT drives for now
		//

		if(isinvalid(lpCDS) || isnetwork(lpCDS) || issubst(lpCDS))
			break;

		// Don't touch CD-ROMS this year
		//
		if(isCDROM(lpCDS))
			break;

// !!
//
// These next 4 ALWAYS return FALSE right now
//

		// How can we tell?
		//
//		if(ismemory(lpCDS)  || isREMSERV(lpCDS))
//			break;

		// Don't handle compressed drives for now.
		//
//		if(IsStacked(DOSDrive) || IsDblSpaced(DOSDrive))
//			break;

		// OK presumably we're talking about a real mechanical FAT drive,
		// get ptr to the Drive Parameter Block
		//
		lpDPB = lpCDS->lpDPB;
		if((lpDPB == 0L) || ((long)(lpDPB) == -1L))	// if goofy DPB ptr, forget it
			break;

		return lpDPB;
	}
	return 0L;
}

#if 0

Test if DPB.hiclus > 0xFF6 to determine if 12 or 16 bit FAT.

//
// !!! may need to implememnt
//
// IOCTL call int 21h fn 440Dh sub-fn 0860h to get BPS, SPT, HEADS
//


// Given a DOS drive number, returns the corresponding Drive
// Data Table (DDT).
//
// ->	DOS drive # (0 <= DOSDrive <= 25d)
// <- lpDDT for drive or 0L if not found
//
LPDDT PASCAL GetDDTFromDOSDrive( ushort DOSDrive )
{
	union REGS regs;
	struct SREGS sregs;
	LPDDT iosys;

	ASSERT( DOSDrive <= 25 );

	// Make a call to dos and get a pointer to the IO->SYS tables
	//
	regs.x.ax = 0x0803;
	int86x( 0x2F, &regs, &regs, &sregs );

	//	Move the segment and offset of the pointer we need into two
	//	unsigned integers! Then use DOS' MK_FP to create a FAR pointer
	//	to the table!
	//
	iosys = (LPDDT) MK_FP( sregs.ds, regs.x.di);

	// IOSYS had beeter be valid at this point!
	//
	ASSERT( iosys );

	while (1)
	{
		if( iosys->dos_drive == DOSDrive )
		{
			return iosys;
		}

		// 0xFFFF will be in the lower WORD of ->nextTable when
		//	this is the last table entry!
		//
		if((ushort)(iosys->nextptr) == 0xFFFF)
			return 0L;

		iosys = iosys->nextptr;
	}
}
#endif


// Given a file handle, returns a ptr to the corresponding
// System File Table (SFT). The max handle number can be increased
// beyond the normal value of 20 by relocating the handle array.
//
// -> file handle
// <-	lpSFT for file or 0L if it couldn't be found
//
LPSFT PASCAL HandleToSFT(int handle)
{
	LPDCB lpDCB;
	LPSFT lpSFT;
	struct PSP far * lpPSP;
	int SFTIndex;
	uchar far * lpPSPHandleTableAddr;

	ASSERT(_psp  != 0);
	ASSERT(lpLOL != 0L);

	lpPSP 		= ((struct PSP far *)(MK_FP(_psp,0)));

	ASSERT(handle < lpPSP->hcount);

	lpPSPHandleTableAddr = lpPSP->htbl_addr;
	ASSERT(lpPSPHandleTableAddr !=  0L);
	ASSERT((long)(lpPSPHandleTableAddr) != -1L);

	SFTIndex		= (int)(lpPSPHandleTableAddr[handle]);
	lpDCB 		= lpLOL->first_dcb;
	lpSFT 		= 0L;

	while(1)
	{
		ASSERT(lpDCB    != 0L);
		ASSERT(SFTIndex >= 0);

		/* we have found it */
		if(SFTIndex < lpDCB->numsft)
		{
			lpSFT = MK_FP(FP_SEG(lpDCB),lpDCB->sft+SFTIndex);
			break;
		}

		/* check if we are at the end of the list */
		if((ushort)lpDCB->next == 0xFFFF)
			break;

		/* check the next device control block */
		SFTIndex -= lpDCB->numsft;
		lpDCB 	 = lpDCB->next;
	}

	return lpSFT;
}


// Searches through all System File Tables (SFTs) looking for open
// files.
//
// ->	void
// <-	TRUE	if an open file was found
//		FALSE if no files are open
//
bool	OpenFiles(void)
{
	LPDCB lpDCB;
	LPSFT lpSFT;
	int i;

	lpLOL			= GetLOLPtr();

	ASSERT(lpLOL != 0L);

	lpDCB 		= lpLOL->first_dcb;
	lpSFT 		= 0L;

	while(1)
	{
		ASSERT(lpDCB != 0L);

		lpSFT = MK_FP(FP_SEG(lpDCB),lpDCB->sft);

		// Get each SFT in this DCB. If it has any open handles
		// then at least one file is open
		//
		for(i = 0; i < lpDCB->numsft; i++,lpSFT++)
		{
			if(lpSFT->ioctl & SFT_DEVICE)
				continue;

			if(lpSFT->handles)
			{
				if((lpSFT->ioctl & 0x3F) == DOSDrive)
					return TRUE;
			}
		}

		// check if we are at the end of the list
		//
		if((ushort)lpDCB->next == 0xFFFF)
			break;

		// link to the next device control block
		//
		lpDCB 	 = lpDCB->next;
	}
	return FALSE;
}
