;
;   $Id: DEFUTILA.ASM 1.4 2002/10/16 21:45:08Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
;   DEFIO.ASM  Created 03-13-97 Dennis Edwards
;
;   Copyright (c) 1989-2002 Datalight, Inc.
;   All Rights Reserved.
;   Datalight Confidential and Proprietary
;
;   DOS/BIOS I/O module for the Datalight disk DEFRAGmentation utility.
;
;   This module contains a variety of DOS I/O functions listed below.
;   They were collected for a variety of reasons:
;
;	1) There is no RTL support for long file names except
;	   under WIN 32. These routines allow migration to long
;	   file names by using unique names for functions that
;	   could support vectored LFN support in the future.
;	   This capability could be useful both in house and to
;	   our customers.
;
;	2) The Borland RTL is inconsistent from version to
;	   version and simple low level DOS I/O routines now
;	   considered obsolete.
;
;	3) File attributes such as "immediate commit" are not
;	   supported by the RTL. And mixing RTL file access
;	   methods is a questionable practice.
;
;	SupportsLongFileNames
;
;	GetDrive
;	SetDrive
;       GetFreeSpaceOnDrive
;
;	GetDir
;	ChangeDir
;
;	OpenOrCreateFile
;	CloseFile
;	CommitFile
;	DeleteFile
;	ReadFile
;	WriteFile
;	SeekFile
;
;	GetFileAttributes
;	SetFileAttributes
;	GetFileDateAndTime
;	SetFileDateAndTime
;
;	SetDTA
;	FindFirst
;	FindNext
;	Rename
;
;	Each of the above functions performs similarly,
;	returning TRUE if the function was successful and FALSE
;	if it failed. On failure _errno will be set to the DOS
;	error code returned. Any return values are left in
;	variables specified in the call.
;
;	No "X" registers are ever saved.
;
;   Also here are various helper routines:
;
;	KbdInputPending
;	ReadKbd
;	TTYOut
;
;	RebootTheMachine
;
;

include ..\include\datalght.inc

InitModel compact, pascal, use16


; Segment used to address BIOS data area
;
ABS_DATA	segment at 0h

; BIT values for <Ctrl><Alt><Del> keystroke
;
CTRL_SHIFT	=	4
ALT_SHIFT	=	8
DEL_SCAN_CODE	=	83

	; the BIOS keyboard shift state flag
	;
	org	417h
kbd_shift_flag	db	?

ABS_DATA	ends


USE_ENHANCED_KBD_CALLS	= 0	;; set to 10h to use enhanced kbd BIOS calls

	extrn	_errno:word

.Code
	assume	cs:@curseg, ds:@data, es:nothing, ss:nothing


; Determines if long file name support is available.
;
; ->	void
; <-	AX	TRUE (1) if long file names are supported by the DOS
;
comment ~
bool PASCAL SupportsLongFileNames(void)
end_comment ~
ProcBegin	SupportsLongFileNames,public
ProcSetup

	mov	ax, 71A0h
	int	21h
	jc	SLFN_yes
	cmp	ax, 7100h
	je	SLFN_no

   SLFN_yes:
	mov	ax, 1
	jmp	short SLFN_exit

   SLFN_no:
	xor	ax, ax

   SLFN_exit:
ProcEnd


comment ~
bool PASCAL GetDrive(uchar far * lpDrive)
end_comment ~
ProcBegin	GetDrive,public
ProcUses	es
ProcArgs	@@lpDrive:dword
ProcSetup

	mov	ah, 19h
	int	21h
	les	bx, @@lpDrive
	mov	es:[bx], al
	mov	ax, 1

    GDF_exit:
ProcEnd


comment ~
bool PASCAL GetFreeSpaceOnDrive(uchar drive,
				ushort *lpSectorsPerCluster,
				ushort *lpFreeClusters,
				ushort *lpBytesPerSector,
				ushort *lpTotalClusters)
end_comment ~
ProcBegin	GetFreeSpaceOnDrive,public
ProcUses	ds,si
ProcArgs	@@drive:byte,@@lpSPC:dword,@@lpFree:dword,@@lpBPS:dword,@@lpTotal:dword
ProcSetup

	mov	ah, 36h
	mov	dl, @@drive
	int	21h

	lds	si, @@lpSPC
	mov	[si], ax
	lds	si, @@lpFree
	mov	[si], bx
	lds	si, @@lpBPS
	mov	[si], cx
	lds	si, @@lpTotal
	mov	[si], dx
ProcEnd


comment ~
bool PASCAL SetDrive(uchar drive)
end_comment ~
ProcBegin	SetDrive,public
ProcUses	ds
ProcArgs	@@drive:byte
ProcSetup

	mov	ah, 0Eh
	mov	dl, @@drive
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL GetDir(uchar drive, char far *lpPathName)
end_comment ~
ProcBegin	GetDir,public
ProcUses	ds,si
ProcArgs	@@lpPathName:dword, @@drive:byte
ProcSetup

	mov	ah, 47h
	lds	si, @@lpPathName
	mov	dl, @@drive
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL ChangeDir(char far *lpPathName)
end_comment ~
ProcBegin	ChangeDir,public
ProcUses	ds
ProcArgs	@@lpPathName:dword
ProcSetup

	mov	ah, 3Bh
	lds	dx, @@lpPathName
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL OpenOrCreateFile(char far *lpName, ushort openMode,
			     ushort attributes, ushort far * lpAction,
			     short far *lpHandle)
end_comment ~
ProcBegin	OpenOrCreateFile,public
ProcUses	ds,si
ProcArgs	@@lpName:dword, @@openMode:word, @@attributes:word,	@@lpAction:dword,  @@lpHandle:dword
ProcSetup

	mov	ah, 6Ch
	mov	bx, @@openMode
	mov	cx, @@attributes
	lds	si, @@lpAction
	mov	dx, [si]
	lds	si, @@lpName
	int	21h
	jc	OF_err

	lds	bx, @@lpAction
	mov	[bx], cx
	lds	bx, @@lpHandle
	mov	[bx], ax
	mov	ax, 1
	jmp	short OF_exit

    OF_err:
	lds	bx, @@lpHandle
	mov	word ptr [bx], -1
	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0

    OF_exit:
ProcEnd


comment ~
bool PASCAL CloseFile(ushort handle)
end_comment ~
ProcBegin	CloseFile,public
ProcUses	ds
ProcArgs	@@handle:word
ProcSetup

	mov	ah, 3Eh
	mov	bx, @@handle
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL CommitFile(ushort handle)
end_comment ~
ProcBegin	CommitFile,public
ProcUses	ds
ProcArgs	@@handle:word
ProcSetup

	mov	ah, 68h
	mov	bx, @@handle
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL ReadFile(ushort handle, uchar far *lpBuffer, ushort far *lpByteCount)
end_comment ~
ProcBegin	ReadFile,public
ProcUses	ds
ProcArgs	@@handle:word, @@lpBuffer:dword, @@lpByteCount:dword
ProcSetup

	mov	ah, 3Fh
	lds	bx, @@lpByteCount
	mov	cx, [bx]
	mov	bx, @@handle
	lds	dx, @@lpBuffer
	int	21h
	jc    RF_err

	lds	bx, @@lpByteCount
	mov	[bx], ax
	xor	ax, ax

	 RF_err:
	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL WriteFile(ushort handle, uchar far *lpBuffer, ushort far * lpByteCount)
end_comment ~
ProcBegin	WriteFile,public
ProcUses	ds
ProcArgs	@@handle:word, @@lpBuffer:dword, @@lpByteCount:dword
ProcSetup

	mov	ah, 40h
	lds	bx, @@lpByteCount
	mov	cx, [bx]
	mov	bx, @@handle
	lds	dx, @@lpBuffer
	int	21h
	jc	WF_err

	lds	bx, @@lpByteCount
	mov	[bx], ax
	xor	ax, ax

	 WF_err:
	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL DeleteFile(char far *lpName)
end_comment ~
ProcBegin	DeleteFile,public
ProcUses	ds
ProcArgs	@@lpName:dword
ProcSetup

	mov	ah, 41h
	lds	dx, @@lpName
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL SeekFile(ushort handle, long far *lpDistance, uchar from)
end_comment ~
ProcBegin	SeekFile,public
ProcUses	es
ProcArgs	@@handle:word, @@lpDistance:dword, @@from:byte
ProcSetup

	mov	ah, 42h
	mov	al, @@from
	les	bx, @@lpDistance
	mov	dx, word ptr es:[bx]
	mov	cx, word ptr es:[bx][2]
	mov	bx, @@handle
	int	21h
	jc	SF_err

	les	bx, @@lpDistance
	mov	es:[bx],    ax
	mov	es:[bx][2], dx
	mov	ax, 1
	jmp	short SF_exit

    SF_err:
	mov	bx, SEG _errno
	mov	es, bx
	mov	es:_errno, ax
	xor	ax, ax

    SF_exit:
ProcEnd


comment ~
bool PASCAL GetFileAttributes(char far *lpName, ushort far *lpAttributes)
end_comment ~
ProcBegin	GetFileAttributes,public
ProcUses	ds
ProcArgs	@@lpName:dword, @@lpAttributes:dword
ProcSetup

	mov	ax, 4300h
	lds	dx, @@lpName
	int	21h
	jc	GFA_err

	lds	bx, @@lpAttributes
	mov	[bx], cx
	mov	ax, 1
	jmp	short GFA_exit

	 GFA_err:
	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	xor	ax, ax

	 GFA_exit:
ProcEnd


comment ~
bool PASCAL SetFileAttributes(char far *lpName, ushort attributes)
end_comment ~
ProcBegin	SetFileAttributes,public
ProcUses	ds
ProcArgs	@@lpName:dword, @@attributes:word
ProcSetup

	mov	ax, 4301h
	mov	cx, @@attributes
	lds	dx, @@lpName
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd

;
; ---	Set DTA,  Find first/next and Rename
;

comment ~
void PASCAL SetDTA(void far *newDTA)
end_comment ~
ProcBegin	SetDTA,public
ProcUses	ds
ProcArgs	@@newDTA:dword
ProcSetup

	lds	dx, @@newDTA
	mov	ah, 1Ah
	int	21h
ProcEnd


; find first, DTA must have already be set
;
comment ~
bool PASCAL FindFirst(char far *lpPath, ushort attr)
end_comment ~
ProcBegin 	FindFirst,public
ProcUses	ds
ProcArgs 	@@path:dword, @@attr:word
ProcSetup

	mov	cx, @@attr
	lds	dx, @@path
	mov	ah, 4Eh
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


; find next, DTA must have been set and FindFirstFile() been called
;
comment ~
bool PASCAL FindNext(void)
end_comment ~
ProcBegin 	FindNext,public
ProcUses	ds
ProcSetup

	mov	ah, 4Fh
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


comment ~
bool PASCAL	Rename(char far *lpOldName, char far *lpNewName)
end_comment ~
ProcBegin	Rename, public
ProcUses	es,ds,di,si
ProcArgs	@@lpOldName:dword, @@lpNewName:dword
ProcSetup

	mov	ah, 56h
	lds	dx, @@lpOldName
	les	di, @@lpNewName
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


;
; --- 	get/set file date/time
;
comment ~
bool PASCAL GetFileDateAndTime(short handle, ushort *lpDate, ushort *lpTime)
end_comment ~
ProcBegin	GetFileDateAndTime,public
ProcUses	es
ProcArgs	@@handle:word, @@lpDate:dword, @@lpTime:dword
ProcSetup

	mov	ax, 5700h
	mov	bx, @@handle
	int	21h
	jc	GFDAT_err

	les	bx, @@lpDate
	mov	es:[bx], dx
	les	bx, @@lpTime
	mov	es:[bx], cx
	mov	ax, 1
	jmp	short GFDAT_exit

	 GFDAT_err:
	mov	bx, SEG _errno
	mov	es, bx
	mov	es:_errno, ax
	xor	ax, ax

	 GFDAT_exit:
ProcEnd


comment ~
bool PASCAL SetFileDateAndTime(short handle, ushort date, ushort time)
end_comment ~
ProcBegin	SetFileDateAndTime,public
ProcUses	ds
ProcArgs	@@handle:word, @@date:word, @@time:word
ProcSetup

	mov	ax, 5701h
	mov	bx, @@handle
	mov	cx, @@time
	mov	dx, @@date
	int	21h

	mov	bx, SEG _errno
	mov	ds, bx
	mov	_errno, ax
	mov	ax, 0
	cmc
	adc	ax, ax
ProcEnd


;
; --- BIOS I/O routines used by critical error handler and friends ---
;


; checks for pending keyboard input
;
comment ~
bool PASCAL KbdInputPending(void)
end_comment ~
ProcBegin	KbdInputPending,public
ProcSetup

	mov	ah, 1
	or	ah, USE_ENHANCED_KBD_CALLS
	int	16h
	mov	ax, 0
	jz    	CK_exit

	inc	ax

    CK_exit:
ProcEnd


; waits for and returns the next extended scan code from the BIOS kbd buffer
;
comment ~
ushort PASCAL ReadKbd(void)
end_comment ~
ProcBegin 	ReadKbd,public
ProcSetup

	xor	ax, ax
	or	ah, USE_ENHANCED_KBD_CALLS
	int	16h
ProcEnd


; Displays VidChar at the current cursor location on the CRT using
; the BIOS video TTY output function.
;
comment ~
void PASCAL TTYOut(uchar VidChar)
end_comment ~
ProcBegin 	TTYOut,public
ProcArgs        @@VidChar:byte
ProcSetup

	mov	al, @@VidChar
	mov	ah, 0Eh
	int	10h
ProcEnd




end

