//
//   $Id: DELTREE.C 1.9 2002/10/16 21:45:44Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  DELTREE.C
//
//  Copyright (c) 1989-2002 Datalight, Inc.
//	 All Rights Reserved.
//	 Datalight Confidential and Proprietary
//
//  History:
//
//  03-29-02 bt Rewrote
//  05-05-97 de Created
//
#define __FILENUM__ 1


//  Datalight libraries
//
#include "stdc.h"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "dlapp.h"
#include "dlpath.h"
#include "dlerrcod.h"
#include "fatworld.h"
#include "dosver.h"


//  Program message strings
//
#define DELTREE 1
#include "dosmsg.c"


//  Prototypes
//
#include "dldeltre.pt"
#include "deltree.pt"


//  Declaration to change the default stack size
//
unsigned _stklen = 1536;


//  Global string to build file name and path
//
char szTargetSpec[LFN_MAX_PATH];


//  Command line parsing switches.  Global because ARGDEF cant supply
//  addresses of local variables
//
bool notPrompting = FALSE;
bool fHideSignon  = FALSE;


//  Globals for knowing when we are in a FAT32 or LFN environment
//
bool fIsFat32World;
bool fLFNsSupported;


//  Contains a pointer to the Global LFN search name in the datalight
//  libraries
//
char * szLFN;


//  Display the program signon
//
void SignOn(void)
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signOnMsg,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signOnMsg,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Display the signon screen, then the usage screen, then exit
//
void Usage(void)
{
	SignOn();
	printf(usageMsg);
	exit(ERR_PARAMETER_MISSING);
}


//  Get only a 'y' or 'n' and an enter.  Displays the result as well
//
uchar GetKey(void)
{
	uchar ucCharacter1;
	uchar ucReturnCharacter = 0;


	while( TRUE )
	{
		dlGetKey( (char *)&ucCharacter1 );
		ucCharacter1 = tolower( ucCharacter1 );
		if( (ucCharacter1 == 'y') || (ucCharacter1 == 'n') )
		{
			ucReturnCharacter = ucCharacter1;
			printf( "%c\b", ucReturnCharacter );
		}
		else if( (ucReturnCharacter) && (ucCharacter1 == '\r') )
		{
			printf("\n");
			break;
		}
		else if( ucCharacter1 == 0 )
		{
			dlGetKey( (char *)&ucCharacter1 );
		}
	}

	return ucReturnCharacter;
}


//  Displays a question and get yes or no answer.
//  Returns TRUE if the answer was affirmative/yes
//
bool ConfirmDelete(char *question, char *name, char *allowed)
{
	uchar ucResponse;


	printf(question,name,allowed);
	ucResponse = GetKey();
	if( ucResponse == 'y' )
	{
		return TRUE;
	}
	else
	{
		return FALSE;
	}
}


int main(int argc, char *argv[])
{
	static ARGDEF delTreeOpts[] =
	{
		{ BOOL_SWITCH, &notPrompting, 'y' },
		{ BOOL_SWITCH, &notPrompting, 'c' },
		{ BOOL_SWITCH, &fHideSignon,  '[' },
		{ 0 }
	};
	unsigned uIndex;
	bool fSuccessful;
	FIND FindData;
	char * szNamePart;
	char * szTemp;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Determine if LFNs are support
	//
	fLFNsSupported = LFNPresent();


	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	SetParseFlags( PARSEF_NOARGSOK );


	//  Display copyright and version info then process user options.
	//
	argc = ParseCmdline(argc-1, &argv[1], delTreeOpts);


	//  if not at least one path name exit
	//
	if(argc == 0)
	{
		printf( szMissingParms );
		exit(ERR_PARAMETER_MISSING);
	}


	//  Display the program signon message
	//
	SignOn();


	//  Remove any quotes from LFN arguments
	//
	for( uIndex = 1; uIndex <= argc; uIndex++ )
	{
		LFNStripArgQuotes( argv[uIndex] );
		if( strlen(argv[uIndex]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}


	//  Process each file and path specified on the command line
	//
	for( uIndex = 1; uIndex <= argc; uIndex++ )
	{
		//  Copy the name into a local variable so that the string
		//  can be modified
		//
		strcpy(szTargetSpec, argv[uIndex]);


		//  Find the name in the user supplied path
		//
		szNamePart = strrchr( szTargetSpec, '\\' );
		if( szNamePart == NULL )
		{
			szNamePart = strrchr( szTargetSpec, ':' );
			if( szNamePart == NULL )
			{
				szNamePart = szTargetSpec;
			}
			else
			{
				szNamePart++;
				if( *szNamePart == 0 )
				{
					strcat( szNamePart, "*.*" );
				}
			}
		}
		else
		{
			szNamePart++;
			if( *szNamePart == 0 )
			{
				strcat( szNamePart, "*.*" );
			}
		}


		//  The "." and ".." are implied wild-cards but DOS won't
		//  find-first 'em so append the *.* to the name
		//
		szTemp = szNamePart;
		if( *szTemp == '.' )
		{
			while( *szTemp == '.' )
			{
				szTemp++;
			}
			if( *szTemp == 0 )
			{
				strcat( szNamePart, "\\*.*" );
				szNamePart = szTemp + 1;
			}
		}


		//  Search for the first occurance
		//
		fSuccessful = SmartFindFirst( (A_ANYTHING & ~A_VOLUME),
				szTargetSpec, &FindData );
		if(!fSuccessful)
		{
			SmartFindClose( &FindData );
			continue;
		}


		//  Get the address where all the LFN version of the find first name
		//  will go
		//
		if( fLFNsSupported )
		{
			GetSmartFindLFNAddress( &szLFN );
		}


		//  We assume that we need to expand wildcards in the input
		//
		while(fSuccessful)
		{
			//  if so make sure we don't process "." and ".." in subdirs
			//
			if(*(FindData.name) == '.')
			{
				goto NEXT;
			}


			//  Build the name string on the supplied path
			//
			if( fLFNsSupported )
			{
				strcpy( szNamePart, szLFN );
			}
			else
			{
				strcpy( szNamePart, FindData.name );
				strlwr( szNamePart );
			}


			//  use different routines for trees and single files
			//
			if((FindData.attr & A_SUBDIR) == A_SUBDIR)
			{
				//  verify tree removal
				//
				if(!notPrompting)
				{
					if(!ConfirmDelete(dirDeletion,szTargetSpec,ynString))
					{
						goto NEXT;
					}
				}


				//  Display the name to be deleted
				//
				printf(deletingMsg,szTargetSpec);


				//  Clear the attributes that would prevent deleting the file
				//
				if((FindData.attr & (A_HIDDEN|A_SYSTEM|A_READONLY)) != A_NORMAL)
				{
					SmartSetFileAttributes( szTargetSpec, A_NORMAL );
				}


				//  prune the dir tree
				//
				if(!DeleteTree(szTargetSpec))
				{
					printf(cannotRmDirMsg,szTargetSpec);
				}
			}
			else
			{
				//  verify file deletion
				//
				if(!notPrompting)
				{
					if(!ConfirmDelete(fileDeletion,szTargetSpec,ynString))
					{
						goto NEXT;
					}
				}


				//  Display the name to be deleted
				//
				printf(deletingMsg,szTargetSpec);


				//  Clear the attributes that would prevent deleting the file
				//
				if((FindData.attr & (A_HIDDEN|A_SYSTEM|A_READONLY)) != A_NORMAL)
				{
					SmartSetFileAttributes( szTargetSpec, A_NORMAL );
				}


				//  Delete the specified file
				//
				if( !SmartDelete(szTargetSpec) )
				{
					printf(cannotDelFileMsg,szTargetSpec);
				}
			}

NEXT:
			//  Try to find the next wild card match...
			//  (if none specified we'll just drop out of this loop)
			//
			fSuccessful = SmartFindNext( &FindData );
		}


		//  Close the find handle
		//
		SmartFindClose( &FindData );
	}


	return ERR_NONE;
}
