//
//   $Id: DISKCOMP.C 1.5 2002/10/16 21:46:42Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  DISKCOMP.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//	 04-25-2001  aj  Created
//


//  BORLAND INCLUDES
//
#include <dir.h>
#include <stdio.h>


//  DATALIGHT LIBRARIES
//
#include "datalght.h"

#include "dostruct.h"

#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "fatworld.h"
#include "dlapp.h"
#include "dllfn.h"
#include "dlpath.h"


//  Program strings
//
#define DISKCOMP 1
#include "dosmsg.c"


//  Diskcomp header files
//
#include "DISKCOMP.h"


//  PROTOTYPES
//
#include "DISKCOMP.pt"
#include "DISKCOMN.pt"


#define TEMP_DOS_DRIVE_NUMBER		0


//
//  VARIABLES
//


//  Flag indicating that we are using DOS 7.1
//
bool fIsFat32World = FALSE;
bool fInWindows = FALSE;


//  Flags for the command line switches
//  (accessed through macros in DISK.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];


//  Start of DISKCOMP
//
void main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;
	uchar ucFirstDriveNumber = 0;
	uchar ucSecondDriveNumber = 0;
	uchar ucGetChar;
	bool fCopying = TRUE;
	unsigned uDosVersion;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Make sure that the DOS version is at least 5.0 becuase
	//  IOCTL calls were not supported prior to 5.0
	//
	dlKernelSupportVersion( &uDosVersion );
	if( uDosVersion < 0x0500 )
	{
		DisplayErrorAndExit( ERR_NOTDOS5 );
	}


	//  Initialize SaveSegment to 0
	//
	SaveSegment( 0 );


	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1],
			&ucFirstDriveNumber, &ucSecondDriveNumber );


	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();


	//  Install a control break handler so that we can abort cleanly
	//  if the user presses CTRL-C
	//
	InstallCTRLCHandler();


	//  Install a critical error handler so that we can abort cleanly
	//
	InstallInt24Handler();


	//  Perform disk compare
	//
	while( fCopying )
	{
		//  Compare the disks
		//
		ucExitCode = DiskComp( ucFirstDriveNumber+1, ucSecondDriveNumber+1 );
		if( ucExitCode == ERR_NONE )
		{
			//  Determine if the user wants to copy another disk
			//
			while( TRUE )
			{
				printf(szCompPrompt);
				GetKey(&ucGetChar);
				printf("\n");

				if( toupper(ucGetChar) == 'Y' )
				{
					fCopying = TRUE;
					break;
				}

				if( toupper(ucGetChar) == 'N' )
				{
					fCopying = FALSE;
					break;
				}
			}
		}
		else
		{
			break;
		}
	}


	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();


	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();


	//  Set the appropriate error code and exit
	//
	DisplayErrorAndExit( ucExitCode );
}


//
//
uchar DiskComp( uchar ucFirstDriveNumber, uchar ucSecondDriveNumber )
{
	static DRIVE_PARMS DriveParms;
	bool fSuccessful;
	uchar ucErrorCode;
	uchar ucDriveLetter;
	uchar ucGetChar;
	bool fSingleDrive;


	//  Determine if the source and destination drives are the same
	//
	fSingleDrive = IsSingleDrive( ucFirstDriveNumber, ucSecondDriveNumber );


	//  Prompt for first diskette
	//
	ucDriveLetter = GetDriveLetter( ucFirstDriveNumber );
	printf(szInsertSource, ucDriveLetter);
	if(( ucFirstDriveNumber == ucSecondDriveNumber ) ||
			( fSingleDrive == TRUE ))
	{
		printf(szAnyKey);
		GetKey(&ucGetChar);
	}
	else
	{
		ucDriveLetter = GetDriveLetter( ucSecondDriveNumber );
		printf(szInsertTarget, ucDriveLetter);
		printf(szAnyKey);
		GetKey(&ucGetChar);
	}


	//  TODO: Add a check here to make sure we have the right drive number
	//

	//  Get the drive parameters for the first drive
	//
	DriveParms.Parms.specFunc = 1;
	fSuccessful = FALSE;

	while( !fSuccessful )
	{
		fSuccessful = dlGetParametersForBlockDevice( ucFirstDriveNumber,
		  		&DriveParms );
		if( !fSuccessful )
		{
			ErrDiskread( ucFirstDriveNumber );
		}
	}


	//  Build the track layout table
	//
	if( fIsFat32World )
	{
		SetTrackLayout( &DriveParms.uSectorsInTrack, &DriveParms.Parms.bpb );
	}
	else
	{
		SetTrackLayout( &((char *)(&DriveParms))[0x26],
				&DriveParms.Parms.bpb );
	}


	//  Indicate what has changed in the device parameters
	//
	DriveParms.Parms.specFunc = 4;


	//  Set the drive parameters for the destination drive
	//
	fSuccessful = dlSetParametersForBlockDevice( ucSecondDriveNumber,
			&DriveParms );
	if( !fSuccessful )
	{
		return ERR_DISKREAD;
	}

	ucErrorCode = CompDisk( ucFirstDriveNumber, ucSecondDriveNumber,
			DriveParms.Parms.cylinders, DriveParms.Parms.bpb.sec_per_trak,
			DriveParms.Parms.bpb.sides );
	if( ucErrorCode )
	{
		return ucErrorCode;
	}

	return ERR_NONE;
}


uchar CompDisk( uchar ucFirstDriveNumber, uchar ucSecondDriveNumber,
		unsigned uCountCylinders, unsigned uSectorsPerTrack,
		unsigned uCountSides )
{
	static DRIVE_PARMS DriveParms;
	uchar ucErrorCode;
	uchar ucDriveLetter;
	uchar ucGetChar;
	unsigned uSegment;
	unsigned uAvailPara;
	unsigned uMaxTrackComp = 0;
	unsigned uAllocPara;
	unsigned uStartTrack;
	unsigned uMaxSides = uCountSides;
	unsigned uMaxSecs = uSectorsPerTrack;
	ulong ulBytesAvail;
	ulong ulBytesInTrack;
	ulong ulTrackBufferAddress;
	ulong ulBytesInAllHeadsOnTrack;
	bool fSuccessful = FALSE;
	bool fFirstPass = TRUE;
	bool fSingleDrive;
	bool fCompareOk = TRUE;


	//  Set up for command line switches
	//
	if( DoCompOneSide() )
	{
		uMaxSides = 1;
	}
	if( DoCompFirst8() )
	{
		uMaxSecs = 8;
	}


	//  Get the flag that indicates that there is only one physical
	//  drive for two logical drives.
	//
	fSingleDrive = IsSingleDrive( ucFirstDriveNumber, ucSecondDriveNumber );


	//  Determine how much memory is available
	//
	dlAllocMem( 0xffff, &uSegment );
	uAvailPara = uSegment;


	//  Determine how many tracks can be stored
	//  in the available amount of memory
	//
	ulBytesAvail = ((ulong)uAvailPara << 4);
	ulBytesInTrack = ((ulong)uSectorsPerTrack * BYTES_PER_SECTOR);
	ulBytesInAllHeadsOnTrack = ulBytesInTrack * (ulong)uCountSides;
	if( ulBytesInAllHeadsOnTrack )
	{
		uMaxTrackComp = (unsigned)(ulBytesAvail / ulBytesInAllHeadsOnTrack);
	}


	//  Determine if there is enough memory to allocate
	//  at least one track worth of bytes
	//
	if( uMaxTrackComp )
	{
		//  Determine how much memory to allocate
		//
		uAllocPara = (unsigned)((ulBytesInAllHeadsOnTrack *
				uMaxTrackComp) >> 4);


		//  Allocate the memory
		//
		fSuccessful = dlAllocMem( uAllocPara, &uSegment );
		if( !fSuccessful )
		{
			return ERR_ALLOCMEM;
		}


		//  Store the segment of the allocated memory block in a place
		//  where other functions, like CTRL-BREAK handler, have access
		//  to it.
		//
		SaveSegment( uSegment );


		//  Determine the physical memory address of the track buffer
		//
		ulTrackBufferAddress = ((ulong)uSegment << 4);
	}


	//  Display the statistics
	//
	printf(szDisplayStats, uCountCylinders, uMaxSecs, uMaxSides);


	uStartTrack = 0;
	while( uStartTrack < uCountCylinders )
	{
		//  Make sure we aren't going to compare more tracks than
		//  exist on the disk.
		//
		if( (uStartTrack + uMaxTrackComp) >= uCountCylinders)
		{
	  		uMaxTrackComp = uCountCylinders - uStartTrack;
		}

		//  Read a number of tracks
		//
		fSuccessful = ReadWriteANumberOfTracks( IS_READ, uCountSides,
		 		uMaxSecs, uMaxTrackComp, uStartTrack,
	  	  		ucFirstDriveNumber, ulTrackBufferAddress, DoCompOneSide() );
		if( !fSuccessful )
		{
			return ERR_DISKREAD;
		}


		//  Determine if we should swap disks
		//  and swap disks.
		//
		if((ucFirstDriveNumber == ucSecondDriveNumber) ||
				(fSingleDrive))
		{
			SwapDisk( SWAP_SECOND, ucSecondDriveNumber );
		}


		//  Determine if this is the first pass
		//
		if( fFirstPass )
		{
			fFirstPass = FALSE;

			//  Get the drive parameters of the second drive
			//
			DriveParms.Parms.specFunc = 1;
			fSuccessful = FALSE;

			while( !fSuccessful )
			{
				fSuccessful = dlGetParametersForBlockDevice( ucSecondDriveNumber,
	 		  			&DriveParms );
				if( fSuccessful )
				{
					if( 	DriveParms.Parms.cylinders != uCountCylinders 			||
				  		 	DriveParms.Parms.bpb.sec_per_trak != uSectorsPerTrack ||
				  			DriveParms.Parms.bpb.sides != uCountSides )
					{
				  		DisplayError( ERR_NOTCOMPAT );
				  		return ERR_COMPEND;
					}
				}
				else
				{
			  		ErrDiskRead( ucSecondDriveNumber );
				}
			}
		}


		//  Compare a number of tracks
		//
		fSuccessful = CompareANumberOfTracks( uCountSides,
				uSectorsPerTrack, uMaxTrackComp, uStartTrack,
	  	  		ucSecondDriveNumber, ulTrackBufferAddress );
		if( !fSuccessful )
		{
			fCompareOk = FALSE;
		}

		uStartTrack += uMaxTrackComp;


		//  Determine if we should swap disks
		//  and swap disks.
		//
		if((uStartTrack != uCountCylinders) &&
				((ucFirstDriveNumber == ucSecondDriveNumber) || (fSingleDrive)))
		{
			SwapDisk( SWAP_FIRST, ucFirstDriveNumber );
		}
	}

	//  Free the track buffer
	//
	fSuccessful = dlFreeMem( uSegment );
	if( !fSuccessful )
	{
		return ERR_FREEMEM;
	}
	SaveSegment( 0 );


	//  If no differences in compare, print compare OK
	//
	if( fCompareOk )
	{
		printf(szCmpOk);
	}


	//  Return that the diskcomp was successful
	//
	return ERR_NONE;
}


//  Compare a number of tracks
//
bool CompareANumberOfTracks( unsigned uCountSides,
		unsigned uSectorsPerTrack,	unsigned uMaxTracks, unsigned uStartTrack,
		uchar ucDriveNumber, ulong ulTrackBufferAddress )
{
	static IOCTL_READWRITE ParmBlock;
	static uchar pTempTrackBuffer[0x12 * BYTES_PER_SECTOR];
	lpBOOTSECTOR lpBootSector;
	pBOOTSECTOR pBootSector;
	unsigned uTrack;
	unsigned uHead;
	unsigned uMaxSides = uCountSides;
	unsigned uMaxSecs = uSectorsPerTrack;
	unsigned uSegment;
	unsigned uOffset;
	unsigned uDrive = ucDriveNumber;
	unsigned uEndTrack = uStartTrack + uMaxTracks;
	bool fSuccessful;
	bool fSingleDrive;
	uchar ucDriveLetter;
	uchar ucCategoryCode = DEFAULT_CATEGORY_CODE;
	int iError = 0;
	bool fReturn = TRUE;


	//  Find out the real drive number
	//
	if( ucDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDrive = 1;
		}
	}


	//  And the drive letter
	//
	ucDriveLetter = GetDriveLetter( ucDriveNumber );


	//  Determine if the number of sides being copied is 1
	//
	if( DoCompOneSide() )
	{
		uMaxSides = 1;
	}

	if( DoCompFirst8() )
	{
		uMaxSecs = 8;
	}


	//  Setup the IOCTL read parameter block
	//
	ParmBlock.ucSpecialFunction = 0;
	ParmBlock.uSector = 0;
	ParmBlock.uNumberOfSectors = uMaxSecs;


	//  Continue reading tracks until the requested amount has
	//  been read
	//
	for( uTrack = uStartTrack; uTrack < uEndTrack; uTrack++ )
	{
		ParmBlock.uTrack = uTrack;
		for( uHead = 0; uHead < uMaxSides; uHead++ )
		{
			ParmBlock.uHead = uHead;


			//  Normalize pointer for next iteration
			//
			uOffset = (unsigned)(ulTrackBufferAddress & 0xF);
			uSegment = (unsigned)(ulTrackBufferAddress >> 4);
			ParmBlock.lpBuffer = MK_FP( uSegment, uOffset );

			_fmemcpy( pTempTrackBuffer, ParmBlock.lpBuffer,
		  			(uMaxSecs * BYTES_PER_SECTOR) );

			fSuccessful = dlGenericIOCTLForBlockDevices( uDrive,
					ucCategoryCode, IOCTL_READ, &ParmBlock );
			if( !fSuccessful )
			{
				printf(szReadError, ucDriveLetter, uHead, uTrack);
			}

			if( (uHead == 0) && (uTrack == 0) )
			{
				lpBootSector = ParmBlock.lpBuffer;
				pBootSector = (pBOOTSECTOR)(pTempTrackBuffer);

				lpBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber = 0;
				pBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber = 0;
			}

			iError = _fmemcmp( pTempTrackBuffer, ParmBlock.lpBuffer,
					(uMaxSecs * BYTES_PER_SECTOR) );
			if( iError != 0 )
			{
				fReturn = FALSE;
				printf(szCmpError, uHead, uTrack);
			}


			//  Update pointer
			//
			ulTrackBufferAddress += (uMaxSecs * BYTES_PER_SECTOR);
		}
	}

	return fReturn;
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[],
		uchar *ucFirstDriveNumber, uchar *ucSecondDriveNumber )
{
	uchar ucDriveLetter;
	static ARGDEF DISKOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],		'[', 0 },
		{ BOOL_SWITCH, &fSwitches[SINGLE_SIDE],      '1', 0 },
		{ BOOL_SWITCH, &fSwitches[FIRST8_SECTORS],	'8', 0 },
		{ 0 }
	};


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], DISKOpts);


	if( argc != 2 )
	{
		Usage();
	}


	if( (argv[0][1] == ':') && (argv[0][2] == 0) &&
		 (argv[1][1] == ':') && (argv[1][2] == 0))
	{
		ucDriveLetter = toupper(argv[0][0]);
		if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
		*ucFirstDriveNumber = ucDriveLetter - 'A';


		//  Determine if the drive is valid.  If the drive is not valid
		//  then display an error message for drive not valid and exit
		//
		if( !CheckIfCanUseDrive(*ucFirstDriveNumber) )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}


		ucDriveLetter = toupper(argv[1][0]);
		if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
		*ucSecondDriveNumber = ucDriveLetter - 'A';


		//  Determine if the drive is valid.  If the drive is not valid
		//  then display an error message for drive not valid and exit
		//
		if( !CheckIfCanUseDrive(*ucSecondDriveNumber) )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
	}
	else
	{
		Usage();
	}
}


//  Determine if the drive is valid
//
bool CheckIfCanUseDrive( uchar ucDosDriveNumber )
{
	bool fSuccessful;
	bool fIsRemote;
	bool fIsSubsted;
	unsigned uDosVersion;


	//  Determine if the drive is local or remote.  If the drive
	//  is not local that the drive is not valid, otherwise
	//  return that it is valid
	//
	fSuccessful = dlIsDriveRemote( ucDosDriveNumber+1, &fIsRemote );
	if( !fSuccessful )
	{
		return FALSE;
	}
	if( fIsRemote )
	{
		return FALSE;
	}


	//  Determine if the drive is substed.  If the drive
	//  substed, the drive is not valid. Otherwise, return
	//  that it is valid
	//
	fSuccessful = dlIsDriveSubst( ucDosDriveNumber+1, &fIsSubsted );
	if( !fSuccessful )
	{
		return FALSE;
	}
	if( fIsRemote )
	{
		return FALSE;
	}


	return TRUE;
}


//  Display the usage screen
//
void Usage( void )
{
	DisplaySignonMessage();
	printf( szUsage );
	exit( ERR_USAGE );
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	char * szSignOnMessage = "DISKCOMP %s %s\n%s\n\n";


	if( !DoHideSignon() )
	{
		/* display sign-on */
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Display an error message
//
void DisplayError( uchar ucIndex )
{
	unsigned uArrayEntries;


	//  Make sure that there is an error message string associated
	//  with this index
	//
	uArrayEntries = ARRAY_ENTRIES( szErrorMessages );
	uArrayEntries--;
	if( ucIndex < uArrayEntries )
	{
		//  Display the error message string
		//
		printf(szErrorMessages[ucIndex]);
	}
	else
	{
		printf(szUnkErrorCode, ucIndex );
	}
	printf("\n");
}


//  Display an error message and exit with the appropriate error code
//
void DisplayErrorAndExit( uchar ucErrorCode )
{
	bool fSuccessful;
	unsigned uAllocatedMemorySegment;


	//  Display the error message associated with this error code
	//
	DisplayError( ucErrorCode );


	//  Free the allocated memory block, if there was one
	//
	if( GetSegment() )
	{
		dlFreeMem( GetSegment() );
	}


	//  Uninstall our interrupt handlers
	//
	UninstallInt24Handler();
	UninstallCTRLCHandler();


	//  Terminate the program setting the appropriate exit code
	//
	exit( ucErrorCode );
}


//  Install a control break handler so that we can abort cleanly if
//  the user presses CTRL-C
//
void InstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	SaveOldCtrlCHandler( *lplpOldCtrlCHandler );
	*lplpOldCtrlCHandler = CTRLCHandler;
}


//  Uninstall the control break handler
//
void UninstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	*lplpOldCtrlCHandler = GetOldCtrlCHandler();
}


//  Control break handler
//
void CTRLCHandler( void )
{
	//  Free mem
	//
	if( GetSegment() )
	{
		dlFreeMem( GetSegment() );
	}


	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();

	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();

	//  Exit
	//
	exit( MS_ERR_CTRL_BREAK );
}


//  Return the address of the old control break handler
//
void far * GetOldCtrlCHandler( void )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	return *plpOldCtrlCHandler;
}


//  Save the address of the old control break handler
//
void SaveOldCtrlCHandler( void far * lpOldCtrlCHandler )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	*plpOldCtrlCHandler = lpOldCtrlCHandler;
}


//  Return the address of the variable which contain the old control
//  break handler
//
void far * * GetOldCtrlCHandlerAddress( void )
{
	static void far * lpOldCtrlCHandler;

	return &lpOldCtrlCHandler;
}


//  Int 24 handler
//
#pragma argsused
 void interrupt far Int24Handler(BP, DI, SI, DS, ES, DX, CX, BX, AX,
		IP, CS, FLAGS)
{
	(uchar)AX = 3;
}


//  Install a critical error handler so that we can abort cleanly
//
void InstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;


	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	SaveOldCritHandler( *lplpOldCritHandler );
	*lplpOldCritHandler = (void far *)Int24Handler;
}


//  Uninstall the critical error handler
//
void UninstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;

	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	*lplpOldCritHandler = GetOldCritHandler();
}


//  Return the address of the old critical error handler
//
void far * GetOldCritHandler( void )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	return *plpOldCritHandler;
}


//  Save the address of the old critical error handler
//
void SaveOldCritHandler( void far * lpOldCritHandler )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	*plpOldCritHandler = lpOldCritHandler;
}


//  Return the address of the variable which contain the old critical
//  error handler
//
void far * * GetOldCritHandlerAddress( void )
{
	static void far * lpOldCritHandler;

	return &lpOldCritHandler;
}


//  Save the segment
//
void SaveSegment( unsigned uSegment )
{
	unsigned * puSegment;

	puSegment = GetSegmentPointer();
	*puSegment = uSegment;
}


//  Return the segment
//
unsigned GetSegment( void )
{
	unsigned * puSegment;

	puSegment = GetSegmentPointer();
	return *puSegment;
}


//  Return the address of the variable that contains the segment
//
unsigned * GetSegmentPointer( void )
{
	static unsigned uSegment;
	return &uSegment;
}

