//
//   $Id: DISKCOPY.C 1.11 2002/10/16 21:47:06Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  DISKCOPY.C
//
//  Copyright (c) 2001-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//	 03-01-2001  aj  Created
//


//  DATALIGHT LIBRARIES
//
#include "datalght.h"

#include "dostruct.h"

#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "fatworld.h"
#include "dlapp.h"
#include "dlpath.h"
#include "stdc.h"


//  Program strings
//
#define DISKCOPY 1
#include "dosmsg.c"


//  Diskcopy header files
//
#include "DISKCOPY.h"


//  PROTOTYPES
//
#include "DISKCOPY.pt"
#include "DISKCOMN.pt"


#define TEMP_DOS_DRIVE_NUMBER		0


//
//  VARIABLES
//


//  Flag indicating that we are using DOS 7.1
//
bool fIsFat32World = FALSE;
bool fInWindows = FALSE;


//  Flags for the command line switches
//  (accessed through macros in DISK.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];

//  Swap filename
//
char szSwapName[30];


//  Start of DISKCOPY
//
void main( int argc, char *argv[] )
{
	uchar ucExitCode = ERR_NONE;
	uchar ucSrcDriveNumber = 0;
	uchar ucDestDriveNumber = 0;
	uchar ucGetChar;
	bool fCopying = TRUE;
	unsigned uDosVersion;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();


	//  Make sure that the DOS version is at least 5.0 becuase
	//  IOCTL calls were not supported prior to 5.0
	//
	dlKernelSupportVersion( &uDosVersion );
	if( uDosVersion < 0x0500 )
	{
		DisplayErrorAndExit( ERR_NOTDOS5 );
	}


	//  Initialize SaveSegment to 0
	//
	SaveSegment( 0 );


	//  Initialize SaveHandle to 0
	//
	SaveHandle( 0 );


	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1],
			&ucSrcDriveNumber, &ucDestDriveNumber );


	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();


	//  Install a control break handler so that we can abort cleanly
	//  if the user presses CTRL-C
	//
	InstallCTRLCHandler();


	//  Install a critical error handler so that we can abort cleanly
	//
	InstallInt24Handler();


	//  Perform diskcopy
	//
	while( fCopying )
	{
		//  Copy the disk
		//
		ucExitCode = DiskCopy( ucSrcDriveNumber+1, ucDestDriveNumber+1 );
		if( ucExitCode == ERR_NONE )
		{
			//  Determine if the user wants to copy another disk
			//
			while( TRUE )
			{
				printf(szCopyPrompt);
				GetKey(&ucGetChar);
				printf("\n");

				if( toupper(ucGetChar) == 'Y' )
				{
					fCopying = TRUE;
					break;
				}

				if( toupper(ucGetChar) == 'N' )
				{
					fCopying = FALSE;
					break;
				}
			}
		}
		else
		{
			break;
		}
	}


	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();


	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();


	//  Set the appropriate error code and exit
	//
	DisplayErrorAndExit( ucExitCode );
}


//
//
uchar DiskCopy( uchar ucSrcDriveNumber, uchar ucDestDriveNumber )
{
	static DRIVE_PARMS DriveParms;
	bool fSuccessful;
	uchar ucErrorCode;
	uchar ucDriveLetter;
	uchar ucGetChar;
	bool fSingleDrive;
	unsigned uSwapFileHandle = 0;
	static char szUnique[30];


	strcpy(szUnique, "C:\\");
	//  Determine if the source and destination drives are the same
	//
	fSingleDrive = IsSingleDrive( ucSrcDriveNumber, ucDestDriveNumber );


	//  If this is not a memory only copy, make sure we can
	//  open a swap file.
	//
	if ( !DoMemOnly() && fSingleDrive )
	{
		fSuccessful = dlCreateUniqueFile( A_NORMAL, szUnique, &uSwapFileHandle );
		if( fSuccessful )
		{
			SaveHandle( uSwapFileHandle );
			strcpy(szSwapName, szUnique);
		}
		else
		{
			printf( szMemOnly );
		}
	}


	//  Prompt for source diskette
	//
	ucDriveLetter = GetDriveLetter( ucSrcDriveNumber );
	printf(szInsertSource, ucDriveLetter);
	if(( ucSrcDriveNumber == ucDestDriveNumber ) ||
			( fSingleDrive == TRUE ))
	{
		printf(szAnyKey);
		GetKey(&ucGetChar);
	}
	else
	{
		ucDriveLetter = GetDriveLetter( ucDestDriveNumber );
		printf(szInsertTarget, ucDriveLetter);
		printf(szAnyKey);
		GetKey(&ucGetChar);
	}


	//  Get the drive parameters for the source drive
	//
	DriveParms.Parms.specFunc = 1;
	fSuccessful = FALSE;

	while( !fSuccessful )
	{
		fSuccessful = dlGetParametersForBlockDevice( ucSrcDriveNumber,
		  		&DriveParms );
		if( !fSuccessful )
		{
			ErrDiskread( ucSrcDriveNumber );
		}
	}

	//  Get the disk info
	//
	GetInfo( ucSrcDriveNumber );


	//  Build the track layout table
	//
	if( fIsFat32World )
	{
		SetTrackLayout( &DriveParms.uSectorsInTrack, &DriveParms.Parms.bpb );
	}
	else
	{
		SetTrackLayout( &((char *)(&DriveParms))[0x26],
				&DriveParms.Parms.bpb );
	}


	//  Indicate what has changed in the device parameters
	//
	DriveParms.Parms.specFunc = 4;


	//  Set the drive parameters for the destination drive
	//
	fSuccessful = dlSetParametersForBlockDevice( ucDestDriveNumber,
			&DriveParms );
	if( !fSuccessful )
	{
		return ERR_DISKWRITE;
	}

	ucErrorCode = CopyDisk( ucSrcDriveNumber, ucDestDriveNumber,
			DriveParms.Parms.cylinders, DriveParms.Parms.bpb.sec_per_trak,
			DriveParms.Parms.bpb.sides );
	if( ucErrorCode )
	{
		return ucErrorCode;
	}

	return ERR_NONE;
}


uchar CopyDisk( uchar ucSrcDriveNumber, uchar ucDestDriveNumber,
		unsigned uCountCylinders, unsigned uSectorsPerTrack,
		unsigned uCountSides )
{
	static DRIVE_PARMS DriveParms;
	uchar ucErrorCode;
	uchar ucDriveLetter;
	uchar ucGetChar;
	unsigned uSegment;
	unsigned uAvailPara;
	unsigned uMaxTrackCopy = 0;
	unsigned uAllocPara;
	unsigned uStartTrack;
	unsigned uSwapFileHandle;
	unsigned uMaxSides;
	ulong ulBytesAvail;
	ulong ulBytesInTrack;
	ulong ulTrackBufferAddress;
	ulong ulBytesInAllHeadsOnTrack;
	bool fSuccessful = FALSE;
	bool fFirstPass = TRUE;
	bool fCopy = TRUE;
	bool fSingleDrive;



	uSwapFileHandle = GetHandle();

 	uMaxSides = uCountSides;
 	if( DoCopyOneSide() )
	{
		uMaxSides = 1;
	}

	//  Get the flag that indicates that there is only one physical
	//  drive for two logical drives.
	//
	fSingleDrive = IsSingleDrive( ucSrcDriveNumber, ucDestDriveNumber );


	//  Determine how much memory is available
	//
	dlAllocMem( 0xffff, &uSegment );
	uAvailPara = uSegment;


	//  Determine how many tracks can be stored
	//  in the available amount of memory
	//
	ulBytesAvail = ((ulong)uAvailPara << 4);
	ulBytesInTrack = ((ulong)uSectorsPerTrack * BYTES_PER_SECTOR);
	ulBytesInAllHeadsOnTrack = ulBytesInTrack * (ulong)uCountSides;
	if( ulBytesInAllHeadsOnTrack )
	{
		uMaxTrackCopy = (unsigned)(ulBytesAvail / ulBytesInAllHeadsOnTrack);
	}


	//  Determine if there is enough memory to allocate
	//  at least one track worth of bytes
	//
	if( uMaxTrackCopy )
	{
		//  Determine how much memory to allocate
		//
		uAllocPara = (unsigned)((ulBytesInAllHeadsOnTrack *
				uMaxTrackCopy) >> 4);
		if( (uSwapFileHandle) && (fSingleDrive) )
		{
			uAllocPara = (unsigned)(ulBytesInTrack >> 4);
		}


		//  Allocate the memory
		//
		fSuccessful = dlAllocMem( uAllocPara, &uSegment );
		if( !fSuccessful )
		{
			return ERR_ALLOCMEM;
		}


		//  Store the segment of the allocated memory block in a place
		//  where other functions, like CTRL-BREAK handler, have access
		//  to it.
		//
		SaveSegment( uSegment );


		//  Determine the physical memory address of the track buffer
		//
		ulTrackBufferAddress = ((ulong)uSegment << 4);
	}


	//  Display the statistics
	//
	printf(szDisplayStats, uCountCylinders, uSectorsPerTrack, uMaxSides);


	if( (!uSwapFileHandle) || (!fSingleDrive) )
	{
		uStartTrack = 0;
		while( uStartTrack < uCountCylinders )
		{
			//  Make sure we aren't going to copy more tracks than
			//  exist on the disk.
			//
			if( (uStartTrack + uMaxTrackCopy) >= uCountCylinders)
			{
		  		uMaxTrackCopy = uCountCylinders - uStartTrack;
			}

			//  Read a number of tracks
			//
			fSuccessful = ReadWriteANumberOfTracks( IS_READ, uCountSides,
			 		uSectorsPerTrack, uMaxTrackCopy, uStartTrack,
	  		  		ucSrcDriveNumber, ulTrackBufferAddress, DoCopyOneSide() );
			if( !fSuccessful )
			{
				return ERR_DISKREAD;
			}


			//  Determine if we should swap disks
			//  and swap disks.
			//
			if((ucSrcDriveNumber == ucDestDriveNumber) ||
					(fSingleDrive))
			{
				SwapDisk( SWAP_DEST, ucDestDriveNumber );
			}


			//  Determine if this is the first pass
			//
			if( fFirstPass )
			{
				fFirstPass = FALSE;

				//  Get the drive parameters of the Target
				//
				DriveParms.Parms.specFunc = 1;
 				fSuccessful = dlGetParametersForBlockDevice( ucDestDriveNumber,
		 		 		&DriveParms );
				if( fSuccessful )
				{
					if( 	DriveParms.Parms.cylinders != uCountCylinders 			||
						 	DriveParms.Parms.bpb.sec_per_trak != uSectorsPerTrack ||
							DriveParms.Parms.bpb.sides != uCountSides )
					{
						DisplayError( ERR_NOTCOMPAT );
						return ERR_COPYEND;
					}
				}
				else
				{
			 		//  Format disk
			 		//
					printf(szFormatting);
  			  		ucErrorCode = FormatDisk( ucDestDriveNumber, uCountCylinders,
	  						uCountSides );
					if( ucErrorCode )
					{
						return ERR_FORMAT;
					}
				}
			}


			//  Write a number of tracks
			//
			fSuccessful = ReadWriteANumberOfTracks( IS_WRITE, uCountSides,
					uSectorsPerTrack, uMaxTrackCopy, uStartTrack,
	  		  		ucDestDriveNumber, ulTrackBufferAddress, DoCopyOneSide() );
			if( !fSuccessful )
			{
				return ERR_DISKWRITE;
			}
			else
			{
				uStartTrack += uMaxTrackCopy;
			}


			//  Determine if we should swap disks
			//  and swap disks.
			//
			if((uStartTrack != uCountCylinders) &&
					((ucSrcDriveNumber == ucDestDriveNumber) || (fSingleDrive)))
			{
				SwapDisk( SWAP_SRC, ucSrcDriveNumber );
			}
  		}

		//  Display the serial number of the disk that was just copied
		//
		printf(szDisplaySerial,
				(unsigned)((GetSerialNumber() & 0xFFFF0000L) >> 16),
			  	(unsigned)((GetSerialNumber() & 0x0000FFFFL)) );
	}
	else
	{
		//  Read disk to swap
		//
		fSuccessful = ReadDiskToSwap( uCountCylinders, uCountSides,
				uSectorsPerTrack, ucSrcDriveNumber, ulTrackBufferAddress );
		if( !fSuccessful )
		{
			return ERR_DISKREAD;
		}

		//  Determine if we should swap disks
		//  and swap disks.
		//
		if((ucSrcDriveNumber == ucDestDriveNumber) ||
				(fSingleDrive))
		{
			SwapDisk( SWAP_DEST, ucDestDriveNumber );
		}


		//  Determine if this is the first pass
		//
		if( fFirstPass )
		{
			fFirstPass = FALSE;

			//  Get the drive parameters of the Target
			//
 			DriveParms.Parms.specFunc = 1;
 			fSuccessful = dlGetParametersForBlockDevice( ucDestDriveNumber,
		 	 		&DriveParms );
			if( fSuccessful )
			{
				if( 	DriveParms.Parms.cylinders != uCountCylinders 			||
					 	DriveParms.Parms.bpb.sec_per_trak != uSectorsPerTrack ||
						DriveParms.Parms.bpb.sides != uCountSides )
				{
					DisplayError( ERR_NOTCOMPAT );
					return ERR_COPYEND;
				}
			}
			else
			{
				//  Format disk
				//
				printf(szFormatting);
  		  		ucErrorCode = FormatDisk( ucDestDriveNumber, uCountCylinders,
	 					uCountSides );
				if( ucErrorCode )
				{
					return ERR_FORMAT;
				}
			}
		}

		while ( fCopy )
		{
			//  Write disk from swap
			//
			fSuccessful = WriteDiskFromSwap( uCountCylinders, uCountSides,
					uSectorsPerTrack, ucDestDriveNumber, ulTrackBufferAddress );
			if( !fSuccessful )
			{
				return ERR_DISKWRITE;
			}


			//  Display the serial number of the disk that was just copied
			//
			printf(szDisplaySerial,
				(unsigned)((GetSerialNumber() & 0xFFFF0000L) >> 16),
			  	(unsigned)((GetSerialNumber() & 0x0000FFFFL)) );


			//  Prompt to make another copy
			//
			while( TRUE )
			{
				printf(szAnotherDupe);
				GetKey(&ucGetChar);
				printf("\n");

		 		if( toupper(ucGetChar) == 'Y' )
				{
					ucDriveLetter = GetDriveLetter( ucDestDriveNumber );
					printf(szInsertTarget, ucDriveLetter);
					printf(szAnyKey);
					GetKey(&ucGetChar);

					fCopy = TRUE;
					break;
				}


				if( toupper(ucGetChar) == 'N' )
				{
					fCopy = FALSE;
					break;
				}
			}
		}
	}


	//  Free the track buffer
	//
	fSuccessful = dlFreeMem( uSegment );
	if( !fSuccessful )
	{
		return ERR_FREEMEM;
	}
	SaveSegment( 0 );


	//  Close the swapfile handle
	//
	if( uSwapFileHandle )
	{
		fSuccessful = dlCloseFile( (int)uSwapFileHandle );
		if( !fSuccessful )
		{
			return ERR_NOCLOSE;
		}
		SaveHandle( 0 );

		//  Delete the swapfile
		if( szSwapName[0] )
		{
			fSuccessful = dlDeleteFile(szSwapName);
			if( !fSuccessful )
			{
				return ERR_NODEL;
			}
		}
	}


	//  Return that the diskcopy was successful
	//
	return ERR_NONE;
}


//  Verify a disk is formatted
//
bool VerifyFormat( unsigned ucDestDriveNumber, unsigned uCountCylinders,
		unsigned uCountSides, unsigned uSectorsPerTrack )
{
	static IOCTL_READWRITE ParmBlock;
	uchar ucCategoryCode = DEFAULT_CATEGORY_CODE;
	unsigned uDestDrive = ucDestDriveNumber;
	unsigned uTrack;
	unsigned uHead;
	unsigned uMaxSides = uCountSides;
	bool fSuccessful;
	bool fSingleDrive;

	if( ucDestDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDestDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDestDrive = 1;
		}
	}

	if( DoCopyOneSide() )
	{
		uMaxSides = 1;
	}


	ParmBlock.ucSpecialFunction = 2;
	ParmBlock.uSector = 0;
	ParmBlock.uNumberOfSectors = uSectorsPerTrack;

	for( uTrack = 0; uTrack < uCountCylinders; uTrack++ )
	{
		ParmBlock.uTrack = uTrack;
		for( uHead = 0; uHead < uMaxSides; uHead++ )
		{
			ParmBlock.uHead = uHead;
			fSuccessful = dlGenericIOCTLForBlockDevices( uDestDrive,
					ucCategoryCode, IOCTL_VERIFY, &ParmBlock );
			if( !fSuccessful )
			{
				return fSuccessful;
			}
		}
	}

	return fSuccessful;
}


//  Read disk to swapfile
//
bool ReadDiskToSwap( unsigned uCountCylinders, unsigned uCountSides,
		unsigned uSectorsPerTrack, uchar ucDriveNumber,
		ulong ulTrackBufferAddress )
{
	static IOCTL_READWRITE ParmBlock;
	unsigned uTrack;
	unsigned uHead;
	unsigned uMaxSides = uCountSides;
	unsigned uSegment;
	unsigned uOffset;
	unsigned uBytesWrite;
	unsigned uBytesWritten;
	unsigned uDrive = ucDriveNumber;
	ulong ulTempTrackBufferAddress;
	bool fSuccessful;
	bool fSingleDrive;
	unsigned uSwapFileHandle;

	uchar ucDriveLetter;
	uchar ucCategoryCode = DEFAULT_CATEGORY_CODE;

	long lDistance = 0;


	//
	//
	uSwapFileHandle = GetHandle();


	//  Find out the real drive number
	//
	if( ucDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDrive = 1;
		}
	}


	//  And the driveletter
	//
	ucDriveLetter = GetDriveLetter( ucDriveNumber );


	//  Print reading message
	//
	printf(szReadSrc);


	//  Get Segment & Offset and make a pointer
	//
	uOffset = (unsigned)(ulTrackBufferAddress & 0xF);
	uSegment = (unsigned)(ulTrackBufferAddress >> 4);


	if( DoCopyOneSide() )
	{
		uMaxSides = 1;
	}

	ParmBlock.ucSpecialFunction = 0;
	ParmBlock.uSector = 0;
	ParmBlock.uNumberOfSectors = uSectorsPerTrack;
	ParmBlock.lpBuffer = MK_FP( uSegment, uOffset );

	uBytesWrite = (uSectorsPerTrack * BYTES_PER_SECTOR);
	uBytesWritten = uBytesWrite;


	for( uTrack = 0; uTrack < uCountCylinders; uTrack++ )
	{
		ParmBlock.uTrack = uTrack;

		for( uHead = 0; uHead < uMaxSides; uHead++ )
		{
			ParmBlock.uHead = uHead;


			fSuccessful = dlGenericIOCTLForBlockDevices( uDrive,
					ucCategoryCode, IOCTL_READ, &ParmBlock );
			if( !fSuccessful )
			{
				printf(szReadError, ucDriveLetter, uHead, uTrack);
			}

 			//  Write to swap
 			//
  			fSuccessful = dlWriteFile( (int)uSwapFileHandle,
					&uBytesWritten, ParmBlock.lpBuffer);
			if ( !fSuccessful )
			{
				return FALSE;
			}
			if ( uBytesWritten != uBytesWrite )
			{
				return FALSE;
			}
		}
	}


	dlCommitFile( (int)uSwapFileHandle );


	//  Reset the swapfile position
	//
	fSuccessful = dlSeekFile(SEEK_SET, (int)uSwapFileHandle, &lDistance);
	if( !fSuccessful )
	{
		return FALSE;
	}

	return TRUE;
}


//  Write disk from swapfile
//
bool WriteDiskFromSwap( unsigned uCountCylinders, unsigned uCountSides,
		unsigned uSectorsPerTrack, uchar ucDestDriveNumber,
		ulong ulTrackBufferAddress )
{
	static IOCTL_READWRITE ParmBlock;
	lpBOOTSECTOR lpBootSector;
	unsigned uTrack;
	unsigned uHead;
	unsigned uMaxSides = uCountSides;
	unsigned uSegment;
	unsigned uOffset;
	unsigned uBytesReadRef;
	unsigned uBytesRead;
	unsigned uDestDrive = ucDestDriveNumber;
	ulong ulTempTrackBufferAddress;
	bool fSuccessful;
	bool fSingleDrive;
	unsigned uSwapFileHandle;
	int iError;
	uchar ucDriveLetter;
	uchar ucCategoryCode = DEFAULT_CATEGORY_CODE;
	static uchar pTempTrackBuffer[0x12 * BYTES_PER_SECTOR];
	long lDistance = 0;


	//
	//
	uSwapFileHandle = GetHandle();


	//  Find out the real drive number
	//
	if( ucDestDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDestDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDestDrive = 1;
		}
	}
	//  And the drive letter
	//
	ucDriveLetter = GetDriveLetter( ucDestDriveNumber );

	//  Print message
	//
	printf(szWriteTarg);

	//  Get Segment & Offset and make a pointer
	//
	uOffset = (unsigned)(ulTrackBufferAddress & 0xF);
	uSegment = (unsigned)(ulTrackBufferAddress >> 4);


	if( DoCopyOneSide() )
	{
		uMaxSides = 1;
	}

	ParmBlock.ucSpecialFunction = 0;
	ParmBlock.uSector = 0;
	ParmBlock.uNumberOfSectors = uSectorsPerTrack;
	ParmBlock.lpBuffer = MK_FP( uSegment, uOffset );

	uBytesReadRef = (uSectorsPerTrack * BYTES_PER_SECTOR);
	uBytesRead = uBytesReadRef;


	for( uTrack = 0; uTrack < uCountCylinders; uTrack++ )
	{
		ParmBlock.uTrack = uTrack;

		for( uHead = 0; uHead < uMaxSides; uHead++ )
		{
			ParmBlock.uHead = uHead;

 			//  Read from swap
 			//
  			fSuccessful = dlReadFile( (int)uSwapFileHandle,
					&uBytesRead, ParmBlock.lpBuffer);
			if ( !fSuccessful )
			{
				return FALSE;
			}
			if ( uBytesRead != uBytesReadRef )
			{
				return FALSE;
			}

			//  Write to disk
			//
			if( (uTrack == 0) && (uHead == 0) )
			{
				lpBootSector = ParmBlock.lpBuffer;
				lpBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber =
						CreateNewSerialNumber();
				SaveSerialNumber( lpBootSector->BootSectorInfo.BootInfo16.ulVolumeSerialNumber );
			}

			fSuccessful = dlGenericIOCTLForBlockDevices( uDestDrive,
					ucCategoryCode, IOCTL_WRITE, &ParmBlock );
			if( !fSuccessful )
			{
				printf(szWriteError, ucDriveLetter, uHead, uTrack);
			}

			if( DoVerifyCopy() )
			{
				_fmemcpy( pTempTrackBuffer, ParmBlock.lpBuffer,
						(uSectorsPerTrack * BYTES_PER_SECTOR) );
				fSuccessful = dlGenericIOCTLForBlockDevices( uDestDrive,
						ucCategoryCode, IOCTL_READ, &ParmBlock );
				if( !fSuccessful )
				{
					printf(szReadError, ucDriveLetter, uHead, uTrack);
				}
				iError = _fmemcmp( pTempTrackBuffer, ParmBlock.lpBuffer,
						(uSectorsPerTrack * BYTES_PER_SECTOR) );
				if( iError != 0 )
				{
					printf(szCmpError, ucDriveLetter, uHead, uTrack);
				}
			}
		}
	}

	//  Reset the swapfile position
	//
	fSuccessful = dlSeekFile(SEEK_SET, (int)uSwapFileHandle, &lDistance);
	if( !fSuccessful )
	{
		return FALSE;
	}

	return TRUE;
}


//  Format disk
//
uchar FormatDisk( uchar ucDriveNumber, unsigned uCountCylinders,
		unsigned uCountSides )
{
	unsigned uHead;
	unsigned uTrack;
	uchar ucSpecialFunction = 0;
	bool fSuccessful;
	bool fSingleDrive;
	unsigned uMaxSides;
	unsigned uDestDrive = ucDriveNumber;


	if( ucDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDestDrive = 1;
		}
	}


	uMaxSides = uCountSides;
	if( DoCopyOneSide() )
	{
		uMaxSides = 1;
	}


	for( uTrack = 0; uTrack < uCountCylinders; uTrack++ )
	{
		for( uHead = 0; uHead < uMaxSides; uHead++ )
		{
			fSuccessful = dlIOCTLFormatAndVerifyTrack( uDestDrive, uHead,
					uTrack, ucSpecialFunction, fIsFat32World );
			if( !fSuccessful )
			{
				return ERR_FORMAT;
			}
		}
	}


	return ERR_NONE;
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[],
		uchar *ucSrcDriveNumber, uchar *ucDestDriveNumber )
{
	uchar ucDriveLetter;
	static ARGDEF DISKOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ BOOL_SWITCH, &fSwitches[SINGLE_SIDE],		'1', 0 },
		{ BOOL_SWITCH, &fSwitches[VERIFY_COPY],		'V', 0 },
		{ BOOL_SWITCH, &fSwitches[MEM_ONLY],			'M', 0 },
		{ 0 }
	};


	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], DISKOpts);


	if( argc != 2 )
	{
		Usage();
	}


	if( (argv[0][1] == ':') && (argv[0][2] == 0) &&
		 (argv[1][1] == ':') && (argv[1][2] == 0))
	{
		ucDriveLetter = toupper(argv[0][0]);
		if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
		*ucSrcDriveNumber = ucDriveLetter - 'A';


		//  Determine if the drive is valid.  If the drive is not valid
		//  then display an error message for drive not valid and exit
		//
		if( !CheckIfCanUseDrive(*ucSrcDriveNumber) )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}


		ucDriveLetter = toupper(argv[1][0]);
		if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
		*ucDestDriveNumber = ucDriveLetter - 'A';


		//  Determine if the drive is valid.  If the drive is not valid
		//  then display an error message for drive not valid and exit
		//
		if( !CheckIfCanUseDrive(*ucDestDriveNumber) )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}
	}
	else
	{
		Usage();
	}
}


//  Display an error message
//
void DisplayError( uchar ucIndex )
{
	unsigned uArrayEntries;


	//  Make sure that there is an error message string associated
	//  with this index
	//
	uArrayEntries = ARRAY_ENTRIES( szErrorMessages );
	uArrayEntries--;
	if( ucIndex < uArrayEntries )
	{
		//  Display the error message string
		//
		printf(szErrorMessages[ucIndex]);
	}
	else
	{
		printf(szUnkErrorCode, ucIndex );
	}
	printf("\n");
}


//  Get Info of drive
//  1 based drivenumber
//
void GetInfo( uchar ucDriveNumber )
{
	static DISK_INFO DiskInfo;
	bool fSuccessful = FALSE;
	bool fSingleDrive;
	unsigned uDrive = ucDriveNumber;
	char * szVolumeLabel;

	if( ucDriveNumber != 1 )
	{
		fSingleDrive = IsSingleDrive( ucDriveNumber, 1 );
		if( fSingleDrive )
		{
			uDrive = 1;
		}
	}

	fSuccessful = dlGenericIOCTLForBlockDevices( uDrive,
			DEFAULT_CATEGORY_CODE, IOCTL_GETINFO, &DiskInfo );
	if( !fSuccessful )
	{
		DisplayErrorAndExit( ERR_DISKREAD );
	}

	szVolumeLabel = GetVolumeLabelAddress();
	strcpy(szVolumeLabel, DiskInfo.cVolumeLabel);
}


//  Return the address of the volume label string
//
char * GetVolumeLabelAddress( void )
{
	static char szVolumeLabel[12] = "NO NAME    ";
	return szVolumeLabel;
}


//  Return the address of the FAT type string
//
char * GetFatTypeString( void )
{
	static char szFatTypeString[]="FAT12";
	return szFatTypeString;
}


//  Determine if the drive is valid
//
bool CheckIfCanUseDrive( uchar ucDosDriveNumber )
{
	bool fSuccessful;
	bool fIsRemote;
	bool fIsSubsted;
	unsigned uDosVersion;


	//  Determine if the drive is local or remote.  If the drive
	//  is not local that the drive is not valid, otherwise
	//  return that it is valid
	//
	fSuccessful = dlIsDriveRemote( ucDosDriveNumber+1, &fIsRemote );
	if( !fSuccessful )
	{
		return FALSE;
	}
	if( fIsRemote )
	{
		return FALSE;
	}


	//  Determine if the drive is substed.  If the drive
	//  substed, the drive is not valid. Otherwise, return
	//  that it is valid
	//
	fSuccessful = dlIsDriveSubst( ucDosDriveNumber+1, &fIsSubsted );
	if( !fSuccessful )
	{
		return FALSE;
	}
	if( fIsRemote )
	{
		return FALSE;
	}


	return TRUE;
}


//  Display the usage screen
//
void Usage( void )
{
	DisplaySignonMessage();
	printf( szUsage );
	exit( ERR_USAGE );
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	char * szSignOnMessage = "DISKCOPY %s %s\n%s\n\n";


	if( !DoHideSignon() )
	{
		/* display sign-on */
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Display an error message and exit with the appropriate error code
//
void DisplayErrorAndExit( uchar ucErrorCode )
{
	bool fSuccessful;
	unsigned uAllocatedMemorySegment;


	//  Display the error message associated with this error code
	//
	DisplayError( ucErrorCode );


	//  Free the allocated memory block, if there was one
	//
	if( GetSegment() )
	{
		dlFreeMem( GetSegment() );
	}


	//  Close swapfile handle, if there was one
	if( GetHandle() )
	{
		dlCloseFile( (int)GetHandle() );
	}


	//  Delete swapfile, if there was one
	if( szSwapName[0] )
	{
		dlDeleteFile( szSwapName );
	}


	//  Uninstall our interrupt handlers
	//
	UninstallInt24Handler();
	UninstallCTRLCHandler();


	//  Terminate the program setting the appropriate exit code
	//
	exit( ucErrorCode );
}


//  Install a control break handler so that we can abort cleanly if
//  the user presses CTRL-C
//
void InstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	SaveOldCtrlCHandler( *lplpOldCtrlCHandler );
	*lplpOldCtrlCHandler = CTRLCHandler;
}


//  Uninstall the control break handler
//
void UninstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	*lplpOldCtrlCHandler = GetOldCtrlCHandler();
}


//  Control break handler
//
void CTRLCHandler( void )
{
	//  Free mem
	//
	if( GetSegment() )
	{
		dlFreeMem( GetSegment() );
	}


	//  Close the swapfile handle
	//
	if( GetHandle() )
	{
		dlCloseFile( (int)GetHandle() );
	}


	//  Delete the swapfile
	if( szSwapName )
	{
		dlDeleteFile(szSwapName);
	}


	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();

	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();

	//  Exit
	//
	exit( MS_ERR_CTRL_BREAK );
}


//  Return the address of the old control break handler
//
void far * GetOldCtrlCHandler( void )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	return *plpOldCtrlCHandler;
}


//  Save the address of the old control break handler
//
void SaveOldCtrlCHandler( void far * lpOldCtrlCHandler )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	*plpOldCtrlCHandler = lpOldCtrlCHandler;
}


//  Return the address of the variable which contain the old control
//  break handler
//
void far * * GetOldCtrlCHandlerAddress( void )
{
	static void far * lpOldCtrlCHandler;

	return &lpOldCtrlCHandler;
}


//  Int 24 handler
//
#pragma argsused
 void interrupt far Int24Handler(BP, DI, SI, DS, ES, DX, CX, BX, AX,
		IP, CS, FLAGS)
{
	(uchar)AX = 3;
}


//  Install a critical error handler so that we can abort cleanly
//
void InstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;


	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	SaveOldCritHandler( *lplpOldCritHandler );
	*lplpOldCritHandler = (void far *)Int24Handler;
}


//  Uninstall the critical error handler
//
void UninstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;

	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	*lplpOldCritHandler = GetOldCritHandler();
}


//  Return the address of the old critical error handler
//
void far * GetOldCritHandler( void )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	return *plpOldCritHandler;
}


//  Save the address of the old critical error handler
//
void SaveOldCritHandler( void far * lpOldCritHandler )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	*plpOldCritHandler = lpOldCritHandler;
}


//  Return the address of the variable which contain the old critical
//  error handler
//
void far * * GetOldCritHandlerAddress( void )
{
	static void far * lpOldCritHandler;

	return &lpOldCritHandler;
}


//  Save the segment
//
void SaveSegment( unsigned uSegment )
{
	unsigned * puSegment;

	puSegment = GetSegmentPointer();
	*puSegment = uSegment;
}


//  Return the segment
//
unsigned GetSegment( void )
{
	unsigned * puSegment;

	puSegment = GetSegmentPointer();
	return *puSegment;
}


//  Return the address of the variable that contains the segment
//
unsigned * GetSegmentPointer( void )
{
	static unsigned uSegment;
	return &uSegment;
}


//  Save the swap file handle
//
void SaveHandle( unsigned uHandle )
{
	unsigned * puHandle;

	puHandle = GetHandlePointer();
	*puHandle = uHandle;
}


//  Return the swap file handle
//
unsigned GetHandle( void )
{
	unsigned * puHandle;

	puHandle = GetHandlePointer();
	return *puHandle;
}


//  Return the address of the variable that contains the handle
//
unsigned * GetHandlePointer( void )
{
	static unsigned uHandle;
	return &uHandle;
}



