//
//   $Id: FAKEDISK.C 1.5 2002/10/16 21:48:58Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/********************************************************************
	$DOCMODULE:FAKEDISK.C

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This module provides interface routine for the FDISK unit tests.
	By using a file called FDISK.TM to do all the interaction with
	the BIOS (reading and writing sectors, determining # of drives),
	unit test cases can be easily created.

	A minimal FDISK.TM file might be:
	NumFixedDisks: 1
	Disk80Parms: 1024 256 63

*********************************************************************/
#define __FILENUM__	2
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dlmath.h"

FILE * fpFake;			/* the file used to read/write "hard disk" */
char szLine[120];

#define DEF_SECT_SIZE 512

/*
	$DOCPROC:FakeOpen

	Make sure the "fake hard disk" file is open. Used for debugging
	purposes.
*/
void DbgFakeOpen(void)
{
	static char szDebugFileName[] = "FDISK.TM";
	if(!fpFake)
	{
		/* open it for read/write in text mode */
		fpFake = fopen(szDebugFileName, "r+t");

		/* doesn't exist? create it for read/write in next mode */
		if(!fpFake)
			fpFake = fopen(szDebugFileName, "w+t");

		/* if it didn't work, we've got problems */
		SHIPASSERT(fpFake);
	}
}


/*
	$DOCPROC:FileDump

	Dump bytes to a file.
*/
void FileDump(FILE *fp, void *buff, int length)
{
	int start=0, len, i;
	uchar *buffer = (uchar *)buff;

	while(length>0)
	{
		fprintf(fp, "%04X: ",start);
		len=min(length,16);
		for(i=0;i<16;++i)
		{
			fprintf(fp, i<len ? "%02X " : "   ",buffer[start+i]);
			if(i==7)
				fprintf(fp,i<len ? "| " : "  ");
		}
		fprintf(fp," ");
		for(i=0;i<16;++i)
		{
			if(i<len)
				fprintf(fp,isprint(buffer[start+i]) ? "%c" : ".",buffer[start+i]);
			else
				fprintf(fp," ");
		}
		fprintf(fp,"\n");

		length -= len;
		start  += len;
	}
}

/*
	$DOCPROC:NextArg
	Return the next argument in a string.
*/
char *NextArg(char *sz)
{
	/* skip the arg we are on */
	while(*sz && !DL_IS_SPACE(*sz))
		++sz;
	/* skip the space between */
	while(*sz && DL_IS_SPACE(*sz))
		++sz;
	return sz;
}


/*
	$DOCPROC:DbgNumFixedDisks

	Returns # of fixed disks, as if BIOS Int 13h were called.
*/
int DbgNumFixedDisks(void)
{
	static char szTag[] = "NumFixedDisks";
	int	iNumFixedDisks = 0;
	char	*szNum;

	DbgFakeOpen();
	fseek(fpFake, 0L, SEEK_SET);
	while(fgets(szLine, sizeof(szLine)-1, fpFake))
	{
		/* found the tag? get the # of disks */
		if(strncmp(szLine, szTag, sizeof(szTag)-1) == 0)
		{
			szNum = NextArg(szLine);
			iNumFixedDisks = atoi(szNum);
			break;
		}
	}

	return iNumFixedDisks;
}

/*
	$DOCPROC:DbgGetFixedDiskSize

	Returns this size of the disk, or FALSE if failed.
*/
bool DbgGetFixedDiskSize(int hdrv, int *pCyls, int *pHeads, int *pSects)
{
	static char szTagTemplate[] = "Disk%XParms";
	char	szTag[20];
	char	*szNum;

	/* create a tag for this drive */
	sprintf(szTag, szTagTemplate, hdrv);

	DbgFakeOpen();
	fseek(fpFake, 0L, SEEK_SET);
	while(fgets(szLine, sizeof(szLine)-1, fpFake))
	{
		/* found the tag? get the # of disks */
		if(strncmp(szLine, szTag, strlen(szTag)) == 0)
		{
			szNum   = NextArg(szLine);
			*pCyls  = atoi(szNum);
			szNum   = NextArg(szNum);
			*pHeads = atoi(szNum);
			szNum   = NextArg(szNum);
			*pSects = atoi(szNum);
			return TRUE;
		}
	}

	return FALSE;
}

// tag to find a sector (used on both write and find sector)
static char szSectorTag[] = "Disk%XSector";
static char szSectorLine[] = "%s: %d %d %d\n";

/*
	$DOCPROC:DbgFindSector

	Checks in the file FDISK.TM for the given sector. Matches
	standard track/head/sector of a BIOS call.
*/
int DbgFindSector(int hdrv, int iCyl, int iHead, int iSect)
{
	char	szTag[20];
	int	iGotCyl, iGotHead, iGotSect;
	char	*szNum;

	/* open the file if it hasn't been already */
	DbgFakeOpen();

	/* create the tag */
	sprintf(szTag,szSectorTag,hdrv);

	fseek(fpFake, 0L, SEEK_SET);
	while(fgets(szLine, sizeof(szLine)-1, fpFake))
	{
		/* is this a sector line? */
		if(strncmp(szLine, szTag, strlen(szTag)) == 0)
		{
			// yes, but is it the right sector?
			szNum = NextArg(szLine);
			iGotCyl  = atoi(szNum);
			szNum    = NextArg(szNum);
			iGotHead = atoi(szNum);
			szNum    = NextArg(szNum);
			iGotSect = atoi(szNum);
			SHIPASSERT(iGotSect);		/* must have all 3 parameters */

			// all parameters must match
			if(iGotCyl == iCyl && iGotHead == iHead && iGotSect == iSect)
			{
				// bug in BC2: fseek required to do a write after a read
				fseek(fpFake, 0L, SEEK_CUR);
				return TRUE;
			}
		}
	}

	// bug in BC2: fseek required to do a write after a read
	fseek(fpFake, 0L, SEEK_CUR);

	// sector not found
	return FALSE;
}

/*
	$DOCPROC:DbgWriteSector

	Write a sector out to the file FDISK.TM instead of to the actual
	hard disk. Used to debug and unit test FDISK.
*/
void DbgWriteSector(int hdrv, int iCyl, int iHead, int iSect, char *pBuffer)
{
	char szTag[20];

	/* if sector is not in the file, make a header */
	if(!DbgFindSector(hdrv, iCyl, iHead, iSect))
	{
		fseek(fpFake, 0L, SEEK_END);
		sprintf(szTag, szSectorTag, hdrv);
		sprintf(szLine, szSectorLine, szTag, iCyl, iHead, iSect);
		fputs(szLine, fpFake);
	}

	// write the sector data
	FileDump(fpFake, pBuffer, DEF_SECT_SIZE);
	fflush(fpFake);
}

/*
	$DOCPROC:DbgReadSector

	Read a sector from the file FDISK.TM instead of from the actual
	hard disk. Used to debug and unit test FDISK.
*/
void DbgReadSector(int hdrv, int iCyl, int iHead, int iSect, char *pBuffer)
{
	int iRow,iCol,iExtra;

	/* is the sector already in the file? Then read it! */
	if(DbgFindSector(hdrv,iCyl,iHead,iSect))
	{
		/* 16 bytes per line */
		for(iRow=0; iRow<DEF_SECT_SIZE/16; ++iRow)
		{
			fgets(szLine, sizeof(szLine)-1, fpFake);
			iExtra = 0;
			for(iCol=0; iCol<16; ++iCol)
			{
				if(iCol == 8)
					iExtra = 2;
				*pBuffer++ = atoh(szLine + 6 + iCol*3 + iExtra);
			}
		}
	}

	/* not found in file, assume all FFs */
	else
		memset(pBuffer, 0xFF, DEF_SECT_SIZE);
}

/*
	Write a string to the debug file.
*/
void DbgWriteString(char *szStr)
{
	/* open the file if it hasn't been already */
	DbgFakeOpen();
	fputs(szStr, fpFake);
	fflush(fpFake);
}


#if UNIT_TEST

/*
	A simple ASSERT function.
*/
void InternalError(int iFileNum, int iLineNum)
{
	printf("FakeDisk Assert: ", iFileNum, iLineNum);
	exit(1);
}

char aBuffer[512];
char aBuffer2[512];

/*
	Test the Dbg read and write routines.
*/
int main(void)
{
	int i, iNumDisks;
	int iCyls, iHeads, iSects;

	iNumDisks = DbgNumFixedDisks();
	ASSERT(iNumDisks);

	for(i=0; i<iNumDisks; ++i)
	{
		if(!DbgGetFixedDiskSize(0x80+i, &iCyls, &iHeads, &iSects))
			ABORT();

		memset(aBuffer, i+1, sizeof(aBuffer));
		DbgWriteSector(0x80+i,0,0,1,aBuffer);

		DbgReadSector(0x80+i,0,0,1,aBuffer2);

		ASSERT(memcmp(aBuffer, aBuffer2, sizeof(aBuffer)) == 0);
	}

	printf("FakeDisk Unit Tests passed.\n");
	return 0;
}
#endif
