;
;   $Id: HBOOT.ASM 1.8 2003/07/29 16:23:38Z thomd Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;--------------------------------------------------------------------
; $DOCMODULE:HBOOT.ASM
;
; Copyright (c) 1998-2002 Datalight, Inc.
; All Rights Reserved.
; Datalight Confidential and Proprietary.
;
; This master boot record will dual boot. If Alt-F2 is pressed
; prior to boot, this sector will load the Super-Boot DOS partition,
; otherwise it will load the normal DOS partition. Both load at
; 0:7C00.
;
; 1) BIOS loads us at 0000:7C00
; 2) Relocate ourselves to 60:xxx
; 3) Check for Alt-F2 - if so, use the Super-Boot boot partition
; 4) Checks for a bootable partition.
; 5) Loads bootable partition at 0000:7C00
; 6) Passes control to the boot code
; 7) Windows NT that requires DS:SI points to the bootable partition entry
;
;-----------------------------------------------------------------*/

; User selectable options
HOTKEY         equ 6900h   ; hotkey value to Super-Boot (Alt-F2 = 6900h)
SUPERBOOT_ID   equ   98h   ; ID of Super-Boot partition (default=98h)
DELAY_TICKS    equ   27    ; The number of ticks to wait for a keypress (1.5s)

; Do not change
BOOT_OFFSET    equ 7C00h   ; Normal BIOS Boot Offset
RELOC_OFFSET   equ 0600h   ; offset to relocate to
RELOC_SEG      equ   60h
RETRIES        equ    5    ; Try to read boot sector up to 5 times.
ID_OFF         equ    4    ; byte 4 in partition record


_TEXT   SEGMENT   para public 'CODE'
   assume   cs:_TEXT, ds:_TEXT

BootCode:

;-------------------------------------------------------------------
;
;   RELOCATING STARTUP CODE
;
;   This code relocates itself to 70:0 and jumps to it.
;
   cli
   xor   ax, ax
   mov   ss, ax
   mov   sp, 7C00h
   mov   si, sp

   ; zero segment registers
   push  ax
   pop   es
   push  ax
   pop   ds

   ; move code to segment 60
   sti
   cld
   mov   di,0600h
   mov   cx,0100h
   repnz movsw
   db   0EAh
   dw   061Dh
   dw   0000h      ; jmp 0000:061dh

   mov   ax, RELOC_SEG
   mov   ds, ax

;-------------------------------------------------------------------
;
;   Because we are adding a delay to allow the user to press the
;   the super boot hot-key, we will first make sure that a superboot
;   partition exists.
;

   ; scan partition table for a superboot partition
   mov   si, offset table1 + 4
   mov   al, bSuperBootID
   mov   cl, 04
chknextpart:
   cmp   byte ptr [si], al
   je    chkforhotkey
   add   si, 10h
   loop  chknextpart
   jmp   short FindBootPart

;-------------------------------------------------------------------
;
;   Was hotkey pressed? If so, look for Super-Boot partition.
;

chkforhotkey:

   ; Post a debug number indicating that we are waiting for the
   ; superboot hotkey
   ;
   mov   ax, 0e21h
   xor   bx, bx
   mov   cx, 1
   int   10h

   ; Get a time reference point
   ;

   xor   ax, ax
   int   1ah

   ; Determine how long we will wait
   ;

   mov   di, dx
   add   di, DELAY_TICKS

waitforhotkey:

   ; Check if the user pressed a key
   ;

   mov   ah, 1
   int   16h
   jz    NoKey      ; no key waiting

   ; Check if this is the hot-key
   ;

   cmp   ax, wHotKey
   jnz   NoKey      ; didn't press Alt-F2

   ; Point SI to the superboot partition
   ;

   sub   si, 4

   ; Boot from this partition
   ;

   mov   bx,1
   jmp   short FoundBoot

NoKey:

   ; Get a time reference point
   ;

   xor   ax, ax
   int   1ah

   ; Check if time has expired
   ;

   cmp   di, dx
   jnl   waitforhotkey

;-------------------------------------------------------------------
;
;   FIND A BOOTABLE PARTITION IN PARTITION TABLE
;
;   Is there ANY bootable partition?
;

FindBootPart:
   ; scan partition table for a bootable partition
   mov   si, offset table1
   mov   bl, 04

   ; looking for SuperBoot?
TryTryAgain:

   ; looking for normal bootable partition
   cmp   byte ptr [si], 80h ; is this partition bootable?
   jz    FoundBoot
   cmp   byte ptr [si], 00h ; or is it the expected 0?
   jnz   BadTable

NextTableEntry:
   add   si, 10h
   dec   bl
   jnz   TryTryAgain   ; if at first you don't succeed

   ; no bootable partitions
   mov   si, offset szErrNoPart
NoPartitions:
   jmp   PrintString

   ; load registers that BIOS INT 13H needs
FoundBoot:
   mov   bp, si      ; save ptr to partition entry

   ; make sure the rest of the partition table is valid
ValidateTable:
   add   si, 10h
   dec   bl
   jz    LBA_check
   cmp   byte ptr [si], 00h
   jz    ValidateTable
   cmp   byte ptr [si], 80h
   jz    ValidateTable

   ; Invalid Partition table
BadTable:
   mov   si, offset szErrBadPart
   jmp   PrintString


;-------------------------------------------------------------------
;
;   LBA DRIVE CONFIRMATION
;
;   If BIOS supports LBA, we'll use it to load the sector
;
LBA_check:
   mov   di,RETRIES     ; 5 retries

ReadAgain:
   mov   si,bp          ; restore partition table ptr for CHS calc or load

   mov   ah,41h
   mov   bx,055AAh
   mov   dl,80h         ; restore trashed DL - first hard drive
   stc
   int   13h            ; LBA extensions - installation check
   jc    Load_CHS

;-------------------------------------------------------------------
;
;   LBA SECTOR LOAD
;
;   Use an LBA call to load the absolute boot sector from the drive
;
   mov   ax,WORD PTR [si+08h]
   mov   WORD PTR [DAP_addr], AX
   mov   ax,WORD PTR [si+0Ah]
   mov   WORD PTR [DAP_addr+2], AX

   mov   ah,42h               ; LBA extensions - Extended Read
   mov   dl,80h
   mov   si,offset DAP_size   ; pointer to Disk Address Packet
   int   13h

   jmp short Int_13_Results   ; Handle success or error conditions

;-------------------------------------------------------------------
;
;   DO THE APPROPRIATE MATHEMATICS to LOAD CHS VALUES
;
;   We don't trust the partition table CHS values, so instead we
;   will convert the value at offset 08h in partition record
;   ("sectors preceding partition")
;
Load_CHS:

   ; get the current drive 80h parameters
   ;
   mov   dl,80h
   mov   ah,08h
   int   13h

   ; save off the values for the necessary quotients
   ; NOTE: increment the heads to get a "true" value for max_heads
   ;
   and   cx,003Fh       ; CX = sectors_per_track
   xor   bx,bx
   mov   bl,dh
   inc   bx             ; BX = max_heads

   ; dword sector to CHS translation is calculated with these formulas:
   ;
   ;   sector = (dword sector) % sectors_per_track + 1
   ;    temp  = (dword sector) / sectors_per_track
   ;     head = temp % max_heads
   ; cylinder = temp / max_heads
   ;

   ; start with the dword sector in DX:AX
   ;
   mov   ax,[si+08h]
   mov   dx,[si+0Ah]

   ; now do the div/mod routine with sectors_per_track
   ;
   div   cx

   ; AX now contains the "temp" value and DX contains the "sector"
   ; now do the div/mod routine with max_heads
   ;
   push  dx
   xor   dx,dx
   div   bx
   pop   bx

   ; AX now contains the "cylinder" value & DX contains the "head" value
   ; BX contains the "sector" value
   ;

   ; now rope the values back into CX and DH for the upcoming int 13h
   ;  CH = low 8 bits of the "cylinder" value
   ;
   mov   ch,al

   ;  CL = the high two bits of the "cylinder" value
   ;       added to the "sector" value
   ;
   mov   al,0
   ror   ax,1
   ror   ax,1
   add   al,bl
   inc   al
   mov   cl,al

   ;  DH = the head value, and DL = the drive number (80h)
   ;
   xchg  dh,dl
   mov   dl,80h

;-------------------------------------------------------------------
;
;   READ IN THE PARTITION'S BOOT SECTOR
;
;   This code tries to load the real boot sector at 0000:7C00.
;   It assumes that CX and DX contain the drive (DL), head (DH),
;   sector (CL) and track (CH) usable by Int 13H.
;
Read_CHS:
   mov   bx,BOOT_OFFSET ; es:bx - buffer to load (0:7C00)
   mov   ax, 0201h   ; read 1 sector
   int   13h

Int_13_Results:
   jnc   Loaded
   xor   ax, ax      ; reset controller
   int   13h
   dec   di      ; loop up to 5 times
   jnz   ReadAgain

LoadError:
   ; failed. show the load error
   mov   si, offset szErrRead
   jmp   PrintString

   ; OK, now boot sector is loaded. verify that the boot
   ; sector has the AA55 ID at end of sector.
Loaded:
   mov   di, 7DFEh
   cmp   word ptr es:[di], 0AA55h
   jz    sectorOK
   mov   si, offset szErrNotFormatted
   jmp   PrintString

   ; restore pointer to Partition Table and jmp to boot code
sectorOK:

   ; set ds = 0
   xor   si, si
   mov   ds, si

   ; 0:bp and 0:si must point to bootable partition entry
   mov   si, bp         ; ds:offset to part table entry
   add   si, RELOC_SEG * 10h   ; add in segment
   mov   bp, si         ; bp and si should be the same

   ; di should point to SP
   mov   di, sp

   mov   bx,BOOT_OFFSET ; es:bx - buffer to jump to (0:7C00)

   push  es
   push  bx
   retf

;-------------------------------------------------------------------
;   Print_String
;   Entry   ds:si points to an ASCIIZ string
;   Exit   does not exit
PrintString:
   mov   bx, 7
   mov   ax, 0E0Dh   ; CR
   int   10h
   mov   ax, 0E0Ah   ; LF
   int   10h
NextChar:
   lodsb
   cmp   al, 00h
   jz    halt
   mov   ah, 0Eh      ; print a char TTY
   int   10h
   jmp   NextChar

   ; machine is now dead, just halt
   sti
halt:
   jmp   halt

   ; error messages
   szErrNoPart       db "No bootable partitions",0
   szErrBadPart      db "Invalid partition table",0
   szErrRead         db "Read Error",0
   szErrNotFormatted db "Partition not formatted",0

EndOfCode LABEL
   ; variables

   org   1A0h
   ; 8 bytes reserved for CA and AIC.  This is effectively an inter-
   ; process communication area located within the MBR.
   db    8 dup (?)

   ; Disk Address Packet, used by LBA extended read
   ;
   org   1ABh

   DAP_size db 10h            ; size of this packet
            db 00h            ; reserved
   DAP_xfer dw 01h            ; number of blocks to transfer (1)
   DAP_buff dw BOOT_OFFSET    ; dword pointer -> transfer buffer
   DAP_bseg dw 00h            ;
   DAP_addr dq 00h            ; Absolute Sector Address

   ; Super-Boot ID and hot-key at specific offsets so they can
   ; be updated by FDISK.
public   bSuperBootID, wHotKey
   org   1BBh
   bSuperBootID   db SUPERBOOT_ID      ; 1BBh
   wHotKey      dw HOTKEY      ; 1BCh

   org   1BEh
   table1   db   10h dup (?)
   table2   db   10h dup (?)
   table3   db   10h dup (?)
   table4   db   10h dup (?)
   sectorID db   55h, 0AAh

_TEXT   ENDS

   END   BootCode
