//
//   $Id: FORMAT.C 1.19 2003/07/03 17:11:26Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
//  FORMAT.C
//
//  Copyright (c) 1999-2002 Datalight, Inc.
//  All Rights Reserved.
//  Datalight Confidential and Proprietary
//
//  HISTORY:
//
//  09-08-1999  bt  Created
//


#define FORMAT 1


//  Conditional compile switch for supporting FAT32 drives
//
#ifdef ENABLEF32
#define SUPPORT_FAT32_DRIVES		TRUE
#else
#define SUPPORT_FAT32_DRIVES		FALSE
#endif


//  Enable debug information
//
#define FMT_DEBUG		FALSE


//
//  [ DATALIGHT LIBRARIES ]
//

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dosver.h"
#include "dosmsg.c"
#include "fatworld.h"
#include "fatlib.h"
#include "dlapp.h"
#include "windows.h"
#include "stdc.h"


//
//  [ FORMAT HEADER FILES ]
//

#include "format.h"
#include "newsys.h"


//
//  [ PROTOTYPES ]
//
#include "format.pt"
#include "syscomn.pt"



#define TEMP_DOS_DRIVE_NUMBER		0



//
//  [ VARAIABLES ]
//

bool fIsFat32World = FALSE;
bool fInWindows = FALSE;

//  Flags for the command line switches
//  (accessed through macros in FORMAT.H)
//
bool fSwitches[MAX_BOOLEAN_SWITCHES];

//  Boot sector locations
//
extern uchar rgucBootSector[];
extern uchar rgucBootSectorEnd[];
extern uchar rgucBootSector32[];
extern uchar rgucBootSector32End[];

extern bool fIsFat32Binary;

//  This is placed here only for debugging purposes.  Only one
//  function should use it.
//
char * szUserSizeArray[] = {
	"360",
	"1.2",
	"720",
	"1.44",
	"2.88",
};


//  Start of FORMAT
//
void main( int argc, char *argv[] )
{
	uchar ucExitCode;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
#if SUPPORT_FAT32_DRIVES
	fIsFat32World = dlIsFat32World();
#else // #if SUPPORT_FAT32_DRIVES
	fIsFat32World = FALSE;
#endif // #if SUPPORT_FAT32_DRIVES

	//  Determine if we are in windows
	//
	fInWindows = dlInWindows();

	//  Parse the command line
	//
	DoParseCommandLine( argc-1, &argv[1] );

	//  Determine if we should display the signon message and if
	//  so display it
	//
	DisplaySignonMessage();

	//  Determine if we are going to systemize this drive
	//  after we format it.  If so load the system file
	//  into memory just in case they reside on the drive
	//  we are going to format.
	//
	if(DoCopySystem())
	{
		//  Read in the COMMAND.COM and ROM-DOS.SYS files.  We are also
		//  passing in the name and path for SYS.COM so that we can
		//  look for the system files at the same location.
		//
		GetSystemFiles( argv[0] );
	}

	//  Install a control break handler so that we can abort cleanly
	//  if the user presses CTRL-C
	//
	InstallCTRLCHandler();

	//  Install a critical error handler so that we can abort cleanly
	//
	InstallInt24Handler();

	//  Format these disks/drives until the user does not
	//  want to format any more disks/drives.
	//
	ucExitCode = FormatDisks();

	if(DoCopySystem())
	{
      //  Free the memory allocated by GetSystemFiles()
      //
      FreeAllMemory();
	}

	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();

	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();

	//  Set the DOS error code and exit
	//
	if( ucExitCode == 0 )
	{
		exit( ucExitCode );
	}
	else
	{
		exit( MS_ERR_FATAL_ERROR );
	}
}


//  Format these disks/drives until the user does not
//  want to format any more disks/drives.
//
uchar FormatDisks( void )
{
	bool fMultiDiskFormat = TRUE;
	uchar ucExitCode;


	while( fMultiDiskFormat )
	{
		//  Display a message showing the size of the disk
		//  we are formating.
		//
		PromptForNewDisk();

 		//  Determine what the BPB will be for the drive we
		//  are formatting and save it in the boot sector
		//
		ucExitCode = SetBPBForDrive();
		if( ucExitCode )
		{
			if( ( ucExitCode == ERR_READ ) || ( ucExitCode == ERR_NOQUICK ) )
			{
				printf( szCannotQuickFormat );
				if( getyesno() )
				{
					//  Change mode to an unconditional format
					//
					fSwitches[QUICK_FORMAT] = FALSE;
					ucExitCode = SetBPBForDrive();
				}
				else
				{
					exit( MS_ERR_NO_PROCEED );
				}
			}

			if( ucExitCode )
			{
				DisplayError( ucExitCode );
			}
		}

		if( !ucExitCode )
		{
			//  Display format message
			//
			DisplayFormatMessage();

			//  Format the drive
			//
			ucExitCode = PerformFormat();
         if( Autotest() )
         {
            fMultiDiskFormat = FALSE;
				//  Copy the system files to the drive if we are making
				//  this disk bootable
				//
				WriteTheSystemFiles();
         }
         else
         {
   			if( ucExitCode )
   			{
   				//  Display that format is done
   				//
   				printf(szFormatTerminated);
   			}
   			else
   			{
   				//  Display that format is done
   				//
   				printf(szFormatComplete);

   				//  Copy the system files to the drive if we are making
   				//  this disk bootable
   				//
   				WriteTheSystemFiles();

   				//  Get the volume label
   				//
   				volume_label();

   				//  Display the statistics about the disk
   				//
   				DisplayStatistics();
   			}
         }
		}

		//  Determine if we are prompting the user to format another
		//  disk and if so display a message inquiring if the user
		//  wants to format another disk in the same drive.
		//
		if( fMultiDiskFormat )
		{
			fMultiDiskFormat = FormatAnotherDisk();
		}
	}
	return ucExitCode;
}


//  Format the drive
//
uchar PerformFormat( void )
{
	uchar ucExitCode = 0;
	bool	fSuccessful;
	uchar ucDosDriveNumber;
	uchar ucAccessFlags;


	//  Get the dos drive number of the drive we are formatting
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Determine if this is DOS 7.1.  If this is DOS 7.1 then lock
	//  the drive so that we can have direct disk access
	//
	if( fIsFat32World || fInWindows )
	{
		//  (drive: 0=default,1=A,...)
		//
		fSuccessful = dlLockLogicalVolume( ucDosDriveNumber+1,
				0, 0 );
		if(!fSuccessful)
		{
			DisplayErrorAndExit( ERR_LOCK );
		}
	}

	//  Determine if DOS is allowing us to write to this device
	//
	dlGetAccessFlags( ucDosDriveNumber+1, fIsFat32World, &ucAccessFlags );
	if( ucAccessFlags != 1 )
	{
		if( IsHardDrive() )
		{
			//  Let windows know we can start using this device
			//
			dlSetAccessFlags( ucDosDriveNumber+1, 1, fIsFat32World );
		}
	}

	//  Create a buffer to record bad cluster numbers and
	//  also buffers for the FAT sectors
	//
	AllocateBadClusterList();
	AllocateFATBuffers();

	//  Make sure that we have the memory to format the drive
	//
	if( !ucExitCode )
	{
		ucExitCode = FormatDrive();
	}

	//  Release the memory allocated for the FAT buffers
	//
	FreeBadClusterList();
	FreeFATBuffers();

	//  Determine if this is DOS 7.1.  If this is DOS 7.1 then unlock
	//  the drive
	//
	if( fIsFat32World || fInWindows )
	{
		//  (drive: 0=default,1=A,...)
		//
		dlUnlockLogicalVolume( ucDosDriveNumber+1 );

		//  Reset the drive
		//
		ResetIOCTL();
	}
	else
	{
		//  Reset the drive
		//
		ResetIOCTL();
	}

	//  Reset drives.  Resets one drive if it is a hard drive,
	//  Resets all drives if floppy (because B: may equal A:)
	//
	ResetDisks();

	return ucExitCode;
}


//  Format the drive/disk (does quick formats too)
//
uchar FormatDrive( void )
{
	uchar ucReturnCode = 0;
	unsigned uTrack;
	unsigned uEndTrack;
	pDRIVE_PARMS pDriveParms;
	bool fSuccessful;
	uchar ucDosDriveNumber;
	ulong ulSerialNumber;


	//  Get the dos drive number of the drive we are formatting
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Reset the drive
	//
	ResetIOCTL();

	//  (drive: 0=default,1=A,...)
	//
	if( fInWindows )
	{
		fSuccessful = dlLockLogicalVolume( ucDosDriveNumber+1,
				0, 4 );
		if(!fSuccessful)
		{
			DisplayErrorAndExit( ERR_LOCK );
		}
	}

	//  Determine if this is a quick format.  If this is a
	//  quick format perform a quick format on the drive.
	//  Otherwise, do a regular format on the drive.
	//
	if(IsQuickFormat())
	{
		//  Perform a quick format of the drive
		//
		ucReturnCode = QuickFormatDrive();
	}
	else
	{
		//  Get the address of the storage area for the device parameters
		//
		pDriveParms = GetDriveParameterAddress();
		uEndTrack = pDriveParms->Parms.cylinders - 1;

		//  Starting with the first track, repeat until all of
		//  the tracks on the drive have been formatted
		//
		for( uTrack=0; uTrack <= uEndTrack; uTrack++ )
		{
			//  Format this track using all of the heads for
			//  this drive.  If there is an error formatting
			//  a track outside of the data area this function
			//  will display an error message and return an
			//  error.
			//
			ucReturnCode = FormatTrackOnAllHeads( uTrack );

			//  Determine if there was an error formatting the tracks.
			//  If so, we should not continue formatting.  Otherwise,
			//  Display the percentage of what we have formatted on
			//  the drive.
			//
			if( ucReturnCode )
			{
				break;
			}
		}
		if( !ucReturnCode )
		{
			printf("\n");
		}
	}

	//  (drive: 0=default,1=A,...)
	//
	if( fInWindows )
	{
		fSuccessful = dlUnlockLogicalVolume( ucDosDriveNumber+1 );
		if(!fSuccessful)
		{
			DisplayErrorAndExit( ERR_UNLOCK );
		}

		//  Reset the drive
		//
		ResetIOCTL();
	}

	if( !ucReturnCode )
	{
		//  Resize the bad cluster list and the allocated buffers
		//
		ResizeBuffers();

		//  Generate a serial number and update it in the boot sector
		//  before it gets committed
		//
		ulSerialNumber = GetNewSerialNumber();
		SaveSerialNumber( ulSerialNumber );

		//  Write the boot sector
		//
		ucReturnCode = WriteTheBootSector();
		if( ucReturnCode )
		{
			DisplayError( ucReturnCode );
			return ucReturnCode;
		}

		//  Set the serial number
		//
		WriteTheSerialNumber( ulSerialNumber );

		//  Write the root directory
		//
		WriteTheRootDirectory();

		//  Write the FAT
		//
		WriteTheFAT();

		//  Reset the drive
		//
		ResetIOCTL();

		//  Set DPB for the format
		//
		SetNewDPB();

		//  Reset drives.  Resets one drive if it is a hard drive,
		//  Resets all drives if floppy (because B: may equal A:)
		//
		ResetDisks();

		//  Compute the free space for the drive we just formatted
		//
		ComputeFreeSpace();
	}

	return ucReturnCode;
}


//  Format this track using all of the heads for
//  this drive.  If there is an error formatting
//  a track outside of the data area this function
//  will display an error message and return an
//  error.
//
uchar FormatTrackOnAllHeads( unsigned uTrack )
{
	uchar ucReturnCode = 0;
	uchar ucErrorCode;
	unsigned uHead;
	unsigned uEndHead;
	unsigned uEndTrack;  // for percentage
	pDRIVE_PARMS pDriveParms;
	unsigned uPercent = 0;
	static unsigned uOldPercent = 101;
	uchar ucSpecialFunction = 2;


	if( uTrack == 0 )
	{
		ucSpecialFunction = 1;
	}

	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();
	uEndHead = pDriveParms->Parms.bpb.sides;
	uEndTrack = pDriveParms->Parms.cylinders - 1;

	if(!IsHardDrive())
	{
		ucSpecialFunction = 0;
	}

	//  Starting with the first head, repeat until this track
	//  number on all the heads has been formatted
	//
	for( uHead=0; uHead < uEndHead; uHead++ )
	{
		//  Format this track at this head location on the drive
		//
		ucErrorCode = FormatSpecifiedTrack( uTrack, uHead, ucSpecialFunction );
		if( ucSpecialFunction == 1 )
		{
			ucSpecialFunction++;
		}
		else
		{
			if( ucSpecialFunction != 0 )
			{
				uHead++;
			}
		}

		//  Determine if there was an error formatting this track.
		//
		if( !ucErrorCode )
		{
			if( uTrack != uEndTrack )
			{
				uPercent = (unsigned)((ulong)uTrack * 100 / uEndTrack);
			}
			else
			{
				uPercent = 100;
			}
			if( uPercent != uOldPercent )
			{
				printf( format_percent, uPercent);
				uOldPercent = uPercent;
				if( uOldPercent == 100 )
				{
					uOldPercent++;
				}
			}
		}
		else
		{
			//  Determine if this was a write protection error.  If this
			//  was not a write protection error
			//
			if( ucErrorCode == BAD_TRACK_ERROR )
			{
				//  Determine if this track lies in the data area
				//  of the disk.  If this track lies in the data
				//  area of the disk, mark all the clusters which
				//  use this track as bad
				//
				ucReturnCode = MarkClustersInTrackBad( uTrack, uHead );
				if( ucReturnCode == 1 )
				{
					DisplayError(ERR_CANTFORMAT);
					ucReturnCode = 0xff;
					break;
				}
				else
				{
					if( ucReturnCode != 0 )
					{
						DisplayError(ERR_BADCLUSTERS);
						ucReturnCode = 0xff;
						break;
					}
				}
			}
			else
			{
				//  We can not format a write protected disk/drive. Return
				//  that there was an error formatting this track
				//
				printf("\n");
				DisplayError(ucErrorCode);
				ucReturnCode = 0xff;
				break;
			}
		}
	}

	return ucReturnCode;
}


//  Format this track at this head location on the drive
//
uchar FormatSpecifiedTrack( unsigned uTrack, unsigned uHead,
		uchar ucSpecialFunction )
{
	uchar ucReturnCode = 0;
	bool fSuccessful;
	uchar ucDosDriveNumber;
	unsigned uExtendedErrorCode;
	uchar ucErrorClass;
	uchar ucRecomendAction;
	uchar ucErrorLocus;


	//  Get the drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	fSuccessful = dlIOCTLFormatAndVerifyTrack( ucDosDriveNumber + 1,
			uHead, uTrack, ucSpecialFunction, fIsFat32World );
	if( !fSuccessful )
	{
		//  Determine what to do about this error
		//
		dlGetExtendedErrorInfo( &uExtendedErrorCode, &ucErrorClass,
				&ucRecomendAction, &ucErrorLocus );
		if( (!uExtendedErrorCode) || (uExtendedErrorCode == 0x13) )
		{
			ucReturnCode = ERR_WRITEPROTECT;
		}
		else if( uExtendedErrorCode == 0x15 )
		{
			ucReturnCode = ERR_DRVNOTREADY;
		}
		else
		{
			ucReturnCode = BAD_TRACK_ERROR;
		}
	}

	return ucReturnCode;
}



//  Perform a quick format of the drive
//
uchar QuickFormatDrive( void )
{
	uchar ucReturnCode = 0;
	PDPB pDPB;
	ulong ulMaxCluster;
	ulong ulEntry;
	ulong ulClusterValue = 0;


	//  Set the buffer mode
	//
	SetBufferMode( BUFFER_READ, BUFFER_NO_WRITE );

	//  Get the address of our local DPB
	//
	pDPB = GetDPBAddress();

	//  Determine how many clusters we need to set
	//
	ulMaxCluster = pDPB->hiclus;
	if( ulMaxCluster == 0)
	{
		ulMaxCluster = pDPB->ulMaxCluster;
	}

	//  Find all the clusters which are marked as BAD
	//
	for( ulEntry = 2; ulEntry <= ulMaxCluster; ulEntry++ )
	{
		GetFatEntry( pDPB, ulEntry, &ulClusterValue );
		if( ulClusterValue == BAD_CLUSTER )
		{
			//  Record that this entry is a bad cluster
			//
			ucReturnCode = RecordBadCluster(ulEntry);
			if( ucReturnCode )
			{
				break;
			}
		}
	}

	return ucReturnCode;
}


//  Record that this cluster should be marked BAD
//
uchar RecordBadCluster( ulong ulClusterNumber )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	ulong far * lpEntry;
	unsigned uEntryNumber;

	pBadClusterInfo = GetBadClusterInfoPointer();
	lpEntry = MK_FP( pBadClusterInfo->uListSegment, 0);
	uEntryNumber = pBadClusterInfo->uNumberOfEntries;

	if( uEntryNumber == MAX_BAD_CLUSTERS )
	{
		return ERR_BADCLUSTERS;
	}
	else
	{
		lpEntry[uEntryNumber] = ulClusterNumber;
		pBadClusterInfo->uNumberOfEntries = uEntryNumber + 1;

		return NO_ERROR;
	}
}


//  Return the address of the bad cluster management structure
//
pBAD_CLUSTER_INFO GetBadClusterInfoPointer( void )
{
	static BAD_CLUSTER_INFO BadClusterInfo;
	return &BadClusterInfo;
}


//  Get the first bad cluster number in the list
//
ulong GetFirstBadClusterNumber( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	ulong far * lpEntry;
	unsigned uEntryNumber = 0;

	pBadClusterInfo = GetBadClusterInfoPointer();
	if(pBadClusterInfo->uNumberOfEntries)
	{
		lpEntry = MK_FP( pBadClusterInfo->uListSegment, 0);
		pBadClusterInfo->uNextLookup = uEntryNumber + 1;

		return lpEntry[uEntryNumber];
	}
	else
	{
		return BAD_CLUSTER;
	}
}


//  Get the next bad cluster number in the list
//
ulong GetNextBadClusterNumber( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	ulong far * lpEntry;
	unsigned uEntryNumber = 0;

	pBadClusterInfo = GetBadClusterInfoPointer();
	if(pBadClusterInfo->uNumberOfEntries)
	{
		uEntryNumber = pBadClusterInfo->uNextLookup;
		lpEntry = MK_FP( pBadClusterInfo->uListSegment, 0);
		pBadClusterInfo->uNextLookup = uEntryNumber + 1;

 		return lpEntry[uEntryNumber];
	}
	else
	{
		return BAD_CLUSTER;
	}
}


//  Resize the bad cluster list and then reallocate the memory
//  for the sector buffers.  This is to optimize the write
//  buffers
//
void ResizeBuffers( void )
{
	FreeFATBuffers();
	ResizeBadClusterList();
	AllocateFATBuffers();
}


//  This is a buffering optimization.  Release what ever memory
//  that is not being used by the bad cluster list so that it
//  can be used in the sector buffers
//
void ResizeBadClusterList( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	unsigned uParagraphs;
	unsigned uAvailPargraphs;
	unsigned uNumberOfEntries;
	unsigned uListSegment;


	pBadClusterInfo = GetBadClusterInfoPointer();
	uListSegment = pBadClusterInfo->uListSegment;
	uNumberOfEntries = pBadClusterInfo->uNumberOfEntries;
	if(uNumberOfEntries)
	{
		uParagraphs = uNumberOfEntries/4;
		if( (uNumberOfEntries % 4) != 0 )
		{
			uParagraphs++;
		}
	}
	else
	{
		uParagraphs = 1;
	}

	dlResizeMem( uParagraphs, uListSegment, &uAvailPargraphs );
}


//  Release the memory allocated for the bad sector list
//
void FreeBadClusterList( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	unsigned uListSegment;
	bool fSuccessful;

	pBadClusterInfo = GetBadClusterInfoPointer();
	uListSegment = pBadClusterInfo->uListSegment;
	fSuccessful = dlFreeMem( uListSegment );
	if(!fSuccessful)
	{
		DisplayErrorAndExit( ERR_RELEASE );
	}
}


//  Allocate a block of memory to contain a bad cluster list
//
void AllocateBadClusterList( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	unsigned uListSegment;
	bool fSuccessful;

	fSuccessful = dlAllocMem( (MAX_BAD_CLUSTERS >> 4), &uListSegment );
	if( !fSuccessful )
	{
		DisplayErrorAndExit( ERR_NOMEMORY );
	}

	pBadClusterInfo = GetBadClusterInfoPointer();
	pBadClusterInfo->uListSegment     = uListSegment;
	pBadClusterInfo->uNumberOfEntries = 0;
}


//  Set the serial number (THIS IS WHERE MSDOS DOS IT)
//
void WriteTheSerialNumber( ulong ulSerialNumber )
{
	static DISK_INFO DiskInfo;
//	bool fSuccessful;
	char * szVolumeLabel;
	char * szFatTypeString;
	uchar ucDosDriveNumber;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Initialize the fields
	//
	DiskInfo.uInfoLevel = 0;
	DiskInfo.ulSerialNumber = ulSerialNumber;
	memset( &(DiskInfo.cVolumeLabel[0]), 0x20, 11 );
	memset( &(DiskInfo.cFatTypeString[0]), 0x20, 8 );

	//  Set the volume label
	//
	szVolumeLabel = GetVolumeLabelAddress();
	memcpy( &(DiskInfo.cVolumeLabel[0]), szVolumeLabel,
			strlen(szVolumeLabel) );

	//  Set the fat id string
	//
	szFatTypeString = GetFatTypeString();
	memcpy( &(DiskInfo.cFatTypeString[0]), szFatTypeString,
			strlen(szFatTypeString) );

	//  Set the serial number and other misc info
	//
	//  (drive: 0=default,1=A,2=B,...)
	//
	if( fIsFat32World )
	{
		dlGenericIOCTLForBlockDevices( ucDosDriveNumber+1,
				0x48, 0x46,	&DiskInfo );
	}
	else
	{
		dlGenericIOCTLForBlockDevices( ucDosDriveNumber+1,
				0x08, 0x46,	&DiskInfo );
	}
}


//  Record the serial number for easy referencing
//
void SaveSerialNumber( ulong ulSerialNumber )
{
	ulong * pulSerialNumber;

	pulSerialNumber = GetSerialNumberPointer();
	*pulSerialNumber = ulSerialNumber;
}


//  Return the serial number for the disk we just formatted
//
ulong GetSerialNumber( void )
{
	ulong * pulSerialNumber;

	pulSerialNumber = GetSerialNumberPointer();
	return *pulSerialNumber;
}


//  Return the address of the variable which contains the serial number
//
ulong * GetSerialNumberPointer( void )
{
	static ulong ulSerialNumber;
	return &ulSerialNumber;
}


//  Return the address of the volume label string
//
char * GetVolumeLabelAddress( void )
{
	static char szVolumeLabel[12] = "NO NAME    ";
	return szVolumeLabel;
}


//  This is used to set whether the buffer sceme zeros a buffer
//  or reads it from disk.
//
void SetBufferMode( bool fDoRead, bool fDoWrite )
{
	pFATBUF pBufferInfo;


	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Set the read and write flags
	//
	pBufferInfo->fDoRead  = fDoRead;
	pBufferInfo->fDoWrite = fDoWrite;
}


//  Return a new serial number
//
ulong GetNewSerialNumber( void )
{
//	union REGS r;
	unsigned cx1,cx2;
	unsigned dx1,dx2;
	DOSDate DOSDate;
	DOSTime DOSTime;


#if 0
	r.h.ah = 0x2A;		/* get date */
	intdos(&r,&r);
	cx1 = r.x.cx;
	dx1 = r.x.dx;
	r.h.ah = 0x2C;		/* get time */
	intdos(&r,&r);
	cx2 = r.x.cx;
	dx2 = r.x.dx;
#else
	dlGetDOSDate( &DOSDate );
	cx1 = DOSDate.year;
	dx1 = *((unsigned *)(&DOSDate));
	dlGetDOSTime( &DOSTime );
	cx2 = *((unsigned *)(&DOSTime));
	dx2 = *((unsigned *)((char *)(&DOSTime))+2);
#endif

	/* add date & time together to get serial # */
	return (ulong)(MK_FP(dx1+dx2,cx1+cx2));
}


//  Return the address of the FAT type string
//
char * GetFatTypeString( void )
{
	static char * szFatTypeString = "FAT12";
	PDPB pDPB;


	//  Get our DPB for this drive
	//
	pDPB = GetDPBAddress();

	//  Check if this is a FAT32 drive
	//
	if(IsFat32Drive(pDPB))
	{
		szFatTypeString[3] = '3';
		szFatTypeString[4] = '2';
	}
	else
	{
      // Clusters are counted from 0, so HICLUS12 and HICLUS16 are the
      // largest possible values. But pDPB->hiclus is a count of clusters
      // so we have to subtract one when comparing ... or add one to HICLUS
      //
		if( pDPB->hiclus > HICLUS12+1 )
		{
			szFatTypeString[3] = '1';
			szFatTypeString[4] = '6';
		}
		else
		{
			szFatTypeString[3] = '1';
			szFatTypeString[4] = '2';
		}
	}

	return szFatTypeString;
}


//  Give DOS 7.1 a new DPB for this drive
//
void SetNewDPB( void )
{
	uchar ucDosDriveNumber;
	struct {
		unsigned uPacketSize;
		unsigned uStructVersion;
		ulong ulFunction;
		ulong	ulDPBFreeCount;
		ulong ulDPBNextFree;
		ulong ulUnused[2];
	} ParmBlock;


	if( fIsFat32World )
	{
		//  Get the drive number
		//
		ucDosDriveNumber = GetDosDriveNumber();

		//  Set the DPB parameters for DOS 7.1
		//
		ParmBlock.uPacketSize = 0x18;
		ParmBlock.uStructVersion = 0;
		ParmBlock.ulFunction = 0;
		ParmBlock.ulDPBFreeCount = 0xFFFFFFFFL;
		ParmBlock.ulDPBNextFree  = 0xFFFFFFFFL;
		ParmBlock.ulUnused[0] = 0;
		ParmBlock.ulUnused[1] = 0;
		dlSetDBPForFormat( ucDosDriveNumber+1, &ParmBlock );
	}
	else
	{
		//
		//  We don't need to do this for DOS 6.22
		//
	}
}


/*
	The error is critical in nature (unsupported drive, etc)
	so abort the format altogether
*/
void DisplayErrorAndExit( uchar ucMessageIndex )
{
	uchar ucDosDriveNumber;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	if( fIsFat32World || fInWindows )
	{
		//  (drive: 0=default,1=A,...)
		//
		dlUnlockLogicalVolume( ucDosDriveNumber+1 );
	}

	DisplayError(ucMessageIndex);
	exit( MS_ERR_FATAL_ERROR );
}


//  Display an error message
//
void DisplayError( uchar ucIndex )
{
	uchar ucDosDriveNumber;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	printf("\n");
	printf(err_msg[ucIndex],ucDosDriveNumber+'A');
}


//  Display a message inquiring if the user  wants to format
//  another disk in the same drive.
//
bool FormatAnotherDisk( void )
{
	bool fFormatAnother = FALSE;

	if( !IsHardDrive() )
	{
		if(IsQuickFormat())
		{
			printf(quick_str);
		}
		printf(another_str);
		fFormatAnother = getyesno();
		if( fFormatAnother )
		{
			printf("\n");
		}
	}

	return fFormatAnother;
}


//  Write both copies of the FAT
//
void WriteTheFAT( void )
{
	PDPB pDPB;
	ulong ulMaxCluster;
	ulong ulEntry;
	ulong ulClusterValue;
	ulong ulNextBadCluster;
	ulong ulRootCluster;


	//  Set the buffer mode
	//
	SetBufferMode( BUFFER_NO_READ, BUFFER_WRITE );

	//  Get the address of our local DPB
	//
	pDPB = GetDPBAddress();

	//  It can take a while for FAT32 drives to write out the FAT.
	//  Display a message indicating what we are doing at the moment.
	//
	if( IsFat32Drive(pDPB) )
	{
		printf( szWritingFat );
	}

	//  Determine how many clusters we need to set
	//
	ulMaxCluster = pDPB->hiclus;
	if( ulMaxCluster == 0)
	{
		ulMaxCluster = pDPB->ulMaxCluster;
	}

	if(IsFat32Drive(pDPB))
	{
		ulRootCluster = pDPB->ulRootDirStartCluster;
	}
	else
	{
		ulRootCluster = ulMaxCluster + 1;
	}

	//  Initialize the first two clusters of the FAT
	//
	PutFatEntry( pDPB, 0, 0x0fffff00L | (pDPB->media), &ulClusterValue );
	PutFatEntry( pDPB, 1, 0x0fffffffL, &ulClusterValue );

	ulNextBadCluster = GetFirstBadClusterNumber();

	//  Mark all entries which are not marked as a bad as free/available
	//
	for( ulEntry = 2; ulEntry <= ulMaxCluster; ulEntry++ )
	{
		if( ulEntry == ulNextBadCluster )
		{
			PutFatEntry( pDPB, ulEntry, BAD_CLUSTER, &ulClusterValue );
			ulNextBadCluster = GetNextBadClusterNumber();
		}
		else if(ulEntry == ulRootCluster)
		{
			//  Mark this entry as used by the root directory
			//
			PutFatEntry( pDPB, ulEntry, 0x0FFFFFFFL, &ulClusterValue );
		}
		else
		{
			PutFatEntry( pDPB, ulEntry, 0, &ulClusterValue );
		}
	}

	//  Commit any fat sectors in the buffers
	//
	FlushBuffers();

	if( IsFat32Drive(pDPB) )
	{
		printf( szComplete );
	}
}


//  Write the boot sector to the drive we are formatting
//
uchar WriteTheBootSector( void )
{
	PBPB pBPB;
	PDPB pDPB;
	void far * lpBuffer;
	pFATBUF pBufferInfo;
//	LPFSINFO lpFSINFO;
//	pBAD_CLUSTER_INFO pBadClusterInfo;
	ulong ulRootCluster;


	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Get the address of our local DPB
	//
	pDPB = GetDPBAddress();

	//  Get the BPB address
	//
	pBPB = GetBPBAddress();

	//  Get the address of the bad cluster list managment structure
	//
//	pBadClusterInfo = GetBadClusterInfoPointer();

	//  Set the buffer mode
	//
	SetBufferMode( BUFFER_NO_READ, BUFFER_WRITE );

	//  Load the correct sectors into the buffers
	//
	LoadBuffers( 0 );

	//  Set the address of the buffer
	//
	lpBuffer = MK_FP( GetBufferSegment(), 0 );

	//  Determine if this is a FAT32 drive
	//
	if(IsFat32Drive(pDPB))
	{
		//  Determine which cluster will the root directory start on
		//
		ulRootCluster = GetFAT32RootCluster();

		//  Save the root directory cluster number in the BPB and DPB
		//
		pBPB = GetBPBAddress();
		pBPB->ulRootDirStartCluster = ulRootCluster;
		pDPB->ulRootDirStartCluster = ulRootCluster;

		//  Set the root cluster number in the device parameters
		//
		UpdateDeviceParameters();

		//  Reset the drive
		//
		ResetIOCTL();

		//  Copy the boot sectors into the buffers
		//
		_fmemcpy( lpBuffer, rgucBootSector32, (512*3) );

		//  Set the copies and distance
		//
		pBufferInfo->uNumberOfSectors = 3;
		pBufferInfo->uCopies = 1;
		FlushBuffers();
		pBufferInfo->ulFirstSectorNumber = pBPB->uBackUpBootSector;
		pBufferInfo->uNumberOfSectors = 3;
		pBufferInfo->uCopies = 1;
		FlushBuffers();
	}
	else
	{
		//  Copy the boot sector into the buffer
		//
		_fmemcpy( lpBuffer, rgucBootSector, 512 );

		//  Set the copies and copy distance
		//
		pBufferInfo->uNumberOfSectors = 1;
		pBufferInfo->uCopies = 1;

		//  Reset the drive
		//
//		ResetIOCTL();  // Not like MSDOS7 boot
		FlushBuffers();
	}

	return 0;
}


//  Return the address of a (format created) DPB
//
PDPB GetDPBAddress( void )
{
	static DPB DPB;

	return &DPB;
}


//  Create a DPB so that we have all of the drive areas calculated
//
void CreateDPB( void )
{
	PDPB 		pDPB;
	PBPB		pBPB;
	unsigned bps;
	unsigned res;
	ulong    dirstart;
	ulong    datastart;
	ulong    ulFreeClusters = 0;
	ulong		ulNextFreeCluster = 0;
	ulong    ulMaxCluster;
	unsigned entries;
	ulong    sectors;
	uchar    spc;
	uchar    num_fats;
	uchar    shift;
	uchar		ucDosDriveNumber;


	// Get the address of our local DPB
	//
	pDPB = GetDPBAddress();

	// Get the BPB from the boot sector
	//
	pBPB = GetBPBAddress();

	// validate pointers to structures
	//
	ASSERT(pBPB);
	ASSERT(pDPB);

	//  Get the DOS drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Save the dos drive number in the DPB
	//
	pDPB->drive = ucDosDriveNumber;

	//  Make local copies of some of the variables in the BPB
	//
	bps      = pBPB->sec_size;
	res      = pBPB->res_sec;
	spc      = pBPB->sec_per_clus;
	entries  = pBPB->uDirEntries;
	num_fats = pBPB->num_fats;

	//  Determine the shift factor for the cluster
	//
	spc = spc >> 1;
	for(shift=0; spc; ++shift)
	{
		spc >>= 1;
	}

	//
	//  These calculations are the same for both FAT16 and FAT32 drives
	//

	sectors      	= (pBPB->num_sec ? (long)pBPB->num_sec : pBPB->lnumsec);
	ulFreeClusters	= 0xFFFFFFFFL;

	// FAT32 disk? then set the fields up for FAT32
	if( !pBPB->uFatSectors )
	{
		datastart    = res + (pBPB->ulSectorsPerFat * num_fats);
		ulMaxCluster = ((sectors - datastart) >> shift) + 1;

		//  These fields are assigned zero for a FAT32 DPB
		//
		pDPB->dir_ent   = 0;
		pDPB->fat_len   = 0;
		pDPB->hiclus    = 0;

		//  This field is not used for FAT32 drives and should be
		//  assigned -1
		//
		pDPB->dir_start	= Negative1;

		//  Set the FAT32 fields of the DPB
		//
		pDPB->uExtendedFlags        = pBPB->uExtendedFlags;
		pDPB->uFSInfoSector         = pBPB->uFSInfoSector;
		pDPB->uBackUpBootSector     = pBPB->uBackUpBootSector;
		pDPB->ulSectorsPerFAT       = pBPB->ulSectorsPerFat;
		pDPB->ulRootDirStartCluster = pBPB->ulRootDirStartCluster;
		pDPB->ulFirstDataSector     = datastart;
		pDPB->ulMaxCluster          = ulMaxCluster;
	}
	else
	{
		dirstart     = res + (pBPB->uFatSectors * num_fats);
		datastart    = dirstart + ((entries + ((bps>>5) - 1)) << 5) / bps;
		ulMaxCluster = ((sectors - datastart) >> shift) + 1;

		pDPB->dir_ent	 = entries;
		pDPB->dir_start = LOWORD(dirstart);
		pDPB->hiclus    = LOWORD(ulMaxCluster);
		pDPB->fat_len   = pBPB->uFatSectors;
	   pDPB->newclus   = 0;
	}

	//
	//  These fields get set the same for FAT32 or FAT16
	//

	pDPB->bytes_per_sec	= bps;
	pDPB->ms_spc			= pBPB->sec_per_clus - 1;
	pDPB->fat_start		= res;
	pDPB->num_fats			= num_fats;
	pDPB->media				= pBPB->media;
	pDPB->shift          = shift;

	//  This field gets set with with the low word of the same variable
	//  for both FAT32 and FAT16 drives.  See the notes in the DPB
	//  structure before changing these assignments.
	//
	pDPB->data_start  = LOWORD(datastart);

	//  For a FAT32 drive we need to retrieve the number of free clusters
	//  and the next available cluster number from the disk.  The DPB needs
	//  to be fully initilized and the sda_dpbptr needs to point to the
	//  dpb.  For romdos init this is done in dev_init.
	//
	if( !pBPB->uFatSectors )
	{
		// determine freespace and next free cluster
		//
		pDPB->uFreeClustersHigh = HIWORD(ulFreeClusters);
		pDPB->ulNextFreeCluster = ulNextFreeCluster;
		pDPB->newclus           = LOWORD(ulNextFreeCluster);
	}
	else
	{
		pDPB->rebuild = 0xFF;
	}

	//  This field gets set with with the low word of the same variable
	//  for both FAT32 and FAT16 drives.  See the notes in the DPB
	//  structure before changing these assignments.
	//
	pDPB->freeclus = LOWORD(ulFreeClusters);
}


//  Install a control break handler so that we can abort cleanly if
//  the user presses CTRL-C
//
void InstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	SaveOldCtrlCHandler( *lplpOldCtrlCHandler );
	*lplpOldCtrlCHandler = CTRLCHandler;
}


//  Uninstall the control break handler
//
void UninstallCTRLCHandler( void )
{
	void far * far * lplpOldCtrlCHandler;

	lplpOldCtrlCHandler = MK_FP( 0, 0x23*4 );
	*lplpOldCtrlCHandler = GetOldCtrlCHandler();
}


//  Control break handler
//
void CTRLCHandler( void )
{
	uchar ucDosDriveNumber;


	//  Get the drive number for the drive we are formatting
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Uninstall the critical error handler
	//
	UninstallInt24Handler();

	//  Uninstall the control break handler
	//
	UninstallCTRLCHandler();

	//  Reset the drive parameters using IOCTL calls
	//
	ResetIOCTL();

	if( fIsFat32World || fInWindows )
	{
		dlUnlockLogicalVolume( ucDosDriveNumber+1 );
	}

	if( fInWindows )
	{
		dlUnlockLogicalVolume( ucDosDriveNumber+1 );
	}

	//  Display an error message and exit
	//
	DisplayError( ERR_ABORT );
	exit( MS_ERR_CTRL_BREAK );
}


//  Return the address of the old control break handler
//
void far * GetOldCtrlCHandler( void )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	return *plpOldCtrlCHandler;
}


//  Save the address of the old control break handler
//
void SaveOldCtrlCHandler( void far * lpOldCtrlCHandler )
{
	void far * * plpOldCtrlCHandler;

	plpOldCtrlCHandler = GetOldCtrlCHandlerAddress();
	*plpOldCtrlCHandler = lpOldCtrlCHandler;
}


//  Return the address of the variable which contain the old control
//  break handler
//
void far * * GetOldCtrlCHandlerAddress( void )
{
	static void far * lpOldCtrlCHandler;

	return &lpOldCtrlCHandler;
}


//  Int 24 handler
//
#pragma argsused
 void interrupt far Int24Handler(BP, DI, SI, DS, ES, DX, CX, BX, AX,
		IP, CS, FLAGS)
{
	(uchar)AX = 3;
}


//  Install a critical error handler so that we can abort cleanly
//
void InstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;


	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	SaveOldCritHandler( *lplpOldCritHandler );
	*lplpOldCritHandler = (void far *)Int24Handler;
}


//  Uninstall the critical error handler
//
void UninstallInt24Handler( void )
{
	void far * far * lplpOldCritHandler;

	lplpOldCritHandler = MK_FP( 0, 0x24*4 );
	*lplpOldCritHandler = GetOldCritHandler();
}


//  Reset the drive parameters using IOCTL calls
//
bool ResetIOCTL( void )
{
	pDRIVE_PARMS pDriveParms;
	uchar ucDosDriveNumber;
	bool fSuccessful;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();

	//  Indicate what has changed in the device parameters
	//
	pDriveParms->Parms.specFunc = 5;

	//  Get the drive parameters for the drive we are formating
	//  using an IOCTL call
	//
	if( fIsFat32World )
	{
		fSuccessful = dlSetParameters32ForBlockDevice( ucDosDriveNumber+1,
				pDriveParms );
	}
	else
	{
		fSuccessful = dlSetParametersForBlockDevice( ucDosDriveNumber+1,
				pDriveParms );
	}

	return fSuccessful;
}


//  Return the address of the old critical error handler
//
void far * GetOldCritHandler( void )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	return *plpOldCritHandler;
}


//  Save the address of the old critical error handler
//
void SaveOldCritHandler( void far * lpOldCritHandler )
{
	void far * * plpOldCritHandler;

	plpOldCritHandler = GetOldCritHandlerAddress();
	*plpOldCritHandler = lpOldCritHandler;
}


//  Return the address of the variable which contain the old critical
//  error handler
//
void far * * GetOldCritHandlerAddress( void )
{
	static void far * lpOldCritHandler;

	return &lpOldCritHandler;
}


//  Determine if we can format this drive.
//
bool CanFormatDrive( void )
{
	uchar ucDosDriveNumber;
	bool	fSuccessful;
   bool  fDriveIsRemote = TRUE;
   bool  fDriveIsSubst = TRUE;
	unsigned uDosVersion;
	pDRIVE_PARMS pDriveParms;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

   //  We pass in a variable already set to TRUE, in case the function
   //  fails with a DOSERROR and doesn't change the state.
   //  Ergo, if the drive isn't valid, both flags will be TRUE!
   //
	dlIsDriveSubst( ucDosDriveNumber+1, &fDriveIsSubst );
	dlIsDriveRemote( ucDosDriveNumber+1, &fDriveIsRemote );

	if( fDriveIsSubst || fDriveIsRemote )
	{
		return FALSE;
	}

	//  Make sure that the DOS version is at least 5.0 becuase
	//  IOCTL calls were not supported prior to 5.0
	//
	dlGetDosVersion( &uDosVersion );
	if( uDosVersion < 0x0500 )
	{
		return FALSE;
	}

	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();

	//  Get the drive parameters for the drive we are formating
	//  using an IOCTL call
	//
	if( fIsFat32World )
	{
		fSuccessful = dlGetParameters32ForBlockDevice( ucDosDriveNumber+1,
				pDriveParms );

      if(fSuccessful)
      {
   		SetTrackLayout( &pDriveParms->uSectorsInTrack,
   		      &pDriveParms->Parms.bpb );
      }
      else
      {
         // Reset our flag - FAT32 is not supported on this drive
         //
      	fIsFat32World = FALSE;

   		fSuccessful = dlGetParametersForBlockDevice( ucDosDriveNumber+1,
   				pDriveParms );
   		SetTrackLayout( &((char *)(pDriveParms))[0x26],
   				&pDriveParms->Parms.bpb );
      }
	}
	else
	{
		fSuccessful = dlGetParametersForBlockDevice( ucDosDriveNumber+1,
				pDriveParms );
		SetTrackLayout( &((char *)(pDriveParms))[0x26],
				&pDriveParms->Parms.bpb );
	}

	return fSuccessful;
}


//  Build a track layout table
//
void SetTrackLayout( void * pTrackParms, PBPB pBPB )
{
	unsigned uSectorsPerTrack;
	unsigned uSectorSize;
	unsigned uSectorIndex;
	struct Parms{
		unsigned uSectorsInTrack;
		struct TRACK_TABLE TrackInfo[MAX_SPT];
	};
	struct Parms * pParms;


	//  Create some local variables
	//
	pParms = pTrackParms;
	uSectorsPerTrack = pBPB->sec_per_trak;
	uSectorSize = pBPB->sec_size;

	//  Set the number of sectors in a track
	//
	pParms->uSectorsInTrack = uSectorsPerTrack;

	//  Create the track layout table
	//
	for( uSectorIndex=0; uSectorIndex < uSectorsPerTrack; uSectorIndex++)
	{
		pParms->TrackInfo[uSectorIndex].uNumber = uSectorIndex+1;
		pParms->TrackInfo[uSectorIndex].uSize = uSectorSize;
	}
}


//  Return the address of the storage area for the IOCTL
//  drive parameters.
//
pDRIVE_PARMS GetDriveParameterAddress( void )
{
	static DRIVE_PARMS DriveParms;

	return &DriveParms;
}


//  Determine if the drive is valid
//
bool CheckIfValidDrive( void )
{
	uchar ucDosDriveNumber;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Determine if the drive is local or remote.  If the drive
	//  is not local that the drive is not valid, otherwise
	//  return that it is valid
	//
	return IsLocalDrive( ucDosDriveNumber );
}


//  Set the dos drive number.  (0=A, 1=B, etc...)
//
void SetDosDriveNumber( uchar ucDosDriveNumber )
{
	uchar * pucDosDriveNumber;

	pucDosDriveNumber = GetDosDriveNumberPointer();

	*pucDosDriveNumber = ucDosDriveNumber;
}


//  Return the DOS drive number (0=A, 1=B, etc...)
//
uchar GetDosDriveNumber( void )
{
	uchar * pucDosDriveNumber;

	pucDosDriveNumber = GetDosDriveNumberPointer();

	return *pucDosDriveNumber;
}


//  Return the address of the variable which contains the dos
//  drive number.  (0=A, 1=B, etc...)
//
uchar * GetDosDriveNumberPointer( void )
{
	static uchar ucDosDriveNumber = TEMP_DOS_DRIVE_NUMBER;

	return &ucDosDriveNumber;
}


//  Determine if the drive is local or remote.  If the drive
//  is not local, the drive is not valid.  Otherwise return
//  that the drive is valid.
//
bool IsLocalDrive( uchar ucDriveNumber )
{
	bool fSuccessful;
	bool fIsRemote;


	//  Return TRUE if the drive is REMOTE.
	//  Used (0=default,1=A,2=B, etc) for the drive number.
	//
	fSuccessful = dlIsDriveRemote( ucDriveNumber+1, &fIsRemote );
	if( !fSuccessful )
	{
		DisplayErrorAndExit( ERR_NOT13 );
	}

	return !fIsRemote;
}


//  Determine if we should display the signon message and if
//  so display it
//
void DisplaySignonMessage( void )
{
	char * szSignOnMessage = "FORMAT %s %s\n%s\n\n";

	if( !DoHideSignon() )
	{
		/* display sign-on */
		if( fIsFat32World )
		{
			printf(szSignOnMessage,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(szSignOnMessage,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


//  Determine what the BPB will be for the drive we
//  are formatting and save it in the boot sector
//
uchar SetBPBForDrive( void )
{
	bool fCreateNewBPB = FALSE;
	uchar ucReturnCode;


	//  In case someone is formatting B: when B: is actually A:
	//
	if( !IsHardDrive() )
	{
		dlSetLogicalDriveMap( (GetDosDriveNumber() + 1) );
	}

	//  Determine if we are quick formatting the drive
	//
	if(IsQuickFormat())
	{
		//  Retrieve the BPB from the boot sector of the hard drive
		//  we are formatting and save it in our boot sector
		//
		ucReturnCode = ReadBPBIntoBootSector();
		if( ucReturnCode )
		{
			return ucReturnCode;
		}

		//  Determine if the BPB is valid.  If the BPB is invalid
		//  display a message stating that this drive can not be
		//  quick formatted and exit.
		//
		if(!IsBPBValid())
		{
			if( QuickFormatUnformattedDrive() )
			{
				//  Create a new BPB
				//
				fCreateNewBPB = TRUE;
			}
			else
			{
				return ERR_NOQUICK;
			}
		}
		else
		{
			fCreateNewBPB = FALSE;
		}

	}
	else
	{
		//  Determine if the drive is a hard drive
		//
		if(IsHardDrive())
		{
			//  Always build a new BPB for hard drives
			//
			fCreateNewBPB = TRUE;
		}
		else
		{
			//  Determine if the user specified a format size.  If so,
			//  we don't need to check the BPB that is on the disk/drive.
			//  We only need to create a new one.
			//
			if(DoUserSize())
			{
				//  Create a new BPB with the user specified size
				//
				fCreateNewBPB = TRUE;
			}
			else
			{
				//  Retrieve the BPB from the boot sector of the hard drive
				//  we are formatting and save it in our boot sector
				//
				ucReturnCode = ReadBPBIntoBootSector();
				if( ucReturnCode )
				{
					fCreateNewBPB = TRUE;
				}
				else
				{
					//  Determine if we recognize this BPB type.
					//
					if( CheckBPBType() != BPB_TYPE_UNKNOWN )
					{
						//  The BPB is already set to a known BPB
						//
						fCreateNewBPB = FALSE;
					}
					else
					{
						//  Determine if the BPB is valid.  If the BPB is invalid
						//  we will need to build a new BPB.  Otherwise, we will
						//  use this one
						//
						if( !IsBPBValid() )
						{
							//  Create a new BPB if this BPB is invalid
							//
							fCreateNewBPB = TRUE;
						}
						else
						{
							//  This BPB appears to be valid we will try and
							//  use it.
							//
							fCreateNewBPB = FALSE;
						}
					}
				}
			}
		}
	}

	//  Determine if we need to create new BPB
	//
	if( fCreateNewBPB )
	{
		//  Create a new BPB
		//
		CreateNewBPB();
	}

	//  Copy the BPB into the device parameters
	//
	UpdateDeviceParameters();

	//  Calculate some of the fixed locations and sizes of key
	//  items on the drive
	//
	CreateDPB();

	//  Set the volume label, bios drive number, FAT ID string, etc.
	//
	UpdateBootSectorInfo();

	return 0;
}


//  Update the device parameters with the new BPB
//
void UpdateDeviceParameters( void )
{
	pDRIVE_PARMS pDriveParms;
	PBPB pBPB;


	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();

	//  Get the BPB for this drive
	//
	pBPB = GetBPBAddress();

	//  Determine if the is a FAT32 drive
	//
	if( pBPB->uFatSectors == 0 )
	{
		memcpy( &(pDriveParms->Parms.bpb), pBPB, sizeof(BPB) );
	}
	else
	{
		//  Determine if this is a FAT32 world
		//
		if( fIsFat32World )
		{
			memset( &(pDriveParms->Parms.bpb), 0, sizeof(BPB) );
			pDriveParms->Parms.bpb.uFSInfoSector = 0xffff;
			pDriveParms->Parms.bpb.uBackUpBootSector = 0xffff;
		}

		memcpy( &(pDriveParms->Parms.bpb), pBPB, sizeof(struct BootSectorBPB16) );
	}

	if( fIsFat32World )
	{
		SetTrackLayout( &pDriveParms->uSectorsInTrack, &pDriveParms->Parms.bpb );
	}
	else
	{
		SetTrackLayout( &((char *)(pDriveParms))[0x26],
				&pDriveParms->Parms.bpb );
	}
}


//  Build a new BPB from scratch
//
void CreateNewBPB( void )
{
	struct BootSectorBPB16 * pBPBType;
	pDRIVE_PARMS pDriveParms;
//	LPBPB lpBPB;
	unsigned uUserType;


	//  Determine if the user specified a size
	//
	if(DoUserSize())
	{
		//  Get the type (for the size) that the user specified
		//
		uUserType = GetUserType();

		//  Get the BPB array so we can look up this type of BPB
		//
		pBPBType = GetBPBTypeArray();

		//  Save the BPB in the boot sector
		//
		SaveBPB( (LPBPB)(&pBPBType[uUserType]) );
	}
	else
	{
		//  Get the address of the storage area for the device parameters
		//
		pDriveParms = GetDriveParameterAddress();

		if( pDriveParms->Parms.bpb.uFatSectors == 0 )
		{
			pDriveParms->Parms.bpb.uFSInfoSector = 1;
			pDriveParms->Parms.bpb.uBackUpBootSector = 6;
		}

		//  Save the BPB from the drive parameters in the boot sector
		//
		SaveBPB( &(pDriveParms->Parms.bpb) );
	}
}


//  Determine if we recognize this type of BPB (in the boot sector)
//  and return the type that we recognized it as.
//
uchar CheckBPBType( void )
{
	PBPB pBPB;
	struct BootSectorBPB16 * pBPBType;
	unsigned uBPBIndex;
	int		iResult;


	//  Get the address of the BPB
	//
	pBPB = GetBPBAddress();

	//  Get the address of the hard coded BPBs
	//
	pBPBType = GetBPBTypeArray();

	//  Determine if this BPB matches any of our hard coded BPBs
	//
	for( uBPBIndex = 0; uBPBIndex < BPB_TYPE_UNKNOWN; uBPBIndex++ )
	{
		iResult = memcmp( pBPB, &pBPBType[uBPBIndex],
				sizeof(struct BootSectorBPB16));
		if( iResult == 0 )
		{
			break;
		}
	}

	return uBPBIndex;
}


//  Determine if the BPB (in the boot sector) is valid
//
bool IsBPBValid( void )
{
	PBPB pBPB;
	uchar ucMedia;


	//  Get the BPB
	//
	pBPB = GetBPBAddress();

	//  Make sure the media ID is valid
	//
	ucMedia = pBPB->media & 0xF0;
	if( ucMedia != 0xF0)
	{
		//  Invalid BPB
		//
		return FALSE;
	}

	//  Make sure the sector size is valid
	//
	if(( pBPB->sec_size % 128 ) || (!pBPB->sec_size))
	{
		//  Invalid BPB
		//
		return FALSE;
	}

	//  Check the total sectors
	//
	if(( pBPB->num_sec == 0 ) && ( pBPB->lnumsec == 0 ) )
	{
		//  Invalid BPB
		//
		return FALSE;
	}

	//  Check if this is a FAT32 BPB
	//
	if( pBPB->uFatSectors == 0 )
	{
		//
		//  BPB is a FAT32 BPB.  Check to make sure that the
		//  BPB contains some value for the starting cluster
		//  of the root directory.
		//

		if( pBPB->ulRootDirStartCluster < 2 )
		{
			return FALSE;
		}

		//  There is no longer a limit to the number of root directory
		//  entries you can have.  In a FAT32 BPB this field is unused
		//  and should be zero.
		//
		if( pBPB->uDirEntries != 0 )
		{
			return FALSE;
		}
	}

	//  Seems to be a valid BPB
	//
	return TRUE;
}


//  Read the boot sector off the drive we are formatting and
//  store it in the appropriate boot sector (FAT12/FAT16 or FAT32
//  boot sectors)
//
uchar ReadBPBIntoBootSector( void )
{
	unsigned uSegment;
	bool fSuccessful;
	uchar ucDosDriveNumber;
	lpBOOTSECTOR lpBootSector;
	unsigned uAbsErrorCode;


	//  Get the dos drive number of the drive we are formatting
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Allocate a buffer to read in the boot sector
	//
	fSuccessful = dlAllocMem( (MAX_SECTOR_SIZE >> 4)*2, &uSegment );
	if( !fSuccessful )
	{
		DisplayErrorAndExit( ERR_NOMEMORY );
	}
	lpBootSector = MK_FP(uSegment,0);

	//  Determine if we are in a DOS 7.1 enviornment
	//
	if( fIsFat32World )
	{
		//  Read in the boot sector (drive 0=default,1=A,...)
		//
		fSuccessful = dlAbsRead32( ucDosDriveNumber+1, 2, 0, lpBootSector );
	}
	else
	{
		//  Read in the boot sector (drive 0=A,1=B,...)
		//
		fSuccessful = dlAbsDiskRead( ucDosDriveNumber, 1, 0, lpBootSector );
	}
	if(!fSuccessful)
	{
		uAbsErrorCode = __dlCheckAbsRdWrError();
		if( (uchar)uAbsErrorCode == 0x02 )
		{
			return ERR_DRVNOTREADY;
		}
		return ERR_READ;
	}

	//  Save this BPB into our boot sector
	//
	SaveBPB( &(lpBootSector->BootSectorInfo.BootInfo32.BootSectorBPB) );

	//  Release the buffer for the boot sector
	//
	fSuccessful = dlFreeMem( uSegment );
	if( !fSuccessful )
	{
		DisplayErrorAndExit( ERR_RELEASE );
	}

	return 0;
}


//  Save this BPB into our boot sector
//
void SaveBPB( LPBPB lpBPB )
{
	unsigned uBPBSize;


	//  Determine whether this is a FAT16 or a FAT32 BPB in order
	//  to find out the size of the BPB
	//
	if((lpBPB->uFatSectors == 0) && fIsFat32World )
	{
		//  Make sure we are sysing a FAT32 drive with a FAT32 kernel
		//
		if( DoCopySystem() && (!fIsFat32Binary) )
		{
			printf(szErrorNotFat32Kernel);
			exit( MS_ERR_FATAL_ERROR );
		}

		uBPBSize = sizeof(BPB);
		_fmemcpy( &rgucBootSector32[BOOT_BPB], lpBPB, uBPBSize );
		SetWhichBootSectorToUse( USE_FAT32_BOOT_SECTOR );
	}
	else
	{
		uBPBSize = sizeof(struct BootSectorBPB16);
		_fmemcpy( &rgucBootSector[BOOT_BPB], lpBPB, uBPBSize );
		SetWhichBootSectorToUse( USE_FAT16_BOOT_SECTOR );
	}
}


//  Set address of the variable which contains the address of the
//  boot sector to the address of the boot sector we will be
//  using.
//
void SetWhichBootSectorToUse( uchar ucType )
{
	void * * ppBootSector;

	ppBootSector = GetBootSectorPointer();
	if( ucType == USE_FAT32_BOOT_SECTOR )
	{
		*ppBootSector = &rgucBootSector32[0];
	}
	else
	{
		*ppBootSector = &rgucBootSector[0];
	}
}

//  Get the address of the BPB
//
PBPB GetBPBAddress( void )
{
	void * * ppBootSector;
	pBOOTSECTOR pBootSector;

	ppBootSector = GetBootSectorPointer();
	pBootSector  = *ppBootSector;
	return (PBPB)&(pBootSector->BootSectorInfo);
}


//  Return the address of the variable which contains the
//  address of the boot sector
//
void ** GetBootSectorPointer( void )
{
	static void * pBootSector;

	return &pBootSector;
}


//  Determine if the drive we are formatting is a fixed disk
//
bool IsHardDrive( void )
{
	pDRIVE_PARMS pDriveParms;


	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();

	//  Determine if the device type is a fixed disk
	//
	return (pDriveParms->Parms.devType == 5);
}


//  Copy the system files to the drive if we are making
//  this disk bootable
//
void WriteTheSystemFiles( void )
{
	uchar ucDosDriveNumber;
	bool  fFat32Drive;
	PDPB pDPB;


	if(DoCopySystem())
	{
		//  Get the dos drive number
		//
		ucDosDriveNumber = GetDosDriveNumber();

		//  Determine if we are sysing a FAT32 drive
		//
		pDPB = GetDPBAddress();
		fFat32Drive = IsFat32Drive(pDPB);
		CopySystemFiles( ucDosDriveNumber + 'A', DoConfirmDelete(),
				DoHideSystemFiles(), fFat32Drive );

		printf(szSystemTransfered);
	}
}


//  Place the volume label, default serial number (0 like MSDOS), and
//  bios drive number in the boot sector
//
void UpdateBootSectorInfo( void )
{
	union BootInformation * pBS;
	char * szVolumeLabel;
	char * szFatTypeString;


	//  Get the BPB in the boot sector
	//
	pBS = (union BootInformation *)GetBPBAddress();

	//  Determine if this is a FAT32 drive
	//
	if( pBS->BootInfo16.BootSectorBPB.uFatSectors == 0 )
	{
		if(IsHardDrive())
		{
			pBS->BootInfo32.uPhysicalDriveNumber = 0x80;
		}
		else
		{
			pBS->BootInfo32.uPhysicalDriveNumber = 0x00;
		}

		pBS->BootInfo32.ulVolumeSerialNumber = 0;

		memset( &(pBS->BootInfo32.rgucVolumeLabel[0]), 0x20, 11 );
		memset( &(pBS->BootInfo32.rgucFileSystemID[0]), 0x20, 8 );

		//  Set the volume label
		//
		szVolumeLabel = GetVolumeLabelAddress();
		memcpy( &(pBS->BootInfo32.rgucVolumeLabel[0]), szVolumeLabel,
				strlen(szVolumeLabel) );

		//  Set the fat id string
		//
		szFatTypeString = GetFatTypeString();
		memcpy( &(pBS->BootInfo32.rgucFileSystemID[0]), szFatTypeString,
				strlen(szFatTypeString) );
	}
	else
	{
		if(IsHardDrive())
		{
			pBS->BootInfo16.uPhysicalDriveNumber = 0x80;
		}
		else
		{
			pBS->BootInfo16.uPhysicalDriveNumber = 0x00;
		}

		pBS->BootInfo16.ulVolumeSerialNumber = 0;

		memset( &(pBS->BootInfo16.rgucVolumeLabel[0]), 0x20, 11 );
		memset( &(pBS->BootInfo16.rgucFileSystemID[0]), 0x20, 8 );

		//  Set the volume label
		//
		szVolumeLabel = GetVolumeLabelAddress();
		memcpy( &(pBS->BootInfo16.rgucVolumeLabel[0]), szVolumeLabel,
				strlen(szVolumeLabel) );

		//  Set the fat id string
		//
		szFatTypeString = GetFatTypeString();
		memcpy( &(pBS->BootInfo16.rgucFileSystemID[0]), szFatTypeString,
				strlen(szFatTypeString) );
	}
}


//  Required by the FAT libraries but not used in format
//
#pragma argsused
bool PASCAL SetBufferCopiesAndDistance( void * pBufferHandle,
		unsigned uCopiesOfThisSector, ulong ulDistanceToNextSector )
{
	return TRUE;
}


//  Required by the FAT libraries but not used in format
//
#pragma argsused
bool PASCAL MarkBufferDirty( void * pBufferHandle )
{
	return TRUE;
}


//  Return the address of the buffer for the specified handle
//
bool PASCAL GetBufferDataAddress( void * pBufferHandle,
		void far * * plpBufferAddress )
{
	PBPB 		pBPB;
	unsigned uBufferSegment;
	unsigned uBufferOffset;
	unsigned uBufferHandle;


	//  Get the BPB address
	//
	pBPB = GetBPBAddress();

	//  Get the starting segment (offset is 0) of the fat buffers
	//
	uBufferSegment = GetBufferSegment();

	//  Determine the offset which contains this FAT sector
	//
	uBufferHandle = *(unsigned *)pBufferHandle;
	uBufferOffset = (pBPB->sec_size) * uBufferHandle;

	//  Set the address of the request sector in the return variable
	//
	*plpBufferAddress = MK_FP( uBufferSegment, uBufferOffset);

	return TRUE;
}


//  Return a handle for the specified sector
//
#pragma argsused
bool PASCAL ReadSectorBuffer( uchar ucDOSDriveNumber,
		ulong ulAbsSector, void ** ppBufferHandle )
{
	static unsigned uBufferHandle;
	pFATBUF pBufferInfo;

	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Make sure we have the requested sector in the buffers
	//
	if( (pBufferInfo->uNumberOfSectors == 0) ||
			(ulAbsSector < pBufferInfo->ulFirstSectorNumber) ||
			((pBufferInfo->ulFirstSectorNumber +
			pBufferInfo->uNumberOfSectors) <= ulAbsSector) )
	{
		FlushBuffers();
		LoadBuffers( ulAbsSector );
	}

	//  Return a handle to the buffer
	//
	uBufferHandle = (unsigned)(ulAbsSector -
			pBufferInfo->ulFirstSectorNumber);
	*ppBufferHandle = &uBufferHandle;

	return TRUE;
}


//  Return the address of the buffer management structure
//
pFATBUF GetBufferInfoAddress( void )
{
	static FATBUF BufferInfo;

	return &BufferInfo;
}


//  Write all the FAT sector buffers to disk.
//
void FlushBuffers( void )
{
	PDPB pDPB;
	ulong ulStartingSector;
	ulong ulCopyDistance;
	unsigned uNumberOfSectors;
	uchar ucCopies;
	uchar ucDosDriveNumber;
//	unsigned uTemp;
	unsigned uBufferSegment;
	unsigned uBufferOffset;
	void far * lpBuffer;
	bool fSuccessful;
	pFATBUF pBufferInfo;


	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Determine the number of sectors we need to write
	//
	uNumberOfSectors = pBufferInfo->uNumberOfSectors;

	//  Determine if the buffer is empty.  If the buffer is
	//  not empty then we need to commit the sectors.  Otherwise,
	//  we don't need to do anything.
	//
	if( (pBufferInfo->fDoWrite) && (uNumberOfSectors) )
	{
		//  Get the DPB address
		//
		pDPB = GetDPBAddress();

		//  Determine the drive number for which we are buffering
		//
		ucDosDriveNumber = pDPB->drive;

		//  Get the address of the buffer
		//
		uBufferSegment = GetBufferSegment();
		uBufferOffset  = 0;
		lpBuffer = MK_FP( uBufferSegment, uBufferOffset );

		//  Determine which the starting sector of the FAT buffer
		//
		ulStartingSector = pBufferInfo->ulFirstSectorNumber;

		//  Determine the number of copies for each sector
		//
		ucCopies = pBufferInfo->uCopies;

		//  Determine the distance between copies
		//
		ulCopyDistance = pDPB->fat_len;
		if( !ulCopyDistance )
		{
			ulCopyDistance = pDPB->ulSectorsPerFAT;
		}

		//  Determine if this is a DOS 7.1 enviornment
		//
		if( fIsFat32World )
		{
			while( ucCopies )
			{
				//  Write the sectors for this copy (drive 0=default,1=A,...)
				//
				fSuccessful = dlAbsWrite32( ucDosDriveNumber+1, uNumberOfSectors,
						ulStartingSector, lpBuffer, ABS32WRITE_UNKNOWN );
				if(!fSuccessful)
				{
					printf("\nERROR: %04x\n", __dlCheckAbsRdWrError() );
					DisplayErrorAndExit( ERR_WRITE );
				}

				//  Mark that we wrote this copy
				//
				ucCopies--;

				//  Determine the starting sector for the next copy
				//
				ulStartingSector += ulCopyDistance;
			}
		}
		else
		{
			while( ucCopies )
			{
				//  Write the sectors for this copy (drive 0=A,1=B,...)
				//
				fSuccessful = dlAbsDiskWrite( ucDosDriveNumber, uNumberOfSectors,
						ulStartingSector, lpBuffer );
				if(!fSuccessful)
				{
					DisplayErrorAndExit( ERR_WRITE );
				}

				//  Mark that we wrote this copy
				//
				ucCopies--;

				//  Determine the starting sector for the next copy
				//
				ulStartingSector += ulCopyDistance;
			}
		}

		//  Record that all these buffers are written
		//
		pBufferInfo->uNumberOfSectors = 0;
	}
}


//  Load the next block of FAT sectors
//
void LoadBuffers( ulong ulStartingSector )
{
	void far * lpBuffer;
	PDPB pDPB;
	pFATBUF pBufferInfo;
	unsigned uNumberOfSectors;
	ulong ulStartOfNextFAT;
	bool fSuccessful;
	uchar ucDosDriveNumber;
	unsigned uSectorSize;
//	unsigned uTemp;
	unsigned uBufferSegment;
	unsigned uBufferOffset;
	bool fIsFatSectors;


	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Get the address of the buffer
	//
	uBufferSegment = GetBufferSegment();
	uBufferOffset = 0;
	lpBuffer = MK_FP( uBufferSegment, uBufferOffset );

	//  Get the number of buffers
	//
	uNumberOfSectors = GetNumberOfBuffers();

	//  Get the DPB address
	//
	pDPB = GetDPBAddress();

	//  Determine the drive number for which we are buffering
	//
	ucDosDriveNumber = pDPB->drive;

	//  Determine the starting sector of the next copy of the FAT
	//
	ulStartOfNextFAT = pDPB->fat_start;
	if( pDPB->fat_len )
	{
		ulStartOfNextFAT += pDPB->fat_len;
	}
	else
	{
		ulStartOfNextFAT += pDPB->ulSectorsPerFAT;
	}

	//  Determine if these are fat sectors
	//
	if( ulStartOfNextFAT <= ulStartingSector )
	{
		fIsFatSectors = FALSE;
		pBufferInfo->uCopies = 1;
	}
	else
	{
		fIsFatSectors = TRUE;
		pBufferInfo->uCopies = 2;
	}

	//  Determine how many sectors we are loading
	//
	if( fIsFatSectors )
	{
		if( (ulStartingSector + uNumberOfSectors) >= ulStartOfNextFAT )
		{
			uNumberOfSectors = (unsigned)(ulStartOfNextFAT - ulStartingSector);
		}
	}

	//  Save what is in the buffers
	//
	pBufferInfo->uNumberOfSectors    = uNumberOfSectors;
	pBufferInfo->ulFirstSectorNumber = ulStartingSector;

	//  Save the sector size into a local variable
	//
	uSectorSize   = pDPB->bytes_per_sec;

	//  Determine if this is a quick format.  If it is a quick format
	//  we will need to read the sectors off the disk.  Otherwise, we
	//  only need to zero the buffer.
	//
	if(pBufferInfo->fDoRead)
	{
		//  Determine if this is a DOS 7.1 enviornment
		//
		if( fIsFat32World )
		{
			//  Read the sectors into the buffer (drive 0=default,1=A,...)
			//
			fSuccessful = dlAbsRead32( ucDosDriveNumber+1, uNumberOfSectors,
					ulStartingSector, lpBuffer );
		}
		else
		{
			//  Read the sectors into the buffer (drive 0=A,1=B,...)
			//
			fSuccessful = dlAbsDiskRead( ucDosDriveNumber, uNumberOfSectors,
					ulStartingSector, lpBuffer );
		}
		if(!fSuccessful)
		{
			DisplayErrorAndExit( ERR_READ );
		}
	}
	else
	{
		while( uNumberOfSectors )
		{
			//  Fill this sector with zeros
			//
			_fmemset( MK_FP( uBufferSegment, uBufferOffset), 0, uSectorSize );

			//  Record that we filled this sector
			//
			uNumberOfSectors--;

			//  Determine the next buffer offset
			//
			uBufferOffset += uSectorSize;
		}
	}
}


//  Zero the root directory and commit it
//
void WriteTheRootDirectory( void )
{
//	void far * lpBuffer;
	PDPB pDPB;
	pFATBUF pBufferInfo;
	ulong ulStartingSector;
	unsigned uNumberOfSectors;
//	uchar ucDosDriveNumber;
//	unsigned uSectorSize;


	//  Set the buffer mode
	//
	SetBufferMode( BUFFER_NO_READ, BUFFER_WRITE );

	//  Get the DPB address
	//
	pDPB = GetDPBAddress();

	//  Get the address of the buffer management structure
	//
	pBufferInfo = GetBufferInfoAddress();

	//  Determine if this is a FAT32 drive
	//
	if(IsFat32Drive(pDPB))
	{
		//  Determine the number of sectors in the root
		//
		uNumberOfSectors = pDPB->ms_spc+1;

		//  Save the root directory cluster number in the BPB
		//
		ulStartingSector = pDPB->ulRootDirStartCluster;

		//  Convert this cluster to a sector value
		//
		ulStartingSector -= 2;
		ulStartingSector *= uNumberOfSectors;
		ulStartingSector += pDPB->ulFirstDataSector;
	}
	else
	{
		//  Determine the number of sectors in the root directory
		//
		uNumberOfSectors = pDPB->data_start - pDPB->dir_start;

		//  Determine what the starting sector for the root is.
		//
		ulStartingSector = pDPB->dir_start;
	}

	//  Load these sectors into the buffers (will be zerod)
	//
	LoadBuffers( ulStartingSector );

	//  Save what is in the buffers we will
	//
	pBufferInfo->uNumberOfSectors    = uNumberOfSectors;

	//  Flush the directory sectors
	//
	FlushBuffers();
}


//  Return the starting cluster for the root
//  directory of a FAT32 drive
//
ulong GetFAT32RootCluster( void )
{
	pBAD_CLUSTER_INFO pBadClusterInfo;
	ulong far * lpEntry;
	unsigned uEntryNumber;
	unsigned uEntryIndex = 0;
	ulong ulRootCluster = 2;

	pBadClusterInfo = GetBadClusterInfoPointer();
	uEntryNumber = pBadClusterInfo->uNumberOfEntries;
	lpEntry = MK_FP( pBadClusterInfo->uListSegment, 0);
	while( uEntryNumber < uEntryIndex )
	{
		if( ulRootCluster == lpEntry[uEntryIndex] )
		{
			ulRootCluster++;
			uEntryIndex++;
		}
		else
		{
			break;
		}
	}

	return ulRootCluster;
}


//  Return the number of sector buffers which were allocated
//
unsigned GetNumberOfBuffers( void )
{
	unsigned * puNumberOfBuffers;

	puNumberOfBuffers = GetNumberOfBuffersPointer();
	return *puNumberOfBuffers;
}


//  Return the address of the variable which contains the
//  number of allocated sector buffers
//
unsigned * GetNumberOfBuffersPointer( void )
{
	static unsigned uNumberOfBuffers = 0;

	return &uNumberOfBuffers;
}


//  Set the number of allocated sector buffers
//
void SetNumberOfBuffers( unsigned uNumberOfBuffers )
{
	unsigned * puNumberOfBuffers;

	puNumberOfBuffers = GetNumberOfBuffersPointer();
	*puNumberOfBuffers = uNumberOfBuffers;
}


//  Return the starting segment of the allocated buffers
//
unsigned GetBufferSegment( void )
{
	unsigned * puBufferSegment;

	puBufferSegment = GetBufferSegmentPointer();
	return *puBufferSegment;
}


//  Set the starting segment of the allocated buffers
//
void SetBufferSegment( unsigned uBufferSegment )
{
	unsigned * puBufferSegment;

	puBufferSegment = GetBufferSegmentPointer();
	*puBufferSegment = uBufferSegment;
}


//  Return the address of the variable which contains the
//  starting segment of the allocated buffers
//
unsigned * GetBufferSegmentPointer( void )
{
	static unsigned uBufferSegment;

	return &uBufferSegment;
}


//  Create a buffer for the FAT sectors
//
void AllocateFATBuffers( void )
{
	bool fSuccessful;
	unsigned uFreeMemPara;
	unsigned uSectorPara;
	unsigned uNumberOfBuffers;
	unsigned uRequiredBuffers;
	unsigned * puBufferSegment;
	unsigned * puNumberOfBuffers;
	PDPB pDPB;


	//  Retrieve the (our local copy of the) DPB for the
	//  drive we are formatting
	//
	pDPB = GetDPBAddress();

	//  Determine the amount of available conventional memory
	//
	dlAllocMem( MAX_PARAGRAPH_SIZE, &uFreeMemPara );

	//  Determine if there is more than 64k available.  If so,
	//  change the varaiable so that it reflects that we only
	//  have 64k available.  We don't want to use more than
	//  64k anyway.
	//
	if( uFreeMemPara > 2048 )
	{
		uFreeMemPara = 2048;
	}

	//  Calculate how many buffers we can have in this amount
	//  of memory.
	//
	uSectorPara = pDPB->bytes_per_sec >> 4;
	uNumberOfBuffers = uFreeMemPara / uSectorPara;

	//  Determine the minimum number of buffer we need
	//
	if( IsFat32Drive(pDPB) )
	{
		uRequiredBuffers = (pDPB->ms_spc)+1;
	}
	else
	{
		uRequiredBuffers = pDPB->data_start - pDPB->dir_start;
	}

	//  Make sure that we have the required number of buffers
	//
	if( uRequiredBuffers > uNumberOfBuffers )
	{
		DisplayErrorAndExit( ERR_NOMEMORY );
	}

	//  Get the address of the variable which contains the allocated
	//  buffer segment
	//
	puBufferSegment = GetBufferSegmentPointer();

	//  Actually allocate the memory
	//
	fSuccessful = dlAllocMem( (unsigned)(uSectorPara*uNumberOfBuffers),
			puBufferSegment );
	if(!fSuccessful)
	{
		DisplayErrorAndExit( ERR_NOMEMORY );
	}

	//  Save the number of buffers
	//
	puNumberOfBuffers  = GetNumberOfBuffersPointer();
	*puNumberOfBuffers = uNumberOfBuffers;
}


//  Release the memory allocated for the FAT buffers
//
void FreeFATBuffers( void )
{
	bool fSuccessful;
	unsigned * puBufferSegment;
	unsigned * puNumberOfBuffers;


	//  Get the address of the variable which contains the allocated
	//  buffer segment
	//
	puBufferSegment = GetBufferSegmentPointer();

	//  Free the memory
	//
	fSuccessful = dlFreeMem( *puBufferSegment );
	if(!fSuccessful)
	{
		DisplayErrorAndExit( ERR_RELEASE );
	}

	//  Erase the buffer segment
	//
	*puBufferSegment = 0;

	//  Save the number of buffers
	//
	puNumberOfBuffers  = GetNumberOfBuffersPointer();
	*puNumberOfBuffers = 0;
}


//  Prompt the user to insert a disk if this is a removable
//  drive/disk
//
void PromptForNewDisk( void )
{
	uchar ucDosDriveNumber;

	//  Get the drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	if( DoConfirmFormat() && !Autotest() )
	{
		if( IsHardDrive() )
		{
			//  Display a message for the user
			//
			printf( szWarnLostData, ucDosDriveNumber + 'A');

			//  Make sure the user wants to
			//
			if(!getyesno())
			{
				exit( MS_ERR_NO_PROCEED );
			}
		}
		else
		{
			//  Display a message for the user
			//
			printf( newdisk_str, ucDosDriveNumber + 'A');

			//  Make the user press enter
			//
			getyesno();
		}
	}
}


/*
	return TRUE for a yes, 0 for anything else
*/
int getyesno(void)
{
	static char s[20];

	getstring(s);
	printf("\n\n");
	return (toupper(s[2]) == 'Y');
}


/*
	get a string, clearing buffer first
*/
void getstring(char * string)
{
	string[0] = 12;
	string[1] = 0;
	dlFlushAndGetLine( (uchar *)(string) );
}


void DisplayFormatMessage( void )
{
	PBPB pBPB;
	ulong ulTotalSectors;
	char * szSizeString;
	int	btype;				/* 0=360, 1=1.2M, 2=720, 3=1.44M, etc... */


	//  Get the BPB
	//
	pBPB = GetBPBAddress();

	//  Get the size string
	//
	ulTotalSectors = pBPB->num_sec;
	if( !ulTotalSectors )
	{
		ulTotalSectors = pBPB->lnumsec;
	}
	btype = CheckBPBType();
	szSizeString = drtype_name(btype, ulTotalSectors, pBPB->sec_size);

	//  Display the formatting message
	//
	if(IsQuickFormat())
	{
		printf(quick_str);
	}

	printf( format_type_str, szSizeString );
}


//  Return the address of a string which contain the scaled disk size
//
char * drtype_name(int btype, ulong numberOfSectors, unsigned bytesPerSector)
{
	static char s[20];
	static char *name[] = { "360K", "1.2M", "720K", "1.44M", "2.88M" };
	char * fixed = "";  // no "Fixed" like MSDOS 7.1

	/* size must be calculated */
	if(btype >= BPB_TYPE_UNKNOWN)
	{
		get_3digit(s, numberOfSectors, bytesPerSector);
		if(IsHardDrive())
			strcat(s,fixed);
		return s;
	}
	else
		return name[btype];
};


//  Return 3 significant digits of the size
//
void get_3digit(char *s, ulong numberOfSectors, unsigned bytesPerSector)
{
	static char sizemodchar[] = { 0, 'K', 'M', 'G' };
	int sizemod = 0;
	int intval;
	ulong val;
	unsigned uTemp;

	if( numberOfSectors > 100000L )
	{
		uTemp = 1024 / bytesPerSector;
		val = numberOfSectors / uTemp;
		sizemod++;
	}
	else
	{
		val = numberOfSectors * bytesPerSector;
	}

	while(val >= 100000L)
	{
		val /= 1024;
		++sizemod;
	}
	if(val >= 10000L)
	{
		intval = (int)(val / 100);
		++sizemod;
		sprintf(s,"%u.%u%c",intval/10,intval%10,sizemodchar[sizemod]);
	}
	else if(val >= 1000L)
	{
		intval = (int)val/10;
		++sizemod;
		sprintf(s,"%u.%02u%c",intval/100,intval%100,sizemodchar[sizemod]);
	}
	else
		sprintf(s,"%u%c",(int)val,sizemodchar[sizemod]);
}


/*
	Write out a volume label.  This label works with
	any kind of label.
*/
void volume_label( void )
{
	uchar ucDosDriveNumber;
	int i,j;
	XFCB XFCB;
	char * szVolume;
	char * szBadChar = NULL;
	unsigned uVolumeLength;


	//  Get the drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	szVolume = GetVolumeLabelAddress();

	if(!DoHaveLabel())
	{
		do
		{
			printf( getvol_str, ucDosDriveNumber + 'A');
			getstring(volume_str);
			printf("\n");
			if(volume_str[1])
			{
            while( volume_str[2] == ' ' )
            {
            	strcpy(volume_str+2,volume_str+3);
               (char)volume_str[1]--;
            }

				// Make sure volume string is UPPER cased
				//
				for(i=0,j=2; i < volume_str[1]; i++,j++)
				{
					if( szBadChar == NULL )
					{
						szBadChar = strchr( illegalVolChars,
								volume_str[j]);
					}
					volume_str[j] = toupper(volume_str[j]);
				}

				/* set up the string */
				if(volume_str[1] < 11)
					memset(volume_str+2+volume_str[1],' ',11-volume_str[1]);

				szVolume = &volume_str[2];
			}
		}while( szBadChar != NULL );
	}

	/* if user entered a volume name, place it on the disk */
	if((DoHaveLabel()) || (volume_str[1]))
	{
		// Set the volume label MSDOS style
		//

		memset( &XFCB, 0, sizeof(XFCB) );
		memcpy( XFCB.name, "???????????", 11 );
		XFCB.uXFCBIndicator = 0xFF;
		XFCB.uAttributes = 0x08;
		XFCB.drive = ucDosDriveNumber+1;
		dlFCBDelete( (PFCB)&XFCB );

		uVolumeLength = strlen( szVolume );
		while( uVolumeLength < 11 )
		{
			strcat( szVolume, " " );
			uVolumeLength++;
		}

		memset( &XFCB, 0, sizeof(XFCB) );
		memcpy( XFCB.name, szVolume, 8 );
		memcpy( XFCB.ext, &szVolume[8], 3 );
		XFCB.uXFCBIndicator = 0xFF;
		XFCB.uAttributes = 0x08;
		XFCB.drive = ucDosDriveNumber+1;
		dlFCBCreate( (PFCB)&XFCB );
	}
}


//  Determine if this track is in the data area.  If it is not
//  in the data area return FALSE.  Otherwise, mark all the
//  clusters in this track as bad and return TRUE.
//
uchar MarkClustersInTrackBad( unsigned uTrack, unsigned uHead )
{
	PDPB pDPB;
	pDRIVE_PARMS pDriveParms;
	uchar ucSPT;
	ulong ulDataStart;
	ulong ulSectorNumber;
	unsigned uTotalHeads;
	unsigned uNumberOfClusters;
	ulong ulClusterEntry;
	uchar ucSPC;
	uchar ucReturnCode = 0;


	//  Get our DPB
	//
	pDPB = GetDPBAddress();
	ucSPC = pDPB->ms_spc + 1;
	if(IsFat32Drive(pDPB))
	{
		ulDataStart = pDPB->ulFirstDataSector;
	}
	else
	{
		ulDataStart = pDPB->data_start;
	}

	//  Get the address of the storage area for the device parameters
	//
	pDriveParms = GetDriveParameterAddress();
	ucSPT = pDriveParms->Parms.bpb.sec_per_trak;
	uTotalHeads = pDriveParms->Parms.bpb.sides;

	//  Determine the sector number for this track
	//
	ulSectorNumber = ((ulong)uTrack * uTotalHeads * ucSPT) +
			((ulong)uHead * ucSPT);

	//  Make sure this sector is in the data area
	//
	if( ulSectorNumber < ulDataStart )
	{
		return 1;
	}

	//  Determine the first cluster which is bad
	//
	ulClusterEntry = (ulSectorNumber - ulDataStart)/ucSPC;
	ulClusterEntry += 2;

	//  Determine the number of clusters which are bad
	//
	uNumberOfClusters = (unsigned)ucSPT / ucSPC;
	if( (unsigned)ucSPT % ucSPC )
	{
		uNumberOfClusters++;
	}

	//  Mark all of these clusters as bad
	//
	while( uNumberOfClusters && (!ucReturnCode) )
	{
		ucReturnCode = RecordBadCluster(ulClusterEntry);
		uNumberOfClusters--;
		ulClusterEntry++;
	}

	return ucReturnCode;
}


//  Display the statistics about the disk
//
void DisplayStatistics( void )
{
	ulong stat[NUMSTAT];
	pBAD_CLUSTER_INFO pBadClusterInfo;
	unsigned uClusterSize;
	PDPB pDPB;
	unsigned uIndex;
	ulong ulNumberOfClusters;
	uchar ucDosDriveNumber;
	GDF32 DiskSpace;
	ulong ulSerialNumber;
	unsigned uSPC;
//	unsigned uTemp;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Get our DPB for this drive
	//
	pDPB = GetDPBAddress();
	uSPC = pDPB->ms_spc + 1;
	uClusterSize = uSPC * pDPB->bytes_per_sec;

	//  reset all statistics
	//
	for(uIndex=0; uIndex<NUMSTAT; uIndex++)
		stat[uIndex] = 0L;

	//  Determine the total disk space
	//
	ulNumberOfClusters = pDPB->hiclus;
	if( !ulNumberOfClusters )
	{
		ulNumberOfClusters = pDPB->ulMaxCluster;
	}

	//  pDPB->ulMaxCluster or pDPB->hiclus is the maximum possible
	//  cluster number for this drive.  The first two clusters are
	//  not really usable space so TotalCluster = MaxCluster - 2.
	//  Now TotalCluster is still a zoro base value so convert it
	//  to a 1 based value by TotalCluster += 1.  Overall result
	//
	//  -2+1 = -1
	//
	stat[0] = ulNumberOfClusters-1;		// in clusters

#if FMT_DEBUG
	//  Display the total number of clusters
	//
	printf("Total Clusters: %lu\n", stat[0] );
#endif

	//  Get total bad sectors
	//
	pBadClusterInfo = GetBadClusterInfoPointer();
	stat[1] = (ulong)pBadClusterInfo->uNumberOfEntries;  // in clusters

#if FMT_DEBUG
	//  Display the total number of bad clusters
	//
	printf("Total Bad Clusters: %lu\n", stat[1] );
#endif

	//  Free Space Left
	//
	SmartGetFreeSpace( ucDosDriveNumber+1, &DiskSpace );
	stat[3] = DiskSpace.ulAvailableClusters;

#if FMT_DEBUG
	//  Display the remaining free clusters
	//
	printf("Total Free Clusters: %lu\n", stat[3] );
#endif

	if(IsFat32Drive(pDPB))
	{
		//  For FAT32 drives, add a cluster for the root directory.
		//  MSDOS 7.1 displays it in this fashion.
		//
		stat[3] += 1;
	}

	//  Calculate the space used by system files
	//
	stat[2] = stat[0] - stat[3] - stat[1];

#if FMT_DEBUG
	//  Display the remaining free space
	//
	printf("Total space used by system: %lu\n", stat[2] );
#endif

	printf("\n");
	for(uIndex=0;uIndex<NUMSTAT;uIndex++)
	{
		if(!stat[uIndex])
			continue;
		DisplayScaledNumber( stat[uIndex], uClusterSize );
		printf(stat_msg[uIndex]);
	}

	printf("\n");
	printf(stat_str3,(ulong)uClusterSize,stat_msg[4]);
	printf(stat_str3,DiskSpace.ulAvailableAllocationUnits, stat_msg[5]);

	ulSerialNumber = GetSerialNumber();
	printf( volume_serial, ((unsigned *)(&ulSerialNumber))[1],
			((unsigned *)(&ulSerialNumber))[0] );
}



void DisplayScaledNumber( ulong ulVal1, unsigned uVal2 )
{
	unsigned uTemp;
	ulong   ulTemp;


	//  Determine if the value exceeds 999,999,999
	//
	ulTemp = ulVal1;
	ulTemp /= 1024;
	ulTemp *= uVal2;
	if( ulTemp > 976562L )
	{
		if( uVal2 > 1024 )
		{
			ulTemp = ulVal1;
			uVal2 /= 1024;
			ulTemp *= uVal2;
		}

		//  Convert the value into Mb
		//
		uTemp = (unsigned)(ulTemp % 1024);
		uTemp = uTemp / 10;
		ulTemp /= 1024;

		printf(stat_str2,ulTemp,uTemp);
	}
	else
	{
		ulVal1 *= uVal2;
		printf(stat_str1,ulVal1);
	}
}


//  Process all the command line arguments
//
void DoParseCommandLine( int argc, char *argv[] )
{
	static char szUserSize[5];
	static char szUserLabel[14];
   static char szAutotest[9];       // will receive 'utotest'
	char * szVolume;
	uchar ucDriveLetter;
	unsigned uIndex;
	unsigned uCount;
	static ARGDEF FormatOpts[] =
	{
		{ BOOL_SWITCH, &fSwitches[DONT_CONFIRM],     'c', 0 },
		{ BOOL_SWITCH, &fSwitches[DONT_HIDE_SYSTEM], 'h', 0 },
		{ BOOL_SWITCH, &fSwitches[QUICK_FORMAT],     'q', 0 },
		{ BOOL_SWITCH, &fSwitches[HIDE_SIGNON],      '[', 0 },
		{ BOOL_SWITCH, &fSwitches[COPY_SYSTEM],      's', 0 },
		{ BOOL_SWITCH, &fSwitches[FORCE_360K],       '4', 0 },
		{ BOOL_SWITCH, &fSwitches[FORCE_720K],       '7', 0 },
		{ BOOL_SWITCH, &fSwitches[UNCONDITIONAL],    'u', 0 },
		{ BOOL_SWITCH, &fSwitches[QF_UNFORMATTED],   'z', 0 },
		{ STRING_SWITCH, szUserSize,                 'f', 5 },
		{ STRING_SWITCH, szUserLabel,                'v', 14 },
      { STRING_SWITCH, szAutotest,                 'a', 9 },
		{ 0 }
	};


	//  Set a flag to indicate if the volume label was set
	//
	szUserLabel[0] = 0xff;

	//  Set a default option for linking SYS functions
	//
	fSwitches[CONFIRM_DELETE] = 1;  // for syscomn.c

	//  Preprocess the arguments.  This is needed because
	//  BORLAND breaks up the following command line into
	//  to seperate arguments and removes the quotes:
	//
	//  /v:"hello"
	//
	argc = ArgFix( argc, &argv[0] );

	//  Process all the boolean switches in the command line
	//
	argc = ParseCmdline(argc, &argv[0], FormatOpts);

	// if not at least the drive letter
	//
	if(argc != 1)
		Usage();

	//  The remaining option should be the drive letter
	//
	if( (argv[0][1] == ':') && (argv[0][2] == 0))
	{
		ucDriveLetter = toupper(argv[0][0]);
		if( (ucDriveLetter < 'A') || (ucDriveLetter > 'Z') )
		{
			Usage();
		}
		SetDosDriveNumber(ucDriveLetter - 'A');

		//  Determine if the drive is valid.  If the drive is not valid
		//  then display an error message for drive not valid and exit
		//
		if( !CheckIfValidDrive() )
		{
			DisplayErrorAndExit( ERR_BADDRIVE );
		}

		//  Determine if we can format this drive.  If we can not format
		//  this drive then display an error message unable to format
		//  and exit
		//
		if( !CanFormatDrive() )
		{
			DisplayErrorAndExit( ERR_NOT13 );
		}
	}
	else
	{
		Usage();
	}

	//
	//  Check if the user is specifying a size
	//

	if( fSwitches[FORCE_360K] )
	{
		fSwitches[USER_SIZE] = 1;
		SetUserType( 0 );
	}

	if( fSwitches[FORCE_720K] )
	{
		if(DoUserSize())
		{
			Usage();
		}
		fSwitches[USER_SIZE] = 1;
		SetUserType( 2 );
	}

	if( szUserSize[0] != 0 )
	{
		if(DoUserSize())
		{
			Usage();
		}
		fSwitches[USER_SIZE] = 1;

		for( uIndex = 0; uIndex < 5; uIndex++ )
		{
			if( 0 == stricmp(szUserSize, szUserSizeArray[uIndex]) )
			{
				break;
			}
		}

		if( uIndex == 5 )
		{
			Usage();
		}

		SetUserType( uIndex );

	}

	//  Determine if the user specified a size.
	//
	if(DoUserSize())
	{
		//  Determine if the drive is a hard drive
		//
		if(IsHardDrive())
		{
			//  Determine if the user specified a format size.  If so,
			//  you can not format a hard drive with a user specified
			//  size.  Display an error message and exit.
			//
			if(DoUserSize())
			{
				DisplayErrorAndExit( ERR_BADSIZE );
			}
		}

		//  Determine if we are quick formatting the drive
		//
		if(IsQuickFormat())
		{
			//  You can't specify a quick format size.
			//
			DisplayErrorAndExit( ERR_QUICKSIZE );
		}
	}


	//
	//  Check if the user is specifying a size
	//

	szVolume = GetVolumeLabelAddress();
	uIndex = 0;
	uCount = 0;
	if( szUserLabel[uIndex] != (char)0xff )
	{
      if( szUserLabel[0] == ' ' )
      {
         // Most appropriate would be to fail with "Invalid Volume ID"
         // but failing that, let's throw the
         //
         Usage();
      }
      else
      {
   		while((szUserLabel[uIndex] != 0) && (uCount < 12))
   		{
   			if( 0 != strchr(illegalVolChars, szUserLabel[uIndex]) )
   			{
   				Usage();
   			}

   			szVolume[uCount] = toupper(szUserLabel[uIndex]);
   			uCount++;
   			uIndex++;
   		}
   		szVolume[uCount] = 0;
   		fSwitches[HAVE_LABEL] = TRUE;  // boolean flag
      }
	}

   if( szAutotest[0] != 0 )
   {
      strupr( szAutotest );
      if( strcmp( szAutotest, "UTOTEST" ) )
      {
         // Invalid switch error - we must have the whole thing!
      	printf("\n");
      	printf(err_msg[ERR_INVALIDSWITCH],szAutotest);
         exit(0);
      }
      else
      {
         fSwitches[AUTOTEST] = TRUE;
      }
   }
}


//  Preprocess the arguments.  This is needed because
//  BORLAND breaks up the following command line into
//  to seperate arguments and removes the quotes:
//
//  /v:"hello"
//
int ArgFix( int argc, char *argv[] )
{
	unsigned uLength;
	unsigned uArgIndex = 0;
	unsigned uTemp;

	while( uArgIndex < argc )
	{
		uLength = strlen(argv[uArgIndex]);
		if((argv[uArgIndex][uLength-1] == ':') && (uLength == 3))
		{
			if( toupper(argv[uArgIndex][uLength-2]) == 'V' )
			{
				if((uArgIndex+1) < argc)
				{
					argv[uArgIndex][uLength] = '"';
					uTemp = uArgIndex+1;
					while( uTemp < argc )
					{
						argv[uTemp] = argv[uTemp+1];
						uTemp++;
					}
					argc--;
				}
			}
		}
		uArgIndex++;
	}

	return argc;
}


//  Return the BPB type for the size which the user
//  specified to format with.
//
unsigned GetUserType( void )
{
	unsigned * puUserType;

	puUserType = GetUserTypePointer();
	return *puUserType;
}


//  Save the specified user type (format BPB type)
//
void SetUserType( unsigned uUserType )
{
	unsigned * puUserType;

	puUserType = GetUserTypePointer();
	*puUserType = uUserType;
}


//  Return the address of the variable which contains the user type
//
unsigned * GetUserTypePointer( void )
{
	static unsigned uUserType;
	return &uUserType;
}


//  Return the address of the first BPB in the hard coded BPB array
//
struct BootSectorBPB16 * GetBPBTypeArray( void )
{
	static struct BootSectorBPB16 HardCodedBPB[] = {
		{ 512,2,1,2,112, 720,0xFD,2, 9,2,0 },		/* 360K  disk (FD)  0 */
		{ 512,1,1,2,224,2400,0xF9,7,15,2,0 },		/* 1.2M  disk (F9)  1 */
		{ 512,2,1,2,112,1440,0xF9,3, 9,2,0 },		/* 720K  disk (F9)  2 */
		{ 512,1,1,2,224,2880,0xF0,9,18,2,0 },		/* 1.44M disk (F0)  3 */
		{ 512,2,1,2,240,2880,0xF0,9, 9,2,0 },		/* Unknown floppy   4 */
	};

	return &HardCodedBPB[0];
}


//  Show the usage for a bad command line parameter
//
void OptError(char * szMsg)
{
	unsigned uIndex = 0;

	while( (szMsg[uIndex] != ' ') && (szMsg[uIndex] != 0) )
	{
		if( szMsg[uIndex] == '/' )
		{
			uIndex++;
			if( szMsg[uIndex] != 0 )
			{
				uIndex++;
			}
			break;
		}
		uIndex++;
	}
	szMsg[uIndex] = 0;

	printf("Invalid Switch - %s\n", szMsg );
//	Usage();
	exit(255);
}


//  Display the usage screen
//
void Usage( void )
{
	fSwitches[HIDE_SIGNON] = 0;
	DisplaySignonMessage();
	printf( szUsage );
	exit( MS_ERR_FATAL_ERROR );
}


/*
	$DOCAPI:SmartGetFreeSpace

	This is a 'smart' version of get free space. This function will check first
	to see if the disk is a FAT32 disk, and if so, will call the FAT32 get disk
	free space, otherwise this function will build the FAT12/16 free space info
	into the FAT32 structure so that only one structure is used to retrieve the
	free space for a drive. This function returns TRUE upon success and FALSE
	upon failure.

	iDrive = (0=default,1=A,2=B,...)!!!!

*/
//  Note:  Library free space won't work on windows because the DPB didn't
//         get updated
//
bool PASCAL SmartGetFreeSpace( int iDrive, PGDF32 pDiskSpace )
{
	bool		fResult;
	char * szDriveString = "A:\\";

	ASSERT(pDiskSpace);

	if(iDrive == 0)
	{
		dlGetCurrentDrive(&iDrive);
	}
	else
	{
		iDrive--;
	}

	// if we are in a DOS 7.1 enviornment then we can use the get
	// extended free space function
	//
	if( fIsFat32World )
	{
		// The FAT32 Get Free Space function requires a drive string
		// instead of a number, so we must first convert this drive number
		//
		*szDriveString = iDrive + 'A';

		// This member must specifically be initialized to ZERO always
		//
		pDiskSpace->uLevel = 0;

		// Get the drive space structure filled out
		//
		fResult = dlGetDriveFreeSpace32(szDriveString, pDiskSpace);
	}
	else
	{
	unsigned	uFreeClusters;			// Used to calculate FAT16 disk values
	unsigned	uTotalClusters;		// Used to calculate FAT16 disk values
	unsigned	uBytesPerSector;		// Used to calculate FAT16 disk values
	unsigned	uSectorsPerCluster;	// Used to calculate FAT16 disk values

		// Get some info on this drive so we can calculate
		fResult = __dlGetFreeSpaceOnDrive(iDrive+1, &uSectorsPerCluster, &uFreeClusters,
						&uBytesPerSector, &uTotalClusters);

		if(fResult)
		{
			// Initialize as many elements of this structure as possible
			pDiskSpace->uLevel                     = 0;
			pDiskSpace->uSize                      = sizeof(GDF32);
			pDiskSpace->ulSectorsPerCluster        = (ulong) uSectorsPerCluster;
			pDiskSpace->ulBytesPerSector           = (ulong) uBytesPerSector;

			pDiskSpace->ulAvailableClusters        = (ulong) uFreeClusters;
			pDiskSpace->ulAvailableAllocationUnits = (ulong) uFreeClusters;
			pDiskSpace->ulTotalClusters            = (ulong) uTotalClusters;
			pDiskSpace->ulTotalAllocationUnits     = (ulong) uTotalClusters;

			pDiskSpace->ulAvailablePhysicalSectors = (ulong) uFreeClusters *
					uSectorsPerCluster;
			pDiskSpace->ulTotalPhysicalSectors     = (ulong) uTotalClusters *
					uSectorsPerCluster;
		}
	}
	return fResult;
}


void ResetDisks( void )
{
	if( (fIsFat32World) && (IsHardDrive()) )
	{
		uchar ucDosDriveNumber;

		ucDosDriveNumber = GetDosDriveNumber();
		asm{
			mov ax, 710dh
			mov cx, 1h
			xor dx, dx
			mov dl, ucDosDriveNumber;
			int 21h
		};
	}
	else
	{
		//  Reset the drive (like MSDOS)
		//
		dlResetDisk();
	}
}


//  Compute free space for drive
//
void ComputeFreeSpace( void )
{
	PDPB pDPB;
	GDF32 DiskSpace;
	uchar ucDosDriveNumber;


	//  Get the dos drive number
	//
	ucDosDriveNumber = GetDosDriveNumber();

	//  Get the address of our local DPB
	//
	pDPB = GetDPBAddress();

	if( IsFat32Drive(pDPB) )
	{
		//  Display a message indicating that calculating free space
		//  could take a while
		//
		printf( szCalcFreeSpace );
	}

	//  Calculate free space
	//
	SmartGetFreeSpace( ucDosDriveNumber+1, &DiskSpace );

	if( IsFat32Drive(pDPB) )
	{
		//  Display a message indicating that we are done
		//
		printf( szComplete );
	}

}
