//
//   $Id: MCOMMIFC.C 1.4 2002/10/16 21:54:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*

	MCOMMIFC.C created 08/08/97  by Dennis Edwards

	Copyright (c) 1997-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	The interface translation module between the COMM.EXE terminal
	as well as the RSZ.EXE protocol driver program and the MCOMM
	Comm Lib use for ZMODEM file transfers.

	01-10-98 (DE)
		1. Added "file not found" and total file count to transfer stats
			for RSZ

*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <dos.h>

#if !COMM
#include "datalght.h"
#include "dlbiosio.h"
#include "rsz.h"
#endif

#include "comm.h"
#include "ansidrv.h"
#include "extra.h"
#include "colors.h"
#include "zmdos.h"

ASYNC comPort;
int handshaking = 0;
int doZRecover;
char parmString[12] = "";
long PrevSecs = -1L;        /* used by getSeconds to handle midnight */
long PrevHunds = -1L;       /* used by getHunds to handle midnight */

int comm_open(long baud, char *parms, int portNumber, int iIrqNumber)
{
	int stat;
	unsigned uPortAddress;

#if COMM
	void ___init_video(void);
	___init_video();
#endif

	// let MCOMM know where our bufffer is
	//
   comPort.RxSize   = RXD_BUFF_SIZE;
   comPort.TxSize   = TXD_BUFF_SIZE;
   comPort.RingSeg  = FP_SEG(rxdBuffer);
   comPort.RingOfst = FP_OFF(rxdBuffer);

	// hooks int 1C for time base
	//
	tickhookset(1);

	// open port, use existing parms if no baud rate specified
	//
	if(baud!=0)
	{
		sprintf(parmString,"%ld%s",(baud==115)?115200L:(long)baud,parms);
	}

	// Limit COM port number to {0,1,2,3}. *ASSUME* that COM1/COM3
	// and COM2/COM4 share IRQs 4 and 3, respectively get the
	// port's I/O base address from the BIOS data area and try to
	// open it. Error out if the port can't be opened.
	//
	if((portNumber > 3) || (portNumber < 0))
		portNumber = 0;

	uPortAddress = GetCommPortBaseAddress(portNumber);
	stat = async_open(&comPort,
							uPortAddress,
							IRQMASK(iIrqNumber),
							VCTR(iIrqNumber),
							parmString);
	if(stat != 0)
		return 3;

	// set up requested handshaking
	//
	async_xflow(&comPort,0);  					// don't handshake if not requested
	async_msrflow(&comPort,0);
	async_rts(&comPort,1);						// don't block invisible handshakers
	async_dtr(&comPort,1);
	if(handshaking == HARDWARE_HANDSHAKING)	// use CTS/RTS is user wants it
		async_msrflow(&comPort,(B_CTS | B_RTS));
	if(handshaking == SOFTWARE_HANDSHAKING)	// use XON/XOFF if user wants it
		async_xflow(&comPort, B_XUSE);

	return stat;
}

#if COMM
int fileOptions = 3;
int comm_setup(long baud, char *parms)
{
	sprintf(parmString,"%ld%s",(baud==115)?115200L:(long)baud,parms);
   return async_setbpds(&comPort, parmString);
}
void cl_twait(unsigned ticks)
{ long timeout;

	timeout = get_ticker() + ticks;
	while(get_ticker() < timeout)
		{;}
}
#else
int SetHandShakingType(int newMethod)
{ int oldMethod;

		oldMethod = handshaking;
		handshaking = newMethod;
		return oldMethod;
}
#pragma argsused;
int SetFileManagementType(int newMethod, bool crashRecoveryOption)
{ int oldMethod;

		oldMethod = fileOptions;
		fileOptions = newMethod;
		doZRecover = crashRecoveryOption;
		return oldMethod;
}
#endif

int comm_close(void)
{ int stat;

	// unhook int 1C
	//
	stat = async_close(&comPort);
	tickhookset(0);
	return stat;
}

void send_brk(void)
{
   async_sndbrk(&comPort, 1); 	// turn on break signal
	tdelay(1);					// wait a timer tick
	async_sndbrk(&comPort, 0);	// turn off break signal
}

int comm_poll(void)
{int c;

	c = async_rx(&comPort);
	if(c & B_RXEMPTY)
		return -1;
	return (c & 0xFF);
}

int comm_tgetc(int ticks)
{ int c;
  long timeout;

	c = comm_poll();
	if(c >= 0)
		return c;

	timeout = get_ticker() + ticks;

	while(get_ticker() < timeout)
	{
		c = comm_poll();
		if(c >= 0)
			break;
	}
	return c;
}

void comm_putc(char c)
{
	async_tx(&comPort,c);
}

/*/////////////////////////////////////////////////////////
//  Configuration structure.  Structure is used to allow //
//  the defaults to be changed modifying the EXE.        //
//  "TXZMCFG:" is the tag to search for to find the start//
//  of the structure in the EXE.                         //
//////////////////////////////////////////////////////// */
struct PROTCONFIG
{
    char    tag[8], DLPath[80], ULPath[80];
    long    LocBaud, MinFifoBaud;
    int     ComBase, IRQ, Vctr;
    int     h_VBufSize, h_BufSize, b_VBufSize, b_BufSize;
    int     TxWindow, ZExistOpts, XYExistOpts;
    int     FifoTxLvl, FifoRxLvl;
    char    IgnCarrier, MsrFlow, KeepTime, EscCtl, OvlyIO;
    char    Color[10], Mono[10];
} cfg =
{
    "TXZMCFG:", "", "",
    0L, 1L,
    0x3f8, IRQ4, VCTR4,
    2048, 0, 20508, 20480,
    0, 1, 2,
    8, 8,
    1, 0, 1, 0, 1,
    {  WHT|BLU_B,  H_GRN|BLU_B,  H_RED|BLU_B,    YLW|BLU_B,
       GRY|BLU_B,    YLW|BLU_B,  H_RED|BLU_B,  H_MAG|BLU_B,
       CYN,              CYN_B  },
    { WHT, WHT, RVRS, H_WHT, WHT, H_WHT, H_WHT, H_WHT, WHT, WHT_B }
};

/*--------------------------------------------+
|  zmodem message types (defined in ZMDOS.H)  |
+---------------------------------------------*/
#define M_RHDR      0       // received header
#define M_SHDR      1       // sent header
#define M_BLK0      2       // block 0 data processed (name, size, etc.)
#define M_CLASH     3       // file name clash occurred (use ExistOpts)
#define M_FILE      4       // start of transfer, FilePos = 1st position
#define M_EOF       5       // end of transfer (1 file)
#define M_DATA      6       // sent or received file data packet
#define M_FLOW      7       // change in XOFF or CTS flow status
#define M_IDLE      8       // waiting for character or for tx to empty
#define M_RESET     9       // reset to 'first file' condition

#if !COMM

#define M_DONE		  99		 // transfer completed, errors or no

// RSZ.C
//
void StatusMessagePrinter(char *msg);
char messageString[128];
extern bool outputIsFile;

#define IDLE_CALLS_BEFORE_UPDATE	15000

#pragma argsused
void ZMsg(int type,...)
{
	static bool amWaiting = FALSE;
	static unsigned fileCount = 0, wasWaitingCount = 0, idleCalls = 0;
	unsigned long fSize;
	int dotSize;

	if(quiet)
		return;

	if(amWaiting && (type != M_IDLE))
	{
		if(!outputIsFile)
			StatusMessagePrinter("\b \b");
	}
   switch (type)
   {

	case M_BLK0:       	// block 0 data processed (name, size, etc.)
		//
		// show file sequence number and path name
		//
		sprintf(messageString,"%2d %s ",++fileCount,PathName);
		StatusMessagePrinter(messageString);
		amWaiting = FALSE;
		break;

	case M_CLASH:       	// file name clash occurred (use ExistOpts)
      if(TFlag.F.ExistOpts == 0)
		{
			// file already exists, show we skipped it
			//
			amWaiting = FALSE;
			StatusMessagePrinter("skipped.\r\n");
			break;

		}
		else if(TFlag.F.ExistOpts == 2)
		{
			// file already exists, show we duped it under a new name
			//
			amWaiting = FALSE;
			StatusMessagePrinter("duped.\r\n");
			break;
		}
		//
		// fall through...
		//
	case M_FILE:       	// start of transfer, FilePos = 1st position
		if(FilePos > 0)
		{
			// Show restart file position for crash recoveries
			//
			sprintf(messageString," \(%,ld\) ",FilePos);
			StatusMessagePrinter(messageString);
		}
		amWaiting = FALSE;
		break;

	case M_EOF:       	// end of transfer (1 file)
		//
		// show file size in Kbytes
		//
      fSize = (TFlag.F.Receiving) ? RxdFileSize : FileSize;
		dotSize  = ( int )( fSize % 1024L );
		dotSize /= 102;
		if(fSize/1024+dotSize==0)
			sprintf(messageString,"\r\n\t%d bytes\r\n",fSize);
		else
		{
			fSize   /= 1024;
			sprintf(messageString,"\r\n\t%,ld.%d Kbytes\r\n",fSize,dotSize);
		}
		StatusMessagePrinter(messageString);
		amWaiting = FALSE;
		break;

	case M_DATA:       	// sent or received file data packet
		//
		// print a dot for every packet
		//
		amWaiting = FALSE;
		if(!outputIsFile)
			StatusMessagePrinter(".");
		break;

	case M_IDLE:       	// waiting for character or for tx to empty
		//
		// show the spinner thingy while idle
		//
		if(!amWaiting)
		{
			amWaiting = TRUE;
			wasWaitingCount = 1;
			if(!outputIsFile)
				StatusMessagePrinter("-");
			else
				break;
		}
		if(++idleCalls >= IDLE_CALLS_BEFORE_UPDATE)
			idleCalls = 0;
		else
			break;

		if(outputIsFile)
			break;

		switch((wasWaitingCount++ & 3))
		{
		case 0:
			StatusMessagePrinter("\b-");
			break;
		case 1:
			StatusMessagePrinter("\b\\");
			break;
		case 2:
			StatusMessagePrinter("\b|");
			break;
		case 3:
			StatusMessagePrinter("\b/");
			break;
		default:
			break;
		}
		break;

	case M_RESET:       	// reset to 'first file' condition
		fileCount  = 0;
		//
		// fall through...
		//
   case M_RHDR:       	// received header
   case M_SHDR:       	// sent header
	case M_FLOW:       	// change in XOFF or CTS flow status
		amWaiting = FALSE;
		break;
	case M_DONE:			// done
		sprintf(messageString,"\r\n%2d files total\n\r",fileCount);
		StatusMessagePrinter(messageString);
	default:
		amWaiting = FALSE;
		break;
	}
}

#else

long tfBytes, tCPS;         /* total bytes transferred, average CPS */
int tFiles;                 /* number of files transferred */
char buf[256];              /* general purpose buffer */
char *OverrideName = NULL;  /* cmdline rx filename used if not NULL */
char OverrideFlag = '\0';   /* if '/', above specifies complete pathname */
char *color;                /* pointer to list of colors */
int txtcolor;               /* color of most screen message output */
char minsecs[10];           /* seconds to min:secs buffer */
/* variables & dummy function to disable display of file transfer screen */
int qf = 0;
#pragma argsused;
void DoNothing(int r, int c, int a, char *s) {}
void (*DspMsgAt[2])(int, int, int, char *) = { d_msgat, DoNothing };
long getHunds(void);
long getSeconds(void);
long LocBaud = 0L;          /* CPU to modem/device baud rate */
int combase, irq, vctr;     /* port address, IRQ number, & vector */
char OvlyIO;                /* overlay disk and serial I/O flag */
long efficiency;
/*/////////////////////////////////////////////////////////
//  Screen data structure                                //
//////////////////////////////////////////////////////// */
typedef struct
{
    int row, col, color, count;
    char *text;
} SCREENDATA;


/*/////////////////////////////////////////////////////////
//  DrawBox                                              //
//:dra////////////////////////////////////////////////// */
void DrawBox(int row, int col, int nrows, int ncols, int color, int style)
{
    static char boxchars[][6] =
    {   {'', '', '', '', '', ''},
        {'', '', '', '', '', ''},
        {'', '', '', '', '', ''},
        {'', '', '', '', '', ''}
    };
    char *box = boxchars[style];
    int temp = v_color;

    v_color = color;
    --ncols;
    scrlupat(row, col, row + nrows - 1, col + ncols, nrows);
    --nrows;
    ++col, --ncols, --ncols;
    d_nchat(row, col, box[4], color, ncols, 1);
    d_nchat(row + nrows, col, box[4], color, ncols, 1);
    d_nchat(row, col, box[5], color, nrows, 0);
    d_nchat(row, col + ncols, box[5], color, nrows, 0);
    d_chat(row, col, box[0]);
    d_chat(row, col + ncols, box[1]);
    d_chat(row + nrows, col, box[2]);
    d_chat(row + nrows, col + ncols, box[3]);
    v_color = temp;
}

/*/////////////////////////////////////////////////////////
//  DspZmodemScrn                                        //
//:dsp////////////////////////////////////////////////// */
void DspZmodemScrn(void)
{
    static SCREENDATA zscrn[] =
    {
        {  1,  6, 2,   0, " Zmodem File " },
        {  2, 71, 0, -20, "" },
        { 15,  5, 0,  64, "" },
        {  2, 73, 0,   0, "F T" },
        {  3, 73, 4, -19, "" },
        {  3, 75, 4, -19, "" },
        {  3,  5, 0,   0, "File name :" },
        {  5,  5, 0,   0, "Estimated time :" },
        {  6,  5, 0,   0, "Elapsed time   :" },
        {  7,  5, 0,   0, "File CPS rate  :" },
        {  5, 37, 0,   0, "File position      :" },
        {  6, 37, 0,   0, "Expected file size :" },
        {  7, 37, 0,   0, "Beginning offset   :" },
        {  9, 51, 0,   0, " Hdr Data " },
        { 10, 23, 0,   0, "Header Name" },
        { 10, 40, 0,   0, "Type" },
        { 10, 52, 0,   0, "Hex" },
        { 10, 61, 0,   0, "Decimal" },
        { 11,  5, 0,   0, "Last hdr recvd :" },
        { 12,  5, 0,   0, "Last hdr sent  :" },
        { 14,  5, 0,   0, "Crc :" },
        { 14, 37, 0,   0, "Flow :" },
        { 17,  5, 0,   0, "Total files queued :" },
        { 18,  5, 0,   0, "Total bytes queued :" },
        { 17, 42, 0,   0, "Estimated time   :" },
        { 18, 42, 0,   0, "Accumulated time :" },
        { 20,  5, 0,   0, "Files transferred  :" },
        { 21,  5, 0,   0, "Bytes transferred  :" },
        { 21, 42, 0,   0, "Average CPS rate :" },
        { -1, -1, 0,   0, "" }
    };
    register SCREENDATA *sd;

    if (qf != 0)
        return;
    v_color = WHT;
    cls();
    v_color = txtcolor;
    DrawBox(1, 1, 22, 78, color[0], 0);
    for (sd = zscrn; sd->row >= 0; ++sd)
    {
        if (sd->count == 0)
            d_msgat(sd->row, sd->col, color[sd->color], sd->text);
        else if (sd->count > 0)
        {
            d_nchat(sd->row, sd->col, *sd->text, color[sd->color],
             sd->count, 1);
        }
        else
        {
            d_nchat(sd->row, sd->col, *sd->text, color[sd->color],
             -sd->count, 0);
        }
    }
    d_msgat(1, 19, color[2], (TFlag.F.Receiving) ? "Receive " : "Send ");
    loc(23, 0);
}


/*/////////////////////////////////////////////////////////
//  InitDefaults - init protocol global variables        //
//:ini////////////////////////////////////////////////// */
void InitDefaults(void)
{
    combase = cfg.ComBase, irq = cfg.IRQ, vctr = cfg.Vctr;
    LocBaud = cfg.LocBaud;
    OvlyIO = cfg.OvlyIO;
    TFlag.F.ExistOpts = cfg.ZExistOpts;
    TFlag.F.IgnCarrier = cfg.IgnCarrier;
    TFlag.F.KeepTime = cfg.KeepTime;
    TFlag.F.EscCtl = cfg.EscCtl;
    TxWindow = cfg.TxWindow;
}

void ___init_video(void)
{
    color = ((initvid() & 0xff) == CO80) ? cfg.Color : cfg.Mono;
    txtcolor = (int)color[1] & 0xff;
    InitDefaults();
}


/*///////////////////////////////////////////////
//  ConvertSecs                                //
//:con//////////////////////////////////////// */
char *ConvertSecs(long secs)
{
    sprintf(minsecs, "%01ld:%02ld   " , uldiv(secs, 60L), ulmod(secs, 60L));
    return(minsecs);
}

/*///////////////////////////////////////////////
//  vDisplay                                   //
//:vdi//////////////////////////////////////// */
void vDisplay(int row, int col, char *format, ...)
{
    va_list arg_ptr;
    char lbuf[80];

    va_start(arg_ptr, format);
    vsprintf(lbuf, format, arg_ptr);
    DspMsgAt[qf](row, col, txtcolor, lbuf);
}
/*/////////////////////////////////////////////////////////
//  ZMsg -- zmodem message handler                       //
//:zms///////////////////////////////////////////////:z: */
void ZMsg(int type, ...)
{
    static char *HdrType[] =
    {
        "Garbage Count", "Long Packet",   "Garbled Packet",   "Bad Crc",
        "Timed Out",     "Unknown Hdr",   "Sync Error",       "Memory Error",
        "File Error",    "Lost Carrier",  "Remote Abort",     "Local Abort",
        "ZRQINIT",       "ZRINIT",        "ZSINIT",           "ZACK",
        "ZFILE",         "ZSKIP",         "ZNAK",             "ZABORT",
        "ZFIN",          "ZRPOS",         "ZDATA",            "ZEOF",
        "ZFERR",         "ZCRC",          "ZCHALLENGE",       "ZCOMPL",
        "ZCAN",          "ZFREECNT",      "ZCOMMAND",         "ZSTDERR"
    };
    static char *HdrStyle[] = { "ZBIN", "ZHEX", "ZBIN32", "" };
    static char *CrcStyle[] = { "ZCRCE", "ZCRCG", "ZCRCQ", "ZCRCW" };
    static char MsgMask[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 };

    static long fStartSec, tStartSec = -1L, fStartHunds, tStartHunds = -1L;
    static long fCPS, fBytes, tBytes;
    static long fBarInc, tBarInc, fBarBytes, tBarBytes;
    static long fSize, fStartPos, updBytes, lastPos, orgTBytes;
    static long lastSec, temp, flowhalts;
    static int tBarRow, fBarRow;
    static char gotBlock, crcType = ZCRCG, firstfile = 1;

    FILE *flog;
    va_list argptr;
    int i, j, row, hdr;
    long secs;
    char *p1;

    if (MsgMask[type] == '\0')
        return;
    switch (type)
    {
      case M_RHDR:              /* received header */
      case M_SHDR:              /* sent header */
        row = (type == M_RHDR) ? 11 : 12;
        va_start(argptr, type);
        if ((hdr = va_arg(argptr, int)) < 0)
            ZHdr.Data = 0L;
        i = hdr + 12;
        j = va_arg(argptr, int) - 'A';
        if (j < 0 || j > 3)
            j = 3;
        vDisplay(row, 22, "%-17s%-9s%08lX%11ld", HdrType[i], HdrStyle[j],
         ZHdr.Data, ZHdr.Data);
        break;
      case M_CLASH:             /* file name clash occurred */
        if (TFlag.F.ExistOpts != 2)
            break;  /* not new name */
      case M_BLK0:              /* sent or received block 0 */
        if (OverrideName != NULL && TFlag.F.Receiving && type != M_CLASH)
        {
            if (OverrideFlag == '/')
                strcpy(PathName, OverrideName);
            else
                strcpy(NameExt, GetNameExt(OverrideName));
        }
        i = sprintf(buf, "%-52s", PathName) - 52;
        DspMsgAt[qf](3, 17, color[3], &(strupr(buf))[i]);
        if (type == M_CLASH)
            break;
        v_color = txtcolor;
        flowhalts = 0L;
        if (qf == 0)
        {
            scrlupat(5, 22, 6, 32, 0);
            scrlupat(5, 58, 7, 68, 0);
        }
        fSize = (TFlag.F.Receiving) ? RxdFileSize : FileSize;
        vDisplay(6, 58, "%-8ld", fSize);
        fStartHunds = getHunds();
        fStartSec = uldiv(fStartHunds, 100L);
        if (firstfile)
        {
            firstfile = tFiles = 0;
            tBytes = tBarBytes = 0L;
            lastSec = tStartSec = fStartSec;
            tStartHunds = fStartHunds;
            tBarRow = 21;
            if ((tBarInc = uldiv(TotalBytes, 19L)) == 0)
                ++tBarInc;
            orgTBytes = TotalBytes;
            if ((tCPS = uldiv(ulmul(ConnectBaud, 955L), 10000L)) == 0L)
                ++tCPS;
            vDisplay(17, 61, "%-8s",
             ConvertSecs(uldiv(TotalBytes, tCPS) + (TotalFiles >> 2)));
        }
        if ((fBarInc = uldiv(fSize, 19L)) == 0)
            ++fBarInc;
        fBarBytes = 0L, fBarRow = 21;
        if (qf == 0)
            d_nchat(3, 73, '', color[4], 19, 0);
        if (orgTBytes != 0L)
        {
            j = 3 + (int)((orgTBytes & 0xffff0000L) ?
             uldiv(TotalBytes, uldiv(orgTBytes, 19L)) :
             uldiv(ulmul(TotalBytes, 19L), orgTBytes));
            while (j < tBarRow && tBarRow > 2)
                DspMsgAt[qf](tBarRow--, 75, color[6], "");
            vDisplay(17, 26, "%-4d", TotalFiles - 1);
            vDisplay(18, 26, "%-8ld", TotalBytes - fSize);
        }
        else
        {
            tBarInc = fBarInc, tBarBytes = 0L, tBarRow = 21;
            if (qf == 0)
                d_nchat(3, 75, '', color[4], 19, 0);
        }
        break;
      case M_EOF:               /* end of file transfer */
        va_start(argptr, type);
        i = va_arg(argptr, int);
        if (i == ZEOF || i == ZRINIT)
        {
            vDisplay(20, 26, "%d", ++tFiles);
            tfBytes = tBytes;
            if (((p1 = getenv("TXZMLOG")) != NULL)
             && ((flog = fopen(p1, "at")) != NULL))
            {
                sprintf(buf,
                 "%c %6ld %5ld bps %4ld cps %3u errors %5ld %4d %s %d\n",
                 TFlag.F.Receiving ? 'Z' : 'z', FilePos, ConnectBaud, fCPS,
                 ErrCnt, flowhalts, BlkLen, NameExt, (SerNbr) ? SerNbr : -1);
                fputs(buf, flog);
                fclose(flog);
            }
        }
        if (orgTBytes != 0L)
        {
            temp = TotalBytes - fSize;
            j = 3 + (int)((orgTBytes & 0xffff0000L) ?
             uldiv(temp, uldiv(orgTBytes, 19L)) :
             uldiv(ulmul(temp, 19L), orgTBytes));
            while (j < tBarRow && tBarRow > 2)
                DspMsgAt[qf](tBarRow--, 75, color[6], "");
        }
        crcType = ZCRCG;
        break;
      case M_FILE:              /* start of transfer, 1st FilePos set */
        lastPos = fStartPos = FilePos;
        fBytes = 0L;
        vDisplay(5, 22, "%-8s", ConvertSecs(uldiv(fSize - fStartPos, tCPS)));
        vDisplay(7, 58, "%-8ld", fStartPos);
        if (fStartPos == 0L)
            break;
        updBytes = fStartPos;
      case M_DATA:              /* sent or received file data packet */
        if (type != M_FILE)
        {
            gotBlock = 1;
            updBytes = FilePos - lastPos;
            lastPos = FilePos;
            fBytes += updBytes, tBytes += updBytes;
            vDisplay(5, 58, "%-8ld", FilePos);
            vDisplay(21, 26, "%-8ld", tBytes);
            va_start(argptr, type);
            vDisplay(14, 11, "%s-%s", (BinHdr == ZBIN32) ? "32" : "16",
             CrcStyle[(crcType = (char)va_arg(argptr, int)) - ZCRCE]);
        }
        fBarBytes += updBytes;
        tBarBytes += updBytes;
        while (fBarBytes >= fBarInc && fBarRow > 2)
            fBarBytes -= fBarInc, DspMsgAt[qf](fBarRow--, 73, color[5], "");
        while (tBarBytes >= tBarInc && tBarRow > 2)
            tBarBytes -= tBarInc, DspMsgAt[qf](tBarRow--, 75, color[6], "");
        break;
      case M_FLOW:
        va_start(argptr, type);
        switch (i = va_arg(argptr, int))
        {
          case 0:           /* XOFF cleared (XON received) */
          case 1:           /* XOFF received */
            DspMsgAt[qf](14, 44, color[3], (i) ? "XOFF" : "    ");
            break;
          case 2:           /* CTS signal raised */
          case 3:           /* CTS signal lowered */
            DspMsgAt[qf](14, 49, color[3], (i == 3) ? "CTS" : "   ");
            break;
          case 4:           /* port error */
            DspMsgAt[qf](14, 55, color[3]|BLNK, "PORT RESET");
            tdelay(18);
            DspMsgAt[qf](14, 55, color[3], "          ");
        }
        flowhalts += (i & 1);
        break;
      case M_IDLE:
        if (firstfile || ((secs = getSeconds()) == fStartSec))
            break;
        if (gotBlock || crcType == ZCRCE)
        {
            gotBlock = 0;
            if (fBytes & 0xffff0000L)
                fCPS = uldiv(fBytes, secs - fStartSec);
            else
                fCPS = uldiv(ulmul(fBytes, 100L), getHunds() - fStartHunds);
            if (tBytes & 0xffff0000L)
                tCPS = uldiv(tBytes, secs - tStartSec);
            else
                tCPS = uldiv(ulmul(tBytes, 100L), getHunds() - tStartHunds);
            if (tCPS == 0)
                ++tCPS;
            vDisplay(7, 22, "%-6ld", fCPS);
            vDisplay(21, 61, "%-6ld", tCPS);
        }
        if (secs != lastSec)
        {
            lastSec = secs;
            DspMsgAt[qf](6, 22, txtcolor, ConvertSecs(secs - fStartSec));
            DspMsgAt[qf](18, 61, txtcolor, ConvertSecs(secs - tStartSec));
        }
        break;
      case M_RESET:
        if (cfg.MsrFlow && !async_cts(&comPort))
            DspMsgAt[qf](14, 49, color[3], "CTS");
        crcType = ZCRCG, firstfile = 1;
        tStartSec = PrevSecs = PrevHunds = -1L;
        break;
    }
}

void pop_up_stats(void)
{
	efficiency=0L;
   pushscrn(0, 0, 24, 80);
   DspZmodemScrn();
   ZMsg(M_RESET);
}

void pop_down_stats(void)
{
   if (ConnectBaud)
      efficiency = uldiv(ulmul(tCPS, 1000L), ConnectBaud);
   sprintf(buf, "\r\nCPS: %ld (%d files, %ld bytes)  Efficiency %ld%%",
   tCPS, tFiles, tfBytes, efficiency);
   if (qf == 0)
   {
      loc(23,1), v_color = WHT;
      d_str(&buf[2]);
   }
   d_str(", press ENTER...");
	b_getch();
   popscrn();
}

#endif // !COMM

/*/////////////////////////////////////////////////////////
//  SetFIFOLevels - set FIFO levels                      //
//:set////////////////////////////////////////////////// */
void SetFIFOLevels(int rxFIFOlevel, int txFIFOlevel)
{
    if (comPort.Stat3 & B_FIFO)
    {
        async_FIFOrxlvl(&comPort, rxFIFOlevel);
        async_FIFOtxlvl(&comPort, txFIFOlevel);
    }
}

int SendZMODEM(char *pzFiles)
{  int status;
#if COMM
	int temp;
#endif

	TFlag.F.Receiving = 0;

#if COMM
	temp = v_color;
	pop_up_stats();
#endif

	comm_putc('r');
	comm_putc('z');
	comm_putc('\r');

   SetFIFOLevels(cfg.FifoRxLvl, cfg.FifoTxLvl);
	TFlag.F.ExistOpts = fileOptions;
	if(doZRecover)
		TFlag.F.ExistOpts = 1;
	async_xflow(&comPort, B_XUSE);
   status = ZmodemSend(&comPort, pzFiles);
#if !COMM
	if(status == ERR_NO_FILES)
		StatusMessagePrinter("File not found.\r\n");
	ZMsg(M_DONE);
#endif
	async_xflow(&comPort, 0);
   SetFIFOLevels(1, 1);

#if COMM
	pop_down_stats();
  	v_color = temp;
#endif

	return status;
}

int ReceiveZMODEM(void)
{	int status;
#if COMM
	int temp;
#endif

	TFlag.F.Receiving = 1;

#if COMM
	temp = v_color;
	pop_up_stats();
#endif

   SetFIFOLevels(cfg.FifoRxLvl, cfg.FifoTxLvl);
	TFlag.F.ExistOpts = fileOptions;
	if(doZRecover)
		TFlag.F.ExistOpts = 1;
	async_xflow(&comPort, B_XUSE);
   status = ZmodemRecv(&comPort);
#if !COMM
	ZMsg(M_DONE);
#endif
	async_xflow(&comPort, 0);
   SetFIFOLevels(1, 1);

#if COMM
	pop_down_stats();
  	v_color = temp;
#endif

	return status;
};

// end of mcommifc.c
