//
//   $Id: MODE.C 1.9 2003/06/10 17:40:56Z thomd Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	MODE.C	created 08/14/89	by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary


	USAGE:

	MODE COMn: baud [,parity [,databits [,stop bits [,p]]]]
	MODE LPTn: = COMn:

	baud    = 300,1200,2400,4800,9600
	parity  = N, E, O
	databits= 7, 8
	stopbits= 1, 2


	NOTES:

	We do not support setting the typmatic rate.


	HISTORY:

   01/25/2002 jmb added support for printer redirection to COM ports.

	10-22-1998 BT  Made the screen output the same as Microsofts.

	10-20-1998 BT  Added the ability to change the resident portions
						variables.

	10-20-1998 BT  Added the ability to detect if mode is already
						resident.

	10-20-1998 BT  Made it so it returned an error code of zero when
						program ran successfully and the resident portion
						was already installed.

	10-20-1998 BT  Commented the code in TRAP14.ASM and added
						an interrupt 2fh filter.

	10-20-1998 BT  Removed some functions that were not used.

	02-06-1997 DE	Added "43" and "50" line "VideoMode"s

	08-02-1994 RLS	Added support for Video Modes.  This enables MODE to
			   		change the current Video mode.  Example: MODE VideoMode
				   	where VideoMode is "40", "80", "BW80", "BW40", "CO80",
					   "CO40", or "MONO"!

**********************************************************/

//  Borland libraries
//
#include <ctype.h>
#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <bios.h>

//  Datalight libraries
//
#include "dosver.h"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "fatworld.h"
#include "mode.h"

//  The maximum command line length plus the null terminator
//
#define MAX_COMMAND_LINE	128 + 1


#define MODE 1
#include "dosmsg.c"

char signon[] = "\nMODE %s %s\n%s\n\n";

char cmdline[128];


unsigned char ucInt14Retries;
unsigned uResidentCodeSegment;
unsigned gRedirectedLpt;
unsigned gRedirectedCom;


int port     = 0;
int baud     = 9600;
int parity   = 'e';
int databits = 7;
int stopbits = 1;

int baud_array[8] = { 110, 150, 300, 600, 1200, 2400, 4800, 9600 };
int par_array [4] = { 'n', 'o', 'n', 'e' };
int data_array[4] = { 5,6,7,8 };
int stop_array[2] = { 1,2 };

char szCommandLine[MAX_COMMAND_LINE]="";

bool fIsFat32World;

//  Function prototypes
//
#include "mode.pt"
void InstallIntHandlers( void );
void SetRetries( void );
void main(int argc,char *argv[]);
void do_com(char *p, int maxport);
void do_help(void);
void show_device(void);
void _show_device(unsigned char far *fp);
void terminate_and_stay_res(void);
void fdump(char far *fp);
void install_mode(void);
void errout(int err,char *p);
void do_lpt(char *p, int maxport);
void lpt_redirect(int lpt, int com);
int VerifyWithMCBChain( unsigned uCodeSegment );
int InstalledCheck( void );
int numports(void);
int alldigits(char *s);
int inarray(int *array,int val,int num);
int name_cmp(char *name, char far *dev_name);
int installed(void);
char * getfield(char *dst,char *p);
char * skipwhite(char *p);
char * skipequals(char *p);
struct device far *find_device(char *name);
void SetRedirect(int lpt, int com );


void DisplaySignOnMessage( void )
{
	if( fIsFat32World )
	{
		printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

void main(int argc,char *argv[])
{
	char *p;
//	char s[20];
	int  i, maxport;
	int iParam;
   bool fAddColon = FALSE;
//	unsigned parms;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	//  Save command line now in case it is displayed before the
	//  program goes resident.
	//
	for( iParam = 1; iParam <= argc; iParam++ )
	{
		strcat( szCommandLine, argv[iParam] );
		if( iParam == 1 )
		{
      	if(strnicmp(szCommandLine,"COM",3) == 0)
         {
	         if (!strchr(szCommandLine, ':'))
	         {
	            fAddColon = TRUE;            // used later on
	   			strcat( szCommandLine, ":" );
	         }
         }
			strcat( szCommandLine, " " );
		}
	}
	strupr( szCommandLine );

	/* if no arguments */
	if(argc<2)
	{
		DisplaySignOnMessage();
		do_help();
	}

	for(i=1; i<argc; ++i)
	{
		if(argv[i][0] == '/')
		{
			switch(toupper(argv[i][1]))
			{
				case 'H':
				case '?':
					DisplaySignOnMessage();
					do_help();
					exit(1);
				case 'S':
					show_device();
					exit(1);
				default:
					DisplaySignOnMessage();
					printf(bad_opt, argv[i]);
					do_help();
					exit(1);
			}
		}
	}

	/* get the whole command line */
	*cmdline = '\0';
	for(i=1; i<argc; ++i)
   {
		strcat(cmdline,argv[i]);
      if ((i==1)&&(fAddColon))
      {
         strcat(cmdline,":");
      }
   }
	strupr(cmdline);

	if(SetVideoMode(cmdline))		/* If it's a Video Modfication */
		exit(0);			/* Then were all done! */

	/* check equipment for # of COM ports */
	maxport = numports();
	if(!maxport)
		errout(6,"");

	/* parse the command line */
	p = skipwhite(cmdline);
	if(strncmp(p,"COM",3) == 0)
   {
		do_com(p, maxport);
   }
	else if(strncmp(p, "LPT", 3) == 0)
   {
		do_lpt(p, maxport);

   	//  output to screen the command line
   	//
   	printf("\n%s\n",szCommandLine);
   }
	else
   {
		errout(0, p);
   }

	exit(0);
}


/*
	MODE "COM1:9600,n,8,1,p"
*/
void do_com(char *p, int maxport)
{
	char s[20];
//	int  i;
	int IsAlreadyResident;
	unsigned parms;
	unsigned char retries = 0;

	getfield(s,p+3);
	if(*s < '1' || *s > maxport+'0')
		errout(1,s);
	port = *s - '1';

	p = skipwhite(p+4);
   if(*p == ':')
   {
      p++;
   }

	/* get the baud rate -- there must be a baud rate */
	p = getfield(s,p);
	baud = atoi(s);
	baud = inarray(baud_array,baud,8);
	if(baud<0 || !alldigits(s))
		errout(2,s);

	/* get the parity (optional) */
	if(*p != ',')
	{
		p = getfield(s,p);
		if(*s)
			parity = *s;
	}
	parity = inarray(par_array,parity,4);
	if(parity < 0)
		errout(3,s);

	/* get the databits (optional) */
	if(*p != ',')
	{
		p = getfield(s,p);
		if(*s)
			databits = *s-'0';
	}
	databits = inarray(data_array,databits,4);
	if(databits < 0)
		errout(4,s);

	/* get the stopbits (optional) */
	if(*p != ',')
	{
		p = getfield(s,p);
		if(*s)
			stopbits = *s-'0';
	}
	stopbits = inarray(stop_array,stopbits,4);
	if(stopbits < 0)
		errout(5,s);

	/* get printer usage flag (optional) */
	if(*p != ',')
	{
		getfield(s,p);
		if(toupper(*s) == 'P')
			retries = 100;
	}

	parms = (baud << 5) + (parity << 3) + (stopbits <<2) + databits;
	bioscom(0,parms,port);

	/* set retry flag for printer option */
	if(retries)
	{
		IsAlreadyResident = InstalledCheck();
		if( IsAlreadyResident )
		{
			ucInt14Retries = retries;
			SetRetries();
		}
		else
		{
			ucInt14Retries = retries;
			printf(res_installed);
			printf("\n%s\n",szCommandLine);
			InstallIntHandlers();
		}
	}

   printf( "\nCOM%d: %d,%c,%d,%d", port+1, baud_array[baud],
            par_array[parity], data_array[databits], stop_array[stopbits] );
   if( retries )
   {
      printf( ",P\n" );
   }
   else
   {
      printf( ",-\n" );
   }
}


//  This function checks to see if the resident part of
//  mode is already installed.  If it is the function
//  saves the segment of the resident part and returns
//  a boolean true.  Otherwise the function returns a
//  boolean false.
//
int InstalledCheck( void )
{
	int iReturnValue = 1;
	int iSearchReturnValue = 1;

	asm{
		push	bx
		mov	ax, 4900h
		int	2fh
		mov	cx, bx
		pop	bx
		cmp	ah, 0ffh
		jne	NotInstalled
		mov	uResidentCodeSegment, cx
		jmp	Done
	}

NotInstalled:
	asm{
		mov	ax, 0
		mov	iReturnValue, ax
	};

Done:

	if( iReturnValue )
	{
		//  Double check that we are resident by searching
		//  the mcb chain for mode
		//
		iSearchReturnValue = VerifyWithMCBChain( uResidentCodeSegment );
		if( !iSearchReturnValue )
		{
			iReturnValue = 0;
		}
	}

	return iReturnValue;
}


//  This function checks the mcb of the code segment it was passed
//  to see if it is actually mode.  This function returns a boolean
//  true if it is modes mcb and a boolean false otherwise.
//
int VerifyWithMCBChain( unsigned uCodeSegment )
{
	int iReturnValue = 0;
	int iCmpResult;
	char szModeString[] = "MODE";
	struct MCB far * fpMCB;


	fpMCB = MK_FP( (uCodeSegment - 0x11), 0 );
	iCmpResult = _fstrcmp( fpMCB->name, szModeString );
	if( iCmpResult == 0 )
	{
		iReturnValue = 1;
	}

	return iReturnValue;
}


int numports(void)
{
	int ports;

	geninterrupt(0x11);
	ports = _AX;
	ports = ((ports >> 9) & 0x7);

	return ports;
}

int alldigits(char *s)
{
	int i=0;
	while(*s && *s>='0' && *s<='9')
		++s, ++i;
	if(*s)
		return 0;
	return i;
}

#define alnum(c) ((c>='a' && c<='z') || (c>='A' && c<='Z') || (c>='0' && c<='9'))
char * getfield(char *dst,char *p)
{
	p = skipwhite(p);
	while(alnum(*p))
		*dst++ = *p++;
	*dst = '\0';
	if(*p == ',')
		++p;
	return skipwhite(p);
}

int inarray(int *array,int val,int num)
{
	int i;

	for(i=0; i<num; ++i)
		if(val == array[i])
			return i;
	return -1;
}

void errout(int err,char *p)
{
	printf(err_invalid,error_msg[err],p);
	exit(err+1);
}

char *skipwhite(char *p)
{
	while(DL_IS_SPACE(*p))
		++p;
	return p;
}

/*
	Handle Mode commands that start with "LPT".
	There are two types of LPT commands redirection.
	LPT1:=COM2
*/
void do_lpt(char *p, int maxport)
{
	int lpt, com;
	char *skipequals();

	lpt = p[3] - '1';
	if(lpt < 0 || lpt >= 3)
	{
		printf(bad_printer, p);
		exit(1);
	}

	p += 4;
	if(*p == ':')
		++p;

	p = skipequals(p);

	if(strncmp(p, "COM", 3) == 0)
	{
		com = p[3] - '1';
		if(com < 0 || com >= maxport)
		{
			printf(bad_port, com+1);
			exit(1);
		}
		lpt_redirect(lpt, com);
	}
	else
		printf(bad_redir,p);
}

char * skipequals(char *p)
{
	while(*p && (*p==' ' || *p=='\t' || *p=='='))
		++p;
	return p;
}

void lpt_redirect(int lpt, int com)
{
	char *l = "LPT#    ", *c = "COM#    ";
	struct device far *lpt_dev, far *com_dev;
	int IsAlreadyResident;

	l[3] = '1' + lpt;
	c[3] = '1' + com;
	lpt_dev = find_device(l);
	com_dev = find_device(c);
	if(lpt_dev == NULL || com_dev== NULL)
	{
		printf(bad_dev);
		exit(1);
	}
//	lpt_dev->strategy_offset = com_dev->strategy_offset;
//	lpt_dev->interrupt_offset = com_dev->interrupt_offset;

	IsAlreadyResident = InstalledCheck();
	if( IsAlreadyResident )
	{
      gRedirectedLpt = lpt;
      gRedirectedCom = com;
		SetRedirect(lpt, com);
	}
	else
	{
      gRedirectedLpt = lpt;
      gRedirectedCom = com;
		printf(res_installed);
		printf("\n%s\n",szCommandLine);
		InstallIntHandlers();
	}

	printf(reroute_str, lpt+1, com+1);

}

struct device far *find_device(char *name)
{
	struct dos_vars far *p, far *get_dos_vars();
	struct device far *dev;
//	char far *fp;
//	int i;
	p = get_dos_vars();

	dev = (struct device far *) &p->devices;
	do
	{
		if(name_cmp(name, dev->name))
			return dev;			/* Found it */
		dev = dev->next;
	} while(FP_OFF(dev) != 0xFFFF);
	return NULL;
}

int name_cmp(char *name, char far *dev_name)
{
	int i;
	for(i=0; i<8; ++i)
	{
		if(name[i] != dev_name[i])
			return FALSE;
	}
	return TRUE;
}


void do_help(void)
{
	printf(USAGE);
	exit(1);
}

void show_device(void)
{
	struct dos_vars far *p, far *get_dos_vars();
	struct device far *dev;
//	char far *fp;
//	int i;
	p = get_dos_vars();

	dev = (struct device far *) &p->devices;
	do
	{
		_show_device((unsigned char far *)dev);
		dev = dev->next;
	} while(FP_OFF(dev) != 0xFFFF);
}

void _show_device(unsigned char far *fp)
{
	int i;
	struct device far *dev = (struct device far *)fp;
	if(dev->attribute & 0x8000)
	{
		for(i=0;i<8; ++i)
			printf("%c", dev->name[i]);
	}
	else
		printf("BLK DEV ");

	printf("at %04X:%04X  ", FP_SEG(fp), FP_OFF(fp));
	printf("Attrib = %04X Strategy = %04X Interrupt # %04X\n",
		dev->attribute, dev->strategy_offset, dev->interrupt_offset);
}


/*
 * SetVideo Mode
 * 	Takes the Video Mode name and video mode to that mode!
 * Return
 *	TRUE if valid Video Mode
 * 	FALSE if non-valid Video Mode
 */
int SetVideoMode(char *ModeName)
{
	int i;
	union REGS regs;
	static struct { char *Str; int VideoMode; } DispAdptr[] = {
		{ "L25",    3 },
		{ "L43",   -1 },  // EGA 43 line mode: i = 1 => 350 scan lines
		{ "L50",   -1 },	// VGA 50 line mode: i = 2 => 400 scan lines
		{ "40",		1 },
		{ "80",		3 },
		{ "BW80",	3 },
		{ "BW40",	0 },
		{ "CO80",	3 },
		{ "CO40",	1 },
		{ "MONO",	7 },
	};
	strupr(ModeName);
	for(i=0; i < (sizeof(DispAdptr) / sizeof(DispAdptr[0])); ++i)
	{
		if(strcmp(ModeName, DispAdptr[i].Str) == 0)
		{
			if(DispAdptr[i].VideoMode >= 0)
			{
				// select a 25 line display mode
				//
				regs.x.ax = (unsigned)(DispAdptr[i].VideoMode);
				int86(0x10, &regs, &regs);
		   }
			else
			{
//				int j;

			   // Select the appropriate number of scan lines
				// then switch to mode 3 and select the 8x8 font
				// and the alternate EGA/VGA BIOS 43/50 line
				// PrtScr function.
				//
				// NOTE: i MUST be 1 (350 lines) or 2 (400 lines)
				//
				asm {
					push	bx
					mov	ax, i				;// set scan lines
					mov	ah, 12h
					mov	bl, 30h
					int   10h

					mov	ax, 3				;// mode 3
					int 	10h

					mov	ax, 1112h		;// select 8x8 font
					xor	bx, bx
					int	10h

					mov	ah, 12h			;// select alternate PrtScr function
					mov	bl, 20h
					int	10h
					pop	bx
				}
			}
			return 1;
		}
	}
	return 0;
}
