//
//   $Id: PRINTF.C 1.4 2002/10/16 21:55:48Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	PRINTF.C  created 04/18/96  by Drew Gislason

	Copyright (c) 1996-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Contains printf() in all its glory (well, some of it
	anyway).

	Supports
		%c    char
		%d    integer
		%i    same as %d
		%u    unsigned
		%x    unsigned in hex form, UPPERCASE
		%s    char * (in ascii)
		%p    near pointer (same as %04x)
		%ld   long integer
		%li   same as %ld
		%lu   unsigned long
		%lx   unsigned long in hex form, UPPERCASE
		%ls	char far * (in ascii)
		%lp	far pointer (same as %04X:%04X)

	Also support "count", "fill", "left justify" and "commas"
	modifiers:
		%11d		fill with up to 11 spaces
		%011d		fill with up to 11 0s
		%-5d		left justify with fill
		%,5d		use commas for thousands separator

**********************************************************/
#include <string.h>
#include "printf.h"

#define TRUE 1
#define FALSE 0

// where does the printf end up?  Handle 1 is stdout which can be
// redirected from the cmdline.  Handle 2 is stderr, which can't.
#define STDOUT 1
#define STDERR 2

#define IsDigit(c) ((c) >= '0' && (c) <= '9')

// can be changed for foreign country support
#if OPT_THOUSAND
char printf_thousands_sep = ',';
static char sep;								// fill number separator
#endif

// various printf modifiers, set during a given call to printf
#if OPT_FILL
static char fill;		// fill character
static int	count;	// fill count
static char left;		// fill alignment
#endif

static int printfHandle = STDOUT;
static int charCount;

// the output buffer
static char buffer[BUFSIZE];
static int position;

#if SPRINTF
// special output case: if this variable is set, then we are
// sending output to a user buffer instead of the screen.
static char * pUserBuffer;
#endif


#if SELF_CONTAINED
/*
	common routine used by itoa, ltoa and ultoa.

	Change a number from a word to a decimal or hexidecimal
	number.  Radix can be anything from 2 to 16.  The "pOutStr"
	is guaranteed not to be filled with any more chars than is
	absolutely neccessary to represent the number.
*/
void _xultoa(unsigned long value, char *pOutStr, int radix)
{
	static char hex[] = "0123456789ABCDEF";
	int first, last, size;

	// invalid string, do nothing
	// should ASSERT here
	if(!pOutStr)
		return;

	// invalid radix, do nothing
	// should ASSERT here
	if(radix < 2 || radix > 16)
		return;

	// convert number to string form. since we don't know
	// up front how big it will be, we will need to reverse
	// it later.
	size = 0;
	do
	{
		// record this digit
		pOutStr[size] = hex[(int)(value % radix)];

		// on to next digit
		value /= radix;
		++size;

	} while(value);

	// null terminate the string
	pOutStr[size] = 0;

	// reverse the numbers, as they came out backwards
	first = 0;
	last = size - 1;
	while(first < last)
	{
		char c;

		// swap the first and last
		c = pOutStr[first];
		pOutStr[first] = pOutStr[last];
		pOutStr[last] = c;

		// on to next set of characters
		++first, --last;
	}
}

/*
	The standard itoa function.
*/
 char * itoa(int value, char *pStr, int radix)
{
	char * pOrgStr = pStr;

	// handle negative values for base 10
	if(radix == 10 && value < 0)
	{
		*pStr++ = '-';
		value = -value;
	}

	// do the real work
	_xultoa((unsigned)value, pStr, radix);

	// return the original string
	return pOrgStr;
}

/*
	The standard ltoa function.
*/
 char * ltoa(long value, char *pStr, int radix)
{
	char * pOrgStr = pStr;

	// handle negative values for base 10
	if(radix == 10 && value < 0)
	{
		*pStr++ = '-';
		value = -value;
	}

	// do the real work
	_xultoa((unsigned long)value, pStr, radix);

	// return the original string
	return pOrgStr;
}

/*
	The standard ultoa function.
*/
 char * ultoa(unsigned long value, char *pStr, int radix)
{
	// do the real work
	_xultoa(value, pStr, radix);

	// return the original string
	return pStr;
}

#endif	// SELF_CONTAINED


/*
	Flush the printf buffer to
*/
void _FlushBuffer(void)
{
	// nothing to do? just return
	if(!position)
		return;

#if SPRINTF
	// flush the buffer to user space
	if(pUserBuffer)
	{
		memmove(pUserBuffer, buffer, position);
		pUserBuffer += position;
	}

	// flush the buffer to DOS handle
	else
#endif // SPRINTF
	{
		// workaround for BUG in BC 2.0 compiler that screws up
		// on inline assembly.  If a variable is on an odd boundary,
		// generates the instruction from the wrong address!
		unsigned r_bx = printfHandle;
		unsigned r_cx = position;
		unsigned r_dx = (unsigned)(buffer);

		// call DOS to display the string
		asm	push	bx
		asm	mov	bx, r_bx
		asm	mov	cx, r_cx
		asm	mov	dx, r_dx
		asm	mov	ah, 40h
		asm	int	21h
		asm	pop	bx
	}

	// update # of characters flushed
	charCount += position;

	// buffer is now empty
	position = 0;
}

/*
	Display a single character to the "current" output device,
	usually STDOUT.
*/
void _putc(char c)
{
	// no room left in output buffer? then flush it!
	if(position >= BUFSIZE)
		_FlushBuffer();

	// add character to end of output buffer
	buffer[position++] = c;
}

/*
	Put a whole string to the "current" output device.
	See _putc.
*/
void _puts(char *s)
{
	while(*s)
	{
		// expand LFs into CR/LF pairs
		if(*s == '\n')
			_putc('\r');

		// put the character
		_putc(*s++);
	}
}

/*
	Put a number or string, paying attention to fill bytes,
	count and alignment.
*/
void _PutStringWithFill(char *pStr)
{
#if OPT_FILL || OPT_THOUSAND
	int len;

	// determine length of string
	len = strlen(pStr);
#endif

#if OPT_THOUSAND
	// expand the string with "thousands" separators
	if(sep)
	{
		int numLen;

		// only need a thousands separator every 3rd number
		numLen = len;
		while(numLen > 3)
		{
			numLen -= 3;

			// open up string and put in thousands ","
			memmove(pStr+numLen+1, pStr+numLen, len-numLen+1);
			pStr[numLen] = sep;

			// string is now 1 byte longer
			++len;
		}
	}
#endif // OPT_THOUSAND

#if OPT_FILL
	// left aligned string? Then spit out the meat BEFORE the fill
	if(left)
		_puts(pStr);

	// any characters to fill?
	while(len < count)
	{
		_putc(fill);
		--count;
	}

	// right aligned string? Then spit out the meat AFTER the fill
	if(!left)
#endif
		_puts(pStr);
}

/*
	Just as with standard printf(), this function is the real
	work-horse.  It prints the string to a buffer, and flushes
	to screen or user space.  This function supports a subset
	of the standard C printf().  See module header for a list
	of what's supported.

	Note: to fool the Datalight prototyping program, the
	function header must be indented.
*/
 int vprintf(const char *fmt, void *xargs)
{
	int *args = (int *)xargs;
	char szTempStr[16];		// big enough to hold a long "-2,147,483,648"

	// we will be counting the # of character displayed.
	// start with none.
	charCount = 0;

	// keep going until the end of the format string
	while(*fmt)
	{
		// we can't ASSERT here, so just return if fmt string is bad
		if(!fmt)
			break;

		// general output
		if(*fmt != '%')
		{
			// special case: display a CR before a LF to screen,
			// but NOT when using sprintf().
			if(*fmt == '\n')
			{
#if SPRINTF
				if(!pUserBuffer)
#endif
					_putc('\r');
			}

			// output the character
			_putc(*fmt);
		}

		/* this is a formatted output */
		else
		{
			// skip the "%" part of the formatted output
			++fmt;

#if OPT_FILL
			// determine fill alignment (left or right)
			left = FALSE;
			if(*fmt == '-')
			{
				++fmt;
				left = TRUE;
			}
#endif // OPT_FILL

#if OPT_THOUSAND
			// Use thousands separator?
			sep = 0;
			if(*fmt == ',')
			{
				++fmt;
				sep = printf_thousands_sep;
			}
#endif // OPT_THOUSAND

#if OPT_FILL
			// is there a fill count?
			count = 0;
			if(IsDigit(*fmt))
			{
				// should we fill with 0s instead of spaces?
				if(*fmt == '0')
					fill='0', ++fmt;
				else
					fill=' ';

				// determine count of bytes to fill
				while(IsDigit(*fmt))
				{
					count *= 10;
					count += *fmt++ - '0';
				}
			}
#endif // OPT_FILL

			// what kind of thing is this?
			switch(*fmt)
			{
#if FMT_C
				// character (%c)
				case 'c':
					// get the character
					szTempStr[0] = (char)(*args++);
					szTempStr[1] = 0;

					// we do it in string form so we can fill if needed
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_C

#if FMT_D
				// integer (%d)
				case 'd':
				case 'i':

					// convert to number string
					itoa(*args++, szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_D

#if FMT_U
				// unsigned (%u)
				case 'u':
					// convert to number unsigned string
					ltoa((unsigned)(*args++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_U

#if FMT_X
				// hex (%x)
				case 'x':
				case 'X':
					// convert to number unsigned string
					itoa(*args++, szTempStr, 16);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
					break;
#endif // FMT_X

#if FMT_S
				// string (%s)
				case 's':

					// simply display the string
#if (__SMALL__ || __TINY__)
					_PutStringWithFill((char *)(*args++));
#else
					{
						char far *lpStr;
						lpStr = (char far *)(*(((long *)args)++));
						_PutStringWithFill(lpStr);
					}
#endif
					break;
#endif // FMT_S

#if FMT_P
				// pointer in hex form (%p)
				case 'p':

					// set fill and count, as if user did %04x
					fill='0'; count=4;

					// get and display the near pointer
					itoa(*args++, szTempStr, 16);
					_PutStringWithFill(szTempStr);

					break;
#endif // FMT_P

				// some printf() things are longs
				case 'l':

					// skip the 'l', to get to the type 'd', 's', etc..
					++fmt;

#if FMT_LD
				// long integer (%ld)
				if(*fmt == 'd' || *fmt == 'i')
				{
					// convert to long number string
					ltoa(*(((long *)args)++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LD

#if FMT_LU
				// long integer (%lu)
				if(*fmt == 'u')
				{
					// convert to long number string
					ultoa(*(((long *)args)++), szTempStr, 10);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LU

#if FMT_LX
				// long hex number (%lx)
				if(*fmt == 'x' || *fmt == 'X')
				{
					// convert to long number string
					ltoa(*(((long *)args)++), szTempStr, 16);

					// display it, may expand szTempStr with commas
					_PutStringWithFill(szTempStr);
				}
#endif // FMT_LX

#if FMT_LS
				// far string (%ls)
				if(*fmt == 's')
				{
					char far *lpStr;

					// determine start of string
					lpStr = (char far *)(*(((long *)args)++));

					// display each character
					while(*lpStr)
					{
						_putc(*lpStr);
						++lpStr;
					}
				}
#endif // FMT_LS

#if FMT_LP
				// far pointer (%lp) (eg 0300:0FAE)
				if(*fmt == 'p')
				{
					// set fill and count, as if user did %04x
					fill='0'; count=4;

					// get and display segment part of number
					itoa(args[1], szTempStr, 16);
					_PutStringWithFill(szTempStr);

					// separate by a colon
					_putc(':');

					// get and display offset part of number
					itoa(args[0], szTempStr, 16);
					_PutStringWithFill(szTempStr);

					// skip the far pointer	argument
					args += 2;
				}
#endif // FMT_LP

					// done with case '%l'
					break;

				// unknown formatting character, just spit it out
				default:
					_putc(*fmt);
			}
		}

		// try next character
		++fmt;
	}

	// flush the last of the output to stdout or wherever it's going
	_FlushBuffer();

	// return the # of characters printed
	return charCount;
}

#if PRINTF
/*
	Yes, it's printf() (or a subset, anyway).  Displays
	output to STDOUT.

	Note: to fool the Datalight prototyping program, the
	function header must be indented.
*/
 int printf(const char *fmt, int arglist)
{
	printfHandle = STDOUT;
	return vprintf((void *)fmt, (void *) &arglist);
}
#endif PRINTF

#if EPRINTF
/*
	This function is just like printf(), but displays
	output to STDERR instead of STDOUT.

	Note: to fool the Datalight prototyping program, the
	function header must be indented, and the "public"
	version of the function header here in the comment
	must be in the leftmost column.

int eprintf(const char *fmt, ...)
*/
 int eprintf(const char *fmt, int arglist)
{
	printfHandle = STDERR;
	return vprintf((void *)fmt, (void *) &arglist);
}
#endif EPRINTF

#if SPRINTF
/*
	Standard sprintf.
	output to STDERR instead of STDOUT.

	Note: to fool the Datalight prototyping program, the
	function header must be indented.
*/
 int sprintf(char *pMemory, char *fmt, int arglist)
{
	int charCount;

	// inform the Flush function to flush to memory, not screen
	pUserBuffer = pMemory;

	// spit the output to memory
	charCount = vprintf((void *)fmt, (void *) &arglist);

	// null terminate the string
	_putc(0);
	_FlushBuffer();

	// no longer in sprintf
	pUserBuffer = 0;

	// return # of characters copied, excluding the null
	return charCount - 1;
}
#endif SPRINTF
