//
//   $Id: DLFILE.H 1.4 2002/10/16 21:55:48Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
//
// DLFILE.H Created 5-6-97 Dennis Edwards
// DOS file I/O routines for the Datalight C utility libarary.
//
// Copyright (c) 1989-2002 Datalight, Inc.
// All Rights Reserved.
// Datalight Confidential and Proprietary
//
// Most of the routines in this module are simply wrappers
// around DOS int 21h calls. In general the calls return
// TRUE if they worked, FALSE in not. Any values are returned
// via pointers passed as parameters in the call.
//
// In case of an unsuccessful call, you may retrieve the last
// DOS error return code using the LastDOSError() function or
// get and reset the value with ResetDOSError(). DLERRCOD.H/.INC
// describes each of the possible DOS return codes.
//
// See any DOS programming reference for more information.
//
//
#ifndef __DLFILE_H
#define __DLFILE_H


// Access to the last error code value
//
int PASCAL LastDOSError(void);      // get and leave unchanged
int PASCAL ResetDOSError(void);     // get and reset


// file/directory attributes (in a DIR) used by DOSExtendedOpen(),
// DOSOpen(), DOSCreate(), DOSFindFirst/Next, etc.
//
#define A_NORMAL     0x00           // OK for Open or Creat
#define A_READONLY   0x01           // OK for Open or Creat
#define A_HIDDEN     0x02           // OK for Open or Creat
#define A_SYSTEM     0x04           // OK for Open or Creat
#define A_VOLUME     0x08
#define A_LONGNAME   0x0F           // a long filename
#define A_SUBDIR     0x10
#define A_ARCHIVE    0x20           // OK for Open or Creat
#define A_DEVICE     0x40


//THOM  this section is already in DOSDEF.H
//THOM
//THOM// openMode values for DOSExtendedOpen()
//THOM
//THOM#define O_RDONLY     0x0000
//THOM#define O_WRONLY     0x0001
//THOM#define O_RDWR       0x0002
//THOM#define O_RESERVED   0x0008   /* must not be used - reserved */
//THOM#define O_DENYOPEN   0x0010
//THOM#define O_DENYWRITE  0x0020
//THOM#define O_DENYREAD   0x0030
//THOM#define O_DENYNONE   0x0040
//THOM#define O_SHARE      0x0070   /* share bits */
//THOM#define O_SHARESHFT  4
//THOM#define O_NOCHILD    0x0080
//THOM#define O_NOINT24    0x2000
//THOM#define O_WRITENOW   0x4000

//  action values for DOSExtendedOpen()
//
#define ACT_EXIST_OPEN     0x0001
#define ACT_EXIST_REPLACE  0x0002
#define ACT_NEXIST_CREATE  0x0010

// action values returned by DOSExtendedOpen()
//
#define ACTION_OPENED      1
#define ACTION_CREATED     2
#define ACTION_TRUNCATED   3

bool PASCAL DOSExtendedOpen(char far *lpName, unsigned openMode,
                                unsigned attributes, unsigned far * lpAction,
                                int far *lpHandle);
bool PASCAL DOSOpen(char far *lpName, unsigned attributes,
                                int far *lpHandle);
bool PASCAL DOSCreate(char far *lpName, unsigned attributes,
                                int far *lpHandle);
bool PASCAL DOSClose(unsigned handle);
bool PASCAL DOSCommit(unsigned handle);
bool PASCAL DOSRead(unsigned handle, uchar far *lpBuffer,\
                                unsigned far *lpByteCount);
bool PASCAL DOSWrite(unsigned handle, uchar far *lpBuffer,\
                                unsigned far * lpByteCount);
bool PASCAL DOSDelete(char far *lpName);
bool PASCAL	DOSRename(char far *lpOldName, char far *lpNewName);

bool PASCAL GetFileDateAndTime(int handle, unsigned far *lpDate,
											unsigned far *lpTime);
bool PASCAL SetFileDateAndTime(int handle, unsigned date, unsigned time);

//  Location codes for DOSSeek
//
#define SEEK_CUR    1
#define SEEK_END    2
#define SEEK_SET    0

bool PASCAL DOSSeek(unsigned handle, long far *lpDistance, uchar from);

#endif  // __DLFILE_H
