//
//   $Id: MOVE.C 1.10 2002/10/16 21:55:48Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*--------------------------------------------------------------

	MOVE.C Created 05-05-97 Dennis Edwards

	Copyright (c) 1989-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary

	This is the main source file for the MOVE.EXE utility.

	After the switches have been processed by ParseCmdLine(), the
	MOVE command line arguments may appear as below (argv[0] is
	still the program name.)


	arg #:    1     2     3     4   5   6=argc

	value: |s,t,| |u?| |,v*,w| |,| |x| |d|


	The destination argument is syntacticly defined as the last
	argument not preceeded by a comma, provided that at least two
	arguments exist.

	We want to determine if a destination path was specified and,
	if so then

	1) Whether the destination contains wild cards
	2) Whether the destination exists and, if so,
	3) Is the existing destination a file or a dir


	The source argument(s) are considered wild if

	1) more than one is listed
	2) any one them contains a wild card
	3) "." or ".." are interpretted as wild-cards but "X." is
	   not wild. "d:" and "\" are also implied wild-cards.


	The MOVE states can be represented by a 2x2x2 matrix in which
	cell [1][0][1] resolves to a decision tree. The first level
	2x2 matrix ([1][?][?]) is defined by the source input
	specification of which two questions must be asked: 1) does
	the input specify multiple files (either through wild cards
	or enumeration) and, if not, then 2) does the input represent
	a file or a directory.  The source must exist on disk in
	order to MOVE it.


						multiple source specs	single source spec

	exists as file(s)		move/copy 		   move/copy/(rename?)

	exists as dir(s)			error						rename dir


	The significance of the destination specification is
	determined by the location in the above matrix.


							destination dir name		depends, see below

							error							new name of renamed dir


	The decision tree for cell [1][0][1] is:


	entry exists?					NO: move to specified path name
											 possibly making dirs to put
											 it in (if user says ok).

	exists as file?				YES: prompt user for over-write

	else exists as dir, append name part from source onto destination

	does entry exist in dir?	NO: move to dir

	does entry exit as file?	YES: prompt user to over-write

	else entry exists as a dir which is an error


	Most of the code in this module goes into figuring out which
	of	the above eight squares the program is living in.

--------------------------------------------------------------*/

#define __FILENUM__ 1
#define	MOVE	1


#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"

#include "dosver.h"
#include "dlapp.h"
#include "dlpath.h"
#include "dlerrcod.h"
#include "fatworld.h"
#include "stdc.h"

#include "dlrespon.h"
#include "movemsg.h"

#define MOVE_ATTRIBS			(A_ANYTHING & ~A_VOLUME & ~A_HIDDEN & ~A_SYSTEM)
#define COPY_BUFFER_SIZE	0x8000

#define INVALID_MOVE_CHAR  (char) 0x07

#define TURBO_DEBUGGER FALSE

//
// Globals, globals, globals
//
// These are used by several of the path routines
//


// The file I/O buffer for copying files
//
char copyBuffer[COPY_BUFFER_SIZE];

// findfirst/next structs, one for source, one for destination
//
FIND destFound, srceFound;

// state variables used to figure out how to handle the input
//
bool destExists, destIsSubdir, sourceIsWild, sourceIsSubdir, sourceExists;

// the current index into the argv[] array as returned by
// ParseCmdLine()...
//
int curArgIndex;

// Storage for the path up to the file and for full file names.
// Source and dest are likely in different trees.
//
char szDestinationSpec[LFN_MAX_PATH], szSourceSpec[LFN_MAX_PATH];
char szDestinationPath[LFN_MAX_PATH], szSourcePath[LFN_MAX_PATH];

// This is turned off by the /Y switch (and back on with /-Y)
// May also be altered by the value of the COPYCMD environment
// variable (if it exists).
//
bool notPrompting = FALSE;
bool fHideSignon  = FALSE;


// The argument definitions
//
ARGDEF MoveOpts[] =
{
	{ BOOL_SWITCH, &notPrompting, 'y' },
	{ BOOL_SWITCH, &notPrompting, 'c' },
	{ BOOL_SWITCH, &fHideSignon,  '[' },
	{ 0 }
};

bool fIsFat32World;

void SignOn(void)
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signOnMsg,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signOnMsg,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}


void Usage(void)
{
	SignOn();
	printf(usageMsg);
	exit(ERR_MS_1);
}


void InternalError(int fileNum, int lineNum)
{
	printf("ASSERT %u.%u\n",fileNum,lineNum);
	exit(ERR_INTERNAL);
}


// Determines whether or not the user input describes multiple
// files.
//
// Note: argc will have the value returned by ParseCmdLine()
//			NOT the value establised by the startup code...
//			argv will have the values established by ParseCmdline()
//			(all switch args will have been removed).
//
bool HaveMultipleSourceSpecs(int argc, char *argv[])
{
   char *pArgStr;

	// can't have multiple destinations...
	//
	if(argc > 2)
		return TRUE;

	// only have one input to consider...
	//
	pArgStr = argv[1];

	// "." or ".." are both wild
	//
	if(*pArgStr == '.')
	{
		// don't get fooled by ".\xxx" or "..\." or ...
		//
		if((pArgStr[1] == '\0') || (pArgStr[2] == '\0'))
			return TRUE;
	}

	// "d:" is wild
	//
	if((pArgStr[1] == ':') && (pArgStr[2] == '\0'))
		return TRUE;

	// look for specification separators along with "real" wild-cards
	//
	while(*pArgStr)
	{
		if(   (*pArgStr == '*') || (*pArgStr == '?')
         || (*pArgStr == INVALID_MOVE_CHAR))
      {
			return TRUE;
      }
		++pArgStr;
	}
	--pArgStr;

	// finaly, check the end of the string for "." or ".."
	// watch out for "d:\path\name."
	//
	if(*pArgStr == '.')
	{
		--pArgStr;
		if((*pArgStr == '\\') || (*pArgStr == '.'))
	  		return TRUE;
	}
	return FALSE;
}


// converts:
//
// d:\path\file.*	=> d:\path\file
// d:\path\*.*		=>	d:\path
//
void RemoveWildCardsFromName(char *pPath)
{
	if(LastChar(pPath) == '*')
	{
		pPath[strlen(pPath) - 2] = '\0';			// delete ".*" from end of path
		if(LastChar(pPath) == '*')
			pPath[strlen(pPath) - 2] = '\0';		// delete "\*" from end of path
	}
}


// This attempts to determine the source path from a source spec
// and returns TRUE if successful, FALSE if not.
//
//	examples:	\ 			=> d:\*.*
//					. or d:	=> d:\path\*.*
//					subdir	=> d:\path\subdir
//					file		=> d:\path\file
//
bool SourceSpecToSourcePath(char szSpec[])
{
   int len;
   char last, nextToLast;
   bool HasWildcards = FALSE;

   // There are certain things that we want to convert into
	// wild-cards
	//
   HasWildcards = IsWild(szSpec);
	if(!HasWildcards)
	{
		len        = strlen(szSpec);
		last       = szSpec[len-1];
		nextToLast = szSpec[len-2];

		if(last == '.')
		{
			if(len == 1)
			{
				strcpy(szSpec,"*.*");										// "."
			}
			else if(nextToLast == '.')
			{
				strcat(szSpec,"\\*.*");     								// ".."
			}
			else if((nextToLast == ':') || (nextToLast == '\\'))
			{
				strcat(szSpec,"*.*");										// ":."
			}																		// "\."
		}
		else if(last == ':')
		{
			strcat(szSpec,"*.*");											// "d:"
		}
		else if(last == '\\')
		{
			strcat(szSpec,"*.*");											// "\"
		}
	}

   return TRUE;
}


// Tries to Extract the next source pathname from the command
// line input. Returns TRUE if successful, FALSE if no (more)
// valid source specifications.
//
// Note: Validating the existence of the input spec has the
//       side effect of setting up the findnext block for
//       wild card processing of files to be MOVEd
//
// Note: argc will have the value returned by ParseCmdLine()
//			NOT the value establised by the startup code...
//			argv will have the values established by ParseCmdline()
//			(all switch args will have been removed).
//
bool SourcePathExists(int argc, char *argv[])
{
   bool srceSpecFound;
   char *pArgStr;
   int i;

	if(curArgIndex >= argc)
		return FALSE;

	// for all (remaining) source specs in the argument list
	//
	do
	{
		// reset the source spec index and get a ptr to the next arg
		//
		i             = 0;
		srceSpecFound = FALSE;
		pArgStr 	     = argv[curArgIndex];

		// while chars left in the current arg string
		// copy them into source spec
		do
		{
			// if we hit an argument separator, terminate the argstring
			// and record the start of the next arg
			//
			if(*pArgStr == INVALID_MOVE_CHAR)
			{
				argv[curArgIndex] = ++pArgStr;
				break;
			}
			else
				szSourceSpec[i++] = toupper(*pArgStr++);

		} while(*pArgStr);
		szSourceSpec[i]   = '\0';

		// if we processed the whole arg string, point to the next one
		// if the source spec is empty, go on to the next arg string
		//
		if(*pArgStr == '\0')
		{
			++curArgIndex;
			if(*szSourceSpec == '\0')
				continue;
		}

      SmartExpandPath( FULLPATH_DEFNAME, szSourceSpec, szSourcePath );

		if(SourceSpecToSourcePath(szSourceSpec))
		{
			// see if the source path exists and complain if not
			//
			srceSpecFound = SmartFindFirst( MOVE_ATTRIBS, szSourcePath, &srceFound );
			if(!srceSpecFound)
			{
				printf(moveNotFoundMsg, szSourcePath);
				continue;
			}

			// was source spec a subdir other than "."?
			//
			sourceIsSubdir=(((srceFound.attr & A_SUBDIR)==A_SUBDIR) &&
								  (*srceFound.name != '.'));

		}
	} while(!srceSpecFound  &&  (curArgIndex < argc));

	return srceSpecFound;
}


// Determines if there is destination specification on the command
// line and if so if it represents a single file or directory;
// if it does not then this will display an error message and exit.
//
// Note: argc will have the value returned by ParseCmdLine()
//			NOT the value establised by the startup code...
//			argv will have the values established by ParseCmdline()
//			(all switch args will have been removed).
//			argv[argc] will be the destination spec.
//
void ValidateDestinationSpec(int argc, char *argv[])
{
	static char szDestinationSpecTemp[LFN_MAX_PATH];
	bool fSuccessful;
	unsigned uLength;


	// Is a destination spec present in the input? It won't be if:
	// 1) There are not at least two arguments
	// 3) if argv[argc-1][len-1] == ','
	// In which case we have a fatal missing parameter error.
	//
	if(argc < 2)
	{
		Usage();							// calls exit(ERR_PARAMATER_MISSING);
	}
	if(argv[argc-1][strlen(argv[argc-1])-1] == ',')
	{
		Usage();
	}

	// The destination is specified in argv[argc]. If it has wild cards
	// then it is an invalid specification.
	//
	if(IsWild(argv[argc]))
	{
		printf(not1to1Msg);  // Like MS-DOS
		exit(ERR_MS_1);
	}

	// We have a syntactically valid destination spec
	// Copy it into our array, making it all upper case
	//
	strcpy( szDestinationSpecTemp, argv[argc] );
	strupr( szDestinationSpecTemp );


	// See if the destination spec exists and, if so whether it
	// is a file or a dir.
	//
	if( (szDestinationSpecTemp[1] == ':') &&
			(szDestinationSpecTemp[2] == 0x00) )
	{
		szDestinationSpecTemp[2] = '.';
		szDestinationSpecTemp[3] = 0x00;
	}

	uLength = strlen( szDestinationSpecTemp );
	if( uLength >= 2 )
	{
		if( ( szDestinationSpecTemp[uLength-1] == '\\' ) &&
				( szDestinationSpecTemp[uLength-2] != ':' ) )
		{
			szDestinationSpecTemp[uLength-1] = 0;
		}
	}

	fSuccessful = SmartExpandPath( FULLPATH_DEFNAME, szDestinationSpecTemp,
	      szDestinationSpec );
	if( !fSuccessful )
	{
		strcpy( szDestinationSpec, szDestinationSpecTemp );
	}

	if( (szDestinationSpec[1] == ':') && (szDestinationSpec[2] == '\\') &&
			(szDestinationSpec[3] == 0x00) )
	{
		destExists = TRUE;
		destIsSubdir = TRUE;
	}
	else
	{
		destExists = SmartFindFirst( A_ANYTHING, szDestinationSpec, &destFound );
		if(destExists)
			destIsSubdir = ((destFound.attr & A_SUBDIR) == A_SUBDIR);
	}
}


// Given a syntactically valid destination spec, try to make it
// into a full path string that may or may not exist. If it
// doesn't exist then we don't yet know if the "name" part
// represents a file or a subdirectory; we figure that out
// late in main().
//
void DestinationSpecToDestinationPath(char szSpec[])
{
	int drive;

	bool fSuccessful;

	if(destExists)
	{
      strcpy( szDestinationPath, szSpec );
	}
	drive = (int)(toupper(szSpec[0]) - 'A');

	// If drive is invalid or can't get current dir for the
	// drive, abort...
	//
	if(!IsValidDrive(drive))
	{
		printf(badDriveMsg,szSpec);
		exit(ERR_MS_1);
	}

	fSuccessful = SmartFindFirst( A_ANYTHING,szDestinationSpec,&destFound );
	if( fSuccessful )
	{
		if((destFound.attr & A_SUBDIR) == A_SUBDIR)
		{
			strcpy( szDestinationPath, szDestinationSpec );
		}
	}
	else
	{
		strcpy( szDestinationPath, szDestinationSpec );
		RemoveName(szDestinationPath);
	}
}


// This copies one file into another, creating/overwriting
// the destination as neccessary
//
int CopyFile(char *pSrcePath, char *pDestPath)
{
   int action, srceHandle, destHandle, status, pActionTaken;
   unsigned read, written;
   struct DOSFileTime sTime;
   struct DOSFileDate sDate;

   dlResetDOSError();

	action = ACT_EXIST_OPEN;
   if (!SmartCreateOpenFile( O_RDONLY, A_NORMAL, action,
                             pSrcePath, &srceHandle, &pActionTaken ))
	{
		printf(moveStatusMsg,cannotOpenSrce);
		return dlResetDOSError();
	}
	action = ACT_EXIST_REPLACE | ACT_NEXIST_CREATE;
   if (!SmartCreateOpenFile( (O_WRITENOW | O_RDWR), A_NORMAL, action,
                             pDestPath, &destHandle, &pActionTaken ))
	{
		dlCloseFile(srceHandle);
		printf(moveStatusMsg,cannotOpenDest);
		return dlResetDOSError();
	}

	status = ERR_NONE;
   dlGetFileDateAndTime(srceHandle, &sTime, &sDate);
	do
	{
		read = COPY_BUFFER_SIZE;

		if(dlReadFile(srceHandle, &read, copyBuffer))
		{
			written = read;
			if(dlWriteFile(destHandle, &written, copyBuffer))
			{
				if(written == COPY_BUFFER_SIZE)
					continue;
				else if(read == written)
					break;
				else
					status = dlCheckDOSError();
			}
			else
				status = dlCheckDOSError();
		}
		else
			status = dlCheckDOSError();

	} while(status == ERR_NONE);

   dlSetFileDateAndTime(destHandle, &sTime, &sDate);
	dlCloseFile(srceHandle);
	dlCloseFile(destHandle);

   dlResetDOSError();
	return status;
}


// The moves a file from one location to another, using renames
// if possible
//
int MoveIt(char *pSrcePath, char *pDestPath)
{
   static int response = NO, status;
   FIND foundDest, foundSrce;
   bool srceFound, destFound, wasMoved;

   dlResetDOSError();

	// conditionally over-write destination file
	//
   srceFound = SmartFindFirst( MOVE_ATTRIBS, pSrcePath, &foundSrce );
   destFound = SmartFindFirst( A_ANYTHING, pDestPath, &foundDest );
	if(destFound)
	{
		// if the destination pathname does not represent a file...
		//
		if((foundDest.attr & A_ANYFILE) != foundDest.attr)
		{
			printf(movingMsg,pSrcePath,pDestPath);
			printf(moveStatusMsg,cannotOpenDest);
			return(ERR_MS_1);
		}

		if(notPrompting || (response == ALL))
			response = ALL;
		else
			response = ValidateResponse(overwritePrompt,pDestPath,ynaString);

		if(response == NO)
			return ERR_NONE;

		printf(movingMsg,pSrcePath,pDestPath);

		status = CopyFile(pSrcePath,pDestPath);
		if(status == ERR_NONE)
		{
         SmartSetFileAttributes( pSrcePath, 0 );
         SmartWildcardDelete( pSrcePath );
         SmartSetFileAttributes( pSrcePath, foundSrce.attr );
			printf(moveStatusMsg,okMsg);
		}
		else if(status == ERR_HANDLE_DISK_FULL)
		{
			printf(moveStatusMsg,cannotWriteDest);
         SmartWildcardDelete( pDestPath );
		}
		else
		{
			printf(moveStatusMsg,cannotOpenDest);
         SmartWildcardDelete( pDestPath );
		}
      if( status != ERR_NONE )
      {
         status = ERR_MS_1;
      }
		return status;
	}

	printf(movingMsg,pSrcePath,pDestPath);

	if(!srceFound)
	{
		printf(moveStatusMsg,cannotOpenSrce);
		return(dlCheckDOSError());
	}
	if(foundSrce.attr & A_READONLY)
		SmartSetFileAttributes( pSrcePath, (foundSrce.attr & ~A_READONLY) );

	wasMoved = SmartRenameFileOrDirectory( pSrcePath, pDestPath );
	if(wasMoved)
	{
  		SmartSetFileAttributes( pDestPath, foundSrce.attr );
		printf(moveStatusMsg,okMsg);
		return(ERR_NONE);
	}
	status = CopyFile(pSrcePath,pDestPath);
	if(status == ERR_NONE)
	{
		SmartSetFileAttributes( pSrcePath, 0 );
      SmartWildcardDelete( pSrcePath );
		SmartSetFileAttributes( pDestPath, foundSrce.attr );
		printf(moveStatusMsg,okMsg);
	}
	else if(status == ERR_HANDLE_DISK_FULL)
	{
		printf(moveStatusMsg,cannotWriteDest);
      SmartWildcardDelete( pDestPath );
	}
	else
	{
		printf(moveStatusMsg,cannotOpenDest);
      SmartWildcardDelete( pDestPath );
	}
   if( status != ERR_NONE )
   {
      status = ERR_MS_1;
   }
	return status;
}


/*
	$DOCPROC:LFNReplaceCommas
   Replace comma parameter separators with invalid character (0x07)

	<"hello,world","goodbye,world"> becomes <"hello,world""goodbye,world">

   ***note*** this function could easily strip quotes as well...

*/
void PASCAL LFNReplaceCommas(char *pszArg)
{
	bool bInQuote = FALSE;

   do
   {
      if (*pszArg == '"')
      {
			bInQuote = !bInQuote;
      }
      else if (   (*pszArg == ',')        // looking for separating commas
               && (!bInQuote)             // that are not part of any LFN
              )
      {
         *pszArg = INVALID_MOVE_CHAR;     // replace with an invalid char
      }
   } while (*(++pszArg));
}


// MOVE
//
int main(int argc, char *argv[])
{
   int response;
   char far * lpCopyCmd;
	char szName[LFN_MAX_PATH];
   int status, i;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

#if TURBO_DEBUGGER
	for(i=1; i<argc; i++)
	{
      argv[i]=argv[i+1];
   }
   argc--;
#endif
	// Display copyright and version info then process user options.
	//
	curArgIndex = 1;
	lpCopyCmd = GetEnvEntry("COPYCMD");
	if(lpCopyCmd)
	{
		while(*lpCopyCmd)
		{
			if(toupper(*lpCopyCmd) == 'Y')
			{
				if(*(lpCopyCmd-1) != '-')
					notPrompting = TRUE;
				break;
			}
			++lpCopyCmd;
		}
	}
	for(i=1; i <= argc; i++)
	{
      LFNReplaceCommas( argv[i] );
   }
	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(ERR_MS_1);
		}
	}

	//  This is for the command line parsing library function.  It
	//  indicates that no command line options are ok
	//
	SetParseFlags( PARSEF_NOARGSOK );

	argc = ParseCmdline(argc-1, &argv[1], MoveOpts);

	// if not at least one path name exit
	//
	if(argc == 0)
	{
		printf( szMissingParms );
		exit(ERR_MS_1);
	}

	SignOn();

	// won't return if invalid...
	//
	ValidateDestinationSpec(argc, argv);

	// Have a valid spec, make it into a full path string
	// that may or may not exist. If it doesn't exist then
	// we don't yet know if the "name" part represents a file
	// or a subdirectory.
	//
	DestinationSpecToDestinationPath(szDestinationSpec);

	// Did the user specify multiple input files
	//
	sourceIsWild = HaveMultipleSourceSpecs(argc, argv);

	// Cannot concatinate files...
	//
	if(sourceIsWild && destExists && !destIsSubdir)
	{
		printf(not1to1Msg);
		exit(ERR_MS_1);
	}

	// we have to make a directory in which to move the files...
	//
	if(sourceIsWild && !destExists)
	{
		response = ValidateResponse(makeDirPrompt,szDestinationSpec,ynString);
		if(response == NO)
      {
	   	printf(not1to1Msg);
   		exit(ERR_MS_1);
      }
		else if(!MakeDirTree(szDestinationSpec))
		{
			printf(couldNotMakeMsg);
			exit(ERR_MS_1);
		}
      // Overwrite with new directory name
      strcpy( szDestinationPath, szDestinationSpec );
	}

	// must have something to move and, if multiple files
	// were specified, then subdirectories don't count...
	//
	do
	{
		sourceExists  = SourcePathExists(argc, argv);
		if(!sourceExists)
		{
		   printf(moveNotFoundMsg,argv[1]);
			exit(ERR_MS_1);
		}
		if((sourceIsWild && sourceIsSubdir) ||
		   (srceFound.attr & A_VOLUME))
		{
			printf(movingMsg,szSourcePath,szDestinationPath);
			printf(moveStatusMsg,cannotOpenSrce);
		}
	} while(sourceIsWild && sourceIsSubdir);

	// only a single source dir was specified
	//
	if(!sourceIsWild)
	{
		// renaming a subdir and something already has that name?
		//
		if(destExists && sourceIsSubdir)
		{
			printf(alreadyExistsMsg);
			exit(ERR_MS_1);
		}

		// renaming a subdir...
		//
		if(sourceIsSubdir)
			return MoveIt(szSourcePath,szDestinationSpec);

		// moving a single file...
		//
		if(destExists)
		{
			if(destIsSubdir)
			{
				// The destination name exists as a subdir
				// See if there is a dir entry with the same
				// source name in the destination subdir
				//
				CopyNamePart(szName,szSourcePath);
				AddName(szDestinationPath,szName);
				destExists = SmartFindFirst( A_SUBDIR, szDestinationPath,
                  &destFound );

				// if there is an entry with the same name and
				// that entry is a subdir then we're cooked...
				//
				if(destExists)
				{
					destIsSubdir = (destFound.attr & A_SUBDIR);
					if(destIsSubdir)
					{
						printf(movingMsg,szSourcePath,szDestinationPath);
						printf(moveStatusMsg,cannotOpenDest);
						exit(ERR_DESTINATION_EXIST);
					}
				}
			}

			// The file either doesn't exist in the destination
			// dir specified on the command line or it exists as
			// a file in one of two subdirs...
			//
			return MoveIt(szSourcePath,szDestinationPath);
		}

		// the destination file name does not exist, see if
		// the directory it goes in exists...
		//
		CopyNamePart(szName,szDestinationSpec);
		destExists = SmartFindFirst( A_SUBDIR, szDestinationPath, &destFound );

		// if the destination dir for the file doesn't exist,
		// prompt the user to create it...
		//

		if( (szDestinationPath[1] == ':') && (szDestinationPath[2] == '\\') &&
				(szDestinationPath[3] == 0x00) )
		{
			destExists = TRUE;
			destIsSubdir = TRUE;
		}
		else if(!destExists)
		{
			response=ValidateResponse(makeDirPrompt,szDestinationSpec,ynString);
			if(response == NO)
				exit(ERR_NONE);
			else if(!MakeDirTree(szDestinationSpec))
			{
				printf(couldNotMakeMsg);
				exit(ERR_DESTINATION_EXIST);
			}
         // Overwrite with new directory name
         strcpy( szDestinationPath, szDestinationSpec );
		}
		AddName(szDestinationPath, szName);

		// MOVE the file into the dir with the specified name
		//
		return MoveIt(szSourcePath,szDestinationPath);
	}

	// one or more files where specified for the source
	// (possibly implicitly by ".", etc.). Move all the
	// files to the destination dir using renames if
	// possible.
	//

	do
	{
		sourceExists = SmartFindFirst( A_ANYFILE ,szSourcePath, &srceFound );
		RemoveName( szSourcePath );

		while(sourceExists)
		{
         // We have to expand the LongName with the proper directory,
         // so plug it onto the szSourcePath and expand ...
			AddName( szSourcePath, srceFound.name );
         SmartExpandPath( FULLPATH_LONGNAME, szSourcePath, szSourcePath );

         // ... then copy it and append it to the szDestinationPath also.
			CopyNamePart( szName, szSourcePath );
			AddName( szDestinationPath,szName );

			status = MoveIt( szSourcePath, szDestinationPath );
			if( status != ERR_NONE )
			{
            status = ERR_MS_1;
				return status;
			}

			RemoveName( szSourcePath );
			RemoveName( szDestinationPath );

			sourceExists = SmartFindNext( &srceFound );
		}
		sourceExists = TRUE;

	} while( SourcePathExists( argc, argv ));

	return ERR_NONE;
}
