//
//   $Id: POWER.H 1.4 2002/10/16 21:57:04Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*
 * Datalight Power Management Software
 * Header file for POWER
 *
 * Copyright(c) 1993-2002 Datalight Inc.
 * All Rights Reserved
 * Datalight Confidential and Propritary
 */


#include <dos.h>

#define NULL 0

#define SIMULATE	FALSE		/*  Simulate APM int 15  */
#define DEBUG		FALSE  	/* Build a Debugging Version of Power */
#define DISPLAY	FALSE		/* Display diagnostic messages */
#define SOUND		FALSE		/* Audible diagnostics */


#define PERCENT_DIFF 7		/* Percentage diff in idle int 16 calls */

/*
 * LastUsed field contains the Ticks value at the time the devices was last used
 * InactiveCount field contains the # of ticks required to put the device into
 *		StandbyState
 *	Status = Ready, Standby, Syspend, Off
 */
struct device {
	unsigned LastUsed;
	unsigned InactiveCount;
	unsigned Status;
};


/*
 * Power Structure that a pointer is retrived from that the TSR/Device Driver
 * version of power.  For Datalight's ROM-DOS power this structure is
 * in POWER.C starting with the variable TotalTicks!
 */
struct PowerStat {
	long TotalTicks;
	long IdleTicks;
	long OtherValues[4];
};


// Interrupt Definitions
//
#define KEYSTROKE_INTERRUPT	0x09
#define SERIAL_INTERRUPT_1		0x0B
#define SERIAL_INTERRUPT_2		0x0C
#define SCREEN_INTERRUPT		0x10
#define DISK_INTERRUPT			0x13
#define SERIAL_INTERRUPT		0x14
#define KEYBOARD_INTERRUPT 	0x16
#define PRINTER_INTERRUPT		0x17
#define TIMER_INTERRUPT 		0x1C
#define DOS_IDLE_INTERRUPT 	0x28
#define DOS_REDIR_INTERRUPT 	0x2F


//
// The following definitions are from Rev 1.1 of the Advanced Power
// Management (APM) specification from Intel/Microsoft (1993).
//

// Power State Definitions
// Standby mode will typically respond to IRQs. Suspend mode will typically
// require an NMI (from an RTC, for example) to wake the system.
//
#define READY			0			// full speed operation
#define STANDBY		1			// reduced power ops, resumes operation quickly
#define SUSPEND		2			// CPU stopped, eventually resumes operation.
#define OFF				3			// Pwr supply off. Resets when power comes on.


// Device Definitions
// E000h-EFFFh define OEM devices, all other values reserved.
//
#define APM_DEV_SYSTEM		   0x0000
#define APM_DEV_ALL			   0x0001
#define APM_DEV_DISPLAY		   0x0100
#define APM_DEV_DISK			   0x0200
#define APM_DEV_PRINTER		   0x0300
#define APM_DEV_SERIAL		   0x0400
#define APM_DEV_NETWORK		   0x0500
#define APM_DEV_PCMCIA		   0x0600

// The APM standard devices numbers are defined above and, in most
// cases, are used in the same order to index into an array of the
// above device strucs. The keyboard is an exception.
//
#define DEV_SYSTEM		   0x0000
#define DEV_DISPLAY		   0x0001
#define DEV_DISK			   0x0002
#define DEV_PRINTER		   0x0003
#define DEV_SERIAL		   0x0004
#define DEV_KEYBOARD		   0x0005

// APM BIOS Functions offered via int 15h
//
#define APM_INSTALLATION_CHECK		0x5300	// ask "anybody home?"
#define APM_REAL_MODE_CONNECT       0x5301	// say "talk to me in real mode"
#define APM_16BIT_PM_CONNECT        0x5302	// protected mode interface
#define APM_32BIT_PM_CONNECT        0x5303	// protected mode interface
#define APM_INTERFACE_DISCONNECT    0x5304	// say "go away"
#define APM_CPU_IDLE                0x5305	// say "the system is idle"
#define APM_CPU_BUSY                0x5306	// say "give me max power"
#define APM_SET_POWER_STATE         0x5307	// say "..."
#define APM_SET_PWR_MGMT_FUNC      	0x5308	// cx=0: disable, cx=1: enable
#define APM_RESTORE_DEFAULTS        0x5309	// to power on values
#define APM_GET_PWR_STATUS          0x530A	// of AC line and battery
#define APM_GET_PWR_MGMT_EVENT      0x530B	// ask "what should I do next"
#define APM_GET_PWR_MGMT_STATE		0x530C	// ask "what is on-line now"
#define APM_SET_DEV_PWR_MGMT_FUNC   0x530D	// cx=0: disable, cx=1: enable
#define APM_GET_DRIVER_VERSION		0x530E


// APM BIOS events return by calls to int 15h fn 530Bh
// In addition to the events below, a number of reserverd events are
// defined. Each of these events must be broadcast to APM aware applications
// but POWER has no other responsibility for handling them. These events are
//
// 000Ch - 00FFh	Reserved system events
// 0100h - 01FFh  Reserved device events
// 0200h - 02FFh  Reserved OEM events
//
// All other event numbers (0300h - FFFFh) are reserved.
//
#define SYSTEM_STANDBY_REQUEST 		0x0001
#define SYSTEM_SUSPEND_REQUEST		0x0002
#define END_NORMAL_SUSPEND_NOTICE   0x0003
#define END_CRITICAL_SUSPEND_NOTICE 0x0004
#define BATTERY_LOW_NOTICE          0x0005
#define POWER_STATE_CHANGE_NOTICE   0x0006
#define UPDATE_TIME_NOTICE          0x0007
#define CRITICAL_SUSPEND_NOTICE     0x0008
#define USER_STANDBY_REQUEST        0x0009
#define USER_SUSPEND_REQUEST        0x000A
#define END_NORMAL_STANDBY_NOTICE   0x000B


// APM Get status On line response
//
#define APM_STAT_ON_LINE 0x100

extern char *USAGE;
extern char *SIGNON;
extern char  *messages[];
extern unsigned TotalTicks;					/* BIOS Tick Count */
extern unsigned Pwr_Mgt_Ticks;
extern unsigned Power_Down_Display;
extern unsigned Power_Down_Disk;
extern unsigned Power_Down_Printer;
extern unsigned Power_Down_Serial;

extern void far interrupt (*old_int_9)();		/* Orginal keystroke interrupt */
extern void far interrupt int_9();				/* Power interrupt 9 handler */

extern void far interrupt (*old_int_B)();		/* Orginal Serial Interrupt 1  */
extern void far interrupt int_B();				/* Power interrupt B handler */

extern void far interrupt (*old_int_C)();		/* Orginal Serail Interrupt 2 */
extern void far interrupt int_C();				/* Power interrupt C handler */

extern void far interrupt (*old_int_10)();	/* Orginal BIOS Screen interrupt */
extern void far interrupt int_10();				/* Power interrupt 10 handler */

extern void far interrupt (*old_int_13)();	/* Orginal BIOS Disk interrupt */
extern void far interrupt int_13();				/* Power interrupt 13 handler */

extern void far interrupt (*old_int_14)();	/* Orginal BIOS Serial interrupt */
extern void far interrupt int_14();				/* Power interrupt 14 handler */

extern void far interrupt (*old_int_16)();	/* Orginal BIOS Keyboard interrupt */
extern void far interrupt int_16();				/* Power interrupt 16 handler */

extern void far interrupt (*old_int_17)();	/* Orginal BIOS Printer interrupt */
extern void far interrupt int_17();				/* Power interrupt 17 handler */

extern void far interrupt (*old_int_1C)();	/* Orginal BIOS Tick interrupt */
extern void far interrupt int_1C();				/* Power interrupt 1C handler */

extern void far interrupt (*old_int_28)();	/* Orginal DOS Idle interrupt */
extern void far interrupt int_28();				/* Power interrupt 28 handler */

extern void far interrupt (*old_int_2F)();	/* Orginal DOS Redir interrupt */
extern void far interrupt int_2F();				/* Power interrupt 28 handler */


