//
//   $Id: REM8250.C 1.4 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*******************************************************************

	REM8250.C  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	Low level interrupt driven serial communications for the
	8250 UART.  Uses no DOS calls nor BIOS calls.  Assumes a
	completely "dumb" machine.

	When porting to another UART, the routine names MUST
	remain compatible.  Call the new file REMxxxx.C.

********************************************************************/
#define  __FILENUM__	8250

int handshaking = 0;

#include <stdio.h>
#include <dos.h>					/* enable and disable */

#include "dlbiosio.h"

#include "remote.h"
#include "rem8250.h"
#include "rem8250.pt"

#if IBM
void comm_save_state(void);
void comm_restore_state(void);
#endif

unsigned short loop_count;	      /* # of loops before timeout */
volatile int RcvdPacket;			/* A packet is here */
comport cp;			     				/* Definition of communications port */
uchar *pacbuf = cp.buffer;	      /* a pointer to the packet data */
extern int pacsize;
extern int ISR_timeout;

void setportvect(void far *routine);
void interrupt far comm_char_isr(void);

/*
	ISR (Interrupt Service Routine)

	Interrupt driven PACKET receive.  This routine reads the
	whole packet in one interrupt call.  It also processes
	the packet before returning.
*/
void interrupt comm_packet_isr(void)
{
	uchar *p = cp.buffer;
	uchar count;
	unsigned short timeout = 0xffff;

	if(!RcvdPacket)
	{
		/* Read the count */
		inportb(LINESTATUS);					    /* Clear pending status   */
		*p++ = count = inportb(DATAREG);  /* Read 1st part of Count */

		outportb(INTENABLE,0x00);       /* disable receive interrupt */

		if(count)
		   comm_putc(0);						   /* acknowledge */

		/* Read the rest of the Packet */
		while(count--)
		{

			/* loop untill char received or time out */
			while((!(inportb(LINESTATUS) & LS_DRDY)) && timeout)
			       timeout--;		       // dec counter

			if (!timeout)
			{
		       ISR_timeout = 1;
		       goto time_out;
			}

			*p++ = inportb(DATAREG);
		}

		RcvdPacket = 1;						 /* Let the outside world know */

time_out:

		outportb(INTENABLE,0x01);   /* re-enable receive interrupt */
	}

	if( cp.uiIrqNumber > 7 )
	{
		outportb(PIC_EOIPORT2,PIC_EOIVAL); // non-specific EOI
	}

	outportb(PIC_EOIPORT,PIC_EOIVAL); // non-specific EOI
}


#define NOT_IN_ASM 0
#if     NOT_IN_ASM

#define OVER_RUN   0
#if     OVER_RUN
int overrun = 0;
#endif
/*
	comm_char_isr

	This ISR is an interrupt driven CHARACTER receive routine.
	It stuffs the character received into a ring buffer.
*/
void interrupt far comm_char_isr(void)
{ unsigned char stat;

   /* get the data */
again:
   cp.buffer[cp.head++] = inportb(DATAREG);

	/* check for wrap around */
   if(cp.head >= cp.bufsize)
   	cp.head= 0;

   /* check for next character */
   stat = inportb(LINESTATUS);

#if OVER_RUN
	if(stat & LS_OVERRUN)
		++overrun;
#endif
	if(stat & LS_DRDY)
		goto again;

	/* send an end-of-interrupt to the 8259 */
	if( cp.uiIrqNumber > 7 )
	{
		outportb(PIC_EOIPORT2,PIC_EOIVAL); // non-specific EOI
	}

	outportb(PIC_EOIPORT,PIC_EOIVAL); // non-specific EOI
}
#endif


/*
	comm_open

	Enable the 8250/16x50 UART for interrupt driven receive and
	polling transmit.  Initialized a structure defining the
	port,   and set up the interrupt vector for receive.

	port	 (0,1,2,3)
	irq    (3,4,5,6,7,8,9,10,11,12,13,14,15)
	baud	 (300,1200,2400..38400,57600,115)
	parms  (N81, E71, see rem8250.h)
*/
bool comm_open(unsigned port,unsigned uiIrqNumber, unsigned baud,unsigned parms)
{
	unsigned char ucMaskPic1;
	unsigned char ucMaskPic2;
	unsigned char val;
	bool bSuccessful = TRUE;

	/* Check for defaults */
	if((port>3) || (port < 0)) port = 0;		/* use 0 if invalid port */
	if( ( uiIrqNumber < 3 ) || ( uiIrqNumber > 15 ) )
	{
		if( ( port == 0 ) || ( port == 2 ) )
		{
			uiIrqNumber = 4;
		}
		else
		{
			uiIrqNumber = 3;
		}
	}
	if(!parms) parms = N81;    					/* use N81 if invalid parms */
	if(!baud)   baud = 9600;

	/* set up the port address values */
	cp.base = GetCommPortBaseAddress(port);
	if(cp.base == 0)
		return FALSE;

	cp.uiIrqNumber = uiIrqNumber;
	cp.port = port;
	cp.head = cp.tail = 0;
	cp.bufsize = BUFSIZE;

#if IBM
    /* save state of serial port */
    comm_save_state();
#endif


	if( uiIrqNumber < 8 )
	{
		ucMaskPic1 = (1 << (unsigned char)uiIrqNumber);
		ucMaskPic2 = 0;
	}
	else
	{
		ucMaskPic1 = 2;
		ucMaskPic2 = (1 << ((unsigned char)uiIrqNumber - 8));
	}

	baud= makebaud(baud);

	if( uiIrqNumber < 8 )
	{
		/* disable interupts for this port (just in case) */
		val=inportb(PIC_MASKPORT);
		val|= ucMaskPic1;
		outportb(PIC_MASKPORT,val);
	}

	/* disable interupts for this port (just in case) */
	val=inportb(PIC_MASKPORT2);
	val|= ucMaskPic2;
	outportb(PIC_MASKPORT2,val);


	/* initialize the serial port parameters */
	outportb(LINECONTROL,0x80);		     /* set high bit, so we can set baud */
	outportb(BAUDLOW,baud & 0xFF);	  /* baud rate */
	outportb(BAUDHIGH,baud >> 8);
	outportb(LINECONTROL,parms);		    /* stop/data/parity bits */

	outportb(FIFOCONTROL,1);		// try to enable 16650 FIFO's
	if(inportb(INTID)&0xC0)			// if they don't come on its an 8250
	{
		// Set 16550 FIFO trigger depth to 14 and clear all FIFOs
		//
		outportb(FIFOCONTROL,0xC7);
	}

	// clear all pending data and interrupts
	//
	inportb(MODEMSTATUS);    					//	clear an delta MSR ints
	val=inportb(INTID);	   			// clear any Tx ints
	while(inportb(LINESTATUS) & LS_DRDY)	// clear any LSR ints
	{
		inportb(DATAREG);							// drain all Rx data
	}


	/* enable interrupts */
	comm_to_char();	      /* point vector to character ISR */

	val=inportb(PIC_MASKPORT);
	val&= ~ucMaskPic1;				/* set bit to 0 to enable */
	outportb(PIC_MASKPORT,val);			/* enable interrupt mask */

	val=inportb(PIC_MASKPORT2);
	val&= ~ucMaskPic2;				/* set bit to 0 to enable */
	outportb(PIC_MASKPORT2,val);			/* enable interrupt mask */

	/* enable interrupts (bit3), and pins high */
	cp.modemcontrol = MC_INT | MC_RTS | MC_DTR;
	outportb(MODEMCONTROL,cp.modemcontrol);
	outportb(INTENABLE,0x01);    /* enable receive interrupt */

	return bSuccessful;
}

/*
 *  Disable interrupts for this serial port.
 */
void comm_close(void)
{
	unsigned char ucMaskPic1;
	unsigned char ucMaskPic2;
	unsigned char val;

	if( cp.uiIrqNumber < 8 )
	{
		ucMaskPic1 = (1 << (unsigned char)cp.uiIrqNumber);
		ucMaskPic2 = 0;

		/* disable interrupt mask for this channel (8259) */
		val = inportb(PIC_MASKPORT);
		val |= ucMaskPic1;				/* set bit to disable */
		outportb(PIC_MASKPORT,val);
	}
	else
	{
		ucMaskPic1 = 2;
		ucMaskPic2 = (1 << ((unsigned char)cp.uiIrqNumber - 8));

		/* disable interrupt mask for this channel (8259) */
		val = inportb(PIC_MASKPORT2);
		val |= ucMaskPic2;				/* set bit to disable */
		outportb(PIC_MASKPORT2,val);
	}

	/* disable interrupt line as well */

#if IBM
	/* restore original state */
	comm_restore_state();
#endif
}


/*
	makebaud

	From an unsigned value, make a baud rate divisor
	for the 8250 on a PC.
*/
static int makebaud(unsigned baud)
{
	switch(baud)
	{
		case   300: baud = B300;   break;
		case  1200: baud = B1200;  break;
		case  2400: baud = B2400;  break;
		case  4800: baud = B4800;  break;
		case  9600: baud = B9600;  break;
		case 19200: baud = B19200; pacsize = REM_PACSIZE / baud; break;
		case 38400L: baud = B38400; pacsize = REM_PACSIZE / baud; break;
		case 57600L: baud = B57600; pacsize = REM_PACSIZE / baud; break;
		case 115:   baud = B115200;pacsize = REM_PACSIZE / baud; break;
		default:		baud = B9600;  break;
	}

	return baud;
}


/*
	comm_to_packet

	Used only in COMM.  If COMM is being used as both the CONSOLE
	and the remote disk handler, we may possible need to switch
	between "character" adn "packet" modes.  This allows for the
	feature.  See also comm_to_char().
*/
void comm_to_packet(void)
{
	RcvdPacket = 0;
	setportvect(comm_packet_isr);
}

/*
	comm_to_char

	Used only in COMM.  If COMM is being used as both the CONSOLE
	and the remote disk handler, we may possible need to switch
	between "character" adn "packet" modes.  This allows for the
	feature.  See also comm_to_packet().
*/
void comm_to_char(void)
{
	cp.head= cp.tail= 0;
	setportvect(comm_char_isr);
}

/*
	setportvect

	Set a vector to point to a routine.
*/
void setportvect(void far *routine)
{
	unsigned uiInterruptNumber;

	disable();
	uiInterruptNumber =
			(cp.uiIrqNumber<8) ? (cp.uiIrqNumber+8) : (0x70+(cp.uiIrqNumber-8));
	*(long far *)MK_FP(0,(uiInterruptNumber * 4)) = (long)routine;
	enable();
}

/*
	getportvect

	Set a vector to point to a routine.
*/
void far *getportvect(int port)
{
	unsigned uiInterruptNumber;
   long ptr;

	if((port>3) || (port < 0))
      	port = 0;	       /* use 0 if invalid port */

	/* if no base yet, figure it out */
	if(!cp.base)
	{
		cp.base= GetCommPortBaseAddress(port);
		if(cp.base == 0)
			return (void far *)(0L);
	}
	disable();
	uiInterruptNumber =
			(cp.uiIrqNumber<8) ? (cp.uiIrqNumber+8) : (0x70+(cp.uiIrqNumber-8));
	ptr = *(long far *)MK_FP(0,( uiInterruptNumber *4));
	enable();

	return (void far *)(ptr);
}


/*
	comm_poll

	Grab a character out of the ring buffer.

	Returns
		-1	      no character available
		0-n     the character, pulling it out of the ring buffer.
*/
int comm_poll(void)
{
#if USE_CTS
	static int wait = 0;
	int diff, head;
#endif
	uchar c;

	 if(handshaking)
	 {
#if USE_CTS
	// if head-tail is starting to fill up!
	head = cp.head;
	if(head < cp.tail)
		head += BUFSIZE;
	if(!wait && head - cp.tail > BUFSIZE/2)
	{
		// hold off until all characters have been received
		cp.modemcontrol &= ~MC_RTS;
		outportb(MODEMCONTROL,cp.modemcontrol);

		wait = 1;
//	      printf("wait, head %d, tail %d\n",head,cp.tail);	// debugging!
	}
#endif

#if OVER_RUN
	if(overrun)
	{
		printf("Overrun %d!\n",overrun);
		overrun = 0;
	}
#endif
	 }

	if(cp.head != cp.tail)
	{
		c = cp.buffer[cp.tail++];
		if(cp.tail >= cp.bufsize)
			cp.tail=0;
		return c;
	}


#if USE_CTS
	// no characters? enable RTS
	if(wait)
	{
		cp.modemcontrol |= MC_RTS;
		outportb(MODEMCONTROL,cp.modemcontrol);
		wait=0;
		// printf(".");	 // debugging!
	}
#endif

	return -1;
}

/*
	comm_putc

	Send a character out the UART.  This is a "polled"
	transmit, not interrupt driven.
*/
void comm_putc(uchar c)
{
#if USE_CTS
	static int wait;
#endif
	uchar val;

	 if(handshaking)
	 {
#if USE_CTS
	// wait until other end says it's ok to send
	wait = 0;
	while(1)
	{
		val = inportb(MODEMSTATUS);
		if(val & MS_CTS)
			break;
		if(!wait)
			printf("wait ");
		wait = 1;
	}

	if(wait)
		printf(".");
#endif
	 }

	while(TRUE)
	{
		val=inportb(LINESTATUS);
		if(val & LS_TRDY)			       /* trans buffer is empty */
		{
			outportb(DATAREG,c);
			break;
		}
	}
}


/*****    Debugging    *****/

#if 0

#define nybblechar(cc) ((cc) < 10 ? (cc)+'0' : (cc)-10+'A')

/*
	debugging
*/
biosdump(unsigned char *buff,int len)
{
	int i;
	for(i=0; i<len; ++i)
	{
		/* print CR/LF if needed */
		if((i%16) == 0)
		{
			_AX = 0xE0D;
			_BX = 7;
			geninterrupt(0x10);
			_AX = 0xE0A;
			geninterrupt(0x10);
		}

		_AX = 0xE00 | nybblechar(*buff>>4);
		geninterrupt(0x10);

		_AX = 0xE00 | nybblechar(*buff&0xF);
		_BX = 7;
		geninterrupt(0x10);

		_AX = 0xE20;
		geninterrupt(0x10);

		++buff;
	}
}

/*
	debugging routine
*/
prchar(uchar c)
{
	_AX = 0x0E00 | c;
	_BX = 7;
	geninterrupt(0x10);
}
#endif



/****    For use in REMDISK to restore the port    ****/

#if IBM

long comm_vec;				/* the previous vector */
int  comm_int_on1;		/* were interrupts on for this vector? */
int  comm_int_on2;		/* were interrupts on for this vector? */

/*
	Remarks Save the "state" of the interrupt (before we
				trash it).
*/
void comm_save_state(void)
{
	unsigned char ucMaskPic1;
	unsigned char ucMaskPic2;
	unsigned char val;
	unsigned uiInterruptNumber;

	uiInterruptNumber =
			(cp.uiIrqNumber<8) ? (cp.uiIrqNumber+8) : (0x70+(cp.uiIrqNumber-8));
	comm_vec = *(long far *)MK_FP(0,uiInterruptNumber*4);

	comm_int_on1 = 0;
	comm_int_on2 = 0;

	if( cp.uiIrqNumber < 8 )
	{
		ucMaskPic1 = (1 << (unsigned char)cp.uiIrqNumber);
		ucMaskPic2 = 0;

		/* if the bit in the mask is off, the serial interrupt was enabled */
		val=inportb(PIC_MASKPORT);
		comm_int_on1 = !(val & ucMaskPic1);
	}
	else
	{
		ucMaskPic1 = 2;
		ucMaskPic2 = (1 << ((unsigned char)cp.uiIrqNumber - 8));

		/* if the bit in the mask is off, the serial interrupt was enabled */
		val=inportb(PIC_MASKPORT2);
		comm_int_on2 = !(val & ucMaskPic2);
	}



}


/*
	Remarks Restore the original state of the interrupt.
				Enables the CPU interrupt call.
*/
void comm_restore_state(void)
{
	unsigned char ucMaskPic1;
	unsigned char ucMaskPic2;
	unsigned char val;

	/* set the interrupt vector to point to the original ISR */
	if(comm_vec)
	{
		setportvect((void far *)comm_vec);

		if( cp.uiIrqNumber < 8 )
		{
			ucMaskPic1 = (1 << (unsigned char)cp.uiIrqNumber);
			ucMaskPic2 = 0;

			/* set the interrupt mask state */
			val=inportb(PIC_MASKPORT);
			if(comm_int_on1)
				val &= ~ucMaskPic1;				/* set bit to 0 to enable */
			else
				val |= ucMaskPic1;				/* set bit to 1 to disable */
			outportb(PIC_MASKPORT,val);		/* enable/disable interrupt mask */
		}
		else
		{
			ucMaskPic1 = 2;
			ucMaskPic2 = (1 << ((unsigned char)cp.uiIrqNumber - 8));

			/* set the interrupt mask state */
			val=inportb(PIC_MASKPORT2);
			if(comm_int_on2)
				val &= ~ucMaskPic2;				/* set bit to 0 to enable */
			else
				val |= ucMaskPic2;				/* set bit to 1 to disable */
			outportb(PIC_MASKPORT2,val);		/* enable/disable interrupt mask */
		}

	}

}
#endif

/*
	modemstat_DSR

	Returns
		0			DSR not set
		other		DSR set
*/
int modemstat_DSR( void )
{
	if(handshaking)
		return( inportb(MODEMSTATUS) & MS_DSR );
	else
		return MS_DSR;
}
