//
//   $Id: REMDISK2.C 1.5 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**********************************************************************

	REMDISK2.C  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This file is used on the TARGET, in the Remote Disk Driver

	Used by the Remote Disk, this file contains the
	routines to convert the device call into packet
	calls.

	Command Line:
	DEVICE = REMDISK.SYS /B9600 /COM1

	/Bxxxx[+]       set baud rate, '+' indicates packet mode (default 19200)
	/COMn           use comport n (default COM1)
	/BIOS           communicate through the BIOS (INT 10H, INT 16H)
	/Txxx           set timeout to xxx seconds

	History:

	01-30-1999 bt  Renamed all the bpb_def to BPB so that it
						would be using the BPB structure in dostruct.h
	01-30-1999 bt  Renamed rdsk_bpb to std_bpb for memdisk support
	01-30-1999 bt  Added fields for volume label, serial number, and fat
						type string to support memdisk code.

***********************************************************************/
#include <mem.h>
#include <dos.h>
#include <stdlib.h>

#include "datalght.h"

#include "remote.h"
#include "remdisk.h"

#include "remote.pt"
#include "remdisk2.pt"
#include "remdisk3.pt"
#include "print.pt"

#define DEBUG 0

#define isdigit(cc) ((cc) >= '0' && (cc) <= '9')

void bios_puts(char *s);


int opt_unload;
int opt_showhelp;

/* NOTE: these must all be initialized so they get put in _DATA segment */
int      rdsk_debug       = 0;
uchar    rdsk_chg         = 0;			/* change-of-disk status */
uchar    rdsk_big         = 0;			/* is this a BIG (32M+) disk or small */
uchar    rdsk_err         = 0;			/* what type of error occurred? */
uchar    rdsk_drive       = 0;
uchar    rdsk_mode        = DEFMODE;	/* remote xfer mode: CHAR or PACKET */
int      rdsk_port        = 0;			/* what port to use */
unsigned	rdsk_uiIrqNumber = 0;			/* what IRQ  to use */
#if IBM
extern unsigned rdsk_baud;                      // in REMTEXT.C
#else
unsigned        rdsk_baud = DEFBAUD;            /* what baud rate */
#endif

// rem8250.c -- TRUE if RTS-CTS and DTR-DSR are used for flow control
//
extern int handshaking;

// 01-30-1999 bt  Renamed rdsk_bpb to std_bpb for memdisk support
//
struct BPB std_bpb =
{
512,0x10,1,2,0x200,0,0xF8,0xFC,0x3F,0x10,0x3F,0xFBFC1L
};

//  01-30-1999 bt  Added fields for volume label, serial number, and fat
//  type string to support memdisk code.
//
char std_volume[] = "REMDISK    ";
unsigned long std_serial = 0x12345678L;
char std_fattype[] = "FAT12    ";

// 01-30-1999 bt  Renamed rdsk_bpb to std_bpb for memdisk support
//
struct BPB *std_bpbptr = &std_bpb;

uchar           rdsk_buff[540];         /* just over 512 bytes */

/*
        rem_packet

        Send a packet then receive and process it's response.
        The brunt of the work lies here.  This is called from
        the device driver.

        Parameters
                cmd             the command (B,C,D,R,W,X)
                sector  the sector to Read/Write
                dataoff the offset of buffer to Read/Write into/from
                dataseg the segment of buffer to Read/Write into/from
                num             the number of sectors to Read/Write

        Legal Commands

        B                               get bpb from Remote Server
        C                               detect media change
        D                               done with this device call
        R[sect]         read a sector from Server
        W[sect]         write sector to Server, num is # of packets
        X                               tell Server to exit
*/

void rem_packet(char cmd, unsigned long sector, unsigned dataoff, unsigned dataseg, int num)
{
        unsigned bps = std_bpb.sec_size;
        int len;
        uchar err=0;
        uchar done=0;

        /* no errors yet */
        rdsk_err = 0;

        if(rdsk_debug)
                bios_puts("P");

        /* give user the chance to re-establish communications */
        if(enter_rdsk())
        {
                if(rdsk_debug)
                        bios_puts("1");
                goto not_ready;
        }

        /* tell SERVER end that we are about to send packet(s) */
        remote_putc((uchar)(PREPACKET));

        if(rdsk_debug)
                bios_puts("2");

        if(remote_getpacket(rdsk_buff) < 0)
        {

                if(rdsk_debug)
                        bios_puts("3");
                goto not_ready;
        }

        /* read/write packets 1-n sectors */
        if(cmd == 'R')
        {
                while(num-- && !err)
                {
                        if(rdsk_debug)
                              printf("%ld ",sector);

//                      printf("%ld ",sector);

                        err = rem_rd_sect(sector,dataoff,dataseg);
                        dataoff += bps;
                        ++sector;

                }
        }

        else if(cmd == 'W')
        {
                while(num-- && !err)
                {
                        err = rem_wr_sect(sector,dataoff,dataseg);
                        dataoff += bps;
                        ++sector;

                }
        }

        /* simple packet: build and send it */
        else
        {
                if(rdsk_debug)
                        bios_puts("4");

                PAC_CMD = cmd;

                // send the command
sendit:
                len = remote_sendpacket(rdsk_buff,1);
                if(len < 0)
                {
                        if(rdsk_debug)
                                bios_puts("4");
                        goto not_ready;
                }

                len = remote_getpacket(rdsk_buff);
                if(len < 0)
                {
                        if(rdsk_debug)
                                bios_puts("5");
not_ready:
                        rdsk_err = DEVERR_NOTRDY;
                        return;
                }
        }

        if(rdsk_debug)
                bios_puts("6");

        /* we got a packet: act on the command */
        switch(PAC_CMD)
        {
                case 'c':
                        rdsk_chg = PAC_CHGCODE;         /* has the disk changed (yes/no/maybe) */
                        break;
                case 'b':
                        memmove(&std_bpb,PAC_BPB,sizeof(struct BPB));
                        rdsk_big = (std_bpb.num_sec == 0);
                        break;
        }

        if(rdsk_debug)
                bios_puts("7");

        /* tell HOST we are done with this device call */
        if(!done)
        {
                PAC_CMD = 'D';
                done = 1;

                if(rdsk_debug)
                        bios_puts("8");

                goto sendit;
        }

        if(rdsk_debug)
                bios_puts("9");

        /* give user chance to restore "other" communications */
        leave_rdsk();

        if(rdsk_debug)
        {
                if(rdsk_err)
                        bios_puts("e");
                else
                        bios_puts("ok");
        }

        return;
}


/*
        Read a sector from the Remote HOST Disk.
*/
int rem_rd_sect(unsigned long sector,unsigned dataoff,unsigned dataseg)
{
//        int i,num;
        int len;

        /* Tell COMM to read us a sector */
        PAC_CMD = 'R';
        PAC_SECTOR = sector;

        /* find out how many */
        len = remote_sendpacket(rdsk_buff,5);
        if(len > 0)
        {
                len = remote_getpacket(rdsk_buff);
        }

        if(len <= 0)
        {
errout:
                rdsk_err = DEVERR_NOTRDY;
                return -1;
        }

        --len;
		  _fmemcpy( MK_FP(dataseg,dataoff),
		  		MK_FP(_DS,(unsigned)(rdsk_buff)+1 ), len );
        return 0;
}


/*
        Write a sector to the Remote HOST Disk.
*/
int rem_wr_sect(unsigned long sector,unsigned dataoff,unsigned dataseg)
{
		  int len;
        unsigned bps = std_bpb.sec_size;

        PAC_CMD    = 'W';
        PAC_SECTOR = sector;
		  _fmemcpy( MK_FP( _DS, (unsigned)(rdsk_buff)+5 ),
				MK_FP( dataseg, dataoff ), bps );

        /* send the packet and wait for acknowledge */
        len = remote_sendpacket(rdsk_buff,5 + bps);
        if(len > 0)
                len = remote_getpacket(rdsk_buff);
        if(len < 0)
        {
errout:
                rdsk_err = DEVERR_NOTRDY;
                return -1;
        }

        return 0;
}

/*
        []-------------------------------[]
         |   INITIALIZATION FUNCTIONS    |
        []-------------------------------[]
*/

void far_strupr(char far * cmd)
{
        while(*cmd >= ' ' || *cmd=='\t')
        {
                if(*cmd>='a' && *cmd<='z')
                        *cmd = *cmd-'a'+'A';
                ++cmd;
        }
}

/*
        parse_cmdline

        Parses the device command line for Baud and Port
        parameters.  Used only upon initialization.
*/
void parse_cmdline(uchar drive, char far * cmd)
{
	char * cmd_copy = "     ";
	unsigned i;

	// set up defaults
	rdsk_drive = drive;
   rdsk_port = 0;

	if(rdsk_baud==115 || rdsk_baud >= 19200)
		rdsk_mode = M_PACKET;
	else       /* char mode at 9600 or less */
		rdsk_mode = M_CHAR;

	opt_unload = 0;

	far_strupr(cmd);
	while(*cmd >= ' ' || *cmd=='\t')
	{
		if(*cmd == '?')
		{
			++opt_showhelp;
			return;
		}

		if(*cmd == '/' || *cmd == '-')
		{
			++cmd;

			/* unload previous version */
			if(*cmd == 'U')
			{
				++opt_unload;
				return;
			}

			if(*cmd == '?')
			{
				++opt_showhelp;
				return;
			}

			/* debugging */
			if(*cmd == 'D')
				++rdsk_debug;

			/* check for Baud and Packet Mode */
			else if(*cmd == 'B')
			{ long temp;

				++cmd;
				rdsk_baud = 0;
				temp = 0L;
				while(isdigit(*cmd))
				{
					temp *= 10L;
					temp += (long)(*cmd - '0');
					++cmd;
				}
				if(temp == 115200L)
					temp = 115L;
				rdsk_baud = (unsigned)(temp);

				/* MUST be packet mode past 19200 baud */
				if(rdsk_baud==115 || rdsk_baud >= 19200)
					rdsk_mode = M_PACKET;
				else       /* char mode at 9600 or less */
					rdsk_mode = M_CHAR;

				if((*cmd == 'k') || (*cmd == 'K'))
					++cmd;

				if(*cmd == '+')
				{
					rdsk_mode = M_PACKET;
					++cmd;
				}
				else if (*cmd == '-')
				{
					rdsk_mode = M_CHAR;
					++cmd;
				}
			}
			// Check for TIMEOUT setting
			else if(*cmd == 'T')
			{
				// Convert the timeout from characters to a
				// number, moving the pointer as we go.
				++cmd;
				REM_TIMEOUT = 0;
				while(isdigit(*cmd))
				{
					REM_TIMEOUT *= 10;
					REM_TIMEOUT += (*cmd - '0');
					++cmd;
				}

				// If the data is going to overflow,
				// set it back to a reasonable value.
				if( REM_TIMEOUT > 0xFFFF / 18 )
				REM_TIMEOUT = 0xFFFF / 18;

				// Convert seconds to ticks
				REM_TIMEOUT *= 18;

				// Don't allow user to specify less timeout than
				// the default.
				if( REM_TIMEOUT < REMDISK_TIMEOUT )
					REM_TIMEOUT = REMDISK_TIMEOUT;
			}
			// check for HARDWARE HANDSHAKING option
			//
			else if(*cmd == 'H')
			{
				handshaking = 1;
			}
			/* check for COM port */
			else if(strn_fcmp(cmd,"COM",3) == 0)
			{
				rdsk_port = *(cmd+3) - '1';
				cmd += 4;
			}

			//  check for irq number option
			//
			else if(strn_fcmp(cmd,"IRQ",3) == 0)
			{
				for( i=0; i<5; i++)
				{
					cmd_copy[i]=cmd[i+3];
				}
				rdsk_uiIrqNumber = (unsigned)(atoi( ( cmd_copy ) ) );
				cmd += 4;
			}
		}

		else
			++cmd;
	}
}

/*
        strn_fcmp

        Compare a portion of a far string with a near string.
        Used only upon initialization.
*/
int strn_fcmp(char far *s1,char *s2,int len)
{
        while(len)
        {
                if(*s1 != *s2)
                        return *s1 - *s2;
                --len;
        }
        return 0;
}

/*
        Convert an unsigned to a string.
        Return the number of digits in the string.
*/
int utoa(unsigned val,char *s)
{
        unsigned div=10000;
        unsigned char found=0, startdigit=5;

        while(div >= 1)
        {
                *s = (val / div);
                if(*s || found)
                {
                        *s++ += '0';
                        found=1;
                }
                else
                        --startdigit;
                val %= div;
                div /= 10;
        }
        *s = 0;
        return startdigit;
}


/*
        bios_puts

        print a string via BIOS calls
*/
void bios_puts(char *s)
{
        while(*s)
        {
                _AL = *s;
                _BX = 7;
                _AH = 0xE;
                geninterrupt(0x10);
                ++s;
        }
}


#if DEBUG
/*
        bios_fputs

        print a string via BIOS calls
*/
void bios_fputs(char far * s)
{
        while(*s >= ' ')
        {
                _AL = *s;
                _BX = 7;
                _AH = 0xE;
                geninterrupt(0x10);
                ++s;
        }
}
#endif
