//
//   $Id: REMOTE.C 1.4 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/**********************************************************

	REMOTE.C  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This file contains the routines for transferring
	packets back and forth.  It can do it in one of two
	ways: M_CHAR, and M_PACKET.

***********************************************************/
#include <mem.h>
#include <dos.h>                /* for enable(), disable() */

#include "datalght.h"

#include "remote.h"
#include "remote.pt"
#include "print.pt"

int ISR_timeout;
int rem_mode;                           /* the current communications mode */
char *rem_devname;              /* device mode only, the name set by user */
void far *rem_old_isr;  /* the old ISR before remdisk got control */

int pacsize;

void send_nul_packet(void);

extern int rdsk_debug;

void fail(int n)
{
        if(rdsk_debug)
                printf("failure at %d\n",n);
}

/*
        remote_open

        Open a device/port for remote communications.  This
        is called from COMM, RDEB and REMDISK.SYS.

        mode            (character, packet, device, bios)
        port            0=COM1, 1=COM2
        baud            300,1200..38400,57600,115
        parms           N81, E71, etc...

        returns
                -1              failed to open specified device/port
                0               worked
*/
int remote_open(int mode,unsigned port,unsigned uiIrqNumber,
		unsigned baud,unsigned parms)
{
        int err=0;

        rem_mode = mode;
        switch(rem_mode)
        {
                case M_CHAR:
                case M_PACKET:
                        if(!comm_open(port,uiIrqNumber,baud,parms))
                        	err = -1;

                        if(rem_mode == M_PACKET)
                                comm_to_packet();
                        break;
                default:
                        err = -1;
        }
        return err;
}

/*
        remote_close

        Close the device/port used for remote communiations.
*/
void remote_close(void)
{
        switch(rem_mode)
        {
                case M_CHAR             : /* fall thru */
                case M_PACKET   : comm_close(); break;
        };
}

/*
        remote_getpacket

        Get a packet.  We don't care how it arrives, just return
        either timeout or got it.  It can come in via the BIOS,
        a DEVICE or the COMM port in PACKET or CHAR mode.

        Returns
                -1              timeout error
                0-n     length of packet (0-32K)

*/
int remote_getpacket(uchar *buff)
{
        extern uchar *pacbuf;

        int i,len,c,sublen;
        uchar * orgbuf = buff;
        char cksum = 0;                /* init value */
// static  long test = 0L;

        if (!modemstat_DSR())
               return -1;

        /* get a packet of any given length */
        switch(rem_mode)
        {
                /* Character style input */
                case M_CHAR:

                        /* get the 2 byte length of the packet */
                        len = c = remote_tgetc(comm_poll);
                        if(c < 0)
                        {
                                fail(1);
                                goto errout;
                        }
                        c = remote_tgetc(comm_poll);
                        if(c < 0)
                        {
                                fail(2);
                                goto errout;
                        }
                        len |= (c<<8);

                        if(rdsk_debug > 1)
                                printf("receiving %d bytes  ",len);

                        /* get the data of the packet */
                        for(i=0; i<len; ++i)
                        {
                                *buff++ = c = remote_tgetc(comm_poll);
                                if(c < 0)
                                {
                                        fail(3);
                                        goto errout;
                                }
                        }

                        for (i=0;i<len;i++)     /* calculate cksum */
                        {
                             cksum += orgbuf[i];
                        }


//                      test++;
//                      if ((test %10) == 9)
//                              cksum = 1;

                        if (cksum && rdsk_debug)
                               printf("cksum failed!");

//                      if (cksum)
//                             printf("cksum failed! %d",cksum);

                        if (cksum)              /* if failed cksum */
                                goto errout;

                        break;

                /*
                        Packet style input

                        When a sub-packet is received of less than REM_PACKET in
                        size, we are done receiving sub-packets... we have the
                        whole packet.
                */
                case M_PACKET:

                        len = 0;
                        while(1)
                        {
                                sublen = receive_subpacket(buff);

                                /* check for error */
                                if(sublen < 0)
                                {
                                        fail(4);
                                        goto errout;
                                }

                                /* increment the buffer */
                                buff += sublen;
                                len += sublen;

                                /* check if we are done or must do it again */
                                if(sublen < pacsize)
                                        break;

                                send_nul_packet();
                        }


                        for (i=0;i<len;i++)     /* calculate cksum */
                        {
                             cksum += orgbuf[i];
                        }

//                      test++;
//                      if ((test %10) == 9)
//                              cksum = 1;

                        if (cksum && rdsk_debug)
                               printf("cksum failed!");

//                      if (cksum)
//                             printf("cksum failed! %d",cksum);

                        if (cksum)              /* if failed cksum */
                                goto errout;


                        break;
                }

        if (len)
          len--;       /* adj for cksum */

        return len;

errout:
        return -1;
}


/*
        remote_sendpacket

        Send a packet.  We don't care how it goes, just return
        either timeout or sent it.

        Returns
                -1              timeout error
                0-n     length of packet (0-32K)
*/
int remote_sendpacket(uchar *buff,unsigned len)
{
        int i,sublen,oldlen;
        char cksum = 0;                /* init value */


        if (!modemstat_DSR())
                return -1;

        /* send a packet */
        switch(rem_mode)
        {

                /* character style output */
                case M_CHAR:

                        /* calculate cksum value */
                        for(i=0; i<len; i++)
                        {
                              cksum += buff[i]; /* calculate cksum */
                        }

                        /*********************************************/
                        /* invert and add cksum at end               */
                        /* by inverting the cksum, the receiver only */
                        /* has to look for an ending value of 0 to   */
                        /* know that the cksum passed.               */
                        /*********************************************/
                        buff[i] = -cksum;

                        len++;                    /* account for the cksum */

                        if(rdsk_debug > 1)
                                printf("sending len %d  ",len);

                        comm_putc((uchar)len);
                        comm_putc((uchar)(len>>8));
                        for(i=0; i<len; ++i)
                                comm_putc(*buff++);

                        len--;                   /* remove cksum on return */
                        break;

                /*
                        packet style output

                        Send the packet in smaller chunks because we don't
                        want interrupts off.
                */
                case M_PACKET:

                        oldlen = len;           /* save length, to be restored later */

                        /* calculate cksum value */
                        for(i=0; i<len; i++)
                        {
                              cksum += buff[i]; /* calculate cksum */
                        }
                        /*********************************************/
                        /* invert and add cksum at end               */
                        /* by inverting the cksum, the receiver only */
                        /* has to look for an ending value of 0 to   */
                        /* know that the cksum passed.               */
                        /*********************************************/
                        buff[i] = -cksum;

                        len++;                    /* account for the cksum */

                        while(1)
                        {
                                /* figure out a sublength of this packet */
                                if(len == pacsize)
                                        sublen = len-2;
                                else if(len > pacsize)
                                        sublen = pacsize;
                                else
                                        sublen = len;

                                /* send the subpacket */
                                i = send_subpacket(buff,sublen);

                                /* check for failure */
                                if(i < 0)
                                {
                                        fail(10);
                                        return -1;
                                }

                                /* check if we are done sending */
                                if(sublen < pacsize)
                                        break;

                                /* update buffer pointer */
                                buff += sublen;
                                len  -= sublen;

                                /* we must receive an ACKnowlege */
                                i = receive_subpacket(buff);

                                if(i < 0)
                                {
                                        fail(11);
                                        return -1;
                                }
                        }


                        len = oldlen;           /* restore the length */
                        break;
        }

        return len;
}


/*
        []----------------------------[]
         |              Lower Level Routines            |
         |      to be used internally only |
        []----------------------------[]
*/

/*
        Returns count or -1 on error.
*/
int receive_subpacket(uchar *buff)
{
        int status;
        long ticks,cl_start();
        uchar count,*p;
        extern uchar *pacbuf;

        /* is a packet here already? */
        disable();
        status = RcvdPacket;                                    /* has a packet come in? */
        RcvdPacket = 0;
        enable();

        /* we must begin timeout counter */
        if(!status)
                ticks = cl_start();

        while(!status)
        {
                disable();
                status = RcvdPacket;                                    /* has a packet come in? */
                RcvdPacket = 0;
                enable();

                if(status)
                        break;

                disable();
                if(ISR_timeout)
                {
                       ISR_timeout = 0;
                       enable();
                       return -1;
                }
                enable();


                if(cl_timer(ticks) >= REM_TIMEOUT)      /* have we timed out? */
                        return -1;
        }

        /* move the data from the ISR buffer to user buffer */
        p = pacbuf;
        count = *p;
        memmove(buff,p+1,count);

        return count;
}

/*
        Returns count or -1 on error.
*/
int send_subpacket(uchar *buff,int count)
{
        int i;

        /* send packet length and get acknowledge */
        comm_putc(count);

        /* get a nul packet */
        if(receive_subpacket(pacbuf) < 0)
                return -1;

        /* send the packet */
        for(i=0; i<count; ++i)
                comm_putc(*buff++);

        return count;
}

void send_nul_packet(void)
{
        comm_putc(0);
}


/*
        remote_putc

        Put a character over the serial port or device.
*/
void remote_putc(uchar c)
{
        comm_putc(c);
}


/*
        remote_tgetc

        Get a character.  Check for timeout.  Used by M_BIOS
        and M_CHAR modes of communication.

        Returns
                -1              no character, a timeout has occured.
                0-n     the character.
*/
int remote_tgetc(vfuncptr charin)
{
        int c;
        long lcounter;

        c = (*charin)();
        if(c < 0)
        {
                lcounter = cl_start();
                c = (*charin)();
        }
        while(c < 0)
        {

                disable();
                if(ISR_timeout)
                {
                       ISR_timeout = 0;
                       enable();
                       return -1;
                }
                enable();

                if(cl_timer(lcounter) >= REM_TIMEOUT)
                        return -1;
                c = (*charin)();
        }
        return c;
}

/*
        wait_for_receive

        For M_PACKET mode only, wait until a packet has come
        in or we have waited long enough (a timeout).
*/
int wait_for_receive(void)
{
        int gotapacket;
        long ticks;

        /* check for a packet immediately */
        disable();
        gotapacket = RcvdPacket;
        RcvdPacket = 0;
        enable();

        if(!gotapacket)
                ticks = cl_start();

        while(!gotapacket) {

                disable();
                gotapacket = RcvdPacket;
                RcvdPacket = 0;
                enable();

                if(gotapacket)
                        break;

                disable();
                if(ISR_timeout)
                {
                       ISR_timeout = 0;
                       enable();
                       return -1;
                }
                enable();

                if(cl_timer(ticks) >= REM_TIMEOUT)      /* have we timed out? */
                        return -1;
                }

        return 1;
}



/*
        []-------------------------[]
         |    THE CLOCK PACKAGE    |
        []-------------------------[]
*/


#ifdef __TURBOC__
/* The BIOS ticks 18.2 times a second */
long biostime();
#endif

/*
        cl_lwait

        Wait the specified number of ticks.  Each tick occurs
        18.2 times a second.
*/
void cl_lwait(long ticks)
{
        long lcounter;
        lcounter=cl_start();
        while(cl_timer(lcounter)<ticks)
                ;       /* wait */
}

/*
        cl_start

        Return the number of BIOS ticks.  Used to start a timer.
*/
long cl_start(void)
{
        long time;
        _AH = 0;
        geninterrupt(0x1A);
        time = (long)MK_FP(_CX,_DX);
        return time;
}

/*
        cl_timer

        Return the number of ticks since the counter was set
*/
long cl_timer(long tick_cnt)
{
        long now_ticks = cl_start();
        if(now_ticks < tick_cnt)
                now_ticks += 1572480L;  /* midnight wrap around */
        return now_ticks - tick_cnt;
}


