//
//   $Id: REMOTE.H 1.5 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	REMOTE.H  created 02/05/90  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

	This file should be included when using the remote_xx()
	routines for packet transfers.

**********************************************************/


/* Some generic types */
typedef unsigned long ulong;
typedef int (*vfuncptr)(void);
typedef int (*ifuncptr)(int);

/*
   BIOS PARAMETER BLOCK (BPB)
   Description: This structure is stored on FAT disks in
   sector 0, at offset 11H.

   Reference: Adv MS-DOS, p 269  -or-  DOS Ref, p 222
*/
struct BPB
{
   unsigned   sec_size;      /* 00 */
   uchar      sec_per_clus;  /* 02 */
   unsigned   res_sec;       /* 03 */
   uchar      num_fats;      /* 05 */
   unsigned   dir_ent;       /* 06 */
   unsigned   num_sec;       /* 08 */
   uchar      media;         /* 0A */
   unsigned   fat_size;      /* 0B */
   unsigned   sec_per_trak;  /* 0D  Floppy/Harddisk extensions */
   unsigned   sides;         /* 0F       ignored by DOS        */
   long       hidden;        /* 11    for device use only      */
   long       lnumsec;       /* 15 */
};

// The Drive Parameter Block (DPB) structure
// See MSDOS Prog Ref pg 262
//
struct	DPB {

	unsigned char dpbDrive, dpbUnit;
	unsigned dpbSectorSize;
	unsigned char dpbClusterMask, dpbClusterShift;
	unsigned dpbFirstFAT;
	unsigned char dpbFATCount;
	unsigned dpbRootEntries, dpbFirstSector, dpbMaxCluster,
				dpbFatSize, dpbDirSector;
	void far *dpbDriverAddr;
	unsigned char dpMedia, dpbFirstAccess;
	struct DPB far *dpbNextDPB;
	unsigned dpbNextFree, dbpbFreeCnt;
};
typedef struct DPB DPB;
typedef struct DPB far * LPDPB;

/* Modes for remote_open() */
#define M_CHAR          0
#define M_PACKET        1

#if IBM
//#define DEFMODE M_CHAR
//#define DEFBAUD 19200  // IBM - RAY removed becuase defined in REMTEXT.C
#define DEFMODE M_PACKET
#else
#define DEFMODE M_PACKET
#define DEFBAUD 115
#endif


#define PREPACKET       0xDA                    /* the prepacket character */
#define ERR                     -1
#define ERR_NONE        0
#define ERR_DONE        1
#define ERR_TERM        2


/* Timeout period for getting/sending packet */
// IBM - RAY shorten timeout periods.
//#define REMDISK_TIMEOUT (7*18)  // in ticks (18.2/sec)
#define REMDISK_TIMEOUT (3*18)  // in ticks (18.2/sec)
//#define REMSERV_TIMEOUT (18)       // slave doesn't need much of a timeout
#define REMSERV_TIMEOUT (18*2)       // slave doesn't need much of a timeout
extern unsigned REM_TIMEOUT;

// IBM-RAY - changed max size because it causes interrupts to be disabled too
// long.  Interrupts should only be disabled for a MAX of 5ms.
//#define REM_PACSIZE     132                     // maximum size of a packet
#define REM_PACSIZE       64                  // maximum size of a packet
extern int rem_mode;


/* Prototypes from REM8250.C */
extern volatile int RcvdPacket;
extern uchar * pacbuf;
void interrupt far comm_char_isr(void);
void interrupt far comm_packet_isr(void);
bool comm_open(unsigned port,unsigned uiIrqNumber, unsigned baud,unsigned parms);
void comm_close(void);
void comm_to_packet(void);
void comm_to_char(void);
int comm_poll(void);
void comm_putc(uchar c);
int modemstat_DSR( void );

/* Prototypes from REMOTE.C */
int remote_open(int mode,unsigned port,unsigned uiIrqNumber,
		unsigned baud,unsigned parms);
void remote_close(void);
int remote_getpacket(uchar *buff);
int remote_sendpacket(uchar *buff,unsigned len);
int remote_tgetc(vfuncptr charin);
void cl_lwait(long ticks);
long cl_start(void);
long cl_timer(long tick_cnt);
