//
//   $Id: REMSERV.C 1.5 2002/10/16 21:58:28Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*****************************************************

	REMSERV.C  created 03/09/93  by Drew Gislason

	Copyright (c) 1990-2002 Datalight, Inc.
	All Rights Reserved.
	Datalight Confidential and Proprietary.

        REMSERV.EXE  d: [/Bxxxx][+|-] [/COMn]

        /Bxxxx  Set the baud rate.  Choose from 300, 1200, 2400...19200,
                38400, 57600, 115.  A '+' after the baud specifies 'packet'
                style transfers. By default, packet transfers will be used
					 for all baud rates higher than 9600. You may over-ride this
					 default default behaviour by appending a '-' to the baud
					 rate and so force polling I/O. Polling I/O must be used
					 inside an MS Windows 95 DOS Box.

        /COMn   Set the communications port.  Choose 1 or 2.

        /Txxx   Set timeout to xxx seconds

        Example:   REMSERV C: /B9600- /COM2

        The default is /B115K+ /COM1

        The workhorse routine is:
        remdisk_emulate(drive,mode,port,baud);

******************************************************/

#define 	NOT_FAT16_ERR	4

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "dosver.h"
#include "datalght.h"
#include "fatworld.h"

#include "remote.h"
#include "remserv.pt"
#include "print.pt"
#include "remhost.pt"

extern char USAGE1[];
extern char USAGE[];
extern char using_str[];
extern char mouse_range_str[];
extern char exit_str[];
extern char terminate_str[];
extern char invalid_drv_str[];
extern char bad_opt_str[];
extern char irq_range_str[];
extern char packet_err_str[];
extern char invalid_cmd_str[];
extern char not_FAT16_str[];

/* Default Settings */
int drive               = 1;                            /* 1=A, 2=B, 3=C, etc... */
int mode                        = DEFMODE;              /* choose M_CHAR or M_PACKET */
unsigned port   = 0;                            /* 0=COM1, 1=COM2 */
unsigned uiIrqNumber = 0;
#if IBM
extern unsigned baud;                   // IBM baud rate can be found in REMTEXT.C
#else
unsigned baud   = DEFBAUD;              /* 300,1200...115K */
#endif

#if IBM
int mouse_startx;
int mouse_stopx;
int mouse_starty;
int mouse_stopy;
#endif

unsigned REM_TIMEOUT = REMSERV_TIMEOUT;

bool fIsFat32World;

extern int rdsk_debug;
extern ifuncptr  rdsk_charfunc;

// rem8250.c -- TRUE if RTS-CTS and DTR-DSR are used for flow control
//
extern int handshaking;

typedef struct tagBOOTSECTOR
{
	char jmpInstruction[3];
	char nameOfOEM[8];
	struct BPB bsBPB;
} BOOTSECTORTYPE;

int _absread(int drive, int num, long start, void * buff);

// This routine determines if the specified drive is a FAT32 drive
// (1=A, 2=B,...).
//
// returns: -1:	if absread failed
//				 0:	if drive is FAT16 or other.
//				 1:   if drive is FAT32
//
int IsDriveFAT32(int drive)
{ char bootSectorBuffer[512];
  BOOTSECTORTYPE *pBootSector;
  unsigned int sectorsInThisFAT;

	// Read the boot sector off of the user specified disk.
	// Cast it.
	//
   if(_absread(drive-1,1,0,bootSectorBuffer) != 0)
		return -1;
	pBootSector = (BOOTSECTORTYPE*)(&bootSectorBuffer[0]);

	// Assume that we have a FAT drive of some kind. If the BPB
	// lists the number of sectors in the FAT as 0 then we have
	// a FAT32 drive.
	//
	sectorsInThisFAT = pBootSector->bsBPB.fat_size;
	return ((sectorsInThisFAT == 0) & 1);
}


int printchar(int c)
{
	if(c>0 && c<'~')
		printf("%c",c);
	else
		printf("<%02X>",c);
	return 0;
}

char * skipwhite(char *s)
{
	while(*s==' ' || *s=='\t')
		++s;
	return s;
}

char * skipdigits(char *s)
{
	while(isdigit(*s))
		++s;
	return s;
}

void signon(void)
{
	char * signon = "REMSERV %s %s\n%s\n";

	if( fIsFat32World )
	{
		printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
}

int do_args(int argc, char *argv[])
{
	int i;
	char *p;
	char c;
	char *p2;

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	/* get flags /COMn /IRQn /B# /D /X#,# /Y#,# */
	for(i=1; i<argc; ++i)
	{
		p = argv[i];
		while(p)
		{
			p = strchr(p, '/');
			if(p)
			{
				c = toupper(p[1]);
				if(c == 'D')
				{
					++rdsk_debug;
					p += 2;
				}

#if IBM
				// /X#,# sets mouse click range (for X axis)
				else if(c == 'X')
				{
					// check options
					p2 = strchr(p, ',');
					if(!p2)
						goto badopt;
					if(!isdigit(p[2]) || !isdigit(p2[1]))
						goto badopt;

					mouse_startx = atoi(p+2);
					mouse_stopx  = atoi(p2+1);
					++p;
				}

				// /Y#,# sets mouse click range (for Y axis)
				else if(c == 'Y')
				{
					// check options
					p2 = strchr(p, ',');
					if(!p2)
						goto badopt;
					if(!isdigit(p[2]) || !isdigit(p2[1]))
						goto badopt;

					mouse_starty = atoi(p+2);
					mouse_stopy  = atoi(p2+1);
					++p;
				}
#endif
				else if(c == 'B')
				{ long temp;

					/* get baud rate */
					if(!isdigit(p[2])) goto badopt;
					temp = atol(p+2);
					if(temp == 115200L)
					temp = 115L;

	  				// by default, use packet mode at
					// 19200 baud or greater and
					// char mode at 9600 baud or less
					//
					baud = (unsigned)(temp);
					if(baud==115 || baud >= 19200)
			 			mode = M_PACKET;
					else
			 			mode = M_CHAR;

					// but allow user to over-ride the
					// default mode setting by appending
					// a '+' or a '-' to the baud rate
					// spec. Using polling (instead of
					// interrupt driven receive) allows
					// us to work in a Win95 DOS box.
					//
					p = skipdigits(p+2);
					if(*p == '+')
					{
					 	mode = M_PACKET;
						++p;
					}
					else if(*p == '-')
					{
						mode = M_CHAR;
						++p;
					}
				}
				else if(c == 'T')
				{
					// get timeout setting
					if(!isdigit(p[2])) goto badopt;
					REM_TIMEOUT = atoi(p+2);
					// Don't allow it to overflow
					if( REM_TIMEOUT > 0xFFFF / 18 )
					REM_TIMEOUT = 0xFFFF / 18;
					// Convert seconds to ticks
					REM_TIMEOUT *= 18;
					// Don't allow user to specify less timeout than
					// the default.
					if( REM_TIMEOUT < REMSERV_TIMEOUT )
					REM_TIMEOUT = REMSERV_TIMEOUT;
					// Skip to the next argument
					p = skipdigits(p+2);
				}
				// use HARDWARE HANDSHAKING option
				//
				else if(c == 'H')
				{
					handshaking = 1;
					p += 2;
				}
				else if(c == 'C')
				{
					c = p[4];
					if(toupper(p[2])!='O' || toupper(p[3])!='M' || !isdigit(c))
						goto badopt;
					if(c<'1' || c>'4')
						goto badopt;
					port = c-'1';
					p += 4;
				}
				else if(c == 'I')
				{
					uiIrqNumber = (unsigned)(atoi(p+4));
					if( ( uiIrqNumber < 3 ) || ( uiIrqNumber > 15 ) )
						goto badopt;
					p += 4;
				}
				else if(c == 'S')
				{
					printf_off();
					p += 2;
				}
				else
				{
badopt:
					signon();
					if( c != '?' )
					{
						printf(bad_opt_str,p);
					}
					printf(USAGE);
					return (3);
				}
			}
		}
	}
	return 0;
}

/*
		  Since we are using the "quick" startup code, there will be
		  only 2 arguments: argv[1] will be all one string with
		  spaces " A: /COM1", etc...
*/
int main(int argc,char *argv[])
{
	char * cmdline;

	// make sure all arguments are proper
	if(do_args(argc,argv))
		return 3;


	//  set default irq for the specified port if none was specified
	//
	if( uiIrqNumber == 0)
	{
		if( ( port == 0 ) || ( port == 2 ) )
		{
			uiIrqNumber = 4;
		}
		else
		{
			uiIrqNumber = 3;
		}
	}

	//  range check the irq
	//
	if( ( uiIrqNumber < 3 ) || ( uiIrqNumber > 15 ) )
	{
		printf(irq_range_str);
		return 2;
	}


	signon();

	/* user must specify drive */
	cmdline = skipwhite(argv[1]);
	if(argc < 2 || *cmdline == '?' || (*cmdline=='/' && cmdline[1]=='?'))
	{
		printf(USAGE);
		return 1;
	}

	drive = toupper(*cmdline);
	asm {
		push	ds
		mov 	ah, 32h
		mov	dx, drive
		sub 	dx, 'A'
		inc	dx
		int 	21h
		pop	ds
		cmp	al, 0FFh				;// drive valid?
		je	   invalid_drive  	;// no
	}

	if(drive < 'A' || drive > 'Z')
	{
invalid_drive:
		printf(invalid_drv_str,drive,USAGE1);
		return 2;
	}
	drive = drive - 'A' + 1;	/* 1=A, 2=B, etc... */

	if(IsDriveFAT32(drive))
	{
		printf(not_FAT16_str,drive+'@');
		return NOT_FAT16_ERR;
	}

	/* Turn ON debugging so we can see the disk operating */
	rdsk_charfunc = printchar;	      /* output for non-packet bytes */

	/* display current baud/port/drive */
	printf(using_str, port+1,uiIrqNumber,baud,(baud==115 ? "K" : ""),
		(mode == M_PACKET ? "+" : "-"),drive+'@',REM_TIMEOUT/18);
#if IBM
	if(mouse_stopx || mouse_stopy)
		printf(mouse_range_str,mouse_startx,mouse_stopx,mouse_starty,mouse_stopy);
#endif
	printf(exit_str);

	/*
	  this routine does all the work, making a drive on THIS
	  system available to the REMOTE system
	*/
	remdisk_emulate(drive,mode,port,uiIrqNumber,baud);

	printf(terminate_str);

	return 0;
}

