//
//   $Id: SORT.C 1.6 2002/10/16 22:01:36Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	SORT.C

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	History:
	02-23-2000  js Added LFN Compatability

**********************************************************/

#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dllfn.h"
#include "fatworld.h"
#include "dosver.h"
#include "stdc.h"


#define SORT 1
#include "dosmsg.c"

#define MAX_STRINGS 2000
#define MAX_BUF_SIZE 32000

char signon[] = "SORT %s %s\n%s\n\n";

int iInputHandle=STDIN, iOutputHandle=STDOUT;

char buffer[MAX_BUF_SIZE];
char * strings[MAX_STRINGS];
unsigned buf_size = MAX_BUF_SIZE;
unsigned offset=0;
unsigned count=0, i;
unsigned str_index = 0;				/* Default value */
int iret;

int reverse = FALSE;

int found;

char * FileName;

bool fIsFat32World = FALSE;
bool fHideSignon = FALSE;

char *read_text_line(int iFileHandle, char *szReadBuf, int iMaxRead);
void do_args(int argc, char *argv[]);
void do_help(void);

void main(int argc, char *argv[])
{
	unsigned uDevAttributes;
	char *ret, *tmp;
	int iActionTaken;
	unsigned uByteCount;
	bool fSuccessful;


	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}

	do_args(argc, argv);

	dlGetDeviceAttributes( STDIN, &uDevAttributes );
	if( (uDevAttributes & 0x80) && (argc > 1) )
	{
		fSuccessful = SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0, FILE_OPEN,
			 FileName, &iInputHandle, &iActionTaken);

		if(!fSuccessful)
		{
			fprintf(STDERR, cant_open_str, FileName);
			exit(1);
		}
	}

	do
	{
		// Read in a line
		uByteCount = buf_size - offset;
		ret=read_text_line(iInputHandle, buffer+offset, uByteCount);

		// If this is the first line and there's no line to get, quit
		if( !uByteCount && !count )
		{
			fprintf( STDERR, nothing_to_sort_str );
			exit( 1 );
		}

		strings[count] = buffer + offset;
		offset += strlen(buffer+offset) + 1;
		if(MAX_STRINGS == count)
		{
			printf(out_of_mem);
			exit(1);
		}
		++count;
	} while(ret != NULL);

	if(count)				/* Don't decrement 0 */
		--count;
	else
		count = 1;

	do
	{					/* Sort File */
		found = FALSE;

		if( count == 0 )
		{
			break;
		}

		for(i=0; i < count-1; ++i)
		{
			iret = strcmpi(strings[i] + str_index, strings[i+1] + str_index);
			if(reverse)
			{
				if(iret < 0)
				{
					tmp = strings[i];
					strings[i] = strings[i+1];
					strings[i+1] = tmp;
					found = TRUE;
				}
			}
			else
			{
				if(iret > 0)
				{
					tmp = strings[i];
					strings[i] = strings[i+1];
					strings[i+1] = tmp;
					found = TRUE;
				}
			}
		}
	} while(found);

	/* Output File */
	for(i=0; i<count; ++i)
		printf("%s", strings[i]);
	exit(0);
}

/*
	$DOCPROC: read_text_line

	char * read_text_line(int iFileHandle, char * szReadBuf, int iMaxRead);

	Reads one line (CR/LF terminated) from the file.  Meant to be a
	substitute for fgets().

	On Entry:
	iFileHandle = DOS file handle of file to read from
	szReadBuff = Buffer into which text is read
	iMaxRead = Maximum number of bytes that should be read

*/
char * read_text_line(int iFileHandle, char * szReadBuf, int iMaxRead)
{
	int iReadCount=0;
	unsigned uBytes;
	bool fSuccessful;

	while(iReadCount < iMaxRead)
	{
		uBytes=1;
		fSuccessful = dlReadFile(iFileHandle, &uBytes, szReadBuf);
		if(!fSuccessful)
			return(NULL);

		if(uBytes == 0 && iReadCount == 0)
			return(NULL);

		// If the end of the file was reached, terminate string and return.
		if(uBytes == 0)
		{
			*szReadBuf = NULL;
			return(szReadBuf-iReadCount);
		}

		// If last 2 bytes read where a newline code(CRLF), terminate string
		// and return.
		if(*szReadBuf == '\n' && *(szReadBuf-1) == '\r')
		{
			*(szReadBuf-1) = '\n';
			*szReadBuf = NULL;
			return(szReadBuf-iReadCount);
		}

		iReadCount++;
		szReadBuf++;
	}
	return(0);
}


void do_args(int argc, char *argv[])
{
	int i;

	for(i=1; i<argc; ++i)
	{
		if(argv[i][0] == '/')
		{
			switch(argv[i][1])
			{
				case '?':
				case 'h':
				case 'H':
					do_help();

				case 'r':
				case 'R':
					reverse = TRUE;
					break;

				case '[':
					fHideSignon = TRUE;
					break;

				case '+':
					str_index = atoi(argv[i] + 2);
					if(offset > 80)
						offset = 0;
					break;

				default:
					printf(bad_opt, argv[i]);
					do_help();
			}
		}
		else
			FileName = argv[i];
	}
}

void DisplaySignOnMessage( void )
{
	if( !fHideSignon )
	{
		if( fIsFat32World )
		{
			printf(signon,SHORTVERSION,BUILDVERSION,COPYRIGHT);
		}
		else
		{
			printf(signon,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
		}
	}
}

void do_help(void)
{
	DisplaySignOnMessage();
	printf(USAGE);
	exit(1);
}
