;
;   $Id: BOOT.INC 1.4 2002/10/16 22:01:36Z peterb Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
;
;
;   BOOT.INC
;
;   Copyright (c) 1989-2002 Datalight, Inc.
;   All Rights Reserved.
;   Datalight Confidential and Proprietary
;
;   History
;
;   01/27/99 bt  Original
;
;   These are constants for the BOOT.ASM file which is the boot sector
;   code for FAT12 and FAT16 drives.
;
;


;
;
; [ CONSTANTS ]
;

;
; This is the offset that the boot sector gets loaded.  It can not be
; changed.  It is defined as a constant for clearity.
;

PROGRAM_OFFSET			EQU	7c00h

;
; This is the stack size in bytes
;

STACK_SIZE			EQU	200h

;
; This is the segment which ROMDOS will be loaded (offset 0000h)
;

DEF_LOADADDR			EQU	1000H	; 1000:0 (used to be 70:0)

;
; This is the offset that the partition table will be loaded.  This
; gets loaded and used only once.
;

PARTITION_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET + 0200h)

;
; This is the offset that the root directory will be loaded.  This
; gets loaded and used only once.  The entire root directory gets
; loaded starting at this address.
;

ROOT_DIR_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET + 0200h)

;
; This is the offset that the other half of the boot sector gets
; loaded.  Do not change this offset without changing all the
; relative offsets in the code.
;

OTHER_HALF_OF_LOADER_OFFSET	EQU	(PROGRAM_OFFSET + 0200h)

;
; This is the offset that the FAT gets loaded.  We only load one
; sector of the FAT at any given time.  This sector size memory
; block gets reloaded with different sectors of the FAT.
;

FAT_BUFFER_OFFSET		EQU	(PROGRAM_OFFSET - STACK_SIZE - 0600h)

;
; This is the offset of the INT-1Eh vector.  The segment is zero.  This
; is the location of the DBT (Disk Base Table)
;

DBT_VECTOR			EQU	(1eh * 4)

;
; This is the size of the DBT (Disk Base Table) in bytes
;

DBT_SIZE			EQU	0bh

;
; This is the offset for the new location of the DBT (Disk Base Table).
; The segment is 0000h.
;

NEW_DBT_OFFSET			EQU	1000h

;
; This is the offset of the INT-21h vector.  The segment is zero.  This
; is the location of the DBT (Disk Base Table)
;

INT21_VECTOR			EQU	(21h * 4)

;
; These are constants used for clearity when assigning the variable.
; The are used to for 12-bit and 16-bit fats.
;

FAT_12_BIT_END			EQU	0ff8h
FAT_16_BIT_END			EQU	0fff8h
FAT_12_BIT_ENTRIES		EQU	0400h ; (index into 3 sectors)
FAT_16_BIT_ENTRIES		EQU	0300h ; (index into 3 sectors)

;
; We are using these constants because using the offset compiler
; command caused the compiler to interprete the offsets as a word
; and not as a byte.  This caused some indexed variable instructions
; to be longer than they needed to be.
;

;
; These are offsets into the BPB (BIOS Parameter Block) at the begining
; of the boot sector.
;

OFFSET_SPC			EQU	0dh
OFFSET_RES			EQU	0eh
OFFSET_NOF			EQU	10h
OFFSET_RDE			EQU	11h
OFFSET_TS1			EQU	13h
OFFSET_SPF			EQU	16h
OFFSET_SPT			EQU	18h
OFFSET_HEAD			EQU	1ah
OFFSET_HID			EQU	1ch
OFFSET_DRV			EQU	24h
OFFSET_FID			EQU	36h

;
; These are offsets into a partition table entry
;

PARTITION_ID			EQU	4
PARTITION_HID			EQU	8

;
; This is a variable which contains the partition ID if the
; partition is a FAT32 partition which uses the extended BIOS
; function call interface for reads and writes.  Thus, it
; will be either 90h or 0ch.
;

OFFSET_PARTITION_ID		EQU	2

;
; This location is used as a variable to indicate whether the rest of
; boot sector code has been loaded.
;

OFFSET_BS_PART_LOADED		EQU	0

;
; This is the offset for the ROMDOS load address variable
;

;OFFSET_LOAD_ADDR		EQU	3eh

;
; We a using some code space for variable locations.  The
; code space was already used and will not be used again
; by the time we start writing variables there.
;

;
; This is the variable location to save the DBT address
;

;OLD_DBT_ADDRESS			EQU	OFFSET_LOAD_ADDR + 4
OLD_DBT_ADDRESS				EQU	3eh

;
; This is the physical sector number for the start of the data
; area on the disk. It is set at run time.
;

DATA_START			EQU	OLD_DBT_ADDRESS + 4

;
; This is the physical sector number for the start of the FAT
; on the disk. It is set at run time.
;

FAT_START			EQU	DATA_START + 4
