//
//   $Id: TRANSFER.C 1.7 2002/10/16 22:04:26Z peterb Exp $
//
//   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
//
//   Copyright (c) 1993 - 2002 Datalight, Inc.
//   All Rights Reserved
//
//   Datalight, Inc. is a Washington corporation with an address at
//   21520 30TH DR SE MS110 Bothell, WA, 98021
//   Tel: 800.221.6630  Fax: 425.951.8091
//   Http://www.datalight.com
//
//   This software, including without limitation all source code and
//   documentation, is the confidential, trade secret property of
//   Datalight, Inc., and is protected under the copyright laws of
//   the United States and other jurisdictions. Portions of the
//   software may also be subject to one or more the following US
//   patents: US#5860082, US#6260156.
//
//   In addition to civil penalties for infringement of copyright
//   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
//   penalties for violation of (a) the restrictions on circumvention
//   of copyright protection systems found in 17 U.S.C. 1201 and
//   (b) the protections for the integrity of copyright management
//   information found in 17 U.S.C. 1202.
//
//   U.S. Government Restricted Rights. Use, duplication,
//   reproduction, or transfer of this commercial product and
//   accompanying documentation is restricted in accordance with
//   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
//
//   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
//   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
//   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
//   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
//   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
//   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
//   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
//   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
//   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
//   THIS SOFTWARE IN ERROR.
//
/*********************************************************

	TRANSFER.C  created 02/15/89  by Drew Gislason

   Copyright (c) 1989-2002 Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

	Tabs are 3 spaces per column.

	Usage = TRANSFER {options} filename

	Options
		/S      Send a file
		/R      Receive a file
		/B#     Set baud rate (default 9600)
      /BC	  Use BIOS console I/O (vs DOS for CTTY)
		/COM#   Use comport for transfer (default CONsole)
		/IRQ#   Specify different IRQ
		/Q      Quiet mode

	Example

		C>TRANSFER /r /com1 A:\TEST.EXE

	Transfer uses the XMODEM protocol to send/receive files.
	Sets ERRORLVL = 0 on success.


	History:
	03-24-2000	JS
		- Added LFN compatibility

	01-30-1999  BT
		- Added dostruct.h include

	08-19-1998  BT
		- Added support for IRQ's 5-15

**********************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <mem.h>
#include <dos.h>
#include <dir.h>

#define TRANSFER 1


#include "dosver.h"
#include "dosmsg.c"
#include "datalght.h"
#include "dostruct.h"
#include "dosdef.h"
#include "dldosio.h"
#include "dl8250.h"
#include "fatworld.h"
#include "dllfn.h"
#include "dldosio.h"
#include "tranprot.pt"


#define SEC 				18		/* 18 clock ticks per second (approx) */
#define PACKET_RETRIES  10		// limit once a xfer is initiated
#define STARTUP_RETRIES 40		// limit before 1st. packet begins

#define timeout_getc() tran_tgetc(1*SEC)

#define ERR -1

/* some XMODEM protocol bytes */
#define SOH 0x01	/* ctrl('A') */
#define STX 0x02	/* ctrl('B') */
#define EOT 0x04	/* ctrl('D') */
#define ACK 0x06	/* ctrl('F') */
#define NAK 0x15	/* ctrl('U') */
#define CAN 0x18	/* ctrl('X') */


int bios_kbd(void);		// returns char from kbd buffer

uchar buffer[1032];			// the file transfer I/O buffer

#define PACKET_SIZE		132
#define STDIN				0
#define STDOUT				1
#define BINARY_MODE		0x20 	// raw vs. cooked mode (for CON I/O)
#define NOT_THE_CONSOLE (~0x03)

// We want to have at least 7 seconds worth of receiver buffer space
// (allows for two back-back garbage collections on a Card Trick
// drive) before we start overwriting data in the receive buffer.
//
// 132bytes * 10bits/byte / 9600baud = .1375s (about. yeah, yeah: bps != baud)
// 10packets * .1375sec/packet + (10-1)timeouts * 1sec/timeout = 10.375s
// Note: timeout is dominant term, T(115200) = 1.0s+0.0114583333s
//
#define RX_BUFF_SIZE	(10*PACKET_SIZE) //10packets * 132bytes/packet = 1320bytes
uchar rxdBuffer[RX_BUFF_SIZE];	// the interrupt driven COM receive buffer



/* options */
unsigned baudrate = 9600;
int comN;
int iIrqNumber=0;
int sendfile;
int usecom;
int useBIOS;
int quiet;


int comm_opened;			/* is the comm port currently opened? */
int got_opt_send_rec;	/* user MUST specify either /R or /S */

bool fIsFat32World;

void dohelp(void)
{
	char * szSignOn = "TRANSFER %s %s\n%s\n\n";


	if( fIsFat32World )
	{
		printf(szSignOn,SHORTVERSION,BUILDVERSION,COPYRIGHT);
	}
	else
	{
		printf(szSignOn,SHORTVERSION622,BUILDVERSION,COPYRIGHT);
	}
	printf(usage);
}

#define done_exit(n)		{ retval=n; goto done; }


#define USER_CAN 1
#define BAD_BLK  2

int main(int argc,char *argv[])
{
	int iFileHandle=0, iActionTaken;
	int i;
	char c;
	char * filename;
	int err;
	int retval = 0;
	int oldBreakLevel;
	unsigned oldInputDevData, oldOutputDevData;
	unsigned newInputDevData, newOutputDevData;

#if VALLEY_ENGINEERING
	// kludge for valley engineering
	char far * bitptr = MK_FP(0x40,0xD0);
	char oldbit;
	oldbit = *bitptr;
	*bitptr &= ~0x10;		// disable continous reset
#endif

	//  Determine if we are operating in a DOS 7.1 enviornment
	//
	fIsFat32World = dlIsFat32World();

	for(i=1; i <= argc; i++)
	{
		LFNStripArgQuotes( argv[i] );
		if( strlen(argv[i]) >= LFN_MAX_PATH )
		{
			printf("Error: Path too long.\n");
			exit(255);
		}
	}

	// remember current break level (1 = check for ^Cs) and turn
	// break checking off, set STDIN and STDOUT to raw mode.
	//
	dlGetBreakLevel(&oldBreakLevel);
	dlGetDeviceAttributes(STDIN, &oldInputDevData);
	dlGetDeviceAttributes(STDOUT, &oldOutputDevData);

	newInputDevData = ((oldInputDevData | BINARY_MODE) & NOT_THE_CONSOLE);
	newOutputDevData = ((oldOutputDevData | BINARY_MODE) & NOT_THE_CONSOLE);

	dlSetBreakLevel(0);
	dlSetDeviceAttributes(STDIN, newInputDevData);
	dlSetDeviceAttributes(STDOUT, newOutputDevData);

	/* display help */
	if(argc < 2)
	{
		dohelp();
		done_exit(1);
	}
	/* determine the command line options */
	for(i=1; i<argc; ++i)
	{
		if(*argv[i] == '-' || *argv[i] == '/')
		{
			c = toupper(argv[i][1]);
			switch(c)
			{ long temp;

				/* help screen */
				case '?': dohelp(); done_exit(1);

				case 'Q': quiet=1; break;

				/* /B9600 set baud rate */
				case 'B':
					if(toupper(argv[i][2]) == 'C')
					{
						useBIOS = TRUE;
						usecom  = FALSE;
					}
					else
					{
						usecom  = TRUE;
						useBIOS = FALSE;
						temp = atol(argv[i]+2);
						if(temp == 115200L)
							baudrate = 115;
						else
							baudrate = (unsigned)(temp);

						#if DEBUG
							printf("\nbaudrate: %u\n",baudrate);
						#endif
					}
					break;

				/* /COM1  set com port */
				case 'C':
					usecom  = TRUE;
					useBIOS = FALSE;
					comN = atoi(argv[i]+4) - 1;
					if(comN < 0 || comN > 3)
						comN = 0;
					break;

				/* /IRQ set irq number */
				case 'I':
					iIrqNumber = atoi(argv[i]+4);
					break;

				/* send a file */
				case 'S':
					got_opt_send_rec  = sendfile = TRUE;
					break;

				/* receive a file */
				case 'R':
					got_opt_send_rec = TRUE;
					sendfile = FALSE;
					break;

				default:
					dohelp();
					printf(bad_opt_str,argv[i]);
					done_exit(1);
			}
		}
	}

	// set the default irq
	if( iIrqNumber == 0 )
	{
		if( ( comN == 0 ) || ( comN == 2 ) )
		{
			iIrqNumber = 4;
		}
		else
		{
			iIrqNumber = 3;
		}
	}

	// irq range must be between 3 and 15
	if( ( iIrqNumber < 3 ) || ( iIrqNumber > 15 ) )
	{
		dohelp();
		printf(irq_msg);
		done_exit(100);
	}

	// user MUST specify either /R (receive) or /S (send)
	if(!got_opt_send_rec)
	{
		dohelp();
		printf(sendrec_msg);
		done_exit(12);
	}

	/* find the filename */
	filename = NULL;
	for(i=1; i<argc; ++i)
	{
		if(!(*argv[i] == '-' || *argv[i] == '/'))
		{
			filename = argv[i];
			break;
		}
	}

	/* no filename found */
	if(!filename)
	{
		dohelp();
		printf(filename_msg);
		done_exit(2);
	}

	/* open communications (used only for serial I/O) */
	if(usecom)
	{
		comm_opened = comm_open(baudrate, 0u, comN, iIrqNumber, RX_BUFF_SIZE, rxdBuffer);
		if(!comm_opened)
		{
			printf(bad_opt_str,argv[i]);
			done_exit(1);
		}
	}

	/* send a file */
	if(sendfile)
	{
		printf(sending_msg,filename);
		iFileHandle = -1;
		SmartCreateOpenFile(OPEN_ACCESS_READONLY, 0, FILE_OPEN,
			filename, &iFileHandle, &iActionTaken);
		if(iFileHandle == -1)
		{
			printf(open_fail_msg,filename);
			done_exit(3);
		}
		else
		{
			err = xmodem_txd(iFileHandle);
			dlCloseFile(iFileHandle);
			if(err)
			{
				if(err == USER_CAN)
					printf(user_can_msg,filename);
				else
					printf(send_fail_msg,filename);
				done_exit(4);
			}
		}
	}

	/* receive a file */
	else
	{
		printf(receiving_msg,filename);
		iFileHandle = -1;
		SmartCreateOpenFile(OPEN_ACCESS_WRITEONLY, 0,
				FILE_CREATE | FILE_TRUNCATE, filename,	&iFileHandle,
				&iActionTaken);
		if(iFileHandle == -1)
		{
			printf(create_fail_msg,filename);
			done_exit(5);
		}
		else
		{
			err = xmodem_rxd(iFileHandle);
			dlCloseFile(iFileHandle);
			if(err)
			{
				if(err == USER_CAN)
					printf(user_can_msg,filename);
				else
					printf(receive_fail_msg,filename);
				printf(deleting_msg,filename);
				unlink(filename);
				done_exit(6);
			}
		}
	}

done:

	if(comm_opened)
		comm_close();

#if VALLEY_ENGINEERING
	*bitptr = oldbit;
#endif

	// restore old Break (^C) checking and restore STDIN and STDOUT
	// to ASCII mode
	//
	dlSetBreakLevel(oldBreakLevel);
	dlSetDeviceAttributes(STDIN, oldInputDevData);
	dlSetDeviceAttributes(STDOUT,oldOutputDevData);
	return retval;
}



/*
	Transmit the file via XMODEM protocol.  Returns 0 if
	worked, error code if failed.
*/
int xmodem_txd(int iHandle)
{
	int block=0,errors=0,maxErrors;
	int i,c;
	unsigned bytes;
	int err=FALSE;

	tran_flush();
	maxErrors = STARTUP_RETRIES;		// giver user extra time to get started

	/* wait to begin the transmit */
	while(TRUE)
	{
		c = timeout_getc();
		if(c==CAN && (c=timeout_getc())==CAN)
		{
			err = USER_CAN;
			goto exitout;
		}

		if(c==NAK)
			break;

		if(usecom)
		{
			if(bios_kbd() == ESC)
			{
				err = USER_CAN;
				goto exitout;
			}
		}

		if(c == ERR)
		{
			if(++errors >= maxErrors)
			{
				err = BAD_BLK;
				goto exitout;
			}
		}
	}
	tran_flush();	// purge any extra NAKs

	/* transmit the blocks
	*/
	c=ACK;
	errors = 0;
	maxErrors = PACKET_RETRIES;
	while(TRUE)
	{
		// Build a new packet IFF prev packet was ACKed.
		// Pred(1st packet) ACKed by definition.
		//
		if(c==ACK)
		{
			// Read next bytes from file, padding packet data field
			// to length with ASCII SUB (CP/M EOF) chars.
			//
			bytes = 128;
			dlReadFile(iHandle, &bytes, (char *)&buffer[3]);
			if(bytes==0)
				break;

			errors   =  0;			// new packet has no errors
			++block;					// get next packet sequence number
			block   %=  256;

			buffer[0]=  SOH;		// write the synch byte then
			buffer[1]=  block;	// the packet number and its 1s complement
			buffer[2]= ~block;

			if(bytes<128)
			{
//  ETR 3953  BT  3-19-02
//
//  We found that other popular communications programs padded with a NUL
//  instead of a ^Z
//
#if 0
				/* set extra bytes to ^Zs */
				memset(&buffer[3+bytes],26,128-bytes); // should be last block!
#else
				//  Set to NUL
				//
				memset(&buffer[3+bytes],0,128-bytes); // should be last block!
#endif
			}

			// 8 bit chksum catches 95% of single bit errors. Corrects none.
			//
			buffer[131]=checksum(&buffer[3],128);
		}

		// (re)send the current block
		//
		for(i=0;i<132;++i)
			tran_putc(buffer[i]);

		/* make sure the other end received it OK
		*/
		c=timeout_getc();

		if(c==NAK)
			continue;

		if(c==ACK)
		{
			if(usecom && !quiet)
				printf(trans_block_char);
			/* acknowledged */;
		}

		else if(c==CAN && (c=timeout_getc())==CAN)
		{
			err = USER_CAN;
			goto exitout;
		}

		else
		{
			if(c == ERR)
			{
				if(++errors >= maxErrors)
				{
					err = BAD_BLK;
					goto exitout;
				}
			}
		}
		if(usecom)
		{
			if(bios_kbd() == ESC)
			{
				err = USER_CAN;
				goto exitout;
			}
		}
	}

doneout:
	/* end the transmission */
	errors = 0;
	while(TRUE)
	{
		tran_putc(EOT);
		c=timeout_getc();
		if(c == ACK)
			break;
		if(c == ERR)
		{
			if(++errors >= maxErrors)
			{
				err = BAD_BLK;
				goto exitout;
			}
		}
		if(usecom)
		{
			if(bios_kbd() == ESC)
			{
				err = USER_CAN;
				goto exitout;
			}
		}
	}

exitout:
	return err;
}

/* abort_str is 5 ^Xs and 5 ^Cs with backspaces */
char *abort_str =
	"\030\030\030\030\030\010\010\010\010\010"
	"\003\003\003\003\003\010\010\010\010\010";

/*
	Receive a file via XMODEM protocol.  Returns 0 if
	worked, error code if failed.
*/
int xmodem_rxd(int iHandle)
{
	uchar block=1;
	int errors,timeout,maxErrors;
	int c,i;
	unsigned bytes;
	unsigned uWriteSize;
	int err = FALSE;
	char * pTemp;


	// Begin the transmission, we are waiting to synch up with
	// the transmitter's frames.
	//
	tran_flush();
	tran_putc(NAK);
	maxErrors = STARTUP_RETRIES;		// giver user extra time to get started
	errors    = 0;

	/* get the blocks */
	while(TRUE)
	{
		while(TRUE)
		{
			c=timeout_getc();
			if(c==SOH || c==STX)		/* SOH=128 byte blocks, STX=1024 */
			{
				// We have our synch bytes! The transmission is now underway.
				//
				errors    = 0;
				maxErrors = PACKET_RETRIES; // transfer going, tighten threshold
				break;
			}

			if(c==CAN && (c=timeout_getc())==CAN)
			{
				err = USER_CAN;
				goto exitout;
			}

			if(c==EOT)
			{
				err = FALSE;
				tran_putc(ACK);
				tran_putc(ACK);
				goto exitout;
			}

			if(c == ERR)
			{
				if(++errors >= maxErrors)
				{
					err = BAD_BLK;
					goto exitout;
				}
				tran_putc(NAK);
			}

			if(usecom)
			{
				if(bios_kbd() == ESC)
				{
					err = USER_CAN;
					goto exitout;
				}

			}
		}

		/* get the block (normal or 1K protocol)
		*/
		bytes= (c==SOH ? 132 : 1028);
		for(i=1,timeout=FALSE;i<bytes && !timeout;++i)
		{
			c=tran_tgetc(SEC);

			if(c==ERR)
			{
				timeout=TRUE;
				++errors;
			}
			else
				buffer[i]=c;
		}

		/* check if a valid block
		*/
		if(!timeout && validblock(buffer,block,bytes))
		{
//  ETR 3953  BT  3-19-02
//
//  Trying to remove the ^Z padding from a block as the below code
//  does will cause corruption in files that contain a non-padding
//  ^Z character.  Our testing showed that other communication
//  programs do not perform this step.  Thus we removed it to correct
//  the file corruption issue.
//
//  A downside of using the XMODEM protocol is that the file size will
//  be rounded up to the next 128 (or 1024) byte boundary.
//
#if 0
			//  Find an EOF in the buffer data and that marks the
			//  end of the file
			//
			pTemp = memchr( (char *)&buffer[3], 26, 128 );
			if( pTemp )
			{
				uWriteSize = pTemp - ((char *)&buffer[3]);
			}
			else
#endif
			{
				uWriteSize = 128;
			}
			dlWriteFile(iHandle, &uWriteSize, (char *)&buffer[3]);

			err = FALSE;
			errors = 0;
			++block;
			block %= 256;
			tran_putc(ACK);

			if(usecom && !quiet)
				printf(receive_block_char);
		}
		else
		{
			err = BAD_BLK;
			tran_putc(NAK);
		}
	}	/* end of the while loop */

exitout:
	if(err == USER_CAN)
	{
		tran_puts(abort_str);
		tran_flush();
	}
	return err;
}


/*
	Is this block valid?
*/
int	validblock(uchar *buffer, uchar block, int bytes)
{
	if(block != buffer[1] || buffer[1] != (uchar)~buffer[2])
		return FALSE;
	if(checksum(buffer+3,bytes-4)!=buffer[bytes-1])
		return FALSE;
	return TRUE;
}


/*
	Send a string to the other machine.
*/
void	tran_puts(char *s)
{
	while(*s)
	{
		tran_putc(*s);
		++s;
	}
}


/*
	Get a character with timeout.  If "ticks" amount of
	time passes before a character is available, return
	-1.  If a character is available, return it.
*/
int tran_tgetc(int ticks)
{
	long tm;
	long get_ticks();
	int c;

	c = tran_poll();
	if(c != ERR)
		return c;

	tm=get_ticks() + (long)(ticks);
	do
	{
		c = tran_poll();
		if(c != ERR)
			return c;
	} while(tm >= get_ticks());
	return ERR;
}

#if 0
/*
	Get a character.  Wait indefinitely for the charcter.
*/
int tran_getc(void)
{
	int c;

	do
	{
		c = tran_poll();
	} while(c < 0);
	return c;
}
#endif


/*
	Flush the input buffer (throw all characters away).
*/
void tran_flush(void)
{
	while(tran_poll() != ERR)
		/* throw it away */;
}


/*
	Poll the BIOS or COM port for a character.
*/
int tran_poll(void)
{
// int c;

	if(usecom)
	{
		return comm_poll();
	}
	else
	{
		if(useBIOS)
			return bios_poll();
		else
			return CON_poll();
	}
}


/*
	Send a character to the BIOS or COM port.
*/
void tran_putc(char c)
{
	if(usecom)
	{
		comm_putc(c);
	}
	else
	{
		if(useBIOS)
			bios_putc(c);
		else
	 		CON_putc(c);
	}
}


/*
	Returns the byte checksum.
*/
uchar	checksum(uchar *p, int len)
{
	uchar c=0;
	while(len)
	{
		c+= *p;
		++p;
		--len;
	}
	return c;
}

/*
	Returns the char portion of the next extended scan code
	pending in the BIOS keyboard buffer or ERR if no such
	char currently exists.
*/
int bios_kbd(void)
{
	#define ZF 0x40

	_AH = 1;
	geninterrupt(0x16);		// buffer empty?
	if(_FLAGS & ZF)
   	return ERR;				// nothing in buffer
	_AH = 0;
	geninterrupt(0x16);
	_AH = 0;                // ignore scan code
	return _AX;             // return the character
}
