;
;   $Id: EMM386.ASM 1.19 2003/06/23 17:32:00Z thomd Exp $
;
;   *** DO NOT REMOVE THE FOLLOWING NOTICE ***
;
;   Copyright (c) 1993 - 2002 Datalight, Inc.
;   All Rights Reserved
;
;   Datalight, Inc. is a Washington corporation with an address at
;   21520 30TH DR SE MS110 Bothell, WA, 98021
;   Tel: 800.221.6630  Fax: 425.951.8091
;   Http://www.datalight.com
;
;   This software, including without limitation all source code and
;   documentation, is the confidential, trade secret property of
;   Datalight, Inc., and is protected under the copyright laws of
;   the United States and other jurisdictions. Portions of the
;   software may also be subject to one or more the following US
;   patents: US#5860082, US#6260156.
;
;   In addition to civil penalties for infringement of copyright
;   under applicable U.S. law, 17 U.S.C. 1204 provides criminal
;   penalties for violation of (a) the restrictions on circumvention
;   of copyright protection systems found in 17 U.S.C. 1201 and
;   (b) the protections for the integrity of copyright management
;   information found in 17 U.S.C. 1202.
;
;   U.S. Government Restricted Rights. Use, duplication,
;   reproduction, or transfer of this commercial product and
;   accompanying documentation is restricted in accordance with
;   FAR 12.212 and DFARS 227.7202 and by a License Agreement.
;
;   IN ADDITION TO COPYRIGHT AND PATENT LAW, THIS SOFTWARE IS
;   PROTECTED UNDER A SOURCE CODE AGREEMENT, NON-DISCLOSURE
;   AGREEMENT (NDA), OR SIMILAR BINDING CONTRACT BETWEEN DATALIGHT,
;   INC. AND THE LICENSEE ("BINDING AGREEMENT"). YOUR RIGHT, IF ANY,
;   TO COPY, PUBLISH, MODIFY OR OTHERWISE USE THE SOFTWARE,IS SUBJECT
;   TO THE TERMS AND CONDITIONS OF THE BINDING AGREEMENT, AND BY
;   USING THE SOFTWARE IN ANY MANNER, IN WHOLE OR IN PART, YOU AGREE
;   TO BE BOUND BY THE TERMS OF THE BINDING AGREEMENT. CONTACT
;   DATALIGHT, INC. AT THE ADDRESS SET FORTH ABOVE IF YOU OBTAINED
;   THIS SOFTWARE IN ERROR.
;
comment ~ -------------------------------------------------

   EMM386.ASM  created 06/29/94  by Drew Gislason

   Copyright (c) 1994-2002 by Datalight, Inc.
   All Rights Reserved.
   Datalight Confidential and Proprietary

   This device driver provides enough support for ROM-DOS
   to implement basic EMM386 support including Upper Memory
   Blocks (UMBs).  The commands LOADHIGH and DEVICEHIGH
   require this support.

   Syntax:
   DEVICE = HIMEM.SYS
   DEVICE = EMM386.EXE {I=xxxx-yyyy} {X=xxxx-yyyy} [FRAME=seg[,memK]]
            [ROM=nnnn-mmmm] [D=xxx] [MAX=memK] [LOW] [PS2]

   Where xxxx and yyyy define a range of memory.  I= includes UMBs in
   that range, X= excludes UMBs in that range. ROM= creates ROM shadowing
   in that range.  HIMEM.SYS (or another XMS manager) must be installed
   prior to EMM386.

   Examples:
   DEVICE = EMM386.EXE
   DEVICE = EMM386.EXE I=C800-D7FF I=F000-F7FF FRAME=E000
   DEVICE = EMM386.EXE ROM=F000-FFFF

   Supported Calls:
   ----------------

   40H - Get Manager Status
   41H - Get Page Frame Segment
   42H - Get Number of Pages
   43H - Get Handle and Allocate Memory
   44H - Map Memory
   45H - Release Handle and Memory
   46H - Get EMM Version
   47H - Save Mapping context (only one save at a time now)
   48h - Restore Mapping context
   4BH - Get number of EMM handles
   4CH - Get Pages Owned by Handle
   4DH - Get Pages for All Handles
   4EH - Get or set page map
   4FH - Get, Set partial page map
   50H - Map/Unmap Multiple Pages (subfunction 0 is done)
   51H - Reallocate Pages
   52H - Get Set handle attributes
   53H - Get Set Handle Name
   54H - Get Handle directory
   55H - Alter page map and jump
   56H - Alter page map and call
   57H - Move Exchange Memory region
   580xH - Get Mappable Physical Addresses
   59H - Get EMS Hardware Information
   5A0xH - Allocate standard/raw pages
   5BH - Alternate map register set
   5CH - Prepare for a warm boot
   5DH - Enable Disable OS function set
The functions below are defined as "EEMS" functions and are NOT supported
by MS-EMM386 or Datalight's EMM386.
        60H - Get physical windows array                 EEMS
        68H - Get addresses of all page frames in system EEMS
        69H - Map page into frame                        EEMS
        6AH - Page Mapping                               EEMS

   DExxH - Selected VCPI calls


   GLOSSARY:
   ---------
        FRAME           One of 4 16K pages located in a 64k block

        HANDLE          Identifier to reference multiple-page
                        allocations.

        FRAME TABLE     Mapping information for the frames.
                        Consists of one word for each 16K page of
                        physical memory.  The word is zero if the
                        page is available, set = handle if the
                        page is allocated.  Handles FFF0 - FFFF
                        are reserved for use by VCPI memory
                        allocation.  Specifically, handles FFF0 -
                        FFFF are for memory allocated via VCPI
                        function DE04.  The four LSBs indicate
                        which 4K pages within the 16K EMS page
                        are allocated.  A 1 indicates the
                        corresponding page is allocated, a 0
                        indicates the corresponding page is free.
                        Handle FFF0 is normally never used since
                        it indicates no pages are allocated and
                        thus the page should be returned to the
                        EMS pool with a handle of 0.  Handle
                        8000h indicates the block is not present.


        _frame_pages    Word indicating the number of entries in
                        the frame table, which = total number of
                        16K pages possible for EMS.

        _frame_table    An array of 10 near pointers, pointing to the various
                        blocks in the frame table.  Each pointer indicates the
                        start of a block and the end of a previous block.  If
                        a pointer is less than the previous pointer, the
                        block with an indicated size of less than or equal to
                        zero is not present.

        _frame_seg      The segment at the start of the 64K frame
                        segment

        _frame_memK     Total memory specified on command line to be reserved
                        for EMS in Kilobytes.

	_frame_maxK	The maximum number of Kbytes that can be allocated
			from XMS to be used as EMS.

        _frame_memptr   An array of 9 32 bit addresses indicating the start of
                        each block of XMS memory used for EMS.

        _frame_avail    A word indicating the number of 16K pages
                        allocated from XMS and available in the EMS pool.

        incremental_XMS A word indicating the number of kilobytes of excess
                        XMS memory to request, if and when the current EMS
                        pool is exhausted.  This is intended to allocate all
                        available XMS memory in 8 allocations after
                        initialization.

        vcpi_remainder  The total number of available 4 K pages
                        which are contained in 16K pages which
                        have been allocated for use by the VCPI
                        host.  Since VCPI memory allocations are
                        done in 4K chunks, there are often pieces
                        of memory available within partially used
                        16K EMS pages.

   HISTORY:
   --------
   05/19/2000 dg  Added USING_GEOS switch to indicate special GEOS
       version during sign-on. see also mapper.asm.
   06-08-98  (DE)
   1.  SPR 1688 Added PS2 command line switch to force use of
       PS2 style (port 92h) A20 control.

   03-13-98  (DE)
   1.  User arguments now over-ride auto-detection of ROM/RAM.
   2.  Added MAX=nnn option to specify max EMS in Kbytes
   3.  Added LOW switch to prevent putting EMM386 into first
       sufficient UMB. This option uses about 20K of low memory.

   11-25-97  (DE)
   1.  Cleaned up isseparator() and isterminator() to work more
       efficiently (this IS an .ASM file, yes?) and added the
       '/','-' and '=' chars to the list of separators.
   2.  Added command line processing for the "RAM" option (with or
       without a range argument). Defined DEFAULT_UMB_START (0C800h)
       and DEFAULT_UMB_END (0F000h) in EMM386.INC as the segment
       values to be used when no range arguments are given.
       Nothing is done with these command line arguments.

   11-19-97  (DE)
   1.  Changed enable, disable and free DMA reg set (fn 5b06,5b07
       5b08) to not generate an error when called with values of
       0 for DMA reg set number and DMA channel. Per LIM 4.0 spec.
       Cleaned up the 5Dxx fns to use symbolics and renamed
       f5d_disabled to fnAccessState to make sure I caught all
       references.
   2.  Changed the alloc AMRS call to save the current state of
       the mapping registers in the new AMRS per the LIM 4.0 spec.
       Added new symbolic constants to define the AMRS fn errors
       and AMRS limits. The save area size is now defined by the
       SIZE of a struc and allocation dimension is now
       determined by a single symbolic constant.
   3.  Yet more factoring of fn 57xx: F57MovexToVMMRegs now
       processes the EMS_MOVEX_STRUC and converts into validated
       registers ready for either VMM_2 or VMM_A calls. Fixed a
       bug resulting from the previous factoring exercise.
   4.  Created a new EMM386.INC file that has structure definitions
       and equates pertaining to EMS and VMM page mapping stuff.

   11-17-97  (DE)
   1.  Took out "reverse move" code from fn 5700h. It didn't
       work at all and wasn't needed anyway as VMM fn 2 handles
       overlapped blocks nicely.
   2.  Made default DEBUG level 0, leaving out the "unsupported
       int 67 fn" message by default. Also changed unsupported
       function error (84h) to unsupported subfunction error
       (87h) where appropriate.
   3.  Changed fn 52xxh to work like MS EMM386

   11-16-97  (DE)
   1.  Factored out fn 57xx tests for block overlap into the
       CheckF57Overlap routine and changed the return code for
       exchange fn overlap from 92h to 97h. 92h is a warning
       status code for the move fn only, 97h is a fatal error.
       Added move struct definition for use in fn 57xx.
   2.  Added a new CheckF57MemRange function that validates the
       arguments passed by the user to fn 57xxh. This routine
       returns several new error codes and so avoids things that
       caused those functions to fail (len > EMS block size, 1M
       DOS mem wrap, DOS mem/page frame overlap, offset into EMS
       page >= 16k).
   3.  Renamed emm_57_descr_to_abs to F57DescripToPhysAddr, made
       it into a proc and changed the error code to always be
       80h (the return error codes were never checked anyway).
       This error code assumes that the argument validation code
       is working.
   4.  Corrected comments on fn 58xxh and 5900h as well as the
       contents of the structure returned by fn 5900h.

   11-14-97  (DE)
   1.  Corrected alloc and realloc functions to discriminate
       between the "want more than is free" (87h) and the "want
       more than there ever was" (88h) EMS error codes.
   2.  Added code to validate passed handle on entry to
       functions 44h (map/unmap a single page) and 50h
       (map/unmap multiple pages).
   3.  The "save the LIM 3.2 mapping context to a handle"
       function (47h) had a bug that allowed it to save multiple
       contexts to the same handle. Factored out the "find a handle
       in buffer_1" code from fns 47h,48h,53h and 54h into a new
       routine called FindHandleInBuffer1 and made fn 47h use it.
   4.  The free handle function (45h) had a bug that allowed it
       to delete a handle when it had a saved context from a
       function 47h call. Used the FindHandleInBuffer1 routine
       to verify that there is no saved mapping context before
       allowing the handle to be freed.
   5.  Added a check for handle 0 to the check_handle routine so
       that handle 0 would not be validated as handle 255 (the
       BT routine underflowed). This does not correct the
       problem of treating handle 0 as invalid.
   6.  The "get pages allocated to all handles function" did not
       count (or return info on) handles that had 0 pages
       allocated to them. It should have and does now. It still
       doesn't handle handle 0.
   7.  The get/set partial EMS context functions (4Fxxh) should
       allow a page count of 0 as a NOP instead of generating an
       error. They do now.
   8.  The map and call/jump functions (55xxh/56xx) should
       allow a page count of 0 as a NOP instead of generating an
       error. They do now.
   9.  Substituted symbolic constants for some magic numbers.
       Declared some code chunks to be procs.

   11-11-97  (DE)
   1.  Added the DEBUG target to the make file which includes the debug
       printf code processing stuff from the Winlight project. The
       files from the Winlight project are in the new VMMDEBUG dir.
   2.  Fixed functions 47/48 so that they properly saved/restored
       LIM 3.2 EMS contexts when the XMS blocks are noncontiguous.
       Note that I throw away the upper 4 bits of the logical page's
       physical address leaving access to only 256M of memory.

   06/21/97: (DKH)
   1.  Fixed a problem with PIC # 2 interrupts during hlt instructions.  Also
       fixed a problem in which the machine crashes if there are no UMBs.
   06/15/97: (DKH)
   1.  Fixed a bug when linking a UMB at A000 to the DOS MCBs.
   04/26/97: (DKH)
   1.  Fixed some problems when rebooting with a UMB stack.
   04/25/97: (DKH)
   1.  Added I= and X= to the command line processing.  C800-F000 is an
       assumed UMB, bios extention detection and ROM/RAM detection added for
       UMB areas from c000-f000.
   2.  Fixed a bug where function 45h sometimes trashed dx.
   04/18/97: (DKH)
   1.  Fixed a bug in function 5800h, the logical page numbers were incorrect.
   04/17/97: (DKH)
   1.  Finished the XMS-EMS smart allocation code.
   2.  Fixed a bug whereby calls to map_memory were trashing the high
       words of certain registers, most notably edx.
   02/17/97: (DKH)
   1.  Moved VCPI mode switch stack to DEVDATA segment for compatibility
       with 16 bit Rational systems extenders.
   02/14/97: (DKH)
   1.  Added support for function 5700h
   2.  Fixed code which adjusts BIOS reboot vector to search for eah.
   07/25/96: (DKH)
   1. Added VCPI support
   2. Added the "ROM=" shadow RAM feature
   10/31/95: (RME)
   1. Added functions 5800/01H "Get Mappable Physical Addresses"
   10/16/95: (RME)
   1. Added function 5900H "Get EMS Hardware Information"
   10/11/94: (DAG)
   1. Added "Unsupported Int67 Call xxxx"
   2. Added 4DH, and 86H(disable V86)
   3. Extend conventional RAM with range
   10/05/94: (DAG)
   1. Allow EMM386 to move into upper memory
   06/29/94: (DAG)
   1. Created first try for UMBs.
   2. Implemented functions 40H-46H Basic EMS functionality
      (at Zetron's request)
   02/24/95: (DAG)
   1. Support A20 off by trapping I/O access and remapping HMA
   2. i67f5000H

---------------------------------------------------------- ~
__FILENUM__	equ <1>


;  ETR 3801 BT 11-26-01
;
;  Allow the ability to relocate the EBDA to allow for expanding
;  conventional memory when the A000 block is included
;
ENABLE_MOVEEBDA		EQU	1


include macros.inc
include dosver.inc
include emm386.inc

ifndef debug
DEBUG = 0
endif
include datalght.inc
include wlassert.inc
include intelx86.inc
include dlmacros.inc
include vmm.inc
include log.inc


JUMPS	;; need this for relative jumps around debugging macros


;/////////////////////////////////////////////////////////////////////////
;//	User/Programmer Options						//
;/////////////////////////////////////////////////////////////////////////

SIGNON	equ	1	; set to FALSE for no signon
public MAXRANGES
MAXRANGES equ	8	; maximum # of ranges for remapping
MAXHANDLES equ	256	; 256 maximum for emmimp structure.
;MAXHANDLES must be an even multiple of 16!!!!
PAGESPERFRAME equ 4	; # of pages in each frame (4x16K pages)

TSR	  equ	0	; is this program a TSR and device driver (1)
			; or device driver only (0)
TRAPFFFF  equ	1	; trap FAR JMP at FFFF:0? (1=yes, 0=no)

;STACKSIZE equ	1024	; size of stack in bytes
STACKSIZE equ	4096	; size of stack in bytes

; DEBUG level 0 = none
; DEBUG level 1 = unsupported call message
; DEBUG level 2 = debugging using the LOG.EXE TSR
; DEBUG = 1

; Set CALL_67F86 to 0 to remove call 86H (disable V86 mode)
CALL_67F86 = 0

;/////////////////////////////////////////////////////////////////////////
;//	Macros, etc..							//
;/////////////////////////////////////////////////////////////////////////

printm	macro	name
	mov	DX, offset DGROUP:name
	call	print_dos
	endm

PRERROR	macro	name
	mov	DX, offset DGROUP:error_msg
	call	print_dos
	mov	DX, offset DGROUP:name
	call	print_dos
	endm

RANGE	STRUC
	range_start	dw	?	; segment
	range_len	dw	?	; in paragraphs
RANGE	ENDS

; Request header fields
cmdlen	=	00H	; length of this command
unit	=	01H	; sub unit specifier
cmdc	=	02H	; command code
status	=	03H	; status
endaddr	=	0EH	; end address
cmdline	=	12H	; cmdline ptr

; ptr to arguments
P	equ	4


;/////////////////////////////////////////////////////////////////////////
;//	Segment Ordering						//
;/////////////////////////////////////////////////////////////////////////

_DEVDATA segment para public 'DEVDATA'
_DEVDATA ends

_ENDDEV	 segment para public 'DEVDATA'
_ENDDEV  ends

cBegCode
cEndCode

; paragraph align DGROUP
BEGDATA	SEGMENT PARA PUBLIC '_BEGDATA'
	DGROUP GROUP BEGDATA
public datastart
datastart label word
BEGDATA	ENDS

cBegData
cEndData

; For the Microsoft C/Assembly users...
; too many segments - yuck!
IFNDEF	??version
	DGROUP	GROUP	CONST, MSG
CONST SEGMENT WORD PUBLIC 'CONST'
CONST ENDS
MSG SEGMENT WORD PUBLIC 'MSG'
MSG ENDS
ENDIF

cBegUdata
begbss	label word
cEndUdata

_ENDSEG	SEGMENT PARA PUBLIC 'ENDSEG'
	public endsegmarker
endsegmarker label word
_ENDSEG ENDS


;/////////////////////////////////////////////////////////////////////////
;//	Device Driver Header						//
;/////////////////////////////////////////////////////////////////////////

_DEVDATA segment
	assume CS:_DEVDATA, DS:NOTHING

	public	_emmx
_emmx:
	dd	-1		; 0 link to next device
	dw	0c000H		; 4 character device
	dw	strategy	; 6 strategy routine
	dw	service		; 8 service routine
dev_name        db	"EMMDATA0",0	; 10 this name must be changed if EMS

	; these variables make _DEVDATA not ROMable
	public	cmd_ptr
cmd_ptr		dw	?
_text7  dw      _TEXT

.386p
align 4
public dev_cr3,dev_gdt_ptr
dev_cr3         dd      0
dev_gdt_ptr     df      0
ems_enabled     db      0               ;0 for disabled, 1 for enabled

	public	dev_int67,_text1,_text2,_text3,_text4,_text5,_text6,_text7
dev_int67:
                db      0eah            ;far jump
                dw      offset _TEXT:int_67_handler
_text1          dw      _TEXT

strategy:
	mov	word ptr CS:[cmd_ptr], offset _TEXT:emm_strat
	jmp	dword ptr CS:[cmd_ptr]

service:
	mov	word ptr CS:[cmd_ptr], offset _TEXT:emm_serve
	jmp	dword ptr CS:[cmd_ptr]

v86_dis:        ;jmp to call_vmm with al = d
        mov     al,0dh          ;function to disable vmm

        db      0eah            ;far jump to call_vmm
        dw      offset _TEXT:call_VMM
_text6  dw      _TEXT

public v86_dis2
v86_dis2:       ;control returns here in protected mode, cr0 in eax
;si contains the real mode ss, esp already has the real mode sp
        mov     cr0,eax                         ;now in real mode
        db      0eah                            ;far jump
        dw      offset v86_dis_jmp              ;offset for ip
        dw      _DEVDATA                        ;real mode code segment

v86_dis_jmp:
        mov     ss,si
;Caution!  a bug in the 486 25 MHz B step chips causes cache memory to shut down
;when cr3 is loaded with 10h or 18h.  EAX at this point is 10h with that chip
;so we cannot use eax in the instruction below.
        mov     cr3,esp         ;flush tlb to remove any remapping still stored
;we have fully restored the system to real mode, return to caller.
        retf

;The function below must be located in this segment, since it must still
;be here when V86 mode is disabled.  The portion of EMM which is relocated in
;the UMB is not accessible once V86 mode is disabled.
v86_en_dis:     ;call with ax = 0 to disable V86 mode, ax = 1 to enable
        or      ax,ax
        jz short v86_dis
;enable V86 mode
        mov     dx,ss
        mov     ebx,dev_cr3
        mov     eax,cr0
        or      al,1
        mov     cr0,eax                         ;now in protected mode
        mov     cr3,ebx                         ;page directory
        or      eax,80000000h
        mov     cr0,eax                         ;paging is enabled
;UMBs are now back in place, we can access the gdt now
        lgdt    fword ptr dev_gdt_ptr
;jump to the main body of EMM386 to finish up the job.
        db      0eah            ;far jump to en_v86 in protected mode.
        dw      offset _TEXT:en_v86
        dw      CODE_0          ;protected mode _TEXT segment

win_installed   db      0       ;set to non-zero if winlight is active

_2f_1605:       ;return the address of the V86 enable, disable function
;also, change the name of the device driver to EMMXXXX0
        mov     si,offset v86_en_dis
        push    cs
        pop     ds
assume ds:_DEVDATA
        mov dword ptr ds:dev_name[3],'XXXX'     ;set device name
        mov byte ptr ds:win_installed,al        ;set win_installed to non-zero
;unhook the UMBs
        push    ax
        pushf
        mov     ax,5803h
        xor     bx,bx
        ; This call generates a compiler optimization warning, but
        ; must be a far call for compatibility with MS Windows 3.1
        ;
        call	far ptr _21_5803           ;returns below with flags popped off
        pop     ax
        xor     cx,cx                   ;indicate windows may load
;es:bx is supposed to point to the startup info structure.  From experiments
;with MS-Windows and MS-EMM386, it appears this data has no effect on MS-
;Windows.  We will ignore es:bx for now.
        iret
assume ds:nothing

_2f_1606:       ;Winlight is terminating, change device driver name if required
        mov     cs:win_installed,0      ;indicate windows is no longer installed
        cmp     cs:ems_enabled,0
        jne short int2f_chain           ;jmp if ems is enabled
;we must alter the device driver name to indicate EMS services are not
;available.
        mov dword ptr cs:dev_name[3],'ATAD'     ;set device name
        jmp short int2f_chain

int_2f:         ;check for function 1605h, Winlight init broadcast
        cmp     ax,1605h
        je short _2f_1605
;check for function 1606h, windows exit broadcast
        cmp     ax,1606h
        je short _2f_1606
int2f_chain:
                db      0eah            ;far jump
int2f_ptr       dd      ?               ;jmp to original handler

int_21:         ;check for function 5803h
        cmp     ax,5803h
        je short _21_5803
int21_chain:
                db      0eah            ;far jump
int21_ptr       dd      ?               ;jmp to original handler

_21_5803:       ;handles function 5803h.
                db      0eah            ;far jump
                dw      offset _TEXT:int_21_5803
_text2          dw      _TEXT

int_15:         ;handles int 15 functions
        cmp     ax,4f53h
        je short _15_4f
        cmp     ah,87h
        je short _15_87
int_15_chain    db      0eah            ;far jmp
int15_ptr       dd      ?               ;jmp to original handler

_15_4f: db      0eah                    ;far jump
        dw      offset _TEXT:int_15_4f
_text3  dw      _TEXT                   ;space for code segment

_15_87: db      0eah                    ;far jump
        dw      offset _TEXT:int_15_87
_text4  dw      _TEXT                   ;space for code segment

;handler for int 4bh ah != 81h, as specified in VDS specification
_4b_iret:       iret

int_4b:         ;check for VDS functions
        cmp     ah,81h
        je short _4b_81

                db      0eah            ;far jmp
;this vector may be changed during initialization
int4b_ptr       dw      offset _4b_iret
                dw      _DEVDATA

_4b_81: db      0eah
        dw      offset _TEXT:int_4b_81
_text5  dw      _TEXT

align 4
comment &
        The 16 bit Rational systems extender requires a stack in memory which
        is mapped 1:1.  This is the only segment which will work when EMM
        relocates to a UMB.
&
public vcpi_stack
vcpi_stack      db      20 dup (0)

IF TSR
	public	req_hdr
	req_hdr		db (cmdline+4) dup (?)
ENDIF
.186
_DEVDATA	ends


; Segment _TEXT must start with a word of FFFFs to end the
; I/O permission map  - used by the Remapper code.
cBegCode
	public	FFFFs
FFFFs	dw	-1
cEndCode


;/////////////////////////////////////////////////////////////////////////
;//	Data Area							//
;/////////////////////////////////////////////////////////////////////////

cBegData
;the array below is used by function f4f to support function f4e.
f4e_list        dw      LIM32_MAPPABLE_PAGES    ; segment count in list
_frame_seg      dw      0                       ; segment for mappable page 0
                dw      16*1024/16              ; segment for mappable page 1
                dw      32*1024/16              ; segment for mappable page 2
                dw      48*1024/16              ; segment for mappable page 3



; space for 8 AMRS mapping state save areas
;
f5b_save_area   DL_EMS_CONTEXT	(MAX_AMRS+1) dup(<>)

f5b_reg_sets    db      1               ;handle zero is always used
f5b_current_set db      0               ;current register set number
f5b_ptr         dd      0               ;es:di value for register set zero


; Bit record used to maintain the access state of OS/E functions
; 59h,5Bxxh and 5Dxxh. See EMM386.INC for bit value definitions.
;
fnAccessState   db      0

; These two words hold the access key that is issued, verified
; and updated by fn 5Dxxh. The key is initialized to the DOS
; time at device initialization and is updated (by adding to
; both words whatever value happens to be in the BP register)
; whenever the key is returned by the OS/E.
;
key1            dw      0
key2            dw      0



	public	signon_msg
signon_msg	label byte
	db	13,10,"EMM386 "
IF USING_GEOS
	db	"for Geos "
ENDIF
	db FULLVERSION,13,10
	db	COPYRIGHT,13,10,"$"

IFE TSR
device_only	label byte
   db 13,10
   db "Syntax:",13,10
   db "DEVICE = HIMEM.SYS",13,10
   db "DEVICE = EMM386.EXE {I=xxxx-yyyy} {X=xxxx-yyyy} [FRAME=seg[,memK]]",13,10
   db "         [ROM=nnnn-mmmm] [D=xxx] [MAX=memK] [LOW] [PS2]",13,10
   db 13,10
   db "Where xxxx and yyyy define a range of memory.  I= includes UMBs in",13,10
   db "that range, X= excludes UMBs in that range. ROM= creates ROM shadowing",13,10
   db "in that range.  HIMEM.SYS (or another XMS manager) must be installed",13,10
   db "prior to EMM386.",13,10
   db 13,10
   db "Examples:",13,10
   db "DEVICE = EMM386.EXE",13,10
   db "DEVICE = EMM386.EXE I=C800-D7FF I=F000-F7FF FRAME=E000",13,10
   db "DEVICE = EMM386.EXE ROM=F000-FFFF",13,10
   db "$"
ENDIF

	public	bss_zeroed
bss_zeroed db	0

	public	error_msg
error_msg	label byte
	db	13,10,"EMM386 Error: ","$"

	public	already_installed_msg
already_installed_msg 	label byte
	db	"Already installed.",13,10,"$"

	public	no_xms_msg
no_xms_msg	label byte
	db	"Requires HIMEM - not installed.",13,10,"$"

insuf_mem_msg  label byte
db      "Insufficient extended memory for this command line",13,10,"$"


	public	not_a_386_msg
not_a_386_msg	label byte
	db	"Requires an 80386 or later.",13,10,'$'

	public	failed_remap_msg
failed_remap_msg	label byte
	db	"Failed to remap memory!",13,10,'$'

	public	not_in_real_mode_msg
not_in_real_mode_msg label byte
	db	"Already in protected mode.",13,10,'$'

	public	could_not_control_A20_msg
could_not_control_A20_msg label byte
	db	"Could not control A20 line.",13,10,'$'

error_msgs	dw 	offset not_a_386_msg
		dw	failed_remap_msg
		dw	not_in_real_mode_msg
		dw	could_not_control_A20_msg
total_msgs	dw	($-(offset error_msgs))/2
init_err_code	dw	0

	public	framestr
framestr	label byte
	db	"FRAME="

	; used to zero BSS
	saveDI	dw	?
	saveES	dw	?
	saveDX	dw	?
	saveCX	dw	?
	saveAX	dw	?


IF ENABLE_MOVEEBDA
fRelocateEBDA	db	1	; Boolean to relocate EBDA by default
fDidRelocate	db	0

szEBDAStr	label byte
	db	"NOMOVEXBDA"
ENDIF ; IF ENABLE_MOVEEBDA

cEndData

cBegUdata

	public	ptrsav
	ptrsav	dd	?

IF STACKSIZE
	public	user_stack, stacktop, stackbottom
	user_BX		dw	?
	user_CX		dw	?
	user_DS		dw	?
        user_stack      dd      ?               ;ss:sp if real mode
                        dw      ?               ;space for ss if protected mode
	stackbottom	db	STACKSIZE dup (?)
	stacktop	dw	?
ENDIF

public win_page_table_ptr
win_page_table_ptr      dd      ?               ;pointer to ems page tables

	; data segment of DOS
	public	DOS_DS
DOS_DS	dw	?

	;----------------------------------
	;  Internal memory allocation
	public	malloc_ptr
malloc_ptr	dw	?

	;----------------------------------
	;  (EMS) Expanded Memory Service Data
	public	_frame_seg, _frame_memK, _frame_memptr, _frame_maxK
	public	_frame_pages, _frame_avail, _frame_table
_frame_memK	dw	?
_frame_memptr	dd	9 dup (?)
_frame_table    dw      10 dup (?)              ; ptrs to frame table
_frame_pages	dw	?	; total # of pages possible
_frame_avail	dw	?	;
_frame_maxK	dw	?
	public xms_handles
xms_handles     dw      9 dup (?)               ; xms handles



	public	incremental_XMS
incremental_XMS dw      ?
vcpi_remainder  dw      ?

	;----------------------------------
	;  Misc Data
public vds_buff_memk
vmm_memptr      dd      0       ;pointer passed to VMM for physical address
rom_memptr      dd      0       ;pointer to physical memory for ROM=
num_rom         dw      0       ;number of ROM= ranges
rom_start       dw 4 dup (0)    ;segment address for start and stop of shadow
rom_stop        dw 4 dup (0)
rom_memK        dw      0

BUFFER_1_MAP_CONTEXT	equ	0
BUFFER_1_HANDLE_NAME	equ	1
comment &
        Buffer_1 serves as a storage location for functions such as 47, 48, 53
        and 54h.  Format of data in this buffer is as follows:

        function 47,48
              offset  bytes   description
                0       1       handle
                1       1       0, marks it as function 47 and 48
                2       8       state of the page frame (index of page's
				entry in the FRAME TABLE array)

        function 53,54
              offset  bytes   description
                0       1       handle
                1       1       1, marks it as function 53 and 54
                2       8       handle name

        Each entry consumes 10 bytes.
&
buffer_1        dw      40 dup (0)      ;allow for 8 entries
vds_buff_memk   dw      0       ;kbytes of memory to allocate for vds services

;each bit in the array below represents a handle, a 0 means the handle
;is free, a 1 indicates the handle has been allocated.
public handle_list
handle_list     dw      MAXHANDLES/16 dup (0)

	;----------------------------------
	;  (UMB) Upper Memory Block Data
	public	_num_ranges, _range_memK, _ranges, _range_memptr
	public	_first_range
_num_ranges	dw	?
_range_memK	dw	?
_ranges		RANGE MAXRANGES dup (<>)	; ranges of memory
_range_memptr	dd	?			; pointer to range memory
_first_range	dw	?	; 1st range might be 1 if conventional
				; RAM has been expanded

comment &
        Variables below are used to map the first megabyte in 4K pages for I=
        and X= on the command line.  Ranges specified without an identifier in
        front of them are treated as I= ranges.  Each bit in the arrays below
        corresponds to a 4K page.  A 1 indicates that page was specified on
        the command line.  When we are finished, X= takes precedence over I=.

        The range from C800-F000 is autodetected and additional bits may be
        set in the i_cmd_range depending on what is found in that area.
&
i_range         db      32 dup (?)      ;include range
x_range         db      32 dup (?)      ;exclude range

	;-----------------------------------
	;  Data to Help Move EMM Code to Upper Memory
	public	_total_memK, _phys_memptr
	public	dest_move_seg, src_move_seg, moved
_total_memK	dw	?
_phys_memptr	dd	?
dest_move_seg	dw	?	; to CS when moved
src_move_seg	dw	?	; from CS when moved
moved		db	?

	;-----------------------------------
	;  Data to redirect FAR JMP at FFFF:0
IF TRAPFFFF
_reboot_memptr	dd	?	; memory for rebooting
reboot_tmpptr	dd	?	; pointer for temporary memory for top 4K
ENDIF

	public	xms_handler, xms_handle
xms_handler     dd      ?               ; far pointer to XMS handler
xms_handle      dw      8 dup (?)       ; handles for XMS memory that EMM owns

	public 	_donot_load_high
_donot_load_high dw	?

cEndUdata


;/////////////////////////////////////////////////////////////////////////
;//	Device Driver Service						//
;/////////////////////////////////////////////////////////////////////////
cBegCode
.386

	public	use_PS2_A20
use_PS2_A20 db  0


extrn   _TEXT_W:abs, _TEXT_E:abs

	; note: These code-segment variables make the code
	; not ROMable

	public	old_FFFF0
	old_FFFF0 dd	?	; old JMP from FFFF:0

	uOldMemTop	dw	0 ; old top of memory

	public	emm_strat, emm_serve, emm_init, cmderr
emm_strat:
	push	DS
	call	getDGROUP

	; we may need to zero BSS.  Done only once
	cmp	[bss_zeroed], 0
	jnz short already_done
	call	zero_bss
already_done:

	mov	word ptr [ptrsav], BX
	mov	word ptr [ptrsav+2], ES
	pop	DS
	retf

;---------------------------------------
;  Zero the BSS segment.  Done only once
zero_bss:
	; record the fact that we've zeroed it
	mov	[bss_zeroed], 1

	; store registers we will use
	mov	[saveAX], AX
	mov	[saveCX], CX
	mov	[saveDX], DX
	mov	[saveES], ES
	mov	[saveDI], DI

	pop	DX		; pop return addr

	; zero BSS
	mov	CX, DS
	mov	ES, CX
	mov	DI, offset DGROUP:begbss
	mov	CX, offset DGROUP:endsegmarker
	sub	CX, DI
	xor	AX, AX
	cld
	rep	stosb

	push	DX		; restore return addr

	; restore used registers
	mov	DI, [saveDI]
	mov	ES, [saveES]
	mov	DX, [saveDX]
	mov	CX, [saveCX]
	mov	AX, [saveAX]

	ret

emm_serve:
	pushf
	push	AX
	push	BX
	push	CX
	push	DX
	push	SI
	push	DI
	push	DS
	push	ES

	call	getDGROUP
IF STACKSIZE
	call	switch_stacks
ENDIF
	; get pointer to i/o packet
	les	BX, [ptrsav]

	; get output status to appease TLINK
	cmp	byte ptr ES:[BX.cmdc], 0AH
	jz	short exit

        ; Check for the read command
        cmp byte ptr es:[bx.cmdc],3     ;check for ioctl write
	je 	cmd_read

	; Check for the INIT(0) command
	cmp	byte ptr ES:[BX.cmdc], 0	; command code
	jnz short cmderr

	; call the init routine, returns 0 if OK
	call	emm_init

IF DEBUG GE 1
   call  BIOSCRLF
ENDIF
        or      ax,ax
	jz	short exit		; worked
	; fall thru to cmderr...

;
;  An error occurred.  Tell DOS not to install by returning
;  a bogus end address.
;
cmderr:
	les	BX, [ptrsav]
	mov	word ptr ES:[BX.endaddr], 0
	mov	word ptr ES:[BX.endaddr+2], CS
	mov	AL, 3		; unknown cmd error
err_exit:
	mov	AH, 81H		; mark error return
	jmp	short err1

exit:	mov	AH, 01H

IF ENABLE_MOVEEBDA
	;  Determine if the XBDA needs relocating
	;
	cmp	fRelocateEBDA, 0
	jnz short DoRelocate
ENDIF ; IF ENABLE_MOVEEBDA

err1:	les	BX, [ptrsav]
	mov	word ptr ES:[BX.status], AX	; mark as complete

IF STACKSIZE
	call	restore_stack
ENDIF
	pop	ES
	pop	DS
	pop	DI
	pop	SI
	pop	DX
	pop	CX
	pop	BX
	pop	AX
	popf
	retf

IF ENABLE_MOVEEBDA
DoRelocate:

	;  Check if we already relocated
	;
	cmp	fDidRelocate, 0
	jne	err1

	;  Relocate the XBDA to follow emm driver
	;
	push	ds
	push	es
	push	bx
	push	cx
	push	si
	push	di


	;  Set a flag indicating that we already relocated
	;
	mov	bl, 1
	mov	fDidRelocate, bl


	les	BX, [ptrsav] ; Get the address of the device packet

	push	es

	les	di, dword ptr ES:[BX.endaddr]	; EBDA destination in ES:DI
	add	di, 0fh
	shr	di, 4
	mov	ax, es
	add	ax, di
	mov	es, ax
	xor	di, di

	xor	si, si
	mov	ds, si			; zero ds and si
	mov	ds, word ptr ds:[40eh]	; EBDA source in DS:SI
	mov	cx, word ptr ds:[0h]	; EBDA size
	shl	cx, 8
	cld
	rep	movsd

	mov	cx, es		; save the destination segment in CX
	add	di, 0fh		; Round up to nearest paragraph
	shr	di, 4		; convert to paragraphs
	add	di, cx		; New ending segment in DI

	pop	es
	mov	word ptr ES:[BX.endaddr+2], di	; update ending segment in packet
	xor	di, di
	mov	word ptr ES:[BX.endaddr], di	; update ending offset in packet
	mov	es, di
	mov     word ptr es:[40eh], cx	; Update EBDA segment in BDA

	pop	di
	pop	si
	pop	cx
	pop	bx
	pop	es
	pop	ds

	jmp	err1

ENDIF ; IF ENABLE_MOVEEBDA

cmd_read:       ;the only read we support is to supply the address of
                ;the emmimp structure.
        les     di,es:[bx.endaddr]
        mov     eax,cs:emmimp_addr
        stosd
        mov     ax,0b01h
        stosw
        jmp     exit

stack_msg       db      13,10,'Reentrant stack error in EMM386.',24h
switch_err:    ;fatal stack error
        mov     dx,offset _TEXT:stack_msg
        jmp     fatal_err

stack_alloc     db      0               ; nonzero if stack has been allocated

comment &
------------------------------------------------
  Switch from user to our stack.
  Called only from real or V86 mode.
  No registers (other than SS:SP) affected.

  Re-entrant problems are checked as follows:  Check
  value of stack_alloc and increment it.  If we are
  already on our stack, return with no other changes.  If
  we are not on our stack, and original stack_alloc was
  non-zero, jmp to fatal error.  If original stack_alloc
  was zero, switch stacks and return.

&

	public	switch_stacks
switch_stacks:
	mov	[user_BX], BX
	mov	[user_CX], CX
	mov	BX, SS
	mov	CX, DS
	cmp	BX, CX
	jz	short no_switch
;we are not on our stack, check if stack is in use.
        cmp     stack_alloc,0
        jne     switch_err
;stack is not in use, make the switch.
	pop	BX		; retrieve return address
;save the current stack
	mov	word ptr [user_stack+2], SS
	mov	word ptr [user_stack], SP
	mov	SS, CX
	mov	SP, offset DGROUP:stacktop
	push	BX		; store return address
no_switch:      ;restore original bx and cx
	mov	BX, [user_BX]
	mov	CX, [user_CX]
        inc     stack_alloc
        ret

comment &
------------------------------------------------
  Switch from our stack to user stack.
  Assumes DS = DGROUP, real mode only.

  No registers (other than SS:SP) affected

  Re-entrant problems are handled as follows:  Decrement
  stack_alloc, if zero, switch stacks, if not zero, do
  not switch stacks.

&
	public	restore_stack
restore_stack:
        dec     stack_alloc
        jz short unswitch
;This call is not the complement to the original call to switch stacks.
        ret

unswitch:
;This call is the complement to the original call to switch stacks.
        pop     user_BX                 ;pop return address
        lss     sp,user_stack
        jmp     user_BX                 ;return to caller

switch_errp:    ;fatal stack error
        mov     dx,offset _TEXT:stack_msg
        jmp     fatal_errp

comment &
        Equivalent to switch_stack only for protected mode.  Returns with SS
        equivalent to _TEXT_W.  Remember, we might arrive here with a 32 bit
        stack!
&
assume ds:_TEXT
public switch_stackp
switch_stackp:
        push    ds
        push    ebx
        push    cx
        mov     cx,cs
        add     cx,8
        mov     bx,ss
        cmp     bx,cx
        jz short no_switchp
;we are not on our stack, check if stack is in use.
        cmp     cs:stack_alloc,0
        jne     switch_errp
;stack is not in use, make the switch.
        db      66h
        mov     bx,offset _TEXT:stacktop        ;new esp
        dw      0                               ;high word of esp

        push    cx                      ;new ss
        mov     cx,ss                   ;old ss
        pop     ss                      ;load new ss
        xchg    esp,ebx                 ;load new esp, save old esp in ebx
        mov     ds,cx
        push    dword ptr ds:[ebx+6]            ;ds and ret
        push    dword ptr ds:[ebx+2]            ;ebx
        push    word ptr ds:[ebx]               ;cx
        add     ebx,10                          ;pop all garbage from old stack
        push    ebx                             ;old stack pointer is on stack
;we must trick the assembler and linker into using ss to write to user_stack
assume ss:DGROUP        ;ss is actually equivalent to _TEXT_W
        mov     bx,offset _TEXT:datastart
        pop dword ptr ss:user_stack[bx]         ;save old esp
        mov word ptr ss:user_stack[bx+4],ds     ;save old ss
assume ss:_TEXT
no_switchp:     ;restore original ebx and cx
        pop     cx
        pop     ebx
        pop     ds
        inc     ss:stack_alloc
assume ss:nothing
        ret

assume ds:nothing,ss:_TEXT
restore_stackp:
        dec     ss:stack_alloc
        jz short unswitchp
;This call is not the complement to the original call to switch stacks.
        ret

unswitchp:
;This call is the complement to the original call to switch stacks.
        push    ebx
        mov     bx,offset _TEXT:datastart
assume ss:DGROUP        ;ss is actually _TEXT
        push    word ptr ss:user_stack[bx+4]            ;old ss
        mov     ebx,dword ptr ss:user_stack[bx]         ;old esp
assume ss:nothing
        pop     ss                      ;load old ss
        xchg    esp,ebx                 ;load old sp
        push    word ptr cs:[bx+4]      ;ret address
        push    dword ptr cs:[bx]       ;ebx
        pop     ebx
        ret

assume ds:DGROUP

.186
cEndCode

;/////////////////////////////////////////////////////////////////////////
;//	Utilities                                                       //
;/////////////////////////////////////////////////////////////////////////

cBegData

if DEBUG ge 1
err_emm		db	"EMM386 Error: ",0
unsupp_str	db	"Unsupported i67 call ",0
unsupp2_str	db	"!",13,10,0
endif ; DEBUG

if 0
IF DEBUG GE 2	; debug level 2 - show entry to mapper calls
v86init_str	db	"V86Init",13,10
		db	"Phys %lx, from %lx, to %lx, size %x",13,10,0
v86off_str	db	"V86off",13,10
		db	"from %lx, to %lx, size %x",13,10,0
remap386_str	db	"ReMap386",13,10
		db	"ExtAdr %lx, RealAdr %lx, size %lx",13,10,0
ENDIF ; DEBUG
endif

cEndData

cBegCode
.386

;-------------------------------------------
;  Get DGROUP into DS relative to current CS
;  Affects no registers.
	public	getDGROUP
getDGROUP proc near
	push	AX
	push	CX
	mov	AX, offset _TEXT:datastart
	mov	CL, 4
	shr	AX, CL
	mov	CX, CS
	add	AX, CX
	mov	DS, AX
	pop	CX
	pop	AX
	ret
getDGROUP endp

	public	getDEVDATA
getDEVDATA proc near
	mov	AX, SEG _DEVDATA
	ret
getDEVDATA endp


IF DEBUG GE 1

;-----------------------------
;  On Entry
;	AL = character
BIOSPUTC_WAIT:
	call	BIOSPUTC
	push	AX
	xor	AX, AX
	int	16H
	pop	AX
	ret

;-----------------------------
;  On Entry
;	AL = character
public BIOSPUTC
BIOSPUTC:
	push	AX
	push	BX
	mov	AH, 0EH
	mov	BX, 7
	int	10H
	pop	BX
	pop	AX
	ret

;-----------------------------
;  On Entry
;	DS:SI = string
BIOSPUTS:
	push	ax
	push	SI
again_puts:
	mov	AL, [SI]
	cmp	AL, 0
	jz short	done_puts
	call	BIOSPUTC
	inc	SI
	jmp	again_puts
done_puts:
	pop	SI
	pop	ax
	ret

;-----------------------------
;  On Entry
;	AL = digit
BIOSPUTDIGIT:
	push	AX
	push	BX
	and	AL, 0FH
	cmp	AL, 0AH
	jae short hexdigit
	add	AL, '0'
	jmp	short digit10
hexdigit:
	add	AL, 'A'-0AH
digit10:
	mov	BX, 7
	mov	AH, 0EH
	int	10H
	pop	BX
	pop	AX
	ret

;-----------------------------
;  On Entry
;	AX = 16 bit hex #
BIOSPUTX:
	push	CX
	push	DX
	mov	DX, AX
	mov	CL, 12
bpx_again:
	mov	AX, DX
	shr	AX, CL
	push	CX
	push	DX
	call	BIOSPUTDIGIT
	pop	DX
	pop	CX
	sub	CL, 4
	cmp	CL, 0
	jge	bpx_again
	pop	DX
	pop	CX
	ret

;-----------------------------
;  On Entry
;	DX:AX = 32 bit hex #
	public	BIOSPUTLX
BIOSPUTLX:
	push	AX
	mov	AX, DX
	call	BIOSPUTX
	pop	AX
	call	BIOSPUTX
	ret

	public	BIOSCRLF
BIOSCRLF:
	push	AX
	mov	AL, 13
	call	BIOSPUTC
	mov	AL, 10
	call	BIOSPUTC
	pop	AX
	ret

;-----------------------------
; Dump DS:i_range
;   this display is of the bits in 0-15 order, left to right
;
DUMP_IRANGE:
      ; debug code -- display the found "x" range
      ;
      push  AX
      push  BX
      push  CX
      push  DX
      push  SI
      pushf

      mov   AL,'I'
      call  BIOSPUTC
      mov   AL,'='
      call  BIOSPUTC

      mov   si,offset DGROUP:i_range
      mov   cx,16
      CLD
debug_i_loop:
      LODSW
      mov   BX,16
      mov   DX,AX
      push  CX

      MOV   DX,AX
      XOR   AX,AX
      MOV   BX,AX
      MOV   CX,16
debug_i2_loop:
      BT    DX,BX
      RCL   AX,1
      INC   BX
      loop debug_i2_loop

      pop   CX
      cmp   CX,6
      ja short debug_i3
      push  AX
      mov   AL,'.'
      call  BIOSPUTC
      pop   AX
debug_i3:
      call  BIOSPUTX
      loop  debug_i_loop

      popf
      pop   SI
      pop   DX
      pop   CX
      pop   BX
      pop   AX
ret

;-----------------------------
; Dump DS:x_range
;   this display is of the bits in 0-15 order, left to right
;
DUMP_XRANGE:
      ; debug code -- display the found "x" range
      ;
      push  AX
      push  BX
      push  CX
      push  DX
      push  SI
      pushf

      mov   AL,'X'
      call  BIOSPUTC
      mov   AL,'='
      call  BIOSPUTC

      mov   si,offset DGROUP:x_range
      mov   cx,16
      CLD
debug_x_loop:
      LODSW
      mov   BX,16
      mov   DX,AX
      push  CX

      MOV   DX,AX
      XOR   AX,AX
      MOV   BX,AX
      MOV   CX,16
debug_x2_loop:
      BT    DX,BX
      RCL   AX,1
      INC   BX
      loop debug_x2_loop

      pop   CX
      cmp   CX,6
      ja short debug_x3
      push  AX
      mov   AL,'.'
      call  BIOSPUTC
      pop   AX
debug_x3:
      call  BIOSPUTX
      loop  debug_x_loop

      popf
      pop   SI
      pop   DX
      pop   CX
      pop   BX
      pop   AX
ret

ENDIF ; DEBUG

;---------------------------------------------
;  This function is a misnomer in that it is
;  really the size of code + data rounded to
;  the nearest 4K.
;
;  On Exit
;	AX = size (in bytes) of code + data
;
	public	CodeSize4K
CodeSize4K proc near
	mov	AX, offset _TEXT:datastart
	add	AX, [malloc_ptr]
	add	AX, 4*1024 - 1		; round up to nearest 4K
	and	AX, 0F000H
	ret
CodeSize4K endp

;---------------------------------------------
;  Allocate memory from our own internal pool.
;  On Entry
;     AX = bytes to allocate
;  On Exit
;     AX = near ptr to memory
	public	malloc
malloc	proc near
	push	DX
	mov	DX, [malloc_ptr]
	add	[malloc_ptr], AX
	xchg	DX, AX
	pop	DX
	ret
malloc	endp

cEndCode

;/////////////////////////////////////////////////////////////////////////
;//	EMS Support							//
;/////////////////////////////////////////////////////////////////////////

EMSERR_INTERNAL		equ    080h	; general error
EMSERR_HARDWARE		equ    081h	; hardware failure
EMSERR_BADHANDLE 	equ    083h	; no pages allocated to this handle
EMSERR_BADFUNC		equ    084h	; unsupported function #
EMSERR_NOHANDLES 	equ    085h	; no handles left
EMSERR_SAVED_CONTEXT	equ    086h	; attempt to free handle w/ saved context
EMSERR_GT_TOTAL		equ    087h	; too many pages requested
EMSERR_GT_FREE		equ    088h	; too many pages requested
EMSERR_NONE		equ    089h	; no pages requested
EMSERR_BADPAGE		equ    08Ah	; page larger than # that handle owns
EMSERR_BADPHYS		equ    08Bh	; invalid physical page
EMSERR_MAPFULL		equ    08Ch	; mapping state save area full
EMSERR_CONTEXT_EXISTS	equ    08Dh	; error saving mapping context
EMSERR_NO_SUCH_CONTEXT 	equ    08Eh	; error restoring mapping context
EMSERR_BAD_SUBFUNC	equ    08Fh	; unsupported sub-function #
EMSERR_NO_CAPABILITY    equ    091h	; the EMM does not have that capability
EMSERR_MOVE_OVERLAP	equ    092h	; move overlap is merely a warning
EMSERR_LEN_GT_BLOCK	equ    093h	; move/xchg region len > allocated len
EMSERR_DOS_EMS_OLAP 	equ    094h	; move/xchg EMS and DOS memory overlap
EMSERR_BAD_PAGE_OFFSET	equ    095h	; move/xchg logical page offset>16K
EMSERR_LEN_GT_1M	equ    096h	; move/xchg region len > 1M
EMSERR_XCHG_OVERLAP	equ    097h	; exchange overlap is a fatal error
EMSERR_BAD_MEM_TYPE	equ    098h	; move/xchg bad memory type ID
EMSERR_AMRS_UNDEFINED   equ    09Ah	; AMRS # > 7 (or >n for a=n on cmd ln)
EMSERR_NO_MORE_AMRS     equ    09Bh	; no more AMRS left
EMSERR_DMARS_UNSUPPORTED equ   09Ch	; don't support DMA reg sets
EMSERR_AMRS_NOT_ALLOCED equ    09Dh	; AMRS # has not been allocated
EMSERR_NO_HANDLE_NAME	equ    0A0h	; named handle not found
EMSERR_DUP_HANDLE_NAME	equ    0A1h	; handle name already exists
EMSERR_DOS_1M_WRAP	equ    0A2h	; move/xhcg DOS memory region > 1M
EMSERR_DATA_CORRUPT	equ    0A3h	; an EMS data struct had invalid data
EMSERR_ACCESS_DENIED	equ    0A4h	; OS/E denied acces to function

cBegCode

 extrn _ReMap386:near, _V86Init:near, _V86off:near, __x386_code_16_des:word
 extrn irq0:word, call_VMM:near, int_15_87:near,int_4b_81:near,vds_init:near
 extrn en_v86:near,CODE_0:abs,emmimp_addr:dword,__x386_ptor:word
 extrn __x386_rtop:word,fatal_err:near,fatal_errp:near
IF TRAPFFFF
; extern void write_protect_mem(ulong addr);
	extrn _write_protect_mem:near
ENDIF

;---------------------------------------------------
;  Returns a handle # in DX.  Affects no other registers.
;  Note that handles FFF0H - FFFFH are reserved for VCPI use.
;  Returns zero if all handles are in use.
	public	get_new_handle
get_new_handle proc near
        push    ax
        push    cx
        push    si
        mov     si,offset DGROUP:handle_list
        mov     cx,MAXHANDLES/16        ;number of words to check
        cld
handle_loop:
        lodsw
        not     ax                      ;invert for bsf instruction
        bsf     dx,ax
        loopz   handle_loop             ;loop until we find a handle
        jz short handle_error
;a handle has been found
        sub     si,offset DGROUP:handle_list[2] ;lodsw added 2 to si
        shl     si,3                    ;multiply by 8, 8 handles/byte
        add     dx,si                   ;dx is now bit index into handle_list
        bts     handle_list,dx
        inc     dx                      ;don't allow a handle of zero
handle_ret:
        pop     si
        pop     cx
        pop     ax
        ret

handle_error:   ;no handle found
        xor     dx,dx                   ;return a handle of zero
        jmp     handle_ret

get_new_handle endp

;---------------------------------------------------
;  void alloc_pages(int handle, int pages)
;  On Entry
;	DX = handle, BX = # pages to allocate
;  On Exit
;	Returns nothing, affects no registers.
;
	public	alloc_pages
alloc_pages	proc near
	push	AX
	push	BX
	push	CX
	push	DI
	push	ES

	; allocate each page
	;
	cld
	mov	CX, [_frame_pages]
	mov	AX, DX
	mov	DI, [_frame_table]
	push	DS
	pop	ES

apage0:	or	BX, BX
	jz short alloc_done
	cmp	word ptr [DI], 0
	jz short got_free_page
	inc	DI
	inc	DI
	loop	apage0
	jmp	short alloc_done	; no more pages, give up!

got_free_page:
	stosw
	dec	BX
	dec	word ptr [_frame_avail]
	loop	apage0

alloc_done:
	pop	ES
	pop	DI
	pop	CX
	pop	BX
	pop	AX
	ret
alloc_pages	endp


;---------------------------------------------------
;  On Entry
;	DX = handle
;  Frees all pages associated with that handle and converts EMS memory
;  to XMS memory if possible.  Updates _frame_avail.
	public	free_pages
free_pages	proc near
        push    ax
        push    bx
        push    cx
        push    dx
        push    di
        push    si
        push    es

	; zero all entries that the handle refers to
	xor	AX, AX
	mov	CX, [_frame_pages]
	mov	BX, [_frame_table]
fpage0:	cmp	word ptr [BX], DX
	jnz short fpage1
	mov	word ptr [BX], 0
	inc	AX
fpage1:	inc	BX
	inc	BX
	loop	fpage0
        add     _frame_avail,ax                 ;update available EMS memory

;remove any entries in buffer_1 containing this handle
        mov     bx,offset DGROUP:buffer_1
        mov     cx,8                    ;max entries in the buffer

;low byte in buffer_1 contains the handle.
fpage3:
        cmp     ds:[bx],dl              ;check if handle
        jne short fpage2                ;jmp if not this handle
        mov     ds:[bx],dh              ;zero the entry
fpage2:
        add     bx,10
        loop    fpage3

;now check if we can convert any memory from EMS to XMS
;The first block is reserved memory and must never be deallocated
        mov     bx,16                    ;index for variables
        push    ds
        pop     es
        xor     si,si
fpage_loop1:
        mov     dx,xms_handles[bx]      ;get an XMS handle
        or      dx,dx
        jz short fpage4                 ;jmp if no handle
        mov     di,_frame_table[bx]     ;bottom of block
        mov     cx,_frame_table[bx+2]   ;top of block
        sub     cx,di
        shr     cx,1                    ;convert to words in the block
        xor     ax,ax
        cld
        repe    scasw
        jne short fpage4                ;jmp if block is still in use
;This block is no longer in use and can be deallocated.
        mov     si,bx                   ;save for defragmentation
        mov     xms_handles[bx],ax      ;indicate no longer used
        mov     ah,0dh
        call    xms_handler	;unlock the block
        mov     ah,0ah
        call    xms_handler	;free the block
        mov     bx,_frame_table[2+si]   ;top end of this block
        sub     bx,_frame_table[si]     ;number of bytes in the block
        shr     bx,1                    ;number of entries
        sub     _frame_avail,bx         ;adjust _frame_avail down
        mov     bx,si                   ;some versions of himem trash bx
fpage4:
        sub     bx,2
        jnz     fpage_loop1
comment &
        To prevent fragmentation problems in the frame table, we will slide
        all used block downwards so we have contiguous used blocks at the
        bottom of the frame table and the max size possible free block at the
        top.  Bx is zero at this point, si contains the lowest new block freed
        or zero if no blocks were freed.
&
        or      si,si
        jz short fpage7                 ;jmp if no blocks were freed.
;si contains an index indicating the lowest block freed.
        mov     di,_frame_table[si]     ;lowest free area in the table
;see if there are blocks in use above this one
        mov     bx,si
fpage_loop2:
        xor     ax,ax
        add     bx,2
        cmp     bx,16
        ja short fpage6                  ;jmp if no more free blocks
        cmp     xms_handles[bx],ax
        je      fpage_loop2             ;jmp if the next block is also free
;bx is the index for the next used block, move it down
        xchg    xms_handles[bx],ax      ;free the next block, get the handle
        mov     xms_handles[si],ax      ;save the handle
        mov     cx,_frame_table[bx+2]   ;top of the block
        push    si
        mov     si,_frame_table[bx]     ;bottom of the block
        sub     cx,si                   ;number of bytes to move
        shr     cx,1                    ;number of words to move
        cld
        rep     movsw                   ;move the entries in the frame table.
        pop     si
        mov     _frame_table[si+2],di   ;adjust the frame table pointers
        shl     si,1
        shl     bx,1                    ;adjust for dwords
        push    _frame_memptr[bx]       ;entry for block we are moving
        pop     _frame_memptr[si]       ;new location.
        shr     si,1                    ;adjust for indexing words
        shr     bx,1
        add     si,2                    ;point to next available block
        jmp     fpage_loop2

fpage6:         ;from di to the top of the frame table is free, si is index
        xor     ax,ax
fpage_loop3:
        add     si,2
        mov     _frame_table[si],ax     ;mark end of valid frame table
        cmp     si,18
        jb      fpage_loop3

        mov     ah,80h                  ;ax was zero, now 8000h
        mov     cx,_frame_pages         ;number of entries in the frame table
        shl     cx,1                    ;convert to bytes in table
        add     cx,_frame_table         ;pointer to top of table
        sub     cx,di                   ;number of bytes free
        shr     cx,1                    ;number of words free
        rep     stosw                   ;mark all remaining entries as invalid
fpage7:
        pop     es
        pop     si
        pop     di
        pop     dx
        pop     cx
        pop     bx
        pop     ax
        ret
free_pages	endp

;---------------------------------------------------
;  On Entry
;       DX = handle, BX = logical page #
;  On Exit
;       EDX = physical address of page, CF = 0 or CF = 1 if error
;
find_page:
        push    BX
        push    CX
        mov     AX,BX                   ; page # in AX
        mov     CX,[_frame_pages]
        mov     BX,[_frame_table]
findp0: cmp     word ptr [BX],DX
        jz 	short is_it_this_page

findp1: inc     BX
        inc     BX
        loop    findp0
        jmp 	short bad_page_number

is_it_this_page:
        or      AX, AX                	; we got it!
        jz 	short found_page_number
        dec     AX
        jmp     findp1               	; go on to next page, please

found_page_number:
        xchg    ax,bx                   ; ax = frame table entry offset, bx = 0
        mov     cx,9
findp2:
        cmp     ax, _frame_table[bx+2]
        jb 	short findp3

        inc     bx
        inc     bx
        loop    findp2

bad_page_number:
        stc
        jmp short find_page_done

	; bx is 2 * the XMS block number corresponding to the desired page.
	; ax is offset of page entry in the frame table
	;
findp3:
        sub     ax, _frame_table[bx]

	; AX is twice the 0 based index of a 16K logical EMS page
	; contained within a specific XMS block. Get the address
	; of the start of that XMS block from the _frame_memptr
	; array and then calculate the physical address of the
	; logical page. Note that the values stored in the
	; _frame_memptr array are aligned on 4K raw (CPU) pages
	; (the 12 LSBs are 0).
	;
        shl     bx,1            ;convert to index into memptr array (dwords)
        movzx   edx,ax
        shl     edx,13          ;multiply by 8K
        add     edx, dword ptr _frame_memptr[bx]

	; CF must = 0 at this point.
	;
find_page_done:
        pop        CX
        pop        BX
        ret

	; call with bx = index into frame table
        ; returns physical address in dx:ax
	;
get_phys:
        push    bx
        push    cx
        xor     ax,ax
        jmp     found_page_number

;---------------------------------------------------
;  On Entry
;	DX = handle
;  On Exit
;	BX = # of pages owned by handle
get_pages	proc	near
	push	CX
	push	DI

	xor	BX, BX			; no pages to start with
	mov	CX, [_frame_pages]
	mov	DI, [_frame_table]
getp0:	cmp	word ptr [DI], DX
	jnz short gp_not_ours
	inc	BX
gp_not_ours:
	inc	DI
	inc	DI
	loop	getp0

	pop	DI
	pop	CX
	ret
get_pages	endp


;---------------------------------------------------
;  The int 67H functions MUST be contiguous in the
;  following list, starting at function 40H.
;
	public i67_functions
	public i67f40,i67f42,i67f43,i67f44,i67f45,i67f46
	public i67f4C,i67f4D,i67f50,i67f53,i67f58,i67f59

i67_functions label byte
	db	40H
	dw	offset i67f40
	db	41H
	dw	offset i67f41
	db	42H
	dw	offset i67f42
	db	43H
	dw	offset i67f43
	db	44H
	dw	offset i67f44
	db	45H
	dw	offset i67f45
	db	46H
	dw	offset i67f46
	db	47H
	dw	offset i67f47
	db	48H
	dw	offset i67f48
	db	4BH
	dw	offset i67f4B
	db	4CH
	dw	offset i67f4C
	db	4DH
	dw	offset i67f4D
	db	4EH
	dw	offset i67f4E
	db	4FH
	dw	offset i67f4F
	db	50H
	dw	offset i67f50
IF SUPPORT_FN51 GE 1
	db	51H     		; hang machine on a realloc
	dw	offset i67f51
ENDIF
	db	52H
	dw	offset i67f52
	db	53H
	dw	offset i67f53
	db	54H
	dw	offset i67f54
	db	55H
	dw	offset i67f55
	db	56H
	dw	offset i67f56
	db	57H
	dw	offset i67f57
	db	58H
	dw	offset i67f58
	db	59H
	dw	offset i67f59
	db	5AH
	dw	offset i67f5A
	db	5BH
	dw	offset i67f5B
	db	5CH
	dw	offset i67f5C
	db	5DH
	dw	offset i67f5D
IF CALL_67F86
	db	86H
	dw	offset i67f86
ENDIF
	db	0DEH		; VCPI calls
	dw	offset i67fDE
	db	0		; end of list

;---------------------------------------------------
;  Do we support this int 67H function?
;
;  Returns SI=funcptr (supported), or SI=0 (not supported)
;  Destroys AX.
;
	public	is67func
is67func:
	mov	SI, offset i67_functions
isloop:	cmp	byte ptr CS:[SI], 0
	jz	short notfound
	cmp	byte ptr CS:[SI], AH
	jz	short gotfunc
	add	SI, 3
	jmp	short isloop
gotfunc:
	inc	SI
	ret
notfound:
	xor	SI, SI
	ret

;---------------------------------------------------
;  The actual INT 67H handler
;
;  All the EMS/LIM 4.0 that we support are handled
;  through here.
;
	public	int_67_handler
int_67_handler:

if DEBUG ge 2
	cmp	ah, 0DEh
	je      ignore_vcpi
	LOGPRINTF <'i67: AX= %x, BX= %x, CX= %x, DX= %x'>,ax,bx,cx,dx
   ignore_vcpi:
endif
	cld
	push	DS
	push	SI
	call	getDGROUP
	call	is67func
	or	SI, SI
	jz short i67_not_supported_halt
	jmp	CS:[SI]

i67_not_supported_halt:
IF DEBUG GE 1
	push	AX			; save call #
	mov	SI, GOFFSET err_emm
	call	BIOSPUTS
	mov	SI, GOFFSET unsupp_str
	call	BIOSPUTS
	pop	AX			; restore and print call #
	push	ax
	call	BIOSPUTX
	mov	SI, GOFFSET unsupp2_str
	call	BIOSPUTS
	pop	ax

	cmp	ah, 51h
	jne short pause

	cli
    hang:
	jmp	hang

    pause:
	xor	AX, AX
	int	16H
ENDIF ; DEBUG
i67_not_supported:
	mov	AH, EMSERR_BADFUNC
	jmp	short i67done

i67_subfn_not_supported:
	mov	ah, EMSERR_BAD_SUBFUNC
	jmp	short i67done

	;-------------------------------------
	; 40H - Get Manager Status
i67f40:	;jmp	short i67ok

i67ok:	mov	AH, 0
i67done:
if DEBUG ge 2
	LOGPRINTF <'...: AX= %x'>,ax
;	or	ah, ah
;	jz	i67wentok
;	push	ax
;	xor	ax, ax
;	int	16h
;	pop	ax
i67wentok:
endif
	pop	SI
	pop	DS
	iret


	;-------------------------------------
	; 41H - Get Page Frame Segment
i67f41:	mov	BX, [_frame_seg]
	or	BX, BX
	jnz	i67ok
	mov	AH, EMSERR_INTERNAL
	jmp	i67done

	;-------------------------------------
	; 42H - Get Page Counts
i67f42:
        call    get_xms_to_ems          ;returns bx = convertable pages
        add     bx,_frame_avail
        mov     dx,_frame_pages
        jmp     i67ok

	;-------------------------------------
	; 43H - Allocate Memory (and Handle)
i67f43:
	; do nothing if asking for 0 pages
	;
	or	BX, BX
	jnz short alloc1
	mov	AH, EMSERR_NONE
	jmp	i67done
alloc1:
	; do we have enough pages available?
	cmp	BX, [_frame_avail]
	jbe short alloc2

	;not enough pages currently available, can we allocate some XMS?
	;
        call    expand_ems              ;returns CF = 0 if success
        jnc short alloc2
	mov	AH, EMSERR_GT_FREE
	cmp	bx, [_frame_pages]
	jbe short alloc_err
	mov	ah, EMSERR_GT_TOTAL

alloc_err:
	jmp	i67done

alloc2:
	; determine a handle #
	;
	call	get_new_handle	; returned in DX, returns 0 if no handles
        or      dx,dx
        jnz short alloc3                ;jmp if a handle was found

	;all handles are in use, return error
	;
        mov     ah,EMSERR_NOHANDLES
        jmp     i67done

alloc3:
	; allocate the requested pages
	;
	call	alloc_pages

alloc4:
	jmp	i67ok


comment &
        Gets the total XMS memory which can be converted to EMS memory and
        returns it in BX.  Checks for alignment troubles, room in the frame
        table, etc.
&
get_xms_to_ems:
        xor     bx,bx
        cmp     _frame_seg,bx
        jne short get_xms_3
        ret                             ;return zero if no EMS support
get_xms_3:
        pusha
        mov     bp,sp
        mov     cx,8
        mov     si,offset DGROUP:_frame_table[2]
get_xms_loop:
        lodsw
        cmp     ax,ds:[si]
        jae short get_xms_1             ;jmp if we found an available pointer
        loop    get_xms_loop
;no more blocks available in the frame table.
get_xms_zero:   ;no more memory available
        xor     bx,bx
	xor	si,si
        jmp short get_xms_ret

get_xms_1:
;ax points to the current top of the last used block in the frame table
        sub     ax,_frame_table         ;total number of bytes currently used
        shr     ax,1                    ;number of entries used
        mov     cx,_frame_pages         ;total space in pages
        sub     cx,ax                   ;available entries in pages
        mov     ah,8
        call    xms_handler             ;returns largest block in ax
        mov     dx,ax
        mov     si,ax
        mov     ah,9
        call    xms_handler             ;allocate memory
        dec     ax
        jnz     get_xms_zero
        mov     ah,0ch                  ;lock memory
        push    dx
        call    xms_handler
        pop     dx
        push    bx                      ;save lsbs of physical address
        mov     ah,0dh
        call    xms_handler             ;unlock the block
        mov     ah,0ah
        call    xms_handler             ;free the XMS memory
        pop     bx
        and     bx,3                    ;check 4K alignment
        jz short get_xms_2
;start address is not 4K aligned, bx has the 2 lsbs expressed in kbytes
        lea     si,[si+bx-4]
get_xms_2:      ;si is 4K pages in XMS, cx is available frame table entries
        shr     si,4                    ;convert kbytes to 16K pages
;use the smaller of the two
        cmp     si,cx
        jb short get_xms_ret
        mov     si,cx
get_xms_ret:
        mov     ss:[bp+8],si            ;return value for bx
        mov     sp,bp
        popa
        ret

comment &
        Call with BX = total number of EMS pages we are trying to allocate.
        Expands the EMS memory pool by allocating XMS memory and adding it to
        the Frame table.  Preserves all registers, returns CF = 0 if success,
        1 if failure.
&
expand_ems:
        pusha
        sub     bx,_frame_avail         ;total new pages needed
        mov     bp,bx
        jb      expand_ret              ;jmp if we already have enough memory
        call    get_xms_to_ems          ;returns total possible new pages in bx
        cmp     bx,bp
        jb short expand_err             	;jmp if not enough memory

        mov     cx,8
        mov     si,offset DGROUP:_frame_table[2]

expand_loop:
        lodsw
        cmp     ax,ds:[si]
        jae short expand_1              ;jmp if we found an available pointer
        loop    expand_loop
	; we should never reach this point since get_xms_to_ems has succeeded.

expand_err:
        stc
        popa
        ret

expand_err2:    ;handle is pushed on the stack, deallocate XMS block
        pop     dx
        mov     ah,0dh          ;unlock the block
        call    xms_handler
        mov     ah,0ah
        call    xms_handler     ;free the block
        jmp     expand_err

expand_1:       ;we found an available block in the frame table.
                ;ax points to the first available entry in the frame table
                ;si points to location for end marker of new block
                ;bp is the number of bytes needed, bx is the max possible to add
        mov     di,ax
;We have the min acceptable, can we get incremental_XMS more?
        add     bp,incremental_XMS
;now take the smallest of bp or bx
        cmp     bp,bx
        jb short expand_2
        mov     bp,bx                   ;use the smaller one
expand_2:       ;bp contains the number of enries we will add to the frame table
        mov     dx,bp
        shl     dx,4                    ;convert to kilobytes
        mov     ah,9
        call    xms_handler
        dec     ax
        jnz     expand_err
;XMS memory has been allocated, lock it and see if we have 4K alignment
        mov     ah,0ch
        push    dx              ;save handle
        call    xms_handler
        dec     ax
        jnz short expand_err2
        test    bx,0fffh        ;check 4K alignment
        jz short expand_3
;The memory block is not 4K aligned.  We must expand the memory block by 4K
;to allow for this.
        pop     dx              ;restore the handle
        mov     ah,0dh          ;unlock the block
        call    xms_handler
        dec     ax
        push    dx              ;some versions of himem trash dx
        jnz     expand_err2
        mov     ah,0fh          ;reallocate XMS block, make it bigger
        mov     bx,bp           ;number of 16K pages
        shl     bx,4            ;convert to K bytes
        add     bx,4            ;add a 4K page
        call    xms_handler
        dec     ax
        jnz     expand_err2
        pop     dx              ;restore handle
        mov     ah,0ch
        push    dx              ;save handle
        call    xms_handler
        dec     ax
        jnz short expand_err2

	; Our XMS blocks (which hold our logical EMS pages) are 4K aligned.
	;
expand_4K_align_XMS_blocks:
        and     bx,0f000h       ;4K align the physical address

expand_3:       ;no more errors possible, start modifying frame_table, etc.
        pop     ax              ;get xms handle
        push    dx
        push    bx              ;32 bit physical address of new block
        push    ax              ;xms handle
        mov     cx,bp           ;number of 16K pages to add to frame table
        add     _frame_avail,bp ;increase indicated available EMS pages.
        xor     ax,ax
        cld
        push    es
        push    ds
        pop     es
        rep     stosw           ;add the entries to the frame table
        pop     es
        mov     ds:[si],di      ;end marker for new block
        sub     si,offset DGROUP:_frame_table[2]
;si is an index for xms_handles
        pop     xms_handles[si]
        shl     si,1            ;convert to index for frame_memptr
        pop     _frame_memptr[si]
expand_ret:
        popa
        clc
        ret

	;-------------------------------------
	; 44H - Map/Unmap Memory
i67f44:
        call    check_handle		; can't map memory w/out a handle
        jc      f45_err
	call	map_memory
	; return error code to user
	jmp	i67done

;---------------------------------------------------
;  Map/Unmap memory
;
;  On Entry
;	AL = physical page (0-3)
;	BX = logical page (0-n, FFFF=Unmap)
;  	DX = handle
;  Returns
;	AH = 0 (worked) or # (error)
;
	public	map_memory
map_memory:
	; is the physical page OK?
	cmp	AL, PAGESPERFRAME
	jb short map1
	mov	AH, EMSERR_BADPHYS
	ret
map1:
        push    ebx
        push    ecx
        push    edx
        mov     cl,al           ;save EMS physical page
	; a logical page of FFFF unmaps the Physical Page
        cmp     bx,-1
        jne short map2
;unmap the page, restore it to 1:1 mapping
        mov     dl,bl           ;set dl to -1 as a flag
        jmp short map3

map2:
	; does the logical page exist for this handle?
        call    find_page
;CF = 1 if error, physical address in edx if found
        mov     AH,EMSERR_BADPAGE
        jc short map_mem_done           ;jmp if error

map3:
        movzx   ebx,cl                  ; mappable page number
        shl     bx,1                    ;index for words
        mov     bx,_frame_seg[bx]       ;get the segment to remap
        shl     ebx,4                   ;convert to linear address
        or      dl,dl                   ;dl is nonzero if bx was -1
        jz short map4
        mov     edx,ebx                 ;map 1:1 if original bx was -1
map4:
        push    cx                      ;save original al
        mov     ecx,16*1024     ;size
        mov     al,1
        call    call_VMM                ;returns CF = 1 if failure
        pop     ax                      ;restore al
        mov     ah,0                    ;indicate no error
	; did everything work?
        jnc short map_mem_done
        mov     AH,EMSERR_HARDWARE
map_mem_done:
        pop     edx
        pop     ecx
        pop     ebx
        ret

; looks for a handle in the buffer_1 array. This buffer is used to hold
; both LIM 3.2 mapping contexts and handle names
;
; ->:	dh	buffer_1 entry type of interest, one of:
;
;		    BUFFER_1_MAP_CONTEXT (0) if looking for a named context
;		    BUFFER_1_HANDLE_NAME (1) if looking for a named handle
;
;	dl	handle to look for
;	ds	DGROUP
;
; <-:	CF 	set:   handle not found, DI trashed
;	CF	clear: handle found, DS:DI points to handle's slot in buffer_1
;
FindHandleInBuffer1	proc

	push	cx

        mov     di,offset DGROUP:buffer_1
        mov     cx,8                    ;max of 8 entries in f47_buf

    FHIB1_loop:
        cmp     dx,ds:[di]              ;check handle from buffer_1/fn type
        je short FHIB1_exit             	;jmp if we found the specified handle
        add     di,10                   ;increment to next entry in buffer_1
        loop    FHIB1_loop
	stc

    FHIB1_exit:
	pop	cx
	ret
FindHandleInBuffer1	endp


	;-------------------------------------
	; 45H - Free Memory (and Handle)
i67f45:
        call    check_handle		; first make sure handle is valid
        jc short f45_err                 ; error if invalid handle

	; Now make certain the specified handle is not the key
	; to a saved LIM 3.2 EMS mapping context. Can't delete
	; a handle with a saved context.
	;
	; dh = 0 (BUFFER_1_MAP_CONTEXT) since handle already validated
	; dl = valid handle number <= 255
	;
	push	di
	call	FindHandleInBuffer1
	pop	di
	jc short f45_free_handle
	mov	ah, EMSERR_SAVED_CONTEXT
	jmp	i67done

    f45_free_handle:
        push    dx
        dec     dx                      ; handles are off by one
        btr     handle_list,dx          ; deallocate the handle (clear its bit)
        pop     dx			; CF set IFF handle was in use
        jnc short f45_err               	; jmp if invalid handle

	call	free_pages
	jmp	i67ok

	; This label is used by other functions
	;
    f45_err:
	mov	AH, EMSERR_BADHANDLE
	jmp	i67done

	;-------------------------------------
	; 46H - Get EMM Version
i67f46:
	mov	AX, 40H		; EMS version 4.0
	jmp	i67done

	public i67f47
	;-------------------------------------
	; 47H - Save Mapping Context
i67f47:
        call    check_handle
        jc      f45_err                 ;jmp if invalid handle

        push    bx
        push    cx
        push    edx
        push    di
        push    es
        push    ax
        push    ds
        pop     es              ;es and ds are both DGROUP

	; First make certain the specified handle is not already
	; the key to a saved LIM 3.2 EMS mapping context. Can't
	; save multiple LIM 3.2 mapping contexts to the same handle.
	;
	; dh = 0 (BUFFER_1_MAP_CONTEXT) since handle already validated
	; dl = valid handle number <= 255
	;
	call	FindHandleInBuffer1
	jc short f47_alloc_buffer_1_slot
        pop     ax
	mov	ah, EMSERR_CONTEXT_EXISTS
	jmp short f47_done

	; The specified handle is not currently stored in the buffer,
	; try to alloc a new spot
	;
    f47_alloc_buffer_1_slot:
        call    alloc_buff
        mov     si,di
        jnc short f47_1

	; there are no more entries available in the buffer_1
	;
    f47_err1:
        pop     ax
        mov     ah, EMSERR_MAPFULL
        jmp short f47_done


	; es:di points to buffer_1 entry to use.
	;
    f47_1:
        mov     ax,dx
        stosw                   	;save the handle/fn type
        mov     cx,_frame_seg
        shr     cx,8                    ;convert frame_seg to phys addr >> 12
        mov     bx,4                    ;4 EMS pages in the frame

	; CX holds the CPU page table index of the 1st mappable
	; page in the LIM 3.2 page frame. VMM function 6 reads
	; the page table entry for that mappable page and returns
	; in EDX the starting physical address of the memory it
	; contains.
	;
	; NOTE:	The combination of word entries in the buffer_1
	;	array to store the physical address of the EMS
	;	page combined with the 4K alignment of our XMS
	;	blocks restrict us to:
	;
	; 	2^(12+16) = 2^28 = 256M of EMS.
	;
    f47_loop:
        mov     al,6
        call    call_vmm

	shr	edx, 12 		; lose 12 LSBs, all of which are zero
        mov     ax,dx
        stosw
        add     cx,4                    ; point to next mappable page
        dec     bx			; count this one
        jnz     f47_loop

        pop     ax
        mov     ah,0                    ;indicate no error

    f47_done:
        pop     es
        pop     di
        pop     edx
        pop     cx
        pop     bx
        jmp     i67done


	;-------------------------------------
	; 48H - Restore Mapping Context
i67f48:
        call    check_handle
        jc      f45_err                 ;jmp if invalid handle
        push    ebx
        push    ecx
        push    edx
        push    si
        push    ax

	; First make certain the specified handle is the key to
	; a saved LIM 3.2 EMS mapping context.
	;
	; dh = 0 (BUFFER_1_MAP_CONTEXT) since handle already validated
	; dl = valid handle number <= 255
	;
	push	di
	call	FindHandleInBuffer1	; returns buffer_1 slot in ds:di
	mov	si, di			; put possible found slot offset in si
	pop	di
	jc short f48_err 		; it is an error if entry not found

	; ds:si points to the buffer_1 entry to use
	;
f48_l1:
	mov	word ptr ds:[si], 0	; delete the mapping context now
	add	si, 2			; skip over handle and slot type ID
        mov     ecx,16*1024             ; size for VMM call
        movzx   ebx,word ptr _frame_seg ; real mode segment of starting point
        shl     ebx,4                   ; convert to 32 bit
        mov     ah,4                    ; alter 4 EMS pages

	; On entry to call_VMM, EDX will have no more than 28
	; significant bits and a 4K alignment (see comments on
	; function 47h.)
	;
f48_loop:
        mov     al,1                    ;Alter page directory entries
        movzx   edx,word ptr ds:[si]    ;page #
	shl	edx,12
        call    call_VMM                ;remap pages
        add     si,2                    ;point to next entry in buffer_1
        add     ebx,ecx                 ;point to next linear address
        dec     ah
        jnz     f48_loop
        pop     ax
        mov     ah,0

f48_done:
        pop     si
        pop     edx
        pop     ecx
        pop     ebx
        jmp     i67done

	; handle was not found in the buffer, indicate failure
	;
f48_err:
        pop     ax
        mov     ah, EMSERR_NO_SUCH_CONTEXT
        jmp     short f48_done


;Call with handle in dx, returns CF = 1 if invalid handle
;
check_handle	proc
	or	dx, dx			; OR op clears CF
	jz short CH_handle_validated	; we don't allow access to handle 0
	or	dh, dh			; handle number must be <=255
	jnz short CH_handle_validated	; return CF set if handle invalid

	push    dx
        dec     dx
        bt      handle_list,dx		; sets CF if handle's bit is 1
        pop     dx

    CH_handle_validated:
        cmc
        ret
check_handle	endp


	;-------------------------------------
	; 4BH - Get number of handles currently in use.
i67f4B:
        push    cx
        mov     cx,MAXHANDLES-1
        xor     bx,bx
loop4b:
        bt      handle_list,cx
        jnc short f4b_1
        inc     bx                      ;increment number of handles in use
f4b_1:
        dec     cx
        jns     loop4b          ;we need a loop that includes cx=zero
        pop     cx
        jmp     i67ok

	;-------------------------------------
	; 4CH - Get Pages For This Handle
i67f4C:
        call    check_handle
        jc      f45_err                 ;jmp if bad handle
	call	get_pages
	jmp	i67ok

	;-------------------------------------
	; 4DH - Get Pages For All Handles
	; On Entry
	;   ES:DI = ptr to buffer
	;  On Exit
	;   BX = # of handles
i67f4D:
        mov     bx,_frame_seg
        or      bx,bx
        jz      i67ok           ; jmp if no frame was specified

	push	CX
	push	DX
	xor	CX, CX		; CX = total # of handles allocated
	xor	DX, DX		; DX = current handle #

dump_again:
	inc	DX		; start at handle # 1
	cmp	DX, MAXHANDLES	; don't look for more handles than we allocated
	ja short done_dump

	call	check_handle	; don't consider unopened handles
	jc	dump_again

	call	get_pages	; handle may own 0 pages and that's OK
	inc	CX		; count this open handle
	mov	AX, DX
	stosw			; store handle # in user space
	mov	AX, BX
	stosw			; store # of pages for this handle
	jmp	dump_again

done_dump:
	mov	BX, CX		; store # of handles in BX
	pop	DX
	pop	CX
	jmp	i67ok

comment &
        4EH - Page Map Array (Get - Set - Get/Set - Size )

        al = 0 if getting mapping registers.
                es:di points to array to receive data.

        al = 1 if setting mapping registers.
                ds:si points to array holding data.

        al = 2 if getting and setting registers.
                es:di points to array to receive current settings.
                ds:si points to array holding new settings.

        al = 3  if getting the size of page-mapping array.
                returns bytes in al.

        Page map contains a byte describing the number of entries, followed by
        a series of entries.  Each entry contains a byte specifying the
        physical page # followed by a word which is bits 12 - 28 of the page
        table entry.
&
i67f4E:
        push    ax
        or      al,al
        jz short f4e_0
        dec     al
        jz      f4f_1
        dec     al
        jz short f4e_2
        dec     al
        jz short f4e_3
f4e_err:
        pop     ax
	jmp	i67_subfn_not_supported

f4e_0:  ;put page map at es:di
        mov     si,offset DGROUP:f4e_list       ;list of mappable segments
        push    bp
        jmp short f4f_0a

f4e_2:  ;first save page map to es:di, then restore the page map at ds:si
        mov     si,offset DGROUP:f4e_list       ;list of mappable segments
        call    get_page_map
        jmp     f4f_1

f4e_3:  ;return size of page map array in al
        pop     ax
        mov     al,0dh                  ;13 bytes in a complete page map
        jmp     i67ok

comment &
 4FH - Get - Set Partial Page Map

        For function 4f00h, ds:si points to a structure as follows:
        word    number of entries
        word    entry1
        word    entry2......
        each entry is a real mode segment which must be aligned with logical
        pages in the page frame or must be a "mappable segment" as returned by
        function 5800h.
&

i67f4F:
        push    ax
        or      al,al
        jz 	short f4f_0
        dec     al
        jz      f4f_1
        dec     al
        jnz 	short f4e_err

	; get size of partial page map
	;
f4f_2:
        mov     si,bx                   ;number of mappable segments
        imul    si,3                    ;3 bytes/segment
        inc     si                      ;one extra byte
        pop     ax                      ;original ax is garbage
        mov     ax,si                   ;return value in al
        jmp     i67ok

f4f_0:  ; get partial page map, ds:si list of segments
        ; es:di save area for page map
	;
        push    bp
        mov     bp,sp
        lds     si,ss:[bp+4]            ;restore original ds:si

f4f_0a:
        call    get_page_map
        pop     bp
        pop     ax
        jc short f4f_0_err               ; jmp if error
        jmp     i67ok

f4f_0_err:
        mov     ah, EMSERR_DATA_CORRUPT
        jmp     i67done



comment &

 Given an EMS_MAPPABLE_SEG_LIST returns an appropriate DL_PAGE_MAP_LIST
 that can later be used by set_page_map to restore the mapping context
 of the specified mappable pages. See EMM386.INC for a description of
 the data structures.

 NOTE: 	The actual translation from mappable EMS page addresses
	to physical page addresses is done by the VMM by reading
	the active page table for the V86 machine.


 ->:	ds:si 	ptr to the EMS_MAPPABLE_SEG_LIST
	es:di	ptr to the location at which to store the DL_PAGE_MAP_LIST

 <-:	CF	clear on success, set if unmappable segment found in list

&
assume fs:DGROUP,ds:nothing
get_page_map	proc near
        push    bx
        push    cx
        push    di
	push	bp
        push    fs
        push    ds
        call    getDGROUP
        push    ds
        pop     fs
        pop     ds
        cld
        lodsw                           ; ax = number of segment list entries
	or	ax, ax
	jz	short get_page_done	; getting 0 pages is a NOP not an error

        mov     bp,ax			; BP will be our outer loop counter
        stosb

    get_page_loop1:
        lodsw                           ; get first/next segment in the list
        xor     bx, bx
        mov     cx, f4e_list            ; get number of our mappable segments

    get_page_loop2:
        cmp     ax, _frame_seg[bx]      ; search for this segment
        je 	short get_page_found    ; jmp if we found the specified segment
        lea     bx, [bx+2]
        loop    get_page_loop2

    	; invalid segment, indicate error and return
	;
    get_page_err:
    	stc
	jmp	short get_page_done

	;ax is segment, bx is 2 * physical page number
	;
    get_page_found:
        shr     bx, 1
        movzx   cx, ah          ;linear address shr 12
        mov     al, bl
        stosb                   ;physical page number stored in page map
        push    edx
        mov     al, 6
        call    call_VMM        ;returns physical address in edx
        shr     edx, 12          ;eliminate 12 lsbs
        mov     ax, dx
        stosw                   ;second part of entry
        dec     bp
        pop     edx
        jnz     short get_page_loop1  	; loop until bp = 0
        clc           		        ; indicate success

    get_page_done:
        pop     fs
	pop	bp
        pop     di
        pop     cx
        pop     bx
        ret
get_page_map	endp

comment &

 Sets the EMS mapping context according to the specified DL_PAGE_MAP_LIST.

 NOTE:	The VMM establishes the new mapping context by altering
	entries in the current page table for the V86 machine.


 ->:	ds:si	ptr to the DL_PAGE_MAPLIST containing the new mapping context

 <-:

        Maps the page frame as indicated by the page map pointed to by ds:si.
        Returns ah = error code, trashes bp.
&
set_page_map:
        push    ebx
        push    ecx
        push    edx
	push	bp
        push    fs
        push    ds
        call    getDGROUP
        push    ds
        pop     fs
        pop     ds
        cld
        lodsb                           ; al = number of entries in list
        cmp     al, byte ptr f4e_list   ; check if valid number of entries
        movzx   bp, al
        ja 	short set_p_err         ; jmp if too many entries

set_p_loop1:
        lodsb                           ; physical page number
        cmp     al,byte ptr f4e_list    ; check if valid page number
        jae 	short set_p_err

        movzx   ebx,al
        shl     bx,1
        mov     bx,_frame_seg[bx]       ; get the segment to map in
        shl     ebx, 4                  ; convert to 32 bit address
        mov     ecx, 16*1024            ; map a 16K region
        lodsw                           ; physical address shr 12
        movzx   edx, ax
        shl     edx, 12                 ; physical address
        mov     al, 1
        call    call_VMM		; map in the page
        dec     bp			; one less page to map...
        jnz     set_p_loop1

        pop     fs
	pop	bp
        pop     edx
        pop     ecx
        pop     ebx
        mov     ah, 0                   ; no error
        ret

set_p_err:      ;corrupt page map
        pop     fs
	pop	bp
        pop     edx
        pop     ecx
        pop     ebx
        mov     ah, EMSERR_DATA_CORRUPT
        ret


	assume ds:DGROUP,fs:nothing


f4f_1:  ;restore a page map indicated by ds:si
        push    bp
        mov     bp,sp
        lds     si,ss:[bp+4]            ;restore original ds:si
        call    set_page_map
        mov     bp,sp
        mov     ss:[bp+3],ah            ;save error code
        pop     bp
        pop     ax
        jmp     i67done

        ;-------------------------------------
        ; 500xH - Map/Unmap Multiple Pages
        ; ds:si points to a map, cx = number of entries.

i67f50:
        call    check_handle		; can't map memory w/out a handle
        jc      f45_err

	cmp	AL, 1
        jbe 	short f50_x
        jmp     i67_subfn_not_supported

f50_x:  ;set mapping according to page map
	; ReMap386 may need lots of stack space
	call	switch_stacks

	; save registers we will change
	push	ES
	push	DI
	push	BX
	push	CX
        push    bp
        mov     bp,ax           ;save al in bp

	; load user DS:SI into ES:DI
	les	DI, [user_stack]
	les	DI, ES:[DI]		; user DS:SI now in ES:DI

	; no error if nothing to do
	;
	mov	AH, 0
	jcxz short f50_done

	; get pages
	;
f50_again:
        test    bp,1
        mov     AX,ES:[DI+2]    ;physical page if al=0, segment if al=1.
        jnz short f50_1         	;jmp if segments rather than logical pages

f50_1_ret:
        or      ah,ah
	mov	BX, ES:[DI]	; logical page
        jnz short f50_bad_mappable_page_err
	add	DI, 4		; on to next entry

	; actually DO the map/unmap
	;
	call	map_memory
	or	AH, AH
        jnz short f50_done
	loop	f50_again	; continue if no error


	; restore registers we trashed
	;
f50_done:
        pop     bp
	pop	CX
	pop	BX
	pop	DI
	pop	ES

	; switch back to user stack and return error code to user
	;
	call	restore_stack
        jmp     i67done

f50_bad_mappable_page_err:
        mov     ah, EMSERR_BADPHYS
        jmp     f50_done

f50_1:  ;ax contains a segment, convert to a logical page.
        push    cx
        mov     cx,f4e_list     ;number of mappable segments
        xor     bx,bx
f50_1_loop:
        cmp     ax,ds:_frame_seg[bx]
        je short f50_1_found    ;jmp if we found the segment
        add     bx,2
        loop    f50_1_loop
;error, segment not found
        pop     cx
        mov     ah,8bh
        jmp     f50_done

f50_1_found:    ;bx contains physical page << 1
        shr     bx,1            ;convert to physical page
        mov     ax,bx
        pop     cx
        jmp     f50_1_ret


comment ~

	BAD BUG!

	Function 51h does not work if the handle is being
	reallocated upward from m to n pages and if there are
	free pages in the frame_table which have a lower index
	than that of the handle's logical page 0 at the time of
	the reallocation request.

	The problem is that the first free page in the
	frame_table becomes the handle's new logical page 0
	instead of logical page m+1 as it should. This means
	that the data in all logical pages 0..m is effectively
	lost.

end_comment ~

	;-------------------------------------
	; 51H - Reallocate Pages, dx = handle, bx = requested number of pages
	; returns bx = total number of pages currently allocated, regardless of error.
i67f51:
        call    check_handle
        jc      f45_err
        mov     si,bx           ;save number of pages requested
        call    get_pages       ;returns bx = number of pages currently owned
        sub     si,bx           ;requested minus currently owned
        jb short f51_dealloc    	;jmp if reducing the number of pages

	; The request is to increase the number of pages
	; allocated to the handle in dx.
	;
        cmp     si,[_frame_avail]
        jb short f51_a

	; not enough pages currently available, can we allocate some XMS?
	;
        mov     bx,si           ;number of additional pages needed
        call    expand_ems      ;returns CF = 0 if success
        jnc short f51_a

        mov     ah, EMSERR_GT_FREE
	cmp	bx, [_frame_pages]
	jbe short f51_return_cur_pages
	mov	ah, EMSERR_GT_TOTAL

f51_return_cur_pages:
        call    get_pages       ;return bx = number of pages for this handle
        jmp     i67done

f51_a:
        mov     bx,si           ;number of additional pages to allocate
        call    alloc_pages     ;allocate the pages

f51_ret:
        call    get_pages       ;return bx = number of pages for this handle
        jmp     i67ok

f51_dealloc:    ;Reduce the number of pages owned by the handle.
        push    di
        neg     si
;si = number of pages to return to the EMS pool.  We must deallocate from
;the top down so that logical page 1 remains logical page 1.
        add     _frame_avail,si
        mov     di,_frame_pages         ;number of entries
        shl     di,1                    ;two bytes/entry
        add     di,_frame_table         ;point to just above top of table
f51_b:  sub     di,2                    ;point to next entry down
        cmp     word ptr [di],dx
        jnz     f51_b                   ;jmp if not this handle
        mov     word ptr [di],0         ;deallocate the page
        dec     si                      ;one less page to deallocate
        jnz     f51_b                   ;loop if not finished
        pop     di
        jmp     f51_ret

i67f52: ;get/set handle attributes, only volatile handles are
        ;supported by MS-EMM386.  We do not support non-volatile handles
        ;either.  Function 5202h calls can be handled by returning ax = 0.
	;fns 5200h and 5201h must return 91h as does MS EMM386.EXE
	;
	cmp	al, 2
	mov	ax, 0			; only volatile handles supported
	ja	i67_subfn_not_supported
	je short f5202
	mov	ah, EMSERR_NO_CAPABILITY
    f5202:
        jmp     i67done

comment &
        Function 5300h, get handle name, es:di points to buffer.
        Function 5301h, set handle name, ds:si points to buffer.
        DX = handle for both functions.
&
i67f53: cmp     AL,2
        jae     i67_subfn_not_supported       ;jmp if invalid subfunction
        call    check_handle
        jc      f45_err                 ;jmp if bad handle
        push    bx
        or      al,al
        jnz short f53_1

	; function 5300h, get handle name, es:di points to 8
	; byte user buffer in which to receive name.
	;
        call    find_handle
        jc short f53_err1               	; jmp if handle is not named
        push    si                      ; bx points to the entry in buffer_1
        push    di
        lea     si,[bx+2]               ; point to name
        movsd
        movsd                           ; move 8 bytes
        pop     di
        pop     si
        pop     bx
        jmp     i67ok

f53_err1:
        mov     ah, EMSERR_NO_HANDLE_NAME  ; error, no such handle name

f53_err2:
        pop     bx
        jmp     i67done

f53_err3:
        pop     di
        pop     es
        jmp     f53_err2

	; function 5301h, set handle name, ds:si points to 8 byte buffer.
	; first check for duplicate handle name.
	;
f53_1:
        push    es
        push    di
        mov     bx,sp
        les     di,ss:[bx+6]            ; original ds:si is now in es:di
        call    find_name
        mov     ah, EMSERR_DUP_HANDLE_NAME
        jnc     f53_err3                ; jmp if find_name was successful

        call    find_handle
        jnc short f53_1_a               	; jmp if handle is already named
        call    alloc_buff		; returns CF set if no more slots

        mov     ah, EMSERR_INTERNAL     ; internal error if we jump below
        jc 	f53_err3               	; jmp if no more room in buffer

	; es:di points to buffer entry to use. we must load the
	; original ds:si
	;
f53_1_b:
        mov     bx,sp
        lds     si,ss:[bx+6]            ;restore original ds:si
        inc     dh                      ;set dh = 1
        mov     es:[di],dx              ;set handle and attribute bytes
        add     di,2                    ;point to space for handle name
        mov     bx,ds:[si]              ;get the first word
        movsw
        or      bx,ds:[si]              ;or with the second word
        movsw
        or      bx,ds:[si]              ;or with the third word
        movsw
        or      bx,ds:[si]              ;or with the fourth word
        movsw                           ;move the final word

	; The EMS specification dictates that a null name (a
	; name of all 0s) is the same as no name. If we just set
	; the EMS handle name to a null name then we must set
	; must deallocate the space for saving the name by
	; zeroing out the name's handle and attribute in the
	; buffer_1 array. If the new name is null then ZF==1 and
	; BX==0.
	;
        jnz short f53_1_c
        mov     es:[di-10],bx           ;zero the handle and attribute

f53_1_c:
        mov     dh,0                    ;restore dx to a valid handle
        pop     di
        pop     es
        pop     bx
        jmp     i67ok

f53_1_a:
        push    ds
        pop     es
        mov     di,bx
        jmp     f53_1_b

comment &
        5400 Get handle directory, es:di points to buffer,return al = number
                of entries in directory.
        5401 Search for named handle, ds:si points to name, returns dx =
                handle.
        5402 Get total number of handles, MAXHANDLES.
&
i67f54:
        cmp     al,2
        ja      i67_subfn_not_supported
        je      short f54_2
        or      al,al
        jnz short f54_1

	; function 5400h, get handle directory in buffer pointed to by es:di
	;
        push    bx
        push    dx
        push    di
        push    bp
        push    eax
        xor     dx,dx
        xor     bp,bp

f54_loop:
        inc     dl                      ;check next handle
        jz short f54_loop_done          ;check 255 handles
        call    check_handle
        jc short f54_loop               ;jmp if handle is invalid

	; handle is valid, we must enter it in the array and
	; increment counter in bp
	;
        inc     bp
        mov     ax,dx
        stosw

	; now we need to enter an 8 byte name, it will be all zeros
	; if handle is un-named
	;
        call    find_handle             ;returns bx pointing to name
        jnc short f54_name_found

	; the handle is not named
	;
        xor     eax,eax
        stosd
        stosd                           ;8 bytes of zeroes for un-named handle
        jmp     f54_loop

f54_name_found:
        lea     si,[bx+2]               ; bx points to entry, point si to name
        movsd
        movsd                           ;transfer 8 bytes for handle name
        jmp     f54_loop

f54_loop_done:
        pop     eax
        mov     ax,bp                   ;number of handles in directory
        pop     bp
        pop     di
        pop     dx
        pop     bx
        jmp     i67done                 ;ax is already set to its return value

f54_1:  ;Search for a named handle pointed to by ds:si, return handle in dx
        push    bp
        mov     bp,sp
        push    bx
        push    es
        push    di
        les     di,ss:[bp+2]            ;original ds:si
        call    find_name
        jc short f54_1_err              ;jmp if name was not found

	; bx points to the entry containing the name
	;
        movzx   dx,byte ptr ds:[bx]     ;get the handle, zero the attribute byte
        mov     ah,0                    ;indicate success

f54_1_done:
        pop     di
        pop     es
        pop     bx
        pop     bp
        jmp     i67done

      	;  the specified name was not found
f54_1_err:
        mov     ah, EMSERR_NO_HANDLE_NAME
        jmp     f54_1_done

f54_2:  ; get total number of handles in the system, allocated and unallocated.
	;
        mov     bx, MAXHANDLES
        jmp     i67ok

comment &
        Allocates a 10 byte buffer in buffer_1.  Returns a pointer to the
        buffer in es:di and CF=0 if success.  Returns CF=1 if failure.  Must
        be called with DS = DGROUP.  Note, this does NOT actually mark the
        buffer as allocated, the caller must insert a nonzero handle in the
        buffer to mark it as allocated.
&
alloc_buff:
        push    cx
        push    ds
        pop     es
        mov     di,offset DGROUP:buffer_1[-10]
        mov     cx,8                    ;number of entries in buffer_1
alloc_loop:
        lea     di,[di+10]              ;point to next entry
        cmp     ch,ds:[di]              ;sets CF=1 and ZF=0 if nonzero
        loopne  alloc_loop
        pop     cx
        ret

comment &
        Checks to see if the handle in dx is a named handle.  If it is named,
        it returns with bx pointing to the entry in buffer_1.  Returns CF = 0
        if success, CF = 1 if failure.  Must be called with DS = DGROUP
&
find_handle:
	push	di
	mov	dh, BUFFER_1_HANDLE_NAME
	call	FindHandleInBuffer1
	mov	bx, di
        mov     dh,0                    ;high word of handles is always zero
        pop     di
        ret

comment &
        Searches for the name pointed to by es:di.  Returns bx pointing to the
        entry in buffer_1 and CF=0 if success.  Returns CF=1 if failure.
        Must be called with DS = DGROUP.
&
find_name:
        push    ax
        push    cx
        push    si
        push    bp
        mov     bx,offset DGROUP:buffer_1
        mov     cx,8                    ;number of entries in buffer_1
        mov     bp,di                   ;storage for di
        cld
find_n_loop:
        mov     ax,ds:[bx]              ;get handle and attribute
;it is a valid named handle entry if ah=1 and al != 0
        dec     ah
        jnz short find_n_1              ;jmp if not a named handle
        or      al,al
        jz short find_n_1               ;jmp if not allocated
;the entry is a valid named handle entry, check if the name matches.
        lea     si,[bx+2]               ;point to name
        push    cx
        mov     cx,2                    ;2 dwords in the name
        repe    cmpsd
        pop     cx                      ;restore cx
        mov     di,bp                   ;restore di
        je short find_n_2
find_n_1:
        add     bx,10                   ;point to next entry
        loop    find_n_loop
;There are no matches.
        stc                             ;indicate error
find_n_2:       ;Must preserve CF!
        pop     bp
        pop     si
        pop     cx
        pop     ax
        ret

f55_err:
        mov     ss:[bp-3],ah            ;error code
        pop     eax
        pop     bp
        jmp     i67done

comment &
        Alter page map and jump.  If al = 0, the page map is physical page
        numbers, if al = 1, the page map contains real mode segment addresses.
        DX = handle, ds:si points to a structure as follows:

        dd      far pointer to target address
        db      number of entries in page map
        dd      far pointer to page map
&
i67f55:
        cmp     al,1
        ja      i67_subfn_not_supported ;jmp if invalid subfunction
        push    bp
        mov     bp,sp
        push    eax
        lds     si,ss:[bp+2]            ;load original ds:si
        add     si,4                    ;point to number of entries to alter
        call    map55
        or      ah,ah                   ;check if error
        lds     si,ss:[bp+2]            ;restore original ds:si
        lodsd                           ;target address
;CAUTION!! must preserve flags from or ah,ah above for instruction below.
        jnz short f55_err
;re-mapping has been done, return to the target address
        mov     ss:[bp+6],eax           ;save target address on stack
        pop     eax
        pop     bp
        jmp     i67ok

comment &
&
i67f56:
        cmp     al,2
        ja      i67_subfn_not_supported
        je short f56_2
        push    bp
        mov     bp,sp
        push    eax
        lds     si,ss:[bp+2]            ;load original ds:si
        add     si,4                    ;point to number of entries to alter
        call    map55
        or      ah,ah                   ;check if error
        lds     si,ss:[bp+2]            ;restore original ds:si
;CAUTION!! must preserve flags from or ah,ah above for instruction below.
        jnz short f55_err
;re-mapping has been done, prepare the stack for the far call.
        lodsd                           ;target address
        push    dx                      ;save handle for return.
        push    cs
        push     offset f56_ret         ;return address
        push    word ptr ss:[bp+10]     ;flags
        push    eax                     ;far address to call
        lds     si,ss:[bp+2]            ;restore original ds:si
        mov     eax,ss:[bp-4]           ;restore original eax
        mov     ah,0                    ;error code
        mov     bp,ss:[bp]              ;restore original bp
        iret                            ;call function, returns below

f56_ret:
        push    bp
        mov     bp,sp
        add     bp,8                    ;point to old bp on stack
        pop word ptr ss:[bp]            ;save new bp
        xchg    ss:[bp-6],dx            ;get handle, save new dx
        xchg    eax,ss:[bp-4]           ;get original eax, save new eax
        push    dword ptr ss:[bp+2]     ;original ds:si
        mov     ss:[bp+2],si
        mov     ss:[bp+4],ds            ;save new ds:si
        pop     si
        pop     ds
        add     si,9                    ;point to number of entries for restore
        call    map55                   ;returns ah = error code
        pop     dx
        pop     eax
        pop     bp
        jmp     i67done

f56_2:  ;get the stack space required for function 5600 and 5601.
;MS-EMM returns incorrect number, this matches 386MAX
        mov     bx,22
        jmp     i67ok

comment &
        map55 below is used for functions 55 and 56h.  Call with ds:si
        pointing to a byte indicating the number of entries, followed by a
        dword pointer to a page map consisting of word pairs, the first word
        being the logical page number, the second being either the physical
        page number or the physical segment corresponding to a mappable page.
        AL specifies physical page numbers if 0, segments if 1.  DX contains
        the handle.  Trashes ds:si and ax.  Ah returns 0 if success, error
        code if page map or handle is invalid.
&
map55	proc	near

        push    bx
        push    cx
        or      al,al
        lodsb                   ;number of entries in the page map
        movzx   cx,al
        lds     si,ds:[si]      ;pointer to page map
;CAUTION!!!! flags from or al,al must be preserved for instruction below!!!
        jnz 	short map55_segs_check_count

	; Mapping in zero pages is a NOP, not an error
	;
    	jcxz short map55_done

	; the page map contains mappable page numbers
	;
    map55_phy_loop:
        lodsw                   ;logical page number
        mov     bx,ax
        lodsw                   ;physical page number
        push    ds
        call    getDGROUP
        call    map_memory      ;returns ah = 0 if okay
        or      ah,ah
        pop     ds
        loope   map55_phy_loop
	jmp	short map55_done

    	; The page map contains real mode segments rather than
	; logical pages. Mapping in zero pages is a NOP, not an error.
	;
    map55_segs_check_count:
    	jcxz short map55_done

    map55_segs:
        push    ds
        lodsw                   ;logical page number
        push    ax
        lodsw                   ;physical segment
        call    getDGROUP

	; search for real segment in mappable segments
	;
        mov     bx,6            ;correspond to physical page number 3

    map55_seg_loop:
        cmp     _frame_seg[bx],ax
        je 	short map55_seg_found
        sub     bx,2            ;point to next lower physical page
        jae     map55_seg_loop  ;loop until we find it

	; we have an error, the specified physical segment is
	; not a mappable segment.
	;
        pop     ax
        pop     ds
        mov     ah,8bh          ;invalid physical page
        jmp 	short map55_done

    map55_seg_found:
        pop     ax              ;logical page number
        shr     bx,1            ;convert to physical page number
        xchg    ax,bx
        call    map_memory      ;returns ah = 0 if okay
        pop     ds
        or      ah,ah
        loope   map55_segs

    map55_done:
        pop     cx
        pop     bx
        ret
map55	endp


comment & ----------------------------------------------
        570xh - Move or exchange memory region
           0h - move memory
           1h - exchange memory region.
           ds:si points to EMS_MOVEX_STRUCT structure defined above
&
i67f57:
        sti                     	; big transfers may take a lot of time
        pop     si
        pop     ds              	; restore original pointers
        push    ebx
        push    ecx
        push    edx
        push    edi
        push    eax

	; verify 0 < (length of region to move/exchange) <= 1M
	;
	mov	ecx, ds:[si].regionSize
	or	ecx, ecx
	je short f57_ret			; 0 length is a NOP
	cmp	ecx, 1024*1024
	jbe short f57_check_mem_types

	mov	cl, EMSERR_LEN_GT_1M	; length cannot be > 1M error
	jmp short f57_ret

	; verify both mem types in {0,1}
	;
    f57_check_mem_types:
	mov	ah, 1
        cmp     al, ah			; verify supported sub-function
	ja short f57_bad_subfunc		; jump if not move or exchange fn

	cmp	ds:[si].sourceMemType, ah
	ja short f57_bad_memtype
	cmp	ds:[si].destMemType, ah
	jbe short f57_verify_source_args

    f57_bad_memtype:
	mov	cl, EMSERR_BAD_MEM_TYPE	; either mem type not in {0,1}
	jmp	short f57_ret

    f57_bad_subfunc:
	mov	cl, EMSERR_BAD_SUBFUNC	; invalid subfunction
	jmp	short f57_ret

	; Validate both EMS and conventional (DOS) memory regions
	; See comments on CheckF57MemRange for details.
	;
    f57_verify_source_args:
	lea	bx, ds:[si].sourceMemType
	call	CheckF57MemRange
        jc short f57_ret			; cl = error code if error

	lea	bx, ds:[si].destMemType
	call	CheckF57MemRange
        jc short f57_ret			; cl = error code if error

	; When we get here we have validated the user structure
	; with the one exception that we have not checked for
	; overlap. Overlap checks are done at the beginning of fn
	; 5701h and at the end of 5700h if no other error was
	; discovered.
	;
    f57_args_validated:
	pop	eax			; get back original AX
	push	eax			; save it again
        xor     edi, edi                ; edi is counter for bytes moved so far
        cmp     al, 1
        je short f57_do_exchange         ; jmp if exchange fn (vs. move fn)

comment &
        We have to be very careful doing this transfer since the expanded
        memory is likely to not be contiguous. It is quite possible that page
        breaks in the source do not line up with page breaks in the
        destination.
&
    f57_loop:
        mov     ecx, ds:[si].regionSize ; total number of bytes in move
        sub     ecx, edi                ; number of bytes left to move
	or	ecx, ecx
        jz short f57_done

	call	F57MovexToVMMRegs	; set up regs for VMM call

	; use VMM fn 2, copy memory
	; ecx = size in bytes, ebx = source, edx = destination
	;
        mov     al,2
        call    call_VMM
        jmp     f57_loop

comment &
	When we get here, the error code (which may be 0 -- no
	error) is in CL.  If the error code is zero, we must
	still check for overlapping ranges.  The copy code
	protects against trashing the data in the case of
	overlapping ranges, but the EMS specification dictates
	that we must return an error code of 92h, to warn that
	the source has been overwritten.
&
    f57_done:
	or	cl, cl               	; if there was an error
	jnz	short f57_ret		; return it's error code

	call	CheckF57Overlap		; check for overlap
	jnc	short f57_ret
        mov     cl, EMSERR_MOVE_OVERLAP	; return warning status code if found

    f57_ret:
        pop     eax
        mov     ah, cl
        pop     edi
        pop     edx
        pop     ecx
        pop     ebx
        iret

f57_do_exchange:
	call	CheckF57Overlap		; check for overlapping regions.
	jnc	short f57_1_loop
        mov     cl, EMSERR_XCHG_OVERLAP	; they're an exchange fn fatal error
        jmp     f57_ret

    f57_1_loop:
        mov     ecx, ds:[si].regionSize ; total number of bytes in xchg
        sub     ecx, edi                ; number of bytes left to xchg
        or      ecx, ecx
        jz      f57_ret

	call	F57MovexToVMMRegs	; set up regs for VMM call

	; use VMM fn 0Ah, exchange memory
	; ecx = size in bytes, ebx = source, edx = destination
	;
        mov     al, 0Ah
        call    call_VMM
        jmp     f57_1_loop


; This helper routine for function 5700h and 5701h processes the
; EMS_MOVEX_STRUC and returns with registers set up for either a
; VMM fn 2 or a VMM fn A call to move or exchange memory
; respectively.  The bytes to move, returned in ECX, are
; guaranteed not to violate either alignment or allocation
; granularity contstraints.
;
; ->:	ecx	bytes left to move
;	edi	bytes already moved
;	ds:si	ptr to EMS_MOVEX_STRUC passed in by user
;
; <-:	CF 	set if either source or dest F57DESCRIP struc invalid
;		else CF clear and...
;
;	ebx	physical address of source
;	edx	physical address of destination
;	ecx	maximum bytes to move in this go
;	edi	updated to hold bytes moved should the VMM call succeed
;
F57MovexToVMMRegs	proc

        lea     bx, ds:[si].destMemType    	; point to dest description
        call    F57DescripToPhysAddr
	jc short f57MTVMMR_ret
        mov     edx, eax                	; edx = dest physical address
        lea     bx, ds:[si].sourceMemType  	; point to source description
        call    F57DescripToPhysAddr
	jc short f57MTVMMR_ret
        mov     ebx, eax			; ebx = source physical address

comment &

???	Why 4K? Alignment is fer sher 4K but because we lie
	about our raw page size, allocation granularity is
	always 16K! Isn't this several times as many VMM
	calls as we need?

	    1) We allocate 4K extra memory from XMS and 4K align
	       the block by chopping of any significant 12 LSBs
	       from the block address.
	    2) We partition the XMS block into (XMS block
	       size/16K) and map these partitions onto entries
	       in the page table. Each page table entry
	       represents a contiguous 16K block within some
	       large contiguous XMS block.
	    3) Only VCPI calls allocate partial page table
	       entries such VCPI allocations are taken out of a
	       page that isn't available to EMS and which
	       doesn't change the physical alignment or position
	       of subsequent pages in the page table so EMS
	       pages are still 4K aligned with 16K granularity
	    4) There is no VCPI function to move/exchange memory
	    5) The VMM fn 1 call to map memory simply loops,
	       writing adjacent physical addresses into adjacent
	       page table entries for as many raw CPU pages as
	       specified. EMS pages are always specified as
	       multiples of 16K.

        We must determine how many bytes we can transfer before we run into a
        4 kilobyte boundary. We must not cross 4K boundaries in a single
        transfer because the EMS physical pages are 4K aligned, NOT 16 K!  ECX
        contains the number of bytes we have left to move. We must pick the
        number of bytes left to move, the 4 K boundary associated with the
        source or the 4k boundary dictated by the destination, whichever of
        the three is lowest for the number of bytes to move.
&
        call    next_4k         ;returns with ecx = lowest of two limits
        mov     ax, dx          ;low word of destination
        call    next_4k         ;returns with ecx = lowest of three limits
        add     edi, ecx        ;number of bytes moved so far (clears CF)

    f57MTVMMR_ret:
	ret
F57MovexToVMMRegs	endp

comment &
        Calculates the number of bytes between ax and the next 4 K boundary,
        returns that number in cx if it is less than cx.  CX is adjusted
        downwards if required to keep the next move from crossing a 4K
        boundary.
&
next_4k:
        not     ax
        and     ax, 0fffh
        inc     ax
        movzx   eax, ax
        cmp     eax, ecx
        ja 	short next_4k_ret
        mov     ecx, eax         ;take the smaller of the two

    next_4k_ret:
        ret

; Given a pointer to an F57DESCRIP structure this routine
; returns the corresponding physical address of the start of the
; block it defines, regardless of whether that block is in EMS or
; in conventional memory. The F57DESCRIP struc is defined
; above as a subset of the EMS_MOVEX_STRUC passed by the user to
; fn 57xxh.
;
; ->:	DS:BX	ptr to the F57DESCRIP for the memory block
; 		to be moved/exchanged.
;	EDI	starting offset within that block
;
; <-:	CF clear -- EAX = physical address within the specified block
;	CF set   -- CL  = error code EMSERR_INTERNAL
;
;		Note that since erroneous page and handle
;		specifications are supposed to be detected
;		before the actual move/exchange operation
;		begins: Assuming that works, this error code
;		indicates a failure of the EMM page accounting
;		mechanism.
;
F57DescripToPhysAddr proc

	push    edx

        movzx   edx, word ptr ds:[bx].memOffset  ; get specified offset
        xor     eax, eax
        cmp     ds:[bx].memType, al
        je 	short F57DTPA_block_is_in_conventional_memory

    F57DTPA_block_is_in_EMS_memory:
        push    bx
        mov     eax, edi                ; desired offset into specified block
        add     eax, edx                ; add offset into first logical page

	; eax has the desired offset into the entire EMS block.
	;
        push    ax                      ; save low 16 bits of offset
        shr     eax, 14                 ; convert offset into logical page #
        mov     dx, ds:[bx].memHandle
        mov     bx, ds:[bx].memPageOrSegment
        add     bx, ax                  ; logical page of desired location
        push    ds
        call    getDGROUP
        call    check_handle
        jc 	short F57DTPA_error
        call    find_page               ; returns edx = physical address
        mov     eax,edx
        jc 	short F57DTPA_error   	; jmp if invalid logical page number
        pop     ds

	; EAX contains the physical address of the start of the
	; specified logical page. The remainder below 16K is on
	; the stack in word form (note: DH & 0x3F == DX & 0x3FFF)
	;
        xor     edx, edx
        pop     dx			; dx= low 16 bits of EMS block offset
        and     dh, 3Fh                 ; use bottom 14 bits for offset
        add     eax, edx                ; add the remainder
        pop     bx
	jmp	short F57DTPA_done

    F57DTPA_error:
        pop     ds      		; restore ds
        pop     ax      		; pop garbage
        pop     bx                      ; restore bx
	mov	cl, EMSERR_INTERNAL	; EMM software error
	stc				; set CF to signal error
	jmp	short F57DTPA_exit	; give up

    F57DTPA_block_is_in_conventional_memory:
        movzx   eax, word ptr ds:[bx].memPageOrSegment
        shl     eax, 4                  ; convert segment to physical address
        add     eax, edx                ; add in offset of start of block
        add     eax, edi                ; add in offset into the block

    F57DTPA_done:
	clc

    F57DTPA_exit:
        pop     edx
        ret
F57DescripToPhysAddr endp

; Checks for overlapping source and destination regions in an
; EMS_MOVEX_STRUCT used by fns 5700h (move memory region) and
; 5701h (exchange memory regions). Overlaps cause a warning
; status code to be returned for the former fn (ah=92h) but are
; a fatal error condition for the latter fn (ah=97h).
;
; We determine overlap by comparing the absolute value of the
; difference between the source and destination start addresses
; to the length of the region to be moved/exchanged. If the
; length is greater we have overlap and the subtraction will set
; the CF indicator.
;
; ->:	ds:si	ptr to the EMS_MOVEX_STRUC passed in by user
; <-:	CF	set if regions overlap.
;
CheckF57Overlap	proc	near

	push	eax
	push	bx
	push	edx
	push	edi

	mov	al, ds:[si].sourceMemType
	cmp	al, ds:[si].destMemType
	clc
	jnz	short CF57O_done	; can't overlap if different mem types

	or	al, al
	jz	short CF57O_do_check	; jump if both conventional mem types

	mov	ax, ds:[si].sourceHandle ; source and dest are both EMS memory
	cmp	ax, ds:[si].destHandle
	clc
	jnz	short CF57O_done	; can't overlap if different handles

    CF57O_do_check:
        xor     edi, edi
        lea     bx, ds:[si].destMemType
        call    F57DescripToPhysAddr
	jc short CF57O_done		; internal error if CF set
        mov     edx, eax                ; edx = physical dest start address

        lea     bx, ds:[si].sourceMemType
        call    F57DescripToPhysAddr    ; eax = physical source start address
	jc short CF57O_done		; internal error if CF set

	sub     eax, edx                ; get absolute difference between them
        jns 	short CF57O_compare_length_to_distance
        neg     eax			;

    CF57O_compare_length_to_distance:	; eax is absolute address difference
	cmp     eax, ds:[si]		; set CF if length > difference

    CF57O_done:
	pop	edi
	pop	edx
	pop	bx
	pop	eax
	ret
CheckF57Overlap	endp

; This routine checks the EMS_MOVEX_STRUCT used by fns 5700h
; (move memory region) and 5701h (exchange memory regions) for a
; variety of errors that will cause the call to fail if detected.
;
; For EMS memory verify that:
;   	handle is valid
;	offset < 16K
;	logical page starting page k is valid
;	logical page k+length/16K+((offset>0)?1:0) is valid
;
; For DOS memory verify that
;	start + offset is not within a page frame
;	start + offset + len is not within a page frame
;	start + offset + len <= 1M
;
; ->:	ds:si	ptr to the EMS_MOVEX_STRUC passed in by user
; 	ds:bx	ptr to the F57DESCRIP to verify (this is the effective
;		address of ds:[si].sourceMemPtr or of ds:[si].destMemPtr)
; <-:	CF set  memory specification invalid,
;		error code in CL
;
CheckF57MemRange proc	near
	push	ds
	push	es
	push	ds
	pop	es
	call	getDGROUP

	cmp	es:[bx].memType, 0
    	jz	short C57MR_verify_DOS_mem

    C57MR_verify_EMS_mem:
	mov	dx, es:[bx].memHandle
        call    check_handle
	mov	cl, EMSERR_BADHANDLE
        jc      C57MR_done

	cmp	es:[bx].memOffset, 16*1024
	jae short C57MR_page_offset_invalid

	push	bx
	call	get_pages		; returns BX = pages owned by handle
	mov	cx, bx
	pop	bx
	cmp	cx, es:[bx].memPageOrSegment
	jc short C57MR_invalid_logical_page

	mov	eax, es:[si]            ; get length of block
	shr	eax, 14			; length / 16K = number of pages
	xor	dx, dx
	sub	dx, es:[bx].memOffset	; set CF if offset non-zero
	adc	ax, es:[bx].memPageOrSegment ; ax=last (partial) logical page #
	cmp	cx, ax
	jnc     short C57MR_done	; block size OK

    C57MR_len_exceeds_block_size:
	mov	cl, EMSERR_LEN_GT_BLOCK	; len of move passes end of EMS block
	jmp	short C57MR_done

    C57MR_invalid_logical_page:
	mov	cl, EMSERR_BADPAGE
	jmp	short C57MR_done

    C57MR_verify_DOS_mem:
	movzx	eax, es:[bx].memPageOrSegment
	shl	eax, 4
	movzx	ebx, es:[bx].memOffset
	add	eax, ebx		; get DOS region[0] physical address
	mov	ecx, es:[si].regionSize ; get n
	mov	ebx, eax
	add	ebx, ecx		; get DOS region[n+1]
	cmp	ebx, 1024*1024		; check region[n] < 1M
	jbe short C57MR_check_first_page_frame

    C57MR_DOS_mem_wraps_at_1M:
	mov	cl, EMSERR_DOS_1M_WRAP	; DOS mem can't exceed 1M
	stc
	jmp	short C57MR_done

    C57MR_page_offset_invalid:
	mov	cl, EMSERR_BAD_PAGE_OFFSET
	stc
	jmp	short C57MR_done

    C57MR_DOS_mem_overlaps_page_frame:
	mov	cl, EMSERR_DOS_EMS_OLAP ; DOS mem can't overlap EMS
	stc
	jmp	short C57MR_done

	; If we change Datalight's EMM386 to have multiple page
	; frames then we must verify that none of them overlaps
	; with a DOS memory block specified to a move/exchange
	; function: That condition is an error for a reason. This
	; label is where we can set up registers so that the code
	; at the label below can perform its arithmetic without
	; knowing about the "keeping track of multiple page frame
	; data structures" in any way.
	;
    C57MR_check_first_page_frame:
	movzx	ebx, [_frame_seg]
	shl	ebx, 4			; ebx = start of page frame

	; This is the label where a specific page frame actually
	; gets checked for overlap using the same technique as
	; before (if length > abs(distance) then overlap). It
	; assumes that we check page frames in increasing address
	; order...
	;
	; eax = physical address of bottom byte of conventional memory block
	; ebx = physical address of bottom byte of page frame
	; ecx = length of DOS memory block to move/exchange
	;
    C57MR_get_distance:
	sub	ebx, eax                ; get distance between addresses
	jae short C57MR_check_overlap
	neg	ebx			; we want the magnitude of the distance

    C57MR_check_overlap:
	cmp	ebx, ecx                ; error if block len > distance
	jc	C57MR_DOS_mem_overlaps_page_frame

	; If we have multiple page frames then we have to see if
	; there are more to evaluate at this point. If there are
	; then, after setting up up the EBX register to point to
	; the bottom of the next page frame (and leaving EAX
	; unchanged), we loop up to C57MR_get_distance and see if
	; that page frame suffers from DOS range overlap syndrome...
	;
    C57MR_check_next_page_frame:
	;
	; single page frame, nothing to do
	;
    C57MR_mem_OK:
	xor	cx, cx			; clear CF, zero cl

    C57MR_done:
	pop	es
	pop	ds
	ret
CheckF57MemRange endp


comment &

This function either fills in a user buffer with an array of
EMS_MAPPABLE_PAGE_STRUCT structures (fn 5800h) or returns the
number of entries in the array (fn 5801h). See EMM386.INC for
the definition of the EMS_MAPPABLE_PAGE_STRUCT.  The array of
structs is called the Mappable Page Array (MPA) and there is one
entry for each location in the lower 1M of memory where an EMS
page may be mapped in. The entries in the MPA define the current
correlation between EMS mappable page numbers and the segment
addresses at which those numbered pages are located. A logical
page (which is physically in extended memory) may be mapped into
the lower 1M address space (and so become accessible to a V86
program) by specifying either a mappable page number or the
address of a mappable page as an argument to fns 44h/50h.
Mappable page numbers {0,1,2,3} specially designate (in
increasing address order) the four pages in the LIM 3.2 EMS page
frame; aside from that there is no special meaning assigned to
mappable page numbers and the EMM may apply them in any way that
is convenient. The entries in the MPA must be returned to the
user ordered by increasing segment value.

	;-------------------------------------
	; 580xH	- Get Mappable Page Array
	;
	;    0H	- Get Mappable Page Array (MPA)
	;	- Each entry is an EMS_MAPPABLE_PAGE_STRUCT
	;    1H	- Get number of entries in the MPA
&

i67f58:	cmp	AL, 0
	je short f5800
	cmp	AL, 1
	je short f5801
	jmp	i67_subfn_not_supported

f5800:
	push	DI

	mov	SI, [_frame_seg]
	mov	CX, PAGESPERFRAME
	cld
next_entry:
	mov	AX, SI			; Save segment of this page
	stosw
        mov     ax,4
        sub     ax,cx
	stosw
	add	SI, 400H		; 16k/page = 400H paras per page
	loop	next_entry

	pop	DI
f5801:
	mov	CX, PAGESPERFRAME
	jmp	i67ok


comment &

The first entry is defined as the raw page size in paragraphs,
that is the granularity of the page mapping hardware. In reality
this is 4K but, like all other flavors of EMM386, we lie and say
it is 16K (400h*10h=4000h=16K) to avoid having to deal with more
than one logical page size.

The second and third entries specify the number of Alternate
Mapping Register Sets supported by the EMM and amount of storage
required to store the mapping context (see the comments on
get_page_map for the internal structure of our context save
areas).

The last two entries do not apply to any flavor of EMM386 as
they require special hardware to perform DMA into logical pages
which are not mapped into the 1M address space. We do support
the VDMA spec which provides similar functionality.

	;-------------------------------------
	; 59H - Get EMS Hardware Information
	;
&

i67f59:
	test	fnAccessState, ACCESS_DENIED_BY_OSE
        jnz     AccessDeniedToFuntion  	; access to this fn disabled by OS/E

        cmp     al,1
        ja      i67_subfn_not_supported
        je      i67f42
        mov word ptr es:[di],400h       ; 1K paragraphs per page
        mov word ptr es:[di+2],7h       ; 7 AMRS
        mov word ptr es:[di+4],0dh      ; size of page map array
        mov word ptr es:[di+6],0        ; How many DMA registers?
        mov word ptr es:[di+8],0        ; DMA through function 5BH?
        jmp     i67ok

comment &
        Allocate standard/raw pages.  al = 0 if allocate standard pages, 1 if
        allocate raw pages.  BX = number of pages to allocate.  MS-DOS uses a
        raw page size of 16K which is equal to the standard page size, so we
        treat them identically.
&
i67f5A:
        jmp     alloc1  ;same as function 43, but skip checking for zero pages

comment &

        Alternate map register set.  Emulate 7 alternate register sets, we do
        not support alternate DMA register sets.  The code below performs in a
        similar manner to MS-EMM386.

&
i67f5B:
	test	fnAccessState, ACCESS_DENIED_BY_OSE
        jnz     AccessDeniedToFuntion	; access to these fns disabled by OS/E

        push    ax
        or      al, al
        jz 	short f5b_0		; get Alt. Mapping Reg. Set (AMRS)
        dec     al
        jz 	short f5b_1		; set AMRS
        dec     al
        jz      f5b_2			; get AMRS save area size
        dec     al
        jz      f5b_3			; alloc AMRS
        dec     al
        jz      f5b_4			; free AMRS
        dec     al
        jz      f5b_5			; alloc DMA Reg. Set (DMARS)
        dec     al
        jz      f5b_6			; enable DMARS on DMA channel
        dec     al
        jz      f5b_7_8			; disable DMARS
        dec     al
        jz      f5b_7_8			; free DMARS
        pop     ax
        jmp     i67_subfn_not_supported ;undefined subfunction

	; get alternate map register set, returns bl = active number
        ; if number is zero (no hardware AMRS), returns an es:di value also.
	;
    f5b_0:
        mov     bl, f5b_current_set    	; bl = current register set number
        or      bl, bl			; if nonzero don't return pointer
        jnz 	short f5b_done

	les     di, f5b_ptr         	; es:di = value from last 5B01h call

    f5b_done:
        pop     ax
        jmp     i67ok

    f5b_err9d:
        pop     ax
        mov     ah, EMSERR_AMRS_NOT_ALLOCED
        jmp     i67done

	; set alternate map register set, bl = number
        ; if bl = 0, es:di points to state save area
	; first save current context in the current save area
	;
    f5b_1:
        push    es
        push    di
        movzx   di, f5b_current_set
        or      di, di
        jnz 	short f5b_1_c

	; current set is zero, check if it is the original ptr which is 0
	;
        les     di, f5b_ptr
        mov     ax, es
        or      ax, di
        jnz 	short f5b_1_d           ; jmp if es:di is a nonzero pointer

	; di contains an index into the f5b_save_area
	;
    f5b_1_c:
        push    ds
        pop     es
        imul    di, SIZE DL_EMS_CONTEXT
        add     di, offset DGROUP:f5b_save_area

        ; es:di points to a page map to save
	;
    f5b_1_d:
        mov     si, offset DGROUP:f4e_list
        call    get_page_map
        pop     di      	        ; restore original es:di
        pop     es
        cmp     bl, MAX_AMRS
        movzx   si, bl
        ja 	short f5b_err9d		; jmp if invalid handle

        or      si, si
        jz 	short f5b_1_0           ; jmp if special case of bl=0

        bt 	word ptr f5b_reg_sets, si
        jnc 	short f5b_err9d

        mov     f5b_current_set, bl      ; save new register set
        imul    si,13                    ; make index into f5b_save_area

    f5b_1_a:
        add     si,offset DGROUP:f5b_save_area

    f5b_1_b:
        call    set_page_map
        jmp     f5b_done

        ; bl = 0, es:di may be a pointer
	;
    f5b_1_0:
        mov     f5b_current_set, bl     ; save new register set number
        push    es
        push    di
        pop     f5b_ptr                 ; save new pointer
        les     di, f5b_ptr
        xor     si, si                  ; in case we jump to f5b_1_a
        mov     ax, es
        or      ax, di
        jz 	short f5b_1_a           ; jmp if es:di = 0

        push    es                      ; es:di points to a state save area
        pop     ds
        mov     si, di
        jmp     f5b_1_b

	; get alternate map state save size.
	;
    f5b_2:
        mov     dx, SIZE DL_EMS_CONTEXT
        jmp     f5b_done

	; allocate alternate map register set, return number in bl
	;
    f5b_3:
        xor     ax, ax

    f5b_3_loop:
        inc     ax
        bts 	word ptr f5b_reg_sets,ax 	; unallocated AMRS?
        jnc 	short f5b_3_found		; CF clear if so

        cmp     al, MAX_AMRS
        jb      f5b_3_loop              	; max al of MAX_AMRS

	; all alternate map register sets have been allocated
	;
        pop     ax
        mov     ah, EMSERR_NO_MORE_AMRS
        jmp     i67done

	; Have a newly allocated AMRS number in al, save the
	; current mapping context into that AMRS' save area
	; and return to the caller.
	;
    f5b_3_found:
        mov     bl, al                  ; new AMRS handle
        push    es
        push    di

        push    ds
        pop     es
        mov     di, ax
        imul    di, SIZE DL_EMS_CONTEXT
        add     di, offset DGROUP:f5b_save_area
	mov     si, offset DGROUP:f4e_list
	call    get_page_map

        rep     movsb
        pop     di
        pop     es
        jmp     f5b_done

	; deallocate alternate map register set, specified in bl
	;
    f5b_4:
        movzx   ax, bl
        cmp     bl, MAX_AMRS
        ja 	short f5b_4_err

        or      bl, bl
        jz      f5b_done                    ; deallocating AMRS #0 not an error
        btr 	word ptr f5b_reg_sets,ax    ; check and reset the bit
        jc      f5b_done

	; invalid handle in bl
	;
    f5b_4_err:
        pop     ax
        mov     ah, EMSERR_AMRS_NOT_ALLOCED
        jmp     i67done

	; allocate DMA register set, not really supported, but simulare MS
	;
    f5b_5:
        mov     bl, 0

    f5b_5a:
        pop     ax
        jmp     i67ok

	; enable DMA on alternate map register set if dl (DMA channel #)
	; is non-zero then MS EMM386 generates an error otherwise it
	; returns success if bl <>0.
	;
    f5b_6:
	or	dl, dl
	jnz short f5b_no_DMARS

	; for 5b06,5b07 and 5b08 MS EMM386 returns success if bl = 0
	; or an error otherwise
	;
    f5b_7_8:
	or	bl, bl
	jz	f5b_5a

    f5b_no_DMARS:
        pop     ax
        mov     ah, EMSERR_DMARS_UNSUPPORTED
        jmp     i67done



	; To prepare system for a warm boot, do nothing.
	;
i67f5C:
        jmp     i67ok



	; For fn 5Dxx, functions are disabled and access key was
	; invalid. For fns 59 and 5Bxx, access was disallowed by
	; the Operating System/Environment (OS/E). No other
	; functions are affected.
	;
AccessDeniedToFuntion:
        mov     ah, EMSERR_ACCESS_DENIED
        jmp     i67done


	; Fns 	5D00h	- enable OS/E function set
	;	5D01h	- disable OS/E function set
	; 	5D02h	- reenable OS/E function set and return access key
	;
	; conditionally enable or disable access to certain
	; operating system functions.
	;
i67f5D:
	test	fnAccessState, ACCESS_KEY_ISSUED
        jz 	short f5d_a		; jmp if first access

	; not the first access, is the key correct?
	;
        cmp     bx, key1
        jne     AccessDeniedToFuntion
        cmp     cx, key2
        jne     AccessDeniedToFuntion

	; key is correct, allow access to function 5Dxxh
	;
    f5d_a:
        push    ax
        or      al,al
        jz      short f5d_0
        dec     al
        jz 	short f5d_1
        dec     al
        jnz     f4e_err         	; jmp if invalid subfunction.

	; Function 5D02, return access key to EMM and re-enable
	; access to OS/E specfic functions. We first alter the
	; key and then reset the access state variable to its initial
	; value and so allow full access to the OS/E functions.
	;
        add     key1, bp
        add     key2, bp
        mov     fnAccessState, FULL_ACCESS_ALLOWED
        jmp     f5b_done

	; enable functions and return key
	;
    f5d_0:
	mov	fnAccessState, ACCESS_KEY_ISSUED

    f5d_b:
        mov     bx, key1
        mov     cx, key2
        jmp     f5b_done

	; disable functions and return key
	;
    f5d_1:
	mov	fnAccessState, (ACCESS_KEY_ISSUED or ACCESS_DENIED_BY_OSE)
        jmp     f5d_b



IF CALL_67F86
	;-------------------------------------
	; 86H - Disable V86 Mode
i67f86:
	push	BX
	push	CX
	push	DX
	push	ES
IF STACKSIZE
	call	switch_stacks
ENDIF
	call	disable_V86
IF STACKSIZE
	call	restore_stack
ENDIF
	pop	ES
	pop	DX
	pop	CX
	pop	BX
	jmp	i67ok
ENDIF ; CALL_67F86

prot_mode_ret:
;vcpi handler returns here after handling the call in real mode.
        push    eax
        call word ptr __x386_rtop
        pop     eax
        pop     gs
        pop     fs
        pop     es
        pop     ds
        call    restore_stackp
        popf                    ;restore original interrupt state
        db      66h             ;do a 32 bit far return to the client
        retf

vcpi_functions  dw      offset vcpi_0
                dw      offset vcpi_1
                dw      offset vcpi_2
                dw      offset vcpi_3
                dw      offset vcpi_4
                dw      offset vcpi_5
                dw      offset vcpi_6
                dw      offset vcpi_7
                dw      offset vcpi_8
                dw      offset vcpi_9
                dw      offset vcpi_a
                dw      offset vcpi_b
                dw      offset vcpi_c

public prot_mode_entry
prot_mode_entry:        ;first check for prot to V86 mode call.
        cmp     al,0ch
        je vcpi_d       ;jmp if mode switch call.
;switch to real mode
        pushf           ;preserve state of IF
        cli
        call    switch_stackp
        push    ds
        push    es
        push    fs
        push    gs
        push    eax
        call word ptr __x386_ptor
        pop     eax
        call    getDGROUP
;prepare an iret structure on the stack
        pushf
        push    cs
        push    offset prot_mode_ret
        push    ds
        push    si

;VCPI calls, support subfunctions 0 - C
i67fDE:
        cmp     al,0ch
        ja      i67_subfn_not_supported
        movzx   si,al
        shl     si,1                            ;address a table of words
        jmp word ptr cs:vcpi_functions[si]      ;jmp to the desired function

vcpi_0:                         ;VCPI detection call
        mov     bx,100h         ;version 1.0
        jmp short vcpi_okay

comment &
        Get protected mode interface:
        Call with es:di = pointer to 4K page table buffer
                  ds:si = pointer to three descriptors in GDT
        Returns:  ah = 0
                  di advanced to end of page table buffer
                  ebx = offset of protected mode entry point
&
vcpi_1:
        push    edi
        push    esi
        xor     esi,esi
        mov     esi,es
        movzx   edi,di
        shl     esi,4
        add     edi,esi                 ;absolute pointer for page table
        mov     al,4
        call    call_VMM                ;copy the page table
        mov     si,sp
        lds     si,ss:[si+8]            ;load original ds:si
        mov     di,si
        push    es
        push    ds
        pop     es                      ;es:di points to code des in gdt
        push    cs
        pop     ds
        mov     si,offset _TEXT:__x386_code_16_des
        mov     cx,6
        cld
        rep     movsd                   ;move 3 gdt entries
        pop     es
        pop     esi
        pop     edi
        db      66h                     ;make it ebx below
        mov     bx,offset prot_mode_entry
        dw      0                       ;high word of offset for ebx
        mov     ax,1
        add     di,1088                 ;advance to top of HMA in page table
        jmp     i67done

vcpi_2:         ;return ah = 0, edx = physical address of highest 4k block
        movzx   edx,word ptr ds:_frame_pages    ;max number of 16K pages
        shl     edx,14                          ;number of bytes
        add     edx,ds:_frame_memptr            ;add start of XMS memory
vcpi_okay:
        mov     ah,0
        jmp     i67done

vcpi_3:         ;return number of free 4K pages in edx, return ah = 0
        push    eax
        push    bx
        call    get_xms_to_ems                  ;returns 16K pages in bx
        movzx   edx,ds:_frame_avail             ;number of 16 K pages
        xor     eax,eax
        add     dx,bx                           ;add in the convertible pages
        movzx   eax,ds:vcpi_remainder           ;number of 4 K pages reserved
        lea     edx,[edx*4+eax]
        pop     bx
        pop     eax
        jmp     vcpi_okay

vcpi_4:         ;allocate a 4k page, return ah = 0, edx = physical address
                ;or ah = 88h if failure
        push    bx
        push    cx
        cmp     ds:vcpi_remainder,0
        je short no_remainder
;There should be some memory available in partially used 16K pages marked
;with handles of FFF1 - FFFE.
        mov     si,ds:_frame_table
        cld
        mov     cx,ds:_frame_pages
vcpi_4_loop1:
        lodsw                           ;load a handle from the frame table
        cmp     ax,0fff0h               ;check if VCPI handle
        jae short vcpi_4_1              ;jmp if a VCPI handle was found
vcpi_4_3:
        loop    vcpi_4_loop1
vcpi_4_err1:    ;no remainder was found, this means there was an error
        mov     ds:vcpi_remainder,0
        jmp short no_remainder

vcpi_4_1:       ;A VCPI handle was found and is in ax.
                ;Check if any 4K pages are available.
        cmp     al,0ffH
        je short vcpi_4_3               ;jmp if no unused memory
        dec     ds:vcpi_remainder
vcpi_4_2:
;AX contains the handle, si points to 2 bytes past desired frame table entry
        push    esi
        lea     bx,[si-2]
        xor     esi,esi
        not     ax
        bsf     si,ax           ;find first set bit
        not     ax              ;convert set bit to cleared bit
        bts     ax,si           ;set the bit to allocate memory
        mov     ds:[bx],ax      ;place the new handle
        shl     esi,12          ;multiply by 4096 to get offset in 16K block
        call    get_phys
;edx contains the physical address of the EMS page
        add     edx,esi         ;add offset into EMS page
        pop     esi
        pop     cx
        pop     bx
        mov     al,4
        jmp     vcpi_okay

vcpi_4_err:
        pop     cx
        pop     bx
        mov     ax,8804h
        jmp     i67done

vcpi_4_a:       ;We need to expand the EMS pool if possible.  This requires
                ;that we be in real mode so we can call himem.sys.  BX = 0 if
                ;real mode.
        mov     bx,1
        call    expand_ems
        jc      vcpi_4_err

no_remainder:   ;lets make a remainder if possible.
        cmp     ds:_frame_avail,0
        je      vcpi_4_a                ;jmp if no memory left in EMS pool
        push    di
        mov     di,ds:_frame_table
        push    es
        push    ds
        pop     es                      ;es:di points to frame table
        xor     ax,ax
        mov     cx,ds:_frame_pages
        repne   scasw                   ;search for first available 16K page
        pop     es
        mov     si,di                   ;one word past available entry
        pop     di
        jne     vcpi_4_err              ;jmp if no memory was found
        mov     ax,0fff0h               ;VCPI handle
        mov     ds:vcpi_remainder,3     ;update to reflect a new 16K page
        dec     ds:_frame_avail         ;update to indicate one less page
        jmp     vcpi_4_2

vcpi_5:         ;free a 4K page, physical address passed in edx
                ;return ah = 0 if success, ah = 8ah if failure
        push    bx
        push    ecx
        mov     cx,9            ;maximum of 9 frame_table blocks
        xor     bx,bx
        mov     si,_frame_table
vcpi_5_loop:
        mov     ax,_frame_table[2+bx]
        cmp     si,ax
        ja short vcpi_5_err     ;jmp if no more blocks
        shl     bx,1            ;make index into dwords
        push    edx
        sub     edx,_frame_memptr[bx]
        jb short vcpi_5_a
;The desired page is not below the start of the block, is it below the end?
;edx contains the offset into the XMS block if this is the block.
        shr     edx,13          ;make index into the frame table
        add     dx,si           ;make pointer into frame table
        cmp     dx,ax           ;cmp to pointer to end of block in table
        jb short vcpi_5_b       ;jmp if page is within this block
vcpi_5_a:
;not this block, try the next one
        pop     edx
        shr     bx,1
        add     bx,2
        mov     si,ax
        loop    vcpi_5_loop
vcpi_5_err:
        pop     ecx
        pop     bx
        mov     ax,8a05h
        jmp     i67done

vcpi_5_b:
;we have found the desired block, si points to start of block, bx is dword index
        pop     ecx                             ;original edx value
        mov     edx,ecx
        sub     ecx,ds:_frame_memptr[bx]        ;make offset into EMS block
        shr     ecx,12                          ;4K index
        movzx   bx,cl                           ;save 2 LSBs
        and     bl,3                            ;keep 4K index within 16K block
        shr     cx,1                            ;convert to index into frame table
        and     cl,0feh                         ;clear LSB to word align
        add     si,cx                           ;si points to handle
        mov     ax,ds:[si]                      ;get the handle
        cmp     ax,0fff0h               ;handle should be >
        jbe     vcpi_5_err
        btr     ax,bx                   ;clear the bit for this page
        jnc     vcpi_5_err              ;jmp if page was already free
        cmp     al,0f0h         ;check if entire 16K page is free
        jne short vcpi_5_2
        inc     ds:_frame_avail
        xor     ax,ax                   ;indicate 16K page is free
        sub     ds:vcpi_remainder,4     ;allow for freeing the 16K page
vcpi_5_2:
        mov     ds:[si],ax              ;save handle in frame table
        inc     ds:vcpi_remainder       ;indicate 1 more page available
        pop     ecx
        pop     bx
        mov     ax,5
        jmp     vcpi_okay

vcpi_6:         ;Get physical address of 4K page in First Megabyte
                ;call with cx = page number (linear address shr 12)
                ;returns edx = physical address of 4k page, ah = 0
                ;returns ah = 8bh if invalid page number
        or      ch,ch
        jnz short vcpi_6_err
        call    call_VMM        ;returns edx = physical address, CF = 0
                                ;or CF = 1 if failure
        jnc     vcpi_okay
vcpi_6_err:
        mov     ah,8bh
        jmp     i67done

vcpi_7:         ;Read CR0, return ah = 0, ebx = CR0
        call    call_VMM
        jmp     vcpi_okay

vcpi_8:         ;Read Debug registers, es:di = pointer to array of 8 dwords,
                ;DR0 is first in array, DR4 and DR5 are not included.
vcpi_9  label near      ;Load Debug Registers, opposite of function 8.
        push    edi
        push    esi
        xor     esi,esi
        mov     si,es
        movzx   edi,di
        shl     esi,4                   ;convert segment to absolute
        add     edi,esi                 ;edi has 32 bit address of array
        call    call_VMM
        pop     esi
        pop     edi
        jmp     vcpi_okay

vcpi_a:         ;return PIC mappings bx = int for IRQ0, cx = int for IRQ8
        push dword ptr irq0
        pop     bx
        pop     cx
        jmp     vcpi_okay

vcpi_b:         ;set PIC mappings, bx = int for IRQ0, cx = int for IRQ8
        call    call_VMM
        jmp     vcpi_okay

vcpi_c:         ;switch to protected mode
                ;call with esi = linear address of data structure
                ;in first megabyte.  CR3,ptr to GDT ptr, ptr to
                ;IDTR ptr, LDT selector, cs:eip to return to.
        pop     si              ;restore original si
        mov     al,5
        call    call_VMM        ;no return from this call.

vcpi_d:         ;switch to V86 mode
;note that si has NOT been pushed on the stack and has NOT been trashed, this
;function is called only from protected mode and does not have the same things
;pushed on the stack as other functions called from protected mode.
        mov     al,0ch
        call    call_VMM        ;no return from this call.
assume ds:DGROUP

comment &
        We must hook int 15 function 4fh to detect an alt-cntrl-del.  Many
        BIOSes jmp directly to the reboot code, bypassing the reboot vector at
        FFFF:0.  Thus our hook of the reboot vector is not adequate.
&
int_15_4f:
	push	AX
	push	ES
	mov	AX, 40H
	mov	ES, AX
	mov	AL, ES:[17H]	; check BIOS flags for ctrl-alt
	and	AL, 0CH
	cmp	AL, 0CH
        stc
	pop	ES
	pop	AX
	je short reboot1
        jmp far ptr int_15_chain        ;chain to next handler

reboot1:        ;chain and then disable V86 on return from int 15 chain
        pushf
        call far ptr int_15_chain
;int 15 chain is done processing the call, disable V86 mode
;mask all hardware interrupts
        icmd_reg        equ     20h     ; int ctlr command reg addr (20h)
        imsk_reg        equ     21h     ; int ctlr mask reg         (21h)
        icmd_reg_slv    equ     0A0h    ; slv int ctlr command reg addr (A0h)
        imsk_reg_slv    equ     0A1h    ; slv int ctlr mask reg         (A1h)

        push    ax
        pushf
        mov al,0ffh             ; mask ALL ints
        out imsk_reg,al
        out imsk_reg_slv,al
        mov ax,0820h            ; issue non-specific eoi's
clrints:
        out icmd_reg,al
        out icmd_reg_slv,al
        dec ah
        jnz clrints

	push	bx
	push	ds
	xor	bx, bx
	mov	ds, bx
	mov	bx, cs:uOldMemTop
	mov	ds:[413h], bx
	pop	ds
	pop	bx
	popf
        pop     ax
        call    disable_4_reboot
        stc
        retf 2                  ;do an iret preserving CF

; Disable V86 mode so we can reboot.
	public	disable_4_reboot
disable_4_reboot:
	push	AX
	push	BX
	push	CX
	push	DX
	push	ES
	push	DS
	call	getDGROUP
IF STACKSIZE
	call	switch_stacks
ENDIF
	call	disable_V86
comment &
        We have a difficult problem to deal with at this time.  On some
        systems, the stack will be located in a UMB.  This stack will
        disappear when we switch back to real mode.  We will check to see if
        the original stack is in a UMB, if it is, we will take a short cut and
        jmp directly to the reset vector.  This is a bit dangerous since there
        could be someone who hooked int 15h ahead of emm386 and was waiting
        for the call to return before preparing for a reboot.  If we are on a
        UMB stack, that code will never be called, because the stack is gone
        and we don't know enough about the structures on the stack to be
        certain we can move it and still have things work.
&

        mov     ax,word ptr user_stack          ;sp
        shr     ax,4
        add     ax,word ptr user_stack[2]       ;ss
;ax has the segment address of the stack
        shr     ax,8                            ;convert to page number
        bt word ptr i_range,ax
        jnc short r_stack_ok                    ;jmp if stack is not UMB
;The stack is in the UMB, we must take a short cut, call the reboot vector.
	mov     ax,40h
        mov     ds,ax
        mov word ptr ds:[72h],1234h             ;request a warm boot
	mov	bx, cs:uOldMemTop		; Old top of memory
	mov	word ptr ds:[13h], bx		; Old top of memory
        jmp dword ptr cs:[old_FFFF0]

r_stack_ok:

IF STACKSIZE
	call	restore_stack
ENDIF
	pop	DS
	pop	ES
	pop	DX
	pop	CX
	pop	BX
	pop	AX
	ret

	public	reboot_FFFF0
reboot_FFFF0:
	xor	bx,bx
	mov	es,bx
	mov	bx, cs:uOldMemTop
	mov	es:[413h], bx
	call	disable_4_reboot
	jmp	dword ptr cs:[old_FFFF0]

cEndCode

;/////////////////////////////////////////////////////////////////////////
;//	Upper Memory Block Support					//
;/////////////////////////////////////////////////////////////////////////

; DOS locations
FIRST_MCB	equ	24H	; segment in DOS DS of 1st MCB
UMB_LINK	equ	89H
UMB_ARENA	equ	8CH

; Memory Control Block fields
MCB_ID		equ	0
MCB_OWNER	equ	1
MCB_SIZE	equ	3	; offset for size part of MCB
MCB_INIT_ID	equ	7	; offset for RD specific ID used during BOOT
MCB_NAME	equ	8	; offset for name part of MCB

SYSTEM_ID	equ	8	; SYSTEM "owner" of MCBs
FREE		equ	0	; owner of a free MCB

cBegCode


int_21_5803:   ;int 21h handler for function 5803h
	push	DS
	push	ES
	push	AX
	push	BX

	; get DOS data segment
	call	getDGROUP
	mov	DS, [DOS_DS]

	; set or clear the link
	or	BX, BX
	jz short clear_link

comment &
        The function call is to link in the UMBs, some machines have problems
        with UMBs when windows installs, we don't understand the problem, but
        the following code seems to fix it.
&
	; set the link if windows is not installed
        push    _DEVDATA
        pop     es
assume es:_DEVDATA
        cmp     es:win_installed,0      ;check if windows is installed
assume es:nothing
        jne 	short done_link
	cmp	byte ptr DS:[UMB_LINK], 0
	jnz short done_link
	call	hook_umbs
	mov	BL, 1
	jmp	short set_link

	; clear link
clear_link:
	cmp	byte ptr DS:[UMB_LINK], 0
	jz short done_link
	call	unhook_umbs
	mov	BL, 0
set_link:
	mov	byte ptr DS:[UMB_LINK], BL

	; restore registers
done_link:
	pop	BX
	pop	AX
	pop	ES
	pop	DS

	; clear the carry flag
	push	BP
	mov	BP, SP
	and	byte ptr [BP+6], 0FEH
	pop	BP
	iret


comment ~ ; prototype ;
void hook_umbs(void)
~
c_func	hook_umbs
	push	DS		; save DOS data segment
	call	getDGROUP
	call	get_last_mcb	; get pointer to last MCB
	pop	DS		; restore DOS data segment
	mov	byte ptr ES:[MCB_ID], 'M'
	ret
c_endp	hook_umbs

comment ~ ; prototype ;
void unhook_umbs(void)
~
c_func	unhook_umbs

	; assumes DS = DOS DS upon entry

	; find the MCB that includes the arena
	mov	ES, DS:[FIRST_MCB]
	mov	AX, ES
un_again:
	add	AX, ES:[MCB_SIZE]
	inc	AX
	cmp	AX, DS:[UMB_ARENA]
	jae short un_found
	mov	ES, AX
	jmp	short un_again
un_found:
	; unhook it from the memory list
	mov	byte ptr ES:[MCB_ID], 'Z'
	ret
c_endp	unhook_umbs

	public	get_next_mcb
get_next_mcb	proc	near
	mov	AX, ES
	add	AX, word ptr ES:[MCB_SIZE]
	inc	AX
	mov	ES, AX
	ret
get_next_mcb	endp

	public	get_last_mcb
get_last_mcb	proc	near
	mov	ES, [DOS_DS]
	mov	ES, ES:[FIRST_MCB]
again_mcb:
	cmp	byte ptr ES:[MCB_ID], 'Z'
	jz short got_last_mcb
	call	get_next_mcb
	jmp	again_mcb
got_last_mcb:
	ret
get_last_mcb	endp

comment ~ ; prototype ;
void create_umbs(void)
~
c_func	create_umbs

	; no ranges? don't do anything
	mov	AX, word ptr [_first_range]
	cmp	AX, word ptr [_num_ranges]
	jb short crum1
        ret

	; trap INT 21H to support Get/Set Link, functions 58xxH
crum1:	mov	al,21H
	mov	CX, offset _DEVDATA:int21_ptr
	mov	DX, offset _DEVDATA:int_21
	call	set_interrupt

	; shrink last MCB by 1 to make room for UMB chain
	call	get_last_mcb
	dec	word ptr ES:[MCB_SIZE]
	mov	byte ptr es:[MCB_ID], 'M' ; start with UMBs linked in
	call	get_next_mcb		; ES now points past MCB chain

	; record arena (1st UMB) in DOS data area
	;
	push	DS
	mov	DS, [DOS_DS]
	mov	DS:[UMB_ARENA], ES
	pop	DS

	; add each range to the list
	; for(i=first_range; i<num_ranges; ++i)
	mov	CX, word ptr [_first_range]
crum2:	cmp	CX, [_num_ranges]
	jae short crum9

	; size = ranges[i].start - (FP_SEG(mcb)+1);
	mov	AL, SIZE RANGE
	mul	CL
	mov	BX, GOFFSET _ranges
	add	BX, AX
	mov	AX, [BX.range_start]
	mov	DX, ES
	inc	DX
	sub	AX, DX		; AX = size

	; set_mcb(FP_SEG(mcb),'M', SYSTEM_ID, size);
	mov	byte ptr ES:[MCB_ID], 'M'
	mov	word ptr ES:[MCB_OWNER], SYSTEM_ID
	mov	word ptr ES:[MCB_SIZE], AX
	mov	word ptr ES:[MCB_NAME], 'CS'	; in str form "SC"
	mov	byte ptr ES:[MCB_NAME+2], ' '
	mov	byte ptr ES:[MCB_NAME+3], ' '
	mov	byte ptr ES:[MCB_NAME+4], ' '
	mov	byte ptr ES:[MCB_NAME+5], ' '
	mov	byte ptr ES:[MCB_NAME+6], ' '
	mov	byte ptr ES:[MCB_NAME+7], ' '
	call	get_next_mcb

	; size = ranges[i].len-1, id='Z';
	; if(i < num_ranges-1)
	;    --size, id='M';
	mov	AX, [BX.range_len]
	dec	ax
	mov	DL, 'Z'
	push	BX
	mov	BX, [_num_ranges]
	dec	BX
	cmp	CX, BX
	jae short crum3
	dec	AX		; --size
	mov	DL, 'M'		; id = 'M'
crum3:	pop	BX

	; set_mcb(FP_SEG(mcb), id, FREE, size);
	mov	byte ptr ES:[MCB_ID], DL
	mov	word ptr ES:[MCB_OWNER], FREE
	mov	word ptr ES:[MCB_SIZE], AX
	call	get_next_mcb

	; ++i
	inc	CX
	jmp	crum2
crum9:
	; If in CONFIG.SYS processing now, find the last MCB in the
	; inner MCB chain
	;
	mov	es, [DOS_DS]
	mov	es, es:[FIRST_MCB]
	cmp	byte ptr es:[MCB_INIT_ID], 0
	jne short crum10

	; Link the UMBs to the inner MCB chain so we can load
	; devices high
	;
	mov	ax, es
	inc	ax
	mov	es, ax
	call	again_mcb
	dec	word ptr es:[MCB_SIZE]
	mov	byte ptr es:[MCB_ID], 'M'

crum10:
	; Set the "UMBs are linked in right now" flag
	;
	mov	es, [DOS_DS]
	mov	byte ptr es:[UMB_LINK], 1
	ret
c_endp	create_umbs

cEndCode


;/////////////////////////////////////////////////////////////////////////
;//	Device Driver Initialization					//
;/////////////////////////////////////////////////////////////////////////
cBegCode

if DEBUG ge 2
extrn	VmmInitTimerWait:proc
extrn	VmmInitRMErrorHandler:proc
extrn	VmmInitPMErrorHandler:proc

extrn 	VmmTestPrintf:proc
extrn 	VmmTestAsserts:proc
endif

;----------------------------------------------
; bool emm_init(void)
;
emm_init:

IF 0
IF USING_GEOS
   mov   byte ptr [signon_msg+0],13    ;
   mov   byte ptr [signon_msg+1],10    ; re-enable our signon
   mov   byte ptr [signon_msg+2],'E'   ;
ENDIF
ENDIF

IF SIGNON
	; print a sign-on message
	printm	signon_msg
ENDIF
	; store DOS data segment
	mov	AH, 52H
	int	21H
	mov	[DOS_DS], ES

	; do NOT install if a driver is already installed
	mov	AX, 3567H	; get int vector
	int	21H		; returns vector in ES:BX
	mov	DI, 11		; check ES:0BH
	mov	SI, offset _DEVDATA:_emmx + 11
	mov	CX, 7
	push	DS
	call	getDEVDATA
	mov	DS, AX
	cld
	rep	cmpsb
	pop	DS
	jnz short ok2install	; names are diffent, ok to install
	PRERROR	already_installed_msg
	jmp	emm_initerr
ok2install:

	; do NOT install if no XMS support
	mov	AX, 4300H	; installation check
	int	2FH
	cmp	AL, 80H
	jz short xms_is_supported
	PRERROR	no_xms_msg	; no XMS, can't install
	jmp	emm_initerr
xms_is_supported:
	mov	AX, 4310H	; get address of XMS handler
	int	2FH
	mov	word ptr [xms_handler], BX
	mov	word ptr [xms_handler+2], ES

	; create our own internal memory cache
	mov	word ptr [malloc_ptr], offset DGROUP:endsegmarker

	; parse the command line
	les	BX, dword ptr [ptrsav]
	les	BX, ES:[BX.cmdline]
	push	ES
	push	BX
	; parse_cmdline(char far *s)
	call	_parse_cmdline
        pop     eax                     ;pop garbage off the stack

	xor	ax, ax
	mov	es, ax
	mov	ax, es:[413h]		; Top of memory used by int 12h
	mov	cs:uOldMemTop, ax

IF ENABLE_MOVEEBDA

	;  Determine if we are relocating the EBDA
	;
	cmp	fRelocateEBDA, 0
	jz	DoneEBDAPrep


	;  Determine if there is an EBDA
	;
	push	ax
	push	bx
	push	cx
	push	dx
	push	es


	xor	ax, ax
	mov	es, ax
	mov	ax, word ptr es:[40eh]
	mov	bx, word ptr es:[413h]
	shl	bx, 6			; convert base memory to para (in k now)

	mov	es, ax
	mov	dx, word ptr es:[0h]
	mov	cx, dx
	shl	cx, 6			; convert size to para

	; Only relocate it if it at the top of memory
	;
	cmp	ax, bx
	je short PrepareForRelocate

	pop	es
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	mov	fRelocateEBDA, 0
	jmp short DoneEBDAPrep


PrepareForRelocate:

	;  Increase the base memory
	;
	xor	ax, ax
	mov	es, ax
	add	es:[413h], dx

	;  Add the size of the XBDA to the last MCB for both
	;  the inner and outer MCB chains
	;
	mov	ES, [DOS_DS]
	mov	es, ES:[FIRST_MCB]
	cmp	byte ptr es:[MCB_INIT_ID], 0
	jne short DoNextAddToMCB
DoAddToMCB:
	mov	ax, es
	inc	ax
	mov	es, ax
	call	again_mcb
	add	word ptr es:[MCB_SIZE], cx
DoNextAddToMCB:
	call	get_last_mcb	; get pointer to last MCB
	add	word ptr es:[MCB_SIZE], cx

	;  Move the EBDA down one paragraph so that
	;  an MCB header can follow it.
	;
	push	si
	push	di
	push	ds
	xor	si, si
	mov	ds, si
	mov	di, word ptr ds:[40eh]	; EBDA source in DS:SI
	dec	word ptr ds:[40eh]
	mov	es, word ptr ds:[40eh]	; EBDA dest in ES:DI
	mov	ds, di
	mov	di, si
	mov	cx, word ptr ds:[0h]
	shl	cx, 8			; Set DWORD move count
	cld				; Set forward move
	rep	movsd			; Move the EBDA down one para
	pop	ds
	pop	di
	pop	si

	pop	es
	pop	dx
	pop	cx
	pop	bx
	pop	ax

DoneEBDAPrep:

ENDIF ; IF ENABLE_MOVEEBDA

if DEBUG ge 2
	; Initialize the VMM debugging code to work with the LOG.EXE TSR
	;
	call	VmmInitTimerWait
	call	VmmInitRMErrorHandler
	call	VmmInitPMErrorHandler

	LOGPRINTF <'Allocating XMS'>
endif
	; actually allocate the requested XMS memory
	call	alloc_from_xms              ;returns 0 if success, non-zero if failure
        or      ax,ax
        jnz     emm_initerr

	; create the expanded memory table
	; note: any allocation must be done BEFORE move to
	; Upper Memory (V86Init) takes place.
	; determine size of table in bytes
	;
        mov     ax,_frame_pages                 ;each entry is 2 bytes
        mov     dx,ax
        shl     ax,1
        jz short crems_done                     ;jmp if no EMS
	; allocate the memory from internal pool
        call    malloc
        mov     _frame_table,ax
        mov     di,ax
        mov     cx,_frame_avail
        sub     dx,cx
	; mark page entries in first block as free
        push    ds
        pop     es
        xor     ax,ax
        cld
        rep     stosw
; now mark the entries in the remaining blocks as not present, handle = 8000h
        mov     cx,dx                           ;balance of table
        mov     _frame_table[2],di              ;mark end of first block
        mov     ax,8000h
        rep     stosw                           ;finish initializing frame table
crems_done:

	; special case: if 1st range is JUST above last_mcb,
	; then enlarge DOS MCB chain.
	; no ranges? don't do anything
        cmp word ptr [_num_ranges],0
        je short expcon9

	; if 1st range is JUST above last_mcb, then expand
	; DOS MCB chain to encompass that range.
	call	get_last_mcb
	mov	AX, ES
	add	AX, word ptr ES:[MCB_SIZE]
	inc	AX		; AX = just past end of MCBs (ie A000H)
	cmp	AX, _ranges.range_start	; same as 1st range?
	jnz short expcon9

	; 1st RANGE IS just above MCB chain, indicate so.
	; let world know we used the 1st range
	inc	[_first_range]
expcon9:
        push    es              ;save pointer to MCB

IF TRAPFFFF
	; a memory pointer used by trap_ffff
	; note: must be the last use of malloc_ptr at the data at
	; malloc ptr will be trashed
	mov	AX, [malloc_ptr]
	mov	word ptr [reboot_tmpptr], AX
	mov	word ptr [reboot_tmpptr+2], DS
ENDIF

	; remap XMS memory to specified ranges
	push	[_num_ranges]
	mov	AX, GOFFSET _ranges
	push	AX
	push	word ptr [_range_memptr+2]
	push	word ptr [_range_memptr]
	; int remap_ranges(unsigned long mem,struct range ranges[], int limit)

if DEBUG ge 2
	LOGPRINTF <'Remapping ranges'>
endif
	call	_remap_ranges       ;can't comment this out
	add	SP, 8
	or	AX, AX
        pop     es      ;pointer to MCB
	jnz short remapped

	mov	bx, init_err_code
	or	bx, bx
	jle short default_error_code

	cmp	bx, total_msgs
	jbe short error_code_ok

    default_error_code:
	mov	bx, ERR_CANT_REMAP

    error_code_ok:
	dec	bx
	shl	bx, 1
	PRERROR	error_msgs[bx]
	call	free_xms
	jmp	emm_initerr
remapped:

;enlarge DOS MCB chain if _first_range is non-zero
        cmp byte ptr _first_range,0     ;can we expand conventional memory?
        je short first_range_0
;expand conventional memory, pointer to MCB is in es.
	push	cx

	call	get_last_mcb
        mov     cx,_ranges.range_len
        add     ES:[MCB_SIZE],cx
	mov	ax, es
	inc	ax
	mov	es, ax
	call	again_mcb
        add     ES:[MCB_SIZE],cx

	xor	ax, ax
	mov	es, ax
	shr	cx, 6
	add	word ptr es:[413h], cx

	pop	cx

first_range_0:

   ; this is a fixup for a FRAME defined within the conventional memory chain
   ;
   cmp   _frame_seg, 0
   je end_of_conventional_frame_handling

   ; The FRAME might be defined within conventional memory.
   ; Walk the MCB chain until we either reach the end or
   ;  have the MCB that the FRAME is defined within.
   ;
   mov   ES, [DOS_DS]               ; initial conditions for MCB search
   mov   ES, ES:[FIRST_MCB]
   mov   DX, _frame_seg             ; initial conditions for end of frame
   add   DX, 1000h

test_this_MCB:
   mov   AX, ES

IF DEBUG GE 1
   mov   AL, ' '
   call  BIOSPUTC
   mov   AL, ES:[MCB_ID]         ; dump the MCB_ID
   call  BIOSPUTC
   mov   AX, ES                  ; cleverly restore AX ;)
   call  BIOSPUTX                ; dump the MCB segment
ENDIF

   add   AX, ES:[MCB_SIZE]          ; get the end of this MCB
   cmp   DX, AX                     ; ...compare with the end of _frame_seg
   jbe short adjust_mem_for_MCB     ; _frame_seg within this MCB
   cmp   byte ptr ES:[MCB_ID],'Z'
   je short end_of_conventional_frame_handling  ; _frame_seg past last MCB
   call  get_next_MCB
   jmp short test_this_MCB          ; and loop

adjust_mem_for_MCB:
   ; To handle this situation, we will split the MCB in two,
   ; allocating the second half (and the frame) to the system.
   ;

   push  ES                         ; save ES - points to current MCB

   MOV   DX, _frame_seg             ; the new MCB location, _frame_seg - 1
   DEC   DX

   MOV   CX, ES                     ; length of new MCB is
   ADD   CX, ES:[MCB_SIZE]          ; (ES + ES:[MCB_SIZE] - DX)
   SUB   CX, DX

   MOV   BL, ES:[MCB_ID]            ; ID of new MCB = ID of current MCB
   MOV   ES, DX

   MOV   BYTE PTR ES:[MCB_ID], BL
   MOV   WORD PTR ES:[MCB_SIZE], CX

   MOV   WORD PTR ES:[MCB_OWNER], SYSTEM_ID
   MOV   WORD PTR ES:[MCB_NAME], 'CS'     ; in str form "SC"
   MOV   BYTE PTR ES:[MCB_NAME+2], ' '
   MOV   BYTE PTR ES:[MCB_NAME+3], ' '
   MOV   BYTE PTR ES:[MCB_NAME+4], ' '
   MOV   BYTE PTR ES:[MCB_NAME+5], ' '
   MOV   BYTE PTR ES:[MCB_NAME+6], ' '
   MOV   BYTE PTR ES:[MCB_NAME+7], ' '

   pop   ES                         ; restore pointer to current MCB

   ; Now reduce the "current" MCB length to account for the missing memory
   ;
   MOV   CX, DX                     ; DX still has value of _frame_seg - 1
   MOV   BX, ES
   SUB   CX, BX
   DEC   CX

   MOV   BYTE PTR ES:[MCB_ID], 'M'  ; first MCB is now middle of a chain
   MOV   WORD PTR ES:[MCB_SIZE], CX ; adjust the size as appropriate

end_of_conventional_frame_handling:
	xor	ax, ax
	mov	es, ax

	; create the Upper Memory Block Chain
	; (traps any INTs necessary)
if DEBUG ge 2
	LOGPRINTF <'Creating UMBs'>
endif
	call	create_umbs

if DEBUG ge 2
	LOGPRINTF <'Trapping Int 67h'>
endif

;trap INT 67H for EMM support
        mov     al,67h
        xor     cx,cx                           ;don't save old vector
        mov     dx,offset _DEVDATA:dev_int67    ;offset for new handler
        call    set_interrupt
        cmp     _frame_seg,cx
        jz short no_emm
;we have ems memory, initialize the first entry in f5b_save_area
        push    ds
        pop     es
        mov     di,offset DGROUP:f5b_save_area
        mov     si,offset DGROUP:f4e_list
        call    get_page_map

	; initialize the random key for function 5Dxxh
	;
        mov     ah, 2Ch
        int     21h             ; get DOS time in cx:dx
        mov     key1,cx
        mov     key2,dx

        push    _DEVDATA
        pop     fs

	assume fs:_DEVDATA, ds:DGROUP

        mov dword ptr fs:dev_name[3],'XXXX'     ;Set device name to EMMXXXX0
        mov     fs:ems_enabled,1                ;indicate EMS is enabled

no_emm:
;trap INT 4BH for VDS support
        mov     al,4bh
        xor     cx,cx                           ;don't save old vector
        mov     dx,offset _DEVDATA:int_4b       ;offset of new handler
        call    set_interrupt

;trap int 2fh for Winlight support.
        mov     al,2fh
        mov     cx,offset _DEVDATA:int2f_ptr    ;storage for old vector
        mov     dx,offset _DEVDATA:int_2f       ;offset of new handler
        call    set_interrupt

if DEBUG ge 2
	LOGPRINTF <'Trapping Int 15h'>
endif
	; trap INT 15H in case user reboots
	mov	al,15H
	mov	cx,offset _DEVDATA:int15_ptr
	mov	dx,offset _DEVDATA:int_15
	call	set_interrupt

IF TRAPFFFF
if DEBUG ge 2
	LOGPRINTF <'Trapping reboot JMP'>
endif
	; trap JMP at FFFF:0 so use can reboot via software
	call	trap_FFFF
ENDIF

if DEBUG ge 2
	LOGPRINTF <'Shadowing ROMs'>
endif
        ;implement shadow ram feature if requested on the command line
        call    shadow_ram

if DEBUG ge 2
	LOGPRINTF <'Initializing VDS'>
endif
        call    vds_init                ;initialize the vds driver
;adjust src_move_seg for possible reboot
	mov	ax, cs:uOldMemTop	;get conventional memory size (resized if we relocated EBDA)
        sub     ax,20                   ;drop down 20 Kbytes from top
        shl     ax,6                    ;convert kbytes to segment address
        mov     src_move_seg,ax

	; determine size of this driver
	; if we moved, only the device header (thru ENDDEV) need be saved
	mov	AX, SEG _ENDDEV
	cmp	[moved], 0
	jnz short yes_moved

	; if we did not move, we must save everything
	mov	AX, [malloc_ptr]
	add	AX, 0FH		; we will be rounding to nearest para
	mov	CL, 4
	shr	AX, CL
	mov	CX, DS
	add	AX, CX		; AX = end seg
yes_moved:
	les	BX, [ptrsav]
	mov	word ptr ES:[BX.endaddr], 0
	mov	word ptr ES:[BX.endaddr+2], AX

if DEBUG ge 2
	LOGPRINTF <'EMM Init OK. Driver end address is %x:0'>,ax
endif
	; everthing worked (0=OK)
	xor	AX, AX
	ret

emm_initerr:
if DEBUG ge 2
	LOGPRINTF <'EMM Init failed.'>
endif
	mov	AX, 1
	ret

shadow_ret:
        ret

comment &
        This procedure checks the parameters, rom_start and rom_stop, if they
        are 0, it does nothing.  If they are nonzero, it allocates memory and
        implements a shadow ram feature for the specified range.
&

shadow_ram:
        mov     si,num_rom              ;number of ROM ranges
        shl     si,1                    ;make index into word array
shadow_loop1:
        or      si,si
        jz short shadow_ret
;there are one or more ranges for the ROM=feature
        sub     si,2
        mov     edx,rom_memptr          ;address of physical memory to use
        movzx   ebx,rom_start[si]       ;segment address of ROM
        shl     ebx,4                   ;32 bit address of ROM
        movzx   ecx,rom_stop[si]        ;segment address of stop
        shl     ecx,4                   ;32 bit stop address
;if the stop address was zero, we must set it to 1 meg
        jnz short stop_okay
        mov     ecx,1024*1024           ;takes care of stop at ffffh
stop_okay:
        sub     ecx,ebx                 ;total bytes to move
        add     rom_memptr,ecx          ;update memptr
;ecx = size, ebx = start address, edx = physical memory
        mov     al,2
        call    call_VMM                ;copy ROM into extended memory
        mov     al,1
        call    call_VMM                ;remap extended RAM into ROM space
        xchg    ecx,ebx                 ;move start into ecx
shadow_loop2:
        mov     al,10h
        call    call_VMM                ;write protect a 4K page
        sub     ebx,4096                ;subtract 4K from remaining size
        jbe     shadow_loop2
        jmp     shadow_loop1

IF TRAPFFFF	; {

;----------------------------------------------------
;  This routine remaps some RAM into the BIOS ROM
;  space at the top 4K of memory (FF000-FFFFF).
;  It then reroutes the JMP at FFFF:0 to point to our
;  RAM location.
	public	trap_FFFF
trap_FFFF	proc	near
	; save registers
	push	SI
	push	DI
	push	DS

if DEBUG ge 2
	mov	DX, word ptr [_reboot_memptr+2]
	mov	AX, word ptr [_reboot_memptr]
	LOGPRINTF <'trap_FFFF: reboot_memptr = %x:%x'>,dx,ax
endif
	; make a copy of the BIOS ROM area at 4K from top of mem
	les	DI, [reboot_tmpptr]	; ES:DI = dst in our data area
	mov	AX, 0FF00H
	mov	DS, AX
	xor	SI, SI			; DS:SI = FF00:0 (top 4K)
	mov	CX, 1024		; copy top 4K to our RAM

	; save these pointers, as we'll need 'em again
	push	DS
	push	SI
	push	ES
	push	DI

if DEBUG ge 2
	mov	DX, ES
	mov	AX, DI
	LOGPRINTF <'trap_FFFF: BIOS destination address = %x:%x'>,dx,ax
endif
	cld
	rep	movsd

;now find the far jmp at the top of the BIOS, there may be a cli or
;other instruction there, look for the "EAH".
        mov     al,0eah         ;far jump instruction opcode
        mov     cx,16           ;put a limit on how far we search
        sub     di,cx           ;point to ffff:0
        repne   scasb           ;search for eah.
;di points to the far address in the BIOS for a reboot.
        push    cs
        push    offset reboot_FFFF0
        pop     eax
        xchg    eax,es:[di]             ;get the original address
        mov     cs:old_FFFF0,eax        ;save original

	; remap memory into top 4K (disabling interrupts)
	; ReMap386(ProtAddr,RealAddr,Size);
	call	getDGROUP
;	cli		; disable interrupts so BIOS does not get called!
	xor	DX, DX
	mov	AX, 4*1024
	push	DX
	push	AX				; size = 4K
	mov	DX, 000FH
	mov	AX, 0F000H
	push	DX
	push	AX				; RealAddr = FF000H
	push	word ptr [_reboot_memptr+2]
	push	word ptr [_reboot_memptr]	; ProtAddr
	call	_ReMap386
	cli		; disable interrupts so BIOS does not get called!
	add	SP, 12

	; get pointers back, reversing src and dst
	pop	SI
	pop	DS
	pop	DI
	pop	ES

	; copy 4K into top of memory
	mov	CX, 1024
	cld
	rep	movsd
	sti			; OK to have interrupts now

if DEBUG ge 2
	LOGPRINTF <'trap_FFFF: new top of ROM = %x:%x.'>,dx,ax
	LOGPRINTF <'trap_FFFF: Write protecting ROM copy.'>
endif
	; write protect the memory
	; write_protect_mem(addr)
	mov	DX, 000FH
	mov	AX, 0F000H
	push	DX
	push	AX				; RealAddr = FF000H
	call	_write_protect_mem
	pop	CX
	pop	CX

	pop	DS
	pop	DI
	pop	SI
	ret
trap_FFFF	endp

ENDIF		; } TRAPFFFF

;-----------------------------------
;   On Entry
;       AL = int #
;       CX = offset of storage for old vector, zero if don't store
;       DX = offset of new handler
;       All storage and handlers must be in the _DEVDATA segment
;
	public	set_interrupt
set_interrupt	proc	near
        push    DS
        push    _DEVDATA
        pop     ds
        or      cx,cx
        jz short no_save
        mov     ah,35h
        int     21h
        push    es
        push    bx
        mov     bx,cx           ;get offset for old vector storage
        pop dword ptr ds:[bx]   ;store the old vector
no_save:
        mov     ah,25h
        int     21h             ;set the vector to ds:dx
        pop     ds
        ret
set_interrupt	endp

cEndCode

;/////////////////////////////////////////////////////////////////////////
;//	XMS Support 							//
;/////////////////////////////////////////////////////////////////////////
cBegCode

comment ~ ; prototype ;
void alloc_from_xms(void)
~
c_func	alloc_from_xms
	; allocate the memory from XMS
        mov     DX,16                   ;16K for 386 data, etc...
        add     DX,_range_memK          ; enough for ranges
        add     DX,_frame_memK          ; enough for frame pages
;must add enough for page tables in case windows starts.
        add     dx,word ptr win_page_table_ptr
        add     dx,rom_memK             ;enough for ROM ranges
        add     dx,vds_buff_memk        ;enough for VDS services
	mov	[_total_memK], DX
	mov	AH, 9			; Allocate XMS Memory
        call    xms_handler
;ax = 1 if success
        dec     ax              ;0 if success
        jnz short insuf_mem       ;jmp if failure
	mov	[xms_handle], DX

	; lock the XMS memory (which also determines address)
	mov	AH, 0CH			; Lock XMS Memory
	call	dword ptr [xms_handler]	; addr returned in DX:BX
        push    dx
        push    bx
        pop     ebx                     ;address in ebx

	; store physical address of memory
        mov     _phys_memptr,ebx
        movzx   eax,word ptr vds_buff_memk
        shl     eax,10                  ;multiply by 1024
        add     ebx,eax
        mov     vmm_memptr,ebx
        add     ebx,3000H               ;leave 12K space for 386 I/O map

	; note: "physical" and "range" memory MUST be together
	; in case we move EMM code into 1st range

	; store range memory address
        mov     _range_memptr,ebx
        movzx   eax,word ptr _range_memK        ;add size of ranges
        shl     eax,10                          ;convert kbytes to bytes
        add     ebx,eax

IF TRAPFFFF
	; store BIOS reboot memory address
        mov     _reboot_memptr,ebx
        add     ebx,1000h
ENDIF

        ;store address of RAM used to implement ROM= feature.
        mov     rom_memptr,ebx
        movzx   eax,word ptr rom_memK
        shl     eax,10                  ;convert Kbytes to bytes
        add     ebx,eax
;save address of ems page tables
        movzx   eax,word ptr win_page_table_ptr
        shl     eax,10                  ;convert kbytes to bytes
        mov     win_page_table_ptr,ebx
        add     ebx,eax
	; store frame memory address
        mov     _frame_memptr,ebx
        xor     ax,ax                   ;return 0 for success
	ret

insuf_mem:
        PRERROR  insuf_mem_msg
        ret                             ;ax is already nonzero


c_endp	alloc_from_xms

comment ~ ; prototype ;
void free_xms(void)
~
c_func	free_xms
	mov	AH, 0DH		; unlock XMS
	mov	DX, [xms_handle]
	call	dword ptr [xms_handler]
	mov	AH, 0AH
	call	dword ptr [xms_handler]
	ret
c_endp	free_xms

;--------------------------------
;  On Entry
;	AX = paras
;  On Exit
;	DX:AX = (long) value
;
	public	para2long
para2long	proc near
        xor     DX,DX
        shld    DX,AX,4
        shl     ax,4
        ret

para2long	endp


comment ~ ; prototype ;
/*
   Remap memory into all the ranges.
   Returns TRUE if worked, FALSE if failed
*/
int remap_ranges(ulong physmem, struct range ranges[], int limit)
~
	public _remap_ranges
_remap_ranges	proc	near
	push	BP
	mov	BP, SP
	push	SI			; SI is i
	push	DI
	mov	DI, P[BP+4]		; DI is ranges (pointer)

;
;	/*
;	   Determine if we can move EMM to upper memory.
;	   Basically, if any range is big enough, stuff
;	   it there, and shrink that range.
;	/*
;
	mov	[dest_move_seg], CS	; default values
	mov	[src_move_seg], CS

	cmp	byte ptr _donot_load_high, 1
	jne	short try_and_load_high

	jmp	short done_stuffing


try_and_load_high:
	mov	SI, [_first_range]	; SI = starting range #
	mov	AL, SIZE RANGE
	mul	byte ptr [_first_range]
	xchg	BX, AX			; BX = index into ranges
	call	CodeSize4K		; get code+data size (in bytes)
	mov	CL, 4
	shr	AX, CL			; AX = paras for code+data

; DEBUGGING!!
IF 0
	jmp	done_stuffing		; don't move EMM code
ENDIF

again_stuff:
	cmp	SI, P[BP+6]		; num_ranges
	jae short done_stuffing
	cmp	[DI+BX.range_len], AX	; enough space?
	ja short got_stuff		; Range must be BIGGER than codesize
	add	BX, SIZE RANGE
	inc	SI
	jmp	short again_stuff

got_stuff:
	inc	[moved]			; we're moving code, set to TRUE
	mov	DX, [DI+BX.range_start]
	mov	[dest_move_seg], DX	; record "to" segment
	sub	[DI+BX.range_len], AX	; subtract paras
	add	[DI+BX.range_start], AX
	mov	DX, 16
	mul	DX			; DX:AX = bytes of code
	add	word ptr P[BP], AX
	adc	word ptr P[BP+2], DX	; update physmem ptr to acct for code
done_stuffing:

;
;	/* Initialize V86 mode for remapping */
;	/* V86Init(ulong physaddr, ulong from, ulong to, unsigned size4K)
;	if(V86Init(physaddr,from,to,size4K) != 0)
;	   return 0;
;
	call	CodeSize4K
	push	AX			; size4K

if DEBUG ge 2
	LOGPRINTF <'remap_ranges: V86Init -- physaddr= %lx'>,vmm_memptr
	LOGPRINTF <'              from= %x, to= %x, size4K= %x'>,src_move_seg,dest_move_seg,ax
endif
	mov	AX, [dest_move_seg]
	call	para2long
	push	DX
	push	AX			; to
	mov	AX, [src_move_seg]
	call	para2long
	push	DX
	push	AX			; from
	push	word ptr [vmm_memptr+2]
	push	word ptr [vmm_memptr]	; physaddr
	call	_V86Init
	add	SP, 14
	or	AX, AX
	jz short remap_ok
	jmp short remap_failed
remap_ok:

if DEBUG ge 2
	LOGPRINTF <'remap_ranges: CS= %x, DS= %x, SS= %x'>,cs,ds,ss
endif

;
;	/* Remap each piece of memory */
;	for(i=0; i<limit; ++i)
;	   if(ReMap386(ProtAddr,RealAddr,Size) != 0)
;	      return 0;
;
	xor	SI, SI			; SI is i (i=0)
	xor	BX, BX
reloop:	cmp	SI, P[BP+6]		; limit
	jae short done_reloop
	push	BX	; save ranges index in case ReMap trashes it

	; ReMap386(ProtAddr,RealAddr,Size);
if DEBUG ge 2
	push	bx
	push	cx
	push	esi
endif
	mov	AX, [DI+BX.range_len]
	call	para2long
	push	DX				; push Size (in bytes)
	push	AX
if DEBUG ge 2
	mov	cx, dx
	mov	bx, ax
endif
	mov	AX, [DI+BX.range_start]
	call	para2long
	push	DX				; push RealAddr
	push	AX
	push	word ptr P[BP+2]		; push ProtAddr
	push	word ptr P[BP]
if DEBUG ge 2
	mov	esi,P[bp]
	LOGPRINTF <'remap_ranges: ProtAddr= %lx, RealAddr= %x:%x, Size= %x%x'>,esi,dx,ax,cx,bx
endif
	call	_ReMap386
	add	SP, 12
if DEBUG ge 2
	pop	esi
	pop	cx
	pop	bx
endif
	or	AX, AX
	jnz short remap_failed2

	; on to next range
	pop	BX			; restore ranges index
	mov	AX, [DI+BX.range_len]
	call	para2long
	add	word ptr P[BP], AX
	adc	word ptr P[BP+2], DX
	add	BX, size RANGE
	inc	SI			; ++i
if DEBUG ge 2
	jmp	reloop
else
	jmp	short reloop
endif
done_reloop:
	mov	AX, 1
if DEBUG ge 2
	jmp	remap_done
else
	jmp	short remap_done
endif
remap_failed2:

if DEBUG ge 2

	LOGPRINTF <'remap_ranges: Remap failed, code: %x'>,ax
endif
	push	ax
	call	disable_V86
	pop	ax

remap_failed:
if DEBUG ge 2
	LOGPRINTF <'remap_ranges: V86 mode disabled'>
endif
	mov	init_err_code, ax
	xor	AX, AX
remap_done:
	pop	DI
	pop	SI
	mov	SP, BP
	pop	BP
	ret
_remap_ranges	endp

; convert a linear address (DX:AX) into a segment address (AX)
; destroys CX, DX
	public	linear2seg
linear2seg:
	mov	CL, 4
	shr	AX, CL
	mov	CL, 12
	shl	DX, CL
	or	AX, DX
	ret

;--------------------------------------
; void V86off(ulong from, ulong to, unsigned size4K)
disable_V86 proc near
if DEBUG ge 2
	push	bx
	push	cx
	push	si
endif
	call	CodeSize4K
	push	AX				; size4K
if DEBUG ge 2
	mov	si, ax
endif
	mov	AX, [src_move_seg]           ;this is where we are moving to.
	call	para2long
	push	DX
	push	AX			; to
if DEBUG ge 2
	mov	cx, dx
	mov	bx, ax
endif
	mov	AX, [dest_move_seg]          ;this is actually where emm is now
	call	para2long
	push	DX
	push	AX			; from
if DEBUG ge 2
	LOGPRINTF <'disable_V86: from= %x%x to= %x%x, size4K= %x'>,dx,ax,cx,bx,si
endif
	call	_V86off
	add	SP, 10
if DEBUG ge 2
	pop	si
	pop	cx
	pop	bx
endif
	ret
disable_V86 endp


comment ~ ; prototype ;
/*
   Copy code from one linear address to another.  Assumes
   that both addresses are in the 1Meg+64K real mode
   address space.  Cannot fail.
*/
void CopyCode(ulong from, ulong to)
~
c_func	CopyCode
	push	BP
	mov	BP, SP
	push	SI
	push	DI
	push	DS

	; determine length of code and data
	; since we're called from MAPPER code
	call	CodeSize4K
	xchg	CX, AX			; CX = bytes to copy

	; set DS:SI to "from" address
	mov	AX, P[BP]
	mov	DX, P+2[BP]
	call	linear2seg
	mov	DS, AX
	xor	SI, SI

	; set ES:DI to "to" address
	mov	AX, P+4[BP]
	mov	DX, P+6[BP]
	call	linear2seg
	mov	ES, AX
	xor	DI, DI

	; actually move all of code and data
	cld
	rep	movsb

	pop	DS
	pop	DI
	pop	SI
	pop	BP
	ret
c_endp	CopyCode


;----------------------------------------
; On Entry
;	DS:DX points to a string to print
	public	print_dos
print_dos:
	mov	AH, 9
	int	21h
	ret
cEndCode

;/////////////////////////////////////////////////////////////////////////
;//	Terminate and Stay Resident Code				//
;/////////////////////////////////////////////////////////////////////////
cBegCode
_astart:


IF TSR
	; set up our stack
	mov	AX, DGROUP
	mov	DS, AX
	cli
	mov	SS, AX
	mov	SP, GOFFSET stacktop
	sti

	; zero the uninitialized C data (BSS)
	call	zero_bss

	; call the originial INIT routine
	mov	SI, ES				; save ES (PSP)
	call	getDEVDATA
	mov	ES, AX
	mov	BX, offset _DEVDATA:req_hdr
	mov	byte ptr ES:[BX.cmdc], 0
	mov	word ptr ES:[BX.cmdline], 81H
	mov	word ptr ES:[BX+cmdline+2], SI	; PSP:81H

	; upper case the DOS command line to be like a device arg list
	;
	push	ax		; save what we'll trash
	push	cx
	push	si
	push	di
	push	ds
	push	es

	mov	ds, si
	mov	es, si
	mov	si, 80h
	xor	cx, cx
	mov	cl, [si]	; cx = command tail char count
	inc	si
	mov	di, si		; ds:si = es:di = start of command tail text

    upper_case_cmdln_loop:
	lodsb             	; read char
	cmp	al, 'a'
	jb	upper_case_next_char
	cmp	al, 'z'
	ja	upper_case_next_char

	sub	al, ('a'-'A')   ; convert char to upper case

    upper_case_next_char:
	stosb			; write char
	loop	upper_case_cmdln_loop

	pop	es		; done, restore regs
	pop	ds
	pop	di
	pop	si
	pop	cx
	pop	ax

	; initialize the device
	push	CS
	call	emm_strat
	; we can (and must) call emm_init directly, not through emm_serve
	call	emm_init
	or	AX, AX
	jnz	tsr_error

	; hook into device chain
	assume	DS:_DEVDATA
	push	DS
	call	getDEVDATA
	mov	DS, AX
	mov	AH, 52H
	int	21H		; returned in ES:BX
	mov	AX, ES:[48H]
	mov	DX, ES:[48H+2]	; get pointer to 1st device
	mov	word ptr DS:[_emmx], AX
	mov	word ptr DS:[_emmx+2], DX
	cli
	mov	word ptr ES:[48H], offset _DEVDATA:_emmx
	mov	word ptr ES:[48H+2], DS
	sti
	pop	DS
	assume	DS:DGROUP

	; Terminate and Stay Resident
	call	getDEVDATA
	mov	ES, AX
	mov	BX, offset _DEVDATA:req_hdr
	mov	DX, ES:[BX.endaddr+2]	; end address (calcd by init)
	sub	DX, AX

TODO
1) Must relocate the XBDA to the end of the allocated code
2) Must add XBDA para size to "keep" memory value returned to DOS

	add	DX, 10H		; for PSP
	mov	AX, 3100H
	int	21H
ELSE

	call	getDGROUP

IF 0
IF USING_GEOS
   mov   byte ptr [signon_msg+0],13    ;
   mov   byte ptr [signon_msg+1],10    ; re-enable our signon
   mov   byte ptr [signon_msg+2],'E'   ;
ENDIF
ENDIF

IF SIGNON
	; print a sign-on message
	printm	signon_msg
ENDIF
	printm	device_only

ENDIF ; TSR

	; Error! Terminate
tsr_error:
	mov	AX, 4C01H
	int	21H
cEndCode


;/////////////////////////////////////////////////////////////////////////
;//	Command Line Parsing						//
;/////////////////////////////////////////////////////////////////////////
cBegCode

comment ~ ; prototype ;
int toupper(char c)
~
	public _toupper
_toupper	proc	near
	push	bp
	mov	bp,sp
	mov	dl,byte ptr [bp+4]
   ;
   ;	{
   ;		if(c>='a' && c<='z')
   ;
	cmp	dl,97
	jl	short @2@98
	cmp	dl,122
	jg	short @2@98
   ;
   ;			c = c-'a'+'A';
   ;
	mov	al,dl
	add	al,224
	mov	dl,al
@2@98:
   ;
   ;		return c;
   ;
	mov	al,dl
	cbw
	jmp	short @2@122
@2@122:
   ;
   ;	}
   ;
	pop	bp
	ret
_toupper	endp


comment ~ ; prototype ;
int isterminator(char c)
~
	public _isterminator
_isterminator	proc	near
	push	bp
	mov	bp,sp
	mov	dl,byte ptr [bp+4]
	xor	ax,ax
   ;
   ;	{
   ;		return (c<' ' && c!='\t');
   ;
	cmp	dl, ' '
	jge	short @3@98
	cmp	dl, 9
	je	short @3@98
	inc	ax
@3@98:
   ;
   ;	}
   ;
	pop	bp
	ret
_isterminator	endp

comment ~ ; prototype ;
int isseparator(char c)
~
	public _isseparator
_isseparator	proc	near
	push	bp
	mov	bp,sp
	mov	dl,byte ptr [bp+4]
	xor	ax, ax
   ;
   ; {
   ;	return (c==' ' || c=='\t' || c==',' || c=='/' || c=='-' || c=='=');
   ;
	cmp	dl, ' '
	je	short @4@98
	cmp	dl, 9
	je	short @4@98
	cmp	dl, ','
	je	short @4@98
	cmp	dl, '/'
	je	short @4@98
	cmp	dl, '-'
	je	short @4@98
	cmp	dl, '='
	jne	short @4@122
@4@98:
	inc	ax
@4@122:
   ;
   ; }
   ;
	pop	bp
	ret
_isseparator	endp

comment ~ ; prototype ;
int isdigit(char c)
~
	public _isdigit
_isdigit	proc	near
	push	bp
	mov	bp,sp
	mov	dl,byte ptr [bp+4]
   ;
   ;	{
   ;		return (c>='0' && c<='9');
   ;
	cmp	dl,48
	jl	short @5@98
	cmp	dl,57
	jg	short @5@98
	mov	ax,1
	jmp	short @5@122
@5@98:
	xor	ax,ax
@5@122:
	jmp	short @5@146
@5@146:
   ;
   ;	}
   ;
	pop	bp
	ret
_isdigit	endp

comment ~ ; prototype ;
int isxdigit(char c)
~
	public _isxdigit
_isxdigit	proc	near
	push	bp
	mov	bp,sp
   ;
   ;	{
   ;		c = toupper(c);
   ;
	mov	al,byte ptr [bp+4]
	push	ax
	call	_toupper
	pop	cx
	mov	byte ptr [bp+4],al
   ;
   ;		return (isdigit(c) || (c>='A' && c<='F'));
   ;
	mov	al,byte ptr [bp+4]
	push	ax
	call	_isdigit
	pop	cx
	or	ax,ax
	jne	short @6@98
	cmp	byte ptr [bp+4],65
	jl	short @6@122
	cmp	byte ptr [bp+4],70
	jg	short @6@122
@6@98:
	mov	ax,1
	jmp	short @6@146
@6@122:
	xor	ax,ax
@6@146:
	jmp	short @6@170
@6@170:
   ;
   ;	}
   ;
	pop	bp
	ret
_isxdigit	endp


comment ~ ; prototype ;
unsigned far_atoi(char far *s)
~
	public _far_atoi
_far_atoi	proc	near
	push	bp
	mov	bp,sp
	push	si
   ;
   ;	{
   ;		unsigned val=0;
   ;
	xor	si,si
	jmp	short @7@74
@7@50:
   ;
   ;		while(isdigit(*s))
   ;		{
   ;			val*= 10;
   ;
	mov	dx,10
	mov	ax,si
	imul	dx
	mov	si,ax
   ;
   ;			val+= *s-'0';
   ;
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	cbw
	add	ax,65488
	add	si,ax
   ;
   ;			++s;
   ;
	inc	word ptr [bp+4]
@7@74:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isdigit
	pop	cx
	or	ax,ax
	jne	short @7@50
   ;
   ;		}
   ;		return val;
   ;
	mov	ax,si
	jmp	short @7@122
@7@122:
   ;
   ;	}
   ;
	pop	si
	pop	bp
	ret
_far_atoi	endp

comment ~ ; prototype ;
unsigned far_atoh(char far *s)
~
	public _far_atoh
_far_atoh	proc	near
	push	bp
	mov	bp,sp
	push	si
   ;
   ;	{
   ;		unsigned h=0;
   ;
	xor	si,si
	jmp	short @8@170
@8@50:
   ;
   ;		char c;
   ;
   ;		while(isxdigit(*s))
   ;		{
   ;			h*= 16;
   ;
	mov	dx,16
	mov	ax,si
	imul	dx
	mov	si,ax
   ;
   ;			h+= (*s>='0' && *s<='9' ? *s-'0' : toupper(*s) - 'A' + 10);
   ;
	les	bx,dword ptr [bp+4]
	cmp	byte ptr es:[bx],48
	jl	short @8@122
	les	bx,dword ptr [bp+4]
	cmp	byte ptr es:[bx],57
	jg	short @8@122
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	cbw
	add	ax,65488
	jmp	short @8@146
@8@122:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_toupper
	pop	cx
	add	ax,65481
@8@146:
	add	si,ax
   ;
   ;			++s;
   ;
	inc	word ptr [bp+4]
@8@170:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isxdigit
	pop	cx
	or	ax,ax
	jne	short @8@50
   ;
   ;		}
   ;		return h;
   ;
	mov	ax,si
	jmp	short @8@218
@8@218:
   ;
   ;	}
   ;
	pop	si
	pop	bp
	ret
_far_atoh	endp

comment ~ ; prototype ;
char far *nextparm(char far *s)
~
	public _nextparm
_nextparm	proc	near
	push	bp
	mov	bp,sp
	jmp	short @9@74
@9@50:
   ;
   ;	{
   ;		while(!isterminator(*s) && !isseparator(*s))
   ;			++s;
   ;
	inc	word ptr [bp+4]
@9@74:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isterminator
	pop	cx
	or	ax,ax
	jne	short @9@122
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isseparator
	pop	cx
	or	ax,ax
	je	short @9@50
@9@122:
	jmp	short @9@170
@9@146:
   ;
   ;		while(!isterminator(*s) && isseparator(*s))
   ;			++s;
   ;
	inc	word ptr [bp+4]
@9@170:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isterminator
	pop	cx
	or	ax,ax
	jne	short @9@218
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isseparator
	pop	cx
	or	ax,ax
	jne	short @9@146
@9@218:
   ;
   ;		return s;
   ;
	mov	dx,word ptr [bp+6]
	mov	ax,word ptr [bp+4]
	jmp	short @9@242
@9@242:
   ;
   ;	}
   ;
	pop	bp
	ret
_nextparm	endp


comment ~ ; prototype ;
int fmemcmpi(char far *s1, char far * s2, int n)
~
	public _fmemcmpi
_fmemcmpi	proc	near
	push	bp
	mov	bp,sp
	dec	sp
	dec	sp
	push	si
	mov	si,word ptr [bp+12]
	jmp	short @10@74
@10@50:
   ;
   ;	{
   ;		char c1,c2;
   ;		while(--n>0 && (c1=toupper(*s1))==(c2=toupper(*s2)))
   ;			++s1, ++s2;
   ;
	inc	word ptr [bp+4]
	inc	word ptr [bp+8]
@10@74:
	dec	si
	jle	short @10@122
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_toupper
	pop	cx
	mov	byte ptr [bp-1],al
	push	ax
	les	bx,dword ptr [bp+8]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_toupper
	pop	cx
	mov	byte ptr [bp-2],al
	pop	dx
	cmp	dl,al
	je	short @10@50
@10@122:
   ;
   ;		if(c1 == c2)
   ;
	mov	al,byte ptr [bp-1]
	cmp	al,byte ptr [bp-2]
	jne	short @10@194
   ;
   ;			return 0;
   ;
	xor	ax,ax
	jmp	short @10@266
@10@194:
   ;
   ;		else if(c1 < c2)
   ;
	mov	al,byte ptr [bp-1]
	cmp	al,byte ptr [bp-2]
	jge	short @10@242
   ;
   ;			return -1;
   ;
	mov	ax,65535
	jmp	short @10@266
@10@242:
   ;
   ;		return 1;
   ;
	mov	ax,1
@10@266:
   ;
   ;	}
   ;
	pop	si
	mov	sp,bp
	pop	bp
	ret
_fmemcmpi	endp

comment ~ ; prototype ;
unsigned getnum(char far **sptr, int base)
~
	public _getnum
_getnum	proc	near
	push	bp
	mov	bp,sp
	push	si
	push	di
	mov	si,word ptr [bp+4]
   ;
   ;	{
   ;		unsigned val;
   ;		if(base == 16)
   ;
	cmp	word ptr [bp+6],16
	jne	short @11@74
   ;
   ;			val = far_atoh(*sptr);
   ;
	push	word ptr [si+2]
	push	word ptr [si]
	call	_far_atoh
	pop	cx
	pop	cx
	mov	di,ax
	jmp	short @11@98
@11@74:
   ;
   ;		else
   ;			val = far_atoi(*sptr);
   ;
	push	word ptr [si+2]
	push	word ptr [si]
	call	_far_atoi
	pop	cx
	pop	cx
	mov	di,ax
@11@98:
	jmp	short @11@146
@11@122:
   ;
   ;		while(isxdigit(**sptr))
   ;			++(*sptr);
   ;
	inc	word ptr [si]
@11@146:
	les	bx,dword ptr [si]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isxdigit
	pop	cx
	or	ax,ax
	jne	short @11@122
   ;
   ;		return val;
   ;
	mov	ax,di
	jmp	short @11@194
@11@194:
   ;
   ;	}
   ;
	pop	di
	pop	si
	pop	bp
	ret
_getnum	endp


comment ~ ; prototype ;
void parse_cmdline(char far * s)
~
	public	_parse_cmdline
_parse_cmdline	proc	near
	push	bp
	mov	bp,sp
	push	si
	push	di

	sub	sp,64	; make space for i_cmd, x_cmd
	; [bp-4]
i_cmd	equ	word ptr [bp-4-32]
x_cmd	equ	word ptr i_cmd-32
	; clear i_cmd, x_cmd
	mov	ax,0
	mov	cx,32
	push	ss
	pop	es
	lea	di,x_cmd
	cld
	rep	stosw

comment &
autodetection code moved here to give cmdline parameters precedence
over autodetection results. If user knows what he doing - make so.

        We have an assumed i_range of C800-F000,
	set those bits in i_range before proceeding.
&
        cld
        mov     di,offset DGROUP:i_range[25]
        mov     cx,5
        mov     al,-1
        push    ds
        pop     es
        rep     stosb           ;assumed range of C800-F000 is now in place

;We must search for BIOS extentions in the C000-FFFF range.

        mov     dx,0C000h               ; scan start segment
xtd_scan:
        mov     es,dx                   ; get target seg
        mov     ax,es:[0]               ; get id field
        cmp     ax,0AA55h               ; is it 55AAh
        jnz short next_page               ; else try next 2K
;Now do a checksum
        xor     cx,cx
        mov     ch,es:[2]                       ;get length byte
;length byte is number of 512 byte pages in extention, we have it in ch which
;means cx has the number of words in the extention.
        xor     ax,ax
ext_loop1:
        xor     si,si
ext_loop:
        add     al,es:[si]
        inc     si
        add     al,es:[si]
        inc     si
        loopnz  ext_loop        ;loop until cx = 0 or until si wraps
        jcxz short check_sum_done
;si has wrapped, add 64K to es
        mov     si,es
        add     si,4096         ;64K in segment arithmetic
        mov     es,si
        jmp     ext_loop1

check_sum_done:
;check sum is now in al, it should = zero.
        or      al,al
        mov     es,dx                   ;es may have been trashed
        jnz short next_page             ;jmp if not a valid BIOS extention
;This is a valid BIOS extention, mark it in the x_range
        movzx   cx,byte ptr es:[2]      ; get length in 512 byte pages
        add     cx,7
        shr     cx,3                    ;round up to nearest 4K
;cx contains the number of 4K pages in the extention
        movzx   bx,dh                   ;round start down to 4K
ext_loop2:
        bts word ptr x_range,bx
        inc     bx
        loop    ext_loop2

next_page:
        add     dx,0080h
        jnc     xtd_scan

;check ranges to see if they are all 1s or all 0s
        mov     bx,0c8h         ;start of range to check
det_loop1:
        bt word ptr x_range,bx
        jc short det_a
;check the 4K page specified by bx
        mov     ax,bx
        shl     ax,8
        mov     es,ax
        xor     di,di
        mov     ax,es:[di]      ;get first byte in the range
        mov     cx,2048
        repe    scasb
        jne short det_b         ;jmp if block is not all the same
;now check if block is all 1s or all 0s
        mov     cx,ax
        or      ax,ax
        jz short det_c          ;jmp if all 0s
        inc     ax
        jnz short det_b         ;jmp if not all 1s
det_c:
;check if we can write to the block
        not     cx              ;invert value stored in block
        mov     ax,cx
        pushf
        cli
        xchg    ax,es:[0]       ;try to invert all bits in the first word
        xchg    ax,es:[0]       ;restore the word in case it is RAM
        popf
comment &
        We were checking to insure that AX was equal to the value of the
        block, some machines do not have positive enough pull ups to allow
        this test to work.  The test below should detect genuine RAM.
&
        cmp     ax,cx           ;ax should NOT equal cx unless the block is RAM
        jne short det_a         ;jmp if okay

;Block is not perfect empty space, don't risk it.
det_b:  ;the range specified in bx must be marked as an X range
        bts word ptr x_range,bx
det_a:
        inc     bl
        jnz short det_loop1     ;loop through 1 meg
	jmp short @12@98

   ;
   ;	{
   ;		static char framestr[] = "FRAME=";
   ;		unsigned end;
   ;
   ;		// skip device name
   ;		s = nextparm(s);
   ;
   ;	push	word ptr [bp+8]
   ;	push	word ptr [bp+6]
   ;	call	_nextparm
   ;	pop	cx
   ;	pop	cx
   ;	mov	word ptr [bp+8],dx
   ;	mov	word ptr [bp+6],ax
   ;	jmp	@12@434
   ;
@12@98:

IF ENABLE_MOVEEBDA

	; Check for the NOMOVEXBDA switch
	;
	push	10
	push	word ptr [bp+6]
	push	word ptr [bp+4]
	push	ds
	push	offset DGROUP:szEBDAStr
	call	_fmemcmpi
	add	sp,10
	or	ax,ax
	jne	short CmdFrame


	add	word ptr [bp+6],10	; set position to next arg in command line

	mov	fRelocateEBDA, 0	; Don't relocate EBDA

	jmp	@12@410			; Do the next argument


CmdFrame:

ENDIF ; IF ENABLE_MOVEEBDA

   ;
   ;
   ;		// parse the whole string
   ;		while(!isterminator(*s))
   ;		{
   ;			// is it the FRAME designation?
   ;			if(fmemcmpi(framestr,s,sizeof(framestr)-1) == 0)
   ;
	push	6
	push	word ptr [bp+6]
	push	word ptr [bp+4]
	push	ds
	push	offset DGROUP:framestr
	call	_fmemcmpi
	add	sp,10
	or	ax,ax
	jne	short check_vds
   ;
   ;			{
   ;				s += sizeof(framestr)-1;	// skip the "frame=" part
   ;
	add	word ptr [bp+4],6
   ;
   ;				frame_seg = getnum(&s,16);
   ;
	mov	ax,16
	push	ax
	lea	ax,word ptr [bp+4]
	push	ax
	call	_getnum
	pop	cx
	pop	cx
        mov     _frame_seg,ax
        add     _frame_seg[2],ax
        add     _frame_seg[4],ax
        add     _frame_seg[6],ax                ;for function 4eh
   ;
   ;
   ;				// determine how much memory to devote to EMS
	mov	word ptr DGROUP:_frame_memK,0        ;assume no memory reserved for EMS
   ;
   ;				if(*s == ',')
   ;
	les	bx,dword ptr [bp+4]
	cmp	byte ptr es:[bx],','
	jne	short @12@170
   ;
   ;				{
   ;					++s;
   ;
	inc	word ptr [bp+4]
   ;
   ;					frame_memK = getnum(&s,10);
   ;
        push    10              ;use base 10 numbers
	lea	ax,word ptr [bp+4]
	push	ax
	call	_getnum
	pop	cx
	pop	cx
	mov	word ptr DGROUP:_frame_memK,ax
@12@170:
   ;
   ;				}
   ;				frame_memK &= ~0xF;		// truncate memory to nearest 16K chunk
   ;
	and	word ptr DGROUP:_frame_memK,65520
   ;
   ;			}
   ;
	jmp	@12@410

check_vds:      ;see if the string is "D="
        les     bx,[bp+4]               ;pointer to string
        cmp word ptr es:[bx],'=D'       ;cmp to D=
	jne short check_max		;jmp if not "D="

        add word ptr [bp+4],2   ;point to end of string, start of number
        push    10              ;use base 10 numbers
        lea     ax,word ptr [bp+4]
        push    ax              ;pointer to the pointer
        call	_getnum
        add     sp,4            ;pop garbage
;ax is the number of K bytes to use for DMA buffering
        add     ax,3
        and     al,0fch         ;round up to nearest 4K
        cmp     ax,256
        jb short vds_okay
        mov     ax,256          ;allow no larger than 256K buffer
vds_okay:
        mov     vds_buff_memk,ax
        jmp     @12@410

check_max:
        les     bx,[bp+4]       	;pointer to string
        cmp dword ptr es:[bx],'=XAM'    ;cmp to MAX=
        jne short check_low         	;jmp if not "MAX="

        add 	word ptr [bp+4],4   	;point to end of string, start of maxK
        push    10   		        ;use base 10 numbers
	lea	ax,word ptr [bp+4]
	push	ax
	call	_getnum
	pop	cx
	pop	cx
	mov	word ptr DGROUP:_frame_maxK, ax
        jmp     @12@410


check_low:
	les	bx,[bp+4]
	cmp	word ptr es:[bx],'OL'
	jne short check_PS2
	cmp	byte ptr es:[bx+2], 'W'
	jne short check_PS2

	add	word ptr [bp+4], 3
	mov	byte ptr DGROUP:_donot_load_high, 1
	jmp	@12@410

check_PS2:
	les	bx, [bp+4]
	cmp	word ptr es:[bx], 'SP'
	jne short check_ram
	cmp	byte ptr es:[bx+2], '2'
	jne short check_ram

	add	word ptr [bp+4], 3
	mov	byte ptr cs:use_PS2_A20, 1
	jmp	@12@410


	; see if the string is "RAM"
	; The "RAM" substring may or may not be followed by an
	; '=' char and a range description
	;
	; NOTE:	Right now, we do nothing here except parse the
	;	command line. Any (assumed) arguments are just
	;	thrown away.
	;
check_ram:
        les     bx,[bp+4]       		; get pointer to substring
        cmp 	word ptr es:[bx],'AR'    	; cmp to "RA"
        jne 	check_rom       		; jmp if not "RAM"
	cmp	byte ptr es:[bx+2], 'M'
        jne 	check_rom       		; jmp if not "RAM"

	add 	word ptr [bp+4],3		; skip past token
	cmp	byte ptr es:[bx+3], '=' 	; is a range being specified?
	je short check_ram_find_start_of_range   ; process it if so

	; If no range value is specified for the RAM option,
	; then MS EMM386 uses all of the "adapter space" for UMBs
	; and the LIM 3.2 page frame.
	;
    check_ram_no_range_spec:
	mov	ax, DEFAULT_UMB_START
	mov	bx, DEFAULT_UMB_END
	jmp	short check_ram_have_range_values

    check_ram_find_start_of_range:
        inc 	word ptr [bp+4]     	; skip separator char
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]	; al = current cmd line cahr
	push	ax
	call	_isseparator	; is cur char a separator?
	pop	cx
	or	ax,ax
	jnz	short check_ram_find_start_of_range ; skip cur separator char

	les	bx,dword ptr [bp+4]	; cur char not separator char
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isxdigit 	; error if cur char isn't a hex digit
	pop	cx
	or	ax,ax
	je	short check_ram_syntax_error	; handle ill formed range spec

    check_ram_get_start_of_range:
        lea     ax,[bp+4]       	; pointer to the start spec substring
        push    16              	; use base 16 numbers
        lea     ax,word ptr [bp+4]
        push    ax              	; pass pointer to the pointer
        call 	_getnum
        pop     cx
	pop	cx

	; AX = starting segment of RAM range
	; The pointer argument on the stack now points to just past
	; the "start of range" spec substring.
	;
    check_ram_have_start_of_range:
    	push	ax			; save the start of range value

	; there must be at least one separator char between the start
	; and the stop range specs...
	;
    check_ram_find_end_of_range:
        inc 	word ptr [bp+4]     	; skip separator char
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]	; al = current cmd line cahr
	push	ax
	call	_isseparator	; is cur char a separator?
	pop	cx
	or	ax,ax
	jne	short check_ram_find_end_of_range ; skip cur separator char

	les	bx,dword ptr [bp+4]	; cur char not separator char
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isxdigit 	; error if cur char isn't a hex digit
	pop	cx
	or	ax,ax
	pop	ax				; AX = start seg value
	je	short check_ram_syntax_error	; handle ill formed range spec

	push	ax			; save start of range value again

    check_ram_get_end_of_range:
        lea     ax,[bp+4]       	; pointer to the end spec substring
	push	16
        push    ax
        call 	_getnum
        pop     cx
        pop     cx

	; AX = ending segment value of the RAM range, the
	; starting segment value of the range is on the top of
	; the stack. The command tail pointer argument on the
	; stack now points to just past the "end of range" spec
	; substring.
	;
    check_ram_have_end_of_range:
	mov	bx, ax
	pop	ax

	; The 'RAM=' option was found on the command line
	; AX = the (assumed) "start of range" segment value
	; BX = the (assumed) "end of range" segment value
	;
	; If the range is assumed (because RAM appeared by itself on
	; the command line) then the RAM range will overlap the LIM 3.2
	; page frame, ROM regions and Included/eXcluded regions.
	;
    check_ram_have_range_values:
	;
	; We do nothing with these values yet.
	;
	nop

    check_ram_done:
        jmp     @12@410

	; The 'RAM=' option was found on the command line but the
	; the argument was ill formed. The pointer at [bp+6] points
	; to the offending portion of the command line.
	;
    check_ram_syntax_error:
	;
	; whatawedonow?
	;
	nop

	; The code at the top of the argument processing loop
	; checks to see if we are on a non-separator char and, if so,
	; assumes we are on part of an argument skips over the
	; remainder of that "argument" to find some separators.
	; Back up to some white space.
	;
    check_ram_backup_to_a_sep:
        dec 	word ptr [bp+4]     	; skip separator char
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]	; al = current cmd line cahr
	push	ax
	call	_isseparator	; is cur char a separator?
	pop	cx
	or	ax,ax
	jz	short check_ram_backup_to_a_sep ; find a separator char
	jmp	@12@410


check_rom:      ;see if the string is "ROM="
        les     bx,[bp+4]       ;pointer to string
        cmp dword ptr es:[bx],'=MOR'    ;cmp to ROM=
        jne short @12@194       ;jmp if not "ROM="
        add word ptr [bp+4],4   ;point to end of string, start of range
        push    16              ;use base 16 numbers
        lea     ax,word ptr [bp+4]
        push    ax              ;pointer to the pointer
        call	_getnum
        pop     cx              ;pop pointer only
        xor     al,al           ;round segment down to nearest 4K
        mov     bx,DGROUP:num_rom
;if we have already done 4 ROM ranges, dec num_rom before continuing
        cmp     bx,4
        jb short num_rom_okay
        dec     bx
        mov     num_rom,bx      ;set num_rom back to 3
num_rom_okay:
        shl     bx,1
        mov     DGROUP:rom_start[bx],ax
        inc word ptr [bp+4]     ;point to rom_stop
        lea     ax,[bp+4]       ;pointer to the string
        push    ax
        call	_getnum
        pop     cx
        pop     cx
        mov     bx,DGROUP:num_rom
        shl     bx,1
        inc     ax              ;bump it up if ending in FFF
        xor     al,al           ;round segment address down to 4k
        mov     DGROUP:rom_stop[bx],ax
        sub     ax,DGROUP:rom_start[bx] ;calculate number of paragraphs
;If the top end was 1 meg, we just subtracted the start address from zero.
;This actually works for calculating the size.
        shr     ax,6            ;convert paragraphs to K
        add     DGROUP:rom_memK,ax
        inc     DGROUP:num_rom
        jmp     @12@410

@12@194:
   ;			// is it a memory range?
	les	bx,dword ptr [bp+4]
	mov	bx,es:[bx]
	push	bx
	call	_isxdigit
	pop	cx
        lea     si,i_cmd	;<PPP>
	or	ax,ax
	jnz short @@0
;check for I=
        and     bl,not 20h              ;convert to upper case
        cmp     bx,'=I'
        je short @@0_a
;check for X=
        lea     si,x_cmd	;<PPP>
        cmp     bx,'=X'
        jne short @12@410
@@0_a:
;It was I= or X=, dump the identifiers, find the number.  Si = either offset
;of i_range or offset x_range.
        add word ptr [bp+4],2           ;dump the 'I=' or 'X='
@@0:
   ;				// get start of range (must be on a 4K boundary)
        push    16
        lea     ax,[bp+4]
        push    ax
        call    _getnum         ;returns ax = number
        pop     cx
        pop     cx
        xor     al,al                   ;round down to nearest 4K
        push    ax              ;save for now
        jmp short @12@290
@12@266:
   ;
   ;
   ;				// get end of range (must be on a 4K boundary)
   ;				// allow both D000-DFFF and D000-E000 to work
   ;				while(!isterminator(*s) && !isxdigit(*s))
   ;					++s;
   ;
	inc	word ptr [bp+4]
@12@290:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isterminator
	pop	cx
	or	ax,ax
	jne	short @12@338
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isxdigit
	pop	cx
	or	ax,ax
	je	short @12@266
@12@338:
   ;
   ;				end = getnum(&s,16);
   ;
	mov	ax,16
	push	ax
	lea	ax,word ptr [bp+4]
	push	ax
	call	_getnum
        pop     ecx
;ax has the end segment address, round up to nearest 4K
        add     ax,255
        jnc short rl_1
;We have wrapped over the top, ah is now zero, set it back down
        dec     ah
rl_1:
        mov     al,0
        pop     bx              ;start of block
;check if start and stop addresses are appropriate
        sub     ax,bx           ;is top below or equal to start?
        jbe short @12@410       ;jmp if not appropriate
        or      bx,bx           ;is the start zero?
        jz short @12@410        ;jmp if not appropriate
        movzx   cx,ah           ;number of 4K pages in range
        movzx   bx,bh           ;4K index for start
r_loop:
        bts	ss:[si],bx      ;<PPP> set the proper bit in the range array
        inc     bx
        loop    r_loop

@12@410:
   ;
   ;				}
   ;			}
   ;
   ;			// on to the next parameter
   ;			s = nextparm(s);
   ;
	push	word ptr [bp+6]
	push	word ptr [bp+4]
	call	_nextparm
	pop	cx
	pop	cx
	mov	word ptr [bp+6],dx
	mov	word ptr [bp+4],ax
@12@434:
	les	bx,dword ptr [bp+4]
	mov	al,byte ptr es:[bx]
	push	ax
	call	_isterminator
	pop	cx
	or	ax,ax
        je      @12@98                  ;loop until we find the termination

; The command line has been parsed. Now apply I= and X= parameters.

	xor	si,si	; index/offset
apply_ix_cmd:
	mov	ax,i_cmd[si]
	or	word ptr i_range[si],ax	; apply I=
	not	ax
	and	word ptr x_range[si],ax	; I= takes over autodetect
	mov	ax,x_cmd[si]
	or	word ptr x_range[si],ax	; apply X= over I=
	add	si,2
	cmp	si,32
	jb	apply_ix_cmd

IF DEBUG GE 1
   ; after we've applied I= and X= to the initial conditions
   ;
      call  BIOSCRLF
      call  DUMP_IRANGE
      call  BIOSCRLF
      call  DUMP_XRANGE
ENDIF

comment &
        There are a number of things which can over ride the i_range bits.
        First, we must take care of the FRAME segment and the ROM= ranges.
&
        movzx   bx,byte ptr _frame_seg[1]
        mov     cx,16
det_h:
        bts word ptr x_range,bx
        inc     bx
        loop    det_h


IF DEBUG GE 1
      call  BIOSCRLF
      call  DUMP_XRANGE
ENDIF

;set the bits corresponding to the ROM=ranges
        mov     cx,num_rom
        jcxz short det_k
        xor     si,si           ;si is 2 * index
det_loop4:
        movzx   bx,byte ptr rom_start[si+1]
        movzx   dx,byte ptr rom_stop[si+1]
        or      dx,dx
        jnz short det_loop5     ;is dx zero?
;dx is zero, rom_stop is 1 meg, adjust dx accordingly
        inc     dh
det_loop5:
        bts word ptr x_range,bx
        inc     bx
        cmp     bx,dx
        jb      det_loop5
        add     si,2
        loop    det_loop4

det_k:

IF DEBUG GE 1
      call  BIOSCRLF
      call  DUMP_XRANGE
ENDIF

;The x_range and i_range arrays have been initialized, any bit in the x_range
;that has been set must be cleared in the i_range.
        mov     si,offset DGROUP:x_range
        mov     di,offset DGROUP:i_range
        mov     cx,16
det_loop2:
        lodsw                   ;get a word from the x_range
        not     ax
        and     ds:[di],ax      ;clear the required bits
        add     di,2
        loop    det_loop2
;i_range now accurately specifies the ranges, fill in the _range array
        xor     dx,dx           ;_range_memK
        xor     bx,bx           ;bit index for i_range
        xor     si,si           ;_num_ranges*4
det_loop3:
        bt word ptr i_range,bx
        jc short det_d          ;jmp if UMB
det_i:
        inc     bl              ;point to next 4K page
        jnz short det_loop3     ;loop through 1 meg
;done, exit the loop
        jmp short det_l

det_d:
;bx points to a block to be used for UMB, will it fit in the current range?
        or      dx,dx
        jz short det_e         ;jmp if the first time through
        mov     ax,word ptr _ranges[si]         ;get the start segment
        add     ax,word ptr _ranges[si+2]       ;add the size
;ax contains the segment address of the top of the current range.
        cmp     ah,bl
        jne short det_g                 ;jmp if not contiguous
;things are contiguous, increase the size of the range by 4K or 256 paragraphs
det_j:
        inc byte ptr _ranges[si+3]      ;increment the high byte
;increase size of _range_memK by 4K
        add     dx,4
        jmp     det_i

det_g:  ;start a new range
        add     si,4            ;point to next range
        cmp     si,MAXRANGES*4
        jae short det_f         ;jmp if no more ranges allowed
det_e:  ;si is a pointer to a new range, initialize the start address
        mov byte ptr _ranges[si+1],bl   ;start address in segment form
        jmp     det_j

det_l:
        or      dx,dx
        jz short det_f          ;don't increment si if no UMBs
        add     si,4            ;change from index to num_ranges*4

det_f:  ;finished initializing the _range array, save _range_memK
        mov     _range_memK,dx
        shr     si,2            ;divide by 4
        mov     _num_ranges,si

IF DEBUG GE 1
      call  BIOSCRLF
      call  DUMP_IRANGE
ENDIF

comment &
        Examine the results and see if we have enough memory to do everything
        requested.  Prioritize as follows:
                1.  UMBs
                2.  ROM=
                3.  VDS
                4.  EMS
&


        mov     AH,8
        call    xms_handler             ;sizeK returned in AX
        and     al,0fch                 ;round down to nearest 4K
	and	dl,0fch

	push	bx
	push	cx

	mov	bx, _frame_maxK		; if no max= on the command line
	or	bx, bx
	jz	short pcmdln_limit_max

    pcmdln_have_max:
	cmp	bx, dx
	jbe	short pcmdln_limit_min

    pcmdln_limit_max:
	mov	bx, dx			; use all the EMS we can get

    pcmdln_limit_min:
	cmp	bx, _frame_memK		; if max < min
	jae	short pcmdln_set_max

	mov	_frame_memK, bx		; then set min to max

    pcmdln_set_max:
	mov	_frame_maxK, bx		; update EMS maxK variable
	pop	cx
	pop	bx

        mov     si,ax                   ;save largest XMS block in si
        sub     si,_range_memK
        ja short @12@506                ;jmp if enough for UMBs

;we don't have enough memory for UMBs, use what we have.
        mov     _range_memK,ax
;all others must now be set to zero.
        xor     ax,ax
        mov     rom_memK,ax             ;no ROM=
        mov     rom_start,ax
no_vds:
        xor     ax,ax
        mov     vds_buff_memk,ax
no_ems:
        xor     ax,ax
        mov     _frame_memK,ax          ;zero EMS
	mov	_frame_maxK,ax
        mov     _frame_seg,ax           ;no page frame
        jmp     short cmd_ret

@12@506:
        mov     ax,si                   ;save in case of error
        sub     si,rom_memK             ;allow for rom_memK
        ja short l1                     ;jmp if enough for ROM=
;not enough for ROM=, use what we have.
        mov     rom_memK,ax
        mov     rom_start,ax            ;disable ROM=
        jmp     no_vds
l1:
        mov     ax,si                   ;save in case of error
        sub     si,vds_buff_memk        ;allow for vds buffer
        ja short l2                     ;jmp if enough for VDS buffer
;not enough memory for DMA buffer, use what we have
        mov     vds_buff_memk,ax
        jmp     no_ems
l2:
   ;		// adjust size of EMS memory (in case there's not enough)
   ;		if(frame_memK > totalxms)
   ;
        mov     ax,si                   ;save in case of error
;calculate the memory needed for this size _frame_memK
;we need extra memory for the EMS page tables if windows starts
        xor     edx,edx
        cmp     dx,_frame_seg
        mov     dx,ax                   ;total available XMS memory in kbytes
        jz short cmd_ret                 ;jmp if no EMS
;allow 64 bytes for system handle and round up to nearest 4K
        add     edx,64+4096
        shr     edx,10                  ;convert to k bytes
        and     dl,0fch                 ;4K align
        mov word ptr win_page_table_ptr,dx      ;save for now
        mov     cx,ax                   ;total XMS memory
        sub     cx,dx
        jc      no_ems                  ;jmp if no EMS memory available
;dx contains the memory needed for page tables, cx contains the max possible
;EMS memory.

	cmp	cx,_frame_maxK
	jbe short @12@5539

	mov	cx,_frame_maxK

@12@5539:
        cmp     cx,_frame_memK          ;do we have enough for command line?
        ja 	short @12@554           ;jmp if enough memory
;not enough for requested EMS, use what we have.
        mov     _frame_memK,cx
	mov	_frame_maxK,cx
@12@554:
;frame_pages = max possible EMS memory in pages
        shr     cx,4                    ;convert kbytes to EMS pages.
        mov     _frame_pages,cx
;frame_avail = _frame_memK expressed in 16K pages
        mov     ax,_frame_memK
        shr     ax,4                    ;divide by 16
        mov     _frame_avail,ax
;calculate incremental_XMS.
        sub     si,_frame_memK          ;si is max EMS - reserved for EMS
;si is the number of kbytes which can be swapped from XMS to EMS.
;we want to divide that amount by about 8.
        shr     si,7                    ;divide by 8 and convert to 16K pages
        mov     incremental_XMS,si
cmd_ret:

	add	sp,64	; remove locals

	pop	di
	pop	si
	pop	bp

	ret
_parse_cmdline	endp


cEndCode


;/////////////////////////////////////////////////////////////////////////
;//	Scratch Pad							//
;//	Stuff that's no longer used					//
;/////////////////////////////////////////////////////////////////////////

IF 0
	public	old_temp_21
old_temp_21	dd	?

tmp48	db	'a'
tmp4A	db	'A'

;
;  The following is a temporary INT 21H handler to
;  pretend to allocate memory from DOS during device
;  init time.
;
temp_21:
	cmp	AH, 48H		; allocate memory
	jz	got48
	cmp	AH, 4AH		; reallocate memory
	jnz	is_old_21
	inc	CS:[tmp4A]
	jmp	short doit
got48:	inc	CS:[tmp48]
doit:	mov	word ptr CS:[end_size], BX
	mov	AX, seg _ENDSEG
	push	BP
	mov	BP, SP
	and	byte ptr [BP+6], 0FEH	; clear carry flag
	pop	BP
	iret

is_old_21:
	jmp	dword ptr CS:[old_temp_21]

trap21	proc	near
	mov	AX, 3521H
	int	21h
	mov	word ptr [old_temp_21], BX
	mov	word ptr [old_temp_21+2], ES
	mov	AX, 2521H
	mov	DX, offset temp_21
	int	21H
	ret
trap21	endp

restore21	proc	near
	push	DS
	lds	DX, CS:[old_temp_21]
	mov	AX, DS
	or	AX, AX
	jz	restore_nothing
	mov	AX, 2521H
	int	21h
restore_nothing:
	pop	DS
	ret
restore21	endp

test21	proc	near
mov	BX, 20H
mov	AH, 48h
int	21h

mov	BX, 18H
mov	AH, 4Ah
int	21h

mov	AH, 2h
mov	DL, 'A'
int	21h

ret
test21	endp

ENDIF	; temp 21 stuff


	END	_astart
